/*
** Module   : P2JLocaleHelper.java
** Abstract : tool which provides rudimentary conversion services from
**            Progress collation-table-style format to a Unix locale
**            definition file format
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 ECF 20060629   @27966 Created initial version. Provides services to
**                           read a Progress 16x16 collation-table-style
**                           table format and convert it into format
**                           suitable for cut-and-paste into a Unix locale
**                           definition file.
** 002 ECF 20121126          Replaced uses of StringHelper.sprintf with
**                           String.format;  removed com.goldencode.p2j.util
**                           import statement accordingly.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

import java.io.*;
import java.util.*;

/**
 * Provides services to read a Progress 16x16 collation-table-style table
 * format and convert it into format suitable for cut-and-paste into a
 * Unix locale definition file.  Currently provides only a command line
 * interface and outputs results to standard out.
 * <p>
 * For the meaning of the emitted output, see <a target="_NEW"
 * href="http://www.opengroup.org/onlinepubs/009695399/">IEEE Std 1003.1,
 * 2004 Edition, Chapters 6-7</a>.
 * <p>
 * Note that this helper class is not compiled by default into the main P2J
 * jar file.  It is a standalone class.
 */
public class P2JLocaleHelper
{
   /** Constant indicating uppercase codepoint mapping conversion */
   private static final int UPPERCASE = 1;
   
   /** Constant indicating lowercase codepoint mapping conversion */
   private static final int LOWERCASE = -1;
   
   /**
    * Parse a 16x16 table of base-10, 3-digit, left-padded integers.  For
    * each integer encountered, create a <code>Cell</code> object and invoke
    * the <code>processCell</code> method for that object.  The return value
    * of the <code>processCell</code> method is stored in a list object,
    * which is returned from this method.
    *
    * @param   reader
    *          A buffered reader opened on the stream of table input.
    * @param   processor
    *          An object which processes each table cell in an implementation
    *          specific manner.
    *
    * @return  A <code>List</code> of the results returned by the given
    *          cell processor.
    *
    * @throws  IOException
    *          if an error occurs reading table from the underlying media.
    * @throws  NumberFormatException
    *          if a cell's data cannot be parsed as a base-10 integer value.
    */
   private static List readTable(BufferedReader reader,
                                 CellProcessor processor)
   throws IOException,
          NumberFormatException
   {
      List list = new ArrayList(256);
      String nextLine;
      for (int i = 0; (nextLine = reader.readLine()) != null; i++)
      {
         StringTokenizer tok = new StringTokenizer(nextLine);
         for (int j = 0; tok.hasMoreTokens(); j++)
         {
            int index = i * 16 + j;
            int value = Integer.parseInt(tok.nextToken());
            list.add(processor.processCell(new Cell(index, value)));
         }
      }
      
      return list;
   }
   
   /**
    * Parse a 16x16 Progress-style collation table.  For each table cell
    * encountered, create a <code>CollationElement</code> and store it in
    * this method's return list.  Iterate through the list and emit a snippet
    * of output which is suitable for cutting-and-pasting into the LC_COLLATE
    * section of a Unix locale definition file.
    * <p>
    * See <u>IEEE Std 1003.1, 2004 Edition, section 7.3.2</u>.
    *
    * @param   reader
    *          A buffered reader opened on the stream of table input.
    *
    * @throws  IOException
    *          if an error occurs reading table from the underlying media.
    * @throws  NumberFormatException
    *          if a cell's data cannot be parsed as a base-10 integer value.
    */
   private static void processCollationTable(BufferedReader reader)
   throws IOException,
          NumberFormatException
   {
      CellProcessor cp = new CellProcessor()
      {
         public Object processCell(Cell cell)
         {
            return new CollationElement(cell);
         }
      };
      
      List list = readTable(reader, cp);
      
      Collections.sort(list);
      Iterator iter = list.iterator();
      while (iter.hasNext())
      {
         Element e = (Element) iter.next();
         System.out.println(e.toString());
      }
   }
   
   /**
    * Parse a 16x16 Progress-style table of upper- or lowercase values.  For
    * each table cell encountered, create a <code>CaseMapping</code> object
    * and store it in a list, for further processing as follows:
    * <ul>
    * <li>for a lower-to-uppercase conversion table (i.e. <code>caseType ==
    *     UPPERCASE</code>), emit two snippets of the <code>LC_CTYPE</code>
    *     category of a Unix locale definition file:  <code>upper</code> and
    *     <code>toupper</code>.
    * <li>for an upper-to-lowercase conversion table (i.e. <code>caseType ==
    *     LOWERCASE</code>), emit two snippets of the <code>LC_CTYPE</code>
    *     category of a Unix locale definition file:  <code>lower</code> and
    *     <code>tolower</code>.
    * </ul>
    * <p>
    * See also <u>IEEE Std 1003.1, 2004 Edition, section 7.3.1</u>.
    *
    * @param   reader
    *          A buffered reader opened on the stream of table input.
    * @param   caseType
    *          A constant which must be <code>UPPERCASE</code> or
    *          <code>LOWERCASE</code>, indicating the meaning of the table
    *          being read.
    *
    * @throws  IOException
    *          if an error occurs reading table from the underlying media.
    * @throws  NumberFormatException
    *          if a cell's data cannot be parsed as a base-10 integer value.
    * @throws  UnsupportedOperationException
    *          if <code>caseType</code> is not one of the required constants.
    */
   private static void processCaseMapTable(BufferedReader reader,
                                           final int caseType)
   throws IOException,
          NumberFormatException
   {
      switch (caseType)
      {
         case UPPERCASE:
         case LOWERCASE:
            break;
         default:
            throw new UnsupportedOperationException();
      }
      
      CellProcessor cp = new CellProcessor()
      {
         public Object processCell(Cell cell)
         {
            switch (caseType)
            {
               case UPPERCASE:
                  return new CaseMapping(new Element(cell.value),
                                         new Element(cell.index));
               case LOWERCASE:
                  return new CaseMapping(new Element(cell.index),
                                         new Element(cell.value));
            }
            
            return null;
         }
      };
      
      List list = readTable(reader, cp);
      
      // Produce upper/lower snippet.
      Iterator iter = list.iterator();
      System.out.println(caseType == UPPERCASE ? "upper" : "lower");
      int i = 0;
      while (iter.hasNext())
      {
         CaseMapping map = (CaseMapping) iter.next();
         
         Element e = map.getElement(caseType);
         if (e == null)
         {
            continue;
         }
         
         if (i > 0)
         {
            System.out.print(";");
            if (i % 8 == 0)
            {
               System.out.println("/");
            }
         }
         System.out.print(e.toString());
         i++;
      }
      System.out.println();
      
      // Produce toupper/tolower snippet.
      iter = list.iterator();
      System.out.println(caseType == UPPERCASE ? "toupper" : "tolower");
      i = 0;
      while (iter.hasNext())
      {
         CaseMapping map = (CaseMapping) iter.next();
         
         Element e = map.getElement(caseType);
         if (e == null)
         {
            continue;
         }
         
         if (i > 0)
         {
            System.out.print(";");
            if (i % 3 == 0)
            {
               System.out.println("/");
            }
         }
         System.out.print(map.toString(caseType));
         i++;
      }
      System.out.println();
   }
   
   /**
    * Objects which process the data in a table cell must implement this
    * interface.
    */
   private interface CellProcessor
   {
      /**
       * Process the data in a single cell of a data table.
       *
       * @param   cell
       *          Object which contains an index and a data value.
       *
       * @return  An implementation-specific Object.
       */
      public Object processCell(Cell cell);
   }
   
   /**
    * An abstraction of the data in a single table cell.  Contains an index,
    * which represents a flat model of a two-dimensional table.  That is,
    * each index is calculated as follows:
    * <pre>
    *    <current row index> * <number of columns> + <current column index>
    * </pre>
    * where the row and column indexes are zero-based indices.
    * <p>
    * Also contains the integral data value parsed from the table cell at
    * that index.
    */
   private static class Cell
   {
      /** Flat index of table cell */
      private final int index;
      
      /** Data value for the table cell */
      private final int value;
      
      /**
       * Convenience constructor which assigns index and value.
       *
       * @param   index
       *          Flat index of cell.
       * @param   value
       *          Data value of cell.
       */
      Cell(int index, int value)
      {
         this.index = index;
         this.value = value;
      }
   }
   
   /**
    * Abstraction of a code point value which knows how to format itself in a
    * manner suitable for inclusion in a Unix locale definition.
    */
   private static class Element
   {
      /** Code point value */
      protected final int codePoint;
      
      /**
       * Constructor.
       *
       * @param   codePoint
       *          Code point value.
       */
      Element(int codePoint)
      {
         this.codePoint = codePoint;
      }
      
      /**
       * Test equivalence by comparing code point values.
       *
       * @param   o
       *          Another <code>Element</code> object.
       */
      public boolean equals(Object o)
      {
         Element e = (Element) o;
         
         return (codePoint == e.codePoint);
      }
      
      /**
       * Get a formatted string which represents this element's code point in
       * a format suitable for inclusing in a Unix locale definition.
       *
       * @return  Formatted representation of this element.
       */
      public String toString()
      {
         Integer cp = new Integer(codePoint);
         return String.format("<U%04X>", new Object[] { cp });
      }
   }
   
   /**
    * An encapsulation of the code point and collation weight of a character.
    */
   private static class CollationElement
   extends Element
   implements Comparable
   {
      /** Collation weight association with this element's code point */
      protected final int weight;
      
      /**
       * Constructor which extracts the code point and collation weight from
       * a table cell.
       *
       * @param   cell
       *          Table cell.
       */
      CollationElement(Cell cell)
      {
         super(cell.index);
         
         this.weight = cell.value;
      }
      
      /**
       * Compare this element to another to determine how they sort relative
       * to one another.  Elements with lower collation weights sort first.
       *
       * @param   o
       *          Another <code>CollationElement</code> to compare with this
       *          one.
       *
       * @return  A positive number if this element sorts first;
       *          a negative number if this element sorts last;
       *          <code>0</code> if both elements sort the same.
       */
      public int compareTo(Object o)
      {
         CollationElement e = (CollationElement) o;
         
         return (weight - e.weight);
      }
   }
   
   /**
    * An object which defines the mapping between a single uppercase and
    * lowercase code point pair.  If both the uppercase and lowercase
    * elements encapsulated by this object represent the same code point,
    * this indicates that there is no concept of case for this code point.
    */
   private static class CaseMapping
   {
      /** Uppercase code point element for this mapping */
      private final Element upper;
      
      /** Lowercase code point element for this mapping */
      private final Element lower;
      
      /**
       * Constructor.
       *
       * @param   upper
       *          Uppercase code point element for this mapping.
       * @param   lower
       *          Lowercase code point element for this mapping.
       */
      CaseMapping(Element upper, Element lower)
      {
         this.upper = upper;
         this.lower = lower;
      }
      
      /**
       * Get the element from this mapping which represents the case
       * indicated by the <code>caseType</code> parameter.
       *
       * @param   caseType
       *          Constant indicating which end of the mapping is desired:
       *          <code>UPPERCASE</code> or <code>LOWERCASE</code>.
       *
       * @return  Element which represents the code point of the requested
       *          case, or <code>null</code> if both ends of the mapping
       *          represent the same code point.
       *
       * @throws  UnsupportedOperationException
       *          if <code>caseType</code> is not one of the required
       *          constants.
       */
      Element getElement(int caseType)
      {
         if (upper.equals(lower))
         {
            return null;
         }
         
         switch (caseType)
         {
            case UPPERCASE:
               return upper;
            case LOWERCASE:
               return lower;
            default:
               throw new UnsupportedOperationException();
         }
      }
      
      /**
       * Get a formatted string representation of this case mapping, suitable
       * for use within the <code>toupper</code> or <code>tolower</code>
       * subsections of the <code>LC_CTYPE</code> category of a Unix locale
       * definition.
       *
       * @throws  UnsupportedOperationException
       *          if <code>caseType</code> is not one of the required
       *          constants.
       */
      String toString(int caseType)
      {
         Object[] parms = null;
         
         switch (caseType)
         {
            case UPPERCASE:
               parms = new Object[] { lower, upper };
               break;
            case LOWERCASE:
               parms = new Object[] { upper, lower };
               break;
            default:
               throw new UnsupportedOperationException();
         }
         
         return String.format("(%s,%s)", parms);
      }
   }
   
   /**
    * Command line harness for the locale definition services provided by
    * this class.
    *
    * @param   args
    *          <ol>
    *            <li><code>args[0]</code> - operation mode:
    *            <ul>
    *              <li>C (convert collation table)
    *              <li>U (convert uppercase table)
    *              <li>L (convert lowercase table)
    *            </ul>
    *            <li><code>args[1]</code> - name of the file containing the
    *                tabular data to be converted
    *          </ol>
    */
   public static void main(String[] args)
   {
      try
      {
         if (args.length != 2)
         {
            System.err.println("Usage:  P2JLocaleHelper <mode> <filename>");
            System.err.println("where:");
            System.err.println("   <mode> is one of the following:");
            System.err.println("        C (collation table)");
            System.err.println("        U (uppercase table)");
            System.err.println("        L (lowercase table)");
            System.err.println("   <filename> is the filename containing the table data to be read");
            
            System.exit(-1);
         }
         
         BufferedReader reader = null;
         try
         {
            reader = new BufferedReader(new FileReader(args[1]));
            
            if (args[0].equalsIgnoreCase("C"))
            {
               processCollationTable(reader);
            }
            else if (args[0].equalsIgnoreCase("U"))
            {
               processCaseMapTable(reader, UPPERCASE);
            }
            else if (args[0].equalsIgnoreCase("L"))
            {
               processCaseMapTable(reader, LOWERCASE);
            }
            else
            {
               throw new UnsupportedOperationException(args[0]);
            }
         }
         finally
         {
            if (reader != null)
            {
               reader.close();
            }
         }
      }
      catch (Exception exc)
      {
         exc.printStackTrace();
      }
   }
}