/*
** Module   : AstWalkListener.java
** Abstract : An interface for objects who need to be notified of events
**            during an AST walk
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 ECF 20050316   @20420 Created initial version. Defines interface
**                           for AST tree walk event notification to a
**                           listener object.
** 002 ECF 20050428   @20921 Added nextChild method. Notifies of a lateral
**                           transition from a left sibling to its next,
**                           right sibling. The parent AST is passed as a
**                           parameter. This mimics recursive behavior.
** 003 GES 20050613   @21472 Added next child index support.
** 004 GES 20090429   @42071 Moved packages.
** 005 GES 20090518   @42370 Import change.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.ast;

/**
 * This interface should be implemented by classes which need to be notified
 * of significant events within an AST tree walk. Currently, these include:
 * <ul>
 * <li>the tree walk traverses vertically from a parent node to a child node
 *     (a descent event);
 * <li>the tree walk traverses vertically from a child node to a parent node
 *     (an ascent event);
 * <li>the tree walk traverses laterally from a left child node to its next,
 *     right neighbor (a next child event).
 * </ul>
 */
public interface AstWalkListener
{
   /**
    * Called whenever a transition from a parent node to its first child
    * takes place during an AST walk. This will be called <em>before</em> the
    * next AST node in the tree traversal is visited.
    *
    * @param   ast
    *          Source AST node which represents the parent in the parent to
    *          child transition of the current AST walk.
    */
   public void descent(Aast ast);
   
   /**
    * Called whenever a transition from a child node to its parent takes
    * place during an AST walk. This will be called <em>before</em> the
    * next AST node in the tree traversal is visited.
    *
    * @param   ast
    *          Source AST node which represents the parent in the child to
    *          parent transition of the current AST walk.
    */
   public void ascent(Aast ast);
   
   /**
    * Called whenever a transition from a child node to its next right
    * sibling takes place during an AST walk. This will be called
    * <em>before</em> the next AST node in the tree traversal is visited.
    * The parent node of the children is in scope at the time of
    * notification. This effectively mimics recursion by notifying the
    * listener when one child node has been processed, before the next is
    * visited. This allows the listener to perform parent-level processing
    * between visits to child nodes.
    * <p>
    * This method is <em>not</em> invoked after the last child is visited,
    * as this is the purpose of the {@link #ascent} method.
    *
    * @param   ast
    *          Source AST node which represents the parent of the children
    *          between which the lateral transition is taking place in the
    *          current AST walk.
    * @param   index
    *          The 0-based index of the child which is about to be processed.
    */
   public void nextChild(Aast ast, int index);
}