/*
** Module   : DumpBitSet.java
** Abstract : print a list of the bits in an antlr.collections.BitSet and
**            each bits associated token type name   
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20041223   @19080 First version with support for arbitrary
**                           BitSet field names in arbitrary parser or
**                           lexer classes.
** 002 GES 20050131   @19465 Proper lexer support added (no need for
**                           the _tokenNames[] in this case, each bit
**                           represents the character of the same int
**                           value).  Simplified the command line args
**                           as a result of this change.
** 003 GES 20090428   @42018 Moved this class to a new package.
** 004 EVL 20160225          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.ast;

import antlr.collections.*;
import antlr.collections.impl.*;
import java.lang.reflect.*;
import java.text.*;

/**
 * Prints out a list of each bit that is turned on in an
 * <code>antlr.collections.BitSet</code> and its associated token type name.
 * This is a helper class that is very useful for diagnosing logic problems
 * with ANTLR based lexers or parsers.  ANTLR implements a large number of
 * decisions based on its own <code>BitSet</code> class and the 
 * <code>member</code> method.  Without being able to dump the contents of
 * such a bitset, such decision logic is obscured.  With this tool, one can
 * easily print any <code>BitSet</code> to <code>STDOUT</code>.
 * <p>
 * For command line syntax, see {@link #main}.
 */
public class DumpBitSet
{
   /** Formats a token type (integer) padded on left with zeros. */
   private NumberFormat tfmt = NumberFormat.getInstance();
   
   /** Help text for command line users. */
   private static String[] syntax =
   {
      new String(""),
      new String("Syntax:\n"),
      new String("   java DumpBitSet <containing_class> <bitset_field>"),
      new String(""),
      new String("Where:\n"),
      new String("  <containing_class> = the name of the class containing" +
                 " the BitSet to dump"),
      new String("  <bitset_field> = field name of the BitSet to dump")
   };
   
   /**
    * Uses reflection to obtain a reference to the <code>_tokenNames</code>
    * <code>String[]</code> that maps integer token types to the symbolic
    * names used in the ANTLR grammars.
    * <p>
    * Note that this field is <code>public static</code> and is always named
    *  the same, so it is hard coded into this method for convenience.
    *
    * @param     className
    *            Fully qualified (incuding package) classname of the parser
    *            or lexer in which to find the array.
    * @return    The _tokenNames <code>String[]</code>.
    * 
    * @throws   ClassNotFoundException 
    *           If an respective error occurs. 
    * @throws   NoSuchFieldException 
    *           If an respective error occurs. 
    * @throws   IllegalAccessException 
    *           If an respective error occurs. 
    * @throws   SecurityException 
    *           If an respective error occurs. 
    */
   public static String[] lookupTokenNames(String className)
   throws ClassNotFoundException,
          NoSuchFieldException,                              
          IllegalAccessException,
          SecurityException
   {   
      Class parser  = Class.forName(className);
      Field names   = parser.getField("_tokenNames");
      
      return (String[])names.get(null);      
   }
   
   /**
    * Uses reflection to obtain a reference to the <code>BitSet</code>
    * that is to be dumped.  These 
    * <p>
    * Note that such <code>BitSet</code> fields are <code>public static</code>
    * so all that is needed is the name of the field.
    *
    * @param     className
    *            Fully qualified (incuding package) classname of the parser
    *            or lexer in which to find the bit set.
    * @param     setName
    *            The name of the <code>BitSet</code> field to dump. 
    * @return    The <code>BitSet</code> field reference.
    * 
    * @throws   ClassNotFoundException 
    *           If an respective error occurs. 
    * @throws   NoSuchFieldException 
    *           If an respective error occurs. 
    * @throws   IllegalAccessException 
    *           If an respective error occurs. 
    * @throws   SecurityException 
    *           If an respective error occurs. 
    */
   public static BitSet lookupBitSet(String className,
                                     String setName)
   throws ClassNotFoundException,
          NoSuchFieldException,                              
          IllegalAccessException,
          SecurityException
   {   
      Class parser  = Class.forName(className);
      Field bits    = parser.getField(setName);
      
      return (BitSet)bits.get(null);      
   }
   
   /**
    * Core worker routine that dumps the <code>BitSet</code> to 
    * <code>STDOUT</code>.
    *
    * @param    testSet
    *           The <code>BitSet</code> to dump.
    * @param    tokenNames
    *           The array of associated token names.
    */
   private void dumpWorker(BitSet testSet, String[] tokenNames)
   {
      // setup our integer format
      tfmt.setMinimumIntegerDigits(4);
      tfmt.setGroupingUsed(false);
      
      // convert the BitSet into an array of int
      int[] list = testSet.toArray();
      
      // loop through int array and for each one, dereference the
      // associated name in the token name String[], print the formatted
      // result
      for (int i = 0; i < list.length; i++)
      {
         System.out.print(tfmt.format(list[i]) + "     ");
         if (tokenNames != null)                            
         {
            System.out.println("<"                  +
                               tokenNames[list[i]]  +
                               ">");
         }
         else
         {
            System.out.println("\'" + (char)list[i] + "\'");
         }
      }      
   }
   
   /**
    * Helper method to print out syntax help messages for the command line
    * user.
    */
   private static void printSyntax()
   {
      for (int i = 0; i < syntax.length; i++)            
      {
         System.out.println(syntax[i]);
      }
   }
   
   /**
    * Provides a command line interface for an end user to dump an arbitrary
    * public static <code>BitSet</code> field of an arbitrary public parser
    * or lexer created by ANTLR.
    * <p>
    * Syntax:
    * <pre>
    * java DumpBitSet &lt;bitsetclass&gt; &lt;setname&gt;
    * </pre>
    * Where:
    * <ul>
    *   <li> &lt;bitsetclass&gt; is the fully qualified class name of the
    *        class in which to find the <code>public static BitSet</code>
    *   <li> &lt;setname&gt; name of the <code>public static BitSet</code> to
    *        dump
    * </ul>
    *
    * @param   args 
    *          List of command line arguments.
    */   
   public static void main( String[] args )
   {
      // missing arguments, print syntax help
      if (args.length < 2)
      {
         printSyntax();
         System.exit(-1);
      }
      
      try
      {
         // behave differently if we are a parser (uses _tokenNames[])
         Class   request = Class.forName(args[0]).getSuperclass();
         boolean parser  = request.equals(Class.forName("antlr.LLkParser"));
         
         // use reflection to access the fields necessary
         BitSet   testSet    = lookupBitSet(args[0], args[1]); 
         String[] tokenNames = null;
         
         if (parser)
            tokenNames = lookupTokenNames(args[0]);
         
         DumpBitSet dbs = new DumpBitSet();
         dbs.dumpWorker(testSet, tokenNames);
      }
      
      catch (Exception e)
      {
         e.printStackTrace();
      }
   }
}