/*
** Module   : ShadowNode.java
** Abstract : AST nodes created from the original input but not considered
**            part of the tree structure itself
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 GES 20100922 First version with public data for simplicity.
** 002 GES 20101001 Added helper/accessor methods and moved data private. Added
**                  support to store the source token.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.ast;

import antlr.*;

/**
 * Container for a node that is associated with a given tree (AST), but it
 * exists outside of the primary structure of the tree. An example of such a
 * node is a comment.  This is often something that is important to maintain
 * along with a given input tree.  But it is also something which should not
 * be considered part of the structure of the program.
 * <p>
 * Each shadow node has a left-side node ID and a right-side node ID.  One or
 * both of these IDs must be valid.  These are linkages to the nodes before and
 * after that shadow node in the original input stream.  These are kept here
 * to eliminate the need to walk the tree to determine which nodes were near
 * the shadow node.  It is possible that the left or right IDs reference other
 * shadow nodes.
 * <p>
 * The <code>source</code> member is normally <code>null</code> after the
 * all AST shadow nodes are completely built.  It is used for temporary
 * purposes.
 */
public class ShadowNode
{
   /** AST node identifier of the node to the left. */
   private Long left = null;
   
   /** AST node identifier of the node to the right. */
   private Long right = null;
   
   /** AST shadow node being stored. */
   private Aast node = null;
   
   /** Original token which generated this instance. */
   private CommonHiddenStreamToken source = null;
   
   /**
    * Obtain the ID of the left side node.
    *
    * @return    The ID or <code>null</code> if there is no left side node.
    */
   public Long getLeft()
   {
      return left;
   }
   
   /**
    * Set the ID of the left side node.
    *
    * @param    left
    *           The ID or <code>null</code> if there is no left side node.
    */
   public void setLeft(Long left)
   {
      this.left = left;
   }
   
   /**
    * Obtain the ID of the right side node.
    *
    * @return    The ID or <code>null</code> if there is no right side node.
    */
   public Long getRight()
   {
      return right;
   }
   
   /**
    * Set the ID of the right side node.
    *
    * @param    right
    *           The ID or <code>null</code> if there is no right side node.
    */
   public void setRight(Long right)
   {
      this.right = right;
   }
   
   /**
    * Obtain the AST being stored.
    *
    * @return    The AST node.
    */
   public Aast getNode()
   {
      return node;
   }
   
   /**
    * Set the AST node.
    *
    * @param    node
    *           The AST or <code>null</code> if there is no node.
    */
   public void setNode(Aast node)
   {
      this.node = node;
   }
   
   /**
    * Obtain the nearest non-shadow AST in the given direction.
    *
    * @param    right
    *           <code>true</code> to obtain the right adjacent node and
    *           <code>false</code> for the left adjacent node.
    *
    * @return   The ID of the AST node or <code>null</code> if no non-shadow
    *           node exists in that direction.
    */
   public Long getAdjacentNonShadow(boolean right)
   {
      TreeLocal local = node.getLocal();
      
      ShadowNode last   = this;
      Long       lastid = right ? last.right : last.left;
      ShadowNode next   = local.getShadowNode(lastid);
      
      while (next != null)
      {
         last   = next;
         lastid = right ? last.right : last.left;
         next   = local.getShadowNode(lastid);
      }
      
      // last should now point to the last shadow node in that direction
      return right ? last.right : last.left;
   }
   
   /**
    * Obtain the token used to create the AST being stored.
    *
    * @return    The AST node.
    */
   public CommonHiddenStreamToken getSource()
   {
      return source;
   }
   
   /**
    * Set the token used to create the AST node.
    *
    * @param    source
    *           The token or <code>null</code> if there is no node.
    */
   public void setSource(CommonHiddenStreamToken source)
   {
      this.source = source;
   }
}