/*
** Module   : TreeLocal.java
** Abstract : tree-specific data container and methods
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 GES 20100922 First version.
** 002 GES 20101001 Added helper/accessor methods. Added new mappings for
**                  using the contained node ID and the source token as keys.
** 003 EVL 20160222 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.ast;

import java.util.*;
import antlr.*;

/**
 * Container for data that is specific to a given tree, but not to any one
 * node.  This tree-local data is common to all nodes in the same tree. This
 * also contains methods to manage the access/lookup/traversal of this data.
 */
public class TreeLocal
{
   /** Mapping of shadow nodes by left-side identifier. */
   private Map<Long, ShadowNode> shadowLeft = new HashMap<Long, ShadowNode>();
   
   /** Mapping of shadow nodes by right-side identifier. */
   private Map<Long, ShadowNode> shadowRight = new HashMap<Long, ShadowNode>();
   
   /** Mapping of shadow nodes by node ID. */
   private Map<Long, ShadowNode> shadows = new HashMap<Long, ShadowNode>();
   
   /** Mapping of shadow nodes by the source token. */
   private Map<CommonHiddenStreamToken, ShadowNode> shadowTokens = null;
   
   /** All shadow nodes, in the order read from the original input stream. */
   private List<ShadowNode> shadowList = new ArrayList<ShadowNode>();
   
   /**
    * Get the size of the shadow list.
    *
    * @return    The integer size of the shadow list.
    */
   public int shadowListSize()
   {
      return shadowList.size();
   }
   
   /**
    * Add a shadow node to the list and map the left and right lookups as
    * needed.
    *
    * @param    sn
    *           The shadow node to add.
    */
   public void addShadowNode(ShadowNode sn)
   {
      shadowList.add(sn);
      insertMappings(sn);
   }
   
   /**
    * Add a shadow node to the list at the given index position and map the
    * left and right lookups as needed.
    *
    * @param    idx
    *           Index position at which to add the node.
    * @param    sn
    *           The shadow node to add.
    */
   public void addShadowNode(int idx, ShadowNode sn)
   {
      shadowList.add(idx, sn);
      insertMappings(sn);
   }
   
   /**
    * Access the shadow node at the given index position.
    *
    * @param    idx
    *           0-based index.
    *
    * @return   The shadow node at that index or <code>null</code> if no such
    *           element exists.
    */
   public ShadowNode getShadowNode(int idx)
   {
      if (idx < 0 || idx >= shadowList.size())
         return null;
      
      return shadowList.get(idx);
   }
   
   /**
    * Access the shadow node which contains the AST with the given ID.
    *
    * @param    id
    *           AST ID contained in the shadow node.
    *
    * @return   The shadow node or <code>null</code> if no shadow node exists
    *           which contains that AST.
    */
   public ShadowNode getShadowNode(Long id)
   {
      return shadows.get(id);
   }
   
   /**
    * Access the shadow node which was created from the given token.
    *
    * @param    token
    *           Source token.
    *
    * @return   The shadow node or <code>null</code> if no such element exists.
    */
   public ShadowNode getShadowNode(CommonHiddenStreamToken token)
   {
      return (shadowTokens == null) ? null : shadowTokens.get(token);
   }
   
   /**
    * Access the list of all shadow nodes.
    *
    * @return   All shadow nodes.
    */
   public Iterator<ShadowNode> getShadowNodeList()
   {
      return shadowList.iterator();
   }
   
   /**
    * Access a shadow node that exists to the left of the identified node.
    *
    * @param    id
    *           The ID of the node to the RIGHT of the shadow node. 
    *
    * @return   The shadow node that is left of the current node or 
    *           <code>null</code> if no such node exists.
    */
   public ShadowNode getLeftShadowNode(Long id)
   {
      return getShadowNodeWorker(id, false);
   }
   
   /**
    * Access a shadow node that exists to the right of the identified node.
    *
    * @param    id
    *           The ID of the node to the LEFT of the shadow node. 
    *
    * @return   The shadow node that is right of the current node or 
    *           <code>null</code> if no such node exists.
    */
   public ShadowNode getRightShadowNode(Long id)
   {
      return getShadowNodeWorker(id, true);
   }
   
   /**
    * Map the left and right lookups for the given node.
    *
    * @param    sn
    *           The shadow node to map.
    */
   private void insertMappings(ShadowNode sn)
   {
      shadows.put(sn.getNode().getId(), sn);
      
      // the right-side identifier is mapped as the key here so that when one
      // is looking for the "shadow node to the left" of this id, this node
      // will be found
      if (sn.getRight() != null)
         shadowLeft.put(sn.getRight(), sn);
         
      // the left-side identifier is mapped as the key here so that when one
      // is looking for the "shadow node to the right" of this id, this node
      // will be found
      if (sn.getLeft() != null)
         shadowRight.put(sn.getLeft(), sn);
      
      CommonHiddenStreamToken source = sn.getSource();
      
      // if the source code is available, add a mapping based on that instance
      if (source != null)
      {
         if (shadowTokens == null)
         {
            shadowTokens = new HashMap<CommonHiddenStreamToken, ShadowNode>();
         }
         shadowTokens.put(source, sn);
      }
   }
   
   /**
    * Access the shadow node on the left or right of the current node.
    *
    * @param    id
    *           The ID of the node to the LEFT of the shadow node. 
    * @param    right
    *           <code>true</code> to obtain the right side node,
    *           <code>false</code> to obtain the left side node.
    *
    * @return   The node found or <code>null</code> if no such node exists.
    */
   private ShadowNode getShadowNodeWorker(Long id, boolean right)
   {
      return (right ? shadowRight : shadowLeft).get(id);
   }
}