/*
** Module   : ClassConstant.java
** Abstract : Representation of CONSTANT_Class_info class file structure
**
** Copyright (c) 2001-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20041130 ADD   @      Production version from JTA, reworked
**                               into a new package and with minor cleanups
**                               in javadoc.
** 002 ECF 20050527 CHG   @      Expression engine rewrite. Minor cleanups and
**                               addition of a toString method.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.expr;

import java.io.*;

/**
 * This class represents the <code>CONSTANT_Class_info</code> structure
 * defined by the Java Virtual Machine Specification. This is a constant
 * in a class file's constant pool which refers to a class within the JVM.
 * <p>
 * The constructor for this class is private. To create an instance,
 * the static {@link #create} method is used. This method will ensure that
 * the constant is properly added to the constant pool.
 */
final class ClassConstant
extends Constant
{
   /** Constant for the name of the class represented by this constant */
   private Utf8Constant nameConstant;
   
   /**
    * Return an instance of this class, either newly created, or retrieved
    * from the specified constant pool.
    *
    * @param   pool
    *          Constant pool from which to retrieve instance of this class,
    *          if one with the same name already has been created.
    * @param   name
    *          Name of the class represented by this constant.
    *
    * @return  Instance of this class, either newly created, or retrieved
    *          from the specified constant pool.
    */
   static ClassConstant create(ConstantPool pool, String name)
   {
      return (ClassConstant) pool.get(new ClassConstant(pool, name));
   }
   
   /**
    * Create a new instance of this class, given a class name.
    *
    * @param   pool
    *          Constant pool from which to retrieve the Utf8Constant which
    *          represents the name of the class.
    * @param   name
    *          Name of the class represented by the new constant.
    */
   private ClassConstant(ConstantPool pool, String name)
   {
      super(CLASS);
      
      nameConstant = Utf8Constant.create(pool, name);
   }
   
   /**
    * Return the name of the class represented by this constant.
    *
    * @return  Class name.
    */
   public String getName()
   {
      return nameConstant.getText();
   }
   
   /**
    * Return the unique hash code for this object.
    *
    * @return  Unique hash code for this object.
    */
   public int hashCode()
   {
      return (int) (hashBase + nameConstant.hashCode());
   }
   
   /**
    * This method returns true if this object is equal to the specified
    * object.  Equality is defined as both objects being instances of this
    * class and the class name being the same.
    *
    * @param   o
    *          Another ClassConstant object to check for equality.
    *
    * @return  True if objects are equal, else false.
    */
   public boolean equals(Object o)
   {
      if (!(o instanceof ClassConstant))
      {
         return false;
      }
      
      return nameConstant.equals(((ClassConstant) o).nameConstant);
   }
   
   /**
    * This method compares this object to the specified object to determine
    * the natural ordering among the two.  The superclass' <code>compareTo()
    * </code> method is first called, which sorts the objects according to
    * the constant types.  If this comparison indicates the objects compare
    * as equal, the class names are then compared.
    *
    * @param   o
    *          Another object to compare to this one, presumably another
    *          subclass of <code>Constant</code>.
    *
    * @return  0 if objects compare as equal, a negative number if this
    *          object compares as less than the specified object, or a
    *          positive number if this object compares as greater than the
    *          specified object.
    */
   public int compareTo(Object o)
   {
      int comp = super.compareTo(o);
      if (comp == 0)
      {
         comp = nameConstant.compareTo(((ClassConstant) o).nameConstant);
      }
      
      return comp;
   }
   
   /**
    * Return a string representation of this constant, which is simply the
    * name of the class it represents.
    *
    * @return  Name of the class this constant represents.
    */
   public String toString()
   {
      return nameConstant.toString();
   }
   
   /**
    * This method writes the index of the name constant out to the
    * specified output stream.
    *
    * @param   out
    *          Output stream to which text is written.
    *
    * @throws  IOException
    *          If error occurs writing to stream.
    */
   protected void writeContent(DataOutput out)
   throws IOException
   {
      out.writeShort(nameConstant.getCpIndex());
   }
}