/*
** Module   : ClassFile.java
** Abstract : Writes a binary java class file to a stream
**
** Copyright (c) 2001-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20041130   @21323 Production version from JTA, reworked
**                           into a new package and with minor cleanups
**                           in javadoc.
** 002 ECF 20050505   @21324 Expression engine rewrite. Replaced
**                           references to ExpressionCompiler with
**                           Compiler.
** 003 ECF 20150726          Replaced StringBuffer with StringBuilder.
** 004 EVL 20160222          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.expr;

import java.io.*;
import java.util.*;

/**
 * This class encapsulates knowledge about the Java class file format as
 * defined in <u>The Java Virtual Machine Specification, Second Edition</u>
 * by Tim Lindholm and Frank Yellin.  It contains methods to store and
 * retrieve data specific to a class file, and to write a binary class file
 * to an output stream.  The class file written is compliant with the JVM
 * specification referenced above.
 */
public class ClassFile
implements WritableData,
           CompilerConstants
{
   /** Public access flag. */
   public static final int ACC_PUBLIC = 0x0001;
   
   /** Private access flag. */
   public static final int ACC_PRIVATE = 0x0002;
   
   /** Protected access flag. */
   public static final int ACC_PROTECTED = 0x0004;
   
   /** Static access flag. */
   public static final int ACC_STATIC = 0x0008;
   
   /** Final access flag. */
   public static final int ACC_FINAL = 0x0010;
   
   /** Super access flag. */
   public static final int ACC_SUPER = 0x0020;
   
   /** Synchronized access flag. */
   public static final int ACC_SYNCHRONIZED = 0x0020;
   
   /** Volatile access flag. */
   public static final int ACC_VOLATILE = 0x0040;
   
   /** Transient access flag. */
   public static final int ACC_TRANSIENT = 0x0080;
   
   /** Native access flag. */
   public static final int ACC_NATIVE = 0x0100;
   
   /** Interface access flag. */
   public static final int ACC_INTERFACE = 0x0200;
   
   /** Abstract access flag. */
   public static final int ACC_ABSTRACT = 0x0400;
   
   /** Strict access flag. */
   public static final int ACC_STRICT = 0x0800;
   
   /** Fully qualified name of the Object class. */
   public static final String OBJECT_CLASSNAME = "java/lang/Object";
   
   /** Magic number which is first four bytes of a valid class file */
   private static final int MAGIC_NUMBER = 0xCAFEBABE;
   
   /** Minor version expected by JVM. */
   private static final int MINOR_VERSION = 3;
   
   /** Major version expected by JVM. */
   private static final int MAJOR_VERSION = 45;
   
   /** Constant pool for this class. */
   private ConstantPool pool = new ConstantPool();
   
   /** Bitwise or'ed access flags. */
   private int accessFlags = 0;
   
   /** Class constant representing current class being compiled. */
   private ClassConstant thisClass = null;
   
   /** Class constant representing parent of class being compiled. */
   private ClassConstant superClass = null;
   
   /** List of interfaces implemented by current class. */
   private List interfaces = new ArrayList();
   
   /** List of fields contained within current class. */
   private List fields = new ArrayList();
   
   /** List of methods contained within current class. */
   private List methods = new ArrayList();
   
   /** List of attributes associated with current class. */
   private List attrs = new ArrayList();
   
   /**
    * Constructor.  Assumes <code>java/lang/Object</code> as superclass
    * of class being compiled.
    *
    * @param   className
    *          Fully qualified name of class to be compiled, in the form
    *          <code>package_qualifier/class_name</code>.
    */
   public ClassFile(String className)
   {
      this(className, OBJECT_CLASSNAME);
   }
   
   /**
    * Constructor.  Creates class constants for current class and its
    * parent class, and adds them to the constant pool.
    *
    * @param   className
    *          Fully qualified name of class to be compiled, in the form
    *          <code>package_qualifier/class_name</code>.
    * @param   superName
    *          Fully qualified name of superclass of class to be compiled,
    *          in the form <code>package_qualifier/class_name</code>.
    */
   public ClassFile(String className, String superName)
   {
      thisClass = ClassConstant.create(pool, className);
      if (superName == null)
      {
         superName = OBJECT_CLASSNAME;
      }
      superClass = ClassConstant.create(pool, superName);
   }
   
   /**
    * Get a reference to this class file's constant pool.
    *
    * @return  Constant pool for current class.
    */
   public ConstantPool getConstantPool()
   {
      return pool;
   }
   
   /**
    * Return class constant object for the current class' parent.
    *
    * @return  Class constant representing superclass of current class.
    */
   public ClassConstant getSuperClass()
   {
      return superClass;
   }
   
   /**
    * Return the fully qualified name of the current class, substituting
    * <code>&#46;</code> characters for <code>/</code> characters.    
    *
    * @return  Name of current class, qualified by package name.
    */
   public String getFullyQualifiedName()
   {
      String name = thisClass.getName();
      StringBuilder buf = new StringBuilder();
      int size = name.length();
      for (int i = 0; i < size; i++)
      {
         char ch = name.charAt(i);
         if (ch == '/')
            buf.append('.');
         else
            buf.append(ch);
      }
      
      return buf.toString();
   }
   
   /**
    * This method adds a default constructor with the specified access
    * flags.  The constructor does nothing more than invoking the
    * superclass' default constructor.
    *
    * @param   accessFlags
    *          Access flags to assign to constructor.
    * @param   compiler
    *          The compiler to use.
    *
    * @return  <code>MethodInfo</code> object representing the constructor.
    */
   public MethodInfo addDefaultConstructor(int accessFlags,
                                           Compiler compiler)
   {
      MethodInfo mi = new MethodInfo(pool, "<init>", "()V", accessFlags);
      CodeAttribute code = new CodeAttribute(pool);
      code.addInstruction(new Instruction(_aload_0));
      int[] tmp = {0};
      Instruction instruc = new Instruction(_invokespecial, tmp);
      MethodrefConstant mConst = MethodrefConstant.create(
         pool, superClass.getName(), "<init>", "()V");
      instruc.addArgument(new ConstantArgument(mConst));
      code.addInstruction(instruc);
      code.addInstruction(new Instruction(_return));
      code.setMaxLocals(1);
      code.setMaxStack(compiler);
      mi.addAttribute(code);
      addMethod(mi);
      
      return mi;
   }
   
   /**
    * Set the access flags for this class file.
    *
    * @param   accessFlags
    *          Bitwise or'ed combination of access flags.
    */
   public void setAccessFlags(int accessFlags)
   {
      this.accessFlags = accessFlags;
   }
   
   /**
    * Add an interface to the constant pool of the current class.  The
    * interface is represented in the constant pool as a class constant.
    *
    * @param   name
    *          Fully qualified name of the interface to be added, in the form
    *          <code>package_qualifier/interface_name</code>.
    *
    * @return  Class constant which represents the interface.
    */
   public ClassConstant addInterface(String name)
   {
      ClassConstant rv = ClassConstant.create(pool, name);
      interfaces.add(rv);
      return rv;
   }
   
   /**
    * Add a method to the current class file.
    *
    * @param   methodInfo
    *          Object which describes the method.
    */
   public void addMethod(MethodInfo methodInfo)
   {
      methods.add(methodInfo);
   }
   
   /**
    * Write the class file to a binary output stream in the format defined
    * by the JVM specification.  The following items are written to the
    * stream, in order:
    * <ol>
    * <li>File header
    * <li>Constant pool
    * <li>Access flags
    * <li>Class information for this class and its parent
    * <li>Interfaces implemented, if any
    * <li>Fields, if any (currently none implemented)
    * <li>Methods, if any
    * <li>Class file level attributes, if any (currently none implemented)
    * </ol>
    *
    * @param   out
    *          Data output stream.
    *
    * @throws  IOException
    *          if any error writing to output stream.
    */
   public void write(DataOutput out)
   throws IOException
   {
      // Write header.
      out.writeInt(MAGIC_NUMBER);
      out.writeShort(MINOR_VERSION);
      out.writeShort(MAJOR_VERSION);
      
      // Write constant pool.
      pool.write(out);
      
      // Write access flags.
      out.writeShort(accessFlags);
      
      // Write class info.
      out.writeShort(thisClass.getCpIndex());
      out.writeShort(superClass.getCpIndex());
      
      // Write interfaces.
      out.writeShort(interfaces.size());
      Iterator iter = interfaces.iterator();
      while (iter.hasNext())
      {
         ClassConstant next = (ClassConstant) iter.next();
         out.writeShort(next.getCpIndex());
      }
      
      // Write fields.
      out.writeShort(fields.size());
      
      // Write methods.
      out.writeShort(methods.size());
      iter = methods.iterator();
      while (iter.hasNext())
      {
         MethodInfo next = (MethodInfo) iter.next();
         next.write(out);
      }
      
      // Write attributes.
      out.writeShort(attrs.size());
   }
}