/*
** Module   : Constant.java
** Abstract : 
**
** Copyright (c) 2001-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20041130 ADD   @xxxxx Production version from JTA, reworked
**                               into a new package and with minor cleanups
**                               in javadoc.
** 005 EVL 20160222              Javadoc fixes to make compatible with Oracle Java 8 for
**                               Solaris 10.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.expr;


import java.io.*;


/**
 * This is the parent class to all of the various types of constants
 * supported by the Java class file format.  It provides services to sort
 * constants according to their natural order, to ensure unique hash
 * codes are generated across different constant types, and to store and
 * retrieve a constant's index in its constant pool.  It also provides
 * a framework which all constants use when writing themselves to a
 * binary format during the compilation process.  Finally, it ensures
 * that each constant is aware of its width for purposes of the constant
 * pool, based upon the type of the constant.
 */
abstract class Constant implements Comparable, WritableData
{

   /** Class constant value. */
   protected static final int CLASS = 7;

   /** Fieldref constant value. */
   protected static final int FIELDREF = 9;

   /** Methodref constant value. */
   protected static final int METHODREF = 10;

   /** InterfaceMethodref constant value. */
   protected static final int INTERFACEMETHODREF = 11;

   /** String constant value. */
   protected static final int STRING = 8;

   /** Integer constant value. */
   protected static final int INTEGER = 3;

   /** Float constant value. */
   protected static final int FLOAT = 4;

   /** Long constant value. */
   protected static final int LONG = 5;

   /** Double constant value. */
   protected static final int DOUBLE = 6;

   /** NameAndType constant value. */
   protected static final int NAMEANDTYPE = 12;

   /** Utf8 constant value. */
   protected static final int UTF8 = 1;

   /** Natural order for each constant type within the constant pool. */
   private static final int[] NATURAL_ORDER = {
      12,   // CONSTANT_Utf8
      20,   // Undefined
       6,   // CONSTANT_Integer
       7,   // CONSTANT_Float
       8,   // CONSTANT_Long
       9,   // CONSTANT_Double
       2,   // CONSTANT_Class
       1,   // CONSTANT_String
       3,   // CONSTANT_Fieldref
       4,   // CONSTANT_Methodref
       5,   // CONSTANT_InterfaceMethodref
      11    // CONSTANT_NameAndType
   };

   /** Width in constant pool "slots" (index positions) for each type. */
   private static final int[] WIDTH = {
      1,   // CONSTANT_Utf8
      0,   // Undefined
      1,   // CONSTANT_Integer
      1,   // CONSTANT_Float
      2,   // CONSTANT_Long
      2,   // CONSTANT_Double
      1,   // CONSTANT_Class
      1,   // CONSTANT_String
      1,   // CONSTANT_Fieldref
      1,   // CONSTANT_Methodref
      1,   // CONSTANT_InterfaceMethodref
      1    // CONSTANT_NameAndType
   };

   /** Constant value tag for this constant object. */
   protected int tag;

   /** Base number used when determining a hashcode for subclasses. */
   protected int hashBase;

   /** Index of this constant within the constant pool. */
   private int cpIndex;


   /** 
    * Constructor.  Stores tag value and generates a hash code base number,
    * based upon the tag value.
    *
    * @param   tag
    *          One of the type value constants.
    */
   Constant(int tag)
   {
      this.tag = tag;
      hashBase = tag << 4;
   }


   /** 
    * This method determines the relative ordering of the current constant
    * with the constant specified, based upon the natural ordering of the
    * constant types within the constant pool.  This ordering is relevant
    * to the maintenance of the constant pool.
    *
    * @param   o
    *          A <code>Constant</code> object with which to compare this
    *          <code>Constant</code> for purposes of sorting.
    *
    * @return  A negative value if this constant is considered less than
    *          the specified constant, 0 if they are considered to have
    *          the same sort order, or a positive value if this constant
    *          is considered greater than the specified constant.
    */
   public int compareTo(Object o)
   {

      int thisOrder = NATURAL_ORDER[tag - 1];
      int thatOrder = NATURAL_ORDER[((Constant) o).tag - 1];
      int comp = thisOrder - thatOrder;
      return comp;

   }


   /** 
    * Return the width of this constant in terms of constant pool "slots".
    *
    * @return  Width of the constant.
    */
   final int getWidth()
   {
      return WIDTH[tag - 1];
   }


   /** 
    * Return the index assigned to this constant within the constant pool.
    * Note that this index is not assigned until the constant pool is
    * indexed, which does not happen until it is about to be written into
    * a class file.  If this method is called before the constant pool is
    * indexed, it will always return 0.
    *
    * @return  Index of this constant in the constant pool.
    */
   final int getCpIndex()
   {
      return cpIndex;
   }


   /** 
    * Assign the index of this constant within the constant pool.  This
    * method is intended only to be called by the constant pool itself
    * at the time indices are being determined.
    *
    * @param   cpIndex
    *          Constant pool index to be assigned to this constant.
    */
   final void setCpIndex(int cpIndex)
   {
      this.cpIndex = cpIndex;
   }


   /** 
    * Write the binary format of this constant to the specified output
    * stream.  This will always begin with a single byte, which is the
    * tag value identifying the constant type.  Subsequent writes are
    * left to the subclass to implement.
    *
    * @param   out
    *          Output stream to which to write the binary representation
    *          of this constant.
    *
    * @exception  IOException
    *             if any error writing to stream.
    */
   public void write(DataOutput out) throws IOException {
      out.writeByte(tag);
      writeContent(out);
   }


   /** 
    * Return the string representation of this constant, primarily for
    * debugging purposes.  This representation simply consists of the
    * tag value and constant pool index value.
    *
    * @return  String representation of this constant.
    */
   public String toString()
   {
      return "Tag = " + tag + ", index = " + cpIndex;
   }


   /** 
    * This method must be implemented by a subclass to provide the
    * appropriate behavior when writing the binary representation of
    * this constant to an output stream.
    *
    * @param   out
    *          Output stream to which to write the portion of the binary
    *          representation of this constant which is specific to the
    *          subclass.
    *
    * @exception  IOException
    *             if any error writing to stream.
    */
   protected abstract void writeContent(DataOutput out) throws IOException;


}