/*
** Module   : ConstantArgument.java
** Abstract : An instruction argument which is a constant
**
** Copyright (c) 2001-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20041130 ADD   @xxxxx Production version from JTA, reworked
**                               into a new package and with minor cleanups
**                               in javadoc.
** 002 ECF 20050201 CHG   @xxxxx Minor formatting cleanup.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.expr;

import java.io.*;

/**
 * This class represents an instruction argument which is a constant.
 */
public final class ConstantArgument
extends Argument
implements WritableData
{
   /** Constant represented by this argument */
   private Constant constant;
   
   /**
    * This constructor assumes a constant pool index width of two bytes,
    * which is the norm for most opcodes.
    *
    * @param   constant
    *          Constant object whose constant pool index will comprise
    *          the data of this instruction argument.
    */
   public ConstantArgument(Constant constant)
   {
      this(constant, SHORT);
   }
   
   /**
    * Constructor.
    *
    * @param   constant
    *          Constant object whose constant pool index will comprise
    *          the data of this instruction argument.
    * @param   width
    *          Width of this argument's data, in bytes.
    */
   public ConstantArgument(Constant constant, int width)
   {
      super(width);
      this.constant = constant;
   }
   
   /**
    * This method writes this argument's data to the specified output
    * stream.  The argument's data is simply the constant pool index of
    * the constant, written as a numeric value of the appropriate
    * width.
    *
    * @param   out
    *          Data output stream to which to write argument data.
    *
    * @exception  IOException
    *             if any error occurs while writing.
    */
   public void write(DataOutput out)
   throws IOException
   {
      int index = constant.getCpIndex();
      
      switch (width)
      {
         case BYTE:
            out.writeByte(index);
            break;
         case SHORT:
            out.writeShort(index);
            break;
         case INT:
            out.writeInt(index);
            break;
         default:
            break;
      }
   }
   
   public String toString()
   {
      return constant.toString();
   }
}