/*
** Module   : TestDriver.java
** Abstract : Test driver and sample client for compiled expressions package.
**
** Copyright (c) 2001-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20041209 ADD   @18948 Created initial version for illustrative and
**                               test purposes. Note that this implementation
**                               has a dependency on the p2j.util package (for
**                               properties file loading support). This should
**                               be replaced if a backport is required.
** 002 ECF 20041213 CHG   @19003 Turn off debug by default.
** 003 ECF 20050201 CHG   @19665 Added support for functions to leverage the
**                               changes in the expression compiler classes.
**                               Added support for Boolean data. Now implement
**                               SymbolResolver instead of VariableResolver.
** 004 ECF 20050215 CHG   @19827 Replaced function processing added for #003
**                               (@19665) with an implementation which uses
**                               the new function framework in this package.
**                               All function support is now provided by a
**                               separate, inner class which is registered as
**                               a function library with the FunctionRegistry
**                               class. This class now extends
**                               AbstractSymbolResolver, which handles all of
**                               the function resolution and invocation
**                               transparently. This has allowed us to drop
**                               all of the resolveToXXXX method variants for
**                               functions.
** 005 ECF 20050222 CHG   @20086 Added some runtime variables to the inner,
**                               SymbolLibrary class (previously named
**                               FunctionLibrary) to test new runtime variable
**                               support in the SymbolRegistry and superclass,
**                               AbstractSymbolResolver. Also modified
**                               resolveToLong method to call superclass'
**                               implementation first.
** 006 ECF 20050303 ADD   @20154 Added boolTest method to SymbolLibrary class.
**                               This is a regression test for fix H005 to the
**                               ExpressionCompiler (@20153).
** 007 ECF 20050323 ADD   @20473 Added isTest3 method to SymbolLibrary class.
**                               This is a regression test for fix H002 to the
**                               ReflectiveVariable class (@20472).
** 008 ECF 20050608 CHG   @21464 Disabled temporarily.
** 009 EVL 20160223              Javadoc fixes to make compatible with Oracle Java 8 for
**                               Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.expr;

import java.text.*;
import java.util.*;
import com.goldencode.p2j.util.*;

/**
 * Test driver and sample client code for the compiled expressions package.
 * Manages a simple database of employee records configured in a properties
 * format and permits expressions to be evaluated against all or some of the
 * records. Accepts either arithmetic or logical expression types. Expressions
 * may contain variables, numeric constants, string literals, string
 * constants, including formated dates and times, and any of the operators
 * supported by the compiled expressions package.
 * <p>
 * This class is accessible both from the command line (see {@link #main} for
 * required arguments) and programatically (see the following methods:
 * <ul>                                   
 * <li><code>initialize</code></li>
 * <li><code>computeResults</code> (of an arithmetic expression)</li>
 * <li><code>compute</code> (an arithmetic expression)</li>
 * <li><code>evaluateResults</code> (of a logical expression)</li>
 * <li><code>evaluate</code> (a logical expression)</li>
 * </ul>
 * <p>
 * Where this class must interface to the expression compiler, the data type
 * constants from {@link SymbolResolver} are used. However, this class
 * supports the concept of application-defined user types. This abstraction
 * of data typing permits application code to do its own interpretation of
 * expression tokens at compile time (for instance when the compiler calls
 * <code>resolveConstant</code>), or when initializing variables from
 * the properties file. Ultimately, however, any expression token which will
 * be resolved at expression compile time or runtime must be mapped to a
 * <code>SymbolResolver</code> type constant.
 * <p>
 * As an example, this test driver supports the notion of an application
 * defined <code>date</code> data type. This permits the implementation of
 * more advanced parsing of formated date strings during initialization,
 * or during expression compilation, when a string constant formated as a
 * date is encountered. The application defined <code>date</code> data type
 * is mapped to a long internally. When a formated date
 * string or a variable of type <code>date</code> must be resolved, it is
 * converted to a <code>Long</code> value representing the number of
 * milliseconds since 1970.01.01.
 * <p>
 * A sample set of properties follows:
 * <pre>
 * #
 * # Data for a sample variable resolver. Sets up a simple database of employee
 * # records.
 * #
 * 
 * 
 * #
 * # Acceptable formats for date/time constants in expressions. These determine
 * # how date and time information are parsed by the application.
 * #
 * 
 * date.format.0 = yyyy.MM.dd
 * date.format.1 = MM/dd/yyyy
 * date.format.2 = MMM dd, yyyy
 * date.format.3 = HH:mm:ss
 * date.format.4 = yyyy/MM/dd HH:mm:ss
 * 
 * #
 * # Boolean constants. These will be reserved keywords in the application.
 * # 'true' and 'false' are directly supported by the compiler.
 * #
 * 
 * boolean.constant.yes   = true
 * boolean.constant.no    = false
 * boolean.constant.T     = true
 * boolean.constant.F     = false
 * boolean.constant.yup   = true
 * boolean.constant.nope  = false
 * 
 * #
 * # Variable data types. The meanings of these data types are specific to the
 * # testcase application, which uses them as clues for parsing.
 * #
 * 
 * # employee name
 * type.name = string
 * # employee date of birth
 * type.dob = date
 * # number of overtime hours (fractional)
 * type.overtime = double
 * # employee's city
 * type.city = string
 * # employee's union affiliation status
 * type.union = boolean
 * # time shift begins (millis since midnight)
 * type.begin = date
 * # time shift ends (millis since midnight)
 * type.end = date
 * 
 * #
 * # "Employee" records. Groups of variables which have the same index number
 * # in their property key name comprise the same record.
 * #
 * 
 * emp.1.name     = Larry
 * emp.1.dob      = 1950.01.14
 * emp.1.overtime = 0.
 * emp.1.city     = Boston
 * emp.1.union    = true
 * emp.1.begin    = 09:00:00
 * emp.1.end      = 17:30:00
 * 
 * emp.2.name     = Moe
 * emp.2.dob      = 1960.05.21
 * emp.2.overtime = 5.5
 * emp.2.city     = St. Louis
 * emp.2.union    = true
 * emp.2.begin    = 06:00:00
 * emp.2.end      = 15:00:00
 * 
 * emp.3.name     = Curly
 * emp.3.dob      = 1945.11.13
 * emp.3.overtime = 24.25
 * emp.3.city     = New York
 * emp.3.union    = false
 * emp.3.begin    = 07:30:00
 * emp.3.end      = 17:15:00
 * </pre>
 *
 * @author  ECF
 */
public class TestDriver
//extends SymbolResolver
{
   /** Expression class files are dumped to current directory when true */
   private static final boolean debug = false;
   
   /** Date format keyword prefix in initialization properties */
   private static final String PFX_DATE_FORMAT = "date.format.";
   
   /** Boolean constant keyword prefix in initialization properties */
   private static final String PFX_BOOL_CONSTANT = "boolean.constant.";
   
   /** Variable data type keyword prefix in initialization properties */
   private static final String PFX_TYPE = "type.";
   
   /** Employee variable keyword prefix in initialization properties */
   private static final String PFX_VAR_EMP = "emp.";
   
   /** Application defined boolean data type */
   private static final String APPTYPE_BOOLEAN = "boolean";
   
   /** Application defined long data type */
   private static final String APPTYPE_LONG = "long";
   
   /** Application defined date data type */
   private static final String APPTYPE_DATE = "date";
   
   /** Application defined double data type */
   private static final String APPTYPE_DOUBLE = "double";
   
   /** Application defined string data type */
   private static final String APPTYPE_STRING = "string";
   
   /** Map of app-defined type names to SymbolResolver type constants */
   private static final Map TYPE_CONSTANTS = new HashMap();
   
   static
   {
      // Map application-defined data types to SymbolResolver data type
      // constants.
      /*
      TYPE_CONSTANTS.put(APPTYPE_BOOLEAN, new Integer(TYPE_BOOLEAN));
      TYPE_CONSTANTS.put(APPTYPE_LONG   , new Integer(TYPE_LONG));
      TYPE_CONSTANTS.put(APPTYPE_DATE   , new Integer(TYPE_LONG));
      TYPE_CONSTANTS.put(APPTYPE_DOUBLE , new Integer(TYPE_DOUBLE));
      TYPE_CONSTANTS.put(APPTYPE_STRING , new Integer(TYPE_STRING));
      */
   }
   
   /** List of acceptable date formats, in order of use when parsing */
   private List dateFormats = new ArrayList();
   
   /** Map of symbol names to app-defined data type names */
   private Map symbolAppTypeMap = new HashMap();
   
   /** Map of reserved keywords to constant values */
   private Map reservedMap = new HashMap();
   
   /** List of data records, in index order */
   private List records = new ArrayList();
   
   /** Data record currently being used in an expression evaluation */
   private Map currentRecord = null;
   
   /** Expression compiler */
//   private ExpressionCompiler compiler = new ExpressionCompiler();
   
   /**
    * Default constructor.
    */
   public TestDriver()
   throws NoSuchMethodException
   {
      super();
      
//      registerLibrary(SymbolLibrary.class, new SymbolLibrary());
   }
   
   /**
    * Initialize the driver by storing all acceptable date format strings,
    * boolean constants, application-defined data type to variable mappings,
    * and variable values. This is done in two iteration passes through the
    * specified <code>Properties</code>. Two passes are necessary because
    * we iterate through the keys in <code>props</code> in an indeterminate
    * order. It is necessary to know about constant values, date format
    * strings, and the variables' data type mappings (all collected during
    * the first pass) in order to parse their values correctly in the second
    * pass.
    * <p>
    * <strong>Important:</strong> in order for this test driver to do
    * anything useful at expression runtime, this method must be called
    * beforehand to populate the driver with data. Otherwise, no variables
    * or constants used in expressions will be resolvable.
    * <p>
    * See the {@link TestDriver class description} for a sample of the
    * properties format which is expected by this method.
    *
    * @param   props
    *          Configuration properties
    */
   private void initialize(final Properties props)
   {
      // true and false constants are reserved.
      reservedMap.put("true", Boolean.TRUE);
      reservedMap.put("false", Boolean.FALSE);
      
      // First pass - read dependency data: date format strings, boolean
      // constants, and variable data types.
      Map dateFormatMap = new TreeMap();
      Iterator iter = props.keySet().iterator();
      while (iter.hasNext())
      {
         String key = (String) iter.next();
         
         if (key.startsWith(PFX_DATE_FORMAT))
         {
            String val = (String) props.getProperty(key);
            readDateFormat(key, val, dateFormatMap);
         }
         else if (key.startsWith(PFX_BOOL_CONSTANT))
         {
            String val = (String) props.getProperty(key);
            readBooleanConstant(key, val);
         }
         else if (key.startsWith(PFX_TYPE))
         {
            String val = (String) props.getProperty(key);
            readDataType(key, val);
         }
      }
      
      // Copy date format elements into list in the order in which they are
      // used to parse date/time constants.
      iter = dateFormatMap.keySet().iterator();
      while (iter.hasNext())
      {
         dateFormats.add(dateFormatMap.get(iter.next()));
      }
      
      // Second pass - read variable data and parse based on the information
      // collected in the first pass.
      Map dataRecordMap = new TreeMap();
      iter = props.keySet().iterator();
      while (iter.hasNext())
      {
         String key = (String) iter.next();
         
         if (key.startsWith(PFX_VAR_EMP))
         {
            String val = (String) props.getProperty(key);
            readDataRecordVariable(key, val, dataRecordMap);
         }
      }
      
      // Copy data records into list in the index order specified in the
      // properties.
      iter = dataRecordMap.keySet().iterator();
      while (iter.hasNext())
      {
         records.add(dataRecordMap.get(iter.next()));
      }
   }
   
   /**
    * Read and parse a date format string configuration property. Create a
    * <code>SimpleDateFormat</code> object for later parsing of date strings,
    * and store this in <code>map</code>. The object's key in <code>map</code>
    * is the numeric index extracted from the end of the property's key name.
    *
    * @param   key
    *          Configuration property key name.
    * @param   value
    *          Configuration property value as a string.
    * @param   map
    *          Map into which <code>SimpleDateFormat</code> object will be
    *          stored.
    *
    * @throws  IllegalArgumentException
    *          if <code>key</code> represents an unexpected key name.
    * @throws  NumberFormatException
    *          if a valid numeric index cannot be extracted from
    *          <code>key</code>.
    */
   private void readDateFormat(String key, String value, Map map)
   throws IllegalArgumentException,
          NumberFormatException
   {
      int index = parseIndex(key, PFX_DATE_FORMAT.length());
      SimpleDateFormat format = new SimpleDateFormat(value);
      map.put(new Integer(index), format);
   }
   
   /**
    * Read and parse a boolean constant string configuration property. The
    * value assigned to this constant should be <code>0</code> for
    * <code>false</code> and non-zero (<code>1</code>, conventionally) for
    * <code>true</code>. The last portion of the property key name represents
    * the name of the symbolic constant, which is stored as a reserved key
    * word.
    *
    * @param   key
    *          Configuration property key name.
    * @param   value
    *          Configuration property value as a string.
    *
    * @throws  IllegalArgumentException
    *          if <code>key</code> represents an unexpected key name.
    */
   private void readBooleanConstant(String key, String value)
   throws IllegalArgumentException
   {
      int start = PFX_BOOL_CONSTANT.length();
      String symbol = parseToken(key, start);
      symbolAppTypeMap.put(symbol, APPTYPE_BOOLEAN);
      reservedMap.put(symbol, new Boolean(value));
   }
   
   /**
    * Read and parse a data type configuration property. The property defines
    * the mapping of an application-defined data type to a variable (i.e., an
    * attribute of a data record).
    *
    * @param   key
    *          Configuration property key name.
    * @param   value
    *          Configuration property value as a string.
    *
    * @throws  IllegalArgumentException
    *          if <code>key</code> represents an unexpected key name.
    */
   private void readDataType(String key, String value)
   throws IllegalArgumentException
   {
      int start = PFX_TYPE.length();
      String symbol = parseToken(key, start);
      symbolAppTypeMap.put(symbol, value);
   }
   
   /**
    * Read and parse a data variable configuration property. The property
    * defines the value contained within a variable, which will be resolved
    * at runtime if an expression refers to it. Variables are grouped by
    * index into employee records within the configuration properties. Each
    * variable with the same numeric index within the property key name is
    * an attribute within the same data record. The first time a new index
    * is encountered, a new data record (simply a <code>Map</code> object)
    * is constructed and stored by index within <code>map</code>.
    *
    * @param   key
    *          Configuration property key name.
    * @param   value
    *          Configuration property value as a string.
    * @param   map
    *          Map which holds data records.
    *
    * @throws  IllegalArgumentException
    *          if <code>key</code> represents an unexpected key name or the
    *          variable name extracted from <code>key</code> collides with
    *          the name of a reserved keyword.
    * @throws  NumberFormatException
    *          if a valid numeric index cannot be extracted from
    *          <code>key</code>.
    */
   private void readDataRecordVariable(String key, String value, Map map)
   throws IllegalArgumentException,
          NumberFormatException
   {
      int start = PFX_VAR_EMP.length();
      int len = parseToken(key, start).length();
      Integer index = new Integer(parseIndex(key, start));
      String symbol = parseToken(key, start + len);
      if (reservedMap.containsKey(symbol))
      {
         throw new IllegalArgumentException(
            "'" + symbol + "' is a reserved keyword");
      }
      
      Map record = (Map) map.get(index);
      if (record == null)
      {
         record = new HashMap();
         map.put(index, record);
      }
      Object datum = parseDatum(symbol, value);
      record.put(symbol, datum);
   }
   
   /**
    * Extract a numeric value from the string <code>key</code>. The portion
    * of <code>key</code> which contains the numeric value is expected to
    * begin at character index <code>start</code>. It is expected to end
    * either at the next '<code>.</code>' (dot) character encountered or at
    * the end of <code>key</code>.
    *
    * @param   key
    *          String from which numeric value is to be extracted.
    * @param   start
    *          Starting character of numeric value within <code>key</code>.
    *
    * @return  Numeric value extracted from <code>key</code>.
    *
    * @throws  IllegalArgumentException
    *          if <code>start</code> is an out of bounds index for
    *          <code>key</code>.
    * @throws  NumberFormatException
    *          if a valid numeric index cannot be extracted from
    *          <code>key</code>.
    */
   private int parseIndex(String key, int start)
   throws IllegalArgumentException,
          NumberFormatException
   {
      return Integer.parseInt(parseToken(key, start));
   }
   
   /**
    * Extract a token substring of <code>key</code> which begins at character
    * index <code>start</code> and ends either at the next dot separator
    * character encountered or at the end of <code>key</code>.
    *
    * @param   key
    *          String from which token is to be extracted.
    * @param   start
    *          Starting character of token within <code>key</code>.
    *
    * @return  String token extracted from <code>key</code>.
    *
    * @throws  IllegalArgumentException
    *          if <code>start</code> is an out of bounds index for
    *          <code>key</code>.
    */
   private String parseToken(String key, int start)
   throws IllegalArgumentException
   {
      if (start < 0 || start > key.length())
      {
         throw new IllegalArgumentException("Invalid start index: " + start);
      }
      
      String endOfKey = key.substring(start);
      StringTokenizer tok = new StringTokenizer(endOfKey, ".");
      if (!tok.hasMoreTokens())
      {
         throw new IllegalArgumentException(key);
      }
      
      return tok.nextToken();
   }
   
   /**
    * Parse <code>value</code> into an object of the correct type, using
    * <code>key</code> as a cue to its data type. First check if
    * <code>value</code> represents a reserved keyword and if so, return
    * its constant value. Next, parse the value based on <code>key</code>'s
    * data type. If <code>key</code> is <code>null</code>, or its data type
    * cannot be determined, try parsing <code>value</code> as a date, then
    * default to treating it as a string literal.
    *
    * @param   key
    *          The key associated with the value to be parsed.
    * @param   value
    *          String to be looked up as a reserved keyword or to be parsed
    *          into an object of the correct type.
    *
    * @return  Result of parsing <code>value</code>, or <code>null</code>
    *          if it could not be parsed propertly.
    */
   private Object parseDatum(String key, String value)
   {
      // Check whether value is a reserved keyword with a constant value
      // and return that constant value if found.
      Object datum = reservedMap.get(value);
      if (datum != null)
      {
         return datum;
      }
      
      try
      {
         // Parse the datum value according to its application defined data
         // type.
         String appDataType = (String) symbolAppTypeMap.get(key);
         if (appDataType != null)
         {
            if (APPTYPE_DATE.equals(appDataType))
            {
               Date date = parseDate(stripEnclosing(value, '\''));
               if (date != null)
               {
                  datum = new Long(date.getTime());
               }
            }
            else if (APPTYPE_DOUBLE.equals(appDataType))
            {
               datum = new Double(value);
            }
            else if (APPTYPE_LONG.equals(appDataType))
            {
               datum = new Long(value);
            }
            else if (APPTYPE_STRING.equals(appDataType))
            {
               datum = stripEnclosing(value, '\'');
            }
         }
         else
         {
            // Try it as a date.
            Date date = parseDate(stripEnclosing(value, '\''));
            if (date != null)
            {
               datum = new Long(date.getTime());
            }
            
            // Last resort: interpret it as a string literal.
            if (datum == null)
            {
               datum = stripEnclosing(value, '\'');
            }
         }
      }
      catch (NumberFormatException exc)
      {
         // allow null to be returned
      }
      
      return datum;
   }
   
   /**
    * Try to parse <code>source</code> as a date string using each of the
    * acceptable date format strings provided by the configuration. The
    * order in which format patterns are attempted is that specified in the
    * configuration.
    *
    * @param   source
    *          Text to be recognized as a formated date.
    *
    * @return  The date represented by <code>source</code>, or
    *          <code>null</code> if no format pattern could be applied
    *          successfully.
    */
   private Date parseDate(String source)
   {
      Date date = null;
      Iterator iter = dateFormats.iterator();
      while (iter.hasNext())
      {
         SimpleDateFormat fmt = (SimpleDateFormat) iter.next();
         try
         {
            date = fmt.parse(source);
            break;
         }
         catch (ParseException exc)
         {
            // current format string could not be applied; try next one...
         }
      }
      
      return date;
   }
   
   /**
    * Strip specified enclosing character, if any, from the beginning and end
    * of a string. The character is only removed if it is is both the first
    * and last character of <code>text</code>.
    *
    * @param   text
    *          Text from which to remove character.
    * @param   c
    *          Character to be removed.
    *
    * @return  The adjusted string.
    */
   private String stripEnclosing(String text, char c)
   {
      int begin = text.indexOf(c);
      int end = text.lastIndexOf(c);
      if (begin == 0 && end == text.length() - 1 && begin < end)
      {
         text = text.substring(begin + 1, end);
      }
      
      return text;
   }
   
   /**
    * Query the data type associated with the variable represented by
    * <code>varName</code>.
    *
    * @param   varName
    *          Name of variable whose type is being queried.
    *
    * @return  The {@link SymbolResolver} type constant associated with
    *          the specified variable.
    */
   /*
   public int querySymbolType(String varName)
   {
      int rv = super.querySymbolType(varName);
      if (rv != TYPE_UNKNOWN)
      {
         return rv;
      }
      
      String typeName = (String) symbolAppTypeMap.get(varName);
      Integer type = (Integer) TYPE_CONSTANTS.get(typeName);
      if (type == null)
      {
         return TYPE_UNKNOWN;
      }
      
      return type.intValue();
   }
   */
   
   /**
    * Resolve the string constant represented by <code>constant</code> by
    * parsing it using {@link #parseDatum}.
    *
    * @param   constant
    *          String value of constant to be resolved.
    * @param   other
    *          Other object being compared with <code>constant</code>,
    *          if any. This is used to determine the data type of
    *          <code>constant</code>.
    *
    * @return  The object that results from parsing <code>constant</code>,
    *          using <code>other</code> as a guide to its data type.
    */
   /*
   public Object resolveConstant(String constant, Object other)
   {
      String key = null;
      if (other instanceof Variable)
      {
         key = ((Variable) other).getName();
      }
      else if (other instanceof String)
      {
         key = (String) other;
      }
      
      return parseDatum(key, constant);
   }
   */
   
   /**
    * Resolve the variable represented by <code>name</code> to a
    * <code>String</code> object which represents the contents of the
    * variable. The resulting object will be used in a binary comparison
    * within a logical expression. This method is called at runtime, each
    * time a variable is encountered in an expression, which needs to be
    * substituted by a string.
    * <p>
    * This method first tries to look up <code>name</code> in a list of
    * reserved keyword constants and returns the keyword's constant value if
    * found. Otherwise, it delegates variable resolution responsibility to
    * the current data record.
    *
    * @param   name
    *          Name of the variable to be resolved.
    *
    * @return  A string which represents the contents of the variable
    *          <code>name</code>, or <code>null</code> if
    *          <code>name</code> cannot be resolved to a numeric value.
    *          Note that if <code>null</code> is returned, a comparison
    *          which attempts to match this variable for equality will
    *          return <code>false</code>.
    */
   /*
   public String resolveToString(String name)
   {
      Object value = reservedMap.get(name);
      if (value == null)
      {
         value = currentRecord.get(name);
      }
      
      return (String) value;
   }
   */
   
   /** 
    * Resolve the variable represented by <code>name</code> to a
    * <code>Double</code> object which represents the contents of the
    * variable. This method is called at runtime, each time a variable is
    * encountered in an expression, which needs to be substituted by a
    * <code>double</code> value.
    * <p>
    * This method first tries to look up <code>name</code> in a list of
    * reserved keyword constants and returns the keyword's constant value if
    * found. Otherwise, it delegates variable resolution responsibility to
    * the current data record.
    *
    * @param   name
    *          Name of the variable to be resolved.
    *
    * @return  A <code>Double</code> object which represents the contents of
    *          the variable <code>name</code>, or <code>null</code> if
    *          <code>name</code> cannot be resolved to a numeric value.
    *          Note that if <code>null</code> is returned, a comparison
    *          which attempts to match this variable for equality will
    *          return <code>false</code>.
    */
   /*
   public Double resolveToDouble(String name)
   {
      Object value = reservedMap.get(name);
      if (value == null)
      {
         value = currentRecord.get(name);
      }
      
      return (Double) value;
   }
   */
   
   /**
    * Resolve the variable represented by <code>name</code> to a
    * <code>Long</code> object which represents the contents of the
    * variable. This method is called at runtime, each time a variable is
    * encountered in an expression, which needs to be substituted by a
    * <code>long</code> value.
    * <p>
    * This method first tries to look up <code>name</code> in a list of
    * reserved keyword constants and returns the keyword's constant value if
    * found. If not found, it delegates variable resolution responsibility
    * to the current data record.
    *
    * @param   name
    *          Name of the variable to be resolved.
    *
    * @return  A <code>Long</code> object which represents the contents of
    *          the variable <code>name</code>, or <code>null</code> if
    *          <code>name</code> cannot be resolved to a numeric value.
    *          Note that if <code>null</code> is returned, a comparison
    *          which attempts to match this variable for equality will
    *          return <code>false</code>.
    */
   /*
   public Long resolveToLong(String name)
   {
      Object value = super.resolveToLong(name);
      
      if (value == null)
      {
         value = reservedMap.get(name);
      }
      
      if (value == null)
      {
         value = currentRecord.get(name);
      }
      
      return (Long) value;
   }
   */
   
   /**
    * Resolve the variable represented by <code>name</code> to a
    * <code>Boolean</code> object which represents the contents of the
    * variable. This method is called at runtime, each time a variable is
    * encountered in an expression, which needs to be substituted by a
    * <code>boolean</code> value.
    * <p>
    * This method first tries to look up <code>name</code> in a list of
    * reserved keyword constants and returns the keyword's constant value if
    * found. If not found, it delegates variable resolution responsibility
    * to the current data record.
    *
    * @param   name
    *          Name of the variable to be resolved.
    *
    * @return  A <code>Boolean</code> object which represents the contents
    *          of the variable <code>name</code>, or <code>null</code> if
    *          <code>name</code> cannot be resolved to a numeric value.
    *          Note that if <code>null</code> is returned, a comparison
    *          which attempts to match this variable for equality will
    *          return <code>false</code>.
    */
   /*
   public Boolean resolveToBoolean(String name)
   {
      Object value = super.resolveToBoolean(name);
      
      if (value == null)
      {
         value = reservedMap.get(name);
      }
      
      if (value == null)
      {
         value = currentRecord.get(name);
      }
      
      return (Boolean) value;
   }
   */
   
   /**
    * Get the number of data records managed by the current instance of the
    * test driver.
    *
    * @return  Count of data records.
    */
   public int getRecordCount()
   {
      return records.size();
   }
   
   /**
    * Validate that the range specified by <code>first</code> and
    * <code>last</code> corresponds with the number of data records loaded
    * into the current instance of the test driver. Also make sure that
    * <code>first</code> is not greater than <code>last</code>.
    *
    * @param   first
    *          One-based index of first record to be used.
    * @param   last
    *          One-based index of last record to be used.
    *
    * @throws  IllegalArgumentException
    *          if specified range of records is invalid.
    */
   private void checkRange(int first, int last)
   throws IllegalArgumentException
   {
      if (first < 1)
      {
         throw new IllegalArgumentException(
            "Starting record index out of bounds: " + first);
      }
      
      if (last > getRecordCount())
      {
         throw new IllegalArgumentException(
            "Ending record index out of bounds: " + last);
      }
      
      if (first > last)
      {
         throw new IllegalArgumentException(
            "First record index cannot be higher than last record index");
      }
   }
   
   /**
    * Compile a list of the numeric (<code>Double</code>) results of computing
    * the specified expression <code>expr</code> against the specified range
    * of data records.
    *
    * @param   expr
    *          Text representation of the expression in infix notation
    *          (e.g., <code>A + ((B + 5) * 12)</code>). The expression
    *          must be an arithmetic expression (i.e., written so as to
    *          compute a numeric result).
    * @param   first
    *          One-based index of the first data record to apply to the
    *          expression.
    * @param   last
    *          One-based index of the last data record to apply to the
    *          expression.
    *
    * @return  List of <code>Double</code> results in the order in which the
    *          expressions were computed against the data records.
    *
    * @throws  IllegalArgumentException
    *          if specified range of records is invalid.
    * @throws  UnresolvedSymbolException
    *          if any symbol cannot be resolved at runtime, when the result
    *          of the expression is being computed.
    * @throws  ExpressionException
    *          if there is a syntax or other error while compiling the
    *          expression.
    */
   /*
   public List computeResults(String expr, int first, int last)
   throws IllegalArgumentException,
          UnresolvedSymbolException,
          ExpressionException
   {
      checkRange(first, last);
      
      List results = new ArrayList(last - first + 1);
      for (int i = first; i <= last; i++)
      {
         results.add(new Double(compute(expr, i)));
      }
      
      return results;
   }
   */
   
   /**
    * Compute the specified arithmetic expression against the data stored in
    * the employee record at <code>recordIndex</code>. This instance of
    * <code>TestDriver</code> is passed to the expression compiler as the
    * {@link SymbolResolver} which will be used to resolve constant
    * references in the expression at compile time, and variable references
    * in the expression at runtime.
    *
    * @param   expr
    *          Text representation of the expression in infix notation
    *          (e.g., <code>A + ((B + 5) * 12)</code>). The expression
    *          must be an arithmetic expression (i.e., written so as to
    *          compute a numeric result).
    * @param   recordIndex
    *          One-based index of the data record to apply to the expression.
    *
    * @return  The numeric result of computing the arithmetic expression.
    *
    * @throws  IllegalArgumentException
    *          if <code>recordIndex</code> is out of bounds.
    * @throws  UnresolvedSymbolException
    *          if any symbol cannot be resolved at runtime, when the result
    *          of the expression is being computed.
    * @throws  ExpressionException
    *          if there is a syntax or other error while compiling the
    *          expression.
    */
   /*
   public double compute(String expr, int recordIndex)
   throws IllegalArgumentException,
          UnresolvedSymbolException,
          ExpressionException
   {
      checkRange(recordIndex, recordIndex);
      
      ArithmeticExpression ae = compiler.compileArithmetic(expr, this, debug);
      currentRecord = (Map) records.get(recordIndex - 1);
      
      return ae.compute();
   }
   */
   
   /**
    * Compile a list of logical (<code>Boolean</code>) results of evaluating
    * the specified expression <code>expr</code> against the specified range
    * of data records.
    *
    * @param   expr
    *          Text representation of the expression in infix notation
    *          (e.g., <code>A + ((B + 5) * 12) != C</code>). The expression
    *          must be a logical expression (i.e., written so as to evaluate
    *          to <code>true</code> or <code>false</code>).
    * @param   first
    *          One-based index of the first data record to apply to the
    *          expression.
    * @param   last
    *          One-based index of the last data record to apply to the
    *          expression.
    *
    * @return  List of <code>Boolean</code> results in the order in which the
    *          expressions were computed against the data records.
    *
    * @throws  IllegalArgumentException
    *          if specified range of records is invalid.
    * @throws  ExpressionException
    *          if there is a syntax or other error while compiling the
    *          expression.
    */
   /*
   public List evaluateResults(String expr, int first, int last)
   throws IllegalArgumentException,
          ExpressionException
   {
      checkRange(first, last);
      
      List results = new ArrayList(last - first + 1);
      for (int i = first; i <= last; i++)
      {
         results.add(evaluate(expr, i) ? Boolean.TRUE : Boolean.FALSE);
      }
      
      return results;
   }
   */
   
   /**
    * Compute the specified logical expression against the data stored in
    * the employee record at <code>recordIndex</code>. This instance of
    * <code>TestDriver</code> is passed to the expression compiler as the
    * {@link SymbolResolver} which will be used to resolve constant
    * references in the expression at compile time, and variable references
    * in the expression at runtime.
    *
    * @param   expr
    *          Text representation of the expression in infix notation
    *          (e.g., <code>A + ((B + 5) * 12) != C</code>). The expression
    *          must be a logical expression (i.e., written so as to evaluate
    *          to <code>true</code> or <code>false</code>).
    * @param   recordIndex
    *          One-based index of the data record to apply to the expression.
    *
    * @return  The result of evaluating <code>expr</code>: <code>true</code>
    *          or <code>false</code>.
    *
    * @throws  IllegalArgumentException
    *          if <code>recordIndex</code> is out of bounds.
    * @throws  ExpressionException
    *          if there is a syntax or other error while compiling the
    *          expression.
    */
   /*
   public boolean evaluate(String expr, int recordIndex)
   throws IllegalArgumentException,
          ExpressionException
   {
      checkRange(recordIndex, recordIndex);
      
      LogicalExpression le = compiler.compileLogical(expr, this, debug);
      currentRecord = (Map) records.get(recordIndex - 1);
      
      return le.evaluate();
   }
   */
   
   /**
    * Abort the program and display help text.
    *
    * @param   rc
    *          Process return code to be passed to the operating system.
    */
   private static void endWithHelp(int rc)
   {
      System.out.println(
         "Usage: TestDriver <properties> <type> <expression> [<first> [<last>]]");
      System.out.println();
      System.out.println(
         "where: <properties> is the name of the properties file;");
      System.out.println(
         "       <type>       is 'A' for an arithmetic expression,");
      System.out.println(
         "                       'L' for a logical expression;");
      System.out.println(
         "       <expression> is an expression in double quotes");
      System.out.println(
         "       <first>      is the 1-based index of the first record to access");
      System.out.println(
         "       <last>       is the 1-based index of the last record to access");
      
      System.exit(rc);
   }
   
   /**
    * Command line entry point into program. Expects the following arguments:
    * <ol>
    * <li><code>properties</code> (required) - configuration properties
    *     file</li>
    * <li><code>type</code> (required)
    * <ul>
    * <li><code>A</code> - specifies arithmetic (numeric) expression</li>
    * <li><code>L</code> - specifies logical (boolean) expression</li>
    * </ul>
    * </li>
    * <li><code>expression</code> (required) - an arithmetic or logical
    *     expression in double quotes (type must match <code>type</code>
    *     argument and variable and constant references must be those which
    *     have been configured in <code>properties</code>)</li>
    * <li><code>first</code> (optional) - 1-based index of the first data
    *     record to apply to <code>expression</code></li>
    * <li><code>last</code> (optional) - 1-based index of the last data
    *     record to apply to <code>expression</code>; can only be specified
    *     if preceded by <code>first</code></li>
    * </ol>
    */
   public static void main(String args[])
   {
      System.out.println("Under construction...");
      /*
      try
      {
         // Check for proper number of arguments.
         if (args.length < 3 || args.length > 5)
         {
            endWithHelp(-1);
         }
         
         String file = args[0];
         String type = args[1];
         String expr = args[2];
         
         // Initialize the test driver with data from the properties file.
         Properties props = Utils.loadProperties(file);
         TestDriver driver = new TestDriver();
         driver.initialize(props);
         
         // Determine the first record to use.
         int first = 1;
         if (args.length > 3)
         {
            first = Integer.parseInt(args[3]);
         }
         
         // Determine the last record to use.
         int last = driver.getRecordCount();
         if (args.length > 4)
         {
            last = Integer.parseInt(args[4]);
         }
         
         // Compile and run the appropriate expression type against the list
         // of records specified.
         List results = null;
         if ("A".equals(type))
         {
            results = driver.computeResults(expr, first, last);
         }
         else if ("L".equals(type))
         {
            results = driver.evaluateResults(expr, first, last);
         }
         else
         {
            throw new IllegalArgumentException(
               "Invalid expression type: " + type);
         }
         
         // Display results.
         System.out.println("Expression");
         System.out.print("   ");
         System.out.println(expr);
         System.out.println("results:");
         Iterator iter = results.iterator();
         for (int i = first; iter.hasNext(); i++)
         {
            System.out.println("[" + i + "] " + iter.next());
         }
      }
      catch (Exception exc)
      {
         exc.printStackTrace();
      }
      */
   }
   
   /**
    * This class illustrates a sample library of user functions available to
    * the {@link TestDriver} user. To add new functionality, the only change
    * required is to define a new method in this class.
    * <p>
    * The function name corresponds directly with the method name.
    */
   private class SymbolLibrary
   {
      /** Backing member for runtime variable "test1"; exposed via accessor */
      private long test1 = 85L;
      
      /** Backing member for runtime variable "test2"; exposed directly */
      public long test2 = 1000L;
      
      /**
       * Return the boolean value of the argument passed in. This is a
       * regression test for ExpressionCompiler fix H005.
       *
       * @param   test
       *          <code>true</code> or <code>false</code>.
       *
       * @return  The value of <code>test</code>.
       */
      public boolean boolTest(boolean test)
      {
         return test;
      }
      
      /**
       * Return the value of the runtime variable "test1".
       *
       * @return  Value currently held in the backing member variable.
       */
      public long getTest1()
      {
         return test1;
      }
      
      /**
       * Return <code>true</code> as the value of variable <code>test3</code>.
       *
       * @return  <code>true</code>.
       */
      public boolean isTest3()
      {
         return true;
      }
      
      /**
       * Return the current time as milliseconds since midnight 1970/01/01.
       *
       * @return  Current time in milliseconds since the epoch.
       */
      public Long now()
      {
         return new Long((new Date()).getTime());
      }
      
      /**
       * Return a pseudo-random number <code>x</code> in the range
       * <code>0 &le; x &lt; mult</code> if <code>mult</code> is positive, or
       * <code>0 &ge; x &gt; mult</code> if <code>mult</code> is negative.
       *
       * @param   mult
       *          Multiplier to apply to a pseudo-random number <code>x</code>
       *          in the range <code>0 &le; x &lt; 1.0</code> to achieve the
       *          desired range.
       *
       * @return  A pseudo-random number as described above.
       */
      public Double random(double mult)
      {
         return new Double(Math.random() * mult);
      }
      
      /**
       * Lowercase the specified text.
       *
       * @param   text
       *          Text to be lowercased.
       *
       * @return  New string which represents the text of <code>text</code>
       *          in lowercase.
       */
      public String lower(String text)
      {
         return text.toLowerCase();
      }
      
      /**
       * Uppercase the specified text.
       *
       * @param   text
       *          Text to be uppercased.
       *
       * @return  New string which represents the text of <code>text</code>
       *          in uppercase.
       */
      public String upper(String text)
      {
         return text.toUpperCase();
      }
      
      /**
       * Determine if the employee record is associated with a prehistoric
       * person. For our purposes, prehistoric is defined as having a
       * birthday before the common era.
       *
       * @return  (Wrapped) <code>true</code> if person is prehistoric, else
       *          <code>false</code>.
       */
      /*
      public Boolean prehistoric()
      {
         Calendar cal = Calendar.getInstance();
         cal.set(Calendar.ERA, GregorianCalendar.BC);
         cal.set(Calendar.YEAR, 0);
         cal.set(Calendar.DAY_OF_YEAR, 1);
         Date bce = cal.getTime();
         Date dob = new Date(resolveToLong("dob").longValue());
         
         return (dob.before(bce) ? Boolean.TRUE : Boolean.FALSE);
      }
      */
   }
}