/*
** Module   : SessionPresenter.java
** Abstract : SessionPresenter is a controller that implements business methods to manage sessions.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------Description----------------------
** 001 SBI 20170601 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/


package com.goldencode.p2j.admin.client.application.home.console.sessions;

import java.util.logging.Level;

import javax.inject.Named;

import com.goldencode.p2j.admin.SessionInfo;
import com.goldencode.p2j.admin.client.*;
import com.goldencode.p2j.admin.client.application.home.*;
import com.goldencode.p2j.admin.shared.AdminServiceAsync;
import com.google.gwt.regexp.shared.RegExp;
import com.google.gwt.user.client.rpc.AsyncCallback;
import com.google.gwt.user.client.ui.HasEnabled;
import com.google.inject.*;
import com.google.web.bindery.event.shared.*;
import com.gwtplatform.mvp.client.*;
import com.gwtplatform.mvp.client.annotations.*;
import com.gwtplatform.mvp.client.presenter.slots.*;
import com.gwtplatform.mvp.client.proxy.*;
import com.gwtplatform.mvp.shared.proxy.PlaceRequest;


/**
 * SessionPresenter is a controller that implements business methods to manage sessions.
 */
public class SessionPresenter extends BasePresenter<SessionPresenter.MyView, SessionPresenter.MyProxy>
implements SessionUIHandler
{
   /** The attachment point for modal dialogs on this view*/
   public static final NestedSlot MODAL_CONTENT = new NestedSlot();
   
   /** Administration server interface */
   private final AdminServiceAsync adminService;
   
   /**
    * This view interface that must be implemented by the concrete view according to GWTP
    * framework. Defines the exposed business methods.
    */
   public interface MyView
   extends View, HasUiHandlers<SessionUIHandler>
   {
      /**
       * Sets the session table model.
       * 
       * @param    sessions
       *           The array of sessions
       * @param    filter
       *           The filter criteria
       */
      void setSessions(SessionInfo[] sessions, RegExp filter);
      
      /**
       * Gets a reference to the Filter View.
       * 
       * @return   The Filter View
       */
      FilterSessionsView getFilterView();
      
      /**
       * Gets a reference to the Terminate Session Alert.
       * 
       * @return   The Terminate Session Alert
       */
      TerminateSessionAlert getTerminateSessionAlert();
      
      /**
       * Tests if the tested session is selected.
       * 
       * @param    session
       *           The tested session
       * 
       * @return   True iff the tested session is selected, otherwise false.
       */
      boolean isSessionSelected(SessionInfo session);
   }

   /**
    * Defines the proxy place for GWTP framework.
    */
   @ProxyStandard
   @NameToken(NameTokens.SESSIONS)
   public interface MyProxy
   extends ProxyPlace<SessionPresenter>
   {}

   /**
    * Defines Session presenter.
    * 
    * @param    eventBus
    *           The event bus
    * @param    view
    *           The associated view
    * @param    proxy
    *           The proxy place
    * @param    placeManager
    *           The place manager
    * @param    adminService
    *           The administration service
    */
   @Inject
   SessionPresenter(EventBus eventBus,
                    MyView view,
                    MyProxy proxy,
                    PlaceManager placeManager,
                    AdminServiceAsync adminService)
   {
      super(eventBus, view, proxy, HomePresenter.SLOT_CONTENT, placeManager);
      this.adminService = adminService;
      getView().setUiHandlers(this);
   }

   /**
    * Retrieves the current session list and updates the session table model according to
    * the given filter criteria.
    * 
    * @param    filter
    *           The filter criteria
    */
   private void filterSessionList(RegExp filter)
   {
      this.adminService.getSessionList(new AsyncCallback<SessionInfo[]>()
      {
         
         @Override
         public void onSuccess(SessionInfo[] result)
         {
            getView().setSessions(result, filter);
         }
         
         @Override
         public void onFailure(Throwable caught)
         {
            //TODO notify an admin user
            logger.log(Level.INFO, "Can't get network sessions.", caught);
         }
      });
   }
   
   /**
    * Executes an asynchronous action to retrieve the current session list from the server and
    * to update the current sessions table view.
    */
   @Override
   public void updateSessionList()
   {
      filterSessionList(filterExp);
   }

   /**
    * Loads the data before the associated view will be displayed.
    * 
    * @param    request
    *           The current request
    */
   @Override
   public void prepareFromRequest(PlaceRequest request)
   {
      super.prepareFromRequest(request);
      setFilter(request.getParameter(FILTER, ""));
      updateSessionList();
   }
   
   /**
    * Sets the current search string used as a filter to display the target results.
    * 
    * @param    value
    *           The new search string used as a filter.
    */
   @Override
   public void setDataSelectionFilter(String value)
   {
      setFilter(value);
      
      filterSessionList(filterExp);
   }
   
   /**
    * Executes an asynchronous action to terminate the given session and on its completion
    * to update the current sessions table view.
    */
   @Override
   public void terminateSession(final SessionInfo session)
   {
      if (!getView().isSessionSelected(session))
      {
         updateSessionList();
         return;
      }
      this.adminService.terminateSession(session.id, new AsyncCallback<Void>()
      {
         
         @Override
         public void onSuccess(Void result)
         {
            updateSessionList();
         }
         
         @Override
         public void onFailure(Throwable caught)
         {
            updateSessionList();
            //TODO notify an admin user
            logger.log(Level.INFO, "Can't terminate network session " + session.id, caught);
         }
      });
   }
}