/*
** Module   : ComObject.java
** Abstract : The base class for all COM objects, with minimal management.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 OM  20170530 First commit.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.comauto;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.util.*;

/**
 * Base class for COM objects that will be stored using {@link comhandle}s. Does minimal
 * management of the COM Object.
 */
public abstract class ComObject
implements WrappedResource
{
   /**
    * The resource id. Used by {@code WrappedResource} management. 
    */
   private Long resourceId = null;
   
   /**
    * The parent {@code ControlFrameWidget} widget. This container is used during the dispatch of 
    * the events generated by this COM Object.
    */
   private ControlFrameWidget parentControlFrame = null;
   
   /**
    * Build a new COM Object.
    */
   public ComObject()
   {
   }
   
   /**
    * Initialization of the object. Allows calling {@link handle#resourceId} after the object was
    * fully created and avoid calling virtual methods (namely {@link #valid()}) with uninitialized
    * fields.
    */
   public void init()
   {
      /*resourceId =*/ handle.resourceId(this);
   }
   
   /**
    * Reports if this object is unknown.  
    *
    * @return   <code>true</code> if object is unknown.
    */
   @Override
   public boolean unknown()
   {
      return false;
   }
   
   /**
    * Get this resource's ID, if is already set.
    *
    * @return   The resource's ID or <code>null</code> if not set.
    */
   @Override
   public Long id()
   {
      return resourceId;
   }
   
   /**
    * Set this resource's ID.
    *
    * @param    id
    *           The resource's ID.
    */
   @Override
   public void id(long id)
   {
      resourceId = id;
   }
   
   /**
    * Obtain the name of this COM object.
    * 
    * @return  the name of this COM object.
    */
   public abstract String getName();
   
   /**
    * Perform actual delete of all resources. At the time of this call, it is assumed the resource
    * is valid for deletion (the handle and the resource are both valid).
    */
   public abstract void release();
   
   /**
    * Obtain the {@code ControlFrameWidget} container for this COM, if any.
    * 
    * @return  container for this COM.
    */
   public ControlFrameWidget getParentControlFrame()
   {
      return parentControlFrame;
   }
   
   /**
    * Sets the {@code ControlFrameWidget} container for this COM.
    *
    * @param   pcf
    *          The container for this COM.
    */
   public void setParentControlFrame(ControlFrameWidget pcf)
   {
      this.parentControlFrame = pcf;
   }
}
