/*
** Module   : BaseHelper.java
** Abstract : provides common compatibility helper features
**
** Copyright (c) 2007-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20070822 NEW   @34962 Provides common compatibility helper
**                               features.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.e4gl;     

import java.util.*;

/**
 * Provides common compatibility helper features. This class is a complete,
 * working implementation of the compatibility interface. However, it does
 * not define full compatibility with any of the known implementations.
 */
public class BaseHelper
implements CompatibilityHelper
{
   /** Configuration to be honored. */
   protected Options opts = null;
   
   /** Replacement mappings for escaping unsafe HTML characters. */
   protected Map<String, String> unsafe = new LinkedHashMap<String, String>();
   
   /** Replacement mappings for removing HTML entities. */
   protected Map<String, String> entities = new HashMap<String, String>();
   
   /**
    * Provides the known configuration to the helper and allows it to
    * prepare for use. This must be called before any other method can be
    * used.
    *
    * @param    opts
    *           The configuration.
    */
   public void initialize(Options opts)
   {
      this.opts = opts;
      
      // order is critical here
      unsafe.put("~", "~~");
      unsafe.put(";", "~;");
      unsafe.put("\\{", "~{");
      unsafe.put("}", "~}");
      unsafe.put("'", "~'");
      
      if ("unix".equalsIgnoreCase(opts.getOpsys()))
      {
         // each \ character must be duplicated 4X because it is an escape
         // character for both Java and regex
         unsafe.put("\\\\", "\\\\\\\\");
      }
      
      entities.put("&amp;", "&");
      entities.put("&#38;", "&");
      entities.put("&quot;", "\"");
      entities.put("&#34;", "\"");
      entities.put("&lt;", "<");
      entities.put("&#60;", "<");
      entities.put("&gt;", ">");
      entities.put("&#62;", ">");
      entities.put("&apos;", "\'");
      entities.put("&#39;", "\'");
   }
   
   /**
    * Return the filename suffix (extension) to be used.
    *
    * @return   The filename extension to use based on the known options.
    */
   public String suffix()
   {
      return opts.isInclude() ? ".i" : ".w";
   }
   
   /**
    * Return the text for the header of the generated file. This is output
    * before any parsed content.
    *
    * @return   Always the empty string.
    */
   public String header()
   {
      return "";
   }
   
   /**
    * Return the text used to start a line of raw HTML output.
    *
    * @param    content
    *           <code>true</code> if there is HTML content that is about to
    *           be emitted on the line, <code>false</code> to handle the
    *           no content case (just start the line, without any
    *           content prefix text).
    *
    * @return   The text to use based on the known options.
    */
   public String openHTML(boolean content)
   {
      return String.format("{&OUT} %s", content ? "'" : "");
   }
   
   /**
    * Return the text used to end a line of raw HTML output.
    *
    * @param    content
    *           <code>true</code> if there is content that needs to be
    *           finished on the line, <code>false</code> to handle the
    *           no content case (just terminate the line, without any
    *           content termination text).
    *
    * @return   The text to use based on the known options.
    */
   public String closeHTML(boolean content)
   {
      return content ? "~n'." : ".";
   }
   
   /**
    * Convert any unsafe characters (in the given HTML text) into equivalents
    * that are safe to embed in a 4GL string literal and return that converted
    * text.
    *
    * @return   The replacement text.
    */
   public String escapeHTML(String html)
   {
      return replace(unsafe, html);
   }
   
   /**
    * Convert any HTML entities (in the given E4GL text) into the equivalent
    * characters.
    *
    * @return   The replacement text with all conversion applied.
    */
   public String entityConvert(String e4gl)
   {
      return replace(entities, e4gl);
   }
   
   /**
    * Return the text used to start the output of an embedded 4GL expression.
    *
    * @return   The text to use based on the known options.
    */
   public String openExprEscape()
   {
      return "'";
   }
   
   /**
    * Return the text used to end the output of an embedded 4GL expression.
    *
    * @return   The text to use based on the known options.
    */
   public String closeExprEscape()
   {
      return "'";
   }
   
   /**
    * Return the text used to start a comment that removes the start or end
    * tag for a statement or expression escape.
    *
    * @param    expr
    *           <code>true</code> if this is an expression escape,
    *           <code>false</code> for a statement escape.
    *
    * @return   The text to use based on the known options. <code>null</code>
    *           if the escape should be removed instead of commented.
    */
   public String openComment(boolean expr)
   {
      return "/* ";
   }
   
   /**
    * Return the text used to end a comment that removes the start or end tag
    * for a statement or expression escape.
    *
    * @param    expr
    *           <code>true</code> if this is an expression escape,
    *           <code>false</code> for a statement escape.
    *
    * @return   The text to use based on the known options. <code>null</code>
    *           if the escape should be removed instead of commented.
    */
   public String closeComment(boolean expr)
   {
      return " */";
   }
   
   /**
    * Return the text used to start a comment that removes the META or
    * WSMETA element with an HTTP-EQUIV attribute that is set to
    * CONTENT-TYPE.
    *
    * @return   The text to use based on the known options. Return the empty
    *           string if the element should be left behind instead of
    *           commented.
    */
   public String openContentTypeComment()
   {
      return opts.isKeepContentType() ? "" : "<!-- E4GL Disabled: ";
   }
   
   /**
    * Return the text used to end a comment that removes the META or
    * WSMETA element with an HTTP-EQUIV attribute that is set to
    * CONTENT-TYPE.
    *
    * @return   The text to use based on the known options. Return the empty
    *           string if the element should be left behind instead of
    *           commented.
    */
   public String closeContentTypeComment()
   {
      return opts.isKeepContentType() ? "" : " -->";
   }
   
   /**
    * Return the text for the footer of the generated file. This is output
    * after all parsed content.
    *
    * @return   Always the empty string.
    */
   public String footer()
   {
      return "";
   }
   
   /**
    * Use the mapped keys to convert any matching characters (in the given
    * text) into the equivalent mapped value. Then return that converted
    * text.
    *
    * @return   The replacement text.
    */
   protected String replace(Map<String, String> replace, String txt)
   {
      Iterator<String> keylist = replace.keySet().iterator();
      
      while (keylist.hasNext())
      {
         String key = keylist.next();
         txt = txt.replaceAll(key, replace.get(key)); 
      }
      
      return txt;
   }
}