/*
** Module   : WebSpeedHelper.java
** Abstract : creates WebSpeed-specific output text
**
** Copyright (c) 2007-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20070822 NEW   @34968 Creates WebSpeed-specific output text.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.e4gl;

import java.util.*;

/**
 * Assists in creating WebSpeed-specific output text. Note that this class
 * does not emit any WebSpeed include files that are normally inserted in
 * the header of every generated file. This eliminates the use of PSC written
 * technology in the resulting file, at the expense of strict compatibility.
 */
public class WebSpeedHelper
extends BaseHelper
{
   /** Standard WebSpeed footer template. */
   private static final String FOOTER_SPEC =
      "/************************* END OF HTML *************************/"   +
      NEWLINE                                                               +
      "/*"                                                                  +
      NEWLINE                                                               +
      "** File: %s"                                                         +
      NEWLINE                                                               +
      "** Generated on: %TF %2$TT"                                          +
      NEWLINE                                                               +
      "** By: WebSpeed Embedded SpeedScript Preprocessor"                   +
      NEWLINE                                                               +
      "** Version: %s"                                                      +
      NEWLINE                                                               +
      "** Source file: %s"                                                  +
      NEWLINE                                                               +
      "** Options: %s"                                                      +
      NEWLINE                                                               +
      "**"                                                                  +
      NEWLINE                                                               +
      "** WARNING: DO NOT EDIT THIS FILE.  Make changes to the original"    +
      NEWLINE                                                               +
      "** HTML file and regenerate this file from it."                      +
      NEWLINE                                                               +
      "**"                                                                  +
      NEWLINE                                                               +
      "*/"                                                                  +
      NEWLINE                                                               +
      "%s"                                                                  +
      NEWLINE                                                               +
      "/* end */"                                                           +
      NEWLINE;
      
   /** Standard WebSpeed local-e4gl-options procedure template. */
   private static final String LOCAL_OPTS_SPEC =
      "/********************* Internal Definitions ********************/"   +
      NEWLINE                                                               +
      NEWLINE                                                               +
      "/* This procedure returns the generation options at runtime."        +
      NEWLINE                                                               +
      "   It is invoked by src/web/method/e4gl.i included at the start"     +
      NEWLINE                                                               +
      "   of this file. */"                                                 +
      NEWLINE                                                               +
      "PROCEDURE local-e4gl-options :"                                      +
      NEWLINE                                                               +
      "  DEFINE OUTPUT PARAMETER p_version AS DECIMAL NO-UNDO"              +
      NEWLINE                                                               +
      "    INITIAL %s."                                                     +
      NEWLINE                                                               +
      "  DEFINE OUTPUT PARAMETER p_options AS CHARACTER NO-UNDO"            +
      NEWLINE                                                               +
      "    INITIAL \"%s\"."                                                 +
      NEWLINE                                                               +
      "  DEFINE OUTPUT PARAMETER p_content-type AS CHARACTER NO-UNDO"       +
      NEWLINE                                                               +
      "    INITIAL \"%s\"."                                                 +
      NEWLINE                                                               +
      "END PROCEDURE."                                                      +
      NEWLINE;
      
   /** WebSpeed E4GL version string. */
   private static final String VERSION = "2.0";
      
   /**
    * Provides the known configuration to the helper and allows it to
    * prepare for use. This must be called before any other method can be
    * used.
    *
    * @param    opts
    *           The configuration.
    */
   public void initialize(Options opts)
   {
      super.initialize(opts);
      
      unsafe.put("\'", "\'\'");
   }
   
   /**
    * Return the text for the header of the generated file. This is output
    * before any parsed content.
    *
    * @return   The text to use based on the known options.
    */
   public String header()
   {
      return String.format("/*E4GL-%s*/ ", opts.isInclude() ? "I" : "W");
   }
   
   /**
    * Return the text used to start a comment that removes the start or end
    * tag for a statement or expression escape.
    *
    * @param    expr
    *           <code>true</code> if this is an expression escape,
    *           <code>false</code> for a statement escape.
    *
    * @return   The text to use based on the known options. <code>null</code>
    *           if the escape should be removed instead of commented.
    */
   public String openComment(boolean expr)
   {
      return " /*Tag=";
   }
   
   /**
    * Return the text used to end a comment that removes the start or end
    * tag for a statement or expression escape.
    *
    * @param    expr
    *           <code>true</code> if this is an expression escape,
    *           <code>false</code> for a statement escape.
    *
    * @return   The text to use based on the known options. <code>null</code>
    *           if the escape should be removed instead of commented.
    */
   public String closeComment(boolean expr)
   {
      return "*/ ";
   }
   
   /**
    * Return the text for the footer of the generated file. This is output
    * after all parsed content.
    *
    * @return   The text to use based on the known options.
    */
   public String footer()
   {
      String procedure = "";
      
      if (!opts.isInclude())
      {
         String contentType = null;
         
         // content-type will often have a semicolon which may be munged
         // by the 4GL preprocessor if not escaped
         contentType = opts.getContentType(); 
         contentType = contentType.replaceAll(";", "~;");
         
         // create the text of the local-e4gl-options procedure   
         procedure = String.format(LOCAL_OPTS_SPEC,
                                   VERSION,
                                   opts.getWSOptions(),
                                   contentType);
      }
                                     
      return String.format(FOOTER_SPEC,
                           opts.getOutput(),
                           new Date(),
                           VERSION,
                           opts.getInput(),
                           opts.getWSOptions(),
                           procedure);
   }
}