/*
** Module   : FWDTimer.java
** Abstract : Interface for Timer FWD extension to 4GL.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 OM  20170530 First commit.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.extension;

import com.goldencode.p2j.util.*;

/**
 * Interface of the FWD TIMER Extension. An {@code FWDTimer} object can be scheduled to run
 * repetitive tasks at fixed intervals. This is not fully compatible with ABL standard language,
 * it is an extension to 4GL specific to FWD.
 */
@FWDResource
@LegacyResource(resource = FWDResource.FWD_TIMER)
public interface FWDTimer
{
   /**
    * Returns the state of the ENABLED attribute
    *
    * @return    Current state of the ENABLED attribute.
    */
   @LegacyAttribute(name = "ENABLED")
   public logical isEnabled();
   
   /**
    * Setter for ENABLED attribute.
    *
    * @param    value
    *           The new value of ENABLED attribute.
    */
   @LegacyAttribute(name = "ENABLED", setter = true)
   public void setEnabled(boolean value);
   
   /**
    * Setter for ENABLED attribute.
    *
    * @param    value
    *           The new value of ENABLED attribute.
    */
   @LegacyAttribute(name = "ENABLED", setter = true)
   public void setEnabled(logical value);
   
   /**
    * Obtain the current {@code interval} between two consecutive events.
    *
    * @return  the current value of {@code interval} property.
    */
   @LegacyAttribute(name = "Interval")
   public integer getInterval();
   
   /**
    * Sets the current {@code interval} between two consecutive events. If the timer is already
    * {@code enabled}, it will be automatically armed and will fire after {@code interval} millis 
    * elapses. Setting it to 0, will stop the timer. Setting to a negative value will lead to
    * an error message to be displayed.
    *
    * @param   interval
    *          The new interval between two timer events, in milliseconds.
    */
   @LegacyAttribute(name = "Interval", setter = true)
   public void setInterval(NumberType interval);
   
   /**
    * Sets the current {@code interval} between two consecutive events. If the timer is already
    * {@code enabled}, it will be automatically armed and will fire after {@code interval} millis 
    * elapses. Setting it to 0, will stop the timer. Setting to a negative value will lead to
    * an error message to be displayed.
    *
    * @param   interval
    *          The new interval between two timer events, in milliseconds.
    */
   @LegacyAttribute(name = "Interval", setter = true)
   public void setInterval(long interval);
   
   /**
    * Gets the legacy name of the callback procedure to be called at fixed interval when running
    * in FWD extension mode (not as a COM).
    *
    * @return  the legacy name of the callback procedure.
    */
   @LegacyAttribute(name = "Callback")
   public character getCallback();
   
   /**
    * Sets the legacy name of the callback procedure to be called at fixed interval when running
    * in FWD extension mode (ignored when in COM mode). If set to {@code unknown} the timer will
    * continue 'ticking' but since there is no known peer to call, nothing will happen. This
    * property can be set while the timer is active, without resetting it.
    *
    * @param   callback
    *          the new legacy name of the callback procedure.
    */
   @LegacyAttribute(name = "Callback", setter = true)
   public void setCallback(character callback);
   
   /**
    * Sets the legacy name of the callback procedure to be called at fixed interval when running
    * in FWD extension mode (ignored when in COM mode). If set to {@code unknown} the timer will
    * continue 'ticking' but since there is no known peer to call, nothing will happen. This
    * property can be set while the timer is active, without resetting it.
    *
    * @param   callback
    *          the new legacy name of the callback procedure.
    */
   @LegacyAttribute(name = "Callback", setter = true)
   public void setCallback(String callback);
   
   /**
    * Starts the timer. It has the exact effect like setting the {@code enabled} property to 
    * {@code true}. Note that timer will not be activated unless the {@code interval} property
    * was set to a strictly positive value.
    */
   @LegacyMethod(name = "Start")
   public void start();
   
   /**
    * Force the timer to cease firing events. It has the exact effect like setting the 
    * {@code enabled} property to {@code false}. 
    * <p>
    * The name of the method was selected intentionally for the FWD extension since {@code STOP}
    * would create some syntactic incompatibility with standard ABL attributes.
    */
   @LegacyMethod(name = "Cease")
   public void cease();
   
   /**
    * Undocumented attribute found by chance. Might as well be either one of:
    * <br> - an attribute of Control Frame COM Object (from {@link ControlFrameComObject});
    * <br> - an attribute of any COM-HANDLE (from {@link comhandle});
    * <br> - id "0" some kind of unknown value for {@code comhandle}-s.
    *
    * @return  Always some kind of "0".
    */
   @LegacyAttribute(name = "COM-HANDLE")
   default public character comHandle()
   {
      return new character("0");
   }
}
