/*
** Module   : CommonDriver.java
** Abstract : common services for a command line driver
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 GES 20060815   @28611 First version which centralizes common services for client and
**                           server drivers.
** 002 GES 20070111   @31774 Match interface changes. Simplified the usage as a subclass.  Now
**                           the subclass has all knowledge of the application that must be
**                           started, there is no generic class load and init as before. This
**                           simplifies the downstream processing in the net package.
** 003 GES 20070112   @31812 Removed initial debug level (it wasn't being used downstream) and
**                           added application args.
** 004 GES 20080118   @36863 Output any abend info to STDERR not STDOUT. For servers, STDERR is
**                           the main log file and for clients STDOUT is interactive and can't
**                           be the destination. If STDERR on the client is redirected to file,
**                           then a record of any abend will be available.
** 005 GES 20090722   @43330 Flush STDERR just before exit to ensure that the output is written
**                           to disk.
** 006 GES 20090723   @43357 Avoid logging STOP conditions.
** 007 GES 20110928          Moved isCauseStop into a common class so it can be easily shared.
** 008 GES 20141228          Rework to eliminate application arguments and to extend the client
**                           parameters to more use cases.
** 009 EVL 20160225          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.main;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.util.*;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;

/**
 * Common services used by command line drivers. 
 */
abstract class CommonDriver
{
   /**
    * The main text of the syntax display, with each element being displayed
    * on a separate line.
    */
   protected String[] helptxt = null;
   
   /** Default configuration filename. */
   protected String defCfg = null;
   
   /** 
    * The maximum number of configuration files (and corresponding passwords)
    * to honor.
    */
   protected int maxCfg = 1;
   
   /** Tracks the number of times a given parameter was specified. */
   private Map instances = new HashMap();
   
   /** Configuration file(s) to process. */
   private String[] cfg = new String[] { null, null };
   
   /** Passwords used to decrypt corresponding configuration files. */
   private char[][] pwd = new char[][] { null, null };
   
   /** Number of configuration files actually specified. */
   private int files = 0;
   
   /** Number of passwords actually specified. */
   private int pws = 0;
   
   /** List of overrides. */
   private List overrd = new LinkedList();

   /**
    * Print syntax on <code>STDOUT</code>.
    * <p>
    * <b>This method does NOT return! It calls <code>System.exit</code>
    * with the given exit code.</b>
    *
    * @param    err
    *           An optional message that is displayed first or
    *           <code>null</code>.
    * @param    rc
    *           The exit code to use with <code>System.exit</code>.
    */
   protected void syntax(String err, int rc)
   {
      if (err != null)
      {
         System.out.println(err);
         System.out.println();
      }
      
      if (helptxt != null)
      {
         for (int i = 0; i < helptxt.length; i++)
         {
            System.out.println(helptxt[i]);
         }
      }
      
      System.exit(rc);
   }
   
   /**
    * Add any new values and replace any duplicated values in the
    * configuration from a list of strings.  These strings can have a long 
    * form of "category:group:key=value" or in the short form of "key=value" 
    * (which uses the most recently specified category and group).
    *
    * @param    bc
    *           The configuration object to update.
    * @param    list
    *           The list of override strings.
    */
   protected void processOverrides(BootstrapConfig bc, List list)
   {
      Iterator iter   = list.iterator();
      String   curcat = null;
      String   curgrp = null;
      
      while (iter.hasNext())
      {
         // get and parse the next string
         String over  = (String)iter.next();
         String cat   = curcat;
         String grp   = curgrp;
         int    eq    = over.indexOf('=');
         String front = over.substring(0, eq);    // [[cat:]grp:]key
         String value = over.substring(eq + 1);   // value
         String key   = null;
         int    co    = front.indexOf(':');
         
         if (co != -1)
         {
            grp = front.substring(0, co);
            curgrp = grp;
            key = front.substring(co + 1);
         }
         else
         {
            cat = curcat;
            grp = curgrp;
            key = front;
         }
         
         co = key.indexOf(':');
         
         if (co != -1)
         {
            front = key;
            cat = grp;
            curcat = grp;
            grp = front.substring(0, co);
            curgrp = grp;
            key = front.substring(co + 1);
         }
         
         if (cat == null || cat.length() == 0)
            syntax("Category MUST be specified.", -9);
         
         if (grp == null || grp.length() == 0)
            syntax("Group MUST be specified.", -10);

         bc.setConfigItem(cat, grp, key, value);
      }
   }

   /**
    * Check that the number of parameters of a given type are not specified 
    * more than the maximum number of times and that any required following
    * argument does exist.
    * <p>
    * On failure of any test, this method displays the syntax message and
    * exits the process (it does not return to the caller in this case).
    *
    * @param    args
    *           The command line arguments.
    * @param    current
    *           The current element being processed.
    * @param    max
    *           The maximum number instances of the parameter that are 
    *           allowed in a single command line.
    * @param    follow
    *           <code>true</code> to force testing for the existence of a
    *           single following argument.
    * @param    name
    *           The parameter name to use in error messages.
    */
   protected void testFollowing(String[] args,
                                       int      current,
                                       int      max,
                                       boolean  follow,
                                       String   name)
   {
      String  parm = args[current];
      Integer num  = (Integer) instances.get(parm);
      int     i    = 0;
      
      if (num != null)
      {
         i = num.intValue();
      }
      
      // incrment the number of instances
      i++;
      
      if (i > max)
      {
         syntax("Too many '" + parm + "' options.", -1);
      }
      
      if (args.length <= current + 1)
      {
         syntax(parm + " should be followed by " + name + " argument.", -2);
      }
      
      // save off the number of instances
      instances.put(parm, new Integer(i));
   }
   
   /**
    * Process arguments for any data or options that are specific to the
    * subclass.  This method is called in a loop from {@link #process}.
    *
    * @param    args
    *           The command line arguments.
    * @param    i
    *           The current element being processed.
    *
    * @return   The number of array elements that have been processed by
    *           this method and which MUST be skipped (no processing in
    *           the base class).
    */
   protected int processArguments(String[] args, int i)
   {
      int skip = 0;
      
      // process password
      if (args[i].equals("-p"))
      {
         testFollowing(args, i, maxCfg, true, "password");
         
         pwd[pws] = args[i + 1].toCharArray();
         pws++;
         skip = 2;
      }
      
      // process overrides (they must always have >= 1 '=' characters
      // and 0 through 2 ':' characters to the left of the first '='
      // character
      if (args[i].indexOf('=') != -1)
      {
         overrd.add(args[i]);
         skip = 1;
      }
      
      // process config file
      if (skip == 0)
      {
         if (files == maxCfg)
            syntax("Too many configuration files.", -3);
         
         cfg[files] = args[i];
         files++;
         skip = 1;
      }
      
      return skip;
   }
   
   /**
    * Provide interactive password prompt if requested.
    */
   protected void processPassword()
   {
      // don't need unrelated passwords
      if (files < pws)
         syntax("Unrelated password specified.", -11);

      // check and get the password interactively
      for (int i = 0; i < pws; i++)
      {
         if (pwd[i].length == 1 && pwd[i][0] == '-')
         {
            pwd[i] = null;
         }
         else if (pwd[i].length == 1 && pwd[i][0] == '?')
         {
            try
            {
               pwd[i] = Utils.prompt("Enter password for '" + cfg[i] + "':");
            }
            catch (IOException ex)
            {
               System.exit(-5);
            }
         }
      }
   }
   
   /**
    * Allows a subclass to check if the given configuration matches any
    * required values.  Edits may be made if needed.
    *
    * @param    bc
    *           The configuration to be checked.
    */
   protected void checkConfig(BootstrapConfig bc)
   throws ConfigurationException
   {
   }
   
   /**
    * Runs the primary application using the preprocessed configuration and
    * target object.
    * 
    * @param    bc
    *           The configuration to use.
    */
   abstract void start(BootstrapConfig bc)
   throws Exception;
   
   /**
    * The command line entry point.
    * 
    * @param     args
    *            The array of command-line parameters.
    */
   public void process(String[] args)
   {
      // honor help request
      if (args.length == 1 && args[0].equals("-?"))
      {
         syntax(null, 0);
      }

      // process arguments
      for (int i = 0; i < args.length; i++)
      {
         // allow the subclass to implement its own arguments processing
         // (this superclass' processArguments() call may be called by the
         // subclass as needed)
         int skip = processArguments(args, i);
         
         // if the subclass processed some number of arguments, then skip
         // these and iterate
         if (skip > 1)
         {
            // we will be skipping "the current one" automatically by 
            // iterating so we only need to manually skip one less than
            // the skip value
            i += --skip;
         }
      }

      // error checking and interactive pw prompt if requested
      processPassword();
      
      // try standard configuration file if none is given
      if (cfg[0] == null)
      {
         File std = new File(defCfg);
         if (std.exists())
            cfg[0] = defCfg;
      }

      BootstrapConfig bc = null;
      
      try
      {
         // instantiate bootstrap configuration
         bc = new BootstrapConfig(cfg[0], pwd[0], cfg[1], pwd[1]);
         
         checkConfig(bc);
         processOverrides(bc, overrd);
      }

      catch (ConfigurationException ce)
      {
         ce.printStackTrace(System.err);
         System.err.flush();
         System.exit(-12);
      }

      // run the application
      try
      {
         start(bc);
      }
      
      catch (Throwable thr)
      {
         // STOP conditions don't need to be reported, they are an expected
         // mechanism to terminate
         if (ClientCore.isCauseStop(thr))
         {
            System.exit(0);
         }
         
         thr.printStackTrace(System.err);
         System.err.flush();
         System.exit(-13);
      }
   }
}