/*
** Module   : ProcessClientBuilder.java
** Abstract : build and start a platform process for P2J process.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140206 First version.
** 002 MAG 20140224 Add outputToFile command line parameter.
** 003 MAG 20140722 Implemented the remote launcher.
** 004 GES 20160219 Moved code to the parent class allowing reuse of the native secure connection
**                  setup processing.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.main;

import java.util.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.security.SecurityManager;

/**
 * This class is used to start P2J client processes using the {@code spawn} tool, with no
 * password authentication. The command line parameters for the spawned process are constructed 
 * from default parameters stored in directory and some specific parameters stored inside a 
 * {@link ClientBuilderParameters} structure.  A {@link ProcessBuilderOptions} instance is used
 * to store specific P2J process configuration.
 * <p>
 * The syntax used to start the native process is 
 * {@code spawn 0 <secure-port> <server-hostname> <server-alias> <uuid>}.
 * 
 * @see   ClientBuilder
 *        For details about the native tool used to spawn client processes.
 */
public class ProcessClientBuilder
extends ClientBuilder
{
   /**
    * Constructor.
    * 
    * @param   config
    *          The specific parameters for this client.
    */
   public ProcessClientBuilder(ClientBuilderParameters config)
   {
      super(config);
   }

   /**
    * Mark this P2J process to start in batch mode (add the {@code client:mode:batch=true} 
    * argument to the command).
    */
   protected void addClientOptions()
   {
      command.add("client:mode:batch=true");
   }
   
   /**
    * Get the spawn arguments.  P2J client processes call the {@code spawn} tool using the
    * {@code spawn 0 <secure-port> <server-hostname> <server-alias> <uuid>} syntax.
    * 
    * @return  See above.
    */
   @Override
   protected List<String> getSpawnArguments()
   {
      SessionManager sessMgr = SessionManager.get();
      BootstrapConfig cfg    = sessMgr.config();
      int securePort         = cfg.getInt("net", "server", "secure_port", -1);

      SecurityManager sm = SecurityManager.getInstance();
      
      List<String> cmd = new ArrayList<>();
      
      initNativeSecureConnection(cmd, Integer.toString(securePort), sm.getServerAlias());
      
      // optional parameters should be added at the end of the list.
      // output file name is sent as two parameters -O and <outputToFile>.
      String outputToFile = config.getOutputToFile();
      if (outputToFile != null)
      {
         cmd.add("-O");
         cmd.add(outputToFile);
      }

      return cmd;
   }
}