/*
** Module   : SecureFileHandler.java
** Abstract : FileHandler subclass which overrides security-sensitive methods
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ------------------Description-------------------
** 001 ECF 20090827   @43789 Created initial version. FileHandler subclass
**                           which overrides security-sensitive methods.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.main;

import java.io.*;
import java.util.logging.*;

/**
 * A <code>FileHandler</code> subclass which overrides security-sensitive
 * methods with safe implementations for secure file logging.
 */
public final class SecureFileHandler
extends FileHandler
{
   /**
    * Constructor which uses the superclass' constructor.
    * 
    * @param   pattern
    *          The pattern for naming the output file.
    * @param   limit
    *          The maximum number of bytes to write to any one file.
    * @param   count
    *          The maximum number of files to create.
    * @param   append
    *          <code>true</code> to append to a file, else <code>false</code>.
    * @param   level
    *          The permanent log level of this handler.
    * @param   formatter
    *          The formatter to be used to format log messages.  Since
    *          {@link FileHandler#getFormatter()} cannot be overridden with a
    *          do-nothing implementation here, this formatter must be
    *          immutable to prevent a security risk.
    *          
    * @throws  IOException
    *          if there is any I/O problem opening the file.
    * @throws  SecurityException
    *          if a security manager exists and if the caller does not have
    *          <code>LoggingPermission("control")</code>.
    */
   public SecureFileHandler(String pattern,
                            int limit,
                            int count,
                            boolean append,
                            Level level,
                            Formatter formatter)
   throws IOException,
          SecurityException
   {
      super(pattern, limit, count, append);
      
      super.setLevel(level);
      super.setFormatter(formatter);
   }
   
   /**
    * Override the superclass' implementation to do nothing.
    * 
    * @param   newLevel
    *          The new logging level.
    */
   public void setLevel(Level newLevel)
   throws SecurityException
   {
   }
   
   /**
    * Override the superclass' implementation to do nothing.
    * 
    * @param   newFormatter
    *          The <code>Formatter</code> to use (may not be null).
    */
   public void setFormatter(Formatter newFormatter)
   throws SecurityException
   {
   }
   
   /**
    * Override the superclass' implementation to return <code>null</code>.
    * 
    * @return  <code>null</code>.
    */
   public Filter getFilter()
   {
      return null;
   }
   
   /**
    * Override the superclass' implementation to do nothing.
    * 
    * @param   newFilter
    *          Not used.
    */
   public void setFilter(Filter newFilter)
   throws SecurityException
   {
   }
   
   /**
    * Override the superclass' implementation to return <code>null</code>.
    * 
    * @return  <code>null</code>.
    */
   public ErrorManager getErrorManager()
   {
      return null;
   }
   
   /**
    * Override the superclass' implementation to do nothing.
    * 
    * @param   em
    *          Not used.
    */
   public void setErrorManager(ErrorManager em)
   {
   }
   
   /**
    * Override the superclass' implementation to do nothing.
    * 
    * @param   encoding
    *          Not used.
    */
   public void setEncoding(String encoding) 
   throws SecurityException,
          UnsupportedEncodingException
   {
   }
   
   /**
    * Override the superclass' implementation to do nothing.
    * 
    * @see  #doClose()
    */
   public void close()
   {
   }
   
   /**
    * Flush and close all files by invoking the superclass'
    * <code>close()</code> method.  This method is package private to ensure
    * only authorized callers can close the file handler.
    */
   void doClose()
   {
      super.close();
   }
}