/*
** Module   : AddressPair.java
** Abstract : container for addresses during INIT procedure
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 SIY 20050120 NEW   @19489 Created an container for the addresses
**                               (integers).
** 002 SIY 20050315 CHG   @20374 Organized imports, fixed spelling and other 
**                               minor cleanups.
** 003 GES 20070115 CHG   @31827 Doc and formatting cleanups.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.net;

import java.io.Serializable;

/**
 * This is a simple container as a payload during the address exchange
 * procedure (INIT_STANDARD/INIT_CONVERSATION/INIT_REPLY messages).
 * <p>
 * The class contains two addresses, one of which represents the local node
 * address while the other represents the address of the remote node.
 * <p>
 * By convention, the node which initiates a connection sends an 
 * INIT_STANDARD or INIT_CONVERSATION message. The node which accepts a
 * connection processes that message and sends an INIT_REPLY message.
 * <p>
 * The member names of this class support this convention and represent
 * addresses from the point of view of the node which sends the INIT_STANDARD
 * or INIT_CONVERSATION message (nodeAddress represents the address of
 * the node which sends the INIT message and remoteAddress represents the 
 * address of the node which sends the INIT_REPLY message).
 * 
 * @author SIY
 */
class AddressPair
implements Serializable
{
   /**
    * Stores local node address from the point of view of the node which
    * sends the initialization request message.
    */
   private int nodeAddress;

   /**
    * Stores remote node address from the point of view of the node which
    * sends the initialization request message.
    */
   private int remoteAddress;

   /**
    * Default constructor which instantiates an <code>AddressPair</code>
    * object with both addresses set to 0.
    */
   AddressPair()
   {
      this(0, 0);
   }

   /**
    * Constructs an instance using two provided addresses.
    * 
    * @param    node
    *           Local node address.
    * @param    remote
    *           Remote node address.
    */
   AddressPair(int node, int remote)
   {
      nodeAddress = node;
      remoteAddress = remote;
   }

   /**
    * Returns the local node address.
    * 
    * @return   Local node address.
    */
   int getNodeAddress()
   {
      return nodeAddress;
   }

   /**
    * Returns the remote node address.
    * 
    * @return   Remote node address.
    */
   int getRemoteAddress()
   {
      return remoteAddress;
   }

   /**
    * Sets the local node address.
    * 
    * @param    address
    *           Address which will be set as the local node address.
    */
   void setNodeAddress(int address)
   {
      nodeAddress = address;
   }

   /**
    * Sets the remote node address.
    * 
    * @param    address
    *           Address which will be set as the remote node address.
    */
   void setRemoteAddress(int address)
   {
      remoteAddress = address;
   }

   /**
    * Convenience method to swap the local and remote addresses.
    */
   void swapAddresses()
   {
      int tmp = nodeAddress;
      nodeAddress = remoteAddress;
      remoteAddress = tmp;
   }
}