/*
** Module   : StateSynchronizer.java
** Abstract : an interface for objects which manage incremental and 
**            bidirectional synchronization of state between two sides of a
**            P2J session
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20060725 NEW   @28196 Created initial version an interface for 
**                               objects which manage incremental and
**                               bidirectional synchronization of state 
**                               between two sides of a P2J session.
**
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.net;

import java.io.*;

/**
 * An interface for objects which manage incremental and bidirectional 
 * synchronization of state between two sides of a P2J session.  This is
 * useful to avoid "notification" trips which maintain state synchronously
 * on the other side of the connection but which doesn't need to trigger any 
 * actual processing that has end-user impact.  In other words, if the state
 * change can be deferred until the next time a real trip is needed to the
 * other side of the connection, then the state changes can be "batched" up
 * and "piggybacked" on the next real message.  This can have large
 * performance advantages in cases where the notification trips are frequent
 * (e.g. possibly because they are done in tight loops) but the real worker
 * trips are infrequent.
 * <p>
 * It is up to the implementer of this interface to determine if partial 
 * or incremental changes can be sent or whether all changes must be sent 
 * each time.  If at all possible, it is best to only send the changes and 
 * when no changes have been made, to send nothing.  This is significantly
 * more efficient at the protocol level.
 *
 * @author GES
 */
public interface StateSynchronizer
{
   /**
    * Called by the protocol driver to obtain the latest state that should
    * be sent to the other side of the connection.  This is called just 
    * before a "real" message is sent and any changes provided will be
    * piggybacked into that message.
    * <p>
    * To enable incremental changes, it is guaranteed that this method will
    * be called every time any message is sent to the other side of the
    * connection.  In addition, this will only be called once per message.
    * <p>
    * In a system in which control flow is "sent" to the other side of the 
    * connection (a type of "conversation" or single-threaded synchronous
    * mode), it is further guaranteed that any changes that are made at the
    * other side after this is called will result in a call to this side's
    * {@link #applyChanges} method before this method is ever called again
    * (when control flow returns to this side of the connection). This allows
    * one to ensure that on the call to this method, any "dirty" flags can be
    * safely cleared and the state of this side will be updated with
    * the results of the other side's changes before this method must ever
    * again check to see if there are changes to send.
    *
    * @return   The payload of state changes to be sent to the other side or
    *           <code>null</code> if no changes need to be sent.
    */
   public Serializable getChanges();

   /**
    * Called by the protocol driver upon reading a message that contains
    * state changes that need to be applied to this side of the connection.
    * This is called just after a new message is read from the connection
    * and only if that message had piggybacked state changes attached. This
    * is guaranteed to be called before the containing message is enqueued
    * for further "real" processing on this node.
    *
    * @param    changes
    *           The payload of state changes that must be locally applied.
    */
   public void applyChanges(Serializable changes);
}
