/*
** Module   : SyncMessage.java
** Abstract : state synchronization extensions to a standard message
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20060726 NEW   @28197 Created initial version with state 
**                               synchronization extensions to a standard
**                               message.
** 002 GES 20060731 CHG   @28257 Removed extraneous call to the parent's
**                               readObject/writeObject methods since the
**                               Java infrastructure does this for us.
** 003 GES 20060801 CHG   @28290 Converted to Externalizable to improve
**                               efficiency because Serializable has unneeded
**                               processing (for versioning...) that is
**                               executed on "inflation".
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.net;

import java.io.*;

/**
 * This class provides state synchronization extensions to a standard 
 * message.  In particular, there is an additional "payload" that is
 * provided to allow any state changes that need to be synchronized to be
 * "piggybacked" into any existing message.
 * 
 * @author GES
 */
public class SyncMessage
extends Message
{
   /** State to be synchronized. */
   private Serializable state = null;

   /**
    * Default constructor, used during de-serialization.
    */
   public SyncMessage()
   {
   }

   /**
    * Constructs an instance using an already existing message.
    *
    * @param    msg
    *           A valid message.  Must not be <code>null</code>.
    * @param    state
    *           The state to be piggybacked into the given message.
    */
   public SyncMessage(Message msg, Serializable state)
   {
      super(msg);
      this.state = state;
   }

   /**
    * Get the state from the message.
    * 
    * @return   The state carried in this message.
    */
   public Serializable getState()
   {
      return state;
   }

   /**
    * Set the state in the message.
    * 
    * @param    state
    *           The state to be carried in this message.
    */
   public void setState(Serializable state)
   {
      this.state = state;
   }

   /**
    * Replacement for the default object reading method. The piggybacked
    * state is read from the input source. Note that the parent's version of
    * this method is called first because as an <code>Externalizable</code>
    * it won't be called by the serialization infrastructure.
    * 
    * @param    in
    *           Input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      state = (Serializable) in.readObject();
   }

   /**
    * Replacement for the default object writing method. The piggybacked
    * state is written to the output source. Note that the parent's version of
    * this method is called first because as an <code>Externalizable</code>
    * it won't be called by the serialization infrastructure.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeObject(state);
   }
}