/*
** Module   : PropsWorker.java
** Abstract : provides functions for accessing J2SE system properties
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 GES 20050228 NEW   @20118 Created initial version supporting a simple
**                               query mechanism that is data type specific.
**                               There is also a function that allows the
**                               caller to check if a named property exists.
** 002 ECF 20050303 CHG   @20169 Replaced getIntegerProperty with
**                               getLongProperty, since Integer return types
**                               are not supported by the expression engine.
** 003 ECF 20050526 CHG   @21290 Changes to support new expression engine
**                               implementation. Changed library registration
**                               mechanism based on superclass' modifications
**                               to support single library per pattern worker
**                               limit.
** 004 EVL 20160223              Javadoc fixes to make compatible with Oracle Java 8 for
**                               Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.pattern;

/**
 * Provides a set of user-defined functions for accessing system properties.
 *
 * @author    GES
 * @version   1.0.1
 */
public class PropsWorker
extends AbstractPatternWorker
{
   /**
    * Default constructor which calls the super constructor and registers
    * libraries with the pattern engine.
    */
   public PropsWorker()
   {
      super();
      setLibrary(new Props());
   }
   
   /**
    * This is a pattern engine library that provides easy access to system
    * properties.
    */
   public class Props
   {
      /**
       * Determines if a named system property exists or not. 
       *
       * @param    prop
       *           The system property to lookup.
       *
       * @return   <code>true</code> if a system property of the specified
       *           name exists, <code>false</code> otherwise.
       */
      public boolean isProperty(String prop)
      {
         return (System.getProperty(prop) != null);
      }
      
      /**
       * Return a named system property as a string value. 
       *
       * @param    prop
       *           The system property to lookup.
       *
       * @return   The value of the system property or <code>null</code> if
       *           no such property exists.
       */
      public String getStringProperty(String prop)
      {
         return System.getProperty(prop);
      }
      
      /**
       * Return a named system property as a <code>Double</code> value. 
       *
       * @param    prop
       *           The system property to lookup.
       *
       * @return   The value of the system property or <code>null</code> if
       *           no such property exists.
       */
      public Double getDoubleProperty(String prop)
      throws NumberFormatException
      {
         String ddd = System.getProperty(prop);
         
         if (ddd == null)
            return null;
         
         return new Double(ddd);
      }
      
      /**
       * Return a named system property as a <code>Long</code> value. 
       *
       * @param    prop
       *           The system property to lookup.
       *
       * @return   The value of the system property or <code>null</code> if
       *           no such property exists.
       */
      public Long getLongProperty(String prop)
      throws NumberFormatException
      {
         String lll = System.getProperty(prop);
         
         if (lll == null)
            return null;
         
         return new Long(lll);
      }
      
      /**
       * Return a named system property as a <code>Boolean</code> value. 
       *
       * @param    prop
       *           The system property to lookup.
       *
       * @return   The value of the system property or <code>null</code> if
       *           no such property exists.
       */
      public Boolean getBooleanProperty(String prop)
      throws NumberFormatException
      {
         String bbb = System.getProperty(prop);
         
         if (bbb == null)
            return null;
         
         return new Boolean(bbb);
      }
   }
}