/*
** Module   : AbstractReversible.java
** Abstract : Partial implementation of API of abstract Reversible class
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 ECF 20080925   @39952 Created initial version. Partial, default
**                           implementation of API of abstract Reversible
**                           class.
** 002 ECF 20081020   @40206 Refactored rollback processing. We now avoid
**                           improperly flushing intermediate changes to the
**                           database.
** 003 SVL 20081127   @40697 tempDMOs increment (into prepareRollback()) and
**                           and decrement (into rollback()) DMO use count.
** 004 ECF 20090601   @42578 Minor change. Required by API change in
**                           Persistence class.
** 005 CA  20091202   @44465 Reworked setDmo to be compatible with the new 
**                           rollback mechanism.
** 006 ECF 20140305          Replaced Apache commons logging with J2SE logging.
** 007 ECF 20140823          Added getIdentifier method.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.util.logging.*;
import com.goldencode.p2j.util.ErrorManager;

/**
 * Abstract implementation of {@link Reversible}.  Adds storage of the buffer
 * associated with the reversible action, and a record identifier which is
 * used to retrieve the DMO to be acted upon.  The identifier is used instead
 * of the DMO itself in order to save the memory overhead.  The DMO is only
 * retrieved (using the identifier) when it is actually needed.
 */
abstract class AbstractReversible
extends Reversible
{
   /** Buffer which generated this reversible */
   private final RecordBuffer buffer;
   
   /** Identifier for the database record being acted upon */
   private RecordIdentifier ident = null;
   
   /** A DMO stored temporarily during rollback processing */
   private Persistable tempDMO = null;
   
   /**
    * Constructor.
    * 
    * @param   buffer
    *          Record buffer which generated this reversible.
    * @param   dmo
    *          Object representing record upon which an action may be
    *          reversed.  May not be <code>null</code>.
    */
   public AbstractReversible(RecordBuffer buffer, Persistable dmo)
   {
      setDMO(dmo);
      this.buffer = buffer;
   }

   /**
    * Roll back the database change represented by this object.
    * <p>
    * Subclasses must implement this method in a manner appropriate to
    * the database actions they represent.
    * 
    * @param   rollForward
    *          If <code>true</code>, this method is being invoked in order to
    *          re-roll forward a non-undoable change, after a database-level
    *          rollback.  This indicates the modification is made within a
    *          new transaction invoked after the database-level rollback.
    *          If <code>false</code>, this method is being invoked in order to
    *          roll back an undoable change within the context of an
    *          application-level, full transaction or sub-transaction
    *          rollback.
    * @param   transaction
    *          <code>true</code> if this rollback occurs at a full transaction
    *          boundary, else <code>false</code>.
    *
    * @throws  PersistenceException
    *          if an error occurs at the database or in the persistence
    *          layer when reversing the action.
    */
   protected abstract void rollbackWorker(boolean rollForward, boolean transaction)
   throws PersistenceException;
   
   /**
    * Roll back the database change represented by this object.
    * <p>
    * This implemetation simply invokes {@link #rollbackWorker(boolean,
    * boolean)} and ensures the temporary DMO stored by {@link
    * #prepareRollback(Persistable)} is nulled out afterward to enable it to
    * be garbage collected.
    * 
    * @param   rollForward
    *          If <code>true</code>, this method is being invoked in order to
    *          re-roll forward a non-undoable change, after a database-level
    *          rollback.  This indicates the modification is made within a
    *          new transaction invoked after the database-level rollback.
    *          If <code>false</code>, this method is being invoked in order to
    *          roll back an undoable change within the context of an
    *          application-level, full transaction or sub-transaction
    *          rollback.
    * @param   transaction
    *          <code>true</code> if this rollback occurs at a full transaction
    *          boundary, else <code>false</code>.
    *
    * @throws  PersistenceException
    *          if an error occurs at the database or in the persistence
    *          layer when reversing the action.
    */
   protected final void rollback(boolean rollForward, boolean transaction)
   throws PersistenceException
   {
      try
      {
         rollbackWorker(rollForward, transaction);
      }
      finally
      {
         if (tempDMO != null)
         {
            buffer.decrementDMOUseCount(tempDMO);
            tempDMO = null;
         }
      }
   }

   /**
    * Indicate whether this <code>Reversible</code> implementation stores its
    * DMO locally, as opposed to storing an identifier and having to retrieve
    * the DMO from the backing database, or not making a DMO available at all.
    * 
    * @return  <code>true</code> if the DMO is stored locally, else
    *          <code>false</code>.
    */
   protected boolean storesDMO()
   {
      return false;
   }
   
   /**
    * Get the record identifier of the underlying DMO.
    * 
    * @return  Identifier or <code>null</code> if none is stored.
    */
   protected RecordIdentifier getIdentifier()
   {
      return ident;
   }
   
   /**
    * Get the underlying DMO.  If the record identifier matches the record
    * currently stored in the associated record buffer, that record is
    * returned.  Otherwise, the identifier is used to retrieve the record in
    * its current state from the database.
    *
    * @return  DMO or <code>null</code> if no DMO is associated with this
    *          reversible.
    */
   protected Persistable getDMO()
   {
      if (ident == null)
      {
         return null;
      }
      
      if (tempDMO != null)
      {
         return tempDMO;
      }
      
      Persistable bufferDMO = buffer.getCurrentRecord();
      Serializable id = ident.getRecordID();
      
      if (bufferDMO != null && bufferDMO.getId().equals(id))
      {
         return bufferDMO;
      }
      
      Persistence persistence = buffer.getPersistence();
      
      try
      {
         return persistence.quickLoad(ident);
      }
      catch (PersistenceException exc)
      {
         LOG.log(Level.SEVERE, "Unable to load reversible DMO [" + ident + "]", exc);
      }
      
      return null;
   }
   
   /**
    * Set the underlying DMO for this reversible action.  Stores only an
    * identifier for the record, not the DMO itself.
    * 
    * @param   dmo
    *          The underlying DMO; may not be <code>null</code> if {@link
    *          #isNullDMOAllowed()} returns <code>false</code>.
    * 
    * @throws  NullPointerException
    *          if <code>dmo</code> is <code>null</code> and this is
    *          disallowed.
    */
   protected void setDMO(Persistable dmo)
   {
      if (dmo == null)
      {
         if (isNullDMOAllowed())
         {
            ident = null;
            
            return;
         }
         
         throw new NullPointerException("Reversible DMO may not be null");
      }
      
      Class<?> dmoClass = DatabaseManager.normalizeDMOClass(dmo.getClass());
      String entity = dmoClass.getName();
      
      ident = new RecordIdentifier(entity, dmo.getId());
      
      if (buffer != null && (tempDMO != null || isNewlyCreated()))
      {
         // for records created in current block, the DMO may be null. if a
         // change is needed, this ensures the change will be allowed
         boolean inc = false;
         if (dmo != tempDMO)
         {
            inc = true;
            if (tempDMO != null)
            {
               try
               {
                  // when switching, get rid of the old DMO
                  buffer.decrementDMOUseCount(tempDMO);
               }
               catch (PersistenceException exc)
               {
                  ErrorManager.recordOrThrowError(exc);
               }
            }
         }
         
         tempDMO = dmo;
         
         if (inc)
         {
            // this is a new DMO, record its usage
            buffer.incrementDMOUseCount(tempDMO);
         }
      }
   }
   
   /**
    * Prepare for this reversible to be rolled back.  If this object's
    * temporary DMO has not yet been set, it will be set to the given DMO.  If
    * that DMO is <code>null</code>, the {@link #getDMO()} method will be used
    * to retrieve the backing DMO from the database.
    * 
    * @param   dmo
    *          If non-<code>null</code>, the caller is providing a DMO which
    *          it thinks is the appropriate object to act upon during rollback
    *          processing.
    */
   protected void prepareRollback(Persistable dmo)
   {
      if (tempDMO == null)
      {
         if (dmo == null)
         {
            dmo = getDMO();
         }

         tempDMO = dmo;

         if (tempDMO != null)
            buffer.incrementDMOUseCount(tempDMO);
      }
   }
   
   /**
    * Get the ID of the underlying DMO.
    *
    * @return  DMO ID.
    */
   protected Serializable getId()
   {
      return (ident == null ? null : ident.getRecordID());
   }
   
   /**
    * Get the unqualified interface name of the DMO being managed by this
    * reversible.
    * 
    * @return  DMO interface name.
    */
   protected String getDMOName()
   {
      return buffer.getDMOName();
   }
   
   /**
    * Get the record buffer which generated this reversible.
    * 
    * @return  Record buffer.
    */
   protected RecordBuffer getBuffer()
   {
      return buffer;
   }
   
   /**
    * Indicate whether a DMO passed to {@link #setDMO(Persistable)} may be
    * <code>null</code>.  This implementation returns <code>false</code>;  it
    * should be overridden if a subclass requires different behavior.
    * 
    * @return  <code>false</code>.
    */
   protected boolean isNullDMOAllowed()
   {
      return false;
   }
}