/*
** Module   : ConversionProfile.java
** Abstract : Enumeration of conversion configuration profiles for use with runtime conversion.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20140405 Created initial version for use with ConversionPool.
** 002 OM  20150120 Added POSTPROCESS_FIND_QUERY and POSTPROCESS_OPEN_QUERY items. Removed the
**                  deprecated FIX_FIND and FIX_QUERY profiles.
** 003 ECF 20150327 Added P2O_POST profile.
** 004 ECF 20150407 Added BREW_DMO_ASM profile.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

/**
 * An enumeration of conversion configuration profiles which can be loaded and performed by the
 * pattern engine at runtime to convert dynamic temp-tables and queries.
 */
enum ConversionProfile
{
   /** Convert schema AST to p2o AST */
   P2O("schema/p2o", false),
   
   /** Finish conversion of schema AST to p2o AST (relations and indexes) */
   P2O_POST("schema/p2o_post", false),
   
   /** Generate Hibernate ORM document */
   HIBERNATE("schema/hibernate", false),
   
   /** Create DMO interface and implementation class */
   JAVA_DMO("schema/java_dmo", false),
   
   /** Annotations */
   ANNOTATIONS("annotations/annotations", true),
   
   /** Business logic base structure */
   BASE_STRUCTURE("convert/base_structure", true),
   
   /** Core conversion */
   CORE_CONVERSION("convert/core_conversion", true),
   
   /** Anti-parse Java AST to source code */
   BREW("convert/brew", false),
   
   /** Compile DMO Java AST directly to bytecode using ASM */
   BREW_DMO_ASM("schema/brew_dmo_asm", false),
   
   /** In-memory post-process the FIND query JAST to help interpret it faster */
   POSTPROCESS_FIND_QUERY("runtime/postprocess_find_query", false),
   
   /** In-memory  post-process the OPEN query JAST to help interpret it faster */
   POSTPROCESS_OPEN_QUERY("runtime/postprocess_open_query", false);
   
   /** Profile name */
   private final String name;
   
   /** Flag indicating whether conditional rule-set processing should be enabled */
   private final boolean conditional;
   
   /**
    * Constructor.
    * 
    * @param   name
    *          Profile name.
    * @param   conditional
    *          Flag indicating whether conditional rule-set processing should be enabled.
    */
   private ConversionProfile(String name, boolean conditional)
   {
      this.name = name;
      this.conditional = conditional;
   }
   
   /**
    * Get the string representation of this enum value.
    * 
    * @return  String representation.
    */
   public String toString()
   {
      return name;
   }
   
   /**
    * Get the profile name.
    * 
    * @return  Profile name.
    */
   String getName()
   {
      return name;
   }
   
   /**
    * Get the conditional flag.
    * 
    * @return  Conditional flag.
    */
   boolean isConditional()
   {
      return conditional;
   }
}