/*
** Module   : IndexHelper.java
** Abstract : Helper object to manage database index information
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 ECF 20080513   @38618 Created initial version. Helper object to manage
**                           database index information.
** 002 ECF 20080606   @38644 Support API change in DatabaseManager.
** 003 ECF 20080610   @38680 Fixed lookup of index by sort phrase. We were not
**                           properly handling primary key column in indexes.
**                           Added debug logging.
** 004 SVL 20080730   @39224 Sorters are stored by sort clauses, not by
**                           indexes. 
** 005 CA  20080819   @39461 Support API change in DMOSorter.
** 006 ECF 20090508   @42125 Added maps of index names to property lists.
**                           Needed for validation improvements.
** 007 ECF 20090513   @42166 Fixed getIndexesForProperty(). Must not return
**                           null.
** 008 ECF 20090604   @42588 Fixed getIndexesForEntity(). Was not taking
**                           uniqueness flag into account properly when
**                           caching results.
** 009 ECF 20090707   @43100 Fixed lookupIndexesForDMO(). Instead of omitting
**                           merely the index containing only the primary key,
**                           we were omitting any index containing the primary
**                           key.
** 010 ECF 20131028          Import change.
** 011 VMN 20131128          Added map of entity to primary index and method getPrimaryIndex().
** 012 VMN 20140423          Added enhance schema name conversion support for hint "escape"
**                           attribute.
** 013 OM  20140509          Replaced Apache logging. Applied latest coding-style.
**                           Added intermediary implementation of getLegacyIndexName().
** 014 SVL 20140602          Added WHOLE_INDEX_PREFIX.
** 015 OM  20140617          Removed getLegacyIndexName() as it was a duplicate of 
**                           TableMapper.getLegacyIndexName().
** 016 ECF 20150201          Fixed lookupIndexForSort to properly match a non-unique index which
**                           has had the table's primary key appended to its components.
** 017 ECF 20150801          Minor performance improvements. Some uses of String.intern() were
**                           redundant with those of clients of this class.
** 018 ECF 20160517          Reimplemented getPrimaryIndex to return null instead of throwing an
**                           exception if no primary index is found.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.util.*;
import java.util.logging.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.util.*;

/**
 * Helper object to manage database index information.  One instance of this
 * class exists for each permanent, primary database.  Provides methods to
 * retrieve:
 * <ul>
 *   <li>all indexed properties for a DMO type
 *   <li>all index names for the table backing a DMO type
 *   <li>all indexes in which the column backing a DMO property participates
 *   <li>the index name associated with an index-driven query's sort phrase
 *   <li>a {@link DMOSorter} instance appropriate to an index-driven query's
 *       sort phrase
 * </ul>
 * <p>
 * It is anticipated that the lists of index names returned by some of the
 * methods in this class will be used for purposes of serialization of
 * index resources via a {@link LockManager}.  Therefore, unless otherwise
 * noted, these lists will be sorted in ascending, alphabetical order, to
 * help prevent deadlocks in situations where all such indexes will be locked
 * by a particular context.  That is, if all contexts acquiring locks on these
 * indexes are careful to do so in the same order, deadlocks should not occur.
 */
public final class IndexHelper
{
   /** Prefix for index information string. */
   public static final String WHOLE_INDEX_PREFIX = "WHOLE-INDEX,";
   
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(IndexHelper.class);
   
   /** Cache of instances, mapped by database */
   private static final Map<Database, IndexHelper> cache = new HashMap<>();
   
   /** Primary (non-dirty) database with which this helper is associated */
   private final Database database;
   
   /** Map of sort clauses to index names */
   private final Map<SortKey, String> indexBySort = new HashMap<>();
   
   /** Map of property names to lists of indexes containing them */
   private final Map<PropertyKey, List<String>> indexesByProperty = new HashMap<>();
   
   /** Map of sort clauses to <code>DMOSorter</code> objects */
   private final Map<SortKey, DMOSorter> sorters = new HashMap<>();
   
   /** Map of entity keys to lists of all index names */
   private final Map<EntityIndexKey, List<String>> indexesByEntity = new HashMap<>();
   
   /** Map of entity to primary index */
   private final Map<String, P2JIndex> primaryIndexByEntity = new HashMap<>();
   
   /** Map of entity keys to maps of index names to index property names */
   private final Map<EntityIndexKey, Map<String, List<String>>> indexMapsByEntity =
      new HashMap<>();
   
   /** Map of entity keys to indexed properties */
   private final Map<EntityIndexKey, Set<String>> propsByEntity = new HashMap<>();
   
   /**
    * Private constructor.  Instances of this class are created only by the
    * public factory method {@link #get(com.goldencode.p2j.persist.Database)}.
    * 
    * @param   database
    *          Database for which index information is needed.
    */
   private IndexHelper(Database database)
   {
      this.database = database;
   }
   
   /**
    * Get an instance of this class for the given database.  If an instance
    * does not exist in the shared cache, one is created and stored for future
    * requests.
    * 
    * @param   database
    *          Database for which index information is needed.
    * 
    * @return  Instance of this class for the given database.
    */
   public static IndexHelper get(Database database)
   {
      IndexHelper helper = null;
      
      synchronized (cache)
      {
         helper = cache.get(database);
         if (helper == null)
         {
            helper = new IndexHelper(database);
            cache.put(database, helper);
         }
      }
      
      return helper;
   }
   
   /**
    * Retrieve a list of the names of all indexes in which the database
    * column associated with the given DMO property participates.
    * 
    * @param   entity
    *          A DMO entity name which represents a database table.
    * @param   property
    *          A DMO property which represents a database column.
    * 
    * @return  A list of index names in ascending alphabetical order.  The
    *          list will not contain duplicates, but it may be empty.  It will
    *          never be <code>null</code>.
    *          
    * @throws  PersistenceException
    *          if there is any error querying index metadata from the JDBC
    *          driver of the associated database.
    */
   public List<String> getIndexesForProperty(String entity, String property)
   throws PersistenceException
   {
      List<String> indexes = null;
      property = property.intern();
      PropertyKey key = new PropertyKey(entity, property);
      
      synchronized (indexesByProperty)
      {
         indexes = indexesByProperty.get(key);
      }
      
      if (indexes == null)
      {
         indexes = lookupIndexesForProperty(entity, property);
         if (indexes == null)
         {
            indexes = Collections.emptyList();
         }
         else
         {
            Collections.sort(indexes);
            indexes = Collections.unmodifiableList(indexes);
         }
         
         synchronized (indexesByProperty)
         {
            List<String> old = indexesByProperty.put(key, indexes);
            if (old != null)
            {
               indexes = old;
               indexesByProperty.put(key, old);
            }
         }
      }
      
      return indexes;
   }
   
   /**
    * Given a sort phrase from an index-driven query, return the name of the
    * index represented by the sort phrase.
    * 
    * @param   buffer
    *          Record buffer which contains information about the DMO entity
    *          associated with the sort phrase.
    * @param   sort
    *          A sort phrase which includes one or more DMO property
    *          references, each qualified by a DMO alias and followed by a
    *          keyword indicating sort direction.
    * 
    * @return  Name of the index associated with the given sort phrase, or
    *          <code>null</code> if no matching index can be found.
    * 
    * @throws  PersistenceException
    *          if there is any error querying index metadata from the JDBC
    *          driver of the associated database.
    */
   public String getIndexForSort(RecordBuffer buffer, String sort)
   throws PersistenceException
   {
      String index = null;
      String entity = buffer.getEntityName();
      SortKey key = new SortKey(entity, sort);
      
      synchronized (indexBySort)
      {
         index = indexBySort.get(key);
         if (index == null)
         {
            index = lookupIndexForSort(buffer, sort);
         }
         else if ("".equals(index))
         {
            index = null;
         }
      }
      
      return index;
   }
   
   /**
    * Retrieve primary index of given database table.
    *
    * @param   entity
    *          A DMO entity name which represents a database table.
    *
    * @return  Primary index.
    *
    * @throws  PersistenceException
    *          if there is any error in retrieving primary index or more than one primary index
    *          is found.
    */
   public P2JIndex getPrimaryIndex(String entity)
   throws PersistenceException
   {
      P2JIndex index;
      
      synchronized (primaryIndexByEntity)
      {
         index = primaryIndexByEntity.get(entity);
         
         // nulls are permitted and mean the entity has no primary index
         if (index == null && primaryIndexByEntity.containsKey(entity))
         {
            return null;
         }
      }
      
      if (index == null)
      {
         String indexName = null;
         Class<?> dmoClass = DBUtils.dmoClassForEntity(entity);
         LegacyIndexes lIndexes = dmoClass.getAnnotation(LegacyIndexes.class);
         if (lIndexes != null)
         {
            for (LegacyIndex lIndex : lIndexes.value())
            {
               if (lIndex.primary())
               {
                  if (indexName == null)
                  {
                     indexName = lIndex.converted();
                  }
                  else
                  {
                     throw new PersistenceException("Table '" + entity +
                                                    "' has two primary indexes: '" +
                                                    indexName + "' and '" +
                                                    lIndex.name() +"'");
                  }
               }
            }
         }
         if (indexName == null)
         {
            throw new PersistenceException("Table '" + entity + "' has no primary index");
         }
         else
         {
            indexName = DBUtils.getPrefixedParameter(indexName, "idx__");
            String schema = DatabaseManager.getSchemaByClass(dmoClass);
            Class<?> dmoIface = DMOIndex.getDMOInterface(schema, dmoClass);
            Iterator<P2JIndex> iter = DMOIndex.databaseIndexes(schema, database, dmoIface);
            while (iter.hasNext())
            {
               P2JIndex ind = iter.next();
               if (indexName.equals(ind.getName()))
               {
                  index = ind;
                  break;
               }
            }
         }
         
         synchronized (primaryIndexByEntity)
         {
            // cache index, even if none found (i.e., null)
            primaryIndexByEntity.putIfAbsent(entity, index);
         }
      }
      
      return index;
   }
   
   /**
    * Retrieve a {@link DMOSorter} instance which is able to sort DMO
    * instances according to the sort criteria specified by the given sort
    * phrase.
    * 
    * @param   buffer
    *          Record buffer which contains information about the DMO entity
    *          associated with the sort phrase.
    * @param   sort
    *          A sort phrase which includes one or more DMO property
    *          references, each qualified by a DMO alias and followed by a
    *          keyword indicating sort direction.
    * 
    * @return  DMO sorter object which can sort DMOs according to the given
    *          index.
    *
    * @throws  PersistenceException
    *          if there is any error querying index metadata from the JDBC
    *          driver of the associated database.
    */
   public DMOSorter getSorterForSortPhrase(RecordBuffer buffer, String sort)
   throws PersistenceException
   {
      SortKey key = new SortKey(buffer.getEntityName(), sort);
      
      synchronized (indexBySort)
      {
         String index = indexBySort.get(key);
         if (index == null)
         {
            // This will add an appropriate sorter to the sorters map, if possible.
            lookupIndexForSort(buffer, sort);
         }
         
         return sorters.get(key);
      }
   }
   
   /**
    * Retrieve a list of all index names associated with the table represented
    * by the given DMO entity name, or the sublist which represents unique
    * indexes only.
    * 
    * @param   entity
    *          A DMO entity name which represents a database table.
    * @param   unique
    *          If <code>true</code>, only unique indexes are returned;  if
    *          <code>false</code>, all indexes are returned.
    * 
    * @return  The names of all or only unique indexes for the given entity,
    *          sorted in ascending alphabetical order.  The list will not
    *          contain duplicates, but it may be empty.  It will never be
    *          <code>null</code>.
    *          
    * @throws  PersistenceException
    *          if there is any error querying index metadata from the JDBC
    *          driver of the associated database.
    */
   public List<String> getIndexesForEntity(String entity, boolean unique)
   throws PersistenceException
   {
      List<String> list = null;
      EntityIndexKey key = new EntityIndexKey(entity, unique);
      
      synchronized (indexesByEntity)
      {
         list = indexesByEntity.get(key);
      }
      
      if (list == null)
      {
         list = lookupIndexesForDMO(entity, unique);
         if (list != null)
         {
            Collections.sort(list);
            list = Collections.unmodifiableList(list);
         }
         else
         {
            list = Collections.emptyList();
         }
         
         synchronized (indexesByEntity)
         {
            List<String> old = indexesByEntity.put(key, list);
            if (old != null)
            {
               indexesByEntity.put(key, old);
            }
         }
      }
      
      return list;
   }
   
   /**
    * Retrieve a mapping of index names to their component property names, for
    * indexes associated with the table represented by the given DMO entity
    * name.
    * 
    * @param   entity
    *          A DMO entity name which represents a database table.
    * @param   unique
    *          If <code>true</code>, only unique indexes are returned;  if
    *          <code>false</code>, only non-unique indexes are returned.
    * 
    * @return  A mapping of index names to ordered lists of the names of their
    *          component properties, for the given DMO entity.  The map
    *          represents only unique indexes, or only non-unique indexes,
    *          depending upon the <code>unique</code> parameter.  The map may
    *          be empty, but it will never by <code>null</code>.  The property
    *          names will not be wrapped in any functions.  The map will
    *          iterate in ascending alphabetical order by index name.
    *          
    * @throws  PersistenceException
    *          if there is any error querying index metadata from the JDBC
    *          driver of the associated database.
    */
   public Map<String, List<String>> getIndexMapForEntity(String entity, boolean unique)
   throws PersistenceException
   {
      Map<String, List<String>> map = null;
      EntityIndexKey key = new EntityIndexKey(entity, unique);
      
      synchronized (indexMapsByEntity)
      {
         map = indexMapsByEntity.get(key);
      }
      
      if (map == null)
      {
         map = lookupIndexMapForDMO(entity, unique);
         if (map != null)
         {
            map = Collections.unmodifiableMap(map);
         }
         else
         {
            map = Collections.emptyMap();
         }
         
         synchronized (indexMapsByEntity)
         {
            Map<String, List<String>> old = indexMapsByEntity.put(key, map);
            if (old != null)
            {
               indexMapsByEntity.put(key, old);
            }
         }
      }
      
      return map;
   }
   
   /**
    * Retrieve the set of all property names for a particular DMO entity,
    * which represent database columns that participate in an index on the
    * entity's backing table, or the subset which represent database columns
    * that participate in a unique index.
    * <p>
    * Note:  the returned set will <em>not</em> include the reserved primary
    * key property <code>id</code>.
    * 
    * @param   entity
    *          A DMO entity name which represents a database table.
    * @param   unique
    *          If <code>true</code>, only properties associated with unique
    *          indexes are returned;  if <code>false</code>, all indexed
    *          properties are returned.
    * 
    * @return  Set of indexed property names.  The set may be empty but it
    *          will never be <code>null</code>.
    * 
    * @throws  PersistenceException
    *          if there is any error querying index metadata from the JDBC
    *          driver of the associated database.
    */
   public Set<String> getIndexedProperties(String entity, boolean unique)
   throws PersistenceException
   {
      Set<String> set = null;
      EntityIndexKey key = new EntityIndexKey(entity, unique);
      
      synchronized (propsByEntity)
      {
         set = propsByEntity.get(key);
      }
      
      if (set == null)
      {
         set = lookupIndexedProperties(entity, unique);
         if (set != null)
         {
            set = Collections.unmodifiableSet(set);
         }
         else
         {
            set = Collections.emptySet();
         }
         
         synchronized (propsByEntity)
         {
            Set<String> old = propsByEntity.put(key, set);
            if (old != null)
            {
               propsByEntity.put(key, old);
            }
         }
      }
      
      return set;
   }
   
   /**
    * Collect a list of all index names associated with the table represented
    * by the given DMO entity name.
    * 
    * @param   entity
    *          A DMO entity name which represents a database table.
    * @param   unique
    *          If <code>true</code>, only unique indexes are returned;  if
    *          <code>false</code>, all indexes are returned.
    * 
    * @return  The names of all indexes for the given entity, in no particular
    *          order, or <code>null</code> if none are found.
    *          
    * @throws  PersistenceException
    *          if there is any error querying index metadata from the JDBC
    *          driver of the associated database.
    */
   private List<String> lookupIndexesForDMO(String entity, boolean unique)
   throws PersistenceException
   {
      ArrayList<String> list = null;
      Class<?> dmoClass = DBUtils.dmoClassForEntity(entity);
      String schema = DatabaseManager.getSchemaByClass(dmoClass);
      Class<?> dmoIface = DMOIndex.getDMOInterface(schema, dmoClass);
      Iterator<P2JIndex> iter = DMOIndex.databaseIndexes(schema, database, dmoIface);
      
      outer:
      while (iter.hasNext())
      {
         P2JIndex index = iter.next();
         
         if (unique && !index.isUnique())
         {
            continue;
         }
         
         // Special case:  omit index on (only) primary key.
         Iterator<P2JIndexComponent> comps = index.components();
         while (comps.hasNext())
         {
            P2JIndexComponent comp = comps.next();
            String compName = comp.getName();
            
            if (DatabaseManager.PRIMARY_KEY.equals(compName))
            {
               continue outer;
            }
            else
            {
               break;
            }
         }
         
         if (list == null)
         {
            list = new ArrayList<>();
         }
         
         list.add(index.getName());
      }
      
      if (list != null)
      {
         list.trimToSize();
      }
      
      if (LOG.isLoggable(Level.FINE))
      {
         LOG.log(Level.FINE,
                 "Indexes " + (unique ? "[unique]" : "[all]") +
                       " for entity '" + entity + "':  " + list
         );
      }
      
      return list;
   }
   
   /**
    * Generate a mapping of index names to their component property names, for
    * indexes associated with the table represented by the given DMO entity
    * name.
    * 
    * @param   entity
    *          A DMO entity name which represents a database table.
    * @param   unique
    *          If <code>true</code>, only unique indexes are returned;  if
    *          <code>false</code>, only non-unique indexes are returned.
    * 
    * @return  A mapping of index names to ordered lists of the names of their
    *          component properties, for the given DMO entity.  The map
    *          represents only unique indexes, or only non-unique indexes,
    *          depending upon the <code>unique</code> parameter.  The map may
    *          be empty, but it will never by <code>null</code>.  The property
    *          names will not be wrapped in any functions.  The map will
    *          iterate in ascending alphabetical order by index name.
    *          
    * @throws  PersistenceException
    *          if there is any error querying index metadata from the JDBC
    *          driver of the associated database.
    */
   private Map<String, List<String>> lookupIndexMapForDMO(String entity, boolean unique)
   throws PersistenceException
   {
      Map<String, List<String>> map = null;
      ArrayList<String> props = null;
      Class<?> dmoClass = DBUtils.dmoClassForEntity(entity);
      String schema = DatabaseManager.getSchemaByClass(dmoClass);
      Class<?> dmoIface = DMOIndex.getDMOInterface(schema, dmoClass);
      Iterator<P2JIndex> iter = DMOIndex.databaseIndexes(schema, database, dmoIface);
      Map<String, String> mappedCols =
            DatabaseManager.getColumnToPropertyMap(database, dmoClass, null, true);
      while (iter.hasNext())
      {
         P2JIndex index = iter.next();
         
         if (index.isUnique())
         {
            if (!unique)
            {
               continue;
            }
         }
         else if (unique)
         {
            continue;
         }
         
         props = new ArrayList<>();
         Iterator<P2JIndexComponent> comps = index.components();
         while (comps.hasNext())
         {
            P2JIndexComponent comp = comps.next();
            String compName = comp.getName();
            
            if (DatabaseManager.PRIMARY_KEY.equals(compName))
            {
               continue;
            }
            
            String propName = mappedCols.get(compName);
            props.add(propName);
         }
         
         if (!props.isEmpty())
         {
            props.trimToSize();
            if (map == null)
            {
               map = new TreeMap<>();
            }
            map.put(index.getName(), props);
         }
      }
      
      if (LOG.isLoggable(Level.FINE))
      {
         LOG.log(Level.FINE,
                 "Indexes " + (unique ? "[unique]" : "[non-unique]") +
                       " for entity '" + entity + "':  " + map
         );
      }
      
      return map;
   }
   
   /**
    * Collect a list of the names of all indexes in which the database
    * column associated with the given DMO property participates.
    * 
    * @param   entity
    *          A DMO entity name which represents a database table.
    * @param   property
    *          A DMO property which represents a database column.
    * 
    * @return  A list of index names in no particular order, or
    *          <code>null</code> if none are found.
    *          
    * @throws  PersistenceException
    *          if there is any error querying index metadata from the JDBC
    *          driver of the associated database.
    */
   private List<String> lookupIndexesForProperty(String entity, String property)
   throws PersistenceException
   {
      ArrayList<String> list = null;
      Class<?> dmoClass = DBUtils.dmoClassForEntity(entity);
      String schema = DatabaseManager.getSchemaByClass(dmoClass);
      Class<?> dmoIface = DMOIndex.getDMOInterface(schema, dmoClass);
      Iterator<P2JIndex> iter = DMOIndex.databaseIndexes(schema, database, dmoIface);
      if (iter.hasNext())
      {
         Map<String, String> mappedCols =
            DatabaseManager.getColumnToPropertyMap(database, dmoClass, null, true);
         
         while (iter.hasNext())
         {
            P2JIndex index = iter.next();
            Iterator<P2JIndexComponent> comps = index.components();
            while (comps.hasNext())
            {
               P2JIndexComponent next = comps.next();
               String colName = next.getName();
               String propName = mappedCols.get(colName);
               
               if (propName != null && propName.equals(property))
               {
                  if (list == null)
                  {
                     list = new ArrayList<>();
                  }
                  list.add(index.getName());
                  
                  break;
               }
            }
         }
      }
      
      if (list != null)
      {
         list.trimToSize();
      }
      
      if (LOG.isLoggable(Level.FINE))
      {
         LOG.log(Level.FINE, "Indexes for '" + entity + ":" + property + "':  " + list);
      }
      
      return list;
   }
   
   /**
    * Given a sort phrase from an index-driven query, find the name of the
    * index represented by the sort phrase.  If a matching index is found, a
    * {@link DMOSorter} instance is created and cached by the sort clause, for
    * later retrieval using the
    * {@link #getSorterForSortPhrase(RecordBuffer, String)} method.
    * 
    * @param   buffer
    *          Record buffer which contains information about the DMO entity
    *          associated with the sort phrase.
    * @param   sort
    *          A sort phrase which includes one or more DMO property
    *          references, each qualified by a DMO alias and followed by a
    *          keyword indicating sort direction.
    * 
    * @return  Name of the index associated with the given sort phrase, or
    *          <code>null</code> if no matching index can be found.
    * 
    * @throws  PersistenceException
    *          if there is any error querying index metadata from the JDBC
    *          driver of the associated database.
    */
   private String lookupIndexForSort(RecordBuffer buffer, String sort)
   throws PersistenceException
   {
      String indexName = null;
      SortKey key = new SortKey(buffer.getEntityName(), sort);
      Class<?> dmoClass = buffer.getDMOImplementationClass();
      String schema = buffer.getSchema();
      Class<?> dmoIface = buffer.getDMOInterface();
      Iterator<P2JIndex> iter = DMOIndex.databaseIndexes(schema, database, dmoIface);
      if (iter.hasNext())
      {
         List<SortCriterion> scList = SortCriterion.parse(sort, buffer, false);
         Map<String, String> mappedCols =
            DatabaseManager.getColumnToPropertyMap(database, dmoClass, null, true);
         
         while (iter.hasNext())
         {
            boolean match = true;
            P2JIndex index = iter.next();
            Iterator<P2JIndexComponent> comps = index.components(true);
            Iterator<SortCriterion> crits = scList.iterator();
            while (true) // loop iterators, exit only by breaks
            {
               boolean moreComps = comps.hasNext();
               boolean moreCrits = crits.hasNext();
               
               if (moreComps != moreCrits)
               {
                  match = false;
                  break;
               }
               
               if (!moreComps || !moreCrits)
               {
                  break;
               }
               
               P2JIndexComponent next = comps.next();
               String compName = next.getName();
               SortCriterion crit = crits.next();
               String critName = crit.getUnqualifiedName();
               String propName = mappedCols.get(compName);
               
               if (!critName.equals(propName))
               {
                  match = false;
                  break;
               }
            }
            
            if (match)
            {
               indexName = index.getName().intern();
               indexBySort.put(key, indexName);
               sorters.put(key, new DMOSorter(buffer.getDatabase(), dmoIface, scList));
               
               break;
            }
         }
      }
      
      if (LOG.isLoggable(Level.FINE))
      {
         LOG.log(Level.FINE,
                 "Index for sort phrase '" + sort + "':  " +
                 (indexName != null ? indexName : "?") + " [" + buffer.getTable() + "]");
      }
      
      if (indexName == null)
      {
         indexBySort.put(key, "");
      }
      
      return indexName;
   }
   
   /**
    * Collect the set of all indexed property names associated with the table
    * represented by the given DMO entity name.
    * <p>
    * Note:  the returned set will <em>not</em> include the reserved primary
    * key property <code>id</code>.
    * 
    * @param   entity
    *          A DMO entity name which represents a database table.
    * @param   unique
    *          If <code>true</code>, only properties associated with unique
    *          indexes are returned;  if <code>false</code>, all indexed
    *          properties are returned.
    * 
    * @return  The names of all indexes for the given entity, or
    *          <code>null</code> if none are found.
    *          
    * @throws  PersistenceException
    *          if there is any error querying index metadata from the JDBC
    *          driver of the associated database.
    */
   private Set<String> lookupIndexedProperties(String entity, boolean unique)
   throws PersistenceException
   {
      Set<String> union = null;
      Class<?> dmoClass = DBUtils.dmoClassForEntity(entity);
      String schema = DatabaseManager.getSchemaByClass(dmoClass);
      Class<?> dmoIface = DMOIndex.getDMOInterface(schema, dmoClass);
      Iterator<P2JIndex> iter = DMOIndex.databaseIndexes(schema, database, dmoIface);
      Map<String, String> mappedCols =
         DatabaseManager.getColumnToPropertyMap(database, dmoClass, null, true);
      
      while (iter.hasNext())
      {
         P2JIndex index = iter.next();
         
         if (unique && !index.isUnique())
         {
            continue;
         }
         
         Iterator<P2JIndexComponent> comps = index.components();
         while (comps.hasNext())
         {
            P2JIndexComponent next = comps.next();
            
            String colName = next.getName();
            if (DatabaseManager.PRIMARY_KEY.equals(colName))
            {
               continue;
            }
         
            String propName = mappedCols.get(colName);
            if (propName == null)
            {
               throw new NullPointerException(
                  "Cannot map column '" + colName + "' to a property in entity '" + entity + "'");
            }
            
            if (union == null)
            {
               union = new HashSet<>();
            }
            
            union.add(propName);
         }
      }
      
      if (LOG.isLoggable(Level.FINE))
      {
         LOG.log(Level.FINE,
                 "Indexed properties " + (unique ? "[unique]" : "[all]") +
                     " for entity '" + entity + "':  " + union );
      }
      
      return union;
   }
   
   /**
    * A hashable key which encapsulates a DMO entity and property name.
    */
   private static class PropertyKey
   {
      /** DMO entity name */
      private final String entity;
      
      /** DMO property name */
      private final String property;
      
      /**
       * Constructor.
       * 
       * @param   entity
       *          DMO entity name.
       * @param   property
       *          DMO property name.
       */
      PropertyKey(String entity, String property)
      {
         this.entity = entity;
         this.property = property;
      }
      
      /**
       * Override the default implementation.  Consistent with {@link
       * #equals(Object)}.
       * 
       * @return  Hash code.
       * 
       * @see     java.lang.Object#hashCode()
       */
      public int hashCode()
      {
         int result = 17;
         result = 37 * result + entity.hashCode();
         result = 37 * result + property.hashCode();
         
         return result;
      }
      
      /**
       * Override the default implementation.
       * 
       * @param   o
       *          Another instance of this class to compare with this one.
       * 
       * @return  <code>true</code> if <code>o</code> is equivalent to this
       *          object, else <code>false</code>.
       *          
       * @see     java.lang.Object#equals(java.lang.Object)
       */
      public boolean equals(Object o)
      {
         if (!(o instanceof PropertyKey))
         {
            return false;
         }
         
         PropertyKey that = (PropertyKey) o;
         
         return (entity.equals(that.entity) &&
                 property.equals(that.property));
      }
   }
   
   /**
    * A hashable key which encapsulates a DMO entity and a sort phrase.
    */
   private static class SortKey
   {
      /** DMO entity name */
      private final String entity;
      
      /** Sort phrase */
      private final String sort;
      
      /**
       * Constructor.
       * 
       * @param   entity
       *          DMO entity name.
       * @param   sort
       *          Sort phrase.
       */
      SortKey(String entity, String sort)
      {
         this.entity = entity;
         this.sort = sort;
      }
      
      /**
       * Override the default implementation.  Consistent with {@link #equals(Object)}.
       * 
       * @return  Hash code.
       * 
       * @see     java.lang.Object#hashCode()
       */
      public int hashCode()
      {
         int result = 17;
         result = 37 * result + entity.hashCode();
         result = 37 * result + sort.hashCode();
         
         return result;
      }
      
      /**
       * Override the default implementation.
       * 
       * @param   o
       *          Another instance of this class to compare with this one.
       * 
       * @return  <code>true</code> if <code>o</code> is equivalent to this
       *          object, else <code>false</code>.
       *          
       * @see     java.lang.Object#equals(java.lang.Object)
       */
      public boolean equals(Object o)
      {
         if (!(o instanceof SortKey))
         {
            return false;
         }
         
         SortKey that = (SortKey) o;
         
         return (entity.equals(that.entity) && sort.equals(that.sort));
      }
      
      /**
       * Provide a string representation of this object's state.
       * 
       * @return  String representation of this object.
       * 
       * @see java.lang.Object#toString()
       */
      public String toString()
      {
         return "SortKey [" + entity + ":" + sort + "]";
      }
   }
   
   /**
    * A hashable key which encapsulates a DMO entity and a boolean indicating
    * unique vs. non-unique indexes.
    */
   private static class EntityIndexKey
   {
      /** DMO entity name */
      private final String entity;
      
      /** Sort phrase */
      private final boolean unique;
      
      /**
       * Constructor.
       * 
       * @param   entity
       *          DMO entity name.
       * @param   unique
       *          Uniqueness flag.
       */
      EntityIndexKey(String entity, boolean unique)
      {
         this.entity = entity;
         this.unique = unique;
      }
      
      /**
       * Override the default implementation.  Consistent with {@link
       * #equals(Object)}.
       * 
       * @return  Hash code.
       * 
       * @see     java.lang.Object#hashCode()
       */
      public int hashCode()
      {
         int result = 17;
         result = 37 * result + entity.hashCode();
         result = 37 * result + (unique ? 1 : 0);
         
         return result;
      }
      
      /**
       * Override the default implementation.
       * 
       * @param   o
       *          Another instance of this class to compare with this one.
       * 
       * @return  <code>true</code> if <code>o</code> is equivalent to this
       *          object, else <code>false</code>.
       *          
       * @see     java.lang.Object#equals(java.lang.Object)
       */
      public boolean equals(Object o)
      {
         if (!(o instanceof EntityIndexKey))
         {
            return false;
         }
         
         EntityIndexKey that = (EntityIndexKey) o;
         
         return (entity.equals(that.entity) && unique == that.unique);
      }
   }
}