/*
** Module   : LocalTempTableHelper.java
** Abstract : Helper class which generates local temp table DDL
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 CA  20080424  @38109  Created initial version. Generates DDL for local temp tables with
**                           local indexes.
** 002 CA  20080430  @38157  Changes due to original incorrect index creation for the "dirty"
**                           database.
** 003 ECF 20080512  @38503  Changed getIndexes(). Use new signature for
**                           DMOIndex.databaseIndexes().
** 004 ECF 20080611  @38696  Modified getComputedCols() to handle new PersistenceException.
**                           DMOIndex now throws an exception if index metadata cannot be queried.
** 005 ECF 20140822          Added implementation of isNotNullAllowed. Replaced Apache commons
**                           logging with J2SE logging.
** 006 EVL 20160223          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.util.*;
import java.util.List;
import java.util.Map;
import java.util.logging.*;
import org.hibernate.mapping.*;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.util.*;

/**
 * Helper class which generates and caches SQL phrases to create and drop
 * temporary tables, and to create indexes in those tables.
 * <p>
 * This class is responsible for local temporary tables support: the created
 * tables and indices will be local to the session which created them.
 */
final class LocalTempTableHelper
extends TempTableHelper
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(LocalTempTableHelper.class);
   
   /**
    * Constructor.
    *
    * @param   dialect
    *          Hibernate database dialect helper.
    * @param   dmoIface
    *          DMO interface which provides the public API to the temp
    *          table.
    */
   LocalTempTableHelper(P2JDialect dialect, Class<?> dmoIface)
   {
      super(dialect, dmoIface, DatabaseManager.TEMP_TABLE_DB);
   }

   /**
    * This method will return the needed create table string, which is the 
    * "create local temporary table" string.
    * 
    * @return  The create table string.
    */
   protected String getCreateTableStatement()
   {
      return "create local temporary table ";
   }

   /**
    * This method will return an additional suffix to make the index unique
    * across sessions.
    * 
    * @return The additional suffix.
    */
   protected String getAdditionalIndexSuffix()
   {
      StringBuilder buf = new StringBuilder();
      buf.append("__${");
      buf.append(UNIQUE_SUFFIX_KEY);
      buf.append("}");
      
      return buf.toString();
   }

   /**
    * Populate the list of dialect-specific SQL statements used to create
    * any indexes required by the temp table(s).
    * <p>
    * If there are no indices for the required table, a statement to create an 
    * index for the {@link TemporaryBuffer#MULTIPLEX_FIELD_NAME multiplex 
    * column} will be created.
    * 
    * @param   dialect
    *          Hibernate database dialect helper.
    * @param   dmoIface
    *          DMO interface which provides the public API to the temp
    *          table.
    * @param   primary
    *          Hibernate database table configuration for primary temp
    *          table.
    * @param   secondaries
    *          Optional list of Hibernate database table configurations
    *          for secondary temp tables.  These are the tables which
    *          support temp table fields with extents, if any.  May be
    *          <code>null</code> if no such tables exist.
    * @param   iter
    *          The iterator for the collection which contains the needed 
    *          indices. 
    *          
    * @return  List of SQL statements.  Will be empty if no indices are
    *          defined.
    */
   protected List<String> generateSQLCreateIndexList(P2JDialect dialect,
                                                     Class<?> dmoIface, 
                                                     Table primary, 
                                                     List<Table> secondaries,
                                                     Iterator<P2JIndex> iter)
   {
      List<String> stmts = new ArrayList<String>();
      
      // Handle indices for primary table first.
      if (!iter.hasNext())
      {
         // If no other indices exist, create an index on the multiplex ID
         // of the primary table.
         String tableName = primary.getName();
         StringBuilder buf = new StringBuilder("idx_mpid__");
         buf.append(tableName);
         String idxName = buf.toString();
         P2JIndex idx = new P2JIndex(tableName, idxName, false);
         idx.addComponent(TemporaryBuffer.MULTIPLEX_FIELD_NAME);
         stmts.add(generateSQLCreateIndex(dialect, idx, primary, false, true));
         
         ((ArrayList<String>) stmts).trimToSize();
      }
      else
      {
         stmts = super.generateSQLCreateIndexList(dialect, dmoIface, primary, secondaries, iter);
      }
      
      return stmts;
   }
   
   /**
    * Populate the list of dialect-specific SQL statements used to drop
    * any indexes required by the temp table(s).
    * <p>
    * If there are no indexes for the required table, a statement to drop the 
    * index associated with the {@link TemporaryBuffer#MULTIPLEX_FIELD_NAME 
    * multiplex column} will be created.
    * 
    * @param   dialect
    *          Hibernate database dialect helper.
    * @param   dmoIface
    *          DMO interface which provides the public API to the temp
    *          table.
    * @param   primary
    *          Hibernate database table configuration for primary temp
    *          table.
    * @param   secondaries
    *          Optional list of Hibernate database table configurations
    *          for secondary temp tables.  These are the tables which
    *          support temp table fields with extents, if any.  May be
    *          <code>null</code> if no such tables exist.
    * @param   iter
    *          The iterator for the collection which contains the needed 
    *          indexes.
    *          
    * @return  List of SQL statements.  Will be empty if no indices are
    *          defined.
    */
   protected List<String> generateSQLDropIndexList(P2JDialect dialect,
                                                   Class<?> dmoIface, 
                                                   Table primary, 
                                                   List<Table> secondaries,
                                                   Iterator<P2JIndex> iter)
   {
      List<String> stmts = new ArrayList<String>();
      
      if (!iter.hasNext())
      {
         // If no other indices exist, drop the index on the multiplex ID
         // of the primary table.
         String tableName = primary.getName();
         StringBuilder buf = new StringBuilder("idx_mpid__");
         buf.append(tableName);
         String idxName = buf.toString();
         stmts.add(generateSQLDropIndex(dialect, idxName));

         ((ArrayList<String>) stmts).trimToSize();
      }
      else 
      {
         stmts = super.generateSQLDropIndexList(dialect, dmoIface, primary, secondaries, iter);
      }
      
      return stmts;
   }

   /**
    * This method which will return the computed columns for the required
    * DMO, as retrieved from the dmo_index.xml file.
    * 
    * @param   schema
    *          The schema used for this DMO.
    * @param   dmoIface
    *          The DMO for which to retrieve the computed columns.
    * 
    * @return  A map with the computed columns associated with this DMO.
    */
   protected Map<String, P2JIndexComponent> getComputedCols(String schema, Class<?> dmoIface)
   {
      try
      {
         return DMOIndex.getIndexedCharacterColumns(schema, database, dmoIface);
      }
      catch (PersistenceException exc)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, "Error getting computed columns", exc);
         }
         
         return Collections.emptyMap();
      }
   }
   
   /**
    * This method will return the indexes associated with a certain DMO, as
    * retrieved from the dmo_index.xml file.
    * 
    * @param   schema
    *          The schema used for this DMO.
    * @param   dmoIface
    *          The DMO for which to retrieve the indexes.
    * 
    * @return  An iterator for the indexes associated with this DMO.
    */
   protected Iterator<P2JIndex> getIndexes(String schema, Class<?> dmoIface)
   {
      try
      {
         return DMOIndex.databaseIndexes(schema, database, dmoIface);
      }
      catch (PersistenceException exc)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, "Error querying index data for " + dmoIface.getName(), exc);
         }
         
         throw new StopConditionException(exc);
      }
   }
   
   /**
    * Indicate whether non-nullable columns are permitted.
    * 
    * @return  <code>true</code> if non-nullable columns are permitted, else <code>false</code>.
    */
   protected boolean isNotNullAllowed()
   {
      return true;
   }
}