/*
** Module   : PreselectQuery.java
** Abstract : A query which enables stateful, cursored navigation of a server side result set
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 ECF 20051115   @23396 Created initial version, based on prototype
**                           CursoredQuery. Backing support for queries
**                           which must support the Progress, row-level
**                           access semantic against a preselected result
**                           set.
** 002 ECF 20051129   @23575 Added lenient-off-end mode. When in this
**                           mode, QueryOffEndExceptions thrown by the
**                           underlying buffer during next and previous
**                           operations are suppressed.
** 003 ECF 20051216   @23750 Added scrolling support. Modified retrieval
**                           logic to allow for repositioning. Added many
**                           new methods.
** 004 ECF 20060105   @23831 Inserted abstraction layer between query and
**                           its scrollable results. Added the Results
**                           interface and its default implementation,
**                           SimpleResults. This allows subclasses to
**                           provide custom implementations by overriding
**                           the createResults method.
** 005 ECF 20060202   @24358 Added temp table support. Use buffer's
**                           multiplex ID as a query substitution
**                           parameter when executing HQL. Parameters are
**                           now managed separately, by component, to
**                           enable this processing.
** 006 ECF 20060208   @24721 Added support for client-side where clause
**                           processing. Results and SimpleResults are now
**                           top level classes. Modified all single-table
**                           constructor variants to accept an instance of
**                           WhereExpression. Modified Component to handle
**                           argument resolution more correctly. Results
**                           can now be provided by an external entity
**                           (CompoundQuery typically).
** 007 ECF 20060302   @25019 Add support for foreign key joins. Added new
**                           constructor and addComponent method variants
**                           which mirror existing versions, but accept an
**                           additional, inverse DMO parameter. Updated
**                           Component class, HQL generation, and query
**                           execution to integrate natural joins.
** 008 ECF 20060422   @25657 Added setIterationType method. Sets retrieval
**                           mode (first, last, next, etc.) for the most
**                           recently added query component.
** 009 ECF 20060423   @25661 Allow multiple client-side where clause
**                           expressions to be registered. This is needed
**                           because each table in a multi-table query can
**                           contain its own client-side where component.
** 010 ECF 20060424   @25670 Fix to several constructor variants. Swapped
**                           positions of sort and inverse parameters to
**                           match the natural order in which these
**                           parameters are emitted in converted code.
** 011 ECF 20060424   @25695 Added new constructor variants. Each existing
**                           variant now has a counterpart which does not
**                           expect a WhereExpression parameter.
** 012 ECF 20060424   @25699 Roll back previous change. Fixed ordering of
**                           some parameters in internal constructor
**                           calls.
** 013 ECF 20060425   @25718 Added setLenientOffEnd method. Required for
**                           conversion of certain OPEN QUERY constructs.
** 014 ECF 20060501   @25915 Added isPreselect method. Indicates query's
**                           preselection behavior.
** 015 ECF 20060531   @26850 Replaced EndConditionException with
**                           QueryOffEndException. The latter is a more
**                           specific version of the former, used to
**                           indicate a query has run off the end of its
**                           result set.
** 016 ECF 20060602   @26944 Optimization. HQL queries now will always
**                           attempt to load the full record, rather than
**                           just the primary key ID. This avoids the
**                           overhead of a separate load step.
** 017 ECF 20060616   @27412 Fix to regression caused by #016 (@26944).
**                           FilteredResults needs to store DMO instance
**                           within an array.
** 018 ECF 20060705   @27828 Fixed ClassCastException. Made fetch method
**                           tolerant of a primary key ID result (instead
**                           of DMO) in a result set.
** 019 ECF 20060713   @28051 Integrated HQLPreprocessor. An instance of
**                           this class is created for each component
**                           added to the query. This affects how the
**                           overall HQL query is composed.
** 020 ECF 20060719   @28116 Rolled back #016 (@26944). With the addition
**                           of the change broker architecture, it is
**                           actually possible to avoid loading the record
**                           at all in limited cases. In these cases, only
**                           a Hibernate proxy is created, which is far
**                           more efficient than retrieving the entire
**                           record up front. Additionally, #016 was
**                           problematic for locking, because the record
**                           could be changed by another user after it was
**                           fetched, but before the lock was acquired.
** 021 ECF 20060722   @28135 Changes to make inheritance of this class
**                           more flexible. In particular, this set of
**                           changes accommodates the AdaptiveQuery
**                           subclass.
** 022 ECF 20060725   @28188 Fixed regressions and implemented
**                           Finalizable. The latter is necessary to
**                           support retry, and to prevent a session from
**                           closing prematurely and discarding the
**                           query's results.
** 023 ECF 20060730   @28270 Client-side where clause processing changes.
**                           Moved FilteredResults into a top-level class.
** 024 ECF 20060801   @28352 Refactored execute() method and order by
**                           clause preparation code. Query substitution
**                           parameter preparation was moved into its own
**                           polymorphic method. Integrated SortCriterion
**                           class to replace code that was duplicative
**                           with that class.
** 025 ECF 20060803   @28396 Further refactoring. Added getRecordBuffers()
**                           method. Made results private again.
** 026 ECF 20060807   @28462 Fixed assembleOrderByClauseNoPreamble(). DMO
**                           alias needed to be trimmed of leading space.
** 027 ECF 20060829   @28954 Fixed join defect in constructor. We were
**                           creating a join and passing it to the wrong
**                           version of addComponent().
** 028 ECF 20060830   @28991 Additional join fix. The fix has been moved
**                           from a constructor to the most specific
**                           variant of addComponent().
** 029 ECF 20060831   @29087 Added support for sorting by an indexed
**                           property. Specifically, support sorting by
**                           a property of a DMO's associated list of
**                           composite properties.
** 030 ECF 20060901   @29139 Fixed NPE. Methods which accept a list of
**                           sort criteria must handle null list case.
** 031 ECF 20060911   @29461 Added debug code. Report DMOs found by the
**                           query if trace logging is enabled.
** 032 ECF 20060925   @29963 Check for a pending error before retrieving a
**                           record. Retrieval is not attempted if an
**                           error is pending in the ErrorManager.
** 033 ECF 20061003   @30154 Backed out #032 (@29963). This logic is only
**                           necessary for FindQuery and AdaptiveFind.
** 034 ECF 20061010   @30310 Minor change to initialization logic. Joins
**                           across different Persistence objects, rather
**                           than across databases, are now disallowed.
** 035 ECF 20061019   @30528 Fixed reposition support to better handle
**                           varying subclass behavior. This class now
**                           implements the SessionListener interface.
**                           Upon receiving a sessionClosing notification,
**                           a PreselectQuery will cache all of its
**                           results in a SimpleResults object and replace
**                           its existing results with that cache.
** 036 ECF 20061023   @30607 Moved errorIfNull logic to AbstractQuery base
**                           class. Method setErrorIfNull() is now
**                           required by the P2JQuery interface and is
**                           implemented in AbstractQuery instead of here.
** 037 ECF 20061027   @30752 Replaced Object with DataModelObject for DMO
**                           arguments to methods/constructors. Required
**                           for compile-time type safety.
** 038 ECF 20061102   @30897 Fixed issue querying row count. The cached
**                           row count must be reset to its default value
**                           when a new result set is retrieved. This
**                           forces the size() method to recalculate the
**                           proper row count.
** 039 ECF 20061102   @30959 Added support for reposition notification and
**                           auto-fetch features. Cleaned up reposition
**                           logic so that fetches after a reposition are
**                           (hopefully) correct now.
** 040 ECF 20061115   @31209 Changed call to set current buffer record.
**                           RecordBuffer API has changed to permit proper
**                           management of placeholder snapshot record.
**                           Added isResetSnapshot() method to allow
**                           subclasses to override this behavior.
** 041 ECF 20061120   @31326 Modified Component inner class. Expose
**                           restriction properties determined during HQL
**                           where clause preprocessing.
** 042 ECF 20061204   @31576 Integrated RepositionCache. This replaced a
**                           simple hash map approach which failed in
**                           cases where a lookup used an ID array which
**                           had fewer IDs than joined tables in a query.
**                           Also emulate Progress reposition quirk which
**                           repositions to last record in result set when
**                           no match is found for the target recid.
** 043 ECF 20061206   @31587 Integrated changes to AbstractQuery's
**                           afterReposition() signature. Failed attempts
**                           to reposition by ID cause afterReposition()
**                           to be invoked in notify-only mode.
** 044 ECF 20070225   @32227 Plug results resource leak. Clean up results
**                           resources when enclosing block ends. This was
**                           causing both prepared statements and result
**                           sets to remain unclosed in the backing JDBC
**                           connection.
** 045 ECF 20070225   @32232 Fixed regression caused by #044 (@32227).
**                           Finalizable which cleans up results resources
**                           must not be hard-coded to nearest enclosing
**                           scope where results are created. Subclasses
**                           may need to register this object to the
**                           global scope instead.
** 046 ECF 20070228   @32247 Optimized no-lock and temp table fetches. In
**                           these cases, it is not necessary to fetch the
**                           record in two steps. The no-lock case makes
**                           no guarantee that the record in memory has
**                           not been updated by other sessions. The temp
**                           table case guarantees that no other session
**                           can make an update to the record in memory,
**                           because a temp table record is inherently
**                           scoped to a session.
** 047 ECF 20070308   @32334 Fixed behavior for first/last. These should
**                           never raise an error condition, only a query
**                           off end condition.
** 048 ECF 20070328   @32648 Changed to support HQLPreprocessor caching.
**                           Use factory method to retrieve an instance of
**                           the preprocessor rather than invoking the
**                           constructor directly.
** 049 ECF 20070412   @32984 Accommodate method name change in
**                           RecordBuffer. setCurrentRecord --> setRecord.
** 050 ECF 20070415   @33002 Fixed locking problem. DMOs fetched "whole"
**                           with lock type NONE must be re-loaded with a
**                           lock if called for by the query component or
**                           lock type override in the retrieval method.
**                           In certain circumstances, these were being
**                           left without a lock.
** 051 ECF 20070416   @33026 Integrated user interrupt handling.
** 052 ECF 20070502   @33371 Changed to support HQLPreprocessor changes.
**                           The signature of that class' primary factory
**                           has changed.
** 053 ECF 20070504   @33409 Fixed load() implementation. We now retry
**                           reposition if it fails, assuming current
**                           result set did not contain target records.
** 054 ECF 20070508   @33459 Fixed sessionClosing() and session listener
**                           registration. Was not properly scoping
**                           registration for scrolling queries.
** 055 ECF 20070511   @33470 Fixed resetResults(). Additional state needed
**                           to be reset to default values.
** 056 ECF 20070517   @33670 Fixed abend in sessionClosing(). Check that
**                           all necessary buffers are still in scope
**                           before attempting to cache preselected
**                           results in a SimpleResults object. This
**                           prevents a fatal error when trying to
**                           retrieve results for a temp table which has
**                           been dropped by the time we get the session
**                           closing notification.
** 057 ECF 20070518   @33683 Changed access for getRecordBuffers(). Made
**                           public (was protected), since this is now a
**                           method of the Joinable interface.
** 058 ECF 20070706   @34402 Fixed selection of scope for SessionListener
**                           registration. Implemented new SessionListener
**                           method deregisteredSessionListener(). Also
**                           replaced StringBuffer with StringBuilder.
** 059 ECF 20070709   @34422 Fixed sessionClosing(). Was unsafe for
**                           certain non-scrolling results which do not
**                           support reset.
** 060 ECF 20070711   @34448 Extracted ScrollingResults inner class into a
**                           separate, top-level class.
** 061 ECF 20070716   @34510 Implement new cleanup() method as required by
**                           P2JQuery interface. Check whether query is
**                           standalone before registering a cleaner with
**                           the TM. This prevents us from cleaning up too
**                           aggressively in case we are wrapped by a
**                           QueryWrapper or contained in a CompoundQuery
**                           that is so wrapped.
** 062 ECF 20070720   @34611 Fixed fetch(). Always reload a DMO, even if
**                           it is in a temp table or has no lock. This is
**                           necessary to handle the case of fetching a
**                           cached DMO which may have been detached from
**                           its session previously. For a persistent DMO,
**                           the additional overhead should be small,
**                           since Hibernate only needs to go to the first
**                           level cache to find the object.
** 063 ECF 20070720   @34621 Implemented a better fix for fetch() than
**                           #062 (@34611). If DMO is a temp table, simply
**                           try to reassociate it with the Hibernate
**                           session, rather than going all the way back
**                           to the database (if DMO was detached). Can't
**                           do this for permanent tables, though.
** 064 ECF 20070821   @34907 Added support for temporary silent error
**                           mode. This allows query substitution parms to
**                           be processed safely during initialization of
**                           a query.
** 065 ECF 20070827   @34996 Optimized certain temp table queries. Only
**                           augment where clause with a temp buffer's
**                           multiplex ID when necessary. Use multiplex ID
**                           as a constant in where clause rather than as
**                           a substitution parameter.
** 066 ECF 20070927   @35276 Fixed reposition logic. Any reposition must
**                           first release all buffers. Changes to methods
**                           to get current row and report whether query
**                           is off-end.
** 067 ECF 20071030   @35634 Fixed fetch() to prevent loading stale temp
**                           table records. It is possible for a record to
**                           be evicted from the session cache while still
**                           held in a result set, such that the result
**                           set will return a stale record that is not a
**                           detached instance, even if the session is
**                           flushed before the fetch. We now force a
**                           reload of fetched temp table records to avoid
**                           this problem.
** 068 ECF 20071130   @36122 Removed isRefreshSnapshot() method. The
**                           RecordBuffer placeholder snapshot is no
**                           longer cleared by any query type when the
**                           query does not find a record to place in the
**                           buffer. Integrated generics.
** 069 ECF 20071206   @36267 Fixed off-end tracking. Replaced boolean flag
**                           with OffEnd enum to discern between off-end
**                           directions (i.e., off back vs off front of
**                           result set). Current off-end status and
**                           active sort index are reported to the record
**                           buffer when fetching a record.
** 070 ECF 20071218   @36463 Fixed order by clause generation. Was failing
**                           for sort criteria which included references
**                           to specific extents of indexed properties.
** 071 ECF 20080120   @37009 Removed restriction of dealing only with IDs.
**                           This query can now handle both Persistables
**                           and primary key IDs when retrieving data.
**                           This is necessary when the full record is
**                           needed immediately, not just an ID, such as
**                           in presort cases. Also added open() method to
**                           set proper semantics for a converted OPEN
**                           QUERY statement, and optionally to reposition
**                           to the first result.
** 072 ECF 20080222   @37144 Fixed NPE in sessionClosing().
** 073 ECF 20080310   @37481 Made changes necessary to support embedded,
**                           H2 database.
** 074 SVL 20080418   @38036 Added support for the inverseSorting flag.
** 075 SVL 20080421   @38080 Added support for legacy key join helpers.
** 076 SVL 20080425   @38119 lastRepo assignment has been fixed into
**                           reposition(int).
** 077 ECF 20080502   @23396 Backed out #076 (@38119). The reposition(int)
**                           implementation was generally correct and this
**                           change caused regressions. Fixed open() to
**                           execute the query immediately instead of
**                           doing this lazily. We now reposition on open
**                           only if the query is being browsed. Removed
**                           obsolete method isRepositionOnOpen().
** 078 SVL 20080508   @38238 Fixed incorrect record values after
**                           first-next, forward/backward(0) and
**                           last-prev execution sequences.
** 079 ECF 20080508   @38243 Fixed regression in reposition(int) which did
**                           not handle all non-positive rows correctly.
**                           Refactored open() method to use parent class.
**                           Override base lock type for temporary tables.
**                           This allows some downstream optimizations.
** 080 ECF 20080509   @38252 Fixed regression in open(). Removed check for
**                           standalone flag.
** 081 SVL 20080526   @38311 Preselect locking implemented.
** 082 SVL 20080604   @38534 Modified according to the new P2JQuery
**                           interface. previous() fixed.
** 083 ECF 20080604   @38608 Added placeholder for dirty database check.
** 084 SVL 20080611   @38685 Behavior of previous() fixed for the case 
**                           when the off-end is FRONT.
** 085 CA  20080618   @38869 When the JOINed foreign buffer does not refer
**                           any record, the query must return no results
**                           as it can not be executed.
** 086 CA  20080619   @38871 If executed within an OPEN QUERY statement, 
**                           the query must be removed from the session's
**                           listeners only when the enclosed external
**                           scope is exit. This fixes the case when the
**                           OPEN QUERY is executed inside a procedure.
** 087 ECF 20080709   @39086 Created cacheResults() method. Abstracted
**                           from sessionClosing() and made protected so
**                           subclasses could access this functionality.
** 088 ECF 20080710   @39090 Modified sessionClosing() and cacheResults().
**                           The latter no longer throws
**                           IllegalStateException.
** 089 ECF 20080714   @39127 Fixed NPE in debug logging.
** 090 ECF 20080723   @39173 Removed retry handling. This is now the
**                           responsibility of subclasses.
** 091 CA  20080729   @39218 Added back the retry handling. Now the retry
**                           will rely on the reposition mechanism to 
**                           retrieve the same record, but only when the
**                           query is associated with a FOR block.
** 092 SVL 20080730   @39237 Component.multiplexID field was removed.
**                           RecordBuffer's multiplex ID is used instead.
** 093 ECF 20080730   @39325 Added rowid support. Refactored reposition*()
**                           methods.
** 094 SVL 20080808   @39348 Fixed cacheResults(). Did not handle case
**                           when query has just executed but has not yet
**                           retrieved any results.
** 095 ECF 20080822   @39556 Fix for join validation. Abstracted check
**                           into separate method, verifyJoins(). Disable
**                           join buffer check for server-based joins.
**                           Other changes in support of global DMO change
**                           events.
** 096 CA  20080819   @39463 Support API change in ScrollingResults.
** 097 CA  20080828   @39610 Fix NPE caused by repositionByID(rowid, ...).
** 098 ECF 20081009   @40081 Added support for inlining query substitution
**                           parameters. Integrated HQLPreprocessor API
**                           changes and the ParameterIndices class.
** 099 ECF 20081030   @40308 Fixed regression introduced by #098 (@40081).
**                           Refactored Component inner class and code in
**                           top-level class which uses it.
** 100 ECF 20081105   @40311 Implemented registerRecordChangeListeners().
**                           Required by the Joinable interface.
** 101 ECF 20081106   @40324 Added _isOffEnd(). Required by P2JQuery
**                           interface change.
** 102 ECF 20081107   @40405 Moved multiplex ID position in order by
**                           clause. It used to be the first element, but
**                           that sometimes resulted in a less efficient
**                           index being selected for a query plan. Now it
**                           is either the last element, or second to last
**                           (if the primary key is part of the clause).
** 103 CA  20081210   @40839 Do not cleanup if the query was explicitly 
**                           closed.
** 104 CA  20081211   @40859 Fixed regression introduced by H103: on 
**                           sessionClosing event, if the query is closed,
**                           return true.
** 105 ECF 20081215   @40914 Removed flushBuffers() method. This work is
**                           now done when a component is added to the
**                           query. The previous invocation, during record
**                           retrieval, was too late. A validation error
**                           ignores silent error mode, since the flushing
**                           of the previous record in the buffer is not
**                           masked by silent error mode.
** 106 ECF 20090202   @41248 Rolled back #081 (@38311). Prelock mechanism
**                           has been disabled.
** 107 ECF 20090313   @41600 Changed createResults(). Added parameter to
**                           FilteredResults constructor.
** 108 SVL 20090316   @41610 assembleOrderByClauseNoPreamble() refactored,
**                           sort variable is protected instead of private 
**                           now.
** 109 GES 20090424   @41941 Import change.
** 110 ECF 20090512   @42153 Modified addComponent(). Now invokes flushAll()
**                           instead of flush() on backing buffer. This
**                           ensures all like buffers are allowed to flush
**                           their transient records properly.
** 111 ECF 20090603   @42599 Fixed fetch(). A dirty record in the backing
**                           buffer must not be replaced by the primary
**                           database's version of that record.
** 112 ECF 20090604   @42606 Added support for Progress isolation leak quirk.
**                           Modified execute() method to publish uncommitted
**                           changes with dirty share manager if necessary.
** 113 ECF 20090609   @42631 Fixed execute(). We now evict, if possible, those
**                           DMOs read into the Hibernate Session, whose
**                           uncommitted changes are leaked to other sessions
**                           early.
** 114 ECF 20090619   @42772 Added executeScroll(). Executes default variant
**                           of Persistence.scroll(String, Object[], Type[]).
**                           Children classes which need to invoke different
**                           variants must override this method.
** 115 ECF 20090630   @42985 Rewrote reposition logic. Moved from integer to
**                           recid type in reposition methods.
** 116 ECF 20090702   @43035 Fixed sessionClosing() and regressions in query
**                           (in the Progress OPEN QUERY sense) behavior.
**                           Added isNativeScrolling() to indicate that a base
**                           PreselectQuery instance used for a converted 4GL
**                           query behaves as if that query were defined as
**                           scrolling. Added error handling in caching
**                           results during a sessionClosing event.
** 117 ECF 20090703   @43051 Reimplemented load() method. This method now sets
**                           the backing buffer to unknown mode and throws
**                           MissingRecordException if a target record cannot
**                           be loaded. Also fixed cleanup(); the query is now
**                           flagged as closed once it runs.
** 118 ECF 20090707   @43101 Fixed defect related to processing reposition
**                           notifications. Make sure we don't release the
**                           buffers' current records and change other state
**                           when processing nested reposition calls during
**                           reposition notifications.
** 119 ECF 20090708   @43107 Partial rollback of #118 (@43101). Some state
**                           must be changed when processing nested reposition
**                           calls during reposition notifications.
** 120 ECF 20090711   @43192 Use new ScrollingResults c'tor.
** 121 ECF 20090716   @43220 Refactored for lazy record buffer initialization.
**                           Initialize backing buffers when their scopes have
**                           been opened. Otherwise allow initialization to be
**                           deferred.
** 122 SVL 20090723   @43344 Added setNonScrolling() function.
** 123 ECF 20090724   @43419 Made getPersistence() protected.
** 124 ECF 20090729   @43448 Fixed executeQuery(). Surrounded creation of
**                           ScrollingResults with push/pop calls for implicit
**                           transaction.
** 125 CA  20090731   @43468 Register the query for off-end listener 
**                           registration after block initialization (done
**                           only for block types FOR and FOR EACH). Added
**                           getOffEndListeners, which returns a list of all
**                           off-end listeners for all used buffers.
** 126 ECF 20090810   @43588 Changed to accommodate SessionListener API
**                           change. sessionClosing() method was replaced with
**                           sessionEvent().
** 127 ECF 20090811   @43600 Fixed resource cleanup. Refactored registration
**                           of a cleaner object with the TM, so that
**                           subclasses could use it. Ensured cleaner is only
**                           registered once.
** 128 SVL 20091030   @44296 Fixed forward(), backward(), currentRow() and next() functions.
** 129 OM  20130829          Added change detection to buffer on current() methods.
**                           Updated calls to match the new Persistence.load() signatures.
** 130 ECF 20131028          Import change.
** 131 CA  20131013          Added no-op deleted() method, required by the changes in Finalizable
**                           interface.
** 132 SVL 20140210          Use HQLExpression instead of HQL string.
** 133 SVL 20140604          Added support for INDEX-INFORMATION.
** 134 ECF 20140814          Replaced RecordBuffer.flushAll with flush in addComponent method.
** 135 SVL 20140909          Fixed null case for peekNext/peekPrevious. Changed some access
**                           modifiers. Added deleteResultListEntry.
** 136 SVL 20141209          Update row count after deleteResultListEntry.
** 137 SVL 20150212          Fixed last().
** 138 SVL 20150331          Added deleteResultListEntry(boolean).
** 139 OM  20150421          Changed the hqlPreprocessor of each components to expires on when
**                           resetArgs() is called based on nullArguments fingerprint.
**                           Replaced Apache logging with J2SE logging.
** 140 ECF 20150801          Implemented new Joinable methods necessary for compound query
**                           optimization. Refactored to remove inner class Component, which is
**                           now the top level class QueryComponent.
** 141 SVL 20150821          isScrolling was made public.
** 142 ECF 20150906          Reimplemented silent error mode processing of query substitution
**                           parameters. Fixed NPE related to outer joins.
** 143 ECF 20151214          Prevent releasing the current buffer record(s) under certain
**                           circumstances, when no record is found by the query.
** 144 EVL 20160217          Clean up comments from symbols invalid for Solaris to compile.
** 145 SVL 20160314          Added support for parameterFilter.
** 146 ECF 20160320          Force full records to be queried on the first pass.
** 147 ECF 20160512          Deregister as SessionListener during cleanup instead of letting TM
**                           call cleaner. Fixed coreFetch issue with null record.
** 148 ECF 20160610          Reverted #146.
** 149 SVL 20160608          The query can handle null IDs (for scrolling compound queries
**                           with OUTER join). Reposition validation moved away.
** 150 ECF 20160614          Added getEntities to collecte entity names associated with this
**                           query; modified executeScroll to pass these to Persistence.scroll.
** 151 SVL 20160721          Proper reset on query reopen.
** 152 OM  20160720          Added support for loading template records.
** 153 ECF 20160725          Fixed buffer lookup during HQL preprocessing.
** 154 ECF 20160823          Improved coreFetch to better detect missing records and avoid
**                           deadlock.
** 155 ECF 20160901          DirtyShareContext.isDirtyDelete signature change.
** 156 GES 20160804          Switched from WhereExpression to lambda usage.
**         20160919          All initialization is now done with a default constructor followed
**                           by a call to initialize(). This enables query instance members in
**                           converted code to be local variables that are effectively final.
**                           This is needed to support lambdas without moving queries to be
**                           instance members. Queries as instance members breaks recursion
**                           use cases.
** 157 SVL 20161007          Fix for the case when REPOSITION-TO-ROWID takes more ids than the
**                           actual query table count.
** 158 ECF 20161014          Fixed implementation of current() method in the event there is no
**                           current record in the buffer to reload, and this is not considered
**                           an error in the context of the query (e.g., query represents an
**                           outer join and record can be null).
** 159 OM  20170118          Fixed javadoc (Solaris compatibility).
**     OM  20170209          Added return value for repositioning methods.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.util.*;
import java.util.function.*;
import java.util.logging.*;
import org.hibernate.*;
import org.hibernate.type.*;
import com.goldencode.p2j.persist.dirty.*;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.LogHelper;
import com.goldencode.p2j.util.ErrorManager;
import com.goldencode.util.*;

/**
 * Query which issues an HQL query to the database server and uses a
 * scrollable database cursor to navigate over the set of results.  This
 * query is used to back converted Progress queries which either explicitly
 * or implicitly cause a result list to be preselected.  Both single- and
 * multi-table queries are supported.  Because this class implements the
 * <code>Joinable</code> interface, instances may be added as components of
 * a {@link CompoundQuery}.
 * <p>
 * The query is lazily executed the first time a record retrieval/navigation
 * method is invoked.  Only the database identifiers (i.e., primary keys) of
 * the records are retrieved by the initial query.  On each subsequent fetch,
 * a row in this result list is retrieved, where each column contains the
 * primary key for the record associated with the corresponding query
 * component.  The record for each primary key is retrieved and stored in its
 * associated buffer.
 * <p>
 * At this time, outer joins are not supported.  Furthermore, components may
 * only specify NEXT as their iteration type at this time.  FIRST and LAST
 * currently are not supported.
 * <p>
 * When used to represent a Progress query (created with DEFINE or OPEN
 * QUERY), lenient-off-end must be enabled at construction.  This mode
 * suppresses end condition exceptions for <code>next</code> and
 * <code>previous</code> commands which run off the end of the query's
 * results.
 * <p>
 * This query may be integrated with a client-side where clause expression,
 * by passing a non-<code>null</code> <code>Supplier&lt;logical&gt;</code> argument
 * to one of the single-table constructor variants.  Note that additional
 * components may not be added to an instance so created, since this type of
 * query must act against no more than one record buffer.  If a multi-table
 * preselect query is called for, <code>CompoundQuery</code> must be used,
 * though its components may be instances of this class.
 */
public class PreselectQuery
extends AbstractQuery
implements Joinable,
           QueryConstants,
           Finalizable,
           SessionListener
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(PreselectQuery.class.getName());
   
   /** Has query been repositioned between result rows? */
   protected boolean betweenRows = true;
   
   /** Has the query run off either end of its result set? */
   protected OffEnd offEnd = OffEnd.NONE;
   
   /** HQL order by clause used across all query components */
   protected String sort;
   
   /** Components of the query;  at least 1 is required */
   protected List<QueryComponent> components;
   
   /**
    * Determines if the current row was deleted by DELETE-RESULT-LIST-ENTRY. Note that if the
    * query is off-end, it also can be "deleted".
    */
   protected boolean currentRowDeleted = false;
   
   /** List of client-side where clause expressions */
   private List<Supplier<logical>> whereExprList = new ArrayList<>(1);
   
   /** Persistence object which this query's components use */
   private Persistence persistence = null;
   
   /** Flag indicating results must be scrollable */
   private boolean scrolling = false;
   
   /** Cache of primary key lists to result set row numbers */
   private RepositionCache repoCache = new RepositionCache();
   
   /** HQL query string to retrieve primary key ID arrays */
   private String hql = null;
   
   /** Result set for this query */
   private Results results = null;
   
   /** Does this query have any arguments which are runtime resolvable? */
   private boolean resolvableArgs = false;
   
   /** <code>false</code> if running off end of results should raise error */
   private boolean lenientOffEnd = false;
   
   /** Number of total results returned */
   private int rowCount = -1;
   
   /** Is query registered for session event notification? */
   private boolean regWithSession = false;
   
   /** Has query been registered for resource cleanup? */
   private boolean regCleaner = false;
   
   /** Force query to retrieve full records, rather than primary keys only */
   private boolean fullRecords = false;
   
   /**
    * Default constructor. Initialization is deferred until <code>initialize()</code> is
    * called.
    */
   public PreselectQuery()
   {
   }
   
   /**
    * Initializer logic which is typically used for a multi-table query. Details
    * specific to each table component of the query are added via one of the
    * <code>addComponent</code> method variants.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    sort
    *           HQL order by clause to be applied across all components of the
    *           query.
    *
    * @return   This query instance.
    */ 
   public PreselectQuery initialize(String sort)
   {
      this.sort = sort;
      components = new ArrayList<>(1);
      
      TransactionManager.registerOffEndQuery(this);
      
      return this;
   }
   
   /**
    * Initializer logic which is typically used for a multi-table query. Details
    * specific to each table component of the query are added via one of the
    * <code>addComponent</code> method variants.
    * <p>
    * This variant is used by code which is converted from an OPEN QUERY
    * statement, which permits advancing off the end of a result set without
    * raising a <code>QueryOffEndException</code>.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    lenientOffEnd
    *           <code>true</code> if advancing past either end of the result
    *           set should NOT raise a <code>QueryOffEndException</code>;
    *           <code>false</code> if doing so should raise the exception.
    * @param    sort
    *           HQL order by clause to be applied across all components of the
    *           query.
    *
    * @return   This query instance.
    */
   public PreselectQuery initialize(boolean lenientOffEnd, String sort)
   {
      initialize(sort);
      this.lenientOffEnd = lenientOffEnd;
      
      return this;
   }
   
   /**
    * Initializer logic designed for a single-table preselect query.
    * The where clause must expect no substitution parameters.  A share lock
    * will be applied to retrieved records.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           DMO proxy which defines buffer and record type.
    * @param    where
    *           HQL where clause.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           HQL order by clause.
    *
    * @return   This query instance.
    */
   public PreselectQuery initialize(DataModelObject   dmo,
                                    String            where,
                                    Supplier<logical> whereExpr,
                                    String            sort)
   {
      return initialize(dmo,
                        where,
                        whereExpr,
                        sort,
                        (DataModelObject) null,
                        null,
                        LockType.SHARE);
   }
   
   /**
    * Initializer logic designed for a single-table preselect query.
    * The where clause must expect no substitution parameters.  A share lock
    * will be applied to retrieved records.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           DMO proxy which defines buffer and record type.
    * @param    where
    *           HQL where clause.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           HQL order by clause.
    * @param    indexInfo
    *           Index information string as it is returned by INDEX-INFORMATION.
    *           May be <code>null</code> if it is not an OPEN QUERY case or if
    *           you don't need debug information about selected indexes.
    *
    * @return   This query instance.
    */
   public PreselectQuery initialize(DataModelObject   dmo,
                                    String            where,
                                    Supplier<logical> whereExpr,
                                    String            sort,
                                    String            indexInfo)
   {
      return initialize(dmo, where, whereExpr, sort, indexInfo, null, null, LockType.SHARE);
   }
   
   /**
    * Initializer logic designed for a single-table preselect query.
    * The where clause must expect no substitution parameters.  A share lock
    * will be applied to retrieved records.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           DMO proxy which defines buffer and record type.
    * @param    where
    *           HQL where clause.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           HQL order by clause.
    * @param    inverse
    *           DMO to which this query should join via a foreign relation.
    *           May be <code>null</code>.
    *
    * @return   This query instance.
    */
   public PreselectQuery initialize(DataModelObject   dmo,
                                    String            where,
                                    Supplier<logical> whereExpr,
                                    String            sort,
                                    DataModelObject   inverse)
   {
      return initialize(dmo, where, whereExpr, sort, inverse, null, LockType.SHARE);
   }
   
   /**
    * Initializer logic designed for a single-table preselect query.
    * The where clause must expect no substitution parameters.  A share lock
    * will be applied to retrieved records.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           DMO proxy which defines buffer and record type.
    * @param    where
    *           HQL where clause.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           HQL order by clause.
    * @param    indexInfo
    *           Index information string as it is returned by INDEX-INFORMATION.
    *           May be <code>null</code> if it is not an OPEN QUERY case or if
    *           you don't need debug information about selected indexes.
    * @param    inverse
    *           DMO to which this query should join via a foreign relation.
    *           May be <code>null</code>.
    *
    * @return   This query instance.
    */
   public PreselectQuery initialize(DataModelObject   dmo,
                                    String            where,
                                    Supplier<logical> whereExpr,
                                    String            sort,
                                    String            indexInfo,
                                    DataModelObject   inverse)
   {
      return initialize(dmo, where, whereExpr, sort, indexInfo, inverse, null, LockType.SHARE);
   }
   
   /**
    * Initializer logic designed for a single-table preselect query.
    * A share lock will be applied to retrieved records.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           DMO proxy which defines buffer and record type.
    * @param    where
    *           HQL where clause.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           HQL order by clause.
    * @param    args
    *           Substitution parameters to be inserted into placeholders within
    *           the where clause.
    *
    * @return   This query instance.
    */
   public PreselectQuery initialize(DataModelObject   dmo,
                                    String            where,
                                    Supplier<logical> whereExpr,
                                    String            sort,
                                    Object[]          args)
   {
      return initialize(dmo, where, whereExpr, sort, (DataModelObject) null, args, LockType.SHARE);
   }
   
   /**
    * Initializer logic designed for a single-table preselect query.
    * A share lock will be applied to retrieved records.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           DMO proxy which defines buffer and record type.
    * @param    where
    *           HQL where clause.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           HQL order by clause.
    * @param    indexInfo
    *           Index information string as it is returned by INDEX-INFORMATION.
    *           May be <code>null</code> if it is not an OPEN QUERY case or if
    *           you don't need debug information about selected indexes.
    * @param    args
    *           Substitution parameters to be inserted into placeholders within
    *           the where clause.
    *
    * @return   This query instance.
    */
   public PreselectQuery initialize(DataModelObject   dmo,
                                    String            where,
                                    Supplier<logical> whereExpr,
                                    String            sort,
                                    String            indexInfo,
                                    Object[]          args)
   {
      return initialize(dmo, where, whereExpr, sort, indexInfo, null, args, LockType.SHARE);
   }
   
   /**
    * Initializer logic designed for a single-table preselect query.
    * A share lock will be applied to retrieved records.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           DMO proxy which defines buffer and record type.
    * @param    where
    *           HQL where clause.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           HQL order by clause.
    * @param    inverse
    *           DMO to which this query should join via a foreign relation.
    *           May be <code>null</code>.
    * @param    args
    *           Substitution parameters to be inserted into placeholders within
    *           the where clause.
    *
    * @return   This query instance.
    */
   public PreselectQuery initialize(DataModelObject   dmo,
                                    String            where,
                                    Supplier<logical> whereExpr,
                                    String            sort,
                                    DataModelObject   inverse,
                                    Object[]          args)
   {
      return initialize(dmo, where, whereExpr, sort, inverse, args, LockType.SHARE);
   }
   
   /**
    * Initializer logic designed for a single-table preselect query.
    * A share lock will be applied to retrieved records.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           DMO proxy which defines buffer and record type.
    * @param    where
    *           HQL where clause.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           HQL order by clause.
    * @param    indexInfo
    *           Index information string as it is returned by INDEX-INFORMATION.
    *           May be <code>null</code> if it is not an OPEN QUERY case or if
    *           you don't need debug information about selected indexes.
    * @param    inverse
    *           DMO to which this query should join via a foreign relation.
    *           May be <code>null</code>.
    * @param    args
    *           Substitution parameters to be inserted into placeholders within
    *           the where clause.
    *
    * @return   This query instance.
    */
   public PreselectQuery initialize(DataModelObject   dmo,
                                    String            where,
                                    Supplier<logical> whereExpr,
                                    String            sort,
                                    String            indexInfo,
                                    DataModelObject   inverse,
                                    Object[]          args)
   {
      return initialize(dmo, where, whereExpr, sort, indexInfo, inverse, args, LockType.SHARE);
   }
   
   /**
    * Initializer logic designed for a single-table preselect query.
    * The where clause must expect no substitution parameters.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           DMO proxy which defines buffer and record type.
    * @param    where
    *           HQL where clause.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           HQL order by clause.
    * @param    lockType
    *           Lock type to apply to records retrieved by this query.
    *
    * @return   This query instance.
    */
   public PreselectQuery initialize(DataModelObject   dmo,
                                    String            where,
                                    Supplier<logical> whereExpr,
                                    String            sort,
                                    LockType          lockType)
   {
      return initialize(dmo, where, whereExpr, sort, (DataModelObject) null, null, lockType);
   }
   
   /**
    * Initializer logic designed for a single-table preselect query.
    * The where clause must expect no substitution parameters.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           DMO proxy which defines buffer and record type.
    * @param    where
    *           HQL where clause.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           HQL order by clause.
    * @param    indexInfo
    *           Index information string as it is returned by INDEX-INFORMATION.
    *           May be <code>null</code> if it is not an OPEN QUERY case or if
    *           you don't need debug information about selected indexes.
    * @param    lockType
    *           Lock type to apply to records retrieved by this query.
    *
    * @return   This query instance.
    */
   public PreselectQuery initialize(DataModelObject   dmo,
                                    String            where,
                                    Supplier<logical> whereExpr,
                                    String            sort,
                                    String            indexInfo,
                                    LockType          lockType)
   {
      return initialize(dmo, where, whereExpr, sort, indexInfo, null, null, lockType);
   }
   
   /**
    * Initializer logic designed for a single-table preselect query.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           DMO proxy which defines buffer and record type.
    * @param    where
    *           HQL where clause.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           HQL order by clause.
    * @param    inverse
    *           DMO to which this query should join via a foreign relation.
    *           May be <code>null</code>.
    * @param    lockType
    *           Lock type to apply to records retrieved by this query.
    *
    * @return   This query instance.
    */
   public PreselectQuery initialize(DataModelObject   dmo,
                                    String            where,
                                    Supplier<logical> whereExpr,
                                    String            sort,
                                    DataModelObject   inverse,
                                    LockType          lockType)
   {
      return initialize(dmo, where, whereExpr, sort, inverse, null, lockType);
   }
   
   /**
    * Initializer logic designed for a single-table preselect query.
    * The where clause must expect no substitution parameters.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           DMO proxy which defines buffer and record type.
    * @param    where
    *           HQL where clause.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           HQL order by clause.
    * @param    indexInfo
    *           Index information string as it is returned by INDEX-INFORMATION.
    *           May be <code>null</code> if it is not an OPEN QUERY case or if
    *           you don't need debug information about selected indexes.
    * @param    inverse
    *           DMO to which this query should join via a foreign relation.
    *           May be <code>null</code>.
    * @param    lockType
    *           Lock type to apply to records retrieved by this query.
    *
    * @return   This query instance.
    */
   public PreselectQuery initialize(DataModelObject   dmo,
                                    String            where,
                                    Supplier<logical> whereExpr,
                                    String            sort,
                                    String            indexInfo,
                                    DataModelObject   inverse,
                                    LockType          lockType)
   {
      return initialize(dmo, where, whereExpr, sort, indexInfo, inverse, null, lockType);
   }
   
   /**
    * Initializer logic designed for a single-table preselect query.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           DMO proxy which defines buffer and record type.
    * @param    where
    *           HQL where clause.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           HQL order by clause.
    * @param    args
    *           Substitution parameters to be inserted into placeholders within
    *           the where clause.
    * @param    lockType
    *           Lock type to apply to records retrieved by this query.
    *
    * @return   This query instance.
    */
   public PreselectQuery initialize(DataModelObject   dmo,
                                    String            where,
                                    Supplier<logical> whereExpr,
                                    String            sort,
                                    Object[]          args,
                                    LockType          lockType)
   {
      return initialize(dmo, where, whereExpr, sort, (DataModelObject) null, args, lockType);
   }
   
   /**
    * Initializer logic designed for a single-table preselect query.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           DMO proxy which defines buffer and record type.
    * @param    where
    *           HQL where clause.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           HQL order by clause.
    * @param    indexInfo
    *           Index information string as it is returned by INDEX-INFORMATION.
    *           May be <code>null</code> if it is not an OPEN QUERY case or if
    *           you don't need debug information about selected indexes.
    * @param    args
    *           Substitution parameters to be inserted into placeholders within
    *           the where clause.
    * @param    lockType
    *           Lock type to apply to records retrieved by this query.
    *
    * @return   This query instance.
    */
   public PreselectQuery initialize(DataModelObject   dmo,
                                    String            where,
                                    Supplier<logical> whereExpr,
                                    String            sort,
                                    String            indexInfo,
                                    Object[]          args,
                                    LockType          lockType)
   {
      return initialize(dmo, where, whereExpr, sort, indexInfo, null, args, lockType);
   }
   
   /**
    * Initializer logic designed for a single-table preselect query.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           DMO proxy which defines buffer and record type.
    * @param    where
    *           HQL where clause.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           HQL order by clause.
    * @param    inverse
    *           DMO to which this query should join via a foreign relation.
    *           May be <code>null</code>.
    * @param    args
    *           Substitution parameters to be inserted into placeholders within
    *           the where clause.
    * @param    lockType
    *           Lock type to apply to records retrieved by this query.
    *
    * @return   This query instance.
    */
   public PreselectQuery initialize(DataModelObject   dmo,
                                    String            where,
                                    Supplier<logical> whereExpr,
                                    String            sort,
                                    DataModelObject   inverse,
                                    Object[]          args,
                                    LockType          lockType)
   {
      return initialize(dmo, where, whereExpr, sort, null, inverse, args, lockType);
   }
   
   /**
    * Initializer logic designed for a single-table preselect query.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           DMO proxy which defines buffer and record type.
    * @param    where
    *           HQL where clause.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           HQL order by clause.
    * @param    indexInfo
    *           Index information string as it is returned by INDEX-INFORMATION.
    *           May be <code>null</code> if it is not an OPEN QUERY case or if
    *           you don't need debug information about selected indexes.
    * @param    inverse
    *           DMO to which this query should join via a foreign relation.
    *           May be <code>null</code>.
    * @param    args
    *           Substitution parameters to be inserted into placeholders within
    *           the where clause.
    * @param    lockType
    *           Lock type to apply to records retrieved by this query.
    *
    * @return   This query instance.
    */
   public PreselectQuery initialize(DataModelObject   dmo,
                                    String            where,
                                    Supplier<logical> whereExpr,
                                    String            sort,
                                    String            indexInfo,
                                    DataModelObject   inverse,
                                    Object[]          args,
                                    LockType          lockType)
   {
      components = new ArrayList<>(2);
      
      setErrorIfNull(false);
      addComponent(dmo, where, indexInfo, inverse, args, lockType, NEXT, false);
      
      this.sort = sort;
      if (whereExpr != null)
      {
         addWhereExpression(whereExpr);
      }
      
      TransactionManager.registerOffEndQuery(this);
      
      return this;
   }
   
   /**
    * Set the behavior of this query as a record retrieval request moves off
    * the end of its results.  If set to lenient mode, invocations of
    * <code>next</code> and <code>previous</code> which do not find a record
    * do not raise a <code>QueryOffEndException</code>.  By default, this
    * would normally raise the exception.
    *
    * @param   lenientOffEnd
    *          <code>true</code> to suppress <code>QueryOffEndException</code>
    *          for <code>next</code> and <code>previous</code> requests;
    *          <code>false</code> to permit these exceptions.
    */
   public void setLenientOffEnd(boolean lenientOffEnd)
   {
      this.lenientOffEnd = lenientOffEnd;
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    *   <li>no inverse join DMO
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    */
   public void addComponent(DataModelObject dmo, String where)
   {
      addComponent(dmo, where, (String) null);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    *   <li>no inverse join DMO
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   indexInfo
    *          Index information string as it is returned by INDEX-INFORMATION.
    *          May be <code>null</code> if it is not an OPEN QUERY case or if
    *          you don't need debug information about selected indexes.
    */
   public void addComponent(DataModelObject dmo, String where, String indexInfo)
   {
      addComponent(dmo, where, indexInfo, null, null, LockType.SHARE, NEXT, false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    */
   public void addComponent(DataModelObject dmo, String where, DataModelObject inverse)
   {
      addComponent(dmo, where, null, inverse);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   indexInfo
    *          Index information string as it is returned by INDEX-INFORMATION.
    *          May be <code>null</code> if it is not an OPEN QUERY case or if
    *          you don't need debug information about selected indexes.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            String indexInfo,
                            DataModelObject inverse)
   {
      addComponent(dmo, where, indexInfo, inverse, null, LockType.SHARE, NEXT, false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    *   <li>no inverse join DMO
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    */
   public void addComponent(DataModelObject dmo, String where, Object[] args)
   {
      addComponent(dmo, where, (String) null, args);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    *   <li>no inverse join DMO
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   indexInfo
    *          Index information string as it is returned by INDEX-INFORMATION.
    *          May be <code>null</code> if it is not an OPEN QUERY case or if
    *          you don't need debug information about selected indexes.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    */
   public void addComponent(DataModelObject dmo, String where, String indexInfo, Object[] args)
   {
      addComponent(dmo, where, indexInfo, null, args, LockType.SHARE, NEXT, false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            DataModelObject inverse,
                            Object[] args)
   {
      addComponent(dmo, where, null, inverse, args);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   indexInfo
    *          Index information string as it is returned by INDEX-INFORMATION.
    *          May be <code>null</code> if it is not an OPEN QUERY case or if
    *          you don't need debug information about selected indexes.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            String indexInfo,
                            DataModelObject inverse,
                            Object[] args)
   {
      addComponent(dmo, where, indexInfo, inverse, args, LockType.SHARE, NEXT, false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    *   <li>no inverse join DMO
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    */
   public void addComponent(DataModelObject dmo, String where, LockType lockType)
   {
      addComponent(dmo, where, (String) null, lockType);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    *   <li>no inverse join DMO
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   indexInfo
    *          Index information string as it is returned by INDEX-INFORMATION.
    *          May be <code>null</code> if it is not an OPEN QUERY case or if
    *          you don't need debug information about selected indexes.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            String indexInfo,
                            LockType lockType)
   {
      addComponent(dmo, where, indexInfo, null, null, lockType, NEXT, false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            DataModelObject inverse,
                            LockType lockType)
   {
      addComponent(dmo, where, null, inverse, lockType);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   indexInfo
    *          Index information string as it is returned by INDEX-INFORMATION.
    *          May be <code>null</code> if it is not an OPEN QUERY case or if
    *          you don't need debug information about selected indexes.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            String indexInfo,
                            DataModelObject inverse,
                            LockType lockType)
   {
      addComponent(dmo, where, indexInfo, inverse, null, lockType, NEXT, false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    *   <li>no inverse join DMO
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            Object[] args,
                            LockType lockType)
   {
      addComponent(dmo, where, (String) null, args, lockType);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    *   <li>no inverse join DMO
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   indexInfo
    *          Index information string as it is returned by INDEX-INFORMATION.
    *          May be <code>null</code> if it is not an OPEN QUERY case or if
    *          you don't need debug information about selected indexes.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            String indexInfo,
                            Object[] args,
                            LockType lockType)
   {
      addComponent(dmo, where, indexInfo, null, args, lockType, NEXT, false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            DataModelObject inverse,
                            Object[] args,
                            LockType lockType)
   {
      addComponent(dmo, where, null, inverse, args, lockType);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   indexInfo
    *          Index information string as it is returned by INDEX-INFORMATION.
    *          May be <code>null</code> if it is not an OPEN QUERY case or if
    *          you don't need debug information about selected indexes.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            String indexInfo,
                            DataModelObject inverse,
                            Object[] args,
                            LockType lockType)
   {
      addComponent(dmo, where, indexInfo, inverse, args, lockType, NEXT, false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>no inverse join DMO
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    * @param   iteration
    *          Iteration/retrieval mode for this component of the query.  At
    *          this time, only <code>NEXT</code> is supported.  Future
    *          support for <code>FIRST</code>, <code>LAST</code>, and
    *          <code>UNIQUE</code> is expected.
    * @param   outer
    *          <code>true</code> if an outer join is required.  At this time,
    *          only inner joins are supported.  Future support for outer
    *          joins is expected.
    *
    * @throws  UnsupportedOperationException
    *          if <code>outer</code> is <code>true</code> (temporary);
    *          if <code>iteration</code> is not <code>NEXT</code> (temporary);
    *          if <code>dmo</code>'s associated database is different than
    *          that of a previously added component's database (cross-database
    *          preselect joins are not possible).
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            Object[] args,
                            LockType lockType,
                            int iteration,
                            boolean outer)
   {
      addComponent(dmo, where, (String) null, args, lockType, iteration, outer);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>no inverse join DMO
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   indexInfo
    *          Index information string as it is returned by INDEX-INFORMATION.
    *          May be <code>null</code> if it is not an OPEN QUERY case or if
    *          you don't need debug information about selected indexes.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    * @param   iteration
    *          Iteration/retrieval mode for this component of the query.  At
    *          this time, only <code>NEXT</code> is supported.  Future
    *          support for <code>FIRST</code>, <code>LAST</code>, and
    *          <code>UNIQUE</code> is expected.
    * @param   outer
    *          <code>true</code> if an outer join is required.  At this time,
    *          only inner joins are supported.  Future support for outer
    *          joins is expected.
    *
    * @throws  UnsupportedOperationException
    *          if <code>outer</code> is <code>true</code> (temporary);
    *          if <code>iteration</code> is not <code>NEXT</code> (temporary);
    *          if <code>dmo</code>'s associated database is different than
    *          that of a previously added component's database (cross-database
    *          preselect joins are not possible).
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            String indexInfo,
                            Object[] args,
                            LockType lockType,
                            int iteration,
                            boolean outer)
   {
      QueryComponent comp = new QueryComponent(this,
                                               (BufferReference) dmo,
                                               null,
                                               where,
                                               indexInfo,
                                               lockType,
                                               args,
                                               iteration,
                                               outer);
      addComponent(comp);
   }
   
   /**
    * Add a component to the query.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    * @param   iteration
    *          Iteration/retrieval mode for this component of the query.  At
    *          this time, only <code>NEXT</code> is supported.  Future
    *          support for <code>FIRST</code>, <code>LAST</code>, and
    *          <code>UNIQUE</code> is expected.
    * @param   outer
    *          <code>true</code> if an outer join is required.  At this time,
    *          only inner joins are supported.  Future support for outer
    *          joins is expected.
    *
    * @throws  UnsupportedOperationException
    *          if <code>outer</code> is <code>true</code> (temporary);
    *          if <code>iteration</code> is not <code>NEXT</code> (temporary);
    *          if <code>dmo</code>'s associated database is different than
    *          that of a previously added component's database (cross-database
    *          preselect joins are not possible).
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            DataModelObject inverse,
                            Object[] args,
                            LockType lockType,
                            int iteration,
                            boolean outer)
   {
      addComponent(dmo, where, null, inverse, args, lockType, iteration, outer);
   }
   
   /**
    * Add a component to the query.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   indexInfo
    *          Index information string as it is returned by INDEX-INFORMATION.
    *          May be <code>null</code> if it is not an OPEN QUERY case or if
    *          you don't need debug information about selected indexes.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    * @param   iteration
    *          Iteration/retrieval mode for this component of the query.  At
    *          this time, only <code>NEXT</code> is supported.  Future
    *          support for <code>FIRST</code>, <code>LAST</code>, and
    *          <code>UNIQUE</code> is expected.
    * @param   outer
    *          <code>true</code> if an outer join is required.  At this time,
    *          only inner joins are supported.  Future support for outer
    *          joins is expected.
    *
    * @throws  UnsupportedOperationException
    *          if <code>outer</code> is <code>true</code> (temporary);
    *          if <code>iteration</code> is not <code>NEXT</code> (temporary);
    *          if <code>dmo</code>'s associated database is different than
    *          that of a previously added component's database (cross-database
    *          preselect joins are not possible).
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            String indexInfo,
                            DataModelObject inverse,
                            Object[] args,
                            LockType lockType,
                            int iteration,
                            boolean outer)
   {
      BufferReference proxy = (BufferReference) dmo;
      AbstractJoin join = processJoin(proxy, (BufferReference) inverse);
      
      QueryComponent comp = new QueryComponent(this,
                                               proxy,
                                               join,
                                               where,
                                               indexInfo,
                                               lockType,
                                               args,
                                               iteration,
                                               outer);
      addComponent(comp);
   }
   
   /**
    * Register all <code>RecordChangeListener</code>s associated with this
    * query (if any), with the context-local <code>ChangeBroker</code> at the
    * indicated scope.
    * <p>
    * This implementation does nothing.
    * 
    * @param   scope
    *          Scope at which listeners should be registered with the change
    *          broker.
    */
   public void registerRecordChangeListeners(int scope)
   {
   }
   
   /**
    * Respond to a clean (i.e., transaction commit) session event by caching
    * all preselected results and using that cache to back the query's results
    * for all future fetches.  If all results already are cached, do nothing.
    * <p>
    * Respond to a session closing event by clearing all results, unless they
    * were just cached as described above.
    *
    * @param   event
    *          Session event type.
    *
    * @return  <code>true</code>, indicating this query should be
    *          deregistered as a session listener after this method returns;
    *          else <code>false</code>.
    * 
    * @throws  PersistenceException
    *          if there is an error caching results.
    */
   public boolean sessionEvent(SessionListener.Event event)
   throws PersistenceException
   {
      if (closed)
      {
         return true;
      }
      
      switch (event)
      {
         case TRANSACTION_COMMITTING:
            if (results != null)
            {
               try
               {
                  SimpleResults cached = cacheResults(results);
                  if (results != cached)
                  {
                     results.sessionClosing();
                  }
                  setResults(cached);
               }
               catch (Exception exc)
               {
                  resetResults();
                  
                  throw new PersistenceException("Error caching results", exc);
               }
            }
            break;
            
         case SESSION_CLOSING_NORMALLY:
            if (results != null && !(results instanceof SimpleResults))
            {
               resetResults();
            }
            break;
            
         case SESSION_CLOSING_WITH_ERROR:
            resetResults();
            break;
            
         default:
            return false;
      }
      
      return true;
   }
   
   /**
    * Invoked when this session listener is removed from the list of
    * registered session listeners.
    */
   public void deregisteredSessionListener()
   {
      regWithSession = false;
   }
   
   /**
    * Provides a notification that the scope in which this query was created
    * is ending.  This is a no-op implementation required by the interface.
    */
   public void finished()
   {
   }
   
   /**
    * Provides a notification that the external program scope in which the object is registered is
    * is being deleted and the object's reference will be lost after this method is called.
    * <p>
    * This is a no-op implementation required by the interface.
    */
   public void deleted()
   {
   }
   
   /**
    * Provides a notification that the block whose scope in which the object
    * is registered is about to iterate and attempt another pass.  This is a
    * no-op implementation required by the interface.
    */
   public void iterate()
   {
   }
   
   /**
    * Provides a notification that the block whose scope in which the object
    * is registered is about to retry the same iteration (all loop control
    * data is unchanged).  This implementation sets the state of the query as
    * if a reposition to the current row has just occurred.  This will keep
    * the subsequent call to {@link #next()} at the top of the block body from
    * advancing to the next row.
    * <p>
    * This implementation retries the current record only if the query is 
    * associated with a FOR block (see {@link BlockType}).
    */
   public void retry()
   {
      BlockType type = TransactionManager.getBlockType();
      
      // only queries associated with a FOR block allows the retry of the 
      // current record
      switch (type)
      {
         case FOR_LOOP:
         case FOR_LOOP_WHILE:
         case FOR_LOOP_TO:
         case FOR_LOOP_TO_WHILE:
         case FOR_BLOCK:
         case FOR_BLOCK_WHILE:
         case FOR_BLOCK_TO:
         case FOR_BLOCK_TO_WHILE:
            betweenRows = true;
            break;
      }
      
      if (LOG.isLoggable(Level.FINE))
      {
         LOG.log(Level.FINE, "Preparing query for retry [" + this + "]");
      }
   }
   
   /**
    * Create a string representation of this object's state, primarily for debug purposes.
    * 
    * @return  String representation of this object.
    */
   public String toString()
   {
      return "[" + getTableCount() + " table(s)] HQL: " + hql;
   }
   
   /**
    * Add a client-side where expression to this query.  Each initial query
    * result found by the server must be accepted by this expression before
    * becoming a part of the query's final result set.
    *
    * @param   whereExpr
    *          Client-side where clause expression.  May not be
    *          <code>null</code>.
    *
    * @throws  NullPointerException
    *          if <code>whereExpr</code> is <code>null</code>.
    */
   public void addWhereExpression(Supplier<logical> whereExpr)
   {
      if (whereExpr == null)
      {
         throw new NullPointerException("Client-side where expression may not be null");
      }
      
      whereExprList.add(whereExpr);
   }
   
   /**
    * Set the iteration type for the most recently added query component.
    *
    * @param   iteration
    *          Retrieval mode from {@link QueryConstants}. Must not be FIRST or LAST (use
    *          {@link #setIterationType(int, String)} instead for these types.
    *
    * @throws  IllegalStateException
    *          If no component has yet been added.
    * @throws  IllegalArgumentException
    *          If <code>iteration</code> type is <code>FIRST</code> or <code>LAST</code>.
    */
   public void setIterationType(int iteration)
   {
      setIterationType(iteration, null);
   }
   
   /**
    * Set the iteration type for the most recently added query component.
    *
    * @param   iteration
    *          Retrieval mode from {@link QueryConstants}.
    * @param   sort
    *          Order by clause which should be associated with the most recently added query
    *          component. Must be non-<code>null</code> for iteration types <code>FIRST</code>
    *          and <code>LAST</code>.
    *
    * @throws  IllegalStateException
    *          If no component has yet been added.
    * @throws  IllegalArgumentException
    *          If <code>iteration</code> type is <code>FIRST</code> or <code>LAST</code>, but
    *          <code>sort</code> is <code>null</code>, or if <code>sort</code> represents an
    *          invalid HQL order by expression.
    */
   public void setIterationType(int iteration, String sort)
   {
      if (components.isEmpty())
      {
         throw new IllegalStateException(
            "Must add a query component before setting its iteration type");
      }
      
      QueryComponent comp = components.get(components.size() - 1);
      
      try
      {
         comp.setIteration(iteration, sort);
      }
      catch (PersistenceException exc)
      {
         // should only happen if we have a runtime programming error or conversion error
         // (or with bad, hand-written business logic)
         throw new IllegalArgumentException(exc);
      }
   }
   
   /**
    * Retrieve the first composite row of results for the query.  The
    * underlying buffer for each query component is updated with the
    * appropriate record.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @throws  ErrorConditionException
    *          if the query or the fetch of any record fails, or if no records
    *          are available and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void first()
   {
      first(null);
   }
   
   /**
    * Retrieve the first composite row of results for the query.  The
    * underlying buffer for each query component is updated with the
    * appropriate record.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @param   lockType
    *          Lock type to apply to records retrieved (overrides default
    *          lock type set for each query component).
    *
    * @throws  ErrorConditionException
    *          if the query or the fetch of any record fails, or if no records
    *          are available and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void first(LockType lockType)
   {
      execute();
      
      boolean missing = false;
      do
      {
         try
         {
            boolean avail = (missing ? results.next() : results.first());
            missing = false;
            offEnd = (avail ? OffEnd.NONE : OffEnd.BACK);
            betweenRows = false;
            fetch(avail, lockType, isErrorIfNull());
         }
         catch (MissingRecordException exc)
         {
            missing = true;
         }
         catch (QueryOffEndException exc)
         {
            offEnd = OffEnd.BACK;
            exc.fillInStackTrace();
            throw exc;
         }
      } while (missing);
   }
   
   /**
    * Retrieve the last composite row of results for the query.  The
    * underlying buffer for each query component is updated with the
    * appropriate record.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @throws  ErrorConditionException
    *          if the query or the fetch of any record fails, or if no records
    *          are available and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void last()
   {
      last(null);
   }
   
   /**
    * Retrieve the last composite row of results for the query.  The
    * underlying buffer for each query component is updated with the
    * appropriate record.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @param   lockType
    *          Lock type to apply to records retrieved (overrides default
    *          lock type set for each query component).
    *
    * @throws  ErrorConditionException
    *          if the query or the fetch of any record fails, or if no records
    *          are available and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void last(LockType lockType)
   {
      execute();
      
      boolean missing = false;
      do
      {
         try
         {
            boolean avail = (missing ? results.previous() : results.last());
            missing = false;
            offEnd = (avail ? OffEnd.NONE : OffEnd.BACK);
            betweenRows = false;
            fetch(avail, lockType, isErrorIfNull());
         }
         catch (MissingRecordException exc)
         {
            missing = true;
         }
         catch (QueryOffEndException exc)
         {
            offEnd = OffEnd.BACK;
            exc.fillInStackTrace();
            throw exc;
         }
      } while (missing);
   }
   
   /**
    * Retrieve the next composite row of results for the query.  The
    * underlying buffer for each query component is updated with the
    * appropriate record.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @throws  ErrorConditionException
    *          if the query or the fetch of any record fails, or if no records
    *          are available and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void next()
   {
      next(null);
   }
   
   /**
    * Retrieve the next composite row of results for the query.  The
    * underlying buffer for each query component is updated with the
    * appropriate record.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @param   lockType
    *          Lock type to apply to records retrieved (overrides default
    *          lock type set for each query component).
    *
    * @throws  ErrorConditionException
    *          if the query or the fetch of any record fails, or if no records
    *          are available and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void next(LockType lockType)
   {
      execute();
      
      boolean missing;
      do
      {
         missing = false;
         try
         {
            boolean avail = ((betweenRows && results.getRowNumber() != -1) || results.next());
            offEnd = (avail ? OffEnd.NONE : OffEnd.BACK);
            fetch(avail, lockType, false);
         }
         catch (QueryOffEndException exc)
         {
            offEnd = OffEnd.BACK;
            if (!lenientOffEnd)
            {
               throw exc;
            }
         }
         catch (MissingRecordException exc)
         {
            missing = true;
         }
         finally
         {
            betweenRows = false;
         }
      } while (missing);
   }
   
   /**
    * Retrieve the previous composite row of results for the query.  The
    * underlying buffer for each query component is updated with the
    * appropriate record.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @throws  ErrorConditionException
    *          if the query or the fetch of any record fails, or if no records
    *          are available and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void previous()
   {
      previous(null);
   }
   
   /**
    * Retrieve the previous composite row of results for the query.  The
    * underlying buffer for each query component is updated with the
    * appropriate record.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @param   lockType
    *          Lock type to apply to records retrieved (overrides default
    *          lock type set for each query component).
    *
    * @throws  ErrorConditionException
    *          if the query or the fetch of any record fails, or if no records
    *          are available and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void previous(LockType lockType)
   {
      boolean setToLast = treatPreviousAsLast();
      execute();
      
      boolean missing;
      do
      {
         missing = false;
         try
         {
            boolean avail = (setToLast ? results.last() : results.previous());
            if (avail)
            {
               offEnd = OffEnd.NONE;
            }
            else
            {
               offEnd = (setToLast ? OffEnd.BACK : OffEnd.FRONT);
            }
            fetch(avail, lockType, false);
         }
         catch (QueryOffEndException exc)
         {
            offEnd = (setToLast ? OffEnd.BACK : OffEnd.FRONT);
            if (!lenientOffEnd)
            {
               throw exc;
            }
         }
         catch (MissingRecordException exc)
         {
            missing = true;
         }
         finally
         {
            betweenRows = false;
         }
      } while (missing);
   }
   
   /**
    * Retrieve a unique record (only valid for a single-table query).
    *
    * @throws  ErrorConditionException
    *          if the query fails or if more than one result is found which
    *          matches the query criteria.
    * @throws  UnsupportedOperationException
    *          if this method is invoked on a multi-table query.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void unique()
   {
      unique(null);
   }
   
   /**
    * Retrieve a unique record (only valid for a single-table query),
    * overriding the default lock type of the underlying query component.
    *
    * @param   lockType
    *          Lock type which should override that of the query component.
    *
    * @throws  ErrorConditionException
    *          if the query fails or if more than one result is found which
    *          matches the query criteria.
    * @throws  UnsupportedOperationException
    *          if this method is invoked on a multi-table query.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void unique(LockType lockType)
   {
      if (components.size() > 1)
      {
         throw new UnsupportedOperationException(
            "Unique retrieval only valid for a single-table preselect query");
      }
      
      boolean setToFirst = (results == null);
      execute();
      boolean avail = true;
      if (setToFirst)
      {
         avail = results.first();
      }
      
      if (!results.isFirst() || !results.isLast())
      {
         try
         {
            QueryComponent comp = components.get(0);
            RecordBuffer buffer = comp.getBuffer();
            getPersistence().uniqueResultViolation(buffer, null);
         }
         catch (PersistenceException exc)
         {
            ErrorManager.recordOrThrowError(exc);
            
            return;
         }
      }
      
      offEnd = (avail ? OffEnd.NONE : OffEnd.BACK);
      
      try
      {
         fetch(avail, lockType, false);
      }
      catch (MissingRecordException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
      catch (QueryOffEndException exc)
      {
         offEnd = OffEnd.BACK;
         exc.fillInStackTrace();
         throw exc;
      }
   }
   
   /**
    * Reload the current composite row of results for the query.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @throws  ErrorConditionException
    *          if the reload triggers an error.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void current()
   {
      current(null);
   }
   
   /**
    * Reload the current composite row of results for the query, overriding
    * the lock type to apply to each record.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @param   lockType
    *          Lock type which should override that of any query component.
    *
    * @throws  ErrorConditionException
    *          if the reload triggers an error.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void current(LockType lockType)
   {
      if (_isOffEnd() || currentRowDeleted)
      {
         ErrorManager.displayError(4114, "No query record is available", false);
         return;
      }
      
      execute();
      
      Iterator<QueryComponent> iter = components();
      while (iter.hasNext())
      {
         QueryComponent comp = iter.next();
         LockType lt = (lockType == null ? comp.getLockType() : lockType);
         
         final RecordBuffer buffer = comp.getBuffer();
         buffer.armCurrentChanged();
         buffer.reload(lt, isErrorIfNull());
         buffer.updateCurrentChanged();
      }
   }
   
   /**
    * Load one or more records into their associated buffers, given an array
    * of primary key ID values, or the actual DMOs themselves.  The IDs/DMOs
    * are arranged to coincide with the tables underlying the query, from
    * left-most to right-most (in the sense of how they are joined by the
    * query).
    * <p>
    * In the event an expected value cannot be loaded (e.g., the record has
    * been deleted or is otherwise no longer available), this method throws
    * {@link MissingRecordException}, after setting the associated buffer(s)
    * into {@link RecordBuffer#setUnknownMode() unknown mode}. Note that a
    * record being locked by another session does not constitute a "missing"
    * record. Note also that this exception is not thrown until all the
    * elements of <code>data</code> have been processed to the extent
    * possible.
    *
    * @param   data
    *          Array of primary key IDs or DMOs, one per table involved in the
    *          query.
    * @param   lockType
    *          Lock type which should by used.  Set to <code>null</code> to
    *          allow query to use its current lock type.
    * @param   silentIfNullId
    *          Do not raise {@link MissingRecordException} if some of the provided
    *          IDs are <code>null</code>. <code>null</code> IDs are valid for
    *          queries with OUTER join.
    *
    * @throws  MissingRecordException
    *          if any record cannot be loaded, because it is no longer
    *          available (e.g., deleted, etc.).
    * @throws  PersistenceException
    *          if there is an error loading data.
    */
   public void load(Object[] data, LockType lockType, boolean silentIfNullId)
   throws PersistenceException
   {
      boolean retry = (results != null);
      
      int len = data.length;
      Serializable[] ids = new Serializable[len];
      for (int i = 0; i < len; i++)
      {
         Object datum = data[i];
         ids[i] = (datum instanceof Persistable
                   ? ((Persistable) datum).getId()
                   : (Serializable) datum);
      }
      
      try
      {
         if (!repositionByID(ids, true))
         {
            if (retry)
            {
               // Query did not have the target record(s) cached in its current
               // result set, so reset the results and try again.
               resetResults();
               if (!repositionByID(ids, true))
               {
                  throw new MissingRecordException();
               }
            }
            else
            {
               throw new MissingRecordException();
            }
            
            // TODO do not throw MissingRecordException if silentIfNullId is true and an ID
            // is null
         }
         
         fetch(offEnd == OffEnd.NONE, lockType, false);
      }
      catch (MissingRecordException exc)
      {
         setEmptyBuffersUnknown();
         
         throw exc;
      }
      finally
      {
         betweenRows = false;
      }
   }
   
   /**
    * Get all the off-end listeners associated with this query.
    * 
    * @return  the off-end listeners associated with this query.
    */
   public List<QueryOffEndListener> getOffEndListeners()
   {
      List<QueryOffEndListener> listeners = new ArrayList<>();
      
      for (QueryComponent component : components)
      {
         listeners.add(component.getBuffer().getQueryOffEndListener());
      }
      
      return listeners;
   }
   
   /**
    * Assemble an array of primary key IDs for the current record(s) in the
    * buffers underlying this query, from left-most to right-most (in the
    * sense of how the query joins the associated tables).  This method
    * essentially takes a lightweight snapshot of the currently loaded
    * record(s), so that this may be restored later.
    *
    * @return  Array of primary key IDs, one per table involved in the query.
    *
    * @see     #load(Object[], LockType, boolean)
    */
   public Object[] getRow()
   {
      if (results == null)
      {
         return null;
      }
      
      Object[] srcData = results.get();
      if (srcData == null)
      {
         return null;
      }
      
      int size = components.size();
      Object[] rowData = new Object[size];
      System.arraycopy(srcData, 0, rowData, 0, size);
      
      return rowData;
   }
   
   /**
    * Reposition the cursor such that a request to retrieve the next result
    * will retrieve the result which matches the specified primary key ID.  A
    * request to retrieve the previous result will retrieve the result
    * immediately previous to this, if any, in the results list.
    *
    * @param   id
    *          A primary key ID, representing the target record (in the case
    *          of a join, this ID is associated with the left-most record in
    *          the join).
    * 
    * @throws  ErrorConditionException
    *          if the specified ID represents the unknown value.
    */
   public void repositionByID(recid id)
   {
      if (!validateRepositionByID(id))
      {
         return;
      }
      
      Serializable[] serIDs = new Serializable[] { id.longValue() };
      
      if (!repositionByID(serIDs, false))
      {
         ErrorManager.recordOrThrowError(7331, "Cannot reposition query to recid/rowid(s) given");
      }
   }
   
   /**
    * Reposition the cursor such that a request to retrieve the next result
    * will retrieve the result which matches the specified array of primary
    * key IDs.  A request to retrieve the previous result will retrieve the
    * result immediately previous to this, if any, in the results list.
    *
    * @param   id1
    *          A primary key ID, representing the target record (in the case
    *          of a join, this ID is associated with the left-most record in
    *          the join).
    * @param   joinIDs
    *          All remaining primary key IDs, if any, arranged from left to
    *          right to coincide with the records being joined by the
    *          underlying query.
    * 
    * @throws  ErrorConditionException
    *          if any of the specified IDs represents the unknown value.
    */
   public void repositionByID(rowid id1, rowid...joinIDs)
   {
      int len = joinIDs.length;
      Serializable[] serIDs = new Serializable[len + 1];
      
      serIDs[0] = id1.getValue();
      
      for (int i = 0; i < len; i++)
      {
         rowid next = joinIDs[i];
         
         serIDs[i + 1] = next.getValue();
      }
      
      if (!repositionByID(serIDs, false))
      {
         ErrorManager.recordOrThrowError(
            7331,
            "Cannot reposition query to recid/rowid(s) given");
      }
   }
   
   /**
    * Reposition the cursor to the specified row in the result set.  The row
    * indicates the position between two results, such that a call to
    * retrieve the next record will get the result at the specified row,
    * and a call to retrieve the previous result will get the result before
    * the new position.
    *
    * @param   row
    *          1-based index of the target position.
    *
    * @throws  ErrorConditionException
    *          if <code>row</code> represents unknown value.
    */
   public void reposition(NumberType row)
   {
      if (validateReposition(row))
      {
         reposition(row.intValue());
      }
   }
   
   /**
    * Reposition the query to the specified row in the result set.  The row
    * indicates the position between two results, such that a call to
    * retrieve the next record will get the record after the new position,
    * and a call to retrieve the previous record will get the record before
    * the new position.
    *
    * @param   row
    *          1-based index of the target position.
    *
    * @throws  ErrorConditionException
    *          if there is an error repositioning the cursor.
    */
   public void reposition(int row)
   {
      prepareReposition();
      
      if (results == null)
      {
         execute();
      }
      
      row--;   // adjust for 1-based row
      int currentRow = results.getRowNumber();
      if ((row >= 0 && currentRow == row) || results.setRowNumber(row))
      {
         offEnd = OffEnd.NONE;
         betweenRows = true;
      }
      else
      {
         offEnd = (row < 0 ? OffEnd.FRONT : OffEnd.BACK);
         betweenRows = false;
      }
      
      afterReposition(true);
   }
   
   /**
    * Advance the current cursor position forward by the specified number of rows.  This will
    * always leave the cursor between two results.  For example, given a request to move 1 row
    * forward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>, this would move it
    *       ahead <i>one and a half positions</i>, to just beyond the following result;
    *   <li>if the cursor currently is positioned <i>between two results</i>, this would move it
    *       ahead <i>one position</i>, to just beyond the following result.
    * </ul>
    *
    * @param   rows
    *          Number of rows to scroll the cursor forward.
    *
    * @return  {@code true} if operation was successful and the cursor was moved the exact amount
    *          of rows as specified by the parameter. If {@code false} is returned then it is
    *          possible that cursor's position was moved but not with {@code rows} rows.
    *
    * @throws  ErrorConditionException
    *          if {@code rows} represents the unknown value; if there is an error repositioning
    *          the cursor.
    */
   public boolean forward(NumberType rows)
   {
      if (validateReposition(rows))
      {
         return forward(rows.intValue());
      }
      return false;
   }
   
   /**
    * Advance the current cursor position forward by the specified number of rows.  This will
    * always leave the cursor between two results.  For example, given a request to move 1 row
    * forward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>, this would move it
    *       ahead <i>one and a half positions</i>, to just beyond the following result;
    *   <li>if the cursor currently is positioned <i>between two results</i>, this would move it
    *       ahead <i>one position</i>, to just beyond the following result.
    * </ul>
    *
    * @param   rows
    *          Number of rows to scroll the cursor forward.
    *
    * @return  {@code true} if operation was successful and the cursor was moved the exact amount
    *          of rows as specified by the parameter. If {@code false} is returned then it is
    *          possible that cursor's position was moved but not with {@code rows} rows.
    *
    * @throws  ErrorConditionException
    *          if there is an error repositioning the cursor.
    */
   public boolean forward(int rows)
   {
      if (rows == 0)
      {
         return true;
      }
      
      if (rows < 0)
      {
         return backward(-rows);
      }
      
      prepareReposition();
      
      if (results == null)
      {
         execute();
      }
      
      // Increment the requested number of rows by one if we are:
      // 1. on a record (i.e. not between rows) OR
      // 2. into front offend,
      // since the resulting position must be before the record *following*
      // the targeted jump.
      if (!betweenRows || (results.getRowNumber() == -1 && offEnd != OffEnd.BACK))
      {
         rows++;
      }
      
      boolean avail = results.scroll(rows);
      if (avail)
      {
         offEnd = OffEnd.NONE;
         betweenRows = true;
      }
      else
      {
         offEnd = OffEnd.BACK;
         betweenRows = false;
      }
      
      afterReposition(true);
      return avail;
   }
   
   /**
    * Move the current cursor position backward by the specified number of rows.  This will always
    * leave the cursor between two results.  For example, given a request to move 1 row backward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>, this would move it
    *       back <i>one half position</i>, to just before the current result;
    *   <li>if the cursor currently is positioned <i>between two results</i>, this would move it
    *       back <i>one position</i>, to just before the current result.
    * </ul>
    *
    * @param   rows
    *          Number of rows to scroll the cursor backward.
    *
    * @return  {@code true} if operation was successful and the cursor was moved the exact amount
    *          of rows as specified by the parameter. If {@code false} is returned then it is
    *          possible that cursor's position was moved but not with {@code rows} rows.
    *
    * @throws  ErrorConditionException
    *          if {@code rows} represents the unknown value; if there is an error repositioning
    *          the cursor.
    */
   public boolean backward(NumberType rows)
   {
      if (validateReposition(rows))
      {
         return backward(rows.intValue());
      }
      return false;
   }
   
   /**
    * Move the current cursor position backward by the specified number of rows.  This will always
    * leave the cursor between two results.  For example, given a request to move 1 row backward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>, this would move it
    *       back <i>one half position</i>, to just before the current result;
    *   <li>if the cursor currently is positioned <i>between two results</i>, this would move it
    *       back <i>one position</i>, to just before the current result.
    * </ul>
    *
    * @param   rows
    *          Number of rows to scroll the cursor backward.
    *
    * @return  {@code true} if operation was successful and the cursor was moved the exact amount
    *          of rows as specified by the parameter. If {@code false} is returned then it is
    *          possible that cursor's position was moved but not with {@code rows} rows.
    *
    * @throws  ErrorConditionException
    *          if there is an error repositioning the cursor.
    */
   public boolean backward(int rows)
   {
      if (rows == 0)
      {
         return true;
      }
      
      if (rows < 0)
      {
         return forward(-rows);
      }
      
      prepareReposition();
      
      if (results == null)
      {
         execute();
      }
      
      // Decrement the requested number of rows by one if we are:
      // 1. on a record (i.e. not between rows) AND
      // 2. NOT into back offend,
      // since the resulting position must be before the record *following*
      // the targeted jump.
      if (!betweenRows && offEnd != OffEnd.BACK)
      {
         rows--;
      }
      
      boolean avail = results.scroll(-rows);
      if (avail)
      {
         offEnd = OffEnd.NONE;
         betweenRows = true;
      }
      else
      {
         offEnd = OffEnd.FRONT;
         betweenRows = false;
      }
      
      afterReposition(true);
      return avail;
   }
   
   /**
    * Return the number of results found for this query.
    * <p>
    * Note that the first time this method is invoked, the database server
    * may have to scroll to the end of the result set, which may be a
    * time-consuming process for a large result set.
    *
    * @return  Size of result set or value if the query is not open.
    */
   public integer size()
   {
      if (results == null)
      {
         return new integer();
      }
      
      if (rowCount < 0)
      {
         int row = results.getRowNumber();
         results.last();
         rowCount = results.getRowNumber() + 1;
         
         if (row == -1)
         {
            if (offEnd == OffEnd.BACK)
            {
               results.next();
            }
            else
            {
               results.reset();
            }
         }
         else
         {
            results.setRowNumber(row);
         }
      }
      
      return (new integer(rowCount));
   }
   
   /**
    * Progress-compatible function that returns the 1-based index of the
    * current row in the result set.
    *
    * @return  Unknown value if the query is not open or result set is empty.
    *          1 if the query is positioned before the first record.
    *          A value 1 greater than the number of rows in the query result
    *          set if the query is positioned beyond the last record.
    *          Otherwise returns the number of rows in the query result set.
    *
    */
   public integer currentRow()
   {
      if (results == null ||
          (offEnd == OffEnd.NONE && results.getRowNumber() == -1 && !isScrolling()))
      {
         return (new integer());
      }
      else
      {
         int size = size().intValue();
         int row = results.getRowNumber();
         
         if (size == 0)
         {
            return (new integer());
         }
         else if (offEnd == OffEnd.BACK)
         {
            return (new integer(size + 1));
         }
         else if (row == -1)
         {
            // We are at the front offend. Return unknown value if the query
            // is an adaptive one and we haven't retrieved any results using 
            // it.
            return !isNativelyPreselect() && (offEnd != OffEnd.FRONT)
                   ? new integer()
                   : new integer(1);
         }
         else
         {
            return (new integer(row + 1));
         }
      }
   }
   
   /**
    * Get the 1-based index of the current row in the result set.
    *
    * @return  Current row or the unknown value if the query has not yet
    *          executed.
    */
   public integer currentRowImpl()
   {
      if (results == null || offEnd != OffEnd.NONE)
      {
         return (new integer());
      }
      
      return (new integer(results.getRowNumber() + 1));
   }
   
   /**
    * Indicate whether the query has run off either end of its result set,
    * due to scrolling beyond the last record or before the first record.
    *
    * @return  <code>true</code> if no record can be retrieved at the
    *          cursor's current position;  otherwise <code>false</code>.
    */
   public logical isOffEnd()
   {
      boolean isOffEnd = (offEnd != OffEnd.NONE);
      if (!isOffEnd && results == null)
      {
         return (new logical());
      }
      
      return (new logical(isOffEnd));
   }
   
   /**
    * Indicate whether the query has run off either end of its result set,
    * due to scrolling beyond the last record or before the first record.
    *
    * @return  <code>true</code> if no record can be retrieved at the
    *          cursor's current position;  otherwise <code>false</code>.
    */
   public boolean _isOffEnd()
   {
      boolean isOffEnd = (offEnd != OffEnd.NONE);
      if (!isOffEnd && results == null)
      {
         return false;
      }
      
      return isOffEnd;
   }
   
   /**
    * Indicate whether the query will retrieve full records, rather than primary keys only.
    * 
    * @return  <code>true</code> to return full records; <code>false</code> to return primary
    *          keys only.
    */
   public boolean isFullRecords()
   {
      return fullRecords;
   }
   
   /**
    * Force the query to retrieve full records, rather than primary keys only.
    */
   public void setFullRecords()
   {
      this.fullRecords = true;
   }
   
   /**
    * Reset the query by clearing its state to a known default.  This is
    * used to ensure that subsequent requests for FIRST and LAST record
    * retrievals operate on a known query/buffer state.
    * <p>
    * This implementation nulls out the current result set, so that the next
    * record retrieval request triggers a new server-side query.  Optionally,
    * it also resets resolvable substitution arguments and triggers where
    * expression arguments to be resolved, if a client-side where cause
    * expression is defined.
    *
    * @param   resolveArgs
    *          <code>true</code> if query substitution arguments should be
    *          resolved/reset as part of the reset;  <code>false</code> if
    *          existing argument values should be used for the next query
    *          execution.
    */
   public void reset(boolean resolveArgs)
   {
      resetResults();
      
      if (resolveArgs && resolvableArgs)
      {
         for (QueryComponent comp : components)
         {
            comp.resetArgs();
         }
      }
   }
   
   /**
    * Set each buffer backing a query component to unknown mode.
    *
    * @see     RecordBuffer#setUnknownMode
    */
   public void setUnknownRecord()
   {
      Iterator<QueryComponent> iter = components();
      while (iter.hasNext())
      {
         QueryComponent comp = iter.next();
         comp.getBuffer().setUnknownMode();
      }
   }
   
   /**
    * Get the number of tables joined by this query.
    *
    * @return  Number of joined tables.
    */
   public int getTableCount()
   {
      return components.size();
   }
   
   /**
    * Fetch the array of primary key IDs associated with the first result in
    * this query's result list.
    *
    * @return  Array of record IDs reflecting the first query result, or
    *          <code>null</code> if there is no such result.
    */
   public Object[] peekFirst()
   {
      execute();
      results.first();
      
      return getRow();
   }
   
   /**
    * Fetch the array of primary key IDs associated with the last result in
    * this query's result list.
    *
    * @return  Array of record IDs reflecting the last query result, or
    *          <code>null</code> if there is no such result.
    */
   public Object[] peekLast()
   {
      execute();
      results.last();
      
      return getRow();
   }
   
   /**
    * Fetch the array of primary key IDs associated with the next result in
    * this query's result list.
    *
    * @return  Array of record IDs reflecting the next query result, or
    *          <code>null</code> if there is no such result.
    */
   public Object[] peekNext()
   {
      execute();
      if (results.next())
      {
         return getRow();
      }
      else
      {
         return null;
      }
   }
   
   /**
    * Fetch the array of primary key IDs associated with the previous result
    * in this query's result list.
    *
    * @return  Array of record IDs reflecting the previous query result, or
    *          <code>null</code> if there is no such result.
    */
   public Object[] peekPrevious()
   {
      execute();
      if (results.previous())
      {
         return getRow();
      }
      else
      {
         return null;
      }
   }
   
   /**
    * Indicate whether this query preselects all of its results.
    *
    * @return  <code>true</code>.
    */
   public boolean isPreselect()
   {
      return true;
   }
   
   /**
    * Set the given results object into this query.
    *
    * @param   results
    *          New result set for this query.
    */
   public void setResults(Results results)
   {
      this.results = results;
   }
   
   /**
    * Access the record buffers manipulated by this query, in order from
    * left-most (outermost) table of the query's join (if any) to right-most
    * (innermost) table.
    * 
    * @return  Array of one or more record buffers.
    */
   @Override
   public RecordBuffer[] getRecordBuffers()
   {
      RecordBuffer[] buffers = new RecordBuffer[getTableCount()];
      Iterator<QueryComponent> iter = components();
      for (int i = 0; iter.hasNext(); i++)
      {
         buffers[i] = iter.next().getBuffer();
      }
      
      return buffers;
   }
   
   /**
    * Get the query substitution arguments associated with this query, if any.
    * 
    * @return  Substitution arguments, in order, or <code>null</code> if there are none.
    */
   @Override
   public Object[] getArgs()
   {
      return components.get(0).getRawArgs();
   }
   
   /**
    * Get the original, HQL where clause associated with this query.
    * 
    * @return  Original where clause, or <code>null</code> if there is none.
    */
   @Override
   public String getOriginalWhere()
   {
      return components.get(0).getOriginalWhere();
   }
   
   /**
    * Get the legacy natural join, if any, associated with this query.
    * 
    * @return  Legacy join object, or <code>null</code> if none.
    */
   @Override
   public AbstractJoin getJoin()
   {
      return components.get(0).getJoin();
   }
   
   /**
    * Force this query to operate in dynamic retrieval mode, if it supports this mode.
    * 
    * @throws  UnsupportedOperationException
    *          always; this query type does not support dynamic retrieval.
    */
   @Override
   public void forceDynamicOperation()
   {
      throw new UnsupportedOperationException("PreselectQuery has no dynamic operation mode!");
   }
   
   /**
    * Get the sort phrase associated with this query.
    * 
    * @return  Sort phrase.
    */
   public String getSortPhrase()
   {
      return sort;
   }
   
   /**
    * Create an adaptive query component based on the information in this query (which is
    * presumed to be a single-table query), and given a list of adaptive query components which
    * represent outer, nested query loops, which will perform a server-side join to the
    * innermost, nested query component.
    * 
    * @param   joinList
    *          List of adaptive components representing nested query loops which will contain
    *          the adaptive query component returned by this method. The last component in the
    *          list represents the nested query loop immediately containing the returned query
    *          component, with which the join will be made.
    * @param   query
    *          Adaptive query which will manage the server-side join of <code>joinList</code> and
    *          the returned query component.
    * @param   iteration
    *          Iteration type: FIRST, LAST, or NEXT.
    * @param   outer
    *          <code>true</code> if the join should be a left outer join. This type of join is
    *          not supported at the time of this writing, so this parameter should always be
    *          <code>false</code>. It is here to support a planned, future enhancement.
    * @param   fallback
    *          A compound query component to use in the event the multi-table adaptive query
    *          switches from preselect to dynamic retrieval mode.
    * 
    * @return  An adaptive query component suitable for a server-side join, based on the
    *          information in this query.
    */
   @Override
   public AdaptiveComponent makeAdaptiveServerJoinComponent(List<AdaptiveComponent> joinList,
                                                            AdaptiveQuery query,
                                                            int iteration,
                                                            boolean outer,
                                                            CompoundComponent fallback)
   {
      QueryComponent original = components.get(0);
      RecordBuffer buffer = original.getBuffer();
      AbstractJoin join = original.getJoin();
      Object[] args = original.getRawArgs();
      String where = original.getOriginalWhere();
      String indexInfo = original.getIndexInfo();
      LockType lockType = original.getLockType();
      
      QueryComponent.ServerJoinData sjd =
         QueryComponent.prepareServerJoinData(joinList, buffer, join, args);
      BufferReference inverse = (join != null ? join.getInverse() : null);
      AdaptiveComponent comp = new AdaptiveComponent(query,
                                                     buffer,
                                                     sjd.join,
                                                     where,
                                                     sort,
                                                     indexInfo,
                                                     lockType,
                                                     sjd.args,
                                                     iteration,
                                                     outer,
                                                     inverse);
      comp.setReferenceSubs(sjd.referenceSubs);
      comp.setRelatedBuffers(sjd.relatedBuffers);
      comp.setFallback(fallback);
      
      if (iteration == FIRST || iteration == LAST)
      {
         try
         {
            comp.setIteration(iteration, sort);
         }
         catch (PersistenceException exc)
         {
            // should only happen if we have a runtime programming error or conversion error
            // (or with bad, hand-written business logic)
            throw new IllegalArgumentException(exc);
         }
      }
      
      return comp;
   }
   
   /**
    * Create a preselect query component based on the information in this query (which is
    * presumed to be a single-table query), and given a list of preselect query components which
    * represent outer, nested query loops, which will perform a server-side join to the
    * innermost, nested query component.
    * 
    * @param   joinList
    *          List of preselect components representing nested query loops which will contain
    *          the preselect query component returned by this method. The last component in the
    *          list represents the nested query loop immediately containing the returned query
    *          component, with which the join will be made.
    * @param   iteration
    *          Iteration type: FIRST, LAST, or NEXT.
    * @param   outer
    *          <code>true</code> if the join should be a left outer join. This type of join is
    *          not supported at the time of this writing, so this parameter should always be
    *          <code>false</code>. It is here to support a planned, future enhancement.
    * 
    * @return  A preselect query component suitable for a server-side join, based on the
    *          information in this query.
    */
   @Override
   public QueryComponent makePreselectServerJoinComponent(List<QueryComponent> joinList,
                                                          int iteration,
                                                          boolean outer)
   {
      QueryComponent original = components.get(0);
      RecordBuffer buffer = original.getBuffer();
      AbstractJoin join = original.getJoin();
      Object[] args = original.getRawArgs();
      String where = original.getOriginalWhere();
      String indexInfo = original.getIndexInfo();
      LockType lockType = original.getLockType();
      
      QueryComponent.ServerJoinData sjd =
         QueryComponent.prepareServerJoinData(joinList, buffer, join, args);
      QueryComponent comp = new QueryComponent(this,
                                               buffer,
                                               sjd.join,
                                               where,
                                               indexInfo,
                                               lockType,
                                               sjd.args,
                                               iteration,
                                               outer);
      comp.setReferenceSubs(sjd.referenceSubs);
      comp.setRelatedBuffers(sjd.relatedBuffers);
      
      if (iteration == FIRST || iteration == LAST)
      {
         try
         {
            comp.setIteration(iteration, sort);
         }
         catch (PersistenceException exc)
         {
            // should only happen if we have a runtime programming error or conversion error
            // (or with bad, hand-written business logic)
            throw new IllegalArgumentException(exc);
         }
      }
      
      return comp;
   }
   
   /**
    * Indicate whether this query has a client-side where clause expression associated with it.
    * 
    * @return  <code>true</code> if there is a where expression; else <code>false</code>.
    */
   @Override
   public boolean hasWhereExpression()
   {
      return !whereExprList.isEmpty();
   }
   
   /**
    * Clean up query resources when this query's scope ends. This implementation cleans up any
    * open results and deregisters this session listener.
    * 
    * @see  P2JQuery#cleanup()
    */
   public void cleanup()
   {
      if (!closed)
      {
         if (results != null)
         {
            results.cleanup();
         }
         
         closed = true;
         
         if (regWithSession)
         {
            regWithSession = false;
            persistence.deregisterSessionListener(this);
         }
      }
   }
   
   /**
    * Open the prepared query and reposition it to the first result if the
    * query is being browsed.
    */
   public void open()
   {
      resetResults();
      
      setErrorIfNull(false);
      setLenientOffEnd(true);
      execute();
      
      super.open();
   }
   
   /**
    * Override parent's no-op implementation to track whether client code has
    * set this as a scrollable query.  While this setting does not explicitly
    * affect this query's ability to respond to reposition requests (this
    * implementation is inherently scrollable), it does provide a hint to the
    * implementation regarding the use and scope of this query.  This method
    * must be invoked before any results are retrieved.
    */
   public void setScrolling()
   {
      scrolling = true;
   }
   
   /**
    * Disable results list scrolling/repositioning (if it was enabled by
    * default). This method must be invoked before any results are retrieved. 
    */
   public void setNonScrolling()
   {
      scrolling = false;
   }
   
   /**
    * Get the off-end status of this query, which indicates whether the query
    * has run off either end of its results.
    * 
    * @return  One of the {@link OffEnd} constants <code>NONE</code>,
    *          <code>FRONT</code>, or <code>BACK</code>.
    */
   public OffEnd getOffEnd()
   {
      return offEnd;
   }
   
   /**
    * Conversion of INDEX-INFORMATION attribute (KW_IDX_INFO).
    *
    * Getter of INDEX-INFORMATION attribute
    *
    * @param    n
    *           An integer expression that evaluates to the level of join for which you want index
    *           information.
    *
    * @return   A character string consisting of a comma-separated list of the index or indexes
    *           the query uses at the level of join specified.
    */
   @LegacyAttribute(name = "INDEX-INFORMATION")
   public character indexInformation(NumberType n)
   {
      return new character(components.get(n.intValue() - 1).getIndexInfo());
   }
   
   /**
    * Indicate whether this query is a preselect query by its nature.
    *
    * @return  <code>true</code>.
    */
   public boolean isNativelyPreselect()
   {
      return true;
   }
   
   /**
    * Deletes the current row of a query's result list. Implements the 4GL
    * DELETE-RESULT-LIST-ENTRY() method.
    *
    * @param  allowBetweenRows
    *         If <code>true</code> then the cursor can be positioned between rows (the next
    *         row will be deleted). <code>false</code> for conventional DELETE-RESULT-LIST-ENTRY()
    *         mode where the cursor should be positioned on a row (otherwise <code>false</code> is
    *         returned).
    *
    * @return <code>true</code> on success.
    */
   @Override
   public logical deleteResultListEntry(boolean allowBetweenRows)
   {
      // In allowBetweenRows mode we may want to delete multiple records without fetching them,
      // so currentRowDeleted state is ignored.
      if (currentRowDeleted && !allowBetweenRows)
      {
         return new logical(false);
      }
      
      SimpleResults cached = cacheResults(results);
      if (results != cached)
      {
         setResults(cached);
      }
      
      if (!_isOffEnd() &&
          cached.getNumberOfLoadedRows() > 0 &&
          (allowBetweenRows || !betweenRows))
      {
         int row = cached.getRowNumber();
         if (row >= 0)
         {
            int width = getTableCount();
            Serializable[] rsIDs = new Serializable[width];
            for (int i = 0; i < width; i++)
            {
               rsIDs[i] = cached.getID(i);
            }
            
            repoCache.remove(rsIDs);
         }
         currentRowDeleted = true;
         cached.deleteRow();
         betweenRows = true;
         rowCount--;
         return new logical(true);
      }
      
      return new logical(false);
   }

   /**
    * Indicate whether this query's <code>scrolling</code> flag has been set.
    * 
    * @return  <code>true</code> if this is a scrolling query;  else
    *          <code>false</code>.
    */
   public boolean isScrolling()
   {
      return scrolling || isNativelyScrolling();
   }
   
   /**
    * Indicate whether this query type is scrolling by its nature.
    * <p>
    * Subclasses which are not natively scrolling should override this method.
    * 
    * @return  <code>true</code>.
    */
   protected boolean isNativelyScrolling()
   {
      return true;
   }
   
   /**
    * Discard the current result set, clear the reposition cache, and reset
    * result-related state to default values.
    */
   protected void resetResults()
   {
      if (results != null)
      {
         results.cleanup();
      }
      setResults(null);
      offEnd = OffEnd.NONE;
      betweenRows = true;
      rowCount = -1;
      clearRepoCache();
      currentRowDeleted = false;
   }
   
   /**
    * Get an iterator on the query components.
    *
    * @return  Iterator on components.
    */
   protected Iterator<QueryComponent> components()
   {
      return components.iterator();
   }
   
   /**
    * Indicate whether this query will throw a
    * <code>QueryOffEndException</code> when it runs off the end of its
    * results.
    *
    * @return  <code>true</code> if an exception will be thrown, else
    *          <code>false</code>.
    */
   protected boolean isLenientOffEnd()
   {
      return lenientOffEnd;
   }
   
   /**
    * Indicate whether a call to move to the previous record in the result
    * list should be treated as a call to move to the last record in the list.
    * This will be the case if either there are no results available, or if
    * the query is not off the front end of its results, but the row number
    * is -1 (indicating no results have yet been retrieved).
    * 
    * @return  <code>true</code> to treat a previous request as a request to
    *          move to the last record in the list;  else <code>false</code>.
    */
   protected boolean treatPreviousAsLast()
   {
      return (results == null || (offEnd != OffEnd.FRONT && results.getRowNumber() == -1));
   }
   
   /**
    * Walk through each component and set its backing buffer to unknown mode,
    * such that the invocation of any of its DMO-backed methods returns
    * unknown value.
    */
   protected void setEmptyBuffersUnknown()
   {
      Iterator<QueryComponent> iter = components();
      while (iter.hasNext())
      {
         RecordBuffer buffer = iter.next().getBuffer();
         if (buffer.getCurrentRecord() == null)
         {
            buffer.setUnknownMode();
         }
      }
   }
   
   /**
    * Given a primary DMO buffer and an inverse DMO buffer which together
    * define a natural join, create the <code>AbstractJoin</code> helper
    * object which is used to implement the join.
    * <p>
    * It is intended that this method be called when adding a query component.
    * If no previous components have been added, a <code>DynamicJoin</code>
    * is created, and it is assumed that the inverse DMO is provided as a
    * query substitution parameter.  If a component previously has been added,
    * a <code>ServerJoin</code> is created, and it is assumed that the inverse
    * DMO will be resolved by a server-side join, rather than being provided
    * as a query substitution parameter.
    * <p>
    * If no inverse buffer is provided, no join is created.
    *
    * @param   primary
    *          Buffer reference to the primary DMO.
    * @param   inverse
    *          Buffer reference to the joined DMO, or <code>null</code> if no
    *          join is required.
    *
    * @return  A join helper object, or <code>null</code> if no natural join
    *          is required.
    */
   protected AbstractJoin processJoin(BufferReference primary, BufferReference inverse)
   {
      if (inverse == null)
      {
         return null;
      }
      
      AbstractJoin join = null;
      
      if (getTableCount() == 0)
      {
         // If we have an inverse proxy and this is the first component being
         // added, the inverse is being provided as a query substitution
         // parameter.
         join = Persistence.isForeignKeysEnabled()
                ? new DynamicJoin(primary.buffer(), inverse.buffer())
                : new DynamicLegacyKeyJoin(primary.buffer(), inverse.buffer());
      }
      else
      {
         // If we have an inverse proxy and this is not the first component
         // being added, the inverse is provided as a related DMO from a
         // related table.  In this case, the inverse was not provided as a
         // query substitution parameter.
         join = Persistence.isForeignKeysEnabled()
                ? new ServerJoin(primary.buffer(), inverse.buffer())
                : new ServerLegacyKeyJoin(primary.buffer(), inverse.buffer());
      }
      
      return join;
   }
   
   /**
    * Assemble the select clause for an HQL projection query, where the only
    * field retrieved for each buffer is the primary key. This uses the
    * Hibernate reserved keyword "id" to represent the primary key. Form of
    * the clause is:
    * <pre>
    *    select {buffer0.id} [, {buffer1.id} [...]]
    * </pre>
    *
    * @param   buf
    *          String buffer into which clause is assembled.
    */
   protected void assembleSelectClause(StringBuilder buf)
   {
      boolean needSelect = !fullRecords;
      
      Iterator<QueryComponent> iter = components();
      while (!needSelect && iter.hasNext())
      {
         QueryComponent comp = iter.next();
         if (!isClientWhere() && !fullRecords && comp.isIdOnly())
         {
            needSelect = true;
         }
      }
      
      if (!needSelect)
      {
         return;
      }
      
      buf.append("select ");
      
      iter = components();
      for (int i = 0; iter.hasNext(); i++)
      {
         QueryComponent comp = iter.next();
         if (i > 0)
         {
            buf.append(", ");
         }
         
         RecordBuffer buffer = comp.getBuffer();
         buf.append(buffer.getDMOAlias());
         if (!isClientWhere() && !fullRecords && comp.isIdOnly())
         {
            buf.append(".id");
         }
      }
      
      buf.append(" ");
   }
   
   /**
    * Assemble the from clause for an HQL query. The class name of each
    * DMO is decapitalized to produce an alias for that object. Form of
    * the clause is:
    * <pre>
    *    from {Buffer0} as {buffer0} [, {Buffer1} as {buffer1} [, ...]]
    * </pre>
    * <p>
    * <b>Known Limitation:</b>
    * At this time, only theta-style joins are supported and as such, only
    * inner joins are supported (there is no way to specify an outer join
    * using theta-join syntax in HQL).
    *
    * @param   buf
    *          String buffer into which clause is assembled.
    * @param   sortCriteria
    *          List of <code>SortCriterion</code> objects which describe the
    *          sort behavior for this query.
    * 
    * @throws  PersistenceException
    *          if there is an error creating the <code>HQLPreprocessor</code>
    *          from which the composite join data are gathered.
    */
   protected void assembleFromClause(StringBuilder buf, List<SortCriterion> sortCriteria)
   throws PersistenceException
   {
      buf.append("from ");
      
      Iterator<QueryComponent> iter = components();
      for (int i = 0; iter.hasNext(); i++)
      {
         QueryComponent comp = iter.next();
         RecordBuffer buffer = comp.getBuffer();
         if (i > 0)
         {
            buf.append(", ");
         }
         buf.append(buffer.getDMOImplementationName());
         buf.append(" as ");
         buf.append(buffer.getDMOAlias());
      }
      
      buf.append(" ");
      
      // add any ANSI joins required by the HQL preprocessor of each component.
      iter = components();
      while (iter.hasNext())
      {
         QueryComponent comp = iter.next();
         Iterator<String> joinsIter = comp.compositeJoins();
         while (joinsIter.hasNext())
         {
            buf.append(joinsIter.next());
            buf.append(" ");
         }
      }
      
      // add any ANSI joins required by sort criteria.
      if (sortCriteria != null)
      {
         for (SortCriterion crit : sortCriteria)
         {
            String join = crit.getAnsiJoin();
            if (join != null)
            {
               buf.append(join);
               buf.append(" ");
            }
         }
      }
   }
   
   /**
    * Assemble the restriction (i.e., where) clause for an HQL query. If more
    * than one component contains a where clause, they are combined with a
    * logical "and". Form of the clause is:
    * <pre>
    *    where ({component0_clause})
    *     [and ({component1_clause})
    *     [...]]
    * </pre>
    * <p>
    * Please refer to the Hibernate reference documentation for permissible
    * syntax of this clause.
    *
    * @param   sortCriteria
    *          List of <code>SortCriterion</code> objects which describe the
    *          sort behavior for this query.
    *
    * @return  assembled where clause.
    *
    * @throws  PersistenceException
    *          if there is an error creating the <code>HQLPreprocessor</code>
    *          from which the preprocessed where clause is retrieved for each
    *          query component.
    */
   protected HQLExpression assembleWhereClause(List<SortCriterion> sortCriteria)
   throws PersistenceException
   {
      HQLExpression whereClause = new HQLExpression();
      
      Iterator<QueryComponent> iter = components();
      while (iter.hasNext())
      {
         QueryComponent next = iter.next();
         RecordBuffer rb = next.getBuffer();
         boolean multiplex = rb.isMultiplexed();
         AbstractJoin join = next.getJoin();
         HQLExpression subClause = next.getWhere();
         if (subClause != null)
         {
            subClause.trim();
         }
         boolean hasJoin = (join != null);
         boolean hasWhere = subClause != null && !subClause.isEmpty();
         
         // a FIRST/LAST component with a server join will specify the join in a subselect, so
         // we don't want the join HQL to appear in the main where clause
         if (hasJoin && join.isServerJoin())
         {
            int iteration = next.getIteration();
            hasJoin = (iteration != FIRST && iteration != LAST);
         }
         
         boolean hasContent = (hasJoin || multiplex || hasWhere);
         
         if (hasContent)
         {
            if (whereClause.isEmpty())
            {
               whereClause.append("where (");
            }
            else
            {
               whereClause.append("and (");
            }
         }
         
         // Determine whether where clause should be augmented with an
         // equality test for a multiplex ID.  We need to do this when:
         // a) we are querying a multiplexed temp table;  OR
         // b) we are querying any temp table and have a non-empty where
         //    clause
         if (multiplex)
         {
            if (hasJoin || hasWhere)
            {
               whereClause.append("(");
            }
            
            whereClause.append(rb.getDMOAlias());
            whereClause.append(".");
            whereClause.append(TemporaryBuffer.MULTIPLEX_FIELD_NAME);
            whereClause.append(" = ", true);
            
            if (hasJoin || hasWhere)
            {
               whereClause.append(") and ");
            }
         }
         
         // Add subclause for a join.
         if (hasJoin)
         {
            if (multiplex || hasWhere)
            {
               whereClause.append("(");
            }
            
            whereClause.append(join.getHQL());
            
            if (multiplex || hasWhere)
            {
               whereClause.append(")");
            }
            
            if (hasWhere)
            {
               whereClause.append(" and ");
            }
         }
         
         if (hasWhere)
         {
            if (hasJoin || multiplex)
            {
               whereClause.append("(");
            }
            
            whereClause.append(subClause);
            
            if (hasJoin || multiplex)
            {
               whereClause.append(")");
            }
         }
         
         if (hasContent)
         {
            whereClause.append(") ");
         }
      }
      
      if (sortCriteria != null)
      {
         for (SortCriterion crit : sortCriteria)
         {
            String rest = crit.getCompositeRestrictor();
            if (rest != null)
            {
               if (whereClause.isEmpty())
               {
                  whereClause.append("where (");
               }
               else
               {
                  whereClause.append("and (");
               }
               whereClause.append(rest);
               whereClause.append(") ");
            }
         }
      }
      
      return whereClause;
   }
   
   /**
    * Assemble the <code>order by</code> clause for an HQL query.  Uses the
    * form:
    * <pre>
    *   order by {{dmo.property} {asc/desc}} [, ...]
    * </pre>
    * <p>
    * Any text-based sort criteria are augmented with the necessary upper()
    * and rtrim() functions.
    *
    * @param   buf
    *          String buffer into which clause is assembled.
    *
    * @return  List of <code>SortCriterion</code> objects which describe the
    *          sort behavior for this query.
    *
    * @throws  PersistenceException
    *          if the raw sort clause provided at construction contains any
    *          unqualified property names;  if the raw sort clause contains
    *          a reference to an unrecognized record buffer.
    */
   protected List<SortCriterion> assembleOrderByClause(StringBuilder buf)
   throws PersistenceException
   {
      if (sort == null)
      {
         return null;
      }
      
      buf.append("order by ");
      
      return assembleOrderByClauseNoPreamble(buf, sort);
   }
   
   /**
    * Assemble the <code>order by</code> clause for an HQL query, sans the
    * &quot;order by&quot; preamble.  Uses the
    * form:
    * <pre>
    *   {{dmo.property} {asc/desc}} [, ...]
    * </pre>
    * <p>
    * Any text-based sort criteria are augmented with the necessary upper()
    * and rtrim() functions.
    *
    * @param   buf
    *          String buffer into which clause is assembled.
    * @param   sort
    *          Raw sort clause to parse and augment if necessary.
    *
    * @return  List of <code>SortCriterion</code> objects which describe the
    *          sort behavior for this query.
    *
    * @throws  PersistenceException
    *          if <code>sort</code> contains any unqualified property names;
    *          if <code>sort</code> clause contains a reference to an
    *          unrecognized record buffer.
    */
   protected List<SortCriterion>
   assembleOrderByClauseNoPreamble(StringBuilder buf, String sort)
   throws PersistenceException
   {
      List<SortCriterion> sortCriteria = buildSortCriteria(sort);
      
      // Assemble the order by clause, given the list of SortCriterion objects.
      return assembleOrderByClauseNoPreamble(buf, sortCriteria);
   }
   
   /**
    * Returns a list <code>SortCriterion</code> objects build from the given
    * sort clause.
    *
    * @param   sort
    *          Raw sort clause to parse and augment if necessary.
    *
    * @return  List of <code>SortCriterion</code> objects which describe the
    *          sort behavior for this query.
    *
    * @throws  PersistenceException
    *          if <code>sort</code> contains any unqualified property names;
    *          if <code>sort</code> clause contains a reference to an
    *          unrecognized record buffer.
    */
   protected List<SortCriterion> buildSortCriteria(String sort)
   throws PersistenceException
   {
      // Map record buffers used in the sort clause to their identifying
      // aliases.
      Map<String, RecordBuffer> aliases = new HashMap<>();
      Iterator<QueryComponent> iter = components();
      while (iter.hasNext())
      {
         RecordBuffer buffer = iter.next().getBuffer();
         aliases.put(buffer.getDMOAlias(), buffer);
      }
      
      // Parse the sort clause, extracting each qualified DMO property
      // reference, matching it to its buffer (by alias), and creating a
      // SortCriterion object for it.  Also group lists of SortCriterion
      // objects by buffer, so that each query component can store its portion
      // of the sort clause.
      Map<RecordBuffer, List<SortCriterion>> subSortMap = new HashMap<>();
      List<SortCriterion> sortCriteria = new ArrayList<>();
      String[] parts = sort.split(",");
      for (String part1 : parts)
      {
         String part = part1.trim();
         int pos = part.indexOf(".");
         if (pos < 0)
         {
            throw new PersistenceException(
               "Order by clause contains unqualified property name:  " + part);
         }
         
         // Sanity.
         String alias = part.substring(0, pos);
         RecordBuffer buffer = aliases.get(alias);
         if (buffer == null)
         {
            throw new PersistenceException(
               "Order by clause references an unrecognized buffer:  " +alias + " (" + part + ")");
         }
         
         // Find or create the sub-list in which to store SortCriterion
         // objects for the current buffer.
         List<SortCriterion> subSortCrit = subSortMap.get(buffer);
         if (subSortCrit == null)
         {
            subSortCrit = new ArrayList<>();
            subSortMap.put(buffer, subSortCrit);
         }
         
         // Create the SortCriterion instance and add it to the master list
         // and to the appropriate sub-list.
         
         String schema = buffer.getSchema();
         Class<?> dmoIface = buffer.getDMOInterface();
         Class<?> dmoClass = buffer.getDMOImplementationClass();
         
         SortCriterion crit = new SortCriterion(buffer, part, schema, dmoIface, dmoClass);
         
         sortCriteria.add(crit);
         subSortCrit.add(crit);
      }
      
      // Match the sort criteria sub-list for each buffer, generate a normalized sort clause for
      // that sub-list, and store it in the appropriate query component.
      for (Map.Entry<RecordBuffer, List<SortCriterion>> next : subSortMap.entrySet())
      {
         RecordBuffer buffer = next.getKey();
         QueryComponent comp = findComponent(buffer);
         assert (comp != null);
         List<SortCriterion> subSortCrit = next.getValue();
         String subSort = SortCriterion.toRawSortPhrase(subSortCrit);
         SortIndex sortIndex = SortIndex.get(subSort, buffer);
         comp.setSortIndex(sortIndex);
      }
      
      return sortCriteria;
   }
   
   /**
    * Assemble the <code>order by</code> clause for an HQL query, sans the
    * &quot;order by&quot; preamble.  Uses the
    * form:
    * <pre>
    *   {{dmo.property} {asc/desc}} [, ...]
    * </pre>
    * <p>
    * Any text-based sort criteria are augmented with the necessary upper()
    * and rtrim() functions.
    *
    * @param   buf
    *          String buffer into which clause is assembled.
    * @param   sortCriteria
    *          A list of <code>SortCriterion</code> objects which describe
    *          the sorting to be applied.
    *
    * @return  List of <code>SortCriterion</code> objects which describe the
    *          sort behavior for this query.
    */
   protected List<SortCriterion>
   assembleOrderByClauseNoPreamble(StringBuilder buf, List<SortCriterion> sortCriteria)
   {
      RecordBuffer buffer = getRecordBuffers()[0];
      boolean multiplex = (buffer.getMultiplexID() != null);
      List<SortCriterion> workList = null;
      
      if (sortCriteria == null)
      {
         if (multiplex)
         {
            workList = new ArrayList<>(1);
         }
         else
         {
            return null;
         }
      }
      
      if (workList == null)
      {
         workList = new ArrayList<>(sortCriteria.size() + 1);
      }
      
      workList.addAll(sortCriteria);
      
      if (multiplex)
      {
         String schema = buffer.getSchema();
         Class<?> dmoIface = buffer.getDMOInterface();
         Class<?> dmoClass = buffer.getDMOImplementationClass();
         
         StringBuilder sb = new StringBuilder();
         sb.append(buffer.getDMOAlias());
         sb.append(".");
         sb.append(TemporaryBuffer.MULTIPLEX_FIELD_NAME);
         sb.append(" ");
         boolean ascending = true;
         int insertAt = sortCriteria.size();
         if (!sortCriteria.isEmpty())
         {
            // Make sure multiplex ID sort direction matches that of the first
            // sort component, otherwise the matching index will not be used
            // by the DB.
            SortCriterion crit = sortCriteria.get(0);
            ascending = crit.isAscending();
            
            int last = sortCriteria.size() - 1;
            crit = sortCriteria.get(last);
            if ("id".equals(crit.getUnqualifiedName()))
            {
               // If the primary key is the final sort criterion, we want to
               // insert the multiplex criterion just before it.
               insertAt--;
            }
         }
         sb.append(ascending ? "asc" : "desc");
         try
         {
            SortCriterion multiCrit = new SortCriterion(buffer,
                                                        sb.toString(),
                                                        schema,
                                                        dmoIface,
                                                        dmoClass);
            workList.add(insertAt, multiCrit);
         }
         catch (PersistenceException exc)
         {
            LOG.log(Level.SEVERE, "Error adding multiplex ID to order by clause", exc);
         }
      }
      
      Iterator<SortCriterion> iter = workList.iterator();
      for (int i = 0; iter.hasNext(); i++)
      {
         SortCriterion crit = iter.next();
         
         if (i > 0)
         {
            buf.append(", ");
         }
         
         buf.append(crit.toSortExpression(false));
      }
      
      return sortCriteria;
   }
   
   /**
    * Get this query's client-side where expressions.
    *
    * @return  Unmodifiable where expression list.
    */
   protected List<Supplier<logical>> getWhereExpressions()
   {
      return Collections.unmodifiableList(whereExprList);
   }
   
   /**
    * Get an iterator on this query's client-side where expressions.
    *
    * @return  Where expression iterator.
    */
   protected Iterator<Supplier<logical>> whereExpressions()
   {
      return getWhereExpressions().iterator();
   }
   
   /**
    * Compose an HQL query statement to represent the combination of all query
    * components, execute it, and store the scrollable results.  If query has
    * been executed previously and we already have a result set, simply
    * return.
    *
    * @return  <code>true</code> if query was executed for the first time;
    *          <code>false</code> if query was executed previously and we
    *          already have a result set.
    *
    * @throws  ErrorConditionException
    *          if an error occurs executing the query.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   protected boolean execute()
   {
      if (results != null)
      {
         return false;
      }
      
      if (components.isEmpty())
      {
         throw new IllegalStateException(
            "Preselect query must contain at least one query component");
      }
      
      if (!verifyJoins())
      {
         return true;
      }
      
      Persistence persistence = getPersistence();
      
      if (!regWithSession)
      {
         // Register to receive session closing notification.
         persistence.registerSessionListener(this, SessionListener.Scope.NONE);
         regWithSession = true;
      }
      
      Results r = null;
      try
      {
         // assemble HQL statement
         hql = assembleHQL();
         
         // Prepare substitution parameters.
         List<Object> argList = new ArrayList<>();
         List<Type> typeList = new ArrayList<>();
         prepareParameters(argList, typeList);
         Object[] args = argList.toArray();
         Type[] types = typeList.toArray(new Type[typeList.size()]);
         DBUtils.makeTypeArray(args, types);
         
         if (DatabaseStatistics.isEnabled())
         {
            Database database = getRecordBuffers()[0].getDatabase();
            DatabaseStatistics.get().queryExecuted(this, database, hql);
         }
         
         if (LOG.isLoggable(Level.FINE))
         {
            LOG.log(Level.FINE, "HQL:  " + hql);
            StringBuilder buf = new StringBuilder("PARMS:  {");
            for (int i = 0; i < args.length; i++)
            {
               if (i > 0)
               {
                  buf.append(", ");
               }
               Object next = args[i];
               if (next instanceof BaseDataType)
               {
                  buf.append(((BaseDataType) next).toStringMessage());
               }
               else
               {
                  buf.append(String.valueOf(next));
               }
            }
            buf.append("}");
            LOG.log(Level.FINE, buf.toString());
         }
         
         // possibly leak uncommitted updates to other sessions
         Iterator<QueryComponent> iter = components();
         while (iter.hasNext())
         {
            QueryComponent next = iter.next();
            HQLPreprocessor hqlPreproc = next.getHQLPreprocessor();
            List<String> earlyPublish = hqlPreproc.getEarlyPublishEntities();
            if (!earlyPublish.isEmpty())
            {
               RecordBuffer buffer = next.getBuffer();
               DirtyShareContext dirtyContext = buffer.getDirtyContext();
               List<Persistable> dmos = dirtyContext.updateSnapshots(earlyPublish, persistence);
               
               // Make sure DMOs do not bloat the Hibernate session.
               for (Persistable dmo : dmos)
               {
                  buffer.evictDMOIfUnused(dmo);
               }
            }
         }
         
         Serializable templateRowid = getTemplateQueryRowid(args);
         if (templateRowid != null)
         {
            r = new TemplateResults(templateRowid);
            setResults(r);
         }
         else
         {
            // execute the primary query
            r = executeQuery(persistence, hql, args, types);
            r = createResults(r);
            setResults(r);
            
            registerCleaner();
         }
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
      finally
      {
         rowCount = -1;
      }
      
      return true;
   }
   
   /**
    * Indicate whether results resources should be cleaned up upon exiting
    * the global scope or upon exiting the nearest enclosing scope in which
    * the results were created.
    * <p>
    * This default implementation returns <code>false</code>, indicating
    * results should be cleaned up at the nearest enclosing scope.
    * Subclasses which require different behavior must override this method.
    *
    * @return  <code>false</code>.
    */
   protected boolean isTopLevelCleanup()
   {
      return false;
   }
   
   /**
    * Invoked just before query execution, this method checks whether all
    * components with a non-null JOIN have a valid inverse on which to perform
    * the join.  If not, set the query's results to an empty result set and
    * indicate failure.
    * 
    * @return  <code>true</code> if verification succeeded and query execution
    *          should continue;  else <code>false</code>
    */
   protected boolean verifyJoins()
   {
      Iterator<QueryComponent> iter = components();
      while (iter.hasNext())
      {
         QueryComponent comp = iter.next();
         AbstractJoin join = comp.getJoin();
         
         if (join != null && !join.isServerJoin())
         {
            BufferReference ref = join.getInverse();
            Persistable record = ref.buffer().getCurrentRecord();
            
            if (record == null)
            {
               handleMissingJoinRecord();
               
               return false;
            }
         }
      }
      
      return true;
   }
   
   /**
    * Respond to the condition that a component with a non-null JOIN has a
    * missing inverse on which to perform the join.
    * <p>
    * This implementation creates an empty result set.
    */
   protected void handleMissingJoinRecord()
   {
      setResults(new SimpleResults(null));
   }
   
   /**
    * Attempt to cache all results from the specified target result set into
    * a {@link SimpleResults} instance.  If this is a scrolling query, this
    * will cache all results in the target result set, including those already
    * visited.  Otherwise, only the current result and all remaining,
    * unvisited results are cached.
    * 
    * @param   target
    *          Result set which is to be cached, presumably because it will
    *          be invalid momentarily.  If this is already an instance of
    *          <code>SimpleResults</code>, it is simply returned.
    * 
    * @return  Results from <code>target</code>, if any, cached in an instance
    *          of <code>SimpleResults</code>.  If there were no results
    *          available to cache, an empty result set is returned.  If any of
    *          the record buffers backing this query have been closed,
    *          <code>null</code> is returned.
    */
   protected SimpleResults cacheResults(Results target)
   {
      if (target instanceof SimpleResults)
      {
         // Results are already cached.
         return (SimpleResults) target;
      }
      
      // If any buffer associated with this query is closed, do not
      // attempt to cache results.  Instead, clean up resources;  this
      // query is done.
      Iterator<QueryComponent> iter = components();
      while (iter.hasNext())
      {
         QueryComponent next = iter.next();
         RecordBuffer buffer = next.getBuffer();
         if (buffer.getOpenScopeCount() < 1)
         {
            resetResults();
            
            return null;
         }
      }
      
      // Remember current row.
      int currentRow = (isScrolling() ? target.getRowNumber() : 0);
      
      // Cache all preselected results in a SimpleResults object and set
      // that as the query's new backing results.
      ArrayList<Object[]> rows = new ArrayList<>(64);
      
      if (isScrolling())
      {
         target.reset();
         while (target.next())
         {
            rows.add(getRow());
         }
      }
      else if (target.get() != null)
      {
         do
         {
            rows.add(getRow());
         } while (target.next());
      }
      else if (target.getRowNumber() == -1)
      {
         currentRow = -1;
         while (target.next())
         {
            rows.add(getRow());
         }
      }
      
      rows.trimToSize();
      
      // Override SimpleResults.cleanup() to do nothing, so we don't
      // immediately discard results when the current scope ends and our
      // Finalizable cleaner (installed in execute()) does its work.
      SimpleResults cached = new SimpleResults(rows)
      {
         public void cleanup() {}
      };
      
      // Reset to current row.
      cached.setRowNumber(currentRow);
      
      return cached;
   }
   
   /**
    * Prepare query substitution parameters for this query.  One argument
    * must be added to each of the <code>args</code> and <code>types</code>
    * lists for each <code>?</code> placeholder in the HQL query string for
    * this query.
    *
    * @param   args
    *          Argument list to be populated.
    * @param   types
    *          Type list to be populated.  This may be populated with
    *          <code>null</code> values, but it must coincide in size and in
    *          the absolute positions of its elements with the
    *          <code>args</code> list.  Any <code>null</code> element will
    *          be filled in later.
    * 
    * @throws  PersistenceException
    *          if there is an error creating an <code>HQLPreprocessor</code>
    *          during the gathering of query substitution parameters.
    */
   protected void prepareParameters(List<Object> args, List<Type> types)
   throws PersistenceException
   {
      for (QueryComponent next : components)
      {
         RecordBuffer rb = next.getBuffer();
         if (rb.isMultiplexed())
         {
            args.add(rb.getMultiplexID());
            types.add(null);
         }
         
         AbstractJoin join = next.getJoin();
         if (join != null)
         {
            List<Serializable> parameters = join.getParameters();
            List<Type> parameterTypes = join.getParameterTypes();
            
            if (parameterTypes != null)
            {
               for (int i = 0; i < parameterTypes.size(); i++)
               {
                  Type paramType = parameterTypes.get(i);
                  if (paramType != null)
                  {
                     args.add(parameters.get(i));
                     types.add(paramType);
                  }
               }
            }
         }
         
         Object[] cArgs = next.getArgs();
         if (cArgs != null)
         {
            if (parameterFilter != null)
            {
               cArgs = parameterFilter.filterParameters(cArgs);
            }
            
            for (Object arg : cArgs)
            {
               if (arg instanceof Resolvable)
               {
                  args.add(((Resolvable) arg).resolve());
               }
               else
               {
                  args.add(arg);
               }
               types.add(null);
            }
         }
      }
   }
   
   /**
    * Collect the names of the entities (DMO types) associated with this query.
    * 
    * @return  Array of entity names.
    */
   protected String[] getEntities()
   {
      int len = components.size();
      String[] entities = new String[len];
      for (int i = 0; i < len; i++)
      {
         entities[i] = components.get(i).getBuffer().getEntityName();
      }
      
      return entities;
   }
   
   /**
    * Checks if this is a template query. A template query is a rowid query that looks for a
    * special template record that is not found in database. The template record is a record
    * having all fields set to default values. The id of such record is in P2J, by convention
    * negative.
    * In case this is a simple template query, this method returns the rowid of the template for
    * current buffer from the single.
    * <p>
    * <b>Note:</b><br>
    * The current implementation identifies only simple queries for template records:
    * <ol>
    *    <li>this {@code PreselectQuery} has a single component;</li>
    *    <li>the where predicate directly test whether the rowid/recid is equal to a value
    *          (this is done by the {@link HQLPreprocessor} of the component);</li>
    *    <li>the value (constant or passed in as query SUBST parameter) is negative (as noted
    *          above).</li>
    * </ol>
    * Although this kind of queries are sufficient for standard template record load according to
    * P4GL idiom, more complex queries can be issued, with predicates that basically match the
    * pattern but add supplementary AND/OR clauses. This kind of queries are not always of common
    * sense and the result in 4GL is not always as expected.<br>
    * For example:
    *       <p>{@code FOR EACH book WHERE RECID(book) = templateId OR book.isbn NE ?}</p>
    * will iterate only books that have the isbn not null, but will not contain the template
    * record in the result returned.
    * <p>
    * On the other hand, it does not make much sense to:
    * <ul>
    *    <li>iterate over the the template records of a table since there will always be at most
    *       one such record;</li>
    *    <li>add additional OR/AND clauses in WHERE predicate since, if the template record
    *       exists, it is well defined by it rowid, configured in database schema and
    *       immutable.</li>
    * </ul>
    * 
    * @param   args
    *          The actual parameters of the query.
    * 
    * @return  if this query looks for the template record of the component the rowid is returned
    *          and {@code null} otherwise. 
    * 
    */
   protected Serializable getTemplateQueryRowid(Object[] args)
   {
      if (components == null || components.size() != 1)
      {
         return null;
      }
      // TODO: need some attention:
      // - this query type is not only used for FOR EACH constructs. It is the base class for
      //       other query types as well. Need to check if template record changes applied to all
      //       these query types;
      
      HQLPreprocessor prep = null;
      try
      {
         prep = components.get(0).getHQLPreprocessor();
      }
      catch (PersistenceException e)
      {
         return null;
      }
      
      if (!prep.isFindByRowid())
      {
         return null;
      }
      
      long tId = prep.getFindByRowid(args);
      return tId < 0 ? tId : null;
   }
   
   /**
    * Execute the query to create a first level result set.
    * <p>
    * This default implementation returns an object which is a thin wrapper around
    * {@code ScrollableResults}.  Subclasses may override this method if they need a
    * {@code Results} implementation other than {@code ScrollingResults}.
    * <p>
    * In case when the {@code hql} represents a valid template record query (a test on recid /
    * rowid of the record against the unique {@code _file._template} for current record of the
    * single {@code QueryComponent}), a special {@link Results} object is returned with a single
    * element that is the template record for the buffer of the component.
    *
    * @param   persistence
    *          Object which is used to execute the query.
    * @param   hql
    *          HQL query string.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within within the HQL
    *          query string.
    * @param   types
    *          Query substitution parameter types.
    *
    * @return  An instance of {@code ScrollingResults} which wrappers the 
    *          {@code ScrollableResults} object returned by the execution of the query.
    *
    * @throws  PersistenceException
    *          if there is an error executing the query.
    */
   protected Results executeQuery(Persistence persistence,
                                  String hql,
                                  Object[] args,
                                  Type[] types)
   throws PersistenceException
   {
      Serializable templateRowid = getTemplateQueryRowid(args);
      if (templateRowid != null)
      {
         return new TemplateResults(templateRowid);
      }
      
      boolean pop = false;
      try
      {
         persistence.pushImplicitTransaction();
         pop = true;
         
         return new ScrollingResults(persistence, executeScroll(persistence, hql, args, types));
      }
      finally
      {
         if (pop)
         {
            persistence.popImplicitTransaction(null);
         }
      }
   }
   
   
   /**
    * Execute the query to create a <code>ScrollableResults</code> result set.
    *
    * @param   persistence
    *          Object which is used to execute the query.
    * @param   hql
    *          HQL query string.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          within the HQL query string.
    * @param   types
    *          Query substitution parameter types.
    *
    * @return  The <code>ScrollableResults</code> object returned by the
    *          execution of the query.
    *
    * @throws  PersistenceException
    *          if there is an error executing the query.
    */
   protected ScrollableResults executeScroll(Persistence persistence,
                                             String hql,
                                             Object[] args,
                                             Type[] types)
   throws PersistenceException
   {
      ScrollMode scrollMode = isScrolling()
                              ? ScrollMode.SCROLL_INSENSITIVE
                              : ScrollMode.FORWARD_ONLY;
      
      return persistence.scroll(getEntities(), hql, args, types, 0, 0, scrollMode);
   }
   
   /**
    * Optionally wrapper the result set created for the query with a filtering
    * <code>Results</code> implementation (for client-side where processing).
    * It is this object which actually will be used to retrieve the primary
    * key values of records to be loaded from the database.
    * <p>
    * If no filtering is necessary, simply return the results object which
    * was passed in.
    * <p>
    * Subclasses may override this method to insert their own layers into
    * result set processing.
    *
    * @param   results
    *          An object which manipulates the database result set returned
    *          from execution of the query.
    *
    * @return  <code>Results</code> object.
    *
    * @throws  PersistenceException
    *          if there is any error creating results.
    */
   protected Results createResults(Results results)
   throws PersistenceException
   {
      if (isClientWhere())
      {
         return (new FilteredResults(results,
                                     Arrays.asList(getRecordBuffers()),
                                     getWhereExpressions(),
                                     !fullRecords));
      }
      
      return results;
   }
   
   /**
    * Fetch all records for a given result row into their backing buffers.  Each result row
    * contains one column per query component.  Each column contains the unique ID of the record
    * to be fetched or the record itself. If the latter, the record is refetched in case it has
    * changed since the record was first retrieved.  If the scrollable cursor backing this query
    * currently is at a position where no retrieval is possible, the backing buffers will have
    * their current records released.
    * <p>
    * As particular case when the query looks for the template record of the buffer of the single
    * {@link QueryComponent}, the buffer is not populated with a record fetched from backing
    * database, instead the template record for the buffer is automatically created in-memory.
    * This {@code results} object has only one element to iterate on and attempting to iterate on
    * this object will throw {@link QueryOffEndException} and release the buffer. 
    *
    * @param   available
    *          {@code true} if the most recent scrollable cursor move left the cursor at an
    *          available result row;  {@code false} if the cursor cannot read a result at its
    *          current position.
    * @param   lockType
    *          Overriding lock type to apply to fetched records.  If {@code null}, default lock
    *          types are applied.
    * @param   errorIfNull
    *          {@code true} if failed retrievals should raise error.
    * 
    * @throws  MissingRecordException
    *          if the record was expected to be available, but wasn't.  This typically indicates
    *          the record was deleted between the time it initially was found by the query and the
    *          time it was to be fetched to store in the record buffer.
    * @throws  ErrorConditionException
    *          if the fetch of any record fails or if no records are available and the query is
    *          set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail in this case.
    */
   protected void fetch(boolean available, LockType lockType, boolean errorIfNull)
   throws MissingRecordException
   {
      currentRowDeleted = false;
      
      try
      {
         // under certain circumstances, we must bypass releasing the current record(s) (if any)
         // in the buffer(s), so we throw the off-end exception directly instead of setting the
         // current buffer record to null
         if (!available   &&
             rowCount < 0 &&
             !errorIfNull &&
             _isOffEnd()  &&
             preserveBuffersOnEmptyResults())
         {
            getRecordBuffers()[0].throwOffEnd();
         }
         
         // check if a template search was issued. In this case we have only a single component.
         // We will load the template record in the respective buffer.
         if (results instanceof TemplateResults)
         {
            QueryComponent comp = components.get(0);
            RecordBuffer buffer = comp.getBuffer();
            if (available)
            {
               buffer.loadTemplateRecord((Long) results.getID(0));
            }
            else
            {
               // will release the buffer and throw QueryOffEndException 
               buffer.setRecord(null, 
                                lockType, 
                                false, 
                                comp.getSortIndex(), 
                                offEnd, 
                                inverseSorting, 
                                false);
            }
         }
         else
         {
            Object[] data = (available ? results.get() : null);
            coreFetch(data, lockType, errorIfNull, false);
            
            if (available)
            {
               accumulate();
            }
         }
      }
      catch (QueryOffEndException exc)
      {
         if (!lenientOffEnd)
         {
            throw exc;
         }
      }
      catch (MissingRecordException exc)
      {
         throw exc;
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
   }
   
   /**
    * Fetch all records for a given result row into their backing buffers,
    * given an array of data which contains one primary key or DMO per query
    * component.
    *
    * @param   data
    *          An array of primary key IDs or DMOs representing records to be
    *          fetched for the composite row.
    * @param   lockType
    *          Overriding lock type to apply to fetched records.  If
    *          <code>null</code>, default lock types are applied.
    * @param   errorIfNull
    *          <code>true</code> if failed retrievals should raise error.
    * @param   silentIfNullId
    *          If <code>true</code>, do not raise {@link QueryOffEndException}
    *          if some of the provided IDs are <code>null</code>. <code>null</code>
    *          IDs are valid for queries with OUTER join.
    *
    * @throws  MissingRecordException
    *          if the record was expected to be available, but wasn't.  This
    *          typically indicates the record was deleted between the time it
    *          initially was found by the query and the time it was to be
    *          fetched to store in the record buffer.
    * @throws  PersistenceException
    *          if the fetch of any record fails.
    * @throws  ErrorConditionException
    *          if no records are available and the query is set to fail in
    *          this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    */
   protected final void coreFetch(Object[] data,
                                  LockType lockType,
                                  boolean errorIfNull,
                                  boolean silentIfNullId)
   throws PersistenceException
   {
      boolean debug = LOG.isLoggable(Level.FINE);
      
      Iterator<QueryComponent> iter = components();
      for (int i = 0; iter.hasNext(); i++)
      {
         QueryComponent comp = iter.next();
         RecordBuffer buffer = comp.getBuffer();
         boolean dirtyCopy = false;
         Persistable dmo = null;
         LockType lt = lockType;
         if (data != null && data[i] != null && !currentRowDeleted)
         {
            LockType compLock = comp.getLockType();
            
            // Use default lock type if not overridden.
            if (lt == null)
            {
               lt = compLock;
            }
            
            // Individual result is either a DMO (i.e., Persistable) or a primary key ID.
            Object o = data[i];
            Serializable id = null;
            boolean updateLock;
            if (o instanceof Persistable)
            {
               dmo = (Persistable) o;
               id = dmo.getId();
               updateLock = (lt != LockType.NONE);
            }
            else
            {
               id = (Serializable) o;
               updateLock = true;
            }
            
            // If the DMO currently in the buffer has the same ID as the ID
            // we wish to load, but also represents a dirty copy (a copy of
            // uncommitted changes in another session), leave it in the buffer.
            if (buffer.isDirtyCopy())
            {
               Persistable dirtyDMO = buffer.getCurrentRecord();
               if (dirtyDMO != null && id.equals(dirtyDMO.getId()))
               {
                  dirtyCopy = true;
                  dmo = dirtyDMO;
               }
            }
            
            DirtyShareContext dirtyContext = buffer.getDirtyContext();
            String entity = buffer.getEntityName();
            
            if (!dirtyCopy)
            {
               // A record may have been retrieved from a cache of detached
               // objects, OR its state may have been changed (updated or
               // or deleted) since we first retrieved it, so reload it by its ID.
               Persistence p = getPersistence();
               if (updateLock)
               {
                  FibonacciCounter fib = new FibonacciCounter();
                  long timeout = fib.next();
                  
                  do
                  {
                     try
                     {
                        dmo = p.load(buffer, id, lt, timeout, updateLock, false);
                        
                        break;
                     }
                     catch (LockTimeoutException exc)
                     {
                        if (LOG.isLoggable(Level.FINE))
                        {
                           String table = buffer.getTable();
                           RecordIdentifier ident = new RecordIdentifier(table, id);
                           LOG.fine("[" +
                                    Utils.describeContext() +
                                    "] coreFetch lock timeout: " +
                                    timeout +
                                    " ms: " +
                                    ident);
                        }
                        
                        // check if record was deleted in uncommitted transaction in another
                        // session
                        if (dirtyContext.isDirtyDelete(entity, id, true))
                        {
                           String table = buffer.getTable();
                           RecordIdentifier ident = new RecordIdentifier(table, id);
                           
                           throw new MissingRecordException(ident);
                        }
                        
                        // steadily increase the timeout value
                        timeout = fib.next();
                        
                        // log deadlock warning if timeout is getting too long and wait
                        // indefinitely
                        if (timeout > 3600000L)
                        {
                           timeout = 0L;
                           if (LOG.isLoggable(Level.WARNING))
                           {
                              String table = buffer.getTable();
                              RecordIdentifier ident = new RecordIdentifier(table, id);
                              LOG.log(Level.WARNING,
                                      "[" +
                                      Utils.describeContext() +
                                      "] Possible deadlocked record: " + ident,
                                      new Throwable());
                           }
                        }
                     }
                  } while (true);
               }
               else
               {
                  dmo = p.load(buffer, id, lt, 0L, updateLock, false);
               }
            }
            
            if (debug)
            {
               StringBuilder buf = new StringBuilder("DMO:  ");
               if (id != null)
               {
                  buf.append(id);
               }
               else
               {
                  buf.append("<not found>");
               }
               
               LOG.log(Level.FINE, buf.toString());
               
               if (LOG.isLoggable(Level.FINEST))
               {
                  buf.setLength(0);
                  buf.append("DMO Detail:  ");
                  buf.append(dmo);
                  
                  LOG.log(Level.FINEST, buf.toString());
               }
            }
            
            // If we expected this record to be available, but it wasn't, it
            // must have been deleted.  Likewise, if we know it was deleted in
            // another context, this is an exceptional condition.  Throw an
            // exception so the caller can figure out the appropriate recovery.
            if (dmo == null || dirtyContext.isDirtyDelete(entity, id, true))
            {
               String table = buffer.getTable();
               RecordIdentifier ident = new RecordIdentifier(table, id);
               
               throw new MissingRecordException(ident);
            }
         }
         
         // TODO:  confirm or fix behavior:  first null record will raise
         // error or end condition;  remaining buffers will not have their
         // current records released.
         SortIndex sortIndex = comp.getSortIndex();
         buffer.setRecord(dmo,
                          lt,
                          errorIfNull || silentIfNullId,
                          sortIndex,
                          offEnd,
                          inverseSorting,
                          dirtyCopy);  // TODO:  need better dirty check
         if (dmo == null && errorIfNull)
         {
            buffer.errorNotOnFile();
         }
      }
   }
   
   /**
    * Indicate whether this query requires initial results to be filtered
    * through a client-side where clause expression.
    *
    * @return  <code>true</code> if client-side filter is required, else
    *          <code>false</code>.
    */
   protected final boolean isClientWhere()
   {
      return (!getWhereExpressions().isEmpty());
   }
   
   /**
    * Access the current <code>Results</code> object for this query.
    *
    * @return  <code>Results</code> object.
    */
   protected Results getResults()
   {
      execute();
      
      return results;
   }
   
   /**
    * Add a component to the query.
    *
    * @param   comp
    *          Query component to be added.
    *
    * @throws  UnsupportedOperationException
    *          if attempting an outer join (temporary);
    *          if iteration type is not <code>NEXT</code> (temporary);
    *          if the component's DMO's associated database is different than
    *          that of a previously added component's database (cross-database
    *          preselect joins are not possible).
    */
   protected void addComponent(QueryComponent comp)
   {
      if (isClientWhere())
      {
         throw new IllegalStateException(
            "Cannot have a multi-table query with a client-side where clause");
      }
      
      // TODO:  add support for outer joins.
      if (comp.isOuter())
      {
         throw new UnsupportedOperationException(
            "Outer join not supported with preselect query");
      }
      
      switch (comp.getIteration())
      {
         case NEXT:
         case FIRST:
         case LAST:
            break;
         default:
            if (!supportsArbitraryIteration())
            {
               throw new UnsupportedOperationException(
                  "Preselect query supports only NEXT/FIRST/LAST iteration");
            }
            break;
      }
      
      boolean isFirst = components.isEmpty();
      
      RecordBuffer buffer = comp.getBuffer();
      
      if (buffer.getOpenScopeCount() > 0)
      {
         buffer.initialize();
         
         // Flush the record buffer, in case a transient, newly created record
         // is currently in the buffer.
         try
         {
            buffer.flush();
         }
         catch (ValidationException exc)
         {
            ErrorManager.throwError(exc);
         }
         
         if (isFirst)
         {
            // Register for retry support.
            TransactionManager.registerFinalizable(this, false);
         }
         // Verify that no cross-connection join is being attempted.
         else if (!getPersistence().equals(buffer.getPersistence()))
         {
            throw new UnsupportedOperationException(
               "Preselect query cannot join across database connections");
         }
      }
      
      if (!isFirst)
      {
         comp.setNotTop();
      }
      
      // Add the component.
      components.add(comp);
      
      // Update cumulative substitution argument count.
      Object[] args = comp.getRawArgs();
      if (args != null)
      {
         int len = args.length;
         for (int i = 0; !resolvableArgs && i < len; i++)
         {
            if (args[i] instanceof Resolvable)
            {
               resolvableArgs = true;
            }
         }
      }
   }
   
   /**
    * Does this query support non-NEXT iteration components?
    * 
    * @return  <code>false</code>
    */
   protected boolean supportsArbitraryIteration()
   {
      return false;
   }
   
   /**
    * Indicate whether the record buffer(s) current record(s) should be preserved, in the event
    * the query does not find any results.
    * <p>
    * This implementation always returns false. Subclasses should override this behavior if
    * needed.
    * 
    * @return  <code>false</code>.
    */
   protected boolean preserveBuffersOnEmptyResults()
   {
      return false;
   }
   
   /**
    * Clear the reposition cache of visited, virtual composite rows.
    */
   protected void clearRepoCache()
   {
      repoCache.clear();
   }
   
   /**
    * Get the HQL query which was executed to produce this query's results.
    *
    * @return  HQL query or <code>null</code> if the query has not yet been
    *          executed.
    */
   protected String getHQL()
   {
      return hql;
   }
   
   /**
    * Get the persistence service object used by this query.
    * 
    * @return  Persistence object.
    */
   protected Persistence getPersistence()
   {
      if (persistence == null)
      {
         // Since no cross-connection joins are permitted, it is safe to set
         // the query's Persistence instance using the first component only.
         RecordBuffer buffer = components.get(0).getBuffer();
         persistence = buffer.getPersistence();
      }
      
      return persistence;
   }
   
   /**
    * Register a <code>Finalizable</code> with the transaction manager for
    * resource cleanup when this query's scope ends.
    */
   protected void registerCleaner()
   {
      // If we are not a standalone query, we assume our container will
      // register for cleanup.
      if (regCleaner || !isStandalone())
      {
         return;
      }
      
      Finalizable cleaner = new Finalizable()
      {
         public void finished() { cleanup(); }
         public void iterate() {}
         public void retry() {}
         public void deleted() {}
      };
      if (isTopLevelCleanup())
      {
         TransactionManager.registerTopLevelFinalizable(cleaner, true);
      }
      else
      {
         TransactionManager.registerFinalizable(cleaner, false);
      }
      
      regCleaner = true;
   }
   
   /**
    * Reposition the cursor such that a request to retrieve the next result
    * will retrieve the result which matches the specified array of primary
    * key IDs.  A request to retrieve the previous result will retrieve the
    * result immediately previous to this, if any, in the results list.
    *
    * @param   ids
    *          Array of primary key IDs, arranged from left to right to
    *          coincide with the records being joined by the underlying
    *          query.
    * @param   forceFetch
    *          <code>true</code> to override the current fetch-on-reposition
    *          setting to <code>true</code>, in case the underlying
    *          <code>Results</code> implementation attempts to fetch the
    *          record during repositioning;  <code>false</code> to honor the
    *          current setting.
    * 
    * @return  <code>true</code> if reposition was successful;
    *          <code>false</code> if unsuccessful.
    *
    * @throws  ErrorConditionException
    *          if there is an error repositioning the cursor.
    */
   protected boolean repositionByID(Serializable[] ids, boolean forceFetch)
   {
      prepareReposition();
      
      if (results == null)
      {
         execute();
      }

      ids = DBUtils.trimToTableCount(ids, this);
      int newRow = repoCache.getRow(ids);
      if (newRow >= 0)
      {
         int oldRow = results.getRowNumber();
         boolean avail;
         if (forceFetch)
         {
            boolean oldFetch = isFetchOnReposition();
            try
            {
               setFetchOnReposition(true);
               avail = results.setRowNumber(newRow);
            }
            finally
            {
               setFetchOnReposition(oldFetch);
            }
         }
         else
         {
            avail = results.setRowNumber(newRow);
         }
         if (avail)
         {
            offEnd = OffEnd.NONE;
            betweenRows = true;
         }
         else
         {
            int diff = oldRow - newRow;
            offEnd = (diff < 0 ? OffEnd.FRONT : OffEnd.BACK);
            betweenRows = false;
         }
         
         afterReposition(true);
         
         return true;
      }
      
      // Scan all results for a match.  Start from where cache last left off.
      boolean error = false;
      int size = repoCache.size();
      results.setRowNumber(size);
      if (results.get(0) == null)
      {
         repoCache.clear();
         error = !results.first();
      }
      
      int width = getTableCount();
      if (!error)
      {
         do
         {
            boolean match = true;
            Serializable[] rsIDs = new Serializable[width];
            for (int i = 0; i < width; i++)
            {
               Serializable nextID = results.getID(i);
               rsIDs[i] = nextID;
               if (ids.length > i && !ids[i].equals(nextID))
               {
                  match = false;
               }
            }
            
            newRow = results.getRowNumber();
            repoCache.add(rsIDs, newRow);
            
            if (match)
            {
               offEnd = OffEnd.NONE;
               betweenRows = true;
               afterReposition(true);
               
               return true;
            }
            
         } while (results.next());
      }
      
      // No match was found;  reset to old current row.
//      results.setRowNumber(currentRow);
      
      // Here we need to emulate a quirk/defect of Progress' reposition
      // implementation:  if no match was found, we leave the "current"
      // position at the last checked position.  Furthermore, if we are
      // fetching on reposition, we have to load this last record into the
      // associated buffer(s), to truly emulate the GET NEXT semantic.
      if (isFetchOnReposition() && results.last())
      {
         try
         {
            offEnd = OffEnd.NONE;
            fetch(true, null, true);
            afterReposition(false);    // notify listeners
         }
         catch (MissingRecordException | ErrorConditionException exc)
         {
            if (LOG.isLoggable(Level.FINE))
            {
               LOG.log(Level.FINE, "Error fetching record(s) during reposition", exc);
            }
         }
      }
      else
      {
         offEnd = OffEnd.BACK;
         betweenRows = false;
         afterReposition(false);
      }
      
      return false;
   }
   
   /**
    * Prepare to process a reposition request by releasing the current
    * records, if any, in each buffer managed by the query.  This occurs in
    * Progress unconditionally, even if the reposition does not actually move
    * the cursor from its current position.
    */
   protected void prepareReposition()
   {
      // Don't release buffer if we're processing a recursive reposition that
      // was triggered by a reposition notification.
      if (isRepositionNotificationActive())
      {
         return;
      }
      
      for (QueryComponent next : components)
      {
         next.getBuffer().release();
      }
   }
   
   /**
    * Find the query component associated with the given record buffer.
    *
    * @param   buffer
    *          Record buffer.
    *
    * @return  Associated query component, or <code>null</code> if not found.
    */
   private QueryComponent findComponent(RecordBuffer buffer)
   {
      Iterator<QueryComponent> iter = components();
      while (iter.hasNext())
      {
         QueryComponent next = iter.next();
         if (next.getBuffer() == buffer)
         {
            return next;
         }
      }
      
      return null;
   }
   
   /**
    * Assemble the HQL which is submitted to Hibernate in order to execute
    * the database query underlying this object. The HQL generated uses the
    * theta join style and includes a select clause, a from clause, an
    * optional where clause, and an order by clause (mandatory in all cases
    * except a unique find).
    *
    * @return  String
    *          A formatted HQL statement.
    *
    * @throws  PersistenceException
    *          if the sort criteria reference unqualified or unrecognized
    *          buffer names.
    */
   private String assembleHQL()
   throws PersistenceException
   {
      StringBuilder buf = new StringBuilder();
      
      // Need to put together order by clause first, as the sort criteria
      // determined in this analysis may be needed when composing from clause
      // and where clause.
      List<SortCriterion> sortCriteria = assembleOrderByClause(buf);
      String orderBy = buf.toString();
      buf.setLength(0);
      assembleSelectClause(buf);
      assembleFromClause(buf, sortCriteria);
      buf.append(assembleWhereClause(sortCriteria).toFinalExpression());
      buf.append(orderBy);
      
      return buf.toString().trim();
   }
   
   /**
    * Validate a row identifier provided to a reposition request to ensure
    * that it does not represent the unknown value.
    *
    * @param   value
    *          ID to be validated.
    *
    * @return  <code>true</code> if the number is valid;  <code>false</code>
    *          if it is invalid, but we are in silent error mode.
    *
    * @throws  ErrorConditionException
    *          if the number is invalid and we are not in silent error mode.
    */
   private boolean validateRepositionByID(rowid value)
   {
      String msg = "Could not evaluate rowid during REPOSITION";
      
      return validateRepositionImpl(value, 7330, msg);
   }
   
   /**
    * Validate a record identifier provided to a reposition request to ensure
    * that it does not represent the unknown value.
    *
    * @param   value
    *          ID to be validated.
    *
    * @return  <code>true</code> if the number is valid;  <code>false</code>
    *          if it is invalid, but we are in silent error mode.
    *
    * @throws  ErrorConditionException
    *          if the number is invalid and we are not in silent error mode.
    */
   private boolean validateRepositionByID(recid value)
   {
      String msg = "Could not evaluate rowid during REPOSITION";
      
      return validateRepositionImpl(value, 7330, msg);
   }
   
   /**
    * Validate a numeric value provided to a reposition request to ensure
    * that it does not represent the unknown value.
    *
    * @param   value
    *          ID or row number/offset to be validated.
    *
    * @return  <code>true</code> if the number is valid;  <code>false</code>
    *          if it is invalid, but we are in silent error mode.
    *
    * @throws  ErrorConditionException
    *          if the number is invalid and we are not in silent error mode.
    */
   private boolean validateReposition(NumberType value)
   {
      String msg = "Could not evaluate reposition amount for query";
      
      return validateRepositionImpl(value, 3165, msg);
   }
   
   /**
    * Validate a numeric value provided to a reposition request to ensure
    * that it does not represent the unknown value.
    *
    * @param   value
    *          ID or row number/offset to be validated.
    * @param   errorNum
    *          Error number if validation fails.
    * @param   errorMsg
    *          Error message if validation fails.
    *
    * @return  <code>true</code> if the number is valid;  <code>false</code>
    *          if it is invalid.
    *
    * @throws  ErrorConditionException
    *          if the number is invalid and we are not in silent error mode.
    */
   private boolean validateRepositionImpl(BaseDataType value, int errorNum, String errorMsg)
   {
      if (value.isUnknown())
      {
         ErrorManager.displayError(errorNum, errorMsg, false);
         
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, errorMsg);
         }
         
         return false;
      }
      
      return true;
   }
   
   /**
    * Special {@link Results} implementation for template records. Since a query on rowid/recid
    * will return only a single row, this class is built around its {@code Serializable} rowid
    * and provide fast access for a {@link PreselectQuery} aware of this query type to iterate
    * on this single-row result.
    */
   private static class TemplateResults
   implements Results
   {
      /** The id of the template record to be loaded. */
      private final Serializable templateRowid;
      
      /** 
       * The position of the cursor. This set of results always contain a single record, at
       * position 0. The natural starting position is -1, before the first row. The value 1 is
       * just after the last row. The class will maintain one of these three states.
       */
      private int cursor = -1;
      
      /**
       * The constructor. Builds a {@link Results} specific to a template rowid.
       * 
       * @param   templateRowid
       *          The rowid of the template record. Immutable.
       */
      public TemplateResults(Serializable templateRowid)
      {
         this.templateRowid = templateRowid;
      }
      
      /**
       * Move cursor to the first results row.
       *
       * @return  {@code true} if there are any results.
       */
      @Override
      public boolean first()
      {
         cursor = 0;
         return true; 
      }
      
      /**
       * Move cursor to the last results row.
       *
       * @return  {@code true} if there are any results.
       */
      @Override
      public boolean last()
      {
         cursor = 0;
         return true;
      }
      
      /**
       * Move cursor to the next results row.
       *
       * @return  {@code true} if there is a result under the cursor under the move.
       */
      @Override
      public boolean next()
      {
         ++ cursor;
         normalize();
         return isFirst();
      }
      
      /**
       * Move cursor to the previous results row.
       *
       * @return  {@code true} if there is a result under the cursor under the move.
       */
      @Override
      public boolean previous()
      {
         -- cursor;
         normalize();
         return isFirst();
      }
      
      /**
       * Is the cursor on the first row in the results set?
       *
       * @return  {@code true} if the cursor is on the first row.
       */
      @Override
      public boolean isFirst()
      {
         return cursor == 0;
      }
      
      /**
       * Is the cursor on the last row in the results set?
       *
       * @return  {@code true} if the cursor is on the first row.
       */
      @Override
      public boolean isLast()
      {
         return cursor == 0;
      }
      
      /**
       * Get the array of objects at the current result row.
       *
       * @return  Object array or {@code null}.
       */
      @Override
      public Object[] get()
      {
         return (cursor == 0) ? new Object[] { templateRowid } : null; 
      }
      
      /**
       * Get the object at the current result row, at the specified column.
       *
       * @param   column
       *          Zero-based index column of the desired object.
       *
       * @return  Object at {@code column} or {@code null}.
       */
      @Override
      public Object get(int column)
      {
         return (cursor == 0 && column == 0) ? templateRowid : null;
      }
      
      /**
       * Get the primary key ID at the current result row, at the specified
       * column.
       *
       * @param   column
       *          Zero-based index column of the desired ID.
       *
       * @return  ID of the record or {@code null}.
       */
      @Override
      public Serializable getID(int column)
      {
         return (cursor == 0 && column == 0) ? templateRowid : null;
      }
      
      /**
       * Get the row number currently under the cursor.
       *
       * @return  Zero-based index of the current row, or {@code -1} if the cursor is not 
       *          currently on a result.
       */
      @Override
      public int getRowNumber() { return cursor == 0 ? 0 : -1; }
      
      /**
       * Set the row number currently under the cursor.
       *
       * @param   row
       *          Zero-based index of the row to be set as the current row.
       *
       * @return  {@code true} if there is a row at the specified row number; else {@code false}.
       */
      @Override
      public boolean setRowNumber(int row)
      {
         cursor = row;
         normalize();
         return isFirst(); 
      }
      
      /**
       * Scroll the cursor ahead by the specified number of rows.  If the number is negative, the
       * cursor is moved backward.  This may put the cursor off the end (either end) of the
       * query's results.
       *
       * @param   rows
       *          Number of rows to jump ahead or back.
       *
       * @return  {@code true} if there is a row at the new location; else {@code false}.
       */
      @Override
      public boolean scroll(int rows)
      {
         cursor += rows;
         normalize();
         return isFirst();
      }
      
      /**
       * Reset the cursor to its natural starting position, before the first result row.
       */
      @Override
      public void reset()
      {
         cursor = -1;
      }
      
      /**
       * Invoked when the current Hibernate session is about to close.  In this case this is a
       * no-op since the template record is not linked to Hibernate.
       */
      @Override
      public void sessionClosing() { }
      
      /**
       * Clean up and release any resources which this object is holding, such as open result
       * sets. In this case it does nothing as no resources are held.
       */
      @Override
      public void cleanup() { }
      
      /**
       * Internal worker that keeps the correctness of cursor position: always in one of the
       * positions: -1 = before the first record, 0 = the only record is selected, +1, after the
       * last record.
       */
      private void normalize()
      {
         if (cursor > 1)
         {
            cursor = 1;
         }
         else if (cursor < -1)
         {
            cursor = -1;
         }
         // else cursor is -1 (before), 0 (on single record), 1 (after)
      }
   }
}