/*
** Module   : QueryComponent.java
** Abstract : A joinable, single-table query component used in a preselect query.
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20150801 Refactored from PreselectQuery, which previously contained this as an inner
**                  class.
** 002 ECF 20150907 Reimplemented silent error mode processing of query substitution parameters.
** 003 CA  20160518 Change ParameterIndices usage to collect all substitution parameters, as their 
**                  number might change (up or down) after ternary is normalized.
** 004 ECF 20160610 Changed access modifier of isIdOnly to protected.
** 005 OM  20160629 HQLPreprocessor needs the list of fields of the current index.
** 006 ECF 20160720 Fixed buffer lookup during HQL preprocessing.
** 007 ECF 20160825 Added missing javadoc.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.util.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.util.*;

/**
 * A helper container for a query component. A query component defines the buffer to be updated,
 * the type of join to perform with the previous component (if any), the lock type for records
 * fetched into the buffer, the where clause to be applied, and any query substitution parameters.
 */
class QueryComponent
implements QueryConstants
{
   /** Joinable query which contains this query component */
   private final Joinable query;
   
   /** Original where clause, <i>before</i> preprocessing */
   private final String where;
   
   /** Record buffer updated by this component of the query */
   private final RecordBuffer buffer;
   
   /** Helper object for a join via a foreign key relation */
   private final AbstractJoin join;
   
   /** Record lock type */
   private final LockType lockType;
   
   /** Base substitution arguments for this component */
   private final Object[] args;
   
   /** {@code true} if component forms outer join with previous component; else {@code false} */
   private final boolean outer;
   
   /** Whether query should return primary key ID only, or entire DMO */
   private final boolean idOnly;
   
   /** Is this component the top/outermost among the query's components? */
   private boolean top = true;
   
   /**
    * Array of field reference names and nulls used to replace substitution parameters when
    * converting a nested looping (client-side) join to a server-side join
    */
   private String[] referenceSubs;
   
   /** Set of related buffers, if component participates in a server-side join */
   private Set<RecordBuffer> relatedBuffers = null;
   
   /** Iteration/retrieval mode (first, last, next, etc.) */
   private int iteration;
   
   /** Order by clause for FIRST/LAST iteration components */
   private String sort;
   
   /** HQL preprocessor */
   private HQLPreprocessor hqlPreprocessor;
   
   /** Position of {@code null} elements in argument list with which HQL was preprocessed */
   private boolean[] nullArguments = null;
   
   /** Sort index associated with this query component */
   private SortIndex sortIndex;
   
   /** Intermediately resolved arguments for this component */
   private Object[] currentArgs = null;
   
   /** Index information string as it is returned by INDEX-INFORMATION. */
   private String indexInfo = null;
   
   /**
    * Constructor.
    * 
    * @param   query
    *          Enclosing query.
    * @param   proxy
    *          DMO proxy which references the record buffer updated by
    *          this component of the query.
    * @param   join
    *          Helper object for a join via a foreign key relation.  May
    *          be <code>null</code>.
    * @param   where
    *          Restriction criteria in HQL where clause format.
    * @param   indexInfo
    *          Index information string as it is returned by INDEX-INFORMATION.
    *          May be <code>null</code> if it is not an OPEN QUERY case or if
    *          you don't need debug information about selected indexes.
    * @param   lockType
    *          Requested record lock type.
    * @param   args
    *          Default (unresolved) query substitution arguments.
    * @param   iteration
    *          Type of iteration:  FIRST, LAST, NEXT, PREVIOUS, or UNIQUE.
    * @param   outer
    *          <code>true</code> if this component is joined with the
    *          previous component via an outer join;  <code>false</code>
    *          if there is no previous component, or if this component is
    *          joined with the previous component via an inner join.
    */
   protected QueryComponent(Joinable query,
                            BufferReference proxy,
                            AbstractJoin join,
                            String where,
                            String indexInfo,
                            LockType lockType,
                            Object[] args,
                            int iteration,
                            boolean outer)
   {
      if (lockType == null)
      {
         throw new NullPointerException("Lock type must be specified");
      }
      
      RecordBuffer buffer = proxy.buffer();
      AbstractQuery.validateSubstitutionArguments(args);
      AbstractQuery.preprocessSubstitutionArguments(buffer.getBufferManager(), args);
      
      this.query = query;
      this.buffer = buffer;
      this.join = join;
      this.where = where;
      this.args = args;
      this.iteration = iteration;
      this.outer = outer;
      this.indexInfo = indexInfo;
      
      // override lock type if this is a temporary buffer to allow some downstream optimizations
      this.lockType = (buffer.isTemporary() ? LockType.NONE : lockType);
      
      // pull back entire record if possible, to avoid additional round trips to the database
      this.idOnly = !LockType.NONE.equals(this.lockType);
   }
   
   /**
    * Prepare the information needed to convert a nested looping (client-side) join of two
    * tables to a server-side join of those tables.
    * 
    * @param   joinList
    *          List of enclosing query components, from outermost to innermost (if any), from the
    *          original nested query. Does not include the component currently being visited.
    * @param   buffer
    *          Record buffer associated with the query component currently being visited.
    * @param   join
    *          Original join helper object which joins the current query component with an outer
    *          component or external buffer.
    * @param   args
    *          Query substitution arguments for the where clause of the current query component.
    * 
    * @return  An object containing the information needed to convert the join from client-side
    *          to server-side.
    */
   static ServerJoinData prepareServerJoinData(List<? extends QueryComponent> joinList,
                                               RecordBuffer buffer,
                                               AbstractJoin join,
                                               Object[] args)
   {
      Object[] processedArgs = args;
      String[] referenceSubs = null;
      AbstractJoin newJoin = join;
      Set<RecordBuffer> relatedBuffers = null;
      
      if (!joinList.isEmpty())
      {
         relatedBuffers = new HashSet<>();
         for (QueryComponent related : joinList)
         {
            relatedBuffers.add(related.getBuffer());
         }
         
         if (args != null)
         {
            int len = args.length;
            List<Object> survivingArgs = new ArrayList<>();
            referenceSubs = new String[len];
            
            for (int i = 0; i < len; i++)
            {
               Object next = args[i];
               
               if (next instanceof FieldReference)
               {
                  FieldReference ref = (FieldReference) next;
                  if (relatedBuffers.contains(ref.getParentBuffer()))
                  {
                     referenceSubs[i] = ref.toString();
                     
                     continue;
                  }
               }
               
               survivingArgs.add(next);
            }
            
            int survived = survivingArgs.size();
            if (survived < len)
            {
               processedArgs = survivingArgs.toArray();
            }
         }
         
         if (join != null)
         {
            RecordBuffer inverseBuffer = join.getInverse().buffer();
            if (relatedBuffers.contains(inverseBuffer))
            {
               newJoin = new ServerLegacyKeyJoin(buffer, inverseBuffer);
            }
         }
      }
      
      return new ServerJoinData(newJoin, processedArgs, relatedBuffers, referenceSubs);
   }
   
   /**
    * Get the HQL preprocessor associated with this component.
    * 
    * @return  HQL preprocessor.
    * 
    * @throws  PersistenceException
    *          if the preprocessor must first be created, and there is an error during
    *          its creation.
    */
   protected HQLPreprocessor getHQLPreprocessor()
   throws PersistenceException
   {
      if (hqlPreprocessor == null)
      {
         Object[] resolvedArgs = resolveArgs();
         hqlPreprocessor = prepareHQLPreprocessor(resolvedArgs);
         if (resolvedArgs == null || resolvedArgs.length == 0)
         {
            nullArguments = null;
         }
         else
         {
            nullArguments = new boolean[resolvedArgs.length];
            for (int k = 0; k < resolvedArgs.length; k++)
            {
               Object param = resolvedArgs[k];
               if (param instanceof FieldReference)
               {
                  param = ((FieldReference) param).resolve();
               }
               
               nullArguments[k] = param instanceof BaseDataType && 
                                  ((BaseDataType) param).isUnknown();
            }
         }
      }
      
      return hqlPreprocessor;
   }
   
   /**
    * Create the <code>HQLPreprocessor</code> which backs this component. If this component
    * represents a nested FIRST or LAST iteration, the where clause portion associated with this
    * component will be embedded within a subselect. To avoid conflicts with the enclosing
    * where clause, this portion will be emitted without DMO alias qualifiers for its references
    * to the associated buffer.
    * 
    * @param   queryArgs
    *          Query substitution parameters used in the preprocessor's creation.
    * 
    * @return  HQL preprocessor.
    * 
    * @throws  PersistenceException
    *          if there is an error creating the preprocessor.
    */
   protected HQLPreprocessor prepareHQLPreprocessor(Object[] queryArgs)
   throws PersistenceException
   {
      boolean subselect = !top && (iteration == FIRST || iteration == LAST);
      RecordBuffer buffer = getBuffer();
      
      return HQLPreprocessor.get(where,
                                 buffer.getDatabase(),
                                 buffer.getDialect(),
                                 getReferencedBuffers(),
                                 queryArgs,
                                 referenceSubs,
                                 true,
                                 subselect ? buffer.getDMOAlias() : null,
                                 false,
                                 null);
   }
   
   /**
    * Get an array of all record buffers referenced by this query, including those directly
    * involved in the query, as well as those (if any) indirectly referenced by inlined,
    * converted, CAN-FIND expressions, and those related via a server-side join (if any).
    * 
    * @return  All record buffers referenced by this query.
    */
   protected RecordBuffer[] getReferencedBuffers()
   {
      RecordBuffer[] qbufs = query.getReferencedBuffers();
      
      if (relatedBuffers == null || relatedBuffers.isEmpty())
      {
         return qbufs;
      }
      
      int len1 = qbufs.length;
      int len = len1 + relatedBuffers.size();
      RecordBuffer[] all = new RecordBuffer[len];
      
      int i = 0;
      while (i < len1)
      {
         all[i] = qbufs[i];
         i++;
      }
      for (RecordBuffer buf : relatedBuffers)
      {
         all[i] = buf;
         i++;
      }
      
      return all;
   }
   
   /**
    * Get substitution arguments for this component, resolving them first if necessary.
    *
    * @return  Substitution arguments.
    * 
    * @throws  PersistenceException
    *          if there is any error creating the HQL preprocessor, which
    *          is necessary to properly filter/order the arguments.
    */
   protected Object[] getArgs()
   throws PersistenceException
   {
      if (args != null && currentArgs == null)
      {
         currentArgs = filterArgs(getHQLPreprocessor());
      }
      
      return currentArgs;
   }
   
   /**
    * Reset current arguments, so that they will need to be resolved the next time they are
    * requested. The hqlPreprocessor is dropped only if it was inlined (and the inlined nodes
    * need to be re-evaluated with the new arguments) or if the argument list contains some
    * other configuration of null values (they were also replaced with "is [not] null" 
    * hql constructs).
    */
   protected void resetArgs()
   {
      currentArgs = null;
      
      // Since the arguments were reset, the HQL preprocessor may need to be reset, too.
      if (hqlPreprocessor == null)
      {
         return; // nothing to do
      }
      
      // if some hql nodes were inlined (?)
      if (hqlPreprocessor.wasInlined())
      {
         resetHQLPreprocessor();
         return;
      }
      
      // check the size of the parameter list (should always be the same, anyway)
      Object[] params = resolveArgs();
      if ((params == null && nullArguments != null) || (params != null && nullArguments == null))
      {
         resetHQLPreprocessor();
         return;
      }
      
      if (params == null)
      {
         // nullArguments is null too, so we keep the hql
         return;
      }
      
      if (params.length != nullArguments.length)
      {
         resetHQLPreprocessor();
         return;
      }
      
      // if params and nullArguments have the same length check for null pattern
      for (int k = 0; k < params.length; k++)
      {
         Object param = params[k];
         if (param instanceof FieldReference)
         {
            param = ((FieldReference) param).resolve();
         }
         
         if (nullArguments[k] != (param instanceof BaseDataType &&
                                  ((BaseDataType) param).isUnknown()))
         {
            resetHQLPreprocessor();
            return;
         }
      }
   }
   
   /**
    * Reset the HQL preprocessor, so that it needs to be recreated the next
    * time it is requested.
    */
   protected void resetHQLPreprocessor()
   {
      hqlPreprocessor = null;
      nullArguments = null; // keep in sync
   }
   
   /**
    * Get the (mostly) resolved query substitution parameters required by
    * this query component and return them in an array.  The array's
    * elements will match the positional placeholders in the query's where
    * clause <em>after</em> it has been preprocessed.  The preprocessing
    * step may have reordered or dropped certain parameters.  The only
    * parameters which will not have been fully resolved will be those
    * <code>FieldReference</code> parameters which are actually resolved by
    * the query itself, when the query's full HQL is being assembled and
    * when the final array of substitution parameters is prepared.
    * 
    * @param   preproc
    *          HQL preprocessor which maps parameter indices to the
    *          original parameters, in case any were changed during HQL
    *          where clause preprocessing.
    * 
    * @return  Array of filtered, resolved parameters.
    */
   protected Object[] filterArgs(HQLPreprocessor preproc)
   {
      Object[] intermediate = resolveArgs();
      if (intermediate == null)
      {
         return null;
      }
      
      ParameterIndices pi = preproc.getParameterIndices();
      int len = pi.getCount();
      Object[] filteredArgs = new Object[len];
      for (int i = 0, count = 0; count < len; i++)
      {
         Integer nextIndex = pi.getIndex(i);
         if (nextIndex == null)
         {
            // Original substitution parameter was inlined into HQL by
            // preprocessor;  skip it.
            continue;
         }
         
         // Substitution placeholders may have been reordered if this
         // statement was rewritten by the HQL preprocessor. Make sure
         // we access the correct substitution parameter.
         Object next = intermediate[nextIndex];
         
         filteredArgs[count++] = next;
      }
      
      return filteredArgs;
   }
   
   /**
    * Set the array of field reference names and nulls used when converting a nested loop where
    * clause to a server-join where clause.
    * 
    * @param   referenceSubs
    *          Array of field reference names and nulls.
    */
   protected void setReferenceSubs(String[] referenceSubs)
   {
      this.referenceSubs = referenceSubs;
   }
   
   /**
    * Store a set of outer component buffers to which this component is related through a server
    * side join.
    * 
    * @param   relatedBuffers
    *          Set of related buffers; may be {@code null}.
    */
   protected void setRelatedBuffers(Set<RecordBuffer> relatedBuffers)
   {
      this.relatedBuffers = relatedBuffers;
   }
   
   /**
    * Resolve all resolvable query substitution parameters, with the exception of
    * <code>FieldReference</code>s.  The latter are only query substitution parameters when it
    * is necessary to join the enclosing query's results with those of a related query, as
    * would be the case when each query is a component of a {@link CompoundQuery}.
    * 
    * @return  Array of query substitution parameters after all non-<code>FieldReference</code>
    *          resolvable parameters have been resolved.
    */
   protected Object[] resolveArgs()
   {
      if (args == null)
      {
         return null;
      }
      
      int len = args.length;
      Object[] resolvedArgs = new Object[len];
      
      for (int i = 0; i < len; i++)
      {
         Object next = args[i];
         
         if (next instanceof FieldReference)
         {
            resolvedArgs[i] = next;
         }
         else if (next instanceof Resolvable)
         {
            resolvedArgs[i] = ((Resolvable) next).resolve();
         }
         else
         {
            resolvedArgs[i] = ((BaseDataType) next).duplicate();
         }
      }
      
      return resolvedArgs;
   }
   
   /**
    * Indicate whether the query component should return a primary key ID
    * only, or the entire DMO instance.
    *
    * @return  <code>true</code> for a projection query which returns only
    *          the ID;  <code>false</code> to fetch the entire DMO.
    */
   protected boolean isIdOnly()
   {
      return idOnly;
   }
   
   /**
    * Get the query which encloses this component.
    * 
    * @return  Enclosing query.
    */
   protected Joinable getEnclosingQuery()
   {
      return query;
   }
   
   /**
    * Get record buffer for this component of the query.
    *
    * @return  Record buffer.
    */
   RecordBuffer getBuffer()
   {
      return buffer;
   }
   
   /**
    * Get restriction criteria in HQL where clause format for this
    * component (before the HQL preprocessor has preprocessed it).
    *
    * @return  Un-preprocessed where clause.
    */
   String getOriginalWhere()
   {
      return where;
   }
   
   /**
    * Get helper object for a join via a foreign key relation, if any.
    *
    * @return  Join helper, or <code>null</code> if there is none.
    */
   AbstractJoin getJoin()
   {
      return join;
   }
   
   /**
    * Get restriction criteria in HQL where clause format for this component (pre-processed).
    *
    * @return  Where clause.
    * 
    * @throws  PersistenceException
    *          if there is an error creating the {@code HQLPreprocessor} from which the
    *          preprocessed where clause is retrieved.
    */
   HQLExpression getWhere()
   throws PersistenceException
   {
      HQLExpression baseWhere = getHQLPreprocessor().getHQL();
      
      if ((iteration != FIRST && iteration != LAST) || top)
      {
         return baseWhere;
      }
      
      // generate a where expression which tests for equality between the DMO's primary key and
      // the primary key of the record found using the core where clause in a sub-select which
      // finds a single, FIRST or LAST record, as determined by the sort clause of this component:
      //    dmo.id = (select from DMO where <baseWhere> order by <sort> single)
      // 'single' is a custom HQL extension which instructs Hibernate to emit dialect-specific
      // SQL to limit the result to 1 row
      HQLExpression exp = new HQLExpression();
      exp.append(buffer.getDMOAlias());
      exp.append(".id = (select id from ");
      exp.append(buffer.getDMOImplementationName());
      if (baseWhere != null || join != null)
      {
         exp.append(" where ");
         
         if (baseWhere != null)
         {
            exp.append("(");
            exp.append(baseWhere);
            exp.append(")");
         }
         
         if (join != null)
         {
            if (baseWhere != null)
            {
               exp.append(" and ");
            }
            
            // need to strip out DMO aliases
            String joinWhere = join.getHQL().toFinalExpression();
            Object[] joinArgs = join.getUnresolvedParameters().toArray();
            HQLPreprocessor preproc = HQLPreprocessor.get(joinWhere,
                                                          buffer.getDatabase(),
                                                          buffer.getDialect(),
                                                          query.getReferencedBuffers(),
                                                          joinArgs,
                                                          null,
                                                          false,
                                                          buffer.getDMOAlias(),
                                                          false,
                                                          null);
            exp.append("(");
            exp.append(preproc.getHQL());
            exp.append(")");
         }
      }
      exp.append(" order by ");
      exp.append(sort);
      exp.append(" single)");
      
      return exp;
   }
   
   /**
    * Get lock type for records fetched into this component's buffer.
    *
    * @return  Lock type.
    */
   LockType getLockType()
   {
      return lockType;
   }
   
   /**
    * Get unresolved substitution arguments for this component, if any.
    *
    * @return  Substitution arguments, or <code>null</code> if none.
    */
   Object[] getRawArgs()
   {
      return args;
   }
   
   /**
    * Get iteration/retrieval mode (first, last, next, etc.).
    *
    * @return  Retrieval mode from {@link QueryConstants}.
    */
   int getIteration()
   {
      return iteration;
   }
   
   /**
    * Set iteration/retrieval mode (first, last, next, etc.).
    *
    * @param   iteration
    *          Retrieval mode from {@link QueryConstants}.
    * @param   sort
    *          Order by clause which should be associated with this query component. Must be
    *          non-<code>null</code> for iteration types <code>FIRST</code> and
    *          <code>LAST</code>.
    * 
    * @throws  IllegalArgumentException
    *          if <code>iteration</code> type is <code>FIRST</code> or <code>LAST</code>, but
    *          <code>sort</code> is <code>null</code>.
    * @throws  PersistenceException
    *          if <code>sort</code> clause is invalid.
    */
   void setIteration(int iteration, String sort)
   throws PersistenceException
   {
      if (iteration == FIRST || iteration == LAST)
      {
         if (sort == null)
         {
            throw new IllegalArgumentException(
               "Sort cannot be null with FIRST/LAST iteration type");
         }
         
         // generate an order by clause using no DMO aliases, suitable for a sub-select
         List<SortCriterion> criteria = SortCriterion.parse(sort, buffer);
         StringBuilder buf = new StringBuilder();
         for (SortCriterion sc : criteria)
         {
            if (buf.length() > 0)
            {
               buf.append(", ");
            }
            
            buf.append(sc.toSortExpression(iteration == LAST, false, true));
         }
         
         this.sort = buf.toString();
      }
      else
      {
         this.sort = null;
      }
      
      this.iteration = iteration;
   }
   
   /**
    * Mark this query component as being nested within an outer component (i.e., not the top
    * component in a nested looping query).
    */
   void setNotTop()
   {
      this.top = false;
   }
   
   /**
    * Indicate whether this component is joined with the previous component
    * via a left outer join.
    *
    * @return  <code>true</code> if an outer join, <code>false</code> if
    *          an inner join or no join (i.e., no previous component).
    */
   boolean isOuter()
   {
      return outer;
   }
   
   /**
    * Get an iterator on the ANSI-style join subexpressions which must be
    * merged by the enclosing query into the overall HQL query statement
    * in order to properly join associated composite element lists (if
    * any).
    *
    * @return  An iterator on join subexpressions.  The iterator may be
    *          empty, but will not be <code>null</code>.
    * 
    * @throws  PersistenceException
    *          if there is an error creating the
    *          <code>HQLPreprocessor</code> from which the composite join
    *          subexpressions are gathered.
    */
   Iterator<String> compositeJoins()
   throws PersistenceException
   {
      return getHQLPreprocessor().ansiJoins();
   }
   
   /**
    * Get an unmodifiable map of DMO entity names to sets of the names of
    * the DMO properties used as restriction criteria in the preprocessed
    * where clause.
    *
    * @return  Map of DMO entity names to sets of property names, or
    *          <code>null</code> if the where clause for this component
    *          contained no restriction properties.
    * 
    * @throws  PersistenceException
    *          if there is an error creating the
    *          <code>HQLPreprocessor</code> from which the restriction
    *          properties are gathered.
    */
   Map<String, Set<String>> getRestrictionProperties()
   throws PersistenceException
   {
      return getHQLPreprocessor().getRestrictionProperties();
   }
   
   /**
    * Get sort criteria in HQL order by clause format for this component.
    *
    * @return  Sort index associated with this component.
    */
   SortIndex getSortIndex()
   {
      return sortIndex;
   }
   
   /**
    * Set sort criteria in HQL order by clause format for this component.
    * 
    * @param   sortIndex
    *          Sort index.
    */
   void setSortIndex(SortIndex sortIndex)
   {
      this.sortIndex = sortIndex;
   }
   
   /**
    * Get the index information string as it is returned by INDEX-INFORMATION.
    * 
    * @return  Index information string.
    */
   String getIndexInfo()
   {
      return indexInfo;
   }
   
   /**
    * Information required to create a server-side join between two tables.
    */
   static class ServerJoinData
   {
      /** Helper object which generates a where clause snippet to join two tables */
      final AbstractJoin join;
      
      /** Query substitution arguments used in the where clause snippet, which survive the join */
      final Object[] args;
      
      /** Set of related buffers referenced by the join */
      final Set<RecordBuffer> relatedBuffers;
      
      /** Array of field reference names and nulls */
      final String[] referenceSubs;
      
      /**
       * Constructor.
       * 
       * @param   join
       *          Helper object which generates a where clause snippet to join two tables.
       * @param   args
       *          Query substitution arguments used in the where clause snippet, which survive
       *          the join.
       * @param   relatedBuffers
       *          Set of related buffers referenced by the join.
       * @param   referenceSubs
       *          Array of field reference names and nulls, where the names are to be substituted
       *          for the positionally equivalent substitution parameters when converting a
       *          where clause used by a nested loop to a where clause using a server-side join.
       */
      ServerJoinData(AbstractJoin join,
                     Object[] args,
                     Set<RecordBuffer> relatedBuffers,
                     String[] referenceSubs)
      {
         this.join = join;
         this.args = args;
         this.relatedBuffers = relatedBuffers;
         this.referenceSubs = referenceSubs;
      }
   }
}