/*
** Module   : RandomAccessQuery.java
** Abstract : Backing query for converted queries which require dynamic record retrieval
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 ECF 20051007   @23259 Created initial version. Backing query for
**                           converted queries which require a dynamic
**                           record retrieval semantic.
** 002 ECF 20051111   @23386 Adapted to use with CompoundQuery. Implements
**                           Joinable, which is used by CompoundQuery when
**                           adding query components.
** 003 ECF 20051129   @23574 Added lenient-off-end mode. When in this
**                           mode, QueryOffEndExceptions thrown by the
**                           underlying buffer during next and previous
**                           operations are suppressed.
** 004 ECF 20051216   @23747 Added scrolling support. Now extends the new
**                           DynamicQuery base class, which provides
**                           cursoring support services. Modified
**                           retrieval logic to allow for repositioning.
** 005 ECF 20060104   @23828 Added accumulation support. Record retrieval
**                           methods first, last, next, and previous now
**                           update any registered accumulators after
**                           successfully retrieving a record.
** 006 GES 20060202   @24226 Removed main().
** 007 ECF 20060202   @24355 Added temp table support. Add buffer's
**                           multiplex ID as a query substitution argument
**                           for temp table queries.
** 008 ECF 20060208   @24365 Changed arguments processing in constructor.
**                           Store duplicates of BaseDataType arguments
**                           rather than references to the original
**                           objects.
** 009 ECF 20060215   @24723 Added support for client-side where clause
**                           processing. Major changes to execute() were
**                           necessary to support this. All constructors
**                           now accept an optional WhereExpression
**                           parameter. Also improved handling of query
**                           substitution parameter resolution to be more
**                           correct.
** 010 ECF 20060302   @25023 Add support for foreign key joins. Added new
**                           constructor variants which mirror existing
**                           constructors, but accept an additional,
**                           inverse DMO parameter. Modified query
**                           execution to integrate natural joins.
** 011 ECF 20060317   @25171 Fixed defect related to extent fields. Use
**                           mapping of query substitution parameters,
**                           which may have been reordered during HQL
**                           where clause preprocessing.
** 012 ECF 20060330   @25256 Removed final modifier from class. FindQuery
**                           extends this class.
** 013 ECF 20060424   @25696 Added new constructor variants. Each existing
**                           variant now has a counterpart which does not
**                           expect a WhereExpression parameter.
** 014 ECF 20060424   @25700 Roll back previous change.
** 015 ECF 20060529   @26645 Added initializeBuffer method. Resets buffer
**                           during query construction.
** 016 ECF 20060531   @26851 Replaced EndConditionException with
**                           QueryOffEndException. The latter is a more
**                           specific version of the former, used to
**                           indicate a query has run off the end of its
**                           result set.
** 017 ECF 20060610   @27102 Fixed lock processing defect. A record must
**                           not be stored in the buffer after a NO-WAIT
**                           record lock attempt fails.
** 018 ECF 20060621   @27506 Logging improvements.
** 019 ECF 20060713   @28046 Use new DBUtils class.
** 020 ECF 20060720   @28118 Minor changes driven by RecordBuffer naming
**                           changes. RecordBuffer.getLatestRecord() was
**                           changed to getSnapshot().
** 021 ECF 20060728   @28272 Minor change to accommodate RecordBuffer
**                           changes. Enclose RecordBuffer.setTempRecord()
**                           call in a try-catch block.
** 022 ECF 20060802   @28353 Added break value support. Implemented new
**                           getBreakValue() method.
** 023 ECF 20060803   @28404 Added updateBuffer() method. Allows
**                           subclasses to override setting current record
**                           into buffer.
** 024 ECF 20060925   @29964 Check for a pending error before retrieving a
**                           record. Retrieval is not attempted if an
**                           error is pending in the ErrorManager.
** 025 ECF 20061003   @30155 Call prepareFetch() instead of obsolete
**                           method checkPendingError(). Implemented
**                           releaseBuffers() to override do-nothing
**                           AbstractQuery implementation.
** 026 ECF 20061019   @30530 Changed reset() method. If the query is in
**                           scrolling mode, the cursor is reset as part
**                           of this method's processing.
** 027 ECF 20061023   @30609 Moved errorIfNull logic to AbstractQuery base
**                           class. Method setErrorIfNull() is now
**                           required by the P2JQuery interface and is
**                           implemented in AbstractQuery instead of here.
** 028 ECF 20061027   @30755 Replaced Object with DataModelObject for DMO
**                           arguments to methods/constructors. Required
**                           for compile-time type safety.
** 029 ECF 20061115   @31210 Changed call to set current buffer record.
**                           RecordBuffer API has changed to permit proper
**                           management of placeholder snapshot record.
**                           Added isResetSnapshot() method to allow
**                           subclasses to override this behavior.
** 030 ECF 20070216   @32165 Explicitly set errorIfNull to false. This is
**                           required to override the default setting of
**                           true made by AbstractQuery.
** 031 ECF 20070228   @32248 Optimized no-lock and temp table fetches. In
**                           these cases, it is not necessary to fetch the
**                           record in two steps. The no-lock case makes
**                           no guarantee that the record in memory has
**                           not been updated by other sessions. The temp
**                           table case guarantees that no other session
**                           can make an update to the record in memory,
**                           because a temp table record is inherently
**                           scoped to a session.
** 032 ECF 20070308   @32335 Fixed behavior for first/last. Upon failure
**                           to find a record, if not in looping mode,
**                           these methods must report a 'FIND FIRST/LAST
**                           failed...' message.
** 033 ECF 20070412   @32985 Retrofit to use RecordLockContext for record
**                           locking. Accommodate method name change in
**                           RecordBuffer: setCurrentRecord --> setRecord.
** 034 ECF 20070416   @33028 Integrated user interrupt handling.
** 035 ECF 20070504   @33410 Retrofitted load() method to use modified
**                           signature specified by Joinable.
** 036 ECF 20070508   @33460 Fixed executeImpl() to properly handle a
**                           validation exception. New requirement based
**                           on a change to RecordBuffer.flush().
** 037 ECF 20070518   @33684 Added getRecordBuffers() and recordBuffers().
**                           The former is required by the Joinable
**                           interface. The latter is required by the
**                           RecordChangeListener interface.
** 038 EVL 20070609   @34005 Adding explicit import of the class
**                           org.hibernate.type.Type to eliminate conflict
**                           with the same class from java.lang.reflect
**                           package to be able to compile for Java 6.
** 039 ECF 20070718   @34571 Fixed findNext()/findPrevious() and their
**                           callers. These methods were not properly
**                           updating locks when necessary.
** 040 ECF 20070713   @34858 Moved all CAN-FIND specific processing to the
**                           FindQuery subclass. Exposed some private
**                           methods as protected and refactored class to
**                           support this.
** 041 ECF 20070827   @35002 Minor optimization for temp table queries.
**                           Omit multiplex ID from query substitution
**                           parameters. HQLHelper now hard-codes the ID
**                           into the where clause, since it is known at
**                           the time the where clause is generated.
** 042 ECF 20070926   @35246 Fix placeholder record defect. In cases where
**                           the referenceRecord DMO is modified, we need
**                           to null it out and use the buffer's current
**                           snapshot as our placeholder record instead.
** 043 ECF 20071130   @36123 Removed isRefreshSnapshot() method. The
**                           RecordBuffer placeholder snapshot is no
**                           longer cleared by any query type when the
**                           query does not find a record to place in the
**                           buffer. Integrated generics.
** 044 ECF 20071204   @36266 Fixed off-end tracking. Replaced boolean flag
**                           with OffEnd enum to discern between off-end
**                           directions (i.e., off back vs off front of
**                           result set). Current off-end status and
**                           active sort index are reported to the record
**                           buffer when fetching a record. Off-end status
**                           is checked when searching for a new record.
** 045 ECF 20080112   @37006 Removed restriction of dealing only with IDs.
**                           This query can now handle both Persistables
**                           and primary key IDs when retrieving data.
**                           This is necessary when the full record is
**                           needed immediately, not just an ID, such as
**                           in presort cases.
** 046 CA  20080311   @37430 Removed the listener code from c'tor and
**                           added it to the new registerChangeListener()
**                           method.
** 047 ECF 20080310   @37482 Made changes necessary to support embedded,
**                           H2 database.
** 048 SVL 20080418   @38041 Added support for the inverse sorting.
** 049 SVL 20080421   @38082 Added support for legacy key join helper.
** 050 ECF 20080508   @38245 Override base lock type for temporary tables.
**                           This allows some downstream optimizations.
** 051 ECF 20080510   @38610 Implemented support for dirty database
**                           checking. Looks into a shared database of
**                           uncommitted updates and possibly overrides
**                           data found in primary database.
** 052 ECF 20080609   @38652 Removed overly aggressive optimization in
**                           executeImpl(). Queries with non-NO-LOCK types
**                           must also search among uncommitted updates,
**                           and locks must be acquired on dirty records.
** 053 ECF 20080610   @38690 Fixed dirty check processing in executeImpl()
**                           method. We were not correctly handling the
**                           case of finding a dirty version of the same
**                           DMO found in the primary database.
** 054 ECF 20080611   @38712 Fixed processResults(). When acquiring a lock
**                           for a dirty record, the record should be
**                           reloaded, in case it disappeared while we
**                           were waiting on the lock.
** 055 CA  20080618   @38868 When the JOINed foreign buffer does not refer
**                           any record, the query must return no result
**                           as it can not be executed.
** 056 SVL 20080630   @38985 Fixed executeImpl() for the case when a dirty
**                           record is inserted and modified.
** 057 SVL 20080730   @39222 Added dmoSorter instance variable.
** 058 CA  20080815   @39465 Support API change in DBUtils.
** 060 ECF 20081009   @40082 Integrated new support for inlining query
**                           substitution parameters. Queries supported by
**                           this class don't actually have their
**                           parameters inlined, but changes were needed
**                           to integrate HQLPreprocessor API changes and
**                           the ParameterIndices class.
** 061 ECF 20081105   @40312 Implemented registerRecordChangeListeners().
**                           Required by the Joinable interface.
** 062 ECF 20081215   @40917 Modified executeImpl(). A validation error
**                           ignores silent error mode, since the flushing
**                           of the previous record in the buffer is not
**                           masked by silent error mode.
** 063 ECF 20090303   @41599 Fixed to better support being a Joinable
**                           component of a CompoundQuery. Implemented
**                           RecordChangeListener, added getOffEnd()
**                           method. Replaced activeBundleKey Object with
**                           an int to support changes in HQLHelper.
** 064 ECF 20090317   @41627 Removed early calls to getHelper(). This
**                           method must not be invoked until the query is
**                           actually being executed, because of possible
**                           dependencies in HQLHelper.obtain() on the
**                           state of the query's substitution parameters.
**                           Specifically, a FieldReference parameter may
**                           not be backed by a record if this method is
**                           invoked arbitrarily early.
** 065 ECF 20090512   @42154 Modified prepareBuffer(). Now invokes flushAll()
**                           instead of flush() on backing buffer. This
**                           ensures all like buffers are allowed to flush
**                           their transient records properly.
** 066 ECF 20090603   @42596 Added support for Progress isolation leak quirk.
**                           Modified executeImpl() and processResults() to
**                           use new features in HQLPreprocessor and in
**                           DirtyShareContext to leak uncommitted changes to
**                           other sessions and to select appropriate record.
** 067 ECF 20090609   @42635 Fixed executeImpl(). We now evict, if possible,
**                           those DMOs read into the Hibernate Session, but
**                           not loaded into the record buffer. This includes
**                           DMOs whose uncommitted changes are leaked to
**                           other sessions early, and DMOs which are found in
**                           the primary database, but overridden by records
**                           found by the dirty share manager.
** 068 ECF 20090623   @42941 Fixed executeImpl(). Handle the case of a unique
**                           search finding information in the dirty database.
**                           Also added back multiplex ID as a substitution
**                           parameter for temp table queries.
** 069 ECF 20090703   @43058 Reimplemented load() method. This method now sets
**                           the backing buffer to unknown mode and throws
**                           MissingRecordException if a target record cannot
**                           be loaded.
** 070 ECF 20090716   @43221 Enabled lazy record buffer initialization. Added
**                           RecordBuffer.initialize() call in c'tor.
** 071 ECF 20090724   @43421 Added implicit transaction support. In various
**                           methods which execute queries, we push/pop an
**                           implicit transaction.
** 072 CA  20090731   @43470 Register the query for off-end listener 
**                           registration after block initialization (done
**                           only for block types FOR and FOR EACH). Added
**                           getOffEndListeners, which returns a list of all
**                           off-end listeners for all used buffers.
** 073 SVL 20091030   @44293 getOffEnd() function was made public in order to
**                           support API change into P2JQuery. 
** 074 OM  20130829          Added change detection to buffer on current() methods.
**                           Updated calls to match the new Persistence.load() signatures.
** 075 ECF 20131028          Import change.
** 076 OM  20140107          Added TODO for missing fatal error message.
** 077 SVL 20140604          Added support for index information field.
** 078 ECF 20140814          Replaced RecordBuffer.flushAll with flush in prepareBuffer method.
**                           Fixed dirty share logic to work with new flush/validate algorithm.
**                           Replaced Apache commons logging with J2SE logging.
** 079 OM  20150507          Removed listener from global scope of ChangeBroker.
** 080 ECF 20150801          Implemented new Joinable methods necessary for compound query
**                           optimization. Error message fix. Enabled database statistics.
** 081 ECF 20150906          Reimplemented silent error mode processing of query substitution
**                           parameters. Fixed NPE related to outer joins.
** 082 SVL 20151112          Continue normally if reposition failed in load().
** 083 EVL 20160223          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 084 SVL 20160314          Added setter for referenceRecord.
** 085 OM  20160316          Signature of DirtyShareContext.getDirtyInfo() has changed.
**     ECF 20160504          Pass new parameter to Persistence.load. Minor change to isIdOnly.
** 086 SVL 20160608          The query can handle null IDs (for scrolling compound queries
**                           with OUTER join).
** 087 ECF 20160615          Bias toward fetching full records when possible.
** 088 IAS 20160701          Fixed legacy name in the [565] error message.
** 089 IAS 20160714          Use legacy name instead of DMO alias.
** 090 OM  20160715          Added support for template records.
**                           Persistence.popImplicitTransaction() changed signature.
** 091 ECF 20160725          Fixed buffer lookup during HQL preprocessing.
** 092 ECF 20160823          Improved executeImpl to better detect missing records and avoid
**                           deadlock.
** 093 ECF 20160901          DirtyShareContext.isDirtyDelete signature change.
** 094 GES 20160804          Switched from WhereExpression to lambda usage.
**         20160919          All initialization is now done with a default constructor followed
**                           by a call to initialize(). This enables query instance members in
**                           converted code to be local variables that are effectively final.
**                           This is needed to support lambdas without moving queries to be
**                           instance members. Queries as instance members breaks recursion
**                           use cases.
** 095 CA  20161012          Fixed scope processing when the query is from a persistent 
**                           procedure, and is part of a QUERY resource.
** 096 ECF 20161014          Fixed implementation of current() method in the event there is no
**                           current record in the buffer to reload, and this is not considered
**                           an error in the context of the query (e.g., query represents an
**                           outer join and record can be null).
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import java.util.function.*;
import java.util.logging.*;
import org.hibernate.type.Type;
import com.goldencode.p2j.persist.dirty.*;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.LogHelper;
import com.goldencode.p2j.util.ErrorManager;
import com.goldencode.util.*;

/**
 * The backing query for converted queries which require a dynamic record
 * retrieval semantic.  This means that a single record is retrieved at a
 * time and the query manages state about that record's position relative
 * to others which also match the query's search criteria.  Furthermore,
 * any change to that record's data may change its position relative to other
 * matching records, and these changes must be supported while client code is
 * using the query to visit result records.  Client code can jump among
 * results randomly.  These moves are specified as either relative requests
 * (next, previous, current), or as absolute requests (first, last, unique).
 * <p>
 * This class slavishly implements the Progress semantic of fully dynamic
 * query iteration.  That is, each record is searched for and retrieved
 * individually, so that any change to an existing record can be reflected
 * immediately in the remaining navigation performed on the query.  This
 * means that an update to the current record may cause it to reappear later
 * in the query results, if the changes made to the record would cause it to
 * be sorted differently than when it was first retrieved (note that this can
 * cause infinite looping, just as in the pre-conversion, Progress code).
 * <p>
 * To implement the single-record retrieval semantic, this high level query
 * may issue multiple, database-level, select statements to the database for
 * each record retrieval attempt.  For details of how this is performed, see
 * {@link HQLHelper} and {@link HQLBundle}.  The result of a record retrieval
 * is a Data Model Object (DMO), which is stored in a {@link RecordBuffer}.
 * A record buffer is associated with this object at construction.
 * <p>
 * Because of this implementation requirement, <code>RandomAccessQuery</code>
 * represents <i>an extremely inefficient mechanism</i> by which to query a
 * database for a set of records.  As such, <b>it should not be used for new
 * development;  it is intended only to support legacy, converted code</b>,
 * insofar as that code may rely on the dynamic nature of the Progress
 * semantic.
 * <p>
 * This class implements the <code>Joinable</code> interface so that it can
 * be added as a component to a {@link CompoundQuery}.  When used in such a
 * capacity, the query should be configured to raise an end condition instead
 * of an error condition, in the case that a request to get the first, last,
 * or a unique record fails.  This will allow the compound query to continue
 * processing normally in this case.  This is accomplished by invoking {@link
 * #setErrorIfNull} with the parameter <code>true</code>.
 * <p>
 * When used to represent a Progress query (as created with DEFINE or OPEN
 * QUERY), {@link #setLenientOffEnd} must be set to <code>true</code>, which
 * suppresses end condition exceptions for <code>next</code> and
 * <code>previous</code> commands which run off the end of the query's
 * results.
 * <p>
 * <strong>Transaction Isolation (Dirty Reads)</strong>
 * <br>
 * This class supports Progress' peculiar form of transaction isolation, where
 * uncommitted updates, inserts, and deletes that affect database indexes are
 * visible across sessions.  This is emulated using a shared database (the
 * "dirty" database) which temporarily stores information about uncommitted
 * changes made by other contexts.  When those other sessions commit or roll
 * back their changes, this transient information is cleared from the dirty
 * database and is no longer visible.
 * <p>
 * The implementation of this feature requires that the results found in the
 * primary database backing a query be treated as provisional. Results found
 * in the dirty database may override, if it is determined that the "dirty"
 * result is the more correct result, given the query criteria.
 */
public class RandomAccessQuery
extends DynamicQuery
implements Joinable,
           QueryConstants,
           RecordChangeListener
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(RandomAccessQuery.class.getName());
   
   /** Record buffer which backs this query */
   private RecordBuffer buffer;
   
   /** Helper object for dynamic join via a foreign key relation */
   private AbstractJoin join;
   
   /** Client-side where clause expression, if any */
   private Supplier<logical> whereExpr;
   
   /** Base substitution parameters for HQL queries */
   private Object[] args;
   
   /** Lock type to apply to records retrieved */
   private LockType lockType;
   
   /** Original where clause (only stored until helper is built) */
   private String where = null;
   
   /** Original sort clause (only stored until helper is built) */
   private String sort = null;
   
   /** DMO sorter associated with query's sort clause */
   private DMOSorter dmoSorter = null;
   
   /** Name of index in use for this query */
   private String index = null;
   
   /** Index information string as it is returned by INDEX-INFORMATION. */
   private String indexInfo = null;
   
   /** Builds HQL statements for various Hibernate queryies */
   private HQLHelper helper = null;
   
   /** HQL data necessary for the current query operation */
   private HQLBundle activeBundle = null;
   
   /** Key which indicates active bundle choice */
   private int activeBundleKey = NONE;
   
   /** <code>false</code> if running off end of results should raise error */
   private boolean lenientOffEnd = false;
   
   /** Substitution parameter values for the next query execution */
   private Object[] currentArgs = null;
   
   /** Record whose data defines current "place" in query results */
   private Persistable referenceRecord = null;
   
   /** Break value indicating a new sort band if non-<code>null</code> */
   private Object breakValue = null;
   
   /** Force query to retrieve full records, rather than primary keys only */
   private boolean fullRecords = false;
   
   /** Was last record found retrieved from the dirty database? */
   private boolean dirtyCopy = false;
   
   /** RecordChangeListener which nulls out placeholder reference record */
   private RecordChangeListener placeholderCleaner = null;
   
   /** Determines if this will be unregistered from ChangeBroker on cleanup. */
   private boolean unregisterOnCleanup = false;
   
   /**
    * Default constructor. Initialization is deferred until <code>initialize()</code> is
    * called.
    */
   public RandomAccessQuery()
   {
   }
   
   /**
    * Initialization logic which defaults record lock type to <code>LockType.SHARE</code>.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           Data model object which determines the record buffer into which
    *           records are retrieved.
    * @param    where
    *           Where clause, using HQL syntax.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           Order by clause, using HQL syntax.  May be <code>null</code>
    *           only if query is to be navigated only using one of the
    *           <code>unique</code> method variants.
    *            
    * @return   This query instance.
    */
   public RandomAccessQuery initialize(DataModelObject   dmo,
                                       String            where,
                                       Supplier<logical> whereExpr,
                                       String            sort)
   {
      return initialize(dmo, where, whereExpr, sort, (String) null);
   }
   
   /**
    * Initialization logic which defaults record lock type to <code>LockType.SHARE</code>.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           Data model object which determines the record buffer into which
    *           records are retrieved.
    * @param    where
    *           Where clause, using HQL syntax.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    indexInfo
    *           Index information string as it is returned by INDEX-INFORMATION.
    *           May be <code>null</code> if it is not an OPEN QUERY case or if
    *           you don't need debug information about selected indexes.
    * @param    sort
    *           Order by clause, using HQL syntax.  May be <code>null</code>
    *           only if query is to be navigated only using one of the
    *           <code>unique</code> method variants.
    *            
    * @return   This query instance.
    */
   public RandomAccessQuery initialize(DataModelObject   dmo,
                                       String            where,
                                       Supplier<logical> whereExpr,
                                       String            sort,
                                       String            indexInfo)
   {
      return initialize(dmo, where, whereExpr, sort, indexInfo, null, null, LockType.SHARE);
   }
   
   /**
    * Initialization logic which is used when joining to another table using a foreign
    * key.  Defaults record lock type to <code>LockType.SHARE</code>.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           Data model object which determines the record buffer into which
    *           records are retrieved.
    * @param    where
    *           Where clause, using HQL syntax.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           Order by clause, using HQL syntax.  May be <code>null</code>
    *           only if query is to be navigated only using one of the
    *           <code>unique</code> method variants.
    * @param    inverse
    *           DMO to which this query should join via a foreign relation.
    *           May be <code>null</code>.
    *            
    * @return   This query instance.
    */
   public RandomAccessQuery initialize(DataModelObject   dmo,
                                       String            where,
                                       Supplier<logical> whereExpr,
                                       String            sort,
                                       DataModelObject   inverse)
   {
      return initialize(dmo, where, whereExpr, sort, null, inverse);
   }
   
   /**
    * Initialization logic which is used when joining to another table using a foreign
    * key.  Defaults record lock type to <code>LockType.SHARE</code>.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           Data model object which determines the record buffer into which
    *           records are retrieved.
    * @param    where
    *           Where clause, using HQL syntax.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           Order by clause, using HQL syntax.  May be <code>null</code>
    *           only if query is to be navigated only using one of the
    *           <code>unique</code> method variants.
    * @param    indexInfo
    *           Index information string as it is returned by INDEX-INFORMATION.
    *           May be <code>null</code> if it is not an OPEN QUERY case or if
    *           you don't need debug information about selected indexes.
    * @param    inverse
    *           DMO to which this query should join via a foreign relation.
    *           May be <code>null</code>.
    *            
    * @return   This query instance.
    */
   public RandomAccessQuery initialize(DataModelObject   dmo,
                                       String            where,
                                       Supplier<logical> whereExpr,
                                       String            sort,
                                       String            indexInfo,
                                       DataModelObject   inverse)
   {
      return initialize(dmo, where, whereExpr, sort, indexInfo, inverse, null, LockType.SHARE);
   }
   
   /**
    * Initialization logic which defaults record lock type to
    * <code>LockType.SHARE</code> and accepts default substitution parameters.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           Data model object which determines the record buffer into which
    *           records are retrieved.
    * @param    where
    *           Where clause, using HQL syntax.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           Order by clause, using HQL syntax.  May be <code>null</code>
    *           only if query is to be navigated only using one of the
    *           <code>unique</code> method variants.
    * @param    args
    *           Substitution parameters for HQL queries.  These will be used
    *           if not overridden by a record retrieval method.
    *            
    * @return   This query instance.
    *           
    * @throws   IllegalArgumentException
    *           if any substitution argument provided is not of a supported
    *           type.
    */
   public RandomAccessQuery initialize(DataModelObject   dmo,
                                       String            where,
                                       Supplier<logical> whereExpr,
                                       String            sort,
                                       Object[]          args)
   {
      return initialize(dmo, where, whereExpr, sort, (String) null, args);
   }
   
   /**
    * Initialization logic which defaults record lock type to
    * <code>LockType.SHARE</code> and accepts default substitution parameters.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           Data model object which determines the record buffer into which
    *           records are retrieved.
    * @param    where
    *           Where clause, using HQL syntax.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           Order by clause, using HQL syntax.  May be <code>null</code>
    *           only if query is to be navigated only using one of the
    *           <code>unique</code> method variants.
    * @param    indexInfo
    *           Index information string as it is returned by INDEX-INFORMATION.
    *           May be <code>null</code> if it is not an OPEN QUERY case or if
    *           you don't need debug information about selected indexes.
    * @param    args
    *           Substitution parameters for HQL queries.  These will be used
    *           if not overridden by a record retrieval method.
    *            
    * @return   This query instance.
    *           
    * @throws   IllegalArgumentException
    *           if any substitution argument provided is not of a supported
    *           type.
    */
   public RandomAccessQuery initialize(DataModelObject   dmo,
                                       String            where,
                                       Supplier<logical> whereExpr,
                                       String            sort,
                                       String            indexInfo,
                                       Object[]          args)
   {
      return initialize(dmo, where, whereExpr, sort, indexInfo, null, args, LockType.SHARE);
   }
   
   /**
    * Initialization logic which is used when joining to another table using a foreign
    * key.  Defaults record lock type to <code>LockType.SHARE</code> and
    * accepts default substitution parameters.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           Data model object which determines the record buffer into which
    *           records are retrieved.
    * @param    where
    *           Where clause, using HQL syntax.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           Order by clause, using HQL syntax.  May be <code>null</code>
    *           only if query is to be navigated only using one of the
    *           <code>unique</code> method variants.
    * @param    inverse
    *           DMO to which this query should join via a foreign relation.
    *           May be <code>null</code>.
    * @param    args
    *           Substitution parameters for HQL queries.  These will be used
    *           if not overridden by a record retrieval method.
    *            
    * @return   This query instance.
    *           
    * @throws   IllegalArgumentException
    *           if any substitution argument provided is not of a supported
    *           type.
    */
   public RandomAccessQuery initialize(DataModelObject   dmo,
                                       String            where,
                                       Supplier<logical> whereExpr,
                                       String            sort,
                                       DataModelObject   inverse,
                                       Object[]          args)
   {
      return initialize(dmo, where, whereExpr, sort, null, inverse, args);
   }
   
   /**
    * Initialization logic which is used when joining to another table using a foreign
    * key.  Defaults record lock type to <code>LockType.SHARE</code> and
    * accepts default substitution parameters.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           Data model object which determines the record buffer into which
    *           records are retrieved.
    * @param    where
    *           Where clause, using HQL syntax.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           Order by clause, using HQL syntax.  May be <code>null</code>
    *           only if query is to be navigated only using one of the
    *           <code>unique</code> method variants.
    * @param    indexInfo
    *           Index information string as it is returned by INDEX-INFORMATION.
    *           May be <code>null</code> if it is not an OPEN QUERY case or if
    *           you don't need debug information about selected indexes.
    * @param    inverse
    *           DMO to which this query should join via a foreign relation.
    *           May be <code>null</code>.
    * @param    args
    *           Substitution parameters for HQL queries.  These will be used
    *           if not overridden by a record retrieval method.
    *            
    * @return   This query instance.
    *           
    * @throws   IllegalArgumentException
    *           if any substitution argument provided is not of a supported
    *           type.
    */
   public RandomAccessQuery initialize(DataModelObject   dmo,
                                       String            where,
                                       Supplier<logical> whereExpr,
                                       String            sort,
                                       String            indexInfo,
                                       DataModelObject   inverse,
                                       Object[]          args)
   {
      return initialize(dmo, where, whereExpr, sort, indexInfo, inverse, args, LockType.SHARE);
   }
   
   /**
    * Initialization logic which sets an explicit, default, record lock type.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           Data model object which determines the record buffer into which
    *           records are retrieved.
    * @param    where
    *           Where clause, using HQL syntax.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           Order by clause, using HQL syntax.  May be <code>null</code>
    *           only if query is to be navigated only using one of the
    *           <code>unique</code> method variants.
    * @param    lockType
    *           Lock type to apply to records retrieved, if not overridden by
    *           a record retrieval method.
    *            
    * @return   This query instance.
    */
   public RandomAccessQuery initialize(DataModelObject   dmo,
                                       String            where,
                                       Supplier<logical> whereExpr,
                                       String            sort,
                                       LockType          lockType)
   {
      return initialize(dmo, where, whereExpr, sort, (String) null, lockType);
   }
   
   /**
    * Initialization logic which sets an explicit, default, record lock type.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           Data model object which determines the record buffer into which
    *           records are retrieved.
    * @param    where
    *           Where clause, using HQL syntax.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           Order by clause, using HQL syntax.  May be <code>null</code>
    *           only if query is to be navigated only using one of the
    *           <code>unique</code> method variants.
    * @param    indexInfo
    *           Index information string as it is returned by INDEX-INFORMATION.
    *           May be <code>null</code> if it is not an OPEN QUERY case or if
    *           you don't need debug information about selected indexes.
    * @param    lockType
    *           Lock type to apply to records retrieved, if not overridden by
    *           a record retrieval method.
    *            
    * @return   This query instance.
    */
   public RandomAccessQuery initialize(DataModelObject   dmo,
                                       String            where,
                                       Supplier<logical> whereExpr,
                                       String            sort,
                                       String            indexInfo,
                                       LockType          lockType)
   {
      return initialize(dmo, where, whereExpr, sort, indexInfo, null, null, lockType);
   }
   
   /**
    * Initialization logic which is used when joining to another table using a foreign
    * key.  Sets an explicit, default, record lock type.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           Data model object which determines the record buffer into which
    *           records are retrieved.
    * @param    where
    *           Where clause, using HQL syntax.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           Order by clause, using HQL syntax.  May be <code>null</code>
    *           only if query is to be navigated only using one of the
    *           <code>unique</code> method variants.
    * @param    inverse
    *           DMO to which this query should join via a foreign relation.
    *           May be <code>null</code>.
    * @param    lockType
    *           Lock type to apply to records retrieved, if not overridden by
    *           a record retrieval method.
    *            
    * @return   This query instance.
    */
   public RandomAccessQuery initialize(DataModelObject   dmo,
                                       String            where,
                                       Supplier<logical> whereExpr,
                                       String            sort,
                                       DataModelObject   inverse,
                                       LockType          lockType)
   {
      return initialize(dmo, where, whereExpr, sort, null, inverse, lockType);
   }
   
   /**
    * Initialization logic which is used when joining to another table using a foreign
    * key.  Sets an explicit, default, record lock type.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           Data model object which determines the record buffer into which
    *           records are retrieved.
    * @param    where
    *           Where clause, using HQL syntax.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           Order by clause, using HQL syntax.  May be <code>null</code>
    *           only if query is to be navigated only using one of the
    *           <code>unique</code> method variants.
    * @param    indexInfo
    *           Index information string as it is returned by INDEX-INFORMATION.
    *           May be <code>null</code> if it is not an OPEN QUERY case or if
    *           you don't need debug information about selected indexes.
    * @param    inverse
    *           DMO to which this query should join via a foreign relation.
    *           May be <code>null</code>.
    * @param    lockType
    *           Lock type to apply to records retrieved, if not overridden by
    *           a record retrieval method.
    *            
    * @return   This query instance.
    */
   public RandomAccessQuery initialize(DataModelObject   dmo,
                                       String            where,
                                       Supplier<logical> whereExpr,
                                       String            sort,
                                       String            indexInfo,
                                       DataModelObject   inverse,
                                       LockType          lockType)
   {
      return initialize(dmo, where, whereExpr, sort, indexInfo, inverse, null, lockType);
   }
   
   /**
    * Initialization logic which sets an explicit, default, record lock type and
    * accepts default substitution parameters.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           Data model object which determines the record buffer into which
    *           records are retrieved.
    * @param    where
    *           Where clause, using HQL syntax.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           Order by clause, using HQL syntax.  May be <code>null</code>
    *           only if query is to be navigated only using one of the
    *           <code>unique</code> method variants.
    * @param    args
    *           Substitution parameters for HQL queries.  These will be used
    *           if not overridden by a record retrieval method.
    * @param    lockType
    *           Lock type to apply to records retrieved, if not overridden by
    *           a record retrieval method.
    *            
    * @return   This query instance.
    *
    * @throws   IllegalArgumentException
    *           if any substitution argument provided is not of a supported
    *           type.
    */
   public RandomAccessQuery initialize(DataModelObject   dmo,
                                       String            where,
                                       Supplier<logical> whereExpr,
                                       String            sort,
                                       Object[]          args,
                                       LockType          lockType)
   {
      return initialize(dmo, where, whereExpr, sort, (String) null, args, lockType);
   }
   
   /**
    * Initialization logic which sets an explicit, default, record lock type and
    * accepts default substitution parameters.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           Data model object which determines the record buffer into which
    *           records are retrieved.
    * @param    where
    *           Where clause, using HQL syntax.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           Order by clause, using HQL syntax.  May be <code>null</code>
    *           only if query is to be navigated only using one of the
    *           <code>unique</code> method variants.
    * @param    indexInfo
    *           Index information string as it is returned by INDEX-INFORMATION.
    *           May be <code>null</code> if it is not an OPEN QUERY case or if
    *           you don't need debug information about selected indexes.
    * @param    args
    *           Substitution parameters for HQL queries.  These will be used
    *           if not overridden by a record retrieval method.
    * @param    lockType
    *           Lock type to apply to records retrieved, if not overridden by
    *           a record retrieval method.
    *            
    * @return   This query instance.
    *
    * @throws   IllegalArgumentException
    *           if any substitution argument provided is not of a supported
    *           type.
    */
   public RandomAccessQuery initialize(DataModelObject   dmo,
                                       String            where,
                                       Supplier<logical> whereExpr,
                                       String            sort,
                                       String            indexInfo,
                                       Object[]          args,
                                       LockType          lockType)
   {
      return initialize(dmo, where, whereExpr, sort, indexInfo, null, args, lockType);
   }
   
   /**
    * Initialization logic which is used when joining to another table using a foreign
    * key.  Sets an explicit, default, record lock type and accepts default
    * substitution parameters.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           Data model object which determines the record buffer into which
    *           records are retrieved.
    * @param    where
    *           Where clause, using HQL syntax.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           Order by clause, using HQL syntax.  May be <code>null</code>
    *           only if query is to be navigated only using one of the
    *           <code>unique</code> method variants.
    * @param    inverse
    *           DMO to which this query should join via a foreign relation.
    *           May be <code>null</code>.
    * @param    args
    *           Substitution parameters for HQL queries.  These will be used
    *           if not overridden by a record retrieval method.
    * @param    lockType
    *           Lock type to apply to records retrieved, if not overridden by
    *           a record retrieval method.
    *            
    * @return   This query instance.
    *
    * @throws   IllegalArgumentException
    *           if any substitution argument provided is not of a supported
    *           type.
    */
   public RandomAccessQuery initialize(DataModelObject   dmo,
                                       String            where,
                                       Supplier<logical> whereExpr,
                                       String            sort,
                                       DataModelObject   inverse,
                                       Object[]          args,
                                       LockType          lockType)
   {
      return initialize(dmo, where, whereExpr, sort, null, inverse, args, lockType);
   }
   
   /**
    * Initialization logic which is used when joining to another table using a foreign
    * key.  Sets an explicit, default, record lock type and accepts default
    * substitution parameters.
    * <p>
    * This code is intentionally separated from the constructor. The purpose of this separation
    * is to allow construction to occur in the prior scope while initialization can still occur
    * within a method or lambda that is inside the next block's scope.
    *
    * @param    dmo
    *           Data model object which determines the record buffer into which
    *           records are retrieved.
    * @param    where
    *           Where clause, using HQL syntax.  May be <code>null</code>.
    * @param    whereExpr
    *           Client-side where clause expression.  May be <code>null</code>.
    * @param    sort
    *           Order by clause, using HQL syntax.  May be <code>null</code>
    *           only if query is to be navigated only using one of the
    *           <code>unique</code> method variants.
    * @param    indexInfo
    *           Index information string as it is returned by INDEX-INFORMATION.
    *           May be <code>null</code> if it is not an OPEN QUERY case or if
    *           you don't need debug information about selected indexes.
    * @param    inverse
    *           DMO to which this query should join via a foreign relation.
    *           May be <code>null</code>.
    * @param    args
    *           Substitution parameters for HQL queries.  These will be used
    *           if not overridden by a record retrieval method.
    * @param    lockType
    *           Lock type to apply to records retrieved, if not overridden by
    *           a record retrieval method.
    *            
    * @return   This query instance.
    *
    * @throws   IllegalArgumentException
    *           if any substitution argument provided is not of a supported
    *           type.
    */
   public RandomAccessQuery initialize(DataModelObject   dmo,
                                       String            where,
                                       Supplier<logical> whereExpr,
                                       String            sort,
                                       String            indexInfo,
                                       DataModelObject   inverse,
                                       Object[]          args,
                                       LockType          lockType)
   {
      validateSubstitutionArguments(args);
      this.buffer = ((BufferReference) dmo).buffer();
      
      AbstractQuery.preprocessSubstitutionArguments(buffer.getBufferManager(), args);
      
      if (buffer.isReadonly())
      {
         // TODO: readonly buffers cannot be used in any FIND statement
         //       but can be used in CAN-FIND functions
      }
      
      buffer.initialize();
      
      if (!buffer.isTemporary())
      {
         // Get index name, which is needed for dirty checking work.
         try
         {
            IndexHelper indexHelper = buffer.getIndexHelper();
            this.index = indexHelper.getIndexForSort(buffer, sort);
            this.dmoSorter = indexHelper.getSorterForSortPhrase(buffer, sort);
         }
         catch (PersistenceException exc)
         {
            if (LOG.isLoggable(Level.SEVERE))
            {
               LOG.log(Level.SEVERE, "Error locating index for sort phrase [" + sort + "]", exc);
            }
         }
      }
      
      if (inverse != null)
      {
         if (Persistence.isForeignKeysEnabled())
         {
            join = new DynamicJoin(buffer, ((BufferReference) inverse).buffer());
         }
         else
         {
            join = new DynamicLegacyKeyJoin(buffer, ((BufferReference) inverse).buffer());
         }
      }
      else
      {
         join = null;
      }
      this.where = where;
      this.sort = sort;
      this.whereExpr = whereExpr;
      this.args = args;
      this.indexInfo = indexInfo;
      
      // Override lock type if this is a temporary buffer;  this allows
      // some downstream optimizations.
      this.lockType = (buffer.isTemporary() ? LockType.NONE : lockType);
      
      setErrorIfNull(false);
      resolveArgs();
      initializeBuffer();
      registerChangeListener();
      
      TransactionManager.registerOffEndQuery(this);
      
      return this;
   }
   
   /**
    * Force the query to retrieve full records, rather than primary keys only.
    */
   public void setFullRecords()
   {
      this.fullRecords = true;
   }
   
   /**
    * Set the behavior of this query as a record retrieval request moves off
    * the end of its results.  If set to lenient mode, invocations of
    * <code>next</code> and <code>previous</code> which do not find a record
    * do not raise a <code>QueryOffEndException</code>.  By default, this
    * would normally raise the exception.
    *
    * @param   lenientOffEnd
    *          <code>true</code> to suppress <code>QueryOffEndException</code>
    *          for <code>next</code> and <code>previous</code> requests;
    *          <code>false</code> to permit these exceptions.
    */
   public void setLenientOffEnd(boolean lenientOffEnd)
   {
      this.lenientOffEnd = lenientOffEnd;
   }
   
   /**
    * Navigate to the first record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default substitution parameters and
    * lock type.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void first()
   {
      first(currentArgs, lockType);
   }
   
   /**
    * Navigate to the first record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default lock type, but override the
    * default substitution parameters with those provided.
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void first(Object[] values)
   {
      first(values, lockType);
   }
   
   /**
    * Navigate to the first record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default substitution parameters, but
    * override the default lock type with that provided.
    *
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void first(LockType lockType)
   {
      first(currentArgs, lockType);
   }
   
   /**
    * Navigate to the first record which meets the query criteria and retrieve
    * it into its record buffer.  Override both the default substitution
    * parameters, and the default lock type with those provided.
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void first(Object[] values, LockType lockType)
   {
      if (!prepareFetch())
      {
         return;
      }
      
      if (getHelper().getHQLPreprocessor().isFindByRowid()) 
      {
         // direct record access using recid/rowid is behaving like unique for all search types 
         unique(values, lockType);
         return;
      }
      
      // If a lock type is not specified, assume the query's lock type.
      if (lockType == null)
      {
         lockType = this.lockType;
      }
      
      Persistable dmo = null;
      Persistence persistence = buffer.getPersistence();
      boolean pop = false;
      
      try
      {
         persistence.pushImplicitTransaction();
         pop = true;
         
         // if query is scrolling, check the cached result list first
         boolean scrolling = (cursor != null);
         if (scrolling)
         {
            dmo = loadByValue(cursor.getFirst(), lockType);
         }
         
         // if not scrolling or wasn't in the cached result list, perform the query and cache the
         // result (if scrolling)
         if (dmo == null)
         {
            activateFirst();
            dmo = execute(values, lockType, false, true);
            if (scrolling)
            {
               cursor.addResultFirst(getIDs(dmo));
            }
         }
         
         // store result in the record buffer, even if null
         boolean errNull = isErrorIfNull();
         OffEnd offEnd = (dmo == null ? OffEnd.BACK : OffEnd.NONE);
         updateBuffer(dmo, lockType, errNull, offEnd);
         
         // Reset reference record.
         referenceRecord = null;
         
         // Reset dirty flag.
         dirtyCopy = false;
         
         if (dmo == null && errNull)
         {
            errorFindFirstLast();
         }
         
         if (dmo != null)
         {
            // Notify accumulators of an iteration.
            accumulate();
         }
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
      finally
      {
         if (pop)
         {
            persistence.popImplicitTransaction(null);
         }
      }
   }
   
   /**
    * Navigate to the last record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default substitution parameters and
    * lock type.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void last()
   {
      last(currentArgs, lockType);
   }
   
   /**
    * Navigate to the last record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default lock type, but override the
    * default substitution parameters with those provided.
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void last(Object[] values)
   {
      last(values, lockType);
   }
   
   /**
    * Navigate to the last record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default substitution parameters, but
    * override the default lock type with that provided.
    *
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void last(LockType lockType)
   {
      last(currentArgs, lockType);
   }
   
   /**
    * Navigate to the last record which meets the query criteria and retrieve
    * it into its record buffer.  Override both the default substitution
    * parameters, and the default lock type with those provided.
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void last(Object[] values, LockType lockType)
   {
      if (!prepareFetch())
      {
         return;
      }
      
      if (getHelper().getHQLPreprocessor().isFindByRowid())
      {
         // direct record access using recid/rowid is behaving like unique for all search types 
         unique(values, lockType);
         return;
      }
      
      // If a lock type is not specified, assume the query's lock type.
      if (lockType == null)
      {
         lockType = this.lockType;
      }
      
      Persistable dmo = null;
      Persistence persistence = buffer.getPersistence();
      boolean pop = false;
      
      try
      {
         persistence.pushImplicitTransaction();
         pop = true;
         
         // If query is scrolling, check the cached result list first.
         boolean scrolling = (cursor != null);
         if (scrolling)
         {
            dmo = loadByValue(cursor.getLast(), lockType);
         }
         
         // If not scrolling or wasn't in the cached result list, perform the
         // query and cache the result (if scrolling).
         if (dmo == null)
         {
            activateLast();
            dmo = execute(values, lockType, false, true);
            if (scrolling)
            {
               cursor.addResultLast(getIDs(dmo));
            }
         }
         
         // Store result in the record buffer, even if null.
         boolean errNull = isErrorIfNull();
         OffEnd offEnd = (dmo == null ? OffEnd.BACK : OffEnd.NONE);
         updateBuffer(dmo, lockType, errNull, offEnd);
         
         // Reset reference record.
         referenceRecord = null;
         
         // Reset dirty flag.
         dirtyCopy = false;
         
         if (dmo == null && errNull)
         {
            errorFindFirstLast();
         }
         
         if (dmo != null)
         {
            // Notify accumulators of an iteration.
            accumulate();
         }
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
      finally
      {
         if (pop)
         {
            persistence.popImplicitTransaction(null);
         }
      }
   }
   
   /**
    * Navigate to the next record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default substitution parameters and
    * lock type.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the next record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the first record
    * which meets the query criteria.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    * @throws  QueryOffEndException
    *          if no record could be found.
    */
   public void next()
   {
      next(currentArgs, lockType);
   }
   
   /**
    * Navigate to the next record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default lock type, but override the
    * default substitution parameters with those provided.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the next record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the first record
    * which meets the query criteria.
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    * @throws  QueryOffEndException
    *          if no record could be found.
    */
   public void next(Object[] values)
   {
      next(values, lockType);
   }
   
   /**
    * Navigate to the next record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default substitution parameters, but
    * override the default lock type with that provided.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the next record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the first record
    * which meets the query criteria.
    *
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    * @throws  QueryOffEndException
    *          if no record could be found.
    */
   public void next(LockType lockType)
   {
      next(currentArgs, lockType);
   }
   
   /**
    * Navigate to the next record which meets the query criteria and retrieve
    * it into its record buffer.  Override both the default substitution
    * parameters, and the default lock type with those provided.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the next record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the first record
    * which meets the query criteria.
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    * @throws  QueryOffEndException
    *          if no record could be found and <code>lenientOffEnd</code> is
    *          <code>false</code>.
    */
   public void next(Object[] values, LockType lockType)
   {
      if (!prepareFetch())
      {
         return;
      }
      
      if (getHelper().getHQLPreprocessor().isFindByRowid())
      {
         // direct record access using recid/rowid is behaving like unique for all search types 
         unique(values, lockType);
         return;
      }
      
      // If a lock type is not specified, assume the query's lock type.
      if (lockType == null)
      {
         lockType = this.lockType;
      }
      
      Persistable dmo = null;
      Persistence persistence = buffer.getPersistence();
      boolean pop = false;
      
      try
      {
         persistence.pushImplicitTransaction();
         pop = true;
         
         // If query is scrolling, check the cached result list first.
         boolean scrolling = (cursor != null);
         if (scrolling)
         {
            dmo = loadByValue(cursor.getNext(), lockType);
         }
         
         // If not scrolling or wasn't in the cached result list, perform the
         // query and cache the result (if scrolling).
         if (dmo == null)
         {
            boolean isFirst =
               (buffer.getSnapshot() == null &&
                referenceRecord == null      &&
                buffer.getOffEnd(getHelper().getSortIndex(), inverseSorting) == OffEnd.FRONT);
            dmo = findNext(values, lockType, true);
            if (scrolling)
            {
               if (isFirst)
               {
                  cursor.addResultFirst(getIDs(dmo));
               }
               else
               {
                  cursor.addResultNext(getIDs(dmo));
               }
            }
         }
         
         // Store result in the record buffer, even if null.
         OffEnd offEnd = (dmo == null ? OffEnd.BACK : OffEnd.NONE);
         updateBuffer(dmo, lockType, false, offEnd);
         referenceRecord = dmo;
         
         // Notify accumulators of an iteration.
         accumulate();
      }
      catch (QueryOffEndException exc)
      {
         if (!lenientOffEnd)
         {
            exc.fillInStackTrace();
            throw exc;
         }
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
      finally
      {
         // Reset dirty flag.
         dirtyCopy = false;
         
         if (pop)
         {
            persistence.popImplicitTransaction(null);
         }
      }
   }
   
   /**
    * Navigate to the previous record which meets the query criteria and
    * retrieve it into its record buffer.  Use the default substitution
    * parameters and lock type.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the previous record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the last record
    * which meets the query criteria.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    * @throws  QueryOffEndException
    *          if no record could be found.
    */
   public void previous()
   {
      previous(currentArgs, lockType);
   }
   
   /**
    * Navigate to the previous record which meets the query criteria and
    * retrieve it into its record buffer.  Use the default lock type, but
    * override the default substitution parameters with those provided.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the previous record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the last record
    * which meets the query criteria.
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    * @throws  QueryOffEndException
    *          if no record could be found.
    */
   public void previous(Object[] values)
   {
      previous(values, lockType);
   }
   
   /**
    * Navigate to the previous record which meets the query criteria and
    * retrieve it into its record buffer.  Use the default substitution
    * parameters, but override the default lock type with that provided.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the previous record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the last record
    * which meets the query criteria.
    *
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    * @throws  QueryOffEndException
    *          if no record could be found.
    */
   public void previous(LockType lockType)
   {
      previous(currentArgs, lockType);
   }
   
   /**
    * Navigate to the previous record which meets the query criteria and
    * retrieve it into its record buffer.  Override both the default
    * substitution parameters, and the default lock type with those provided.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the previous record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the last record
    * which meets the query criteria.
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    * @throws  QueryOffEndException
    *          if no record could be found and <code>lenientOffEnd</code> is
    *          <code>false</code>.
    */
   public void previous(Object[] values, LockType lockType)
   {
      if (!prepareFetch())
      {
         return;
      }
      
      if (getHelper().getHQLPreprocessor().isFindByRowid())
      {
         // direct record access using recid/rowid is behaving like unique for all search types 
         unique(values, lockType);
         return;
      }
      
      // If a lock type is not specified, assume the query's lock type.
      if (lockType == null)
      {
         lockType = this.lockType;
      }
      
      Persistable dmo = null;
      Persistence persistence = buffer.getPersistence();
      boolean pop = false;
      
      try
      {
         persistence.pushImplicitTransaction();
         pop = true;
         
         // If query is scrolling, check the cached result list first.
         boolean scrolling = (cursor != null);
         if (scrolling)
         {
            dmo = loadByValue(cursor.getPrevious(), lockType);
         }
         
         // If not scrolling or wasn't in the cached result list, perform the
         // query and cache the result (if scrolling).
         if (dmo == null)
         {
            OffEnd offEnd = buffer.getOffEnd(getHelper().getSortIndex(), inverseSorting);
            boolean isLast =
               offEnd == OffEnd.BACK ||
               (buffer.getSnapshot() == null &&
                referenceRecord == null      &&
                offEnd == OffEnd.FRONT);
            dmo = findPrevious(values, lockType, true);
            
            if (scrolling)
            {
               if (isLast)
               {
                  cursor.addResultLast(getIDs(dmo));
               }
               else
               {
                  cursor.addResultPrevious(getIDs(dmo));
               }
            }
         }
         
         // Store result in the record buffer, even if null.
         OffEnd offEnd = (dmo == null ? OffEnd.FRONT : OffEnd.NONE);
         updateBuffer(dmo, lockType, false, offEnd);
         referenceRecord = dmo;
         
         // Notify accumulators of an iteration.
         accumulate();
      }
      catch (QueryOffEndException exc)
      {
         if (!lenientOffEnd)
         {
            exc.fillInStackTrace();
            throw exc;
         }
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
      finally
      {
         // Reset dirty flag.
         dirtyCopy = false;
         
         if (pop)
         {
            persistence.popImplicitTransaction(null);
         }
      }
   }
   
   /**
    * Reload the record most recently loaded into the backing record buffer,
    * if any.  Use the default lock type.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record reload.
    */
   public void current()
   {
      current(lockType);
   }
   
   /**
    * Reload the record most recently loaded into the backing record buffer,
    * if any.  Override the default lock type.
    *
    * @param   lockType
    *          New lock type to apply to the record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record reload, or the lock could
    *          not be transitioned to the requested type.
    */
   public void current(LockType lockType)
   {
      if (!prepareFetch())
      {
         return;
      }
      
      buffer.armCurrentChanged();
      buffer.reload(lockType, isErrorIfNull());
      buffer.updateCurrentChanged();
   }
   
   /**
    * Navigate to a particular record which meets the query criteria and
    * retrieve it into its record buffer.  No more than one record may match
    * the criteria.  Use the default substitution parameters and lock type.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if more than
    *          one record is found which match the criteria.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void unique()
   {
      unique(currentArgs, lockType);
   }
   
   /**
    * Navigate to a particular record which meets the query criteria and
    * retrieve it into its record buffer.  No more than one record may match
    * the criteria.  Use the default lock type, but override the default
    * substitution parameters with those provided.
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if more than
    *          one record is found which match the criteria.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void unique(Object[] values)
   {
      unique(values, lockType);
   }
   
   /**
    * Navigate to a particular record which meets the query criteria and
    * retrieve it into its record buffer.  No more than one record may match
    * the criteria.  Use the default substitution parameters, but override
    * the default lock type with that provided.
    *
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if more than
    *          one record is found which match the criteria.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void unique(LockType lockType)
   {
      unique(currentArgs, lockType);
   }
   
   /**
    * Navigate to a particular record which meets the query criteria and
    * retrieve it into its record buffer.  No more than one record may match
    * the criteria.  Override both the default substitution parameters, and
    * the default lock type with those provided.
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if more than
    *          one record is found which match the criteria.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void unique(Object[] values, LockType lockType)
   {
      if (!prepareFetch())
      {
         return;
      }
      
      getHelper();
      if (!buffer.isTemporary() && 
          helper.isFindTemplate(values, buffer.getDMOImplementationClass()))
      {
         buffer.loadTemplateRecord(helper.getHQLPreprocessor().getFindByRowid(values));
         referenceRecord = null; // reset reference record
         dirtyCopy = false;      // reset dirty flag
         return;
      }
      else if (helper.getHQLPreprocessor().isFindByRowid() &&
               helper.getHQLPreprocessor().getFindByRowid(values) < 0)
      {
         // according to article P12127, looking for an illegal rowid is equivalent to releasing
         // the buffer. Backstage, error 18 is printed in database log but this is not a effect
         // we need to duplicate in P2J. 
         buffer.release();
         referenceRecord = null; // reset reference record
         dirtyCopy = false;      // reset dirty flag
         if (isErrorIfNull())
         {
            buffer.errorNotOnFile();
         }
         return;
      }
      
      if (lockType == null)
      {
         lockType = this.lockType;
      }
      
      Persistence persistence = buffer.getPersistence();
      boolean pop = false;
      
      try
      {
         persistence.pushImplicitTransaction();
         pop = true;
         
         activateUnique();
         Persistable dmo = execute(values, lockType, true, true);
         boolean errNull = isErrorIfNull();
         OffEnd offEnd = (dmo == null ? OffEnd.BACK : OffEnd.NONE);
         updateBuffer(dmo, lockType, errNull, offEnd);
         dirtyCopy = false;
         if (dmo == null && errNull)
         {
            buffer.errorNotOnFile();
         }
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
      finally
      {
         if (pop)
         {
            persistence.popImplicitTransaction(null);
         }
      }
   }
   
   /**
    * Set the buffer backing the query to unknown mode.
    *
    * @see     RecordBuffer#setUnknownMode
    */
   public void setUnknownRecord()
   {
      buffer.setUnknownMode();
   }
   
   /**
    * Reset the query by clearing its state to a known default.  This is
    * used to ensure that subsequent requests for FIRST and LAST record
    * retrievals operate on a known query/buffer state.
    *
    * @param   resolveArgs
    *          <code>true</code> if query substitution arguments should be
    *          resolved as part of the reset;  <code>false</code> if existing
    *          argument values should be used for the next query execution.
    */
   public void reset(boolean resolveArgs)
   {
      if (resolveArgs)
      {
         resolveArgs();
      }
      referenceRecord = null;
      buffer.reset();
      
      if (cursor != null)
      {
         cursor.reset();
      }
   }
   
   /**
    * Get the number of tables joined by this query.
    *
    * @return  Always 1.
    */
   public int getTableCount()
   {
      return 1;
   }
   
   /**
    * Get all the off-end listeners associated with this query.
    * 
    * @return  the off-end listeners associated with this query.
    */
   public List<QueryOffEndListener> getOffEndListeners()
   {
      List<QueryOffEndListener> listeners = 
         new ArrayList<QueryOffEndListener>(1);
      listeners.add(buffer.getQueryOffEndListener());
      
      return listeners;
   }

   /**
    * Load a record into its associated buffer, given an array of size one,
    * containing a primary key ID value of the record to be loaded, or the DMO
    * itself.
    * <p>
    * In the event an expected value cannot be loaded (e.g., the record has
    * been deleted or is otherwise no longer available), this method throws
    * {@link MissingRecordException}, after setting the associated buffer
    * into {@link RecordBuffer#setUnknownMode() unknown mode}. Note that a
    * record being locked by another session does not constitute a "missing"
    * record.
    * 
    * @param   data
    *          Array containing a single primary key ID or DMO.
    * @param   lockType
    *          Lock type which should by used.  Set to <code>null</code> to
    *          allow query to use its current lock type.
    * @param   silentIfNullId
    *          If <code>true</code>, do not raise {@link MissingRecordException}
    *          if some of the provided IDs are <code>null</code>. <code>null</code>
    *          IDs are valid for queries with OUTER join.
    * 
    * @throws  MissingRecordException
    *          if any record cannot be loaded, because it is no longer available
    *          (e.g., deleted, etc.).
    * @throws  PersistenceException
    *          if there is an error loading data.
    *
    * @see     #getRow()
    */
   public void load(Object[] data, LockType lockType, boolean silentIfNullId)
   throws PersistenceException
   {
      try
      {
         LockType lt = (lockType != null ? lockType : this.lockType);
         verifyScrolling();
         Serializable[] ids = new Serializable[1];
         Object datum = data[0];
         ids[0] = (datum instanceof Persistable
                   ? ((Persistable) datum).getId()
                   : (Serializable) datum);
         
         if (cursor.reposition(ids, false))
         {
            cursor.next();
         }
         
         Persistable dmo = loadByValue(data, lt, silentIfNullId);
         OffEnd offEnd = (dmo == null ? OffEnd.BACK : OffEnd.NONE);
         updateBuffer(dmo, lt, silentIfNullId, offEnd);
         
         if (dmo == null)
         {
            buffer.setUnknownMode();
            
            if (!(datum == null && silentIfNullId))
            {
               String table = buffer.getTable();
               RecordIdentifier ident = new RecordIdentifier(table, ids[0]);
               
               throw new MissingRecordException(ident);
            }
         }
      }
      finally
      {
         // Reset dirty flag.
         dirtyCopy = false;
      }
   }
   
   /**
    * Set the given results object into this query.  Only works if cursor is not null.
    *
    * @param   results
    *          New result set for this query.
    */
   public void setResults(Results results)
   {
      if (cursor != null)
      {
         cursor.reset();
         results.reset();
         while (results.next())
         {
            cursor.addResultNext(results.get());
         }
         results.reset();
      }
   }
   
   /**
    * Assemble an array of size one and store in it the primary key ID for
    * the current record in the buffer underlying this query.  This method
    * essentially takes a lightweight snapshot of the currently loaded
    * record, so that this may be restored later.
    *
    * @return  Array containing a single primary key ID.
    *
    * @see     #load
    */
   public Object[] getRow()
   {
      Persistable dmo = (referenceRecord != null ? referenceRecord : buffer.getSnapshot());
      if (dmo == null && !buffer.isUnknownMode())
      {
         throw new NullPointerException("No reference record available");
      }
      
      return (new Object[] { dmo });
   }
   
   /**
    * Fetch the array of primary key IDs associated with the first result in
    * this query's result list.
    * <p>
    * This implementation always returns an array with a single element,
    * (or <code>null</code>), since this query type uses a single record
    * buffer.
    *
    * @return  Array of record IDs reflecting the first query result, or
    *          <code>null</code> if there is no such result.
    */
   public Object[] peekFirst()
   {
      if (!prepareFetch())
      {
         return null;
      }
      
      buffer.release();
      activateFirst();
      
      referenceRecord = execute(currentArgs, LockType.NONE, false, false);
      if (referenceRecord == null)
      {
         return null;
      }
      
      return (new Object[] { referenceRecord });
   }
   
   /**
    * Fetch the array of primary key IDs associated with the last result in
    * this query's result list.
    * <p>
    * This implementation always returns an array with a single element,
    * (or <code>null</code>), since this query type uses a single record
    * buffer.
    *
    * @return  Array of record IDs reflecting the last query result, or
    *          <code>null</code> if there is no such result.
    */
   public Object[] peekLast()
   {
      if (!prepareFetch())
      {
         return null;
      }
      
      buffer.release();
      activateLast();
      
      referenceRecord = execute(currentArgs, LockType.NONE, false, false);
      if (referenceRecord == null)
      {
         return null;
      }
      
      return (new Object[] { referenceRecord });
   }
   
   /**
    * Fetch the array of primary key IDs associated with the next result in
    * this query's result list.
    * <p>
    * This implementation always returns an array with a single element,
    * (or <code>null</code>), since this query type uses a single record
    * buffer.
    *
    * @return  Array of record IDs reflecting the next query result, or
    *          <code>null</code> if there is no such result.
    */
   public Object[] peekNext()
   {
      if (!prepareFetch())
      {
         return null;
      }
      
      buffer.release();
      referenceRecord = findNext(currentArgs, LockType.NONE, false);
      if (referenceRecord == null)
      {
         return null;
      }
      
      return (new Object[] { referenceRecord });
   }
   
   /**
    * Fetch the array of primary key IDs associated with the previous result
    * in this query's result list.
    * <p>
    * This implementation always returns an array with a single element,
    * (or <code>null</code>), since this query type uses a single record
    * buffer.
    *
    * @return  Array of record IDs reflecting the previous query result, or
    *          <code>null</code> if there is no such result.
    */
   public Object[] peekPrevious()
   {
      if (!prepareFetch())
      {
         return null;
      }
      
      buffer.release();
      referenceRecord = findPrevious(currentArgs, LockType.NONE, false);
      if (referenceRecord == null)
      {
         return null;
      }
      
      return (new Object[] { referenceRecord });
   }
   
   /**
    * Get an array of all record buffers managed by this query.
    *
    * @return  All record buffers managed by this query.
    */
   @Override
   public RecordBuffer[] getRecordBuffers()
   {
      return (new RecordBuffer[] { buffer });
   }
   
   /**
    * Get the query substitution arguments associated with this query, if any.
    * 
    * @return  Substitution arguments, in order, or <code>null</code> if there are none.
    */
   @Override
   public Object[] getArgs()
   {
      return args;
   }
   
   /**
    * Get the legacy natural join, if any, associated with this query.
    * 
    * @return  Legacy join object, or <code>null</code> if none.
    */
   @Override
   public AbstractJoin getJoin()
   {
      return join;
   }
   
   /**
    * Get the original, HQL where clause associated with this query.
    * 
    * @return  Original where clause, or <code>null</code> if there is none.
    */
   @Override
   public String getOriginalWhere()
   {
      return where;
   }
   
   /**
    * Force this query to operate in dynamic retrieval mode, if it supports this mode.
    * <p>
    * This implementation does nothing, since this query type always is in dynamic retrieval
    * mode.
    */
   @Override
   public void forceDynamicOperation()
   {
      // no-op
   }
   
   /**
    * Get the sort phrase associated with this query.
    * 
    * @return  Sort phrase.
    */
   public String getSortPhrase()
   {
      return sort;
   }
   
   /**
    * Create an adaptive query component based on the information in this query (which is
    * presumed to be a single-table query), and given a list of adaptive query components which
    * represent outer, nested query loops, which will perform a server-side join to the
    * innermost, nested query component.
    * 
    * @param   joinList
    *          List of adaptive components representing nested query loops which will contain
    *          the adaptive query component returned by this method. The last component in the
    *          list represents the nested query loop immediately containing the returned query
    *          component, with which the join will be made.
    * @param   query
    *          Adaptive query which will manage the server-side join of <code>joinList</code> and
    *          the returned query component.
    * @param   iteration
    *          Iteration type: FIRST, LAST, or NEXT.
    * @param   outer
    *          <code>true</code> if the join should be a left outer join. This type of join is
    *          not supported at the time of this writing, so this parameter should always be
    *          <code>false</code>. It is here to support a planned, future enhancement.
    * @param   fallback
    *          A compound query component to use in the event the multi-table adaptive query
    *          switches from preselect to dynamic retrieval mode.
    * 
    * @return  An adaptive query component suitable for a server-side join, based on the
    *          information in this query.
    */
   @Override
   public AdaptiveComponent makeAdaptiveServerJoinComponent(List<AdaptiveComponent> joinList,
                                                            AdaptiveQuery query,
                                                            int iteration,
                                                            boolean outer,
                                                            CompoundComponent fallback)
   {
      QueryComponent.ServerJoinData sjd =
         QueryComponent.prepareServerJoinData(joinList, buffer, join, args);
      BufferReference inverse = (join != null ? join.getInverse() : null);
      AdaptiveComponent comp = new AdaptiveComponent(query,
                                                     buffer,
                                                     sjd.join,
                                                     where,
                                                     sort,
                                                     indexInfo,
                                                     lockType,
                                                     sjd.args,
                                                     iteration,
                                                     outer,
                                                     inverse);
      comp.setReferenceSubs(sjd.referenceSubs);
      comp.setRelatedBuffers(sjd.relatedBuffers);
      comp.setFallback(fallback);
      
      if (iteration == FIRST || iteration == LAST)
      {
         try
         {
            comp.setIteration(iteration, sort);
         }
         catch (PersistenceException exc)
         {
            // should only happen if we have a runtime programming error or conversion error
            // (or with bad, hand-written business logic)
            throw new IllegalArgumentException(exc);
         }
      }
      
      return comp;
   }
   
   /**
    * Create a preselect query component based on the information in this query (which is
    * presumed to be a single-table query), and given a list of preselect query components which
    * represent outer, nested query loops, which will perform a server-side join to the
    * innermost, nested query component.
    * 
    * @param   joinList
    *          List of preselect components representing nested query loops which will contain
    *          the preselect query component returned by this method. The last component in the
    *          list represents the nested query loop immediately containing the returned query
    *          component, with which the join will be made.
    * @param   iteration
    *          Iteration type: FIRST, LAST, or NEXT.
    * @param   outer
    *          <code>true</code> if the join should be a left outer join. This type of join is
    *          not supported at the time of this writing, so this parameter should always be
    *          <code>false</code>. It is here to support a planned, future enhancement.
    * 
    * @return  A preselect query component suitable for a server-side join, based on the
    *          information in this query.
    */
   @Override
   public QueryComponent makePreselectServerJoinComponent(List<QueryComponent> joinList,
                                                          int iteration,
                                                          boolean outer)
   {
      QueryComponent.ServerJoinData sjd =
         QueryComponent.prepareServerJoinData(joinList, buffer, join, args);
      QueryComponent comp = new QueryComponent(this,
                                               buffer,
                                               sjd.join,
                                               where,
                                               indexInfo,
                                               lockType,
                                               sjd.args,
                                               iteration,
                                               outer);
      comp.setReferenceSubs(sjd.referenceSubs);
      comp.setRelatedBuffers(sjd.relatedBuffers);
      
      if (iteration == FIRST || iteration == LAST)
      {
         try
         {
            comp.setIteration(iteration, sort);
         }
         catch (PersistenceException exc)
         {
            // should only happen if we have a runtime programming error or conversion error
            // (or with bad, hand-written business logic)
            throw new IllegalArgumentException(exc);
         }
      }
      
      return comp;
   }
   
   /**
    * Indicate whether this query has a client-side where clause expression associated with it.
    * 
    * @return  <code>true</code> if there is a where expression; else <code>false</code>.
    */
   @Override
   public boolean hasWhereExpression()
   {
      return (whereExpr != null);
   }
   
   /**
    * Report the record buffer for whose changes this object is interested
    * in listening.  This listener will be registered with the {@link
    * ChangeBroker} to receive notifications of any change to DMOs whose
    * interface types match that returned by the {@link
    * RecordBuffer#getDMOInterface} method of the returned buffer.
    *
    * @return  An iterator on a single object:  the record buffer backing this query.
    */
   public Iterator<RecordBuffer> recordBuffers()
   {
      Iterator<RecordBuffer> iter = new Iterator<RecordBuffer>()
      {
         private boolean avail = true;
         public boolean hasNext() { return avail; }
         public RecordBuffer next() { avail = false;  return buffer; }
         public void remove() { throw new UnsupportedOperationException(); }
      };
      
      return iter;
   }
   
   /**
    * Register all <code>RecordChangeListener</code>s associated with this
    * query (if any), with the context-local <code>ChangeBroker</code> at the
    * indicated scope.
    * 
    * @param   scope
    *          Scope at which listeners should be registered with the change broker.
    */
   public void registerRecordChangeListeners(int scope)
   {
      ChangeBroker broker = ChangeBroker.get();
      broker.addListener(this, scope);
      if (placeholderCleaner != null)
      {
         broker.addListener(placeholderCleaner, scope);
      }
      
      // if these were added as global listeners in ChangeBroker remember to remove them
      // when the query is destroyed. Otherwise, the ChangeBroker will drop them automatically
      // when the respective scope is finished.
      if (scope == 0)
      {
         unregisterOnCleanup = true;
      }
   }
   
   /**
    * Explicitly close the prepared query.
    */
   @Override
   public void close()
   {
      close(false);
   }
   
   /**
    * Explicitly close the prepared query.
    * 
    * @param    inResource
    *           Flag indicating this query is used in a QUERY legacy resource.
    */
   @Override
   public void close(boolean inResource)
   {
      if (inResource && !unregisterOnCleanup)
      {
         // remove it explicitly
         ChangeBroker broker = ChangeBroker.get();
         broker.removeListener(this);
      }
      
      super.close();
   }
   
   /**
    * The cleanup method is overridden to allow the listener(s) to be removed from the
    * {@code ChangeBroker} if they were added in the global scope, otherwise, the 
    * {@code ChangeBroker} will drop them automatically when the respective scope is finished.
    */
   @Override
   public void cleanup()
   {
      if (unregisterOnCleanup)
      {
         ChangeBroker broker = ChangeBroker.get();
         broker.removeFromGlobal(this);
         if (placeholderCleaner != null)
         {
            broker.removeFromGlobal(placeholderCleaner);
         }
         
         unregisterOnCleanup = false;
      }
      
      // let the super class do the rest of the cleanup
      super.cleanup();
   }
   
   /**
    * Respond to a record change event.
    *
    * @param   event
    *          Event which describes the DMO state change.
    *
    * @throws  PersistenceException
    *          if any error occurs while processing the state change event.
    */
   public void stateChanged(RecordChangeEvent event)
   throws PersistenceException
   {
      if (cursor == null)
      {
         return;
      }
      
      if (event.isInsert())
      {
         cursor.reset();
         
         return;
      }
      
      if (event.isDelete())
      {
         int idx = -1;
         RecordBuffer source = event.getSource();
         Iterator<RecordBuffer> iter = recordBuffers();
         for (int i = 0; iter.hasNext() && idx < 0; i++)
         {
            if (source == iter.next())
            {
               idx = i;
            }
         }
         
         // Reset the cache if it contains a reference to the deleted DMO.
         if (idx >= 0 && cursor.contains(event.getDMO().getId(), idx))
         {
            cursor.reset();
         }
      }
   }
   
   /**
    * Get the off-end status of this query, which indicates whether the query
    * has run off either end of its results.
    * 
    * @return  One of the {@link OffEnd} constants <code>NONE</code>,
    *          <code>FRONT</code>, or <code>BACK</code>.
    */
   public OffEnd getOffEnd()
   {
      if (helper == null)
      {
         return OffEnd.NONE;
      }
      
      SortIndex si = helper.getSortIndex();
      
      return buffer.getOffEnd(si, inverseSorting);
   }
   
   /**
    * Conversion of INDEX-INFORMATION attribute (KW_IDX_INFO).
    *
    * Getter of INDEX-INFORMATION attribute
    *
    * @return A character string consisting of a comma-separated list of the index or
    *         indexes the query uses at the level of join specified.
    */
   @LegacyAttribute(name = "INDEX-INFORMATION")
   public character indexInformation(NumberType n)
   {
      return new character(indexInfo);
   }
   
   /**
    * Set the record whose data defines current "place" in query results.
    *
    * @param referenceRecord
    *        The record whose data defines current "place" in query results.
    */
   void setReferenceRecord(Persistable referenceRecord)
   {
      this.referenceRecord = referenceRecord;
   }
   
   /**
    * Register the RecordChangeListener instance with the ChangeBroker.
    */
   protected void registerChangeListener()
   {
      // Listen for changes to our reference record.  If it is modified, we
      // can no longer rely upon it as an accurate placeholder.
      placeholderCleaner = new RecordChangeListener()
      {
         public Iterator<RecordBuffer> recordBuffers()
         {
            return Collections.singletonList(buffer).iterator();
         }
         public void stateChanged(RecordChangeEvent event)
         throws PersistenceException
         {
            if (event.getDMO() == referenceRecord &&
                event.isUpdate()                  &&
                !event.isIndexedPropertyChangeOnly())
            {
               // Rely on buffer snapshot as placeholder instead.
               referenceRecord = null;
            }
         }
      };
      ChangeBroker.get().addListener(placeholderCleaner);
   }
   
   /**
    * Indicate whether this query should be a projection query, retrieving only the primary key
    * of the target record, or whether it should retrieve the entire record as early as possible.
    * 
    * @return  <code>true</code> for a projection query, else <code>false</code>.
    */
   protected boolean isIdOnly()
   {
      // Generally, we want to execute as few queries as possible to retrieve a record. The
      // minimum is one, whereby we get all the columns in one pass. However, we have to
      // consider that for NEXT and PREVIOUS requests, we may execute a query per sort component
      // (plus 1 for the primary key if the sort clause does not represent a unique index).
      // In that case, we want to execute a projection query (i.e., get only the primary key).
      // Also we can only retrieve all the columns in the first pass for a no-lock query,
      // because in order to ensure integrity of a locked record, we have to apply the lock
      // before retrieving the full data. Finally, honor the fullRecords request if possible.
      switch (activeBundleKey)
      {
         // TODO: if a next/previous sort clause contains only 1 component, we could be less
         // strict about forcing a projection query here; however, that would require making
         // a separate call to SortCriterion.parse(String, RecordBuffer), which is somewhat
         // expensive, and we are about to do it momentarily in HQLHelper
         case NEXT:
         case PREVIOUS:
            return true;
         default:
            return !(fullRecords || LockType.NONE.equals(lockType) || buffer.isTemporary());
//            return !(fullRecords || buffer.isTemporary());
//            return !buffer.isTemporary();
//            return true;
//            return !fullRecords;
      }
   }
   
   /**
    * Initialize the associated buffer by resetting its state.  This releases
    * any record currently stored and resets the placeholder record stored in
    * the buffer.
    */
   protected void initializeBuffer()
   {
      buffer.reset();
   }
   
   /**
    * Release the current record, if any, from the buffer which is used by
    * this query.
    *
    * @throws  QueryOffEndException
    *          if the buffer contained a record and it was released.
    */
   protected void releaseBuffers()
   {
      buffer.release();
   }
   
   /**
    * Prepare buffer associated with this query for the query's execution.
    * This resets certain per-query state tracked by the record buffer.
    * 
    * @throws  ValidationException
    *          if the parent's implementation is invoked and a buffer flush
    *          triggers a validation error for the record currently stored in
    *          the buffer.
    */
   protected void prepareBuffer()
   throws ValidationException
   {
      buffer.flush();
      buffer.setLocked(false);
   }
   
   /**
    * Update the record buffer with the given DMO, applying the specified
    * lock type and error-if-null setting.  This implementation causes the
    * record buffer's internal placeholder snapshot to be reset if
    * <code>dmo</code> is null.
    *
    * @param   dmo
    *          Record to be stored in buffer.
    * @param   lockType
    *          Lock type associated with the record, which will be remembered
    *          by the buffer.
    * @param   errorIfNull
    *          If <code>true</code>, setting a <code>null</code> value as the
    *          current record in an underlying buffer will raise an error
    *          condition (if not in silent error mode); if <code>false</code>,
    *          this action will raise an end condition instead.
    * @param   offEnd
    *          Enum indicating whether query is off-end, and if so, in which
    *          direction.
    */
   protected void updateBuffer(Persistable dmo,
                               LockType lockType,
                               boolean errorIfNull,
                               OffEnd offEnd)
   {
      SortIndex sortIndex = getHelper().getSortIndex();
      buffer.setRecord(dmo,
                       lockType,
                       errorIfNull,
                       sortIndex,
                       offEnd,
                       inverseSorting,
                       dirtyCopy);
   }
   
   /**
    * Return an object which represents a property value which defines the
    * end of a <i>sort band</i>.  A <i>sort band</i> is a series of one or
    * more records which share the same value for a particular property,
    * where that property is used as the coarsest sorting criterion defined
    * by this query's <code>order by</code> clause.
    * <p>
    * If the break value for this query currently is non-<code>null</code>,
    * this indicates that the most recently retrieved record has crossed a
    * <i>sort band</i> boundary, and that a new band has begun.  This value
    * will only be non-<code>null</code> at the point at which such a record
    * is found, and will be <code>null</code> for all other records.
    * <p>
    * Note:  break values are only tracked for invocations of
    * <code>next</code> and <code>previous</code>.  Other retrievals will
    * results in break value being <code>null</code>.
    *
    * @return  Break value if dynamic query crossed a sort band boundary,
    *          else <code>null</code>.
    */
   protected Object getBreakValue()
   {
      return breakValue;
   }
   
   /**
    * Get the record buffer associated with this query.
    * 
    * @return  Record buffer.
    */
   protected RecordBuffer getBuffer()
   {
      return buffer;
   }
   
   /**
    * Retrieve this query's arguments, as currently resolved.
    * 
    * @return  Current arguments.
    */
   protected Object[] getCurrentArgs()
   {
      return currentArgs;
   }
   
   /**
    * Retrieve this query's lock type.
    * 
    * @return  Lock type
    */
   protected LockType getLockType()
   {
      return lockType;
   }
   
   /**
    * Execute each query stored in the active HQL bundle in turn until we
    * find a result or have no more statements to execute.  Each statement is
    * decreasingly specific.
    * <p>
    * The latest record to be stored in the record buffer, if any, marks our
    * current position in the query's results.  Its data is used to substitute
    * into the placeholder parameters in the query statements' augmented where
    * clauses.  These are added to the <code>values</code> provided as part of
    * the base query.
    * <p>
    * <b>Client-Side Where Clause Handling</b><br>
    * If this query involves a client-side where clause expression, an
    * additional layer of criteria checking takes place.  In addition to the
    * search described above, each returned result is temporarily stored in
    * the associated record buffer, and the where expression is executed.
    * If the expression indicates a match, the record is returned.  Otherwise,
    * additional records are retrieved and tested until either a match is
    * found or no more records are available in the requested navigation
    * direction.
    * <p>
    * Unique queries combined with client-side where clauses require special
    * handling and may perform particularly poorly as a result.  In this
    * case, every record returned by the first-pass query must be tested
    * against the client-side where expression, even once a result has been
    * found.  This must be done to ensure <i>only one</i> record matches the
    * specified criteria.  The only exception is the case where a second
    * record is actually found, in which case the uniqueness requirement has
    * been violated, and the remainder of the scan is aborted due to the
    * error.
    *
    * @param   values
    *          Array of substitution parameter values for the base where
    *          clause (i.e., the un-augmented portion of the clause).
    * @param   lockType
    *          Type of lock to acquire for the found record.
    * @param   unique
    *          <code>true</code> if there should be no more than one match
    *          for the conditions specified;  <code>false</code> if multiple
    *          matches are possible.
    * @param   updateLock
    *          <code>true</code> if the status of the lock on the retrieved
    *          record should be modified;  else <code>false</code>.  This is
    *          set to <code>true</code> for actual retrievals, and to
    *          <code>false</code> when only detecting whether a record would
    *          be found, or when peeking a record.
    *
    * @return  The record resulting from the query, or <code>null</code> if
    *          no record was found.
    *
    * @throws  ErrorConditionException
    *          if a recoverable error occurred while
    */
   protected Persistable execute(Object[] values,
                                 LockType lockType,
                                 boolean unique,
                                 boolean updateLock)
   {
      // if the JOINed buffer has no record, then we must return NULL
      if (join != null)
      {
         BufferReference buffer = join.getInverse();
         Persistable record = buffer.buffer().getCurrentRecord();
         
         if (record == null)
         {
            return null;
         }
      }
      
      if (whereExpr == null)
      {
         return executeImpl(values, lockType, unique, updateLock);
      }
      
      Persistable dmo = null;
      Persistable oldRefRec = referenceRecord;
      HQLBundle bundle0 = activeBundle;
      HQLBundle bundle1;
      HQLBundle bundle2;
      
      // Prepare the HQL bundles which will be used in the iterative,
      // first pass scan.
      getHelper();
      switch (activeBundleKey)
      {
         case NEXT:
            bundle1 = bundle2 = helper.next();
            break;
         case PREVIOUS:
            bundle1 = bundle2 = helper.previous();
            break;
         case FIRST:
         case UNIQUE:
            bundle1 = helper.first();
            bundle2 = helper.next();
            break;
         case LAST:
            bundle1 = helper.last();
            bundle2 = helper.previous();
            break;
         default:
            throw new IllegalStateException("No active HQL bundle");
      }
      
      Persistence persistence = buffer.getPersistence();
      try
      {
         Persistable uniqueDMO = null;
         buffer.pushTempContext();
         
         for (int i = 0; ; i++)
         {
            activeBundle = (i == 0 ? bundle1 : bundle2);
            
            // Find a candidate record, but do not lock it yet.
            dmo = executeImpl(values, LockType.NONE, false, false);
            if (dmo == null)
            {
               break;
            }
            referenceRecord = dmo;
            
            // Execute where expression.
            buffer.setTempRecord(dmo);
            if (whereExpr.get().booleanValue())
            {
               if (unique)
               {
                  if (uniqueDMO == null)
                  {
                     uniqueDMO = dmo;
                  }
                  else
                  {
                     persistence.uniqueResultViolation(buffer, null);
                  }
               }
               else
               {
                  break;
               }
            }
         }
         
         if (unique)
         {
            dmo = uniqueDMO;
         }
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
         
         return null;
      }
      finally
      {
         try
         {
            buffer.popTempContext();
         }
         catch (PersistenceException exc)
         {
            ErrorManager.recordOrThrowError(exc);
         }
         referenceRecord = oldRefRec;
         activeBundle = bundle0;
      }
      
      // If a record was found and successfully passed the client-side where
      // expression filter, lock it now, if necessary.
      if (dmo != null && !lockType.equals(LockType.NONE) && updateLock)
      {
         try
         {
            String table = buffer.getTable();
            RecordIdentifier ident = new RecordIdentifier(table, dmo.getId());
            RecordLockContext lockCtx = buffer.getPersistenceContext().getRecordLockContext();
            lockCtx.lock(ident, lockType);
         }
         catch (LockUnavailableException exc)
         {
            ErrorManager.recordOrThrowError(
               445,
               buffer.getLegacyName() + " record is locked",
               exc);
            
            // Cannot return record if it could not be locked.
            dmo = null;
         }
      }
      
      return dmo;
   }
   
   /**
    * This method is invoked when a locking exception is caught during query
    * execution.  This implementation delegates handling of the error to the
    * <code>ErrorManager</code>.
    * <p>
    * Subclasses which require different handling must override this method.
    * 
    * @param   exc
    *          Cause exception.
    */
   protected void handleExecuteException(LockUnavailableException exc)
   {
      buffer.setLocked(true);
      ErrorManager.recordOrThrowError(445, buffer.getLegacyName() + " record is locked", exc);
   }
   
   /**
    * This method is invoked when a validation exception is caught during
    * query execution.  This implementation delegates handling of the error
    * to the <code>ErrorManager</code>.
    * <p>
    * Subclasses which require different handling must override this method.
    * 
    * @param   exc
    *          Cause exception.
    */
   protected void handleExecuteException(ValidationException exc)
   {
      ErrorManager.recordOrThrowError(exc);
   }
   
   /**
    * This method is invoked when a persistence exception is caught during
    * query execution.  This implementation delegates handling of the error
    * to the <code>ErrorManager</code>.
    * <p>
    * Subclasses which require different handling must override this method.
    * 
    * @param   exc
    *          Cause exception.
    */
   protected void handleExecuteException(PersistenceException exc)
   {
      ErrorManager.recordOrThrowError(exc);
   }
   
   /**
    * Obtain the {@link HQLHelper} object which supports this query, creating it first if
    * necessary.
    *
    * @return  HQL helper.
    */
   protected HQLHelper getHelper()
   {
      if (helper == null)
      {
         helper = HQLHelper.obtain(buffer,
                                   getReferencedBuffers(),
                                   args,
                                   isIdOnly(),
                                   where,
                                   sort,
                                   join);
         
         if (helper.getSortIndex().isInverseSortPhrase(sort))
         {
            inverseSorting = true;
         }
         
         // don't need original where clause or sort clause anymore, so reclaim a little memory
         where = null;
         sort = null;
      }
      
      return helper;
   }
   
   /**
    * Set as active the HQL bundle corresponding with a search for the single
    * record, if any, which matches this query's criteria.
    */
   protected void activateUnique()
   {
      activeBundleKey = UNIQUE;
      activeBundle = getHelper().unique();
   }
   
   /**
    * Set as active the HQL bundle corresponding with a search for the first
    * record matching this query's criteria.
    */
   protected void activateFirst()
   {
      activeBundleKey = FIRST;
      activeBundle = getHelper().first();
   }
   
   /**
    * Set as active the HQL bundle corresponding with a search for the last
    * record matching this query's criteria.
    */
   private void activateLast()
   {
      activeBundleKey = LAST;
      activeBundle = getHelper().last();
   }
   
   /**
    * Set as active the HQL bundle corresponding with a search for the next
    * record matching this query's criteria, relative to the most recently
    * fetched record.
    */
   private void activateNext()
   {
      activeBundleKey = NEXT;
      activeBundle = getHelper().next();
   }
   
   /**
    * Set as active the HQL bundle corresponding with a search for the
    * previous record matching this query's criteria, relative to the most
    * recently fetched record.
    */
   private void activatePrevious()
   {
      activeBundleKey = PREVIOUS;
      activeBundle = getHelper().previous();
   }
   
   /**
    * Resolve query substitution arguments in preparation for query execution.
    * This method is invoked only once, at construction time, for simple
    * queries (i.e., those corresponding with simple FIND statements in the
    * original Progress code).  For iterating queries (i.e., those which
    * correspond with FOR loops in the original Progress code), it is
    * invoked once per each entrance into the loop.  Thus, for an outer loop,
    * it is only invoked once, for an inner loop it is invoked once per each
    * iteration of the <i>outer</i> loop, and so on.
    * <p>
    * This causes all variable (i.e. <code>BaseDataType</code>) arguments to
    * be duplicated, and all <code>Resolvable</code>s, <i>except</i>
    * <code>FieldReference</code>s, to be resolved and the result stored.
    * <code>FieldReference</code>s are simply stored, and are resolved
    * separately, on demand, each time the query is executed.
    */
   private void resolveArgs()
   {
      // Resolve regular query substitution arguments.
      if (args != null)
      {
         int len = args.length;
         currentArgs = new Object[len];
         for (int i = 0; i < len; i++)
         {
            Object next = args[i];
            if (next == null && ErrorManager.isPendingError())
            {
               currentArgs = null;
               
               return;
            }
            if (next instanceof FieldReference)
            {
               currentArgs[i] = next;
            }
            else if (next instanceof Resolvable)
            {
               currentArgs[i] = ((Resolvable) next).resolve();
            }
            else
            {
               currentArgs[i] = ((BaseDataType) next).duplicate();
            }
         }
      }
   }
   
   /**
    * Find the next record matching the search criteria, or the first such
    * record if no previous search has been executed against this query.
    * 
    * @param   values
    *          Array of substitution parameter values for the base where
    *          clause (i.e., the un-augmented portion of the clause).
    * @param   lockType
    *          Type of lock to acquire for the found record.
    * @param   updateLock
    *          <code>true</code> if the status of the lock on the retrieved
    *          record should be modified;  else <code>false</code>.  This is
    *          set to <code>true</code> for actual retrievals, and to
    *          <code>false</code> when only detecting whether a record would
    *          be found, or when peeking a record.
    *
    * @return  Next (or first) record to match the query criteria.
    */
   private Persistable findNext(Object[] values, LockType lockType, boolean updateLock)
   {
      SortIndex sortIndex = getHelper().getSortIndex();
      OffEnd offEnd = buffer.getOffEnd(sortIndex, inverseSorting);
      if (offEnd == OffEnd.FRONT || (buffer.getSnapshot() == null && referenceRecord == null))
      {
         activateFirst();
      }
      else
      {
         activateNext();
      }
      
      return execute(values, lockType, false, updateLock);
   }
   
   /**
    * Find the previous record matching the search criteria, or the last such
    * record if no previous search has been executed against this query.
    *
    * @param   values
    *          Array of substitution parameter values for the base where
    *          clause (i.e., the un-augmented portion of the clause).
    * @param   lockType
    *          Type of lock to acquire for the found record.
    * @param   updateLock
    *          <code>true</code> if the status of the lock on the retrieved
    *          record should be modified;  else <code>false</code>.  This is
    *          set to <code>true</code> for actual retrievals, and to
    *          <code>false</code> when only detecting whether a record would
    *          be found, or when peeking a record.
    *
    * @return  Previous (or last) record to match the query criteria.
    */
   private Persistable findPrevious(Object[] values, LockType lockType, boolean updateLock)
   {
      SortIndex sortIndex = getHelper().getSortIndex();
      OffEnd offEnd = buffer.getOffEnd(sortIndex, inverseSorting);
      if (offEnd == OffEnd.BACK  || (buffer.getSnapshot() == null && referenceRecord == null))
      {
         activateLast();
      }
      else
      {
         activatePrevious();
      }
      
      return execute(values, lockType, false, updateLock);
   }
   
   /**
    * Retrieve a record, given an array of size one, containing a primary key
    * ID value of the record to be retrieved, or the record itself.  Acquire
    * the specified lock type.
    *
    * @param   data
    *          Array containing a single primary key ID or DMO.
    * @param   lockType
    *          Type of lock to be applied to the record loaded.
    *
    * @return  The record associated with the given ID, or <code>null</code>
    *          if there is no record matching the ID.
    */
   private Persistable loadByValue(Object[] data, LockType lockType)
   {
      return loadByValue(data, lockType, false);
   }
   
   /**
    * Retrieve a record, given an array of size one, containing a primary key
    * ID value of the record to be retrieved, or the record itself.  Acquire
    * the specified lock type.
    *
    * @param   data
    *          Array containing a single primary key ID or DMO.
    * @param   lockType
    *          Type of lock to be applied to the record loaded.
    * @param   silentIfNullId
    *          If <code>true</code>, do not raise error if some of the provided
    *          IDs are <code>null</code>. <code>null</code> IDs are valid for
    *          queries with OUTER join.
    * 
    * @return  The record associated with the given ID, or <code>null</code>
    *          if there is no record matching the ID.
    */
   private Persistable loadByValue(Object[] data, LockType lockType, boolean silentIfNullId)
   {
      Persistable dmo = null;
      
      try
      {
         if (data != null)
         {
            Object datum = data[0];
            boolean isDMO = (datum instanceof Persistable);
            
            if (isDMO && lockType.equals(LockType.NONE))
            {
               dmo = (Persistable) datum;
            }
            else
            {
               Serializable id;
               if (isDMO)
               {
                  dmo = (Persistable) datum;
                  id = dmo.getId();
               }
               else
               {
                  id = (Serializable) datum;
               }
               
               if (id == null && silentIfNullId)
               {
                  releaseBuffers();
                  return null;
               }
               else
               {
                  Persistence persistence = buffer.getPersistence();
                  dmo = persistence.load(buffer, id, lockType, 0L, true, false);
               }
            }
         }
      }
      catch (MissingRecordException exc)
      {
         // record must have been deleted in another context
         dmo = null;
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
      
      return dmo;
   }
   
   /**
    * Create an array of size one, containing the primary key ID of the given
    * DMO.
    *
    * @param   dmo
    *          Data record whose ID is to be stored in the array.
    *
    * @return  Single element array or <code>null</code> if <code>dmo</code>
    *          is <code>null</code>.
    */
   private Object[] getIDs(Persistable dmo)
   {
      if (dmo == null)
      {
         return null;
      }
      
      return (new Object[] { dmo.getId() });
   }
   
   /**
    * Record or throw a FIND FIRST/LAST failed error condition for the
    * current DMO type.
    *
    * @throws  ErrorConditionException
    *          if silent error mode is suppressed.
    */
   private void errorFindFirstLast()
   {
      String msg = "FIND FIRST/LAST failed for table " + buffer.getLegacyName();
      ErrorManager.recordOrThrowError(565, msg);
   }
   
   /**
    * Execute each query stored in the active HQL bundle in turn until we
    * find a result or have no more statements to execute.  Each statement is
    * decreasingly specific.
    * <p>
    * The latest record to be stored in the record buffer, if any, marks our
    * current position in the query's results.  Its data is used to substitute
    * into the placeholder parameters in the query statements' augmented where
    * clauses.  These are added to the <code>values</code> provided as part of
    * the base query.
    *
    * @param   values
    *          Array of substitution parameter values for the base where
    *          clause (i.e., the un-augmented portion of the clause).
    * @param   lockType
    *          Type of lock to apply to the found record.
    * @param   unique
    *          <code>true</code> if there should be no more than one match
    *          for the conditions specified;  <code>false</code> if multiple
    *          matches are possible.
    * @param   updateLock
    *          <code>true</code> if the status of the lock on the retrieved
    *          record should be modified;  else <code>false</code>.  This is
    *          set to <code>true</code> for actual retrievals, and to
    *          <code>false</code> when only detecting whether a record would
    *          be found (i.e., the <code>hasXXXX()</code> methods).
    *
    * @return  The record resulting from the query, or <code>null</code> if
    *          no record was found.
    *
    * @throws  ErrorConditionException
    *          if a recoverable error occurred while
    */
   private Persistable executeImpl(Object[] values,
                                   LockType lockType,
                                   boolean unique,
                                   boolean updateLock)
   {
      boolean debug = LOG.isLoggable(Level.FINE);
      
      DirtyShareContext dirtyContext = buffer.getDirtyContext();
      ParameterIndices pi = activeBundle.getParameterIndices();
      int origArgCount = (values == null ? 0 : values.length);
      int baseArgCount = (pi == null ? origArgCount : pi.getCount());
      
      Persistable placeholder = (referenceRecord != null
                                 ? referenceRecord
                                 : buffer.getSnapshot());
      Persistable dmo = null;
      Set<Persistable> primaryDMOs = null;
      
      try
      {
         List<Serializable> parameters = null;
         List<Type> parameterTypes = null;
         
         if (join != null)
         {
            parameters = join.getParameters();
            parameterTypes = join.getParameterTypes();
            
            if (parameterTypes!=null)
            {
               int s = parameterTypes.size();
               baseArgCount += s;
               origArgCount += s;
            }
         }
         
         prepareBuffer();
         Persistence persistence = buffer.getPersistence();
         boolean multiplex = buffer.isMultiplexed();
         
         if (multiplex)
         {
            baseArgCount++;
         }
         
         Object[] baseArgs = new Object[baseArgCount];
         Type[] baseTypes = new Type[baseArgCount];
         int startIndex = 0;
         
         if (multiplex)
         {
            startIndex++;
            baseArgs[0] = buffer.getMultiplexID();
            baseTypes[0] = null;
         }
         
         // Add the join parameters (if available) as the next arguments.
         if (join != null && parameterTypes != null)
         {
            int size = parameterTypes.size();
            for (int i = 0; i < size; i++)
            {
               baseArgs[startIndex] = parameters.get(i);
               baseTypes[startIndex] = parameterTypes.get(i);
               startIndex++;
            }
         }
         
         HQLPreprocessor hqlPreproc = getHelper().getHQLPreprocessor();
         List<String> earlyPublish = hqlPreproc.getEarlyPublishEntities();
         if (!earlyPublish.isEmpty())
         {
            List<Persistable> dmos = dirtyContext.updateSnapshots(earlyPublish, persistence);
            
            // Make sure DMOs do not bloat the Hibernate session.
            for (Persistable next : dmos)
            {
               buffer.evictDMOIfUnused(next);
            }
         }
         
         boolean needsLock = updateLock && lockType != LockType.NONE;
         boolean findByRowid = hqlPreproc.isFindByRowid();
         
         while (true)
         {
            // Iterate all arguments provided for the base portion of the
            // where clause only.  If any of these are FieldReferences,
            // resolve these parameters.
            for (int j = 0, i = startIndex; i < baseArgCount; j++)
            {
               int k = -1;
               if (pi == null)
               {
                  k = j;
               }
               else
               {
                  Integer idx = pi.getIndex(j);
                  if (idx == null)
                  {
                     // Original substitution parameter was inlined into HQL
                     // by preprocessor;  skip it.
                     continue;
                  }
                  k = idx;
               }
               
               // Substitution placeholders may have been reordered if this
               // statement was rewritten by the HQL preprocessor. Make sure
               // we access the correct substitution parameter.
               Object next = values[k];
               
               // Store parameters in the base args array, resolving them
               // first if necessary.
               if (next instanceof Resolvable)
               {
                  baseArgs[i] = ((Resolvable) next).resolve();
               }
               else
               {
                  baseArgs[i] = next;
               }
               
               i++;
            }
            
            // determine call count
            int callCount = (placeholder != null ? activeBundle.gettersSize() : 0);
            
            // Collect all arguments required by the augmented portion of the
            // where clause only, by invoking getters on the latest DMO stored
            // in the buffer.
            Object[] augmentArgs = new Object[callCount];
            Iterator<Method> getters = activeBundle.getters();
            for (int i = 0; getters.hasNext(); i++)
            {
               Method method = getters.next();
               
               // Arguments to this call can be null, because the augmented
               // part of the where is based on non-extent properties only.
               augmentArgs[i] = method.invoke(placeholder);
            }
            
            // Collect the full arrays of parameters and types, including base
            // arguments and augmented.  In the loop below, we will copy the
            // necessary portions of these arrays to temporary ones.
            int fullCount = baseArgCount + callCount;
            Object[] allArgs = new Object[fullCount];
            Type[] allTypes = new Type[fullCount];
            
            // copy base parameters and types into full arrays
            System.arraycopy(baseArgs, 0, allArgs, 0, baseArgCount);
            System.arraycopy(baseTypes, 0, allTypes, 0, baseArgCount);
            
            // copy augmented parameters into full array
            System.arraycopy(augmentArgs, 0, allArgs, baseArgCount, callCount);
            
            // populate allTypes with Hibernate types matching allArgs
            DBUtils.makeTypeArray(allArgs, allTypes);
            
            // possibly overriding result from uncommitted transaction in another session
            DirtyInfo info = null;
            
            // execute HQL statements in turn, until we find a result (or run out of statements)
            Iterator<String> statements = activeBundle.statements();
            while (dmo == null && info == null && statements.hasNext())
            {
               String hql = statements.next();
               int length = baseArgCount + callCount;
               Object[] parms = new Object[length];
               Type[] types = new Type[length];
               
               // copy arguments needed for this pass from full argument list
               System.arraycopy(allArgs, 0, parms, 0, length);
               
               // copy types needed for this pass from full type list
               System.arraycopy(allTypes, 0, types, 0, length);
               
               if (callCount > 0)
               {
                  callCount--;
               }
               
               FibonacciCounter fib;
               long timeout;
               
               needsLock = updateLock && lockType != LockType.NONE;
               if (needsLock)
               {
                  fib = new FibonacciCounter();
                  timeout = fib.next();
               }
               else
               {
                  fib = null;
                  timeout = 0L;
               }
               
               do
               {
                  // find and optionally lock the record
                  try
                  {
                     dmo = persistence.load(buffer,
                                            hql,
                                            parms,
                                            types,
                                            lockType,
                                            timeout,
                                            unique,
                                            updateLock,
                                            false,
                                            findByRowid);
                     
                     // either got the lock or didn't need one
                     needsLock = false;
                     
                     break;
                  }
                  catch (LockTimeoutException exc)
                  {
                     // a DMO was found but could not be locked and we could not determine it
                     // has been deleted in an uncommitted transaction in another session
                     RecordIdentifier ident = exc.getIdentifier();
                     
                     if (LOG.isLoggable(Level.FINE))
                     {
                        LOG.fine("[" +
                                 Utils.describeContext() +
                                 "] FIND lock timeout: "
                                 + timeout
                                 + " ms: "
                                 + ident);
                     }
                     Serializable id = ident.getRecordID();
                     String entity = buffer.getEntityName();
                     RecordIdentifier entityIdent = new RecordIdentifier(entity, id);
                     
                     // try to load the DMO from our session
                     dmo = persistence.quickLoad(entityIdent);
                     
                     // If the record came back null, this indicates another session deleted it
                     // and has since (as in, since we timed out on the lock but before we were
                     // able to quick-load the DMO) committed its transaction. We have to
                     // execute the current query again with the same arguments to try to get
                     // (and lock) another qualifying record. Don't recalculate the timeout in
                     // this case.
                     
                     // if DMO came back non-null, we need to hold a reference to it temporarily
                     // as a placeholder, in case we have to execute another query
                     if (dmo != null)
                     {
                        // evict from the session if not otherwise in use
                        buffer.evictDMOIfUnused(dmo);
                        
                        if (dirtyContext.isDirtyDelete(entity, id, true))
                        {
                           // a record was found but it is being deleted in another session's
                           // uncommitted transaction
                           info = new DirtyInfo();
                           info.setDeleted();
                           
                           // NOTE: we are breaking out of the load/lock loop WITHOUT acquiring
                           // the requested lock! This only works because the logic below forces
                           // the DMO back to null before we can use it. We don't null it here
                           // because we may need it as a placeholder record for another pass.
                           break;
                        }
                     }
                     
                     // prepare for another pass
                     if (dmo == null)
                     {
                        // if same query is to be executed again, reset the timeout value
                        fib.reset();
                        timeout = fib.next();
                     }
                     else
                     {
                        // otherwise, steadily increase the timeout value
                        timeout = fib.next();
                        
                        // log deadlock warning if timeout is getting too long and wait
                        // indefinitely
                        if (timeout > 3600000L)
                        {
                           timeout = 0L;
                           if (LOG.isLoggable(Level.WARNING))
                           {
                              LOG.log(Level.WARNING,
                                      "[" +
                                      Utils.describeContext() +
                                      "] Possible deadlocked record: " + ident,
                                      new Throwable());
                           }
                        }
                     }
                  }
                  catch (MissingRecordException exc)
                  {
                     if (needsLock)
                     {
                        // we got the lock, but the record was deleted in another session's
                        // transaction, which has since been committed; reset the timeout and
                        // try to find another qualifying record
                        fib.reset();
                        timeout = fib.next();
                     }
                     else
                     {
                        break;
                     }
                  }
               } while (true);
               
               if (DatabaseStatistics.isEnabled())
               {
                  Database database = buffer.getDatabase();
                  DatabaseStatistics.get().queryExecuted(this, database, hql);
               }
               
               if (debug)
               {
                  logQueryDetails(hql, parms, dmo);
               }
            }
            
            if (dmo != null           &&
                callCount == 0        &&
                augmentArgs != null   &&
                augmentArgs.length > 0)
            {
               // This won't pick up break value for FIRST and LAST, but we're
               // only concerned with NEXT and PREV anyway.
               breakValue = augmentArgs[0];
            }
            
            // check the dirty share manager for an overriding result
            if (info == null)
            {
               info = dirtyContext.getDirtyInfo(buffer,
                                                index,
                                                activeBundleKey,
                                                activeBundle,
                                                allArgs,
                                                allTypes,
                                                dmo,
                                                hqlPreproc.isFindByRowid());
            }
            
            // if check found nothing of interest, we're done
            if (info == null)
            {
               break;
            }
            
            // keep track of the DMO(s) we found in the primary database; in the event we don't
            // use them otherwise, we'll have to evict them from the session
            if (dmo != null)
            {
               if (primaryDMOs == null)
               {
                  primaryDMOs = new HashSet<>();
               }
               
               primaryDMOs.add(dmo);
            }
            
            // can't trigger revalidation of AdaptiveQuery if dirty database result is in use
            breakValue = null;
            
            // Get the DMO found in the dirty database, if any.  This will be
            // a deep copy of the original, so it is safe to use.
            Persistable dirtyDMO = info.getDirtyDMO();
            
            // If found DMO was deleted, update the placeholder and try again.
            // Also handle the case where the DMO found in the primary
            // database has been modified in such a way that it may no longer
            // satisfy the search criteria.
            if (activeBundleKey != UNIQUE &&
                (info.isDeleted() ||
                 (info.isModified() &&
                  (!info.isInserted() ||
                   (dirtyDMO != null && dirtyDMO.getId().equals(dmo.getId()))))))
            {
               // Test if the record has moved along the index we're walking.
               if (info.isModified() && dirtyDMO != null)
               {
                  Persistable compDirtyDMO = info.getComparableDMO();
                  int comp = dmoSorter.compare(dmo, compDirtyDMO);
                  if (comp == 0)
                  {
                     // Both the primary DMO and dirty DMO sort equivalently
                     // on the index being navigated.  If the dirty DMO does
                     // not represent the full set of uncommitted changes
                     // being tracked in the dirty database, that indicates we
                     // published uncommitted changes early and we must use
                     // the dirty DMO as our result...
                     if (!info.isFullyPublished())
                     {
                        dmo = dirtyDMO;
                        dirtyCopy = true;
                     }
                     
                     // ...otherwise, we use the primary DMO.
                     break;
                  }
               }
               
               // We are about to update the placeholder and try another pass,
               // so if the navigation request was for the first or last
               // record, we have to modify the mode to be next or previous,
               // respectively.  Otherwise, we'll just find the same record in
               // an infinite loop.
               switch (activeBundleKey)
               {
                  case FIRST:
                     activateNext();
                     break;
                  case LAST:
                     activatePrevious();
                     break;
               }
               
               // The found record must be skipped because it was either
               // deleted or modified on the index we're walking.  Update the
               // placeholder to be the record we actually found...
               placeholder = dmo;
               dmo = null;
               
               // ...then try another pass.
               continue;
            }
            
            if (needsLock && dmo != null)
            {
               // If we are here, we determined a provisional DMO we found was deleted in another
               // session's uncommitted transaction and we broke out of the load/lock loop
               // without acquiring the requested lock. The DMO was loaded only for placeholder
               // purposes. Null it out now, since we haven't got the lock; we will either find
               // another qualifying record or give up. A bit redundant with the delete logic
               // in processDirtyResults, but safety first.
               dmo = null;
            }
            
            // determine which of the two candidate DMOs is most appropriate as a result
            dmo = processDirtyResults(persistence, dmo, info, lockType, updateLock);
            
            break;
         }
         
         if (dirtyCopy && primaryDMOs != null)
         {
            // We are not going to use the primary DMO(s) we found, so make
            // sure they are evicted from the session if not in use elsewhere
            // in this context.
            for (Persistable pDMO : primaryDMOs)
            {
               buffer.evictDMOIfUnused(pDMO);
            }
         }
      }
      catch (IllegalAccessException exc)
      {
         throw new RuntimeException(exc);
      }
      catch (InvocationTargetException exc)
      {
         DBUtils.handleException(buffer.getDatabase(), exc);
         
         throw new RuntimeException(exc);
      }
      catch (LockUnavailableException exc)
      {
         handleExecuteException(exc);
         
         // Cannot return record if it could not be locked.
         dmo = null;
      }
      catch (ValidationException exc)
      {
         ErrorManager.throwError(exc);
      }
      catch (PersistenceException exc)
      {
         handleExecuteException(exc);
      }
      finally
      {
         if (dmo == null)
         {
            breakValue = null;
         }
      }
      
      return dmo;
   }
   
   /**
    * Process the results of a dirty index check and return the most
    * appropriate result for the query.  This method may acquire a record lock
    * for the selected result.
    * 
    * @param   persistence
    *          Persistence services object.
    * @param   primaryDMO
    *          Candidate DMO found in the primary database.  Will not be
    *          <code>null</code>.
    * @param   info
    *          Information found during the dirty database check.
    * @param   lockType
    *          Type of lock to apply to the found record.
    * @param   updateLock
    *          <code>true</code> if the status of the lock on the retrieved
    *          record should be modified;  else <code>false</code>.  This is
    *          set to <code>true</code> for actual retrievals, and to
    *          <code>false</code> when only detecting whether a record would
    *          be found (i.e., the <code>hasXXXX()</code> methods).
    * 
    * @return  Either <code>primaryDMO</code> or <code>dirtyDMO</code>,
    *          depending upon which record is most appropriate in terms of the
    *          index being walked, and possibly the most recently found
    *          record, in the case of a relative move (NEXT/PREVIOUS).
    * 
    * @throws  PersistenceException
    *          if any error occurs retrieving a record from the primary
    *          database, after determining its ID in the dirty database.
    * @throws  LockUnavailableException
    *          if a no-wait lock requested on <code>dirtyDMO</code> is
    *          currently unavailable.
    */
   private Persistable processDirtyResults(Persistence persistence,
                                           Persistable primaryDMO,
                                           DirtyInfo info,
                                           LockType lockType,
                                           boolean updateLock)
   throws PersistenceException
   {
      Persistable dmo = primaryDMO;
      Persistable dirtyDMO = info.getDirtyDMO();
      Persistable compDirtyDMO = info.getComparableDMO();
      boolean useDirtyDMO = true;
      
      if (primaryDMO != null && activeBundleKey != UNIQUE)
      {
         // At this point, we have both a candidate DMO from the primary
         // database (dmo), and a candidate DMO from the dirty database
         // (dirtyDMO).  We need to sort them to determine which one is
         // the appropriate "next" record in the index being walked.
         int comp = dmoSorter.compare(primaryDMO, compDirtyDMO);
         if (activeBundleKey == LAST || activeBundleKey == PREVIOUS)
         {
            comp *= -1;
         }
         
         useDirtyDMO = (comp >= 0);
      }
      
      if (useDirtyDMO)
      {
         if (info.isInserted() && (info.isLocal() || LockType.NONE.equals(lockType)))
         {
            // Use the dirty copy if no lock has been requested, or if the record found in the
            // dirty database is local to the session.
            // In the latter case (info.isLocal() == true), the dirty record actually is the
            // local original. This comes into play when a DMO is created in one buffer and is
            // queried by another buffer in the same session, before the first buffer has
            // validated and flushed the record (it is validated as a side effect of the
            // getDirtyInfo() call). The record is not found in primary database (since it has
            // not yet been flushed), but it is found in the dirty database, and the dirty
            // context substitutes the original DMO for the dirty one in getDirtyInfo().
            dmo = dirtyDMO;
            
            // mark the record as dirty and potentially from another context; we do this even in
            // the local case, since this prevents us from trying to reassociate the record with
            // the current Hibernate session more than once, in certain cases (see
            // Persistence$Context.getSession())
            dirtyCopy = true;
         }
         else if (dirtyDMO != null)
         {
            // Just use the primary key of the record found in the dirty
            // database, but reload the record from the primary database.
            // Note:  it might not be there, if the record existed only in
            // another context, and was never committed.
            Serializable id = dirtyDMO.getId();
            try
            {
               dmo = persistence.load(buffer, id, lockType, 0L, updateLock, false);
            }
            catch (MissingRecordException exc)
            {
               dmo = null;
            }
         }
         else if (info.isDeleted() || info.isModified())
         {
            dmo = null;
         }
      }
      
      return dmo;
   }
   
   /**
    * Log debug details about the inputs and results of a query.
    * 
    * @param   hql
    *          HQL query statement.
    * @param   parms
    *          Query substitution parameters
    * @param   dmo
    *          DMO record found, if any.
    */
   private void logQueryDetails(String hql, Object[] parms, Persistable dmo)
   {
      LOG.log(Level.FINE, "HQL:  " + hql);
      
      StringBuilder buf = new StringBuilder("PARMS:  ");
      int len = parms.length;
      if (len == 0)
      {
         buf.append("N/A");
      }
      
      for (int k = 0; k < len; k++)
      {
         if (k > 0)
         {
            buf.append(", ");
         }
         Object next = parms[k];
         if (next instanceof BaseDataType)
         {
            buf.append(((BaseDataType) next).toStringMessage());
         }
         else
         {
            buf.append(String.valueOf(next));
         }
      }
      
      LOG.log(Level.FINE, buf.toString());
      
      buf.setLength(0);
      buf.append("DMO:  ");
      if (dmo != null)
      {
         buf.append(dmo.getId());
      }
      else
      {
         buf.append("<not found>");
      }
      
      LOG.log(Level.FINE, buf.toString());
      
      if (LOG.isLoggable(Level.FINEST))
      {
         buf.setLength(0);
         buf.append("DMO Detail:  ");
         buf.append(buffer.toString(dmo));
         
         LOG.log(Level.FINEST, buf.toString());
      }
   }
}