/*
** Module   : TableMapper.java
** Abstract : Table and field name to converted DMO and property name (bidirectional).
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20131015 Created initial version.
** 002 CA  20131021 Added legacy index support.
** 003 CA  20131029 mapPermanentDMO must register the mapper for the permanent schema in the
**                  permanentDBs map.
** 004 SVL 20131115 Changes to functions that return DMO class.
** 005 SVL 20140106 Fixed mapping logic for temp tables.
** 006 OM  20140117 Added support for indexed access to fields of the buffer.
** 007 SVL 20140123 Added getPropertyName(Class, int).
** 008 VMN 20140122 Added format(), initial(), columnLabel(), label(), mandatory(),
**                  validateMessage(), validateExpression() support.
** 009 ECF 20140301 Added getLegacyFieldToPropertyMap, a helper method optimized to retrieve all
**                  legacy field name to DMO property name mappings at once.
** 010 SVL 20140320 Fixed propertyName(Class<?>, int).
** 011 VMN 20140328 Added support for custom denormalization of fields with extent.
** 012 VMN 20140409 Changes to getLegacyFieldNameMap() method, return LegacyFieldNameMap.
** 013 CA  20140513 Added a weight for the context-local var, to ensure predetermined order during
**                  context reset.
** 014 ECF 20140920 Memory optimizations: intern many strings and re-use LegacyTableInfo objects
**                  which share a DMO class, instead of creating duplicates.
** 015 VMN 20150227 Fixed LegacyFieldInfo instances used in map denorm2j.
** 016 ECF 20150313 Refactored to store statically defined table information in shared caches,
**                  and dynamically defined temp-table information in a context-local cache, to
**                  conserve memory and avoid redundant mapping work for static temp-tables.
** 017 VMN 20150402 Expand using getDenormalizedProperty() for temp-tables.
** 018 SVL 20150503 Added getPropertyName(Buffer, String).
** 019 ECF 20150810 Denormalized schema fixes.
** 020 EVL 20160223 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 021 ECF 20160225 Changes required by new PropertyHelper implementation. Implemented backing
**                  logic for NUM-FIELDS attribute.
** 022 ECF 20160403 Avoid unnecessary context initialization.
** 023 IAS 20160511 Create temporary table LIKE a permanent denormalized one
** 024 SVL 20160610 Support for table-level delete validation.
** 025 SVL 20160626 Added helper functions for detecting table joins at runtime.
** 026 SVL 20160705 Added methods to support INDEX-INFORMATION.
** 027 OM  20161014 Normalized field name in getPropertyName() methods.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import java.lang.reflect.*;
import java.util.*;

import com.goldencode.p2j.convert.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.*;

/**
 * This class performs a bi-directional runtime mapping of DMO implementation classes (permanent 
 * or temporary) to their associated legacy table name. The mapping for a DMO implementation class
 * is held in a {@link LegacyTableInfo} instance.
 * <p>
 * For physical DBs, the mappings for the permanent DBs (connected at server startup) will be 
 * registered at server startup. For permanent DBs connected at application runtime, the mappings
 * will be registered and deregistered at the appropriate DB connection/disconnection times.
 * <p>
 * All the mappings for temporary tables will be registered during application lifetime, as the
 * temp tables get created or dropped.
 * <p>
 * APIs which resolve a legacy name for a converted name may return null, when resolving an index,
 * field or property name; this is to allow checks for synthetic properties or indexes, which do
 * not exist in the legacy code.  In all cases, it is expected that the given table name is for a
 * table which has registered mappings; else, a {@link NullPointerException} will be thrown. 
 * 
 * @param    <K>
 *           The type of the key used by the {@link #p2j progress-to-java} map. For permanent
 *           DMOs, this will be a {@link String}. For temporary DMOs, this will be an 
 *           {@link Integer} value with the <code>UNIQUE-ID</code> identifying the 
 *           {@link TempTable} resource.
 */
public abstract class TableMapper<K>
{
   /** Holds the mappers for all permanent databases by their schema name */
   private static final Map<String, PermanentTableMapper> permanentDBs = new HashMap<>();
   
   /** Holds the mapper for all statically defined temp tables */
   private static final TempTableMapper staticTempDB = new StaticTempTableMapper();
   
   /** Context-local mapper for all dynamic temporary tables */
   private static final ContextLocal<TempTableMapper> dynamicDB =
   new ContextLocal<TempTableMapper>()
   {
      @Override
      public WeightFactor getWeight()
      {
         return WeightFactor.LEVEL_2;
      }
      
      @Override
      protected TempTableMapper initialValue()
      {
         return new DynamicTempTableMapper();
      }
   };
   
   /** The schema name */
   private final String schema;
   
   /**
    * Create a new mapper for the given schema.
    * 
    * @param    schema
    *           The schema name.
    */
   private TableMapper(String schema)
   {
      this.schema = schema;
   }
   
   /**
    * Get new field name for denormalized legacy field name and index.
    *
    * @param    dmoIface
    *           The DMO interface for a permanent DMO.
    * @param    property
    *           The DMO property name.
    * @param    index
    *           zero-based index of this property in denormalization hints or null.
    *
    * @return   New field name.
    */
   public static String getDenormalizedProperty(Class<?> dmoIface, String property, Integer index)
   {
      String schema = DatabaseManager.getSchemaByInterface(dmoIface);
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         return staticTempDB.denormalizedProperty(dmoIface, property, index);
      }
      else
      {
         synchronized (permanentDBs)
         {
            PermanentTableMapper permanentTableMapper = locatePerm(schema);
            
            return permanentTableMapper.denormalizedProperty(dmoIface, property, index);
         }
      }
   }
   
   /**
    * Get the legacy table name for the DMO referenced by the given buffer.
    * 
    * @param    buffer
    *           A buffer instance.
    * 
    * @return   The legacy table name.
    */
   public static String getLegacyName(Buffer buffer)
   {
      RecordBuffer buf = ((BufferImpl) buffer).buffer();
      if (buf.isTemporary())
      {
         TempTable tt = buf.getParentTable();
         String name = tt.name().getValue();
         if (name != null)
         {
            // return the name of the parent table
            return name;
         }
         else
         {
            // name not yet assigned to a static temp table, get it from the mapping
            return getLegacyName(tt);
         }
      }
      else
      {
         return getLegacyName(buf.getDMOInterface());
      }
   }
   
   /**
    * Get the COLUMN-LABEL field attribute for the property associated with the DMO referenced
    * by the given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    *
    * @return   COLUMN-LABEL field attribute or <code>null</code> if there is no such field.
    */
   public static String getLegacyFieldColumnLabel(Buffer buffer, String property)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      
      return legacyFieldInfo == null ? null : legacyFieldInfo.columnLabel;
   }
   
   /**
    * Set the COLUMN-LABEL field attribute for the property associated with the DMO referenced
    * by the given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    * @param    columnLabel
    *           COLUMN-LABEL field attribute.
    */
   public static void setLegacyFieldColumnLabel(Buffer buffer,
                                                String property,
                                                String columnLabel)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      legacyFieldInfo.columnLabel = columnLabel.intern();
   }
   
   /**
    * Get the LABEL field attribute for the property associated with the DMO referenced
    * by the given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    *
    * @return   LABEL field attribute or <code>null</code> if there is no such field.
    */
   public static String getLegacyFieldLabel(Buffer buffer, String property)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      
      return legacyFieldInfo == null ? null : legacyFieldInfo.label;
   }
   
   /**
    * Set the LABEL field attribute for the property associated with the DMO referenced
    * by the given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    * @param    label
    *           LABEL field attribute.
    */
   public static void setLegacyFieldLabel(Buffer buffer, String property, String label)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      legacyFieldInfo.label = label.intern();
   }
   
   /**
    * Get the id of field in the sequence of the fields of the buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    *
    * @return   the id of field in the sequence of the fields of the buffer.
    */
   public static integer getLegacyFieldId(Buffer buffer, String property)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      
      return new integer(legacyFieldInfo == null ? null : legacyFieldInfo.fieldId);
   }
   
   /**
    * Get the LITERAL-QUESTION field attribute for the property associated with the DMO
    * referenced by the given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    *
    * @return   LITERAL-QUESTION field attribute or <code>null</code> if there is no such field.
    */
   public static logical getLegacyFieldLiteralQuestion(Buffer buffer, String property)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      
      return new logical(legacyFieldInfo == null ? null : legacyFieldInfo.literalQuestion);
   }
   
   /**
    * Set the LITERAL-QUESTION field attribute for the property associated with the DMO
    * referenced by the given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    * @param    literalQuestion
    *           LITERAL-QUESTION field attribute.
    */
   public static void setLegacyFieldLiteralQuestion(Buffer buffer,
                                                    String property,
                                                    logical literalQuestion)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      legacyFieldInfo.literalQuestion = literalQuestion.booleanValue();
   }
   
   /**
    * Get the MANDATORY field attribute for the property associated with the DMO
    * referenced by the given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    *
    * @return   MANDATORY field attribute or <code>null</code> if there is no such field.
    */
   public static logical isLegacyFieldMandatory(Buffer buffer, String property)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      
      return new logical(legacyFieldInfo == null ? null : legacyFieldInfo.mandatory);
   }
   
   /**
    * Get the VALIDATE-EXPRESSION field attribute for the property associated with the DMO
    * referenced by the given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    *
    * @return   VALIDATE-EXPRESSION field attribute or <code>null</code> if there is no such
    *           field.
    */
   public static String getLegacyFieldValidateExpression(Buffer buffer, String property)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      
      return legacyFieldInfo == null ? null : legacyFieldInfo.validateExpression;
   }
   
   /**
    * Set the VALIDATE-EXPRESSION field attribute for the property associated with the DMO
    * referenced by the given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    * @param    validateExpression
    *           VALIDATE-EXPRESSION field attribute.
    */
   public static void setLegacyFieldValidateExpression(Buffer buffer,
                                                    String property,
                                                    String validateExpression)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      legacyFieldInfo.validateExpression = validateExpression;
   }
   
   /**
    * Get the VALIDATE-MESSAGE field attribute for the property associated with the DMO
    * referenced by the given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    *
    * @return   VALIDATE-MESSAGE field attribute or <code>null</code> if there is no
    *           such field.
    */
   public static String getLegacyFieldValidateMessage(Buffer buffer, String property)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      
      return legacyFieldInfo == null ? null : legacyFieldInfo.validateMessage;
   }
   
   /**
    * Set the VALIDATE-MESSAGE field attribute for the property associated with the DMO
    * referenced by the given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    * @param    validateMessage
    *           VALIDATE-MESSAGE field attribute.
    */
   public static void setLegacyFieldValidateMessage(Buffer buffer,
                                                    String property,
                                                    String validateMessage)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      legacyFieldInfo.validateMessage = validateMessage;
   }
   
   /**
    * Get the FORMAT field attribute for the property associated with the DMO referenced by the
    * given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    *
    * @return   FORMAT field attribute or <code>null</code> if there is no such field.
    */
   public static String getLegacyFieldFormat(Buffer buffer, String property)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      
      return legacyFieldInfo == null ? null : legacyFieldInfo.format;
   }
   
   /**
    * Set the FORMAT field attribute for the property associated with the DMO referenced by the
    * given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    * @param    format
    *           FORMAT field attribute.
    */
   public static void setLegacyFieldFormat(Buffer buffer, String property, String format)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      legacyFieldInfo.format = format.intern();
   }
   
   /**
    * Get the INITIAL field attribute for the property associated with the DMO referenced by the
    * given buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    *
    * @return   INITIAL field attribute or <code>null</code> if there is no such field.
    */
   public static String getLegacyFieldInitial(Buffer buffer, String property)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      
      return legacyFieldInfo == null ? null : legacyFieldInfo.initial;
   }
   
   /**
    * Get the legacy field name for the property associated with the DMO referenced by the given
    * buffer.
    * 
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    * 
    * @return   The legacy field name or <code>null</code> if there is no such field.
    */
   public static String getLegacyFieldName(Buffer buffer, String property)
   {
      LegacyFieldInfo legacyFieldInfo = getLegacyFieldInfo(buffer, property);
      
      return legacyFieldInfo == null ? null : legacyFieldInfo.legacyName;
   }

   /**
    * Get information about the index associated with the DMO referenced by the given buffer,
    * as a string.
    *
    * @param    buffer
    *           A buffer instance which specifies the DMO class associated with the target index.
    * @param    indexNum
    *           0-based index number in the index definition order.
    *
    * @return   a string describing the target index or <code>null</code> if there is no index
    *           with the given number. Format of the string is described in
    *           {@link Buffer#indexInformation(int)}.
    */
   public static String getLegacyIndexInfoString(Buffer buffer, int indexNum)
   {
      LegacyIndexInfo legacyIndexInfo = getLegacyIndexInfo(buffer, indexNum);

      return legacyIndexInfo == null ? null : legacyIndexInfo.indexInformation();
   }
   
   /**
    * Get the legacy index name of a converted index, defined by the backing table of the 
    * specified buffer.
    * 
    * @param    buffer
    *           A buffer instance.
    * @param    idx
    *           The converted index name.
    *           
    * @return   The legacy index name or <code>null</code> if there is no such index defined.
    */
   public static String getLegacyIndexName(Buffer buffer, String idx)
   {
      RecordBuffer buf = ((BufferImpl) buffer).buffer();
      if (buf.isTemporary())
      {
         return getLegacyIndexName((TempTable) buffer.tableHandle().get(), idx);
      }
      else
      {
         return getLegacyIndexName(buf.getDMOInterface(), idx);
      }
   }
   
   /**
    * Get the legacy table name for the given DMO, which is part of a permanent DB.
    * 
    * @param    dmoIface
    *           The DMO interface for a permanent DMO.
    *           
    * @return   The legacy table name.
    * 
    * @throws   IllegalStateException
    *           If a DMO belonging to the {@link DatabaseManager#TEMP_TABLE_SCHEMA} temp schema
    *           is passed as a parameter; for these cases, {@link #getLegacyName(TempTable)}
    *           or {@link #getLegacyName(Buffer)} must be used.
    */
   public static String getLegacyName(Class<?> dmoIface)
   {
      String schema = DatabaseManager.getSchemaByInterface(dmoIface);
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         final String msg =
            "Legacy table name for a temporary DMO can not be resolved using this API!";
         
         throw new IllegalArgumentException(msg);
      }
      
      synchronized (permanentDBs)
      {
         PermanentTableMapper mapper = locatePerm(schema);
         
         return mapper.legacyName(dmoIface);
      }
   }

   /**
    * Determine if the permanent table has table-level delete validation.
    *
    * @param    dmoIface
    *           The DMO interface for a permanent DMO which defines the target table.
    *
    * @return   <code>true</code> if the table has table-level delete validation.
    *
    * @throws   IllegalStateException
    *           If a DMO belonging to the {@link DatabaseManager#TEMP_TABLE_SCHEMA} temp schema
    *           is passed as a parameter.
    */
   public static boolean hasValidation(Class<?> dmoIface)
   {
      String schema = DatabaseManager.getSchemaByInterface(dmoIface);
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         final String msg = "Table-level validation is not available for temporary tables!";
         throw new IllegalArgumentException(msg);
      }

      synchronized (permanentDBs)
      {
         PermanentTableMapper mapper = locatePerm(schema);

         return mapper.validation(dmoIface);
      }
   }
   
   /**
    * Get the legacy field name for the given property, which is defined by a DMO part of a 
    * permanent DB.
    * 
    * @param    dmoIface
    *           The DMO interface for a permanent DMO.
    * @param    property
    *           The DMO property name.
    *           
    * @return   The legacy field name or <code>null</code> if there is no such field.
    * 
    * @throws   IllegalStateException
    *           If a DMO belonging to the {@link DatabaseManager#TEMP_TABLE_SCHEMA} temp schema
    *           is passed as a parameter; for these cases, {@link #getLegacyFieldName(TempTable, 
    *           String)} or {@link #getLegacyFieldName(Buffer, String)} must be used.
    */
   public static String getLegacyFieldName(Class<?> dmoIface, String property)
   {
      String schema = DatabaseManager.getSchemaByInterface(dmoIface);
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         final String msg =
            "Legacy field names for a property of a temporary DMO can not be resolved using " +
            "this API!";
         throw new IllegalArgumentException(msg);
      }
      
      synchronized (permanentDBs)
      {
         PermanentTableMapper mapper = locatePerm(schema);
         
         return mapper.legacyFieldName(dmoIface, property);
      }
   }
   
   /**
    * Get the value of the extent field
    * 
    * @param dmoIface
    *        The DMO interface for a permanent DMO.
    * @param fieldNameHibernate
    *        the field's Hibernate name
    * 
    * @return the value of the first element of the extent field
    */
   public static BaseDataType getExtentValue(Class<?> dmoIface, String fieldNameHibernate) 
   {
      String schema = DatabaseManager.getSchemaByInterface(dmoIface);
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         return null;
      }
      
      synchronized (permanentDBs)
      {
         LegacyFieldInfo info = locatePerm(schema).j2p().get(dmoIface).
                                       extents.get(fieldNameHibernate);
         return info == null ? null : info.initialValue;
      }
   }
   
   /**
    * Get the value of the non-extent field
    * 
    * @param dmoIface
    *        The DMO interface for a permanent DMO.
    * @param fieldNameHibernate
    *        the field's Hibernate name
    * 
    * @return the value of the first element of the extent field
    */
   public static BaseDataType getFieldValue(Class<?> dmoIface, String fieldNameHibernate) 
   {
      String schema = DatabaseManager.getSchemaByInterface(dmoIface);
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         return null;
      }
      
      synchronized (permanentDBs)
      {
         LegacyFieldInfo info = locatePerm(schema).j2p().get(dmoIface).
                                       j2p.get(fieldNameHibernate);
         return info == null ? null : info.initialValue;
      }
   }
   

   /**
    * Get the 4GL name of the extent field
    * 
    * @param dmoIface
    *        The DMO interface for a permanent DMO.
    * @param fieldNameHibernate
    *        the field's Hibernate name
    * 
    * @return the 4GL name of the extent field
    */
   public static String getExtent4GLName(Class<?> dmoIface, String fieldNameHibernate)
   {
      String schema = DatabaseManager.getSchemaByInterface(dmoIface);
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         return null;
      }
      
      synchronized (permanentDBs)
      {
         LegacyFieldInfo info = locatePerm(schema).j2p().get(dmoIface).
                                       extents.get(fieldNameHibernate);
         
         return info == null ? fieldNameHibernate : info.legacyName;
      }
   }

   /**
    * Get the legacy field for the given property, which is defined by a DMO part of a
    * permanent DB.
    *
    * @param    dmoIface
    *           The DMO interface for a permanent DMO.
    * @param    property
    *           The DMO property name.
    *
    * @return   The legacy field or <code>null</code> if a table belonging to the
    *           {@link DatabaseManager#TEMP_TABLE_SCHEMA} is used or there is no such field.
    *
    */
   public static LegacyFieldInfo getLegacyFieldInfo(Class<?> dmoIface, String property)
   {
      String schema = DatabaseManager.getSchemaByInterface(dmoIface);
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         return null;
      }
      
      synchronized (permanentDBs)
      {
         PermanentTableMapper mapper = locatePerm(schema);
         
         return mapper.legacyFieldInfo(dmoIface, property);
      }
   }
   
   /**
    * Get the legacy fields' list which are defined by a DMO part of a permanent DB.
    *
    * @param    dmoIface
    *           The DMO interface for a permanent DMO.
    * @return   The legacy fields' list if a table belonging to the
    *           {@link DatabaseManager#TEMP_TABLE_SCHEMA} or empty list if it is not.
    *
    */
   public static List<String> getLegacyProperties(Class<?> dmoIface)
   {
      String schema = DatabaseManager.getSchemaByInterface(dmoIface);
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         return Collections.emptyList();
      }
      
      synchronized (permanentDBs)
      {
         PermanentTableMapper mapper = locatePerm(schema);
         List<LegacyFieldInfo> lfi = new ArrayList<>(mapper.j2p().get(dmoIface).j2p.values());
         Collections.sort(lfi, (f1, f2) -> f1.fieldId.compareTo(f2.fieldId));
         List<String> lp = new ArrayList<>(lfi.size());
         String ln = "";
         for (LegacyFieldInfo lf: lfi) 
         {
            if (ln.equals(lf.legacyName)) 
            {
               continue;
            }
            ln = lf.legacyName;
            lp.add("".equals(lf.original) ? lf.javaName : lf.original);
         }
         return lp;
      }
   }
   
   /**
    * Get the legacy name of a converted index, which is defined by the backing table for the 
    * specified buffer, part of a permanent DB.
    * 
    * @param    dmoIface
    *           The DMO interface for a permanent DMO.
    * @param    idx
    *           The converted index name.
    *           
    * @return   The legacy index name or <code>null</code> if there is no such index defined.
    * 
    * @throws   IllegalStateException
    *           If a DMO belonging to the {@link DatabaseManager#TEMP_TABLE_SCHEMA} temp schema
    *           is passed as a parameter; for these cases, {@link #getLegacyFieldName(TempTable, 
    *           String)} or {@link #getLegacyFieldName(Buffer, String)} must be used.
    */
   public static String getLegacyIndexName(Class<?> dmoIface, String idx)
   {
      String schema = DatabaseManager.getSchemaByInterface(dmoIface);
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         final String msg =
            "Legacy index names for a temporary DMO can not be resolved using this API!";
         
         throw new IllegalArgumentException(msg);
      }
      
      synchronized (permanentDBs)
      {
         PermanentTableMapper mapper = locatePerm(schema);
         
         return mapper.legacyIndexName(dmoIface, idx);
      }
   }
   
   /**
    * Get the legacy table name for the given {@link TempTable} resource.
    * 
    * @param    tt
    *           The {@link TempTable} resource.
    *           
    * @return   The legacy table name.
    */
   public static String getLegacyName(TempTable tt)
   {
      TempTableMapper mapper = locateTemp(tt);
      
      synchronized (mapper)
      {
         return mapper.legacyName(tt);
      }
   }
   
   /**
    * Get the legacy field name for the given {@link TempTable} resource and property.
    * 
    * @param    tt
    *           The {@link TempTable} resource.
    * @param    property
    *           The DMO property name.
    * 
    * @return   The legacy field name or <code>null</code> if there is no such field.
    */
   public static String getLegacyFieldName(TempTable tt, String property)
   {
      TempTableMapper mapper = locateTemp(tt);
      
      synchronized (mapper)
      {
         return mapper.legacyFieldInfo(tt, property).legacyName;
      }
   }
   
   /**
    * Get the legacy field for the given {@link TempTable} resource and property.
    *
    * @param    tt
    *           The {@link TempTable} resource.
    * @param    property
    *           The DMO property name.
    *
    * @return   The legacy field or <code>null</code> if there is no such field.
    */
   public static LegacyFieldInfo getLegacyFieldInfo(TempTable tt, String property)
   {
      TempTableMapper mapper = locateTemp(tt);
      
      synchronized (mapper)
      {
         return mapper.legacyFieldInfo(tt, property);
      }
   }
   
   /**
    * Get the legacy index name for the given {@link TempTable} resource and converted index name.
    * 
    * @param    tt
    *           The {@link TempTable} resource.
    * @param    idx
    *           The converted index name.
    *           
    * @return   The legacy index name or <code>null</code> if there is no such index defined.
    */
   public static String getLegacyIndexName(TempTable tt, String idx)
   {
      TempTableMapper mapper = locateTemp(tt);
      
      synchronized (mapper)
      {
         return mapper.legacyIndexName(tt, idx);
      }
   }
   
   /**
    * Get the DMO implementation class for the specified {@link TempTable} resource.
    * 
    * @param    table
    *           The {@link TempTable} resource.
    *           
    * @return   The DMO implementation class.
    */
   public static Class<?> getDMOClass(TempTable table)
   {
      // no reason to search the mappings, get it from the persistence info
      return table.getDMOClass();
   }
   
   /**
    * Resolve the DMO property name for a legacy field name belonging to the specified 
    * {@link TempTable} resource.
    * 
    * @param    table
    *           The {@link TempTable} resource.
    * @param    field
    *           The legacy field name.
    *           
    * @return   The DMO property name associated with the legacy field or <code>null</code> if
    *           there is no such field.
    */
   public static String getPropertyName(TempTable table, String field)
   {
      return getPropertyName(table, field, null);
   }
   
   /**
    * Resolve the DMO property name for a legacy field name belonging to the specified
    * {@link TempTable} resource.
    *
    * @param    table
    *           The {@link TempTable} resource.
    * @param    field
    *           The legacy field name.
    * @param    index
    *           <code>null</code> for usual properties and zero-based index of this property for
    *           denormalized field with extent.
    *
    * @return   The DMO property name associated with the legacy field or <code>null</code> if
    *           there is no such field.
    */
   public static String getPropertyName(TempTable table, String field, Integer index)
   {
      TempTableMapper mapper = locateTemp(table);
      
      synchronized (mapper)
      {
         Integer key = mapper.getLegacyKey(table);
         
         String normalized4glField = TextOps.rightTrim(field).toStringMessage().toLowerCase();
         return mapper.getPropertyName(key, normalized4glField, index);
      }
   }
   
   /**
    * Resolve the DMO property name for a field index name belonging to the specified
    * {@link TempTable} resource.
    *
    * @param    table
    *           The {@link TempTable} resource.
    * @param    fieldId
    *           The field index in table declaring sequence.
    *
    * @return   The DMO property name associated with the legacy field or <code>null</code> if
    *           there is no such field.
    */
   public static String getPropertyName(TempTable table, int fieldId)
   {
      return getPropertyName(table, fieldId, null);
   }
   
   /**
    * Resolve the DMO property name for a field index name belonging to the specified
    * {@link TempTable} resource.
    *
    * @param    table
    *           The {@link TempTable} resource.
    * @param    fieldId
    *           The field index in table declaring sequence.
    * @param    index
    *           <code>null</code> for usual properties and zero-based index of this property for
    *           denormalized field with extent.
    *
    * @return   The DMO property name associated with the legacy field or <code>null</code> if
    *           there is no such field.
    */
   public static String getPropertyName(TempTable table, int fieldId, Integer index)
   {
      TempTableMapper mapper = locateTemp(table);
      
      synchronized (mapper)
      {
         Integer key = mapper.getLegacyKey(table);
         
         return mapper.getPropertyName(key, fieldId, index);
      }
   }
   
   /**
    * Resolve the number of fields belonging to the specified {@link TempTable} resource.
    *
    * @param    table
    *           The {@link TempTable} resource.
    *
    * @return   The DMO property name associated with the legacy field or <code>null</code> if
    *           there is no such field.
    */
   public static int getNumFields(TempTable table)
   {
      TempTableMapper mapper = locateTemp(table);
      
      synchronized (mapper)
      {
         Integer key = mapper.getLegacyKey(table);
         
         return mapper.getNumFields(key);
      }
   }
   
   /**
    * Resolve the converted index name for a legacy index belonging to the temp-table defined
    * by the {@link TempTable} resource.
    * 
    * @param    table
    *           The {@link TempTable} resource.
    * @param    idx
    *           The legacy index name.
    *           
    * @return   The converted index name or <code>null</code> if there is no such index.
    */
   public static String getIndexName(TempTable table, String idx)
   {
      TempTableMapper mapper = locateTemp(table);
      
      synchronized (mapper)
      {
         Integer key = mapper.getLegacyKey(table);
         
         return mapper.getIndexName(key, idx);
      }
   }
   
   /**
    * Get the DMO implementation class for the legacy table name (part of the specified schema).
    * 
    * @param    schema
    *           The schema name.
    * @param    table
    *           The legacy table name.
    *           
    * @return   The DMO implementation class or <code>null</code> if there is no table with the
    *           given name.
    * 
    * @throws   IllegalArgumentException
    *           If a table belonging to the {@link DatabaseManager#TEMP_TABLE_SCHEMA} is used.
    */
   public static Class<?> getDMOClass(String schema, String table)
   {
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         final String msg = "DMO for a temporary table can not be resolved using this API!";
         throw new IllegalArgumentException(msg);
      }
      
      synchronized (permanentDBs)
      {
         PermanentTableMapper mapper = locatePerm(schema);
         
         return mapper.getDMOClass(table.toLowerCase());
      }
   }
   
   /**
    * Get the DMO property name for a legacy field name belonging to the specified legacy table.
    * 
    * @param    schema
    *           The schema name.
    * @param    table
    *           The legacy table name.
    * @param    field
    *           The legacy field name.
    *           
    * @return   The DMO property name associated with the legacy field or <code>null</code> if
    *           there is no such field.
    * 
    * @throws   IllegalArgumentException
    *           If a table belonging to the {@link DatabaseManager#TEMP_TABLE_SCHEMA} is used.
    */
   public static String getPropertyName(String schema, String table, String field)
   {
      return getPropertyName(schema, table, field, null);
   }
   
   /**
    * Get the DMO property name for a legacy field name belonging to the specified legacy table.
    *
    * @param    schema
    *           The schema name.
    * @param    table
    *           The legacy table name.
    * @param    field
    *           The legacy field name.
    * @param    index
    *           <code>null</code> for usual properties and zero-based index of this property for
    *           denormalized field with extent.
    *
    * @return   The DMO property name associated with the legacy field or <code>null</code> if
    *           there is no such field.
    *
    * @throws   IllegalArgumentException
    *           If a table belonging to the {@link DatabaseManager#TEMP_TABLE_SCHEMA} is used.
    */
   public static String getPropertyName(String schema, String table, String field, Integer index)
   {
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         final String msg = "Properties for temporary DMOs can not be resolved using this API!";
         throw new IllegalArgumentException(msg);
      }
      
      synchronized (permanentDBs)
      {
         PermanentTableMapper mapper = locatePerm(schema);
         
         String normalized4glField = TextOps.rightTrim(field).toStringMessage().toLowerCase();
         return mapper.getPropertyName(table.toLowerCase(), normalized4glField, index);
      }
   }
   
   /**
    * Get the DMO property name for a field index belonging to the specified legacy table.
    *
    * @param    schema
    *           The schema name.
    * @param    table
    *           The legacy table name.
    * @param    fieldId
    *           The field index.
    *
    * @return   The DMO property name associated with the legacy field or <code>null</code> if
    *           there is no such field.
    *
    * @throws   IllegalArgumentException
    *           If a table belonging to the {@link DatabaseManager#TEMP_TABLE_SCHEMA} is used.
    */
   public static String getPropertyName(String schema, String table, int fieldId)
   {
      return getPropertyName(schema, table, fieldId, null);
   }
   
   /**
    * Get the DMO property name for a field index belonging to the specified legacy table.
    *
    * @param    schema
    *           The schema name.
    * @param    table
    *           The legacy table name.
    * @param    fieldId
    *           The field index.
    * @param    index
    *           <code>null</code> for usual properties and zero-based index of this property for
    *           denormalized field with extent.
    *
    * @return   The DMO property name associated with the legacy field or <code>null</code> if
    *           there is no such field.
    *
    * @throws   IllegalArgumentException
    *           If a table belonging to the {@link DatabaseManager#TEMP_TABLE_SCHEMA} is used.
    */
   public static String getPropertyName(String schema, String table, int fieldId, Integer index)
   {
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         final String msg = "Properties for temporary DMOs can not be resolved using this API!";
         throw new IllegalArgumentException(msg);
      }
      
      synchronized (permanentDBs)
      {
         PermanentTableMapper mapper = locatePerm(schema);
         
         return mapper.getPropertyName(table.toLowerCase(), fieldId, index);
      }
   }
   
   /**
    * Get the number of fields belonging to the specified legacy table.
    *
    * @param    schema
    *           The schema name.
    * @param    table
    *           The legacy table name.
    *
    * @return   Number of fields in the table.
    *
    * @throws   IllegalArgumentException
    *           If a table belonging to the {@link DatabaseManager#TEMP_TABLE_SCHEMA} is used.
    */
   public static int getNumFields(String schema, String table)
   {
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         final String msg = "Properties for temporary DMOs can not be resolved using this API!";
         throw new IllegalArgumentException(msg);
      }
      
      synchronized (permanentDBs)
      {
         PermanentTableMapper mapper = locatePerm(schema);
         
         return mapper.getNumFields(table.toLowerCase());
      }
   }
   
   /**
    * Get the DMO property name for a field index belonging to the specified legacy table.
    *
    * @param    dmoIface
    *           The DMO interface for a permanent DMO.
    * @param    fieldId
    *           The field index.
    *
    * @return   The DMO property name associated with the legacy field or <code>null</code> if
    *           there is no such field.
    *
    * @throws   IllegalArgumentException
    *           If a table belonging to the {@link DatabaseManager#TEMP_TABLE_SCHEMA} is used.
    */
   public static String getPropertyName(Class<?> dmoIface, int fieldId)
   {
      String schema = DatabaseManager.getSchemaByInterface(dmoIface);
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         final String msg = "Properties for temporary DMOs can not be resolved using this API!";
         throw new IllegalArgumentException(msg);
      }
      
      synchronized (permanentDBs)
      {
         PermanentTableMapper mapper = locatePerm(schema);
         
         return mapper.propertyName(dmoIface, fieldId);
      }
   }
   
   /**
    * Get the DMO property name for the given buffer and legacy field name.
    *
    * @param  buffer
    *         Buffer which belongs to the target table.
    * @param  legacyFieldName
    *         Legacy name of the target field.
    *
    * @return DMO property name.
    */
   public static String getPropertyName(Buffer buffer, String legacyFieldName)
   {
      RecordBuffer recordBuffer = RecordBuffer.get((DataModelObject) buffer);
      handle tableHandle = recordBuffer.tableHandle();
      return tableHandle.isUnknown() ?
             getPropertyName(recordBuffer.getSchema(), recordBuffer.getTable(), legacyFieldName) :
             getPropertyName(tableHandle.unwrapTempTable(), legacyFieldName);
   }


   /**
    * Resolve the converted index name for a legacy index belonging to the specified legacy table.
    * 
    * @param    schema
    *           The schema name.
    * @param    table
    *           The legacy table name.
    * @param    idx
    *           The legacy index name.
    *           
    * @return   The converted index name or <code>null</code> if there is no such index.
    * 
    * @throws   IllegalArgumentException
    *           If a table belonging to the {@link DatabaseManager#TEMP_TABLE_SCHEMA} is used.
    */
   public static String getIndexName(String schema, String table, String idx)
   {
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         final String msg = "Properties for temporary DMOs can not be resolved using this API!";
         throw new IllegalArgumentException(msg);
      }
      
      synchronized (permanentDBs)
      {
         PermanentTableMapper mapper = locatePerm(schema);
         
         return mapper.getIndexName(table.toLowerCase(), idx);
      }
   }
   
   /**
    * Add the legacy mappings for the specified DMO.
    * 
    * @param    schema
    *           The schema to which the DMO belongs.
    * @param    dmoClass
    *           The MDO implementation class.
    */
   static void mapPermanentDMO(String schema, Class<?> dmoClass)
   {
      Class<?> dmoIface = DMOIndex.getDMOInterface(schema, dmoClass);
      
      // - mapping is kept by schema
      // - permamentDBs are kept in a global map
      synchronized (permanentDBs)
      {
         PermanentTableMapper mapper = locatePerm(schema);
         if (mapper == null)
         {
            mapper = new PermanentTableMapper(schema);
            permanentDBs.put(schema,  mapper);
         }
         
         mapper.mapClass(dmoIface, dmoClass);
      }
   }
   
   /**
    * Remove the mappings for the specified schema.
    * 
    * @param    schema
    *           The schema name.
    */
   static void removePermanentSchema(String schema)
   {
      synchronized (permanentDBs)
      {
         permanentDBs.remove(schema);
      }
   }
   
   /**
    * Remove all the temporary tables for this context.
    */
   static void removeDynamicTempTables()
   {
      TempTableMapper ttm = dynamicDB.get(false);
      if (ttm != null)
      {
         ttm.clear();
      }
   }
   
   /**
    * Add the legacy mappings for the given temporary table.
    * 
    * @param    table
    *           The {@link TempTable} resource.
    */
   static void mapTemporaryTable(TempTable table)
   {
      TempTableMapper mapper = locateTemp(table);
      
      synchronized (mapper)
      {
         mapper.mapClass(table);
      }
   }
   
   /**
    * Remove the mappings for the given temporary table.
    * 
    * @param    table
    *           The {@link TempTable} resource to be removed.
    */
   static void removeTemporaryTable(TempTable table)
   {
      if (table._dynamic())
      {
         // only remove the mapping if the table is dynamic; static table mappings are shared
         TempTableMapper mapper = dynamicDB.get(false);
         if (mapper != null)
         {
            mapper.removeClass(table);
         }
      }
   }

   /**
    * Compose a set of legacy field names which represents the intersection of the fields (by name
    * and type) of the two tables. These fields represent the possible, <em>natural</em> join
    * points between the tables.
    *
    * @param   buf1
    *          The record buffer which represents the first table.
    * @param   buf2
    *          The record buffer which represents the second table.
    *
    * @return  The set of legacy field names which are common between the two tables.
    */
   static Set<String> getCommonFields(RecordBuffer buf1, RecordBuffer buf2)
   {
      Set<String> res = new HashSet<>();

      Map<String, String> legacyToProp1 = getLegacyFieldNameMap(buf1).getLegacyField2Name();
      Map<String, String> legacyToProp2 = getLegacyFieldNameMap(buf2).getLegacyField2Name();
      Set<String> legacy1 = legacyToProp1.keySet();
      Set<String> legacy2 = legacyToProp2.keySet();


      Class iface1 = buf1.getDMOInterface();
      Map<String, Method> getters1 = PropertyHelper.legacyGettersByProperty(iface1);
      Map<String, Integer> extents1 = PropertyHelper.extentsByProperty(iface1);

      Class iface2 = buf2.getDMOInterface();
      Map<String, Method> getters2 = PropertyHelper.legacyGettersByProperty(iface2);
      Map<String, Integer> extents2 = PropertyHelper.extentsByProperty(iface2);

      for (String legacyField: legacy1)
      {
         if (!legacy2.contains(legacyField))
         {
            continue;
         }

         // Compare property types.
         Method getter1 = getters1.get(legacyToProp1.get(legacyField));
         ParmType parmType1 = ParmType.fromClass(getter1.getReturnType());

         Method getter2 = getters2.get(legacyToProp2.get(legacyField));
         ParmType parmType2 = ParmType.fromClass(getter2.getReturnType());

         if (parmType1 == null || parmType2 == null || parmType1 != parmType2)
         {
            continue;
         }

         // Shouldn't be an extent field.
         Integer extent1 = extents1.get(legacyField);
         Integer extent2 = extents2.get(legacyField);

         if (extent1 != null || extent2 != null)
         {
            continue;
         }

         res.add(legacyField);
      }

      return res;
   }

   /**
    * Given a table and a map of fields which this table class has in common (by name and type)
    * with another table, find the index which can be used to join this table with the other. The
    * index must be unique and must contain only fields which are in the common map. There should
    * be only one such index in the table, but this restriction is not enforced here; once a
    * suitable index is found, it is returned.
    *
    * @param   buffer
    *          The buffer which represents the table to be searched for a suitable index.
    * @param   commonFields
    *          Map of legacy fields which the given table has in common with another one.
    *
    * @return  The unique index which defines the join between this table and another, based upon
    *          fields described by <code>commonFields</code>, or <code>null</code> if no such
    *          index is found.
    */
   static P2JIndex findJoiningIndex(RecordBuffer buffer, Set<String> commonFields)
   throws PersistenceException
   {
      Iterator<P2JIndex> indexes = DMOIndex.databaseIndexes(buffer.getSchema(),
                                                            buffer.getDatabase(),
                                                            buffer.getDMOInterface());

      Map<String, String> columnToProp =
            DatabaseManager.getColumnToPropertyMap(buffer.getDatabase(),
                                                   buffer.getDMOImplementationClass(),
                                                   null,
                                                   false);
      Map<String, String> propToLegacy = getLegacyFieldNameMap(buffer).getName2LegacyField();

      indexLoop:
      while (indexes.hasNext())
      {
         P2JIndex index = indexes.next();
         if (!index.isUnique())
         {
            continue;
         }

         Iterator<P2JIndexComponent> components = index.components();
         while (components.hasNext())
         {
            P2JIndexComponent component = components.next();
            String propName = columnToProp.get(component.getName());
            if (!commonFields.contains(propToLegacy.get(propName)))
            {
               continue indexLoop;
            }
         }

         return index;
      }

      return null;
   }

   /**
    * Compose a set of all legacy field names defined for the given index.
    *
    * @param   buffer
    *          The buffer which represents the table to which the index belongs.
    * @param   index
    *          The index whose fields are to be included in the set.
    *
    * @return  Set of legacy field names defined for the given index.
    */
   static Set<String> getIndexFieldNames(RecordBuffer buffer, P2JIndex index)
   throws PersistenceException
   {
      Set<String> res = new HashSet<>();
      Map<String, String> columnToProp =
            DatabaseManager.getColumnToPropertyMap(buffer.getDatabase(),
                                                   buffer.getDMOImplementationClass(),
                                                   null,
                                                   false);
      Map<String, String> propToLegacy = getLegacyFieldNameMap(buffer).getName2LegacyField();
      Iterator<P2JIndexComponent> components = index.components();
      while (components.hasNext())
      {
         P2JIndexComponent component = components.next();
         String propName = columnToProp.get(component.getName());
         res.add(propToLegacy.get(propName));
      }

      return res;
   }

   /**
    * A helper method to return bi-directional mapping of (lower case) legacy field names to
    * their DMO property counterparts, optimized to gather information about all fields at once.
    * 
    * @param   buffer
    *          Record buffer for which the mapping is needed.
    * 
    * @return  Bi-directional mapping of legacy field names to property names for the given
    *          buffer.
    */
   static LegacyFieldNameMap getLegacyFieldNameMap(RecordBuffer buffer)
   {
      Map<String, String> legacyField2Name = new HashMap<>();
      Map<String, String> name2LegacyField = new HashMap<>();
      
      LegacyTableInfo tableInfo;
      
      if (buffer.isTemporary())
      {
         TempTable tt = buffer.getParentTable();
         TempTableMapper mapper = locateTemp(tt);
         
         synchronized (mapper)
         {
            Integer key = mapper.getLegacyKey(tt);
            tableInfo = mapper.p2j.get(key);
         }
      }
      else
      {
         synchronized (permanentDBs)
         {
            PermanentTableMapper mapper = locatePerm(buffer.getSchema());
            tableInfo = mapper.j2p.get(buffer.getDMOInterface());
         }
      }
      
      synchronized (tableInfo)
      {
         // it is assumed this method is only called at a safe time, after the mappings have been
         // initialized, otherwise we will get a NullPointerException here, which indicates a
         // programming error, rather than a runtime error
         for (LegacyFieldInfo fieldInfo : tableInfo.p2j.values())
         {
            String legacyName = fieldInfo.legacyName;
            String legacyLowName = legacyName.toLowerCase().intern();
            String propertyName = (fieldInfo.original.length() > 0
                                   ? fieldInfo.original
                                   : fieldInfo.javaName);
            legacyField2Name.put(legacyLowName, propertyName);
            name2LegacyField.put(propertyName, legacyName);
         }
      }
      
      return new LegacyFieldNameMap(legacyField2Name, name2LegacyField);
   }
   
   /**
    * Get the legacy field info for the property associated with the DMO referenced by the given
    * buffer.
    *
    * @param    buffer
    *           A buffer instance.
    * @param    property
    *           The DMO property name.
    *
    * @return   The legacy field info or <code>null</code> if there is no such field.
    */
   private static LegacyFieldInfo getLegacyFieldInfo(Buffer buffer, String property)
   {
      RecordBuffer buf = ((BufferImpl) buffer).buffer();
      if (buf.isTemporary())
      {
         return getLegacyFieldInfo((TempTable) buffer.tableHandle().get(), property);
      }
      else
      {
         return getLegacyFieldInfo(buf.getDMOInterface(), property);
      }
   }

   /**
    * Get information about the index associated with the DMO referenced by the given buffer,
    * as a string.
    *
    * @param    buffer
    *           A buffer instance which specifies the DMO class associated with the target index.
    * @param    indexNum
    *           0-based index number in the index definition order.
    *
    * @return   a string describing the target index or <code>null</code> if there is no index
    *           with the given number. Format of the string is described in
    *           {@link Buffer#indexInformation(int)}.
    */
   public static LegacyIndexInfo getLegacyIndexInfo(Buffer buffer, int indexNum)
   {
      RecordBuffer buf = ((BufferImpl) buffer).buffer();
      if (buf.isTemporary())
      {
         return getLegacyIndexInfo((TempTable) buffer.tableHandle().get(), indexNum);
      }
      else
      {
         return getLegacyIndexInfo(buf.getDMOInterface(), indexNum);
      }
   }

   /**
    * Get information about the index associated with the given temporary table, as a string.
    *
    * @param    tt
    *           Temporary table associated with the target index.
    * @param    indexNum
    *           0-based index number in the index definition order.
    *
    * @return   a string describing the target index or <code>null</code> if there is no index
    *           with the given number. Format of the string is described in
    *           {@link Buffer#indexInformation(int)}.
    */
   public static LegacyIndexInfo getLegacyIndexInfo(TempTable tt, int indexNum)
   {
      TempTableMapper mapper = locateTemp(tt);

      synchronized (mapper)
      {
         return mapper.legacyIndexInfo(tt, indexNum);
      }
   }

   /**
    * Get information about the index associated with the given DMO class, as a string.
    *
    * @param    dmoIface
    *           DMO class associated with the target index.
    * @param    indexNum
    *           0-based index number in the index definition order.
    *
    * @return   a string describing the target index or <code>null</code> if there is no index
    *           with the given number. Format of the string is described in
    *           {@link Buffer#indexInformation(int)}.
    */
   public static LegacyIndexInfo getLegacyIndexInfo(Class<?> dmoIface, int indexNum)
   {
      String schema = DatabaseManager.getSchemaByInterface(dmoIface);
      if (DatabaseManager.TEMP_TABLE_SCHEMA.equals(schema))
      {
         return null;
      }

      synchronized (permanentDBs)
      {
         PermanentTableMapper mapper = locatePerm(schema);

         return mapper.legacyIndexInfo(dmoIface, indexNum);
      }
   }

   /**
    * Locate the appropriate {@link TempTableMapper} for the given temp-table. If it is statically
    * defined, return the shared, static version, else return the context-local version for
    * dynamically defined temp-tables.
    * 
    * @param   tt
    *          Temp-table object.
    * 
    * @return  Appropriate table mapper as described above.
    */
   private static TempTableMapper locateTemp(TempTable tt)
   {
      return tt._dynamic() ? dynamicDB.get() : staticTempDB;
   }
   
   /**
    * Locate the {@link PermanentTableMapper} instance for the given schema.
    * 
    * @param    schema
    *           The schema name.
    *           
    * @return   See above.
    */
   private static PermanentTableMapper locatePerm(String schema)
   {
      synchronized (permanentDBs)
      {
         return permanentDBs.get(schema);
      }
   }
   
   /**
    * Get the legacy table name associated with the given DMO implementation class.
    * 
    * @param    dmoIface
    *           The DMO interface.
    *           
    * @return   The legacy table name.
    */
   protected String legacyName(Class<?> dmoIface)
   {
      LegacyTableInfo tableInfo = j2p().get(dmoIface);
      
      return tableInfo.legacyName;
   }

   /**
    * Determine if the table has table-level delete validation.
    *
    * @param    dmoIface
    *           The DMO interface associated with the given table.
    *
    * @return   <code>true</code> if the table has table-level delete validation.
    */
   protected boolean validation(Class<?> dmoIface)
   {
      LegacyTableInfo tableInfo = j2p().get(dmoIface);

      return tableInfo.validation;
   }
   
   /**
    * Get the legacy field name for given property defined by the specified DMO.
    * 
    * @param    dmoIface
    *           The DMO interface.
    * @param    property
    *           The DMO property name.
    *           
    * @return   The legacy field name or <code>null</code> if there is no such field.
    */
   protected String legacyFieldName(Class<?> dmoIface, String property)
   {
      LegacyTableInfo tableInfo = j2p().get(dmoIface);
      LegacyFieldInfo fieldInfo = tableInfo.j2p.get(property);
      
      return (fieldInfo == null ? null : fieldInfo.legacyName);
   }
   
   /**
    * Get the DMO property name for a field index belonging to the specified legacy table.
    *
    * @param    dmoIface
    *           The DMO interface for a permanent DMO.
    * @param    fieldId
    *           The field index.
    *
    * @return   The DMO property name associated with the legacy field or <code>null</code> if
    *           there is no such field.
    */
   protected String propertyName(Class<?> dmoIface, int fieldId)
   {
      LegacyTableInfo tableInfo = j2p().get(dmoIface);
      LegacyFieldInfo fieldInfo = tableInfo.ip2j.get(fieldId);
      
      return (fieldInfo == null ? null : fieldInfo.javaName);
   }
   
   /**
    * Get legacy field info for a field from the specified legacy table.
    *
    * @param    dmoIface
    *           The DMO interface for a permanent DMO.
    * @param    property
    *           Property name.
    *
    * @return   legacy field info for a field from the specified legacy table or
    *           <code>null</code> if there is no such field.
    */
   protected LegacyFieldInfo legacyFieldInfo(Class<?> dmoIface, String property)
   {
      LegacyTableInfo tableInfo = j2p().get(dmoIface);
      
      return tableInfo.j2p.get(property);
   }

   /**
    * Get information about the index associated with the given DMO class.
    *
    * @param    dmoIface
    *           DMO class associated with the target index.
    * @param    indexNum
    *           0-based index number in the index definition order.
    *
    * @return   an object describing the target index or <code>null</code> if there is no index
    *           with the given number.
    */
   protected LegacyIndexInfo legacyIndexInfo(Class<?> dmoIface, int indexNum)
   {
      LegacyTableInfo tableInfo = j2p().get(dmoIface);

      return indexNum < 0 || indexNum >= tableInfo.idx.size() ?
               null :
               tableInfo.idx.get(indexNum);
   }
   
   /**
    * Get the legacy index name of converted index, which is defined by the backing table for the 
    * specified DMO implementation class.
    * 
    * @param    dmoIface
    *           The DMO implementation class.
    * @param    idx
    *           The converted index name.
    *           
    * @return   The legacy index name or <code>null</code> if there is no such index defined.
    */
   protected String legacyIndexName(Class<?> dmoIface, String idx)
   {
      LegacyTableInfo tableInfo = j2p().get(dmoIface);
      LegacyIndexInfo indexInfo = tableInfo.idxj2p.get(idx);
      
      return (indexInfo == null ? null : indexInfo.legacyName);
   }
   
   /**
    * Remove the mappings for the specified DMO implementation class.
    * 
    * @param    dmoIface
    *           The DMO interface.
    */
   protected void removeClass(Class<?> dmoIface)
   {
      LegacyTableInfo info = j2p().get(dmoIface);
      
      if (info == null)
      {
         // ignore if not found
         return;
      }
      
      j2p().remove(info.dmoIface);
      p2j().remove(info.legacyName);
   }
   
   /**
    * Get the property name for the specified field index. The legacy table info is determined
    * using the specified <code>legacyKey</code>.
    *
    * @param    legacyKey
    *           The key used to identify the {@link LegacyTableInfo}.
    * @param    fieldId
    *           The index of the field in the buffer's sequence.
    * @param    index
    *           <code>null</code> for usual properties and zero-based index of this property for
    *           denormalized field with extent.
    *
    * @return   The DMO property name associated with the field index or <code>null</code> if
    *           there is no such field.
    */
   protected String getPropertyName(K legacyKey, int fieldId, Integer index)
   {
      LegacyTableInfo tableInfo = p2j().get(legacyKey);
      LegacyFieldInfo fieldInfo = tableInfo.ip2j.get(fieldId);
      
      return getPropertyName(tableInfo, fieldInfo, index);
   }
   
   /**
    * Get the number of fields in the table mapped to the given key.
    * 
    * @param   legacyKey
    *          Key to which the target table is matched.
    * 
    * @return  Number of fields in the table.
    */
   protected int getNumFields(K legacyKey)
   {
      LegacyTableInfo tableInfo = p2j().get(legacyKey);
      
      return tableInfo.ip2j.size();
   }
   
   /**
    * Get the property name for the specified table info, field info and null-based index.
    *
    * @param    tableInfo
    *           The key used to identify the {@link LegacyTableInfo}.
    * @param    fieldInfo
    *           The index of the field in the buffer's sequence.
    * @param    index
    *           <code>null</code> for usual properties and zero-based index of this property for
    *           denormalized field with extent.
    *
    * @return   property name for the specified table info, field info and index.
    */
   private static String getPropertyName(LegacyTableInfo tableInfo,
                                         LegacyFieldInfo fieldInfo,
                                         Integer index)
   {
      if (fieldInfo == null)
      {
         return null;
      }
      
      String property;
      
      if (index == null)
      {
         // it is possible to get here for an extent field with index == null, so check for
         // an original name first, but return the java name if there is no original name
         property = fieldInfo.original.length() > 0 ? fieldInfo.original : fieldInfo.javaName;
      }
      else
      {
         property = tableInfo.denorm2j.get(fieldInfo.original).get(index).javaName;
      }
      
      return property;
   }
   
   /**
    * Get the converted index name for the specified legacy index. The legacy table info is 
    * determined using the specified <code>legacyKey</code>.
    * 
    * @param    legacyKey
    *           The key used to identify the {@link LegacyTableInfo}.
    * @param    idx
    *           The legacy index name.
    *           
    * @return   The converted index name or <code>null</code> if there is no such index.
    */
   protected String getIndexName(K legacyKey, String idx)
   {
      LegacyTableInfo tableInfo = p2j().get(legacyKey);
      LegacyIndexInfo indexInfo = tableInfo.idxp2j.get(idx.toLowerCase());
      
      return (indexInfo == null ? null : indexInfo.convertedName);
   }
   
   /**
    * Clear all the mappings.
    */
   protected void clear()
   {
      j2p().clear();
      p2j().clear();
   }
   
   /**
    * Get new field name for denormalized legacy field name and index.
    *
    * @param    dmoIface
    *           The DMO implementation class for a permanent DMO.
    * @param    property
    *           The DMO property name.
    * @param    index
    *           zero-based index of this property in denormalization hints.
    *
    * @return   New field name.
    */
   public String denormalizedProperty(Class<?> dmoIface, String property, Integer index)
   {
      if (index == null)
      {
         return property;
      }
      LegacyTableInfo tableInfo = j2p().get(dmoIface);
      if (tableInfo == null)
      {
         return null;
      }
      Map<Integer, LegacyFieldInfo> map = tableInfo.denorm2j.get(property);
      if (map == null)
      {
         return null;
      }
      LegacyFieldInfo legacyFieldInfo = map.get(index + 1);
      if (legacyFieldInfo == null)
      {
         return null;
      }
      else
      {
         return legacyFieldInfo.javaName;
      }
   }
   
   /**
    * Get the property name for the specified field. The legacy table info is determined using
    * the specified <code>legacyKey</code>.
    * 
    * @param    legacyKey
    *           The key used to identify the {@link LegacyTableInfo}.
    * @param    field
    *           The legacy field name.
    * @param    index
    *           <code>null</code> for usual properties and zero-based index of this property for
    *           denormalized field with extent.
    *
    * @return   The DMO property name associated with the legacy field or <code>null</code> if
    *           there is no such field.
    */
   protected String getPropertyName(K legacyKey, String field, Integer index)
   {
      String normalized4glField = TextOps.rightTrim(field).toStringMessage().toLowerCase();
      LegacyTableInfo tableInfo = p2j().get(legacyKey);
      LegacyFieldInfo fieldInfo = tableInfo.p2j.get(normalized4glField);
      
      return getPropertyName(tableInfo, fieldInfo, index);
   }
   
   /**
    * Get the mapping of legacy table keys to {@link LegacyTableInfo} instances. The type of key
    * is implementation specific.
    * 
    * @return  Map of legacy table keys to table information objects.
    */
   protected abstract Map<K, LegacyTableInfo> p2j();
   
   /**
    * Get the mapping of DMO implementation classes to {@link LegacyTableInfo} instances.
    * 
    * @return  Map of DMO classes to table information objects.
    */
   protected abstract Map<Class<?>, LegacyTableInfo> j2p();
   
   /**
    * Mapper for permanent tables.
    */
   private static class PermanentTableMapper
   extends TableMapper<String>
   {
      /** Mapping of DMO interfaces to {@link LegacyTableInfo} instances */
      private final Map<Class<?>, LegacyTableInfo> j2p = new HashMap<>();
      
      /** Mapping of the legacy table names to {@link LegacyTableInfo} instances. */
      private final Map<String, LegacyTableInfo> p2j = new HashMap<>();
      
      /**
       * Default c'tor.
       * 
       * @param    schema
       *           The schema name.
       */
      private PermanentTableMapper(String schema)
      {
         super(schema);
      }
      
      /**
       * Register the mappings for the specified permanent DMO.
       * 
       * @param   dmoIface
       *          The DMO interface.
       * @param   dmoClass
       *          The DMO implementation class.
       * 
       * @throws  NullPointerException
       *          if the DMO class does not have a {@link LegacyTable} annotation.
       */
      protected void mapClass(Class<?> dmoIface, Class<?> dmoClass)
      {
         if (!dmoClass.isAnnotationPresent(LegacyTable.class))
         {
            final String msg = "DMO class %s must have a LegacyTable annotation!";
            
            throw new NullPointerException(String.format(msg, dmoClass.getName()));
         }
         
         LegacyTable table = dmoClass.getAnnotation(LegacyTable.class);
         
         mapClass(table.name().toLowerCase(), dmoIface, dmoClass);
      }
      
      /**
       * Map the specified DMO implementation class using the specified <code>legacyKey</code>.
       * <p>
       * Before calling this method, make sure the passed DMO has a {@link LegacyTable}
       * annotation.
       * 
       * @param   legacyKey
       *          The legacy key to identify this DMO implementation class.
       * @param   dmoIface
       *          The DMO interface.
       * @param   dmoClass
       *          The DMO implementation class.
       */
      protected void mapClass(String legacyKey, Class<?> dmoIface, Class<?> dmoClass)
      {
         if (j2p.containsKey(dmoIface))
         {
            // already mapped
            return;
         }
         
         LegacyTableInfo tableInfo = new LegacyTableInfo(dmoIface, dmoClass);
         
         p2j.put(legacyKey, tableInfo);
         j2p.put(dmoIface, tableInfo);
      }
      
      /**
       * Get the DMO class for the legacy table identified with the specified
       * <code>legacyKey</code>.
       * 
       * @param    legacyKey
       *           The key used to identify the {@link LegacyTableInfo}.
       * 
       * @return   The DMO implementation class or <code>null</code> if there is no class
       *           corresponding the given key.
       */
      protected Class<?> getDMOClass(String legacyKey)
      {
         LegacyTableInfo tableInfo = p2j.get(legacyKey);
         
         return tableInfo == null ? null : tableInfo.dmoClass;
      }
      
      /**
       * Get the mapping of legacy table names to {@link LegacyTableInfo} instances.
       * 
       * @return  Map of legacy table names to table information objects.
       */
      protected Map<String, LegacyTableInfo> p2j()
      {
         return p2j;
      }
      
      /**
       * Get the mapping of DMO implementation classes to {@link LegacyTableInfo} instances.
       * 
       * @return  Map of DMO classes to table information objects.
       */
      protected Map<Class<?>, LegacyTableInfo> j2p()
      {
         return j2p;
      }
   }
   
   /**
    * Mapper for temporary tables.
    */
   private static abstract class TempTableMapper
   extends TableMapper<Integer>
   {
      /** Mapping of DMO implementation classes to {@link LegacyTableInfo} instances */
      private final Map<Class<?>, LegacyTableInfo> j2p = new HashMap<>();
      
      /** Mapping of the legacy table IDs to their {@link LegacyTableInfo} instances */
      private final Map<Integer, LegacyTableInfo> p2j = new HashMap<>();
      
      /**
       * Default c'tor.
       */
      protected TempTableMapper()
      {
         super(DatabaseManager.TEMP_TABLE_SCHEMA);
      }
      
      /**
       * Register the mappings for the specified temp table.
       * 
       * @param    tt
       *           The temp-table which needs to be mapped.
       *           
       * @throws   NullPointerException
       *           If the associated DMO class does not have a {@link LegacyTable} annotation.
       */
      protected void mapClass(TempTable tt)
      {
         Class<?> dmoClass = tt.getDMOClass();
         if (!dmoClass.isAnnotationPresent(LegacyTable.class))
         {
            final String msg = "DMO class %s must have a LegacyTable annotation!"; 
            throw new NullPointerException(String.format(msg, dmoClass.getName()));
         }
         
         Integer key = getLegacyKey(tt);
         
         if (p2j.containsKey(key))
         {
            // already mapped
            return;
         }
         
         // multiple keys may map to the same DMO implementation class (since these are shared);
         // reuse an existing instance if possible
         Class<?> dmoIface = tt.getDMOInterface();
         LegacyTableInfo tableInfo = j2p.get(dmoIface);
         
         if (tableInfo == null)
         {
            tableInfo = new LegacyTableInfo(dmoIface, dmoClass);
            j2p.put(dmoIface, tableInfo);
         }
         
         p2j.put(key, tableInfo);
      }
      
      /**
       * Remove the mappings for the specified temp table.
       *
       * @param    tt
       *           Temp table object.
       */
      protected void removeClass(TempTable tt)
      {
         Integer key = getLegacyKey(tt);
         p2j.remove(key);
      }
      
      /**
       * Get the legacy table name associated with the given temp table.
       *
       * @param    tt
       *           Temp table object.
       *
       * @return   The legacy table name.
       */
      protected String legacyName(TempTable tt)
      {
         Integer key = getLegacyKey(tt);
         LegacyTableInfo info = p2j.get(key);
         
         return info.legacyName;
      }
      
      /**
       * Get the mapping of temp-table IDs to {@link LegacyTableInfo} instances.
       * 
       * @return  Map of temp-table IDs to table information objects.
       */
      protected Map<Integer, LegacyTableInfo> p2j()
      {
         return p2j;
      }
      
      /**
       * Get the mapping of DMO implementation classes to {@link LegacyTableInfo} instances.
       * 
       * @return  Map of DMO classes to table information objects.
       */
      protected Map<Class<?>, LegacyTableInfo> j2p()
      {
         return j2p;
      }
      
      /**
       * Get the legacy field name for given property defined by the specified temp table.
       *
       * @param    tt
       *           Temp table object.
       * @param    property
       *           The DMO property name.
       *
       * @return   The legacy field name or <code>null</code> if there is no such field.
       */
      private LegacyFieldInfo legacyFieldInfo(TempTable tt, String property)
      {
         Integer key = getLegacyKey(tt);
         LegacyTableInfo tableInfo = p2j.get(key);
         
         return tableInfo.j2p.get(property);
      }

      /**
       * Get information about the index associated with the given temporary table.
       *
       * @param    tt
       *           Temp table object.
       * @param    indexNum
       *           0-based index number in the index definition order.
       *
       * @return   an object describing the target index or <code>null</code> if there is no index
       *           with the given number.
       */
      private LegacyIndexInfo legacyIndexInfo(TempTable tt, int indexNum)
      {
         Integer key = getLegacyKey(tt);
         LegacyTableInfo tableInfo = p2j.get(key);

         return indexNum < 0 || indexNum >= tableInfo.idx.size() ?
                  null :
                  tableInfo.idx.get(indexNum);
      }
      
      /**
       * Get the legacy index name of converted index, which is defined by the specified backing
       * temp table.
       *
       * @param    tt
       *           Temp table object.
       * @param    idx
       *           The converted index name.
       *
       * @return   The legacy index name or <code>null</code> if there is no such index defined.
       */
      protected String legacyIndexName(TempTable tt, String idx)
      {
         Integer key = getLegacyKey(tt);
         LegacyTableInfo tableInfo = p2j.get(key);
         LegacyIndexInfo indexInfo = tableInfo.idxj2p.get(idx);
         
         return indexInfo.legacyName;
      }
      
      /**
       * Get legacy key for the specified temp table.
       *
       * @param  tt
       *         Temp table object.
       *
       * @return legacy key for the specified temp table.
       */
      protected abstract Integer getLegacyKey(TempTable tt);
   }
   
   private static class StaticTempTableMapper
   extends TempTableMapper
   {
      /** Map of temp-table objects to unique keys */
      private final Map<Class<?>, Integer> keys = new HashMap<>();
      
      /** Next available key */
      private int nextKey = 0;
      
      /**
       * Get legacy key for the specified temp table.
       *
       * @param  tt
       *         Temp table object.
       *
       * @return legacy key for the specified temp table.
       */
      protected Integer getLegacyKey(TempTable tt)
      {
         Class<?> dmoIface = tt.getDMOInterface();
         Integer key = keys.get(dmoIface);
         
         if (key == null)
         {
            key = nextKey++;
            keys.put(dmoIface, key);
         }
         
         return key;
      }
   }
   
   private static class DynamicTempTableMapper
   extends TempTableMapper
   {
      /**
       * Get legacy key for the specified temp table.
       *
       * @param  tt
       *         Temp table object.
       *
       * @return legacy key for the specified temp table.
       */
      protected Integer getLegacyKey(TempTable tt)
      {
         return tt.getUniqueID().toJavaIntegerType();
      }
   }
   
   /**
    * A container with the legacy info for a table.
    */
   private static class LegacyTableInfo
   {
      /** Field info kept by legacy name in progress. */ 
      private final Map<String, LegacyFieldInfo> p2j = new HashMap<>();
      
      /**
       * Field info kept by legacy buffer field sequence index in progress.
       * This is needed at runtime by the buffer-field attribute.
       */
      private final Map<Integer, LegacyFieldInfo> ip2j = new HashMap<>();
      
      /** Field info kept by associated DMO property name. */
      private final Map<String, LegacyFieldInfo> j2p = new HashMap<>();
      
      /** Field info about extent fields kept by associated DMO property name. */
      private final Map<String, LegacyFieldInfo> extents = new HashMap<>();

      /** Index info kept by legacy index name. */
      private final Map<String, LegacyIndexInfo> idxp2j = new HashMap<>();
      
      /** Index info kept by converted index name. */
      private final Map<String, LegacyIndexInfo> idxj2p = new HashMap<>();

      /** Index info kept by index number. */
      private final List<LegacyIndexInfo> idx = new ArrayList<>();
      
      /** The legacy table name. */
      private final String legacyName;

      /** Determines if the table has table-level delete validation. */
      private final boolean validation;
      
      /** Indexed map of new field names for denormalized legacy field name. */
      private final Map<String, Map<Integer, LegacyFieldInfo>> denorm2j = new HashMap<>();
      
      /** The DMO interface. */
      private final Class<?> dmoIface;
      
      /** The DMO implementation class. */
      private final Class<?> dmoClass;

      /**
       * Initialize the legacy info for a table by reading the {@link LegacyTable} and 
       * {@link LegacyField} annotation from the DMO implementation class.
       * 
       * @param    dmoClass
       *           The DMO class from where the annotations will be read.
       * 
       * @see      #loadFields(Class)
       */
      public LegacyTableInfo(Class<?> dmoIface, Class<?> dmoClass)
      {
         LegacyTable ltable = dmoClass.getAnnotation(LegacyTable.class);
         
         this.legacyName = ltable.name();
         this.dmoIface = dmoIface;
         this.dmoClass = dmoClass;
         this.validation = ltable.hasValidation();
         
         loadFields(dmoClass);
         
         Class<?>[] classMembers = dmoClass.getDeclaredClasses();
         for (Class<?> cls : classMembers)
         {
            loadFields(cls);
         }
         
         LegacyIndexes lindexes = dmoClass.getAnnotation(LegacyIndexes.class);
         if (lindexes != null)
         {
            LegacyIndexInfo firstIndex = null;
            boolean hasPrimaryIndex = false;

            for (LegacyIndex lidx : lindexes.value())
            {
               LegacyIndexInfo idxInfo = new LegacyIndexInfo(lidx);
               if (firstIndex == null)
               {
                  firstIndex = idxInfo;
               }

               if (idxInfo.effectivePrimary)
               {
                  hasPrimaryIndex = true;
               }
               
               this.idxp2j.put(idxInfo.legacyName.toLowerCase(), idxInfo);
               this.idxj2p.put(idxInfo.convertedName, idxInfo);
               this.idx.add(idxInfo);
            }

            // If there is no index with "primary" annotation, make the first index primary.
            if (!hasPrimaryIndex && firstIndex != null)
            {
               firstIndex.effectivePrimary = true;
            }
         }
      }
      
      /**
       * Load all fields from the specified class and add their mappings.
       * 
       * @param    cls
       *           The DMO class or a composite class with the extent fields.
       * 
       * @throws   NullPointerException
       *           If a DMO property does not have a {@link LegacyField} annotation.
       */
      private void loadFields(Class<?> cls)
      {
         Object instance = null;
         if (cls.getClass().getAnnotation(LegacyTable.class) != null)
         {
            try 
            {
               instance = cls.newInstance();
            } 
            catch (InstantiationException | IllegalAccessException e) 
            {
               throw new RuntimeException("Failed to instantiate " + cls.getName(), e);
            } 
         }

         Field[] fields = cls.getDeclaredFields();
         for (Field field : fields)
         {
            field.setAccessible(true);
            if (!BaseDataType.class.isAssignableFrom(field.getType()))
            {
               // ignore P2J-specific DMO properties. all legacy DMO properties have BaseDataType
               // subclass as their type
               continue;
            }
            BaseDataType initialValue;
            try 
            {
               initialValue = instance == null ? null : (BaseDataType)field.get(instance);
            } 
            catch (IllegalArgumentException | IllegalAccessException e) 
            {
               throw new RuntimeException("Failed to read " + field.getName(), e);
            }

            LegacyField lfield = field.getAnnotation(LegacyField.class);
            
            if (lfield == null)
            {
               final String msg = "Property %s for class %s must have a LegacyField annotation!"; 
               throw new NullPointerException(
                  String.format(msg, field.getName(), dmoClass.getName()));
            }
            
            String columnLabel = lfield.columnLabel();
            String legacyName = lfield.name();
            String javaName = field.getName().intern();
            String format = lfield.format();
            String initial = lfield.initial();
            String label = lfield.label();
            boolean mandatory = lfield.mandatory();
            String validateMessage = lfield.validateMessage();
            String validateExpression = lfield.validateExpression();
            int extentIndex = lfield.index();
            String original = lfield.original();
            
            Integer fieldId = lfield.fieldId();
            
            String pnameLowCase = legacyName.toLowerCase().intern();
            LegacyFieldInfo fieldInfo = p2j.get(pnameLowCase);
            boolean firstSeen = fieldInfo == null; 
            if (firstSeen || fieldInfo.fieldId != fieldId)
            {
               fieldInfo = new LegacyFieldInfo(columnLabel,
                                               fieldId,
                                               format,
                                               initial,
                                               javaName,
                                               label,
                                               legacyName,
                                               mandatory,
                                               original,
                                               validateMessage,
                                               validateExpression,
                                               initialValue);
            }

            if (firstSeen) // for extent field store the first representative info only 
            {
               p2j.put(pnameLowCase, fieldInfo);
            }
            j2p.put(javaName, fieldInfo);
            ip2j.put(fieldId, fieldInfo);
            
            if (original.length() > 0)
            {
               Map<Integer, LegacyFieldInfo> javaNames = denorm2j.get(original);
               if (javaNames == null)
               {
                  javaNames = new HashMap<>();
                  denorm2j.put(original, javaNames);
               }
               javaNames.put(extentIndex, fieldInfo);
               if (extents.get(original) == null) 
               {
                  LegacyFieldInfo extentInfo = new LegacyFieldInfo(columnLabel,
                                                                  fieldId,
                                                                  format,
                                                                  initial,
                                                                  original,
                                                                  label,
                                                                  legacyName,
                                                                  mandatory,
                                                                  "",
                                                                  validateMessage,
                                                                  validateExpression,
                                                                  initialValue);
                  extents.put(original, extentInfo);
               }
            }
         }
      }
   }

   /**
    * A container with a index information, holding index attributes and components.
    */
   private static class LegacyIndexInfo
   {
      /** The legacy name of this index. */
      private final String legacyName;
      
      /** The converted name of this index. */
      private final String convertedName;

      /**
       * Determine if the index is a primary index: the one with the "primary" annotation or, if
       * there is no such one, the first in the definition order.
       */
      private boolean effectivePrimary;

      /** Determine if the index is an unique index. */
      private final boolean unique;

      /** Determine if the index is a word-index index. */
      private final boolean word;

      /** Index components (fields with sorting directions). */
      private final LegacyIndexComponentInfo[] components;

      /**
       * Create a new index information container, holding index attributes and components.
       *
       * @param index
       *        Source index annotation generated at conversion stage.
       */
      LegacyIndexInfo(LegacyIndex index)
      {
         legacyName = index.name();
         convertedName = index.converted();
         effectivePrimary = index.primary();
         unique = index.unique();
         word = index.word();

         LegacyIndexComponent[] components = index.components();
         int size = components.length;
         this.components = new LegacyIndexComponentInfo[size];
         Boolean descending = null;
         for (int i = 0; i < size; i++)
         {
            this.components[i] = new LegacyIndexComponentInfo(components[i]);
            if (descending == null)
            {
               descending = this.components[i].effectiveDescending;
            }
         }
      }

      /**
       * Returns index information as a string. Format of this string is described in
       * {@link Buffer#indexInformation(int)}
       *
       * @return see above.
       */
      public String indexInformation()
      {
         StringBuilder sb = new StringBuilder(legacyName).append(",");
         sb.append(unique ? "1" : "0").append(",");
         sb.append(effectivePrimary ? "1" : "0").append(",");
         sb.append(word ? "1" : "0");

         for (LegacyIndexComponentInfo component : components)
         {
            sb.append(",").append(component.fieldLegacyName);
            sb.append(",").append(component.effectiveDescending ? "1" : "0");
         }

         return sb.toString();
      }
   }

   /**
    * A container with an index component information.
    */
   private static class LegacyIndexComponentInfo
   {
      /** The legacy name of the field participating in the index. */
      private final String fieldLegacyName;

      /** Determines if this component has descending sorting. */
      private boolean effectiveDescending;

      /**
       * Create a new index component information container, holding field name and sorting
       * direction.
       *
       * @param component
       *        Source index component annotation generated at conversion stage.
       *
       */
      LegacyIndexComponentInfo(LegacyIndexComponent component)
      {
         fieldLegacyName = component.fieldName();
         effectiveDescending = component.descending();
      }
   }

   /**
    * A container with a field's legacy and DMO property name and field attributes.
    */
   private static class LegacyFieldInfo
   {
      /** COLUMN-LABEL attribute of this field. */
      private String columnLabel;
      
      /**
       * The id of this field in the sequence of the fields of the buffer. It is needed by the
       * dynamic <code>bufferField(int)</code> attribute of the <code>BufferImpl</code> for
       * emulating the <code>BUFFER-FIELD(integer)</code> buffer handle attribute.
       */
      private final Integer fieldId;
      
      /** FORMAT attribute of this field. */
      private String format;
      
      /** INITIAL attribute of this field. */
      private final String initial;
      
      /** The associated DMO property name. */
      private final String javaName;
      
      /** LABEL attribute of this field. */
      private String label;
      
      /** The legacy name of this field. */
      private final String legacyName;
      
      /** LITERAL-QUESTION attribute of this field. */
      private boolean literalQuestion = false;
      
      /** MANDATORY attribute of this field. */
      public final boolean mandatory;
      
      /** Original property name (not exposed in DMO) for denormalized field with extent. */
      public final String original;
      
      /** VALIDATE-MESSAGE attribute of this field. */
      private String validateMessage;
      
      /** VALIDATE-EXPRESSION attribute of this field. */
      private String validateExpression;
      
      /** the initial value for this field */
      private final BaseDataType initialValue;
      
      /**
       * Create a new field information container, holding the legacy and converted names and
       * field attributes.
       *
       * @param    columnLabel
       *           COLUMN-LABEL attribute.
       * @param    fieldId
       *           The id of field in the sequence of the fields of the buffer.
       * @param    format
       *           FORMAT attribute.
       * @param    initial
       *           INITIAL attribute.
       * @param    javaName
       *           The field's associated DMO property name.
       * @param    label
       *           LABEL attribute.
       * @param    legacyName
       *           The field's legacy name.
       * @param    mandatory
       *           MANDATORY attribute.
       * @param    original
       *           Original property name (actually not exposed in DMO) for denormalized field
       *           with extent.
       * @param    validateMessage
       *           VALIDATE-MESSAGE attribute.
       * @param    validateExpression
       *           VALIDATE-EXPRESSION attribute.
       */
      public LegacyFieldInfo(String columnLabel,
                             Integer fieldId,
                             String format,
                             String initial,
                             String javaName,
                             String label,
                             String legacyName,
                             boolean mandatory,
                             String original,
                             String validateMessage,
                             String validateExpression,
                             BaseDataType initialValue)
      {
         this.columnLabel = columnLabel.intern();
         this.fieldId = fieldId;
         this.format = format.intern();
         this.initial = initial.intern();
         this.javaName = javaName.intern();
         this.label = label.intern();
         this.legacyName = legacyName.intern();
         this.mandatory = mandatory;
         this.original = original.intern();
         this.validateMessage = validateMessage;
         this.validateExpression = validateExpression;
         this.initialValue = initialValue;
      }
   }
}