/*
** Module   : P2JPostgreSQLDialect.java
** Abstract : Implementation of P2JDialect for a PostgreSQL backend
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20060929   @30034 Created initial version. Adds P2JDialect
**                           method implementations to extend Hibernate's
**                           PostgreSQLDialect.
** 002 ECF 20070514   @33483 Added extractColumnName(). Extracts a root
**                           column name from its native representation in
**                           a particular dialect.
** 003 ECF 20070923   @35221 Override superclass' VARCHAR mapping to point
**                           to PG's 'text' data type. This is desirable
**                           because we always want to use 'text' instead
**                           of 'varchar(n)' for string data on the
**                           backend.
** 004 ECF 20071002   @35334 Implemented new P2JDialect methods. Added
**                           createContext() and postTransaction() methods
**                           to enable temp table periodic vacuums.
** 005 ECF 20080311   @37518 Implemented methods for computed columns and
**                           user defined function overloading, which are
**                           not needed by this dialect but required by
**                           the P2JDialect interface.
** 006 ECF 20080430   @38179 Fix to permit schema generation. Defer lookup
**                           of vacuumInterval in directory until first
**                           use. Performing the directory bind during
**                           class initialization fails when generating
**                           schema via Ant task.
** 007 ECF 20080611   @38705 Added isCaseInsensitiveColumn(). Provides a
**                           rough test for case-insensitivity, based only
**                           on an expression which contains a column
**                           name, such as might be extracted from index
**                           metadata.
** 008 ECF 20080805   @39316 Added getCreateSequenceString(). Generates
**                           DDL to create a sequence starting at a
**                           particular value.
** 009 CA  20080815   @39444 Added isConnectionError method - it checks if
**                           a given exception represents a connection 
**                           error.
** 010 ECF 20080930   @40001 Added isQueryRangeParameterInlined().
**                           Indicates whether query substitution
**                           parameters within range checks should be
**                           inlined.
** 011 ECF 20081015   @40096 Added formatDate(). Formats a date into a
**                           string suitable for use in an SQL statement.
** 012 ECF 20090310   @41456 Fixed performance problem with variable limit
**                           clause in a SELECT statement. In cases of
**                           simple and very small (2 or less) record
**                           limits (with no offset) in a SELECT, we now
**                           inline the number directly in the SQL. In the
**                           default implementation, Hibernate would use a
**                           bound parameter, which could result in a poor
**                           query plan (sequential table scan instead of
**                           an index scan).
** 013 GES 20090424   @41944 Import change.
** 014 ECF 20090705   @43225 Appended 'analyze' to 'vacuum' command.
** 015 SVL 20090810   @43575 Added requiresExplicitCastInsideTernary().
**                           Determines whether substitution parameters should
**                           be explicitly casted to their datatypes inside
**                           "then ? else ?" statement.
** 016 GES 20100701          Added a type name override to allow honoring of
**                           precision and scale in numeric types.
** 017 CA  20101026          Added methods to generate the DROP/CREATE INDEX
**                           statements in a dialect specific form. Also, 
**                           added method to generate the ALTER TABLE statement
**                           which sets the formula for the computed columns.
**                           Added methods to prepare/process string metadata 
**                           query parameters and results, to get the database
**                           prepare DDL statements and collation statement.
** 018 CA  20101202          Added method to build the URL for a remote 
**                           connection.
** 019 SVL 20110622          Added usesSingleBackslash.
** 020 OM  20121120          Implemented DDL/SQL generation for sequences.
** 021 SVL 20130331          Upgraded to Hibernate 4.
** 022 ECF 20130510          Modified sequence-related methods to not use quoted names. Replace
**                           deprecated superclass.
** 023 OM  20130605          Added if exists clause to drop index statement.
** 024 OM  20130712          Overrided default (invalid) min/max values for sequences.
**                           Implemented createSequenceHandler().
** 025 VMN 20131005          Added injectComputedColumns. Replaced $p to 50 in
*                            registerColumnType for Types.NUMERIC.
** 026 OM  20130925          Replaced apache.commons logging with p2j.util.LogHelper.
** 027 OM  20130820          Added support for mixed asc/desc indexes.
** 028 OM  20140410          Fixed support case-sensitive fields in indexes.
**                           Added Override annotations. Simplified string concatenations.
** 029 VMN 20140502          Added enhance schema name conversion support for hint "escape"
**                           attribute.
** 030 OM  20140508          Added support for computing the index-key size.
** 031 OM  20140623          Added rtrim() dialect specific implementation. Added query max
**                           parameters count support.
** 032 OM  20140924          Removed usesSingleBackslash.
** 033 OM  20150522          Added getReservedKeywords() static method.
** 034 OM  20151130          rtrim() dialect specific implementation returns the number of
**                           occurrences of the expression processed. Implemented 
**                           supportsBooleanDatatype().
** 035 EVL 20160223          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 036 OM  20160603          Fixed rtrimming to space character only.
** 037 ECF 20160912          Added getSequencePrefetchString.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.dialect;

import java.net.InetAddress;
import java.net.URI;
import java.sql.*;
import java.sql.Date;
import java.text.*;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import com.goldencode.p2j.util.LogHelper;
import org.hibernate.cfg.Configuration;
import org.hibernate.dialect.*;
import org.hibernate.mapping.Column;
import org.hibernate.mapping.Index;
import org.hibernate.mapping.Table;
import com.goldencode.util.*;
import com.goldencode.p2j.directory.DirectoryService;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.sequence.*;
import com.goldencode.p2j.security.ContextLocal;
import com.goldencode.p2j.util.*;

/**
 * Extends Hibernate's <code>PostgreSQLDialect</code> class to add method
 * implementations required by the {@link P2JDialect} interface.
 * <p>
 * Note:  this implementation expects a directory to be available at the time
 * the class is loaded.
 */
public class P2JPostgreSQLDialect
extends PostgreSQL82Dialect
implements P2JDialect
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(P2JPostgreSQLDialect.class.getName());
   
   /** Default minimum interval in seconds between temp table vacuums */
   private static final int DEFAULT_VACUUM_INTERVAL = 30;
   
   /** Debug level logging enabled? */
   private static final boolean debug = LOG.isLoggable(Level.FINE);
   
   /** Flag indicating whether to inline a record limit in an SQL SELECT */
   private static final ContextLocal<Boolean> inlineLimit =
         new ContextLocal<Boolean>()
         {
            protected Boolean initialValue() { return Boolean.FALSE; }
         };
   
   /** Lowest date supported by PostgreSQL (01/01/4713 BC) */
   private static final Date lowDate;
   
   /** Highest date supported by PostgreSQL (12/31/32767 AD) */
   private static final Date highDate;
   
   /** Actual minimum interval in milliseconds between temp table vacuums */
   private static long vacuumInterval = -1L;
   
   static
   {
      // Prepare supported date values.  We use the GMT timezone to define
      // our date range, since this is how com.goldencode.p2j.util.date
      // handles its dates.
      
      TimeZone tz = TimeZone.getTimeZone("GMT");
      GregorianCalendar cal = new GregorianCalendar(tz);
      
      cal.set(Calendar.ERA, GregorianCalendar.BC);
      cal.set(Calendar.YEAR, 4713);
      cal.set(Calendar.MONTH, Calendar.JANUARY);
      cal.set(Calendar.DATE, 1);
      lowDate = new Date(cal.getTime().getTime());
      
      cal.clear();
      
      cal.set(Calendar.ERA, GregorianCalendar.AD);
      cal.set(Calendar.YEAR, 32767);
      cal.set(Calendar.MONTH, Calendar.DECEMBER);
      cal.set(Calendar.DATE, 31);
      highDate = new Date(cal.getTime().getTime());
   }
   
   /**
    * Default constructor.  This constructor is for Hibernate's use.  Use code
    * should not construct instances of this class.
    * <p>
    * We provide this constructor to override the superclass' mapping of 
    * <code>Types.VARCHAR</code>, in order to force this type to map to
    * PostgreSQL's <code>text</code> data type.
    */
   public P2JPostgreSQLDialect()
   {
      super();
      
      // Override superclass' VARCHAR mapping, since we always use text for
      // string data.
      registerColumnType(Types.VARCHAR, "text");
      
      // Override superclass' NUMERIC mapping, since the Hibernate PostgreSQL
      // dialect doesn't honor precision and scale.
      registerColumnType(Types.NUMERIC, "numeric(50,$s)");
   }
   
   /**
    * Get a collection of reserved keywords specific to this dialect.
    *
    * The {@link com.goldencode.p2j.convert.NameConverter} uses thes list to create an exclusion
    * list of names so there will not be schema collisions with identifiers (tables, fields) from
    * the generated schema.
    *
    * @return  A list with PostgreSQL 9.3 reserved keywords in lowercase.
    *
    * @see     <a href="http://www.postgresql.org/docs/9.3/static/sql-keywords-appendix.html">
    *          http://www.postgresql.org/docs/9.3/static/sql-keywords-appendix.html</a>
    */
   public static List<String> getReservedKeywords()
   {
      String[] keywords = {
         "all",
         "analyse",
         "analyze",
         "and",
         "any",
         "array",
         "as",
         "asc",
         "asymmetric",
         "authorization",
         "binary",
         "both",
         "case",
         "cast",
         "check",
         "collate",
         "collation",
         "column",
         "concurrently",
         "constraint",
         "create",
         "cross",
         "current_catalog",
         "current_date",
         "current_role",
         "current_schema",
         "current_time",
         "current_timestamp",
         "current_user",
         "default",
         "deferrable",
         "desc",
         "distinct",
         "do",
         "else",
         "end",
         "except",
         "false",
         "fetch",
         "for",
         "foreign",
         "freeze",
         "from",
         "full",
         "grant",
         "group",
         "having",
         "ilike",
         "in",
         "initially",
         "inner",
         "intersect",
         "into",
         "is",
         "isnull",
         "join",
         "lateral",
         "leading",
         "left",
         "like",
         "limit",
         "localtime",
         "localtimestamp",
         "natural",
         "not",
         "notnull",
         "null",
         "offset",
         "on",
         "only",
         "or",
         "order",
         "outer",
         "over",
         "overlaps",
         "placing",
         "primary",
         "references",
         "returning",
         "right",
         "select",
         "session_user",
         "similar",
         "some",
         "symmetric",
         "table",
         "then",
         "to",
         "trailing",
         "true",
         "union",
         "unique",
         "user",
         "using",
         "variadic",
         "verbose",
         "when",
         "where",
         "window",
         "with",
      };
      
      return Arrays.asList(keywords);
   }
   
   /**
    * Report the minimum number of milliseconds between explicit vacuums.
    * This number is retrieved from the directory, or if not configured there,
    * is defaulted to a standard interval.
    * 
    * @return  Minimum number of milliseconds between vacuums.
    */
   private static long getVacuumInterval()
   {
      if (vacuumInterval < 0)
      {
         // Attempt to read temp-table vacuum interval from temp-table section
         // of directory.  Missing value uses default interval, 0 disables,
         // negative number is an error.
         DirectoryService ds = DirectoryService.getInstance();
         if (!ds.bind())
         {
            throw new RuntimeException("Directory bind failed");
         }
         
         try
         {
            String nodeID = Utils.findDirectoryNodePath(ds, "database", "container", false);
            Integer value = null;
            if (nodeID != null)
            {
               value = ds.getNodeInteger(nodeID + "/" + DatabaseManager.TEMP_TABLE_SCHEMA +
                                             "/p2j/vacuum_interval",
                                         "value");
            }
            
            int seconds = DEFAULT_VACUUM_INTERVAL;
            if (value == null)
            {
               if (LOG.isLoggable(Level.INFO))
               {
                  LOG.log(Level.INFO, "Temp table vacuum interval not specified;  using default");
               }
            }
            else
            {
               seconds = value;
            }
            
            if (seconds <= 0)
            {   
               if (LOG.isLoggable(Level.WARNING))
               {
                  LOG.log(Level.WARNING, "Temp table periodic vacuuming disabled!");
               }
            }
            else if (LOG.isLoggable(Level.INFO))
            {
               LOG.log(Level.INFO, "Temp table vacuum interval set to " + seconds + " seconds");
            }
            
            vacuumInterval = seconds * 1000;
         }
         finally
         {
            ds.unbind();
         }
      }
      
      return vacuumInterval;
   }
   
   /**
    * Indicate whether this dialect requires computed columns to participate
    * within index definitions, instead of embedded expressions.
    * 
    * @return  <code>false</code> to indicate the backing database can embed
    *          an expression within an index definition.
    */
   @Override
   public boolean needsComputedColumns()
   {
      return false;
   }
   
   /**
    * Get the prefix, if any, which should be prepended to computed column and/or property names.
    *
    * @param   caseSensitive
    *          Ignored for this dialect.
    * 
    * @return  <code>null</code> to indicate the backing database does not use computed columns.
    * 
    * @see     #needsComputedColumns()
    */
   @Override
   public String getComputedColumnPrefix(boolean caseSensitive)
   {
      return null;
   }
   
   /**
    * Indicate whether this dialect natively supports the overloading of user
    * defined functions by parameter number and data type.
    * 
    * @return  <code>true</code>.
    */
   @Override
   public boolean supportsFunctionOverloading()
   {
      return true;
   }
   
   /**
    * Indicates whether this dialect supports boolean datatype. Support includes:
    * <ul>
    *    <li>usage of {@code true} and {@code false} literals;
    *    <li>operations with boolean values;
    *    <li>allowing functions to have booleans as parameters and return datatype. 
    * </ul>
    *
    * @return  Always {@code true}, all the above mentioned items are supported.
    */
   @Override
   public boolean supportsBooleanDatatype()
   {
      return true;
   }
   
   /**
    * Create a data object, the purpose of which is specific to a particular
    * dialect implementation, to be stored as a context-local object.  This
    * allows a dialect to establish context-local data which is stored by the
    * persistence runtime framework, without the overhead of a separate,
    * context-local variable.  This object will be handed back to the dialect
    * in callback methods when the services of certain hooks are required.
    * <p>
    * Since the context-local data is specific to a particular dialect
    * implementation, it is optional.  Dialect implementations which do not
    * have a need for this data should simply return <code>null</code>.
    * 
    * @param   persistence
    *          Provider of persistence services.
    * 
    * @return  Dialect-specific, context-local data, or <code>null</code>.
    */
   @Override
   public Object createContext(Persistence persistence)
   {
      Database database = persistence.getDatabase();
      
      return (DatabaseManager.TEMP_TABLE_DB.equals(database) && getVacuumInterval() > 0)
             ? new Context()
             : null;
   }
   
   /**
    * A dialect-specific callback hook which is invoked by the persistence
    * runtime framework after a transaction is complete.  The Hibernate
    * session will have been closed by the time this method is invoked.
    * 
    * @param   persistence
    *          Provider of persistence services.
    * @param   context
    *          Optional, context-local data created previously by {@link
    *          #createContext}.
    * @param   rollback
    *          <code>true</code> if transaction ended with a rollback;
    *          <code>false</code> if transaction ended with a commit.
    * 
    * @throws  PersistenceException
    *          if any persistence-related error occurs.
    */
   @Override
   public void postTransaction(Persistence persistence, Object context, boolean rollback)
   throws PersistenceException
   {
      if (context != null)
      {
         Context local = (Context) context;
         long now = System.currentTimeMillis();
         long diff = now - local.lastVacuum;
         if (diff >= getVacuumInterval())
         {
            int count = 0;
            try
            {
               StringBuilder buf = new StringBuilder();
               Iterator<String> iter = TemporaryBuffer.activeTables();
               for (count = 0; iter.hasNext(); count++, buf.setLength(0))
               {
                  buf.append("vacuum analyze ");
                  buf.append(iter.next());
                  persistence.executeSQL(buf.toString());
               }
            }
            finally
            {
               long end = System.currentTimeMillis();
               long elapsed =  end - now;
               local.vacuumTime += elapsed;
               if (debug)
               {
                  LOG.log(Level.FINE,
                          persistence.message("Vacuum " + count + " temp table(s) [" + diff +
                                              ":" + elapsed + ":" + local.vacuumTime + "]"));
               }
               local.lastVacuum = end;
            }
         }
      }
   }
   
   /**
    * Retrieve the earliest possible date value which is valid for the
    * backing database.
    *
    * @return  Earliest date supported by the backing database.
    */
   @Override
   public Date getLowDate()
   {
      return lowDate;
   }
   
   /**
    * Retrieve the latest possible date value which is valid for the backing
    * database.
    *
    * @return  Latest date supported by the backing database.
    */
   @Override
   public Date getHighDate()
   {
      return highDate;
   }
   
   /**
    * Format a date object as a string which is appropriate for use in an SQL
    * statement.
    * 
    * @param   d
    *          Date to be formatted as a string.
    * 
    * @return  Formatted representation of the given date.
    */
   @Override
   public String formatDate(date d)
   {
      NumberFormat nf = NumberFormat.getInstance(); 
      
      nf.setMinimumIntegerDigits(4);
      nf.setGroupingUsed(false);
      
      StringBuilder buf = new StringBuilder();
      int year = d.getYear();
      
      buf.append(nf.format(Math.abs(year)));
      buf.append('-');
      
      nf.setMinimumIntegerDigits(2);
      buf.append(nf.format(d.getMonth()));
      buf.append('-');
      buf.append(nf.format(d.getDay()));
      
      if (year < 0)
      {
         buf.append(" BC");
      }
      
      return buf.toString();
   }
   
   /**
    * Given a column expression as queried from database metadata, extract the root name of a
    * column. For example, a column name as found within an index in metadata might be represented
    * as:
    * <pre>
    *    upper(rtrim("key"::text))
    * </pre>
    * We are interested only in the {@code key} root name, so we strip away the {@code upper()}
    * and {@code rtrim()} functions the double quotes, and the {@code ::text} suffix.
    *
    * @param   expr
    *          Expression containing column name as natively represented by the database dialect.
    *
    * @return  Root column name.
    */
   @Override
   public String extractColumnName(String expr)
   {
      final String rtrim = "rtrim(";
      final String upper = "upper(";
      
      String root = expr.toLowerCase();
      
      // Remove "upper" function if present (only applicable to text fields).
      if (root.startsWith(upper))
      {
         int len = root.length();
         root = root.substring(upper.length(), len - 1);
      }
      
      // Remove "rtrim" function if present (only applicable to text fields).
      if (root.startsWith(rtrim))
      {
         int commaPos = root.lastIndexOf(",");
         if (commaPos > 0)
         {
            root = root.substring(rtrim.length(), commaPos);
            if (LOG.isLoggable(Level.INFO))
            {
               LOG.log(Level.INFO, "Old rtrim: [" + expr + "].");
            }
         }
         else 
         {
            root = root.substring(rtrim.length(), root.length() - 1);
         }
      }
      
      // Remove data type suffix.
      int delim = root.indexOf("::");
      
      root = (delim >= 0 ? root.substring(0, delim) : root);
      
      // Strip enclosing parentheses.
      while (root.startsWith("(") && root.endsWith(")"))
      {
         root = root.substring(1, root.length() - 1);
      }
      
      // Strip enclosing double quotes or backticks.
      root = StringHelper.stripEnclosing(root, '"');
      root = StringHelper.stripEnclosing(root, '`');
      
      return root;
   }
   
   /**
    * Adjust a character column name appropriate to this dialect, such that it
    * can be inserted into an index.  This involves wrapping the name within
    * an rtrim function, and if <code>ignoreCase</code> is <code>true</code>,
    * an upper function.
    * 
    * @param   name
    *          Base column name.
    * @param   ignoreCase
    *          <code>true</code> to cause case to be ignored, else <code>false</code>.
    * 
    * @return  Processed column name.
    */
   @Override
   public String getProcessedCharacterColumnName(String name, boolean ignoreCase)
   {
      StringBuilder buf = new StringBuilder();
      if (ignoreCase)
      {
         buf.append("upper(");
      }
      buf.append("rtrim(").append(name).append(")");
      if (ignoreCase)
      {
         buf.append(")");
      }
      
      return buf.toString();
   }
   
   /**
    * Report whether the given column name represents an alias or an
    * expression which indicates the column is case insensitive.  Assume the
    * column represented by <code>name</code> is already determined to be the
    * appropriate data type for textual data.
    * 
    * @param   name
    *          Column name, alias, or expression.
    * 
    * @return  <code>true</code> if <code>name</code> represents case-insensitive data.
    */
   @Override
   public boolean isCaseInsensitiveColumn(String name)
   {
      return name.toLowerCase().contains("upper(");
   }
   
   /**
    * Report whether the given column name represents an alias or an expression which indicates
    * the column is computed.  Assume the column represented by <code>name</code> is
    * already determined to be the appropriate data type for textual data.
    *
    * @param   name
    *          Column name, alias, or expression.
    *
    * @return  <code>true</code> if <code>name</code> represents a computed column.
    */
   @Override
   public boolean isComputedColumn(String name)
   {
      return false; // PostgreSQL dialect does not use computed columns
   }
   
   /**
    * Generate a DDL string which will create a global sequence with the given starting value.
    * 
    * @param   name
    *          Sequence name.
    * @param   start
    *          Sequence starting value.
    * 
    * @return  Dialect-specific string to create the sequence, or
    *          <code>null</code> if not supported.
    */
   @Override
   public String getCreateSequenceString(String name, long start)
   {
      return getCreateSequenceString(name) + " start " + start;
   }
   
   /**
    * Generate a DDL string which will create a sequence with given properties.
    * 
    * @param   name
    *          Sequence name. Mandatory not null.
    * @param   init
    *          Sequence starting value. Mandatory(not null).
    * @param   increment
    *          Sequence incrementing value. Mandatory(not null).
    * @param   minVal
    *          Sequence minimum value. Optional (may be null).
    * @param   maxVal
    *          Sequence maximum value. Optional (may be null).
    * @param   cycle
    *          Sequence cycling property. Optional, if missing assumed not cycling.
    * 
    * @return  Dialect-specific string to create the sequence, or <code>null</code> if 
    *          not supported.
    */
   @Override
   public String getCreateSequenceString(String name,
                                         Long init,
                                         Long increment,
                                         Long minVal,
                                         Long maxVal,
                                         Boolean cycle)
   {
      StringBuilder buf = new StringBuilder(getCreateSequenceString(name));
      buf.append(" increment ").append(increment);
      
      if (minVal != null)
      {
         buf.append(" minvalue ").append(minVal);
      }
      else 
      {
         // default MINVALUE in PSQL is 1, must take care to avoid 
         // START value (n) cannot be less than MINVALUE (1) error
         // if (increment > 0 /* && init < 1*/)
         {
            
            //default minvalue is: -9223372036854775807 !!
            minVal = Long.MIN_VALUE;
            buf.append(" minvalue ").append(minVal);
            LOG.log(Level.WARNING, "Fixed MINVALUE = " + minVal + " for PSQL sequence " + name);
         }
      }
      
      if (maxVal != null)
      {
         buf.append(" maxvalue ").append(maxVal);
      }
      else 
      {
         // default MAXVALUE in PSQL is -1, must take care to avoid 
         // START value (n) cannot be greater than MAXVALUE (-1) error
         // if (increment < 0 /* && init > -1 */)
         {
            maxVal = Long.MAX_VALUE;
            buf.append(" maxvalue ").append(maxVal);
            LOG.log(Level.WARNING, "Fixed MAXVALUE = " + maxVal + " for PSQL sequence " + name);
         }
      }
      
      buf.append(" start ").append(init);
      if (cycle != null && cycle)
      {
         buf.append(" cycle");
      }
      buf.append(" cache 1");
      
      return buf.toString();
   }
   
   /**
    * Generate a DDL string which will drop a sequence with given properties.
    * 
    * @param   name
    *          The name of the sequence to be dropped.
    *
    * @return  Dialect-specific string to drop the sequence, or
    *          <code>null</code> if not supported.
    */
   @Override
   public String getDropSequenceString(String name)
   {
      return "drop sequence if exists " + name;
   }
   
   /**
    * Generate the appropriate select statement to retrieve the next value of a sequence.
    * 
    * @param   sequenceName
    *          The name of the sequence.
    * @param   newVal
    *          The new value to be set.
    *
    * @return  Dialect-specific string to query the sequence, or
    *          <code>null</code> if not supported.
    */
   @Override
   public String getSequenceSetValString(String sequenceName, long newVal)
   {
      return "select setval('" + sequenceName + "', " + newVal + ")";
   }
   
   /**
    * Generate the appropriate select statement to reset the value of a sequence.
    * 
    * @param   sequenceName
    *          The name of the sequence to be queried.
    *
    * @return  Dialect-specific string to query the sequence, or
    *          <code>null</code> if not supported.
    */
   @Override
   public String getSequenceCurrValString(String sequenceName)
   {
      // The CURRVAL is relative to current session and will fail if called
      // before a call to NEXTVAL or SETVAL
      
      return "select last_value from " + sequenceName;
   }
   
   /**
    * Returns <code>true</code> if this dialect fully support the sequences as
    * 4GL language (most important: bounded values, cycle). If not fully
    * supported, additional code is needed in P2J sequence class
    * implementation to fix this.
    * 
    * @return true because PostgreSQL dialect fully support 4GL sequences
    */
   @Override
   public boolean hasFullSequenceSupport()
   {
      return true;
   }
   
   /**
    * Generate the appropriate select statement to pre-fetch multiple, ascending values from a
    * sequence. The values need not be contiguous.
    * 
    * @param   sequenceName
    *          The name of the sequence from which the values are to be retrieved.
    * 
    * @return  Dialect-specific SQL expression, or {@code null} if this feature is not supported
    *          by the dialect.
    */
   @Override
   public String getSequencePrefetchString(String sequenceName)
   {
      return "select nextval('" + sequenceName + "') from generate_series(?, ?)";
   }
   
   /**
    * Context local data needed by this object to perform dialect-specific
    * processing.  In this case, we track the last time a temp table vacuum
    * was performed and how much accumulated time we have spent performing
    * such vacuums in this context.
    */
   private static class Context
   {
      /** Time of last temp table vacuum */
      private long lastVacuum = System.currentTimeMillis();
      
      /** Accumulated time spent vacuuming temp tables */
      private long vacuumTime = 0L;
   }

   /**
    * Given an exception, this method will check if it indicates a database
    * connectivity error.
    * <p>
    * For the PostgreSQL dialect, if the state indicates a 08xxx error code
    * or 57P01 or 57P02 codes, the method will return <code>true</code>.
    * <p>
    * A complete list of defined error codes can be found here:
    * http://www.postgresql.org/docs/8.1/static/errcodes-appendix.html
    * 
    * @param   exc
    *          The exception to be checked.
    * 
    * @return  <code>true</code> if the exception indicates a database 
    *          connectivity error.
    */
   @Override
   public boolean isConnectionError(Exception exc)
   {
      boolean error = false;
      
      Throwable next = exc;
      while (next != null && !error)
      {
         if (next instanceof SQLException)
         {
            String state = ((SQLException) next).getSQLState();

            error = state != null && 
                    (state.startsWith("08") || 
                     state.equals("57P01")  ||
                     state.equals("57P02"));
         }
         next = next.getCause();
      }
      
      return error;
   }

   /**
    * Indicate whether a query parameter which is part of a range check should
    * be inlined into the where clause, or whether it should be substituted
    * into a prepared statement.  This will result in an "inlined"
    * substitution parameter within a where clause.  For example, given a
    * where clause of:
    * <pre>
    *    where value &ge; ? and id = ?
    * </pre>
    * along with respective substitution parameters of 45 and 100, the
    * inlined version would read:
    * <pre>
    *    where value &ge; 45 and id = ?
    * </pre>
    * Only the <code>id</code> substitution parameter of 100 would be
    * substituted into the prepared statement.
    * <p>
    * Inlining in such a way can give some database dialects the opportunity
    * to choose a better query plan, than if the parameter being tested with
    * a range check were to be substituted at statement execution time.
    * <p>
    * Note:  this serves as a hint which may be overridden by runtime logic or
    * other configuration settings.
    * 
    * @return  <code>true</code> to inline range check parameters;
    *          <code>false</code> to use normal parameter substitution.
    */
   @Override
   public boolean isQueryRangeParameterInlined()
   {
      return true;
   }
   
   /**
    * Enable/disable limit clause inlining.  Inlining the limit clause will
    * cause the limit/max value to be hard-coded into the SQL statement,
    * rather than being bound as a substitution parameter.
    * <p>
    * Dialects are free to treat this method as a hint.
    * 
    * @param   on
    *          <code>true</code> to enable inlining;  <code>false</code> to
    *          disable it.
    */
   @Override
   public void inlineLimit(boolean on)
   {
      inlineLimit.set(on);
   }

   /**
    * Determines whether substitution parameters should be explicitly casted
    * to their datatypes inside "then ? else ?" statement.
    *
    * @return <code>false</code>.
    */
   @Override
   public boolean requiresExplicitCastInsideTernary()
   {
      return true;
   }
   
   /**
    * Override the parent's implementation to inline small (2 or less) limit
    * values when there is no accompanying offset.  In other cases, we fall
    * back to the original implementation.
    * <p>
    * WARNING:  the implementation of this override has an ugly dependency on
    * the order of the logic in Hibernate's <code>Loader</code> class.
    * Specifically, we rely on the fact that this method is invoked when a
    * query is being prepared, which must happen before parameters are bound,
    * at which point {@link #supportsVariableLimit()} is invoked.  This
    * method sets state in a context local variable which is later read when
    * the latter method is invoked.
    * <p>
    * TODO:  review this implementation when upgrading Hibernate versions. 
    * 
    * @param   sql
    *          Base select statement SQL without the limit/offset clause.
    * @param   offset
    *          Starting offset.
    * @param   limit
    *          Number of records to return.
    *          
    * @return  SQL with limit/offset clause appended.
    * 
    * @see     #supportsVariableLimit()
    */
   @Override
   public String getLimitString(String sql, int offset, int limit)
   {
      String stmt;
      boolean inlinePermitted = inlineLimit.get();
      
      if (inlinePermitted && limit <= 2 && offset == 0)
      {
         stmt = sql + " limit " + limit;
      }
      else
      {
         if (inlinePermitted)
         {
            // Override inline request.
            inlineLimit.set(Boolean.FALSE);
         }
         stmt = super.getLimitString(sql, offset, limit);
      }
      
      return stmt;
   }
   
   /**
    * Indicate whether a variable limit is supported by this dialect.  The
    * answer is dependent upon context local state set within {@link
    * #getLimitString(String, int, int)}.  See the applicable warning in that
    * method's description.  This method is consulted during parameter binding
    * in Hibernate's <code>Loader</code> class.
    * 
    * @return  <code>true</code> if a bound parameter should be used in a
    *          SELECT statement's limit clause;  <code>false</code> if the
    *          limit was inlined by {@link #getLimitString(String, int, int)}.
    */
   @Override
   public boolean supportsVariableLimit()
   {
      return inlineLimit.get() ? false : super.supportsVariableLimit();
   }

   /**
    * Build the DDL to drop an index based upon the specified definition.
    * The SQL is generated by using the database-specific dialect.
    *
    * @param   index
    *          Definition of the index to be dropped.
    *
    * @return  The DDL to drop the given index.
   */
   @Override
   public String getDropIndexString(P2JIndex index)
   {
      return "drop index if exists " + quote(index.getName());
   }

   /**
    * Build the DDL to create an index based upon the specified definition.
    * The SQL is generated by using a database-specific dialect.
    *
    * @param   index
    *          Definition of the index to be created.
    * @param   unique
    *          If <code>true</code>, create the index as a unique index,
    *          <code>false</code> to omit the unique constraint.
    * @param   toFile
    *          <code>true</code> if the string will be output to a file or
    *          <code>false</code> if it will be directly submitted via JDBC.
    *
    * @return  The DDL which was created.
    */
   @Override
   public String getCreateIndexString(P2JIndex index, boolean unique, boolean toFile)
   {
      // Add primary key column to any index which is not explicitly or implicitly unique.
      // TODO:  move this to import ruleset so we don't hardcode primary key column name.
      //        OM: my attempt to move the code from both dialects to rulesets failed as
      //            the index list passes through P2JIndex.normalizeUniqueIndexes() where
      //            'unique' property it's adjusted to its correct value. 
      if (!index.isUnique())
      {
         // the direction of the pk column name is the same with the last component before it.
         // this make this index much easier to fit using the sort clause of the FindQuery, 
         // especially in reverse sorting mode (at runtime I cannot tell is the sorting direction
         // was already reversed during conversion for FIND LAST/FIND PREV structures so I
         // cannot assume the direction of pk id is always ascending; however, if the direction
         // is linked to one of the components of the index, reversing it's sorting direction
         // will also reverse the sort order of the pk id so the query planner will be able to
         // identify the correct index at runtime).
         boolean idDesc = false;
         if (index.size() > 0) // should normally be
         {
            P2JIndexComponent last = index.getComponent(index.size() - 1);
            idDesc = last.isDescending(); 
         }
         index.addComponent(DatabaseManager.PRIMARY_KEY, idDesc, false, false, false);
      }
      
      // Extract index information into a Hibernate-friendly form.
      String tabName = index.getTable();
      Table table = new Table(tabName);
      List<Column> columns = new ArrayList<>();
      Iterator<P2JIndexComponent> iter = index.components();
      while (iter.hasNext())
      {
         P2JIndexComponent next = iter.next();
         String colName = next.getName();
         if (next.isCharType())
         {
            // Special handling for text columns.
            boolean ignoreCase = next.isIgnoreCase();
            StringBuilder buf = new StringBuilder();
            if (ignoreCase)
            {
               buf.append("upper(");
            }
            buf.append("rtrim(").append(quote(colName)).append(")");
            if (ignoreCase)
            {
               buf.append(")");
            }
            colName = buf.toString();
         }
         
         // TODO: asc/desc is not (directly) supported by hibernate
         // to allow asc/desc clauses for the index, we manipulate the column name to add the 
         // "desc" direction, if needed; this works as long as (function-)computed Columns 
         // for the char/string columns also do
         // PostgreSQL version must be 8.3 or above to support "asc/desc" clause
         if (next.isDescending())
         {
            colName = colName + " desc";
         }
         columns.add(new Column(colName));
      }
      
      // Compose the string.
      String createString =
         Index.buildSqlCreateIndexString(this,
                                         quote(index.getName()),
                                         table,
                                         columns.iterator(),
                                         unique,
                                         null,
                                         null);
      return createString;
   }
   
   /**
    * The PostgreSQL dialect doesn't need computed columns for indexed, char
    * columns, so this implementation returns <code>null</code>
    *
    * @param   tableName
    *          Table name possible containing backticks.
    * @param   config
    *          The hibernate's configuration object.
    * @param   implClassName
    *          The full name of the implementing class for the DMO.
    * @param   colName
    *          The column name which needs to have a computed version. This is
    *          not the computed column name, but the name of its "parent".
    * @param   ignoreCase
    *          <code>true</code> if this is a character field which has the
    *          "ignore case" flag set.
    *          
    * @return  <code>null</code> 
    */
   @Override
   public String getAlterComputedColumnString(String tableName,
                                              Configuration config,
                                              String implClassName,
                                              String colName,
                                              boolean ignoreCase)
   {
      return null;
   }
   
   /**
    * Returns the formula for character computed columns. This will be called in two cases: when
    * generating the ddl for permanent databases schema (this will be written to a .sql file) and
    * when creating temp-tables (the code will be sent directly to SQL).
    *
    * @param   colName
    *          The base column name.
    * @param   ignoreCase
    *          {@code true} if this is a char field which has the "ignore case" flag set.
    *
    * @return  Always {@code null} because this dialect does not use computed columns.
    */
   @Override
   public String getComputedColumnFormula(String colName, boolean ignoreCase)
   {
      return null;
   }
   
   /**
    * Prepare the given metadata query parameter by making it lowercase.
    * 
    * @param   param
    *          The metadata query parameter.
    *          
    * @return  the lowercased parameter.
    */
   @Override
   public String prepareMetadataParameter(String param)
   {
      return param == null ? null : param.toLowerCase();
   }

   /**
    * Process the given metadata query result, by making it lowercase.
    * 
    * @param   value
    *          The metadata query result.
    *          
    * @return  the lowercased result.
    */
   @Override
   public String processMetadataResult(String value)
   {
      return value == null ? null : value.toLowerCase();
   }

   /**
    * Check if the given value is a descending clause for a column index.
    * 
    * @param   value
    *          The value to be checked.
    *          
    * @return  <code>true</code> if the value is the "D" or "d" string.
    */
   @Override
   public boolean isMetadataSortDesc(String value)
   {
      return "D".equalsIgnoreCase(value);
   }
   
   /**
    * Check if the given value is an ascending clause for a column index.
    * 
    * @param   value
    *          The value to be checked.
    *          
    * @return  <code>true</code> if the value is the "A" or "a" string.
    */
   @Override
   public boolean isMetadataSortAsc(String value)
   {
      return "A".equalsIgnoreCase(value);
   }

   /**
    * Get a list of database prepare DDL statements.
    * 
    * @param    database
    *           The database which needs to be prepared.
    *            
    * @return   an empty list.
    */
   @Override
   public List<String> getDatabasePrepareStatements(Database database)
   {
      return Collections.EMPTY_LIST;
   }
   
   /**
    * Check if the dialect needs to set an explicit collation during schema
    * DDL generation.
    * 
    * @return  always <code>false</code>.
    */
   @Override
   public boolean explicitSetCollation()
   {
      return false;
   }
   
   /**
    * Get the statement delimiter for this dialect.
    * 
    * @return  the statement delimiter.
    */
   @Override
   public String getDelimiter()
   {
      return ";";
   }
   
   /**
    * Set any database parameters before, to be available when the database
    * driver is loaded.
    * <p>
    * This implementation does nothing.
    */
   @Override
   public void beforeDriverLoad()
   {
      // no-op
   }

   /**
    * Build the URL to be used for a remote connection, based on the given
    * URL and database.
    * 
    * @param    url
    *           The URL as received from the remote server.
    * @param    database
    *           The database to which this URL belongs.
    * 
    * @return   The built URL as to be used for a remote connection.
    *
    * @throws   PersistenceException
    *           if the specified connection URL is invalid.
    */
   @Override
   public String buildRemoteURL(String url, Database database)
   throws PersistenceException
   {
      String origUrl = url;
      try
      {
         url = url.substring("jdbc:".length());
         URI uri = (new URI(url)).normalize();

         // Check for localhost references.
         String host = uri.getHost().toLowerCase();
         if (InetAddress.getByName(host).isLoopbackAddress())
         {
            // Use hostname stored in database object.
            host = database.getSocketAddress().getHostName();
            
            // Generate new URI.
            uri = new URI(uri.getScheme(),
                          uri.getUserInfo(),
                          host,
                          uri.getPort(),
                          uri.getPath(),
                          uri.getQuery(),
                          uri.getFragment());
            
            // Prepend 'jdbc:' prefix.
            url = "jdbc:" + uri;
         }
         else
         {
            // no preparation is needed
            url = origUrl;
         }
      }
      catch (Exception exc)
      {
         throw new PersistenceException("Invalid database connection URL", exc);
      }

      return url;
   }
   
   /**
    * Creates and returns a <code>SequenceHandler</code> for a PostgresSQL database.
    * 
    * @param   ldbName
    *          The logical name of the database to create the <code>SequenceHandler</code> for.
    * 
    * @return  A <code>PostgreSQLSequenceHandler</code> instance.
    */
   @Override
   public SequenceHandler createSequenceHandler(String ldbName)
   {
      // TODO Auto-generated method stub
      return new PostgreSQLSequenceHandler(ldbName);
   }

   /**
    * Report whether the computed columns should be injected on table creating.
    *
    * @return   <code>false</code> because PostgreSQL does not support computed columns.
    */
   @Override
   public boolean injectComputedColumns()
   {
      return false;
   }
   
   /**
    * Some dialects do have constraints about the maximum size of fields that compose an index. 
    * This method reports the dialect-specific index length limit, if any.
    *
    * @return  always 0 as this dialect do not have any known constraints about the maximum size 
    *          of the fields of an index. 
    */
   @Override
   public int getIndexLengthLimit()
   {
      return 0;
   }
   
   /**
    * Computes the size (in bytes) occupied by this value when stored as index-key in an index.
    * Because this dialect does not have index-size constraints then this method returns 0.
    *
    * @param   value
    *          Ignored. 
    *
    * @return  always 0.
    */
   public int computeIndexKeySize(BaseDataType value)
   {
      return 0;
   }
   
   /**
    * Returns the size of the synthetic part of the index. At this moment only the <code>id</code>
    * of the record may be additionally added to an index to add custom properties (eg uniqueness)
    * but other fields could be added. This part of the index is not visible using index 
    * reflection routines from <code>IndexHelper</code>. The synthetic components may differ 
    * between unique and non-unique indexes.
    * <p>
    * See {@link #getCreateIndexString(P2JIndex, boolean, boolean)}
    *
    * @param   unique
    *          <code>true</code> if the index is unique
    *
    * @return  0 if index is unique, 8 = size of an pkid (int8) 
    */
   public int getSyntheticIndexSize(boolean unique)
   {
      return unique ? 0 : 8;
   }
   
   /**
    * Composes a rtrim equivalent call for a specified expr. This is dialect specific.
    * The returned value looks like:
    * <pre>
    *    rtrim(expr)
    * </pre>
    * The rtrimmed expression is appended to the string builder.
    *
    * @param   expr
    *          The expression to be rtrim -med. It's up to the caller if it is
    *          a qualified field or not or any other kind of character-type expression.
    * @param   sb
    *          The {@link StringBuilder} to append the trimmed filed to.
    */
   @Override
   public void addRtrimmedExpression(String expr, StringBuilder sb)
   {
      sb.append("rtrim(").append(expr).append(")");
   }
   
   /**
    * Obtain the number of parameters a query of this dialect is allowed by the SQL server.
    *
    * @return  always 0 because PostgreSQL does no have such limitation.
    */
   @Override
   public int getMaxParameterCount()
   {
      return 0;
   }
}