/*
** Module   : DefaultDirtyShareContext.java
** Abstract : Default implementation of DirtyShareContext
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 ECF 20080420   @38626 Created initial version. Client context's
**                           primary bridge to the DirtyShareManager.
**                           Manages dirty state specific to the current
**                           context.
** 002 ECF 20080610   @38682 Added safety logic to prevent NPEs.
** 003 ECF 20080612   @38719 Fixed getDirtyInfo(). If an uncommitted
**                           delete was made before the backing table in
**                           the dirty database was created, we could
**                           attempt a query nonetheless and bomb.
** 004 ECF 20080627   @39073 Minor optimization. Avoid collection empty
**                           check is some cases.
** 005 CA  20080716   @39122 Renamed method lockUniqueIndexes to 
**                           lockAllIndexes - this will lock all the 
**                           indexes for a certain entity. Modified the
**                           delete(...) and rollbackInsert(...) methods
**                           so the index locking can be disabled.
** 006 ECF 20080815   @39542 Added support for global change events. These
**                           are events that represent index-changing
**                           inserts, updates, or deletes, which are
**                           processed by sessions other than the session
**                           in which the event originated.
** 007 ECF 20081211   @40869 Fixed delete()/rollbackInsert() defect. The
**                           deletion of a record which was inserted in
**                           the same transaction rolls back the insert,
**                           but does not track the record as deleted.
**                           This caused a problem when the insert itself
**                           was later rolled back. Such records are now
**                           tracked via the deadInserts set.
** 008 ECF 20090428   @42129 Added isLocalInsert(). Reports whether a record
**                           is being tracked as newly inserted by this
**                           context.
** 009 ECF 20090513   @42173 Modified update() method. Need to pass new lock
**                           parameter to modified DirtyShareManager update()
**                           method.
** 010 ECF 20090601   @42569 Added support for Progress isolation leak quirk.
**                           Added new methods.
** 011 ECF 20090610   @42639 Modified updateSnapshots() method. Now returns
**                           list of DMOs affected and tracks entities whose
**                           uncommitted changes have been leaked. The list of
**                           such entities is now used during cleanup.
** 012 ECF 20090611   @42661 Optimized updateSnapshots(). In a long-running
**                           transaction with many updates, it was doing much
**                           unnecessary work. We now track which DMOs need to
**                           have their snapshots updated and refresh only
**                           these.
** 013 CA  20090721   @43285 Fixed rollback of dirty records, when a record is
**                           in dead insert mode - the changes map should not 
**                           be checked in this case.
** 014 ECF 20090721   @43307 Modified delete(). If the record being deleted is
**                           already being tracked for changes, it must be
**                           removed from the changes and dirtySnapshots
**                           collections.
** 015 ECF 20090819   @43692 Added new update() variant. Used when it is known
**                           that the target DMO already is being tracked by
**                           the dirty share manager.
** 016 ECF 20110405          Fixed dirty change leak. Remove record identifier
**                           from deadInserts set upon insert to avoid error
**                           during rollbackInsert() for same record later.
**                           Otherwise, a record is leaked in the dirty share
**                           manager for the life of the server process. Also
**                           fixed (separate) minor bug in rollbackInsert()
**                           which was not removing dirtySnapshot identifiers
**                           when their corresponding records were removed
**                           from dirty share tracking.
** 017 ECF 20131028          Import change.
** 018 ECF 20140821          Changed insert API to allow insertion of unvalidated, newly created
**                           records.
** 019 OM  20160307          Added support for EXTENTs in update() methods. Updated signature of
**                           getDirtyInfo().
** 020 ECF 20160608          Removed unnecessary intern calls.
** 021 ECF 20160917          Added ignoreLocal parameter to isDirtyDelete. 
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.dirty;

import java.io.Serializable;
import java.lang.reflect.Method;
import java.util.*;
import org.hibernate.type.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.util.*;

/**
 * Default implementation of a client context's primary interface to store
 * uncommitted index updates for sharing with other contexts, and to read
 * uncommitted index updates made by other contexts.  This is intended to
 * support compatibility with Progress' peculiar transaction isolation
 * behavior of immediately sharing updates to indexed fields across user
 * sessions, but deferring the sharing of updates to non-indexed fields until
 * the enclosing transaction is committed.
 * <p>
 * This implementation maintains state specific to the current context, in
 * order both to filter information appropriately, and to be able to clean up
 * shared data when the current context's transactions are completed.
 * 
 * @see  DirtyShareManager
 * @see  DefaultDirtyShareManager
 * @see  NopDirtyShareContext
 */
public final class DefaultDirtyShareContext
implements DirtyShareContext
{
   /** Object which manages dirty reads for the database */
   private final DirtyShareManager manager;
   
   /** Identifiers of records inserted by this context */
   private Set<RecordIdentifier> inserts = null;
   
   /** Identifiers of records inserted by this context, which deletes have rolled back */
   private Set<RecordIdentifier> deadInserts = null;
   
   /** Buffers holding newly created records, which have not yet been validated/flushed */
   private Map<RecordIdentifier, RecordBuffer> unvalidatedInserts = null;
   
   /** Map of entity names to sets of IDs inserted/updated by this context */
   private Map<String, Set<Serializable>> changes = null;
   
   /** Map of entity names to sets of IDs deleted by this context */
   private Map<String, Set<Serializable>> deletes = null;
   
   /** Set of IDs of DMOs changed since last updateSnapshots() call */
   private Set<RecordIdentifier> dirtySnapshots = null;
   
   /** DMO entities which have had uncommitted updates "leaked" */
   private Set<String> leakedEntities = null;
   
   /**
    * Constructor which stores a reference to the dirty share manager for the
    * associated database.
    * 
    * @param   manager
    *          Dirty share manager for the associated database.
    */
   DefaultDirtyShareContext(DirtyShareManager manager)
   {
      this.manager = manager;
   }
   
   /**
    * Register with a global event clearinghouse an interest in insert,
    * delete, and update events which affect the given DMO entity.
    * 
    * @param   entity
    *          DMO entity for which the caller is interested in tracking
    *          change.
    * 
    * @return  A unique identifier for the registrant, assigned by the event
    *          clearinghouse, used later to collect events and deregister.
    * 
    * @see     #getGlobalEvents(long, long)
    * @see     #deregisterForGlobalEvents(long)
    */
   public Long registerForGlobalEvents(String entity)
   {
      return manager.registerForGlobalEvents(entity);
   }
   
   /**
    * Deregister with a global event clearinghouse an interest in insert,
    * delete, and update events which affect a particular DMO type.
    * 
    * @param   registerID
    *          A unique identifier for the registrant being deregistered.
    * 
    * @see     #getGlobalEvents(long, long)
    * @see     #registerForGlobalEvents(String)
    */
   public void deregisterForGlobalEvents(long registerID)
   {
      manager.deregisterForGlobalEvents(registerID);
   }
   
   /**
    * Retrieve any new events that have been collected by a global event
    * clearinghouse, representing inserts, deletes, and updates made to
    * record types for which the caller previously has registered interest.
    * 
    * @param   registerID
    *          A unique identifier for an object which has registered interest
    *          in record-changing events.
    * @param   lastID
    *          A unique identifier representing the last event retrieved by
    *          the caller.  If less than 0, it is assumed no events have yet
    *          been retrieved since the caller registered interest in these
    *          events.
    * 
    * @return  Array of applicable, new events, or <code>null</code> if no
    *          such events have been collected.
    * 
    * @throws  EventRegistrationException
    *          if <code>registerID</code> is invalid or has expired.
    */
   public GlobalChangeEvent[] getGlobalEvents(long registerID, long lastID)
   throws EventRegistrationException
   {
      return manager.getGlobalEvents(registerID, lastID);
   }
   
   /**
    * Lock or unlock all indexes associated with the given DMO entity.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   lockType
    *          Type of lock to acquire, or <code>NONE</code> to release
    *          existing locks.  <code>NO_WAIT</code> variants will be treated
    *          as regular, blocking variants.
    * 
    * @throws  PersistenceException
    *          if there is an error querying JDBC metadata while retrieving
    *          index information.
    */
   public void lockAllIndexes(String entity, LockType lockType)
   throws PersistenceException
   {
      manager.lockAllIndexes(entity, lockType);
   }
   
   /**
    * Begin tracking a newly inserted record in the dirty database.
    * <p>
    * This record is removed when the adding context's current transaction is
    * committed or rolled back.
    * 
    * @param   buffer
    *          Record buffer responsible for the insert.
    * @param   dmo
    *          Record to be tracked.
    * @param   setter
    *          Optional setter method for a property which must be updated in
    *          the DMO copy before it is inserted into the dirty database.
    *          May be <code>null</code>.
    * @param   datum
    *          Updated value to apply to <code>property</code>.  May be
    *          <code>null</code>.
    * @param   uniqueIndexes
    *          If the insertion is happening at record creation, and the record has not yet been
    *          validated, this will represent the collection of unique indexes that will need to
    *          be validated before the record can safely be persisted. If <code>null</code>, this
    *          indicates the record has been fully validated and should no longer be tracked as
    *          unvalidated.
    * @param   lock
    *          <code>true</code> to acquire write locks to insert the record;
    *          <code>false</code> to assume write locks are acquired by
    *          calling code.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void insert(RecordBuffer buffer,
                      Persistable dmo,
                      Method setter,
                      BaseDataType datum,
                      Collection<String> uniqueIndexes,
                      boolean lock)
   throws PersistenceException
   {
      String entity = buffer.getEntityName();
      Serializable id = dmo.getId();
      RecordIdentifier ident = new RecordIdentifier(entity, id);
      
      if (inserts != null && inserts.contains(ident))
      {
         if (uniqueIndexes == null      &&
             unvalidatedInserts != null &&
             unvalidatedInserts.remove(ident) != null)
         {
            manager.notifyInsertValidated(ident, (Persistable) dmo.deepCopy());
         }
         
         // This record was already inserted during this transaction.
         return;
      }
      
      // make a copy of the DMO to share with other contexts
      Persistable copy = (Persistable) dmo.deepCopy();
      
      if (uniqueIndexes != null)
      {
         if (unvalidatedInserts == null)
         {
            unvalidatedInserts = new HashMap<>();
         }
         unvalidatedInserts.put(ident, buffer);
      }
      
      if (deadInserts != null)
      {
         // Remove this record from the dead inserts set if we were tracking
         // it as such;  this will prevent any problems if we later have to
         // roll back this insert.
         deadInserts.remove(ident);
      }
      
      // Apply pending update if necessary.
      if (setter != null)
      {
         try
         {
            setter.invoke(copy, datum);
         }
         catch (Exception exc)
         {
            throw new PersistenceException(exc);
         }
      }
      
      // Store in database.
      manager.insert(entity, copy, uniqueIndexes, lock);
      
      // Track insert in local context.
      if (inserts == null)
      {
         inserts = new HashSet<>();
      }
      inserts.add(ident);
      trackChange(entity, id);
   }
   
   /**
    * Indicate whether the specified record is being tracked as newly inserted
    * by this context.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key of the record to be checked.
    * 
    * @return  <code>true</code> if this record is newly inserted by this
    *          context, else <code>false</code>.
    */
   public boolean isLocalInsert(String entity, Serializable id)
   {
      if (inserts == null)
      {
         return false;
      }
      
      RecordIdentifier ident = new RecordIdentifier(entity, id);
      
      return inserts.contains(ident);
   }
   
   /**
    * Indicate whether the record identified by <code>ident</code> represents a newly inserted
    * record, in a different context than that of the caller, which has an unvalidated index,
    * <code>index</code>. Assumes appropriate indexes already are locked by the caller.
    * 
    * @param   ident
    *          Record identifier (using entity name and primary key).
    * @param   index
    *          Index to be checked for validation.
    * 
    * @return  <code>true</code> if <code>ident</code> represents a new record in a foreign
    *          context with unvalidated index <code>index</code>.
    */
   public boolean isForeignUnvalidatedIndex(RecordIdentifier ident, String index)
   {
      // check easy case first
      if (unvalidatedInserts != null && unvalidatedInserts.containsKey(ident))
      {
         // tracked by this context, not a foreign one.
         return false;
      }
      
      return manager.isUnvalidatedIndex(ident, index);
   }
   
   /**
    * If the specified record is an unvalidated insert created by a buffer local to this context,
    * get the buffer which created it.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation class associated
    *          with a table being tracked for uncommitted changes.
    * @param   id
    *          Primary key of the record to be checked.
    * 
    * @return  Creating buffer, or <code>null</code> if the record is not a locally created,
    *          unvalidated insert.
    */
   public RecordBuffer getCreatingBuffer(String entity, Serializable id)
   {
      return unvalidatedInserts != null
             ? unvalidatedInserts.get(new RecordIdentifier(entity, id))
             : null;
   }
   
   /*
   public boolean isLocalDelete(String entity, Serializable id)
   {
      if (deletes != null)
      {
         Set<Serializable> deleted = deletes.get(entity);
         if (deleted != null && deleted.contains(id))
         {
            return true;
         }
      }
      
      if (deadInserts != null)
      {
         RecordIdentifier ident = new RecordIdentifier(entity, id);
         if (deadInserts.contains(ident))
         {
            return true;
         }
      }
      
      return false;
   }
   */
   
   /**
    * Rollback the insert of a record previously introduced with the {@link
    * #insert(RecordBuffer, Persistable, Method, BaseDataType, Collection, boolean)}
    * method.
    * 
    * @param   ident
    *          Identifier for the target record, which encapsulates its entity
    *          name and primary key.
    * @param   lock
    *          <code>true</code> to acquire write locks;
    *          <code>false</code> to assume write locks are acquired by
    *          calling code.
    * 
    * @throws  IllegalStateException
    *          if the record was not inserted in this context during the
    *          current transaction
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void rollbackInsert(RecordIdentifier ident, boolean lock)
   throws PersistenceException
   {
      // if the target record was already deleted from the dirty context, then
      // the changes map check is not relevant (it may be empty if only one
      // record is in the dirty context and this is a dead record)
      if ((deadInserts == null || !deadInserts.contains(ident)) && 
          (changes == null || changes.isEmpty() || 
           inserts == null || !inserts.contains(ident)))
      {
         throw new IllegalStateException(
            "No dirty entry for '" + ident + "';  cannot roll back insert");
      }
      
      try
      {
         if (deadInserts != null && deadInserts.remove(ident))
         {
            // This insert occurred within the existing transaction, and was
            // rolled back by a subsequent delete within the transaction, so
            // it already has been removed from the dirty share manager.  We
            // cannot remove it again as this will cause an error.
            
            if (deadInserts.isEmpty())
            {
               deadInserts = null;
            }
            
            return;
         }
         
         // Remove from local context insert tracking.
         inserts.remove(ident);
         if (unvalidatedInserts != null)
         {
            unvalidatedInserts.remove(ident);
         }
         
         // Remove from database.
         manager.rollbackInsert(ident, lock);
      }
      finally
      {
         // Remove from local context change tracking.
         String entity = ident.getTable();
         Serializable id = ident.getRecordID();
         Set<Serializable> ids = changes.get(entity);
         if (ids != null && ids.remove(id))
         {
            if (ids.isEmpty())
            {
               changes.remove(entity);
            }
            
            dirtySnapshots.remove(ident);
         }
      }
   }
   
   /**
    * Track an update to one or more indexed, DMO properties, as indicated by the 
    * {@code properties}, {@code extIndexes} and {@code values} arguments.
    * <p>
    * Any record added to the dirty database as a result of this method is removed when the adding
    * context's current transaction is committed or rolled back.
    * <p>
    * Assumes appropriate indexes have been locked by the caller.
    *
    * @param   buffer
    *          Record buffer responsible for the update.
    * @param   properties
    *          Array of names of those DMO properties which were updated. These should represent
    *          columns which participate in one or more indexes.
    * @param   extIndexes
    *          In case of EXTENT fields this array contains their index. Otherwise -1.
    * @param   values
    *          Updated DMO property values.  Each value in the array corresponds positionally with
    *          its matching property name in the {@code properties} array.
    * @param   validatedUniqueIndexes
    *          Collection of unique indexes which have been checked successfully for this update
    *          to have passed validation.
    * @param   transientIndexValidated
    *          {@code true} if the update was caused by an index validation of a transient record.
    *
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void update(RecordBuffer buffer,
                      String[] properties,
                      int[] extIndexes,
                      BaseDataType[] values,
                      Collection<String> validatedUniqueIndexes, 
                      boolean transientIndexValidated)
   throws PersistenceException
   {
      String entity = buffer.getEntityName();
      Persistable snapshot = buffer.getSnapshot();
      Serializable id = snapshot.getId();
      
      // determine whether the dirty share manager is already tracking changes to this record from
      // a previous insert or update.  If not, begin tracking the record, then update it. This set
      // of calls is assumed to be "atomic", since we must have an exclusive lock on the record.
      if (manager.isTracked(entity, id))
      {
         manager.update(entity,
                        id,
                        properties,
                        extIndexes,
                        values,
                        validatedUniqueIndexes,
                        transientIndexValidated,
                        false);
      }
      else
      {
         Persistable copy = (Persistable) snapshot.deepCopy();
         manager.update(entity,
                        copy,
                        properties,
                        extIndexes,
                        values,
                        validatedUniqueIndexes,
                        false);
      }
      
      // Track update in local context.
      trackChange(entity, id);
   }
   
   /**
    * Track an update to one or more indexed, DMO properties, as indicated by the 
    * {@code properties}, {@code extIndexes} and {@code values} arguments.
    * <p>
    * It is assumed the record identified by <code>entity</code> and
    * <code>id</code> already is present in the dirty database due to a
    * previous update or insert event.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted changes.
    * @param   id
    *          Primary key of the target record.
    * @param   properties
    *          Array of names of those DMO properties which were updated.
    *          These should represent columns which participate in one or more indexes.
    * @param   extIndexes
    *          In case of EXTENT fields this array contains their index. Otherwise -1.
    * @param   values
    *          Updated DMO property values.  Each value in the array
    *          corresponds positionally with its matching property name in the
    *          <code>properties</code> array.
    * @param   lock
    *          <code>true</code> to acquire write locks to update the record;
    *          <code>false</code> to assume write locks are acquired by calling code.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void update(String entity,
                      Serializable id,
                      String[] properties,
                      int[] extIndexes,
                      BaseDataType[] values,
                      boolean lock)
   throws PersistenceException
   {
      manager.update(entity, id, properties, extIndexes, values, null, false, lock);
   }
   
   /**
    * Overwrite the state of the records in the dirty database which match the
    * IDs of all DMOs for the given entities, which have been modified by the
    * current context.
    * <p>
    * This is necessary to support a quirk/bug in Progress which compromises
    * transaction isolation by leaking uncommitted changes (even non-indexed
    * data) under certain conditions.
    * 
    * @param   entities
    *          List of entities, which are the names of the DMO implementation
    *          classes associated with tables being tracked for uncommitted
    *          changes.  Only those records which match these entities will
    *          have their current state leaked.
    * @param   persistence
    *          Persistence service object to be used to retrieved the latest
    *          version of the target records in the current context.
    * 
    * @return  List of all DMOs whose snapshots were updated with the dirty
    *          share manager.  The list may be empty, but will not be
    *          <code>null</code>.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public List<Persistable> updateSnapshots(List<String> entities, Persistence persistence)
   throws PersistenceException
   {
      if (changes == null || dirtySnapshots == null)
      {
         // Nothing to update.
         return Collections.emptyList();
      }
      
      List<Persistable> allDMOs = null;
      
      for (String entity : entities)
      {
         Set<Serializable> changed = changes.get(entity);
         if (changed == null)
         {
            continue;
         }
         
         int size = changed.size();
         if (size > 0)
         {
            if (allDMOs == null)
            {
               allDMOs = new ArrayList<>();
            }
            
            List<Persistable> entityDMOs = new ArrayList<>(size);
            
            Iterator<Serializable> iter = changed.iterator();
            for (int i = 0; iter.hasNext(); i++)
            {
               Serializable id = iter.next();
               RecordIdentifier ident = new RecordIdentifier(entity, id);
               if ((unvalidatedInserts == null || !unvalidatedInserts.containsKey(ident)) &&
                   dirtySnapshots.remove(ident))
               {
                  Persistable dmo = persistence.quickLoad(ident);
                  entityDMOs.add(dmo);
                  allDMOs.add(dmo);
               }
            }
            
            size = entityDMOs.size();
            if (size > 0)
            {
                Persistable[] dmos = entityDMOs.toArray(new Persistable[size]);
                
                manager.updateSnapshots(entity, dmos);
                
                if (leakedEntities == null)
                {
                   leakedEntities = new HashSet<>();
                }
                
                leakedEntities.add(entity);
            }
         }
      }
      
      if (allDMOs == null)
      {
         allDMOs = Collections.emptyList();
      }
      
      return allDMOs;
   }
   
   /**
    * Mark a record as having been deleted in an uncommitted transaction, or
    * unmark a record which previously was so marked.
    * <p>
    * If marking a record as deleted, that was earlier inserted by the current
    * context, rollback the insert, but do not mark the record as deleted.
    * <p>
    * A record is unmarked when the adding context's current transaction is
    * committed or rolled back.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key of deleted record.
    * @param   rollback
    *          If <code>false</code>, the target record is marked as deleted;
    *          if <code>true</code>, it is unmarked.
    * @param   lock
    *          <code>true</code> to acquire write locks to delete the record;
    *          <code>false</code> to assume write locks are acquired by
    *          calling code.
    * 
    * @throws  PersistenceException
    *          if a database error occurs rolling back an earlier insert.
    */
   public void delete(String entity, Serializable id, boolean rollback, boolean lock)
   throws PersistenceException
   {
      RecordIdentifier ident = new RecordIdentifier(entity, id);
      
      if (!rollback)
      {
         // Special case 1:  if we are deleting a record we previously
         // inserted in the same transaction, remove the record, but do not
         // update the shared deleted records state.
         if (inserts != null && inserts.contains(ident))
         {
            rollbackInsert(ident, lock);
            if (deadInserts == null)
            {
               deadInserts = new HashSet<>(8);
            }
            deadInserts.add(ident);
            
            return;
         }
         
         // Special case 2:  if we are deleting a record we previously
         // updated in the same transaction, remove the record, but still
         // update the shared deleted records state.
         if (changes != null)
         {
            Set<Serializable> changed = changes.get(entity);
            if (changed != null && changed.remove(id))
            {
               if (changed.isEmpty())
               {
                  changes.remove(entity);
               }
               dirtySnapshots.remove(ident);
            }
         }
      }
      
      // Update shared state.
      try
      {
         manager.delete(entity, id, rollback, lock);
      }
      finally
      {
         // Track state locally, so we can clean up at end of transaction.
         if (rollback)
         {
            if (deletes != null)
            {
               Set<Serializable> ids = deletes.get(entity);
               if (ids != null)
               {
                  if (ids.remove(id) && ids.isEmpty())
                  {
                     deletes.remove(entity);
                  }
               }
            }
         }
         else
         {
            Set<Serializable> ids = null;
            if (deletes == null)
            {
               deletes = new HashMap<>();
            }
            else
            {
               ids = deletes.get(entity);
            }
            if (ids == null)
            {
               ids = new HashSet<>();
               deletes.put(entity, ids);
            }
            ids.add(id);
         }
      }
   }
   
   /**
    * Indicate whether there are currently any uncommitted changes (inserts,
    * updates, or deletes) in the table represented by the given DMO entity
    * name.
    * 
    * @param   entity
    *          Fully qualified name of the DMO entity to be checked.
    * 
    * @return  <code>true</code> if there any pending, uncommitted changes in
    *          the backing table, else <code>false</code>.
    */
   public boolean isEntityDirty(String entity)
   {
      int dirtyState = manager.getEntityDirtyState(entity);
      
      return (dirtyState != DirtyInfo.NONE);
   }
   
   /**
    * Indicate whether the record specified by the given entity and primary key has
    * been deleted within an uncommitted transaction within a context, optionally
    * ignoring deletes in the current context.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key of the record to be checked.
    * @param   ignoreLocal
    *          Ignore a delete if it occurred in the current context and only report
    *          whether a delete of the given record occurred in a different context.
    * 
    * @return  {@code true} if the specified record has been deleted, else {@code false}.
    */
   public boolean isDirtyDelete(String entity, Serializable id, boolean ignoreLocal)
   {
      if (!manager.isDeleted(entity, id))
      {
         return false;
      }
      
      if (ignoreLocal)
      {
         // It is marked deleted in the shared space, but if this context is the
         // one that deleted it, return false.
         Set<Serializable> deleted = null;
         if (deletes != null)
         {
            deleted = deletes.get(entity);
         }
         
         return (deleted == null || !deleted.contains(id));
      }
      
      return true;
   }
   
   /**
    * Retrieve from the dirty database the record specified by the given
    * entity and primary key, if it exists, and if it was not put there by the
    * current context.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key of the record to be retrieved.
    * 
    * @return  Target record if it exists in the dirty database, and was not
    *          put there by the current context, else <code>null</code>.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public Persistable getDirtyDMO(String entity, Serializable id)
   throws PersistenceException
   {
      // If we're the context making the change, we don't need the dirty
      // copy, since it's the same as the one we have in our current session.
      if (changes != null)
      {
         Set<Serializable> changed = changes.get(entity);
         if (changed != null && changed.contains(id))
         {
            return null;
         }
      }
      
      return manager.getDirtyDMO(entity, id);
   }
   
   /**
    * Execute an HQL query and return the results as a list.  If the query returns multiple
    * results per row, each element in the list will be an array of <code>Object</code>s.
    * <p>
    * No locking is attempted.
    *
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation class associated
    *          with a table being tracked for uncommitted changes. Should be <code>null</code>
    *          for a multi-table query.
    * @param   hql
    *          HQL query statement passed to Hibernate. This method makes no assumptions as to
    *          the types of object(s) returned.
    * @param   params
    *          Substitution values for the query. If none, this should be an empty array.
    * @param   types
    *          Hibernate types for the <code>values</code> array.  If none, this should be an
    *          empty array.
    * @param   maxResults
    *          The maximum number of elements to be returned in the list. If this value is
    *          non-positive, no upper limit is applied.
    * @param   startOffset
    *          The 0-based offset of the first record to retrieve.  If this value is non-positive,
    *          an offset of 0 is used by default.
    * @param   readOnly
    *          <code>true</code> to execute the query in read-only mode, else <code>false</code>.
    *
    * @return  A list of results, or <code>null</code> if no result was found.
    *
    * @throws  PersistenceException
    *          if there was an error executing the query.
    */
   public List list(String entity,
                    String hql,
                    Object[] params,
                    Type[] types,
                    int maxResults,
                    int startOffset,
                    boolean readOnly)
   throws PersistenceException
   {
      // if a specific entity is provided, check its dirty state to provide a quick out for the
      // common case
      if (entity != null)
      {
         int dirtyState = manager.getEntityDirtyState(entity);
         if ((dirtyState & DirtyInfo.CHANGE) == DirtyInfo.NONE)
         {
            return null;
         }
      }
      
      // query the dirty share manager
      return manager.list(hql, params, types, maxResults, startOffset, readOnly);
   }
   
   /**
    * Given certain search criteria and possibly a potential DMO match in the
    * primary database, report any significant information currently being
    * tracked for uncommitted transactions, which might override data (or the
    * lack thereof) found in the primary database.
    * <p>
    * If a {@link DirtyInfo} object is returned by this method, it will
    * contain overriding information, such as a dirty DMO which matched the
    * search criteria, whether that record was newly inserted, or the fact
    * that the candidate record found in the primary search has been deleted
    * or modified in an uncommitted transaction.  A <code>null</code> return
    * indicates no overriding information was available.
    * 
    * @param   buffer
    *          Record buffer associated with the current query.
    * @param   index
    *          Name of index which governs the ordering of data in the current query.
    * @param   queryType
    *          Type of query.
    * @param   bundle
    *          An object containing one or more HQL statements to be executed
    *          against the dirty database to find a record which matches the
    *          current query criteria.  The list is in order of most specific
    *          to least specific search criteria.
    * @param   params
    *          Query substitution parameters which match placeholders in the
    *          most specific HQL statement.
    * @param   types
    *          Hibernate data types of the query substitution parameters.
    * @param   found
    *          DMO, if any, found in the primary database, which matches the
    *          current query criteria.
    * @param   byRowid
    *          The predicate of the query does a lookup for ROWID/RECID of the table. In case of
    *          transient records they are accessible without the need of having a full matched
    *          index.
    *
    * @return  An object which summarizes the findings, or <code>null</code>
    *          if no uncommitted change information was available for the
    *          given arguments.
    *
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public DirtyInfo getDirtyInfo(RecordBuffer buffer,
                                 String index,
                                 int queryType,
                                 HQLBundle bundle,
                                 Object[] params,
                                 Type[] types,
                                 Persistable found,
                                 boolean byRowid)
   throws PersistenceException
   {
      String entity = buffer.getEntityName();
      int dirtyState = manager.getEntityDirtyState(entity);
      
      // quick out (hopefully the common case): if there are no changes pending for the target
      // entity, we can bypass the check altogether
      if (dirtyState == DirtyInfo.NONE)
      {
         return null;
      }
      
      Serializable foundID = (found != null ? found.getId() : null);
      
      // if we found a record in the primary database with a unique find, and we already are
      // tracking that record for changes/insert in this context (and it is validated), the
      // primary database in this session already reflects reality, and we don't need to check
      // the dirty database
      if (foundID != null &&
          queryType == QueryConstants.UNIQUE &&
          (changes != null && changes.containsKey(entity)) &&
          (unvalidatedInserts == null ||
           !unvalidatedInserts.containsKey(new RecordIdentifier(entity, foundID))))
      {
         return null;
      }
      
      // has the record found in primary database been deleted?
      if ((dirtyState & DirtyInfo.DELETE) != DirtyInfo.NONE &&
          foundID != null &&
          isDirtyDelete(entity, foundID, false))
      {
         DirtyInfo info = new DirtyInfo();
         info.setDeleted();
         
         return info;
      }
      
      DirtyInfo info = null;
      
      if ((dirtyState & DirtyInfo.CHANGE) != DirtyInfo.NONE)
      {
         List<String> statements = bundle.getDirtyStatements();
         
         info = manager.getDirtyInfo(entity, foundID, index, statements, params, types, byRowid);
         
         if (info == null)
         {
            return null;
         }
         
         Persistable dirtyDMO = info.getDirtyDMO();
         if (dirtyDMO != null && changes != null)
         {
            // ignore validated changes that are driven by this context; they are already
            // reflected in the primary database
            Serializable id = dirtyDMO.getId();
            RecordIdentifier ident = new RecordIdentifier(entity, id);
            Set<Serializable> changed = changes.get(entity);
            if (changed != null &&
                changed.contains(id) &&
                (unvalidatedInserts == null || !unvalidatedInserts.containsKey(ident)))
            {
               return null;
            }
            
            // check whether the dirty DMO represents an unvalidated record being managed by this
            // context; if so, replace the copy generated by the dirty database with the actual
            // record being managed for the primary database
            if (unvalidatedInserts != null)
            {
               RecordBuffer creatingBuffer = unvalidatedInserts.get(ident);
               if (creatingBuffer != null && creatingBuffer != buffer)
               {
                  Persistable localDMO = creatingBuffer.getCurrentRecord();
                  info.setLocal();
                  info.setDirtyDMOs(localDMO, localDMO);
               }
            }
         }
      }
      
      return info;
   }
   
   /**
    * Clean up all information contributed to the dirty share manager by this
    * context for the current transaction.  This method is called when a full
    * database transaction has been committed or rolled back against the
    * primary database.  It is also invoked when the client context ends.  It
    * removes all records it added to the dirty database for new inserts,
    * updates, and deletes.
    * 
    * @throws  StopConditionException
    *          if a critical database error occurs that would leave the dirty
    *          database in an improper state.
    */
   public void cleanup()
   {
      try
      {
         if (changes != null)
         {
            Set<String> empty = Collections.emptySet();
            Set<String> leaked = (leakedEntities != null ? leakedEntities : empty);
            manager.cleanupChanges(changes, inserts, leaked);
         }
         if (deletes != null)
         {
            manager.cleanupDeletes(deletes);
         }
      }
      catch (PersistenceException exc)
      {
         throw new StopConditionException(exc);
      }
      finally
      {
         changes = null;
         inserts = null;
         deadInserts = null;
         unvalidatedInserts = null;
         deletes = null;
         dirtySnapshots = null;
         leakedEntities = null;
      }
   }
   
   /**
    * Remember that the record represented by the given entity and primary key
    * ID has been modified in the current context.  This is necessary for
    * proper cleanup at the end of the current transaction.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key ID of the modified record.
    */
   private void trackChange(String entity, Serializable id)
   {
      if (changes == null)
      {
         changes = new HashMap<>();
      }
      
      if (dirtySnapshots == null)
      {
         dirtySnapshots = new HashSet<>();
      }
      
      Set<Serializable> ids = changes.get(entity);
      if (ids == null)
      {
         ids = new HashSet<>();
         changes.put(entity, ids);
      }
      ids.add(id);
      
      RecordIdentifier ident = new RecordIdentifier(entity, id);
      dirtySnapshots.add(ident);
   }
}