/*
** Module   : GlobalChangeEvent.java
** Abstract : A DMO change event which is shared across sessions
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20080815 ADD  @39550  Created initial version. A DMO change event
**                               which is shared across sessions.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.event;

import java.io.*;
import java.util.*;
import com.goldencode.p2j.persist.*;

/**
 * A DMO change event which is shared across sessions.
 */
public final class GlobalChangeEvent
extends RecordChangeEvent
{
   /** Globally unique identifier */
   private final long eventID;
   
   /** DMO entity name */
   private final String entity;
   
   /** DMO primary key ID */
   private final Serializable primaryKey;

   /**
    * Constructor used for an update event.
    *
    * @param   eventID
    *          Globally unique event identifier.
    * @param   entity
    *          DMO entity name.
    * @param   dmo
    *          <b>Post-modification</b> state of the updated DMO.
    * @param   properties
    *          A list of the names of properties which were modified.
    */
   public GlobalChangeEvent(long eventID,
                            String entity,
                            Persistable dmo,
                            String[] properties)
   {
      super(Type.UPDATE, null, dmo, null, makePropMap(properties));
      
      this.eventID = eventID;
      this.entity = entity;
      this.primaryKey = (dmo == null ? null : dmo.getId());
   }
   
   /**
    * Constructor used for an insert event.
    *
    * @param   eventID
    *          Globally unique event identifier.
    * @param   entity
    *          DMO entity name.
    * @param   dmo
    *          The inserted DMO.
    */
   public GlobalChangeEvent(long eventID, String entity, Persistable dmo)
   {
      super(Type.INSERT, null, dmo, null, makePropMap(null));
      
      this.eventID = eventID;
      this.entity = entity;
      this.primaryKey = (dmo == null ? null : dmo.getId());
   }
   
   /**
    * Constructor used for a delete event.
    *
    * @param   eventID
    *          Globally unique event identifier.
    * @param   entity
    *          DMO entity name.
    * @param   primaryKey
    *          Primary key ID of the deleted record.
    */
   public GlobalChangeEvent(long eventID,
                            String entity,
                            Serializable primaryKey)
   {
      super(Type.DELETE, null, null, null, makePropMap(null));
      
      this.eventID = eventID;
      this.entity = entity;
      this.primaryKey = primaryKey;
   }
   
   /**
    * Generate a map of property names to nulls.  This is intended to be used
    * for the <code>properties</code> (property name to array index values)
    * parameter required by the superclass' constructor.
    * 
    * @param   props
    *          Property names for which to generate a map.
    * 
    * @return  A hash map whose keys are the given property names, and whose
    *          values are all <code>null</code>.
    */
   private static Map<String, List<Integer>> makePropMap(String[] props)
   {
      if (props == null)
      {
         return Collections.emptyMap();
      }
      
      Map<String, List<Integer>> map = new HashMap<String, List<Integer>>();
      for (String next : props)
      {
         map.put(next, null);
      }
      
      return map;
   }
   
   /**
    * Get the globally unique identifier for this event.
    * 
    * @return  Event ID.
    */
   public long getEventID()
   {
      return eventID;
   }
   
   /**
    * Get the DMO type associated with this event.  This corresponds with the
    * DMO implementation class' fully qualified name.
    * 
    * @return  DMO entity name.
    */
   public String getEntity()
   {
      return entity;
   }
   
   /**
    * Get the primary key ID of the DMO affected by this event.
    * 
    * @return  Primary key ID.
    */
   public Serializable getPrimaryKey()
   {
      return primaryKey;
   }
}