/*
** Module   : DataTypeHelper.java
** Abstract : helper to manage data types of objects referenced by HQL
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 ECF 20080310   @37492 Created initial version. Provides static helper methods to determine
**                           the data types of various constructs found within HQL.
** 002 ECF 20080609   @38681 Fixed initBasicTokens(). Added STRING/TYPE constant mapping to 
**                          tokens map.
** 003 ECF 20080804   @39310 Added rowid support. Map Long and rowid classes, and long primitive,
**                           to HQLTypes.LONG enum.
** 004 SVL 20090225   @41366 Added recid support (mapped to HQLTypes.INTEGER).
** 005 GES 20090518   @42391 Import change.
** 006 ECF 20090817   @43683 Added persistence *Field types to classes map.
** 007 SVL 20120331          Upgraded to Hibernate 4.
** 008 OM  20130501          Added support for 64 bit data types.
** 009 OM  20130527          Added mapping for Timestamp.class as HQLTypes.DATETIME and 
**                           Object[] for HQLTypes.DATETIMETZ (don't know if this is correct).
**                           For the moment this is disabled (1584#150).
** 010 VMN 20141127          Renamed DatetimeTzField.class to DatetimetzField.class.
** 011 CA  20151025          Registered the handle.class type.
** 012 OM  20151130          Added HQLFunction support in initFunction() method.
** 013 ECF 20160114          Removed references to obsolete *Field classes.
** 014 ECF 20160226          Changes required by new PropertyHelper implementation.
** 015 CA  20160226          In a ternary, if 'then' expression is unknown, then resolve the 
**                           datatype from the 'else' expresson.
** 016 ECF 20160720          Fixed buffer lookup during HQL preprocessing.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.hql;

import java.lang.reflect.*;
import java.math.*;
import java.sql.Date;
import java.sql.Timestamp;
import java.util.*;
import org.hibernate.type.Type;
import com.goldencode.ast.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.pl.*;
import com.goldencode.p2j.util.*;

/**
 * A helper to manage data types of objects referenced by HQL statements
 * (specifically, where clauses).  This class provides static helper methods
 * to determine the data type of:
 * <ul>
 *   <li>certain non-volatile tokens (i.e., constants or other token types
 *       whose data types cannot change based on other factors)
 *   <li>standard functions, such as <code>upper()</code>
 *   <li>user defined functions, such as all the built-in functions migrated
 *       from Progress
 *   <li>query substitution parameters
 *   <li>qualified and unqualified references to DMO properties
 * </ul>
 * <p>
 * The {@link HQLTypes} enum is used to rationalize all the various type
 * constructs (token types, Java class types, Hibernate types) into a common
 * construct.  Any request for type information will be in terms of an
 * <code>HQLTypes</code> enum value.
 */
public final class DataTypeHelper
implements HQLParserTokenTypes
{
   /** Map of token types to HQL types */
   private static final Map<Integer, HQLTypes> tokens = new HashMap<>();
   
   /** Map of functions to HQL types */
   private static final Map<FunctionKey, HQLTypes> functions = new HashMap<>();
   
   /** Map of query substitution parameter Hibernate types to HQL types */
   private static final Map<Type, HQLTypes> parameters = new HashMap<>();
   
   /** Map of Java class types to HQL types */
   private static final Map<Class<?>, HQLTypes> classes = new HashMap<>();
   
   /**
    * Private constructor;  access is via static methods.
    */
   private DataTypeHelper()
   {
   }
   
   /**
    * Initialize the data structures needed to enable this helper.  This
    * should be invoked early during initialization of the persistence
    * runtime.
    */
   public static void initialize()
   {
      // Initialize Java data type wrapper mappings.
      initJavaClasses();
      
      // Initialize basic token type mappings.
      initBasicTokens();
      
      // Initialize query parameter type mappings (Hibernate types).
      initQueryParamTypes();
      
      // Initialize standard built-in function return type mappings.
      initFunctions();
   }
   
   /**
    * Get the HQL type associated with a particular Java class.
    * 
    * @param   javaType
    *          A Java class.
    * 
    * @return  The HQL type associated with the given class, or {@link
    *          HQLTypes#UNKNOWN} if no mapping is available.
    */
   public static HQLTypes getTypeClass(Class<?> javaType)
   {
      HQLTypes type = classes.get(javaType);
      
      return (type != null ? type : HQLTypes.UNKNOWN);
   }
   
   /**
    * Get the HQL type associated with a particular token type.  Only those
    * token types whose data type cannot vary (i.e., constants and certain
    * operators) are handled by this method.
    * 
    * @param   tokenType
    *          A token type from {@link HQLParserTokenTypes}.
    * 
    * @return  The HQL type associated with the given token type, or {@link
    *          HQLTypes#UNKNOWN} if no mapping is available.
    */
   public static HQLTypes getTypeToken(int tokenType)
   {
      HQLTypes type = tokens.get(tokenType);
      
      return (type != null ? type : HQLTypes.UNKNOWN);
   }
   
   /**
    * Get the HQL type associated with a particular Hibernate
    * <code>Type</code>, used to identify the data type of a query
    * substitution parameter.
    * 
    * @param   paramType
    *          A Hibernate <code>Type</code>.
    * 
    * @return  The HQL type associated with the given Hibernate type, or
    *          {@link HQLTypes#UNKNOWN} if no mapping is available.
    */
   public static HQLTypes getTypeParameter(Type paramType)
   {
      HQLTypes type = parameters.get(paramType);
      
      return (type != null ? type : HQLTypes.UNKNOWN);
   }
   
   /**
    * Get the HQL type associated with a particular {@link FunctionKey}.
    * 
    * @param   key
    *          A <code>FunctionKey</code> which describes a standard function
    *          or a built-in function.
    * 
    * @return  The HQL type associated with the given function key, or {@link
    *          HQLTypes#UNKNOWN} if no mapping is available.
    */
   public static HQLTypes getTypeFunction(FunctionKey key)
   {
      HQLTypes type = functions.get(key);
      
      return (type != null ? type : HQLTypes.UNKNOWN);
   }
   
   /**
    * Get the HQL type associated with a particular, qualified DMO property,
    * as represented by the given AST.
    * 
    * @param   aliasAst
    *          An AST which represents a DMO alias, whose first (and only)
    *          child represents a property of that DMO.
    * @param   bufferMap
    *          A map of DMO aliases to their corresponding record buffers, where {@code null}
    *          is mapped to the buffer whose property references in the HQL expression are
    *          unqualified (if any).
    * 
    * @return  The HQL type associated with the given property, or {@link
    *          HQLTypes#UNKNOWN} if no mapping is available.
    * 
    * @throws  PersistenceException
    *          if the property does not exist, or if there is some error
    *          determining its data type.
    */
   public static HQLTypes getTypeQualifiedProperty(Aast aliasAst,
                                                   Map<String, RecordBuffer> bufferMap)
   throws PersistenceException
   {
      Aast propAst = aliasAst.getChildAt(0);
      String property = propAst.getText();
      RecordBuffer buffer = lookupBuffer(aliasAst, bufferMap, null);
      Class<?> dmoIface = buffer.getDMOInterface();
      Class<?> javaType = DBUtils.getDMOPropertyType(dmoIface, property);
      
      return getTypeClass(javaType);
   }
   
   /**
    * Get the HQL type associated with a particular, unqualified DMO property,
    * as represented by the given AST.
    * 
    * @param   propAst
    *          An AST which represents a property of a DMO.
    * @param   bufferMap
    *          A map of DMO aliases to their corresponding record buffers, where {@code null}
    *          is mapped to the buffer whose property references in the HQL expression are
    *          unqualified (if any).
    * 
    * @return  The HQL type associated with the given property, or {@link
    *          HQLTypes#UNKNOWN} if no mapping is available.
    * 
    * @throws  PersistenceException
    *          if the property does not exist, or if there is some error
    *          determining its data type.
    */
   public static HQLTypes getTypeUnqualifiedProperty(Aast propAst,
                                                     Map<String, RecordBuffer> bufferMap)
   throws PersistenceException
   {
      Aast parent = propAst.getParent();
      Aast aliasAst = parent != null && parent.getType() == ALIAS ? parent : null;
      RecordBuffer buffer = lookupBuffer(aliasAst, bufferMap, null);
      Class<?> dmoIface = buffer.getDMOInterface();
      String property = propAst.getText();
      Class<?> javaType = DBUtils.getDMOPropertyType(dmoIface, property);
      
      return getTypeClass(javaType);
   }
   
   /**
    * Determine the HQL type of the expression represented by the given AST node, which may be
    * a single leaf node, or the head of a more complex expression.
    * 
    * @param   node
    *          AST node whose HQL type is to be determined.
    * @param   bufferMap
    *          A map of DMO aliases to their corresponding record buffers, where {@code null}
    *          is mapped to the buffer whose property references in the HQL expression are
    *          unqualified (if any).
    * 
    * @return  The HQL type which represents the data type to which the given
    *          expression would evaluate.
    * 
    * @throws  PersistenceException
    *          if a DMO property referenced within the expression does not
    *          exist, or if there is some error determining its data type.
    */
   public static HQLTypes expressionType(Aast node, Map<String, RecordBuffer> bufferMap)
   throws PersistenceException
   {
      HQLTypes c0;
      HQLTypes c1;
      
      int tokType = node.getType();
      switch (tokType)
      {
         case ALIAS:
            return getTypeQualifiedProperty(node, bufferMap);
            
         case PROPERTY:
            return getTypeUnqualifiedProperty(node, bufferMap);
            
         case FUNCTION:
            return getTypeFunction(node, bufferMap);
            
         case SUBST:
            return HQLTypes.valueOf((String) node.getAnnotation("datatype"));
            
         case TERNARY:
            c0 = expressionType(node.getChildAt(1), bufferMap); // case...then
            c1 = expressionType(node.getChildAt(2), bufferMap); // case...else
            if (c0 == HQLTypes.DECIMAL || c1 == HQLTypes.DECIMAL)
            {
               return HQLTypes.DECIMAL;
            }
            return (c0 == HQLTypes.LONG || c1 == HQLTypes.LONG) 
                     ? HQLTypes.LONG 
                     : (c0 == HQLTypes.UNKNOWN ? c1 : c0);
            
         case PLUS:
         case MINUS:
         case MULTIPLY:
         case DIVIDE:
            c0 = expressionType(node.getChildAt(0), bufferMap);  // lvalue
            c1 = expressionType(node.getChildAt(1), bufferMap);  // rvalue
            if (c0 == HQLTypes.DECIMAL || c1 == HQLTypes.DECIMAL)
            {
               return HQLTypes.DECIMAL;
            }
            return (c0 == HQLTypes.LONG || c1 == HQLTypes.LONG)
                   ? HQLTypes.LONG
                   : HQLTypes.INTEGER;
            
         case UN_MINUS:
            return expressionType(node.getChildAt(0), bufferMap);
            
         default:
            return getTypeToken(tokType);
      }
   }
   
   /**
    * Attempt to look up a record buffer by alias from the given map. The DMO alias is extracted
    * from the {@code alias} AST. If there is no matching entry in the buffer map, an exception
    * is thrown.
    * <p>
    * The returned buffer is initialized by side effect.
    * 
    * @param   alias
    *          DMO alias AST. If {@code null}, a {@code null} lookup key is used.
    * @param   bufferMap
    *          A map of DMO aliases to record buffers. If no alias is provided (e.g., for an
    *          unqualified property reference, the appropriate buffer is expected to be mapped
    *          to the {@code null} key.
    * @param   hql
    *          Optional HQL expression from which the alias is parsed.
    * 
    * @return  A record buffer matching the given alias.
    */
   public static RecordBuffer lookupBuffer(Aast alias,
                                           Map<String, RecordBuffer> bufferMap,
                                           HQLExpression hql)
   {
      String aliasName = alias != null ? alias.getText() : null;
      RecordBuffer buffer = bufferMap.get(aliasName);
      
      if (buffer == null)
      {
         throw new IllegalArgumentException(
            "Unrecognized buffer at column " +
            alias.getColumn() +
            " in where clause:  '" +
            aliasName +
            "' [" +
            (hql != null ? hql.toString() : "<unavailable>") +
            "]");
      }
      
      buffer.initialize();
      
      return buffer;
   }
   
   /**
    * Get the HQL type which will be returned when the given function is executed.
    * 
    * @param   function
    *          AST node representing a standard or built-in function expression.
    * @param   bufferMap
    *          A map of DMO aliases to their corresponding record buffers, where {@code null}
    *          is mapped to the buffer whose property references in the HQL expression are
    *          unqualified (if any).
    * 
    * @return  The HQL type associated with the given function, or {@link HQLTypes#UNKNOWN} if
    *          no mapping is available.
    * 
    * @throws  PersistenceException
    *          if a DMO property referenced within the expression does not exist, or if there
    *          is some error determining its data type.
    */
   private static HQLTypes getTypeFunction(Aast function, Map<String, RecordBuffer> bufferMap)
   throws PersistenceException
   {
      String name = function.getText();
      int len = function.getNumImmediateChildren();
      HQLTypes[] hqlSig = new HQLTypes[len];
      for (int i = 0; i < len; i++)
      {
         hqlSig[i] = expressionType(function.getChildAt(i), bufferMap);
      }
      FunctionKey key = new FunctionKey(name, hqlSig);
      
      return getTypeFunction(key);
   }
   
   /**
    * Initialize the mapping of Java classes to {@link HQLTypes} for those
    * classes which are known to be used in HQL where clauses.
    */
   private static void initJavaClasses()
   {
      // Java primitive types.
      classes.put(Boolean.TYPE     , HQLTypes.BOOLEAN);
      classes.put(Integer.TYPE     , HQLTypes.INTEGER);
      classes.put(Long.TYPE        , HQLTypes.LONG);
      classes.put(Double.TYPE      , HQLTypes.DOUBLE);
      
      // Standard Java wrapper types.
      classes.put(String.class     , HQLTypes.TEXT);
      classes.put(Boolean.class    , HQLTypes.BOOLEAN);
      classes.put(Integer.class    , HQLTypes.INTEGER);
      classes.put(Long.class       , HQLTypes.LONG);
      classes.put(BigDecimal.class , HQLTypes.DECIMAL);
      classes.put(Date.class       , HQLTypes.DATE);
      classes.put(Timestamp.class  , HQLTypes.DATETIME);
      //classes.put(Object[].class   , HQLTypes.DATETIMETZ); // datetimetz is represented by a 2-uple (Timestamp, Offset)
      
      // P2J wrapper types.
      classes.put(logical.class    , HQLTypes.BOOLEAN);
      classes.put(integer.class    , HQLTypes.INTEGER);
      classes.put(recid.class      , HQLTypes.INTEGER);
      classes.put(decimal.class    , HQLTypes.DECIMAL);
      classes.put(date.class       , HQLTypes.DATE);
      classes.put(raw.class        , HQLTypes.RAW);
      classes.put(character.class  , HQLTypes.TEXT);
      classes.put(rowid.class      , HQLTypes.LONG);
      classes.put(int64.class      , HQLTypes.LONG);
      classes.put(datetime.class   , HQLTypes.DATETIME);
      classes.put(datetimetz.class , HQLTypes.DATETIME);
      //classes.put(datetimetz.class , HQLTypes.DATETIMETZ);
      classes.put(handle.class     , HQLTypes.TEXT);
   }
   
   /**
    * Initialize the mapping of basic token types to {@link HQLTypes}.  Only
    * those token types whose data types will not change (i.e., constants and
    * certain operators) are mapped here.
    */
   private static void initBasicTokens()
   {
      tokens.put(BOOL_TRUE   , HQLTypes.BOOLEAN);
      tokens.put(BOOL_FALSE  , HQLTypes.BOOLEAN);
      tokens.put(IS_NULL     , HQLTypes.BOOLEAN);
      tokens.put(NOT_NULL    , HQLTypes.BOOLEAN);
      tokens.put(EQUALS      , HQLTypes.BOOLEAN);
      tokens.put(NOT_EQ      , HQLTypes.BOOLEAN);
      tokens.put(GT          , HQLTypes.BOOLEAN);
      tokens.put(LT          , HQLTypes.BOOLEAN);
      tokens.put(LTE         , HQLTypes.BOOLEAN);
      tokens.put(GTE         , HQLTypes.BOOLEAN);
      tokens.put(AND         , HQLTypes.BOOLEAN);
      tokens.put(OR          , HQLTypes.BOOLEAN);
      tokens.put(NOT         , HQLTypes.BOOLEAN);
      tokens.put(NUM_LITERAL , HQLTypes.INTEGER);
      tokens.put(LONG_LITERAL, HQLTypes.LONG);
      tokens.put(DEC_LITERAL , HQLTypes.DECIMAL);
      tokens.put(STRING      , HQLTypes.TEXT);
   }
   
   /**
    * Initialize the mapping of Hibernate <code>Type</code>s to {@link
    * HQLTypes} for those Hibernate <code>Type</code>s (including custom P2J
    * wrapper types) which are known to be used in HQL where clauses as query
    * substitution parameters.
    */
   private static void initQueryParamTypes()
   {
      Map<Class<?>, Type> class2Type = DBUtils.getHibernateTypeMap();
      for (Map.Entry<Class<?>, Type> entry : class2Type.entrySet())
      {
         Class<?> clazz = entry.getKey();
         Type paramType = entry.getValue();
         HQLTypes hqlType = classes.get(clazz);
         
         if (hqlType == null)
         {
            throw new IllegalArgumentException(
               "No HQL type found for Java type " + clazz.getName());
         }
         
         parameters.put(paramType, hqlType);
      }
   }
   
   /**
    * Initialize the mappings of {@link FunctionKey} to {@link HQLTypes} for
    * all standard functions used in HQL (including built-in functions which
    * emulate Progress functionality).
    */
   private static void initFunctions()
   {
      initFunction("upper", HQLTypes.TEXT, HQLTypes.TEXT);
      
      for (Method method : BuiltIns.getStandardFunctions())
      {
         initFunction(method);
      }
   }
   
   /**
    * Given a Java method which represents a user-defined function accessible
    * from HQL, create a mapping of {@link FunctionKey} to {@link HQLTypes} for that function.
    * 
    * @param   method
    *          Java method to be mapped.
    */
   private static void initFunction(Method method)
   {
      HQLFunction hqlFunction = method.getAnnotation(HQLFunction.class);
      if (hqlFunction == null)
      {
         throw new IllegalArgumentException(
               "Method " + method + " is not backing a hql function");
      }
      String name = hqlFunction.name();
      // use the annotation to obtain the correct function name generated in the code
      if (name.isEmpty())
      {
         name = method.getName();
         // if not explicitly set, it matches the name of method
      }
      Class<?> returns = method.getReturnType();
      HQLTypes hqlReturns = classes.get(returns);
      
      if (hqlReturns == null)
      {
         throw new IllegalArgumentException(
            "No HQL type found for user defined function return type " + returns.getName());
      }
      
      Class<?>[] sig = method.getParameterTypes();
      HQLTypes[] hqlSig = new HQLTypes[sig.length];
      for (int i = 0; i < sig.length; i++)
      {
         Class<?> next = sig[i];
         HQLTypes hqlParam = classes.get(next);
         if (hqlParam == null)
         {
            throw new IllegalArgumentException(
               "No HQL type found for user defined function parameter type " + next.getName());
         }
         hqlSig[i] = hqlParam;
      }
      
      FunctionKey key = new FunctionKey(name, hqlSig);
      functions.put(key, hqlReturns);
   }
   
   /**
    * Given the name and signature of a user-defined function accessible from
    * HQL, create a mapping of {@link FunctionKey} to {@link HQLTypes} for that function.
    * 
    * @param   name
    *          Function name
    * @param   returns
    *          Data type that the function returns.
    * @param   paramTypes
    *          Data types of function parameters, if any.
    */
   private static void initFunction(String name, HQLTypes returns, HQLTypes... paramTypes)
   {
      FunctionKey key = new FunctionKey(name, paramTypes);
      functions.put(key, returns);
   }
}