/*
** Module   : ConnectionTableUpdater.java
** Abstract : Manages metadata connect records for a particular, primary database
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20150111 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.meta;

import java.lang.reflect.*;
import java.util.*;
import java.util.logging.*;
import org.hibernate.type.Type;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.util.*;
import com.goldencode.proxy.*;

/**
 * A {@link com.goldencode.p2j.persist.meta.ConnectionListener database connection listener
 * implementation} which receives notifications of database connection events (at a Progress
 * level, not JDBC level) and updates the metadata connect table accordingly. This involves the
 * creation, lookup, and deletion of metadata connect records. Since such events are minimized in
 * real applications (it is a relatively expensive operation), this work is done inline (as
 * opposed to offloaded into a worker thread).
 */
public final class ConnectTableUpdater
implements ConnectionListener
{
   /** Logger */
   private static final Logger log = LogHelper.getLogger(ConnectTableUpdater.class);
   
   /** Metadata connect table name */
   private static final String CONNECT_TABLE = "meta_connect";
   
   /** HQL query to find a connect record by connect ID */
   private static final String HQL_CONNECT = "from MetaConnectImpl m where m.connectId = ?0";
   
   /** Parameter types for connect record query */
   private static final Type[] TYPES_CONNECT =
      DBUtils.makeTypeArray(new Object[] { new int64() });
   
   /** MetaConnectImpl DMO implementation class */
   private static final Class<?> dmoClass;
   
   /** Map of <code>MinimalConnect</code> methods to <code>MetaConnectImpl</code> methods */
   private static final Map<Method, Method> methodMap = new HashMap<>();
   
   static
   {
      String clsName = String.format("%s._meta.impl.MetaConnectImpl", DMOIndex.getBasePackage());
      try
      {
         dmoClass = Class.forName(clsName);
         
         // map MinimalConnect methods to corresponding dmoClass methods; if we try to use
         // MinimalConnect methods with a MetaConnectImpl object, we will raise
         // IllegalArgumentException during Method.invoke(), because our object will not be an
         // instance of MinimalConnect
         for (Method ifcMeth : MinimalConnect.class.getMethods())
         {
            Method dmoMeth = dmoClass.getMethod(ifcMeth.getName(),
                                                ifcMeth.getParameterTypes());
            methodMap.put(ifcMeth, dmoMeth);
         }
      }
      catch (ClassNotFoundException | NoSuchMethodException exc)
      {
         throw new RuntimeException("Error preparing MetaConnect implementation class", exc);
      }
   }
   
   /** Persistence helper object for metadata database */
   private final Persistence persistence;
   
   /** Proxy object for the current metadata connect DMO */
   private final MinimalConnect proxy;
   
   /** Invocation handler which maps proxy method calls to underlying DMO methods */
   private final Handler handler = new Handler();
   
   /** Object on which connect/disconnect updates are synchronized */
   private final Object lock = new Object();
   
   /**
    * Constructor which stores a persistence object for the associated metadata database, and
    * initializes a proxy for the application-specific DMO.
    * 
    * @param   metaDB
    *          Metadata database helper object.
    */
   public ConnectTableUpdater(Database metaDB)
   {
      this.persistence = PersistenceFactory.getInstance(metaDB);
      this.proxy = ProxyFactory.getProxy(dmoClass,
                                         new Class<?>[] { MinimalConnect.class },
                                         handler);
   }
   
   /**
    * Create a new connect record in the metadata database, in response to a connection event.
    * 
    * @param   connectID
    *          Unique identifier for the database connection.
    */
   public void databaseConnected(int64 connectID)
   {
      // fairly coarse synchronization should be OK here; if this becomes a bottleneck, we can
      // replace this with a lock manager
      synchronized (lock)
      {
         Persistable dmo = null;
         boolean commit = false;
         
         try
         {
            commit = persistence.beginTransaction();
            
            // create new DMO instance and back the proxy with it
            dmo = (Persistable) dmoClass.newInstance();
            handler.setDelegate(dmo);
            
            // assign primary key
            Long key = (Long) persistence.nextPrimaryKey(CONNECT_TABLE);
            dmo.setId(key);
            
            // assign connect ID (required by unique index)
            proxy.setConnectId(connectID);
            
            // assign fields with values retrieved from client parameters
            ClientParameters parms = StandardServer.getClientParameters();
            String name = parms.osUserName != null ? parms.osUserName : "";
            proxy.setConnectName(new character(name));
            String device = parms.terminalName != null ? parms.terminalName : "";
            proxy.setConnectDevice(new character(device));
            proxy.setConnectPid(new integer(parms.pid));
            
            // assign other fields which are mandatory and default to unknown
            character emptyString = new character("");
            proxy.setConnectBatch(emptyString);
            proxy.setConnectTime(emptyString);
            proxy.setConnectUsr(new integer(0));
            
            // save the new record
            persistence.save(dmo, key);
            
            // flush explicitly
            persistence.flush();
            
            // commit
            if (commit)
            {
               commit = false;   // to avoid double-rollback in catch block if commit fails
               persistence.commit();
            }
         }
         catch (Exception exc)
         {
            try
            {
               if (commit)
               {
                  persistence.rollback();
               }
            }
            catch (PersistenceException pe)
            {
               // error already logged in finally block
            }
            finally
            {
               if (log.isLoggable(Level.SEVERE))
               {
                  String msg = String.format("Error inserting into _connect table for %s (%s)",
                                             persistence.getDatabase().toString(),
                                             connectID.toStringMessage());
                  log.log(Level.SEVERE, msg, exc);
               }
            }
         }
      }
   }
   
   /**
    * Remove the connect record with the given ID from the metadata database, in response to a
    * disconnection event.
    * 
    * @param   connectID
    *          Unique identifier for the database connection.
    */
   public void databaseDisconnected(int64 connectID)
   {
      // fairly coarse synchronization should be OK here; if this becomes a bottleneck, we can
      // replace this with a lock manager
      synchronized (lock)
      {
         Persistable dmo = null;
         boolean commit = false;
         
         try
         {
            commit = persistence.beginTransaction();
            
            // retrieve DMO with the given connect ID
            Object[] args = new Object[] { connectID };
            List<Persistable> list = persistence.list(HQL_CONNECT,
                                                      args,
                                                      TYPES_CONNECT,
                                                      1,
                                                      0,
                                                      true);
            if (list == null)
            {
               throw new PersistenceException("Unable to retrieve connect metadata record " +
                                              connectID.toStringMessage());
            }
            
            // delete the record
            dmo = list.get(0);
            persistence.delete(dmo);
            
            // flush explicitly
            persistence.flush();
            
            // commit
            if (commit)
            {
               commit = false;   // to avoid double-rollback in catch block if commit fails
               persistence.commit();
            }
         }
         catch (Exception exc)
         {
            try
            {
               if (commit)
               {
                  persistence.rollback();
               }
            }
            catch (PersistenceException pe)
            {
               // error already logged in finally block
            }
            finally
            {
               if (log.isLoggable(Level.SEVERE))
               {
                  String msg = String.format("Error deleting from _connect table for %s (%s)",
                                             persistence.getDatabase().toString(),
                                             connectID.toStringMessage());
                  log.log(Level.SEVERE, msg, exc);
               }
            }
         }
      }
   }
   
   /**
    * Invocation handler which delegates calls to the {@link MinimalConnect} proxy methods to the
    * backing DMO instance.
    */
   private static class Handler
   implements InvocationHandler
   {
      /** DMO delegate whose methods ultimately are invoked */
      private Object delegate = null;
      
      /**
       * Process method invocations on the {@link MinimalConnect} proxy by delegating them to the
       * backing DMO instance.
       * 
       * @param   proxy
       *          Proxy on which the enclosing class invokes methods.
       * @param   method
       *          Method which was invoked.
       * @param   args
       *          Method arguments.
       * 
       * @throws  Throwable
       *          if the delegated method call throws an exception.
       */
      @Override
      public Object invoke(Object proxy, Method method, Object[] args)
      throws Throwable
      {
         Method dmoMethod = methodMap.get(method);
         
         return dmoMethod.invoke(delegate, args);
      }
      
      /**
       * Set the backing DMO delegate instance.
       * 
       * @param   delegate
       *          DMO whose methods will be invoked.
       */
      void setDelegate(Object delegate)
      {
         this.delegate = delegate;
      }
   }
   
   /**
    * Interface which defines the minimally required methods of the <code>MetaConnect</code>
    * interface. This interface defines the proxy used to create and manipulate metadata connect
    * DMO instances. The DMO's methods cannot be invoked directly from code here, since the
    * <code>MetaConnect</code> interface resides in an application-specific package, which is not
    * known at compile time.
    */
   public interface MinimalConnect
   extends Persistable
   {
      /**
       * Setter: connectId
       * 
       * @param   connectId
       *          connectId
       */
      public void setConnectId(NumberType connectId);
      
      /**
       * Setter: User-Id
       * 
       * @param   connectUsr
       *          User-Id
       */
      public void setConnectUsr(NumberType connectUsr);
      
      /**
       * Setter: Name
       * 
       * @param   connectName
       *          Name
       */
      public void setConnectName(Text connectName);
      
      /**
       * Setter: Device
       * 
       * @param   connectDevice
       *          Device
       */
      public void setConnectDevice(Text connectDevice);
      
      /**
       * Setter: Login time
       * 
       * @param   connectTime
       *          Login time
       */
      public void setConnectTime(Text connectTime);
      
      /**
       * Setter: Pid
       * 
       * @param   connectPid
       *          Pid
       */
      public void setConnectPid(NumberType connectPid);
      
      /**
       * Setter: Batch User
       * 
       * @param   connectBatch
       *          Batch User
       */
      public void setConnectBatch(Text connectBatch);
   }
}