/*
** Module   : RemoteIdentityManager.java
** Abstract : IdentityManager implementation which delegates to a remote
**            server.
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 SVL 20080111 ADD  @36796  Created initial version. IdentityManager
**                               implementation which delegates to a remote
**                               server.
** 002 ECF 20080806 CHG  @39307  Modified to conform to new IdentityManager
**                               API.
** 003 SVL 20080901 CHG  @39649  Added reclaimKeys() function. 
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.remote;

import java.io.Serializable;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.id.*;

/**
 * A specialized implementation of {@link
 * com.goldencode.p2j.persist.id.IdentityManager IdentityManager} which
 * delegates requests for database record identity services to the remote P2J
 * server which is authoritative for the target database. This is necessary
 * in cases where business logic dynamically connects to a remote database
 * which is not managed by the current P2J server instance, since unique
 * primary key generation action must be serialized through a identity manager
 * which is authoritative for the target database.
 * <p>
 * An instance of this class works with a local proxy for a remote instance of
 * {@link IdentityManagerMultiplexer} residing at the target server. The
 * remote object may manage multiple databases, each of which is assigned a
 * unique database identifier. When the database is {@link
 * #setPersistence(Persistence) set} into a <code>RemoteIdentityManager</code>
 * instance, the associated identifier is requested from the remote
 * multiplexer.  The ID returned by the multiplexer is used on all subsequent
 * service requests sent by the <code>IdentityManager</code> instance.
 *
 * @see     RemotePersistence
 * @author  SVL
 */
public class RemoteIdentityManager
implements IdentityManager
{
   /** Identity manager multiplexer which will service remote requests */
   private final IdentityManagerMultiplexer multiplexer;

   /** Multiplex ID for the remote database associated with this object */
   private int mpID = -1;

   /**
    * Constructor which stores a proxy to the remote object which will
    * service identity manager requests for the remote database.
    *
    * @param   multiplexer
    *          Identity manager multiplexer which will service remote
    *          requests.
    */
   public RemoteIdentityManager(IdentityManagerMultiplexer multiplexer)
   {
      this.multiplexer = multiplexer;
   }

   /**
    * Set the name of the physical database with which this identity manager
    * is associated.
    *
    * @param   persistence
    *          Persistence service object.
    */
   public void setPersistence(Persistence persistence)
   {
      if (mpID >= 0)
      {
         throw new IllegalStateException("Database can only be set once");
      }
      
      Database database = persistence.getDatabase();
      mpID = multiplexer.getMultiplexID(database.getName());
   }

   /**
    * Return the next primary key for the specified table.
    *
    * @param   table
    *          Table for which the next primary key will be returned.
    *
    * @return  The next primary key for the given table.
    *
    * @throws  PersistenceException
    *          if there is an error determining the next primary key ID.
    */
   public Serializable nextPrimaryKey(String table)
   throws PersistenceException
   {
      return multiplexer.nextPrimaryKey(mpID, table);
   }

   /**
    * Inform the identity manager that the given primary keys can be reused.
    *
    * @param keys
    *        Keys for reuse.
    */
   public void reclaimKeys(Serializable[] keys)
   {
      multiplexer.reclaimKeys(mpID, keys);
   }
}