/*
** Module   : RawUserType.java
** Abstract : Custom user type to wrap P2J raw for Hibernate access
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20050908   @22602 Created initial version. Maps P2J raw to SQL
**                           byte array. Replaces Hibernate binary type as
**                           the mapping for Progress raw values.
** 002 ECF 20051025   @23169 Added isMutable method. Overrides superclass'
**                           implementation to return true, reflecting the
**                           DMO access strategy for raw wrapper user
**                           types.
** 003 ECF 20080806   @39300 Moved to type sub-package of persist.
** 004 SVL 20120331          Upgraded to Hibernate 4.
** 005 GES 20130523          Match signature changes in raw.
** 006 ECF 20140220          Replaced Apache commons logging with J2SE standard logging.
** 007 ECF 20140624          Refactored SQL_TYPES to use TypeHelper.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist.type;

import java.sql.*;
import java.util.logging.*;
import com.goldencode.p2j.util.*;
import org.hibernate.engine.spi.*;

/**
 * Implementation of a custom, Hibernate user type for the {@link
 * com.goldencode.p2j.util.raw P2J raw} data type. This permits Hibernate to
 * convert this data type to and from the SQL VARBINARY data type. The
 * unknown value is mapped to SQL NULL.
 */
public final class RawUserType
extends AbstractWrapperUserType
{
   /** Single element array containing the SQL type constant being mapped */
   private static final int[] SQL_TYPES = { TypeHelper.getSqlType(raw.class) };
   
   /**
    * Returns <code>true</code> to indicate that the P2J data wrapper object
    * is mutable.  DMOs which get and set these objects create a non-final
    * reference.  Getters return the reference directly, and setters replace
    * the reference with that of the object provided by the caller.  Hence,
    * changes to the DMO's state can be made by changing the wrapper's state
    * directly, necessitating dirty checking by Hibernate.
    *
    * @return  <code>true</code>.
    */
   public boolean isMutable()
   {
      return true;
   }
   
   /**
    * Get the array of SQL data type constants (in this case only a single
    * constant), which this classes uses to map a raw value to/from a
    * database record.
    *
    * @return  Array containing SQL type constant.
    */
   public int[] sqlTypes()
   {
      return SQL_TYPES;
   }
   
   /**
    * Gets the Java data type class mapped by this custom type.
    *
    * @return  The {@link com.goldencode.p2j.util.raw P2J raw} class.
    */
   public Class returnedClass()
   {
      return raw.class;
   }
   
   /**
    * Create and return a deep copy of the given <code>raw</code> instance.
    *
    * @param   value
    *          The <code>raw</code> instance to be copied.
    *
    * @return  A new instance of <code>raw</code> which contains the same
    *          internal value as <code>value</code>.
    */
   public Object deepCopy(Object value)
   {
      return ((raw) value).duplicate();
   }
   
   /**
    * Create an instance of the data type mapped by this class, based upon
    * a data value in the specified result set. If the corresponding data
    * value is NULL in the result set, return a <code>raw</code> instance
    * which is initialized to the unknown value.
    *
    * @param   rs
    *          Result set positioned at the record from which we will access
    *          a <code>byte</code> array.
    * @param   names
    *          Array of column names. In this case, it will contain only one
    *          element, which represents the name of the column from which
    *          we will access our data.
    * @param   session
    *          Not used.
    * @param   owner
    *          Owning entity; not used.
    *
    * @return  A <code>raw</code> instance, possibly initialized to the
    *          unknown value.
    *
    * @throws  SQLException
    *          if an error occurs accessing the result set.
    */
   public Object nullSafeGet(ResultSet rs,
                             String[] names,
                             SessionImplementor session,
                             Object owner)
   throws SQLException
   {
      byte[] value = rs.getBytes(names[0]);
      raw wrapper = null;
      
      if (rs.wasNull())
      {
         if (LOG_TRACE)
         {
            LOG.log(Level.FINEST, "returning unknown value as column: " + names[0]);
         }
         wrapper = new raw();
         wrapper.setUnknown();
      }
      else
      {
         wrapper = new raw(value);
         if (LOG_TRACE)
         {
            LOG.log(Level.FINEST, "returning '" + toString(wrapper) + "' as column: " + names[0]);
         }
      }
      
      return wrapper;
   }
   
   /**
    * Set a <code>byte</code> array value into the given prepared statement,
    * based upon the <code>raw</code> instance <code>value</code>.
    *
    * @param   ps
    *          Prepared statement into which we will set our data.
    * @param   value
    *          A <code>raw</code> instance.
    * @param   index
    *          Index in the statement at which to write the data.
    *
    * @throws  SQLException    
    *          if an error occurs setting data into the statement.
    */
   protected void bindNonNull(PreparedStatement ps,
                              BaseDataType value,
                              int index)
   throws SQLException
   {
      ps.setBytes(index, ((raw) value).getByteArray());
   }
}