/*
** Module   : ReportDriver.java
** Abstract : drives the pattern engine based on a list of report definitions
**
** Copyright (c) 2007-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 GES 20070514   @33498 Created initial version that provides
**                           support for command line processing based
**                           on a list of report definitions.
** 002 GES 20070626   @34286 Added support for naming the master report
**                           index and source file index.
** 003 GES 20090424   @41934 Import change.
** 004 GES 20090515   @42212 Another import change.
** 005 GES 20130203          Rewrite to process all reports in a single pattern
**                           engine run. This cuts the total time of a reporting
**                           run down by 80%! As part of this, the input format
**                           (for defining the reports) has been streamlined and
**                           the reports are parsed into a list of report
**                           definition objects.
** 006 ECF 20150715          Replace StringBuffer with StringBuilder.
** 007 GES 20160311          Added supportLvlExpr parsing and assignment into the report def.
** 008 CA  20170118          Added -T option, which is like -S but with a trailing of explicit 
**                           schema AST file names.
** 009 ECF 20170505          Retrofit for use with FWD Analytics web application.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.report;

import java.io.*;
import java.util.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.pattern.*;
import com.goldencode.p2j.util.*;
import com.goldencode.util.*;

/**
 * Special-purpose driver for the pattern engine to process a list of report
 * definitions against a given set of ASTs.  This provides a command line
 * interface, please see {@link #main} for the syntax.
 * <p>
 * The report definitions are stored in an XML file with the following
 * structure:
 * <p>
 * <ul>
 *    <li>list (required root element, contains a list of all reports to
 *        create)
 *    <ul> 
 *       <li>idxname (optional attribute, contains the base file name of the
 *           master report index to be generated, defaults to "index" which
 *           would result in the master report filename being "index.html",
 *           when overridden this value will also be prepended to the
 *           source file cross-reference default filename 
 *           "source_file_index.html")
 *       <li>report (element, 1 or more, represents a single report)
 *       <ul>
 *          <li>name (required attribute, contains the rule-set file name
 *              that creates the associated report)
 *          <li>descr (optional attribute, contains the description of the
 *              report)
 *          <li>variable (element, 0 or more, represents a user-replaceable
 *              input for the report being run)
 *          <ul>
 *             <li>name (required attribute, the name of the variable being
 *                 set)
 *             <li>value (required attribute, the value to assign to the
 *                 variable)
 *          </ul>
 *       </ul>
 *    </ul>
 * </ul>
 * <p>
 * Report definition files have a <code>.rpt</code> filename suffix and must
 * reside in the pattern path.
 */
public class ReportDriver
implements DebugLevels,
           XmlConstants,
           ReportConstants
{
   /** Default admin password system property key */
   private static final String APP_ADMIN_PASS = "appAdminPass";
   
   /** Reporting rule-set. */
   private static final String ruleset = "reports/consolidated_reports";
   
   /** Database mode flag. */
   private static boolean database = false;
   
   /**
    * Run the pattern engine with the given input.
    *
    * @param    files
    *           List of ASTs to process.
    * @param    dbg
    *           Debug level.
    * @param    init
    *           Variable initializers to enable.
    */
   public static void patternEngine(FileList[] files, int dbg, Map<String, Object> init)
   throws Exception
   {
      PatternEngine engine = new PatternEngine();
      for (FileList fl : files)
      {
         if (fl != null)
         {
            engine.addAstSpec(fl);
         }
      }
      PatternEngine.setDebugLevel(dbg);
      engine.setReadOnly(true);
      engine.setVariableInitializers(init);
      engine.run(ruleset);
   }
   
   /**
    * Read the given report definitions and store the list of reports for subsequent processing.
    * Read both the standard reports as well as any user-defined reports that have been exported.
    *
    * @param   rptdef
    *          File name root of the XML file storing the list of report definitions.
    *
    * @return  The list of report definitions.
    * 
    * @throws  ConfigurationException
    *          if there is an error loading the project configuration.
    * @throws  ParserConfigurationException
    *          if there is an error configuring the XML parser.
    * @throws  SAXException
    *          if the XML SAX parser encounters an error.
    * @throws  IOException
    *          if there is an error accessing the file.
    */
   private static List<ReportDefinition> loadReports(String rptdef)
   throws ConfigurationException,
          ParserConfigurationException,
          IOException,
          SAXException
   {
      ConfigurationPersistence cfgPers = new ConfigurationPersistence();
      List<ReportDefinition> reports = new ArrayList<>();
      
      // load standard reports
      File file = ConfigLoader.findConfigurationFile(rptdef, RPT_SUFFIX);
      int typePrimary = cfgPers.readXmlReports(file, reports);
      String userRptBase;
      
      switch (typePrimary)
      {
         case SRC_CODE_CACHE:
            database = false;
            userRptBase = "code";
            break;
         case SRC_SCHEMA:
            database = true;
            userRptBase = "schema";
            break;
         default:
            throw new RuntimeException("Unknown report type: " + typePrimary);
      }
      
      // load user-defined reports, if any
      file = new File(Configuration.home() +
                      File.separator +
                      String.format(AUTO_EXPORT_RPT_SPEC, userRptBase));
      int typeUser = cfgPers.readXmlReports(file, reports);
      
      // sanity: make sure user report type, if any, matches primary report type
      if (typeUser != SRC_UNKNOWN && typeUser != typePrimary)
      {
         throw new RuntimeException("Unexpected report type found in " + file + ": " + typeUser);
      }
      
      return reports;
   }
   
   /**
    * Emit a syntax statement and optional error message to
    * <code>stderr</code>, then exit the process with the specified return
    * code. This method is called from {@link #main} when the user passes
    * command line parameters which are not viable.
    *
    * @param    msg
    *           Optional error message; may be <code>null</code>.
    * @param    rc
    *           Process return code used with <code>System.exit()</code>.
    */
   private static void syntax(String msg, int rc)
   {
      if (msg != null)
      {
         System.err.println(msg);
         System.err.println();
      }
      
      String[] staticText =
      {
         "",
         "Syntax:",
         "",
         "   java -DP2J_HOME=<home> ReportDriver [options] <rptdef> <filelist>",
         "   java -DP2J_HOME=<home> ReportDriver -S[options] <rptdef> <directory> \"<filespec>\"",
         "   java -DP2J_HOME=<home> ReportDriver -T[options] <rptdef> " +
         "<directory> \"<filespec>\" <filelist>",
         "",
         "Where",
         "",
         "   Options",
         "",
         "      Dn = set debug level 'n' (must be a numeric digit between",
         "           " + MSG_NONE + " and " + MSG_TRACE + " inclusive)",
         "      S  = use filespecs instead of an explicit file list",
         "      T  = use filespecs and a trail of explicit files",
         "      N  = no recursion in filespec mode or trail mode",
         "",
         "   rptdef     = the report definition file to use, must be found",
         "                somewhere in the pattern path", 
         "   filelist   = arbitrary list of absolute and/or relative file", 
         "                names to scan (default approach)",
         "   directory  = a relative or absolute path (only used when using",
         "                filespecs)",
         "   filespec   = file filter specification of the filenames in the",
         "                given directory to scan, should be enclosed in",
         "                double quotes if any of the wildcard characters '*'",
         "                or '?' are used (only used when using filespecs)",
         ""
      };
      
      for (int i = 0; i < staticText.length; i++)
      {
         System.err.println(staticText[i]);
      }
      
      System.exit(rc);
   }
   
   /**
    * Provides a command line interface for generating a list of reports
    * using a list of ASTs and a report definition.
    * <p>
    * Syntax:
    * <pre>
    *   java -DP2J_HOME=&lt;home&gt; ReportDriver [options] &lt;rptdef&gt; &lt;filelist&gt;
    *   java -DP2J_HOME=&lt;home&gt; ReportDriver -S[options] &lt;rptdef&gt; &lt;directory&gt; &lt;filespec&gt;
    *   java -DP2J_HOME=&lt;home&gt; ReportDriver -T[options] &lt;rptdef&gt; &lt;directory&gt; &lt;filespec&gt; &lt;filelist&gt;
    * </pre>
    * Where:
    * <ul>
    *   <li> Options
    *   <ul>
    *      <li> Dn = set debug level 'n' (must be a numeric digit between
    *           MSG_NONE (0) and MSG_TRACE (3) inclusive
    *      <li> S  = use filespecs instead of an explicit file list
    *      <li> T  = use filespecs with a trailing explicit file list (optional)
    *      <li> N  = no recursion in filespec mode
    *   </ul>
    *   <li> <code>rptdef</code> is the name of a report definition file
    *        found via the pattern path
    *   <li> explicit file list mode (default approach)
    *   <ul>
    *      <li> &lt;filelist&gt; is arbitrary list of absolute and/or relative file names to scan
    *   </ul> 
    *   <li> filespec mode
    *   <ul>
    *      <li> &lt;directory&gt; is the directory to search for files
    *      <li> &quot;filespec&quot; is the file specification that will be
    *           used to create a list of files to process, should be enclosed
    *           in double quotes if any of the wildcard characters '*' or '?'
    *           are used
    *   </ul>
    * </ul>
    *
    * @param   args
    *          List of command line arguments.
    */
   public static void main(String[] args) 
   {
      // option defaults
      boolean explicit  = true;
      boolean filespec  = false;
      boolean recursion = true;
      int     debug     = MSG_STATUS;
      int     idx       = 0;
      
      // requires (mode + directory + filespec) OR (mode and a filelist)
      if (args.length < 2)
      {
         syntax(null, -1);
      }
      
      // process options
      if (args[0].startsWith("-"))
      {
         idx++;
         String opts = args[0].toLowerCase();
         
         if (opts.indexOf('s') != -1)
         {
            explicit = false;
            filespec = true;
         }
         if (opts.indexOf('t') != -1)
         {
            explicit = true;
            filespec = true;
         }
         
         
         if (opts.indexOf('n') != -1) recursion = false;
         else recursion = true;
         
         int dbg = opts.indexOf('d');
         
         if (dbg != -1)
         {
            char level = opts.charAt(dbg + 1);
            
            if (!Character.isDigit(level))
            {
               syntax("Missing numeric debug level character following D option.", -2);
            }
            else
            {
               debug = Character.digit(level, 10);
               
               if (debug < MSG_NONE || debug > MSG_TRACE)
               {
                  syntax("Debug level must be between " + MSG_NONE +
                         " and " + MSG_TRACE + ".",
                         -3);
               }
            }
         }
      }
      
      // process the report definition
      if (idx >= args.length)
      {
         syntax("Missing report definition parameter.", -4);
      }
      
      try
      {
         List<ReportDefinition> reports = loadReports(args[idx]);
         idx++;
         
         FileList[] files = new FileList[2];
         
         // at this point idx points to the next arg after the options, this
         // must be one of our various forms of filename input
         
         if (filespec)
         {
            // there must be exactly 2 args left, unless explicit mode is used, too
            int remArgs = args.length - idx;
            if ((!explicit && remArgs != 2) || (explicit && remArgs < 2))
            {
               syntax("Missing/incomplete filespec parameter.", -7);
            }
            
            files[0] = new FileSpecList(new File(args[idx]), args[idx+1], recursion);
            idx = idx + 2;
         }
         
         if (explicit)
         {
            // there must be at least 1 arg left
            if (args.length <= idx)
            {
               if (!filespec)
               {
                  // mandatory only if filespec was not included
                  syntax("Missing file list parameter.", -6);
               }
            }
            else
            {
               int      len      = args.length;
               String[] filelist = new String[len - idx];
               
               for (int j = idx; j < len; j++)
               {
                  filelist[j - idx] = args[j];
               }
               
               files[1] = new ExplicitFileList(filelist);
            }
         }
         
         Map<String, Object> vars = new HashMap<>();
         
         String appAdminPass = System.getProperty(APP_ADMIN_PASS);
         if (appAdminPass != null && appAdminPass.length() > 0)
         {
            vars.put(APP_ADMIN_PASS, '"' + StringHelper.stripEnclosing(appAdminPass, '"') + '"');
         }
         
         // add a definition for database mode
         vars.put("databaseMode", database);
         
         // add our reports
         vars.put("reports", reports);
         
         // create a new pattern engine, configure it and run
         patternEngine(files, debug, vars);
      }
      
      catch(Exception excpt)
      {
         excpt.printStackTrace();
      }
   }
}