/*
** Module   : CallGraphLink.java
** Abstract : Encodes a single call graph link (graph database edge).
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 GES 20170530 First version.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.report.server;

/**
 * Encodes a single call graph link (graph database edge).
 */
public class CallGraphLink
{
   /** Graph-unique id for the source node. */ 
   private String source = "";
   
   /** Graph-unique id for the target node. */ 
   private String target = "";
   
   /** Descriptive text for the link. */
   private String text = "";
   
   /** The unique ID of this link. */
   private int id;
   
   /** If present, it represents the ID of a parent program-to-program link. */
   private Integer parentId = null;
   
   /**
    * Construct an instance.
    *
    * @param    id
    *           The link's unique ID.
    * @param    source
    *           Graph-unique id for the source node.
    * @param    target
    *           Graph-unique id for the target node.
    * @param    text
    *           Descriptive text for the link.
    */
   public CallGraphLink(int id, String source, String target, String text)
   {
      this.source = source;
      this.target = target;
      this.text   = text;
      this.id = id;
   }
   
   /**
    * Construct an instance.
    *
    * @param    id
    *           The link's unique ID.
    * @param    parentId
    *           The link's associated program-to-program link ID.
    * @param    source
    *           Graph-unique id for the source node.
    * @param    target
    *           Graph-unique id for the target node.
    * @param    text
    *           Descriptive text for the link.
    */
   public CallGraphLink(int id, int parentId, String source, String target, String text)
   {
      this(id, source, target, text);
      this.parentId = parentId;
   }
   
   /**
    * Obtain the unique id of the source node.
    *
    * @return   The source node's unique id.
    */
   public String getSource()
   {
      return source;
   }

   /**
    * Set the unique id of the source node.
    *
    * @param    source
    *           The new source node id to assign.
    */
   public void setSource(String source)
   {
      this.source = source;
   }

   /**
    * Obtain the unique id of the target node.
    *
    * @return   The target node's unique id.
    */
   public String getTarget()
   {
      return target;
   }

   /**
    * Set the unique id of the target node.
    *
    * @param    target
    *           The new target node id to assign.
    */
   public void setTarget(String target)
   {
      this.target = target;
   }

   /**
    * Obtain the descriptive text for the link.
    *
    * @return   The description of the link.
    */
   public String getText()
   {
      return text;
   }

   /**
    * Set the descriptive text for the link.
    *
    * @param    text
    *           The new descriptive text to assign.
    */
   public void setText(String text)
   {
      this.text = text;
   }
   
   /**
    * Compute the unique hashcode for this instance, using the {@link #source} and {@link #target}
    * fields.
    * 
    * @return   See above.
    */
   @Override
   public int hashCode()
   {
      int result = 17;
      result = result * 37 + source.hashCode();
      result = result * 37 + target.hashCode();
      return result;
   }
   
   /**
    * Check if this instance is the same as the other one (if {@link #source} and {@link #target}
    * are equal).
    * 
    * @param    obj
    *           The instance to compare.
    * @return   See above.
    */
   @Override
   public boolean equals(Object obj)
   {
      if (!(obj instanceof CallGraphLink))
      {
         return false;
      }
      
      CallGraphLink link = (CallGraphLink) obj;
      return source.equals(link.source) && target.equals(link.target);
   }

   /**
    * Get the {@link #id} of this link.
    * 
    * @return   See above.
    */
   public int getId()
   {
      return id;
   }

   /**
    * Set the {@link #id} of this link
    * 
    * @param    id
    *           The unique ID of this link.
    */
   public void setId(int id)
   {
      this.id = id;
   }

   /**
    * Get the {@link #parentId} of this link.  Will be {@code null}, if this is an intra-program
    * or a link with a non-program, external, resource.
    * 
    * @return   See above.
    */
   public Integer getParentId()
   {
      return parentId;
   }

   /**
    * Set the {@link #parentId} of this link
    * 
    * @param    parentId
    *           The ID of this link's program-to-program associated link.
    */
   public void setParentId(Integer parentId)
   {
      this.parentId = parentId;
   }
}