/*
** Module   : astview.js
** Abstract : AST SVG viewer.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 ECF 20170528 First version, inspired by D3 examples.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

/**
 * An SVG viewer which depicts an animated visualization of a Progress abstract syntax tree (AST).
 */
define(function()
{
   /** The top-level object which will be returned from this function. */
   var me = {};
   
   /** Margin used by model when data is small enough to fit completely within viewer */
   var margin = { top: 20, right: 120, bottom: 20, left: 120 };
   
   /** SVG top-level element */
   var svg = null;
   
   /** Master group element which contains all node and link elements */
   var mst = null;
   
   /** Current root node of the model */
   var root = null;
   
   /** Duration in milliseconds of animated transitions */
   var duration = 800;
   
   /** Radius of circle nodes */
   var radius = 7;
   
   /** Absolute x offset of text elements from the circle nodes they describe */
   var textOffset = radius * 2 - 4;
   
   /** Virtual height of the model */
   var h;
   
   /** Actual width of the SVG element (i.e., that of its parent container) */
   var svgWidth;
   
   /** Actual height of the SVG element (i.e., that of its parent container) */
   var svgHeight;
   
   /** Minimum width of a "column" of data nodes in the model */
   var minColWidth;
   
   /** Maximum allowable width of a text element */
   var maxTextWidth;
   
   /**
    * Prepare the AST viewer by initializing the SVG and attaching it to the specified DOM
    * element.
    * 
    * @param   {String} selector
    *          Selector of DOM element to which to attach the SVG viewer.
    * @param   {Number} width
    *          Parent container width.
    * @param   {Number} height
    *          Parent container height.
    */
   me.prepare = function(selector, width, height)
   {
      svgWidth = width;
      svgHeight = height;
      h = height - margin.top - margin.bottom;
      minColWidth = svgWidth / 10;
      
      // create the SVG element and top level group if not already present
      if (!svg || d3.select(selector).select("svg").empty())
      {
         // create SVG element
         svg = d3.select(selector).append("svg")
                    .attr("width", width)
                    .attr("height", height);
         
         // append master group
         mst = svg.append("g")
                  .attr("transform", "translate(" + margin.left + "," + margin.top + ")");
         
         // add the zoom/pan support; if we associate the zoom with the master g element, then we
         // don't get the events needed to zoom, but by using the svg we do get the events
         svg.call(d3.zoom().on("zoom", function() { mst.attr("transform", d3.event.transform) }));
      }
      
      fwd.evt.addListener("astenter", onAstEnter);
      fwd.evt.addListener("astexit", onAstExit);
   }
   
   /**
    * Clean up resources.
    */
   me.cleanup = function()
   {
      fwd.evt.removeListener("astenter", onAstEnter);
      fwd.evt.removeListener("astexit", onAstExit);
   }
   
   /**
    * Show the visualization of the specified AST.
    * 
    * @param   {Object} ast
    *          Abstract syntax tree data structure.
    */
   me.show = function(ast, previousAst)
   {
      // wrap the AST in a d3 hierarchy
      root = d3.hierarchy(ast, function(d) { return d.children; });
      root.x0 = svgHeight / 2;
      root.y0 = 0;
      
      // safety valve to prevent rendering huge trees if the root has many children
      if (root.children && root.children.length > 100)
      {
         root.children.forEach(function(d)
                              {
                                 if (d !== previousAst)
                                 {
                                    collapse(d);
                                 }
                              });
      }
      
      // refresh the model
      refreshState(root);
   }
   
   /**
    * Refresh the state of the model using the given source node, which causes the tree to be
    * reflowed and drawn.
    * 
    * @param   {Object} source
    *          Node which describes the root of the model to be depicted.
    */
   function refreshState(source)
   {
      // adjust layout to use full width for shallow trees, but expand deeper trees beyond the
      // width of the SVG, so individual nodes are legible, and viewing deeper nodes requires
      // panning or zooming out
      var depth = maxDepth();
      var w;
      if (depth >= svgWidth / 100)
      {
         w = depth * minColWidth;
         maxTextWidth = minColWidth - (radius * 2 + 2);
      }
      else
      {
         w = svgWidth - margin.left - margin.right;
         maxTextWidth = w / depth - (textOffset / 2);
      }
      
      // get a tree layout
      var data = d3.tree()
                   .size([h, w])(root);
      
      // get node and link arrays
      var nodes = data.descendants();
      var links = nodes.slice(1);
      
      // associate the AST data with the model for use by nodes
      var node = mst.selectAll("g.ast-node")
                    .data(nodes, function(d) { return d.data.id; });
      
      // create node enter selection for nodes being added to the model
      var nodeEnter = node.enter().append("g")
                          .attr("id", function(d) { return d.data.id; })
                          .attr("class", "ast-node")
                          .attr("transform", function(d)
                               {
                                  return "translate(" + source.y0 + "," + source.x0 + ")";
                               });
      
      // add circles to represent nodes, color coded to their AST types
      nodeEnter.append("circle")
               .attr("r", 1e-6)
               .style("fill", circleFill)
               .style("stroke", circleStroke)
               .attr("cursor", "pointer")
               .on("click", click)
               .on("mouseenter", fireAstEnter)
               .on("mouseleave", fireAstExit);
      
      // add a text element for the AST text label
      nodeEnter.append("text")
               .attr("class", "ast-text")
               .attr("x", textXPos)
               .attr("dy", ".9em")
               .attr("text-anchor", textAnchor)
               .style("fill-opacity", 1e-6);
      
      // add a text element for the AST token type label
      nodeEnter.append("text")
               .attr("class", "ast-token")
               .attr("x", textXPos)
               .attr("dy", tokenYPos)
               .attr("text-anchor", textAnchor)
               .style("fill-opacity", 1e-6);
      
      // merge enter and update nodes to perform update actions on new nodes and those being
      // updated
      var nodeUpdate = nodeEnter.merge(node);
      
      // translate nodes into the positions computed by the tree layout
      nodeUpdate.transition()
                .duration(duration)
                .attr("transform", function(d)
                     { 
                        return "translate(" + d.y + "," + d.x + ")";
                     });
      
      // grow circles from their initial, tiny radius and draw them
      nodeUpdate.select("circle")
                .transition()
                .duration(duration)
                .attr("r", 7)
                .style("fill", circleFill)
                .style("stroke", circleStroke);
      
      // fade in (from initially transparent state) and set the appropriate AST text label;
      // do not set text for synthetic nodes
      nodeUpdate.select(".ast-text")
                .each(function(d)
                     {
                        if (d.data.l0 > 0)
                        {
                           morphLabel(this, d.data.text);
                        }
                     });
      
      // fade in (from initially transparent state) and set the appropriate AST token type label
      nodeUpdate.select(".ast-token")
                .each(function(d) { morphLabel(this, d.data.token); });
      
      // select exiting nodes and translate their position back to that of the root node
      var nodeExit = node.exit()
                         .transition()
                         .duration(duration)
                         .attr("transform", function(d)
                              {
                                 return "translate(" + source.y + "," + source.x + ")";
                              })
                         .remove();
      
      // shrink exiting circles back to their tiny radius
      nodeExit.select("circle")
              .attr("r", 1e-6);
      
      // hide existing text
      nodeExit.selectAll("text")
              .style("fill-opacity", 1e-6);
      
      // associate the AST data with the model for use by links
      var link = mst.selectAll("path.ast-link")
                    .data(links, function(d) { return d.data.id; });
      
      // create link enter selection for links being added to the model
      var linkEnter = link.enter().insert("path", "g")
                          .attr("class", "ast-link")
                          .attr("d", function(d)
                               {
                                  var o = { x: source.x0, y: source.y0 };
                                  return diagonal(o, o);
                               });
      
      // merge enter and update links to perform update actions on new links and those being
      // updated
      var linkUpdate = linkEnter.merge(link);
      
      // transition new and updated link paths into place
      linkUpdate.transition()
                .duration(duration)
                .attr("d", function(d) { return diagonal(d, d.parent); });
      
      // transition exiting link paths back to the source node
      link.exit().transition()
                 .duration(duration)
                 .attr("d", function(d)
                      {
                         var o = {x: source.x, y: source.y};
                         return diagonal(o, o);
                      })
                 .remove();
      
      // remember node positions for a future update cycle
      nodes.forEach(function(d)
                   {
                      d.x0 = d.x;
                      d.y0 = d.y;
                   });
   }
   
   /**
    * Get circle node fill color, varying by AST type and whether node has children.
    * 
    * @return  {String} Fill color name or hex value.
    */
   function circleFill(d)
   {
      return d._children ? (d.data.l0 > 0 ? "lightsteelblue" : "gold") : "#fff";
   }
   
   /**
    * Get circle node stroke color, varying by AST type.
    * 
    * @return  {String} Stroke color name.
    */
   function circleStroke(d)
   {
      return d.data.l0 > 0 ? "steelblue" : "goldenrod";
   }
   
   /**
    * Get the X offset for a text node: left of the node for parents, right of the node for
    * leaves. Works in conjunction with text-anchor setting.
    * 
    * @return  {Number} X offset from node position for start (leaf) or end (parent) of text
    *                   element.
    */
   function textXPos(d)
   {
      return d.children || d._children ? -textOffset : textOffset;
   }
   
   /**
    * Get the Y offset of a token type label text element, which can vary by AST type. Token
    * names are vertically centered for synthetic nodes, because the text label is omitted for
    * these.
    * 
    * @return  {String} Y offset from node position in em units.
    */
   function tokenYPos(d)
   {
      return d.data.l0 > 0 ? "-.25em" : ".35em";
   }
   
   /**
    * Get the text anchor type for node text elements.
    * 
    * @return  {String} "start" for leaf nodes; "end" for parent nodes.
    */
   function textAnchor(d)
   {
      return d.children || d._children ? "end" : "start";
   }
   
   /**
    * Perform concurrent transitions on a text element to modify the fill-opacity and text value.
    * The text transition is quite coarse, having only 2 interpolation states. It is performed
    * as a transition, however, to ensure the text changes while the model is in motion, as
    * nodes and links are being translated. This avoids an abrupt change in text values at the
    * start of an update.
    * 
    * @param   (SVGTextElement}  svgText
    *          Target element.
    * @param   {String}          label
    *          Text value of the label.
    */
   function morphLabel(svgText, label)
   {
      var oLock = {};
      var tLock = {};
      
      var textSel = d3.select(svgText);
      
      // transition fill-opacity of text
      d3.select(oLock)
        .transition()
        .duration(duration)
        .tween("style:fill-opacity", function()
              {
                 var fo = textSel.style("fill-opacity");
                 var interp = d3.interpolateNumber(fo, 1);
                 
                 return function(t) { textSel.style("fill-opacity", interp(t)); };
              });
      
      // change text half-way through the transition
      d3.select(tLock)
        .transition()
        .duration(duration)
        .tween("text", function()
              {
                 var start = textSel.text();
                 var end = astLabel(svgText, label);
                 var interp = function(t) { return t < 0.5 ? start : end; }
                 
                 return function(t) { textSel.text(interp(t)); };
              });
   }
   
   /**
    * Generate the text for the given text element, abbreviating it with ellipses if necessary
    * to fit within the maximum available width.
    * 
    * @param   {SVGTextElement}  svgText
    *          Element which will contain the text label.
    * @param   {String}          label
    *          Original text value of the label.
    * 
    * @return  {String} Text for the label, abbreviated as necessary.
    */
   function astLabel(svgText, label)
   {
      var text = label;
      
      // remember original svg text
      var origText = svgText.innerHTML;
      
      svgText.innerHTML = text;
      var bbox = svgText.getBBox();
      
      for (var i = label.length - 2; i >= 0 && bbox.width > maxTextWidth; i--)
      {
         svgText.innerHTML = text = label.substring(0, i) + "...";
         bbox = svgText.getBBox();
      }
      
      svgText.innerHTML = origText;
      
      return text;
   }
   
   /**
    * Generate a diagonal path to be used to represent a link between nodes.
    * 
    * @param   {Object} s
    *          Coordinate containing x and y position of the source node.
    * @param   {Object} d
    *          Coordinate containing x and y position of the destination node.
    * 
    * @return  {String} Link path descriptor.
    */
   function diagonal(s, d)
   {
      return "M" + s.y + "," + s.x +
             "C" + ((s.y + d.y) / 2) + "," + s.x +
             " " + ((s.y + d.y) / 2) + "," + d.x +
             " " + d.y + "," + d.x;
   }
   
   /**
    * Determine the maximum, relative depth of the tree hierarchy in its current state, taking
    * all expanded and collapsed sub-trees into account.
    * 
    * @return  {Number} Depth of the deepest leaf of the tree, relative to the depth of the root.
    */
   function maxDepth()
   {
      var max = root.depth;
      root.each(function(d) { max = d.children ? Math.max(max, d.depth + 1) : max; });
      
      return max - root.depth + 1;
   }
   
   /**
    * Collapse a branch in the tree model by removing its children (array is stored for later
    * restoration when expanding).
    * 
    * @param   {Object} d
    *          Node to collapse.
    */
   function collapse(d)
   {
      if (d.children)
      {
         d._children = d.children;
         d._children.forEach(collapse);
         d.children = null;
      }
   }
   
   /**
    * Mouse click function which expands or collapses nodes when clicked (no effect for leaf
    * nodes).
    * 
    * @param   {Object} d
    *          Node at the base of the branch to be expanded or collapsed.
    */
   function click(d)
   {
      if (d3.event.shiftKey)
      {
         // make immediate parent node, if any the new root, otherwise do nothing
         var parentAst = d.data.parent;
         if (parentAst)
         {
            me.show(parentAst, d.data);
            fireAstSelect(parentAst.id);
            
            return;
         }
      }
      else if (d3.event.ctrlKey)
      {
         // make the clicked node the root of the tree, unless it already is
         if (d !== root)
         {
            me.show(d.data);
            fireAstSelect(d.data.id);
         }
      }
      else if (d.children)
      {
         // collapse immediate children
         d._children = d.children;
         d.children = null;
      }
      else
      {
         // expand immediate children
         d.children = d._children;
         d._children = null;
      }
      refreshState(d);
   }
   
   /**
    * Fire an event indicating the mouse is hovering over an AST node.
    * 
    * @param   {Object} d
    *          Node entered.
    */
   function fireAstEnter(d)
   {
      fwd.evt.fireEvent("astenter", this, { id : d.data.id });
   }
   
   /**
    * Fire an event indicating the mouse is no longer hovering over an AST node.
    * 
    * @param   {Object} d
    *          Node exited.
    */
   function fireAstExit(d)
   {
      fwd.evt.fireEvent("astexit", this, { id : d.data.id });
   }
   
   /**
    * Fire an event indicating a new hode has been selected as the root of the tree.
    * 
    * @param   {Number} astId
    *          ID of the new root's AST node.
    */
   function fireAstSelect(astId)
   {
      fwd.evt.fireEvent("astselect", this, { id : astId });
   }
   
   /**
    * Handler for the &quot;astenter&quot; event. Redraws the node to indicate it has been
    * highlighted and temporarily expands any abbreviated text.
    * 
    * @param   {Object} e
    *          Event object.
    */
   function onAstEnter(e)
   {
      var g = d3.select("[id='" + e.id + "']");
      
      if (!g.empty())
      {
         g.select("circle")
          .transition()
          .duration(duration / 4)
          .attr("r", radius + 3)
          .style("stroke", circleFill)
          .style("fill", circleStroke);
         
         g.select(".ast-text")
          .text(function(d) { return d.data.l0 > 0 ? d.data.text : ""; });
         
         g.select(".ast-token")
          .text(function(d) { return d.data.token; });
         
         fwd.ast.showInfoPanel(svg, g.datum().data);
      }
   }
   
   /**
    * Handler for the &quot;astexit&quot; event. Redraws the node to indicate it is no longer
    * highlighted and resets any expanded text.
    * 
    * @param   {Object} e
    *          Event object.
    */
   function onAstExit(e)
   {
      var g = d3.select("[id='" + e.id + "']");
      
      g.select("circle")
       .transition()
       .duration(duration / 4)
       .attr("r", radius)
       .style("stroke", circleStroke)
       .style("fill", circleFill);
      
      g.select(".ast-text")
       .text(function(d) { return d.data.l0 > 0 ? astLabel(this, d.data.text) : ""; });
      
      g.select(".ast-token")
       .text(function(d) { return astLabel(this, d.data.token); });
      
      fwd.ast.removeInfoPanel(svg);
   }
   
   /* exports */
   return me;
   
});