/*
** Module   : legend.js
** Abstract : helper to create a legend as part of a larger SVG 
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 GES 20170530 First version.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

//"use strict";

/**
 * Legend functionality.
 */
//fwd.legend
define(function()
{
   "use strict";
   /** The top-level object which will be returned from this function. */
   var me = {};
   
   /** Left margin of each legend entry. */
   var LEGEND_ENTRY_INDENT = 10;
                   
   /** Height of each legend entry. */
   var LEGEND_ENTRY_HEIGHT = 8;
   
   /** Width of each legend entry. */
   var LEGEND_ENTRY_WIDTH = 20;
                   
   /** Vertical space between legend entries. */
   var LEGEND_ENTRY_VERT_GAP = 3;
                   
   /** Horizontal space between legend entries. */
   var LEGEND_ENTRY_HORIZ_GAP = 10;
   
   /**
    * Create the legend for the call graph visualization.
    *
    * @param    svg
    *           The SVG container in which to create the legend.
    * @param    content
    *           An array of legend entry objects that contain color and text members.            
    */
   function createLegend(svg, displayProps, nodes, links)
   {
      var bundle = {};
      
      bundle.icon = svg.append("image")
                       .attr("x", 10)
                       .attr("y", 10)
                       .attr("width",35)
                       .attr("height", 35)
                       .attr("xlink:href", "images/compass.svg")
                       .attr("visibility", "display")
                       .on("click", function() { toggle(bundle, true); });
      
      bundle.legend = svg.append("g")
                         .attr("opacity", 0)
                         .attr("visibility", "hidden")      
                         .on("click", function() { toggle(bundle, false); });
                         
      var rect = bundle.legend.append("rect");
      var bounds =
      {
         "width"  : 0,
         "height" : 0,
      };
      
      var usedNodes = {};
      var ncontent = [];
      
      for (var i = 0; i < nodes.length; i++)
      {
         var type = nodes[i].type;
         
         if (displayProps[type] && !usedNodes[type])
         {
            usedNodes[type] = true;
            ncontent.push(displayProps[type]);
         }
      }
      
      ncontent.sort(function(l1, l2) 
      {
         var t1 = l1.text;
         var t2 = l2.text;
         return t1 > t2 ? 1 : t1 < t2 ? -1 : 0;
      });

      var usedLinks = {};
      var lcontent = [];
      
      for (var i = 0; i < links.length; i++)
      {
         var ltext = links[i].text;
         
         if (displayProps[ltext] && !usedLinks[ltext])
         {
            usedLinks[ltext] = true;
            lcontent.push(displayProps[ltext]);
         }
      }
      
      lcontent.sort(function(l1, l2) 
      {
         var t1 = l1.text;
         var t2 = l2.text;
         return t1 > t2 ? 1 : t1 < t2 ? -1 : 0;
      });

      var baseX = LEGEND_ENTRY_INDENT / 2;
      var baseY = LEGEND_ENTRY_VERT_GAP * 3;
      
      var ntbounds;
      var txt = bundle.legend.append("text")
                             .text("Visible Vertices")
                             .attr("x", baseX + LEGEND_ENTRY_INDENT)
                             .attr("y", baseY + LEGEND_ENTRY_HEIGHT)
                             .style("font", "normal 10px sans-serif")
                             .style("stroke", "black")
                             .attr("stroke-width", .3)
                             .each(function(d) { ntbounds = this.getBBox(); });
      baseY += ntbounds.height;
      
      for (var i = 0; i < ncontent.length; i++)
      {
         var ebounds = legendEntry(bundle.legend, i, ncontent[i], baseX, baseY, false);
         
         bounds.width  = Math.max(bounds.width, ebounds.width);
         bounds.height = ebounds.height;
         bounds.y      = ebounds.y;  // last entry's Y
      }

      txt = bundle.legend.append("text")
                         .text("Visible Links")
                         .attr("x", baseX + LEGEND_ENTRY_INDENT)
                         .attr("y", baseY + (LEGEND_ENTRY_HEIGHT + 2 + LEGEND_ENTRY_VERT_GAP) * (ncontent.length + 1))
                         .style("font", "normal 10px sans-serif")
                         .style("stroke", "black")
                         .attr("stroke-width", .3)
                         .each(function(d) { ntbounds = this.getBBox(); });
      baseY += ntbounds.height;

      for (var i = 0; i < lcontent.length; i++)
      {
         var ebounds = legendEntry(bundle.legend, i + ncontent.length, lcontent[i], baseX, baseY, true);
         
         bounds.width  = Math.max(bounds.width, ebounds.width);
         bounds.height = ebounds.height;
         bounds.y      = ebounds.y;  // last entry's Y
      }
      
      rect.attr("x", LEGEND_ENTRY_INDENT / 2)
          .attr("y", LEGEND_ENTRY_VERT_GAP)
          .attr("width", bounds.width + LEGEND_ENTRY_INDENT)
          .attr("height", (bounds.y - (LEGEND_ENTRY_VERT_GAP * 2)) + bounds.height + (2 * LEGEND_ENTRY_VERT_GAP))
          .attr("fill", "#f8f9f9")
          .attr("stroke", "gray")
          .attr("opacity", 0.8)
          .attr("stroke-width", "1px");
      
      return bundle;
   }
   
   /**
    * Switch the visibility on the legend and the legend icon.
    *
    * @param    bundle
    *           The object containing the legend and icon elements on which to process.
    * @param    active
    *           Boolean value, if <code>true</code> we activate the legend and hide the icon,
    *           vice versa on <code>false</code>.
    */
   function toggle(bundle, active)
   {
      if (active)
      {
         bundle.legend.transition()
                      .duration(1500)
                      .ease(d3.easeExp)
                      .attr("visibility", "display")
                      .attr("opacity", 1);
                      
         bundle.icon.transition()
                    .duration(1500)
                    .ease(d3.easeElastic)
                    .attr("opacity", 0)
                    .attr("visibility", "hidden");
      }
      else
      {
         bundle.icon.transition()
                    .duration(1500)
                    .ease(d3.easeExp)
                    .attr("visibility", "display")
                    .attr("opacity", 1);
                    
         bundle.legend.transition()
                      .duration(1500)
                      .ease(d3.easeElastic)
                      .attr("opacity", 0)
                      .attr("visibility", "hidden");
      }
   }
   
   /**
    * Create a single legend entry.
    *
    * @param    parent
    *           The container for the legend.
    * @param    index
    *           Entry number to create.
    * @param    props
    *           The entry from the callgraph.displayProps array.
    * @param    baseX
    *           X offset at which to start.
    * @param    baseY
    *           Y offset at which to start.
    * @param    link
    *           Flag indicating if this is a link or a node.
    */
   function legendEntry(parent, index, props, baseX, baseY, link)
   {
      var color = props.color;
      var name = props.text;
   
      var entry = parent.append("g");
      
      var x = baseX + LEGEND_ENTRY_INDENT;
      var y = baseY + ((LEGEND_ENTRY_HEIGHT + 2 + LEGEND_ENTRY_VERT_GAP) * index);
      
      var oriX = x;
      
      // now add the gap (even before the first entry)
      y += LEGEND_ENTRY_VERT_GAP;
      
      var rbounds = null;
      
      var widget;
      if (link)
      {
         widget = entry.append("line")
                       .attr("x1", x)
                       .attr("y1", y + LEGEND_ENTRY_HEIGHT / 2)
                       .attr("x2", x + LEGEND_ENTRY_WIDTH)
                       .attr("y2", y + LEGEND_ENTRY_HEIGHT / 2)
                       .attr("stroke", props["stroke"])
                       .attr("stroke-width", props["stroke-width"])
                       .attr("stroke-dasharray", props["stroke-dasharray"])
                       .each(function(d) { rbounds = this.getBBox(); });
      }
      else
      {
         widget = entry.append("rect")
                       .attr("x", x)
                       .attr("y", y)
                       .attr("width", LEGEND_ENTRY_WIDTH)
                       .attr("height", LEGEND_ENTRY_HEIGHT)
                       .attr("fill", color)
                       .attr("stroke", "black")
                       .attr("stroke-width", 2)
                       .each(function(d) { rbounds = this.getBBox(); });
      }
           
      x += (LEGEND_ENTRY_WIDTH + LEGEND_ENTRY_HORIZ_GAP);
      
      var tbounds = null;
      
      var txt = entry.append("text")
                     .text(function(d) { return name; })
                     .attr("x", x)
                     .attr("y", y + LEGEND_ENTRY_HEIGHT - 1)
                     .style("font", "normal 7px sans-serif")
                     .style("stroke", "black")
                     .attr("stroke-width", .3)
                     .each(function(d) { tbounds = this.getBBox(); });
           
      var ebounds =
      {
         "x"      : oriX,
         "y"      : y,
         "width"  : LEGEND_ENTRY_INDENT + rbounds.width + LEGEND_ENTRY_HORIZ_GAP + tbounds.width,
         "height" : LEGEND_ENTRY_VERT_GAP + Math.max(tbounds.height, rbounds.height) + 1,
      }
      
      return ebounds;
   }
   
   me.createLegend = createLegend;
   
   // return our exported interface
   return me;
   
});