/*
** Module   : main.js
** Abstract : top level namespace and generic web client javascript functionality 
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 GES 20170519 First version.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

/**
 * Primary namespace and generic functions for the reporting web client.
 */
define(["dijit/Dialog", "dijit/ConfirmDialog", "dojox/uuid/generateRandomUuid", "dojo/domReady!"],
function(Dialog, ConfirmDialog, generateRandomUuid)
{
   "use strict";
   
   /** The top-level object which will be returned from this function. */
   var me = {};
   
   /** OS detection flags. */
   me.isWindows = navigator.appVersion.indexOf('Win') > -1;

   /** Browser detection flags. */
   me.isFirefox = navigator.userAgent.toLowerCase().indexOf('firefox') > -1;
   me.isChrome  = navigator.userAgent.toLowerCase().indexOf('chrome') > -1;
   me.isSafari  = (navigator.userAgent.toLowerCase().indexOf('safari/') > -1) && !me.isChrome;
   me.isChromium = !!window.chrome;
   me.isIE11 = navigator.userAgent.indexOf("Trident") > -1
                                                      || navigator.userAgent.indexOf("Edge") > -1;
   me.isIe = !!window.ActiveXObject || me.isIE11;
   
   /** The notification templates to prompt a user to do an action */
   var prompts = {};
   
   /** The hint templates to help a user to perform an action */
   var hints = {};
   
   /** The templates parameters */
   var parameters = {};
   
   /** The notification template for the copy action */
   prompts["copy"] = "The application is requesting a COPY action." +
                     " Press {0} to complete the COPY action.";
   
   /** The hint for the copy action */
   hints["copy"] = "Press {0} to complete the COPY action and to close this dialog.";
   
   /** copy template parameter */
   parameters["copy"] = me.isSafari ? "COMMAND-C" : "CTRL-C";
   
   /** The notification template for the paste action */
   prompts["paste"] = "The application is requesting a PASTE action." +
                      " Press {0} to complete the PASTE action.";
   
   /** The hint for the paste action */
   hints["paste"] = "Press {0} to complete the PASTE action and to close this dialog.";
   
   /** copy template parameter */
   parameters["paste"] = me.isSafari ? "COMMAND-V" : "CTRL-V";
   
   /** The notification template for the cut action */
   prompts["cut"] = "The application is requesting a CUT action." +
                    " Press {0} to complete the CUT action.";
   
   /** The hint for the cut action */
   hints["cut"] = "Press {0} to complete the CUT action and to close this dialog.";
   
   /** cut template parameter */
   parameters["cut"] = me.isSafari ? "COMMAND-X" : "CTRL-X";
   
   /** its properties define if copy, cut and paste commands are supported by this browser */
   var queryCommandSupported = {};

   /** Track the active page. */
   var activePage = null;
   
   /**
    * Make the given page the active page.
    *
    * @param    page
    *           The page to activate.
    *
    * @return   (Boolean) <code>true</code> if a new page was activated.
    */
   function activate(page)
   {
      if (page && page != activePage)
      {
         deactivate();
         activePage = page;
         return true;
      }
      
      return false;
   }
   
   me.activate = activate;
   
   /**
    * Clear the state for the currently active page.
    */
   function deactivate()
   {
      if (activePage != null && $(activePage).data("noclear") == undefined)
      {
         $(activePage).empty();
      }
      
      activePage = null;
   }
   
   me.deactivate = deactivate;
   
   /**
    * Block the user interface with an animated, modal overlay to indicate a long-running
    * operation.
    * 
    * @param   {String} message
    *          Optional message.
    * @param   {Boolean/Integer/String/Array} fade
    *          Optional fade in/out options:
    *          - Boolean: true/false
    *          - Integer: number of millis to fade in/out
    *          - String: "fast" or "slow"
    *          - Array: two-element array: [ fade-in option, fade-out option ]
    *          See also https://gasparesganga.com/labs/jquery-loading-overlay/
    */
   function showOverlay(message, fade)
   {
      var options = {};
      
      if (message)
      {
         options.custom =
            $("<div>", { id : "wait", css : { "font-size" : "2em" }, text : message, });
      }
      
      if (fade !== undefined)
      {
         options.fade = fade;
      }
      
      $.LoadingOverlay("show", options);
   }
   
   me.showOverlay = showOverlay;
   
   /**
    * Hide a modal overlay previously used to block user input.
    */
   function hideOverlay()
   {
      $.LoadingOverlay("hide");
   }
   
   me.hideOverlay = hideOverlay;
   
   /**
    * Display a message in a titled dialog.
    * 
    * @param   {String} title
    *          Dialog's title.
    * @param   {String} message
    *          Message to display.
    * @param   {Function} callback
    *          
    * @param   {Number} millis
    *          Optional number of milliseconds to display; if missing, user must dismiss dialog.
    */
   function showDialog(title, message, millis)
   {
      var dialog = new Dialog(
      {
         title   : title,
         content : message,
      });
      
      dialog.show();
      
      if (millis)
      {
         setTimeout(function() { dialog.hide(); }, millis);
      }
   }
   
   me.showDialog = showDialog;
   
   /**
    * Show a confirm dialog with custom title, content, and with OK and Cancel buttons.
    * 
    * @param   {String} title
    *          Dialog title.
    * @param   {String} content
    *          Plain text or HTML content for the dialog's main area.
    * @param   {Function} onExecute
    *          Callback to execute if the user clicks 'OK'.
    * @param   {Function} onCancel
    *          Callback to execute if the user clicks 'Cancel'.
    */
   function showConfirmDialog(title, content, onExecute, onCancel)
   {
      var listeners = new Map();
      
      if (onExecute)
      {
         listeners.set("execute", onExecute);
      }
      
      if (onCancel)
      {
         listeners.set("cancel", onCancel);
      }
      
      var id = generateRandomUuid();
      
      var dialog = createConfirmDialog(id, title, content, listeners);
      
      dialog.show();
   }
   
   me.showConfirmDialog = showConfirmDialog;
   
   /**
    * Create a confirm dialog with a custom title and content, and specified listeners.
    * 
    * @param   {String} id
    *          Dialog's unique identifier.
    * @param   {String} title
    *          Dialog title.
    * @param   {String} content
    *          Plain text or HTML content for the dialog's main area.
    * @param   {Map} listeners
    *          An event to its listener map.
    * 
    * @return  {ConfirmDialog} A target confirm dialog
    */
   function createConfirmDialog(id, title, content, listeners)
   {
      var dialog = dijit.byId(id);
      
      if (dialog)
      {
         dialog.destroyRecursive(false);
      }
      
      dialog = new ConfirmDialog(
      {
         id      : id,
         title   : title,
         content : content,
      });
      
      if (listeners)
      {
         listeners.forEach(function(listener, eventName)
                           {
                              dialog.on(eventName, listener);
                           });
      }
      
      return dialog;
   }
   
   me.createConfirmDialog = createConfirmDialog;
   
   /**
    * Add a container which temporarily hides existing content under the given parent element.
    * The container has a centered title and an optional, left-aligned "back" arrow. When the
    * arrow is clicked, the container removes itself and unhides the original content, then
    * optionally executes a callback function.
    * 
    * @param   {jQuery selection} parent
    *          Selection of the parent element. Should select a single element to which to attach
    *          new container.
    * @param   {String} title
    *          Title content.
    * @param   {boolean} back
    *          True to display the back button, false if back functionality is not required.
    * @param   {boolean} copy
    *          True to display the "copy to clipboard" button, false if this functionality is not
    *          required.
    * @param   {boolean} search
    *          True to display the "Search" button, false if this functionality is not required.
    * @param   {Function} backFunc
    *          Optional callback function to execute upon clicking back button, after redisplaying
    *          old content.
    * 
    * @return  {jQuery selection} Selection of the new container, to which new content should be
    *          attached.
    */
   function addHidingContainer(parent, title, back, copy, search, backFunc)
   {
      var oldContent = parent.children();
      oldContent.hide();
      
      var div = "<div></div>";
      var span = "<span></span>";
      
      var newContent = $(div);
      parent.append(newContent);
      
      var titleAndBackDiv = $(div).addClass("toc-heading row no-padding");
      newContent.append(titleAndBackDiv);
      
      var buttonsContainer;
      
      if (back)
      {
         buttonsContainer = $(div).addClass("col-md-3 no-padding");
         // append back arrow
         var backSpan = $(span).addClass("glyphicon glyphicon-arrow-left")
                               .css({
                                        "float"   : "left",
                                        "margin"  : "4px",
                                        "cursor"  : "pointer",
                                        "z-index" : "2"
                                     })
                               .click(function()
                                     {
                                        newContent.remove();
                                        oldContent.show();
                                        if (backFunc)
                                        {
                                           backFunc();
                                        }
                                     });
         
         buttonsContainer.append(addToolTip(backSpan, "Back"));
      }
      
      if (copy)
      {
         // append duplicate (copy to clipboard) icon
         var copySpan = $(span).addClass("glyphicon glyphicon-duplicate")
                               .css({
                                        "float"   : "left",
                                        "margin"  : "4px",
                                        "cursor"  : "pointer",
                                        "z-index" : "2"
                                     });
         
         buttonsContainer.append(addToolTip(copySpan, "Copy to Clipboard"));
      }
      
      if (search)
      {
         // append Search icon
         var searchSpan = $(span).addClass("glyphicon glyphicon-search")
                                 .css({
                                          "float"   : "left",
                                          "margin"  : "4px",
                                          "cursor"  : "pointer",
                                          "z-index" : "2"
                                       });
         
         buttonsContainer.append(addToolTip(searchSpan, "Find"));
      }
      
      if (buttonsContainer)
      {
         titleAndBackDiv.append(buttonsContainer);
      }
      // append title
      var titleContainer = $(div).addClass((buttonsContainer) ? "col-md-6" : "col-md-12");
      var titleSpan = $(span).css({"text-align" : "center", "z-index" : "1"});
      titleSpan.text(title);
      titleContainer.append(titleSpan);
      titleAndBackDiv.append(titleContainer);
      if (buttonsContainer)
      {
         titleAndBackDiv.append($(div).addClass("col-md-3"));
      }
      
      return newContent;
   }
   
   me.addHidingContainer = addHidingContainer;
   
   /**
    * Adds a tooltip to the target widget.
    * 
    * @param   {jQuery} widget
    *          A jQuery node representing a widget
    * @param   (String} hint
    *          A help text
    * 
    * @return  {jQuery} The target widget.
    */
   function addToolTip(widget, hint)
   {
      widget.attr("title", hint);
      
      return widget;
   }
   
   me.addToolTip = addToolTip;
   
   /**
    * Extract the error messages from an exception chain, from most to least specific, and return
    * an array of these messages. If no messages are available, include a generic one.
    * 
    * @param   throwable
    *          An exception thrown by the Java server; we expect each chained exception to have
    *          fields for a cause and a message.
    * 
    * @return  {String[]} An array of minimum length 1, so it should always be safe to call
    *          errorMessages(err)[0];
    */
   function errorMessages(throwable)
   {
      var msgs = [];
      
      var cause = throwable;
      while (cause)
      {
         if (cause.message)
         {
            msgs.unshift(cause.message);
         }
         cause = cause.cause;
      }
      
      if (msgs.length == 0)
      {
         // exception stack was not helpful; punt with a generic error message
         msgs.push("There was an error.");
      }
      
      return msgs;
   }
   
   me.errorMessages = errorMessages;
   
   /**
    * Calculate the available height remaining in the browser's client window area, after
    * subtracting the vertical space consumed by those elements currently in the DOM's <body>
    * element (not including the loading overlay, if present).
    * 
    * @return  Height in pixels available in the browser's client window area.
    */
   function getAvailableHeight()
   {
      var offset = 0;
      
      // always skip loading overlay, dialogs, and their underlays
      var skip = ":not(.loadingoverlay, .dijitDialog, .dijitDialogUnderlayWrapper)";
      
      $("body").children(skip).each(function(i, child)
      {
         offset += child.offsetHeight;
      });
      var height = $(window).outerHeight() - offset;
      
      return height;
   }
   
   me.getAvailableHeight = getAvailableHeight;
   
   /**
    * Get the platform-specific line ending.
    * 
    * @return  The platform-specific line ending.
    */
   function getPlatformSpecificEOL()
   {
      if (me.isWindows)
      {
         return "\r\n";
      }
      else
      {
         return "\n";
      }
   }
   
   me.getPlatformSpecificEOL = getPlatformSpecificEOL;
   
   /* TODO Common code with FWD web client */
   
   Object.defineProperty(
         me,
         "prompts",
         {
            get: function ()
            {
               return function(cmd)
               {
                  var prompt;
                  var template = prompts[cmd];
                  if (template)
                  {
                     prompt = template.replace("{0}", parameters[cmd]);
                  }
                  
                  return prompt;
               }
            }
         });
   
   Object.defineProperty(
         me,
         "hints",
         {
            get: function ()
            {
               return function(cmd)
               {
                  var hint;
                  var template = hints[cmd];
                  if (template)
                  {
                     hint = template.replace("{0}", parameters[cmd]);
                  }
                  
                  return hint;
               }
            }
         });
   
   ["copy", "paste", "cut"].forEach(
      function(cmd)
      {
         try
         {
            queryCommandSupported[cmd] = document.queryCommandSupported(cmd) === true;
         }
         catch (e)
         {
            queryCommandSupported[cmd] = false;
         }
      });
   
   Object.defineProperty(
      me,
      "commandSupported",
      {
         get: function ()
         {
            return function(cmd)
            {
               return queryCommandSupported[cmd];
            }
         }
      });

   /* exports */
   return me;
   
});
