/*
** Module   : report.js
** Abstract : Base reporting services web client implementation.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 ECF 20170503 First version.
** 002 ECF 20170815 Added bucket count to summary report 'Category' column title.
** 003 ECF 20170923 Fixed invalid default dump type for custom report screen.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

/**
 * Report functionality; various tabular report views, a source and AST viewer, and custom
 * search and report features.
 */
define(["wsbase",
        "filesview",
        "dojo",
        "dojo/dom",
        "dijit/dijit",
        "dijit/form/Form",
        "dijit/form/Button",
        "dijit/form/RadioButton",
        "dijit/form/CheckBox",
        "dijit/form/TextBox",
        "dijit/form/ComboBox",
        "dojox/layout/TableContainer",
        "dijit/layout/ContentPane",
        "dijit/ProgressBar",
        "dojo/store/Memory"],
       function(ws, filesView, dojo, dom, dijit, Form, Button, RadioButton, CheckBox, TextBox,
                ComboBox, TableContainer, ContentPane, ProgressBar, Memory)
{
   "use strict";
   
   /** The top-level object which will be returned from this function. */
   var me = {};
   
   /** Starting value for report-related message codes. */
   var MSG_BASE = ws.MSG_RPT_BASE;
   
   /** Message types, which correspond with those defined in the server. */
   var msgTypes =
   {
      MST_CODE_RPT         :  MSG_BASE + 1,
      MST_SCHEMA_RPT       :  MSG_BASE + 2,
      STD_SUM_RPT          :  MSG_BASE + 3,
      DETAIL_RPT           :  MSG_BASE + 4,
      SOURCE_DATA          :  MSG_BASE + 5,
      DEFINE_RPT           :  MSG_BASE + 6,
      DELETE_RPT           :  MSG_BASE + 7,
      RUN_RPT              :  MSG_BASE + 8,
      SEARCH               :  MSG_BASE + 9,
      SEARCH_HISTORY       :  MSG_BASE + 10,
   }
   
   me.msgTypes = msgTypes;
   
   /** Gap analysis support levels, which correspond with those defined in Java. */
   var supportLevels =
   {
      CVT_LVL_NONE         :  0x0001,
      CVT_LVL_PARTIAL      :  0x0002,
      CVT_LVL_BASIC        :  0x0004,
      CVT_LVL_FULL_RESTR   :  0x0008,
      CVT_LVL_FULL         :  0x0010,
      RT_LVL_NONE          :  0x0100,
      RT_LVL_STUB          :  0x0200,
      RT_LVL_UNTESTED      :  0x0400,
      RT_LVL_PARTIAL       :  0x0800,
      RT_LVL_BASIC         :  0x1000,
      RT_LVL_FULL_RESTR    :  0x2000,
      RT_LVL_FULL          :  0x4000,
   }
   
   /** Source code file types */
   var srcTypes =
   {
      SCHEMA     : 0x0001,
      CODE_CACHE : 0x0002,
      CODE_BASE  : 0x0004,
   }
   
   me.srcTypes = srcTypes;
   
   /** List of criteria that have successfully run for searches or custom reports */
   var searchHistory =
   {
      initialized : false,
      list        : [],
   };
   
   /** Dojo memory store for search condition combo box */
   var searchStore = new Memory(
   {
      "idProperty"   : "id",
      "data"         : searchHistory.list,
   });
   
   /**
    * Show the main report view for source code reports.
    */
   me.showCodeView = function()
   {
      showReportsMainView(msgTypes.MST_CODE_RPT,
                          "#codeView",
                          "Source Code Reports",
                          srcTypes.CODE_CACHE,
                          srcTypes.CODE_CACHE);
   }
   
   /**
    * Show the main report view for schema reports.
    */
   me.showSchemaView = function()
   {
      showReportsMainView(msgTypes.MST_SCHEMA_RPT,
                          "#schemaView",
                          "Schema Reports",
                          srcTypes.SCHEMA,
                          srcTypes.SCHEMA | srcTypes.CODE_CACHE);
   }
   
   /**
    * Show the file filter profiles view.
    */
   me.showFilesView = function()
   {
      filesView.setupAndDisplay();
   }
   
   /**
    * Show the administrative interface view.
    */
   me.showAdminView = function()
   {
   }
   
   /**
    * Show the view of online resources related to the FWD Analytics tool set.
    */
   me.showResourcesView = function()
   {
      var parent = $("#resourcesView");
      
      parent.empty();
      
      var div = "<div></div>";
      
      var containDiv = $(div).attr("id", "containerView").addClass("container-fluid");
      parent.append(containDiv);
      
      // append attach point for first row
      var row1Div = $(div).addClass("row form-panel");
      $(containDiv).append(row1Div);
      
      var content =
         "<div>" +
         "<h3>Documentation</h3>" +
         "<p>Please refer to the " +
         "<a href='https://proj.goldencode.com/projects/p2j/wiki/Code_Analytics'" +
         "target='_blank'>online documentation</a> for the FWD Code Analytics tool set.</p>" +
         "<h3>License</h3>" +
         "<p>This software is part of the <a href='https://www.beyondabl.com' target='_blank'>FWD " +
         "Open Source Software Project</a>. It is released under the " +
         "<a href='https://proj.goldencode.com/projects/p2j/wiki/Licensing' target='_blank'>Affero " +
         "GPL 3.0 license</a>. That license, including any additional terms added by Golden Code " +
         "Development Corporation, governs your use of this software, unless you " +
         "<a href='https://www.goldencode.com/licensing-gcd.html' target='_blank'>have purchased " +
         "a superseding license</a> directly from Golden Code Development Corporation.</p>" +
         "<h3>Attributions</h3>" +
         "<ul><li>Icons made by <a href='http://www.flaticon.com/authors/freepik'" +
         "target='_blank'>Freepik</a> from <a href='https://www.flaticon.com'" +
         "target='_blank'>www.flaticon.com</a>.</li>" +
         "<li>Icon made by <a href='http://www.flaticon.com/authors/dimi-kazak'" +
         "target='_blank'>Dimi Kazak</a> from <a href='https://www.flaticon.com'" +
         "target='_blank'>www.flaticon.com</a>.</li></ul>" +
         "<h3>Copyright Notice</h3>" +
         "<p>Copyright &copy; 2004-2017 Golden Code Development Corporation.</p>" +
         "</div>";
      
      var contentDiv = $(content).addClass("basic-text");
      row1Div.append(contentDiv);
   }
   
   /**
    * Display the ad-hoc search view. Allow the user to define and run a search, and to view the
    * results.
    */
   me.showSearchView = function()
   {
      initSearchHistory();
      
      var parent = $("#searchView");
      
      parent.empty();
      
      var div = "<div></div>";
      
      var containDiv = $(div).attr("id", "containerView").addClass("container-fluid");
      parent.append(containDiv);
      
      // append attach point for first row
      var row1Div = $(div).addClass("row");
      $(containDiv).append(row1Div);
      
      var w = dijit.byId("customSearchPanel");
      if (w)
      {
         w.destroy(true);
      }
      
      var attach = row1Div.append($(div).attr("id", "customSearchPanel")
                                        .addClass("form-panel"));
      
      // create content pane
      var panel = new ContentPane(
      {
         content  : "<p style='font: 1.2em bold'>Custom Search</p>",
      }, "customSearchPanel");
      
      // create the radio button form to select source type for the search
      var typeForm = createSourceTypeForm("Search Type");
      
      // create table container for radio set and text box
      var table = new TableContainer(
      {
         cols         : 1,
         labelWidth   : 200,
         customClass  : "customReportTable",
      });
      
      // function to perform search based on user criteria
      var doSearch = function()
      {
         fwd.showOverlay("Searching");
         
         // remove any previous status message and results
         $("#resultsView").remove();
         
         // get the search criteria expression
         var condition = condBox.get("value");
         
         // should we apply the current filter?
         var applyFilter = filterForm.checkBox.get("checked");
         
         // invoke the search API
         ws.sendMessage(
            msgTypes.SEARCH,
            [ condition, applyFilter, typeForm.sourceType ],
            function(err, data)
            {
               fwd.hideOverlay();
               
               // report an error if there was one
               if (err)
               {
                  fwd.showDialog("Error", fwd.errorMessages(err)[0]);
                  
                  return;
               }
               
               // store successful condition expression in local history
               storeLocalSearchHistory(condition);
               
               // get data returned by server
               var rows = data[0];
               var matches = data[1];
               var files = data[2];
               
               // check for results
               if (!matches)
               {
                  fwd.showDialog("Search Complete",
                                 "No matches found for the given criteria.",
                                 5000);
                  
                  return;
               }
               
               // clear status message and display results
               $("#resultsView").remove();
               var resultsView = $(div).attr("id", "resultsView");
               row1Div.append(resultsView);
               var height = fwd.getAvailableHeight() - 16;
               var sumData =
               {
                  cat   : condition,
                  mCnt  : matches,
                  fCnt  : files,
               };
               showDetailRows(resultsView,
                              "#containerView",
                              height,
                              sumData,
                              rows,
                              typeForm.sourceType,
                              false);
            }
         );
      };
      
      // create widgets
      var filterForm =
         createLabelCheckBox("Only search files included in active file filter",
                             true,
                             "filterCheck",
                             "filter"/*,
                             function(checked) { applyFilter = checked; }*/);
      var condBox = createConditionComboBox();
      var runButton = new Button(
      {
         label   : "Run",
         onClick : doSearch,
      });
      
      // add child widgets to the TableContainer
      table.addChild(filterForm.form);
      table.addChild(typeForm.form);
      table.addChild(condBox);
      
      // add child widgets to the panel, start it, and show it
      panel.addChild(table);
      panel.addChild(runButton);
      panel.startup();
      parent.show();
      
      // set initial focus on criteria field
      condBox.focus();
   }
   
   /**
    * Display the custom report view. Allow the user to define and run a report. Report results
    * are added to the database, so they appear with all other results in the code or schema
    * reports views.
    */
   me.showCustomReportView = function()
   {
      initSearchHistory();
      
      var parent = $("#customReportView");
      
      parent.empty();
      
      var div = "<div></div>";
      
      var containDiv = $(div).attr("id", "containerView").addClass("container-fluid");
      parent.append(containDiv);
      
      // append attach point for first row
      var row1Div = $(div).addClass("row");
      $(containDiv).append(row1Div);
      
      var w = dijit.byId("customReportPanel");
      if (w)
      {
         w.destroy(true);
      }
      
      var attach = row1Div.append($(div).attr("id", "customReportPanel")
                                        .addClass("form-panel"));
      
      var panel = new ContentPane(
      {
         content  : "<p style='font: 1.2em bold'>Define A Custom Report</p>",
      }, "customReportPanel");
      
      var typeForm = createSourceTypeForm("Report Type");
      
      var table = new TableContainer(
      {
         cols         : 1,
         labelWidth   : 200,
         customClass  : "customReportTable",
      });
      
      // Create text boxes and buttons
      var style = "width: 60em;";
      var titleTextBox = new TextBox(
      {
         label       : "Report Title",
         placeHolder : "Enter the report's title",
         style       : style,
      });
      var condBox = createConditionComboBox();
      var mplexTextBox = new TextBox(
      {
         label       : "Multiplex Expression",
         placeHolder : "Enter an expression that evaluates to a string which will define " +
                       "match categories (optional)",
         style       : style,
      });
      var tagsTextBox = new TextBox(
      {
         label       : "Category Tags",
         placeHolder : "Enter a comma-separated list of tags to group this report with similar " +
                       "reports (optional)",
         style       : style,
      });
      var dumpTypeForm = createDumpTypeForm("Match Text Format");
      var dumpExprBox = new TextBox(
      {
         label       : "Match Expression",
         placeHolder : "Enter an expression that evaluates to a string for custom match text " +
                       "(optional)",
         style       : style,
      });
      var dumpLevelBox = new TextBox(
      {
         label       : "Match Text Level",
         placeHolder : "Enter a non-negative number of ancestor levels up the tree from the " +
                       "matched node to extract match text (defaults to 0)",
         style       : style,
      });
      
      var doRun = function()
      {
         // disable the button to prevent multiple clicks while search is running
         this.set("disabled", true);
         
         // gather inputs
         var title = titleTextBox.get("value");
         var condition = condBox.get("value");
         var tagText = tagsTextBox.get("value");
         var tags = [];
         if (tagText != null && tagText.trim().length > 0)
         {
            // trim whitespace from front and back of each tag
            var rawTags = tagText.split(",");
            rawTags.forEach(function(e) { tags.push(e.trim()); });
         }
         var rptdef =
         {
            title          :  title,
            tags           :  tags,
            condition      :  condition,
            conditionDescr :  condition,
            multiplexExpr  :  mplexTextBox.get("value"),
            dumpType       :  dumpTypeForm.dumpType,
            dumpExpr       :  dumpExprBox.get("value"),
            dumpLevel      :  dumpLevelBox.get("value"),
         };
         
         // invoke server API to define report
         ws.sendMessage(
            msgTypes.DEFINE_RPT,
            [ rptdef, typeForm.sourceType ],
            function(err, rid)
            {
               fwd.showOverlay("Running Report");
               
               // invoke server API to run report
               ws.sendMessage(
                  msgTypes.RUN_RPT,
                  [ rid ],
                  function(err)
                  {
                     fwd.hideOverlay();
                     
                     // re-enable the run button
                     runButton.set("disabled", false);
                     
                     if (err)
                     {
                        // report an error
                        fwd.showDialog("Error", fwd.errorMessages(err)[0]);
                        
                        // delete the broken report
                        ws.sendMessage(msgTypes.DELETE_RPT, [ rid ], function(err) {});
                        
                        return;
                     }
                     
                     // store successful condition expression in local history
                     storeLocalSearchHistory(condition);
                     
                     // report success
                     var loc = (typeForm.sourceType == srcTypes.CODE_CACHE ? "Code" : "Schema");
                     var msg = "Please see results in the " + loc + " Reports area.";
                     
                     fwd.showDialog("Report Complete", msg, 5000);
                  }
               );
            }
         );
      };
      
      var runButton = new Button(
      {
         label   : "Run/Add",
         onClick : doRun,
      });
      
      var clearButton = new Button(
      {
         label    : "Clear Fields",
         onClick  : function()
         {
            titleTextBox.set("value", "");
            condBox.set("value", "");
            mplexTextBox.set("value", "");
            tagsTextBox.set("value", "");
            dumpExprBox.set("value", "");
            dumpLevelBox.set("value", "");
         },
      });
      
      // add widgets to the TableContainer
      table.addChild(typeForm.form);
      table.addChild(titleTextBox);
      table.addChild(condBox);
      table.addChild(mplexTextBox);
      table.addChild(tagsTextBox);
      table.addChild(dumpTypeForm.form);
      table.addChild(dumpExprBox);
      table.addChild(dumpLevelBox);
      
      // add widgets to panel, start it, show it
      panel.addChild(table);
      panel.addChild(runButton);
      panel.addChild(clearButton);
      panel.startup();
      parent.show();
      
      // set initial focus on title field
      titleTextBox.focus();
   }
   
   /**
    * Initialize the local store of search history TRPL expressions from the server.
    */
   function initSearchHistory()
   {
      if (searchHistory.initialized)
      {
         return;
      }
      
      searchHistory.initialized = true;
      
      ws.sendMessage(msgTypes.SEARCH_HISTORY, null, function(err, data)
      {
         if (err)
         {
            // not fatal, so just log it
            console.log("Unable to retrieve search history: " + fwd.errorMessages(err)[0]);
            
            return;
         }
         
         if (data)
         {
            // copy elements, rather than just assigning array variable, because the searchHistory
            // content variable already will have been referenced by the time this async response
            // arrives from server
            data.forEach(function(elem, i)
            {
               searchHistory.list.push({ "id" : (i + 1), "expr" : elem });
            });
         }
      });
   }
   
   /**
    * Store a successful condition expression in the client's search history list.
    * 
    * @param   {String} condition
    *          TRPL search condition expression.
    */
   function storeLocalSearchHistory(condition)
   {
      // store successful condition expression in history
      var qr = searchStore.query({"expr" : condition});
      if (qr == undefined || qr.total == 0)
      {
         searchHistory.list.unshift({ "id" : searchHistory.length, "expr" : condition });
      }
   }
   
   /**
    * Create a combo box to manage a history of successful search condition expressions.
    * 
    * @return  {ComboBox} The combo box.
    */
   function createConditionComboBox()
   {
      return new ComboBox(
      {
         label       : "Condition Expression",
         style       : "width: 60em;",
         placeHolder : "Enter an expression that describes the AST for which to search",
         store       : searchStore,
         searchAttr  : "expr",
      });
   }
   
   /**
    * Create a simple form with a label and two radio buttons: "code" and "schema", suitable for
    * placement within a TableContainer.
    * 
    * @param   {String} formLabel
    *          Label for the combined radio set.
    * 
    * @return  {Object} An object with two fields: "form", which is a reference to the Dojo Form
    *          widget; and "sourceType", which indicates the source type constant associated with
    *          the currently selected radio button.
    */
   function createSourceTypeForm(formLabel)
   {
      var f = {};
      
      f.sourceType = srcTypes.CODE_CACHE;
      
      // create the "code" radio button widget
      var codeRadio = new RadioButton(
      {
         checked  :  f.sourceType == srcTypes.CODE_CACHE,
         value    :  "code",
         name     :  "type",
         onClick  :  function() { f.sourceType = srcTypes.CODE_CACHE; },
      });
      
      // create the "schema" radio button widget
      var schemaRadio = new RadioButton(
      {
         checked  :  f.sourceType == srcTypes.SCHEMA,
         value    :  "schema",
         name     :  "type",
         style    :  "margin-left: 10px;",
         onClick  :  function() { f.sourceType = srcTypes.SCHEMA; },
      });
      
      // create the form itself, append the radio button widgets, and assign labels to the radio
      // buttons
      f.form = new Form({ label: formLabel });
      var radioDom = f.form.domNode;
      radioDom.appendChild(codeRadio.domNode);
      dojo.create("label", { innerHTML: "code", for: codeRadio.id }, radioDom);
      radioDom.appendChild(schemaRadio.domNode);
      dojo.create("label", { innerHTML: "schema", for: schemaRadio.id }, radioDom);
      
      return f;
   }
   
   /**
    * Create a simple form with a labeled checkbox, where the label appears to the right of the
    * checkbox.
    * 
    * @param   {String} label
    *          The label to affix to the checkbox.
    * @param   {Boolean} checked
    *          Initial state of the checkbox; true is checked, false is unchecked.
    * @param   {String} value
    *          Checkbox value attribute.
    * @param   {String} name
    *          Checkbox name attribute.
    * 
    * @return  {Object} An object with two properties: form (the Dojo Form object which contains
    *          the checkbox widget and its corresponding label) and checkBox (a reference to
    *          the Dojo CheckBox widget).
    */
   function createLabelCheckBox(label, checked, value, name)
   {
      var f = {};
      
      var checkBox = new CheckBox(
      {
         checked  :  checked,
         value    :  value,
         name     :  name,
      });
      
      // create the form itself, append the checkbox widget, and assign label to it
      f.form = new Form({});
      var formDom = f.form.domNode;
      formDom.appendChild(checkBox.domNode);
      dojo.create("label",
                  {
                     innerHTML : "<span style='padding-left: 5px'>" + label + "</span>",
                     for       : checkBox.id
                  },
                  formDom);
      f.checkBox = checkBox;
      
      return f;
   }
   
   /**
    * Create a simple form with a label and two radio buttons: "parser" and "simple", suitable for
    * placement within a TableContainer.
    * 
    * @param   {String} formLabel
    *          Label for the combined radio set.
    * 
    * @return  {Object} An object with two fields: "form", which is a reference to the Dojo Form
    *          widget; and "dumpType", which indicates the match text dump type string associated
    *          with the currently selected radio button.
    */
   function createDumpTypeForm(formLabel)
   {
      var f = {};
      
      f.dumpType = "parser";
      
      // create the "code" radio button widget
      var parserRadio = new RadioButton(
      {
         checked  :  true,
         value    :  "parser",
         name     :  "type",
         onClick  :  function() { f.dumpType = this.get("value") },
      });
      
      // create the "schema" radio button widget
      var simpleRadio = new RadioButton(
      {
         checked  :  false,
         value    :  "simple",
         name     :  "type",
         style    :  "margin-left: 10px;",
         onClick  :  function() { f.dumpType = this.get("value"); },
      });
      
      // create the form itself, append the radio button widgets, and assign labels to the radio
      // buttons
      f.form = new Form({ label: formLabel });
      var radioDom = f.form.domNode;
      radioDom.appendChild(parserRadio.domNode);
      dojo.create("label", { innerHTML: "parser", for: parserRadio.id }, radioDom);
      radioDom.appendChild(simpleRadio.domNode);
      dojo.create("label", { innerHTML: "simple", for: simpleRadio.id }, radioDom);
      
      return f;
   }
   
   /**
    * Display the main report view specific to the given selector. This includes a high-level
    * overview of all reports specific to the selector (source code or schema), as well as a
    * summary view of the report currently selected in the overview table.
    * 
    * @param   {Number} msgType
    *          Message type identifier.
    * @param   {String} selector
    *          Selector for the DOM element to which this view will be attached.
    * @param   {String} title
    *          Title to display as the overview report heading.
    * @param   {Number} astType
    *          Type of ASTs associated with the detail reports accessible from this report.
    * @param   {Number} srcType
    *          Type of source file associated with this report. May be a bitwise-OR'd combination
    *          of multiple types.
    */
   function showReportsMainView(msgType, selector, title, astType, srcType)
   {
      var parent = $(selector);
      
      parent.empty();
      
      var div = "<div></div>";
      
      var containDiv = $(div).attr("id", "containerView").addClass("container-fluid");
      parent.append(containDiv);
      
      // append attach point for first row
      var row1Div = $(div).addClass("row");
      $(containDiv).append(row1Div);
      
      fwd.showOverlay("Building Overview Report");
      
      ws.sendMessage(msgType, null, function(err, rows)
      {
         fwd.hideOverlay();
         
         var r1c1Div = $(div).addClass("col-md-4").css("padding", "3px");
         row1Div.append(r1c1Div);
         
         var r1c2Div = $(div).addClass("col-md-8").css("padding", "3px");
         row1Div.append(r1c2Div);
         
         // append title
         var tocTitleDiv = $(div).addClass("toc-heading").text(title);
         r1c1Div.append(tocTitleDiv);
         
         var tabDiv = $(div).attr("id", "overviewTable");
         r1c1Div.append(tabDiv);
         
         var reselect = null;
         
         // initialize the tabulator
         tabDiv.tabulator(
         {
            fitColumns           :  true, // fit columns to width of table
            height               :  (fwd.getAvailableHeight() - 16) + "px",
            groupBy              :  "tag",
            groupStartOpen       :  function(value, count, data)
                                    {
                                       return (astType == srcTypes.SCHEMA
                                               ? value == "Tables"
                                               : value == "Project Structure");
                                    },
            selectable           :  1,
            sortable             :  false,
            tooltipsHeader       :  true,
            columns              :
            [
               {
                  title          :  "Report Category / Name",
                  field          :  "title",
                  tooltip        :  true,
                  headerFilter   :  true,
               },
               {
                  title          :  "Matches",
                  field          :  "matches",
                  align          :  "center",
                  width          :  "15%",
                  tooltip        :  function(field, value, data)
                                    {
                                       return conditionTooltip(data);
                                    },
                  formatter      :  function(value, data, cell, row, options, formatterParams)
                                    {
                                       return numberFormat(value);
                                    },
               },
               {
                  title          :  "Files",
                  field          :  "files",
                  align          :  "center",
                  width          :  "15%",
                  tooltip        :  function(field, value, data)
                                    {
                                       return conditionTooltip(data);
                                    },
                  formatter      :  function(value, data, cell, row, options, formatterParams)
                                    {
                                       return numberFormat(value);
                                    },
               },
               {
                  title          :  "Category",
                  field          :  "tag",
                  visible        :  false,
               },
               {
                  field          :  "condition",
                  visible        :  false,
               }
            ],
            rowSelectionChanged  :  function(data, rows)
                                    {
                                       if (data.length == 0 && reselect != null)
                                       {
                                          // if reselect is not null, it means a click has
                                          // deselected a selected row, but we want it to remain
                                          // selected, so we manually force selection on it
                                          tabDiv.tabulator("selectRow", reselect);
                                          reselect = null;
                                       }
                                       else if (reselect == null || $("#detailReport").length)
                                       {
                                          // if row is newly selected or user has replaced the
                                          // summary report with a detail report, (re)display the
                                          // summary report for the selected row
                                          showSummaryReport(r1c2Div, data[0], astType, srcType);
                                       }
                                    },
            rowClick             :  function(e, id, data, row)
                                    {
                                       // A click will both select and de-select a row, but we
                                       // want a selected row to stay selected on subsequent
                                       // clicks, and we don't want to reload a summary report
                                       // if it's already loaded. So, if we determine a click
                                       // has deselected a row, remember the row data and let
                                       // the rowSelectionChanged handler re-select it
                                       var selected = tabDiv.tabulator("getSelectedData");
                                       reselect = selected.length > 0 && selected[0] === data
                                                  ? data
                                                  : null;
                                    },
            tableBuilt           :  function()
                                    {
                                       if (rows.length > 0)
                                       {
                                          var data = rows[0];
                                          tabDiv.tabulator("selectRow", data);
                                       }
                                       
                                       // re-render table on window resize
                                       $(window).resize(function()
                                       {
                                          tabDiv.tabulator("redraw");
                                       });
                                    },
            data                 :  rows,
         });
      });
   }
   
   /**
    * Compose tool tip text summarizing a match condition expression and the number of matches
    * and files associated with it for a given report
    * 
    * @param   {Object} data
    *          Data object describing an overview report row.
    * 
    * @return  {String} Tool tip text.
    */
   function conditionTooltip(data)
   {
      return data.condition +
             "   (" +
             numberFormat(data.matches) +
             " matches in " +
             numberFormat(data.files) + 
             " files)";
   }
   
   /**
    * Display a summary report at the given DOM selector, showing each match category, along with
    * the number of matches and files for which that category was matched across the active file
    * set.
    * 
    * @param   {String} selector
    *          Selector for the DOM element to which this view will be attached.
    * @param   {Object} data
    *          Summary row data from the overview report table; provides a summary report ID,
    *          title, and number of matches and files.
    * @param   {Number} astType
    *          Type of ASTs associated with the detail reports accessible from this report.
    * @param   {Number} srcType
    *          Type of source file associated with this report. May be a bitwise-OR'd combination
    *          of multiple types.
    */
   function showSummaryReport(selector, data, astType, srcType)
   {
      var parent = $(selector);
      
      parent.empty();
      
      var div = "<div></div>";
      
      // append title
      var title = data.title + " (" +
                  numberFormat(data.matches) +
                  " in " +
                  numberFormat(data.files) +
                  " files)";
      var sumTitleDiv = $(div).addClass("toc-heading").text(title);
      parent.append(sumTitleDiv);
      
      // append attach point for tabulator
      var tabDiv = $(div);
      
      var condition = data.condition;
      
      fwd.showOverlay("Building Summary Report");
      
      var parms = [ data.id, data.matches, data.files ];
      ws.sendMessage(msgTypes.STD_SUM_RPT, parms, function(err, reportInfo)
      {
         fwd.hideOverlay();
         
         var isCustomReport = reportInfo[0];
         
         parent.append(tabDiv);
         
         if (isCustomReport)
         {
            var report = reportInfo[1];
            var colDefs = report.colDefs;
            var rows = report.rows;
            
            var len = colDefs.length;
            
            // calculate optimal widths for numeric columns using the rough metric that text
            // columns should be twice as wide as numeric columns
            var divisor = len;
            for (var i = 0; i < len; i++)
            {
               if (colDefs[i].format == null)
               {
                  divisor++;
               }
            }
            var numWidth = (100.0 / divisor) + "%";
            
            var tabCols = [];
            for (var i = 0; i < len; i++)
            {
               var colDef = colDefs[i];
               var fmt = colDef.format;
               var nextTabCol = new Object();
               nextTabCol.title = colDef.heading;
               nextTabCol.field = "" + i;
               if (fmt == null)
               {
                  nextTabCol.format = "string";
               }
               else
               {
                  nextTabCol.align = "center";
                  nextTabCol.width = numWidth;
                  nextTabCol.formatterParams = fmt;
                  nextTabCol.formatter = function(value, data, cell, row, options, format)
                                         {
                                            return numberFormat(value, format);
                                         }
               }
               tabCols.push(nextTabCol);
            }
            
            // initialize the custom report tabulator
            tabDiv.tabulator(
            {
               fitColumns           : true,
               height               : $("#overviewTable").outerHeight() + "px",
               sortable             : false,
               selectable           : false,
               pagination           : "local",
               paginationSize       : 150,
               tooltipsHeader       : true,
               columns              : tabCols,
               data                 : rows,
            });
         }
         else
         {
            var rows = reportInfo[1];
            var reselect = null;
            
            // initialize the summary report tabulator
            tabDiv.tabulator(
            {
               fitColumns           : true, // fit columns to width of table
               height               : $("#overviewTable").outerHeight() + "px",
               sortable             : true,
               selectable           : 1,
               pagination           : "local",
               paginationSize       : 150,
               tooltipsHeader       : true,
               columns              :
               [
                  {
                     title          :  "Category (" + numberFormat(rows.length) + " total)",
                     field          :  "cat",
                     sorter         :  "string",
                     headerFilter   :  true,
                     tooltipHeader  :  function(col)
                                       {
                                          return "Condition matched: " + condition;
                                       },
                     tooltip        :  function(field, value, data)
                                       {
                                          return value + " (" + data.mplex + ")";
                                       },
                  },
                  {
                     title          :  "Conversion",
                     field          :  "cLvl",
                     align          :  "center",
                     width          :  "10%",
                     sorter         :  "number",
                     visible        :  data.support,
                     tooltipHeader  :  "Conversion Support Level",
                     tooltip        :  function(field, value, data)
                                       {
                                          return supportDescription(value);
                                       },
                     formatter      :  function(value, data, cell, row, options, formatterParams)
                                       {
                                          return decodeSupport(value);
                                       },
                  },
                  {
                     title          :  "Runtime",
                     field          :  "rLvl",
                     align          :  "center",
                     width          :  "10%",
                     sorter         :  "number",
                     visible        :  data.support,
                     tooltipHeader  :  "Runtime Support Level",
                     tooltip        :  function(field, value, data)
                                       {
                                          return supportDescription(value);
                                       },
                     formatter      :  function(value, data, cell, row, options, formatterParams)
                                       {
                                          return decodeSupport(value);
                                       },
                  },
                  {
                     title          :  "Matches",
                     field          :  "mCnt",
                     align          :  "center",
                     width          :  "10%",
                     sorter         :  "number",
                     tooltip        :  true,
                     formatter      :  function(value, data, cell, row, options, formatterParams)
                                       {
                                          return numberFormat(value);
                                       },
                  },
                  {
                     title          :  "%",
                     field          :  "mPct",
                     align          :  "center",
                     width          :  "7%",
                     tooltip        :  function(field, value, data)
                                       {
                                          return percentToolTip(value);
                                       },
                     sorter         :  "number",
                     formatter      :  "progress",
                  },
                  {
                     title          :  "Files",
                     field          :  "fCnt",
                     align          :  "center",
                     width          :  "10%",
                     sorter         :  "number",
                     tooltip        :  true,
                     formatter      :  function(value, data, cell, row, options, formatterParams)
                                       {
                                          return numberFormat(value);
                                       },
                  },
                  {
                     title          :  "%",
                     field          :  "fPct",
                     align          :  "center",
                     width          :  "7%",
                     tooltip        :  function(field, value, data)
                                       {
                                          return percentToolTip(value);
                                       },
                     sorter         :  "number",
                     formatter      :  "progress",
                  },
               ],
               data                 :  rows,
               rowSelectionChanged  :  function(data, rows)
                                       {
                                          // if reselect is not null, it means a click has
                                          // deselected a selected row, but we want it to remain
                                          // selected, so we manually force selection on it
                                          if (data.length == 0 && reselect != null)
                                          {
                                             tabDiv.tabulator("selectRow", reselect);
                                             reselect = null;
                                          }
                                       },
               rowClick             :  function(e, id, data, row)
                                       {
                                          // A click will both select and de-select a row, but we
                                          // want a selected row to stay selected on subsequent
                                          // clicks. So, if we determine a click has deselected a
                                          // row, remember the row data and let the
                                          // rowSelectionChanged handler re-select it.
                                          var selected = tabDiv.tabulator("getSelectedData");
                                          reselect = selected.length > 0 && selected[0] === data
                                                   ? data
                                                   : null;
                                          
                                          // display detail report associated with this row
                                          showDetailReport(tabDiv, data, astType, srcType);
                                       },
            });
         }
         
         // re-render table on window resize
         $(window).resize(function()
         {
            tabDiv.tabulator("redraw");
         });
      });
   }
   
   /**
    * Display a detail report at the given DOM selector, showing each file, line, and column
    * number at which a match occurs, as well as the text that was matched.
    * 
    * @param   {Selection} selection
    *          jQuery selection for the DOM element to which this view will be attached.
    * @param   {Object} sumData
    *          Match category row data from the summary report table; provides match category
    *          name, along with the number of matches and files in which that match category
    *          occurs.
    * @param   {Number} astType
    *          Type of AST (schema or code) associated with this detail report.
    * @param   {Number} srcType
    *          Type of source file associated with this report. May be a bitwise-OR'd combination
    *          of multiple types.
    */
   function showDetailReport(selection, sumData, astType, srcType)
   {
      fwd.showOverlay("Building Detail Report");
      
      ws.sendMessage(msgTypes.DETAIL_RPT, [ sumData.cid, astType ], function(err, detailRows)
      {
         fwd.hideOverlay();
         
         var height = $("#overviewTable").outerHeight();
         showDetailRows(selection, "#containerView", height, sumData, detailRows, srcType, true);
      });
   }
   
   /**
    * Show a tabulator showing each file, line, and column number at which a match occurs, as
    * well as the text that was matched.
    * 
    * @param   {jQuery selection} attachDiv
    *          The selection describing the DOM element to which this view will be attached.
    * @param   {String} container
    *          Selector for the container to which the source/ast view will be attached, if the
    *          user drills down into a file.
    * @param   {Number} height
    *          Height to allocate to the tabulator.
    * @param   {Object} sumData
    *          Match category data which contains match category desciption, along with the
    *          number of matches and files in which that match category occurs.
    * @param   {Object[]} rows
    *          Detail row data.
    * @param   {Number} srcType
    *          Type of source file associated with this report. May be a bitwise-OR'd combination
    *          of multiple types.
    * @param   {Boolean} back
    *          True to display a back arrow; false if there is no previous view to return to.
    */
   function showDetailRows(attachDiv, container, height, sumData, rows, srcType, back)
   {
      var title = sumData.cat + " (" +
                  numeral(sumData.mCnt).format("0,0") +
                  " in " +
                  numeral(sumData.fCnt).format("0,0") +
                  " files)";
      
      var newContent = fwd.addHidingContainer(attachDiv, title, back, false, false);
      
      // append attach point for tabulator
      var div = "<div></div>";
      var tabDiv = $(div);
      tabDiv.attr("id", "detailReport");
      newContent.append(tabDiv);
      
      var reselect = null;
      
      // initialize the detail report tabulator
      tabDiv.tabulator(
      {
         fitColumns           : true, // fit columns to width of table
         height               : (height - newContent.outerHeight())  + "px",
         groupBy              : function(data)
                                {
                                   return sumData.mCnt == sumData.fCnt ? sumData.cat : data.file;
                                },
         groupStartOpen       : true,
         sortable             : true,
         selectable           : 1,
         pagination           : "local",
         paginationSize       : 150,
         tooltipsHeader       : true,
         columns              :
         [
            {
               title          :  "File Name",
               field          :  "file",
               tooltip        :  true,
               sorter         :  "string",
               width          :  "28%",
               headerFilter   :  true,
            },
            {
               title          :  "Line",
               field          :  "line",
               align          :  "right",
               sorter         :  "number",
               width          :  "7%",
               formatter      :  function(value, data, cell, row, options, formatterParams)
                                 {
                                    return numberFormat(value);
                                 },
            },
            {
               title          :  "Col",
               field          :  "col",
               align          :  "left",
               sortable       :  false,
               width          :  "5%",
               formatter      :  function(value, data, cell, row, options, formatterParams)
                                 {
                                    return numberFormat(value);
                                 },
            },
            {
               title          :  "Match Text",
               field          :  "text",
               tooltip        :  true,
               sorter         :  "string",
               headerFilter   :  true,
            },
         ],
         data                 :  rows,
         rowSelectionChanged  :  function(data, rows)
                                 {
                                    // if reselect is not null, it means a click has
                                    // deselected a selected row, but we want it to remain
                                    // selected, so we manually force selection on it
                                    if (data.length == 0 && reselect != null)
                                    {
                                       tabDiv.tabulator("selectRow", reselect);
                                       reselect = null;
                                    }
                                 },
         rowClick             :  function(e, id, data, row)
                                 {
                                    // A click will both select and de-select a row, but we
                                    // want a selected row to stay selected on subsequent
                                    // clicks. So, if we determine a click has deselected a
                                    // row, remember the row data and let the
                                    // rowSelectionChanged handler re-select it.
                                    var selected = tabDiv.tabulator("getSelectedData");
                                    reselect = selected.length > 0 && selected[0] === data
                                             ? data
                                             : null;
                                    
                                    // display detail report associated with this row
                                    showSourceAstView(container,
                                                      data.fid,
                                                      data.file,
                                                      data.astid,
                                                      srcType,
                                                      data.line,
                                                      data.col,
                                                      sumData.cat);
                                 },
      });
      
      // re-render table on window resize
      $(window).resize(function()
      {
         tabDiv.tabulator("redraw");
      });
   }
   
   /**
    * Add to the DOM a set of elements to show the source code and AST viewers for a particular
    * source file. The source code viewer displays the source code in a scrollable viewport.
    * The AST viewer is a visualization of the AST associated with the file. The views are
    * linked by events, such that interacting with one affects the state of the other.
    * 
    * @param   {String} selector
    *          Selector identifying parent HTML element to which attach the div element
    *          containing the viewers.
    * @param   {Number} fid
    *          Unique identifier for a source file.
    * @param   {String} file
    *          Name of the file to be displayed.
    * @param   {Number} astId
    *          Unique identifier for the root node of the AST corresponding with the file to
    *          be displayed in the viewers.
    * @param   {Number} srcType
    *          Type of source file associated with this report. May be a bitwise-OR'd combination
    *          of multiple types.
    * @param   {Number} atLine
    *          Optional line number to scroll the source code viewport to initially. If 0 or
    *          null, the source code viewer will open to the start of the file.
    * @param   {Number} atColumn
    *          Optional column number to scroll the source code viewport to initially. Not
    *          actually used currently.
    * @param   {String} description
    *          Descriptive text which is displayed as a heading to the view. Should not be
    *          overly long (a few words or short expression).
    * @param   {Function} backFunc
    *          Optional callback function to execute upon clicking back button, after redisplaying
    *          old content.
    */
   function showSourceAstView(selector,
                              fid,
                              file,
                              astId,
                              srcType,
                              atLine,
                              atColumn,
                              description,
                              backFunc)
   {
      var parent = $(selector);
      
      var title = file;
      if (description)
      {
         title += (" (" + description + ")");
      }
      
      var newContent = fwd.addHidingContainer(parent, title, true, true, true, backFunc);
      
      var div = "<div></div>";
      
      // append attach point for row
      var rowDiv = $(div).addClass("row");
      $(newContent).append(rowDiv);
      
      // append attach point for left column (source code view)
      var srcDiv = $(div).attr("id", "srcView").addClass("col-md-6 no-padding");
      $(rowDiv).append(srcDiv);
      
      // append attach point for right column (AST view)
      var astDiv = $(div).attr("id", "astView").addClass("col-md-6 no-padding");
      $(rowDiv).append(astDiv);
      
      var height = fwd.getAvailableHeight() - 16;
      
      srcDiv.css(
      {
         "height"          : height + "px",
         "display"         : "flex",
         "align-items"     : "center",
         "justify-content" : "center"
      });
      
      astDiv.css(
      {
         "height"          : height + "px",
         "display"         : "flex",
         "align-items"     : "center",
         "justify-content" : "center"
      });
      
      ws.sendMessage(msgTypes.SOURCE_DATA, [ fid, astId, srcType ], function(err, data)
      {
         fwd.ast.composeAst(data[1]);
         var $astView = $("#astView");
         $astView.css("background-color", "#ddd");
         fwd.astview.prepare("#astView", $astView.width(), $astView.height());
         fwd.srcview.show("#srcView", height, data[0], data[1], atLine, atColumn);
      });
   }
   
   // export
   me.showSourceAstView = showSourceAstView;
   
   /**
    * Format a number according to the specified, Numeral.js format phrase.
    * 
    * @param   {Number} value
    *          Number to be formatted.
    * @param   {String} format
    *          Optional, Numeral.js format phrase (defaults to "0,0").
    */
   function numberFormat(value, format)
   {
      if (format == undefined)
      {
         format = "0,0";
      }
      
      return numeral(value).format(format);
   }
   
   // export
   me.numberFormat = numberFormat
   
   /**
    * Format a number as a percentage with two decimal places precision, for use as a tooltip
    * with a tabulator.js, progress format field. The tabulator expects a number from 0-100 when
    * using the progress format, so we need to divide the value by 100 before applying the
    * Numeral.js format phrase, which expects a number from 0-1 for percentage formatting.
    */
   function percentToolTip(value)
   {
      return numeral(value / 100).format("0.00%");
   }
   
   /**
    * Get a verbose description of a support level, suitable for use as a tooltip.
    * 
    * @param   {Number} lvl
    *          A support level constant.
    * 
    * @return  {String} Support level description.
    */
   function supportDescription(lvl)
   {
      var text = "Unknown";
      
      switch (lvl)
      {
         case supportLevels.CVT_LVL_NONE:
         case supportLevels.RT_LVL_NONE:
            text = "None";
            break;
         case supportLevels.RT_LVL_STUB:
            text = "Runtime functionality stubbed out, but not implemented";
            break;
         case supportLevels.RT_LVL_UNTESTED:
            text = "Runtime implementation, but needs testing";
            break;
         case supportLevels.CVT_LVL_PARTIAL:
         case supportLevels.RT_LVL_PARTIAL:
            text = "Partial support implemented";
            break;
         case supportLevels.CVT_LVL_BASIC:               
         case supportLevels.RT_LVL_BASIC:
            text = "Basic support implemented; some limitations";
            break;
         case supportLevels.CVT_LVL_FULL_RESTR:
         case supportLevels.RT_LVL_FULL_RESTR:
            text = "Full support implemented, with permanent restrictions";
            break;
         case supportLevels.CVT_LVL_FULL:
         case supportLevels.RT_LVL_FULL:
            text = "Full support implemented";
            break;
      }
      
      return text;
   }
   
   /**
    * Given a support level constant, return a div tag containing a concise description of the
    * support level, using foreground and background colors appropriate for that level.
    * 
    * @param   {Number} lvl
    *          A support level constant.
    * 
    * @return  {String} Text for a div DOM element to be displayed in a tabulator table cell.
    */
   function decodeSupport(lvl)
   {
      var text = "Unknown";
      
      var back = "rgba(128, 128, 128, 0.65)";
      var fore = "#000000";
      
      switch (lvl)
      {
         case supportLevels.CVT_LVL_NONE:
         case supportLevels.RT_LVL_NONE:
            text = "None";
            // red
            back = "rgba(255, 0, 0, 0.65)";
            fore = "#FFFFFF";
            break;
         case supportLevels.RT_LVL_STUB:
            text = "Stubs";
            // dark orange
            back = "rgba(255, 80, 0, 0.65)";
            fore = "#FFFFFF";
            break;
         case supportLevels.RT_LVL_UNTESTED:
            text = "Untested";
            // medium orange
            back = "rgba(255, 165, 0, 0.65)";
            fore = "#000000";
            break;
         case supportLevels.CVT_LVL_PARTIAL:
         case supportLevels.RT_LVL_PARTIAL:
            text = "Partial";
            // bright yellow
            back = "rgba(255, 255, 0, 0.65)";
            fore = "#000000";
            break;
         case supportLevels.CVT_LVL_BASIC:               
         case supportLevels.RT_LVL_BASIC:
            text = "Basic";
            // light yellow-green
            back = "rgba(191, 255, 0, 0.65)";
            fore = "#000000";
            break;
         case supportLevels.CVT_LVL_FULL_RESTR:
         case supportLevels.RT_LVL_FULL_RESTR:
            text = "Full (R)";
            // yellow-green
            back = "rgba(96, 204, 0, 0.65)";
            fore = "#000000";
            break;
         case supportLevels.CVT_LVL_FULL:
         case supportLevels.RT_LVL_FULL:
            text = "Full";
            // good green
            back = "rgba(0, 153, 0, 0.65)";
            fore = "#FFFFFF";
            break;
      }
      
      return "<div style='color: " + fore +
             "; background-color: " + back + ";'>" + text + "</div>";
   }
   
   /* exports */
   return me;
   
});
