/*
** Module   : srcview.js
** Abstract : Source code viewer.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 ECF 20170526 First version.
**     SBI 20170724 Added copy to clipboard functionality, the find next/prev source functionality. 
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

/**
 * Source code viewer. Renders a source file in a scrolling viewport and interacts with hover and
 * selection events generated by this viewer or by external objects.
 */
define(["dojo/on",
        "dojo/_base/window",
        "dijit/Dialog",
        "dojox/uuid/generateRandomUuid",
        "dojo/domReady!"],
function(on, win, Dialog, generateRandomUuid)
{
   "use strict";
   
   /** The top-level object which will be returned from this function. */
   var me = {};
   
   /** Types of source code selection/highlight */
   var selectType =
   {
      HOVER      : 1,   // mouse hovers over node/text
      MATCH      : 2,   // report-based condition match
      STICKY     : 3,   // survives until next sticky selection
      FIND       : 4    // the next/previous matched result through the source 
   }
   
   /** Array of AST nodes, sorted by location of associated code in a source file */
   var sortedAstNodes = null;
   
   /** Map of AST IDs to AST node objects */
   var astMap = null;
   
   /** Array of lines of source code */
   var sourceLines = null;
   
   /** Number of columns to reserve for the line number gutter */
   var sourceColOffset = 0;
   
   /** Approximate width of an average character in monospace font */
   var sourceCharWidth = 0;
   
   /** Height in pixels of a row in which a single line of code is rendered */
   var sourceRowHeight = 15;
   
   /** The line number of the source code over which the mouse most recently hovered */
   var lastSourceLine = 0;
   
   /** The column number of the source code over which the mouse most recently hovered */
   var lastSourceColumn = 0;
   
   /** The AST which represents the source code over which the mouse most recently hovered */
   var hoverAst = null;
   
   /** Scroller containing source code view */
   var sourceScroller = null;
   
   /** Copy to clipboard and save to file functionality*/
   var downloadHandler = new DownloadHandler();
   
   var searchNextSource = new SearchNextSource();
   
   /**
    * Responsible to download the target source code into the system clipboard or the local file
    * system.
    */
   function DownloadHandler()
   {
      /** Keydown listener's handler */
      var keyDown;
      
      /** Hidden input field */
      var hiddenInput;
      
      /** Copy to clipboard prompt */
      var downloadPrompt;
      
      /** Holds source code */
      var sourceCode;
      
      /** Generates unique id for hidden input field */
      var uuid = generateRandomUuid();
      
      /** Holds used keystrokes */
      var symKeys = {
                      SHIFT : 16,
                      CTRL  : 17,
                      ALT   : 18,
                      C     : 67,
                      S     : 83
                    };
      
      /** Hidden text area field's HTML template */
      var template = "<textarea id='{0}' class='copy-paste-support'></textarea>";
      
      /**
       * Attach click handler to the "Copy to Clipboard" tool bar button.
       */
      function attachCopyToClipboardButton()
      {
         var copySpan = $(".toc-heading span.glyphicon.glyphicon-duplicate");
         if (copySpan.length > 0)
         {
            copySpan.off("click");
            
            copySpan.on("click",
                        function()
                        {
                           displayCopyToClipboardPrompt();
                           focusHiddenInputField();
                        });
            
            var hiddenTextArea = $("#"+ uuid);
            
            if (hiddenTextArea.length == 0)
            {
               var srcView = $("#srcView");
               hiddenTextArea = $(template.replace("{0}", uuid));
               srcView.append(hiddenTextArea);
            }
            
            hiddenTextArea.off("input");
            
            hiddenTextArea.on("input", focusHiddenInputField);
            
            if (keyDown)
            {
               keyDown.remove();
            }
            keyDown = on(document, "keydown", onKeyDown);
            
            document.removeEventListener("copy", copyHandler);
            
            document.addEventListener("copy", copyHandler);
         }
      }
      
      /** Removes keydown and copy/paste listeners */
      function cleanup()
      {
         if (keyDown)
         {
            keyDown.remove();
         }
         
         keyDown = null;
         
         document.removeEventListener("copy", copyHandler);
         
         var hiddenTextArea = $('#' + uuid);
         
         hiddenTextArea.off("input");
         
         hiddenTextArea.remove();
         
         var copySpan = $(".toc-heading > span.glyphicon.glyphicon-duplicate");
         
         copySpan.off("click");
      }
      
      /** Removes keydown and copy/paste listeners */
      this.cleanup = cleanup;
      
      /**
       * Set the source code from an array of line objects {line: , text:}
       * 
       * @param   {Array} lines
       *          An array of source code lines
       */
      function setSourceCode(lines)
      {
         if (!lines)
         {
            return;
         }
         
         if (!Array.isArray(lines))
         {
            console.error("Expected an array but got " + (typeof lines));
            return;
         }
         
         sourceCode = lines.map(function (line) { return line.text; }).join(
               fwd.getPlatformSpecificEOL());
         
         attachCopyToClipboardButton();
      }
      
      this.setSourceCode = setSourceCode;
      
      /**
       * Display copy to clipboard prompt.
       */
      function displayCopyToClipboardPrompt()
      {
         //check if srcview is visible on the screen.
         if (!$("#srcView").is(":visible"))
         {
            return;
         }
         
         if (!downloadPrompt)
         {
            downloadPrompt = new Dialog(
            {
               title: "Copy to Clipboard",
               style: "width: 300px"
            });
            downloadPrompt.set("content", "<p>" + fwd.hints("copy") + "</p>");
         }
         
         downloadPrompt.show();
         
         downloadPrompt.focus();
      }
      
      function displaySaveToFilePrompt()
      {
         // TODO Add FileSaver.min.js zipped package?
      }
      
      /**
       * Keydown event handler.
       * 
       * @param   {KeyboardEvent} event
       *          Keydown event.
       */
      function onKeyDown(event)
      {
         event = dojo.fixEvent(event, document);
         var charOrCode = event.charCode || event.keyCode;
         // ALT
         if (charOrCode == symKeys.ALT)
         {
            event.preventDefault();
            event.stopPropagation();
         }
         else if (charOrCode == symKeys.C && !event.shiftKey)
         {
            // ALT+C
            if (event.altKey && !event.metaKey)
            {
               event.preventDefault();
               displayCopyToClipboardPrompt();
               focusHiddenInputField();
            }// CTRL+C or COMMAND+C
            else if ((event.ctrlKey && !event.altKey && !fwd.isSafari) ||
                     (event.metaKey && fwd.isSafari))
            {
               focusHiddenInputField();
            }
         }
         else if (charOrCode == symKeys.S && event.ctrlKey && event.shiftKey)
         {
            event.preventDefault();
            displaySaveToFilePrompt();
         }
      }
      
      /**
       * Copy handler.
       * 
       * @param   {ClipboardEvent} event
       *          Copy event.
       */
      function copyHandler(event)
      {
         if (!downloadPrompt || !downloadPrompt.open)
         {
            return;
         }
         // copy to clipboard
         if (window.clipboardData) 
         {
            window.clipboardData.clearData();
            window.clipboardData.setData("Text", sourceCode);
         }
         else if (event.clipboardData) 
         {
            event.clipboardData.clearData();
            event.clipboardData.setData("text/plain", sourceCode);
         }
         
         setTimeout(focusHiddenInputField, 0);
         setTimeout(function(){ downloadPrompt.hide(); }, 0);
         event.preventDefault();
      }
      
      /**
       * Focus hidden input field.
       */
      function focusHiddenInputField()
      {
         var hiddenTextArea = $('#' + uuid);
         if (hiddenTextArea.length > 0)
         {
            hiddenTextArea[0].value = ' ';
            hiddenTextArea[0].focus();
            hiddenTextArea[0].select();
         }
      }
   }
   
   /**
    * UI view-controller delegate that encapsulates the "find next source code" functionality.
    */
   function SearchNextSource()
   {
      /** Html span template */
      var span = "<span></span>";
      
      /** Html div template */
      var div = "<div></div>";
      
      /** Html input field template */
      var input = "<input type='text'></input>";
      
      /** Holds UI state: focused widgets states and business logic state */
      var state = new Map();
      
      /** An array of source code lines */
      var sourceByLines;
      
      /** Holds the original input search string*/
      var filter;
      
      /** Holds the case insensitive search string with all chars in their upper cases.*/
      var caseInsensitiveFilter;
      
      /** A button that provides an access to the "find next source code" functionality.*/
      var searchButton;
      
      /** The input search field */
      var searchField;
      
      /**  A button to perform the "find previous source code" search.*/
      var prevSource;
      
      /**  A button to perform the "find next source code" search.*/
      var nextSource;
      
      /** An html span to display the "x of y" notification message */
      var displayMatches;
      
      /**
       * An array of indices of the matched results. Each index is an object with "line" and
       * "column" fields that hold respectively the line and column numbers of its match in
       * the source code. 
       */ 
      var matchedResults = [];
      
      /** Holds a current position within the matched results. */
      var currentMatch = -1;
      
      resetSearchFilterState();
      
      /**
       * Resets the current filter search to its initial state.
       */
      function resetSearchFilterState()
      {
         state.set("line", 0);
         state.set("column", 0);
         matchedResults.length = 0;
         currentMatch = -1;
      }
      
      /**
       * Sets filter result and reset the search if the search filter is changed.
       */
      function setFilterValue()
      {
         if (searchField)
         {
            filter = searchField.val();
            
            var oldFilter = caseInsensitiveFilter;
            
            caseInsensitiveFilter = filter.toUpperCase();
            
            if (caseInsensitiveFilter != oldFilter)
            {
               resetSearchFilterState();
            }
         }
      }
      
      /**
       * Gets the current case insensitive filter string.
       */
      function getFilterValue()
      {
         return caseInsensitiveFilter;
      }
      
      this.getFilterValue = getFilterValue;
      
      /**
       * Create jQuery node from the given html template.
       * 
       * @param   {String} control
       *          An html string
       * @param   {String} clazz
       *          A css classes asigned to new created node
       * @param   {Object} styles
       *          A plain js object holds additional css styles to override the existing classes.
       * 
       * @return  {JQuery} a new node with provided css styles.
       * 
       */
      function createControl(control, clazz, styles)
      {
         return $(control).addClass(clazz).css(styles);
      }
      
      /**
       * Creates a search panel and attaches SearchNextSource listeners to the search button and
       * the seach panel.
       */
      function attachSearchNextSource()
      {
         searchButton = $(".toc-heading span.glyphicon.glyphicon-search");
         
         if (searchButton.length > 0)
         {
            searchButton.off("click");
            
            searchButton.on("click",
                        function()
                        {
                           searchButton.toggleClass("hidden");
                           
                           var searchCSS = {
                                 "float"  : "left",
                                 "cursor" : "text",
                                 "margin" : "2px 1px 2px 4px",
                                 "border" : "1px white solid",
                                 "width"  : "80px",
                                 "height" : "18px",
                                 "font-size" : "10pt",
                                 "color"     : "black",
                                 "overflow"  : "hidden",
                                 "padding"   : "1px",
                                 "text-align" : "left",
                                 "z-index"   : "2"
                               };
                           if (searchField)
                           {
                              searchField.remove();
                           }
                           searchField = createControl(input, "search-field", searchCSS);
                           searchField.attr("placeholder", "Find...");
                           searchField.attr("tabindex", "1");
                           searchField.insertAfter(searchButton);
                           
                           var prevCSS = {
                                 "float"  : "left",
                                 "cursor" : "pointer",
                                 "margin" : "auto",
                                 "z-index" : "2"
                              };
                           if (prevSource)
                           {
                              prevSource.remove();
                           }
                           prevSource = createControl(
                                                   span,
                                                   "glyphicon glyphicon-triangle-top disabled",
                                                   prevCSS);
                           fwd.addToolTip(prevSource, "Find Previous");
                           prevSource.insertAfter(searchField);
                           setButtonEnabled(prevSource, false);
                           
                           var nextCSS = {
                                            "float"  : "left",
                                            "cursor" : "pointer",
                                            "margin" : "auto",
                                            "z-index"   : "2"
                                         };
                           
                           if (nextSource)
                           {
                              nextSource.remove();
                           }
                           
                           nextSource = createControl(
                                                    span,
                                                    "glyphicon glyphicon-triangle-bottom disabled",
                                                    nextCSS);
                           fwd.addToolTip(nextSource, "Find Next");
                           nextSource.insertAfter(prevSource);

                           setButtonEnabled(nextSource, false);
                           
                           if (displayMatches)
                           {
                              displayMatches.remove();
                           }

                           var matchesCSS = {
                                               "float"   : "left",
                                               "margin"  : "2px",
                                               "z-index" : "2",
                                               "text-align" : "left",
                                               "font-size" : "10pt",
                                               "font-weight" : "normal",
                                               "color" : "white"
                                            };
                           displayMatches = createControl(span, "matches", matchesCSS);
                           displayMatches.insertAfter(nextSource);
                           
                           // attach handlers
                           searchField.off("click");
                           searchField.on(
                                 "click",
                                 function()
                                 {
                                    searchField.val(filter);
                                    searchField.focus();
                                    setTimeout(lookupAllMatches, 0);
                                 });
                           
                           searchField.off("change");
                           searchField.on(
                                 "change",
                                 function()
                                 {
                                    setFilterValue();
                                 });
                           
                           var widgets = [searchField, prevSource, nextSource];
                           widgets.forEach(
                                 function(widget)
                                 {
                                    widget.off("blur");
                                    widget.off("focus");
                                    widget.off("keydown");
                                    widget.on("blur", getChangeStateHandler(widget, false));
                                    widget.on("focus", getChangeStateHandler(widget, true));
                                    widget.on("keydown", getKeyDownHandler(widget));
                                 });
                           prevSource.off("click");
                           prevSource.on("click", getPrevSourceClickHandler());
                           nextSource.off("click");
                           nextSource.on("click", getNextSourceClickHandler());
                           searchField.focus();
                        });
         }
      }
      
      /**
       * Hides the search panel if the input focus is moved out of its html controls.
       */
      function hideSearchNextSourcePanel()
      {
         var focused = state.get(searchField);
         focused = focused || state.get(prevSource);
         focused = focused || state.get(nextSource);
         
         if (!focused)
         {
            searchField.toggleClass("hidden");
            prevSource.toggleClass("hidden");
            nextSource.toggleClass("hidden");
            displayMatches.toggleClass("hidden");
            
            searchButton.toggleClass("hidden");
            
            resetSearchFilterState();
         }
      }
      
      /**
       * Creates "change state" event listener to track focus states of a target html widget.
       * 
       * @param   {jQuery} widget
       *          A target widget with a tracked focus state 
       * @param   {Boolean} value
       *          A new widget state that is changed to this value on recieving the target event
       * 
       * @return  {Function} a new handler that implements EventListener interface for "focus" and
       *          "blur" events.
       */
      function getChangeStateHandler(widget, value)
      {
         return function(evt)
         {
            changeState(widget, value);
            
            if (evt.type == "blur")
            {
               setTimeout(hideSearchNextSourcePanel, 0);
            }
         };
      }
      
      /**
       * Creates a "keydown" event listener to track "keydown" events of a target html widget.
       *  
       * @param   {jQuery} widget
       *          A target widget
       * 
       * @return  {Function} a new handler that implements EventListener interface
       *          for "keydown" events of the target widget
       */
      function getKeyDownHandler(widget)
      {
         return function(evt)
         {
            if (evt.keyCode == 13 && evt.target == widget[0])
            {
               if ((evt.target == searchField[0]) || (evt.target == nextSource[0]))
               {
                  doNextSearch();
               }
               else if (evt.target == prevSource[0])
               {
                  doPrevSearch();
               }
            }
            else if (evt.target == searchField[0])
            {
               // disable next and previous source buttons
               setButtonEnabled(prevSource, false);
               setButtonEnabled(nextSource, false);
               // reset filter state
               resetSearchFilterState();
               setTimeout(lookupAllMatches, 0);
            }
         };
      }
      
      /**
       * Performs all matches search and builds indices of matched results.
       */
      function lookupAllMatches()
      {
         setFilterValue();
         
         if (!getFilterValue())
         {
            displayMatches.text("");
            return;
         }
         
         resetSearchFilterState();
         
         var results = nextSearch(Number.MAX_VALUE, 0, 0, matchedResults);
         
         var found = (results.length > 0);
         
         if (found)
         {
            doNextSearch();
         }
         else
         {
            displayMatches.text("");
         }
         
         displayNotFoundBorder(!found);
      }
      
      /**
       * Shows or hides not found warning border around the input field.
       * 
       * @param   {Boolean} notFound
       *          An input parameter to show the warning border or to hide it.
       */
      function displayNotFoundBorder(notFound)
      {
         if (notFound)
         {
            searchField.css({"border" : "2px solid red"});
         }
         else
         {
            searchField.css({"border" : "none"});
         }
      }
      
      /**
       * Creates the "click" event handler for the next source html button.
       * 
       * @return   {Function} a target handler that implements EventListener interface.
       */
      function getNextSourceClickHandler()
      {
         return function(evt)
         {
            if (nextSource.hasClass("disabled"))
            {
               evt.preventDefault();
               evt.stopPropagation;
               return;
            }
            
            doNextSearch();
         };
      }
      
      /**
       * Creates the "click" event handler for the previous source html button.
       * 
       * @return   {Function} a target handler that implements EventListener interface.
       */
      function getPrevSourceClickHandler()
      {
         return function(evt)
         {
            if (prevSource.hasClass("disabled"))
            {
               evt.preventDefault();
               evt.stopPropagation;
               return;
            }
            
            doPrevSearch();
         };
      }
      
      /**
       * Change a focus related state for a target widget.
       * 
       * @param   {jQuery} widget
       *          A target widget 
       * @param   {Boolean} value
       *          A new widget state
       */
      function changeState(widget, value)
      {
         state.set(widget, value);
      }
      
      /**
       * Enables or disables a target html button.
       * 
       * @param   {jQuery} button
       *          A target html button
       * @param   {Boolean} enabled
       *          A new button's state, true value indicates to enable button and false value
       *          to disable it.
       */
      function setButtonEnabled(button, enabled)
      {
         if (enabled)
         {
            button.attr("tabindex", "1");
            button.css({"cursor" : "pointer"});
         }
         else
         {
            button.removeAttr("tabindex");
            button.css({"cursor" : "not-allowed"});
         }
         
         button.toggleClass("disabled", !enabled);
      }
      
      /**
       * Set the source code from an array of line objects {line: , text:}
       * 
       * @param   {Array} lines
       *          An array of source code lines
       */
      function setSourceCode(lines)
      {
         if (!lines)
         {
            return;
         }
         
         if (!Array.isArray(lines))
         {
            console.error("Expected an array but got " + (typeof lines));
            return;
         }
         
         sourceByLines = lines.map(function (line) { return line.text; });
         
         attachSearchNextSource();
      }
      
      this.setSourceCode = setSourceCode;
      
      /**
       * Moves the current position to the next match.
       */
      function doNextSearch()
      {
         if (currentMatch == (matchedResults.length - 1))
         {
            // Move to the current position before the first occurrence of the search string
            
            currentMatch = -1;
         }
         
         currentMatch++;
         
         if (currentMatch < matchedResults.length)
         {
            var nextMatch = matchedResults[currentMatch];
            
            if (nextMatch.atColumn > 0)
            {
               scrollToCode(nextMatch.atLine, nextMatch.atColumn);
               highlightCode(nextMatch.atLine, nextMatch.atColumn, selectType.FIND);
               displayCurrentAndTotalMatches();
            }
         }
         
         var enableButtons = (matchedResults.length > 0);
         
         setButtonEnabled(nextSource, enableButtons);
         setButtonEnabled(prevSource, enableButtons);
         
         var displayNotFound = (matchedResults.length == 0);
         
         setTimeout(
               function()
               {
                  displayNotFoundBorder(displayNotFound);
                  
                  if (displayNotFound)
                  {
                     searchField.focus();
                  }
               }, 0);
      }
      
      /**
       * Perform the next search forward the source code from the given line and column.
       * 
       * @param   {Number} lookup
       *          The look-ahead number
       * @param   {Number} line
       *          The line to start search from.
       * @param   {Number} column
       *          The column to start search from.
       * @param   {Array} matches
       *          An array that holds the matched results
       * 
       * @return  {Array} An array with the search results. Returns a reference to the given
       *          "matches" array, otherwise new array with matched results is created.
       */
      function nextSearch(lookup, line, column, matches)
      {
         var index = -1;
         var i = line;
         var startColumn = column;
         
         /** matches */
         if (!matches)
         {
            matches = [];
         }
         
         /** lookup index */
         var j = 0;
         
         /** lookup limit */
         var l = (lookup && (lookup > 0)) ? lookup : 0;
         
         for(; i < sourceByLines.length && j <= l; i++)
         {
            var text = sourceByLines[i].toUpperCase();
            
            while ((index = text.indexOf(caseInsensitiveFilter, startColumn)) >= 0 )
            {
               matches.push({ "atColumn" : (index + 1), "atLine" : (i + 1) });
               // lookup index
               j++;
               // shift startColumn
               startColumn = index + caseInsensitiveFilter.length;
            }
            // new line
            startColumn = 0; 
         }
         
         return matches;
      }
      
      /**
       * Moves the current position to the previous match.
       */
      function doPrevSearch()
      {
         if (currentMatch <= 0)
         {
         // Move to the current position after the last occurrence of the search string
            currentMatch = matchedResults.length;
         }

         currentMatch--;
         
         if (currentMatch < matchedResults.length)
         {
            var prevMatch = matchedResults[currentMatch];
            
            if (prevMatch.atColumn > 0)
            {
               scrollToCode(prevMatch.atLine, prevMatch.atColumn);
               highlightCode(prevMatch.atLine, prevMatch.atColumn, selectType.FIND);
               displayCurrentAndTotalMatches();
            }
         }
         
         var enableButtons = (matchedResults.length > 0);
         
         setButtonEnabled(nextSource, enableButtons);
         setButtonEnabled(prevSource, enableButtons);
      }
      
      /**
       * Sets the current position an total matches.
       */
      function displayCurrentAndTotalMatches()
      {
         if (matchedResults.length > 0)
         {
            var pos = currentMatch + 1;
            displayMatches.text(pos + " of " + matchedResults.length);
         }
      }
   }
   /**
    * Show the source code view.
    * 
    * @param   {String} selector
    *          Selector for DOM element to which to attache source view.
    * @param   {Number} height
    *          Height in pixels allocated to view.
    * @param   {Object[]} lines
    *          Array of source line objects.
    * @param   {Object[]} nodes
    *          Array of AST node objects.
    * @param   {Number} atLine
    *          Line number in source file to which we should scroll upon showing the view.
    * @param   {Number} atColumn
    *          Column number of character to which we should scroll upon showing the view (not
    *          implemented currently).
    */
   me.show = function(selector, height, lines, nodes, atLine, atColumn)
   {
      sourceLines = lines;
      
      if (nodes)
      {
         astMap = fwd.ast.mapNodes(nodes);
         sortedAstNodes = nodes.sort(compareAstSpans);
      }
      
      // generate appropriate numeric format for line numbers
      var len = (sourceLines.length + 1).toString().length;
      sourceColOffset = len + 1;
      var fmt = "";
      for (var i = 0; i < len; i++)
      {
         fmt += "0";
      }
      
      var maxLineLen = 0;
      sourceLines.forEach(function(line, i)
      {
         maxLineLen = Math.max(maxLineLen, line.text.length);
      });
      
      var srcView = $(selector);
      var div = "<div></div>";
      var viewport = $(div).addClass("viewport")
                           .css(
                           {
                              "height" : srcView.height() + "px",
                              "width"  : "100%",
                           });
      srcView.append(viewport);
      
      var svg = d3.select(".viewport").append("svg")
                  .attr("class", "scroll-svg");
      
      if (nodes)
      {
         var defs = svg.append("defs");
         
         // set up gradients for selection paths
         appendRadialGradient(defs, "srcHoverGradient", "#aaddee", "#3090d0");
         appendRadialGradient(defs, "srcMatchGradient", "#eeaadd", "#d030c0");
         appendRadialGradient(defs, "srcStickyGradient", "#aaeeaa", "#30d030");
         appendRadialGradient(defs, "srcFindGradient", "#ffa500", "#ff0000");
      }
      
      // calculate character width for current font
      var t = "";
      for (var i = 1; i <= 1000; i++)
      {
         t += "0";
      }
      var te = svg.append("text")
                  .text(t)
                  .attr("class", "src-text");
      var bb = te.node().getBBox();
      sourceCharWidth = bb.width / t.length;
      te.remove();
      
      var width = (maxLineLen + len + 2) * sourceCharWidth;
      svg.attr("width", width);
      
      var chartGroup = svg.append("g")
                          .attr("class", "chartGroup");
      chartGroup.append("rect");
      
      var rowEnter = function(rowSelection)
      {
         rowSelection.append("text")
                     .attr("class", "src-text")
                     .style("white-space", "pre")
                     .attr("transform", "translate(3,15)");
         
         // add mouse handlers if we are associated with an AST
         if (nodes)
         {
            rowSelection.on("mousemove", mouseMove)
                        .on("click", mouseClick)
                        .on("mouseout", fireAstExit);
         }
      };
      
      var rowUpdate = function(rowSelection)
      {
         rowSelection.select("text")
                     .text(function (d)
                          {
                             return fwd.rpt.numberFormat(d.line, fmt) + " " + d.text;
                          });
      };
      
      var rowExit = function(rowSelection)
      {
      };
      
      sourceScroller = d3.VirtualScroller()
                         .rowHeight(sourceRowHeight)
                         .enter(rowEnter)
                         .update(rowUpdate)
                         .exit(rowExit)
                         .svg(svg)
                         .totalRows(sourceLines.length + 2)
                         .viewport(d3.select(".viewport"));
      
      sourceScroller.data(sourceLines, function(d) { return d.line; });
      
      chartGroup.call(sourceScroller);
      
      if (nodes)
      {
         fwd.evt.addListener("astenter", onAstEnter);
         fwd.evt.addListener("astexit", onAstExit);
         fwd.evt.addListener("astselect", onAstSelect);
      }
      
      if (atLine && atColumn)
      {
         scrollToCode(atLine, atColumn);
         highlightCode(atLine, atColumn, selectType.MATCH);
      }
      
      // set source code
      downloadHandler.setSourceCode(lines);
      searchNextSource.setSourceCode(lines);
   }
   
   /**
    * Function which should be invoked when it is time to clean up this view's resources.
    */
   me.cleanup = function()
   {
      fwd.evt.removeListener("astenter", onAstEnter);
      fwd.evt.removeListener("astexit", onAstExit);
      fwd.evt.removeListener("astselect", onAstSelect);
   }
   
   /**
    * Fire an event when the mouse enters an area of the source code view which corresponds with
    * an AST node.
    * 
    * @param   {Object} ast
    *          AST node represented at the mouse location.
    * @param   {Selection} selection
    *          D3 selection associated with the mouse location.
    */
   function fireAstEnter(ast, selection)
   {
      if (ast)
      {
         if (selection)
         {
            selection.style("cursor", "pointer");
         }
         fwd.evt.fireEvent("astenter", null, { id : ast.id });
      }
      hoverAst = ast;
   }
   
   /**
    * Fire an event when the mouse exits an area of the source code view which corresponds with
    * an AST node.
    */
   function fireAstExit()
   {
      if (hoverAst)
      {
         d3.select(this).style("cursor", "auto");
         fwd.evt.fireEvent("astexit", null, { id : hoverAst.id });
         hoverAst = null;
      }
   }
   
   /**
    * Callback invoked in response to an "astenter" event. Highlight the area of source code
    * which corresponds with the AST which was entered.
    * 
    * @param   {Object} e
    *          The event, which contains the ID of the AST node entered.
    */
   function onAstEnter(e)
   {
      highlightAst(astMap.get(e.id), selectType.HOVER);
   }
   
   /**
    * Callback invoked in response to an "astexit" event. Remove any previously highlighted area
    * of source code which corresponds with the AST which was exited.
    */
   function onAstExit()
   {
      d3.selectAll(".hoverpath").remove();
   }
   
   /**
    * Callback invoked in response to an "astselect" event. Highlight the area of source code
    * which corresponds with the AST which was selected and scroll to its location in the source
    * view.
    * 
    * @param   {Object} e
    *          The select event, which contains the ID of the AST node entered..
    */
   function onAstSelect(e)
   {
      var ast = astMap.get(e.id);
      var line = ast.l0 > 0 ? ast.l0 : ast.el0;
      var col = ast.c0 > 0 ? ast.c0 : ast.ec0;
      scrollToCode(line, col);
      highlightAst(astMap.get(e.id), selectType.STICKY);
   }
   
   /**
    * Compare two spans of source code, according to their relative positions in the source file.
    * 
    * @param   {Object} a
    *          AST node.
    * @param   {Object} b
    *          AST node.
    * 
    * @return  -1 if span a ends before span b begins, or if span b contains span a entirely;
    *          1 if span b ends before span a begins, or if span a contains span b entirely;
    *          0 for artificial nodes, which all have line and column numbers of 0.
    */
   function compareAstSpans(a, b)
   {
      if (comparePositions(a.c1, a.l1, b.c0, b.l0) < 0)
      {
         // end of a is before start of b
         return -1;
      }
      
      if (comparePositions(a.c0, a.l0, b.c1, b.l1) > 0)
      {
         // end of b is before start of a
         return 1;
      }
      
      var d0 = comparePositions(a.c0, a.l0, b.c0, b.l0);
      var d1 = comparePositions(a.c1, a.l1, b.c1, b.l1);
      
      if ((d0 < 0 && d1 >= 0) || (d0 <= 0 && d1 > 0))
      {
         // a contains b; we want b to sort first in this case
         return 1;
      }
      
      if ((d1 < 0 && d0 >= 0) || (d1 <= 0 && d0 > 0))
      {
         // b contains a; we want a to sort first in this case
         return -1;
      }
      
      // should only happen for artificial nodes
      return 0;
   }
   
   /**
    * Indidate whether the given AST contains the location represented by the given line and
    * column.
    * 
    * @param   {Object} ast
    *          AST node
    * @param   {Number} line
    *          Line number.
    * @param   {Number} column
    *          Column number.
    */
   function astContainsCell(ast, line, column)
   {
      var d0 = comparePositions(ast.c0, ast.l0, column, line);
      var d1 = comparePositions(ast.c1, ast.l1, column, line);
      
      return d0 <= 0 && d1 >= 0;
   }
   
   /**
    * Identify the AST, if any, associated with the given line and column location in the source
    * file.
    * 
    * @param   {Number} line
    *          Line number.
    * @param   {Number} column
    *          Column number.
    */
   function findAstByCell(line, column)
   {
      if (!sortedAstNodes || sortedAstNodes.length == 0)
      {
         return null;
      }
      
      var lower = 0;
      var upper = sortedAstNodes.length - 1;
      
      return findAstInRange(line, column, upper, lower, null);
   }
   
   /**
    * Find the most specific AST node, if any, contained within the specified range, using a
    * recursive, binary search.
    * 
    * @param   {Number} line
    *          Line number.
    * @param   {Number} column
    *          Column number.
    * @param   {Number} upper
    *          Upper bound of the index range being searched within an array of AST nodes sorted
    *          by line and column position.
    * @param   {Number} lower
    *          Lower bound of the index range being searched within an array of AST nodes sorted
    *          by line and column position.
    * @param   {Object} candidate
    *          The most specific candidate AST node found so far in any previous iterations of
    *          the recursive search (may be null).
    * 
    * @return  The most specific AST node corresponding with the given location, or null if none
    *          was found.
    */
   function findAstInRange(line, column, upper, lower, candidate)
   {
      var i = Math.floor((upper + lower) / 2);
      var next = sortedAstNodes[i];
      var d0 = comparePositions(column, line, next.c0, next.l0);
      var contains = astContainsCell(next, line, column);
      
      if (contains || d0 < 0)
      {
         // keep looking in lower range
         upper = i - 1;
         return findAstInRange(line, column, upper, lower, contains ? next : candidate);
      }
      else if (lower > upper)
      {
         return contains ? next : candidate;
      }
      else
      {
         // keep looking in upper range
         lower = i + 1;
         return findAstInRange(line, column, upper, lower, candidate);
      }
   }
   
   /**
    * Compare two, 2-D coordinates by position, assuming a coordinate system whose origin of 0,0
    * is at the upper left of the coordinate space. The coordinate which is lower, or if at the
    * same vertical location, further to the right, is considered greater than the other
    * coordinate. Coordinates at the same location are considered equivalent.
    * 
    * @param   {Number} x0
    *          Horizontal location of first coordinate.
    * @param   {Number} y0
    *          Vertical location of first coordinate.
    * @param   {Number} x1
    *          Horizontal location of second coordinate.
    * @param   {Number} y1
    *          Vertical location of second coordinate.
    * 
    * @return  -1 if x0,y0 is less than x1,y1;
    *          1 if x0,y0 is greater than x1,y1;
    *          0 if they are equivalent.
    */
   function comparePositions(x0, y0, x1, y1)
   {
      if (y0 < y1)
      {
         return -1;
      }
      if (y0 > y1)
      {
         return 1;
      }
      else if (x0 < x1)
      {
         return -1;
      }
      else if (x0 > x1)
      {
         return 1;
      }
      else
      {
         return 0;
      }
   }
   
   /**
    * Define a radial gradient and append it to the "defs" element of the SVG.
    * 
    * @param   {Selection} defs
    *          D3 selection representing the "defs" SVG element.
    * @param   {String} id
    *          Unique identifying selector of the gradient.
    * @param   {String} color1
    *          Color string for the starting color.
    * @param   {String} color2
    *          Color string for the ending color.
    */
   function appendRadialGradient(defs, id, color1, color2)
   {
      var grad = defs.append("radialGradient")
                     .attr("id", id)
                     .attr("fx", "40%")
                     .attr("fy", "40%")
                     .attr("r", "65%")
                     .attr("spreadMethod", "pad");
      grad.append("stop")
          .attr("offset", "0%")
          .attr("stop-color", color1)
          .attr("stop-opacity", 0.3);
      grad.append("stop")
          .attr("offset", "100%")
          .attr("stop-color", color2)
          .attr("stop-opacity", 0.3);
   }
   
   /**
    * Mouse move callback function. May cause a source code highlight change.
    * 
    * @param   {Object} d
    *          Data bound to D3 selection associated with the mouse move.
    */
   function mouseMove(d)
   {
      mouseAction.call(this, d, selectType.HOVER);
   }
   
   /**
    * Mouse click callback function. May cause a source code selection change.
    * 
    * @param   {Object} d
    *          Data bound to D3 selection associated with the mouse click.
    */
   function mouseClick(d)
   {
      mouseAction.call(this, d, selectType.STICKY);
   }
   
   /**
    * Respond to a mouse move or click.
    * 
    * @param   {Object} d
    *          Data bound to D3 selection associated with the mouse event.
    * @param   {Number} type
    *          Select/highlight type.
    */
   function mouseAction(d, type)
   {
      var x = d3.mouse(this)[0];
      var line = d.line;
      var col = Math.min(Math.floor((x / sourceCharWidth) - sourceColOffset + 1), d.text.length);
      if (col > 0)
      {
         highlightCode(line, col, type, d3.select(this));
      }
   }
   
   /**
    * Scroll the source code viewport to the specified line (column currently ignored).
    * 
    * @param   {Number} line
    *          Line number.
    * @param   {Number} column
    *          Column number.
    */
   function scrollToCode(line, column)
   {
      if (sourceScroller == undefined)
      {
         return;
      }
      
      var vpos = sourceScroller.rowHeight() * line - $("#srcView").height() / 5;
      sourceScroller.position(vpos);
      sourceScroller.render(false);
   }
   
   /**
    * Highlight a block of source code located at the given line and column number. The expanse
    * and color of the highlight depend upon the AST node associated with the code and the type
    * of selection to apply.
    * 
    * @param   {Number} line
    *          Line number.
    * @param   {Number} column
    *          Column number.
    * @param   {Number} type
    *          Selection type.
    * @param   {Selection} selection
    *          D3 selection associated with the mouse location, if the request originated
    *          in the source view itself.
    */
   function highlightCode(line, column, type, selection)
   {
      var hover = type == selectType.HOVER;
      
      if (hover && line == lastSourceLine && column == lastSourceColumn)
      {
         return;
      }
      
      if (type == selectType.FIND)
      {
         highlightMatchedCode(line, column);
         return;
      }
      
      var ast = findAstByCell(line, column);
      if (!ast)
      {
         lastSourceLine = null;
         lastSourceColumn = null;
         
         return;
      }
      
      if (hover && ast === hoverAst)
      {
         return null;
      }
      
      lastSourceLine = line;
      lastSourceColumn = column;
      
      switch (type)
      {
         case selectType.HOVER:
            fireAstExit(hoverAst);
            fireAstEnter(ast, selection);
            break;
         case selectType.STICKY:
         case selectType.MATCH:
            highlightAst(ast, type);
            fwd.astview.show(ast);
            break;
      }
   }
   
   /**
    * Highlight a matched source code.
    * 
    * @param   {Number} line
    *          Line number.
    * @param   {Number} column
    *          Column number.
    */
   function highlightMatchedCode(line, column)
   {
      var matchedString = searchNextSource.getFilterValue();
      
      var path = createMatchedSourcePath(line, column, matchedString);
      
      var data = [];
      
      data.push({ "d" : path});
      
      var highlight = d3.select(".scroll-svg").selectAll('path[id="FIND"]').data(data);
      
      highlight.exit().remove();
      
      highlight.enter().append("path")
                   .style("opacity", 1)
                   .attr("id", "FIND")
                   .attr("class", "code-hilite")
                   .attr("fill", "url('#srcFindGradient')")
                   .attr("transform", "translate(3,3)")
                   .merge(highlight)
                   .attr("d", function(row) { return row.d;});
   }
   
   /**
    * Highlight a block of source code represented by the give AST node associated with the code.
    * 
    * @param   {Object} ast
    *          AST node associated with the code to be highlighted.
    * @param   {Number} type
    *          Selection type.
    * @param   {Selection} selection
    *          D3 selection associated with the mouse location, if the request originated
    *          in the source view itself.
    */
   function highlightAst(ast, type)
   {
      var classes = "code-hilite";
      var gradient = null;
      var stroke = null;
      
      switch (type)
      {
         case selectType.HOVER:
            classes += " transient-hilite hoverpath";
            gradient = "#srcHoverGradient";
            stroke = "#3030cc";
            break;
            
         case selectType.STICKY:
            var delay = 400;
            d3.selectAll(".transient-hilite")
              .transition()
              .duration(delay)
              .style("opacity", 1e-6)
              .transition()
              .duration(delay)
              .remove();
            classes += " transient-hilite stickypath";
            gradient = "#srcStickyGradient";
            stroke = "#00aa00";
            break;
            
         case selectType.MATCH:
            gradient = "#srcMatchGradient";
            stroke = "#cc3030";
            break;
         case selectType.FIND:
            gradient = "#srcFindGradient";
            stroke = "#cc3030";
            break;
      }
      
      
      var effPath = createAstSourcePath(ast, true);
      var pEff = d3.select(".scroll-svg")
                   .append("path")
                   .style("opacity", type == selectType.STICKY ? 1e-6 : 1)
                   .attr("class", classes)
                   .attr("d", effPath)
                   .attr("fill", "url('" + gradient + "')")
                   .attr("transform", "translate(3,3)");
      
      // synthetic nodes will have nothing to highlight for the actual path
      var synthetic = ast.l0 == 0;
      var pAct;
      if (!synthetic)
      {
         var actPath = createAstSourcePath(ast, false);
         pAct = d3.select(".scroll-svg")
                  .append("path")
                  .attr("class", classes)
                  .attr("d", actPath)
                  .attr("fill", "rgba(80, 80, 80, 0")
                  .attr("stroke", stroke)
                  .attr("transform", "translate(3,3)");
      }
      
      if (type == selectType.STICKY)
      {
         var delay = 400;
         
         pEff.transition()
             .duration(delay)
             .style("opacity", 1);
         
         if (!synthetic)
         {
            pAct.transition()
                .duration(delay)
                .style("opacity", 1);
         }
      }
   }
   
   /**
    * Define an SVG path associated with an AST node. The path may span multiple lines of text.
    * The path is determined by the starting and ending line and column numbers stored in the AST
    * node, and may or may not include the spans of descendant nodes.
    * 
    * @param   {Object} ast
    *          AST node.
    * @param   {boolean} effective
    *          True to include in the path the spans of all descendant AST nodes; false to use
    *          only the span of the given node.
    */
   function createAstSourcePath(ast, effective)
   {
      /*
       *         7----6
       *     1---0nnnn|
       *     |nnnnnnnn|
       *     |nnnnnnnn|
       *     |nnnnnnnn|
       *     |nnnn4---5
       *     2----3
       */
      // TODO: deal with multiple lines of varying length
      var l0 = effective ? ast.el0 : ast.l0;
      var c0 = effective ? ast.ec0 : ast.c0;
      var l1 = effective ? ast.el1 : ast.l1;
      var c1 = effective ? ast.ec1 : ast.c1;
      var dl = l1 - l0;
      
      var maxLen = 0;
      for (var i = l0 - 1; i < l1; i++)
      {
         maxLen = Math.max(sourceLines[i].text.length, maxLen);
      }
      
      var x0 = (c0 - 1 + sourceColOffset) * sourceCharWidth;
      var y0 = l0 * sourceRowHeight;
      var x1 = dl > 0 ? sourceColOffset * sourceCharWidth : x0;
      var y2 = dl > 0 ? l1 * sourceRowHeight : y0;
      var x3 = (c1 + sourceColOffset) * sourceCharWidth;
      var y4 = (l1 - 1) * sourceRowHeight;
      var x5 = dl > 0 ? (maxLen + sourceColOffset) * sourceCharWidth : x3;
      var y6 = (l0 - 1) * sourceRowHeight;
      var x7 = x0;
      
      var path = "M" + x0 + "," + y0
               + "H" + x1
               + "V" + y2
               + "H" + x3
               + "V" + y4
               + "H" + x5
               + "V" + y6
               + "H" + x7
               + "Z";
      
      return path;
   }
   
   /**
    * Define an SVG path associated with a matched source code.
    * 
    * @param   {Number} line
    *          Line number.
    * @param   {Number} column
    *          Column number.
    * @param   {String} matchedString
    *          A matched string.
    */
   function createMatchedSourcePath(line, column, matchedString)
   {
      /*
       *     1____2
       *     |mmmm|
       *     4____3
       */
      
      var x1 = (column - 1 + sourceColOffset) * sourceCharWidth;
      var y1 = (line - 1) * sourceRowHeight;
      var y2 = y1;
      var x2 = x1 + matchedString.length * sourceCharWidth;
      var x3 = x2;
      var y3 = y2 + sourceRowHeight;
      var x4 = x1;
      var y4 = y3;
      
      var path = "M" + x1 + "," + y1
               + "H" + x2
               + "V" + y3
               + "H" + x4
               + "Z";
      
      return path;
   }

   /* exports */
   return me;
   
});