/*
** Module   : JobTask.java
** Abstract : Implements tasks to be passed to the java timer.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140206 First version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.scheduler;

import java.util.*;
import java.util.logging.*;

import com.goldencode.p2j.util.*;

/**
 * Implements tasks to be passed to the java {@link Scheduler#timer}.
 * <p>
 * When a job is executed, it will delegate the job execution to the 
 * {@link Scheduler#processor job processor}, which will execute the job using the server's
 * context. 
 */
class JobTask
extends TimerTask
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(JobTask.class);

   /** The job name. */
   private final String jobName;
   
   /**
    * Create a new job task to be registered with the {@link Scheduler#timer}.
    */
   public JobTask(String jobName)
   {
      this.jobName = jobName.toLowerCase();
   }

   /**
    * Execute a the job with the given {@link #jobName name}.
    */
   @Override
   public void run()
   {
      // we are executing, remove it from the map
      synchronized (Scheduler.jobRegistry)
      {
         Scheduler.jobRegistry.remove(jobName);
      }

      // get the configured job, reschedule it (if needed) and execute it
      synchronized (Scheduler.jobs)
      {
         Job job = Scheduler.getJob(jobName);
         
         if (!job.isEnabled())
         {
            // the job is not enabled.
            if (LOG.isLoggable(Level.WARNING))
            {
               String msg = String.format("Can not reschedule: job %s is disabled.", jobName);
               LOG.logp(Level.WARNING, "Scheduler$JobTask", "run", LogHelper.generate(msg));
               return;
            }
         }

         // schedule it again (depending on the mode)
         Scheduler.scheduleJob(jobName, this.scheduledExecutionTime());
         
         if (LOG.isLoggable(Level.FINE))
         {
            String msg = String.format("Executing job %s at %tc.",
                                       jobName, this.scheduledExecutionTime());
            LOG.logp(Level.FINE, "Scheduler$JobTask", "run", LogHelper.generate(msg));
         }
         
         // get the core executor
         Runnable core = Scheduler.getRunner(job);
         
         Scheduler.processor.executeJob(core);
      }
   }
}