/*
** Module   : PropertyMapper.java
** Abstract : Classes which map bean properties to configuration info needed
**            for data import
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description-----------------------------
** 001 ECF 20050819   @22203 Moved initial version from ImportWorker.
**                           These classes were previously inner classes
**                           of that top level class, but were extracted
**                           to be shared between database and directory
**                           import tasks.
** 002 ECF 20051005   @22984 Added first class support for raw data type.
**                           Replace byte array processing from RawMapper
**                           and replaced with raw data type.
** 003 ECF 20060326   @25258 Modified to use new setter signatures for
**                           integer and decimal property mappers. These
**                           now accept NumberType instead of the more
**                           specific types.
** 004 ECF 20070905   @35339 Fixed type assertion. Added unknown value as
**                           a permissible type for all property mappers.
** 005 ECF 20080107   @36751 Added RowIDMapper inner class. Handles RECID
**                           and ROWID data (or the lack thereof). Added
**                           isExported() method to base class and added
**                           override in RowIDMapper.
** 006 GES 20080418   @38173 Moved from a lexer approach to using the
**                           equivalent runtime support for the Progress
**                           IMPORT language stmt.
** 007 ECF 20080625   @38930 Fixed RowIDMapper.createPlaceholder(). This
**                           method must return a data wrapper instance
**                           which represents unknown value, rather than
**                           throwing UnsupportedOperationException.
** 008 CA  20101125          Added getter support.
** 009 SVL 20120331          Use Text as the type which is expected by the
**                           setter method associated with the mapped
**                           character property.
** 010 OM  20130419          Added support for int64, longchar, datetime and datetime-tz.
**                           Introduced generics where needed.
** 011 OM  20130618          Use default constructors for unknown values of date_ values.
** 012 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 013 ECF 20160229          Fixed generic type in constructors. These were not picked up earlier
**                           because they only are called from TRPL.
** 014 ECF 20160311          Fixed regression introduced with #013.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.schema;

import java.lang.reflect.*;
import java.util.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.util.*;

/**
 * Maps a DMO class property to a data value read at a specific position
 * in a data export file. Also manages a list of foreign class names
 * with which the current table/class is associated by a foreign key
 * association. This list will be empty if no foreign associations exist.
 * <p>
 * One instance of this class is created for each Progress field (or
 * subscripted data value within an "extent" field) represented in the
 * export data file being imported.
 * <p>
 * This class is the abstract base class for data type specific, concrete
 * implementations. Each of these is a public inner class of this class.
 * Each is static, so it can be instantiated externally, directly, using the
 * semantic:
 * <pre>
 *    PropertyMapper mapper = new PropertyMapper$&lt;data_type&gt;Mapper(...);
 * </pre>
 */
public abstract class PropertyMapper
implements SchemaParserTokenTypes
{
   /** List of classes associated with this property by foreign key */
   private final List<String> joins = new ArrayList<>(1);
   
   /** Name of the property being mapped by this object */
   private String propertyName = null;
   
   /** Method which sets value into DMO property mapped by this object */
   private Method setter = null;
   
   /** Method which gets value for the DMO property mapped by this object */
   private Method getter = null;
   
   /** Array of arguments used when invoking <code>setter</code> */
   private Object[] setterArgs = null;
   
   /** Array of arguments used when invoking <code>getter</code> */
   private Object[] getterArgs = null;
   
   /** Is this property delegated to a foreign class? */
   private boolean delegated = false;
   
   /**
    * Constructor.
    *
    * @param   recordClass
    *          Class of DMO object which contains the mapped property.
    * @param   method
    *          Name of the method to be invoked via reflection to set a
    *          data value into the mapped property within a DMO instance.
    * @param   propertyName
    *          Name of the mapped property.
    * @param   index
    *          If the mapped property represents an individual subscripted
    *          value within a Progress extent field, this value represents
    *          the zero-based index of that value, relative to the start
    *          of the extent (array). It should be negative for a property
    *          which is not part of an extent field.
    *
    * @throws  NoSuchMethodException
    *          if <code>method</code> cannot be found within <code>recordClass</code>.
    */
   protected PropertyMapper(Class<?> recordClass,
                            String method,
                            String propertyName,
                            int    index)
   throws NoSuchMethodException
   {
      if (method != null)
      {
         if (index < 0)
         {
            getterArgs = new Object[0];
            setterArgs = new Object[1];
            Class<?>[] paramTypes = { getPropertyClass() };
            setter = recordClass.getMethod(method, paramTypes);
         }
         else
         {
            getterArgs = new Object[1];
            setterArgs = new Object[2];
            setterArgs[0] = new Integer(index);
            getterArgs[0] = new Integer(index);
            Class<?>[] paramTypes = { Integer.TYPE, getPropertyClass() };
            setter = recordClass.getMethod(method, paramTypes);
         }
         
         // get the DMO interface associated with the DMO implementation class
         Class<?> dmoIface = null;
         for (Class<?> next : recordClass.getInterfaces())
         {
            if (DataModelObject.class.isAssignableFrom(next))
            {
               dmoIface = next;
               
               break;
            }
         }
         
         assert (dmoIface != null);
         
         // find the associated method
         getter = PropertyHelper.pojoGettersByProperty(dmoIface).get(propertyName);
      }
      else
      {
         delegated = true;
      }
      
      this.propertyName = propertyName;
   }
   
   /**
    * Add the name of a foreign class, where the mapped property defines
    * all or part of the basis of a foreign key association with that
    * class.
    *
    * @param   joinClass
    *          Name of the foreign class to which the mapped property
    *          defines an association.
    */
   public final void addJoin(String joinClass)
   {
      joins.add(joinClass);
   }
   
   /**
    * Return the name of the mapped property. Following Java convention,
    * the property name begins with a lowercase letter and is camel cased.
    *
    * @return  Name of mapped property.
    */
   final String getPropertyName()
   {
      return propertyName;
   }
   
   /**
    * Get an iterator over the list of class names with which the mapped
    * property defines a foreign key association.
    *
    * @return  Iterator over joined class names.
    */
   final Iterator<String> joins()
   {
      return joins.iterator();
   }
   
   /**
    * Indicate whether the mapped property is delegated to a class with
    * which the property's containing class AST defines a foreign key
    * association.
    *
    * @return  <code>true</code> if mapped property is delegated to a
    *          "foreign" class, else <code>false</code>.
    */
   final boolean isDelegated()
   {
      return delegated;
   }
   
   /**
    * Invoke a method to set a value into the mapped property of the
    * given DMO instance. The method invoked will be a simple setter if
    * the mapped property does not represent an extent field. Otherwise,
    * the method will be an indexed setter (i.e., the first parameter is
    * is the index at which to set the data value within a list). In the
    * latter case, the index set at construction time is used as the
    * first argument to the setter method.
    *
    * @param   dmo
    *          Data model object in which to set the value.
    * @param   value
    *          Data value to be stored in the mapped property.
    *
    * @throws  IllegalAccessException
    *          if the DMO class or its default constructor is not
    *          accessible.
    * @throws  InvocationTargetException
    *          if there is any error invoking the setter method.
    */
   final void invokeSetter(Object dmo, Object value)
   throws IllegalAccessException,
          InvocationTargetException
   {
      setterArgs[setterArgs.length - 1] = value;
      setter.invoke(dmo, setterArgs);
   }
   
   /**
    * Invoke a method to get the value of the mapped property of the
    * given DMO instance. The method invoked will be a simple getter if
    * the mapped property does not represent an extent field. Otherwise,
    * the method will be an indexed getter (i.e., the first parameter is
    * is the index at which to get the data value within a list). In the
    * latter case, the index set at construction time is used as the
    * first argument to the getter method.
    *
    * @param   dmo
    *          Data model object in which to get the value.
    *
    * @throws  IllegalAccessException
    *          if the DMO class or its default constructor is not
    *          accessible.
    * @throws  InvocationTargetException
    *          if there is any error invoking the setter method.
    */
   final BaseDataType invokeGetter(Object dmo)
   throws IllegalAccessException,
          InvocationTargetException
   {
      return (BaseDataType) getter.invoke(dmo, getterArgs);
   }
   
   /**
    * Report whether the property is expected to be found in exported data.
    * Certain property types are skipped when Progress exports data, and as
    * such, these will not appear in an export file.  This method tells the
    * record loader that the lexer should not expect to find any datum for
    * this property in the current row being imported.
    * <p>
    * By default, this method returns <code>true</code>;  implementations
    * which expect no data for a particular field must override this method
    * to respond accordingly.
    * 
    * @return  <code>true</code>.
    */
   boolean isExported()
   {
      return true;
   }
   
   /**
    * Return the type which is expected by the setter method associated with the mapped property.
    *
    * @return  Class of the object which is to be set into the DMO for the mapped property.
    */
   abstract Class<? extends BaseDataType> getPropertyClass();
   
   /**
    * Instantiate an object which is the proper type to contain a value for
    * the mapped property, for a particular imported record. This storage
    * will be passed to the import processing and will be assigned the
    * value as read from the export file.
    *
    * @return  Placeholder storage of the correct type.
    */
   abstract BaseDataType createPlaceholder();
   
   /**
    * A concrete <code>PropertyMapper</code> implementation for properties
    * which are mapped to the {@link com.goldencode.p2j.util.integer integer}
    * data type.
    */
   public static class IntegerMapper
   extends PropertyMapper
   {
      /**
       * Constructor.
       *
       * @param   recordClass
       *          Class of DMO object which contains the mapped property.
       * @param   method
       *          Name of the method to be invoked via reflection to set a
       *          data value into the mapped property within a DMO instance.
       * @param   propertyName
       *          Name of the mapped property.
       * @param   index
       *          If the mapped property represents an individual subscripted
       *          value within a Progress extent field, this value represents
       *          the zero-based index of that value, relative to the start
       *          of the extent (array). It should be negative for a property
       *          which is not part of an extent field.
       *
       * @throws  NoSuchMethodException
       *          if <code>method</code> cannot be found within
       *          <code>recordClass</code>.
       */
      public IntegerMapper(Class<?> recordClass,
                           String method,
                           String propertyName,
                           int    index)
      throws NoSuchMethodException
      {
         super(recordClass, method, propertyName, index);
      }
      
      /**
       * Return the type which is expected by the setter method associated
       * with the mapped property; in this case, the <code>NumberType</code> class.
       *
       * @return  <code>NumberType</code> class.
       */
      Class<? extends BaseDataType> getPropertyClass()
      {
         return NumberType.class;
      }
      
      /**
       * Instantiate an object which is the proper type to contain a value for
       * the mapped property, for a particular imported record. This storage
       * will be passed to the import processing and will be assigned the
       * value as read from the export file.
       *
       * @return  Placeholder storage of the correct type.
       */
      BaseDataType createPlaceholder()
      {
         return new integer();
      }
   }

   /**
    * A concrete <code>PropertyMapper</code> implementation for properties which are mapped to 
    * the {@link com.goldencode.p2j.util.int64 int64} data type.
    */
   public static class Int64Mapper
   extends PropertyMapper
   {
      /**
       * Constructor.
       *
       * @param   recordClass
       *          Class of DMO object which contains the mapped property.
       * @param   method
       *          Name of the method to be invoked via reflection to set a data value into the
       *          mapped property within a DMO instance.
       * @param   propertyName
       *          Name of the mapped property.
       * @param   index
       *          If the mapped property represents an individual subscripted value within a
       *          Progress extent field, this value represents the zero-based index of that
       *          value, relative to the start of the extent (array). It should be negative for
       *          a property which is not part of an extent field.
       *
       * @throws  NoSuchMethodException
       *          if <code>method</code> cannot be found within <code>recordClass</code>.
       */
      public Int64Mapper(Class<?> recordClass,
                         String method, 
                         String propertyName, 
                         int    index)
      throws NoSuchMethodException
      {
         super(recordClass, method, propertyName, index);
      }
      
      /**
       * Return the type which is expected by the setter method associated with the mapped 
       * property; in this case, the <code>NumberType</code> class.
       *
       * @return  <code>NumberType</code> class.
       */
      Class<? extends BaseDataType> getPropertyClass()
      {
         return NumberType.class;
      }
      
      /**
       * Instantiate an object which is the proper type to contain a value for the mapped 
       * property, for a particular imported record. This storage will be passed to the import 
       * processing and will be assigned the value as read from the export file.
       *
       * @return  Placeholder storage of the correct type.
       */
      BaseDataType createPlaceholder()
      {
         return new int64();
      }
   }

   /**
    * A concrete <code>PropertyMapper</code> implementation for properties
    * which are mapped to the {@link com.goldencode.p2j.util.decimal decimal}
    * data type.
    */
   public static class DecimalMapper
   extends PropertyMapper
   {
      /**
       * Constructor.
       *
       * @param   recordClass
       *          Class of DMO object which contains the mapped property.
       * @param   method
       *          Name of the method to be invoked via reflection to set a
       *          data value into the mapped property within a DMO instance.
       * @param   propertyName
       *          Name of the mapped property.
       * @param   index
       *          If the mapped property represents an individual subscripted
       *          value within a Progress extent field, this value represents
       *          the zero-based index of that value, relative to the start
       *          of the extent (array). It should be negative for a property
       *          which is not part of an extent field.
       *
       * @throws  NoSuchMethodException
       *          if <code>method</code> cannot be found within
       *          <code>recordClass</code>.
       */
      public DecimalMapper(Class<?> recordClass,
                           String method, 
                           String propertyName, 
                           int    index)
      throws NoSuchMethodException
      {
         super(recordClass, method, propertyName, index);
      }
      
      /**
       * Return the type which is expected by the setter method associated
       * with the mapped property; in this case, the <code>NumberType</code> class.
       *
       * @return  <code>NumberType</code> class.
       */
      Class<? extends BaseDataType> getPropertyClass()
      {
         return NumberType.class;
      }
      
      /**
       * Instantiate an object which is the proper type to contain a value for
       * the mapped property, for a particular imported record. This storage
       * will be passed to the import processing and will be assigned the
       * value as read from the export file.
       *
       * @return  Placeholder storage of the correct type.
       */
      BaseDataType createPlaceholder()
      {
         return new decimal();
      }
   }
   
   /**
    * A concrete <code>PropertyMapper</code> implementation for properties
    * which are mapped to the {@link com.goldencode.p2j.util.character
    * character} data type.
    */
   public static class CharacterMapper
   extends PropertyMapper
   {
      /**
       * Constructor.
       *
       * @param   recordClass
       *          Class of DMO object which contains the mapped property.
       * @param   method
       *          Name of the method to be invoked via reflection to set a
       *          data value into the mapped property within a DMO instance.
       * @param   propertyName
       *          Name of the mapped property.
       * @param   index
       *          If the mapped property represents an individual subscripted
       *          value within a Progress extent field, this value represents
       *          the zero-based index of that value, relative to the start
       *          of the extent (array). It should be negative for a property
       *          which is not part of an extent field.
       *
       * @throws  NoSuchMethodException
       *          if <code>method</code> cannot be found within
       *          <code>recordClass</code>.
       */
      public CharacterMapper(Class<?> recordClass,
                             String method, 
                             String propertyName, 
                             int    index)
      throws NoSuchMethodException
      {
         super(recordClass, method, propertyName, index);
      }
      
      /**
       * Return the type which is expected by the setter method associated
       * with the mapped property; in this case, the <code>Text</code> class.
       *
       * @return  <code>Text</code> class.
       */
      Class<? extends BaseDataType> getPropertyClass()
      {
         return Text.class;
      }
      
      /**
       * Instantiate an object which is the proper type to contain a value for
       * the mapped property, for a particular imported record. This storage
       * will be passed to the import processing and will be assigned the
       * value as read from the export file.
       *
       * @return  Placeholder storage of the correct type.
       */
      BaseDataType createPlaceholder()
      {
         return new character();
      }
   }
   
   /**
    * A concrete <code>PropertyMapper</code> implementation for properties
    * which are mapped to the {@link com.goldencode.p2j.util.logical logical}
    * data type.
    */
   public static class LogicalMapper
   extends PropertyMapper
   {
      /**
       * Constructor.
       *
       * @param   recordClass
       *          Class of DMO object which contains the mapped property.
       * @param   method
       *          Name of the method to be invoked via reflection to set a
       *          data value into the mapped property within a DMO instance.
       * @param   propertyName
       *          Name of the mapped property.
       * @param   index
       *          If the mapped property represents an individual subscripted
       *          value within a Progress extent field, this value represents
       *          the zero-based index of that value, relative to the start
       *          of the extent (array). It should be negative for a property
       *          which is not part of an extent field.
       *
       * @throws  NoSuchMethodException
       *          if <code>method</code> cannot be found within
       *          <code>recordClass</code>.
       */
      public LogicalMapper(Class<?> recordClass,
                           String method, 
                           String propertyName, 
                           int    index)
      throws NoSuchMethodException
      {
         super(recordClass, method, propertyName, index);
      }
      
      /**
       * Return the type which is expected by the setter method associated
       * with the mapped property; in this case, the <code>logical</code>
       * class.
       *
       * @return  <code>logical</code> class.
       */
      Class<? extends BaseDataType> getPropertyClass()
      {
         return logical.class;
      }
      
      /**
       * Instantiate an object which is the proper type to contain a value for
       * the mapped property, for a particular imported record. This storage
       * will be passed to the import processing and will be assigned the
       * value as read from the export file.
       *
       * @return  Placeholder storage of the correct type.
       */
      BaseDataType createPlaceholder()
      {
         return new logical();
      }
   }
   
   /**
    * A concrete <code>PropertyMapper</code> implementation for properties
    * which are mapped to the {@link com.goldencode.p2j.util.date date} data
    * type.
    */
   public static class DateMapper
   extends PropertyMapper
   {
      /**
       * Constructor.
       *
       * @param   recordClass
       *          Class of DMO object which contains the mapped property.
       * @param   method
       *          Name of the method to be invoked via reflection to set a
       *          data value into the mapped property within a DMO instance.
       * @param   propertyName
       *          Name of the mapped property.
       * @param   index
       *          If the mapped property represents an individual subscripted
       *          value within a Progress extent field, this value represents
       *          the zero-based index of that value, relative to the start
       *          of the extent (array). It should be negative for a property
       *          which is not part of an extent field.
       *
       * @throws  NoSuchMethodException
       *          if <code>method</code> cannot be found within
       *          <code>recordClass</code>.
       */
      public DateMapper(Class<?> recordClass,
                        String method, 
                        String propertyName, 
                        int    index)
      throws NoSuchMethodException
      {
         super(recordClass, method, propertyName, index);
      }
      
      /**
       * Return the type which is expected by the setter method associated
       * with the mapped property; in this case, the <code>date</code> class.
       *
       * @return  <code>date</code> class.
       */
      Class<? extends BaseDataType> getPropertyClass()
      {
         return date.class;
      }
      
      /**
       * Instantiate an object which is the proper type to contain a value for
       * the mapped property, for a particular imported record. This storage
       * will be passed to the import processing and will be assigned the
       * value as read from the export file.
       *
       * @return  Placeholder storage of the correct type.
       */
      BaseDataType createPlaceholder()
      {
         return new date(); // unknown date
      }
   }
   
   /**
    * A concrete <code>PropertyMapper</code> implementation for properties which are mapped to
    * the {@link com.goldencode.p2j.util.datetime datetime} data type.
    */
   public static class DatetimeMapper
   extends PropertyMapper
   {
      /**
       * Constructor.
       *
       * @param   recordClass
       *          Class of DMO object which contains the mapped property.
       * @param   method
       *          Name of the method to be invoked via reflection to set a data value into the
       *          mapped property within a DMO instance.
       * @param   propertyName
       *          Name of the mapped property.
       * @param   index
       *          If the mapped property represents an individual subscripted value within a 
       *          Progress extent field, this value represents the zero-based index of that 
       *          value, relative to the start of the extent (array). It should be negative for 
       *          a property which is not part of an extent field.
       *
       * @throws  NoSuchMethodException
       *          if <code>method</code> cannot be found within <code>recordClass</code>.
       */
      public DatetimeMapper(Class<?> recordClass,
                            String method, 
                            String propertyName, 
                            int    index)
      throws NoSuchMethodException
      {
         super(recordClass, method, propertyName, index);
      }
      
      /**
       * Return the type which is expected by the setter method associated with the mapped
       * property; in this case, the <code>date</code> class.
       *
       * @return  <code>date</code> class.
       */
      Class<? extends BaseDataType> getPropertyClass()
      {
         return date.class;
      }
      
      /**
       * Instantiate an object which is the proper type to contain a value for the mapped
       * property, for a particular imported record. This storage will be passed to the import
       * processing and will be assigned the value as read from the export file.
       *
       * @return  Placeholder storage of the correct type.
       */
      BaseDataType createPlaceholder()
      {
         return new datetime(); // unkn own datetime
      }
   }
   
   /**
    * A concrete <code>PropertyMapper</code> implementation for properties which are mapped to 
    * the {@link com.goldencode.p2j.util.datetimetz datetimetz} data type.
    */
   public static class DatetimetzMapper
   extends PropertyMapper
   {
      /**
       * Constructor.
       *
       * @param   recordClass
       *          Class of DMO object which contains the mapped property.
       * @param   method
       *          Name of the method to be invoked via reflection to set a
       *          data value into the mapped property within a DMO instance.
       * @param   propertyName
       *          Name of the mapped property.
       * @param   index
       *          If the mapped property represents an individual subscripted
       *          value within a Progress extent field, this value represents
       *          the zero-based index of that value, relative to the start
       *          of the extent (array). It should be negative for a property
       *          which is not part of an extent field.
       *
       * @throws  NoSuchMethodException
       *          if <code>method</code> cannot be found within
       *          <code>recordClass</code>.
       */
      public DatetimetzMapper(Class<?> recordClass,
                              String method, 
                              String propertyName, 
                              int    index)
      throws NoSuchMethodException
      {
         super(recordClass, method, propertyName, index);
      }
      
      /**
       * Return the type which is expected by the setter method associated
       * with the mapped property; in this case, the <code>date</code> class.
       *
       * @return  <code>date</code> class.
       */
      Class<? extends BaseDataType> getPropertyClass()
      {
         return date.class;
      }
      
      /**
       * Instantiate an object which is the proper type to contain a value for
       * the mapped property, for a particular imported record. This storage
       * will be passed to the import processing and will be assigned the
       * value as read from the export file.
       *
       * @return  Placeholder storage of the correct type.
       */
      BaseDataType createPlaceholder()
      {
         return new datetimetz(); // unknown datetime-tz
      }
   }
   
   /**
    * A concrete <code>PropertyMapper</code> implementation for properties
    * which are mapped to a byte array data type.
    */
   public static class RawMapper
   extends PropertyMapper
   {
      /**
       * Constructor.
       *
       * @param   recordClass
       *          Class of DMO object which contains the mapped property.
       * @param   method
       *          Name of the method to be invoked via reflection to set a
       *          data value into the mapped property within a DMO instance.
       * @param   propertyName
       *          Name of the mapped property.
       * @param   index
       *          If the mapped property represents an individual subscripted
       *          value within a Progress extent field, this value represents
       *          the zero-based index of that value, relative to the start
       *          of the extent (array). It should be negative for a property
       *          which is not part of an extent field.
       *
       * @throws  NoSuchMethodException
       *          if <code>method</code> cannot be found within
       *          <code>recordClass</code>.
       */
      public RawMapper(Class<?> recordClass,
                       String method, 
                       String propertyName, 
                       int    index)
      throws NoSuchMethodException
      {
         super(recordClass, method, propertyName, index);
      }
      
      /**
       * Return the type which is expected by the setter method associated
       * with the mapped property; in this case, the <code>BinaryData</code> class.
       *
       * @return  <code>BinaryData</code> class.
       */
      Class<? extends BaseDataType> getPropertyClass()
      {
         return BinaryData.class;
      }
      
      /**
       * Instantiate an object which is the proper type to contain a value for
       * the mapped property, for a particular imported record. This storage
       * will be passed to the import processing and will be assigned the
       * value as read from the export file.
       *
       * @return  Placeholder storage of the correct type.
       */
      BaseDataType createPlaceholder()
      {
         return raw.instantiateUnknownRaw();
      }
   }
   
   /**
    * A concrete <code>PropertyMapper</code> implementation for properties
    * which are mapped to the <code>rec-id</code> or <code>row-id</code> data types.
    */
   public static class RowIDMapper
   extends PropertyMapper
   {
      /**
       * Constructor.
       *
       * @param   recordClass
       *          Class of DMO object which contains the mapped property.
       * @param   method
       *          Name of the method to be invoked via reflection to set a
       *          data value into the mapped property within a DMO instance.
       * @param   propertyName
       *          Name of the mapped property.
       * @param   index
       *          If the mapped property represents an individual subscripted
       *          value within a Progress extent field, this value represents
       *          the zero-based index of that value, relative to the start
       *          of the extent (array). It should be negative for a property
       *          which is not part of an extent field.
       *
       * @throws  NoSuchMethodException
       *          if <code>method</code> cannot be found within
       *          <code>recordClass</code>.
       */
      public RowIDMapper(Class<?> recordClass,
                         String method, 
                         String propertyName, 
                         int    index)
      throws NoSuchMethodException
      {
         super(recordClass, method, propertyName, index);
      }
      
      /**
       * Return the type which is expected by the setter method associated
       * with the mapped property; in this case, the <code>NumberType</code> class.
       *
       * @return  <code>integer</code> class.
       */
      Class<? extends BaseDataType> getPropertyClass()
      {
         return NumberType.class;
      }
      
      /**
       * Instantiate an object which is the proper type to contain a value for
       * the mapped property, for a particular imported record. This storage
       * will be passed to the import processing and will be assigned the
       * value as read from the export file.
       *
       * @return  Placeholder storage of the correct type.  In this case, we
       *          return an <code>integer</code> instance set to unknown value.
       */
      BaseDataType createPlaceholder()
      {
         return new integer();
      }
      
      /**
       * Report whether the property is expected to be found in exported data.
       * Certain property types are skipped when Progress exports data, and as
       * such, these will not appear in an export file.  This method tells the
       * record loader that the lexer should not expect to find any datum for
       * this property in the current row being imported.
       * 
       * @return  <code>false</code>.
       */
      boolean isExported()
      {
         return false;
      }
   }
}