/*
** Module   : SSLCertGenUtil.java
** Abstract : Utility to generate the root CA and SSL certificates for all the accounts in a
**            directory.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140207 First version.
** 002 CA  20140411 Added option to save the root CA in the server certs store too (to be used
**                  as a trusted cert store).
**                  Added option to save the private key for the root CA in an external key store.
** 003 CA  20140707 Allow the account private keys to be saved in external store(s).
**                  Extracted generate() code into distinct methods.
** 004 CA  20160106 Enforce a RSA key length of at least 2048 bits.
**                  Allow server private-keys to be saved in external key-store(s).
** 005 CA  20170630 Allow re-use of the encryption password.  Allow custom common name for 
**                  certificate.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.security;

import java.io.*;
import java.math.*;
import java.security.*;
import java.security.cert.Certificate;
import java.security.cert.*;
import java.util.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.util.*;

/**
 * Generate new root CA and SSL certificates for all the accounts which authenticate via a 
 * certificate and save them and the associated private keys in the directory.  All existing CAs,
 * peer certificates and private keys will be removed.
 * <p>
 * The private keys will be saved under the "/security/certificates/private-keys/" node.
 * <p>
 * The company information (used to build the certificate's subject) will be saved under the
 * "/security/certificates/company/" node.  If the node does not already exist, it will ask the
 * user for the information.
 */
public class SSLCertGenUtil
{
   /** Constant identifying the CAs node in the directory. */
   private static final String ROOT_CA_NODE = "/security/certificates/cas/";

   /** Constant identifying the peer certs node in the directory. */
   private static final String ROOT_PEER_NODE = "/security/certificates/peers/";
   
   /** Constant identifying the security/certificates node in the directory. */
   private static final String ROOT_CERTIFICATES_NODE = "/security/certificates";

   /** Constant identifying the process accounts node in the directory. */
   private static final String ROOT_PROCESSES_NODE = "/security/accounts/processes/";

   /** Constant identifying the user accounts node in the directory. */
   private static final String ROOT_USERS_NODE = "/security/accounts/users/";
   
   /** Constant identifying the company information node in the directory. */
   private static final String ROOT_COMPANY_NODE = "/security/certificates/company/";
   
   /** Constant identifying the common name node in the directory. */
   private static final String ROOT_CERT_COMMON_NAME = "/security/certificates/company/hostnames/";
   
   /** Constant identifying the private keys container node in the directory. */
   private static final String ROOT_PRIVATE_KEYS_NODE = "/security/certificates/private-keys/";

   /** Constant identifying the name of the company node containing the certificate validity. */
   private static final String CERT_VALIDITY_YEARS_NODE = "validityYears";

   /**
    * Constant identifying the name of the company node containing the RSA private key strength.
    */
   private static final String RSA_PRIVATE_KEY_STRENGTH = "rsaKeyStrength";

   /** Constant identifying the name of the company node containing the RSA public exponent. */
   private static final String RSA_PUBLIC_EXPONENT = "rsaPublicExponent";
   
   /** Valid entries for yes/no options. */
   private static final String[] YES_NO = { "yes", "no" };
   
   /** Read data from the standard input. */
   private final BufferedReader reader;

   /** Configuration used to initialize the directory. */
   private final BootstrapConfig cfg;
   
   /** Map identifying the company attributes. */
   private final Map<String, String> company = new HashMap<>();
   
   /** All the collected aliases from the defined accounts. */
   private final Set<String> aliases = new TreeSet<>();
   
   /** All the aliases associated with server processes. */
   private final Set<String> serverAliases = new TreeSet<>();
   
   /** Factory to generate the SSL certificate. */
   private final SSLCertFactory factory;
   
   /** Store where to add the trusted certificates (servers and CAs). */
   private final KeyStore trustCertStore;
   
   /** Store where to add the private keys for the trusted certificates. */
   private final KeyStore trustKeyStore;

   /** Store where to add the certificates associated with non-server accounts. */
   private final KeyStore accCertStore;

   /**
    * Store where to add the private keys for the certificates associated with non-server 
    * accounts.
    */
   private final KeyStore accKeyStore;

   /**
    * The random passwords used to encrypt the private keys in their {@link KeyStore}, per each
    * alias.
    */
   private final Map<String, String> keyEntryPasswords = new HashMap<>();
   
   /** The certificate's validity, in years. */
   private int validity = -1;

   /** Alias for the root CA. */
   private String rootCAAlias = null;

   /** Password for the root CA (if reused). */
   private String rootCAPassword = null;
   
   /** Instance to access the directory. */
   private DirectoryService ds;
   
   /** The master password to encrypt all private keys, in the directory. */
   private String masterPassword = null;

   /** Flag indicating if passwords are reused from the directory. */
   private boolean reusePasswords = false;

   /** Flag indicating if the existing root CA is re-used. */
   private boolean reuseRootCA= false;

   /** The passwords used to encrypt the directory private keys. */
   private Map<String, String> directoryPasswords = new HashMap<>();
   
   /** Array from where to read the input. */
   private String[] inputs = null;
   
   /** Current index in the {@link #inputs}, if set. */
   private int currentInputIdx = 0;

   /**
    * Create a new utility.
    * 
    * @param    config
    *           Configuration for accessing the directory.
    *           
    * @throws   SSLCertGenException
    *           If the utility could not be instantiated.
    */
   public SSLCertGenUtil(BootstrapConfig config)
   throws SSLCertGenException
   {
      Console console = System.console();
      
      if (console != null)
      {
         reader = new BufferedReader(console.reader());
      }
      else
      {
         reader = new BufferedReader(new InputStreamReader(System.in));
      }
      
      cfg = new BootstrapConfig(config);
      if (cfg.getString("directory", "xml", "filename", null) == null)
      {
         throw new NullPointerException("The directory file must be specified!");
      }
      
      // hard-coded to Bouncy Castle
      this.factory = new BCCertFactory();

      // we are using 4 KeyStore instances:
      // 1. key store for the private key for the self-signed, root CA certificate
      // 2. trust-store with the trusted certificates (root CA certificate + server certificates)
      // 3. key store for the private key for each non-serverprocess and user accounts
      // 4. cert store with the process and user certificates

      trustCertStore = createEmptyStore();
      trustKeyStore  = createEmptyStore();
      accCertStore  = createEmptyStore();
      accKeyStore   = createEmptyStore();
   }
   
   /**
    * Main method to (re)generate the root CA, peer certificates and private keys.
    * 
    * @throws    ConfigurationException
    *            In case the {@link #ds directory service} could not be initialized.
    * @throws    IOException
    *            If standard input can not be accessed.
    * @throws    SSLCertGenException
    *            In case of problems during the generation of root CA or peer certificates.
    */
   public void generate()
   throws ConfigurationException,
          IOException,
          SSLCertGenException
   {
      cfg.setConfigItem("directory", "xml", "must_exist", "true");

      String dirFile = cfg.getString("directory", "xml", "filename", null);
      System.out.println("Initializing service for directory " + dirFile + "...");
      this.ds = DirectoryService.createInstance(cfg);

      if (!ds.bind())
         throw new RuntimeException("bind() failed");

      // step 1. read configuration from the /security/certificates/company node
      //         any missing nodes are read from user input and saved
      System.out.println("Reading company configuration...");
      readCompanyConfiguration();

      // step 2. read processes (servers and batch) and users configured with a alias
      readAccounts(ROOT_PROCESSES_NODE);
      readAccounts(ROOT_USERS_NODE);
      
      String answer = null;
      
      // check if we re-use the passwords
      answer = readOption("Re-use the encryption passwords currently in the directory (yes/no): ", 
                          YES_NO);
      reusePasswords = "yes".equals(answer);
      
      // check if we re-use the root CA
      answer = readOption("Re-use the root CA currently in the directory (yes/no): ", 
                          YES_NO);
      reuseRootCA = "yes".equals(answer);
      if (reuseRootCA)
      {
         rootCAPassword = readLine("Enter the root CA private-key encryption password: ");
      }
      
      // step 3. generate root CA and the certificates for each account
      generateRootCA();
      generateCertificates();
      
      // check how to encrypt the private keys
      if (!reusePasswords)
      {
         answer = readOption("Encrypt the in-directory private keys using the same password " +
                             "(yes/no): ", YES_NO);
         
         masterPassword = ("yes".equals(answer)) ? createAES256BitKey() : null;
      }
      
      // step 4. save the certificates and the keys in the directory
      saveRootCA();
      saveCertificates();
      
      if (!ds.unbind())
         throw new RuntimeException("unbind() failed");
      
      answer = readOption("Save the account private keys in external key-store(s) (yes/no): ",
                          YES_NO);
      if ("yes".equalsIgnoreCase(answer))
      {
         savePrivateKeys(false);
      }
      
      answer = readOption("Save the server private keys in external key-store(s) (yes/no): ",
                          YES_NO);
      if ("yes".equalsIgnoreCase(answer))
      {
         savePrivateKeys(true);
      }
      
      // save server's certificates in an external key store
      answer = readOption("Save servers' certificates in an external key store (yes/no): ", 
                          YES_NO);
      
      if ("yes".equalsIgnoreCase(answer))
      {
         saveServerCertificates();
      }
      
      // save the root CA in an external key store
      answer = readOption("Save the root CA private key in an external key store (yes/no): ",
                          YES_NO);
      
      if ("yes".equalsIgnoreCase(answer))
      {
         saveRootCAPrivateKey();
      }
      
      System.out.println("Done.");
      System.out.println("WARNING! Any configuration set at the client.xml or server.xml files " +
                         "or via bootstrap config arguments will have priority over the " +
                         "in-directory keys or certificates.");
      System.out.println("WARNING! The private key encryption passwords for all the account " +
                         "certificates are saved unencrypted in the directory.  The root CA's " +
                         "private key can be safely deleted, if it is not required to issue " +
                         "other certificates using this CA.");
      if (masterPassword != null)
      {
         System.out.println("WARNING! All private keys are encrypted using the same [" +
                            masterPassword + "] password. If needed, delete the 'key-password' " +
                            "nodes from the directory manually, and set this password using the " +
                            "access:password:masterkeyentry bootstrap config at server startup.");
      }
   }
   
   /**
    * Save the root CA private key in an external key store.
    * 
    * @throws    IOException
    *            If standard input can not be accessed.
    * @throws    SSLCertGenException
    *            In case of problems during key store access.
    */
   private void saveRootCAPrivateKey()
   throws IOException,
          SSLCertGenException
   {
      String file = null;
      do
      {
         file = readLine("Enter key store file name: ");
      }
      while (file.length() == 0);

      try
      {
         String kepass = keyEntryPasswords.get(rootCAAlias);
         Key rootKey = this.trustKeyStore.getKey(rootCAAlias, kepass.toCharArray());
         Certificate rootCert = this.trustCertStore.getCertificate(rootCAAlias);
         
         KeyStore store = createEmptyStore();
         store.setKeyEntry(rootCAAlias, rootKey, kepass.toCharArray(),
                           new Certificate[] { rootCert });
         
         String kspass = RandomWordGenerator.create();
         store.store(new FileOutputStream(file), kspass.toCharArray());

         String msg = 
            "The root CA's private key and certificate were saved in the [%s] store.\n" +
            "The key store was encrypted using the [%s] password, while the root " +
            "CA's private key was encrypted using the [%s] password.";
         System.out.println(String.format(msg, file, kspass, kepass));
      }
      catch (KeyStoreException         |
             NoSuchAlgorithmException  |
             UnrecoverableKeyException |
             CertificateException e)
      {
         throw new SSLCertGenException(e);
      }
   }
   
   /**
    * Save the server certificates in an external store.
    * 
    * @throws    IOException
    *            If standard input can not be accessed.
    * @throws    SSLCertGenException
    *            In case of problems during key store access.
    */
   private void saveServerCertificates()
   throws IOException,
          SSLCertGenException
   {
      // save root CA certificate
      String answer = readOption("Include the root CA certificate, to use this store as a " +
                                 "trust store (yes/no)? ", YES_NO);

      boolean saveRootCA = "yes".equalsIgnoreCase(answer);
      
      String file = null;
      do
      {
         file = readLine("Enter key store file name: ");
      }
      while (file.length() == 0);
      
      try
      {
         KeyStore store = createEmptyStore();
         String firstAlias = null;
         for (String alias : serverAliases)
         {
            Certificate cert = trustCertStore.getCertificate(alias);
            
            store.setCertificateEntry(alias, cert);
            
            if (firstAlias == null)
            {
               firstAlias = alias;
            }
         }
         
         if (saveRootCA)
         {
            store.setCertificateEntry(rootCAAlias, trustCertStore.getCertificate(rootCAAlias));
         }
         
         String pass = readPassword(firstAlias);
         store.store(new FileOutputStream(file), pass.toCharArray());
         
         String msg = "The servers' certificates %s were saved in the [%s] key store, " +
                      "using password [%s].";
         System.out.println(String.format(msg, serverAliases.toString(), file, pass));
      }
      catch (KeyStoreException | NoSuchAlgorithmException | CertificateException e)
      {
         throw new SSLCertGenException(e);
      }
   }
   
   /**
    * Get the reused password or generate a new one, for the given alias.
    * 
    * @param    alias
    *           The certificate alias.
    *           
    * @return   See above.
    */
   private String readPassword(String alias)
   {
      if (reusePasswords)
      {
         if (directoryPasswords.containsKey(alias))
         {
            return directoryPasswords.get(alias);
         }
      }
      
      return RandomWordGenerator.create();
   }
   
   /**
    * Save the private keys in external key store(s).
    * 
    * @param    server
    *           <code>true</code> to save the server private keys; <code>false</code> to save
    *           the account private-keys.
    * 
    * @throws   IOException
    *           If standard input can not be accessed.
    * @throws   SSLCertGenException
    *           In case of problems during key store access.
    */
   private void savePrivateKeys(boolean server)
   throws IOException,
          SSLCertGenException
   {
      // save the private keys in external key store
      String answer = readOption("Save the private keys in the same key store (yes/no): ",
                                 YES_NO);
      boolean samePKStore = "yes".equals(answer); 
      KeyStore pkStore = null;
      String pkFile = null;
      
      KeyStore certStore = (server ? trustCertStore : accCertStore);
      KeyStore keyStore = (server ? trustKeyStore : accKeyStore);
      
      for (String alias : aliases)
      {
         if (server != serverAliases.contains(alias))
         {
            // skip, if the alias is not for a server (and server is used) or if alias is for a
            // server (and account is used)
            continue;
         }
         
         if ((samePKStore && pkStore == null) || !samePKStore)
         {
            if (samePKStore)
            {
               do
               {
                  pkFile = readLine("Enter key store file name: ");
               }
               while (pkFile.length() == 0);
            }
            else
            {
               pkFile = String.format("%s-private-key.store", alias);
            }
         
            pkStore = createEmptyStore();
         }
         
         try
         {
            Key         key      = keyStore.getKey(alias, keyEntryPasswords.get(alias).toCharArray());
            Certificate cert     = certStore.getCertificate(alias);
            Certificate rootCert = trustCertStore.getCertificate(rootCAAlias);
            
            String password = (reusePasswords && directoryPasswords.containsKey(alias) 
                                 ? directoryPasswords.get(alias) 
                                 : keyEntryPasswords.get(alias));
            pkStore.setKeyEntry(alias, key, password.toCharArray(), 
                                new Certificate[] { cert, rootCert });
            
            if (!samePKStore)
            {
               String ksPass = readPassword(alias);
               pkStore.store(new FileOutputStream(pkFile), ksPass.toCharArray());
               
               String msg = "The private key for certificate [%s] was saved in the [%s] " +
                            "key store, using store password [%s] and key-entry password [%s].";
               System.out.println(String.format(msg, alias, pkFile, ksPass, password));
            }
            else
            {
               String msg = "Adding the private key for certificate [%s] to store [%s], " +
                            "using key-entry password [%s]...";
               System.out.println(String.format(msg, alias, pkFile, password));
            }
         }
         catch (UnrecoverableKeyException |
                KeyStoreException         |
                NoSuchAlgorithmException  |
                CertificateException e)
         {
            throw new SSLCertGenException(e);
         }
      }
      
      if (samePKStore)
      {
         try
         {
            String ksPass = RandomWordGenerator.create();
            pkStore.store(new FileOutputStream(pkFile), ksPass.toCharArray());

            String msg = "The key store [%s] was encrypted using password [%s].";
            System.out.println(String.format(msg, pkFile, ksPass));
         }
         catch (KeyStoreException | NoSuchAlgorithmException | CertificateException e)
         {
            throw new SSLCertGenException(e);
         }
      }
   }
   
   /**
    * Save all the root CA certificate and its private key in the directory.
    * 
    * @throws    SSLCertGenException
    *            If the data could not be saved.
    */
   private void saveRootCA()
   throws SSLCertGenException
   {
      openBatch(ROOT_CERTIFICATES_NODE);
      
      savePrivateKey(rootCAAlias, trustKeyStore);
      saveCertificate(rootCAAlias, trustCertStore, ROOT_CA_NODE);
      
      closeBatch(ROOT_CERTIFICATES_NODE);
   }
   
   /**
    * Save all the peer certificates and their private keys in the directory.
    * 
    * @throws    SSLCertGenException
    *            If the data could not be saved.
    */
   private void saveCertificates()
   throws SSLCertGenException
   {
      openBatch(ROOT_CERTIFICATES_NODE);

      for (String alias : aliases)
      {
         boolean isServer = serverAliases.contains(alias);
         
         savePrivateKey(alias, (isServer ? trustKeyStore : accKeyStore));
      }

      for (String alias : aliases)
      {
         boolean isServer = serverAliases.contains(alias);
         
         saveCertificate(alias, (isServer ? trustCertStore : accCertStore), ROOT_PEER_NODE);
      }

      closeBatch(ROOT_CERTIFICATES_NODE);
   }
   
   /**
    * Save the certificate for the specified alias in the directory, under the specified parent
    * node.
    * 
    * @param    alias
    *           The alias.
    * @param    certStore
    *           The store from which to retrieve the certificate.
    * @param    parentNode
    *           The parent node where to save the certificate.
    *           
    * @throws   SSLCertGenException
    *           If the data could not be saved.
    */
   private void saveCertificate(String alias, KeyStore certStore, String parentNode)
   throws SSLCertGenException
   {
      String fullId = parentNode + alias;
      deleteNode(fullId);
      
      System.out.println(String.format("Adding certificate for alias [%s] to the %s node...",
                                       alias, fullId));
      
      try
      {
         Certificate cert = certStore.getCertificate(alias);

         addNode(fullId, "bytes", cert.getEncoded());
      }
      catch (KeyStoreException | CertificateEncodingException e)
      {
         throw new SSLCertGenException(e);
      }
      
   }
   
   /**
    * Save the private key for the specified alias in the directory.
    * <p>
    * The private key is encrypted using a random password, which will be saved too in the
    * directory.
    * 
    * @param    alias
    *           The alias.
    * @param    privateKeyStore
    *           The store from which to read the private key.
    *           
    * @throws   SSLCertGenException
    *           If the private key could not be saved.
    */
   private void savePrivateKey(String alias, KeyStore privateKeyStore)
   throws SSLCertGenException
   {
      String fullId = ROOT_PRIVATE_KEYS_NODE + alias;

      if (ds.getNodeClass(ROOT_PRIVATE_KEYS_NODE) == null)
      {
         addNode(ROOT_PRIVATE_KEYS_NODE, "container", null);
      }
      
      String password = keyEntryPasswords.get(alias);
      boolean rootCA = rootCAAlias.equals(alias);

      // determine which password to use: master or distinct
      String dirPassword = (masterPassword != null && !rootCA)
                              ? masterPassword 
                              : rootCA && reuseRootCA ? rootCAPassword : createAES256BitKey(alias);
      
      deleteNode(fullId);
      addNode(fullId, "container", null);
      
      if (!rootCA)
      {
         // do not save the root CA password
         addNode(ROOT_PRIVATE_KEYS_NODE + alias + "/key-password", "bytes", dirPassword.getBytes());
      }
      
      directoryPasswords.put(alias, dirPassword);
      
      try
      {
         Key key = privateKeyStore.getKey(alias, password.toCharArray());

         byte[] encrypted = factory.encryptPrivateKey(key, dirPassword);
         addNode(ROOT_PRIVATE_KEYS_NODE + alias + "/key-entry", "bytes", encrypted);
      }
      catch (UnrecoverableKeyException | KeyStoreException | NoSuchAlgorithmException e)
      {
         throw new SSLCertGenException(e);
      }
      
      final String msg1 = "Private key for alias [%s] was encrypted using the [%s] password " +
                          "and saved in the %s node.\nUse the access:password:keyentry-%s key " +
                          "to set this password in the server's bootstrap config file and " +
                          "delete  the /security/certificates/private-keys/%s node from " +
                          "the directory.";
      final String msg2 = "Private key for root CA [%s] was encrypted using the [%s] password " +
                          "and the password was NOT SAVED in the directory.  Losing this " +
                          "password will prohibit using the root CA to issue new certificates.";
      if (rootCA)
      {
         System.out.println(String.format(msg2, alias, dirPassword));
      }
      else
      {
         System.out.println(String.format(msg1, alias, dirPassword, fullId, alias, alias));
      }
   }
   
   /**
    * Generate peer certificates for all {@link #aliases}.
    * 
    * @throws    SSLCertGenException
    *            If the SSL certificates could not be generated.
    */
   private void generateCertificates()
   throws SSLCertGenException
   {
      for (String alias : aliases)
      {
         boolean isServer = (serverAliases.contains(alias));
         String type = (isServer ? "server" : "account");
         String commonName = alias + " " + type;
         
         System.out.println(String.format("Generating certificate for %s with alias [%s]...",
                                          type, alias));
         
         String customCN = ds.getNodeString(ROOT_CERT_COMMON_NAME + "/" + alias, "value");
         if (customCN != null)
         {
            commonName = customCN;
         }
         
         String password = factory.generateCertificate(alias, validity, 
                                                       commonName, company,
                                                       (isServer ? trustCertStore : accCertStore), 
                                                       (isServer ? trustKeyStore : accKeyStore));

         keyEntryPasswords.put(alias, password);
      }
   }
   
   /**
    * Generate the root CA.
    * 
    * @throws    SSLCertGenException
    *            If the root CA could not be generated.
    * @throws    IOException
    *            If the alias for the root CA could not be read from standard input.
    */
   private void generateRootCA()
   throws SSLCertGenException,
          IOException
   {
      // find the root CA alias
      openBatch(ROOT_CA_NODE);
      
      String[] cas = ds.enumerateNodes(ROOT_CA_NODE);
      if (cas.length == 0)
      {
         // read the root CA alias
         while (rootCAAlias == null || rootCAAlias.length() == 0)
         {
            rootCAAlias = readLine("Enter the root CA alias: ");
            
            if (!aliases.contains(rootCAAlias))
            {
               break;
            }
            
            System.out.println("Alias " + rootCAAlias + " is in use!");
            rootCAAlias = null;
         }
         
         reuseRootCA = false;
         rootCAPassword = null;
      }
      else
      {
         // reuse the root CA alias
         rootCAAlias = cas[0];
      }

      closeBatch(ROOT_CA_NODE);
      
      String msg = reuseRootCA ? "Re-using root CA using [%s] alias..."
                               : "Generating root CA using [%s] alias...";
      System.out.println(String.format(msg, rootCAAlias));

      String password;
      if (reuseRootCA)
      {
         try
         {
            password = rootCAPassword;
            
            // read the root CA from the directory (certificate and private key)
            byte[] bpk = ds.getNodeByteArray(ROOT_PRIVATE_KEYS_NODE + rootCAAlias + "/key-entry", 
                                             "value");
            byte[] bcert = ds.getNodeByteArray(ROOT_CERTIFICATES_NODE + "/cas/" + rootCAAlias, 
                                               "value");
            
            PrivateKey pk = factory.decryptPrivateKey(bpk, rootCAPassword);
            CertificateFactory cf = CertificateFactory.getInstance("X.509");
            X509Certificate cert = (X509Certificate) cf.generateCertificate(new ByteArrayInputStream(bcert));
            
            trustCertStore.setCertificateEntry(rootCAAlias, cert);
            trustKeyStore.setKeyEntry(rootCAAlias,
                                      pk,
                                      rootCAPassword.toCharArray(), 
                                      new Certificate[] { cert });
            
            factory.setRootCA(cert, pk);
         }
         catch (Exception e)
         {
            throw new SSLCertGenException(e);
         }
      }
      else
      {
         password = factory.generateRootCA(rootCAAlias, validity,
                                           rootCAAlias + " - root CA authority", company,
                                           trustCertStore, trustKeyStore);
      }
      
      keyEntryPasswords.put(rootCAAlias, password);
   }
   
   /**
    * Read all the accounts which have an alias specified.
    * 
    * @param    rootNode
    *           The node from which to read the accounts.
    */
   private void readAccounts(String rootNode)
   {
      openBatch(rootNode);
      
      String[] accIds = ds.enumerateNodes(rootNode);
      if (accIds.length == 0)
      {
         System.out.println("No accounts defined in " + rootNode + "!");
         return;
      }
      
      String  alias    = null;
      Boolean isServer = null;
      
      // process existing nodes
      for (int i = 0; i < accIds.length; i++)
      {
         String accID = accIds[i];
         String fullId = new String(rootNode + accID);
         String nodeCls = ds.getNodeClass(fullId);
         
         if ("/meta/class/process".equals(nodeCls))
         {
            isServer = ds.getNodeBoolean(fullId, "server");
         }
         else if ("/meta/class/user".equals(nodeCls))
         {
            // user account
         }
         else
         {
            System.out.println("Account " + fullId + " ignored - not a user or process");
            continue;
         }

         alias    = ds.getNodeString(fullId,  "alias");
         
         if (alias == null || alias.length() == 0)
         {
            System.out.println("Account " + fullId + " ignored - no alias is defined");
            continue;
         }

         if (aliases.contains(alias))
         {
            System.out.println("WARNING: alias " + alias + " is set for multiple accounts!");
            continue;
         }

         System.out.println("Adding account " + fullId);

         aliases.add(alias);
         
         if (isServer != null && isServer.booleanValue())
         {
            serverAliases.add(alias);
         }
      }
      
      closeBatch(rootNode);
   }
   
   /**
    * Read the existing company configuration from the directory or, if missing, read it from
    * the standard input.
    * 
    * @throws    IOException
    *            In case of problems during reading.
    * @throws    SSLCertGenException
    *            If the {@link SSLCertFactory} factory could not be initialized.
    */
   private void readCompanyConfiguration()
   throws IOException,
          SSLCertGenException
   {
      Map<String, String> mandatoryKeys = factory.getMandatorySubjectFields();
      
      while (!ds.openBatch(ROOT_COMPANY_NODE))
      {
         // node does not exist, add it
         openBatch(ROOT_CERTIFICATES_NODE);
         
         // add a new node
         addNode(ROOT_COMPANY_NODE, "container", null);
         
         closeBatch(ROOT_CERTIFICATES_NODE);
      }

      Integer rsaKeyStrength = null;
      BigInteger rsaPublicExponent = null;
      
      // read the existing nodes first
      String[] existing = ds.enumerateNodes(ROOT_COMPANY_NODE);
      for (int i = 0; i < existing.length; i++)
      {
         String oid = existing[i].toUpperCase();
         
         if (oid.equalsIgnoreCase(CERT_VALIDITY_YEARS_NODE))
         {
            validity = ds.getNodeInteger(ROOT_COMPANY_NODE + oid, "value");
            System.out.println(String.format("Using %d years for validity.", validity));
            continue;
         }
         
         if (oid.equalsIgnoreCase(RSA_PRIVATE_KEY_STRENGTH))
         {
            rsaKeyStrength = ds.getNodeInteger(ROOT_COMPANY_NODE + oid, "value");
            System.out.println(String.format("Using %d bits for RSA private key strength.", 
                                             rsaKeyStrength));
            continue;
         }
         
         if (oid.equalsIgnoreCase(RSA_PUBLIC_EXPONENT))
         {
            // the exponent can be a very large or even a hex value, so keep it in a string. 
            String exp = ds.getNodeString(ROOT_COMPANY_NODE + oid, "value");
            rsaPublicExponent = new BigInteger(exp);
            System.out.println(String.format("Using %s for RSA public exponent.", exp));
            continue;
         }
         
         if (oid.equalsIgnoreCase("hostnames"))
         {
            continue;
         }

         String val = ds.getNodeString(ROOT_COMPANY_NODE + oid, "value").trim();

         // existing value (may be empty)
         String txt = mandatoryKeys.get(oid);
         System.out.println(String.format("Using '%s' for [%s] %s.", val, oid,
                                          (txt == null ? "" : txt)));
         
         mandatoryKeys.remove(oid);
         company.put(oid, val);
      }
      
      // get user input for the remainder of the nodes
      for (String oid : mandatoryKeys.keySet())
      {
         String txt = mandatoryKeys.get(oid);
         
         // read the value
         String val = readLine(String.format("%s [%s]: ", txt, oid));
         
         addNode(ROOT_COMPANY_NODE + oid, "string", val);
         
         // save empty values
         System.out.println(String.format("Using '%s' for [%s] %s.", val, oid, txt));
      }

      // check the certificate validity
      if (validity == -1)
      {
         while (true)
         {
            try
            {
               validity = Integer.parseInt(readLine("Enter validity (years): "));
               
               if (validity <= 0)
               {
                  System.out.println("Enter a number of years greater than zero.");
                  continue;
               }
               
               break;
            }
            catch (NumberFormatException e)
            {
               System.out.println("Incorrect year.");
               continue;
            }
         }
         
         addNode(ROOT_COMPANY_NODE + CERT_VALIDITY_YEARS_NODE, "integer", validity);
         
         System.out.println(String.format("Using %d years for validity.", validity));
      }
      
      // get the key strength
      if (rsaKeyStrength == null)
      {
         do
         {
            String msg = "Enter RSA key strength (in bits), greater or equal than " +
                         "%d (blank for default %d): ";
            msg = String.format(msg,
                                SSLCertFactory.MIN_RSA_KEY_STRENGTH, 
                                SSLCertFactory.MIN_RSA_KEY_STRENGTH);
            String line = readLine(msg);
            rsaKeyStrength = (line.length() > 0) ? new Integer(line) 
                                                 : SSLCertFactory.MIN_RSA_KEY_STRENGTH;
            
            if (rsaKeyStrength > SSLCertFactory.MIN_RSA_KEY_STRENGTH)
            {
               addNode(ROOT_COMPANY_NODE + RSA_PRIVATE_KEY_STRENGTH, "integer", rsaKeyStrength);
               
               System.out.println(String.format("Using %d bits for RSA private key strength.", 
                                                rsaKeyStrength));
            }
         }
         while (rsaKeyStrength < SSLCertFactory.MIN_RSA_KEY_STRENGTH);
      }
      
      // get the public exponent
      if (rsaPublicExponent == null)
      {
         String line = readLine("Enter RSA public exponent, blank for default (65337): ");
         if (line.length() > 0)
         {
            rsaPublicExponent = new BigInteger(line);

            addNode(ROOT_COMPANY_NODE + RSA_PUBLIC_EXPONENT,
                    "string", rsaPublicExponent.toString());
            
            System.out.println(String.format("Using %s for RSA public exponent.", 
                                             rsaPublicExponent.toString()));
         }
         else
         {
            rsaPublicExponent = new BigInteger("65337");
         }
      }
      
      factory.init(rsaKeyStrength, rsaPublicExponent);
      
      closeBatch(ROOT_COMPANY_NODE);
   }

   /**
    * Delete the specified node and all of its children.
    * 
    * @param    nodeId
    *           The node ID.
    * 
    * @throws   RuntimeException
    *           If the node or one of its children could not be deleted.
    */
   private void deleteNode(String nodeId)
   {
      if (ds.getNodeClass(nodeId) != null)
      {
         String[] children = ds.enumerateNodes(nodeId);
         for (int i = 0; i < children.length; i++)
         {
            String fullId = nodeId + "/" + children[i];
            
            // remove previous entry
            if (!ds.deleteNode(fullId))
            {
               throw new RuntimeException("deleteNode() failed for the " + fullId + " node");
            }
         }

         if (!ds.deleteNode(nodeId))
         {
            throw new RuntimeException("deleteNode() failed for the " + nodeId + " node");
         }
      }
   }
   
   /**
    * Add a new node to the directory, having the specified class and a single attribute
    * with the same class and specified value.
    * 
    * @param    nodeId
    *           The node ID.
    * @param    cls
    *           The node's and attribute's class.
    * @param    value
    *           The attribute's value.
    * 
    * @throws   RuntimeException
    *           If the node could not be added.
    */
   private void addNode(String nodeId, String cls, Object value)
   {
      NodeAttribute nodeAttr = ds.getClassNodeAttribute(cls, "value");
      Attribute attr = new Attribute(nodeAttr, new Object[] { value });
      addNode(nodeId, cls, new Attribute[] { attr });
   }
   
   /**
    * Add a new node to the directory, having the specified class and attributes.
    * 
    * @param    nodeId
    *           The node ID.
    * @param    cls
    *           The node's class.
    * @param    attrs
    *           The node's attributes.
    * 
    * @throws   RuntimeException
    *           If the node could not be added.
    */
   private void addNode(String nodeId, String cls, Attribute[] attrs)
   {
      if (!ds.addNode(nodeId, cls, attrs))
      {
         throw new RuntimeException("addNode() failed for the " + nodeId + " node");
      }
   }
   
   /**
    * Open a batch editing session for the specified node.
    * 
    * @param    node
    *           The node ID.
    * 
    * @throws   RuntimeException
    *           If the batch editing session could not be opened.
    */
   private void openBatch(String node)
   {
      if (!ds.openBatch(node))
      {
         throw new RuntimeException("openBatch() failed for the " + node + " node");
      }
   }
   
   /**
    * Close a batch editing session.
    * 
    * @param    node
    *           The node ID.
    * 
    * @throws   RuntimeException
    *           If the batch editing session could not be closed.
    */
   private void closeBatch(String node)
   {
      if (!ds.closeBatch(true))
         throw new RuntimeException("closeBatch() failed for " + node);
   }

   /**
    * Read a line of text using the created {@link #reader}.
    * 
    * @param    txt
    *           Description to be written to standard output.
    *           
    * @return   The line of text from standard input.
    * 
    * @throws   IOException
    *           If data could not be read.
    */
   private String readLine(String txt)
   throws IOException
   {
      if (inputs != null)
      {
         return inputs[currentInputIdx++];
      }

      System.out.print(txt);
      
      String line = reader.readLine();
   
      return line.trim();
   }
   
   /**
    * Ask the user to enter one of the specified valid options, using the given message.
    * 
    * @param    msg
    *           The message shown to the user.
    * @param    valid
    *           An array of valid options.
    */
   private String readOption(String msg, String[] valid)
   throws IOException
   {
      String answer = null;
      l1: do
      {
         answer = readLine(msg);
         
         for (String s : valid)
         {
            if (answer.equalsIgnoreCase(s))
            {
               break l1;
            }
         }
      }
      while (true);
      
      return answer;
   }

   /**
    * Create an empty store, to hold either private keys or certificates.
    * 
    * @return   An empty {@link KeyStore} instance.
    * 
    * @throws   SSLCertGenException
    *           If the store could not be generated.
    */
   private KeyStore createEmptyStore()
   throws SSLCertGenException
   {
      try
      {
         KeyStore ks = KeyStore.getInstance("JKS");
         ks.load(null, null);
         return ks;
      }
      catch (NoSuchAlgorithmException | CertificateException | IOException | KeyStoreException e)
      {
         throw new SSLCertGenException(e);
      }
   }

   /**
    * Create a random 256-bit password to be used as an AES encryption key.
    * 
    * @return   See above.
    */
   private String createAES256BitKey()
   {
      return RandomWordGenerator.create(32, 8, 4);
   }
   
   /**
    * Create a random 256-bit password to be used as an AES encryption key.
    * <p>
    * If {@link #reusePasswords} is on, and there is a password in the directory, that will be
    * returned.
    * 
    * @param    alias
    *           The alias to check.
    *           
    * @return   See above.
    */
   private String createAES256BitKey(String alias)
   {
      if (reusePasswords)
      {
         String node = ROOT_PRIVATE_KEYS_NODE + alias + "/key-password";
         byte[] bytes = ds.getNodeByteArray(node, "value");
         if (bytes != null)
         {
            return new String(bytes);
         }
      }
      
      return createAES256BitKey();
   }
   
   /**
    * Command line driver.
    * 
    * @param    args
    *           Application command line parameters. File name is the only one expected.
    *
    * @throws    ConfigurationException
    *            In case the {@link #ds directory service} could not be initialized.
    * @throws    IOException
    *            If standard input can not be accessed.
    * @throws    SSLCertGenException
    *            In case of problems during the generation of root CA or peer certificates.
    */
   public static void main(String[] args)
   throws ConfigurationException,
          IOException,
          SSLCertGenException
   {
      String dirFile = null;

      if (args.length == 0)
      {
         System.out.println("usage: java SSLCertGenUtil <directory.xml> <inputs...>");
         return;
      }
      
      dirFile = args[0];
      String[] inputs = null;
      
      if (args.length > 1)
      {
         inputs = new String[args.length - 1];
         System.arraycopy(args, 1, inputs, 0, args.length - 1);
      }
      
      BootstrapConfig cfg = new BootstrapConfig();
      
      cfg.setConfigItem("directory", "backend", "type",        "xml");
      cfg.setConfigItem("directory", "xml",     "filename",    dirFile);
      cfg.setConfigItem("directory", "xml",     "must_exist", "true");
      cfg.setServer(true);

      SSLCertGenUtil gen = new SSLCertGenUtil(cfg);
      gen.inputs = inputs;
      gen.generate();
   }
}