/*
** Module   : ProgressStatsHelper.java
** Abstract : helper to copy and link to Progress 4GL source code
**
** Copyright (c) 2007-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -------------------Description--------------------
** 001 GES 20070716   @34526 Helper to copy and link to Progress 4GL
**                           source code.
** 002 GES 20130206          Moved parent class to the pattern package.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.uast;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.pattern.*;

/**
 * Helper to copy and link to Progress 4GL source code.
 */
public class ProgressStatsHelper
implements SourceFileHelper
{
   /** Column headings. */
   private static final String[] headings = new String[]
   {
      "Original Source File",
      "Preprocessed Source",
      "Lexer Output",
      "Parser Output",
      "Persisted AST"
   }; 
   
   /**
    * Obtain the list of source file descriptions which is suitable for use
    * as the column headings in the source file index table.
    *
    * @return   Column headings.
    */
   public String[] getHeadings()
   {
      return headings;
   }
   
   /**
    * Provide the list of copy specifications for the given source file and
    * output directory.
    *
    * @param    filename
    *           Original source filename.
    * @param    outDir
    *           Root of the output directory where all copies must be made.
    *
    * @return   The list of source and target copy operations to be made.
    */
   public CopySpec[] getSpecifications(String filename, String outDir)
   {
      // safety first
      if (filename == null)
         return null;
      
      CopySpec[] specs = new CopySpec[5];
      
      String[] sources = new String[5];
      String[] targets = new String[5];
      
      try
      {
         // create the master source filename 
         sources[0] = Configuration.toFile(filename).getAbsolutePath();
      }
      catch (ConfigurationException ce)
      {
         return null;
      }
      
      // now create the rest of the source filenames
      StringBuilder sb = new StringBuilder(sources[0]);
      sb.append(AstGenerator.CACHE_POSTFIX);
      sources[1] = sb.toString();
      
      sb = new StringBuilder(sources[0]);
      sb.append(AstGenerator.LEXER_POSTFIX);
      sources[2] = sb.toString();
      
      sb = new StringBuilder(sources[0]);
      sb.append(AstGenerator.PARSER_POSTFIX);
      sources[3] = sb.toString();
      
      sb = new StringBuilder(sources[0]);
      sb.append(AstGenerator.AST_POSTFIX);
      sources[4] = sb.toString();
      
      // create the target filenames
      sb = new StringBuilder(outDir).append(filename);
      sb.append(".txt");
      targets[0] = sb.toString();
      
      sb = new StringBuilder(outDir).append(filename);
      sb.append(AstGenerator.CACHE_POSTFIX).append(".html");
      targets[1] = sb.toString();
      
      sb = new StringBuilder(outDir).append(filename);
      sb.append(AstGenerator.LEXER_POSTFIX).append(".txt");
      targets[2] = sb.toString();
      
      sb = new StringBuilder(outDir).append(filename);
      sb.append(AstGenerator.PARSER_POSTFIX).append(".txt");
      targets[3] = sb.toString();
      
      sb = new StringBuilder(outDir).append(filename);
      sb.append(AstGenerator.AST_POSTFIX);
      targets[4] = sb.toString();
      
      specs[0] = new CopySpec(sources[0], targets[0]);
      specs[1] = new CopySpec(sources[1],
                              targets[1],
                              CopySpec.Type.TXT_TO_HTML);
      specs[2] = new CopySpec(sources[2], targets[2]);
      specs[3] = new CopySpec(sources[3], targets[3]);
      specs[4] = new CopySpec(sources[4], targets[4]);
      
      return specs;
   }
   
   /**
    * Create a URL to link to the given source file and line number.
    *
    * @param    relPath
    *           Relative path to the root of the output directory where the
    *           source files copies were made.
    * @param    source
    *           Original source filename.
    * @param    line
    *           The line number to which to link.
    *
    * @return   A URL to link to the copied file.
    */
   public String getURL(String relPath, String source, String line)
   {
      StringBuilder sb = new StringBuilder();
      sb.append(relPath)
        .append(source)
        .append(AstGenerator.CACHE_POSTFIX)
        .append(".html#")
        .append(line);
        
      return sb.toString();
   }
}