/*
** Module   : SymbolResolver.java
** Abstract : manages the symbol dictionaries and lookup process
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- --------------------------- Description ---------------------------
** 001 GES 20041104   @18650 First version, with support for resolving:
**                           - optionally abbreviated keywords
**                           - variables with multiple scoping levels
**                             including a special global scope
** 002 GES 20041115   @18651 Added support for:
**                           - functions
**                           - procedures
**                           - labels
** 003 GES 20041208   @18937 Added new methods for lookup of variables
**                           and functions.  These new methods are now
**                           using the old "default" names and the old
**                           methods that handle exact matching are under
**                           new names lookup*Exact().  The new methods 
**                           implement an algorithm that checks to see
**                           if a match is made to a language keyword,
**                           if such a match exists, a 2nd level of
**                           processing occurs. Otherwise the processing
**                           is the same as before.  This new algorithm
**                           is necessary to allow built-in functions and
**                           variables to be properly resolved, since
**                           these symbols are keywords which must support
**                           the reserved/unreserved attribute and the
**                           matching must be done considering
**                           abbreviations.
** 004 GES 20041217   @19041 Added a first pass implementation for the
**                           lookup of database objects (databases,
**                           tables and fields) in both unqualified
**                           and partially/fully qualified forms.  In
**                           addition, table and field abbreviations are
**                           fully supported.  Warning: table support is
**                           hard coded at this time.
** 005 NVS 20041225   @19109 Now the label dictionary is scoped, with one
**                           label per scope (Progress allows only one
**                           label per block). Modified addLabel creates
**                           new scope implicitly. Added deleteLabel()
**                           method which deletes the current scope.
** 006 ECF 20050103   @19188 Reimplemented lookupDatabase, lookupTable,
**                           and lookupField methods to handle exceptions
**                           thrown by SchemaDictionary. In each case, if
**                           an exception is caught, error information is
**                           written to stdout and -1 is returned as the
**                           token type.
** 007 GES 20050103   @19192 Added isTable and isField methods which 
**                           don't throw exceptions. This allows the
**                           parser to test things that may be ambiguous
**                           without being hit over the head and left
**                           bleeding in the gutter.
** 008 GES 20050107   @19270 Widgets are handled just like variables 
**                           (they share the same namespace) in Progress.
**                           However due to the fact that some symbols
**                           do "double duty" as both variables and as
**                           widgets, a separate widget namespace is
**                           kept so that the generic one name to one
**                           token type mapping can be kept for the
**                           majority of the standard name lookups. A
**                           new widgetDict member is added and two
**                           methods (addWidget and lookupWidget) are
**                           provided to provide access.  This new
**                           namespace does participate in the scoping
**                           just like variables.
** 009 GES 20050118   @19336 Added support for modifying the schema
**                           namespace (adding tables, fields and lists
**                           of fields).
** 010 GES 20050119   @19346 Added support for record scoping (driving
**                           the schema's scopes and which tables are
**                           promoted to the current scope).
** 011 GES 20050127   @19449 Added new namespaces for frames, menus (+
**                           sub-menus) and menu-items.
** 012 GES 20050128   @19460 Added new namespace for streams.
** 013 GES 20050201   @19513 Added additional parser token type to
**                           schema dictionary data type conversions
**                           (for variable types like VAR_INT). Also
**                           added dump support for the schema dictionary.
** 014 GES 20050204   @19570 Added loadSchemaDatabase() method to load
**                           an arbitrary non-default database into the
**                           schema namespace.
** 015 GES 20050207   @19624 Removed deprecated method call to schema
**                           namespace.
** 016 GES 20050209   @19737 Modified addTable() method to allow user
**                           control over where a table is created,
**                           either in the global scope or in an internal
**                           procedure or function scope (buffers only).
** 017 GES 20050210   @19749 Added the ability to add a variable using
**                           data contained in a wrapper class. This may
**                           be extended later to store more data in the
**                           dictionary than is possible today. In
**                           addition a new method of adding lists of
**                           variables, widgets and frames to the
**                           respective dictionaries was provided. Also
**                           added a method for adding aliases to the 
**                           schema namespace.
** 018 GES 20050218   @19860 Minor improvements to the loading of
**                           variables, to allow fields and a special
**                           token type of DYNAMIC to be loaded by the
**                           same mechanism.
** 019 GES 20050307   @20217 Added methods to lookup the fully qualified
**                           name of a field or table.
** 020 GES 20050331   @20561 Major additions to store options at the
**                           time that variables and temp-tables are
**                           defined. These options are saved in the
**                           variable dictionary in the case of vars and
**                           are pushed out to the schema dictionary in
**                           the case of temp-tables. All variable and
**                           temp-table definitions that are LIKE another
**                           variable, field or table now honor and
**                           bidirectionally pass the options as needed.
**                           Variable and field references (the lvalue
**                           rule) store annotations for non-default
**                           options. WARNING: INITIAL, LABEL and
**                           VIEW-AS are 3 common options that are not
**                           processed or are only partially processed
**                           at this time.
** 021 ECF 20050331   @20613 Implemented all schema dictionary hooks added
**                           for 020 (@20561).
** 022 GES 20050401   @20596 Added support for a temporary index to allow
**                           the linking of a variable reference to its
**                           original definition at parse time. This is
**                           needed because project-unique IDs can't be
**                           easily generated at parse time, but are only
**                           added after the entire tree is complete.
** 023 GES 20050406   @20651 Massive rewrite to use ASTs to store variable
**                           options and to exchange ASTs with the schema
**                           dictionary and fully handle field and
**                           variable options.  In addition it is now
**                           possible to instantiate this class in a
**                           manner that bypasses the schema dictionary.
**                           This allows the ProgressLexer to be used
**                           without dependencies upon an already
**                           functional schema.
** 024 ECF 20050413   @20701 Implemented all schema dictionary hooks which
**                           needed to be modified as a result of 023
**                           (@20651). TODO: still need to handle VALIDATE
**                           keyword.
** 025 GES 20050414   @20723 Converted function processing into a wrapped
**                           approach that allows storing other attributes
**                           and providing annotation helpers.  This
**                           required some rework of the variable/field
**                           processing.
** 026 GES 20050509   @21132 Modified annotateVariableOptions() to be
**                           safe when called by the class
**                           ProgressExpressionEvaluator which short-
**                           circuits the normal variable dictionary
**                           processing and thus can generate nulls
**                           where they are normally not possible.
** 027 GES 20050822   @22187 Support schema dictionary record type
**                           lookups for fields and buffers.  This
**                           reports the token type of the backing
**                           construct (table, temp-table or work-table).
** 028 GES 20050830   @22431 Added attribute/method namespace support.
** 029 GES 20050922   @22778 Support schema dictionary buffer name
**                           lookups for fields and records.  This is
**                           sometimes (often) different from the fully
**                           qualified schema name and is needed to know
**                           which buffer this reference is uniquely
**                           referencing.
** 030 GES 20051004   @22931 Added database name lookup for tables (this
**                           is sometimes different from the fully
**                           qualified schema name).
** 031 GES 20051013   @23058 Moved field annotation here from parser and
**                           added helper to create a field node based
**                           on an unambiguous field name.
** 032 GES 20051109   @23282 Added more annotations and tempidx support
**                           for non-var widgets (buttons/browse) by
**                           using the wrapper approach.
** 033 ECF 20060425   @25740 Enhanced support for define temp-table. A
**                           field can be like a specific element of an
**                           extent field.
** 034 ECF 20060427   @25823 Enhanced support for defining a variable like
**                           a database field. Enabled specification of a
**                           specific element of an extent field as the
**                           model field.
** 035 GES 20070329   @32655 Extended the concept in H034 to all forms of
**                           variable and field definition using another
**                           variable.
** 036 GES 20070404   @32793 Pass through a marker when adding a scope to
**                           the schema dictionary. This allows the new
**                           scope to be associated with the fact of
**                           whether the scope belongs to an internal
**                           procedure, user-defined function or trigger.
** 037 GES 20070418   @33112 Added support for COM-HANDLE types.
** 038 GES 20070528   @33788 Added access to function definition object.
** 039 GES 20070629   @34346 Changed APIs to pass through enough ASTs
**                           to allow the schema dictionary to set the
**                           line and column info for temp-table ASTs
**                           created during parsing.
** 040 GES 20070907   @35010 Support for 10.1B data types.
** 041 GES 20070911   @35138 Support for 10.1B class and package
**                           namespaces.
** 042 GES 20070923   @35184 Added removeKeyword().
** 043 GES 20070927   @35257 Modified promotion behavior to match new
**                           interfaces in the schema dictionary. Added
**                           table promotion when qualified field
**                           references are encountered.
** 044 GES 20071012   @35529 Added separate namespaces for dataset,
**                           data-source and data-relation resources.
** 045 GES 20071107   @35900 Made most schema related methods safe to
**                           use with a null schema dictionary. This
**                           allows the parser to operate on a wide range
**                           of expressions without any knowledge or
**                           dependency upon the schema dictionary. This
**                           is important in cases of preprocessor
**                           expression evaluation, where text may need
**                           to be parsed as a symbol instead of as a
**                           field name (but if the schema dictionary
**                           was loaded, the preprocessor symbol would
**                           resolve as a field which is an invalid
**                           result).
** 046 GES 20071129   @36187 Removed support for DYNAMIC hints (no longer
**                           needed).
** 047 GES 20071207   @36286 Signature changes to pass more data to the
**                           schema dictionary. Also added a global flag
**                           to enable/disable table promotion.
** 048 ECF 20080704   @39145 Minor memory optimization. Use Boolean
**                           constants instead of instantiating new
**                           Boolean objects.
** 049 GES 20090429   @42065 Match package and class name changes.
** 050 GES 20110712          Made failure to find a class abend more gracefully
**                           (and provide more useful data).
** 051 GES 20110811          Added simple list of handles that could be possible
**                           active-x targets. This should probably be a real
**                           scoped dictionary, but for a start it should handle
**                           virtually all cases.
** 052 GES 20110822          Added support for loading built-in 4GL OO classes/
**                           interfaces and .NET classes/interfaces. Added
**                           static support. Added support for define event
**                           when loading a class via AST. Events are treated
**                           like properties and vars (they share the same
**                           namespace). Reworked class loading for prescanning
**                           and to improve performance.
** 053 GES 20111005          Added safety code to avoid class loading issues
**                           when the OO skeleton class directories don't exist.
** 054 CA  20130225          Annotate the fields with the db name.
** 055 CA  20130307          Fixed abbreviated field name disambiguation - the tables which are
**                           weak reference in the current scope have precedence over others.
** 056 CA  20130307          When multiple there are multiple definitions for the same var, only
**                           the first one has its tempidx set and is saved in the dictionary.
** 057 CA  20130307          Fixed honoring of VALIDATE clause for temp-table fields defined using
**                           the LIKE clause and tables defined using the LIKE clause.
** 058 GES 20130317          Added support for the USING clause when loading from an AST.
** 059 CA  20131029          All the static info is kept context-local, for runtime query
**                           conversion support.
** 060 OM  20131205          Fixed addLocalVariableLike() by creating the new variable with exact
**                           type as the variable passed as parameter.
** 061 GES 20140129          Fixed an exists() use case (this is also an optimization).
** 062 GES 20140428          Added tryLoadClass() and differentiated the builtin annotation for
**                           classes as builtin-cls. Fixed OO lookup logic to match changes in
**                           ClassDefinition processing.  Added methods to more properly support
**                           the keyword ignore list.
** 063 GES 20141125          Added caching functions to allow local vars to be restored in
**                           function defs that use the 4GL feature of making their parms list
**                           optional when a forward function def has already specified the
**                           parms.
** 064 VIG 20141104          Added menu support.
** 065 ECF 20151030          Added methods to retrieve schema ASTs for database tables and
**                           fields, and to manage field widget references by frame.
** 066 GES 20160224          Implemented a configuration parameter to allow overriding the
**                           default location for skeleton classes. Complete rewrite of schema
**                           dictionary resolution to handle the dynamic lookup in the class
**                           hierarchy of protected buffers and temp-tables. Added support for
**                           query resources. Queries, datasets and data-sources are now
**                           looked up in the class hierarchy too.
** 067 ECF 20170303          Honor FOR TEMP-TABLE clause in DEFINE BUFFER/PARAMETER statements.
** 068 CA  20170523          Fix parsing of implicit Progress.Lang.Object super-class inheritance.
** 069 GES 20170706          Added utility functions used to generate the class map.
**     GES 20170708          When not in silent mode, will print the class name of any
**                           recursively parsed class file.
**     CA  20170711          1st level OO class scanning is decoupled from full parse: first, all
**                           references from members or inheritance are resolved, for the entire
**                           graph, and only after that the reached classes are parsed.
**     CA  20170721          Remove all referenced classes/interfaces loaded during a 1st level 
**                           scan, if it fails when loading a class. 
**     CA  20170724          When pre-parsing a class, failures to resolve a class starting with
**                           the second or subsequent parsing level will result in 'sink-ing' all
**                           references to the missing class, and not fail.  Failures will be 
**                           emitted only if the currently parsed class has a direct reference to
**                           a missing class.  This is to emulate the 4GL compiler, which doesn't
**                           resolve the entire class reference graph, but only the first level;
**                           the compromise is needed because FWD requires to resolve all class
**                           references (members and types) at parsing time.
**     CA  20170725          When processing the method block during first level parse, try to 
**                           resolve the expected type of the data member from the currently 
**                           processed class first.
**     CA  20170804          processHierarchy first checks the real class hierarchy if 
**                           ignoreOORefs is set.
** 070 OM  20170802          Dropped possible ActiveX detection. Removed redundant casts. 
** 071 CA  20170825          Initial support for USE-DICT-EXPS: not fully complete for the HELP
**                           behaviour.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.uast;

import java.io.*;
import java.util.*;
import java.util.function.*;
import com.goldencode.ast.*;
import com.goldencode.util.*;
import antlr.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.schema.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.*;

/**
 * Contains a dictionary for each type of language namespace and provides
 * methods to maintain and lookup in the respective namespaces.
 * <p>
 * Provides support for the following namespaces:
 * <ul>
 *    <li> language keywords (optionally abbreviated)
 *    <li> global and local variables with multiple arbitrary levels of
 *         scoping
 *    <li> functions (both built-in and user-defined)
 *    <li> procedures
 *    <li> classes
 *    <li> package names
 *    <li> labels
 *    <li> schema names
 *    <ul>
 *       <li> databases
 *       <li> tables - qualified or not, optionally abbreviated
 *       <li> fields - qualified or not, optionally abbreviated
 *    </ul>
 *    <li> widgets (this namespace is artificially separated from the
 *         variable namespace though normally this is not the case in
 *         Progress)
 *    <li> frames 
 *    <li> menus and sub-menus
 *    <li> menu-items
 *    <li> streams
 *    <li> ProDataSet support
 *    <ul> 
 *       <li> datasets
 *       <li> data-sources
 *       <li> data-relations
 *    </ul> 
 *    <li> object-oriented support
 *    <ul> 
 *       <li> packages
 *       <li> class definitions
 *    </ul> 
 * </ul>
 * <p>
 * All namespaces (except language keywords) provide a mapping between a 
 * symbol and an integer token type as defined by the Lexer and Parser.  It
 * is this token type that is returned from most lookup operations.
 * <p>
 * This class is aware of the parser token types and where necessary it
 * converts other values into the proper types for use in the parser.
 * At this time this feature is only needed for schema name lookups as
 * all other types are already normalized properly in the respective 
 * dictionary.  See {@link #lookupField}.
 * <p>
 * Only the variable, widget and label namespaces utilize multiple scoping
 * levels at this time.  All other namespaces are flat.
 */
public class SymbolResolver
implements ProgressParserTokenTypes
{
   /** Possible access modes for object data members. */
   private static final Set<Integer> ACCESS_MODES = new HashSet<>();
   
   /** Available menu children types. */
   private static final Set<Integer> MENU_CHILDREN_TYPES = new HashSet<>();
   
   /** Available sub-menu children types. */
   private static final Set<Integer> SUBMENU_CHILDREN_TYPES = new HashSet<>();
   
   /** Class or interface definition file extension. */
   public static final String CLASS_EXT = ".cls";
   
   /** Class or interface definition file search specification. */
   public static final String CLASS_SPLAT = "*" + CLASS_EXT;
   
   /** Class or interface definition file extension. */
   public static final String CLASS_MAP_FILE = "./cfg/class_map.xml";
   
   /** Default parent of all classes if no explicit parent is specified. */
   public static final String ROOT_OBJ_NAME = "Progress.Lang.Object";
   
   /** Default parent of all classes if no explicit parent is specified. */
   public static final String ROOT_CLASS_NAME = "Progress.Lang.Class";
   
   /** Relative location for the non-project class/interface 4GL code. */
   public static final String SKELETON_PATH = "./p2j/skeleton/";
   
   /** Relative location for the 4GL object-oriented base definitions. */
   public static final String OO4GL_PATH = "oo4gl/";
   
   /** Relative location for the .NET base definitions. */
   public static final String DOTNET_PATH = "dotnet/";
   
   /** Relative location for project-specific .NET definitions. */
   public static final String ASSEMBLY_PATH = "./assemblies/";
   
   /** Relative location for the non-project class/interface 4GL code. */
   public static String skeletonPath = SKELETON_PATH;
   
   /** Relative location for the 4GL object-oriented base definitions. */
   public static String oo4glPath = SKELETON_PATH + OO4GL_PATH;
   
   /** Relative location for the .NET base definitions. */
   public static String dotnetPath = SKELETON_PATH + DOTNET_PATH;
   
   /** Context-local data. */
   private static final ContextLocal<WorkArea> local = new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      {
         return new WorkArea();
      }
   };
   
   /**
    * Mock class to be able to parse OO references in pre-scan mode, without affecting the lookup,
    * as not all classes might be loaded.
    */
   private final ClassDefinition MOCK_CLASS_DEF = 
   new ClassDefinition(null, null, null, false, false, false)
   {
      @Override
      public void cleanupObjectResources()
      {
         // no-op
      }
      
      @Override
      public void annotateMethodCall(String mname, int access, boolean isStatic, Aast node)
      {
         // no-op
      }
      
      @Override
      public String getName()
      {
         return "MOCK_FWD_CLASS";
      };
      
      @Override
      public java.util.Set<String> getInterfaces() 
      {
         return Collections.emptySet();
      };
      
      @Override
      public SchemaDictionary getSchemaDict()
      {
         return null;
      };
      
      @Override
      public void addDataSet(String name, int type, int access, boolean isStatic)
      {
         // no-op
      };
      
      @Override
      public void addDataSource(String name, int type, int access, boolean isStatic) 
      {
         // no-op
      };
      
      @Override
      public void addMethod(String name, int type, int access, boolean isStatic, Aast node, String qname) 
      {
         // no-op
      };
      
      @Override
      public void addQuery(String name, int type, int access, boolean isStatic) 
      {
         // no-op
      };
      
      @Override
      public void addTable(String name, int type, int access, boolean isStatic, SchemaDictionary schemaDict) 
      {
         // no-op
      };
      
      @Override
      public void addVariable(String name, int type, int access, boolean isStatic, String qname)
      {
         // no-op
      }
      
      @Override
      public int lookupDataSet(String name, int access, boolean isStatic)
      {
         return DATA_SET;
      }
      
      @Override
      public int lookupDataSource(String name, int access, boolean isStatic)
      {
         return DATA_SOURCE;
      };
      
      @Override
      public int lookupMethod(String name, int access, boolean isStatic)
      {
         if (!currentCls.isEmpty())
         {
            int type = currentCls.peek().lookupMethod(name, access, isStatic);
            if (type != -1)
            {
               return type;
            }
         }
         
         return METH_CLASS;
      };
      
      @Override
      public int lookupQuery(String name, int access, boolean isStatic)
      {
         return QUERY;
      };
      
      @Override
      public String lookupMethodClassName(String name, int access, boolean isStatic)
      {
         if (!currentCls.isEmpty())
         {
            String clsName = currentCls.peek().lookupMethodClassName(name, access, isStatic);
            if (clsName != null)
            {
               return clsName;
            }
         }

         return "MOCK_FWD_METHOD";
      };
      
      @Override
      public int lookupTable(String name, int access, boolean isStatic)
      {
         return TABLE;
      };
      
      @Override
      public int lookupVariable(String name, int access, boolean isStatic)
      {
         if (!currentCls.isEmpty())
         {
            int type = currentCls.peek().lookupVariable(name, access, isStatic);
            if (type != -1)
            {
               return type;
            }
         }

         return VAR_CLASS;
      };
      
      @Override
      public String lookupVariableClassName(String name, int access, boolean isStatic)
      {
         if (!currentCls.isEmpty())
         {
            String clsName = currentCls.peek().lookupVariableClassName(name, access, isStatic);
            if (clsName != null)
            {
               return clsName;
            }
         }

         return "MOCK_FWD_VAR";
      };
   };
   
   /** Stores keywords and handles abbreviation support. */
   private KeywordDictionary kwDict = null;
   
   /** Stores variables, both local (scoped) and global. */
   private ScopedSymbolDictionary varDict = null;
   
   /** Cache of local variables used to duplicate function signatures. */
   private Map<String, Set<Map.Entry>> varDefCache = null;
   
   /** Stores function names and return types. */
   private ScopedSymbolDictionary funcDict = null;
   
   /** Stores procedure names. */
   private ScopedSymbolDictionary procDict = null;
   
   /** Stores explicit package names. */
   private Map<String, UsingSpec> explPkgDict = null;
   
   /** Stores package search specifications. */
   private Set<UsingSpec> searchPkgDict = null;
   
   /** Stores label names. */
   private ScopedSymbolDictionary labelDict = null;   
   
   /** Stores database object names (databases, tables and fields). */
   private SchemaDictionary schemaDict = null;
   
   /** Stores widget names. */
   private ScopedSymbolDictionary widgetDict = null;
   
   /** Stores frame names. */
   private ScopedSymbolDictionary frameDict = null;
   
   /** Stores sets of database field widget ASTs, organized by their frame names */
   private ScopedSymbolDictionary<Set<Aast>> fieldWidgetDict = null;
   
   /** Stores the state of USE-DICT-EXPS flag for each frame. */
   private ScopedSymbolDictionary<Boolean> useDictExps = null;
   
   /** Stores menu and sub-menu names. */
   private ScopedSymbolDictionary menuDict = null;
   
   /** Stores menu-item names. */
   private ScopedSymbolDictionary menuItemDict = null;
   
   /** Stores stream names. */
   private ScopedSymbolDictionary streamDict = null;
   
   /** Stores query names. */
   private ScopedSymbolDictionary qryDict = null;
   
   /** Stores dataset names. */
   private ScopedSymbolDictionary dsetDict = null;
   
   /** Stores data-source names. */
   private ScopedSymbolDictionary dsrcDict = null;
   
   /** Stores data-relation names. */
   private ScopedSymbolDictionary drelDict = null;
   
   /** Maps keyword token types to the correct attribute and method type. */
   private Map<Integer, Integer> attributes = null;
   
   /** Provides access to ATTR_CLASS or METH_CLASS class names. */
   private Map<Integer, String> attrTypes = null;
   
   /** Current class or interface definition that is being created. */
   private Stack<ClassDefinition> currentCls = null;
   
   /** AST-specific variable definition index workaround. */
   private int tempIdx = 0;
   
   /** Current PROPATH. */
   private String[] propath = null;
   
   /** The legacy system's filesystem case-sensitivity. */
   private boolean caseSens = true;
   
   /** Used for recursive class or interface loading. */
   private AstGenerator generator = null;
   
   /** Flag indicating whether silent mode is required. */
   private boolean silent = false;
   
   /** Flag to globally enable/disable table promotion. */
   private boolean allow = true;
   
   /** Flag indicating OO refs should be ignored as we are in pre-scan mode. */
   private boolean ignoreOORefs = false;

   static
   {
      locate().classDict = new ScopedSymbolDictionary(null, false);
      
      ACCESS_MODES.add(KW_PUBLIC);
      ACCESS_MODES.add(KW_PROTECTD);
      ACCESS_MODES.add(KW_PRIVATE);
      
      MENU_CHILDREN_TYPES.add(KW_NEW);
      MENU_CHILDREN_TYPES.add(KW_SHARED);
      MENU_CHILDREN_TYPES.add(KW_FGCOLOR);
      MENU_CHILDREN_TYPES.add(KW_BGCOLOR);
      MENU_CHILDREN_TYPES.add(KW_DCOLOR);
      MENU_CHILDREN_TYPES.add(KW_PFCOLOR);
      MENU_CHILDREN_TYPES.add(KW_FONT);
      MENU_CHILDREN_TYPES.add(KW_TITLE);
      MENU_CHILDREN_TYPES.add(KW_MENU_BAR);
      MENU_CHILDREN_TYPES.add(KW_MENU_ITM);
      MENU_CHILDREN_TYPES.add(KW_SUB_MENU);
      MENU_CHILDREN_TYPES.add(KW_SKIP);
      MENU_CHILDREN_TYPES.add(KW_RULE);
      
      SUBMENU_CHILDREN_TYPES.add(KW_FGCOLOR);
      SUBMENU_CHILDREN_TYPES.add(KW_BGCOLOR);
      SUBMENU_CHILDREN_TYPES.add(KW_DCOLOR);
      SUBMENU_CHILDREN_TYPES.add(KW_PFCOLOR);
      SUBMENU_CHILDREN_TYPES.add(KW_FONT);
      SUBMENU_CHILDREN_TYPES.add(KW_MENU_ITM);
      SUBMENU_CHILDREN_TYPES.add(KW_SUB_MENU);
      SUBMENU_CHILDREN_TYPES.add(KW_SKIP);
      SUBMENU_CHILDREN_TYPES.add(KW_RULE);
   }
   
   /**
    * Default constructor that creates an instance with schema lookups
    * <b>disabled</b>.  Each namespace is initialized.
    */
   public SymbolResolver()
   {
      this(false, true, null, true);
   }
   
   /**
    * Constructs an instance with control over disabling schema lookups.
    * This allows an instance to be created in the absence of any schema
    * support.  Each namespace is initialized.
    *
    * @param    schema
    *           <code>true</code> enables schema dictionary support.
    */
   public SymbolResolver(boolean schema)
   {
      this(schema, true, null, true);
   }
   
   /**
    * Constructs an instance with control over disabling schema lookups. This allows an instance
    * to be created in the absence of any schema support.  Each namespace is initialized.
    *
    * @param    schema
    *           <code>true</code> enables schema dictionary support.
    * @param    caseSens
    *           <code>true</code> if the legacy system had a case-sensitive file-system.
    * @param    propath
    *           The current <code>PROPATH</code> to honor or <code>null</code> for the default
    *           of ".".
    * @param    silent
    *           Flag indicating if <code>Sytsem.out</code> output should be suppressed.
    */
   public SymbolResolver(boolean schema, boolean caseSens, String[] propath, boolean silent)
   {
      kwDict          = new KeywordDictionary();
      varDict         = new ScopedSymbolDictionary<>(null, false);
      funcDict        = new ScopedSymbolDictionary<>(null, false);
      varDefCache     = new HashMap<>();
      procDict        = new ScopedSymbolDictionary<>(null, false);
      explPkgDict     = new LinkedHashMap<>();
      searchPkgDict   = new LinkedHashSet<>();
      labelDict       = new ScopedSymbolDictionary<>(null, false);
      widgetDict      = new ScopedSymbolDictionary<>(null, false);
      frameDict       = new ScopedSymbolDictionary<>(null, false);
      fieldWidgetDict = new ScopedSymbolDictionary<>(null, false);
      useDictExps     = new ScopedSymbolDictionary<>(null, false);
      menuDict        = new ScopedSymbolDictionary<>(null, false);
      menuItemDict    = new ScopedSymbolDictionary<>(null, false);
      streamDict      = new ScopedSymbolDictionary<>(null, false);
      qryDict         = new ScopedSymbolDictionary<>(null, false);
      dsetDict        = new ScopedSymbolDictionary<>(null, false);
      dsrcDict        = new ScopedSymbolDictionary<>(null, false);
      drelDict        = new ScopedSymbolDictionary<>(null, false);
      attributes      = new HashMap<>();
      attrTypes       = new HashMap<>();
      currentCls      = new Stack<>();
      
      if (schema)
      {
         // this load is dependent upon a configuration file being available
         try
         {
            schemaDict = new SchemaDictionary();
            
            // add a default scope that represents the scope of the external 
            // procedure
            addSchemaScope(false);
         }
         catch (SchemaException ex)
         {
            // in this case, the schema dictionary is invalid, however all
            // other types of lookups will work fine --> if there is no
            // schema name processing, then consuming this exception is safe
            // if there is schema name processing then get ready for a
            // null pointer exception!
            ex.printStackTrace();
         }
      }
      
      setPropath(propath);
      this.caseSens = caseSens;
      this.silent   = silent;
      initPossibleClasses(this.propath, this.caseSens);
   }
   
   /**
    * Override the default skeleton path if it is configured.
    */
   public static synchronized void initSkeletonPath()
   {
      skeletonPath = Configuration.getParameter("oo-skeleton-path");
      
      // only rework these from the defaults if a skeleton path is overridden
      if (skeletonPath != null)
      {
         if (!skeletonPath.endsWith(File.separator))
         {
            skeletonPath = skeletonPath + File.separator;
         }
         
         oo4glPath = skeletonPath + OO4GL_PATH;
         dotnetPath = skeletonPath + DOTNET_PATH;
      }
   }
   
   /**
    * Obtain the list of all "fakeout" class and interface definitions that
    * exist in the project.  These <code>.cls</code> files represent the
    * Progress built-ins and all .NET classes/interfaces that can be accessed.
    * <p>
    * These files will be listed in the directories represented by the
    * {@link #OO4GL_PATH}, {@link #DOTNET_PATH} and the {@link #ASSEMBLY_PATH}.
    *
    * @return   The list of <code>.cls</code> files, in the same case in which
    *           they exist in the filesystem.
    */
   public static String[] listFakeoutFiles()
   {
      String[] oo4gl    = getClassFileList(oo4glPath, false);
      String[] dotnet   = getClassFileList(dotnetPath, false);
      String[] assembly = getClassFileList(ASSEMBLY_PATH, false);
      
      ArrayList<String> list = new ArrayList<>();
      
      list.addAll(Arrays.asList(oo4gl));
      list.addAll(Arrays.asList(dotnet));
      list.addAll(Arrays.asList(assembly));
      
      return list.toArray(new String[0]);
   }
   
   /**
    * List the 4GL class/interface files in a given path.
    *
    * @param    path
    *           The directory to search.
    * @param    caseSens
    *           <code>true</code> if the legacy system had a case-sensitive
    *           file-system.
    *
    * @return   The list of files that exist.  If there are no files, the
    *           array will have a 0 length.
    */
   public static String[] getClassFileList(String path, boolean caseSens)
   {
      File dir = new File(path);
      
      if (!dir.exists() || !dir.isDirectory())
      {
         return new String[0];
      }
      
      FileList fl = new FileSpecList(dir, CLASS_SPLAT, true, caseSens);
      
      return fl.listFilenames();
   }
   
   /**
    * Store each element of the list as a key in the map and create a filename
    * as the mapped value (using the path + the key). The keys will be
    * lowercased if case-insensitive matching is configured.
    *
    * @param    map
    *           The map to modify.
    * @param    list
    *           The elements to add.
    * @param    path
    *           The root path in which all elements reside.
    * @param    caseSens
    *           <code>true</code> if the legacy system had a case-sensitive
    *           file-system.
    */
   public static void createClassMappings(Map<String, String> map,
                                          String[]            list,
                                          String              path,
                                          boolean             caseSens)
   {
      int begin = path.length();
      
      for (int i = 0; i < list.length; i++)
      {
         int end = list[i].length() - 4;
         String entry = list[i].substring(begin, end);                    
         map.put(caseSens ? entry : entry.toLowerCase(), list[i]);
      }
   }
   
   /**
    * Get the context-local {@link WorkArea} instance.
    * 
    * @return   See above.
    */
   private static WorkArea locate()
   {
      return local.get();
   }
   
   /**
    * Enables schema dictionary lookups when they are disabled at construction
    * and provides external control or sourcing over the schema dictionary.
    *
    * @param    schemaDict
    *           The schema dictionary to use.
    */
   public void setSchemaDictionary(SchemaDictionary schemaDict)
   {
      this.schemaDict = schemaDict;
   }
   
   /**
    * Set the {@link #ignoreOORefs} flag.
    * 
    * @param    state
    *           The state of ignore OO refs mode.
    */
   public void setIgnoreOORefs(boolean state)
   {
      this.ignoreOORefs = state;
   }
   
   /**
    * Gets the <code>PROPATH</code>.
    *
    * @return   The current <code>PROPATH</code> to honor.
    */
   public String[] getPropath()
   {
      return propath;
   }
   
   /**
    * Sets the <code>PROPATH</code>.
    *
    * @param    propath
    *           The current <code>PROPATH</code> to honor or <code>null</code>
    *           for the default of ".".
    */
   public void setPropath(String[] propath)
   {
      this.propath = (propath == null) ? new String[] { "." } : propath;
   }
   
   /**
    * Adds a <code>Keyword</code> to the keyword dictionary.
    *
    * @param   word
    *          <code>Keyword</code> to add.
    */
   public void addKeyword(Keyword word)
   {
      kwDict.addKeyword(word);
   }
   
   /**
    * Removes a <code>Keyword</code> from the keyword dictionary.
    *
    * @param   word
    *          <code>Keyword</code> to remove.
    */
   public void removeKeyword(Keyword word)
   {
      kwDict.deleteKeyword(word);
   }
   
   /**
    * Searches for a <code>Keyword</code> in the keyword dictionary, based
    * on a specific key.
    *
    * @param   key
    *          Text string to match.
    * @return  The <code>Keyword</code> that was found or null if no match
    *          was found.
    */   
   public Keyword lookupKeyword(String key)
   {
      return kwDict.lookupKeyword(key);
   }
   
   /**
    * Increases the scoping level of the variable and widget dictionaries by
    * 1 level.  All symbols must be unique within a scoping level.
    */   
   public void addScope()
   {
      varDict.addScope(null);
      widgetDict.addScope(null);
      fieldWidgetDict.addScope(null);
      useDictExps.addScope(null);
   }
   
   /**
    * Decreases the scoping level of the variable and widget dictionaries by
    * 1 level.  All symbols previously defined in the top-most scoping level
    * are lost and will no longer be found during lookups.
    */      
   public void deleteScope()
   {
      varDict.deleteScope();
      widgetDict.deleteScope();
      fieldWidgetDict.deleteScope();
      useDictExps.deleteScope();
   }
   
   /**
    * Persist schema dictionary information which is specific to the source
    * file currently being parsed.
    *
    * @param    srcFile
    *           Name of the 4GL source code file containing the schema
    *           definitions being persisted.
    * @param    schemaFile
    *           Name of schema persistence file.
    *
    * @throws   AstException
    *           if any error occurs writing persistence file.
    */
   public void persistSchemaData(String srcFile, String schemaFile)
   throws AstException
   {
      if (schemaDict != null)
      {
         schemaDict.persist(srcFile, schemaFile);
      }
   }
   
   /**
    * Adds a variable name and associated token type to the appropriate scope
    * of the variable dictionary, based on the content of the passed in
    * <code>Variable</code> wrapper object.
    *
    * @param    global
    *           Add to global scope if <code>true</code>.
    * @param    var
    *           Variable to add.
    */
   public void addVariable(boolean global, Variable var)
   {
      addWrappedWorker(varDict, global, var.getName(), var);
   }
   
   /**
    * Adds a widget name and associated token type to the appropriate scope
    * of the widget dictionary, based on the content of the passed in
    * <code>Variable</code> wrapper object.
    *
    * @param    global
    *           Add to global scope if <code>true</code>.
    * @param    wid
    *           Widget as represented in a <code>Variable</code> object.
    */
   public void addWidget(boolean global, Variable wid)
   {
      addWrappedWorker(widgetDict, global, wid.getName(), wid);
   }
   
   /**
    * Helper to load variable, widget and frame dictionaries based on a list
    * of <code>Variable</code> objects. This processing will load all 3
    * dictionaries from the same list, however it must be noted that adding
    * a variable does NOT result in a matching widget of the same name in
    * the widget dictionary (even though this is essentially what happens
    * when the parser adds new variables to the variable dictionary). To 
    * acheive this same effect, one must add multiple <code>Variable</code>
    * objects to the list as needed.
    * <p>
    * A field can be added to this lookup by using a field token type instead
    * of a variable token type. This will allow an ambiguous field reference
    * to be made unambiguous since variable lookups are processed
    * <b>before</b> any schema dictionary lookups.
    *
    * @param    varlist
    *           List of variables, widgets, frames or fields to add to the
    *           appropriate dictionary.
    */
   public void loadVariableSymbols(Variable[] varlist)
   {
      for ( int j = 0; j < varlist.length; j++ )
      {
         int type = varlist[j].getTokenType();
         
         if (
              (type >= ProgressParser.BEGIN_VARTYPES   && type <= ProgressParser.END_VARTYPES) ||
              (type >= ProgressParser.BEGIN_FIELDTYPES && type <= ProgressParser.END_FIELDTYPES)
            )
         {
            addVariable(true, varlist[j]);
         }
         else if (type == ProgressParser.WID_FRAME)
         {
            addFrame(varlist[j].getName(), ProgressParser.WID_FRAME);
         }         
         else if (type >= ProgressParser.BEGIN_WIDGETS &&
                  type <= ProgressParser.END_WIDGETS )
         {
            addWidget(true, varlist[j]);
         }
      }
   }
   
   /**
    * Adds a variable name and associated token type to the global scope
    * of the variable dictionary.  Names in this scope can be hidden by
    * duplicate names added to scopes that are closer to the top of the
    * stack of scopes.
    *
    * @param    name
    *           Variable name to add.
    * @param    tokenType
    *           Lexer and Parser token type associated with the name.
    */
   public void addGlobalVariable(String name, int tokenType)
   {
      addGlobalVariable(name, tokenType, null);
   }
   
   /**
    * Adds a variable name and associated token type to the global scope
    * of the variable dictionary.  Names in this scope can be hidden by
    * duplicate names added to scopes that are closer to the top of the
    * stack of scopes.
    *
    * @param    name
    *           Variable name to add.
    * @param    tokenType
    *           Lexer and Parser token type associated with the name.
    * @param    cls
    *           Fully qualified class name of the object instance represented
    *           or <code>null</code> if this variable does not represent an
    *           object instance.
    */
   public void addGlobalVariable(String name, int tokenType, String cls)
   {
      addWrappedWorker(varDict, true, name, new Variable(name, tokenType, cls));
   }
   
   /**
    * Removes any global variable associated with this <code>Keyword</code> from the variable
    * dictionary.  This is only safe to call BEFORE real parsing begins.
    *
    * @param   word
    *          <code>Keyword</code> to remove (if it exists).
    */
   public void removeGlobalVariable(Keyword word)
   {
      varDict.deleteSymbol(word.getFullText());
   }
   
   /**
    * Adds a variable name and associated token type to the current scope
    * of the variable dictionary.  This is the same as the top-most scope.
    * Names in this scope can hide duplicate names added to scopes that are
    * lower on the stack of scopes.
    *
    * @param    name
    *           Variable name to add.
    * @param    tokenType
    *           Lexer and Parser token type associated with the name. 
    * @param    cls
    *           Fully qualified class name of the object instance represented
    *           or <code>null</code> if this variable does not represent an
    *           object instance.
    */
   public void addLocalVariable(String name, int tokenType, String cls)
   {
      if (varDict.getSymbolAtScope(name, 0) == null)
      {
         addWrappedWorker(varDict, false, name, new Variable(name, tokenType, cls));
      }
   }
   
   /**
    * Adds a menu name and associated token type/options to the menu dictionary 
    * based on a referenced variable in a <code>LIKE</code> clause.
    *
    * @param    name
    *           Menu name to add.
    * @param    ast
    *           DEFINE_MENU or DEFINE_SUB_MENU AST node.
    */
   public void addMenuLike(String name, String oldname, Aast ast)
   {
      if (ast == null)
      {
         throw new NullPointerException("Root DEFINE_MENU or DEFINE_SUB_MENU statement is null");
      }
      
      int type = lookupMenu(oldname);
      Variable var   = new Variable(name, type, null);
      
      Variable like  = (Variable) lookupWrapped(menuDict, oldname);
      Aast     fopts = like.getOptions(0);
      
      if (fopts != null)
      {
         var.copyDefaultOptions(fopts);
         
         Collection<Integer> types;
         if (ast.getType() == DEFINE_MENU)
         {
            types = MENU_CHILDREN_TYPES;
         }
         else if (ast.getType() == DEFINE_SUB_MENU)
         {
            types = SUBMENU_CHILDREN_TYPES;
         }
         else
         {
            throw new IllegalArgumentException("Unsupported root ast type");
         }
         
         // copy children asts.
         for (Integer astType : types)
         {
            Variable.overrideIfNotPresent(fopts, ast, astType);
         }
         
         // copy annotations of root ast.
         for (Iterator<String> it = fopts.annotationKeys(); it.hasNext(); )
         {
            String key = it.next();
            if (ast.getAnnotation(key) == null)
            {
               AnnotatedAst anast = (AnnotatedAst)ast;
               Object ann = fopts.getAnnotation(key);
               if (ann instanceof String)
                  anast.putAnnotation(key, (String)ann);
               if (ann instanceof Double)
                  anast.putAnnotation(key, (Double)ann);
               if (ann instanceof Long)
                  anast.putAnnotation(key, (Long)ann);
               if (ann instanceof Boolean)
                  anast.putAnnotation(key, (Boolean)ann);
            }
         }
      }
      
      if (menuDict.getSymbolAtScope(name, 0) == null)
      {
         addWrappedWorker(menuDict, false, name, var);
      }
   }
   
   /**
    * Adds a variable name and associated token type/options to the current
    * scope of the variable dictionary based on a referenced variable or
    * schema field in a <code>LIKE</code> clause.  This is the same as the
    * top-most scope.  Names in this scope can hide duplicate names added to
    * scopes that are lower on the stack of scopes.
    *
    * @param    name
    *           Variable name to add.
    * @param    ast
    *           AST node containing the like clause.
    */
   public void addLocalVariableLike(String name, Aast ast)
   { 
      int     ltype   = ast.getType();
      int     type    = ltype;
      String  oldname = ast.getText();
      int     extent  = getExtentSubscript(ast);
      boolean isFld   = ltype > BEGIN_FIELDTYPES && ltype < END_FIELDTYPES;
      String  cls     = null;
      
      // the local variable type must be converted to the proper
      // VAR_ variant *if* this was a like FIELD_...
      switch (ltype)
      {
         case FIELD_CHAR:
            type = VAR_CHAR;
            break;
         case FIELD_CLASS:
            type = VAR_CLASS;
            // class fields must propagate the fully qualified class name
            cls  = (String) ast.getAnnotation("qualified");
            break;
         case FIELD_COM_HANDLE:
            type = VAR_COM_HANDLE;
            break;
         case FIELD_DATE:
            type = VAR_DATE;
            break;
         case FIELD_DATETIME:
            type = VAR_DATETIME;
            break;
         case FIELD_DATETIME_TZ:
            type = VAR_DATETIME_TZ;
            break;
         case FIELD_DEC:
            type = VAR_DEC;
            break;
         case FIELD_HANDLE:
            type = VAR_HANDLE;
            break;
         case FIELD_INT:
            type = VAR_INT;
            break;
         case FIELD_INT64:
            type = VAR_INT64;
            break;
         case FIELD_LOGICAL:
            type = VAR_LOGICAL;
            break;
         case FIELD_RAW:
            type = VAR_RAW;
            break;
         case FIELD_RECID:
            type = VAR_RECID;
            break;
         case FIELD_ROWID:
            type = VAR_ROWID;
            break;
         default:
            // this MUST be at the end of the list of fields, it will match
            // BLOB and CLOB which can't be variable types
            if (isFld)
            {
               String err = String.format("Unexpected field type %s!",
                                          ast.getDescriptiveTokenText());
               throw new RuntimeException(err);
            }
            else
            {
               // TODO: OM: is this correct when creating LIKE variable (copy its true type)?
               type = lookupVariable(oldname);
            }
            break;
      }
      
      // class variables must propagate the fully qualified class name 
      if (ltype == VAR_CLASS)
         cls = (String) ast.getAnnotation("qualified");
      
      Variable var   = new Variable(name, type, cls);
      Aast     fopts = null;
      
      if (isFld)
      {
         // we must pattern the new variable off a referenced field
         // obtain the field options AST from the schema
         try
         {
            SchemaHelper<Aast> fcopy = (SchemaDictionary dict) ->
            {
               Aast fld = null;
               
               // getFieldCopy() is not safe to call if the field doesn't exist
               if (dict.isField(oldname))
               {
                  fld = dict.getFieldCopy(oldname, extent);
               }
               
               return fld;
            };
               
            Predicate<Aast> notNull = (Aast val) -> { return (val == null); };
               
            fopts = processHierarchy(fcopy, notNull);
         }
         catch (SchemaException exc)
         {
            exc.printStackTrace();
         }
      }
      else
      {
         // we must pattern the variable off another variable
         Variable like = (Variable) lookupWrapped(varDict, oldname);
         fopts = like.getOptions(extent);
      }
      
      if (fopts != null)
      {
         var.copyDefaultOptions(fopts);
      }
      
      if (varDict.getSymbolAtScope(name, 0) == null)
      {
         addWrappedWorker(varDict, false, name, var);
      }
   }
   
   /**
    * Save off the variable dictionary definitions for the current scope into a cache.
    *
    * @param    name
    *           The name by which the saved definitions can be restored.
    */
   public void saveLocalVariables(String name)
   {
      varDefCache.put(name.toUpperCase(), varDict.entrySet(0));
   }
   
   /**
    * Restore any previously cached variable dictionary definitions into the current scope.
    * <p>
    * Any cached definitions will be removed from the cache.
    *
    * @param    name
    *           The name by which the saved definitions can be located.
    *
    * @return   <code>true</code> if there were variable definitions restored.
    */
   public boolean restoreLocalVariables(String name)
   {
      boolean nonEmpty = false;
      
      Set<Map.Entry> locals = varDefCache.remove(name.toUpperCase());
      
      if (locals != null)
      {
         Iterator<Map.Entry> iter = locals.iterator();
         
         while (iter.hasNext())
         {
            Map.Entry entry = iter.next();
            varDict.addEntry(false, entry.getKey(), entry.getValue());
            nonEmpty = true;
         }
      }
      
      return nonEmpty;
   }
   
   /**
    * Remove any previously cached variable dictionary definitions without restoring them
    * to the current scope.
    *
    * @param    name
    *           The name by which the saved definitions can be located.
    */
   public void removeLocalVariables(String name)
   {
      varDefCache.remove(name.toUpperCase());
   }
   
   /**
    * Inspect the variable definition and write any non-default options into 
    * the variable object.
    *
    * @param    name
    *           The variable name which must have its options set.
    * @param    ast
    *           The tree to inspect which defines a Progress variable.
    */
   public void setVariableOptions(String name, Aast ast)
   {
      Variable var = (Variable) lookupWrapped(name, varDict, widgetDict, menuDict, menuItemDict);
      
      if (var == null || ast == null)
      {
         // nothing to do
         return;
      }
      
      // the KW_LIKE should already have been processed by now (in the 
      // addLocalVariableLike() so that defaults are set from that       
      // variable/field and then can be overridden by explicit options      
      // so now we process the explicit overrides
      
      var.setOptions(ast);
   }
   
   /**
    * Stores all non-standard options as annotations to the AST passed as
    * a parameter if this is the original definition, otherwise it stores
    * a subset of values, most important of which is the cross-reference
    * index to allow the original definition to be easily identified and
    * accessed from a reference.  If an option has the default value, it is
    * not saved as an annotation.
    *
    * @param    name
    *           The variable name which must have its options annotated.
    * @param    ast
    *           The AST to annotate.    
    * @param    original
    *           <code>true</code> if this is to annotate the AST of the
    *           original definition, <code>false</code> if the AST is only
    *           a reference.
    */
   public void annotateVariableOptions(String  name,
                                       Aast    ast,
                                       boolean original)
   {
      Variable var = (Variable) lookupWrapped(name, varDict, widgetDict, menuDict, menuItemDict);
      
      // if called from the ProgressExpressionEvaluator (to evaluate an
      // expression) the normal variable dictionary processing is bypassed
      // we annotate
      if (var != null)
         var.annotateOptions(ast, original, this);
   }
   
   /**
    * Creates a new field AST node based on a qualified or unqualified
    * field name (must be unambiguous) and sets the token type, text and
    * all annotations as needed.  This centralizes a service which is
    * normally handled by the parser directly.  It uses the
    * {@link #annotateField} method for the core of the lookup and
    * annotation logic.
    * <p>
    * No table promotion will occur based on this field name, even if it
    * is a qualified name.
    *
    * @param    name
    *           The field name from which to obtain options and type.
    *
    * @return   The created AST node.
    */
   public Aast createField(String name)
   {
      ProgressAst field = new ProgressAst();
      
      field.setText(name);
      annotateField(field, null, true, false);
      
      return field;
   }
   
   /**
    * Stores common data (schemaname, bufname, recordtype) as well as
    * all non-standard options as annotations to the AST passed as
    * a parameter.  If an option has the default value, it is not saved as
    * an annotation.  Optionally sets the type of the given AST based on
    * the field found in the schema dictionary.  The field will be looked
    * up in the dictionary based on the text of the AST node and there
    * must be no ambiguity.
    * <p>
    * If the field cannot be found due to ambiguity (more than 1 match), the
    * method will return with no resulting changes.
    * <p>
    * If no such field exists, some default annotations will be set to
    * "unknown".
    *
    * @param    field
    *           The AST to annotate.
    * @param    table
    *           If the field's table is known, this is the name-node for that table.
    *           Passing this will narrow the lookup just to that table. Pass <code>null</code>
    *           to do a normal schema namespace lookup (with all the ambiguity and other
    *           problems that could occur).
    * @param    setType
    *           If <code>true</code>, the AST node will have its type set
    *           based on the field found in the schema dictionary.
    * @param    promote
    *           <code>true</code> to force the promotion (in the schema
    *           dictionary) of the table if the field's name is qualified.
    */
   public void annotateField(Aast field, Object table, boolean setType, boolean promote)
   {
      String    ftext = field.getText();
      FieldInfo finfo = lookupFieldInfo((NameNode) table, ftext);
      Aast      fback = finfo.getAst();
      
      // make sure we found the backing field
      if (fback != null)
      {
         if (setType)
         {
            field.setType(fback.getType());
         }
         
         // now store an annotation with the fully qualified name and
         // the shorted unabbreviated unique buffer name
         String fullname = finfo.getQualified();
         String bufname  = finfo.getBuffer();
         String dbname   = finfo.getDatabase();
         
         // was the original reference qualified?
         if (promote && ftext.indexOf('.') != -1)
         {
            // qualified references cause table promotion
            promoteTableName(bufname, false, false);
         }
         
         if (fullname != null)
            field.putAnnotation("schemaname", fullname);
         
         if (bufname != null)
            field.putAnnotation("bufname", bufname);
         
         if (dbname != null)
            field.putAnnotation("dbname", dbname);
         
         // store the backing construct (this will be TABLE, TEMP_TABLE or WORK_TABLE)
         int backing = finfo.getRecordType();
         
         if (backing != -1)
            field.putAnnotation("recordtype", new Long(backing));
         
         int    newtype = field.getType();
         String cls     = (newtype == FIELD_CLASS) ? ROOT_OBJ_NAME : null;
         
         // temporary variable object to be our worker to annotate
         Variable var = new Variable(ftext, newtype, cls);
         
         // we must pattern the new variable off a referenced field
         var.copyDefaultOptions(fback);
         var.annotateOptions(field, true, this);
      }
      else
      {
         // if the FIELD_ type was set via a hint or some other
         // unexpected manner, then this will get triggered...
         field.putAnnotation("schemaname", "unknown field");
         field.putAnnotation("bufname", "unknown buffer");
         field.putAnnotation("recordtype", new Long(-1));
      }
   }
   
   /**
    * Searches the variable dictionary for a symbol matching the passed 
    * key or matching a keyword's full text if the symbol matches a language
    * keyword.  This lookup starts at the top-most scoping level and descends
    * through each scope level on the stack until a match is found. The
    * token type of the first match found is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param   actual
    *          Actual text string to match which may be a language keyword
    *          abbreviation.
    *
    * @return  The token type that was found or -1 if no match was found.
    */    
   public int lookupVariable(String actual)
   {
      return lookupWrappedType(lookupWrapped(varDict, actual));
   }
   
   /**
    * Searches a dictionary containing {@link TokenDataWrapper} objects for a
    * symbol matching the passed key or matching a keyword's full text if the
    * symbol matches a language keyword.  This lookup starts at the top-most
    * scoping level and descends through each scope level on the stack until
    * a match is found. The wrapped object of the first match found is
    * returned.
    *
    * @param   ssd
    *          The dictionary to search.
    * @param   actual
    *          Actual text string to match which may be a language keyword
    *          abbreviation.
    *
    * @return  The wrapped object that was found or <code>null</code> if no
    *          match was found.
    */
   public TokenDataWrapper lookupWrapped(ScopedSymbolDictionary ssd, String actual)
   {
      Keyword          kw      = lookupKeyword(actual);
      TokenDataWrapper wrapper = null;
      
      if (kw != null)
      {
         // the actual text matches a keyword
         if (kw.isReserved())
         {
            // since this is a reserved keyword, this must be a 
            // built-in variable/function
            wrapper = (TokenDataWrapper) ssd.lookupSymbol(kw.getFullText());
         }
         else
         {
            // this is an unreserved keyword, this MAY be a 
            // built-in variable/function BUT we must allow a user-defined
            // name to take precedence (note that if the actual text is
            // is the same as the full keyword text, then we may still
            // get a match here that is the built-in, assuming no
            // user-defined name exists --> that is OK too)
            wrapper = (TokenDataWrapper) ssd.lookupSymbol(actual);
            
            // check if there was a match to the actual text
            if (wrapper == null)
            {
               // match the full text of the keyword with the built-in
               wrapper = (TokenDataWrapper) ssd.lookupSymbol(kw.getFullText());
            }
         }
      }
      else
      {
         // there is no keyword, so this must be a user-defined name
         wrapper = (TokenDataWrapper) ssd.lookupSymbol(actual);
      }
      
      return wrapper;
   }
   
   /**
    * Searches a dictionaries containing {@link TokenDataWrapper} objects for a
    * symbol matching the passed key or matching a keyword's full text if the
    * symbol matches a language keyword.  This lookup starts at the top-most
    * scoping level and descends through each scope level on the stack until
    * a match is found. The wrapped object of the first match found is
    * returned.
    *
    * @param   actual
    *          Actual text string to match which may be a language keyword
    *          abbreviation.
    * @param   ssds
    *          An array of dictionaries to search.
    *
    * @return  The wrapped object that was found or <code>null</code> if no
    *          match was found.
    */    
   public TokenDataWrapper lookupWrapped(String actual, ScopedSymbolDictionary<?> ... ssds)
   {
      for (ScopedSymbolDictionary<?> ssd : ssds)
      {
         TokenDataWrapper wrapped = lookupWrapped(ssd, actual);
         
         if (wrapped != null)
         {
            return wrapped;
         }
      }
      
      return null;
   }
   
   /**
    * Adds a user-defined function and its return type to the dictionary
    * of functions. This is a flat dictionary (there is no scoping).  The 
    * function name is used as the key and a wrapped {@link Function} object
    * is the value.  Since the dictionary is flat, all function names must be
    * unique.
    * <p>
    * Due to the Progress forward definition capability, this function only
    * adds to the dictionary the first time this method is called for a
    * given name.
    *
    * @param    name
    *           Function name.
    * @param    tokenType
    *           The token type of the function's return value as defined by
    *           the Lexer and Parser.
    * @param    qname
    *           Fully qualified class name for functions that return an
    *           object instance, otherwise <code>null</code>.
    */
   public void addFunction(String name, int tokenType, String qname)
   {
      // a forward declaration may have already occurred for this function
      // try a lookup
      Function func = (Function) funcDict.lookupSymbol(name);
      
      // is this a new function def OR an override of a builtin function?
      if (func == null || func.isBuiltin())
      {
         func = new Function(name, tokenType, false, qname);
         
         addWrappedWorker(funcDict, false, name, func);
      }
   }
   
   /**
    * Adds a built-in Progress function and its return type to the dictionary
    * of functions. This is a flat dictionary (there is no scoping).  The 
    * function name is used as the key and a wrapped {@link Function} object
    * is the value.  Since the dictionary is flat, all function names must be
    * unique.
    *
    * @param    name
    *           Function name.
    * @param    tokenType
    *           The token type of the function's return value as defined by
    *           the Lexer and Parser.
    * @param    returnsUnknown
    *           Specifies if this built-in function can possibly return
    *           the unknown value.
    * @param    qname
    *           Fully qualified class name for functions that return an
    *           object instance, otherwise <code>null</code>.
    */
   public void addBuiltinFunction(String  name,
                                  int     tokenType,
                                  boolean returnsUnknown,
                                  String  qname)
   {
      Function func = new Function(name, tokenType, true, qname);
      func.setReturnsUnknown(returnsUnknown ? Boolean.TRUE : Boolean.FALSE);
      
      addWrappedWorker(funcDict, false, name, func);
   }
   
   /**
    * Adds a built-in Progress function and its return type to the dictionary
    * of functions. This is a flat dictionary (there is no scoping).  The 
    * function name is used as the key and a wrapped {@link Function} object
    * is the value.  Since the dictionary is flat, all function names must be
    * unique.
    *
    * @param   name
    *          Function name.
    * @param   tokenType
    *          The token type of the function's return value as defined by
    *          the Lexer and Parser.
    * @param   returnsUnknown
    *          Specifies if this built-in function can possibly return
    *          the unknown value.
    */
   public void addBuiltinFunction(String  name,
                                  int     tokenType,
                                  boolean returnsUnknown)
   {
      addBuiltinFunction(name, tokenType, returnsUnknown, null);
   }

   /**
    * Searches the function dictionary for a symbol matching the passed 
    * key or matching a keyword's full text if the symbol matches a language
    * keyword.  This lookup starts at the top-most scoping level and descends
    * through each scope level on the stack until a match is found. The
    * token type of the first match found is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param    actual
    *           Actual text string to match which may be a language keyword
    *           abbreviation.
    *
    * @return   The token type that was found or -1 if no match was found.
    */
   public int lookupFunction(String actual)
   {
      return lookupWrappedType(lookupWrapped(funcDict, actual));
   }
   
   /**
    * Searches the function dictionary for a symbol matching the passed 
    * key or matching a keyword's full text if the symbol matches a language
    * keyword.  This lookup starts at the top-most scoping level and descends
    * through each scope level on the stack until a match is found. The
    * definition of the first match found is returned.
    *
    * @param    actual
    *           Actual text string to match which may be a language keyword
    *           abbreviation.
    *
    * @return   The function definition or <code>null</code> if no match was
    *           found.
    */
   public Function lookupFunctionDef(String actual)
   {
      return (Function) lookupWrapped(funcDict, actual);
   }
   
   /**
    * Removes any function associated with this <code>Keyword</code> from the builtin
    * function dictionary.  This is only safe to call BEFORE real parsing begins.
    *
    * @param   word
    *          <code>Keyword</code> to remove (if it exists).
    */
   public void removeBuiltinFunction(Keyword word)
   {
      funcDict.deleteSymbol(word.getFullText());
   }
   
   /**
    * Stores all non-standard options as annotations to the AST passed as
    * a parameter if this is the original definition, otherwise it stores
    * a subset of values, most important of which is the cross-reference
    * index to allow the original definition to be easily identified and
    * accessed from a reference.  If an option has the default value, it is
    * not saved as an annotation.
    *
    * @param    name
    *           The function name which must have its options annotated.
    * @param    ast
    *           The AST to annotate.
    * @param    original
    *           <code>true</code> if this is to annotate the AST of the
    *           original definition, <code>false</code> if the AST is only
    *           a reference.
    */
   public void annotateFunction(String name, Aast ast, boolean original)
   {
      Function func = (Function) lookupWrapped(funcDict, name);
      
      // special case: preprocessor expression evaluation can resolve to
      // function token types without any function being registered since
      // the CallbackResolver registers the function token types inside
      // the KeywordDictionary; this means that a null pointer can occur
      // here, so we just return without doing any annotations in that
      // case
      if (func == null)
         return;
      
      func.annotateOptions(ast, original, this);
   }
   
   /**
    * Adds a procedure and a token type to the dictionary of procedures.
    * This is a flat dictionary (there is no scoping).  The procedure name
    * is used as the key and the token type is the value.  Since the
    * dictionary is flat, all procedure names must be unique.
    *
    * @param   name
    *          Procedure name.
    * @param   tokenType
    *          The procedure's token type as defined by the Lexer and Parser.
    */   
   public void addProcedure(String name, int tokenType) 
   {
      procDict.addSymbol(false, name, new Integer(tokenType));
   }
   
   /**
    * Searches the procedure dictionary for a symbol matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param   name
    *          Text string to match.
    * @return  The procedure's token type that was found or -1 if no match
    *          was found.
    */
   public int lookupProcedure(String name)
   {
      return lookupWorkerExact(procDict, name);
   }
   
   /**
    * Adds a class name and a token type to the dictionary of classes.
    * This is a flat dictionary (there is no scoping).  The class name
    * is used as the key and the token type is the value.  Since the
    * dictionary is flat, all class names must be unique.
    *
    * @param    symbol
    *           Node defining the fully qualified (including package) class
    *           name.
    * @param    inherits
    *           Node defining the fully qualified name of the superclass or
    *           <code>null</code> for none.
    * @param    imple
    *           Node which has the list of interfaces that are implemented
    *           or <code>null</code> if none exist.
    */
   public void addClass(Aast symbol, Aast inherits, Aast imple)
   {
      ClassDefinition parent = null;
      Set<String>     ifaces = null;
      String          name   = symbol.getText();
      
      WorkArea wa = locate();
      if (wa.preScanPass == 0)
      {
         loadClass(name);
      }
      
      if (inherits != null)
      {
         Aast   child = inherits.getChildAt(0);
         String pname = (String) child.getAnnotation("qualified");
         parent = lookupClass(pname);
         
         if (parent == null)
         {
            // will cause recursion if successful
            loadClass(pname);
            parent = lookupClass(pname);
         }
         
         if (parent == null)
         {
            String err = String.format("Missing class definition for %s!", pname);
            throw new RuntimeException(err);
         }
      }
      else if (!ROOT_OBJ_NAME.equals(name))
      {
         // default parent for all classes
         loadClass(ROOT_OBJ_NAME);
         parent = lookupClass(ROOT_OBJ_NAME);
      }
      
      if (imple != null)
      {
         ifaces = new HashSet<String>();
         
         Aast child = imple.getChildAt(0);
         
         while (child != null)
         {
            ifaces.add((String) child.getAnnotation("qualified"));
            child = (Aast) child.getNextSibling();
         }
      }

      SearchResult found = resolveClassName(name);
      addClass(name, parent, ifaces, found.builtin, found.dotnet);
   }
   
   /**
    * Adds a class name and a token type to the dictionary of classes.
    * This is a flat dictionary (there is no scoping).  The class name
    * is used as the key and the token type is the value.  Since the
    * dictionary is flat, all class names must be unique.
    *
    * @param    name
    *           The fully qualified (including package) class name.
    * @param    parent
    *           The superclass or <code>null</code> for none.
    * @param    ifaces
    *           The list of interfaces that are implemented or
    *           <code>null</code> if none exist.
    * @param    builtin
    *           <code>true</code> if this definition is built-in to Progress.
    * @param    dotnet
    *           <code>true</code> if this definition is a .NET class.
    */
   public void addClass(String          name,
                        ClassDefinition parent,
                        Set<String>     ifaces,
                        boolean         builtin,
                        boolean         dotnet)
   {
      ClassDefinition cls = lookupClass(name);
      boolean loaded = (cls != null);
      
      if (!loaded)
      {
         cls = new ClassDefinition(name, parent, ifaces, builtin, false, dotnet);
      }
      
      // remember this is the class definition being updated (with methods,
      // members and properties)
      currentCls.push(cls);

      if (loaded)
      {
         return;
      }
      
      synchronized (SymbolResolver.class)
      {
         WorkArea wa = locate();
         wa.classDict.addSymbol(false, name, cls);
         
         wa.preScanRefs.add(name);
      }
   }
   
   /**
    * Adds a class name and a token type to the dictionary of classes.
    * This is a flat dictionary (there is no scoping).  The class name
    * is used as the key and the token type is the value.  Since the
    * dictionary is flat, all class names must be unique.
    *
    * @param    symbol
    *           Node defining the fully qualified (including package) class
    *           name.
    */
   public void addInterface(Aast symbol)
   {
      String name = symbol.getText();
      SearchResult found = resolveClassName(name);
      
      addInterface(symbol.getText(), found.builtin, found.dotnet);
   }
   
   /**
    * Adds a class name and a token type to the dictionary of classes.
    * This is a flat dictionary (there is no scoping).  The class name
    * is used as the key and the token type is the value.  Since the
    * dictionary is flat, all class names must be unique.
    *
    * @param    name
    *           The fully qualified (including package) class name.
    * @param    builtin
    *           <code>true</code> if this definition is built-in to Progress.
    * @param    dotnet
    *           <code>true</code> if this definition is a .NET class.
    */
   public void addInterface(String name, boolean builtin, boolean dotnet)
   {
      ClassDefinition iface = lookupClass(name);
      boolean loaded = (iface != null);
      
      if (!loaded)
      {
         iface = new ClassDefinition(name, null, null, builtin, true, dotnet);
      }
      
      // remember this is the interface definition being updated (with
      // methods, members and properties)
      currentCls.push(iface);

      if (loaded)
      {
         return;
      }

      synchronized (SymbolResolver.class)
      {
         WorkArea wa = locate();
         wa.classDict.addSymbol(false, name, iface);
         
         wa.preScanRefs.add(name);
      }
   }
   
   /**
    * Obtain the currently active class or interface definition.
    *
    * @return   The active class or interface definition or <code>null</code>
    *           if there is no active definition.
    */
   public ClassDefinition getCurrentClassDef()
   {
      if (ignoreOORefs)
      {
         return MOCK_CLASS_DEF;
      }
      
      return currentCls.empty() ? null : currentCls.peek();
   }
   
   /**
    * Obtain the name of the currently active class definition.
    *
    * @return   The class name or <code>null</code> if there is no active
    *           class definition.
    */
   public String getCurrentClassName()
   {
      ClassDefinition cls = getCurrentClassDef();
      return (cls == null) ? null : cls.getName();
   }
   
   /**
    * Obtain the parent of the currently active class definition.
    *
    * @return   The parent class definition or <code>null</code> if there is
    *           no explicit parent.
    */
   public ClassDefinition getCurrentParentDef()
   {
      if (ignoreOORefs)
      {
         return MOCK_CLASS_DEF;
      }

      ClassDefinition cls = getCurrentClassDef();
      return (cls == null) ? null : cls.getParent();
   }
   
   /**
    * Obtain the name of the parent of the currently active class definition.
    *
    * @return   The parent class name.
    */
   public String getCurrentParentName()
   {
      ClassDefinition cls = getCurrentParentDef();
      return (cls == null) ? ROOT_OBJ_NAME : cls.getName();
   }
   
   /**
    * Notify the symbol resolver that the current class or interface
    * definition is complete.
    */
   public void classDefinitionComplete()
   {
      currentCls.pop();
   }
   
   /**
    * Attempt to load the class or interface name given. This "eats" any
    * failures since it is not guaranteed that the input does represent a
    * valid class name.
    *
    * @param    name
    *           A potential class or interface name.
    *
    * @return   The fully qualified class name if it can be loaded or <code>null</code> if the
    *           class or interface cannot be loaded.
    */
   public String tryLoadClass(String name)
   {
      String full = null;
      
      try
      {
         full = loadClass(name);
      }
      
      catch (Exception exp)
      {
         // ignore
      }
      
      return full;
   }
   
   /**
    * Attempt to load the class or interface name given. This "eats" any
    * failures since it is not guaranteed that the input does represent a
    * valid class name.
    *
    * @param    name
    *           A potential class or interface name.
    *
    * @return   <code>true</code> if the class or interface can be loaded.
    */
   public boolean canLoadClass(String name)
   {
      return (tryLoadClass(name) != null);
   }
   
   /**
    * Load (using recursion if needed) the hierarchy of class definitions (or
    * a single interface definition) referenced by the given name. This should
    * be called when a reference to a class or interface name occurs (except
    * in the case of the <code>CLASS</code> or <code>INTERFACE</code> language
    * statements). The name can be unqualified (requiring a call to
    * {@link #resolveClassName} to determine the fully qualified name) or it
    * can be fully qualified. The name cannot (just like in Progress) be
    * partially qualified. Once resolved, if the class or interface definition
    * is already loaded, then this will return immediately.  Otherwise the
    * class or interface file will be found via the <code>PROPATH</code> and
    * that file's AST will be loaded.  If the AST does not exist, then the
    * file will first be parsed. During parsing, the <code>CLASS</code> or 
    * <code>INTERFACE</code> language statements will call {@link #addClass}
    * or {@link #addInterface} respectively. This will instantiate the proper
    * {@link ClassDefinition} objects and the contained definitions for
    * methods, member variables and properties will be added by the rules
    * that match those language features. Once parsing is complete, the
    * definition will be fully formed and loaded. Any references to other
    * classes or interfaces in that file will cause recursive parsing/loading
    * using this method, until the entire object hierarchy is loaded. At that
    * point, this chain of methods will return. This recursive loading can
    * also be duplicated by tree walking for an AST that already exists. This
    * processing is the same as that done by parsing but it just avoids the
    * overhead of the full parsing. However, it is triggered by the same
    * language features as noted above. This method uses parsing and/or tree
    * walking as needed and both these techniques may be intermixed in the
    * same recursive chain of class/interface loading.
    *
    * @param    name
    *           Unqualified or fully qualified class or interface name to
    *           load.
    *
    * @return   The fully qualified class or interface name that was loaded.
    */
   public String loadClass(String name)
   {
      SearchResult found = resolveClassName(name);
      
      ClassDefinition cls = lookupClass(found.clsname);
      
      if (cls == null)
      {
         // not loaded yet

         // do we have a cached file name?
         if (found.filename == null)
         {
            // find the class file name in the propath
            if (!findFile(found.clsname, found.type, found))
            {
               String spec = "Cannot find class/interface %s in PROPATH.";
               String err  = String.format(spec, found.clsname);
               
               WorkArea wa = locate();
               if (wa.preScanPass > 1)
               {
                  wa.classDict.addSymbol(false, found.clsname, MOCK_CLASS_DEF);
                  wa.preScanRefs.add(name);

                  // this class is not found, but it must not break parsing - show a warning and
                  // all 'sink' all references to it
                  System.err.println("WARNING: failed " + wa.preScanPass + " level parse. " + err);
                  
                  return found.clsname;
               }
               else
               {
                  throw new RuntimeException(err);
               }
            }
            
            // at this point found.filename must be valid
         }

         if (generator == null)
         {
            generator = new AstGenerator();
            generator.setSilent(silent);
            generator.setPreprocess(true);
            generator.setCache(true);
            generator.setDumpLexer(true);
            generator.setDumpParser(true);
            generator.setAstPersist(true);
         }

         String relative = Configuration.normalizeFilename(found.filename);
         
         WorkArea wa = locate();
         
         if (wa.preScanPass == 0)
         {
            wa.preScanRefs.clear();
         }
         
         wa.preScanPass++;
         
         try
         {
            if (!silent)
            {
               System.out.println("   Recursive level " + wa.preScanPass + 
                                  " parse of class: " + relative);
            }

            generator.preScanClass(relative, found.builtin, found.dotnet);
         }
         catch (AstException   |
                ANTLRException |
                IOException e)
         {
            String msg = String.format("Failed level %d parse of class %s", 
                                       wa.preScanPass,
                                       found.clsname);
            
            for (String clsName : wa.preScanRefs)
            {
               wa.classDict.deleteSymbol(clsName);
            }

            throw new RuntimeException(msg, e);
         }
         finally
         {
            wa.preScanPass--;
            
            if (wa.preScanPass == 0)
            {
               wa.preScanRefs.clear();
               System.out.println("   Done 1st level parsing for " + relative);
            }
         }

         if (lookupClass(found.clsname) == null)
         {
            String spec = "Cannot load class/interface %s from file %s!";
            String err  = String.format(spec, found.clsname, relative);
            throw new RuntimeException(err);
         }
      }
      
      return found.clsname;
   }
   
   /**
    * Searches the class dictionary for a definition matching the passed key.
    *
    * @param   name
    *          Fully qualified class/interface name to match.
    *
    * @return  The class that was found or <code>null</code> if no match was found.
    */
   public ClassDefinition lookupClass(String name)
   {
      if (ignoreOORefs)
      {
         return MOCK_CLASS_DEF;
      }
      
      ClassDefinition cls = null;
      
      synchronized (SymbolResolver.class)
      {
         cls = (ClassDefinition) locate().classDict.lookupSymbol(name);
      }
      
      return cls;
   }
   
   /**
    * Annotate the given method invocation reference with details for the specified method
    * in this class.  This is a no-operation if the method doesn't exist or if the node is
    * <code>null</code>.
    *
    * @param    cname
    *           Class name, may be <code>null</code>.
    * @param    mname
    *           Method name, must NOT be <code>null</code>.
    * @param    isStatic
    *           <code>true</code> if the method is static.
    * @param    node
    *           The AST to be annotated.
    */
   public void annotateObjectMethod(String cname, String mname, boolean isStatic, Aast node)
   {
      ClassDefinition cls = (cname == null) ? getCurrentClassDef() : lookupClass(cname);
      
      int access = calcAccessMode(cname);
      
      if (cls != null)
      {
         cls.annotateMethodCall(mname, access, isStatic, node);
      }
   }
   
   /**
    * Annotate the given node which is a reference to the named class.
    *
    * @param   qname
    *          Fully qualified class name being referenced.
    * @param   node
    *          The node to annotate.
    */
   public void annotateClassRef(String qname, Aast node)
   {
      ClassDefinition cls = null;
      
      if (ignoreOORefs)
      {
         cls = MOCK_CLASS_DEF;
      }
      else
      {
         synchronized (SymbolResolver.class)
         {
            cls = (ClassDefinition) locate().classDict.lookupSymbol(qname);
         }
      }
      
      node.putAnnotation("qualified", qname);
      node.putAnnotation("builtin-cls", cls.isBuiltIn());
      node.putAnnotation("dotnet-cls", cls.isDotNet());
   }
   
   /**
    * Converts any unqualified class or interface names into the fully
    * qualified name (which has a package included). There is no such thing
    * as a partially qualified class or interface name in Progress, so if
    * any "." character is found in the source name, that name is assumed to
    * be fully qualified and is immediately returned. TODO: actually there ARE
    * partially qualified names in v11 and we need to fix this method!
    * <p>
    * Otherwise, the following name resolution actions are attempted:
    * <p>
    * <ol>
    *    <li> If the name matches the unqualified name of any of the
    *         explicitly imported classes (via the USING statement), that
    *         fully qualified class name is returned.
    *    <li> The "*" in each of the imported package specifications is
    *         replaced with the unqualified name (in order) and:
    *    <ol>
    *       <li> If the resulting name matches a class that is already loaded,
    *            that class name is set into the search result, but the other
    *            search result members will NOT be set.
    *       <li> The following paths are searched depending on the USING type:
    *          <ul>
    *             <li> <code>PROPATH</code> for the project's 4GL cls files
    *             <li> <code>OO4GL_PATH</code> for built-in 4GL cls files
    *             <li> <code>ASSEMBLY_PATH</code> for the project's .NET cls
    *                  files
    *             <li> <code>DOTNET_PATH</code> for built-in .NET cls files
    *          </ul>
    *       <li> If a match is found via the search, the members of the search
    *            result will be set as noted below.
    *    </ol>
    * </ol>
    * <p>
    * If an unqualified name is not matched to the above criteria, the name
    * is assumed to be the complete type name and is returned unchanged. This
    * means that no search has been successful.
    * <p>
    * If the filesystem search has been successful, the returned search
    * result will have a non-null <code>filename</code> member.  In addition,
    * the <code>builtin</code> and <code>dotnet</code> members will have been
    * set and can be relied upon.
    *
    * @param    name
    *           Unqualified or fully qualified class or interface name.
    *
    * @return   An object containing data about the search. If the 
    *           <code>filename</code> member is <code>null</code> then only
    *           the <code>name</code> member has been set, though it may not
    *           be valid. If the <code>filename</code> member is NOT
    *           <code>null</code>, then all members of the object are already
    *           set and can be relied upon.
    */
   public SearchResult resolveClassName(String name)
   {
      ClassDefinition cls       = null;
      String          qualified = null;
      UsingSpec       uspec     = null;
      UsingType       utype     = UsingType.UNSPECIFIED;
      SearchResult    result    = new SearchResult();
      
      // remove array brackets if they exist (this assumes they are always
      // at the end AND that there is no other valid content after the
      // opening square bracket); note that brackets can define multi-
      // dimensional arrays like MyClass[,,] for a 3-dimensional array
      int sqBrackets = name.indexOf("[");
      
      if (sqBrackets >= 0)
      {
         name = name.substring(0, sqBrackets);
      }
      
      if (name.indexOf('.') == -1)
      {
         // check the explicit list of class names
         uspec     = explPkgDict.get(name.toLowerCase());
         qualified = (uspec != null) ? uspec.spec : null;
         
         if (qualified == null)
         {
            Iterator<UsingSpec> iter = searchPkgDict.iterator();
            
            // try each of the search specifications in order
            while (iter.hasNext() && qualified == null)
            {
               uspec = iter.next();
               
               // build the next possible qualified class name
               String next = String.format("%s%s", uspec.spec, name);
               
               // short circuit the search logic in the case that we already
               // have loaded this class definition
               cls = lookupClass(next);
               
               if (cls != null)
               {
                  qualified = next;
                  result.builtin = cls.isBuiltIn();
                  result.dotnet  = cls.isDotNet();
               }
               else
               {
                  // search propath for the specific class file name
                  if (findFile(next, uspec.type, result))
                  {
                     qualified = next;
                  }
               }
            }
         }
         
         if (qualified == null)
         {
            // the unqualified name must be the full name
            qualified = name;
         }
      }
      else
      {
         // TODO: actually, this code needs improvement because the following is allowed
         //       (assume that there is a basepath.something.qualifier.Gadget.cls)
         //
         //    USING basepath.something.*.
         //    DEFINE VARIABLE ref AS qualifier.Gadget.
         //
         // To fix this, we need to:
         // 1. check if the name exists as-is
         // 2. if not, do a search as above using searchPkgDict
         
         qualified = name;
      }
      
      if (uspec != null)
         utype = uspec.type;
      
      result.clsname  = qualified;
      result.type     = utype;
      
      return result;
   }
   
   /**
    * Adds a package search specification ("package.*") or a fully qualified
    * class name ("package.class") to the list of known packages. These are
    * used to match unqualified class names with their fully qualified
    * equivalents.
    *
    * @param   name
    *          Package search specification or explicit fully qualified
    *          class name.
    * @param   fromClause
    *          The nodes parsed as the <code>FROM (ASSEMBLY | PROPATH)</code>
    *          clause or <code>null</code> if it wasn't present.
    */
   public void addPackage(String name, Aast fromClause) 
   {
      UsingType type = UsingType.UNSPECIFIED;
      
      if (fromClause != null)
      {
         Aast child = fromClause.getChildAt(0);
         
         if (child != null)
         {
            type = (child.getType() == KW_ASSEMBLY) ? UsingType.DOTNET
                                                    : UsingType.PROPATH;
         }
      }
      
      if (name.endsWith(".*"))
      {
         // remove the asterisk since we will no longer need it (it only
         // serves to denote that this is a search spec versus an explicit
         // fully qualified class name 
         name = name.substring(0, name.length() - 1);
         
         // error out if the directory can't be found in the PROPATH
         if (!pkgExists(name))
         {
            String err = String.format("Non-existent package %s!", name);
            throw new RuntimeException(err);
         }
         
         searchPkgDict.add(new UsingSpec(name, type));
      }
      else
      {
         // find the base name
         int idx = name.lastIndexOf(".");
         
         if (idx < 0)
         {
            throw new RuntimeException("Invalid fully qualified class name!");
         }
         
         String base = name.substring(idx + 1);
         
         // map the lowercased base name to the fully qualified name (allows
         // case-insensitive lookup)
         explPkgDict.put(base.toLowerCase(), new UsingSpec(name, type));
      }
   }
   
   /**
    * Add the named method to the currently active class definition.
    *
    * @param    name
    *           Method name.
    * @param    mtype
    *           Return type for the method.
    * @param    access
    *           Access mode (<code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>).
    * @param    isStatic                          
    *           <code>true</code> if the variable/property is static.
    * @param    node
    *           The AST to be annotated.
    * @param    qname
    *           Fully qualified class name for <code>mtype</code> methods of
    *           <code>METH_CLASS</code> type. <code>null</code> if this
    *           method does not return an object instance.
    */
   public void addObjectMethod(String  name,
                               int     mtype, 
                               int     access,
                               boolean isStatic,
                               Aast    node,
                               String  qname)
   {
      // must be non-null since methods can't be defined outside of classes
      getCurrentClassDef().addMethod(name, mtype, access, isStatic, node, qname);
   }
   
   /**
    * Add the named variable as a data member of the currently active class
    * definition. If no class definition is active, this method does nothing.
    *
    * @param    name
    *           Method name.
    * @param    access
    *           Node of type <code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>.
    * @param    stic
    *           Node of type <code>KW_STATIC</code> if the variable/property is
    *           static.
    * @param    qname
    *           Fully qualified class name for <code>mtype</code> methods of
    *           <code>METH_CLASS</code> type. <code>null</code> if this
    *           method does not return an object instance.
    */
   public void addDataMember(String name, Aast access, Aast stic, String qname)
   {
      ClassDefinition cls = getCurrentClassDef();
      
      if (cls != null)
      {
         addDataMember(name, lookupVariable(name), access, stic, qname);
      }
   }
   
   /**
    * Add the named variable as a data member of the currently active class
    * definition. If no class definition is active, this method does nothing.
    *
    * @param    name
    *           Method name.
    * @param    type
    *           Token type of the member.
    * @param    access
    *           Node of type <code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>.
    * @param    stic
    *           Node of type <code>KW_STATIC</code> if the variable/property is
    *           static.
    * @param    qname
    *           Fully qualified class name for <code>mtype</code> methods of
    *           <code>METH_CLASS</code> type. <code>null</code> if this
    *           method does not return an object instance.
    */
   public void addDataMember(String name, int type, Aast access, Aast stic, String qname)
   {
      ClassDefinition cls = getCurrentClassDef();
      
      if (cls != null)
      {
         int dmode = (type == DEFINE_PROPERTY) ? KW_PUBLIC : KW_PRIVATE;
         int amode = (access == null) ? KW_PUBLIC : access.getType();
         
         addDataMember(name, type, amode, (stic != null), qname);
      }
   }
   
   /**
    * Add the named variable as a data member of the currently active class
    * definition. If no class definition is active, this method does nothing.
    *
    * @param    name
    *           Method name.
    * @param    access
    *           Access mode (<code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>).
    * @param    isStatic
    *           <code>true</code> if the variable/property is static.
    * @param    qname
    *           Fully qualified class name for <code>mtype</code> methods of
    *           <code>METH_CLASS</code> type. <code>null</code> if this
    *           method does not return an object instance.
    */
   public void addDataMember(String name, int access, boolean isStatic, String qname)
   {
      ClassDefinition cls = getCurrentClassDef();
      
      if (cls != null)
      {
         addDataMember(name, lookupVariable(name), access, isStatic, qname);
      }
   }
   
   /**
    * Add the named variable as a data member of the currently active class
    * definition. If no class definition is active, this method does nothing.
    *
    * @param    name
    *           Method name.
    * @param    type
    *           Token type for the member.
    * @param    access
    *           Access mode (<code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>).
    * @param    isStatic                          
    *           <code>true</code> if the variable/property is static.
    * @param    qname
    *           Fully qualified class name for <code>mtype</code> methods of
    *           <code>METH_CLASS</code> type. <code>null</code> if this
    *           method does not return an object instance.
    */
   public void addDataMember(String name, int type, int access, boolean isStatic, String qname)
   {
      ClassDefinition cls = getCurrentClassDef();
      
      if (cls != null)
      {
         cls.addVariable(name, type, access, isStatic, qname);
      }
   }
   
   /**
    * Check if the given class is built-in.
    * 
    * @param    name
    *           The qualified class name.
    *           
    * @return   See above.
    */
   public boolean isBuiltInClass(String name)
   {
      SearchResult found = resolveClassName(name);
      
      return found != null && found.builtin;
   }
   
   
   /**
    * Check if the given class is a .NET class.
    * 
    * @param    name
    *           The qualified class name.
    *           
    * @return   See above.
    */
   public boolean isDotNetClass(String name)
   {
      SearchResult found = resolveClassName(name);
      
      return found != null && found.dotnet;
   }
   
   /**
    * Reports if the given data member is a static member of the given class.
    *
    * @param    cname
    *           Class name to search within or <code>null</code> if this is
    *           a member within the local class definition.
    * @param    name
    *           Member name to lookup.
    *
    * @return   <code>true</code> if the member is static.  <code>false</code> otherwise
    *           including the case where the member does not exist at all.
    */
   public boolean isStaticDataMember(String cname, String name)
   {
      return lookupDataMember(cname, name, true) != -1;
   }
   
   /**
    * Searches the named class' hierarchy for the named data member. This
    * will match both static and instance members.
    *
    * @param    cname
    *           Class name to search within or <code>null</code> if this is
    *           a member within the local class definition.
    * @param    name
    *           Member name to lookup.
    *
    * @return   The token type of the member or -1 if no match was found.
    */
   public int lookupDataMember(String cname, String name)
   {
      int type = lookupDataMember(cname, name, false);
      
      return type == -1 ? lookupDataMember(cname, name, true) : type;
   }
   
   /**
    * Searches the named class' hierarchy for the named data member.
    *
    * @param    cname
    *           Class name to search within or <code>null</code> if this is
    *           a member within the local class definition.
    * @param    name
    *           Member name to lookup.
    * @param    isStatic
    *           If <code>true</code>, only static methods will be looked up.
    *           Otherwise any method can be returned.
    *
    * @return   The token type of the member or -1 if no match was found.
    */
   public int lookupDataMember(String cname, String name, boolean isStatic)
   {
      ClassDefinition cls = (cname == null) ? getCurrentClassDef() : lookupClass(cname);
      
      int access = calcAccessMode(cname);
      
      return (cls == null) ? -1 : cls.lookupVariable(name, access, isStatic);
   }
   
   /**
    * Searches the named class' hierarchy for the named data member and
    * returns the fully qualified class name if that member represents an
    * object instance.
    *
    * @param    cname
    *           Class name to search within or <code>null</code> if this is
    *           a member within the local class definition.
    * @param    name
    *           Member name to lookup.
    * @param    isStatic
    *           <code>true</code> if this is a static reference.
    *
    * @return   Fully qualified class name of the the object instance
    *           represented by the named variable, or <code>null</code> if no
    *           such variable exists.
    */
   public String lookupDataMemberClass(String  cname, String  name, boolean isStatic)
   {
      ClassDefinition cls = (cname == null) ? getCurrentClassDef() : lookupClass(cname);
      
      int access = calcAccessMode(cname);
      
      return (cls == null) ? null : cls.lookupVariableClassName(name, access, isStatic);
   }
   
   /**
    * Reports if the given method is a static member of the given class.
    *
    * @param    cname
    *           Class name to search within or <code>null</code> if this is
    *           a member within the local class definition.
    * @param    name
    *           Member name to lookup.
    *
    * @return   <code>true</code> if the member is static.  <code>false</code> otherwise
    *           including the case where the member does not exist at all.
    */
   public boolean isStaticMethod(String cname, String name)
   {
      return lookupObjectMethod(cname, name, true) != -1;
   }
   
   /**
    * Searches the named class' hierarchy for the named method. This will
    * match any method (static or instance) of the given name.
    *
    * @param    cname
    *           Class name to search within or <code>null</code> if this is
    *           a method call within the local class definition.
    * @param    method
    *           Method name to lookup.
    *
    * @return   The return type of the method or -1 if no match was found.
    */
   public int lookupObjectMethod(String cname, String method)
   {
      int type = lookupObjectMethod(cname, method, false);
      
      return type == -1 ? lookupObjectMethod(cname, method, true) : type;
   }
   
   /**
    * Searches the named class' hierarchy for the named method.
    *
    * @param    cname
    *           Class name to search within or <code>null</code> if this is
    *           a method call within the local class definition.
    * @param    method
    *           Method name to lookup.
    * @param    isStatic
    *           If <code>true</code>, only static methods will be looked up.
    *           Otherwise any method can be returned.
    *
    * @return   The return type of the method or -1 if no match was found.
    */
   public int lookupObjectMethod(String cname, String method, boolean isStatic)
   {
      ClassDefinition cls = (cname == null) ? getCurrentClassDef() : lookupClass(cname);
      
      int access = calcAccessMode(cname);
      
      return (cls == null) ? -1 : cls.lookupMethod(method, access, isStatic);
   }
   
   /**
    * Searches the named class' hierarchy for the named method and returns
    * the fully qualified class name if that method returns an object
    * instance.
    *
    * @param    cname
    *           Class name to search within or <code>null</code> if this is
    *           a method call within the local class definition.
    * @param    method
    *           Method name to lookup.
    * @param    isStatic
    *           <code>true</code> if this is a static reference.
    *
    * @return   Fully qualified class name of the the object instance
    *           represented by the named method, or <code>null</code> if no
    *           such method exists.
    */
   public String lookupObjectMethodClass(String  cname, String method, boolean isStatic)
   {
      ClassDefinition cls = (cname == null) ? getCurrentClassDef() : lookupClass(cname);
      
      int access = calcAccessMode(cname);
      
      return (cls == null) ? null : cls.lookupMethodClassName(method, access, isStatic);
   }
   
   /**
    * Adds an attribute or method keyword token type and the associated token 
    * type of the data type (for attributes) or return type (for methods) to
    * the dictionary of attributes/methods.
    * <p>
    * This is a flat dictionary (there is no scoping).  The keyword's token
    * type is used as the key and the result token type is the value.
    * <p>
    * This form of the method is NOT suitable for adding attributes or
    * methods with <code>ATTR_CLASS</code> or <code>METH_CLASS</code> result
    * types. This is due to the fact that there is no mechanism to
    * associate a class name with the resulting attribute or method.
    *
    * @param    ktype
    *           Keyword token type to be used as the key in the dictionary.
    * @param    rtype
    *           The token type representing the attribute data type or the 
    *           method return type as defined by the Lexer and Parser.
    */
   public void addAttributeOrMethod(int ktype, int rtype)
   {
      addAttributeOrMethod(ktype, rtype, null);
   }
   
   /**
    * Adds an attribute or method keyword token type and the associated token 
    * type of the data type (for attributes) or return type (for methods) to
    * the dictionary of attributes/methods.
    * <p>
    * This is a flat dictionary (there is no scoping).  The keyword's token
    * type is used as the key and the result token type is the value.
    *
    * @param    ktype
    *           Keyword token type to be used as the key in the dictionary.
    * @param    rtype
    *           The token type representing the attribute data type or the 
    *           method return type as defined by the Lexer and Parser.
    * @param    cls
    *           Fully qualified name of the class (user defined type) that
    *           is returned from this method. Must be <code>null</code> if
    *           the resultType is not <code>ATTR_CLASS</code> or
    *           <code>METH_CLASS</code>.
    */
   public void addAttributeOrMethod(int ktype, int rtype, String cls)
   {
      attributes.put(ktype, rtype);
      
      if (cls != null)
         attrTypes.put(ktype, cls);
   }
   
   /**
    * Searches the attribute and method dictionary for a token type matching
    * the passed keyword.  The token type of the method return type or the
    * attribute data type is returned, if a match is found.  A token type
    * of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param    ktype
    *           Token type of the keyword to lookup.
    *
    * @return   The attribute or method token type that was found or -1 if no
    *           match was found.
    */
   public int lookupAttributeOrMethod(int ktype)
   {
      Integer result = attributes.get(ktype);
      
      return result == null ? -1 : result.intValue();
   }
   
   /**
    * Searches the attribute and method dictionary for the fully qualified
    * class name that is the actual type of an <code>ATTR_CLASS</code> or 
    * <code>METH_CLASS</code> matching the given keyword.
    *
    * @param    ktype
    *           Token type of the keyword to lookup.
    *
    * @return   The fully qualified class name that is the actual type 
    *           of an <code>ATTR_CLASS</code> or <code>METH_CLASS</code>.
    *           <code>null</code> if the given keyword is not a valid class
    *           attribute or method.
    */
   public String lookupAttributeOrMethodClass(int ktype)
   {
      int ttype = lookupAttributeOrMethod(ktype);
      
      String result = null;
      
      if (ttype == ATTR_CLASS || ttype == METH_CLASS)
      {
         result = attrTypes.get(ktype);
      }
      
      return result;
   }
   
   /**
    * Adds a label and a token type to the dictionary of labels.  This is
    * a special form of a scoped dictionary where there is only one label name
    * per scope (Progress does not allow lable aliases). The label name is
    * used as the key and the token type is the value.
    *
    * @param   name
    *          Label name.
    * @param   tokenType
    *          The label's token type as defined by the Lexer and Parser.
    */
   public void addLabel(String name, int tokenType)
   {
      labelDict.addScope(null);
      labelDict.addSymbol(false, name, new Integer(tokenType));
   }
   
   /**
    * Searches the label dictionary for a symbol matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param   name
    *          Text string to match.
    * @return  The label's token type that was found or -1 if no match was
    *          found.
    */
   public int lookupLabel(String name)
   {
      return lookupWorkerExact(labelDict, name);
   }
   
   /**
    * Deletes the current label from the dictionary of labels.  This is
    * a special form of a scoped dictionary where there is only one label name
    * per scope (Progress does not allow lable aliases). So, deleting the
    * current label is deleting the current scope.
    */   
   public void deleteLabel()
   {
      labelDict.deleteScope();
   }
   
   /**
    * Adds a widget and a token type to the dictionary of widgets.
    * This dictionary supports scoping that matches the variable namespace.
    * The widget name is used as the key and the token type is the value.
    * <p>
    * Note that this is artificially separated from the variable namespace
    * for implementation simplicity, but in Progress there is only one
    * namespace that handles both variables and widgets.
    *
    * @param   name
    *          Widget name.
    * @param   tokenType
    *          The procedure's token type as defined by the Lexer and Parser.
    */
   public void addWidget(String name, int tokenType)
   {
      Variable var = new Variable(name, tokenType, null);
      
      addWrappedWorker(widgetDict, false, name, var);
   }
   
   /**
    * Searches the widget dictionary for a symbol matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    * <p>
    * Note that this is artificially separated from the variable namespace
    * for implementation simplicity, but in Progress there is only one
    * namespace that handles both variables and widgets.
    *
    * @param   name
    *          Text string to match.
    * @return  The widget's token type that was found or -1 if no match
    *          was found.
    */
   public int lookupWidget(String name)
   {
      return lookupWrappedType(lookupWrapped(widgetDict, name));
   }
   
   /**
    * Adds a frame and a token type to the dictionary of frames. The frame
    * name is used as the key and the token type is the value.
    *
    * @param   name
    *          Frame name.
    * @param   tokenType
    *          The token type as defined by the Lexer and Parser.
    */
   public void addFrame(String name, int tokenType)
   {
      frameDict.addSymbol(false, name, new Integer(tokenType));
   }
   
   /**
    * Get the set of field reference ASTs associated with a particular frame.
    * 
    * @param   frame
    *          Name of the frame for which field references should be retrieved.
    * 
    * @return  Set of field references associated with the given frame.
    */
   public Set<Aast> getFrameFields(String frame)
   {
      Set<Aast> fieldSet = fieldWidgetDict.lookup(frame);
      if (fieldSet == null)
      {
         fieldSet = new HashSet<>();
         fieldWidgetDict.addSymbol(false, frame, fieldSet);
      }
      
      return fieldSet;
   }
   
   /**
    * Map a collection of field reference ASTs to the name of the frame in which they reside as
    * widgets.
    * 
    * @param   frame
    *          Name of the frame contaning the field widgets.
    * @param   fields
    *          Collection of field ASTs to be stored.
    */
   public void addFrameFields(String frame, Collection<Aast> fields)
   {
      Set<Aast> fieldSet = getFrameFields(frame);
      for (Aast fld : fields)
      {
         fieldSet.add(fld);
      }
   }
   
   /**
    * Map a field reference ASTs to the name of the frame in which it resides as a widget.
    * 
    * @param   frame
    *          Name of the frame contaning the field widgets.
    * @param   field
    *          Field AST to be stored.
    */
   public void addFrameField(String frame, Aast field)
   {
      addFrameFields(frame, Collections.singleton(field));
   }
   
   /**
    * Check if the USE-DICT-EXPS option is active for the given frame.
    * 
    * @param    frame
    *           The frame name.
    *           
    * @return   <code>true</code> if USE-DICT-EXPS is active.
    */
   public boolean isUseDictExps(String frame)
   {
      Boolean state = useDictExps.lookup(frame);
      
      return state != null && state;
   }
   
   /**
    * Set the USE-DICT-EXPS for this frame, if not already set.
    * 
    * @param    frame
    *           The frame name.
    */
   public void markUseDictExps(String frame)
   {
      Boolean state = useDictExps.lookup(frame);
      
      if (state == null)
      {
         useDictExps.addSymbol(false, frame, true);
      }
   }
   
   /**
    * Searches the frame dictionary for a symbol matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param   name
    *          Text string to match.
    * @return  The frame's token type that was found or -1 if no match
    *          was found.
    */             
   public int lookupFrame(String name)
   {
      return lookupWorkerExact(frameDict, name);
   }
   
   /**
    * Adds a menu or sub-menu and a token type to the dictionary of menus.
    * The menu or sub-menu name is used as the key and the token type is the
    * value.
    *
    * @param   name
    *          Menu or sub-menu name.
    * @param   tokenType
    *          The token type as defined by the Lexer and Parser.
    */
   public void addMenu(String name, int tokenType) 
   {
      Variable var = new Variable(name, tokenType, null);
      
      addWrappedWorker(menuDict, false, name, var);
   }
   
   /**
    * Searches the menu dictionary for a symbol matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    * <p>
    * This supports both menus and sub-menus.
    *
    * @param   name
    *          Text string to match.
    * @return  The menu's token type that was found or -1 if no match
    *          was found.
    */             
   public int lookupMenu(String name)
   {
      return lookupWrappedType(lookupWrapped(menuDict, name));
   }
   
   /**
    * Adds a menu-item and a token type to the dictionary of menu-items.
    * The menu-item name is used as the key and the token type is the value.
    *
    * @param   name
    *          Menu-item name.
    * @param   tokenType
    *          The token type as defined by the Lexer and Parser.
    */
   public void addMenuItem(String name, int tokenType) 
   {
      Variable var = new Variable(name, tokenType, null);
      
      addWrappedWorker(menuItemDict, false, name, var);
   }
   
   /**
    * Searches the menu-item dictionary for a symbol matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param   name
    *          Text string to match.
    * @return  The menu-item's token type that was found or -1 if no match
    *          was found.
    */             
   public int lookupMenuItem(String name)
   {
      return lookupWrappedType(lookupWrapped(menuItemDict, name));
   }
   
   /**
    * Adds a stream and a token type to the dictionary of streams.
    * The stream name is used as the key and the token type is the value.
    *
    * @param   name
    *          Stream name.
    * @param   tokenType
    *          The token type as defined by the Lexer and Parser.
    */   
   public void addStream(String name, int tokenType) 
   {
      streamDict.addSymbol(false, name, new Integer(tokenType));
   }
   
   /**
    * Searches the stream dictionary for a symbol matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param   name
    *          Text string to match.
    * @return  The stream's token type that was found or -1 if no match
    *          was found.
    */             
   public int lookupStream(String name)
   {
      return lookupWorkerExact(streamDict, name);
   }
   
   /**
    * Adds a query and a token type to the dictionary of queries.
    * The query name is used as the key and the token type is the value.
    *
    * @param   name
    *          Query name.
    * @param   tokenType
    *          The token type as defined by the Lexer and Parser.
    * @param    am 
    *           Access modifiers if not null. Only will be non-null if this is a resource
    *           defined as a member of a class definition.
    * @param    st
    *           Static specifier if not null. Only will be non-null if this is a resource
    *           defined as a member of a class definition.
    */
   public void addQuery(String name, int tokenType, Aast am, Aast st)
   {
      qryDict.addSymbol(false, name, new Integer(tokenType));
      
      ClassDefinition def = getCurrentClassDef();
      
      if (def != null)
      {
         def.addQuery(name, tokenType, decodeAccessMode(am), decodeStatic(st));
      }
   }
   
   /**
    * Searches the query dictionary for a symbol matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param   name
    *          Text string to match.
    * @return  The query's token type that was found or -1 if no match
    *          was found.
    */             
   public int lookupQuery(String name)
   {
      int result = lookupWorkerExact(qryDict, name);
      
      if (result == -1)
      {
         ClassDefinition par = getCurrentParentDef();
         
         if (par != null)
         {
            // TODO: we need to know the context of the calling code here to know whether
            //       this is a static or instance lookup; for now we try both
            result = par.lookupQuery(name, KW_PROTECTD, false);
            
            if (result == -1)
            {
               result = par.lookupQuery(name, KW_PROTECTD, true);
            }
         }
      }
         
      return result;
   }
   
   /**
    * Adds a dataset and a token type to the dictionary of datasets.
    * The dataset name is used as the key and the token type is the value.
    *
    * @param   name
    *          Dataset name.
    * @param   tokenType
    *          The token type as defined by the Lexer and Parser.
    * @param    am 
    *           Access modifiers if not null. Only will be non-null if this is a resource
    *           defined as a member of a class definition.
    * @param    st
    *           Static specifier if not null. Only will be non-null if this is a resource
    *           defined as a member of a class definition.
    */
   public void addDataSet(String name, int tokenType, Aast am, Aast st)
   {
      dsetDict.addSymbol(false, name, new Integer(tokenType));
      
      ClassDefinition def = getCurrentClassDef();
      
      if (def != null)
      {
         def.addDataSet(name, tokenType, decodeAccessMode(am), decodeStatic(st));
      }
   }
   
   /**
    * Searches the dataset dictionary for a symbol matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param   name
    *          Text string to match.
    * @return  The dataset's token type that was found or -1 if no match
    *          was found.
    */             
   public int lookupDataSet(String name)
   {
      int result = lookupWorkerExact(dsetDict, name);
      
      if (result == -1)
      {
         ClassDefinition par = getCurrentParentDef();
         
         if (par != null)
         {
            // TODO: we need to know the context of the calling code here to know whether
            //       this is a static or instance lookup; for now we try both
            result = par.lookupDataSet(name, KW_PROTECTD, false);
            
            if (result == -1)
            {
               result = par.lookupDataSet(name, KW_PROTECTD, true);
            }
         }
      }
         
      return result;
   }
   
   /**
    * Adds a data-source and a token type to the dictionary of data-sources.
    * The data-source name is used as the key and the token type is the value.
    *
    * @param   name
    *          Data-source name.
    * @param   tokenType
    *          The token type as defined by the Lexer and Parser.
    * @param    am 
    *           Access modifiers if not null. Only will be non-null if this is a resource
    *           defined as a member of a class definition.
    * @param    st
    *           Static specifier if not null. Only will be non-null if this is a resource
    *           defined as a member of a class definition.
    */
   public void addDataSource(String name, int tokenType, Aast am, Aast st) 
   {
      dsrcDict.addSymbol(false, name, new Integer(tokenType));
      
      ClassDefinition def = getCurrentClassDef();
      
      if (def != null)
      {
         def.addDataSource(name, tokenType, decodeAccessMode(am), decodeStatic(st));
      }
   }
   
   /**
    * Searches the data-source dictionary for a symbol matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param   name
    *          Text string to match.
    * @return  The data-source's token type that was found or -1 if no match
    *          was found.
    */             
   public int lookupDataSource(String name)
   {
      int result = lookupWorkerExact(dsrcDict, name);
      
      if (result == -1)
      {
         ClassDefinition par = getCurrentParentDef();
         
         if (par != null)
         {
            // TODO: we need to know the context of the calling code here to know whether
            //       this is a static or instance lookup; for now we try both
            result = par.lookupDataSource(name, KW_PROTECTD, false);
            
            if (result == -1)
            {
               result = par.lookupDataSource(name, KW_PROTECTD, true);
            }
         }
      }
         
      return result;
   }
   
   /**
    * Adds a data-relation and a token type to the dictionary of
    * data-relations. The data-relation name is used as the key and the token
    * type is the value.
    *
    * @param   name
    *          Data-relation name.
    * @param   tokenType
    *          The token type as defined by the Lexer and Parser.
    */
   public void addDataRelation(String name, int tokenType) 
   {
      drelDict.addSymbol(false, name, new Integer(tokenType));
   }
   
   /**
    * Searches the data-relation dictionary for a symbol matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param   name
    *          Text string to match.
    * @return  The data-relation's token type that was found or -1 if no match
    *          was found.
    */
   public int lookupDataRelation(String name)
   {
      return lookupWorkerExact(drelDict, name);
   }
   
   /**
    * Loads a list of databases into the schema dictionary.  This is the
    * way that non-default databases are 'connected' or added into the 
    * known namespace.
    *
    * @param    dblist
    *           List of database names to load.
    */
   public void loadSchemaDatabases(String[] dblist)
   {
      if (dblist == null || schemaDict == null)
         return;
      
      try
      {
         for ( int i = 0; i < dblist.length; i++ )
         {
            schemaDict.loadSchema(dblist[i]);
         }
      }
      catch (SchemaException excpt)
      {
         System.err.println(excpt.toString());
      }
   }
   
   /**
    * Adds a list of aliases (each to an associated database name) to the 
    * schema dictionary.
    *
    * @param    list
    *           A list of string arrays, each representing an alias and
    *           associated database to add to the schema dictionary.
    */
   public void loadAliases(List list)
   {
      if (schemaDict != null)
      {
         int len = list.size();
         
         for (int i = 0; i < len; i++)
         {
            String[] pair   = (String[]) list.get(i);
            boolean  result = schemaDict.createAlias(pair[0], pair[1]);
            
            if (!result)
            {
               // todo: yes, more logging here too
               System.err.println("createAlias() failed on " + pair[0] +
                                  " and database " + pair[1]);
            }
         }
      }
   }
   
   /**
    * Returns the shortest unique non-abbreviated buffer name for a field
    * or table reference.  This allows one to calculate the identical buffer
    * name from each reference, such that all references that use the same
    * buffer will share the same buffer name, even if the references are
    * different (due to abbreviations, field vs. table, qualified vs
    * unqualified) in the original source code.
    *
    * @param   name
    *          Text string to match, exactly as found in the original source
    *          code (abbreviated, qualified or inqualified). 
    * @param   field
    *          <code>true</code> if this is a field reference, 
    *          <code>false</code> if this is a table reference.
    *
    * @return  The shortest unabbreviated unique buffer name or
    *          <code>null</code> if no match was found.
    */
   public String lookupBufferName(String name, boolean field)
   {
      String result = null;
      
      try
      {
         if (schemaDict != null)
         {
            SchemaHelper<String> bff =
               (SchemaDictionary dict) -> { return dict.getBufferForField(name); };
               
            SchemaHelper<String> bft =
               (SchemaDictionary dict) -> { return dict.getBufferForTable(name); };
               
            Predicate<String> notNull = (String val) -> { return (val == null); };
               
            result = processHierarchy(field ? bff : bft, notNull);
         }
      }
      
      catch (SchemaException exc)
      {
         exc.printStackTrace();
      }
      
      return result;
   }
   
   /**
    * Returns the logical database name being used for a given table
    * reference.  The resulting name may be different from the database
    * portion of the fully qualified schema name as returned by the
    * {@link #lookupFullTableName} since the same schema can be referenced
    * as multiple database instances.
    *
    * @param   name
    *          Text string to match, exactly as found in the original source
    *          code (abbreviated, qualified or inqualified). 
    *
    * @return  The logical database name or <code>null</code> if no match was
    *          found.
    */
   public String lookupDatabaseName(String name)
   {
      String result = null;
      
      try
      {
         if (schemaDict != null)
         {
            SchemaHelper<String> ldft = (SchemaDictionary dict) -> 
            {
               String dbname = null;
               
               // getLogicalDatabaseForTable() assumes that the node will be found, so we can
               // only call it if we know the node exists in the current dictionary
               if (dict.isTable(name))
               {
                  dbname = dict.getLogicalDatabaseForTable(name);
               }
               
               return dbname;
            };
               
            Predicate<String> notNull = (String val) -> { return (val == null); };
               
            result = processHierarchy(ldft, notNull);
         }
      }
      
      catch (SchemaException exc)
      {
         exc.printStackTrace();
      }
      
      return result;
   }
   
   /**
    * Returns the fully qualified name for a field matching the passed 
    * key.
    * <p>
    * Schema field names can come in 3 forms:
    * <ul>
    *    <li> unqualified (e.g. address1)
    *    <li> partially qualified with a tablename (e.g. customer.address1)
    *    <li> fully qualified with database and table (e.g. 
    *         salesapp.customer.address1)
    * </ul>
    * <p>
    * All 3 forms are accepted as input.  All searches are done on 
    * a case-insensitive basis.  In addition, the Progress 4GL schema
    * supports optional abbreviation of field name portion. These 
    * abbreviations are valid as long as they represent a unique match. The
    * smallest unique match to a table name is the smallest possible 
    * abbreviation.  As such, no ambiguity is allowed in the name passed as 
    * an argument.  There must be only one match that is ever found in the
    * schema namespace.
    *
    * @param   name
    *          Text string to match, in either of the 3 possible forms.
    *
    * @return  The field's fully qualified schema name or <code>null</code>
    *          if no match was found.
    */
   public String lookupFullFieldName(String name)
   {
      String fqn = null;
      
      try
      {
         if (schemaDict != null)
         {
            SchemaHelper<String> ffn =
               (SchemaDictionary dict) -> { return dict.getFullFieldName(name); };
               
            Predicate<String> notNull = (String val) -> { return (val == null); };
               
            fqn = processHierarchy(ffn, notNull);
         }
      }
      catch (SchemaException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      
      return fqn;
   }
   
   /**
    * Searches the schema dictionary for a field matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match, or one that failed due to an
    * ambiguous field name or due to some other problem with the schema
    * dictionary.
    * <p>
    * Schema field names can come in 3 forms:
    * <ul>
    *    <li> unqualified (e.g. address1)
    *    <li> partially qualified with a tablename (e.g. customer.address1)
    *    <li> fully qualified with database and table (e.g. 
    *         salesapp.customer.address1)
    * </ul>
    * <p>
    * All 3 forms are accepted as input.  All searches are done on 
    * a case-insensitive basis.  In addition, the Progress 4GL schema
    * supports optional abbreviation of tablenames and fieldnames. These 
    * abbreviations are valid as long as they represent a unique match. The
    * smallest unique match to a table or field name is the smallest 
    * possible abbreviation.  As such, no ambiguity is allowed in the 
    * name passed as an argument.  There must be only one match that is
    * ever found in the schema namespace.
    *
    * @param    name
    *           Text string to match, in any of the supported 3 forms.
    *
    * @return   The label's token type that was found or -1 if no match was
    *           found.
    */
   public int lookupField(String name)
   {
      int type = -1;
      
      try
      {
         if (schemaDict != null)
         {
            SchemaHelper<Integer> fdt =
               (SchemaDictionary dict) -> { return dict.getFieldDataType(name); };
               
            Predicate<Integer> notNull = 
               (Integer val) -> { return (val == null || val.intValue() == -1) ; };
               
            type = processHierarchy(fdt, notNull);
         }
      }
      catch (SchemaException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      
      return type;
   }
   
   /**
    * Searches the schema dictionary for a field matching the passed 
    * key and returns a boolean reporting if it is a valid field or not.
    * <p>
    * Schema field names can come in 3 forms:
    * <ul>
    *    <li> unqualified (e.g. address1)
    *    <li> partially qualified with a tablename (e.g. customer.address1)
    *    <li> fully qualified with database and table (e.g. 
    *         salesapp.customer.address1)
    * </ul>
    * <p>
    * All 3 forms are accepted as input.  All searches are done on 
    * a case-insensitive basis.  In addition, the Progress 4GL schema
    * supports optional abbreviation of tablenames and fieldnames. These 
    * abbreviations are valid as long as they represent a unique match. The
    * smallest unique match to a table or field name is the smallest 
    * possible abbreviation.  As such, no ambiguity is allowed in the 
    * name passed as an argument.  There must be only one match that is
    * ever found in the schema namespace.
    * <p>
    * <b>This method is safe to use inside a semantic predicate since it
    * will not throw an exception if the name is not found or if there is
    * ambiguity detected!</b>
    *
    * @param   name
    *          Text string to match, in any of the supported 3 forms.
    *
    * @return  true if the name is a valid field, false otherwise.
    */
   public boolean isField(String name)
   {
      boolean result = false;
      
      if (schemaDict != null)
      {
         try
         {
            SchemaHelper<Boolean> isFld =
               (SchemaDictionary dict) -> { return dict.isField(name); };
               
            Predicate<Boolean> whileFalse = (Boolean val) -> { return (val == Boolean.FALSE); };
               
            result = processHierarchy(isFld, whileFalse);
         }
         catch (SchemaException exc)
         {
            // this should not happen
            
            // TODO: log this instead of using stdout...
            exc.printStackTrace();
         }
      }
      
      return result;
   }
   
   /**
    * Returns the fully qualified name for a table matching the passed 
    * key.
    * <p>
    * Schema table names can come in 2 forms:
    * <ul>
    *    <li> unqualified, table name only (e.g. customer)
    *    <li> fully qualified with database and table (e.g. salesapp.customer)
    * </ul>
    * <p>
    * Both forms of input are accepted.  All searches are done on 
    * a case-insensitive basis.  In addition, the Progress 4GL schema
    * supports optional abbreviation of tablenames portion. These 
    * abbreviations are valid as long as they represent a unique match. The
    * smallest unique match to a table name is the smallest possible 
    * abbreviation.  As such, no ambiguity is allowed in the name passed as 
    * an argument.  There must be only one match that is ever found in the
    * schema namespace.
    * <p>
    * In addition to the predefined table names from the schema, the user
    * may define or create buffers, temp-tables and work-tables that all
    * are interchangeable with table names.  This represents a dynamic
    * table namespace which must appear as a single unified space.
    *
    * @param   name
    *          Text string to match, in either of the 2 possible forms.
    *
    * @return  The table's fully qualified schema name or <code>null</code>
    *          if no match was found.
    */
   public String lookupFullTableName(String name)
   {
      String fqn = null;
      
      try
      {
         if (schemaDict != null)
         {
            SchemaHelper<String> ftn =
               (SchemaDictionary dict) -> { return dict.getFullTableName(name); };
               
            Predicate<String> notNull = (String val) -> { return (val == null); };
               
            fqn = processHierarchy(ftn, notNull);
         }
      }
      catch (SchemaException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      
      return fqn;
   }
   
   /**
    * Searches the schema dictionary for a table matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match, or one that failed due to an
    * ambiguous table name or due to some other problem with the schema
    * dictionary.
    * <p>
    * Schema table names can come in 2 forms:
    * <ul>
    *    <li> unqualified, table name only (e.g. customer)
    *    <li> fully qualified with database and table (e.g. salesapp.customer)
    * </ul>
    * <p>
    * Both forms of input are accepted.  All searches are done on 
    * a case-insensitive basis.  In addition, the Progress 4GL schema
    * supports optional abbreviation of tablenames portion. These 
    * abbreviations are valid as long as they represent a unique match. The
    * smallest unique match to a table name is the smallest possible 
    * abbreviation.  As such, no ambiguity is allowed in the name passed as 
    * an argument.  There must be only one match that is ever found in the
    * schema namespace.
    * <p>
    * In addition to the predefined table names from the schema, the user
    * may define or create buffers, temp-tables and work-tables that all
    * are interchangeable with table names.  This represents a dynamic
    * table namespace which must appear as a single unified space.
    *
    * @param    name
    *           Text string to match, in either of the 2 possible forms.
    *
    * @return   The table's token type that was found or -1 if no match was
    *           found.
    */
   public int lookupTable(String name)
   {
      int type = -1;
      
      try
      {
         if (schemaDict != null)
         {
            SchemaHelper<Integer> ftt =
               (SchemaDictionary dict) -> { return dict.getTableType(name); };
               
            Predicate<Integer> notNull = 
               (Integer val) -> { return (val == null || val.intValue() == -1) ; };
               
            type = processHierarchy(ftt, notNull);
         }
      }
      catch (IllegalArgumentException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      catch (SchemaException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      
      return type;
   }
   
   /**
    * Searches the schema dictionary for a table matching the passed 
    * key and returns a boolean reporting if it is a valid table or not.
    * <p>
    * Schema table names can come in 2 forms:
    * <ul>
    *    <li> unqualified, table name only (e.g. customer)
    *    <li> fully qualified with database and table (e.g. salesapp.customer)
    * </ul>
    * <p>
    * Both forms are accepted as input.  All searches are done on 
    * a case-insensitive basis.  In addition, the Progress 4GL schema
    * supports optional abbreviation of tablenames and fieldnames. These 
    * abbreviations are valid as long as they represent a unique match. The
    * smallest unique match to a table or field name is the smallest 
    * possible abbreviation.  As such, no ambiguity is allowed in the 
    * name passed as an argument.  There must be only one match that is
    * ever found in the schema namespace.
    * <p>
    * In addition to the predefined table names from the schema, the user
    * may define or create buffers, temp-tables and work-tables that all
    * are interchangeable with table names.  This represents a dynamic
    * table namespace which must appear as a single unified space.
    * <p>
    * <b>This method is safe to use inside a semantic predicate since it
    * will not throw an exception if the name is not found or if there is
    * ambiguity detected!</b>
    *
    * @param   name
    *          Text string to match, in either of the supported 2 forms.
    *
    * @return  true if the name is a valid field, false otherwise.
    */
   public boolean isTable(String name)
   {
      boolean result = false;
      
      if (schemaDict != null)
      {
         try
         {
            SchemaHelper<Boolean> isTbl = 
               (SchemaDictionary dict) -> { return dict.isTable(name); };
               
            Predicate<Boolean> whileFalse = (Boolean val) -> { return (val == Boolean.FALSE); };
               
            result = processHierarchy(isTbl, whileFalse);
         }
         catch (SchemaException exc)
         {
            // this should not happen
            
            // TODO: log this instead of using stdout...
            exc.printStackTrace();
         }
      }
      
      return result;
   }
   
   /**
    * Searches the schema dictionary for a field matching the passed 
    * name and then will return the token type of the backing construct
    * (table, temp-table or work-table) to which this record belongs.  Note
    * that a token type of -1 is invalid in the Lexer and Parser, so this
    * value is used to denote a search that found no match, or one that
    * failed due to an ambiguous table name or due to some other problem with
    * the schema dictionary.
    * <p>
    * Schema field names can come in 3 forms:
    * <ul>
    *    <li> unqualified (e.g. address1)
    *    <li> partially qualified with a tablename (e.g. customer.address1)
    *    <li> fully qualified with database and table (e.g. 
    *         salesapp.customer.address1)
    * </ul>
    * <p>
    * All 3 forms are accepted as input.  All searches are done on 
    * a case-insensitive basis.  In addition, the Progress 4GL schema
    * supports optional abbreviation of tablenames and fieldnames. These 
    * abbreviations are valid as long as they represent a unique match. The
    * smallest unique match to a table or field name is the smallest 
    * possible abbreviation.  As such, no ambiguity is allowed in the 
    * name passed as an argument.  There must be only one match that is
    * ever found in the schema namespace.
    *
    * @param   name
    *          Text string to match, in any of the supported 3 forms.
    *
    * @return  The backing record's token type that was found or -1 if no
    *          match was found.
    */
   public int lookupFieldRecordType(String name)
   {
      int type = -1;
      
      try
      {
         if (schemaDict != null)
         {
            SchemaHelper<Integer> frt =
               (SchemaDictionary dict) -> { return dict.getFieldRecordType(name); };
               
            Predicate<Integer> notNull = 
               (Integer val) -> { return (val == null || val.intValue() == -1) ; };
               
            type = processHierarchy(frt, notNull);
         }
      }
      catch (IllegalArgumentException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      catch (SchemaException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      
      return type;
   }
   
   /**
    * Searches the schema dictionary for a record matching the passed 
    * name and then will return the token type of the backing construct
    * (table, temp-table or work-table) to which this record belongs. By
    * definition, if one passes a name that references a table, temp-table
    * or work-table directly, then this method returns the same token type
    * value as can be obtained via {@link #lookupTable}. However, if the
    * name references a buffer, then this method will return the type of
    * the backing construct to which the buffer refers.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match, or one that failed due to an
    * ambiguous table name or due to some other problem with schema
    * dictionary.
    * <p>
    * Schema table names can come in 2 forms:
    * <ul>
    *    <li> unqualified, table name only (e.g. customer)
    *    <li> fully qualified with database and table (e.g. salesapp.customer)
    * </ul>
    * <p>
    * Both forms are accepted as input.  All searches are done on 
    * a case-insensitive basis.  In addition, the Progress 4GL schema
    * supports optional abbreviation of tablenames and fieldnames. These 
    * abbreviations are valid as long as they represent a unique match. The
    * smallest unique match to a table or field name is the smallest 
    * possible abbreviation.  As such, no ambiguity is allowed in the 
    * name passed as an argument.  There must be only one match that is
    * ever found in the schema namespace.
    * <p>
    * In addition to the predefined table names from the schema, the user
    * may define or create buffers, temp-tables and work-tables that all
    * are interchangeable with table names.  This represents a dynamic
    * table namespace which must appear as a single unified space.
    *
    * @param   name
    *          Text string to match, in either of the supported 2 forms.
    *
    * @return  The backing record's token type that was found or -1 if no
    *          match was found.
    */
   public int lookupTableRecordType(String name)
   {
      int type = -1;
      
      try
      {
         if (schemaDict != null)
         {
            SchemaHelper<Integer> trt =
               (SchemaDictionary dict) -> { return dict.getTableRecordType(name); };
               
            Predicate<Integer> notNull = 
               (Integer val) -> { return (val == null || val.intValue() == -1) ; };
               
            type = processHierarchy(trt, notNull);
         }
      }
      catch (IllegalArgumentException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      catch (SchemaException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      
      return type;
   }
   
   /**
    * Searches the schema dictionary for a database matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match, or one that failed due to an
    * ambiguous database name or due to some other problem with the schema
    * dictionary.
    * <p>
    * Schema database names are always unqualified (by '.' characters) and
    * cannot be abbreviated.  All searches are done on a case-insensitive
    * basis.  No ambiguity is allowed in the name passed as an argument.
    * There must be only one match that is ever found in the schema namespace.
    * <p>
    * In addition to the predefined table names from the schema, the user
    * may define or create database name aliases that are interchangeable
    * with database names.  This allows a procedure written to use a specific
    * database name to use (at runtime) a database of a different name that 
    * has an identical or at least compatible schema.  This represents a
    * dynamic namespace which must appear as a single unified space.
    * <p>
    * <b>At this time, this method does not support the lookup of aliases
    * since the dictionary does not provide such support.</b>
    * <p> 
    * <b>This method is safe to use inside a semantic predicate since it
    * will not throw an exception if the name is not found or if there is
    * ambiguity detected!</b>
    *
    * @param    name
    *           Text string to match.
    *
    * @return   The database's token type that was found or -1 if no match
    *           was found.
    */
   public int lookupDatabase(String name)
   {
      int type = -1;
      
      // lookup in our local schema dictionary
      if (schemaDict != null && schemaDict.isDatabase(name))
      {
         type = DATABASE;
      }
      
      return type;
   }
   
   /**
    * Look up and return the field info instance which describes a database field in a schema,
    * given the field's name and an optional table name-node.
    * 
    * @param   table
    *          Name node asssociated with the field reference. If not {@code null}, only the
    *          private namespace of the given {@code table} is searched (for an exact match)
    *          of {@code fieldName}. If {@code null}, a lenient search (i.e., abbreviations
    *          permitted) is performed across all scopes in the schema dictionary.
    * @param   name
    *          Field's name. May be qualified or not, but must be unambiguous in the current
    *          parsing context.
    * 
    * @return  Field's info instance, or <code>null</code> if <code>name</code> cannot be
    *          resolved to a field.
    */
   public FieldInfo lookupFieldInfo(NameNode table, String name)
   {
      FieldInfo finfo = null;
      
      try
      {
         SchemaHelper<FieldInfo> fget = (SchemaDictionary dict) ->
         {
            FieldInfo fld = null;
            
            // getFieldInfo() assumes that the node will be found, so we can
            // only call it if we know the node exists in the current dictionary
            if ((table == null && dict.isField(name)) ||
                (table != null && dict.containsNode(table)))
            {
               fld = dict.getFieldInfo(table, name);
            }
            
            return fld;
         };
            
         Predicate<FieldInfo> notNull = (FieldInfo val) -> { return (val == null); };
            
         finfo = processHierarchy(fget, notNull);
      }
      catch (SchemaException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      
      return finfo;
   }
   
   /**
    * Look up and return the AST which represents a database field in a schema, given the field's
    * name.
    * 
    * @param   name
    *          Field's name. May be qualified or not, but must be unambiguous in the current
    *          parsing context.
    * 
    * @return  Field's schema AST, or <code>null</code> if <code>name</code> cannot be resolved
    *          to a field.
    */
   public Aast lookupFieldAst(String name)
   {
      Aast field = null;
      
      try
      {
         SchemaHelper<Aast> fget = (SchemaDictionary dict) ->
         {
            Aast fld = null;
            
            // getField() assumes that the node will be found, so we can
            // only call it if we know the node exists in the current dictionary
            if (dict.isField(name))
            {
               fld = dict.getField(name);
            }
            
            return fld;
         };
            
         Predicate<Aast> notNull = (Aast val) -> { return (val == null); };
            
         field = processHierarchy(fget, notNull);
      }
      catch (SchemaException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      
      return field;
   }
   
   /**
    * Look up and return the AST which represents a database table in a schema, given the table's
    * name.
    * 
    * @param   name
    *          Table's name. May be qualified or not, but must be unambiguous in the current
    *          parsing context.
    * 
    * @return  Table's schema AST, or <code>null</code> if <code>name</code> cannot be resolved
    *          to a table.
    */
   public Aast lookupTableAst(String name)
   {
      Aast table = null;
      
      try
      {
         SchemaHelper<Aast> fget = (SchemaDictionary dict) ->
         {
            Aast tbl = null;
            
            // getTable() is not safe to call if the table doesn't exist in the
            // current schema
            if (dict.isTable(name))
            {
               tbl = dict.getTable(name);
            }
            
            return tbl;
         };
            
         Predicate<Aast> notNull = (Aast val) -> { return (val == null); };
            
         table = processHierarchy(fget, notNull);
      }
      catch (SchemaException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      
      return table;
   }
   
   /**
    * Get an iterator on all of the field ASTs defined for a given table. The fields are iterated
    * in the order in which they were added to the schema dictionary.
    *
    * @param   table
    *          Target table AST.
    *
    * @return  An iterator of fields for the specified table. May be empty, but will not be
    *          <code>null</code>.
    */
   public Iterator<Aast> fields(String table)
   {
      Iterator<Aast> iter = null;
      
      try
      {
         SchemaHelper<Iterator<Aast>> fget = (SchemaDictionary dict) ->
         {
            Aast tableAst = null;
            
            // getTable() is not safe to call if the table doesn't exist in the
            // current schema
            if (dict.isTable(table))
            {
               tableAst = dict.getTable(table);
            }
            
            return (tableAst == null) ? null : dict.fields(tableAst);
         };
            
         Predicate<Iterator<Aast>> notNull = (Iterator<Aast> val) -> { return (val == null); };
            
         iter = processHierarchy(fget, notNull);
      }
      catch (SchemaException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      
      return (iter == null) ? new EmptyIterator<>() : iter;
   }
   
   /**
    * Cleanup any object resources which should not remain in memory.  The only resources
    * that will be used after this point are ones which can be looked up by other classes.
    */
   public void cleanupObjectResources()
   {
      ClassDefinition def = getCurrentClassDef();
      
      if (def != null)
      {
         def.cleanupObjectResources();
      }
   }
   
   /**
    * Allows the addition of a new table of any type (table, buffer,
    * temp-table or work-table) to the schema namespace.  Additions are
    * made either to the global namespace (which is the same as the external
    * procedure scope) or to the current internal procedure or function
    * scope.  This is user controlled by a parameter to this method call. 
    * <p>
    * The database to which these are added is <code>null</code> and is
    * picked up by context in the schema dictionary or the table is
    * simply not associated with a specific database (true for temp-tables
    * and work-tables at least).
    * <p>
    * Any of these token types are accepted as the 2nd parameter:
    * <p>
    * <ul>
    *    <li> TABLE (this should never really occur)
    *    <li> BUFFER
    *    <li> TEMP_TABLE
    *    <li> WORK_TABLE
    * </ul>
    * <p>
    * In order to hide the schema dictionary interfaces from the caller,
    * the <code>NameNode</code> object that can be used to reference
    * the table (for the further addition of fields), is returned as a
    * type <code>Object</code>.  Field additions will require this instance
    * reference to be passed as a parameter.  See {@link #addField}.
    *
    * @param    tablename
    *           The name of the table to add.
    * @param    type
    *           The integer token type representing the type of table. 
    * @param    global
    *           Specifies if the table should be added to the global scope
    *           or to the current procedure/function scope.
    * @param    node
    *           The node that triggered the table creation.
    * @param    am 
    *           Access modifiers if not null. Only will be non-null if this is a resource
    *           defined as a member of a class definition.
    * @param    st
    *           Static specifier if not null. Only will be non-null if this is a resource
    *           defined as a member of a class definition.
    *
    * @return   The reference to the added table which is used in subsequent
    *           field adds.  Returns <code>null</code> on error.
    */
   public Object addTable(String  tablename,
                          int     type,
                          boolean global,
                          Aast    node,
                          Aast    am,
                          Aast    st)
   {
      if (tablename == null || type < BEGIN_RECORDTYPES || type > END_RECORDTYPES)
      {
         return null;
      }
      
      ClassDefinition def = getCurrentClassDef();
      
      // protected temp-tables and buffers are accessible from child classes
      if (def != null && (type == TEMP_TABLE || type == BUFFER))
      {
         int     access   = (am != null) ? am.getType() : -1;
         boolean isStatic = (st != null && st.getType() == KW_STATIC);
         
         def.addTable(tablename, type, access, isStatic, schemaDict);
      }
      
      // make edits in our local schema dictionary
      return (Object) schemaDict.addTableEntry(null, tablename, type, global, node);
   }
   
   /**
    * Allows the addition of a new field of any Progress data type to a 
    * specific table in the schema namespace.  Additions are
    * always made to the global namespace, not to a specific scope.
    * <p>
    * Any of the valid field token types are accepted as the 3rd parameter
    * (e.g. <code>ProgressParserTokenTypes.FIELD_INT</code>).
    * <p>
    * In order to hide the schema dictionary interfaces from the caller,
    * the <code>NameNode</code> object that can be used to reference
    * the table is required.  This is an <code>Object</code> returned from a
    * previous call to {@link #addTable}.  This rule casts that to the
    * proper type before calling the dictionary.
    *
    * @param    table
    *           The table reference in which to add the field.
    * @param    field
    *           The name of the field to add. 
    * @param    ftype
    *           The parser token type representing the field's data type.
    * @param    node
    *           Node in source file from which this new field creation was
    *           triggered.
    *
    * @return   The <code>Object</code> representing the added field or 
    *           <code>null</code> on failure.
    */
   public Object addField(Object table, String field, int ftype, Aast node)
   {
      if (table == null || field == null || ftype < BEGIN_FIELDTYPES || ftype > END_FIELDTYPES)
      {
         return null;
      }
      
      // make edits in our local schema dictionary
      return (Object) schemaDict.addFieldEntry((NameNode)table,
                                               node,
                                               field,
                                               ftype,
                                               true);
   }
   
   /**
    * Allows the addition of a new field of any Progress data type to a 
    * specific table in the schema namespace based on reference to an already
    * existing variable or field.  Additions are always made to the global
    * namespace, not to a specific scope.
    * <p>
    * Depending on the token type stored in the 3rd parameter, 2 possible
    * outcomes can occur:
    * <p>
    * <ol>
    *    <li> If the AST represents a variable type, then the field's options
    *         will be set to any non-default values stored in the variable 
    *         definition.
    *    <li> If the AST is a field, then the schema dictionary will create
    *         the field with its options set the same as the referenced
    *         field.
    * </ol>
    * <p>
    * In order to hide the schema dictionary interfaces from the caller,
    * the <code>NameNode</code> object that can be used to reference
    * the table is required.  This is an <code>Object</code> returned from a
    * previous call to {@link #addTable}.  This rule casts that to the
    * proper type before calling the dictionary.
    *
    * @param    table
    *           The table reference in which to add the field.
    * @param    field
    *           The name of the field to add. 
    * @param    ast
    *           The AST node being used as a reference variable or field.
    *
    * @return   The <code>Aast</code> instance representing the added field or 
    *           <code>null</code> on failure.
    */
   public Aast addFieldLike(Object table, String field, Aast ast)
   {
      if (table == null || field == null || ast == null)
         return null;
      
      int     ftype   = ast.getType();
      int     extent  = getExtentSubscript(ast);
      String  oldname = ast.getText();
      boolean isFld   = ftype > BEGIN_FIELDTYPES && ftype < END_FIELDTYPES;
      
      if (isFld)
      {
         NameNode node = null;
         
         // this field should be LIKE another field, call the schema
         // dictionary to add the field.
         try
         {
            // find the source field's name node, even if it is in a different schema dict
            NameNode source = null;
            
            SchemaHelper<NameNode> ffn =
               (SchemaDictionary dict) -> { return dict.findFieldNode(oldname); };
               
            Predicate<NameNode> notNull = (NameNode val) -> { return (val == null); };
               
            source = processHierarchy(ffn, notNull);
            
            if (source == null)
            {
               throw new SchemaException("Field '" + oldname + "' not found");
            }
            
            // add to our local schema dictionary, using the (possibly remote) source node
            node = schemaDict.addFieldEntry((NameNode) table,
                                            ast,
                                            field,
                                            source,
                                            extent);
         }
         catch (SchemaException exc)
         {
            // TODO: log this instead of using stderr...
            exc.printStackTrace();
         }
         
         return node.getAst();
      }
      
      // OK, we are add a field like a variable, so we add the field and then
      // copy our variable options into it
      int type = -1;
      
      // the VAR_ type must be converted to the proper FIELD_ type
      switch (ftype)
      {
         case VAR_CHAR:
            type = FIELD_CHAR;
            break;
         case VAR_CLASS:
            type = FIELD_CLASS;
            break;
         case VAR_COM_HANDLE:
            type = FIELD_COM_HANDLE;
            break;
         case VAR_DATE:
            type = FIELD_DATE;
            break;
         case VAR_DATETIME:
            type = FIELD_DATETIME;
            break;
         case VAR_DATETIME_TZ:
            type = FIELD_DATETIME_TZ;
            break;
         case VAR_DEC:
            type = FIELD_DEC;
            break;
         case VAR_HANDLE:
            type = FIELD_HANDLE;
            break;
         case VAR_INT:
            type = FIELD_INT;
            break;
         case VAR_INT64:
            type = FIELD_INT64;
            break;
         case VAR_LOGICAL:
            type = FIELD_LOGICAL;
            break;
         case VAR_RAW:
            type = FIELD_RAW;
            break;                     
         case VAR_RECID:
            type = FIELD_RECID;
            break;
         case VAR_ROWID:
            type = FIELD_ROWID;
            break;
         default:
            // this MUST be at the end of the list of variables, it will
            // match LONGCHAR and MEMPTR which cannot be created as a field
            if (ftype > BEGIN_VARTYPES && ftype < BEGIN_VARTYPES)
            {
               String err = String.format("Unexpected variable type %s!",
                                          ast.getDescriptiveTokenText());
               throw new RuntimeException(err);
            }
            break;
      }
      
      if (type == -1)
         return null;
      
      Variable var = (Variable) lookupWrapped(varDict, oldname);
      NameNode node = null;
      try
      {
         // make edits in our local schema dictionary
         node = schemaDict.addFieldEntry((NameNode) table,
                                         ast,
                                         field,
                                         type,
                                         var.getOptions(extent));
      }
      catch (SchemaException exc)
      {
         exc.printStackTrace();
      }
      
      return node.getAst();
   }
   
   /**
    * Inspect the field definition and write any options into the field 
    * object.
    *
    * @param    name
    *           The name of the field which must have its options set.
    * @param    ast
    *           The tree to inspect which defines a Progress field.
    */
   public void setFieldOptions(String name, Aast ast)
   {
      try
      {
         // make edits in our local schema dictionary
         schemaDict.setFieldProperties(name, ast);
      }
      catch (SchemaException exc)
      {
         // TODO: log this instead of using stderr...
         exc.printStackTrace();
      }
   }
   
   /**
    * Allows the addition of a list of fields (as defined in an existing
    * table) to a specific table in the schema namespace.  Additions are
    * always made to the global namespace, not to a specific scope.
    * <p>
    * In order to hide the schema dictionary interfaces from the caller,
    * the <code>NameNode</code> object that can be used to reference
    * the table is required.  This is an <code>Object</code> returned from a
    * previous call to {@link #addTable}.  This rule casts that to the
    * proper type before calling the dictionary.
    *
    * @param    table
    *           The table reference in which to add fields.
    * @param    tablename
    *           The name of the table whose fields should be added.
    *
    * @return   The list of ASTs representing the copied fields.
    */
   public Aast[] addFieldsFrom(Object table, String tablename)
   {
      return addFieldsFrom(table, tablename, false);
   }
   
   /**
    * Allows the addition of a list of fields (as defined in an existing
    * table) to a specific table in the schema namespace.  Additions are
    * always made to the global namespace, not to a specific scope.
    * <p>
    * In order to hide the schema dictionary interfaces from the caller,
    * the <code>NameNode</code> object that can be used to reference
    * the table is required.  This is an <code>Object</code> returned from a
    * previous call to {@link #addTable}.  This rule casts that to the
    * proper type before calling the dictionary.
    *
    * @param    table
    *           The table reference in which to add fields.
    * @param    tablename
    *           The name of the table whose fields should be added.
    * @param    skipGlobal
    *           if (@code true}, do not look in the global schema scope to find
    *           {@code tablename}.
    *
    * @return   The list of ASTs representing the copied fields.
    */
   public Aast[] addFieldsFrom(Object table, String tablename, boolean skipGlobal)
   {
      List<Aast> fields = null;
      try
      {
         NameNode toTable = (NameNode) table;
         
         // find the source field's name node, even if it is in a different schema dict
         SchemaHelper<NameNode> ftfn = (SchemaDictionary dict) ->
         {
            NameNode nn = null;
            
            // findTableFromNode() is not safe to call if the table doesn't exist in the
            // current schema
            if (dict.isTable(tablename))
            {
               nn = dict.findTableFromNode(tablename, toTable, skipGlobal);
            }
            
            return nn;
         };
         
         Predicate<NameNode> notNull = (NameNode val) -> { return (val == null); };
         
         NameNode fromTable = processHierarchy(ftfn, notNull);
         
         if (fromTable == null)
         {
            throw new SchemaException("'From' table not found: " + tablename);
         }
         
         // make edits in our local schema dictionary
         schemaDict.addFieldEntries(tablename, fromTable, toTable, skipGlobal);
         
         fields = new ArrayList<Aast>();
         Aast child = (Aast) toTable.getAst().getFirstChild();
         while (child != null)
         {
            if (child.getType() > BEGIN_FIELDTYPES && child.getType() < END_FIELDTYPES)
            {
               fields.add(child);
            }
            
            child = (Aast) child.getNextSibling();
         }
      }
      catch (SchemaException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      
      return fields == null ? null : fields.toArray(new Aast[0]);
   }
   
   /**
    * Adds an index from the original source table into the temp-table that
    * is passed in.  Additions are always made to the global namespace, not
    * to a specific scope.
    * <p>
    * In order to hide the schema dictionary interfaces from the caller,
    * the <code>NameNode</code> object that can be used to reference
    * the table is required.  This is an <code>Object</code> returned from a
    * previous call to {@link #addTable}.  This rule casts that to the
    * proper type before calling the dictionary.
    *
    * @param    table
    *           The table reference in which to add fields.
    * @param    node
    *           Node in source file from which this new field creation was
    *           triggered.
    * @param    name
    *           The name of the index to add. 
    * @param    primary
    *           Whether or not the <code>PRIMARY</code> keyword was present.
    */
   public void addIndexFrom(Object  table,
                            Aast    node,
                            String  name,
                            boolean primary)
   {
      if (table == null || name == null)
      {
         // nothing to do
         return;
      }
      
      // add "use index" processing to schema dictionary
      try
      {
         // make edits in our local schema dictionary uses (possibly remote) state that has
         // been previously setup
         schemaDict.copyIndex((NameNode) table, node, name, primary);
      }
      catch (SchemaException exc)
      {
         exc.printStackTrace();
      }
   }
   
   /**
    * Adds an index as defined in the source code into the temp-table that
    * is passed in.  Additions are always made to the global namespace, not
    * to a specific scope.
    * <p>
    * In order to hide the schema dictionary interfaces from the caller,
    * the <code>NameNode</code> object that can be used to reference
    * the table is required.  This is an <code>Object</code> returned from a
    * previous call to {@link #addTable}.  This rule casts that to the
    * proper type before calling the dictionary.
    *
    * @param    table
    *           The table reference in which to add fields.
    * @param    name
    *           The name of the index to add. 
    * @param    ast
    *           The AST storing the nodes that define the index.
    */
   public void addIndex(Object table, String name, Aast ast)
   {
      if (table == null || name == null || ast == null)
      {
         // nothing to do
         return;
      }
      
      try
      {
         // make edits in our local schema dictionary
         // add a new index for the given table by the given name and
         // set the options based on our AST
         schemaDict.addIndex((NameNode) table, name, ast);
      }
      catch (SchemaException exc)
      {
         exc.printStackTrace();
      }
   }
   
   /**
    * Sets specific conditions on a temp-table definition based on
    * encountering certain tokens in the source code.  The following tokens
    * are known: <code>NO-UNDO, VALIDATE, RCODE-INFORMATION</code>.
    * <p>
    * In order to hide the schema dictionary interfaces from the caller,
    * the <code>NameNode</code> object that can be used to reference
    * the table is required.  This is an <code>Object</code> returned from a
    * previous call to {@link #addTable}.  This rule casts that to the
    * proper type before calling the dictionary.
    *
    * @param    table
    *           The table reference which is completed.
    * @param    token
    *           The type of the option that was encountered in the source
    *           code.
    * @param    node
    *           The node that triggered the table creation.
    */
   public void setTableOption(Object table, int token, Aast node)
   {
      // access our local schema dictionary
      Aast tableAst = schemaDict.getTable((NameNode) table);
      
      try
      {
         Aast prop = new ProgressAst();
         prop.setType(token);
         prop.setLine(node.getLine());
         prop.setColumn(node.getColumn());
         
         switch (token)
         {
            case KW_NO_UNDO:
            {
               // set NO-UNDO option for this table
               prop.setText("no-undo");
               break;
            }
            
            case KW_VALIDATE:
               // set VALIDATE option for this table
               prop.setText("validate");
               break;
               
            case KW_RCOD_INF:
            {
               // set RCODE-INFORMATION option for this table
               prop.setText("rcode-information");
               break;
            }
         }
         
         // make edits in our local schema dictionary
         schemaDict.setTableProperty(tableAst, prop);
      }
      catch (SchemaException exc)
      {
         exc.printStackTrace();
      }
   }
   
   /**
    * Notifies the schema dictionary that the end of a temp-table definition
    * has been reached. This allows certain default processing to be executed
    * as needed.
    * <p>
    * In order to hide the schema dictionary interfaces from the caller,
    * the <code>NameNode</code> object that can be used to reference
    * the table is required.  This is an <code>Object</code> returned from a
    * previous call to {@link #addTable}.  This rule casts that to the
    * proper type before calling the dictionary.
    *
    * @param    table
    *           The table reference which is completed.
    */
   public void finalizeTableDefine(Object table)
   {
      try
      {
         // notify the local schema dictionary of the "end of temp-table definition"
         schemaDict.finalizeTableDefine((NameNode) table);
      }
      catch (SchemaException exc)
      {
         exc.printStackTrace();
      }
   }
   
   /**
    * Adds a record scope to the schema namespace to the top of the stack
    * of scopes.  Any tables that are promoted (see {@link #promoteTable})
    * within this scope are given precedence in making their field names
    * unambiguous when compared to fields of the same name in tables that
    * are in scopes that are lower on the stack.
    *
    * @param    mark
    *           <code>true</code> to mark the associated scope as one which
    *           is an internal procedure, user-defined function or trigger.
    *           <code>false</code> for any other case including external
    *           procedures and inner blocks (<code>DO, REPEAT or FOR</code>).
    */
   public void addSchemaScope(boolean mark)
   {
      if (schemaDict != null)
      {
         schemaDict.addScope(mark);
      }
   }
   
   /**
    * Deletes the current record scope in the schema namespace (the top of
    * the stack of scopes).  Any tables that were promoted (see
    * {@link #promoteTable}) into this scope are naturally popped off the
    * stack by this action if <code>propagate</code> is <code>false</code>.
    * Otherwise the namespace updates made in this scope will be copied to
    * the namespaces of the enclosing scope.
    *
    * @param    propagate
    *           <code>true</code> to force namespace propagation to the
    *           enclosing scope.
    */
   public void deleteSchemaScope(boolean propagate)
   {
      try
      {
         if (schemaDict != null)
         {
            schemaDict.removeScope(propagate);
         }
      }
      
      catch (SchemaException exp)
      {
         // ignore it, we don't care!
      }
   }
   
   /**
    * Globally enable or disable schema table name promotion.
    * 
    * @param    allow
    *           <code>true</code> to enable, <code>false</code> to disable.
    */
   public void allowPromotion(boolean allow)
   {
      this.allow = allow;
   }
   
   /**
    * Promotes a table (and all its fields) into the current top scope of
    * the schema dictionary's stack of scopes.  Any tables that are promoted
    * within this scope are given precedence in making their field names
    * unambiguous when compared to fields of the same name in tables that
    * are in scopes that are lower on the stack.
    * <p>
    * Scopes are created using {@link #addSchemaScope} and are deleted using 
    * {@link #deleteSchemaScope}.
    * 
    * @param    table
    *           The name of the table (optionally qualified) to promote.
    * @param    force
    *           Override the default promotion behavior and force a refresh
    *           of the promoted table even if it was already promoted
    *           previously.
    * @param    noProp                   
    *           <code>true</code> to force any promoted nodes to have their
    *           no propagate flag set. This disables propagation of these
    *           nodes across the current scope boundary.
    */
   public void promoteTableName(String table, boolean force, boolean noProp)
   {
      try
      {
         if (allow)
         {
            // we only promote if the name is in the local schema, there is no
            // scoping or need to scope the schema dictionaries in the class
            // hierarchy, because the temp-tables are already at the global
            // scope in those dictionaries
            if (schemaDict.isTable(table))
            {
               schemaDict.promoteTable(table, force, noProp);
            }
         }
      }
      catch (SchemaException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
   }
   
   /**
    * Promotes a table (and all its fields) into the current top scope of
    * the schema dictionary's stack of scopes.  Any tables that are promoted
    * within this scope are given precedence in making their field names
    * unambiguous when compared to fields of the same name in tables that
    * are in scopes that are lower on the stack.
    * <p>
    * Scopes are created using {@link #addSchemaScope} and are deleted using 
    * {@link #deleteSchemaScope}.
    * 
    * @param    table
    *           The <code>Object</code> reference representing a table
    *           that had been previously added to the schema dictionary.
    * @param    force
    *           Override the default promotion behavior and force a refresh
    *           of the promoted table even if it was already promoted
    *           previously.
    * @param    noProp
    *           <code>true</code> to force any promoted nodes to have their
    *           no propagate flag set. This disables propagation of these
    *           nodes across the current scope boundary.
    */
   public void promoteTable(Object table, boolean force, boolean noProp)
   {
      promoteTableName(((NameNode) table).getName(), force, noProp);
   }
   
   /**
    * Post-process the given field, in terms of honoring the VALIDATE clause, if present. This 
    * will:
    * <ul>
    *    <li>
    *       If the VALIDATE clause is not present, it will remove the VALEXP and VALMSG children
    *    </li>
    *    <li>
    *       If the VALIDATE clause is present, it will remove all extent and LBRACKET information
    *       from the VALEXP nodes. Also, any reference of the source field in the VALEXP will 
    *       be rewritten to reference the new field.
    *    </li>
    * </ul>
    * 
    * @param    field
    *           The field which needs to be processed. This is a field defined using a 
    *           <code>LIKE table.sourceField</code> clause, with or without the VALIDATE clause. 
    * @param    hasValidate
    *           When <code>true</code>, it indicates the VALIDATE clause was present and the 
    *           VALEXP node will need to be processed.
    * @param    sourceSchemaName
    *           The schema name of the source field
    */
   public void processLikeField(Aast field, boolean hasValidate, String sourceSchemaName)
   {
      try
      {
         if (schemaDict != null)
         {
            // make edits in our local schema dictionary
            schemaDict.processLikeField(field, hasValidate, sourceSchemaName);
         }
      }
      catch (AmbiguousSchemaNameException exc)
      {
         exc.printStackTrace();
      }
   }
   
   /**
    * Prints a report to <code>stderr</code> with the listing of each
    * dictionary.  One record per dictionary entry will be printed and
    * there will be a section for each namespace type.
    */
   public void dump()
   {
      System.err.println("\n-------------Keyword Listing-------------");
      kwDict.dump();
      
      System.err.println("\n-------------Variable Listing-------------");
      varDict.dump();
      
      System.err.println("\n-------------Function Listing-------------");
      funcDict.dump();
      
      System.err.println("\n-------------Procedure Listing-------------");
      procDict.dump();
      
      System.err.println("\n---------------Label Listing---------------");
      labelDict.dump();
      
      System.err.println("\n--------------Widget Listing--------------");
      widgetDict.dump();
      
      System.err.println("\n---------------Frame Listing---------------");
      frameDict.dump();
      
      System.err.println("\n-----------Field Widget Listing-----------");
      fieldWidgetDict.dump();
      
      System.err.println("\n----------- Frame USE-DICT-EXPS Listing-----------");
      useDictExps.dump();
      
      System.err.println("\n---------------Menu Listing---------------");
      menuDict.dump();
      
      System.err.println("\n-------------Menu Item Listing-------------");
      menuItemDict.dump();
      
      System.err.println("\n--------------Stream Listing--------------");
      streamDict.dump();
   }
   
   /**
    * Prints a report to <code>schema.dmp</code> with the listing of all
    * current contents of the Schema Dictionary.
    */
   public void dumpSchema()
   throws IOException,
          FileNotFoundException
   {
      FileOutputStream fout = null;
      PrintStream      ps   = null;
      try
      {
         if (schemaDict != null)
         {
            File file = new File("schema.dmp");
            fout = new FileOutputStream(file);
            ps = new PrintStream(fout);
            schemaDict.dump(ps);
            System.out.println(file.getAbsolutePath() + " dump file created");
         }
      }
      finally
      {
         if (ps != null)
         {
            ps.close();
         }
         else if (fout != null)
         {
            fout.close();
         }
      }
   }
   
   /**
    * Returns the token type from the passed wrapped object or -1 for an
    * is invalid {@link TokenDataWrapper} object.
    *
    * @param   wrapped
    *          The object from which to obtain the token type.
    *
    * @return  The token type that was found or -1 for an invalid wrapper
    *          object.
    */    
   public int lookupWrappedType(TokenDataWrapper wrapped)
   {
      return wrapped != null ? wrapped.getTokenType() : -1;
   }
   
   /**
    * Set the {@link SchemaDictionary#markPreferred} flag, so all info for the tables which are 
    * weak references will be collected in the preferred nodes for the current scope.
    * 
    * @param    preferred
    *           The new value of the flag.
    * 
    * @return   The old value of the {@link SchemaDictionary#markPreferred} field. 
    */
   public boolean markPreferred(boolean preferred)
   {
      boolean old = false;
      
      if (schemaDict != null)
      {
         // make edits in our local schema dictionary
         old = schemaDict.markPreferred(preferred);
      }
      
      return old;
   }
   
   /**
    * Initialize the maps of all possible class and interface definitions that
    * exist in the project. This includes all <code>.cls</code> files which can
    * be found via the <code>PROPATH</code> as well as all Progress built-ins
    * and all .NET classes/interfaces that can be accessed.
    * <p>
    * These files will be listed from the directories represented by the
    * {@link #propath}, {@link #OO4GL_PATH}, {@link #DOTNET_PATH} and the
    * {@link #ASSEMBLY_PATH}.
    * <p>
    * If file-system case-sensitivity is <code>false</code> then all map keys
    * will be lowercased.
    *
    * @param    propath
    *           The propath to scan for initialization. The propath can change
    *           with each instantiation of this class.  If it changes then
    *           the propath class mappings must be reinitialized. This may be
    *           <code>null</code> in which case the map will be empty.
    * @param    caseSens
    *           <code>true</code> if the legacy system had a case-sensitive
    *           file-system.
    */
   private static synchronized void initPossibleClasses(String[] propath, boolean caseSens)
   {
      WorkArea wa = locate();
      
      if (wa.propathCls == null)
      {
         wa.propathCls = new HashMap<String, String>();
         XmlHelper.loadStringMap(wa.propathCls, CLASS_MAP_FILE, true);
         
         // this is the dynamic scanning approach, it is horrifically SLOW
         // which is why we don't use it; the code must also add this check
         // in the above if stmt: " || propath != locate().lastPropath)"
         /*
         wa.lastPropath = propath;
         
         if (propath != null)
         {
            boolean curDir = false;
            
            for (int i = 0; i < propath.length; i++)
            {
               if (propath[i] != null)
               {
                  String srch = propath[i];
                  
                  if (srch.length() == 0 || srch.equals("."))
                  {
                     // only search the current directory once
                     if (curDir)
                        continue;
                     
                     srch   = ".";
                     curDir = true;
                  }
                  
                  createClassMappings(propathCls,
                                      getClassFileList(srch, false),
                                      srch,
                                      false);
               }
            }
         }
         */
      }
      
      // add the built-ins, .net and custom fakeout classes/interfaces; this
      // only happens the first time this method is called
      // TODO: this assumes the caseSens value is the same project-wide AND
      //       that the correct value is passed in the first time the
      //       class is instantiated (which can be by the SchemaLoader)
      if (wa.oo4glCls == null)
      {
         // 4GL classes are matched case-insensitively (even on case-sensitive file systems)
         wa.oo4glCls = new HashMap<String, String>();
         createClassMappings(wa.oo4glCls,
                             getClassFileList(oo4glPath, false),
                             oo4glPath,
                             false);
      }
      if (wa.dotnetCls == null)
      {
         wa.dotnetCls = new HashMap<String, String>();
         createClassMappings(wa.dotnetCls,
                             getClassFileList(dotnetPath, caseSens),
                             dotnetPath,
                             caseSens);
      }
      if (wa.assemblyCls == null)
      {
         wa.assemblyCls = new HashMap<String, String>();
         createClassMappings(wa.assemblyCls,
                             getClassFileList(ASSEMBLY_PATH, caseSens), 
                             ASSEMBLY_PATH,
                             caseSens);
      }
   }
   
   /**
    * List the 4GL class/interface files in a given path.
    *
    * @param    path
    *           The directory to search.
    *
    * @return   The list of files that exist.  If there are no files, the
    *           array will have a 0 length.
    */
   private static ArrayList<File> getClassFiles(String path)
   {
      ArrayList<File> results = new ArrayList<File>();
      
      scan(new File(path), results);
      
      return results;
   }
   
   /**
    * Core implementation of the listing algorithm which uses the
    * <code>File</code> class to make a complete list of the directory 
    * contents, then this list is used to build a filtered list of just
    * those files that match the user-defined regular expression. Note
    * that if recursion was previously requested (on construction) by the
    * caller, this method will recursively call itself for every directory
    * it encounters in the list.  Only files will be returned in the results
    * list and all results are added to the results list that is passed as
    * a parameter.
    *
    * @param    dir
    *           The directory in which to list all files.
    * @param    results
    *           The list to which all matching files must be appended.
    */
   private static void scan(File dir, ArrayList<File> results)
   {
      // we don't use a FilenameFilter here because this would filter out
      // all directories that don't match the regular expression, but this
      // would not be correct, we need to see all directories and we will
      // manually filter the filenames down ourselves
      File[] rawResults  = dir.listFiles();
      
      // separate file and directory results
      for (int i = 0; i < rawResults.length; i++)
      {
         if (rawResults[i].isDirectory())
         {
            // recursively call ourselves to process this subdirectory
            scan(rawResults[i], results);
         }
         else
         {
            // only accept matches that meet the user's criteria (regular
            // expression)
            if (rawResults[i].getName().endsWith(CLASS_EXT))
            {
               // add the file to the results list
               results.add(rawResults[i]);
            }
         }
      }
   }
   
   /**
    * Multi-level lookup routine for processing searches on variable and
    * function dictionaries based on
    * the {@link com.goldencode.p2j.util.ScopedSymbolDictionary} class. All 
    * searches resolve to a symbol that exactly matches the passed key or to
    * a built-in function or variable that is represented by a language
    * keyword.  The token type of the match is returned.  Note that a token  
    * type of -1 is invalid in the Lexer and Parser, so this value is used to 
    * denote a search that found no match.  Abbreviation processing is
    * handled for built-ins ONLY and is naturally done by usage of the 
    * keyword dictionary in the first phase of the lookup process.  The
    * second lookup phase uses any returned keyword to modify the search or 
    * if no keyword is found, then a standard exact search is executed.
    * <p> 
    * The algorithm checks for a keyword match.  If not, a simple actual text
    * lookup is done based on the exact text found in the source. If there is
    * a keyword match, then if the keyword is reserved, the keyword's full
    * text is used in the name lookup.  The result would match the preloaded 
    * built-in name.  If the keyword match is with an unreserved keyword,
    * then the exact text from the source is used for a first lookup.  If this
    * does not succeed, then a second lookup using the full keyword text is
    * used.  This approach allows unreserved keywords to be used as user-
    * defined names that hide built-in names if they exist.  However the
    * built-in names are still found if there is no user-defined name.
    * <p>
    * This usage of the full keyword text for lookup of built-ins is required
    * since only the full text definition exists in the variable dictionary, 
    * rather than the full set of abbreviations.  Instead the abbreviation
    * matching is handled by the keyword dictionary and this method is aware
    * of the difference.      
    *
    * @param   dict
    *          <code>ScopedSymbolDictionary</code> instance to search.
    * @param   actual
    *          Text string to match.
    * @return  The label's token type that was found or -1 if no match was
    *          found.
    * @see     #lookupVariable
    * @see     #lookupFunction
    */
   private int lookupWorker(ScopedSymbolDictionary dict, String actual)
   {
      Keyword kw    = lookupKeyword(actual);
      int     type  = -1;
      
      if (kw != null)
      {
         // the actual text matches a keyword
         if (kw.isReserved())
         {
            // since this is a reserved keyword, this must be a 
            // built-in variable/function
            type = lookupWorkerExact(dict, kw.getFullText());
         }
         else
         {
            // this is an unreserved keyword, this MAY be a 
            // built-in variable/function BUT we must allow a user-defined
            // name to take precedence (note that if the actual text is
            // is the same as the full keyword text, then we may still
            // get a match here that is the built-in, assuming no
            // user-defined name exists --> that is OK too)
            type = lookupWorkerExact(dict, actual);
            
            // check if there was a match to the actual text
            if (type == -1)
            {
               // match the full text of the keyword with the built-in
               type = lookupWorkerExact(dict, kw.getFullText());
            }
         }
      }
      else
      {
         // there is no keyword, so this must be a user-defined name
         type = lookupWorkerExact(dict, actual);
      }
      
      return type;
   }
   
   /**
    * Primary worker routine for processing searches on dictionaries based on
    * the {@link com.goldencode.p2j.util.ScopedSymbolDictionary} class. All 
    * searches are for a symbol exactly matching the passed key.  The token
    * type of the match is returned.  Note that a token type of -1 is invalid 
    * in the Lexer and Parser, so this value is used to denote a search that
    * found no match.  No abbreviation processing will be provided!
    *
    * @param   dict
    *          <code>ScopedSymbolDictionary</code> instance to search.
    * @param   name
    *          Exact text string to match.
    * @return  The label's token type that was found or -1 if no match was
    *          found.
    */
   private int lookupWorkerExact(ScopedSymbolDictionary dict, String name)   
   {
      int     type   = -1;
      Integer result = (Integer) dict.lookupSymbol(name);
      
      if (result != null)
      {
         type = result.intValue();
      }
      
      return type;
   }
   
   /**
    * Centralizes the process of setting an file-unique temporary index for
    * each new wrapped object and then adding that object to the passed-in
    * dictionary.
    *
    * @param    ssd
    *           Symbol dictionary.
    * @param    global
    *           Add to global scope if <code>true</code>.
    * @param    name
    *           Variable name to add.
    * @param    wrapped
    *           Wrappered token object to add.
    */
   private void addWrappedWorker(ScopedSymbolDictionary ssd,
                                 boolean                global,
                                 String                 name,
                                 TokenDataWrapper       wrapped)
   {
      wrapped.setTempIndex(tempIdx);
      tempIdx++;
      ssd.addSymbol(global, name, wrapped);
   }
   
   /**
    * Given an AST which represents a field or variable which possibly has
    * an extent, check for a subscript and return it.  The assumed structure
    * of the AST, if it does have an extent subscript is:
    * <pre>
    *    FIELD/VAR
    *      |
    *      +--LBRACKET
    *           |
    *           +--EXPRESSION
    *                |
    *                +--NUM_LITERAL
    * </pre>
    * where the <code>FIELD/VAR</code> node is the <code>ast</code> parameter
    * passed to this method.
    * <p>
    * Currently, only the <code>NUM_LITERAL</code> token type is supported
    * for the subscript value.  Any other type of expression will raise an
    * exception.
    *
    * @param   ast
    *          Field or variable AST.
    *
    * @return  The 1-based subscript value from the <code>NUM_LITERAL</code>
    *          node depicted above, or 0 if there is no subscript.
    *
    * @throws  UnsupportedOperationException
    *          if a token type other than a simple <code>NUM_LITERAL</code>
    *          is provided for the subscript expression.
    * @throws  IllegalArgumentException
    *          if the <code>NUM_LITERAL</code> token cannot be parsed into an
    *          integral value.
    */
   private int getExtentSubscript(Aast ast)
   {
      int subscript = 0;
      
      if (ast.getFirstChild() != null)
      {
         Aast lbracket = ast.getImmediateChild(LBRACKET, null);
         if (lbracket != null)
         {
            Aast expr = lbracket.getChildAt(0);
            Aast numlit = expr.getImmediateChild(NUM_LITERAL, null);
            
            if (numlit == null)
            {
               String err = "Non-NUM_LITERAL extent value not supported " +
                            "in this context!";
               throw new UnsupportedOperationException(err);
            }
            
            try
            {
               subscript = Integer.parseInt(numlit.getText());
            }
            catch (NumberFormatException exc)
            {
               throw new IllegalArgumentException(
                  "Invalid extent subscript:  " + numlit.getText());
            }
         }
      }
      
      return subscript;
   }
   
   /**
    * Convert any "." characters (in a Progress qualified class or interface
    * name) to the current system's file name separator string.
    *
    * @param    name
    *           Qualified class or interface name.
    *
    * @return   Name that is safe to use with the current file system.
    */
   private String convertSeparator(String name)
   {
      return name.replaceAll("\\.", File.separator); 
   }
   
   /**
    * Search the PROPATH for a specific class or interface. The given name
    * should be a Progress style qualified name (package + class). This will
    * be converted to a file system name and searched for.
    *
    * @param    name
    *           Qualified class or interface name.
    * @param    type
    *           The type of the using clause that is being checked. This
    *           changes the search order and may drop some paths from the
    *           search when possible.
    * @param    result
    *           Object in which to store the search results (the file system
    *           name, the built-in flag and the .NET flag) when the search
    *           is successful. If nothing is found, this object is unchanged.
    *
    * @return   <code>true</code> if the class file was found.
    */
   private boolean findFile(String name, UsingType type, SearchResult result)
   {
      // we tried an approach based on FileSystemDaemon.search() but it is
      // slow, so we switched to a preloaded in-memory cache approach
      
      name = convertSeparator(name);
      
      Map<String, String>[] searchList = null;
      
      boolean[] builtin   = new boolean[4];
      boolean[] dotnet    = new boolean[4];
      
      // 4GL always matches case-insensitively (even on case-sensitive file system)
      // the code below will differentiate using this var
      boolean[] honorCase = new boolean[4];
      
      // choose the search paths based on the type
      switch (type)
      {
         case PROPATH:
            // no need to search .NET stuff
            searchList    = new HashMap[2];
            searchList[0] = locate().propathCls;   
            searchList[1] = locate().oo4glCls;
            
            builtin[0] = false; dotnet[0] = false; honorCase[0] = false;
            builtin[1] = true;  dotnet[1] = false; honorCase[1] = false;
            break;
         case DOTNET:
            // the documentation says there is no need to search PROPATH but
            // working code has found Progress.* using stmts with the FROM
            // ASSEMBLY option, so we add the OO4GL stuff in as a last resort
            // TODO: should this be more general?
            searchList    = new HashMap[3];
            searchList[0] = locate().dotnetCls;    
            searchList[1] = locate().assemblyCls;  
            searchList[2] = locate().oo4glCls;     
            
            builtin[0] = true;  dotnet[0] = true;  honorCase[0] = caseSens;
            builtin[1] = false; dotnet[1] = true;  honorCase[1] = caseSens;
            builtin[2] = true;  dotnet[2] = false; honorCase[2] = false;
            break;
         default:
            // search everything!
            searchList    = new HashMap[4];
            searchList[0] = locate().propathCls;
            searchList[1] = locate().oo4glCls;
            searchList[2] = locate().dotnetCls;
            searchList[3] = locate().assemblyCls;
            
            builtin[0] = false; dotnet[0] = false; honorCase[0] = false;
            builtin[1] = true;  dotnet[1] = false; honorCase[1] = false;
            builtin[2] = true;  dotnet[2] = true;  honorCase[2] = caseSens;
            builtin[3] = false; dotnet[3] = true;  honorCase[3] = caseSens;
            break;
      }
      
      for (int i = 0; i < searchList.length; i++)
      {
         Map<String, String> map  = searchList[i];
         String              srch = honorCase[i] ? name : name.toLowerCase();
         
         result.filename = map.get(srch);
         
         if (result.filename != null)
         {
            result.builtin = builtin[i];
            result.dotnet  = dotnet[i];
            return true;
         }
      }                                          
      
      return false;
   }
   
   /**
    * Test if the given package exists in the PROPATH, or in the 4GL system
    * packages or in the .NET system namespaces.
    *
    * @param    pkg
    *           Package name to search for.
    *
    * @return   <code>true</code> if the package exists in the PROPATH.
    */
   private boolean pkgExists(String pkg)
   {
      pkg = convertSeparator(pkg);
      
      File file = null;
      
      for (int i = 0; i < propath.length; i++)
      {
         if (propath[i].length() == 0 || ".".equals(propath[i]))
         {
            // check a relative file in the current directory
            file = new File(pkg);
         }
         else
         {
            // check a relative file in the specified path
            file = new File(propath[i], pkg);
         }
         
         // 4GL classes are always matched case-insensitively (even on case-sensitive
         // file systems
         if (exists(file, false))
         {
            return true;
         }
      }
      
      // check if the package exists as part of the 4GL system
      file = new File(oo4glPath, pkg);
      
      // 4GL classes are always matched case-insensitively (even on case-sensitive file systems
      if (exists(file, false))
      {
         return true;
      }
      
      // check if the namespace for the system classes in .NET
      file = new File(dotnetPath, pkg);
      
      if (exists(file, caseSens))
      {
         return true;
      }
      
      // check if the namespace for the user assemblies in .NET
      file = new File(ASSEMBLY_PATH, pkg);
      
      if (exists(file, caseSens))
      {
         return true;
      }
      
      return false;
   }
   
   /**
    * Implements a check for the file's existence while honoring the 
    * case-sensitive file-system flag.  If the original file-system was
    * case-sensitive, then a simple <code>exists()</code> method call is
    * made on the file object.  Otherwise a case-insensitive search is
    * undertaken.
    *
    * @param    file
    *           The file to check.
    * @param    caseSens
    *           <code>true</code> if the legacy system had a case-sensitive
    *           file-system.
    *
    * @return   <code>true</code> if the file object exists.
    */
   private boolean exists(File file, boolean caseSens)
   {
      // quick out: does the exact match exist?
      if (file.exists())
      {
         return true;
      }
      
      if (!caseSens)
      {
         // TODO: this is broken!!!!  It was found to fail when the exact case-sensitive match
         //       actually existed (resolved by checking for that first, above).  It is not
         //       known if it fails in other cases too.
         return FileSystemDaemon.getCaseInsensitiveMatch(file, true) != null;
      }
      
      return false;
   }

   /**
    * Map the keyword type to the corresponding field type.
    *
    * @param    ktype
    *           Keyword type.
    *
    * @return   The matching field type or -1 if the input is invalid.
    */
   private int keywordToFieldType(int ktype)
   {
      int ftype = -1;
      
      switch (ktype)
      {
      case KW_BLOB:
         ftype = FIELD_BLOB;
         break;
      case KW_CHAR:
         ftype = FIELD_CHAR;
         break;
      case KW_CLASS:
      case CLASS_NAME:
         ftype = FIELD_CLASS;
         break;
      case KW_CLOB:
         ftype = FIELD_CLOB;
         break;
      case KW_COM_HNDL:
         ftype = FIELD_COM_HANDLE;
         break;
      case KW_DATE:
         ftype = FIELD_DATE;
         break;
      case KW_DATETIME:
         ftype = FIELD_DATETIME;
         break;
      case KW_DATE_TZ:
         ftype = FIELD_DATETIME_TZ;
         break;
      case KW_DEC:
         ftype = FIELD_DEC;
         break;
      case KW_HANDLE:
         ftype = FIELD_HANDLE;
         break;
      case KW_INT:
         ftype = FIELD_INT;
         break;
      case KW_INT64:
         ftype = FIELD_INT64;
         break;
      case KW_LOGICAL:
         ftype = FIELD_LOGICAL;
         break;
      case KW_RAW:
         ftype = FIELD_RAW;
         break;
      case KW_RECID:
         ftype = FIELD_RECID;
         break;
      case KW_ROWID:
         ftype = FIELD_ROWID;
         break;
      case KW_WID_HAND:
         ftype = FIELD_HANDLE;
         break;
      }
      
      return ftype;
   }
   
   /**
    * Process a request for the schema dictionary.  If that request cannot be satisfied from
    * the current schema dictionary instance AND the current scope is processing a class, then
    * the schema dictionaries of the parent hierarchy will be checked until the request is
    * satisfied or there are no more parents in the inheritance chain.
    *
    * @param    worker
    *           The core processing to be executed on the given schema dictionary.
    * @param    cont
    *           A condition which if it returns <code>true</code>, then the next schema
    *           dictionary should be tried.  On <code>false</code>, the current result is
    *           returned.
    *
    * @return   The result of the processing.
    */
   private <V> V processHierarchy(SchemaHelper<V> worker, Predicate<V> cont)
   throws SchemaException
   {
      SchemaException[] exc = new SchemaException[1];
      
      Supplier<V> check = () -> 
      {
         V result = null;
         try
         {
            result = worker.process(schemaDict);
            ClassDefinition def = getCurrentClassDef();
            
            if (def != null)
            {
               // process as long as the condition is true
               while (cont.test(result))
               {
                  // start at the parent, then move up to the parent's parent with each iteration
                  def = def.getParent();
                  
                  if (def == null)
                  {
                     // no more classes in the inheritence hierarchy, we're done
                     break;
                  }
                  
                  SchemaDictionary dict = def.getSchemaDict();
                  
                  // only test this class if there is something to test
                  if (dict != null)
                  {
                     // try again
                     result = worker.process(dict);
                  }
               }
            }
         }
         catch (SchemaException e)
         {
            exc[0] = e;
         }

         return result;
      };
      
      V result = null;
      
      if (ignoreOORefs)
      {
         try
         {
            ignoreOORefs = false;
            
            result = check.get();
            
            if (exc[0] != null)
            {
               throw exc[0];
            }
         }
         finally
         {
            ignoreOORefs = true;
         }
      }
      
      if (result == null || cont.test(result))
      {
         result = check.get();
         
         if (exc[0] != null)
         {
            throw exc[0];
         }
      }
      
      return result;
   }
   
   /**
    * Return the access mode as encoded by the AST type.
    *
    * @param    am
    *           The access mode node.
    *
    * @return   The type of the node or KW_PRIVATE if there is no node.
    */
   private int decodeAccessMode(Aast am)
   {
      return (am != null) ? am.getType() : KW_PRIVATE;
   }
   
   /**
    * Return the static mode as encoded by the AST type.
    *
    * @param    st
    *           The static mode node.
    *
    * @return   <code>true</code> if the type of the node KW_STATIC.
    */
   private boolean decodeStatic(Aast st)
   {
      return (st != null && st.getType() == KW_STATIC);
   }
   
   /**
    * Calculates the access mode for method and data member searches of
    * class definitions, based on the presence and contents of the given
    * class name and the current context. By default the access mode will
    * be <code>KW_PUBLIC</code>. If this request is made within the context
    * of a class definition, then the access mode will be 
    * <code>KW_PRIVATE</code> if <code>cname</code> is <code>null</code> or
    * if <code>cname</code> is the same as the current class definition
    * name. The access mode will be <code>KW_PROTECTD</code> if the
    * <code>cname</code> is not <code>null</code> and it is the equivalent
    * of one of the parent class names in the current class definition's
    * inheritance chain.
    *
    * @param    cname
    *           Fully qualified class name being searched or <code>null</code>
    *           if the current class definition is to be searched.
    *
    * @return   The calculated access mode of <code>KW_PUBLIC</code>,
    *           <code>KW_PROTECTD</code> or <code>KW_PRIVATE</code>.
    */
   private int calcAccessMode(String cname)
   {
      int access = KW_PUBLIC;
      
      ClassDefinition cls = getCurrentClassDef();
      
      if (cls != null)
      {
         if (cname != null)
         {
            String current = cls.getName();
            
            if (cname.equalsIgnoreCase(current))
            {
               // inside a class def and explicitly referencing the current
               // class, this is a private search
               access = KW_PRIVATE;
            }
            else
            {
               ClassDefinition parent = cls.getParent();
               
               // search the parent hierarchy for a matching class name
               while (parent != null)
               {
                  String pname = parent.getName();
                  
                  if (cname.equalsIgnoreCase(pname))
                  {
                     // this is explicit access to a parent class from within
                     // a subclass definition, this is a protected search
                     access = KW_PROTECTD;
                     break;
                  }
                  
                  parent = parent.getParent();
               }
            }
         }
         else
         {
            // we are inside a class def and have no explicit class to
            // search, this is a private search of the current class'
            // hierarchy
            access = KW_PRIVATE;
         }
      }
      
      return access;
   }
   
   /**
    * Describes the type of the USING specification.
    */
   private static enum UsingType
   {
      PROPATH, DOTNET, UNSPECIFIED
   };
   
   /**
    * Functional interface for use with lambdas.
    */
   private interface SchemaHelper<V>
   {
      /**
       * Execute the logic against the given instance of the schema dictionary.
       *
       * @param    dict
       *           The schema dictionary.
       *
       * @return   The resulting value generated.
       */
      public V process(SchemaDictionary dict)
      throws SchemaException;
   }
   
   /**
    * Store the data in a USING statement.
    */
   private static class UsingSpec
   {
      /** Match specification (package/.NET namespace or full classname). */
      private String spec = null;
      
      /** Match type. */
      private UsingType type = UsingType.UNSPECIFIED;
      
      /**
       * Primary constructor.
       *
       * @param    spec
       *           Match specification (package/.NET namespace or full
       *           classname).
       * @param    type
       *           Match type.
       */
      public UsingSpec(String spec, UsingType type)
      {
         this.spec = spec;
         this.type = type;
      }
   }
   
   /**
    * Store the data resulting from a given file system search.
    */
   private static class SearchResult
   {
      /** File system object that was found (includes pathing). */
      private String filename = null;
      
      /** Fully qualified class or interface name relative to PROPATH. */
      private String clsname = null;
      
      /** Flag marking if this is a built-in Progress class. */
      private boolean builtin = false;
      
      /** Flag to differentiate between 4GL classes and .NET classes. */
      private boolean dotnet = false;
      
      /** Search hint. */
      private UsingType type = UsingType.UNSPECIFIED;
   }
   
   /**
    * Container with context-local data.
    */
   private static class WorkArea
   {
      /** Stores class and interface definitions. */
      private ScopedSymbolDictionary classDict = null;
      
      /** Classes in the propath, mapped qualified classname to filename. */
      private Map<String, String> propathCls = null;
      
      /** Classes in the OO4GL_PATH, mapped qualified classname to filename. */
      private Map<String, String> oo4glCls = null;
      
      /** Classes in the DOTNET_PATH, mapped qualified classname to filename. */
      private Map<String, String> dotnetCls = null;
      
      /** Classes in the ASSEMBLY_PATH, mapped qualified classname to filename. */
      private Map<String, String> assemblyCls = null;
      
      /** The most recently used propath for init of the propath class map. */
      private String[] lastPropath = null;
      
      /**
       * The nesting level of {@link SymbolResolver#loadClass} calls when performing 1st-level 
       * load of a class. 
       */
      private int preScanPass = 0;
      
      /** The set of class/interface references processed during 1st level parse of a class. */
      private Set<String> preScanRefs = new HashSet<>(); 
  }
}
