/*
** Module   : BaseSysTable.java
** Abstract : common code for COLOR-TABLE and FONT-TABLE implementation.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.

**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 MAG 20140827 Created initial version.
** 002 MAG 20140906 Added environment notification action.
** 003 CA  20150317 When resolving an entry, allow default values to be specified, if the entry 
**                  is not explicitly set. 
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.util.*;
import java.util.logging.*;

import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.Utils.DirScope;

/**
 * Common code for COLOR-TABLE and FONT-TABLE implementation.
 */
public abstract class BaseSysTable
{
   /** logger */
   private static final Logger LOG = LogHelper.getLogger(BaseSysTable.class.getName());

   /**
    * Read the entire table. If a custom environment is set the table is loaded from a INI file
    * otherwise the table is loaded from server directory. 
    * 
    * @param   node
    *          Server directory node.
    * @param   section
    *          INI file section.
    * @param   key
    *          INI file key.
    *          
    * @return  The table.  As 4GL does, this is 0-indexed.
    */
   static String[] readTable(String node, String section, String key)
   {
      return read(node, section, key, EnvironmentOps.isCustomEnvironment(), null);
   }

   /**
    * Read the entire table. If a custom environment is set the table is loaded from a INI file
    * otherwise the table is loaded from server directory. 
    * 
    * @param   node
    *          Server directory node.
    * @param   section
    *          INI file section.
    * @param   key
    *          INI file key.
    * @param   defaults
    *          The mapping of keys to their standard values.
    *          
    * @return  The table.  As 4GL does, this is 0-indexed.
    */
   static String[] readTable(String node, String section, String key, Map<String, String> defaults)
   {
      return read(node, section, key, EnvironmentOps.isCustomEnvironment(), defaults);
   }

   /**
    * Read the entire table. If a custom environment is set the table is loaded from a INI file
    * otherwise the table is loaded from server directory. 
    * 
    * @param   node
    *          Server directory node.
    * @param   section
    *          INI file section.
    * @param   key
    *          INI file key.
    * @param   customEnv
    *          <code>true</code> if load from custom environment.
    *          
    * @return  The table.  As 4GL does, this is 0-indexed.
    */
   static String[] read(String node, String section, String key, boolean customEnv)
   {
      return read(node, section, key, customEnv, null);
   }
   
   /**
    * Read the entire table. If a custom environment is set the table is loaded from a INI file
    * otherwise the table is loaded from server directory. 
    * 
    * @param   node
    *          Server directory node.
    * @param   section
    *          INI file section.
    * @param   key
    *          INI file key.
    * @param   customEnv
    *          <code>true</code> if load from custom environment.
    * @param   defaults
    *          The mapping of keys to their standard values.
    *          
    * @return  The table.  As 4GL does, this is 0-indexed.
    */
   static String[] read(String              node, 
                        String              section, 
                        String              key, 
                        boolean             customEnv, 
                        Map<String, String> defaults)
   {
      // use an explicit instance, to avoid binding/unbinding for each search
      DirectoryService ds = DirectoryService.getInstance();

      if (!ds.bind() && LOG.isLoggable(Level.SEVERE))
      {
         LOG.severe("BaseSysTable.readTable - could not bind directory.");
         return new String[0];
      }
      
      // items collection
      List<String> items = new ArrayList<>();
      
      // read the entire font table
      int idx = 0;
      while (true)
      {
         String itemKey = key + idx++;
         String defVal = (defaults == null ? null : defaults.get(itemKey));
         String item = readValue(ds, customEnv, node, section, itemKey, defVal);

         if (item == null || item.trim().isEmpty())
         {
            // not found, the font-table reading is finished
            break;
         }

         // add item to collection
         items.add(item);
      }
      
      if (!ds.unbind() && LOG.isLoggable(Level.SEVERE))
      {
         LOG.severe("BaseSysTable.readTable - could not unbind directory.");
      }

      return items.toArray(new String[0]);
   }

   /**
    * Get a value from server directory or from a custom environment for a given key.
    *  
    * @param   node
    *          Server directory node.
    * @param   section
    *          INI file section.
    * @param   key
    *          INI file key.
    * 
    * @return  see above.
    */
   static String readValue(String node, String section, String key)
   {
      return readValue(null, EnvironmentOps.isCustomEnvironment(), node, section, key, null);
   }

   /**
    * Get a value from server directory or from a custom environment for a given key.
    *  
    * @param   node
    *          Server directory node.
    * @param   section
    *          INI file section.
    * @param   key
    *          INI file key.
    * @param   defVal
    *          The default value for this mapping.
    * 
    * @return  see above.
    */
   static String readValue(String node, String section, String key, String defVal)
   {
      return readValue(null, EnvironmentOps.isCustomEnvironment(), node, section, key, defVal);
   }

   /**
    * Get a value from server directory or from a custom environment for a given key.
    *  
    * @param   ds
    *          Directory service instance.
    * @param   customEnv
    *          Indicate to read from server directory or from a custom environment. 
    * @param   node
    *          Server directory node.
    * @param   section
    *          INI file section.
    * @param   key
    *          INI file key.
    * @param   defVal
    *          The default value for this mapping.
    * 
    * @return  see above.
    */
   static String readValue(DirectoryService ds,
                           boolean          customEnv,
                           String           node,
                           String           section,
                           String           key,
                           String           defVal)
   {
      String item = null;
         
      if (customEnv)
      {
         // if there is an explicitly loaded environment, use that
         character chItem = EnvironmentOps.getKeyValue(section, key);
         item = chItem.isUnknown() ? null : chItem.toStringMessage();
         if (item == null)
         {
            item = defVal;
         }
      }
      else
      {
         // otherwise, fallback to the directory configuration
         String dirNode = node + "/" + key;
         item = Utils.getDirectoryNodeString(ds, dirNode, defVal, DirScope.BOTH, null);
      }
      
      return item;
   }
  
   /**
    * Call when a new window is created.
    * 
    * @param   window
    *          Window configuration.
    */
   public static void registerWindow(WindowConfig window)
   {
      ColorTable.registerWindow(window);
   }
   
   /**
    * Set current environment name. A notification from environment API whenever an 
    * <code>USE environment</code> statement is executed.
    * 
    * @param   envName
    *          Environment name.
    */
   public static void use(String envName)
   {
      ColorTable.use(envName);
   }

   /**
    * Unload environment. A notification from environment API whenever an 
    * <code>UNLOAD environment</code> statement is executed.
    * 
    * @param   envName
    *          Environment name.
    */
   public static void unload(String envName)
   {
      ColorTable.unload(envName);
   }
}