/*
** Module   : ClientConfigManager.java
** Abstract : Management of widget configurations, on client-side.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140929 Created initial version.
** 002 CA  20141014 Separated data which needs to survive the appserver agent's context reset into 
**                  a different container.
** 003 HC  20141106 Javadocs fixes.
** 004 CA  20141125 Added API to deactivate tracking of the changed configs.
** 005 HC  20141207 Added Widget.syncWithConfig(), so far limited to window widget and only some
**                  cases.
** 006 CA  20150105 Changes for SIDE-LABEL-HANDLE runtime support.
** 007 VIG 20150122 Added Menu components configs.
** 008 HC  20150323 Removed business logic from config classes.
** 009 HC  20150913 Removed ScrollPaneConfig class.
** 010 EVL 20160127 Added SliderConfig class.
** 011 GES 20160727 Replaced reflection usage with MethodHandle which is faster.
** 012 OM  20170530 Added ControlFrameConfig to config list.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.lang.reflect.*;
import java.lang.invoke.*;
import java.util.*;

import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Management of widget configurations, on client-side.
*/
public class ClientConfigManager
extends ConfigManager
implements WidgetStateListener
{
   /** This list contains all config classes which implement {@link WidgetConfig}.*/
   private static final List<Class<? extends WidgetConfig>> CONFIG_CLASSES;
   
   /**
    * Flag indicating when the client-side can start pushing updates.  The notification to 
    * start pushing updates will come from the server-side, via 
    * {@link ClientExports#activateConfigUpdates()}.
    */
   private boolean canPushUpdates = false;
   
   /** Cache of the {@link ThinClient} instance. */
   private final ThinClient tc;
   
   static
   {
      // add to this list all config classes which implement WidgetConfig - the entire hierarchy 
      // of classes must be added in this list
      List<Class<? extends WidgetConfig>> configs = new ArrayList<>();
      configs.add(BaseConfig.class);
      configs.add(BrowseConfig.class);
      configs.add(BrowseColumnConfig.class);
      configs.add(ButtonConfig.class);
      configs.add(ComboBoxConfig.class);
      configs.add(ControlConfig.class);
      configs.add(ControlSetConfig.class);
      configs.add(ControlTextConfig.class);
      configs.add(EditorConfig.class);
      configs.add(FieldGroupConfig.class);
      configs.add(FillInConfig.class);
      configs.add(FrameConfig.class);
      configs.add(MenuConfig.class);
      configs.add(MenuItemConfig.class);
      configs.add(ImageConfig.class);
      configs.add(LabelConfig.class);
      configs.add(PaneConfig.class);
      configs.add(RadioSetConfig.class);
      configs.add(RectangleConfig.class);
      configs.add(SelectionListConfig.class);
      configs.add(SkipConfig.class);
      configs.add(SliderConfig.class);
      configs.add(SubMenuConfig.class);
      configs.add(TextConfig.class);
      configs.add(ToggleBoxConfig.class);
      configs.add(WidgetConfig.class);
      configs.add(WindowConfig.class);
      configs.add(ControlFrameConfig.class);
      
      CONFIG_CLASSES = Collections.unmodifiableList(configs);
   }

   /**
    * Initialize the client-side by collecting all widget configuration classes (registered in 
    * the {@link ClientConfigManager#CONFIG_CLASSES}) and store them in 
    * the {@link ConfigManager.WorkArea#cfgDefs} map.
    */
   public ClientConfigManager()
   {
      WorkArea wa = survivals.get();
      
      // walk the widget config classes and resolve their configuration
      for (Class<? extends WidgetConfig> cls : CONFIG_CLASSES)
      {
         WidgetConfigDef cfgDef = new WidgetConfigDef(cls);
         
         wa.cfgDefs.put(cls, cfgDef);
      }
      
      this.tc = ThinClient.getInstance();
   }

   /**
    * On client-side, send the widget configurations to the specified server.
    * 
    * @param    server
    *           The server where to  send the configs.
    *           
    * @see      ServerExports#initWidgetConfigurations(Map)
    */
   public void pushConfigDefs(ServerExports server)
   {
      WorkArea wa = survivals.get();
      
      // this is a one-time setup, so we can afford using a map. 
      // key: the config class name
      // value: the list of instance field names
      Map<String, List<String>> configs = new HashMap<>();
      for (WidgetConfigDef cfg : wa.cfgDefs.values())
      {
         String clsName = cfg.cfgCls.getName();
         
         configs.put(clsName, cfg.fieldNames);
      }
      
      server.initWidgetConfigurations(configs);
   }

   /**
    * Inform the client-side that it can no longer send config updates.
    */
   public void deactivateConfigUpdates()
   {
      canPushUpdates = false;
   }
   
   /**
    * Inform the client-side that it can start sending the config updates.
    */
   public void activateConfigUpdates()
   {
      canPushUpdates = true;
   }

   /**
    * Get the widget configuration updates, by comparing the active with the backup configuration.
    * 
    * @return    A list of widget configuration updates to be sent to the server side.
    */
   public WidgetConfigUpdates[] getConfigUpdates()
   {
      // do not send config updates until the server-side is fully initialized.
      if (!canPushUpdates)
         return null;

      return super.getConfigUpdates();
   }

   /**
    * Notify the given widget has been destroyed.  This includes removing its ID from the
    * maps.
    * 
    * @param    w
    *           The deleted widget.
    */
   @Override
   public void notifyWidgetDestroyed(Widget<? extends OutputManager<?>> w)
   {
      WidgetConfig cfg = w.config();
      if (cfg == null)
      {
         // not all widgets have a config
         return;
      }

      WidgetId wid = cfg.id;
      
      notifyWidgetDestroyed(wid);
   }
   
   /**
    * Check if the given ID is associated with a registered widget.
    * 
    * @param    id
    *           The widget ID.
    *           
    * @return   <code>true</code> if a widget exists in the registry, with the given ID.
    */
   protected boolean isRegisteredWidget(WidgetId id)
   {
      return tc.getWidget(id.asInt()) != null;
   }

   /**
    * The method attempts to resolve a {@link ConfigOwner} reference from
    * the widget <code>widgetId</code>.
    * 
    * @param    widgetId
    *           Widget Id.
    *           
    * @return   A {@link ConfigOwner} reference or <code>null</code> if
    *           such cannot be find. 
    */
   @Override
   ConfigOwner<?> resolveConfigOwner(WidgetId widgetId)
   {
      Widget<?> widget = tc.getWidget(widgetId);
      if (widget instanceof ConfigOwner)
      {
         return (ConfigOwner<?>) widget;
      }
      
      return null;
   }

   /**
    * Switch the widget's ID from the old one to the new one.  The references remain unchanged.
    * 
    * @param    oldId
    *           The old widget ID, currently in use by the manager.
    * @param    newId
    *           The new widget ID.
    */
   public void switchId(WidgetId oldId, WidgetId newId)
   {
      switchId(oldId, newId, widgetCfg);
      
      activeConfigs.get(oldId).id = newId;
      switchId(oldId, newId, activeConfigs);
      
      backupConfigs.get(oldId).id = newId;
      switchId(oldId, newId, backupConfigs);
      
      switchId(oldId, newId, dirtyConfigs);
   }
   
   /**
    * Mark all differences between the backup config and the given widget configuration as being
    * "dirty".
    * 
    * @param    cfg
    *           The configuration from which the changes are extracted.
    */
   public void trackAllChanges(WidgetConfig cfg)
   {
      WidgetConfig backup = backupConfigs.get(cfg.id);
      
      WidgetConfigDef wdef = widgetCfg.get(cfg.id);
      
      for (int fid = 0; fid < wdef.fieldCount; fid++)
      {
         String fname = wdef.fieldNames.get(fid);

         try
         {
            Object activeVal = wdef.getters[fid].invokeExact((Object)cfg);
            Object backupVal = wdef.getters[fid].invokeExact((Object)backup);

            boolean changed = !((activeVal == null && backupVal == null) ||
                                (activeVal != null && activeVal.equals(backupVal)));
            
            if (changed)
            {
               addDirtyConfig(cfg, fname);
            }
         }
         catch (Throwable t)
         {
            final String msg =
               "Could not resolve the active/backup values for widget config %s in class %s.";

            throw new RuntimeException(String.format(msg, fname, cfg.getClass().getName()), 
                                                     t);
         }
      }
   }
   
   /**
    * Helper method to replace the entry in the given map, to use the new key. 
    * 
    * @param    oldId
    *           The old widget ID, which has some state registered in the given map.
    * @param    newId
    *           The new widget ID, to which the state needs to be re-put.
    * @param    map
    *           The map which needs to be updated.
    */
   @SuppressWarnings("unchecked")
   private void switchId(WidgetId oldId, WidgetId newId, Map map)
   {
      Object value = map.remove(oldId);
      map.put(newId, value);
   }
}
