/*
** Module   : ClientState.java
** Abstract : contains data to synchronize the server with the client's state
**
** Copyright (c) 2008-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description-----------------------------
** 001 GES 20080613   @38878 Created initial version to contain a simple set of client-side state.
** 002 SIY 20081209   @40856 Added list of IDs of wrapped frames.
** 003 CA  20090413   @41772 Added tracking of changed VISIBLE/HIDDEN  widget attributes. When 
**                           these attributes get changed on client-side, they are sent to the
**                           server for update.
** 004 VIG 20131225          Added function key to client state.
** 005 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI concrete
**                           implementation for the same widget ID.  Refactored the widget 
**                           configuration classes: all fields were made public; these classes 
**                           need to be as dumb as possible, all logic related to setting/getting 
**                           a certain field should be at the widget or at the caller. Refs #2254
** 006 CA  20141021          Added a way to inform the server-side about a virtual LAST-KEY 
**                           generated by client-side from a non-keyboard device (i.e. event from 
**                           window or mouse action).
** 007 CA  20141119          Refactored drawing code to enable multiple OS windows, for GUI
**                           support.  Misc changes/fixes related to drawing on multiple windows.
** 008 CA  20160327          Added support for LAST-EVENT:X and LAST-EVENT:Y attributes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.io.*;

/**
 * This class is a container for client-side state.
 */
public class ClientState
implements Externalizable
{
   /** List of frame IDs to deregister. */
   public int[] deregister = null;
   
   /** Frame DOWN info. */
   public int[] down = null;
   
   /** IDs of DOWN frames.*/
   public int[] downIDs = null;

   /** IDs of wrapped down frames. */
   public int[] wrapped = null;
   
   /** Flag to indicate if there was a pause during the last operation. */
   public boolean hadPause = false;

   /**
    * The code of last event (LAST-EVENT:FUNCTION)
    * <p>
    * If this the type of this field will be "int" there will be following 
    * conflict: not initialized value will be 0, but at the same time
    * 0 - is a valid code of the key (Ctrl-@). The Integer class is used here
    * to distinguish uninitialized value (null) and 0 (valid function code) 
    */
   public Integer functionKey;
   
   /** The value of the LAST-EVENT:X attribute. */
   public Integer eventX;
   
   /** The value of the LAST-EVENT:Y attribute. */
   public Integer eventY;
   
   /** A last-key code which was not set via user input, from the keyboard. */
   public Integer virtualLastKey;
   
   /** The list of widget configuration updates incoming from the client-side. */
   public WidgetConfigUpdates[] configUpdates;
   
   /** The ID of the active window or -1 if unchanged. */
   public int activeWindowId;
   
   /**
    * Construct an empty instance.
    */
   public ClientState()
   {
   }
   
   /**
    * Replacement for the default object reading method.
    * 
    * @param    in
    *           Input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      deregister     = (int[]) in.readObject();
      down           = (int[]) in.readObject();
      downIDs        = (int[]) in.readObject();
      wrapped        = (int[]) in.readObject();
      hadPause       = in.readBoolean();
      functionKey    = (Integer) in.readObject();
      virtualLastKey = (Integer) in.readObject();

      int ulen = in.readInt();
      this.configUpdates = (ulen == 0 ? null : new WidgetConfigUpdates[ulen]);
      for (int i = 0; i < ulen; i++)
      {
         WidgetConfigUpdates cfg = new WidgetConfigUpdates();
         cfg.readExternal(in);

         this.configUpdates[i] = cfg;
      }
      
      activeWindowId  = in.readInt();
      eventX = (Integer) in.readObject();
      eventY = (Integer) in.readObject();
   }

   /**
    * Replacement for the default object writing method.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeObject(deregister);
      out.writeObject(down);
      out.writeObject(downIDs);
      out.writeObject(wrapped);
      out.writeBoolean(hadPause);
      out.writeObject(functionKey);
      out.writeObject(virtualLastKey);
      
      if (configUpdates != null)
      {
         out.writeInt(configUpdates.length);
         for (WidgetConfigUpdates cfg : configUpdates)
         {
            cfg.writeExternal(out);
         }
      }
      else
      {
         out.writeInt(0);
      }
      
      out.writeInt(activeWindowId);
      out.writeObject(eventX);
      out.writeObject(eventY);
   }
}