/*
** Module   : ColorRgb.java
** Abstract : a simple color structure containing RGB values
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.

**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 MAG 20140829 Created initial version.
** 002 MAG 20140911 Added asString method.
** 003 GES 20150424 Added fromString() convenience method and setGuiRgb() method.
** 004 IAS 20160331 Added default constructor and hashCode.
** 005 SVL 20170201 Improved equals().
** 006 OM  20170320 Added conversion constructors from awt Color.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.io.*;

/**
 * A simple color structure containing RGB values. 
 */
public class ColorRgb
implements Externalizable
{
   /** Red value */
   private int red;
   
   /** Green value */
   private int green;
   
   /** Blue value */
   private int blue;
   
   /** Color name */
   private String name;

   /**
    * Constructor.
    */
   public ColorRgb()
   {
      this(0,0,0);
   }

   /**
    * Constructor.
    * 
    * @param   red
    *          Red value.
    * @param   green
    *          Green value.
    * @param   blue
    *          Blue value.
    */
   public ColorRgb(int red, int green, int blue)
   {
      this.red = red;
      this.green = green;
      this.blue = blue;
   }

   /**
    * Constructor.
    * 
    * @param   red
    *          Red value.
    * @param   green
    *          Green value.
    * @param   blue
    *          Blue value.
    * @param   name
    *          Color name.
    */
   public ColorRgb(int red, int green, int blue, String name)
   {
      this(red, green, blue);
      this.name = name;
   }

   /**
    * Copy constructor.
    * 
    * @param   rgb
    *          ColorRgb copy.
    */
   public ColorRgb(ColorRgb rgb)
   {
      this(rgb.red, rgb.green, rgb.blue, rgb.name);
   }
   
   /**
    * Constructor. Creates a color identical with {@code color} parameter.
    *
    * @param   color
    *          An java awt color.
    */
   public ColorRgb(java.awt.Color color)
   {
      this(color.getRed(), color.getGreen(), color.getBlue());
   }
   
   /**
    * Constructor. Creates a color identical with {@code rgbColor} parameter.
    *
    * @param   rgbColor
    *          An integer that hold the color information in ARBG format. Alpha chanel is ignored.
    */
   public ColorRgb(int rgbColor)
   {
      // order is important: BGR!
      this.blue = rgbColor & 0xFF;
      rgbColor >>= 8;
      this.green = rgbColor & 0xFF;
      rgbColor >>= 8;
      this.red = rgbColor & 0xFF;
   }
   
   /**
    * Helper to take a Java compatible color definition and create a new instance of this
    * class.
    * 
    * @param    rgb
    *           Color in octal, hexadecimal (0xFFFFFF or #FFFFFF) or decimal format.  This uses
    *           <code>Integer.decode()</code> to create the integer that is then split into
    *           rgb values.
    * @param    def
    *           Value to copy if the string input is invalid.
    *
    * @return   The new instance.
    */
   public static ColorRgb fromString(String rgb, ColorRgb def)
   {
      ColorRgb result = new ColorRgb(def);
      
      try
      {
         int value = Integer.decode(rgb);
         
         result.setGuiRgb(value);
      }
      
      catch (NumberFormatException nfe)
      {
         // nothing to do, just return the previously created result
      }
      
      return result;
   }

   /**
    * Get red value.
    * 
    * @return  see above.
    */
   public int getRed()
   {
      return red;
   }

   /**
    * Set red value.
    * 
    * @param   red
    *          New red value
    */
   public void setRed(int red)
   {
      this.red = red;
   }

   /**
    * Get green value.
    * 
    * @return  see above.
    */
   public int getGreen()
   {
      return green;
   }

   /**
    * Set green value.
    * 
    * @param green
    *        New green value.
    */
   public void setGreen(int green)
   {
      this.green = green;
   }

   /**
    * Get blue value.
    * 
    * @return  see above.
    */
   public int getBlue()
   {
      return blue;
   }

   /**
    * Set blue value.
    * 
    * @param   blue
    *          New blue value.
    */
   public void setBlue(int blue)
   {
      this.blue = blue;
   }

   /**
    * Get a combination of red, green and blue value P4GL compliant.
    * 
    * @return  see above.
    */
   public int getRgb()
   {
      return this.red | (this.green << 8) | (this.blue << 16);
   }

   /**
    * Set color values form a RGB value using Progress 4GL parsing of the integer.
    * 
    * @param   rgb
    *          A combination of red, green and blue value.
    */
   public void setRgb(int rgb)
   {
      this.red   = rgb & 0xff;
      this.green = (rgb >>  8) & 0xff;
      this.blue  = (rgb >> 16) & 0xff;
   }

   /**
    * Get a combination of red, green and blue value Java compliant.
    * 
    * @return  see above.
    */
   public int getGuiRgb()
   {
      return this.blue | (this.green << 8) | (this.red << 16);
   }

   /**
    * Set color values form a RGB value using Java parsing of the integer.
    * 
    * @param   rgb
    *          A combination of red, green and blue value.
    */
   public void setGuiRgb(int rgb)
   {
      this.blue  = rgb & 0xff;
      this.green = (rgb >>  8) & 0xff;
      this.red   = (rgb >> 16) & 0xff;
   }

   /**
    * Reports if the given object is the same color as this instance. This does not just
    * compare references, but instead it compares the contained rgb values.
    *
    * @param    obj
    *           The object to compare.
    *
    * @return   <code>true</code> if the objects are equivalent, <code>false</code> if the given
    *           instance is <code>null</code>, is not an instance of this class or if its rgb
    *           value is not equivalent.
    */
   @Override
   public boolean equals(Object obj)
   {
      if (this == obj)
         return true;
      
      if (obj == null || !getClass().isAssignableFrom(obj.getClass()))
         return false;

      ColorRgb other = (ColorRgb) obj;

      return (red == other.red) && (green == other.green) && (blue == other.blue);
   }
   
   /**
    * Calculate hash code compatible with equals()
    * 
    * @return hash code
    */
   @Override
   public int hashCode()
   {
      final int prime = 31;
      int result = 1;
      result = prime * result + blue;
      result = prime * result + green;
      result = prime * result + red;
      
      return result;
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */   
   public void readExternal(ObjectInput in) 
   throws IOException, 
          ClassNotFoundException
   {
      red   = in.readInt();
      green = in.readInt();
      blue  = in.readInt();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out) 
   throws IOException
   {
      out.writeInt(red);
      out.writeInt(green);
      out.writeInt(blue);      
   }
   
   /**
    * Return a RGB as string comma separated.
    * 
    * @return  see above.
    */
   public String asString()
   {
      return name == null ? red + "," + green  + "," + blue : name;
   }
   
   /**
    * Return a string with the RGB value in "0xRRGGBB" format where the red, green and
    * blue values are each encoded as 2 hexadecimal digits.
    *
    * @return   The value as a hexadecimal string.
    */
   public String asHexString()
   {
      return String.format("0x%02x%02x%02x", red, green, blue);
   }
   
   /**
    * Get color name.
    * 
    * @return  see above.
    */
   protected String getName()
   {
      return name;
   }
   
   /**
    * Set color name.
    * 
    * @param   name
    *          Color name.
    */
   protected void setName(String name)
   {
      this.name = name;
   }
}