/*
** Module   : ColorSpec.java
** Abstract : Implements Progress color-phrase details.
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 NVS 20051026 NEW   @23274 Created initial version.
** 002 SIY 20080331 CHG   @37736 Improved support for COLOR_VALUE mode.
**                               Ref: #37732
** 003 NVS 20080420 CHG   @38047 Color value that encodes one of the regular
**                               color specifications MESSAGE/INPUT/NORMAL
**                               is not stored as value but converts into the
**                               corresponding specification right away.
** 004 SIY 20080429 CHG   @38143 Return invalid color if it is not present in
**                               color map.
** 005 SIY 20101122              Minor cleanups.
** 006 SIY 20101223              Complete rewrite due to change of color
**                               handling approach.
** 007 CS  20121031              Removed client specific references
** 008 HC  20140119              Implemented runtime support for the COLOR 
**                               protermcap-attribute.
** 009 MAG 20140820              Implements WINDOWS colors.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.util.*;

import org.apache.commons.logging.*;

import com.goldencode.p2j.util.*;

/**
 * Class ColorSpec supports the color-phrase specification.
 * <ul>
 *   <li>- defines color-phrase instance
 *   <li>- NORMAL | INPUT | MESSAGES
 *   <li>- protermcap-attribute
 *   <li>- BLINK- BRIGHT- fgnd-color bgnd-color (WINDOWS only)
 *   <li>- BLINK- RVV- UNDERLINE- BRIGHT- fgnd-color (WINDOWS only).
 *   <li>- VALUE(expression)
 * </ul>
 * The following options are NOT SUPPORTED:
 * <ul>
 *   <li>- dos-hex-attribute
 *         This option is fully documented but not yet implemented.
 *         Testing on OpenEdge for Windows environment found that isn't supported.
 *         Seems to be supported only in P4GL for MS-DOS version.  
 * </ul>
 * <p>
 * This class does not interpret the meaning of the color value.
 *
 * @author   NVS
 * @version  2.0
 */
public class ColorSpec
{
   /** Logger */
   private static final Log LOG = LogFactory.getLog(ColorSpec.class);
   
   /** COLOR_VALUE definition */
   public static final int COLOR_VALUE = 3;

   /** Color name constant for NORMAL color. */
   public static final String COLOR_NORMAL   = "NORMAL";

   /** Color name constant for INPUT color. */
   public static final String COLOR_INPUT    = "INPUT";

   /** Color name constant for MESSAGES color. */
   public static final String COLOR_MESSAGES = "MESSAGES";

   /** Color name constant for BOLD color. */
   public static final String COLOR_BOLD     = "BOLD";

   /** Color name constant for BLINK color. */
   public static final String COLOR_BLINK    = "BLINK";
   
   /** Color name (String) or index (Integer) definition value */
   private Object colorValue = COLOR_NORMAL;

   /** Color mapping table common to all terminals */
   private static Map<String, Color> globalColorMap = new LinkedHashMap<String, Color>();
   
   /** Map of terminal color mappings. */
   private static Map<String, Map<Object, Color>> terminalColorMap;
   
   /** Color mapping table for Windows colors */
   private static Map<Object, Color> windowsColorMap = new LinkedHashMap<>();

   /** Map of WINDOWS color names and indexes. */
   private static Map<String, Integer> windowsColors;

   /** WINDOWS colors abbreviations */
   private static String colorsPattern;

   /** BLINK- BRIGHT- fgnd-color bgnd-color */
   private static String pattern0;

   /** BLINK- RVV- UNDERLINE- BRIGHT- fgnd-color */
   private static String pattern1;

   /** Color table index used on WINDOWS only*/
   private static int index;

   /* Initialize local map */
   static
   {
      // the default mappings, the keys must be upper case!
      globalColorMap.put(COLOR_NORMAL  , Color.DEFAULT);
      globalColorMap.put(COLOR_INPUT   , Color.DEFAULT.withUnderline());
      globalColorMap.put(COLOR_MESSAGES, Color.DEFAULT.withReverse());
      globalColorMap.put(COLOR_BOLD    , Color.DEFAULT.withBold());
      globalColorMap.put(COLOR_BLINK   , Color.DEFAULT.withBlink());

      // initialize Linux/Unix OS colors.
      initLinuxColors();
      
      // initialize Windows OS colors.
      index = globalColorMap.size();
      initWindowsColors();
   }

   /**
    * Copy constructor.
    *
    * @param   color
    *          existing color specification to duplicate
    */
   public ColorSpec(ColorSpec color)
   {
      this.colorValue = color.colorValue;
   }

   /**
    * Main constructor.
    *
    * @param   value
    *          color name.
    */
   public ColorSpec(String value)
   {
      this.colorValue = (value == null) ? null : value.trim().toUpperCase();
   }

   /**
    * Build color specification using integer value.
    *
    * @param   intValue
    *          Color number.
    */
   public ColorSpec(int intValue)
   {
      this.colorValue = intValue;
   }

   /**
    * Convert color specification into real color.
    *
    * @return  color value
    */
   public Color convert()
   {
      // try the global static map first
      Color result = globalColorMap.get(colorValue);
      if (result != null)
      {
         return result;
      }
      // search on legacy client runtime platform color table
      if (EnvironmentOps.OS_WIN.equals(EnvironmentOps.getLegacyPlatform()))
      {
         // search on Windows color map first.
         result = windowsColorMap.get(colorValue);
         if (result == null)
         {
            // if not in the map process it.
            result = processWindowsColor();
         }
      }
      else
      {
         result = processLinuxColor();
      }
      // fall back to the NORMAL color
      return (result == null) ? globalColorMap.get(COLOR_NORMAL) : result;
   }
   
   /**
    * Process Unix/Linux colors.
    * 
    * @return  Instance of Color.
    */
   private Color processLinuxColor()
   {
      Color result = null;
      // now the colors from the terminal specific color map, if the map was initialized
      if (terminalColorMap != null)
      {
         character terminalType = LogicalTerminal.getTerminalType();
         if (terminalType != null && !terminalType.isUnknown())
         {
            String terminalName = terminalType.getValue().toUpperCase();
            Map<Object, Color> terminalColors = terminalColorMap.get(terminalName);
            if (terminalColors != null)
            {
               result = terminalColors.get(colorValue);
            }
         }
      }      
      return result;
   }
   
   /**
    * Process Windows colors.
    * BLINK- BRIGHT- fgnd-color bgnd-color
    * BLINK- RVV- UNDERLINE- BRIGHT- fgnd-color
    * 
    * @return  Instance of Color. 
    */
   private Color processWindowsColor()
   {
      Color color =  globalColorMap.get(COLOR_NORMAL);      

      if (colorValue instanceof String)
      {
         int attribute = ColorAttribute.NORMAL.getValue();
         
         String[] parts = ((String) colorValue).split("/");
         switch (parts.length)
         {
         case 1:
            // BLINK- RVV- UNDERLINE- BRIGHT- fgnd-color
            if (parts[0].matches(pattern1))
            {
               // foreground
               for (int i = 2; i < 6; i++)
               {
                  String attr = parts[0].replaceFirst(pattern1, "$" + i);
                  attribute = processAttribute(attribute, attr);
               }
               // color name
               String fgColorName = parts[0].replaceFirst(pattern1, "$6");
               int fg = getStandardColors(fgColorName);
               if (fg == -1)
               {
                  Integer fgColorIndex = windowsColors.get(fgColorName);
                  fg = (fgColorIndex == null) ? 0 : 
                        ColorPalette.colorToPalette(fgColorIndex.intValue()) + 1;
               }
               else
               {
                  fg = (fg >> 16) & 0xFFFF; 
               }
               // build and save color
               color = makeColor(attribute, fg, 0);
               // add a new entry into Windows color table
               windowsColorMap.put(colorValue, color);
               windowsColorMap.put(index++, color);
            }
         case 2:
            // BLINK- BRIGHT- fgnd-color bgnd-color
            if (parts[0].matches(pattern0) && parts[1].matches(pattern0))
            {
               // foreground
               for (int i = 2; i < 4; i++)
               {
                  String attr = parts[0].replaceFirst(pattern0, "$" + i);
                  attribute = processAttribute(attribute, attr);
               }
               // color name
               String fgColorName = parts[0].replaceFirst(pattern0, "$4");
               int fg = getStandardColors(fgColorName);
               if (fg == -1)
               {
                  Integer fgColorIndex = windowsColors.get(fgColorName);
                  fg = (fgColorIndex == null) ? 0 :
                        ColorPalette.colorToPalette(fgColorIndex.intValue()) + 1;
               }
               else
               {
                  fg = (fg >> 16) & 0xFFFF; 
               }
               
               // background
               for (int i = 2; i < 4; i++)
               {
                  String attr = parts[1].replaceFirst(pattern0, "$" + i);
                  processAttribute(attribute, attr);
               }
               // color name
               String bgColorName = parts[1].replaceFirst(pattern0, "$4");
               int bg = getStandardColors(bgColorName);
               if (bg == -1)
               {
                  Integer bgColorIndex = windowsColors.get(bgColorName);
                  bg = (bgColorIndex == null) ? 0 :
                        ColorPalette.colorToPalette(bgColorIndex.intValue()) + 1;
               }
               else
               {
                  bg = bg & 0xFFFF; 
               }

               // build and save color
               color = makeColor(attribute, fg, bg);
               // add a new entry into Windows color table
               windowsColorMap.put(colorValue, color);
               windowsColorMap.put(index++, color);
            }
         }         
      }
      
      return color;      
   }
   
   /**
    * Init colors for Unix/Linux.  The values are read from protermcap file.
    */
   private static void initLinuxColors()
   {
      // load the color map
      ColorMapLoader loader = new ColorMapLoader();
         
      Map<String, Map<Object, Color>> loadedMap = loader.loadAllMappings();
      if (loadedMap != null)
      {
         // fill the terminalColorMap field, don't use the loaded map directly, we need to
         // normalize the keys
         Map<String, Map<Object, Color>> localTerminalMap = new LinkedHashMap<String, Map<Object, Color>>();
         for (Map.Entry<String, Map<Object, Color>> termEntry : loadedMap.entrySet())
         {
            Map<Object, Color> localColorMap = new LinkedHashMap<Object, Color>();
            for (Map.Entry<Object, Color> colorEntry : termEntry.getValue().entrySet())
            {
               Object key = colorEntry.getKey();
               if (key instanceof String)
               {
                  // normalize the string key
                  localColorMap.put(((String) key).trim().toUpperCase(), colorEntry.getValue());
               }
               else
               {
                  localColorMap.put(key, colorEntry.getValue());
               }
            }
            localTerminalMap.put(termEntry.getKey().trim().toUpperCase(), termEntry.getValue());
         }
         terminalColorMap = localTerminalMap;
      } 
      else
      {
         LOG.info("Terminal color map was not loaded.");
      }      
   }
   
   /**
    * Initialize Windows color names and indexes.  Build patterns used as regular expressions. 
    * BLINK- BRIGHT- fgnd-color bgnd-color
    * BLINK- RVV- UNDERLINE- BRIGHT- fgnd-color
    */
   private static void initWindowsColors()
   {
      windowsColors = new LinkedHashMap<>();
      windowsColors.put("BLACK", 0);
      windowsColors.put("BLA", 0);
      windowsColors.put("BLK", 0);
      windowsColors.put("BLUE", 1);
      windowsColors.put("BLU", 1);
      windowsColors.put("GREEN", 2);
      windowsColors.put("GRE", 2);
      windowsColors.put("GRN", 2);
      windowsColors.put("CYAN", 3);
      windowsColors.put("C", 3);
      windowsColors.put("RED", 4);
      windowsColors.put("MAGENTA", 5);
      windowsColors.put("MA", 5);
      windowsColors.put("BROWN", 6);
      windowsColors.put("BRO", 6);
      windowsColors.put("BRN", 6);
      windowsColors.put("GRAY", 7);
      windowsColors.put("GRA", 7);
      windowsColors.put("GRY", 7);
      windowsColors.put("DARK-GRAY", 8);
      windowsColors.put("D-GRA", 8);
      windowsColors.put("LIGHT-BLUE", 9);
      windowsColors.put("LT-BLU", 9);
      windowsColors.put("LIGHT-GREEN", 10);
      windowsColors.put("LT-GRE", 10);
      windowsColors.put("LIGHT-CYAN", 11);
      windowsColors.put("LT-C", 11);
      windowsColors.put("LIGHT-RED", 12);
      windowsColors.put("LT-RED", 12);
      windowsColors.put("LIGHT-MAGENTA", 13);
      windowsColors.put("LT-MA", 13);
      windowsColors.put("LIGHT-BROWN", 14);
      windowsColors.put("LT-BRO", 14);
      windowsColors.put("YELLOW", 14);
      windowsColors.put("Y", 14);
      windowsColors.put("WHITE", 15);
      windowsColors.put("W", 15);
      
      // build regular expressions
      StringBuilder sb = new StringBuilder("(");
      for (String name : windowsColors.keySet())
      {
         sb.append(name).append("|");
      }
      // add standard colors
      sb.append("NORMAL|INPUT|MESSAGES)?");
      
      colorsPattern = sb.toString();
      pattern0 = "^((.*BLINK-)?(.*BRIGHT-)?)" + colorsPattern;
      pattern1 = "^((.*BLINK-)?(.*RVV-)?(.*UNDERLINE-)?(.*BRIGHT-)?)" + colorsPattern;            
   }
   
   /**
    * Build a Color object.
    * 
    * @param   attribute
    *          Color attributes.
    * @param   fg
    *          Foreground color index.
    * @param   bg
    *          Background color index.
    * 
    * @return  A new Color object instance.
    */
   private Color makeColor(int attribute, int fg, int bg)
   {
      int color = (bg & 0xFFFF) | ((fg << 16) & 0xFFFF0000);
      return new Color(attribute, color);
   }
   
   /**
    * Process color attribute.
    * 
    * @param   attribute
    *          Current attributes.
    * @param   attr
    *          Attribute string.
    * 
    * @return  Update attributes.
    */
   private int processAttribute(int attribute, String attr)
   {
      if (attr != null && !attr.isEmpty())
      {
         if ("BLINK-".equals(attr))
         {
            attribute |= ColorAttribute.BLINK.getValue();
         }
         else if ("RVV-".equals(attr))
         {
            attribute |= ColorAttribute.REVERSE.getValue();
         }
         else if ("UNDERLINE-".equals(attr))
         {
            attribute |= ColorAttribute.UNDERLINE.getValue();
         }
         else if ("BRIGHT-".equals(attr))
         {
            attribute |= ColorAttribute.BOLD.getValue();
         }
      }
      return attribute;
   }
   
   /**
    * Check if color is missing or is a standard color NORMAL, INPUT or MESSAGES
    * If yes return the standard color otherwise -1.
    * 
    * @param   colorName
    *          color name.
    *          
    * @return  see above
    */
   private int getStandardColors(String colorName)
   {
      // color is missing
      if (colorName == null || colorName.trim().isEmpty())
      {
         return globalColorMap.get(COLOR_NORMAL).color();
      }
      // NORMAL color
      else if (COLOR_NORMAL.equals(colorName))
      {
         return globalColorMap.get(COLOR_NORMAL).color();         
      }
      // INPUT color
      else if (COLOR_INPUT.equals(colorName))
      {
         return globalColorMap.get(COLOR_INPUT).color();         
      }
      // MESSAGES color
      else if (COLOR_MESSAGES.equals(colorName))
      {
         return globalColorMap.get(COLOR_MESSAGES).color();         
      }
      // other color
      return -1;
   }
}