/*
** Module   : CommonWindow.java
** Abstract : Define all window-related APIs available via a window handle instance.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 CA  20130112 Created initial version.
** 002 CA  20130927 Added LegacyResource annotation, to determine resource type.
** 003 EVL 20131008 Added legacy attributes and methods annotation.
** 004 MAG 20140512 Added new attributes.
** 005 CA  20140701 Added support for font-related attributes:  MESSAGE-AREA-FONT, 
**                  STATUS-AREA-FONT.
** 006 EVL 20141204 Added LOAD-SMALL-ICON setter.
** 007 HC  20141207 Implemented runtime support for widget coordinate attributes.
** 008 VIG 20141104 Added MENUBAR attribute.
** 009 HC  20150508 Improvements to GUI window scrolling support.
** 010 SVL 20170126 Added support for WINDOW-STATE attribute.
**     HC  20170127 Added support for conversion of WINDOW-DELAYED-MINIMIZE literal.
** 011 HC  20170629 Implemented support of WINDOW:DISABLE-REDRAW attribute, the FWD extension to
**                  4GL.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;
 
/**
 * Provide access to all window-related APIs which can be accessed via a 
 * handle instance.
 */
@LegacyResource(resource = LegacyResource.WINDOW)
public interface CommonWindow
{
   /** Window state: maximized. */
   public static final int WINDOW_STATE_MAXIMIZED = 1;

   /** Window state: minimized. */
   public static final int WINDOW_STATE_MINIMIZED = 2;

   /** Window state: normal. */
   public static final int WINDOW_STATE_NORMAL = 3;

   /** Window state: delayed minimized. */
   public static final int WINDOW_STATE_DELAYED_MINIMIZE = 4;

   /**
    * Get icon file name previously loaded by {@link #loadIcon(character)}.
    * 
    * @return  icon file name.
    */
   @LegacyAttribute(name = "ICON")
   public character getIcon();
   
   /**
    * Get window MENUBAR.
    * 
    * @return  handle.
    */
   @LegacyAttribute(name = "MENUBAR")
   public handle getMenubar();

   /**
    * Set window MENUBAR.
    * 
    * @param  menubar
    *         The menu handle.
    */
   @LegacyAttribute(name = "MENUBAR", setter = true)
   public void setMenubar(handle menubar);
   
   /**
    * Set window MENUBAR.
    * 
    * @param  menubar
    *         The menu widget.
    */
   @LegacyAttribute(name = "MENUBAR", setter = true)
   public void setMenubar(MenuWidget menubar);
   
   /**
    * Gets the FULL-HEIGHT-CHARS writable attribute.
    *
    * @return current value of FULL-HEIGHT-CHARS attribute
    */
   @LegacyAttribute(name = "FULL-HEIGHT-CHARS")
   public decimal getFullHeightChars();
   
   /**
    * Gets the FULL-HEIGHT-PIXELS writable attribute.
    *
    * @return current value of FULL-HEIGHT-PIXELS attribute
    */
   @LegacyAttribute(name = "FULL-HEIGHT-PIXELS")
   public integer getFullHeightPixels();
   
   /**
    * Gets the FULL-WIDTH-CHARS writable attribute.
    *
    * @return current value of FULL-WIDTH-CHARS attribute
    */
   @LegacyAttribute(name = "FULL-WIDTH-CHARS")
   public decimal getFullWidthChars();
   
   /**
    * Gets the FULL-WIDTH-PIXELS writable attribute.
    *
    * @return current value of FULL-WIDTH-PIXELS attribute
    */
   @LegacyAttribute(name = "FULL-WIDTH-PIXELS")
   public integer getFullWidthPixels();
   
   /**
    * Get window maximal height.
    * 
    * @return  maximal height.
    */
   @LegacyAttribute(name = "MAX-HEIGHT")
   public decimal getMaxHeight();

   /**
    * Gets the MAX-HEIGHT-CHARS writable attribute.
    *
    * @return current value of MAX-HEIGHT-CHARS attribute
    */
   @LegacyAttribute(name = "MAX-HEIGHT-CHARS")
   public decimal getMaxHeightChars();

   /**
    * Gets the MAX-HEIGHT-PIXELS writable attribute.
    *
    * @return current value of MAX-HEIGHT-PIXELS attribute
    */
   @LegacyAttribute(name = "MAX-HEIGHT-PIXELS")
   public integer getMaxHeightPixels();

   /**
    * Get window maximal width.
    * 
    * @return  maximal width.
    */
   @LegacyAttribute(name = "MAX-WIDTH")
   public decimal getMaxWidth();

   /**
    * Gets the MAX-WIDTH-CHARS writable attribute.
    *
    * @return   The current value of MAX-WIDTH-CHARS attribute.
    */
   @LegacyAttribute(name = "MAX-WIDTH-CHARS")
   public decimal getMaxWidthChars();

   /**
    * Gets the MAX-WIDTH-PIXELS writable attribute.
    *
    * @return   The current value of MAX-WIDTH-PIXELS attribute.
    */
   @LegacyAttribute(name = "MAX-WIDTH-PIXELS")
   public integer getMaxWidthPixels();

   /**
    * Gets the MIN-HEIGHT-CHARS writable attribute.
    *
    * @return current value of MIN-HEIGHT-CHARS attribute
    */
   @LegacyAttribute(name = "MIN-HEIGHT-CHARS")
   public decimal getMinHeightChars();

   /**
    * Gets the MIN-HEIGHT-PIXELS writable attribute.
    *
    * @return current value of MIN-HEIGHT-PIXELS attribute
    */
   @LegacyAttribute(name = "MIN-HEIGHT-PIXELS")
   public integer getMinHeightPixels();
   
   /**
    * Gets the MIN-WIDTH-CHARS writable attribute.
    *
    * @return current value of MIN-WIDTH-CHARS attribute
    */
   @LegacyAttribute(name = "MIN-WIDTH-CHARS")
   public decimal getMinWidthChars();

   /**
    * Gets the MIN-WIDTH-PIXELS writable attribute.
    *
    * @return current value of MIN-WIDTH-PIXELS attribute
    */
   @LegacyAttribute(name = "MIN-WIDTH-PIXELS")
   public integer getMinWidthPixels();
   
   /**
    * Gets the KEEP-FRAME-Z-ORDER writable attribute.
    *
    * @return   The current value of the KEEP-FRAME-Z-ORDER attribute.
    */
   @LegacyAttribute(name = "KEEP-FRAME-Z-ORDER")
   public logical isKeepFrameZOrder();
   
   /**
    * Check if window has KEEP-Z-ORDER option set.
    * 
    * @return <code>true</code> if window has KEEP-Z-ORDER option set.
    */
   @LegacyAttribute(name = "KEEP-Z-ORDER")
   public logical isKeepZOrder();

   /**
    * Check if window has message area.
    * 
    * @return <code>true</code> if window has message area.
    */
   @LegacyAttribute(name = "MESSAGE-AREA")
   public logical isMessageArea();

   /**
    * Check if window size can be changed.
    * 
    * @return  <code>true</code> if window size can be changed.
    */
   @LegacyAttribute(name = "RESIZE")
   public logical isResize();

   /**
    * Check if window has scrollbars.
    * 
    * @return  <code>true</code> if window has scrollbars.
    */
   @LegacyAttribute(name = "SCROLL-BARS")
   public logical isScrollBars();

   /**
    * Check if window has message area.
    * 
    * @return  <code>true</code> if window has message area.
    */
   @LegacyAttribute(name = "STATUS-AREA")
   public logical isStatusArea();

   /**
    * Load icon for window.
    * 
    * @param   searchPath
    *          Icon file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "LOAD-ICON")
   public logical loadIcon(character searchPath);

   /**
    * Load icon for window.
    * 
    * @param   searchPath
    *          Icon file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "LOAD-ICON")
   public logical loadIcon(String searchPath);

   /**
    * Load small icon(16x16) for window.
    * 
    * @param   searchPath
    *          Icon file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "LOAD-SMALL-ICON")
   public logical loadSmallIcon(character searchPath);

   /**
    * Load small icon(16x16) for window.
    * 
    * @param   searchPath
    *          Icon file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "LOAD-SMALL-ICON")
   public logical loadSmallIcon(String searchPath);
   
   /**
    * Sets the KEEP-FRAME-Z-ORDER writable attribute.
    *
    * @param    keep
    *           The new value for the KEEP-FRAME-Z-ORDER attribute.
    */
   @LegacyAttribute(name = "KEEP-FRAME-Z-ORDER", setter = true)
   public void setKeepFrameZOrder(boolean keep);

   /**
    * Sets the KEEP-FRAME-Z-ORDER writable attribute.
    *
    * @param    keep
    *           The new value for the KEEP-FRAME-Z-ORDER attribute.
    */
   @LegacyAttribute(name = "KEEP-FRAME-Z-ORDER", setter = true)
   public void setKeepFrameZOrder(logical keep);

   /**
    * Enable/disable KEEP-Z-ORDER option.
    * 
    * @param   logical
    *          New flag value.
    */
   @LegacyAttribute(name = "KEEP-Z-ORDER", setter = true)
   public void setKeepZOrder(logical logical);

   /**
    * Set maximal height.
    * 
    * @param   decimal
    *          New option value.
    */
   @LegacyAttribute(name = "MAX-HEIGHT", setter = true)
   public void setMaxHeight(NumberType decimal);

   /**
    * Sets the MAX-HEIGHT-CHARS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-HEIGHT-CHARS attribute.
    */
   @LegacyAttribute(name = "MAX-HEIGHT-CHARS", setter = true)
   public void setMaxHeightChars(NumberType max);

   /**
    * Sets the MAX-HEIGHT-PIXELS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-HEIGHT-PIXELS attribute.
    */
   @LegacyAttribute(name = "MAX-HEIGHT-PIXELS", setter = true)
   public void setMaxHeightPixels(integer max);

   /**
    * Set maximal width.
    * 
    * @param   decimal
    *          New option value.
    */
   @LegacyAttribute(name = "MAX-WIDTH", setter = true)
   public void setMaxWidth(NumberType decimal);

   /**
    * Sets the MAX-WIDTH-CHARS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-WIDTH-CHARS attribute.
    */
   @LegacyAttribute(name = "MAX-WIDTH-CHARS", setter = true)
   public void setMaxWidthChars(NumberType max);

   /**
    * Sets the MAX-WIDTH-CHARS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-WIDTH-CHARS attribute.
    */
   @LegacyAttribute(name = "MAX-WIDTH-PIXELS", setter = true)
   public void setMaxWidthPixels(integer max);

   /**
    * Sets the MIN-HEIGHT-CHARS writable attribute.
    *
    * @param    min
    *           The new value for the MIN-HEIGHT-CHARS attribute.
    */
   @LegacyAttribute(name = "MIN-HEIGHT-CHARS", setter = true)
   public void setMinHeightChars(NumberType min);

   /**
    * Sets the MIN-HEIGHT-PIXELS writable attribute.
    *
    * @param    min
    *           The new value for the MIN-HEIGHT-PIXELS attribute.
    */
   @LegacyAttribute(name = "MIN-HEIGHT-PIXELS", setter = true)
   public void setMinHeightPixels(integer min);

   /**
    * Sets the MIN-WIDTH-CHARS writable attribute.
    *
    * @param    min
    *           The new value for the MIN-WIDTH-CHARS attribute.
    */
   @LegacyAttribute(name = "MIN-WIDTH-CHARS", setter = true)
   public void setMinWidthChars(NumberType min);

   /**
    * Sets the MIN-WIDTH-CHARS writable attribute.
    *
    * @param    min
    *           The new value for the MIN-WIDTH-CHARS attribute.
    */
   @LegacyAttribute(name = "MIN-WIDTH-PIXELS", setter = true)
   public void setMinWidthPixels(integer min);

   /**
    * Enable/disable displaying of message area.
    * 
    * @param   logical
    *          New flag value.
    */
   @LegacyAttribute(name = "MESSAGE-AREA", setter = true)
   public void setMessageArea(logical logical);

   /**
    * Enable/disable window resizing.
    * 
    * @param   logical
    *          New option value.
    */
   @LegacyAttribute(name = "RESIZE", setter = true)
   public void setResize(logical logical);

   /**
    * Enable/disable displaying of scrollbars.
    * 
    * @param   logical
    *          New option value.
    */
   @LegacyAttribute(name = "SCROLL-BARS", setter = true)
   public void setScrollBars(logical logical);

   /**
    * Enable/disable displaying of status area.
    * 
    * @param   logical
    *          New option value.
    */
   @LegacyAttribute(name = "STATUS-AREA", setter = true)
   public void setStatusArea(logical logical);

   /**
    * Get the message-area's font.
    * 
    * @return   The message-area's font.
    */
   @LegacyAttribute(name = "MESSAGE-AREA-FONT")
   public int64 getMessageAreaFont();

   /**
    * Set the message-area's font.
    * 
    * @param    fontNum
    *           The font number, an entry in the font-table.
    */
   @LegacyAttribute(name = "MESSAGE-AREA-FONT", setter = true)
   public void setMessageAreaFont(long fontNum);

   /**
    * Set the message-area's font.
    * 
    * @param    fontNum
    *           The font number, an entry in the font-table.
    */
   @LegacyAttribute(name = "MESSAGE-AREA-FONT", setter = true)
   public void setMessageAreaFont(int64 fontNum);
   
   /**
    * Get the status-area's font.
    * 
    * @return   The status-area's font.
    */
   @LegacyAttribute(name = "STATUS-AREA-FONT")
   public int64 getStatusAreaFont();

   /**
    * Set the status-area's font.
    * 
    * @param    fontNum
    *           The font number, an entry in the font-table.
    */
   @LegacyAttribute(name = "STATUS-AREA-FONT", setter = true)
   public void setStatusAreaFont(long fontNum);

   /**
    * Set the status-area's font.
    * 
    * @param    fontNum
    *           The font number, an entry in the font-table.
    */
   @LegacyAttribute(name = "STATUS-AREA-FONT", setter = true)
   public void setStatusAreaFont(int64 fontNum);
   
   /**
    * Gets the SCREEN-LINES readable attribute.
    *
    * @return   current value of SCREEN-LINES attribute
    */
   @LegacyAttribute(name = "SCREEN-LINES")
   public decimal getScreenLines();

   /**
    * Gets the WINDOW-STATE attribute.
    *
    * @return  current value of WINDOW-STATE attribute (see <code>WINDOW_STATE_*</code>
    *          constants).
    */
   @LegacyAttribute(name = "WINDOW-STATE")
   public int64 getWindowState();

   /**
    * Sets the WINDOW-STATE attribute.
    *
    * @param state
    *        Value of WINDOW-STATE attribute to set (see <code>WINDOW_STATE_*</code>
    *        constants).
    */
   @LegacyAttribute(name = "WINDOW-STATE", setter = true)
   public void setWindowState(long state);

   /**
    * Sets the WINDOW-STATE attribute.
    *
    * @param state
    *        Value of WINDOW-STATE attribute to set (see <code>WINDOW_STATE_*</code>
    *        constants).
    */
   @LegacyAttribute(name = "WINDOW-STATE", setter = true)
   public void setWindowState(int64 state);

   /**
    * Sets the MIN-BUTTON attribute.
    *
    * @param   value
    *          The new attribute value.
    */
   @LegacyAttribute(name = "MIN-BUTTON", setter = true)
   default public void setMinButton(logical value)
   {
      if (value.isUnknown())
      {
         ErrorManager.recordOrShowError(
            4083,
            "**Unable to assign UNKNOWN value to attribute MIN-BUTTON on WINDOW widget",
            true, false, false);

         return;
      }

      setMinButton(value.booleanValue());
   }

   /**
    * Sets the MIN-BUTTON attribute.
    *
    * @param   value
    *          The new attribute value.
    */
   @LegacyAttribute(name = "MIN-BUTTON", setter = true)
   public void setMinButton(boolean value);

   /**
    * Gets the MIN-BUTTON attribute.
    *
    * @return  see above.
    */
   @LegacyAttribute(name = "MIN-BUTTON")
   public logical isMinButton();

   /**
    * Sets the MAX-BUTTON attribute.
    *
    * @param   value
    *          The new attribute value.
    */
   @LegacyAttribute(name = "MAX-BUTTON", setter = true)
   default public void setMaxButton(logical value)
   {
      if (value.isUnknown())
      {
         ErrorManager.recordOrShowError(
         4083,
         "**Unable to assign UNKNOWN value to attribute MAX-BUTTON on WINDOW widget",
         true, false, false);

         return;
      }

      setMaxButton(value.booleanValue());
   }

   /**
    * Sets the MAX-BUTTON attribute.
    *
    * @param   value
    *          The new attribute value.
    */
   @LegacyAttribute(name = "MAX-BUTTON", setter = true)
   public void setMaxButton(boolean value);

   /**
    * Gets the MAX-BUTTON attribute.
    *
    * @return  see above.
    */
   @LegacyAttribute(name = "MAX-BUTTON")
   public logical isMaxButton();

   /**
    * Sets the SHOW-IN-TASKBAR attribute.
    *
    * @param   value
    *          The new attribute value.
    */
   @LegacyAttribute(name = "SHOW-IN-TASKBAR", setter = true)
   default void setShowInTaskbar(logical value)
   {
      if (value.isUnknown())
      {
         ErrorManager.recordOrShowError(
            4083,
            "**Unable to assign UNKNOWN value to attribute SHOW-IN-TASKBAR on WINDOW widget",
            true, false, false);

         return;
      }

      setShowInTaskbar(value.booleanValue());
   }

   /**
    * Sets the SHOW-IN-TASKBAR attribute.
    *
    * @param   value
    *          The new attribute value.
    */
   @LegacyAttribute(name = "SHOW-IN-TASKBAR", setter = true)
   void setShowInTaskbar(boolean value);

   /**
    * Gets the SHOW-IN-TASKBAR attribute.
    *
    * @return  see above.
    */
   @LegacyAttribute(name = "SHOW-IN-TASKBAR")
   logical isShowInTaskbar();

   /**
    * Sets the ALWAYS-ON-TOP attribute.
    *
    * @param    value
    *           <code>true</code> if the option/attribute is set to ON.
    */
   @LegacyAttribute(name = "ALWAYS-ON-TOP", setter = true)
   public void setAlwaysOnTop(boolean value);

   /**
    * Sets the ALWAYS-ON-TOP attribute.
    *
    * @param    value
    *           <code>true</code> if the attribute is set to ON.
    */
   @LegacyAttribute(name = "ALWAYS-ON-TOP", setter = true)
   default public void setAlwaysOnTop(logical value)
   {
      if (value == null || value.isUnknown())
      {
         ErrorManager.recordOrShowError(
            4083,
            "Unable to assign UNKNOWN value to attribute ALWAYS-ON-TOP on WINDOW widget",
            true, false, false);

         return;
      }

      setAlwaysOnTop(value.booleanValue());
   }

   /**
    * Gets the ALWAYS-ON-TOP attribute's value.
    *
    * @return  the value of the attribute
    */
   @LegacyAttribute(name = "ALWAYS-ON-TOP")
   public logical isAlwaysOnTop();

   /**
    * Sets the CONTROL-BOX attribute.
    *
    * @param    value
    *           <code>true</code> if the option/attribute is set to ON.
    */
   @LegacyAttribute(name = "CONTROL-BOX", setter = true)
   public void setControlBox(boolean value);

   /**
    * Sets the CONTROL-BOX attribute.
    *
    * @param    value
    *           <code>true</code> if the attribute is set to ON.
    */
   @LegacyAttribute(name = "CONTROL-BOX", setter = true)
   default public void setControlBox(logical value)
   {
      if (value == null || value.isUnknown())
      {
         ErrorManager.recordOrShowError(
            4083,
            "Unable to assign UNKNOWN value to attribute CONTROL-BOX on WINDOW widget",
            true, false, false);

         return;
      }

      setControlBox(value.booleanValue());
   }

   /**
    * Gets the CONTROL-BOX attribute's value.
    *
    * @return  the value of the attribute
    */
   @LegacyAttribute(name = "CONTROL-BOX")
   public logical isControlBox();

   /**
    * Sets the SMALL-TITLE attribute.
    *
    * @param    value
    *           <code>true</code> if the option/attribute is set to ON.
    */
   @LegacyAttribute(name = "SMALL-TITLE", setter = true)
   public void setSmallTitle(boolean value);

   /**
    * Sets the SMALL-TITLE attribute.
    *
    * @param    value
    *           <code>true</code> if the attribute is set to ON.
    */
   @LegacyAttribute(name = "SMALL-TITLE", setter = true)
   default public void setSmallTitle(logical value)
   {
      if (value == null || value.isUnknown())
      {
         ErrorManager.recordOrShowError(
            4083,
            "Unable to assign UNKNOWN value to attribute SMALL-TITLE on WINDOW widget",
            true, false, false);

         return;
      }

      setSmallTitle(value.booleanValue());
   }

   /**
    * Gets the SMALL-TITLE attribute's value.
    *
    * @return  the value of the attribute
    */
   @LegacyAttribute(name = "SMALL-TITLE")
   public logical isSmallTitle();

   /**
    * Gets the DISABLE-REDRAW attribute.
    * This attribute is a FWD extension, not real 4GL!
    *
    * @return  current value of DISABLE-REDRAW attribute.
    */
   @LegacyAttribute(name = "DISABLE-REDRAW")
   public logical isDisableRedraw();

   /**
    * Sets the DISABLE-REDRAW attribute.
    * This attribute is a FWD extension, not real 4GL!
    *
    * @param   value
    *          Value to set to DISABLE-REDRAW attribute.
    */
   @LegacyAttribute(name = "DISABLE-REDRAW", setter = true)
   default public void setDisableRedraw(logical value)
   {
      if (value == null || value.isUnknown())
      {
         // ignore invalid value
         return;
      }

      setDisableRedraw(value.booleanValue());
   }

   /**
    * Sets the DISABLE-REDRAW attribute.
    * This attribute is a FWD extension, not real 4GL!
    *
    * @param   value
    *          Value to set to DISABLE-REDRAW attribute.
    */
   @LegacyAttribute(name = "DISABLE-REDRAW", setter = true)
   public void setDisableRedraw(boolean value);
}