/*
** Module   : Coordinate.java
** Abstract : Provides elementary operations for the UI coordinate data types.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 HC  20140911 Created initial version.
** 002 MAG 20140926 Added double to integer conversion helper.
** 003 HC  20141021 Implemented window move/position support.
** 004 HC  20141207 Implemented runtime support for widget coordinate attributes.
** 005 HC  20150803 Added support for multiple rounding modes for coordinates conversion.
** 006 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 007 HC  20161018 Added equals() method.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.math.BigDecimal;

import com.goldencode.p2j.ui.client.*;

/**
 * The class provides elementary operations for the UI coordinate data types.   
 */
public class Coordinate
{
   /**
    * Minimum character coordinate value for ChUI environments.
    */
   public static final int MIN_CHUI_CH =    1;

   /**
    * Maximum character coordinate value for ChUI environments.
    */
   public static final int MAX_CHUI_CH =    131072;

   /**
    * Minimum character coordinate value for GUI environments.
    */
   public static final double MIN_GUI_CH = -131072;

   /**
    * Maximum character coordinate value for GUI environments.
    */
   public static final double MAX_GUI_CH =  131072;
   
   /**
    * Minimum pixel coordinate value for a realized widget.
    */
   public static final int MIN_GUI_PIXEL = -32768;
   
   /**
    * Maximum pixel coordinate value for a realized widget.
    */
   public static final int MAX_GUI_PIXEL = 32767;
   
   /** 
    * Maximum pixel width for a realized widget in GUI mode.
    */
   public static final int MAX_GUI_WIDTH_PIXELS = 65527; 

   /** 
    * Maximum pixel height for a realized widget in GUI mode.
    */
   public static final int MAX_GUI_HEIGHT_PIXELS = 65456;
   
   /**
    * Identifiers denoting rounding mode used in the UI logic.
    */
   public enum RoundingMode
   {
      /** Value is rounded up or down depending on the number's decimal part. */
      ROUND,
      
      /** The decimal part is discarded. */
      TRUNCATE
   }
   
   /**
    * Performs half-up rounding of a character coordinate value.
    * The result is meaningful only to ChUI environments.
    * 
    * @param   ch
    *          The coordinate value in character units.
    * 
    * @return  The rounded coordinate value.
    */
   public static double chuiRound(double ch)
   {
      // take a shortcut in case the input is already a whole number 
      int ich = (int) ch;
      if (ich == ch)
      {
         return ich;
      }
      
      return BigDecimal.valueOf(ch).setScale(0, BigDecimal.ROUND_HALF_UP).doubleValue();
   }
   
   /**
    * Performs half-up rounding of a character coordinate value.
    * The result is meaningful only to GUI environments.
    * 
    * @param   ch
    *          The coordinate value in character units.
    * 
    * @return  The rounded coordinate value.
    */
   public static double guiRound(double ch)
   {
      // take a shortcut in case the input is already a whole number 
      int ich = (int) ch;
      if (ich == ch)
      {
         return ich;
      }
      
      return BigDecimal.valueOf(ch).setScale(2, BigDecimal.ROUND_HALF_UP).doubleValue();
   }
   
   /**
    * Sets the input value to the default scale (of two decimal places) for the character 
    * coordinate type. Use this method whenever the value holds an unknown precision 
    * (for example a result of an arithmetic operation) and the default scale is required
    * (for example when comparing two character values).
    * <p>
    * Note that the result of this method - the default scale for a character coordinate value -
    * doesn't depend on the UI mode. It is the same for both GUI and ChUI.
    * 
    * @param   ch
    *          The coordinate value in character units.
    * 
    * @return  The scaled coordinate value.
    */
   public static double scale(double ch)
   {
      validate(ch);
      
      // take a shortcut in case the input is already a whole number 
      int ich = (int) ch;
      if (ich == ch)
      {
         return ich;
      }
      
      return BigDecimal.valueOf(ch).setScale(2, BigDecimal.ROUND_HALF_UP).doubleValue();
   }
   
   /**
    * Sets the input value to the default scale (of two decimal places) for the character 
    * coordinate type. Use this method whenever the value holds an unknown precision 
    * (for example a result of an arithmetic operation) and the default scale is required
    * (for example when comparing two character values).
    * <p>
    * Note that the result of this method - the default scale for a character coordinate value -
    * doesn't depend on the UI mode. It is the same for both GUI and ChUI.
    * 
    * @param   insets
    *          An {@link Insets} instance. 
    * 
    * @return  The scaled {@link Insets} instance.
    */
   public static Insets scale(Insets insets)
   {
      return new Insets(scale(insets.top), scale(insets.left), 
                        scale(insets.bottom), scale(insets.right));
   }
   
   /**
    * Sets the input value to the default scale (of two decimal places) for the character 
    * coordinate type. Use this method whenever the value holds an unknown precision 
    * (for example a result of an arithmetic operation) and the default scale is required
    * (for example when comparing two character values).
    * <p>
    * Note that the result of this method - the default scale for a character coordinate value -
    * doesn't depend on the UI mode. It is the same for both GUI and ChUI.
    * 
    * @param   point
    *          A {@link Point} instance. 
    * 
    * @return  The scaled {@link Point} instance.
    */
   public static Point scale(Point point)
   {
      return new Point(scale(point.x), scale(point.y));
   }
   
   /**
    * Sets the input value to the default scale (of two decimal places) for the character 
    * coordinate type. Use this method whenever the value holds an unknown precision 
    * (for example a result of an arithmetic operation) and the default scale is required
    * (for example when comparing two character values).
    * <p>
    * Note that the result of this method - the default scale for a character coordinate value -
    * doesn't depend on the UI mode. It is the same for both GUI and ChUI.
    * 
    * @param   rect
    *          An {@link Rectangle} instance. 
    * 
    * @return  The scaled {@link Rectangle} instance.
    */
   public static Rectangle scale(Rectangle rect)
   {
      return new Rectangle(scale(rect.top()), scale(rect.left()), 
                           scale(rect.bottom()), scale(rect.right()));
   }
   
   /**
   /**
    * Sets the input value to the default scale (of two decimal places) for the character 
    * coordinate type. Use this method whenever the value holds an unknown precision 
    * (for example a result of an arithmetic operation) and the default scale is required
    * (for example when comparing two character values).
    * <p>
    * Note that the result of this method - the default scale for a character coordinate value -
    * doesn't depend on the UI mode. It is the same for both GUI and ChUI.
    * 
    * @param   dim
    *          An {@link Dimension} instance. 
    * 
    * @return  The scaled {@link Dimension} instance.
    */
   public static Dimension scale(Dimension dim)
   {
      return new Dimension(scale(dim.width), scale(dim.height));
   }

   /**
    * Returns <code>true</code> if the input coordinates equal.
    *
    * @param   coordinate1
    *          A coordinate value.
    * @param   coordinate2
    *          A coordinate value.
    *
    * @return  See above.
    */
   public static boolean equals(double coordinate1, double coordinate2)
   {
      double val1 = scale(coordinate1);
      double val2 = scale(coordinate2);

      return val1 == val2;
   }
   
   /**
    * Validates the input value. The method throws {@link RuntimeException}
    * if the input value is:
    * <ul>
    * <li>equal to <code>Double.NaN,</code></li>
    * <li>infinitely large.</li>
    * </ul> 
    * 
    * @param   ch
    *          Coordinate value in character units.
    *          
    * @throws  RuntimeException
    *          When the <code>ch</code> is not a valid coordinate value.
    */
   private static void validate(double ch)
   {
      if (Double.isInfinite(ch) || Double.isNaN(ch))
      {
         throw new RuntimeException("Invalid character coordinate value.");
      }
   }

   /**
    * Convert a double value to integer. Round integer value.
    * 
    * @param   val
    *          Double value.
    *          
    * @return  Integer rounded value.
    */
   public static int toInt(double val)
   {
      return Double.valueOf(val + 0.5).intValue();
   }
   
   /**
    * The input value is adjusted according to minimum and maximum possible
    * character coordinate values for ChUI environments.
    * <p>
    * If the input value is less than {@linkplain #MIN_CHUI_CH} then {@linkplain #MIN_CHUI_CH}
    * is returned. If the input value is greater than {@linkplain #MAX_CHUI_CH} then 
    * {@linkplain #MAX_CHUI_CH} is returned. Otherwise the input value is returned.
    * 
    * @param   value
    *          Character coordinate value being adjusted.
    *          
    * @return  The adjusted character coordinate value.
    */
   public static double adjustMinMaxChuiChar(double value)
   {
      if (value < MIN_CHUI_CH)
      {
         return MIN_CHUI_CH;
      }
      else if (value > MAX_CHUI_CH)
      {
         return MAX_CHUI_CH;
      }
      
      return value;
   }
   
   /**
    * The input value is adjusted according to minimum and maximum possible
    * character coordinate values for ChUI environments.
    * <p>
    * If the input value is less than {@linkplain #MIN_GUI_CH} then {@linkplain #MIN_GUI_CH}
    * is returned. If the input value is greater than {@linkplain #MAX_GUI_CH} then 
    * {@linkplain #MAX_GUI_CH} is returned. Otherwise the input value is returned.
    * 
    * @param   value
    *          Character coordinate value being adjusted.
    *          
    * @return  The adjusted character coordinate value.
    */
   public static double adjustMinMaxGuiChar(double value)
   {
      if (value < MIN_GUI_CH)
      {
         return MIN_GUI_CH;
      }
      else if (value > MAX_GUI_CH)
      {
         return MAX_GUI_CH;
      }
      
      return value;
   }
   
   /**
    * The input value is adjusted according to minimum and maximum possible
    * character coordinate values for ChUI environments.
    * <p>
    * If the input value is less than {@linkplain #MIN_GUI_PIXEL} then 
    * {@linkplain #MIN_GUI_PIXEL} is returned. If the input value is greater than 
    * {@linkplain #MAX_GUI_PIXEL} then {@linkplain #MAX_GUI_PIXEL} is returned. 
    * Otherwise the input value is returned.
    * 
    * @param   value
    *          Character coordinate value being adjusted.
    *          
    * @return  The adjusted character coordinate value.
    */
   public static int adjustMinMaxGuiPixel(int value)
   {
      if (value < MIN_GUI_PIXEL)
      {
         return MIN_GUI_PIXEL;
      }
      else if (value > MAX_GUI_PIXEL)
      {
         return MAX_GUI_PIXEL;
      }
      
      return value;
   }
}