/*
** Module   : EmbeddedAssignment.java
** Abstract : encodes an embedded assignment such that it can be delegated
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 GES 20060204   @24269 Created initial version, which encodes an embedded assignment such 
**                           that it can be delegated to the user interface processing at a
**                           specific point in the normal assign processing, as defined by the
**                           order of the FrameElement array passed to the helper.
** 002 GES 20110801          JavaDoc improvements.
** 003 GES 20160606          Major cleanup to eliminate the need for Resolvable instances when
**                           assigning values that are literals or simple variables.  Fields
**                           are handled via the FieldReference (which is a Resolvable). Added
**                           lambda support and switched internal implementation to use it.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;
import java.util.function.*;

/**
 * This class encodes an embedded assignment expression such that evaluation can be delegated
 * to the user interface processing at a specific point in the normal assignment processing, 
 * as defined by the order of the <code>FrameElement</code> array passed to the helper.
 * <p>
 * This frame element must otherwise be completely ignored for normal user interface processing. 
 * In particular, it is not associated with any widget nor does it change the state, behavior
 * or appearance of the user interface in any way.
 * <p>
 * Typically, this element is part of a larger array of <code>FrameElement</code> objects, most 
 * of which do affect the state, behavior and appearance of the user interface.  In some cases a
 * user interface helper method handles assignment of data modified by the user back into the
 * original object.  This operation generally occurs at the end of the helper's processing.
 * <p>
 * An embedded assignment is one where the assignee and new value are defined in calling code
 * but the actual assignment is done at a later time in the user interface helper. This
 * delegation approach is used where it is necessary to intermix or insert this assignment in a
 * specific order with other embedded assignments and/or with the frame element assignments that
 * may occur.
 * <p>
 * It is also important to note that the evaluation of the given expression occurs at assignment
 * time and thus honors the current state of the calling class.  This current state includes any
 * of the variables or fields which have been modified by having had edited widget values
 * assigned back by earlier <code>FrameElement</code> objects in the array.  Since all of this
 * assignment processing is done by the user-interface runtime code AND the evaluation of the
 * embedded assignment expressions must be deferred to that time and inserted in the proper
 * order, this class is needed to support that behavior.
 */
public class EmbeddedAssignment
extends FrameElement
{
   /** The target of the assignment operation. */
   private Accessor assignee = null;
   
   /** The expression to be executed at assignment time. */
   private Supplier<BaseDataType> source = null;
   
   /**
    * Create an instance.
    *
    * @param    assignee
    *           The object to assign.  Must NOT be <code>null</code>.
    * @param    val
    *           The value to be assigned on a deferred basis.
    */
   public EmbeddedAssignment(BaseDataType assignee, int val)
   {
      this(new AccessorWrapper(assignee), val);
   }
   
   /**
    * Create an instance.
    *
    * @param    assignee
    *           The object to assign.  Must NOT be <code>null</code>.
    * @param    val
    *           The value to be assigned on a deferred basis.
    */
   public EmbeddedAssignment(Accessor assignee, int val)
   {
      this(assignee, new integer(val));
   }
   
   /**
    * Create an instance.
    *
    * @param    assignee
    *           The object to assign.  Must NOT be <code>null</code>.
    * @param    val
    *           The value to be assigned on a deferred basis.
    */
   public EmbeddedAssignment(BaseDataType assignee, double val)
   {
      this(new AccessorWrapper(assignee), val);
   }
   
   /**
    * Create an instance.
    *
    * @param    assignee
    *           The object to assign.  Must NOT be <code>null</code>.
    * @param    val
    *           The value to be assigned on a deferred basis.
    */
   public EmbeddedAssignment(Accessor assignee, double val)
   {
      this(assignee, new decimal(val));
   }
   
   /**
    * Create an instance.
    *
    * @param    assignee
    *           The object to assign.  Must NOT be <code>null</code>.
    * @param    val
    *           The value to be assigned on a deferred basis.
    */
   public EmbeddedAssignment(BaseDataType assignee, boolean val)
   {
      this(new AccessorWrapper(assignee), val);
   }
   
   /**
    * Create an instance.
    *
    * @param    assignee
    *           The object to assign.  Must NOT be <code>null</code>.
    * @param    val
    *           The value to be assigned on a deferred basis.
    */
   public EmbeddedAssignment(Accessor assignee, boolean val)
   {
      this(assignee, new logical(val));
   }
   
   /**
    * Create an instance.
    *
    * @param    assignee
    *           The object to assign.  Must NOT be <code>null</code>.
    * @param    val
    *           The value to be assigned on a deferred basis.
    */
   public EmbeddedAssignment(BaseDataType assignee, String val)
   {
      this(new AccessorWrapper(assignee), val);
   }
   
   /**
    * Create an instance.
    *
    * @param    assignee
    *           The object to assign.  Must NOT be <code>null</code>.
    * @param    val
    *           The value to be assigned on a deferred basis.
    */
   public EmbeddedAssignment(Accessor assignee, String val)
   {
      this(assignee, new character(val));
   }
   
   /**
    * Create an instance.
    *
    * @param    assignee
    *           The object to assign.  Must NOT be <code>null</code>.
    * @param    val
    *           The value to be assigned on a deferred basis.
    */
   public EmbeddedAssignment(BaseDataType assignee, BaseDataType val)
   {
      this(new AccessorWrapper(assignee), val);
   }
   
   /**
    * Create an instance.
    *
    * @param    assignee
    *           The object to assign.  Must NOT be <code>null</code>.
    * @param    source
    *           The expression to be evaluated at assignment time.
    */
   public EmbeddedAssignment(BaseDataType assignee, Resolvable source)
   {
      this(new AccessorWrapper(assignee), source);
   }
   
   /**
    * Create an instance.
    *
    * @param    assignee
    *           The object to assign.  Must NOT be <code>null</code>.
    * @param    source
    *           The expression to be evaluated at assignment time.
    */
   public EmbeddedAssignment(BaseDataType assignee, Supplier<BaseDataType> source)
   {
      this(new AccessorWrapper(assignee), source);
   }
   
   /**
    * Create an instance.
    *
    * @param    assignee
    *           The object to assign.  Must NOT be <code>null</code>.
    * @param    val
    *           The value to be assigned on a deferred basis.
    */
   public EmbeddedAssignment(Accessor assignee, final BaseDataType val)
   {
      this(assignee, () -> val);
   }
   
   /**
    * Create an instance.
    *
    * @param    assignee
    *           The object to assign.  Must NOT be <code>null</code>.
    * @param    source
    *           The expression to be evaluated at assignment time.
    */
   public EmbeddedAssignment(Accessor assignee, final Resolvable source)
   {
      this(assignee, () -> source.resolve());
   }
   
   /**
    * Create an instance.
    *
    * @param    assignee
    *           The object to assign.  Must NOT be <code>null</code>.
    * @param    source
    *           The expression to be evaluated at assignment time.
    */
   public EmbeddedAssignment(Accessor assignee, Supplier<BaseDataType> source)
   {
      this.assignee = assignee;
      this.source   = source;
   }
   
   /**
    * Determines if the element is only for the purpose of assignment or if the element must be
    * honored as something that can affect the state of the user interface.
    *
    * @return   Always <code>true</code>;
    */
   public boolean assignOnly()
   {
      return true;
   }
   
   /**
    * Get a copy of the referenced object's current runtime value (this does not return the
    * original instance itself).
    * <p>
    * This is unsupported for embedded assignments.
    *
    * @return   The copy of the current value.
    */
   public BaseDataType get()
   {
      throw new UnsupportedOperationException("Embedded assignments can't be read.");
   }
   
   /**
    * Set the referenced object's current runtime value to the given value.
    *
    * @param    val
    *           The current value to be assigned into the referenced object. This value will be
    *           ignored for embedded assignments and can be passed as <code>null</code>.
    */
   public void set(BaseDataType val)
   {
      assignee.set(source.get());
   }
}