/*
** Module   : FieldGroup.java
** Abstract : groups widgets in the frame on the server side
**
** Copyright (c) 2008-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description-----------------------------
** 001 GES 20080318   @37589 Created initial version to group widgets in
**                           a frame on the server.
** 002 NVS 20081211   @40873 getParent() method is overriden here to get
**                           the frame or another field group as the
**                           parent.
** 003 GES 20081212   @40883 Only BaseEntity can accept a field group (at
**                           this time), so safety code was added.
** 004 GES 20081230   @41002 Return a NullWidget instance instead of null
**                           for safety.
** 005 SIY 20101122          Cleaned up imports.
** 006 SIY 20111101          Updated imports.
** 007 CA  20130221          Refactored usage of TYPE attribute.
** 008 CA  20130323          The PARENT attribute must be a CommonWidget, not Object.
** 019 CA  20130927          Added LegacyResource annotation, to determine resource type.
**                           Removed NullWidget usage and fixed getFirst/LastChild to return a 
**                           handle instance.
** 020 CA  20131115          Changes related to properly instantiate dynamic resources.
** 021 VIG 20140314          Added new parameter before to the constructor to support
**                           insertion semantic of new field-group.    
** 022 VIG 20140523          Added getFirstTabItem and setFirstTabItem methods of 
**                           the tab order management.
** 023 VIG 20140612          Removing TODO in getFirstTabItem.
** 024 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 025 CA  20150416          Fixed server-side support for PARENT attribute.
**                           Fixed support for NEXT/PREV-SIBLING in relation with tree widget
**                           structures.
** 026 IAS 20151020          addDynamicWidget method.
** 027 CA  20151026          Added add/remove widget support (to allow nested frames).
** 028 HC  20151213          Improved DIALOG-BOX window parenting logic and related changes.
** 029 SBI 20151213          Removed apache dependency.
** 030 EVL 20160621          Fix for FIRST-TAB-ITEM wrong message for invisible frame.  If client
**                           is unable to provide valid widget P2J use first available field-level
**                           child as last resort.
** 031 HC  20160714          FIELD-GROUP:FIRST-TAB-ITEM must return an unknown handle when no
**                           items in the field group.
** 032 HC  20170131          Added runtime support for dynamic MENU* widgets.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.util.*;
import java.util.function.*;

import com.goldencode.p2j.util.*;
 
/**
 * Groups widgets together into a list of siblings that can be accessed
 * via the first or last entry in the list.
 */
@LegacyResource(resource = LegacyResource.FIELD_GROUP)
public class FieldGroup
extends BaseEntity<FieldGroupConfig>
implements Group,
           WidgetContainer
{
   /** The list of contained widgets. */
   private List<GenericWidget<?>> widgets = null;
   
   /** Function for adding widgets. 
    * 
    *  The FieldGroup is a server-side only object.
    *  However at least in the case of the BACKGROUND attribute they can be changed
    *  (widgets can be added)
    *  To make these changes visible the widgets should be added to the enclosing frame
    *  as well. This Consumer is used for this purpose instead of direct reference to a
    *  GenericFrame object to make this link less tight.
    */ 
    private Consumer<GenericWidget<?>> widgetAdder = null; 
   
   /**
    * Creates an instance to groups related widgets in a frame.
    *
    * @param    widgets
    *           The widgets contained in this instance. This is used as a
    *           read-only resource.
    * @param    beforeFg
    *           The right neighbor of the field-group been created or null if new field-group
    *           will be the last in the chain. If <b>beforeFg</b> is the first element in the chain
    *           the new BaseEntity will become the new first element in the chain.
    */
   public FieldGroup(List<GenericWidget<?>> widgets, FieldGroup beforeFg)
   {
      super(false, new FieldGroupConfig(), beforeFg, true);
      this.widgets = widgets;
      
      if (widgets != null)
      {
         for (GenericWidget<?> next : widgets)
         {
            if (next instanceof BaseEntity)
            {
               BaseEntity<?> be = (BaseEntity<?>) next;
               be.setGroup(this);
            }
         }
      }
   }

   /**
    * Remove the specified widget from this field-group.
    * 
    * @param    widget
    *           The widget to be removed.
    */
   public void removeWidget(GenericWidget<?> widget)
   {
      this.widgets.remove(widget);
      if (widget instanceof BaseEntity)
      {
         BaseEntity<?> be = (BaseEntity<?>) widget;
         be.setGroup(null);
      }
   }
   
   /**
    * Add the specified widget to this field-group.
    * 
    * @param    widget
    *           The widget to be added.
    */
   public void addWidget(GenericWidget<?> widget)
   {
      this.widgets.remove(widget);
      this.widgets.add(widget);
      
      if (widget instanceof BaseEntity)
      {
         BaseEntity<?> be = (BaseEntity<?>) widget;
         be.setGroup(this);
      }
   }
   
   /**
    * Sets the containing frame instance.
    * <p>
    * This updates the field-group's parentId, too.
    * 
    * @param    frame
    *           <code>GenericFrame</code> instance that contains this widget
    */
   @Override
   public void setFrame(GenericFrame frame)
   {
      super.setFrame(frame);
      
      config.parentId = frame.getFrameId();
   }
   
   /**
    * Sets the name of this handle.This method is outside interfaces and only 
    * meant for initializing this within the implementation.
    * 
    * @param name
    *        The handle type to be set.
    */
   public void name(character name)
   {
      if (name == null || name.isUnknown())
      {
         String text = "Unable to assign UNKNOWN value to attribute NAME of FIELD-GROUP widget";
         ErrorManager.recordOrShowError(4083, text, false);
         return;
      }

      super.name(name); 
   }

   /**
    * Gets the PARENT writable attribute.
    * This is an override for the BaseEntity's implementation.
    * If this field group is a member of another field group,
    * that field group is returned as parent.
    * Otherwise, the owning frame is returned as the parent.
    *
    * @return   the parent widget for this widget
    */
   public CommonWidget getParent()
   {
      // explicitly set field group
      if (group != null)
      {
         return group;
      }

      // normal case
      return frame.asWidget();
   }

   /**
    * Accesses the first widget in the list of contained widgets.
    *
    * @return   The first widget or an unknown handle if no widgets are contained.
    */
   @Override
   public handle firstChild()
   {
      return new handle((widgets == null || widgets.size() == 0) ? null : widgets.get(0));
   }

   /**
    * Accesses the last widget in the list of contained widgets.
    *
    * @return   The last widget or an unknown handle if no widgets are contained.
    */
   @Override
   public handle lastChild()
   {
      return new handle((widgets == null) ? null : widgets.get(widgets.size() - 1));
   }
   
   /**
    * Accesses the next widget in the list of contained widgets.
    *
    * @param    child
    *           The widget in the list for which to find the following child.
    *
    * @return   The next widget or a <code>null</code> if there is no subsequent widget.
    */
   public GenericWidget<?> getNextChild(GenericWidget<?> child)
   {
      if (widgets == null || child == null)
         return null;
      
      int oldIdx = widgets.indexOf(child);
      int newIdx = oldIdx + 1;
      
      return (oldIdx < 0 || newIdx >= widgets.size()) ? null : widgets.get(newIdx);
   }

   /**
    * Accesses the previous widget in the list of contained widgets.
    *
    * @param    child
    *           The widget in the list for which to find the preceding child.
    *
    * @return   The previous widget or <code>null</code> if there is no preceding widget.
    */
   public GenericWidget<?> getPrevChild(GenericWidget<?> child)
   {
      if (widgets == null || child == null)
         return null;
      
      int oldIdx = widgets.indexOf(child);
      int newIdx = oldIdx - 1;
      
      return (oldIdx < 0 || newIdx < 0) ? null : widgets.get(newIdx);
   }
   
   /**
    * Getting the FIRST-TAB-ITEM attribute.
    * 
    * @return  The widget handle that is the first in the frame tab order
    */
   public handle getFirstTabItem()
   {
      int widgetId = LogicalTerminal.getFirstTabItem(this); 

      GenericWidget<?> firstWidget = getFrame().getWidgetForId(widgetId);
      
      // sometimes when the frame is not visible the client returns null
      // in this case we try to get the first field-level widget from the group
      if (firstWidget == null)
      {
         firstWidget = getFirstFieldLevelItem();
      }

      if (firstWidget == null)
      {
         // when no item is found, unknown handle must be returned
         return new handle();
      }
         
      return firstWidget.asWidgetHandle();
   }

   /**
    * Setting the FIRST-TAB-ITEM attribute.
    * 
    * @param   wHandle
    *          The handle of widget in this field-group to assign as first in the tab order.
    */
   public void setFirstTabItem(handle wHandle)
   {
      if (wHandle == null || !wHandle._isValid() || wHandle.isUnknown())
      {
         String ms = 
            "**Unable to assign UNKNOWN value to attribute FIRST-TAB-ITEM on FIELD-GROUP widget";
         ErrorManager.recordOrThrowError(4083, ms, false);
         return;
      }
      
      CommonWidget widget = wHandle.unwrapWidget();
      
      if (!widgets.contains(widget))
      {
         String ms = "**Unable to set attribute FIRST-TAB-ITEM for FIELD-GROUP widget";
         ErrorManager.recordOrThrowError(4078, ms, false);
         return;
      }

      LogicalTerminal.setFirstTabItem(this, widget);
   }
   
   /**
    * Extracts field-group widgets.
    * 
    * @return   widgets collection.
    */
   public Collection<GenericWidget<?>> getWidgets()
   {
      return Collections.unmodifiableList(widgets);
   }

   /**
    * Adding new dynamic widget to the current group.
    *
    * @param    dynWidget  The widget to be added to current group.
    */
   public void addDynamicWidget(GenericWidget<?> dynWidget)
   {
      if (widgetAdder != null)
      {
         widgetAdder.accept(dynWidget);
         addWidget(dynWidget);
      }
      else 
      {
         throw new UnsupportedOperationException("The FieldGroup is not expandable");
      }
   }
   
   /**
    * Set the value of the widgetAdder Consumer.
    *
    * @param    widgetAdder
    *           the new value of the widgetAdder.
    */
   public void setWidgetAdder(Consumer<GenericWidget<?>> widgetAdder)
   {
      this.widgetAdder = widgetAdder;
   }

   /**
    * Indicates, that HandleChain implementation of siblings methods should be used.
    * 
    * @return   TRUE  - if the logic of parent class {@link HandleChain} is used,
    *           FALSE - if own siblings implementation logic is used. 
    */
   protected boolean useHandleChainSiblings()
   {
      return true;
   }

   /**
    * Returns the widget types that can act as parent for this widget.
    *
    * @return   see above.
    */
   @Override
   protected String[] getValidParentTypes()
   {
      // parent must be a frame
      return new String[] {LegacyResource.FRAME};
   }
   
   /**
    * Delete the resource.
    * <p>
    * Field groups need to be manually removed from the registry, when they are deleted.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   @Override
   protected boolean resourceDelete()
   {
      boolean delete = super.resourceDelete();
      
      if (delete)
      {
         LogicalTerminal.deregisterWidget(config.id.asInt());
      }
      
      return delete;
   }

   /**
    * Accesses the first field-level widget in the given field group.
    *
    * @return   The first possible field-level widget in this group or
    *           <code>null</code> if there are no field level widgets.
    */
   private GenericWidget<?> getFirstFieldLevelItem()
   {
      if (widgets == null)
         return null;
      
      // scan first available field-level widget
      for (int i = 0; i < widgets.size(); i++)
      {
         GenericWidget<?> widget = widgets.get(i);
         if (widget.config().enabled && !(widget instanceof RectangleWidget))
         {
            return widget;
         }
      }
      
      return null;
   }
}