/*
** Module   : FillInWidget.java
** Abstract : server side FillIn widget implementation
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description-----------------------------
** 001 NVS 20051026   @23490 Created initial version.
** 002 GES 20060114   @23909 Removed totally useless screen-value attr.
** 003 SIY 20060404   @25357 Support for the text group.
** 004 GES 20060803   @28375 Cleanup and refactoring.
** 005 GES 20080320   @37590 Added TYPE attribute.
** 006 SVL 20101006          Added NATIVE attribute.
** 007 SIY 20111101          Updated imports.
** 008 CA  20130221          Refactored usage of TYPE attribute.
** 009 CA  20130927          Added LegacyResource annotation, to determine resource type.
** 010 CA  20131115          Changes related to properly instantiate dynamic resources.
** 011 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 012 SVL 20141102          Make this class implement AutoZapElement.
** 013 GES 20141126          Added SET-SELECTION() method.
** 014 SVL 20150605          Properly implemented AUTO-ZAP and CURSOR-OFFSET.
** 015 IAS 20150606          Added support for the DISABLE-AUTO-ZAP and SELECTED-TEXT-related 
**                           attributes/methods.
** 016 IAS 20150801          Added support for the BLANK, DEBLANK, and PASSWORD-FIELD 
**                           attributes.
** 017 IAS 20151214          LOAD-MOUSE-POINTER support.
** 018 CA  20160303          Made attribute setters be a no-op if the value doesn't change.
** 019 EVL 20161028          Adding change modified flag when screen value is really changing on
**                           the server side by application code for example.
** 020 OM  20170505          Logged error message for removeSelection().
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;
import com.goldencode.util.*;

/**
 * Server side FillIn widget implementation.
 */
@LegacyResource(resource = LegacyResource.FILL_IN)
public class FillInWidget
extends ControlTextWidget<FillInConfig>
implements AutoZapElement, 
           SelectableText,
           PasswordFieldInterface,
           BlankInterface,
           DeblankInterface,
           Hoverable
{
   /**
    * Default constructor.
    */
   public FillInWidget()
   {
      this(false);
   }
   
   /**
    * Default constructor.
    * 
    * @param   dynamic
    *          Flag indicating if this is a static or dynamic resource.
    */
   public FillInWidget(boolean dynamic)
   {
      super(dynamic, new FillInConfig());
   }

   /**
    * Get value of the AUTO-ZAP attribute of the focused fill-in.
    *
    * @return   Value of the AUTO-ZAP attribute of the focused fill-in. If there is no focused
    *           fill-in, <code>true</code> is returned.
    */
   static logical isFocusedAutoZap()
   {
      FillInWidget focused = getFocusedFillIn();
      return focused != null ? new logical(focused.config.autoZap) : new logical(true);
   }

   /**
    * Get focused fill-in.
    *
    * @return focused fill-in widget or <code>null</code> if there is no focused fill-in.
    */
   private static FillInWidget getFocusedFillIn()
   {
      handle h = LogicalTerminal.focus();
      if (!h.isUnknown())
      {
         WrappedResource obj = h.getResource();
         if (obj.getClass().isAssignableFrom(FillInWidget.class))
            return (FillInWidget) obj;
      }

      return null;
   }

   /**
    * Sets the AUTO-RETURN option.
    * 
    * @param    autoReturn
    *           <code>true</code> if the option is set to ON.
    */
   public void setAutoReturn(boolean autoReturn)
   {
      if (config.autoReturn == autoReturn)
      {
         return;
      }
      
      config.autoReturn = autoReturn;
      pushScreenDefinition();
   }

   /**
    * Sets a read only state.
    * 
    * @param   newReadOnly
    *          Is widget read only.
    */
   public void setReadOnly(boolean newReadOnly)
   {
      if (config.readOnly == newReadOnly)
      {
         return;
      }
      
      config.readOnly = newReadOnly;
      pushScreenDefinition();
   }

   /**
    * Start text group of widgets (word wrap mode).
    * 
    * @param   prev
    *          Reference to previous fill-in widget or null if no such
    *          widgets.
    */
   public void startGrouping(FillInWidget prev)
   {
      config.grouping = true;
      
      if (prev != null)
      {
         prev.setNextGroupItem(this);
         config.prevId = prev.getId();
      }
   }

   /**
    * Link widget with the next widget in the text group.
    * 
    * @param   next
    *          Reference to next widget in group.
    */
   public void setNextGroupItem(FillInWidget next)
   {
      if (next != null)
         config.nextId = next.getId();
   }

   /**
    * Reset grouping flag and links to previous and next widget in the text
    * group.
    */
   public void resetGrouping()
   {
      config.grouping = false;
      config.nextId = -1;
      config.prevId = -1;
   }

   /**
    * Set value of the NATIVE attribute.
    *
    * @param   nativ
    *          New value for the attribute.
    */
   public void setNative(boolean nativ)
   {
      config.nativ = nativ;
   }

   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   @Override
   public logical setSelection(int64 start, int64 end)
   {
      return setSelection(start == null || start.isUnknown() ? 0 : start.intValue(), 
            end == null || end.isUnknown() ? 0 : end.intValue());
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   @Override
   public logical setSelection(int64 start, int end)
   {
      return setSelection(start == null || start.isUnknown() ? 0 : start.intValue(), end);
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   @Override
   public logical setSelection(int start, int64 end)
   {
      return setSelection(start, end == null || end.isUnknown() ? 0 : end.intValue());
   }
   
   /**
    * Selects text in a widget between start and end points.
    * 
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   @Override
   public logical setSelection(int start, int end)
   {
      if (!canAccess("SET-SELECTION"))
      {
         return new logical();
      }

      if (start == 0 || end == 0)
      {
         ErrorManager.recordOrShowError(4057, 
               String.format(
               "Attribute SET-SELECTION for the %s %s has an invalid value of 0", 
               type(), widgetName()),
               false);
         return new logical();
      }
      int s1 = Math.min(start, end);
      int s2 = Math.max(start,  end);
      if (s1 < 0 || s2 > config.widthChars)
      {
         return new logical(false);
      }

      config.selectionStart = s1;
      config.selectionEnd = s2;
      config.selectionActive = true;
      pushScreenDefinition();
      return new logical(true);
   }
   
   /**
    * Get SELECTION-TEXT attribute.
    *
    * @return  SELECTION-TEXT value   
    */
   @Override
   public character getSelectionText()
   {
      if (!canAccess("SELECTION-TEXT"))
      {
         return new character();
      }

      if (!config.selectionActive)
      {
         return new character("");
      }

      character sv = getScreenValue();
      String s = sv == null || sv.isUnknown() ? "" : sv.toStringMessage();
      s += StringHelper.repeatChar(' ', (int)config.widthChars - s.length());
      s = config.selectionEnd + 1 >= config.widthChars ?
            s.substring(config.selectionStart) :
            s.substring(config.selectionStart, config.selectionEnd + 1);
            
      return new character(s);
   }

   /**
    * Get SELECTION-END attribute.
    *
    * @return  SELECTION-END value   
    */
   @Override
   public integer getSelectionEnd()
   {
      if (!canAccess("SELECTION-END"))
      {
         return new integer();
      }

      if (!config.selectionActive)
      {
         return new integer();
      }
      return new integer(config.selectionEnd);
   }

   /**
    * Get SELECTION-START attribute.
    *
    * @return  SELECTION-START value   
    */
   @Override
   public integer getSelectionStart()
   {
      if (!canAccess("SELECTION-START"))
      {
         return new integer();
      }

      if (!config.selectionActive)
      {
         return new integer();
      }
      return new integer(config.selectionStart);
   }

   /**
    * Get TEXT-SELECTED attribute.
    *
    * @return  TEXT-SELECTED value   
    */
   @Override
   public logical isTextSelected()
   {
      if (!canAccess("TEXT-SELECTION"))
      {
         return new logical();
      }
      return new logical(config.selectionActive);
   }

   /**
    * Clears text selection
    */
   @Override
   public logical clearSelection()
   {
      boolean wasSelection = config.selectionActive;
      config.selectionActive = false;
      pushScreenDefinition();
      return new logical(wasSelection);
   }
   
   /**
    * Removes selected text. If no text is selected act as pressing DEL key by removing the
    * character to the right of th e cursor.
    * 
    * @return always true/yes
    */
   @Override
   public logical removeSelection()
   {
      if (LogicalTerminal.isChui())
      {
         String msg = "Progress does not support the EDIT-CLEAR attribute for the FILL-IN" +
                      " widget in this display environment";
         ErrorManager.displayError(4088, msg, true);
         
         return new logical();
      }
      LogicalTerminal.removeEditSelection(getId());
      return new logical(true);
   }
   
   /**
    * Sets the DISABLE-AUTO-ZAP option.
    * 
    * @param    disableAutoZap
    *           <code>true</code> if the option is set to ON.
    */
   @Override
   public void setDisableAutoZap(boolean disableAutoZap)
   {
      if (config.disableAutoZap == disableAutoZap)
      {
         return;
      }
      
      config.disableAutoZap = disableAutoZap;
      pushScreenDefinition();
   }

   /**
    * Set DISABLE-AUTO-ZAP attribute. 
    * 
    * @param    value
    *           New value for the attribute.
    */
   @Override
   public void setDisableAutoZap(logical value)
   {
      // if value argument is unknown _literal_ then 4GL generates error 4083 
      // P2J cannot simulate this as long as the argument type remains 'logical'
      // as neither 'unknown' is a subclass of 'logical' nor vice versa  
      // it is possible of course to declare the arg type for this method as BaseDataType,
      // but this will be not consisten with many other similar situations.
      setDisableAutoZap(value == null || value.isUnknown() ? false : value.booleanValue());
   }

   /**
    * Get value of the DISABLE-AUTO-ZAP attribute.
    *
    * @return   value of the attribute.
    */
   @Override
   public logical isDisableAutoZap()
   {
      return new logical(config.disableAutoZap);
   }

   /**
    * Set AUTO-ZAP attribute.
    * 
    * @param    value
    *           New value for the attribute.
    */
   @Override
   public void setAutoZap(final boolean value)
   {
      if (!isFocused())
      {
         return;
      }

      if (config.autoZap == value)
      {
         return;
      }
      
      ConfigManager.syncConfigChanges(new Runnable()
      {
         @Override
         public void run()
         {
            config.autoZap = value;
         }
      });

      pushFillInDefinition();
   }

   /**
    * Set AUTO-ZAP attribute.
    *
    * @param    autoZap
    *           New value for the attribute.
    */
   @Override
   public void setAutoZap(logical autoZap)
   {
      // if autoZap argument is unknown _literal_ then 4GL generates error 4083 
      // see comment for setDisableAutoZap
      setAutoZap(autoZap == null | autoZap.isUnknown() ? false : autoZap.booleanValue());
   }

   /**
    * Get value of the AUTO-ZAP attribute.
    *
    * @return   value of the attribute.
    */
   @Override
   public logical isAutoZap()
   {
      // Oddly enough, 4GL returns zap state of the FOCUSED fill-in, not necessarily this one
      return isFocusedAutoZap();
   }

   /**
    * Set value of the PASSWORD-FIELD attribute.
    *
    * @param    value
    *           Value of the attribute.
    */
   @Override
   public void setPasswordField(boolean value)
   {
      if (config.passwordField == value)
      {
         return;
      }
      
      config.passwordField = value;
      pushScreenDefinition();
   }

   /**
    * Set value of the PASSWORD-FIELD attribute.
    *
    * @param    value
    *           Value of the attribute.
    */
   @Override
   public void setPasswordField(logical value)
   {
      if (value == null || value.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
               "Unable to assign UNKNOWN value to attribute PASSWORD-FIELD on %s %s", 
               type(), widgetName()),
               false);
         return;
      }

      setPasswordField(value.booleanValue());
   }

   /**
    * Get value of the PASSWORD-FIELD attribute.
    *
    * @return   value of the attribute.
    */
   @Override
   public logical isPasswordField()
   {
      return new logical(config.passwordField);
   }

   /**
    * Set value of the BLANK attribute.
    *
    * @param    value
    *           Value of the attribute.
    */
   @Override
   public void setBlank(boolean value)
   {
      if (config.blank == value)
      {
         return;
      }
      
      config.blank = value;
      pushScreenDefinition();
   }

   /**
    * Set value of the BLANK attribute.
    *
    * @param    value
    *           Value of the attribute.
    */
   @Override
   public void setBlank(logical value)
   {
      if (value == null || value.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
               "Unable to assign UNKNOWN value to attribute BLANK on %s %s", 
               type(), widgetName()),
               false);
         return;
      }

      setBlank(value.booleanValue());
   }

   /**
    * Get value of the BLANK attribute.
    *
    * @return   value of the attribute.
    */
   @Override
   public logical isBlank()
   {
      return new logical(config.blank);
   }

   /**
    * Set value of the DEBLANK attribute.
    *
    * @param    value
    *           Value of the attribute.
    */
   @LegacyAttribute(name = "DEBLANK", setter = true)
   public void setDeblank(boolean value) 
   {
      if (config.deblank == value)
      {
         return;
      }
      
      config.deblank = value;
      pushScreenDefinition();
   }

   /**
    * Set value of the DEBLANK attribute.
    *
    * @param    value
    *           Value of the attribute.
    */
   @LegacyAttribute(name = "DEBLANK", setter = true)
   public void setDeblank(logical value)
   {
      if (value == null || value.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
               "Unable to assign UNKNOWN value to attribute DEBLANK on %s %s", 
               type(), widgetName()),
               false);
         return;
      }
      setDeblank(value.booleanValue());
   }

   /**
    * Get value of the DEBLANK attribute.
    *
    * @return   value of the attribute.
    */
   @LegacyAttribute(name = "DEBLANK")
   public logical isDeblank()
   {
      return new logical(config.deblank);
   }

   /**
    * Get value of the CURSOR-OFFSET attribute.
    *
    * @return  value of the attribute.
    */
   public integer getCursorOffset()
   {
      if (isFocused())
      {
         // cursor can take position after the last character of the fill-in
         int offset = Math.min(config.cursorOffset + 1, (int) config.widthChars);
         return new integer(offset);
      }
      else
      {
         return new integer();
      }
   }

   /**
    * Sets the position of the cursor. Corresponds CURSOR-OFFSET attribute.
    *
    * @param    offset
    *           The character index at which the cursor is to be placed.
    */
   public void setCursorOffset(double offset)
   {
      setCursorOffset(new decimal(offset));
   }

   /**
    * Sets the position of the cursor. Corresponds CURSOR-OFFSET attribute.
    *
    * @param    offset
    *           The character index at which the cursor is to be placed.
    */
   public void setCursorOffset(NumberType offset)
   {
      setCursorOffset(offset, name().toStringMessage());
   }

   /**
    * Sets the position of the cursor. Corresponds CURSOR-OFFSET attribute.
    *
    * @param    offset
    *           The character index at which the cursor is to be placed.
    * @param    widgetName
    *           Widget name. Used for error handling.
    */
   public void setCursorOffset(NumberType offset, String widgetName)
   {
      if (!isFocused())
         return;

      if (offset.isUnknown())
      {
         ErrorManager.recordOrShowError(4083,
               "**Unable to assign UNKNOWN value to attribute CURSOR-OFFSET on FILL-IN " +
                  widgetName, false, false);
         return;
      }

      final int val = offset.intValue();
      if (val < 1 || val > config.widthChars)
      {
         ErrorManager.recordOrShowError(4057,
               String.format("**Attribute CURSOR-OFFSET for the FILL-IN %s has an invalid " +
                     "value of %d", widgetName, val),
               false, false);
         return;
      }

      ConfigManager.syncConfigChanges(new Runnable()
      {
         @Override
         public void run()
         {
            config.cursorOffset = val - 1;
         }
      });

      pushFillInDefinition();
   }

   /**
    * Obtain the current value in the screen buffer of the backing data for
    * this widget and return it as a <code>character</code> type. If
    * there is no value in the screen buffer (if it is <code>null</code>
    * which is the representation for an uninitialized value), then return
    * the empty string.
    *
    * @return   The value as a <code>character</code> type or the empty
    *           string if the screen buffer value is uninitialized.
    */
   @Override
   public character getScreenValue()
   {
      character sv = super.getScreenValue();
      if (sv == null || sv.isUnknown() || !config.deblank)
      {
         return sv;
      }
      return new character(sv.toStringMessage().replaceAll("^\\s+", ""));
   }

   /**
    * Set the current value in the screen buffer of the backing data for
    * this widget.  If the given value is <code>null</code> then this
    * widget will be set to the uninitialized value.
    *
    * @param    value
    *           The new value for the widget, use <code>null</code> to set
    *           the value as uninitialized.
    */
   @Override
   public void setScreenValue(character value)   
   {
      // check if we really need to change anything
      character sv = getScreenValue();
      if ((sv == null && value == null) || sv.equalsCaseSensitively(value))
      {
         return;
      }
      
      config.modified = true;
      super.setScreenValue(value);
   }
   
   /**
    * Set the MOUSE-POINTER attribute
    * 
    * @param name
    *        the attribute value
    */
   @Override
   public void setMousePointerName(String name)
   {
      config.mousePointer = name;
   }

   /**
    * Get the MOUSE-POINTER attribute
    * 
    * @return the attribute value 
    *        
    */
   @Override
   public String getMousePointerImpl()
   {
      return config.mousePointer;
   }

   /**
    * Push screen definition instance down to the client. Supports both fill-ins for in-browse
    * editing and normal fill-ins.
    *
    */
   private void pushFillInDefinition()
   {
      if (frame != null)
      {
         pushScreenDefinition();
      }
      else
      {
         ScreenDefinition def = new ScreenDefinition();
         def.addConfig(config);
         LogicalTerminal.pushScreenDefinition(new ScreenDefinition[]{def});
      }
   }

   /**
    * Check if this fill-in is focused.
    *
    * @return <code>true</code> if this fill-in is focused.
    */
   private boolean isFocused()
   {
      handle h = LogicalTerminal.focus();
      if (!h.isUnknown())
      {
         WrappedResource obj = h.getResource();
         if (obj.equals(this))
            return true;
      }
      return false;
   }
}
