/*
** Module   : FontTable.java
** Abstract : Implementation of the FONT-TABLE resource plus other misc font-related APIs.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CA  20140701 Created initial version.
** 002 CA  20140711 Added legacy metrics support and font details for legacy font-tables.
** 003 MAG 20140827 Move common code to BaseSysTable class.
** 004 CA  20141220 Removed some confusing blanks in getLegacyTextMetrics logging messages.
** 005 CA  20150317 Added the standard font table, tracking of fixed-size fonts and font's max 
**                  width. 
** 006 CA  20150507 Fixed the default window title font.
** 007 CA  20150502 The resource ID needs to be kept at the resource, so it can be accessed after
**                  the resource is deleted.
** 008 CA  20150519 Added implicit font size in the font metrics XML.
**                  Refactored so that the font specifications are read in only one trip.
**                  Added API to transfer the entire text metrics DB from the server to the 
**                  client-side, in only one trip.
** 009 HC  20150618 Added ALERT_BOX_FONT system font.
** 010 CA  20150808 Added support for font aliases; the per-user "font-aliases" container can have
**                  any number of mappings of font IDs (i.e. "default-font", "default-fixed-font"
**                  font0, font1, for font table fonts, or system font names) to a specific font.
**                  This will replace the actual font used for drawing.  Scaling is still 
**                  performed for this font, if its driver-level metrics differ than the legacy 
**                  ones for the original, mapped font.
**                  Added support for the "underline" font option.
** 011 CA  20150828 Custom fonts need to be aware if they are bold or not.
** 012 IAS 20150923 Added check for ChUI mode
** 013 CA  20151008 Fix for honoring bold font definitions.
** 014 CA  20151014 If there is no BOLD font definition but there is a non-BOLD font definition,
**                  use the non-BOLD definition and let the style be BOLD.
** 015 SBI 20160107 Changed to set if there is a separate font definition for the bold style.
** 016 EVL 20160415 The default font for alert-box is "Tahoma, size=8".  Not bold.  Adding one
**                  more system font for stack trace window.
** 017 VIG 20160410 Fixed ALERT_BOX_FONT and added system MENU_FONT.
** 018 SVL 20170126 Added SORT_NUMBER_FONT.
** 019 OM  20170329 Added support for Themes.
** 020 CA  20170508 Fixed FONT_ALIASES and CUSTOM_FONTS search: is done both per server and 
**                  account.
**     EVL 20170512 Adding new approach for default font aliases using.  Adding new default
**                  custom fonts list.
** 021 OM  20170720 Added builder for FontChooserDialog. Dropped old stubs. 
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.logging.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;
import org.w3c.dom.Element;
import org.xml.sax.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;
import com.goldencode.p2j.util.LogHelper;
import com.goldencode.p2j.util.Utils.DirScope;
import com.goldencode.proxy.*;
import com.goldencode.util.*;

/**
 * Implementation of the FONT-TABLE system handle (the APIs from the {@link FontTableResource}
 * interface).  In addition, this implements:
 * <ul>
 *    <li>the <code>SYSTEM-DIALOG FONT</code> statement, via the {@link FontChooserDialog}</li>
 *    <li>the <code>PUT-KEY-VALUE FONT ALL</code> statement, via the {@link #saveFontTable} API
 *    </li>
 *    <li>the <code>PUT-KEY-VALUE FONT font-num</code> statement, via the {@link #saveFont} API
 *    </li>
 * </ul>
 * <p>
 * Currently the legacy fonts mentioned in directory file has default replacements that can be
 * loaded from jar file as resources.  Here is the table currently used as replacement rule map.
 * <table summary="">
 *    <tr><th>Legacy proprietary font</th><th>Free replacement</th></tr>
 *    <tr><td>Tahoma</td><td>Wine Tahoma</td></tr>
 *    <tr><td>Courier New</td><td>Liberation Mono</td></tr>
 *    <tr><td>Segoe UI</td><td>Open Sans</td></tr>
 *    <tr><td>Microsoft Sans Serif</td><td>Liberation Sans</td></tr>
 *    <tr><td>MS Sans Serif</td><td>Lato</td></tr>
 *    <tr><td>Fixed Sys Excelsior</td><td>DejaVu Sans</td></tr>
 * </table>
 */
public class FontTable
extends BaseSysTable
{
   /** logger */
   private static final Logger LOG = LogHelper.getLogger(FontTable.class.getName());

   /** The name of the node containing the font-table. */
   public static final String FONT_TABLE = "font-table";

   /** The name of the node containing the system font-table. */
   public static final String SYSTEM_FONT_TABLE = "system-font-table";
   
   /** The name of the node containing the window title font - system font. */
   public static final String WINDOW_TITLE_FONT = "window-title-font";

   /** The name of the node containing the alert box font - system font. */
   public static final String ALERT_BOX_FONT = "alert-box-font";

   /** The name of the node containing the stack trace window font - system font. */
   public static final String STACK_TRACE_WINDOW_FONT = "stack-trace-font";

   /** The name of the node containing the menu font - system font. */
   public static final String MENU_FONT = "menu-font";

   /** The name of the node containing the font for sort numbers in browse column headers. */
   public static final String SORT_NUMBER_FONT = "sort-number-font";
   
   /** The name of the node containing the custom fonts (i.e. bitmap fonts, others). */
   public static final String CUSTOM_FONTS = "custom-fonts";

   /**
    * The name of the node containing the font aliases.  This maps a legacy font defined in the
    * font table or the system font table to a replacement font.
    */
   public static final String FONT_ALIASES = "font-aliases";
   
   /** The name of the node containing the default font. */
   public static final String DEFAULT_FONT = "default-font";

   /** The name of the node containing the default fixed font. */
   public static final String DEFAULT_FIXED_FONT = "default-fixed-font";
   
   /** The fallback font, in case the default fonts are not specified. */
   private static final String FALLBACK_DEFAULT_FONT = "System, size 10";
   
   /** A map containing the legacy font metrics. */
   private static Map<String, LegacyFontMetrics> legacyFontMetrics = new HashMap<>();

   /** A map containing the legacy text metrics. */
   private static Map<String, LegacyTextMetrics> legacyTextMetrics = new HashMap<>();

   /** A set containing the name of the fixed fonts. */
   private static Set<String> fixedFonts = new TreeSet<>();

   /** A map containing the implicit sizes of all fonts. */
   private static Map<String, Integer> implicitSizes = new HashMap<>();

   /** A map containing the custom fonts. */
   private static Map<String, byte[]> customFonts = new HashMap<>();
   
   /** A map to store predefined custom fonts distributed with package jar */
   private static Map<String, String> defaultCustomFonts = new HashMap<>();
   
   /** Context local instance of this class. */
   private static final ContextLocal<WorkArea> local = new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      { 
         return new WorkArea();
      }
   };

   /* initialize the standard font */
   static
   {
      // initialize custom fonts default list
      defaultCustomFonts.put( "wine tahoma#false",     "wine_tahoma.ttf" );
      defaultCustomFonts.put( "wine tahoma#true",      "wine_tahoma_bold.ttf" );
      defaultCustomFonts.put( "liberation mono#false", "LiberationMono-Regular.ttf" );
      defaultCustomFonts.put( "liberation mono#true",  "LiberationMono-Bold.ttf" );
      defaultCustomFonts.put( "open sans#false",       "OpenSans-Regular.ttf" );
      defaultCustomFonts.put( "open sans#true",        "OpenSans-Bold.ttf" );
      defaultCustomFonts.put( "liberation sans#false", "LiberationSans-Regular.ttf" );
      defaultCustomFonts.put( "liberation sans#true",  "LiberationSans-Bold.ttf" );
      defaultCustomFonts.put( "dejavu sans#false",     "DejaVuSans.ttf" );
      defaultCustomFonts.put( "dejavu sans#true",      "DejaVuSans-Bold.ttf" );
      defaultCustomFonts.put( "lato#false",            "Lato-Regular.ttf" );
      defaultCustomFonts.put( "lato#true",             "Lato-Bold.ttf" );
      defaultCustomFonts = Collections.unmodifiableMap(defaultCustomFonts);
      
      // initialize the font metrics
      String[] files = { "font-metrics.xml", "font-metrics-ext.xml" };
      for (String file : files)
      {
         loadFontMetrics(file);
      }
      
      // initialize the text metrics
      loadTextMetrics("text-metrics.xml");
      
      // initialize the custom fonts
      loadCustomFonts();
      
      // freeze the maps
      fixedFonts        = Collections.unmodifiableSet(fixedFonts);
      implicitSizes     = Collections.unmodifiableMap(implicitSizes);
      legacyFontMetrics = Collections.unmodifiableMap(legacyFontMetrics);
      legacyTextMetrics = Collections.unmodifiableMap(legacyTextMetrics);
      customFonts       = Collections.unmodifiableMap(customFonts);
   }

   /**
    * Get a the instance for the FONT-TABLE system handle.  Is obtained using a call to the
    * {@link StaticProxy#obtain(Class, Class[])}, using the {@link FontTableResource} interface
    * and its methods implemented by these classes: {@link FontTable}.
    * 
    * @return   See above.
    */
   public static handle asHandle()
   {
      WorkArea wa = local.get();

      Class<?>[] classes = { FontTable.class };
      FontTableResource proxy = StaticProxy.obtain(FontTableResource.class, classes);
   
      if (wa.id == null)
      {
         wa.id = handle.resourceId(proxy);
      }

      return new handle(proxy);
   }   
   
   /**
    * Get this resource's ID.
    * 
    * @return   The resource's ID.
    */
   public static Long id()
   {
      WorkArea wa = local.get();
      
      return wa.id;
   }
   
   /**
    * Set this resource's ID.
    * <p>
    * This is a no-op for system handles.
    *  
    * @param    id
    *           The resource's ID.
    */
   public static void id(long id)
   {
      // no-op
   }

   /**
    * Implementation for the {@link WrappedResource#valid()} API.
    * 
    * @return   Always true.
    */
   public static boolean valid()
   {
      return true;
   }
   
   /**
    * Implementation for the {@link WrappedResource#unknown()} API.
    * 
    * @return   Always false.
    */
   public static boolean unknown()
   {
      return false;
   }
   
   /**
    * API needed to implement read-only attribute assignment (a 4GL 
    * "feature").
    * 
    * @param    attribute
    *           The attribute's name.
    *           
    * @see      handle#readOnlyError(handle, String)
    */
   public static void readOnlyError(String attribute)
   {
      handle.readOnlyError(asHandle(), attribute);
   }
   
   /**
    * Get the type of its associated handle.
    * 
    * @return   Always return the PSEUDO-WIDGET value.
    */
   public static character getResourceType()
   {
      return new character(LegacyResource.PSEUDO_WIDGET);
   }

   /**
    * Getter for the INSTANTIATING-PROCEDURE attribute.
    * <p>
    * The returned handle will refer a valid external procedure (persistent or not) or will be
    * unknown, in case its instantiating procedure was persistent and has been deleted.
    * 
    * @return   See above.
    */
   public static handle instantiatingProcedure()
   {
      // TODO: implement this
      throw new RuntimeException("INSTANTIATING-PROCEDURE attribute is not implemented.");
   }
   
   /**
    * Get the height in characters unit, of the default font.
    * 
    * @return   See above.
    */
   public static decimal getTextHeightChars()
   {
      return getTextHeightChars(new int64());
   }
   
   /**
    * Get the height in characters unit, of the specified font.  If the font number is unknown,
    * the default font is used.
    * 
    * @param    fontNum
    *           The entry in the font-table.
    * 
    * @return   See above.
    */
   public static decimal getTextHeightChars(long fontNum)
   {
      return getTextHeightChars(new int64(fontNum));
   }
   
   /**
    * Get the height in characters unit, of the specified font.  If the font number is unknown,
    * the default font is used.
    * 
    * @param    fontNum
    *           The entry in the font-table.
    * 
    * @return   See above.
    */
   public static decimal getTextHeightChars(int64 fontNum)
   {
      if (LogicalTerminal.isChui())
      {
         return new decimal(1);
      }

      int pheight = getTextHeightPixels(fontNum).intValue();
      double height = (double) pheight / LogicalTerminal.getClient().getPixelsPerRow();
      
      return new decimal(height, 2);
   }

   /**
    * Get the height in pixels, of the default font.
    * 
    * @return   See above.
    */
   public static integer getTextHeightPixels()
   {
      return getTextHeightPixels(new int64());
   }
   
   /**
    * Get the height in pixels, of the specified font.  If the font number is unknown,
    * the default font is used.
    * 
    * @param    fontNum
    *           The entry in the font-table.
    * 
    * @return   See above.
    */
   public static integer getTextHeightPixels(long fontNum)
   {
      return getTextHeightPixels(new int64(fontNum));
   }

   /**
    * Get the height in pixels, of the specified font.  If the font number is unknown,
    * the default font is used.
    * 
    * @param    fontNum
    *           The entry in the font-table.
    * 
    * @return   See above.
    */
   public static integer getTextHeightPixels(int64 fontNum)
   {
      if (LogicalTerminal.isChui())
      {
         return new integer(8);
      }
      
      int font = (fontNum.isUnknown() ? FontManager.DEFAULT_FONT : fontNum.intValue());
      
      return new integer(LogicalTerminal.getClient().getFontHeightPixels(font));
   }

   /**
    * Get the width of the specified text, in character units, using the default font.
    * 
    * @param    txt
    *           The text.
    *           
    * @return   See above.
    */
   public static decimal getTextWidthChars(String txt)
   {
      return getTextWidthChars(new character(txt), new int64());
   }

   /**
    * Get the width of the specified text, in character units, using the default font.
    * 
    * @param    txt
    *           The text.
    *           
    * @return   See above.
    */
   public static decimal getTextWidthChars(character txt)
   {
      return getTextWidthChars(txt, new int64());
   }

   /**
    * Get the width of the specified text, in character units, using the specified font.  
    * If the font number is unknown, the default font is used.
    * 
    * @param    txt
    *           The text.
    * @param    fontNum
    *           The entry in the font-table.
    *           
    * @return   See above.
    */
   public static decimal getTextWidthChars(String txt, long fontNum)
   {
      return getTextWidthChars(new character(txt), new int64(fontNum));
   }

   /**
    * Get the width of the specified text, in character units, using the specified font.  
    * If the font number is unknown, the default font is used.
    * 
    * @param    txt
    *           The text.
    * @param    fontNum
    *           The entry in the font-table.
    *           
    * @return   See above.
    */
   public static decimal getTextWidthChars(character txt, long fontNum)
   {
      return getTextWidthChars(txt, new int64(fontNum));
   }

   /**
    * Get the width of the specified text, in character units, using the specified font.  
    * If the font number is unknown, the default font is used.
    * 
    * @param    txt
    *           The text.
    * @param    fontNum
    *           The entry in the font-table.
    *           
    * @return   See above.
    */
   public static decimal getTextWidthChars(String txt, int64 fontNum)
   {
      return getTextWidthChars(new character(txt), fontNum);
   }

   /**
    * Get the width of the specified text, in character units, using the specified font.  
    * If the font number is unknown, the default font is used.
    * 
    * @param    txt
    *           The text.
    * @param    fontNum
    *           The entry in the font-table.
    *           
    * @return   See above.
    */
   public static decimal getTextWidthChars(character txt, int64 fontNum)
   {
      if (txt.isUnknown())
      {
         return new decimal(0);
      }
      
      if (LogicalTerminal.isChui())
      {
         return new decimal(TextOps.lengthOf(txt));
      }
      
      int pwidth = getTextWidthPixels(txt, fontNum).intValue();
      double width = (double) pwidth / (double) LogicalTerminal.getClient().getPixelsPerColumn();
      
      return new decimal(width, 2);
   }

   /**
    * Get the width of the specified text, in pixels, using the default font.
    * 
    * @param    txt
    *           The text.
    *           
    * @return   See above.
    */
   public static integer getTextWidthPixels(String txt)
   {
      return getTextWidthPixels(new character(txt), new int64());
   }

   /**
    * Get the width of the specified text, in pixels, using the default font.
    * 
    * @param    txt
    *           The text.
    *           
    * @return   See above.
    */
   public static integer getTextWidthPixels(character txt)
   {
      return getTextWidthPixels(txt, new int64());
   }

   /**
    * Get the width of the specified text, in pixels, using the specified font.  
    * If the font number is unknown, the default font is used.
    * 
    * @param    txt
    *           The text.
    * @param    fontNum
    *           The entry in the font-table.
    *           
    * @return   See above.
    */
   public static integer getTextWidthPixels(String txt, long fontNum)
   {
      return getTextWidthPixels(new character(txt), new int64(fontNum));
   }

   /**
    * Get the width of the specified text, in pixels, using the specified font.  
    * If the font number is unknown, the default font is used.
    * 
    * @param    txt
    *           The text.
    * @param    fontNum
    *           The entry in the font-table.
    *           
    * @return   See above.
    */
   public static integer getTextWidthPixels(character txt, long fontNum)
   {
      return getTextWidthPixels(txt, new int64(fontNum));
   }

   /**
    * Get the width of the specified text, in pixels, using the specified font.  
    * If the font number is unknown, the default font is used.
    * 
    * @param    txt
    *           The text.
    * @param    fontNum
    *           The entry in the font-table.
    *           
    * @return   See above.
    */
   public static integer getTextWidthPixels(String txt, int64 fontNum)
   {
      return getTextWidthPixels(new character(txt), fontNum);
   }

   /**
    * Get the width of the specified text, in pixels, using the specified font.  
    * If the font number is unknown, the default font is used.
    * 
    * @param    txt
    *           The text.
    * @param    fontNum
    *           The entry in the font-table.
    *           
    * @return   See above.
    */
   public static integer getTextWidthPixels(character txt, int64 fontNum)
   {
      if (txt.isUnknown())
      {
         return new integer(0);
      }
      
      if (LogicalTerminal.isChui())
      {
         return new integer(8 * TextOps.lengthOf(txt));
      }
      
      int font = (fontNum.isUnknown() ? FontManager.DEFAULT_FONT : fontNum.intValue());
      String stxt = txt.toStringMessage();
      ClientExports client = LogicalTerminal.getClient();
      
      LegacyTextMetrics ltm = legacyTextMetrics.get(stxt);
      String fontKey = client.getFontKey(font);
      int width = 0;
      
      if (ltm != null && fontKey != null)
      {
         LegacyFontMetrics lfm = ltm.legacyMetrics.get(fontKey);
         if (lfm != null)
         {
            // use the legacy width (in pixels)
            width = lfm.width;
         }
      }
      
      if (width == 0)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            final String msg = 
               "Could not determine width for text [%s] using font [%d] with key [%s].";
            LOG.warning(String.format(msg, stxt, font, fontKey));
         }

         width = client.getTextWidthPixels(stxt, font);
      }
      
      return new integer(width);
   }

   /**
    * Get the number of entries in the font-table.
    * 
    * @return   See above.
    */
   public static integer getNumEntries()
   {
      if (LogicalTerminal.isChui())
      {
         return new integer(0);
      }
      return new integer(LogicalTerminal.getClient().fontTableSize());
   }
   
   /**
    * Set the number of entries in the font-table.
    * 
    * @param    numEntries
    *           The new size of the font-table. 
    */
   public static void setNumEntries(long numEntries)
   {
      setNumEntries(new int64(numEntries));
   }
   
   /**
    * Set the number of entries in the font-table.
    * 
    * @param    numEntries
    *           The new size of the font-table. 
    */
   public static void setNumEntries(int64 numEntries)
   {
      int size = numEntries.isUnknown() ? 0 : numEntries.intValue();
      LogicalTerminal.getClient().adjustFontTableSize(size);
   }

   /**
    * Save the user's font-table to the environment.  This implements the PUT-KEY-VALUE FONT ALL
    * statement.
    */
   public static void saveFontTable()
   {
      // TODO: implement this
      throw new RuntimeException("PUT-KEY-VALUE FONT ALL statement is not implemented.");
   }
   
   /**
    * Save the the specified font, from the user's font-table, to the environment.  This 
    * implements the <code>PUT-KEY-VALUE FONT font-num</code> statement.
    * 
    * @param    fontNum
    *           The font-table entry to be saved.
    */
   public static void saveFont(long fontNum)
   {
      saveFont(new int64(fontNum));
   }
   
   /**
    * Save the the specified font, from the user's font-table, to the environment.  This 
    * implements the <code>PUT-KEY-VALUE FONT font-num</code> statement.
    * 
    * @param    fontNum
    *           The font-table entry to be saved.
    */
   public static void saveFont(int64 fontNum)
   {
      // TODO: implement this
      throw new RuntimeException("PUT-KEY-VALUE FONT <font-num> statement is not implemented.");
   }
   
   /**
    * Factory method for creating a font chooser dialog.
    *
    * @param   fontRef
    *          The color to be presented to user and changed if the dialog returned is accepted.
    *
    * @return  A new color dialog. Calling {@link FontChooserDialog#execute()} on this will
    *          result the client to display a theme specific color chooser dialog. 
    */
   public static FontChooserDialog createFontChooserDialog(int64 fontRef)
   {
      return new FontChooserDialog(fontRef);
   }
   
   /**
    * Factory method for creating a font chooser dialog.
    *
    * @param   fontRef
    *          The color to be presented to user and changed if the dialog returned is accepted.
    *
    * @return  A new color dialog. Calling {@link FontChooserDialog#execute()} on this will
    *          result the client to display a theme specific color chooser dialog. 
    */
   public static FontChooserDialog createFontChooserDialog(long fontRef)
   {
      return new FontChooserDialog(fontRef);
   }
   
   /**
    * Set the theme preferred fonts.
    *
    * @param   themeFonts
    *          The fonts the current installed {@code Theme} on client prefers.
    * @param   themeFontAliases
    *          The font aliases currently installed {@code Theme} on client prefers.
    */
   static void setThemePreferredFonts(Map<String, String> themeFonts,
                                      Map<String, String> themeFontAliases)
   {
      WorkArea wa = local.get();
      if (wa.themePreferredFonts != null)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING,
                  "Default Theme specific font table is redefined. Was the Theme changed?");
         }
      }
      wa.themePreferredFonts = themeFonts;
      // copy default aliases as well
      wa.themePreferredFontAliases = themeFontAliases;
   }
   
   /**
    * Read the default fonts.
    *
    * @return  An array with the font details, with default font on index o and default fixed
    *          font on index 1.
    */
   static FontDetails<?>[] readDefaultFonts()
   {
      readFontAliases();
      
      WorkArea wa = local.get();
      if (wa.defaultFonts != null)
      {
         return wa.defaultFonts;
      }
      
      Map<String, String> stdft = wa.themePreferredFonts;
      if (stdft == null)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING,
                  "Default Theme specific font table not configured. " +
                  "Call 'setThemePreferredFonts()' before 'readDefaultFonts()'.");
         }
      }
      
      String[] fontTable = new String[2];
      
      String node = FONT_TABLE + "/" + DEFAULT_FONT;
      String defFont = stdft == null ? null : stdft.get(DEFAULT_FONT);
      fontTable[0] = Utils.getDirectoryNodeString(null, node, defFont, DirScope.BOTH, null);
      fontTable[0] = safeFont(fontTable[0]);
      
      node = FONT_TABLE + "/" + DEFAULT_FIXED_FONT;
      String defFixedFont = stdft == null ? null : stdft.get(DEFAULT_FIXED_FONT);
      fontTable[1] = Utils.getDirectoryNodeString(null, node, defFixedFont, DirScope.BOTH, null);
      fontTable[1] = safeFont(fontTable[1]);
      
      String[] fontIDs = new String[] { DEFAULT_FONT, DEFAULT_FIXED_FONT };
      
      wa.defaultFonts = buildFontSpecifications(fontIDs, fontTable);
      
      return wa.defaultFonts;
   }
   
   /**
    * Read the font-table with the name of the legacy system fonts.
    * <p>
    * On index 0, is the window title font.
    * TBD: other fonts.
    * 
    * @return  The font table.  As 4GL does, this is 0-indexed.
    */
   static FontDetails<?>[] readSystemFontTable()
   {
      readFontAliases();
      
      WorkArea wa = local.get();
      
      if (wa.systemFonts != null)
      {
         return wa.systemFonts;
      }
      
      Map<String, String> stdft = wa.themePreferredFonts;
      if (stdft == null)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING,
                    "Default Theme specific font table not configured. " +
                    "Call 'setThemePreferredFonts()' before 'readSystemFontTable()'.");
         }
      }
      
      // TODO: other fonts
      
      String[] fontTable = new String[5];
      
      // WINDOW_TITLE_FONT
      String node = SYSTEM_FONT_TABLE + "/" + WINDOW_TITLE_FONT;
      String defWindowFont = stdft == null ? null : stdft.get(WINDOW_TITLE_FONT);
      fontTable[0] = Utils.getDirectoryNodeString(null, node, defWindowFont, DirScope.BOTH, null);
      fontTable[0] = safeFont(fontTable[0]);
      
      // ALERT_BOX_FONT
      node = SYSTEM_FONT_TABLE + "/" + ALERT_BOX_FONT;
      String defAlertFont = stdft == null ? null : stdft.get(ALERT_BOX_FONT);
      fontTable[1] = Utils.getDirectoryNodeString(null, node, defAlertFont, DirScope.BOTH, null);
      fontTable[1] = safeFont(fontTable[1]);
      
      // STACK_TRACE_WINDOW_FONT
      node = SYSTEM_FONT_TABLE + "/" + STACK_TRACE_WINDOW_FONT;
      String defSTFont = stdft == null ? null : stdft.get(STACK_TRACE_WINDOW_FONT);
      fontTable[2] = Utils.getDirectoryNodeString(null, node, defSTFont, DirScope.BOTH, null);
      fontTable[2] = safeFont(fontTable[2]);
      
      // MENU_FONT
      node = SYSTEM_FONT_TABLE + "/" + MENU_FONT;
      String defMenuFont = stdft == null ? null : stdft.get(MENU_FONT);
      fontTable[3] = Utils.getDirectoryNodeString(null, node, defMenuFont, DirScope.BOTH, null);
      fontTable[3] = safeFont(fontTable[3]);
      
      // SORT_NUMBER_FONT
      node = SYSTEM_FONT_TABLE + "/" + SORT_NUMBER_FONT;
      String defSoftFont = stdft == null ? null : stdft.get(SORT_NUMBER_FONT);
      fontTable[4] = Utils.getDirectoryNodeString(null, node, defSoftFont, DirScope.BOTH, null);
      fontTable[4] = safeFont(fontTable[4]);
      
      String[] fontIDs = new String[] { WINDOW_TITLE_FONT,
                                        ALERT_BOX_FONT,
                                        STACK_TRACE_WINDOW_FONT,
                                        MENU_FONT,
                                        SORT_NUMBER_FONT };
      wa.systemFonts = buildFontSpecifications(fontIDs, fontTable);

      return wa.systemFonts;
   }
   
   /**
    * Read the entire font-table.
    * 
    * @return   The font table.  As 4GL does, this is 0-indexed.
    */
   static FontDetails<?>[] readFontTable()
   {
      readFontAliases();
      
      WorkArea wa = local.get();
      
      if (wa.fontTable != null)
      {
         return wa.fontTable;
      }
      
      String[] fontTable = readTable(FONT_TABLE, "fonts", "font", wa.themePreferredFonts);
      String[] fontIDs = new String[fontTable.length];
      for (int i = 0; i < fontTable.length; i++)
      {
         fontIDs[i] = String.format("font%d", i);
      }
      
      wa.fontTable = buildFontSpecifications(fontIDs, fontTable);

      return wa.fontTable;
   }
   
   /**
    * Get the custom definition of this font.  If none available, return <code>null</code>.
    * 
    * @param    name
    *           The font name.
    * @param    bold
    *           Flag indicating if this font is bold.
    *           
    * @return   A byte array with the font's definition, else <code>null</code>.
    */
   static byte[] getCustomFontDefinition(String name, boolean bold)
   {
      String key = name.toLowerCase() + "#" + bold;
      return customFonts.get(key);
   }
   
   /**
    * Get the legacy metrics for the given font, using the font specified by the given key.
    * 
    * @param    key
    *           The font's key.
    *           
    * @return   The legacy metrics, with width on index 0 and height on index 1.
    */
   static int[] getLegacyFontMetrics(String key)
   {
      LegacyFontMetrics lfm = legacyFontMetrics.get(key);
      if (lfm == null)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.severe("Font [" + key + "] has no legacy metrics.");
         }

         return null;
      }
      
      return new int[] { lfm.width, lfm.maxwidth, lfm.height };
   }
   
   /**
    * Get the legacy metrics for the given text, using the font specified by the given key.
    * 
    * @param    text
    *           The text.
    * @param    key
    *           The font's key.
    *           
    * @return   The legacy metrics, with width on index 0 and height on index 1.
    */
   static int[] getLegacyTextMetrics(String text, String key)
   {
      LegacyTextMetrics ltm = legacyTextMetrics.get(text);
      if (ltm == null)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.severe("Text [" + text+ "] has no legacy metrics with font [" + key + "].");
         }

         return null;
      }
      
      LegacyFontMetrics lfm = ltm.legacyMetrics.get(key);
      if (lfm == null)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.severe("Text [" + text+ "] has no legacy metrics with font [" + key + "].");
         }

         return null;
      }
      
      return new int[] { lfm.width, lfm.height };
   }
   
   /**
    * Read the legacy text metrics for all the texts and each font defined in the font table, for
    * this user.
    *
    * @return  A map having as key the font and text and as value the metrics (0 index - width, 
    *          1 index - height).
    */
   static Map<String, int[]> readLegacyTextMetrics()
   {
      Map<String, int[]> legacyTexts = new HashMap<>();
      
      FontDetails<?>[] fontTable = readFontTable();
      FontDetails<?>[] sysFonts = readSystemFontTable();
      FontDetails<?>[] defaultFonts = readDefaultFonts();
      
      Set<String> uniqueKeys = new HashSet<>();
      List<FontDetails<?>> allFonts = new ArrayList<>();
      allFonts.addAll(Arrays.asList(fontTable));
      allFonts.addAll(Arrays.asList(sysFonts));
      allFonts.addAll(Arrays.asList(defaultFonts));
      
      for (FontDetails<?> font : allFonts)
      {
         String fontKey = font.key;
         if (uniqueKeys.contains(fontKey))
         {
            continue;
         }
         uniqueKeys.add(fontKey);
         
         // go through all texts and get metrics for them
         for (LegacyTextMetrics txtMetrics : legacyTextMetrics.values())
         {
            String txt = txtMetrics.text;
            String txtKey = fontKey + "#" + txt;
            
            LegacyFontMetrics lfm = txtMetrics.legacyMetrics.get(fontKey);
            if (lfm != null)
            {
               legacyTexts.put(txtKey, new int[] { lfm.width, lfm.height });
            }
         }
      }
      
      return legacyTexts;
   }
   
   /**
    * Check if the given font face name is for a fixed font, registered in the {@link #fixedFonts}
    * set.
    * 
    * @param   name
    *          The font name.
    *          
    * @return   <code>true</code> if this is a fixed font in 4GL.
    */
   static boolean isLegacyFixedFont(String name)
   {
      return name != null && fixedFonts.contains(name.toLowerCase());
   }
   
   /**
    * Get the font's implicit size, from the {@link #implicitSizes} map.
    * 
    * @param    name
    *           The font name.
    * 
    * @return   The font's implicit size, as used by 4GL.
    */
   static int getImplicitFontSize(String name)
   {
      if (name == null)
      {
         throw new NullPointerException("The font name must be specified!");
      }
      
      if (!implicitSizes.containsKey(name.toLowerCase()))
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING,
                    "No implicit size set for font " + name + "! Using 12 as implicit size.");
         }
         
         return 12;
      }
      
      return implicitSizes.get(name.toLowerCase());
   }
   
   /**
    * Read the font aliases.
    */
   private static void readFontAliases()
   {
      WorkArea wa = local.get();
      
      if (wa.fontAliases != null)
      {
         return;
      }
      
      wa.fontAliases = new HashMap<>();

      DirectoryService ds = DirectoryService.getInstance();
      
      if (!ds.bind())
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.severe("FontTable.readFontAliases - could not bind directory.");
         }
         
         return;
      }
      
      try
      {
         // these can be set per-user
         String path = Utils.findDirectoryNodePath(ds, FONT_ALIASES, "container", 
                                                   DirScope.BOTH, null);
         if (path == null)
         {
            if (LOG.isLoggable(Level.INFO))
            {
               LOG.info("No font aliases are defined in the directory.");
            }
            
            return;
         }
         
         String[] fonts = ds.enumerateNodes(path);
         for (String f : fonts)
         {
            String fontPath = path + "/" + f;
            
            String fontID = f.toLowerCase();
            String font = ds.getNodeString(fontPath, "value");
            
            FontDetails<?> fd = buildFontSpecification(null, font);
            // font aliases might have anti-aliasing mode to be ON
            fd.swingAntiAliasing = wa.swingAntiAliasing;
            
            wa.fontAliases.put(fontID, fd);
         }
      }
      finally
      {
         // now let's check if there are any default aliases that left not redefined
         if (wa.themePreferredFontAliases != null && !wa.themePreferredFontAliases.isEmpty())
         {
            // enumerate all default font aliases
            for (String s : wa.themePreferredFontAliases.keySet())
            {
               String fntAliasID = s.toLowerCase();
               // check if we have redefined default in directory
               if (!wa.fontAliases.containsKey(fntAliasID))
               {
                  // no font alias in directory, use default have got from theme
                  String fntAlias = wa.themePreferredFontAliases.get(fntAliasID);
                  
                  FontDetails<?> fd = buildFontSpecification(null, fntAlias);
                  // font aliases might have anti-aliasing mode to be ON
                  fd.swingAntiAliasing = wa.swingAntiAliasing;
                  
                  wa.fontAliases.put(fntAliasID, fd);
               }
            }
         }
         
         if (!ds.unbind())
         {
            if (LOG.isLoggable(Level.SEVERE))
            {
               LOG.severe("FontTable.readFontAliases - could not unbind directory.");
            }
         }
      }
   }
   
   /**
    * Load the custom fonts.
    */
   private static void loadCustomFonts()
   {
      DirectoryService ds = DirectoryService.getInstance();
      
      if (!ds.bind())
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.severe("FontTable.loadCustomFonts - could not bind directory.");
         }
         
         return;
      }
      
      try
      {
         String path = Utils.findDirectoryNodePath(ds, CUSTOM_FONTS, "container", 
                                                   DirScope.BOTH, null);
         if (path == null)
         {
            if (LOG.isLoggable(Level.WARNING))
            {
               LOG.warning("No custom fonts are defined in the directory.");
            }
            
            return;
         }
         
         String[] fonts = ds.enumerateNodes(path);
         for (String f : fonts)
         {
            String fontPath = path + "/" + f;
            
            String font = ds.getNodeString(fontPath + "/font", "value");
            String file = ds.getNodeString(fontPath + "/file", "value");
            Boolean bold = Boolean.TRUE.equals(ds.getNodeBoolean(fontPath + "/bold", "value"));
            
            // load font data from file system or jar resource
            byte[] bytes = getCustomFontData(font, file);
            if (bytes != null)
            {
               String key = font.toLowerCase() + "#" + bold;
               customFonts.put(key, bytes);
            }
            else
            {
               continue;
            }
         }
      }
      finally
      {
         // now let's check if there are the default fonts left not redefined enumerate all
         // default font aliases
         for (String fntKey : defaultCustomFonts.keySet())
         {
            // check if we have redefined default in directory
            if (!customFonts.containsKey(fntKey))
            {
               // no custom font in directory, use default custom one
               String fntFile = "fonts" + File.separator + defaultCustomFonts.get(fntKey);
               
               // load font data from file system or jar resource
               byte[] bytes = getCustomFontData(fntKey, fntFile);
               if (bytes != null)
               {
                  customFonts.put(fntKey, bytes);
               }
            }
         }
         
         if (!ds.unbind())
         {
            if (LOG.isLoggable(Level.SEVERE))
            {
               LOG.severe("FontTable.loadCustomFonts - could not unbind directory.");
            }
         }
      }
   }

   /**
    * Load the font data from given location.
    * 
    * @param    fntName
    *           The name for custom font.
    * @param    fntPath
    *           The font file to load data.
    *          
    * @return   The byte array of data from fily system or jar resource.  In case of error
    *           returns null.
    */
   private static byte[] getCustomFontData(String fntName, String fntPath)
   {
      // the result
      byte[] bytes = null;
      
      try
      {
         InputStream is = null;
               
         if (new File(fntPath).exists())
         {
            // first try it as an external file
            is = new FileInputStream(new File(fntPath));
         }
         else
         {
            // jar files must use Linux style separator
            if (fntPath.indexOf('\\') != -1)
            {
               fntPath = fntPath.replace("\\", "/");
            }
            // try load it from the jars
            is = ClassLoader.class.getResourceAsStream((fntPath.startsWith("/") ? "" : "/") +
                                                       fntPath);
         }
               
         if (is == null)
         {
            if (LOG.isLoggable(Level.SEVERE))
            {
               LOG.log(Level.SEVERE,
                       "Could not load the custom font [" + fntName + "] from file " + fntPath);
            }

            return null;
         }

         ByteArrayOutputStream baos = new ByteArrayOutputStream();
         bytes = new byte[512];
         while (is.available() > 0)
         {
            int length = is.read(bytes);
            baos.write(bytes, 0, length);
         }
               
         bytes = baos.toByteArray();
      }
      catch (IOException e)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE,
                    "Error loading the custom font [" + fntName + "] from file " + fntPath,
                    e);
         }
         bytes = null;
      }
      finally
      {
         return bytes;
      }
   }
   
   /**
    * Load the text metrics from the specified file.
    * 
    * @param    file
    *           The XML file with the text metrics.
    */
   private static void loadTextMetrics(String file)
   {
      try
      {
         InputStream is = ClassLoader.class.getResourceAsStream("/" + file);
         if (is == null)
         {
            if (LOG.isLoggable(Level.SEVERE))
            {
               LOG.log(Level.SEVERE, "Could not load the legacy text metrics from file " + file);
            }

            return;
         }
         
         String systemID = generateSystemID();

         Document root = XmlHelper.parse(is, true, systemID);
         NodeList texts = root.getElementsByTagName("text");
         int len = texts.getLength();
         for (int i = 0; i < len; i++)
         {
            Element text = (Element) texts.item(i);
            
            String value = text.getAttribute("value");
            LegacyTextMetrics metrics = legacyTextMetrics.computeIfAbsent(value,
                  k -> new LegacyTextMetrics(value));
            
            NodeList fonts = text.getElementsByTagName("font");
            loadMetricsWorker(fonts, metrics.legacyMetrics);
         }
      }
      catch (ParserConfigurationException | IOException | SAXException e)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, "Error loading the legacy text metrics from file " + file, e);
         }
      }
   }
   
   /**
    * Load the font metrics from the specified file.
    * 
    * @param    file
    *           The XML file with the text metrics.
    */
   private static void loadFontMetrics(String file)
   {
      try
      {
         InputStream is = ClassLoader.class.getResourceAsStream("/" + file);
         if (is == null)
         {
            if (LOG.isLoggable(Level.SEVERE))
            {
               LOG.log(Level.SEVERE, "Could not load the legacy font metrics from file " + file);
            }

            return;
         }
         
         String systemID = generateSystemID();
         
         Document root = XmlHelper.parse(is, true, systemID);
         NodeList fonts = root.getElementsByTagName("font");
         loadMetricsWorker(fonts, legacyFontMetrics);
         
         NodeList fontDetailNodes = root.getElementsByTagName("font-details");
         int len = fontDetailNodes.getLength();
         for (int i = 0; i < len; i++)
         {
            Element elFont = (Element) fontDetailNodes.item(i);

            String name = elFont.getAttribute("name");
            name = name.toLowerCase();
            
            boolean fixed = "true".equalsIgnoreCase(elFont.getAttribute("fixed"));
            if (fixed)
            {
               fixedFonts.add(name);
            }
            
            int implicitSize = Integer.parseInt(elFont.getAttribute("implicit-size"));
            implicitSizes.put(name, implicitSize);
         }
      }
      catch (ParserConfigurationException | IOException | SAXException e)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, "Error loading the legacy text metrics from file " + file, e);
         }
      }
   }
   
   /**
    * Load the font metrics from the specified DOM nodes and load them in the specified map.
    * 
    * @param    fonts
    *           The DOM nodes defining the font metrics.
    * @param    m
    *           The map where to load the metrics.
    */
   private static void loadMetricsWorker(NodeList fonts, Map<String, LegacyFontMetrics> m)
   {
      int len = fonts.getLength();
      for (int i = 0; i < len; i++)
      {
         // Get next element and reject if not required type.
         Element elFont = (Element) fonts.item(i);
         
         String name       = elFont.getAttribute("name");
         String ssize      = elFont.getAttribute("size");
         String sbold      = elFont.getAttribute("bold");
         String sitalic    = elFont.getAttribute("italic");
         String sunderline = elFont.getAttribute("underline");
         String savgwidth  = elFont.getAttribute("width");
         String smaxwidth  = elFont.getAttribute("max-width");
         String sheight    = elFont.getAttribute("height");
         
         int     size      = Integer.parseInt(ssize);
         boolean bold      = "true".equalsIgnoreCase(sbold);
         boolean italic    = "true".equalsIgnoreCase(sitalic);
         boolean underline = "true".equalsIgnoreCase(sunderline);
         int     avgwidth  = Integer.parseInt(savgwidth);
         int     maxwidth  = smaxwidth.length() == 0 ? -1 : Integer.parseInt(smaxwidth);
         int     height    = Integer.parseInt(sheight);
         
         String key = FontManager.getFontKey(name, size, bold, italic, underline);
         LegacyFont font = new LegacyFont(name, size, bold, italic, underline);
         LegacyFontMetrics lfm = new LegacyFontMetrics(font, avgwidth, maxwidth, height);
         
         m.put(key, lfm);
      }
   }
   
   /**
    * Generate a system ID to allow us to find the DTD.  This approach assumes that the DTD lives 
    * in the same jar file as this class.
    * 
    * @return    A system ID to be used for XML validation.
    */
   private static String generateSystemID()
   {
      String clsLoc = FontTable.class.getName().replace('.', '/') + ".class";
      URL clsURL = ClassLoader.getSystemResource(clsLoc);
      String path = clsURL.toString();
      String systemID = path.substring(0, path.indexOf(clsLoc));
      
      return systemID;
   }
   
   /**
    * If the given font is <code>null</code> or empty string, return the 
    * {@link #FALLBACK_DEFAULT_FONT}.  Otherwise, return the given font.
    * 
    * @return   See above.
    */
   private static String safeFont(String font)
   {
      return (font == null || font.trim().length() == 0 ? FALLBACK_DEFAULT_FONT : font);
   }
   
   /**
    * Build font specifications for all fonts in the given array.
    * 
    * @param    fontIDs
    *           An array with the fonts' unique ID.
    * @param    fontTable
    *           The fonts.
    *           
    * @return   A font specification, for each font in the array.
    */
   private static FontDetails<?>[] buildFontSpecifications(String[] fontIDs, String[] fontTable)
   {
      FontDetails<?>[] fonts = new FontDetails<?>[fontTable.length];

      for (int f = 0; f < fontTable.length; f++)
      {
         String font = fontTable[f];
         
         fonts[f] = buildFontSpecification(fontIDs[f], font);
      }

      return fonts;
   }
   
   /**
    * Build font specifications for the specified font
    * 
    * @param    fontID
    *           The font unique identification key from the font-table or the system-font-table.
    *           May be <code>null</code> if no alias is required for this font.
    * @param    font
    *           The font.
    *           
    * @return   A font specification.
    */
   private static FontDetails<?> buildFontSpecification(String fontID, String font)
   {
      boolean isBold = false;
      boolean isItalic = false;
      boolean isUnderline = false; // TODO:
      float ptSize = 0;
      String fontName = "";
      
      if (font.contains(","))
      {
         String[] tokens = font.split(",");
         for (int i = 1; i < tokens.length; i++)
         {
            String tok = tokens[i].trim().toLowerCase();
            if (tok.startsWith("size"))
            {
               tok = tok.substring("size".length()).trim();
               while (!Character.isDigit(tok.charAt(0)))
               {
                  tok = tok.substring(1);
               }
               
               ptSize = Float.parseFloat(tok);
            }
            else if (tok.startsWith("bold"))
            {
               isBold = true;
            }
            else if (tok.startsWith("italic"))
            {
               isItalic = true;
            }
            else if (tok.startsWith("underline"))
            {
               isUnderline = true;
            }
         }
         
         fontName = tokens[0].trim();
      }
      else
      {
         fontName = font;
      }

      if (Float.compare(ptSize, 0.0f) == 0)
      {
         // use the implicit size
         ptSize = getImplicitFontSize(fontName);
      }
      
      int size = Math.round(ptSize);

      String key = FontManager.getFontKey(fontName, size, isBold, isItalic, isUnderline);
      
      boolean fixedSize = isLegacyFixedFont(fontName);
      
      int[] metrics = getLegacyFontMetrics(key);
      int legacyWidth    = (metrics == null ? -1 : metrics[0]);
      int legacyMaxWidth = (metrics == null ? -1 : metrics[1]);
      int legacyHeight   = (metrics == null ? -1 : metrics[2]);
      
      byte[] fontDefinition = getCustomFontDefinition(fontName, isBold);
      boolean boldStyle = isBold;
      boolean boldFontDefinition = isBold && fontDefinition != null;
      
      if (isBold)
      {
         if (fontDefinition == null)
         {
            boldStyle = true;

            // check the non-bold definition; if it exists, then let the style use bold AND use
            // that non-bold definition
            fontDefinition = getCustomFontDefinition(fontName, false);
         }
         else
         {
            // there is a bold definition, do not re-bold the font
            boldStyle = false;
         }
      }

      // for bold fonts with custom definition, setting its "bold" style results in incorrect
      // drawing in some drivers (i.e. Web). for this reason, when creating the style, the "bold"
      // is removed, so that the font can be drawn using its look&feel as defined in the .ttf
      FontStyle style = FontStyle.createStyle(boldStyle, 
                                              isItalic, 
                                              isUnderline);
      // legacy fonts do not use anti-aliasing mode
      FontDetails<?> fontDetails = new FontDetails<>(key,
                                                     fontName,
                                                     style, 
                                                     size,
                                                     legacyWidth, 
                                                     legacyMaxWidth, 
                                                     legacyHeight, 
                                                     fontDefinition, 
                                                     fixedSize);
      fontDetails.boldFontDefinition = boldFontDefinition;

      // check for an alias
      if (fontID != null)
      {
         WorkArea wa = local.get();

         FontDetails fontAlias = wa.fontAliases.get(fontID.toLowerCase());
         if (fontAlias != null)
         {
            fontDetails.fontAlias = fontAlias;
         }
      }
      
      return fontDetails;
   }
   
   /**
    * Structure with dDetails about a given font.
    */
   private static class LegacyFont
   {
      /** The font name. */
      private final String name;

      /** The font point size. */
      private final int size;
      
      /** Flag indicating if the font style is bold. */
      private final boolean bold;
      
      /** Flag indicating if the font style is italic. */
      private final boolean italic;
      
      /** Flag indicating if the font style is underline. */
      private final boolean underline;
      
      /**
       * Create a new font with the specified details.
       * 
       * @param    name
       *           The font name.
       * @param    size
       *           The font point size.
       * @param    bold
       *           Flag indicating if the font style is bold.
       * @param    italic
       *           Flag indicating if the font style is italic.
       * @param    underline
       *           Flag indicating if the font style is underline.
       */
      public LegacyFont(String name, int size, boolean bold, boolean italic, boolean underline)
      {
         this.name      = name;
         this.size      = size;
         this.bold      = bold;
         this.italic    = italic;
         this.underline = underline;
      }
   }

   /**
    * Structure containing the legacy metrics, for a certain font.  These metrics may be 
    * associated with the font itself or with a certain text.
    */
   private static class LegacyFontMetrics
   {
      /** The details about the legacy font. */
      private final LegacyFont font;
      
      /** The legacy width. */
      private final int width;
      
      /** The legacy max width of this font. */
      private final int maxwidth;
      
      /** The legacy height. */
      private final int height;
      
      /**
       * Instantiate new metrics details.
       * 
       * @param    font
       *           The associated font.
       * @param    width
       *           The legacy width.
       * @param    maxwidth
       *           The legacy max width of this font.
       * @param    height
       *           The legacy height.
       */
      public LegacyFontMetrics(LegacyFont font, int width, int maxwidth, int height)
      {
         this.font     = font;
         this.width    = width;
         this.height   = height;
         this.maxwidth = maxwidth;
      }
   }

   /**
    * Structure containing the legacy text metrics, for a certain text.
    */
   private static class LegacyTextMetrics
   {
      /** The text to which the metrics belong. */
      private final String text;
      
      /** A font-key to legacy metrics map. */
      private final Map<String, LegacyFontMetrics> legacyMetrics = new HashMap<>();
      
      /**
       * Instantiate a new text metrics.
       * 
       * @param    text
       *           The text.
       */
      public LegacyTextMetrics(String text)
      {
         this.text = text;
      }
   }
   
   /**
    * Builder class that takes care of all options of the {@code SYSTEM-DIALOG FONT} statement.
    * It collects all options through dedicated methods and store them as an internal
    * configuration. When {@link #execute} is called, a dialog box is displayed that allows the
    * user to select and associate a system font with the specified font number.
    * The {@code SYSTEM-DIALOG FONT} statement provides a dialog box appropriate to the graphical
    * environment in which it runs.
    */
   public static class FontChooserDialog
   {
      /**
       * The font reference to be associated by the dialog with the system font the user selects
       * in the dialog.
       */
      private int fontRef = 0;
      
      /**
       * A logical variable to return the status of the user’s dialog interaction. If the
       * user chooses the OK button, the dialog sets logical-variable to TRUE. If the user chooses
       * the Cancel button, the dialog sets logical-variable to FALSE.
       */
      private logical toUpdate = null;
      
      /**
       * Specifies the window from which the dialog box is displayed. The value must be the 
       * handle of a window.
       */
      private handle inWindow = null;
      
      /**
       * Flag for allowing the font dialog to provide only fonts that contain character
       * representations and that do not include graphic symbols.
       */
      private boolean ansiOnly = false;
      
      /**
       * Flag for allowing the font dialog to provide only mono-spaced fonts.
       */
      private boolean fixedOnly = false;
      
      /**
       * Constructor. Create a dialog builder with mandatory parameters. 
       *  
       * @param   fontRef
       *          The index of font to be selected by user. 
       * 
       * @see     #fontRef
       */
      public FontChooserDialog(int64 fontRef)
      {
         this(fontRef.isUnknown() ? 0 : fontRef.intValue());
      }
      
      /**
       * Constructor. Create a dialog builder with mandatory parameters. 
       *
       * @param   fontRef
       *          The index of font to be selected by user. 
       *
       * @see     #fontRef
       */
      public FontChooserDialog(long fontRef)
      {
         if (fontRef > 0 && fontRef <= getNumEntries().intValue())
         {
            this.fontRef = (int) fontRef;
         }
         else
         {
            // negative fonts seem to be allowed, but this corrupts the font-table!
            String msg = "Invalid font number, %d, given to common dialog";
            ErrorManager.recordOrThrowError(3188, String.format(msg, fontRef), false);
         }
      }
   
      /**
       * Force the dialog to provide only fonts that contain character representations and that
       * do not include graphic symbols.
       *
       * @return  The reference to current {@code FontChooserDialog}. Used for chaining.
       * 
       * @see     #ansiOnly
       */
      public FontChooserDialog ansiOnly()
      {
         this.ansiOnly = true;
         return this; 
      }
      
      /**
       * Force the dialog to provide only mono-spaced fonts.
       *
       * @return  The reference to current {@code FontChooserDialog}. Used for chaining.
       *
       * @see     #fixedOnly
       */
      public FontChooserDialog fixedOnly()
      {
         this.fixedOnly = true;
         return this;
      }
      
      /**
       * Has no effect; supported only for backward compatibility.
       * 
       * @param   maxVal
       *          Ignored.
       *
       * @return  The reference to current {@code FontChooserDialog}. Used for chaining. 
       */
      public FontChooserDialog setMaxSize(NumberType maxVal)
      {
         return this;
      }
      
      /**
       * Has no effect; supported only for backward compatibility.
       *
       * @param   maxVal
       *          Ignored.
       *
       * @return  The reference to current {@code FontChooserDialog}. Used for chaining. 
       */
      public FontChooserDialog setMaxSize(long maxVal)
      {
         return this;
      }
      
      /**
       * Has no effect; supported only for backward compatibility.
       *
       * @param   minVal
       *          Ignored.
       *
       * @return  The reference to current {@code FontChooserDialog}. Used for chaining. 
       */
      public FontChooserDialog setMinSize(NumberType minVal)
      {
         return this;
      }
      
      /**
       * Has no effect; supported only for backward compatibility.
       *
       * @param   minVal
       *          Ignored.
       *
       * @return  The reference to current {@code FontChooserDialog}. Used for chaining. 
       */
      public FontChooserDialog setMinSize(long minVal)
      {
         return this;
      }
      
      /**
       * Configures the variable to be set when the selection dialog is closed.
       *
       * @param   update
       *          A variable that will hold the status of the dialog.
       *
       * @return  The reference to current {@code FontChooserDialog}. Used for chaining. 
       *
       * @see     #toUpdate
       */
      public FontChooserDialog update(logical update)
      {
         this.toUpdate = update;
         return this;
      }
      
      /**
       * Configures the dialog to be created as child of a certain windows.
       *
       * @param   window
       *          A handle to a window that will be the parent of the dialog.
       *
       * @return  The reference to current {@code FontChooserDialog}. Used for chaining.
       *
       * @see     #inWindow
       */
      public FontChooserDialog setWindow(handle window)
      {
         if (window.isUnknown())
         {
            String msg = "Could not find widget handle for window from expression";
            ErrorManager.recordOrThrowError(3138, msg, false);
         }
         else
         {
            WrappedResource r = window.getResource();
            if (r instanceof WindowWidget)
            {
               this.inWindow = window;
            }
            else
            {
               String msg = "Wrong type of object handle (%s)  -- must be a window handle";
               ErrorManager.recordOrThrowError(3139,
                                               String.format(msg, r.getClass().toString()),
                                               false);
            }
         }
         return this;
      }
      
      /**
       * Validates the options and tells the client to create the dialog according to options
       * configured and then interpret the results.
       * The client displays a dialog box that allows the user to select and associate a system
       * font with the specified font number. The {@code SYSTEM-DIALOG FONT} statement provides
       * a dialog box appropriate to the graphical environment in which it runs.
       * <p>  
       * This method must be the last in a chaining. 
       */
      public void execute()
      {
         if (inWindow == null)
         {
            inWindow = LogicalTerminal.defaultWindow();
         }
         
         WindowWidget window = (WindowWidget) inWindow.getResource();
         ClientExports lc = LogicalTerminal.getClient();
         boolean res = lc.chooseFont(fontRef, window.getId(), ansiOnly, fixedOnly);
         
         // on return check whether the user accepted or rejected the dialog 
         if (toUpdate != null)
         {
            toUpdate.assign(res);
         }
      }
   }
   
   /**
    * Container for context-local data. 
    */
   private static class WorkArea
   {
      /** The resource's ID. */
      private Long id = null;
      
      /** A cache with the default font specifications, for this context. */
      private FontDetails<?>[] defaultFonts = null;
      
      /** A cache with the system font specifications, for this context. */
      private FontDetails<?>[] systemFonts = null;
      
      /** A cache with the font table specifications, for this context. */
      private FontDetails<?>[] fontTable = null;
      
      /**
       * A cache with the font aliases, for this context.  The key is a font key from the font
       * table (i.e. font1) or a the key for a system font.
       */
      private Map<String, FontDetails<?>> fontAliases = null;
      
      /** A map of standard fonts, to which a key defaults, if not specified in the directory. */
      private Map<String, String> themePreferredFonts = null;
      
      /**
       * A map of standard font aliases, to which a key defaults, if not specified in
       * the directory.
       */
      private Map<String, String> themePreferredFontAliases = null;
      
      /** Flag to use anti-aliasing text rendering on the client side, Swing session. */
      private boolean swingAntiAliasing = Utils.getDirectoryNodeBoolean(
            null, "swing-text/anti-aliasing", true, Utils.DirScope.BOTH, null);
   }
}
