/*
** Module   : FrameWidget.java
** Abstract : server side frame widget class (contains the configuration)
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 NVS 20051026   @23491 Created initial version.
** 002 NVS 20060114   @24097 Frame is created hidden.
** 003 SIY 20060213   @24581 Fixes for the DialogBox support.
** 004 SIY 20060426   @25911 Changed signatures of get/setNextTabItem.
** 005 SIY 20060616   @27537 Added isDown(),setName() methods.
** 006 GES 20060803   @28377 Some cleanup and refactoring.
** 007 ECF 20061207   @31598 Fixed setAttrSpace(). Was calling the wrong method in FrameConfig.
** 008 CA  20090413   @41775 Fixes for FRAME:HIDDEN attribute. On init, HIDDEN flag is off. The
**                           setter must change the HIDDEN flag, not VISIBLE.
** 009 GES 20101005          Added a non-functional COM-HANDLE attribute stub.
** 010 SIY 20101006          Added moveAfterTab() stub.
** 011 SVL 20101006          #10 moved to GenericWidget.
** 012 SIY 20101007          Added setName() with different signature.
** 013 SIY 20111101          Updated imports.
** 014 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 015 CA  20130927          Added LegacyResource annotation, to determine resource type.
** 016 SVL 20131203          Changes caused by renaming of Nameable.get/setName to Nameable.name.
** 017 CA  20131115          Changes related to properly instantiate dynamic resources.
**                           Fixed implicit deletion of static frames.
** 018 CA  20140326          Fixed implicitDeletion in case is called from an Agent, outside of
**                           an external procedure (when the remote persistent procedure is
**                           deleted by the Agent).
** 019 GES 20140627          Inherit from PaneEntity and move all of ContainerEntity into
**                           this class.
** 020 EVL 20140926          More support for dynamic widget cleanup - dynamic frame must be
**                           deregistred in LogicalTerminal.
** 021 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 022 GES 20141124          Added FrameInterface and the centered attr.
** 023 CA  20150312          An unrealized frame needs to be deregistered explicitly when the 
**                           frame gets deleted.
** 024 CA  20150416          Fixed server-side support for PARENT attribute.
**                           Fixed support for NEXT/PREV-SIBLING in relation with tree widget
**                           structures.
** 025 HC  20150509          Javadoc fix.
** 026 CA  20150503          Track the resources created by each procedure, as dynamic resources   
**                           don't get deleted when the creator proc gets deleted, but their  
**                           INSTANTIATING-PROCEDURE attr needs to be unknown.
** 027 GES 20150506          Comment and javadoc cleanup.
** 028 IAS 20150604          Added support for the BOX attribute.
** 029 CA  20150722          CommonFrame.firstChild() now returns handle, so its usage was updated
**                           to be compatible with the new return type.
**                           FIRST-/LAST-CHILD must search FIELD-GROUP children, not frames.
** 030 SVL 20150804          Mark it as implementing Down interface.
** 031 IAS 20150922          Added support for the TOP-ONLY attribute
** 032 OM  20150915          Added stubs for CURRENT-ITERATION setter (attribute is writable for
**                           frames), CANCEL-BUTTON, DEFAULT-BUTTON, CURRENT-ITERATION,
**                           BACKGROUND, BORDER-???-CHARS, BORDER-???-PIXELS and OVERLAY
**                           attributes. Added support for the CANCEL-BUTTON/DEFAULT-BUTTON
**                           attributes
** 033 IAS 20151020          Added support for the BACKGROUND attribute.
**     IAS 20151029          VALIDATE() method implementation
** 034 CA  20151026          Allow support for nested frames. Fixed CURRENT-ITERATION.
**                           Fixed move-to-top/bottom for nested frames.
**                           Fixed VISIBLE,HIDDEN,DOWN attributes when set via the frame's handle.
**                           A fix for HIDDEN attribute: distinguish between HIDE statement and 
**                           setting the HIDDEN attribute to true.  More fixes related to HIDDEN
**                           attribute.
** 035 CA  20151123          Once the frame is attached, if its location is not set, it defaults 
**                           to (0,0).
** 036 HC  20151213          Improved DIALOG-BOX window parenting logic and related changes.
** 037 CA  20160105          Added runtime support for BORDER- attributes.
** 038 HC  20160203          Improved runtime support of FRAME:VIRTUAL* and FRAME:SCROLLABLE
**                           attributes. Frame layout improvements.
** 039 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 040 IAS 20160229          LOAD-MOUSE-POINTER support.
** 041 CA  20160303          Made attribute setters be a no-op if the value doesn't change.
** 042 CA  20160304          resourceDelete() must allow deletion of still visible frames, but for
**                           which their instantiating procedure is unknown.
** 043 HC  20160320          Improved frame virtual size handling.
** 044 CA  20160327          Fixed FRAME:TYPE in case of DIALOG-BOX.
** 045 IAS 20160331          Fixed potential NPE
** 046 HC  20160628          Fixed a case of FRAME:HIDDEN assigned to false not making the frame
**                           visible.
** 047 HC  20160714          Fixed runtime support of FIELD-GROUP:FIRST-CHILD so that the correct
**                           FieldGroup instance is searched.
**     SBI 20160720          Don't log UnimplementedFeature for MOVABLE attribute.
** 048 CA  20160915          Added setter for FRAME:PARENT attribute.  Fixed frame delete for 
**                           dynamic frames.
** 049 HC  20160922          Fixed getFrameHandle() to return frame's parent frame or unknown
**                           in case of no parent.
** 050 CA  20161014          Unified the setters for the FRAME's width/height attributes.
** 051 CA  20161020          Track when configuration is pushed because the HIDDEN attribute was
**                           set.
** 052 OM  20161205          Removed unused setSharedInstantiatingProcedure(dead code).
** 053 CA  20161123          Centralized the logic for the COL/ROW/X/Y setters in GenericFrame, to
**                           use the APIs from FrameWidget.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * This is the server side frame widget class which contains the 
 * client-side configuration for this frame BUT which does NOT implement
 * any of the server side logic for frame processing.
 * <p>
 * WARNING: unlike most of the server-side widget classes, this class
 * intentionally DOES NOT push a modified screen definition to the client
 * when visibility/hidden/down changes occur.  Such changes transfer to the
 * client using alternate paths and since these are frequently changed, using
 * the screen definition for such things is not acceptable.
 */
@LegacyResource(resource = LegacyResource.FRAME)
public class FrameWidget
extends PaneEntity<FrameConfig>
implements FrameInterface,
           TopOnlyInterface,
           BoxInterface,
           CommonHandleTree,
           Down,
           Hoverable
{
   /** FIELD-GROUP containing background widgets */ 
   private FieldGroup background = null;

   /**
    * Default constructor which can be used to create a static frame.
    */
   public FrameWidget()
   {
      this(false);
   }
   
   /**
    * Primary constructor.
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    */
   public FrameWidget(boolean dynamic)
   {
      super(dynamic, new FrameConfig());
      config.hidden = false;
      
      config.scrollable = dynamic;
   }
   
   /**
    * If this is a dialog-box frame, then its type must be <code>DIALOG-BOX</code>.  Otherwise, 
    * its type is <code>FRAME</code>.
    * 
    * @return   See above.
    */
   @Override
   public String type()
   {
      return (config == null ? super.type() : config.dialog ? "DIALOG-BOX" : super.type());
   }
   
   /**
    * Set NAME attribute.
    * 
    * @param   name
    *          New value for the attribute.
    */
   public void name(String name)
   {
      config.name = name;
   }
   
   /**
    * Set NAME attribute.
    * 
    * @param   name
    *          New value for the attribute.
    */
   public void name(character name)
   {
      name(name.isUnknown() ? "" : name.toStringMessage());
   }
   
   /**
    * Gets the NEXT-TAB-ITEM writable attribute.
    *
    * @return   The current value of NEXT-TAB-ITEM attribute.
    */
   public GenericWidget<?> getNextTabItem()
   {
      // TODO: implement this!
      throw new RuntimeException("NEXT-TAB-ITEM is not done yet!");
   }

   /**
    * Sets the NEXT-TAB-ITEM writable attribute.
    *
    * @param    n
    *           The new value for the NEXT-TAB-ITEM attribute.
    */
   public void setNextTabItem(GenericWidget<?> n)
   {
      // TODO: implement this!
      throw new RuntimeException("NEXT-TAB-ITEM is not done yet!");
   }

   /**
    * Gets the ATTR-SPACE writable attribute.
    *
    * @return   The current value of the ATTR-SPACE attribute.
    */
   public logical isAttrSpace()
   {
      return new logical(config.attr_space);
   }

   /**
    * Sets the ATTR-SPACE writable attribute.
    *
    * @param    space
    *           The new value for the ATTR-SPACE attribute.
    */
   public void setAttrSpace(logical space)
   {
      setAttrSpace(space.booleanValue());
   }

   /**
    * Sets the ATTR-SPACE writable attribute.
    *
    * @param    space
    *           The new value for the ATTR-SPACE attribute.
    */
   public void setAttrSpace(boolean space)
   {
      if (config.attr_space == space)
      {
         return;
      }
      
      config.attr_space = space;
      pushScreenDefinition();
   }
   
   /**
    * Get the value of the CENTERED attribute which defines if the frame is to be centered in
    * the width of the parent window.
    *
    * @return   <code>true</code> if the frame is centered.
    */
   public logical isCentered()
   {
      return new logical(config.centered);
   }

   /**
    * Set the value of the CENTERED attribute which defines if the frame is to be centered in
    * the width of the parent window.
    *
    * @param    center
    *           <code>true</code> if the frame is to be centered.
    */
   public void setCentered(boolean center)
   {
      if (config.centered && !center)
      {
         UnimplementedFeature.todo("4GL docs report this should gen an error when setting " +
                                   "from true to false.");
      }
      
      if (config.centered == center)
      {
         return;
      }
      
      config.centered = center;
      pushScreenDefinition();
   }

   /**
    * Set the value of the CENTERED attribute which defines if the frame is to be centered in
    * the width of the parent window.
    *
    * @param    center
    *           <code>true</code> if the frame is to be centered.
    */
   public void setCentered(logical center)
   {
      if (center == null || center.isUnknown())
      {
         UnimplementedFeature.todo("CENTERED attribute unknown behavior is not known.");
         return;
      }
      
      setCentered(center.booleanValue()); 
   }
   
   /**
    * Get the default button of this frame.
    *
    * @return  handle to current default button.
    */
   @Override
   public handle getDefaultButton()
   {
      if (config.defaultButtonId == -1)
      {
         return new handle();
      }
      
      return LogicalTerminal.getWidgetHandleById(config.defaultButtonId);
   }
   
   /**
    * Set the default button of this frame.
    *
    * @param   defaultButton
    *          handle to current default button.
    */
   @Override
   public void setDefaultButton(handle defaultButton)
   {
      if (LogicalTerminal.isChui())
      {
         ErrorManager.recordOrShowError(4088, 
               String.format(
                  "Progress does not support the DEFAULT-BUTTON attribute for the  %s %s" + ""
                        + " in this display environment.", 
                  type(), frame.getName(this)
               ), 
               false);
         return;
      }

      if (defaultButton == null || defaultButton.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
                  "Unable to assign UNKNOWN value to attribute DEFAULT-BUTTON on %s %s", 
                  type(), frame.getName(this)
               ), 
               false);
         return;
      }
      WrappedResource value = defaultButton.getResource();
      if (!(value instanceof ButtonWidget))
      {
         ErrorManager.recordOrShowError(4078, 
               String.format(
                     "Unable to set attribute DEFAULT-BUTTON for %s %s",
                      type(), frame.getName(this)
               ), 
               false);
         return;
      }

      ButtonWidget btn = (ButtonWidget) value;
      if (!btn.config.isDefault)
      {
         ErrorManager.recordOrShowError(4078, 
               String.format(
                     "Unable to set attribute DEFAULT-BUTTON for %s %s",
                      type(), frame.getName(this)
               ), 
               false);
         return;
      }

      if (btn.config.frameId != -1 && btn.config.frameId != getId())
      {
         ErrorManager.recordOrShowError(4078, 
               String.format(
                     "Unable to set attribute DEFAULT-BUTTON for %s %s",
                      type(), frame.getName(this)
               ), 
               false);
         return;
      }

      if (config.realized && btn.config.frameId == -1)
      {
         ErrorManager.recordOrShowError(4040, 
                     "Unable to realize BUTTON widget because it is not in frame",
               false);
         ErrorManager.recordOrShowError(4104, 
               "Unknown error code 1 for attribute ? on the BUTTON widget",
         false);
      }

      if (config.defaultButtonId == btn.getId())
      {
         return;
      }
      
      config.defaultButtonId = btn.getId();
      pushScreenDefinition();
   }
   
   /**
    * Gets the cancel button for this frame.
    *
    * @return  button widget that receives the CHOOSE event when a user cancels the current 
    *          frame or dialog box by pressing the ESC key.
    */
   @Override
   public handle getCancelButton()
   {
      if (config.cancelButtonId == -1)
      {
         return new handle();
      }
      
      return LogicalTerminal.getWidgetHandleById(config.cancelButtonId);
   }
   
   /**
    * Sets the cancel button for this frame.
    *
    * @param    cancelButton
    *           button widget that receives the CHOOSE event when a user cancels the current frame
    *           or dialog box by pressing the ESC key.
    */
   @Override
   public void setCancelButton(handle cancelButton)
   {
      if (LogicalTerminal.isChui())
      {
         ErrorManager.recordOrShowError(4088, 
               String.format(
                  "Progress does not support the CANCEL-BUTTON attribute for the  %s %s" + ""
                        + " in this display environment.", 
                  type(), frame.getName(this)
               ), 
               false);
         return;
      }

      if (cancelButton == null || cancelButton.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
                  "Unable to assign UNKNOWN value to attribute CANCEL-BUTTON on %s %s", 
                  type(), frame.getName(this)
               ), 
               false);
         return;
      }
      
      WrappedResource value = cancelButton.getResource();
      if (!(value instanceof ButtonWidget))
      {
         ErrorManager.recordOrShowError(4078, 
               String.format(
                     "Unable to set attribute CANCEL-BUTTON for %s %s",
                      type(), frame.getName(this)
               ), 
               false);
         return;
      }

      ButtonWidget btn = (ButtonWidget) value;
      if (btn.config.frameId != -1 && btn.config.frameId != getId())
      {
         ErrorManager.recordOrShowError(4078, 
               String.format(
                     "Unable to set attribute CANCEL-BUTTON for %s %s",
                      type(), frame.getName(this)
               ), 
               false);
         return;
      }

      if (config.realized && btn.config.frameId == -1)
      {
         ErrorManager.recordOrShowError(4040, 
                     "Unable to realize BUTTON widget because it is not in frame",
                     false);
         ErrorManager.recordOrShowError(4104, 
                     "Unknown error code 1 for attribute ? on the BUTTON widget",
                     false);
      }

      if (config.cancelButtonId == btn.getId())
      {
         return;
      }
      
      config.cancelButtonId = btn.getId();
      pushScreenDefinition();
   }
   
   /**
    * Conversion of CURRENT-ITERATION attribute (KW_CUR_ITER) of frame widget.
    *
    * Getter of CURRENT-ITERATION attribute.
    *
    * @return current value of CURRENT-ITERATION attribute.
    */
   @Override
   public handle currentIteration()
   {
      return frame.currentIteration();
   }
   
   /**
    * Set the CURRENT-ITERATION widget handle for this frame. This attribute is writable only
    * for frames. 
    *
    * @param   newIter
    *          A {@link handle} for the current iteration of the frame.
    */
   @Override
   public void setCurrentIteration(handle newIter)
   {
      UnimplementedFeature.missing("CURRENT-ITERATION setter is not implemented for frames.");
   }
   
   /**
    * Get the thickness (in character units) of the border at the top of the frame or dialog box.
    *
    * @return  the thickness of the border at the top of the frame.
    * 
    * @see   FrameWidget#getBorderSize(boolean, boolean)
    */
   @Override
   public decimal borderTopChars()
   {
      if ((config.dialog || (config.title != null && config.title.length() > 0)) && 
          !LogicalTerminal.isChui())
      {
         CoordinatesConversion cc = LogicalTerminal.coordinates();
         double baseHeight = getBorderSize(true, true);
         int pixelHeight = cc.heightToNative(baseHeight) + config.titleHeightPixels;
         
         return new decimal(cc.heightFromNative(pixelHeight));
      }
      else
      {
         return new decimal(getBorderSize(true, true));
      }
   }
   
   /**
    * Get the thickness (in character units) of the border at the left of the frame or dialog box.
    *
    * @return  the thickness of the border at the top of the frame.
    * 
    * @see   FrameWidget#getBorderSize(boolean, boolean)
    */
   @Override
   public decimal borderLeftChars()
   {
      return new decimal(getBorderSize(true, false));
   }
   
   /**
    * Get the thickness (in character units) of the border at the right of the frame or dialog
    * box.
    *
    * @return  the thickness of the border at the top of the frame.
    * 
    * @see   FrameWidget#getBorderSize(boolean, boolean)
    */
   @Override
   public decimal borderRightChars()
   {
      return new decimal(getBorderSize(true, false));
   }
   
   /**
    * Get the thickness (in character units) of the border at the bottom of the frame or dialog
    * box.
    *
    * @return  the thickness of the border at the top of the frame.
    * 
    * @see   FrameWidget#getBorderSize(boolean, boolean)
    */
   @Override
   public decimal borderBottomChars()
   {
      return new decimal(getBorderSize(true, true));
   }
   
   /**
    * Get the thickness (in pixels) of the border at the top of the frame or dialog box.
    *
    * @return  the thickness of the border at the top of the frame.
    * 
    * @see   FrameWidget#getBorderSize(boolean, boolean)
    */
   @Override
   public integer borderTopPixels()
   {
      if ((config.dialog || (config.title != null && config.title.length() > 0)) && 
          !LogicalTerminal.isChui())
      {
         CoordinatesConversion cc = LogicalTerminal.coordinates();
         double baseHeight = getBorderSize(true, true);
         
         return new integer(cc.heightToNative(baseHeight) + config.titleHeightPixels);
      }
      else
      {
         return new integer(getBorderSize(false, true));
      }
   }
   
   /**
    * Get the thickness (in pixels) of the border at the left of the frame or dialog box.
    *
    * @return  the thickness of the border at the top of the frame.
    * 
    * @see   FrameWidget#getBorderSize(boolean, boolean)
    */
   @Override
   public integer borderLeftPixels()
   {
      return new integer(getBorderSize(false, false));
   }
   
   /**
    * Get the thickness (in pixels) of the border at the right of the frame or dialog box.
    *
    * @return  the thickness of the border at the top of the frame.
    * 
    * @see   FrameWidget#getBorderSize(boolean, boolean)
    */
   @Override
   public integer borderRightPixels()
   {
      return new integer(getBorderSize(false, false));
   }
   
   /**
    * Get the thickness (in pixels) of the border at the bottom of the frame or dialog box.
    *
    * @return  the thickness of the border at the top of the frame.
    * 
    * @see   FrameWidget#getBorderSize(boolean, boolean)
    */
   @Override
   public integer borderBottomPixels()
   {
      return new integer(getBorderSize(false, true));
   }
   
   /**
    * Sets the dialog box flag.
    *
    * @param dialogBox
    *        new value for the dialog box flag
    */
   public void setDialogBox(boolean dialogBox)
   {
      config.dialog = dialogBox;
   }
   
   /**
    * Gets the dialog box flag.
    *
    * @return <code>true</code> if this frame is a dialog box.
    */
   public boolean isDialogBox()
   {
      return config.dialog;
   }

   /**
    * Sets the down frame size.
    *
    * @param    down
    *           The new number of iterations for this down frame.
    */
   @Override
   public void setDown(int64 down)
   {
      if (down.isUnknown())
      {
         ErrorManager.displayError(4083,
               "**Unable to assign UNKNOWN value to attribute DOWN on FRAME " + widgetName(),
               false);
         return;
      }

      setDown(down.intValue());
   }

   /**
    * Sets the down frame size.
    *
    * @param    down
    *           The new number of iterations for this down frame.
    */
   @Override
   public void setDown(long down)
   {
      if (config.down == down)
      {
         return;
      }
      
      config.down = (int) down;
      pushScreenDefinition();
   }
   
   /**
    * Gets the down frame size.
    *
    * @return   The number of data iterations for the frame (1 if the
    *           frame is not a down frame).
    */
   @Override
   public integer getDown()
   {
      return new integer(config.down);
   }

   /**
    * Return <code>true</code> if frame is DOWN frame.
    * 
    * @return   <code>true</code> if frame is DOWN frame.
    */
   public boolean isDown()
   {
      return (getDown().intValue() != 1);
   }
   
   /**
    * Sets the new value for HIDDEN flag.
    * 
    * @param   hidden
    *          The new boolean value for the HIDDEN attribute.
    */
   public void setHidden(boolean hidden)
   {
      _setHidden(hidden);
      _setVisible(!hidden);
      
      config.hiddenSet = true;
      
      try
      {
         if (frame != null)
         {
            if (!hidden)
            {
               // perform a VIEW only if all parent frames are visible OR this is the root frame
               if (allFramesVisible())
               {
                  frame.view();
               }
               else
               {
                  // but make sure the change is propagated to the client so that
                  // the frame gets displayed when the parent frame is displayed, too
                  pushScreenDefinition();
               }
            }
            else 
               frame.hideWorker(false, true);
         }
      }
      finally
      {
         config.hiddenSet = false;
      }
   }

   /**
    * Sets the Progress VISIBLE attribute.
    * 
    * @param   visible
    *          The new value for the Progress VISIBLE attribute.
    */
   public void setVisible(boolean visible)
   {
      if (frame != null)
      {
         if (visible) 
            frame.view();
         else 
            frame.hideWorker(false, false);

         _setVisible(frame._isVisible());
      }
   }
   
   /**
    * Gets the Progress COM-HANDLE attribute.
    * <p>
    * This is a non-functional stub for now.
    * 
    * @return   The value for the COM-HANDLE attribute.
    */
   public comhandle getComHandle()
   {
      return new comhandle();
   }
   
   /**
    * Hides this frame.
    * <p>
    * This method corresponds to the Progress statement: HIDE FRAME
    * 
    * @param   noPause
    *          <code>true</code> causes omission of any pauses which would
    *          be necessary otherwise.
    */
   @Override
   public void hide(boolean noPause)
   {
      frame.hide(noPause);
   }
   
   /**
    * Views this frame.
    * <p>
    * This method corresponds to the Progress statement: VIEW FRAME
    */
   @Override
   public void view()
   {
      frame.view();
   }

   /**
    * Gets the SELECTED writable attribute.
    *
    * @return    The current value of the SELECTED attribute.
    */
   @Override
   public logical isSelected()
   {
      UnimplementedFeature.partial("SELECTED attribute client-side runtime is not implemented");
      return new logical(config.selected);
   }

   /**
    * Sets the SELECTED writable attribute.
    *
    * @param    sel
    *           The new value for the SELECTED attribute.
    */
   @Override
   public void setSelected(logical sel)
   {
      if (sel == null || sel.isUnknown())
      {
         UnimplementedFeature.todo("SELECTED attribute unknown behavior is not known.");
         return;
      }

      setSelected(sel.booleanValue());
   }

   /**
    * Sets the SELECTED writable attribute.
    *
    * @param    sel
    *           The new value for the SELECTED attribute.
    */
   @Override
   public void setSelected(boolean sel)
   {
      UnimplementedFeature.partial("SELECTED attribute client-side runtime is not implemented");

      if (config.selected == sel)
      {
         return;
      }
      
      config.selected = sel;
      pushScreenDefinition();
   }
   
   /**
    * Queries if the user is allowed to move this widget using direct manipulation.
    * 
    * @return  If <code>true</code>, the widget is movable via direct manipulation.
    */
   public logical isMovable()
   {
      //UnimplementedFeature.partial("MOVABLE attribute client-side runtime is not implemented");
      return new logical(config.movable);
   }
   
   /**
    * Defines if the user is allowed to move this widget using direct manipulation.
    * 
    * @param   movable
    *          <code>true</code> to enable movable direct manipulation.
    */
   public void setMovable(boolean movable)
   {
      //UnimplementedFeature.partial("MOVABLE attribute client-side runtime is not implemented");
      
      if (config.movable == movable)
      {
         return;
      }
      
      config.movable = movable;
      pushScreenDefinition();
   }
   
   /**
    * Defines if the user is allowed to move this widget using direct manipulation.
    * 
    * @param   movable
    *          <code>true</code> to enable movable direct manipulation.
    */
   public void setMovable(logical movable)
   {
      if (movable == null || movable.isUnknown())
      {
         UnimplementedFeature.todo("MOVABLE attribute unknown behavior is not known.");
         return;
      }
      
      setMovable(movable.booleanValue()); 
   }
   
   /**
    * Gets the SCROLLABLE writeable attribute.
    *
    * @return   The current value of the SCROLLABLE attribute.
    */
   public logical isScrollable()
   {
      return new logical(config.scrollable);
   }

   /**
    * Sets the SCROLLABLE writeable attribute.
    *
    * @param    scroll
    *           The new value for the SCROLLABLE attribute.
    */
   public void setScrollable(logical scroll)
   {
      if (scroll == null || scroll.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
                  "Unable to assign UNKNOWN value to attribute SCROLLABLE on %s %s", 
                  type(), frame.getName(this)
               ), 
               false);
         return;
      }
      
      setScrollable(scroll.booleanValue());
   }

   /**
    * Sets the SCROLLABLE writeable attribute.
    *
    * @param    scroll
    *           The new value for the SCROLLABLE attribute.
    */
   public void setScrollable(boolean scroll)
   {
      if (config.scrollable == scroll)
      {
         return;
      }

      config.scrollable = scroll;
      
      if (scroll && !_isRealized())
      {
         if (config.widthChars > config.virtualWidthChars)
         {
            config.virtualWidthChars = config.widthChars; 
         }   
         
         if (config.heightChars > config.virtualHeightChars)
         {
            config.virtualHeightChars = config.heightChars; 
         }   
         
         if (config.widthPixels > config.virtualWidthPixels)
         {
            config.virtualWidthPixels = config.widthPixels; 
         }   
         
         if (config.heightPixels > config.virtualHeightPixels)
         {
            config.virtualHeightPixels = config.heightPixels; 
         }   
      }
      
      pushScreenDefinition();
   }
   
   /**
    * Accesses the first widget in the list of contained widgets.
    *
    * @return   The first widget or <code>null</code> if no widgets are contained.
    */
   @Override
   public handle firstChild()
   {
      // do not use HandleChain.firstResource as that method may return
      // an unexpected field group (like the BACKGROUND)
      return frame.firstChild();
   }
   
   /**
    * Accesses the last widget in the list of contained widgets.
    *
    * @return   The last widget or <code>null</code> if no widgets are contained.
    */
   @Override
   public handle lastChild()
   {
      return HandleChain.lastResource(this, LegacyResource.FIELD_GROUP);
   }
   
   /**
    * Sets the containing frame instance.
    * 
    * @param    frame
    *           The containing instance.
    */
   @Override
   public void setFrame(GenericFrame frame)
   {
      // save this, as it will be overwritten by the super call
      GenericFrame owner = this.frame;

      if (owner != null && config.realized)
      {
         ErrorManager.recordOrShowError(4078,
                  String.format(
                        "Unable to set attribute FRAME for %s %s",
                         type(), frame.getName(this)
                  ),
                  false, false, false);
         return;
      }
      
      super.setFrame(frame);
      
      if (owner != null)
      {
         this.frame = owner;
      }
      
      if (frame.asWidget() != this)
      {
         FieldGroup fg = (FieldGroup) LogicalTerminal.getWidgetForId(config.parentId);
         if (fg != null)
         {
            fg.removeWidget(this);
         }
         
         this.config.parentId = ((GenericWidget) frame.firstChild().getResource()).getId();

         fg = (FieldGroup) LogicalTerminal.getWidgetForId(config.parentId);
         // update the field-group
         fg.addWidget(this);
         
         this.config.frameId = frame.getFrameId();
         frame.addDynamicWidget(this);
         
         // once the frame is attached, if its location is not set, it defaults to (0,0)
         int x = 0;
         int y = 0;
         if (config.initX != BaseConfig.INV_COORD)
         {
            x = config.initX;
         }
         else if (config.x != BaseConfig.INV_COORD)
         {
            x = config.x;
         }
         
         if (config.initY != BaseConfig.INV_COORD)
         {
            y = config.initY;
         }
         if (config.y != BaseConfig.INV_COORD)
         {
            y = config.y;
         }
         
         setXOrYWorker(false, x, true);
         setXOrYWorker(false, y, false);
         pushScreenDefinition();
      }
   }

   /**
    * Indicates an implementation of siblings methods.
    * 
    * @return   Always <code>true</code>. 
    */
   @Override
   protected boolean useHandleChainSiblings()
   {
      return true;
   }

   /**
    * Get the value of the BOX attribute
    *
    * @return   the value of the attribute
    */
   @Override
   public logical isBox()
   {
      return new logical(config.box);
   }

   /**
    * Set the value of the BOX attribute
    *
    * @param   value
    *          the new value of the attribute
    */
   @Override
   public void setBox(logical value)
   {
      if (value == null || value.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
                  "Unable to assign UNKNOWN value to attribute BOX on %s %s", 
                  type(), frame.getName(this)
               ), 
               false);
         return;
      }
      setBox(value.booleanValue());
   }

   /**
    * Set the value of the BOX attribute
    *
    * @param   value
    *          the new value of the attribute
    */
   @Override
   public void setBox(boolean value)
   {
      if (config.realized)
      {
         ErrorManager.recordOrShowError(4053, 
               String.format(
                  "Unable to set BOX because the %s %s has been realised", 
                  type(), frame.getName(this)
               ), 
               false);
         return;
      }
      
      if (config.box == value)
      {
         return;
      }
      
      config.box = value;
      pushScreenDefinition();
  }
   
   /**
    * Sets the TOP-ONLY option/attribute.
    *
    * @param topOnly
    *        <code>true</code> if the option/attribute is set to ON.
    */
   @Override
   public void setTopOnly(boolean topOnly)
   {
      if (config.topOnly == topOnly)
      {
         return;
      }
      
      config.topOnly = topOnly;
      pushScreenDefinition();
   }
   
   /**
    * Sets the TOP-ONLY option/attribute.
    *
    * @param topOnly
    *        <code>true</code> if the attribute is set to ON.
    */
   @Override
   public void setTopOnly(logical topOnly)
   {
      if (topOnly == null || topOnly.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
               "Unable to assign UNKNOWN value to attribute TOP-ONLY on %s %s", 
               type(), widgetName()),
               false);
         return;
      }
      setTopOnly(topOnly.booleanValue());
   }
   
   /**
    * Gets the TOP-ONLY attribute's value.
    *
    * @return the value of the attribute
    */
   @Override
   public logical isTopOnly()
   {
      return new logical(config.topOnly);
   }
   
   /**
    * Sets the OVERLAY option/attribute.
    *
    * @param overlay
    *        <code>true</code> if the option/attribute is set to ON.
    */
   @Override
   public void setOverlay(boolean overlay)
   {
      if (config.overlay == overlay)
      {
         return;
      }
      
      config.overlay = overlay;
      pushScreenDefinition();
   }
   
   /**
    * Sets the OVERLAY attribute.
    *
    * @param overlay
    *        <code>true</code> if the attribute is set to ON.
    */
   @Override
   public void setOverlay(logical overlay)
   {
      if (overlay == null || overlay.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, 
               String.format(
               "Unable to assign UNKNOWN value to attribute OVERLAY on %s %s", 
               type(), widgetName()),
               false);
         return;
      }
      setOverlay(overlay.booleanValue());
   }
   
   /**
    * Get the OVERLAY attribute's value.
    *
    * @return the value of the attribute
    */
   @Override
   public logical isOverlay()
   {
      return new logical(config.overlay);
   }
   
   /**
    * Set the background field group
    *
    * @param background
    *        background fieldgroup
    */
   public void setBackground(FieldGroup background)
   {
      this.background = background;
   }

   /**
    * Get the background field group
    *
    * @return background field group
    */
   public FieldGroup getBackground()
   {
      return background;
   }

   /**
    * Get the handle for the BACKGROUND iteration of the frame.
    *
    * @return  the handle for the background iteration of the frame.
    */
   @Override
   public handle backgroundIteration()
   {
      return new handle(background);
   }
   
   /**
    * Processes validation rules (if they exist), error checking and any other constraint
    * processing that needs to be applied.  This is used to check if a proposed update to an
    * object will pass validation.
    *
    * @param    enabledOnly
    *           If <code>true</code>, VALIDATE will
    *           check enabled fields only. Otherwise, the VALIDATE method validates all fields,
    *           whether enabled or not.
    *
    * @return   {@code true} if the validation is successful.
    */
   @Override
   public logical validateFields(boolean enabledOnly)
   {
      return frame.validateFields(enabledOnly);
   }

   /**
    * Set the MOUSE-POINTER attribute
    * 
    * @param name
    *        the attribute value
    */
   @Override
   public void setMousePointerName(String name)
   {
      config.mousePointer = name;
   }

   /**
    * Get the MOUSE-POINTER attribute
    * 
    * @return the attribute value 
    *        
    */
   @Override
   public String getMousePointerImpl()
   {
      return config.mousePointer;
   }

   /**
    * Sets the PARENT writable attribute.
    *
    * @param    widget
    *           The new value for the PARENT attribute.
    */
   @Override
   public void setParent(GenericWidget<?> widget)
   {
      if (!validateParent(widget))
      {
         // parent is not valid, do not proceed
         return;
      }

      // TODO: if frame is realized, PARENT can't be changed!
      
      if (widget instanceof WindowWidget)
      {
         config.windowID = widget.getId();
      }

      super.setParent(widget);
   }
   
   /**
    * Check if the resource may be implicitly deleted. 
    * 
    * @return   <code>true</code> if the INSTANTIATING-PROCEDURE attribute is unknown (i.e. the 
    *           instantiating procedure has been deleted).
    */
   protected boolean implicitDeletion()
   {
      // shared frames can be declared only in external procedure; they are always marked as for
      // implicit deletion; resourceDelete will take care to delete only the first reference
      return frame.isShared()                          || 
             TransactionManager.getNestingLevel() == 0 || 
             !TransactionManager.isExternalBlock()     || 
             instantiatingProcedure().isUnknown();
   }

   /**
    * Gets the PARENT writable attribute.
    * <p>
    * For FRAMEs, this is either the containing WINDOW or the parent FIELD-GROUP, if attached to 
    * a frame.
    *
    * @return   the parent widget for this frame
    */
   @Override
   public CommonWidget getParent()
   {
      int wid = config.windowID;
      int pid = config.parentId;
      
      if (pid != -1)
      {
         return LogicalTerminal.getWidgetForId(pid);
      }
      else if (wid != -1)
      {
         return LogicalTerminal.getWindowById(wid);
      }
      else
      {
         return null;
      }
   }

   /**
    * Gets the containing frame instance, as a handle.
    *
    * @return   The frame that contains this widget.
    */
   @Override
   public handle getFrameHandle()
   {
      if (isRootFrame())
      {
         return new handle();
      }
      else
      {
         FrameWidget frm = (FrameWidget) LogicalTerminal.getWidgetForId(config.frameId);
         return new handle(frm.frame);
      }
   }

   /**
    * Check if this frame is the root frame (is parented directly to a window).
    * 
    * @return   See above.
    */
   protected boolean isRootFrame()
   {
      return config.frameId == -1 || config.frameId == getId();
   }
   
   /**
    * Determine if all frames up the hierarchy are visible.
    * 
    * @return   <code>true</code> if all frame's ascendants are visible.
    */
   protected boolean allFramesVisible()
   {
      FrameWidget fw = this;
      boolean allVisible = true;
      while (fw != null)
      {
         if (!fw._isVisible())
         {
            allVisible = false;
            break;
         }
         
         if (fw.isRootFrame())
         {
            break;
         }
         
         fw = (FrameWidget) LogicalTerminal.getWidgetForId(fw.config.frameId);
      }
      
      return allVisible;
   }

   /**
    * Perform actual delete of an resource. At the time of this call, it is assumed the resource
    * is valid for deletion (the handle and the resource are both valid).
    */
   @Override
   public void delete()
   {
      if (!resourceDelete())
      {
         return;
      }
      
      deleted = true;
      boolean dynamic = _dynamic();
      
      if (dynamic && _isVisible() && getWindow()._isValid())
      {
         // hide the frame, only if we are still attached to a valid window
         // pause is used only for deleting static frames 
         hide(dynamic);
      }
      
      super.delete();

      // remove the frame, etc
      frame.frameCleanup();
      
      if (dynamic || !config.realized)
      {
         LogicalTerminal.deregisterFrame(getId());
      }

      ConfigManager.getInstance().removeWidgetConfig(config.id);
   }
   
   /**
    * Delete the resource.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   @Override
   protected boolean resourceDelete()
   {
      if (!canDelete())
      {
         return false;
      }
      
      // check frame special cases
      if (!_dynamic()        && 
          _isVisible()       &&  
          !frame.isShared()  &&
          !(frame.isRedirected() || frame.isOutRedirected()) &&
          !instantiatingProcedure().isUnknown())
      {
         // implicit static frame deletion is possible only if is invisible
         return false;
      }
      
      if (frame.isShared() && frame.getScopeNesting() > 0)
      {
         // shared frames can't be deleted until all are out of scope.
         return false;
      }
      
      return true;
   }

   /**
    * Set the instantiating procedure for this resource.
    * <p>
    * This will set the instantiating procedure only if is not shared or this is the procedure
    * where the frame is registered as shared.
    *  
    * @param    hproc
    *           The procedure where the resource was instantiated.
    */
   @Override
   protected void setInstantiatingProcedure(handle hproc)
   {
      if (_dynamic() || !frame.isShared() || frame.getScopeNesting() == 0)
      {
         super.setInstantiatingProcedure(hproc);
      }
   }
   
   /**
    * Returns the widget types that can act as parent for this widget.
    *
    * @return   see above.
    */
   @Override
   protected String[] getValidParentTypes()
   {
      // parent must be a field-group or window
      return new String[] {LegacyResource.FIELD_GROUP, LegacyResource.WINDOW};
   }
   
   /**
    * Worker for the {@link #moveToTop()} and {@link #moveToBottom()} code.
    * 
    * @param    top
    *           <code>true</code> if moving to top, <code>false</code> for bottom.
    * 
    * @return   <code>true</code> if move succeeded.
    */
   @Override
   protected boolean moveToWorker(boolean top)
   {
      if (isRootFrame() || config.parentId == -1)
      {
         handle hw = getWindow();
         if( hw == null || hw.isUnknown())
         {
            return false;
         }

         if (!(((CommonWindow) hw.getResource()).isKeepFrameZOrder()).booleanValue())
         {
            return false;
         }

         return super.moveToWorker(top);
      }
      
      // frame is parented to something else
      return super.moveToWorker(top);
   }
   
   /**
    * Sets the frame's config {@link WidgetConfig#visible visibility flag}.
    * 
    * @param    visible
    *           The frame's visibility state.
    */
   protected void _setVisible(boolean visible)
   {
      config.visible = visible;
      config.assignedVisible = visible;
   }
   
   /**
    * Sets the frame's config {@link WidgetConfig#hidden hidden flag}.
    * 
    * @param    hidden
    *           The frame's hidden state.
    */
   protected void _setHidden(boolean hidden)
   {
      config.hidden = hidden;
   }
   
   /**
    * Checks whether the supplied widget can be used as a valid parent
    * of this widget.
    * 
    * @param    widget
    *           The widget reference to check.
    * 
    * @return   <code>true</code> if the supplied widget can be used
    *           as a valid parent, <code>false</code> otherwise.
    */
   @Override
   protected boolean validateParent(GenericWidget<?> widget)
   {
      boolean result = super.validateParent(widget);
      
      if (result && isDialogBox() && _isRealized())
      {
         String msg = String.format("Unable to set attribute PARENT for DIALOG-BOX %s.",
                                    frame.getName(this));
         ErrorManager.recordOrShowError(4078, msg, false);
         result = false;
      }
      
      return result;
   }

   /**
    * Performs a validation of widget size value.
    * 
    * @param    value
    *           The input value to validate.
    * @param    attr
    *           Denotes the kind of size attribute.
    * 
    * @return   <code>true</code> when valid, <code>false</code> otherwise
    */
   @Override
   protected boolean validateSize(Number value, SizeAttribute attr)
   {
      if (attr.isVirt())
      {
         if (!config.scrollable)
         {
            String attrName = attr.shortName();
            String msg = String.format("**Set SCROLLABLE to TRUE to change %s", attrName);
            ErrorManager.recordOrShowError(4069, msg, !LogicalTerminal.isChui(), false, false);
            return false;
         }
      }
      
      // TODO: when the virt size smaller than the min rectangle of frame content, 
      //       show the following error:
      //       **Unable to set the VIRTUAL-HEIGHT of the FRAME f too small to contain all 
      //       visible widgets. (4070)
      
      return super.validateSize(value, attr);
   }

   /**
    * Sets one of the size char attributes, that is WIDTH-CHARS, HEIGHT-CHARS,
    * MIN-WIDTH-CHARS, MIN-HEIGHT-CHARS, MAX-WIDTH-CHARS, MAX-HEIGHT-CHARS,
    * VIRTUAL-WIDTH_CHARS, VIRTUAL-HEIGHT-CHARS.
    *
    * @param    value
    *           The attribute value. <code>null</code> denotes an unknown attribute value.
    * @param    attr
    *           The attribute being set.
    */
   @Override
   protected void setSizeChars(Double value, SizeAttribute attr)
   {
      super.setSizeChars(value, attr);
      
      config.sizeChanged = false;
   }
   
   
   /**
    * Sets one of the size pixel attributes, that is WIDTH-PIXELS, HEIGHT-PIXELS,
    * MIN-WIDTH-PIXELS, MIN-HEIGHT-PIXELS, MAX-WIDTH-PIXELS, MAX-HEIGHT-PIXELS,
    * VIRTUAL-WIDTH_PIXELS, VIRTUAL-HEIGHT-PIXELS.
    *
    * @param    value
    *           The attribute value. <code>null</code> denotes an unknown attribute value.
    * @param    attr
    *           The attribute being set.
    */
   @Override
   protected void setSizePixels(Integer value, SizeAttribute attr)
   {
      super.setSizePixels(value, attr);
      
      config.sizeChanged = false;
   }
   
   /**
    * Sets the value for the attribute WIDTH-CHARS in the associated config instance.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   @Override
   protected void setWidthCharsWorker(double width)
   {
      if (this.frame.isInsideSetup())
      {
         config.scrollable = true;
      }
      else
      {
         double oldWidth = getWidthChars().doubleValue();

         // safe, comparison of scaled doubles
         config.sizeChanged = (oldWidth != width);
      }

      // interestingly, 4GL doesn't reset the virtual value just when SCROLLABLE is unset but 
      // when SCROLLABLE is unset and WIDTH/HEIGHT is assigned
      if (!config.scrollable)
      {
         config.virtualWidthChars = BaseConfig.INV_COORD;
      }
      else if (width > config.virtualWidthChars)
      {
         // virtual dim follows the frame dim, regardless if frame is realized or not, if that 
         // dim increases
         config.virtualWidthChars = width;
      }
      
      super.setWidthCharsWorker(width);
   }

   /**
    * Sets the value for the attribute HEIGHT-CHARS in the associated config instance.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   @Override
   protected void setHeightCharsWorker(double height)
   {
      if (this.frame.isInsideSetup())
      {
         config.scrollable = true;
      }
      else
      {
         double oldHeight = getHeightChars().doubleValue();

         // safe, comparison of scaled doubles
         config.sizeChanged = (oldHeight != height);
      }
      
      // interestingly, 4GL doesn't reset the virtual value just when SCROLLABLE is unset but 
      // when SCROLLABLE is unset and WIDTH/HEIGHT is assigned
      if (!config.scrollable)
      {
         config.virtualHeightChars = BaseConfig.INV_COORD;
      }
      else if (height > config.virtualHeightChars)
      {
         // virtual dim follows the frame dim, regardless if frame is realized or not, if that 
         // dim increases
         config.virtualHeightChars = height;
      }
      
      super.setHeightCharsWorker(height);
   }

   /**
    * Sets the value for the attribute WIDTH-PIXELS in the associated config instance.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   @Override
   protected void setWidthPixelsWorker(int width)
   {
      if (this.frame.isInsideSetup())
      {
         config.scrollable = true;
      }
      else
      {
         int oldWidth = getWidthPixels().intValue();
         
         config.sizeChanged = (oldWidth != width);
      }
      
      // interestingly, 4GL doesn't reset the virtual value just when SCROLLABLE is unset but 
      // when SCROLLABLE is unset and WIDTH/HEIGHT is assigned
      if (!config.scrollable)
      {
         config.virtualWidthPixels = BaseConfig.INV_COORD;
      }
      else if (width > config.virtualWidthPixels)
      {
         // virtual dim follows the frame dim, regardless if frame is realized or not, if that 
         // dim increases
         config.virtualWidthPixels = width;
      }
      
      super.setWidthPixelsWorker(width);
   }

   /**
    * Sets the value for the attribute HEIGHT-PIXELS in the associated config instance.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   @Override
   protected void setHeightPixelsWorker(int height)
   {
      if (this.frame.isInsideSetup())
      {
         config.scrollable = true;
      }
      else
      {
         int oldHeight = getHeightPixels().intValue();
         
         config.sizeChanged = (oldHeight != height);
      }
      
      // interestingly, 4GL doesn't reset the virtual value just when SCROLLABLE is unset but 
      // when SCROLLABLE is unset and WIDTH/HEIGHT is assigned
      if (!config.scrollable)
      {
         config.virtualHeightPixels = BaseConfig.INV_COORD;
      }
      else if (height > config.virtualHeightPixels)
      {
         // virtual dim follows the frame dim, regardless if frame is realized or not, if that 
         // dim increases
         config.virtualHeightPixels = height;
      }
      
      super.setHeightPixelsWorker(height);
   }
   
   /**
    * Sets the row or column position for this widget.
    *
    * @param    isUnknown
    *           If <code>true</code> the attribute value being set is an unknown value.
    * @param    value
    *           The 1-based row or column number.
    * @param    row
    *           If <code>true</code> <code>value</code> represents a row number, otherwise 
    *           a column number.
    */
   protected void setColumnOrRow(boolean isUnknown, double value, boolean row)
   {
      // TO-DO:
      // value > screen-width - frame-width should produce PROGRESS error 4054
      // "**Unable to set COL.  FRAME f does not fit in WINDOW widget. (4054)"
      if (!validColumnOrRow(isUnknown, value, row))
      {
         // for frames, there are two cases when the value is not a valid coordinate:
         // 1. frame is a child frame (has the FRAME attribute set)
         //   - the frame is not realized: the attribute is adjusted to 1
         //   - frame is realized: the new value is ignored
         
         // 2. frame is a root frame
         //   - the frame is not realized: the frame is placed and the attribute is computed
         //     by the runtime
         //   - frame is realized: the new value is ignored
         
         if (config.realized)
         {
            return;
         }
      }
      
      boolean isChui = LogicalTerminal.isChui();

      value = isChui ? Coordinate.chuiRound(value) : Coordinate.guiRound(value);
      
      if ((row && config.row == value) || (!row && config.column == value))
      {
         return;
      }
      
      // TODO: fixme!
      //       Before the change 'HC 20150323 Removed business logic from config
      //       classes.' the true value of positionChanged
      //       was never picked up because it was being checked in
      //       BaseConfig.setColumn()/setRow() called by BaseConfig.applyConfig()
      //       at which time this value was not yet set (by
      //       FrameConfig.applyConfig()) in the config of the client frame.
      //       After the change 'HC 20150323', which fixes the flag propagation, a
      //       validation depending on positionChanged
      //       starts to fail.
      //       The temporal solution is to keep the previous behavior by leaving
      //       positionChanged unset.
      //    code from GenericWidget.setColum:
      // double assignedColumn = ConfigHelper.getAssignedColumn(config);
      // config.positionChanged = (newColumn != assignedColumn);
      //    code from GenericWidget.setRow:
      // double assignedRow = ConfigHelper.getAssignedRow(config);
      // config.positionChanged = (newRow != assignedRow);
      

      setColumnOrRowWorker(isUnknown, value, row);

      pushScreenDefinition();

      config.positionChanged = false;
   }
   
   /**
    * Get the frame's border size.
    * <p>
    * If <code>NO-BOX</code> option is used at a non-dialog frame, then this method returns 0.  
    * <p>
    * In ChUI, for any kind of frame (dialog, titled, non-titled),  it returns 1 for character 
    * units and 8 for pixel units, for any border.  Note that NO-BOX is ignored for dialog frames.
    * <p>
    * In GUI, for non-dialog non-titled frames or left/right/bottom borders for titled frames, the 
    * reference is 1 pixel (for any border) which is converted to character units, using the 
    * PIXELS-PER-ROW (for top/bottom border) or PIXELS-PER-COlUMN (for left/right border).  
    * Note that this border is not dependent on Windows OS appearance settings.
    * <p>
    * In GUI, for dialog frames, the left/right/bottom border have a reference of 5 units (these
    * can not be changed in Windows OS).
    * <p>
    * Finally, the top border for titled or dialog frames is managed explicitly in 
    * {@link #borderTopChars()} and {@link #borderTopPixels()}.
    * 
    * @param    chars
    *           When <code>true</code>, character units are returned; else, pixel units.
    * @param    row
    *           When <code>true</code>, it means size for the top/bottom border is returned; else,
    *           the size of the left/right border is returned.
    *           
    * @return   The border size, in character units.
    */
   private double getBorderSize(boolean chars, boolean row)
   {
      if (!config.box && !config.dialog)
      {
         return 0;
      }
      
      if (LogicalTerminal.isChui())
      {
         // in ChUI, border width/height is always 1 char unit (aka 8 pixel units), even for 
         // DIALOG-BOX which has NO-BOX set.
         return chars ? 1 : 8;
      }
      else
      {
         // for dialog frames (regardless of NO-BOX), return:
         // - for left/right/bottom border: 5 pixels
         // - for top border: the title-bar height (depending on its title font), but this is 
         //   managed explicitly in the borderTopChars and borderTopPixels APIs.

         // for non-dialog the width/height is always 1 pixel unit, and this must get converted to  
         // chars using the PIXELS-PER-ROW/COLUMN values (it is not dependent on the frame's 
         // font).  case of titled frames, is handled explicitly via config.topBorderPixels
         
         int borderSize = config.dialog ? 5 : 1;
         
         CoordinatesConversion cc = LogicalTerminal.coordinates();
         
         if (!chars)
         {
            return borderSize;
         }
         else
         {
            return row ? cc.heightFromNative(borderSize) : cc.widthFromNative(borderSize);
         }
      }
   }
}