/*
/*
** Module   : PrintingService.java
** Abstract : Printing management class.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description--------------------------------
** 001 OM  20170721 Initial commit with stubbed methods.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;

import java.util.logging.*;

/**
 * This class is a factory for {@code PrinterSetupDialog}s. It exposes a single factory method and
 * some constants used for setting options in the chained methods of the returned dialog class.
 * 
 * @see PrinterSetupDialog
 */
public class PrintingService
{
   /** Constant for page orientation. NONE means that the value is not initialized. */
   public static final int NONE = 0;
   
   /** Constant for page orientation. This value represents a portrait orientation of the page. */
   public static final int PORTRAIT = 1;
   
   /** Constant for page orientation. It represents the landscape orientation of the page. */
   public static final int LANDSCAPE = 2;
   
   /** Constant for page orientation. Both portrait and landscape orientations are acceptable. */
   public static final int ANY = 3;
   
   /** Logger for this class. */
   private static final Logger LOG = LogHelper.getLogger(PrintingService.class.getName());
   
   /**
    * Factory method for a printer setup dialog. 
    * 
    * @return  a new {@link PrinterSetupDialog} that can chain eventual option and and finally
    *          execute on client side.
    */
   public static PrinterSetupDialog createSetupDialog()
   {
      return new PrinterSetupDialog();
   }
   
   /**
    * A class that collects option information about a printer selection statement using chained
    * methods.
    * <p>
    * Each public method emulates an option of the statement. They do a quick validation. If
    * successful they return a reference to this object to allow other method to be chained.
    * <p>
    * The {@code execute()} method does the final checks, send the request with gathered options
    * to client, awaits the user to interact and finally sets up the response and confirmation
    * variable {@code update}.
    */
   public static class PrinterSetupDialog
   {
      /**
       * Specifies the initial value of the Copies field in the Print dialog box.  The user can
       * change this value within the dialog box. This option is supported only with printer
       * drivers that support multi-copy printing. Otherwise, the Copies field is disabled.
       */
      private int copyCount = -1;
      
      /** The orientation of the paper. */
      private int orientation = NONE;
      
      /**
       * A logical variable to return the status of the user’s dialog interaction. If the
       * user chooses the OK button, the dialog sets logical-variable to TRUE. If the user chooses
       * the Cancel button, the dialog sets logical-variable to FALSE.
       */
      private logical toUpdate;
      
      /**
       * Specifies the window from which the dialog box is displayed. The value must be the handle
       * of a window.
       */
      private handle inWindow;
      
      /**
       * Configures the initial value of the Copies field in the Print dialog box.
       * 
       * @param   count
       *          The number of pages in Copies field.
       * 
       * @return  The reference to current {@code PrinterSetupDialog}. Used for chaining.
       * 
       * @see     #copyCount
       */
      public PrinterSetupDialog setCopyCount(int64 count)
      {
         this.copyCount = count.intValue();
         return this;
      }
      
      /**
       * Configures the initial value of the Copies field in the Print dialog box.
       *
       * @param   count
       *          The number of pages in Copies field.
       *
       * @return  The reference to current {@code PrinterSetupDialog}. Used for chaining.
       *
       * @see     #copyCount
       */
      public PrinterSetupDialog setCopyCount(long count)
      {
         this.copyCount = (int) count;
         return this;
      }
      
      /**
       * Configures the orientation of the page in dialog .
       *
       * @param   orientation
       *          The new default page orientation. 
       *
       * @return  The reference to current {@code PrinterSetupDialog}. Used for chaining.
       *
       * @see     #orientation
       */
      public PrinterSetupDialog setOrientation(int orientation)
      {
         if (this.orientation != NONE) 
         {
            // this is a compile-time error in fact (even duplication of same orientation)
            ErrorManager.displayAbend(
                  277, "Phrase or option conflicts with previous phrase or option");
            return null;
         }
         
         this.orientation = orientation;
         return this;
      }
      
      /**
       * Configures the variable to be set when the selection dialog is closed.
       *
       * @param   update
       *          A variable that will hold the status of the dialog.
       *
       * @return  The reference to current {@code PrinterSetupDialog}. Used for chaining. 
       *
       * @see     #toUpdate
       */
      public PrinterSetupDialog update(logical update)
      {
         this.toUpdate = update;
         return this;
      }
      
      /**
       * Configures the dialog to be created as child of a certain windows.
       *
       * @param   window
       *          A handle to a window that will be the parent of the dialog.
       *
       * @return  The reference to current {@code PrinterSetupDialog}. Used for chaining.
       *
       * @see     #inWindow
       */
      public PrinterSetupDialog setWindow(handle window)
      {
         if (window.isUnknown())
         {
            String msg = "Could not find widget handle for window from expression";
            ErrorManager.recordOrThrowError(3138, msg, false);
         }
         else
         {
            WrappedResource r = window.getResource();
            if (r instanceof WindowWidget)
            {
               this.inWindow = window;
            }
            else
            {
               String msg = "Wrong type of object handle (%s)  -- must be a window handle";
               ErrorManager.recordOrThrowError(3139,
                                               String.format(msg, r.getClass().toString()),
                                               false);
            }
         }
         return this;
      }
      
      /**
       * Instruct the client to show the {@code Printer Setup} dialog using the configured options
       * using previously chained methods. When the client returns, update the {@link #toUpdate}
       * variable, if set, to let the programmer whether the client accepted the dialog or
       * rejected it.
       */
      public void execute()
      {
         if (inWindow == null)
         {
            inWindow = LogicalTerminal.defaultWindow();
         }
         
         WindowWidget window = (WindowWidget) inWindow.getResource();
         ClientExports lc = LogicalTerminal.getClient();
         boolean res = lc.setupPrinter(window.getId(), copyCount, orientation);
         
         // on return check whether the user accepted or rejected the dialog 
         if (toUpdate != null)
         {
            toUpdate.assign(res);
         }
      }
   }
}
