/*
** Module   : RectangleConfig.java
** Abstract : RECTANGLE widget options.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SIY 20101002          Created initial version
** 002 SIY 20111101          Moved from CHUI package. Minor 
**                           cleanups/refactorings in preparation to GUI client.
** 003 MAG 20140410          Add EDGE-CHARS and SIZE-PIXELS options. 
**                           EDGE-PIXELS option should be of type INTEGER.
** 004 HC  20140911          GUI coordinate support - data types refactoring.
** 005 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 006 EVL 20141105          Adding support for GROUP-BOX, ROUNDED attributes.
** 007 HC  20150323          Removed business logic from config classes.
** 008 IAS 20160322          Changed the constructor visibility.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.io.*;

/**
 * Rectangle widget configuration.
 */
public class RectangleConfig
extends ControlConfig
{
   /** NO-FILL, FILLED option and attribute */
   public boolean filled = true;
   
   /** GRAPHIC-EDGE option */
   public boolean graphicEdge = false;
   
   /** EDGE-PIXEL value */
   public int edgePixels = 1;
   
   /** EDGE-CHARS value */
   public double edgeChars = 1;
   
   /** GROUP-BOX attribute value. */
   public boolean groupBox = false;
   
   /** ROUNDED attribute value. */
   public boolean rounded = false;
   
   /**
    * Default constructor
    */
   public RectangleConfig()
   {
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   public RectangleConfig(WidgetId id)
   {
      super(id);
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected RectangleConfig(int id)
   {
      super(id);
   }

   /**
    * Set new values from the provided configuration instance.
    * 
    * @param    config
    *           The instance from which to take values.
    */
   @Override
   public void applyConfig(WidgetConfig config)
   {
      super.applyConfig(config);
      
      RectangleConfig cfg = (RectangleConfig) config;
      
      filled      = cfg.filled;
      graphicEdge = cfg.graphicEdge;
      edgePixels  = cfg.edgePixels;
      edgeChars   = cfg.edgeChars;
      groupBox    = cfg.groupBox;
      rounded     = cfg.rounded;
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      filled       = in.readBoolean();
      graphicEdge  = in.readBoolean();
      edgePixels   = in.readInt();
      edgeChars    = in.readDouble();
      groupBox     = in.readBoolean();
      rounded      = in.readBoolean();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeBoolean(filled);
      out.writeBoolean(graphicEdge);
      out.writeInt(edgePixels);
      out.writeDouble(edgeChars);
      out.writeBoolean(groupBox);
      out.writeBoolean(rounded);
   }
}