/*
** Module   : RectangleWidget.java
** Abstract : RECTANGLE widget server side implementation.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20101002 Created initial version
** 002 SIY 20111101 Updated imports.
** 003 CA  20130222 Added NPE protection for cases when a widget is used before is 
**                  attached to a frame (after a CREATE WIDGET statement is used).
** 004 CA  20130927 Added LegacyResource annotation, to determine resource type.
** 005 CA  20131115 Changes related to properly instantiate dynamic resources.
** 006 MAG 20140410 Add RECTANGLE specific options.
** 007 HC  20140911 GUI coordinate support - data types refactoring.
** 008 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 009 HC  20141021 Implemented window move/position support.
** 010 EVL 20141105 Implementing stubs for setter/getter for rectangle specific attributes:
**                  EDGE-CHARS, EDGE-PIXELS, FILLED, GRAPHIC-EDGE, GROUP-BOX, ROUNDED.
** 011 EVL 20141121 Changed return values from boolean to logical.
** 012 HC  20141207 Implemented runtime support for widget coordinate attributes.
** 013 CA  20160303 Made attribute setters be a no-op if the value doesn't change.
** 014 GES 20160909 Implemented rectangle-specific attributes interface.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.util.*;

/**
 * Rectangle widget (server-side).
 */
@LegacyResource(resource = LegacyResource.RECTANGLE)
public class RectangleWidget
extends ControlEntity<RectangleConfig>
implements RectangleInterface
{
   /**
    * Default constructor.
    */
   public RectangleWidget()
   {
      this(false);
   }

   /**
    * Default constructor.
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    */
   public RectangleWidget(boolean dynamic)
   {
      super(dynamic, new RectangleConfig());
   }

   /**
    * Get EDGE-CHARS attribute.
    *          
    * @return  The value of the EDGE-CHARS attribute.
    */
   @Override
   public decimal getEdgeChars()
   {
      return new decimal(config.edgeChars);
   }

   /**
    * Get EDGE-PIXELS attribute.
    * 
    * @return  The value of the EDGE-PIXELS attribute.
    */
   @Override
   public integer getEdgePixels()
   {
      return new integer(config.edgePixels);
   }

   /**
    * Get FILLED attribute.
    * 
    * @return  The value of the FILLED attribute.
    */
   @Override
   public logical getFilled()
   {
      return new logical(config.filled);
   }
   
   /**
    * Get GRAPHIC-EDGE option.
    * 
    * @return  The value of the GRAPHIC-EDGE attribute.
    */
   @Override
   public logical getGraphicEdge()
   {
      return new logical(config.graphicEdge);
   }

   /**
    * Get GROUP-BOX attribute.
    * 
    * @return  The value of the GROUP-BOX attribute.
    */
   @Override
   public logical getGroupBox()
   {
      return new logical(config.groupBox);
   }
   
   /**
    * Get ROUNDED attribute.
    * 
    * @return  The value of the ROUNDED attribute.
    */
   @Override
   public logical getRounded()
   {
      return new logical(config.rounded);
   }
   
   /**
    * Set FILLED attribute or NO-FILL option.
    * 
    * @param   flag
    *          New value for the FILLED attribute.
    */
   @Override
   public void setFilled(boolean flag)
   {
      if (config.filled == flag)
      {
         return;
      }

      config.filled = flag;
      pushScreenDefinition();
   }

   /**
    * Set FILLED attribute NO-FILL option.
    * 
    * @param   flag
    *          New value for the FILLED attribute.
    */
   @Override
   public void setFilled(logical flag)
   {
      setFilled(flag.booleanValue());
   }

   /**
    * Set EDGE-PIXELS option.
    * 
    * @param   pixels
    *          Option value.
    */
   @Override
   public void setEdgePixels(int pixels)
   {
      if (config.edgePixels == pixels)
      {
         return;
      }
      
      config.edgePixels = pixels;
      pushScreenDefinition();
   }

   /**
    * Set EDGE-PIXELS option.
    * 
    * @param   pixels
    *          Option value.
    */
   @Override
   public void setEdgePixels(integer pixels)
   {
      setEdgePixels(pixels.intValue());
   }

   /**
    * Set EDGE-CHARS option.
    *
    * @param   chars
    *          Option value.
    */
   @Override
   public void setEdgeChars(double chars)
   {
      if (config.edgeChars == chars)
      {
         return;
      }
      
      config.edgeChars = chars;
      pushScreenDefinition();
   }

   /**
    * Set EDGE-CHARS option.
    * 
    * @param   chars
    *          Option value.
    */
   @Override
   public void setEdgeChars(NumberType chars)
   {
      setEdgeChars(chars.doubleValue());
   }

   /**
    * Set EDGE-CHARS option.
    * 
    * @param   chars
    *          Option value.
    */
   @Override
   public void setEdgeChars(int chars)
   {
      setEdgeChars((double)chars);
   }

   /**
    * Set GRAPHIC-EDGE option.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setGraphicEdge(boolean flag)
   {
      if (config.graphicEdge == flag)
      {
         return;
      }
      
      config.graphicEdge = flag;
      pushScreenDefinition();
   }

   /**
    * Set GRAPHIC-EDGE option.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setGraphicEdge(logical flag)
   {
      setGraphicEdge(flag.booleanValue());
   }

   /**
    * Set GROUP-BOX attribute.
    * 
    * @param   flag
    *          New value for the GROUP-BOX attribute.
    */
   @Override
   public void setGroupBox(boolean flag)
   {
      if (config.groupBox == flag)
      {
         return;
      }
      
      config.groupBox = flag;
      pushScreenDefinition();
   }

   /**
    * Set GROUP-BOX attribute.
    * 
    * @param   flag
    *          New value for the GROUP-BOX attribute.
    */
   @Override
   public void setGroupBox(logical flag)
   {
      setGroupBox(flag.booleanValue());
   }

   /**
    * Set ROUNDED attribute.
    * 
    * @param   flag
    *          New value for the ROUNDED attribute.
    */
   @Override
   public void setRounded(boolean flag)
   {
      if (config.rounded == flag)
      {
         return;
      }
      
      config.rounded = flag;
      pushScreenDefinition();
   }

   /**
    * Set ROUNDED attribute.
    * 
    * @param   flag
    *          New value for the ROUNDED attribute.
    */
   @Override
   public void setRounded(logical flag)
   {
      setRounded(flag.booleanValue());
   }
}