/*
** Module   : SelectionListWidget.java
** Abstract : server side selection list widget implementation
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description----------------------------
** 001 NVS 20051026   @23495 Created initial version.
** 002 SIY 20060209   @24429 Major refactoring.
** 003 SIY 20060210   @24514 Now isSelected() methods using real selection 
**                           data. Added setSingle() convenience method.
** 004 GES 20060803   @28383 Some cleanup and refactoring.
** 005 GES 20080320   @37594 Added TYPE attribute.
** 006 GES 20110422          Added scrollbar flag getter/setter methods.
** 007 SIY 20111101          Updated imports.
** 008 CA  20130221          Refactored usage of TYPE attribute.
** 009 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 010 OM  20130527          Added int64 support for setters.
** 011 CA  20130927          Added LegacyResource annotation, to determine resource type.
** 012 CA  20131115          Changes related to properly instantiate dynamic resources.
** 013 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 014 SVL 20141102          Make this class implement ScrollbarVerticalElement.
** 015 GES 20141124          Added ScrollbarHorizontalElement and InnerLines. Removed list-items.
**                           Added NUM-ITEMS attribute.
** 016 IAS 20150123          Implemented DELETE() method
** 017 IAS 20150209          Fixed behavior to be compatible with 4GL.
** 018 IAS 20150604          Added support for the REPLACE method, NUM-ITEMS 
**                           and INNER-LINES attribute, realize() operation.
** 019 SVL 20150804          Mark it as implementing Multiple interface.
** 020 EVL 20150821          Getting screen value requires special call in GenericFrame class.
**                           Adding selection-list specific interface.  Fixed getWidthChars().
** 021 CA  20160303          Made attribute setters be a no-op if the value doesn't change.
** 022 CA  20161005          Fixed issues related to SCREEN-VALUE attribute (initial value or 
**                           explicitly set, when the value is empty string or unknown, or the 
**                           widget is uninitialized).
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * Server side selection list widget implementation.
 */
@LegacyResource(resource = LegacyResource.SELECTION_LIST)
public class SelectionListWidget
extends ControlSetEntity<SelectionListConfig>
implements ScrollbarVerticalElement,
           ScrollbarHorizontalElement,
           SelectionListInterface,
           SortedElement,
           InnerLines,
           NumItems,
           Multiple
{
   /**
    * Default constructor.
    */
   public SelectionListWidget()
   {
      this(false);
   }

   /**
    * Default constructor.
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    */
   public SelectionListWidget(boolean dynamic)
   {
      super(dynamic, new SelectionListConfig());
   }
   
   /**
    * Gets the value of the INNER-CHARS attribute which defines the number of character
    * columns in the widget.
    * 
    * @return   The number of character columns.
    */
   public integer getInnerChars()
   {
      return new integer(config.innerChars);
   }
   
   /**
    * Sets the value of the INNER-CHARS attribute which defines the number of character
    * columns in the widget.
    * 
    * @param   chars 
    *          The new value for the attribute.
    */
   public void setInnerChars(double chars)
   {
      if (config.innerChars == chars)
      {
         return;
      }
      
      resetAutoResize();
      
      config.innerChars = (int) chars;
      pushScreenDefinition();
   }
   
   /**
    * Sets the value of the INNER-CHARS attribute which defines the number of character
    * columns in the widget.
    * 
    * @param   chars 
    *          The new value for the attribute.
    */
   public void setInnerChars(NumberType chars)
   {
      if (chars == null || chars.isUnknown())
         return;
      
      setInnerChars(chars.intValue());
   }
   
   /**
    * Gets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget.
    * 
    * @return   The number of lines of text.
    */
   public integer getInnerLines()
   {
      if (!canAccess(null)) // this is how 4GL works
      {
         ErrorManager.recordOrShowError(4104, String.format(
               "Unknown error code 1 for attribute ? on the %s widget", 
               type()),
               false);
      }
      return new integer(Math.max(0, config.innerLines));
   }
   
   /**
    * Sets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget. For the INNER-LINES clause in the VIEW-AS phrase
    * 
    * @param    lines
    *           The new number of lines.
    */
   @Override
   public void setInitialInnerLines(double lines)
   {
      // non-positive values are rejected by the 4GL compiler
      // it also silently replaces too big values with 250.
      config.innerLines = config.initialInnerLines = Math.min((int) lines, 250);
   }

   /**
    * Sets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget.
    * 
    * @param    lines
    *           The new number of lines.
    */
   public void setInnerLines(double lines)
   {
      int ilines = (int)lines;
      if (ilines < 0)
      {
         ErrorManager.recordOrShowError(4082, String.format(
               "Attribute INNER-LINES must be greater than or equal to 0 on %s widget", 
               type()),
               false);
         return;
      }
      if (ilines == 0)
      {
         ErrorManager.recordOrShowError(4486, String.format(
               "Cannot set INNER-LINES on %s widget. Specified INNER-LINES too small", 
               type()),
               false);
         return;
      }

      config.heightChars =  ilines + 2;
      if (config.initialInnerLines > 0)
      {
         if (ilines > config.initialInnerLines)
         {
            ErrorManager.recordOrShowError(4474, String.format(
                  "New height for %s %s is too large to fit in frame", type(),
                  widgetName()), false);
         }
         ilines = Math.min(ilines, config.initialInnerLines + 2);
      }
      else if (frame != null)
      {
         FrameConfig frameConfig = frame.getFrameWidget().config();
         int limit = (int)(frameConfig.heightChars - Math.max(config.row, 0));
         if (frameConfig.box)
            limit -= 2;
         if (ilines > limit)
         {
            ErrorManager.recordOrShowError(4474, String.format(
                  "New height for %s %s is too large to fit in frame", type(),
                  widgetName()), false);
         }
      }
      
      resetAutoResize();
      
      config.innerLines = ilines;
      pushScreenDefinition();
   }

   /**
    * Sets the value of the INNER-LINES attribute which defines the number of text lines
    * in the widget.
    * 
    * @param    lines
    *           The new number of lines.
    */
   public void setInnerLines(NumberType lines)
   {
      if (lines == null || lines.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, String.format(
               "Unable to assign UNKNOWN value to attribute INNER-LINES on %s %s", type(),
               widgetName()), false);
         return;
      }
   
      setInnerLines(lines.intValue());
   }

   /**
    * Gets the HEIGHT-CHARS writable attribute.
    *
    * @return   The current value of the HEIGHT-CHARS attribute.
    */
   @Override
   public decimal getHeightChars()
   {
      if (!canAccess("HEIGHT"))
      {
         return new decimal(0);
      }
      return super.getHeightChars();
   }

   /**
    * Gets the WIDTH-CHARS writable attribute.
    *
    * @return   The current value of the WIDTH-CHARS attribute.
    */
   public decimal getWidthChars()
   {
      if (!canAccess("WIDHTH"))
      {
         return new decimal(0);
      }
      return super.getWidthChars();
   }

   /**
    * Getter for the MULTIPLE read-only attribute.
    * 
    * @return  value of MULTIPLE attribute
    */
   public logical isMultiple()
   {
      return new logical(config.multiple);
   }

   /**
    * Set MULTIPLE attribute.
    * 
    * @param   multiple
    *          New value for the attribute.
    */
   public void setMultiple(boolean multiple)
   {
      if (config.multiple == multiple)
      {
         return;
      }
      
      config.multiple = multiple;
      pushScreenDefinition();
   }

   /**
    * Set MULTIPLE attribute.
    * 
    * @param   multiple
    *          New value for the attribute.
    */
   public void setMultiple(logical multiple)
   {
      setMultiple(multiple.booleanValue());
   }

   /**
    * Convenience method to turning off MULTIPLE attribute.
    * 
    * @param   single
    *          New value for the attribute.
    */
   public void setSingle(boolean single)
   {
      setMultiple(!single);
   }
   
   /**
    * Get the NUM-ITEMS attribute.
    *
    * @return   The numebr of items in the list.
    */
   public integer getNumItems()
   {
      return new integer(config.items == null ? 0 : config.items.length);
   }
   
   /**
    * Get the value of the SCROLLBAR-HORIZONTAL attribute which controls whether the widget must
    * draw the horizontal scroll-bar if the inner width is more than the visible width.
    * 
    * @return   The value of the SCROLLBAR-HORIZONTAL attribute.
    */
   public logical isScrollbarHorizontal()
   {
      return new logical(config.scrollbarHorizontal);
   }
   
   /**
    * Set the value of the SCROLLBAR-HORIZONTAL attribute which controls whether the widget must
    * draw the horizontal scroll-bar if the inner width is more than the visible width.
    *
    * @param   horiz
    *          The new value for the attribute.
    */
   public void setScrollbarHorizontal(logical horiz)
   {
      if (horiz == null || horiz.isUnknown())
      {
         UnimplementedFeature.todo("SCROLLBAR-HORIZONTAL attribute unknown behavior is not " +
                                   "known.");
         return;
      }
      
      setScrollbarHorizontal(horiz.booleanValue());
   }
   
   /**
    * Set the value of the SCROLLBAR-HORIZONTAL attribute which controls whether the widget must
    * draw the horizontal scroll-bar if the inner width is more than the visible width.
    *
    * @param   horiz
    *          The new value for the attribute.
    */
   public void setScrollbarHorizontal(boolean horiz)
   {
      if (config.scrollbarHorizontal == horiz)
      {
         return;
      }
      
      config.scrollbarHorizontal = horiz;
      pushScreenDefinition();
   }
   
   /**
    * Checks whether the selection list must draw the vertical scroll-bar if
    * the inner size is more than visible one.
    * 
    * @return   The value of the SCROLLBAR-VERTICAL attribute.
    */
   public logical isScrollbarVertical()
   {
      return new logical(config.scrollbarVertical);
   }
   
   /**
    * Resets the vertical scroll-bar flag with the new one.
    * 
    * @param    scrollbar
    *           New value for the SCROLLBAR-VERTICAL attribute.
    */
   public void setScrollbarVertical(logical scrollbar)
   {
      if (scrollbar == null || scrollbar.isUnknown())
      {
         UnimplementedFeature.todo("SCROLLBAR-VERTICAL attribute unknown behavior is not " +
                                   "known.");
         return;
      }
      
      setScrollbarVertical(scrollbar.booleanValue());
   }
   
   /**
    * Resets the vertical scroll-bar flag with the new one.
    * 
    * @param    scrollbar
    *           New value for the SCROLLBAR-VERTICAL attribute.
    */
   public void setScrollbarVertical(boolean scrollbar)
   {
      if (config.scrollbarVertical == scrollbar)
      {
         return;
      }

      config.scrollbarVertical = scrollbar;
      pushScreenDefinition();
   }
   
   /**
    * Implements the IS-SELECTED() widget method.
    *
    * @param  item
    *         a string representation of an item to check
    * 
    * @return true if the specified item is selected
    */
   public logical isSelected(character item)
   {
      integer index = lookup(item);
      
      if (index.isUnknown() || index.intValue() == 0)
         return new logical(false);
      
      return isSelected(index);
   }
   
   /**
    * Implements the IS-SELECTED() widget method.
    *
    * @param  index
    *         1-based index of an item to check
    * 
    * @return true if the specified item is selected
    */
   public logical isSelected(int64 index)
   {
      return new logical(isSelected(index.intValue()));
   }

   /**
    * Helper methods with common selected index lookup code.
    *
    * @param  index
    *         1-based index of an item to check
    * 
    * @return <code>true</code> if the specified item is selected
    */
   private boolean isSelected(int index)
   {
      ControlSetItem[] items = config.items;
      
      if (index < 1 || index > items.length)
         return false;

      // adjust index to make it 0-based
      index--;
      
      int [] selection = frame.getSelection(this);
      
      if (selection == null)
         return false;

      // This is a short array, so a sort/binary search does not improve 
      // performance because average number of operations is too small to
      // justify overhead. 
      for (int j = 0; j < selection.length; j++)
      {
         if (selection[j] == index)
            return true;
      }
      
      return false;
   }

   /**
    * Obtain the current value in the screen buffer of the backing data for
    * this widget and return it as a <code>character</code> type. If
    * there is no value in the screen buffer (if it is <code>null</code>
    * which is the representation for an uninitialized value), then return
    * the empty string.
    *
    * @return   The value as a <code>character</code> type or the empty
    *           string if the screen buffer value is uninitialized.
    */
   public character getScreenValue()
   {
      return frame.getScreenValue(this, true);
   }
   
   /**
    * Implements the DELETE() widget method, which removes the list item or
    * items specified.
    * 
    * @param itemList
    *           The item or items to delete.
    * 
    * @return <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(character itemList)
   {
      return delete(itemList, COMPARE_VALUE_IGNORE_CASE, false);
   }

   /**
    * Implements the DELETE() widget method, which removes the list item or
    * items specified.
    * 
    * @param itemList
    *           The item or items to delete.
    * 
    * @return <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(String itemList)
   {
      return delete(itemList, COMPARE_VALUE_IGNORE_CASE, false);
   }

   /**
    * Sets the containing frame instance.
    * 
    * @param    frame
    *           <code>GenericFrame</code> instance that contains this widget
    */
   @Override
   public void setFrame(FrameWidget frame)
   {
      super.setFrame(frame);
      if (config.innerLines < 0)
      {
         config.innerLines = config.items == null ? 3 : config.items.length; 
         config.heightChars = config.innerLines + 2;
      }
   }

   /**
    * Check if the widget's layout is valid
    *   
    * @return
    *       <code>true</code> if the layout is OK.
    *       
    */
   @Override
   public boolean validateLayout()
   {
      FrameConfig frameConfig = frame.getFrameWidget().config();
      int limit = (int)(frameConfig.heightChars - Math.max(config.row, 0));
      if (frameConfig.box)
         limit -= 2;
      if (config.initialInnerLines > 0 ? 
            config.innerLines > config.initialInnerLines :
               config.innerLines > limit)
      {
         if (config.dynamic)
         {
               ErrorManager.recordOrShowError(4474, String.format(
                     "New height for %s %s is too large to fit in frame", type(),
                     widgetName()), false);
         }

         String str = frameConfig.name;
         ErrorManager.recordOrShowError(4041, 
               String.format(
                  "%s %s does not fit in parent FRAME %s", 
                  type(), widgetName(), str == null ? "" : str.substring(0, str.indexOf('#')) 
                ), 
               false);
         return false;
      }
      return true;
   }

   /**
    * Implements the SCROLL-TO-ITEM() widget method.  This method scrolls the selection list so
    * the specified appears at the top of the list.
    *
    * @param    index
    *           The 1-based index of an item to scroll.
    *
    * @return   <code>TRUE</code> in success, <code>FALSE</code> otherwise.
    */
   @Override
   public logical scrollToItem(int index)
   {
      return new logical(scrollToItemInt(index));
   }

   /**
    * Implements the SCROLL-TO-ITEM() widget method.  This method scrolls the selection list so
    * the specified appears at the top of the list.
    *
    * @param    index
    *           The 1-based index of an item to scroll.
    *
    * @return   <code>TRUE</code> in success, <code>FALSE</code> otherwise.
    */
   @Override
   public logical scrollToItem(NumberType index)
   {
      return index.isUnknown() ? new logical(false) : scrollToItem(index.intValue());
   }

   /**
    * Implements the SCROLL-TO-ITEM() widget method.  This method scrolls the selection list so
    * the specified appears at the top of the list.
    *
    * @param    name
    *           The name of an item to scroll.
    *
    * @return   <code>TRUE</code> in success, <code>FALSE</code> otherwise.
    */
   @Override
   public logical scrollToItem(character name)
   {
      // default result is false
      boolean bRes = false;
      
      // if name is unknown - return false
      if (name.isUnknown())
      {
         return new logical(bRes);
      }
      
      // search current item list
      String nameStr = name.toStringMessage();
      if (!nameStr.isEmpty() && config.items.length > 0)
      {
         for (int i = 0; i < config.items.length; i++)
         {
            String itemName = config.items[i].getValue().toStringMessage();
            // letter case is not meaningfull
            if (nameStr.equalsIgnoreCase(itemName))
            {
               // found match by name, 4GL uses 1-based index
               bRes = scrollToItemInt(i + 1);
               break;
            }
         }
      }
      
      return new logical(bRes);
   }

   /**
    * Implements the SCROLL-TO-ITEM() widget method.  This method scrolls the selection list so
    * the specified appears at the top of the list.
    *
    * @param    name
    *           The name of an item to scroll.
    *
    * @return   <code>TRUE</code> in success, <code>FALSE</code> otherwise.
    */
   @Override
   public logical scrollToItem(String name)
   {
      return scrollToItem(new character(name));
   }

   /**
    * Internal worker for getting the SCREEN-VALUE on a per-widget basis.
    * 
    * @param   initialized
    *          Flag indicating if the widget is realized or not.
    * @param   bdt
    *          The current value in the frame's buffer.
    * @param   fmt
    *          The format to be applied, in case this is a formatted value.
    * @param   ignoreFormat
    *          Flag indicating if the format needs to be ignored.
    *          
    * @return  This widget's screen-value
    */
   @Override
   character getScreenValue(boolean initialized, BaseDataType bdt, String fmt, boolean ignoreFormat)
   {
      if (!initialized || bdt.isUnknown())
      {
         return new character();
      }

      return super.getScreenValue(initialized, bdt, fmt, ignoreFormat);
   }
   
   /**
    * Internal worker for setting the SCREEN-VALUE on a per-widget basis.
    * 
    * @param   frameBuf
    *          The frame buffer where to save the value.
    * @param   value
    *          The value to be set via SCREEN-VALUE attribute.
    * @param   inUIStmt
    *          Flag indicating this call originates from a UI statement.
    *          
    * @return  <code>true</code> if the caller can proceed, as the screen-value can be set.
    */
   @Override 
   boolean setScreenValue(ScreenBuffer frameBuf, Object value, boolean inUIStmt)
   {
      if (inUIStmt && frameBuf.getWidgetValue(getId()) != null)
      {
         if (value == null || (value instanceof BaseDataType && ((BaseDataType) value).isUnknown()))
         {
            return false;
         }
         
         if (value instanceof character && ((character) value).toJavaType().length() == 0)
         {
            return false;
         }
      }
      
      if (!inUIStmt                        && 
          value instanceof character       && 
          !((character) value).isUnknown() && 
           ((character) value).toJavaType().length() == 0)
      {
         ((character) value).setUnknown();
      }
      
      return super.setScreenValue(frameBuf, value, inUIStmt);
   }

   /**
    * Internal worker for the SCROLL-TO-ITEM() widget method implementation.
    *
    * @param    index
    *           The 1-based index of an item to scroll.
    *
    * @return   <code>TRUE</code> in success, <code>FALSE</code> otherwise.
    */
   private boolean scrollToItemInt(int index)
   {
      // check for validity
      if (index < 1 || config.items == null || index > config.items.length)
      {
         return false;
      }

      // call the client side for real work
      return LogicalTerminal.scrollToItem(getId(), index);
   }
}