/*
** Module   : ServerConfigManager.java
** Abstract : Management of widget configurations, on server-side.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140929 Created initial version.
** 002 CA  20141014 Separated data which needs to survive the appserver agent's context reset into 
**                  a different container.
** 003 HC  20150323 Removed business logic from config classes.
** 004 CA  20160304 Added isSharedConfig().
** 005 OM  20161207 Removed dead code.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.util.*;

/**
 * Management of widget configurations, on server-side.
 */
public class ServerConfigManager
extends ConfigManager
{
   /**
    * On server-side, synchronize the field IDs for the server-side widget configurations with 
    * the IDs provided by the associated list.
    * 
    * @param    configs
    *           A map of configurations, having as hey a class name and as value a list of field
    *           names (the order in this list gives the field ID).
    */
   public void initWidgetConfigurations(Map<String, List<String>> configs)
   {
      WorkArea wa = survivals.get();
      
      wa.cfgDefs.clear();
      
      // initialize the widget config fields on the remote side
      for (String clsName : configs.keySet())
      {
         List<String> fieldNames = configs.get(clsName);
         
         try
         {
            @SuppressWarnings("unchecked")
            Class<? extends WidgetConfig> cfgCls = 
               (Class<? extends WidgetConfig>) Class.forName(clsName);
            
            WidgetConfigDef cfg = new WidgetConfigDef(cfgCls, fieldNames);
            wa.cfgDefs.put(cfgCls, cfg);
         } 
         catch (ClassNotFoundException e)
         {
            final String msg = "Could not initialize remote widget configs for class %s.";
            throw new RuntimeException(String.format(msg, clsName), e);
         }
      }
   }
   
   /**
    * Check if the given ID is associated with a registered widget.
    * 
    * @param    id
    *           The widget ID.
    *           
    * @return   <code>true</code> if a widget exists in the registry, with the given ID.
    */
   @Override
   protected boolean isRegisteredWidget(WidgetId id)
   {
      return id != null && LogicalTerminal.isWidgetIdInUse(id.asInt());
   }

   /**
    * Determine if the specified config instance can be tracked.
    * <p>
    * Sever-side does not track the config changes.
    * 
    * @param    cfg
    *           The {@link WidgetConfig} instace which needs to be tracked.
    *           
    * @return   See above.
    */
   public boolean trackConfig(WidgetConfig cfg)
   {
      return false;
   }

   /**
    * Register the specified field in the configuration for the specified widget ID as being 
    * dirty.
    * <p>
    * Sever-side does not track the config changes, so this is a no-op.
    * 
    * 
    * @param    cfg
    *           The widget config.
    * @param    field
    *           The changed field.
    */
   public void addDirtyConfig(WidgetConfig cfg, String field)
   {
      // no-op
   }

   /**
    * The method attempts to resolve a {@link ConfigOwner} reference from
    * the widget <code>widgetId</code>.
    * 
    * @param    widgetId
    *           Widget Id.
    *           
    * @return   A {@link ConfigOwner} reference or <code>null</code> if
    *           such cannot be find. 
    */
   @Override
   ConfigOwner<?> resolveConfigOwner(WidgetId widgetId)
   {
      GenericWidget<?> widget = LogicalTerminal.getWidgetForId(widgetId.asInt());
      if (widget instanceof ConfigOwner)
      {
         return (ConfigOwner<?>) widget;
      }
      
      return null;
   }
}