/*
** Module   : ServerExports.java
** Abstract : defines all UI-related API's exported by the server for use
**            by the client  
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description-----------------------------
** 001 SIY 2005xxxx   @23594 Created initial version.
** 002 NVS 20060130   @24149 Server exports taking screen buffer, now 
**                           return screen buffer as well. These are: 
**                           trigger(), validate().
** 003 SIY 20060130   @24156 Minor cleanups.
** 004 NVS 20060221   @24622 Added help() export that locates and runs a
**                           preconfigured help application.
** 005 GES 20060226   @24744 Added a query allowing the client to request
**                           the latest screen buffer from the server.
**                           This is useful in cases where the client
**                           initiates the view() of a frame without an
**                           explicit call from the server (e.g. page
**                           headers and footers in redirected output
**                           mode).
** 006 SIY 20060315   @25139 Changed signatures of BROWSE-related methods.
** 007 NVS 20060320   @25158 Simplified validation protocol by using new
**                           class ValidationResult.
** 008 GES 20060808   @28492 Use int instead of long for the browse 
**                           methods.
** 009 GES 20061017   @30452 Added a new option to trigger().
** 010 GES 20061030   @30788 Changed the return value from trigger() to
**                           eliminate the extra call to the client
**                           (to mark an event as "consume") and to make
**                           the trigger processing synchronous. This
**                           also makes it safer for nested triggers.
** 011 SIY 20061105   @30985 Added getCurrentRow() method.
** 012 NVS 20070118   @31889 EventList class is used instead of WaitList.
** 013 NVS 20070328   @32631 help() request from the client should report
**                           the lastKey variable.
** 014 SIY 20070504   @33426 Updated getRowCount().
** 015 SIY 20111101          Updated imports.
** 016 CA  20130822          Added support of appserver async requests: posting and executing of
**                           server-side events on the client side.
** 017 CA  20130908          Added support for socket runtime.
** 018 HC  20131031          Changed resource ID data type to Long, see issue #2183.
** 019 CA  20131218          Fixed H221 - fixed the signature for createSocket.
** 020 CA  20140711          Enhanced font support, for GUI clients.
** 021 CA  20140728          Added notification that a frame was hidden by the client-side.
** 022 CA  20140805          Added API to load a image from server-side.
** 023 MAG 20140812          Implements COLOR-TABLE, for GUI clients.
** 024 MAG 20140911          Added load environment API.
** 025 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 026 EVL 20141204          Differentiate window icon loading from other images.
** 027 EVL 20141216          Adding SESSION:TOOLTIPS attribute handling.
** 028 EVL 20150114          Adding the API to get application icon data from the server side.
**                           Rework icon/image implementation.
** 029 CA  20150122          Changed trigger API to be aware of the non-widget events (the 
**                           resource ID).
** 030 CA  20150212          Allow events to be captured and processed by the widgets without
**                           server-side or main event-loop notification.  This fixes problems in
**                           the EDITING and WAIT-FOR event loops, for i.e. COMBO-BOX and EDITOR
**                           widgets.
** 031 CA  20150220          Fixed the FRAME-VALUE statement: if HELP is active, it references the
**                           FOCUSed frame/widget as before the HELP is invoked.
** 032 SVL 20150212          Modified getRows signature.
** 033 CA  20150317          Added API to determine if a legacy font is fixed-width.
** 034 SVL 20150411          Added isRowAvailable.
** 035 CA  20150519          Refactored so that the font specifications are read in only one trip.
**                           Added API to transfer the entire text metrics DB from the server to
**                           the client-side, in only one trip.
** 036 SVL 20150605          Added getEditingFillIns.
** 037 SVL 20150625          Added getRowCount().
** 038 CA  20150722          Removed getColumns(), as browse columns are now transferred via frame
**                           definition.
** 039 SVL 20150722          Updated updateRow signature.
** 040 IAS 20150801          Added support for the WIDGET-ENTER and WIDGET-LEAVE attributes.
** 041 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 042 CA  20160418          Added infrastructure for message passing from converted code to a 
**                           customer-specific application or calls from the customer-specific 
**                           application into the converted code (i.e. invoke external programs,
**                           procedures or functions).  Can be used only when the P2J client runs 
**                           embedded in a customer-specific application.
** 043 SVL 20160419          Added moveColumnServer.
** 044 EVL 20160722          Using ScreenBuffer array for editable frames instead of only current
**                           buffer for frame under processing.  The methods involved are:
**                           trigger(), validate(), help(), refresh() and refreshBuffers().
** 045 CA  20161018          Any client-to-server code invocations must maintain the trigger 
**                           registries, not just explicit trigger invocation.
** 046 CA  20170228          Added PUBLISH/SUBSCRIBE/UNSUBSCRIBE extensions for global support and
**                           for external applications.
** 047 OM  20170320          Added support for Themes.
** 048 SVL 20170417          Added getRows() for search by key.
** 049 EVL 20170510          Changing the signature of the theme default fonts passing method.
** 050 SVL 20170813          Changed getRows signature.
** 051 SVL 20170831          getEditingFillIns refactored to getEditorsConfigs.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.util.*;

import com.goldencode.p2j.ui.ColorTable.EnvironmentColorTable;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.util.*;

/**
 * Defines all UI-related API's exported by the server for use by the client.
 */
public interface ServerExports
{
   /**
    * Called from the client-side, it executes the server-side event by calling 
    * {@link ServerEvent#invoke(int)}.
    * 
    * @param    id
    *           The id of the server event.
    */
   public ServerReturnValue invokeServerEvent(int id);
   
   /**
    * Check if the resource with the given id is still valid to be used in a WAIT-FOR statement.
    * It will return <code>false</code> if:
    * <ul>
    *    <li>there is no resource associated with the given id or the resource is unknown.</li>
    *    <li>the SENSITIVE attribute for the resource is <code>false</code>.</li>
    *    <li>the resource is not valid.</li>
    * </ul>
    * In all other cases, it will return <code>true</code>
    * 
    * @param    id
    *           The resource id.
    * 
    * @return   See above.
    */
   public boolean validResource(long id);
   
   /**
    * Signals a trigger to the application.  If a trigger is registered for
    * this event and source widget combination, then it will be executed.
    * 
    * @param    triggerId
    *           The trigger to be called.
    * @param    eventId
    *           The event which was generated.
    * @param    sourceId
    *           The widget which generated the event.
    * @param    otherId
    *           The companion widget id (for ENTER/LEAVE).
    * @param    resourceId
    *           The resource which generated the event.
    * @param    sb
    *           The editable frames <code>ScreenBuffer</code> array
    *
    * @return   A container including the <code>ScreenBuffer</code> 
    *           representing the latest screen values of all widgets and the
    *           consume flag.
    */
   public ServerReturnValue trigger(int          triggerId,
                                    int          eventId,
                                    int          sourceId,
                                    int          otherId,
                                    long         resourceId,
                                    ScreenBuffer[] sb);
   
   /**
    * Signals a validation event to the application.
    * <p>
    * This signal is used to trigger validation execution on the server.
    *
    * @param   sourceId
    *          widget ID of the widget generated the trigger.
    * @param   sb
    *          The <code>ScreenBuffer</code> array for current and other editable buffers.
    * 
    * @return  <code>ValidationResult</code> representing the validation  
    *          result, error message and screen value changes.
    */
   public ValidationResult validate(int sourceId, ScreenBuffer[] sb);
   
   /**
    * Locates and executes a preconfigured HELP application.
    *
    * @param   lastKey
    *          The most recent key event code
    * @param   buffer
    *          The screen buffer array associated with the FOCUS as before HELP is invoked and
    *          other editable frames.
    */
   public void help(int lastKey, ScreenBuffer[] buffer);

   /**
    * Obtains the latest screen buffer associated with the given frame.
    *
    * @param   frameId
    *          The frame for which the screen buffer should be returned.
    *
    * @return  <code>ScreenBuffer</code> array representing the current screen 
    *          values of all widgets in the given frame and other editable frames.
    */
   public ScreenBuffer[] refresh(int frameId);   
   
   //------------------------------------------------------------------------
   // BrowseModel calls
   //------------------------------------------------------------------------
   /**
    * Get configurations of widgets used for in-browse editing (fill-ins, toggle-boxes or
    * combo-boxes).
    *
    * @param    browseId
    *           ID of the browse which should handle the request.
    *
    * @return   Array of configurations of widgets used for in-browse editing (fill-ins,
    *           toggle-boxes or combo-boxes). An array element is configuration of the widget for
    *           the corresponding editable column or <code>null</code> if the column is not
    *           editable. Return value is <code>null</code> if the browse is read-only.
    */
   public WidgetConfig[] getEditorsConfigs(int browseId);

   /**
    * Get the number of rows iterated by the query (for preselect queries - matches result size,
    * for adaptive queries - matches the size of the part of the result set the query has already
    * iterated over).
    *
    * @param   browseId
    *          ID of the browse which should handle the request.
    *
    * @return  number of iterated rows in the result set.
    */
   public int getRowCount(int browseId);

   /**
    * Get number of available rows starting from the given one.
    *
    * @param   browseId
    *          ID of the browse which should handle the request.
    * @param   startRow
    *          Initial row to start count rows from.
    *
    * @return  Number of available rows.
    */
   public int getRowCount(int browseId, int startRow);

   /**
    * Retrieves several subsequent rows from the database.
    *
    * @param    browseId
    *           ID of the browse which should handle the request.
    * @param    startRowIndex
    *           Index of the first row to retrieve (0-based).
    * @param    count
    *           Maximum count of rows to return. <code>-1</code> for no limit.
    * @param    forward
    *           <code>true</code> if rows are retrieved in the forward direction,
    *           <code>false</code> if rows are retrieved in the backward direction,
    * @param    rowDisplayTriggerId
    *           Trigger ID of the ROW-DISPLAY trigger. <code>-1</code> if there is no ROW-DISPLAY
    *           trigger for this browse.
    * @param    sb
    *           Screen buffer for the ROW-DISPLAY trigger. <code>null</code> if there is no
    *           ROW-DISPLAY trigger for this browse.
    * @param    displayStrategy
    *           Specifies type of action for which rows are fetched and defines how rows are
    *           iterated / displayed.
    *
    * @return   Array of rows. Each row is array of cell values, but the first element of the row
    *           array is 0-based index of the row in the browse model.
    *           Note that if rows are fetched in the backward direction, the resulting rows are
    *           returned in the straight order starting from the row with the least row number.
    *           If no query is specified or the query is closed, <code>null</code> is returned.
    */
   public BrowseRow[] getRows(int browseId,
                              int startRowIndex,
                              int count,
                              boolean forward,
                              int rowDisplayTriggerId,
                              ScreenBuffer[] sb,
                              RowDisplayStrategy displayStrategy);

   /**
    * This function is used when search by key is performed in a browse. Retrieves the set of
    * rows which should be in view after search is completed. There are complex rules which
    * determine the top row of the view depending on how the <i>current</i> row is positioned
    * relatively to the current view and how the <i>found</i> row is positioned relatively to the
    * current view and to the current row.
    *
    * @param    browseId
    *           ID of the browse which should handle the request.
    * @param    searchChar
    *           Character to be searched for. Search is performed using only the first character
    *           (in the visual data representation) in the current column. Specifying a character
    *           puts this function in the search mode. This parameter is <code>null</code> for
    *           non-search mode.
    * @param    searchColumn
    *           For search mode: 0-based index of the column to be searched.
    * @param    viewportTopRow
    *           For search mode: 0-based index (in the result set) of the top row in the current
    *           view.
    * @param    viewportBottomRow
    *           For search mode: 0-based index (in the result set) of the bottom row in the
    *           current view.
    * @param    rowDisplayTriggerId
    *           Trigger ID of the ROW-DISPLAY trigger. <code>-1</code> if there is no ROW-DISPLAY
    *           trigger for this browse.
    * @param    sb
    *           Screen buffer for the ROW-DISPLAY trigger. <code>null</code> if there is no
    *           ROW-DISPLAY trigger for this browse.
    *
    * @return   Array of rows. Each row is array of cell values, but the first element of the row
    *           array is 0-based index of the row in the browse model.
    *           If no query is specified or the query is closed, <code>null</code> is returned.
    */
   public BrowseRow[] getRows(int browseId,
                              char searchChar,
                              int searchColumn,
                              int viewportTopRow,
                              int viewportBottomRow,
                              int rowDisplayTriggerId,
                              ScreenBuffer[] sb);
   
   /**
    * Updates selected cells in the given row.
    * 
    * @param   browseId
    *          ID of the browse which should handle the request.
    * @param   rowIndex
    *          Index of the row to update.
    * @param   cellIndices
    *          Cells for update indexes.
    * @param   cellValues
    *          Cells for update new values.
    * @param   onlyNotifyRowReleased
    *          Only notify server that the row has been released. Data is not updated.
    * 
    * @throws  DataValidationException
    *          New values don't passed database validation rules.
    */
   public void updateRow(int            browseId,
                         int            rowIndex,
                         int[]          cellIndices,
                         BaseDataType[] cellValues,
                         boolean        onlyNotifyRowReleased)
   throws DataValidationException;

   /**
    * Inform server about current row in the browse.
    * 
    * @param   browseId
    *          ID of the browse which should handle the request.
    * @param   rowIndex
    *          0-based index of the current row.
    *          
    * @return  New current row.
    */
   public int setCurrentRow(int browseId, int rowIndex);

   /**
    * Get current row in the query.
    *
    * @param   browseId
    *          ID of the browse which should handle the request.
    *          
    * @return  The current row index.
    */
   public int getCurrentRow(int browseId);

   /**
    * Determines if the specified browse row is "available", i.e. it is not deleted.
    *
    * @param  browseId
    *         ID of the browse which should handle the request.
    * @param  rowIndex
    *         0-base index of the target row in the browse model.
    *
    * @return <code>true</code> if the row is "available", i.e. it is not deleted.
    */
   public boolean isRowAvailable(int browseId, int rowIndex);

   /**
    * Implements the server-side MOVE-COLUMN() widget method.<p>
    * Repositions a column in the browse widget.
    *
    * @param  browseId
    *         ID of the browse which should handle the request.
    * @param  src
    *         0-based position of the column to be moved.
    * @param  dest
    *         0-based position to which the column is moved.
    */
   public void moveColumnServer(int browseId, int src, int dest);
   
   /**
    * Called by the P2J client side, it will inform the server that a server-event was generated.
    * Although the event originates from the P2J Client Side, socket events are server-side 
    * events.
    * 
    * @param    resourceId
    *           The resource ID for the socket to which the READ-RESPONSE event belongs.
    *           
    * @return   The ID of the generated server event or <code>-1</code> if a event could not be
    *           posted (i.e. the resource is not SENSITIVE).
    */
   public int readResponseEvent(long resourceId);
   
   /**
    * Called by the P2J client side, it will inform the server that a server event was generated.
    * Although the event originates from the P2J Client Side, socket events are server-side 
    * events.
    * 
    * @param    resourceId
    *           The resource ID for the server socket to which the CONNECT event belongs.
    * @param    clientSocketId
    *           The resource ID of the socket created for the new connected accepted by the
    *           server socket.
    *           
    * @return   The ID of the generated server event or <code>-1</code> if a event could not be
    *           posted (i.e. the resource is not SENSITIVE).
    */
   public int connectEvent(long resourceId, long clientSocketId);
   
   /**
    * Create a socket and associated with the given ID.  The received ID is not a resource ID; 
    * this represents the ID of the associated {@link LowLevelSocketImpl} instance.
    *   
    * @param    id
    *           The ID of the associated {@link LowLevelSocketImpl} instance.
    *           
    * @return   The ID of the created {@link SocketImpl} resource.
    */
   public long createSocket(int id);
   
   /**
    * Check the directory for the UI {@link com.goldencode.p2j.ui.client.gui.theme.Theme}. If
    * any {@code Theme} is configured in registry (in any scope) it is returned, otherwise
    * {@code null} is returned so that the client will chose the default.
    */
   public String getUiTheme();
   
   /**
    * Read the font-table with the name of the legacy system fonts.
    *
    * @return  The font table.  As 4GL does, this is 0-indexed.
    */
   public FontDetails<?>[] readSystemFontTable();
   
   /**
    * Read the entire font-table.
    * 
    * @return   The font table.  As 4GL does, this is 0-indexed.
    */
   public FontDetails<?>[] readFontTable();
   
   /**
    * Set the theme preferred fonts.
    * 
    * @param   themeFonts
    *          The fonts the current installed {@code Theme} on client prefers.
    * @param   themeFontAliases
    *          The font aliases currently installed {@code Theme} on client prefers.
    */
   public void setThemePreferredFonts(Map<String, String> themeFonts,
                                      Map<String, String> themeFontAliases);
   
   /**
    * Read the default fonts.
    * 
    * @return  An array with the font details, with default font on index o and default fixed
    *          font on index 1.
    */
   public FontDetails<?>[] readDefaultFonts();
   
   /**
    * Get the legacy metrics for the given text, using the font specified by the given key.
    * 
    * @param    text
    *           The text.
    * @param    key
    *           The font's key.
    *           
    * @return   The legacy metrics, with width on index 0 and height on index 1.
    */
   public int[] getLegacyTextMetrics(String text, String key);
   
   /**
    * Read the legacy text metrics for all the texts and each font defined in the font table, for
    * this user.
    *
    * @return  A map having as key the font and text and as value the metrics (0 index - width, 
    *          1 index - height).
    */
   public Map<String, int[]> readLegacyTextMetrics();
   
   /**
    * Get color table of the specified environment.
    * 
    * @param   envName
    *          Environment name. Could be <code>null</code> if is default environment. 
    *          
    * @return  see above.
    */
   public EnvironmentColorTable readColorTable(String envName);
   
   /**
    * Get system colors.
    * 
    * @return  System colors
    */
   public ColorTableEntry[] readSystemColorTable();

   /**
    * Load the specified environment as resource.  The environment is stored in an INI file
    * embedded as a resource inside application jar file.
    * 
    * @param    env
    *           The INI file source.
    * 
    * @return   The environment file bytes, or {@code null} if the environment file could not be
    *           loaded.
    */
   public byte[] loadEnvironment(String env);

   /**
    * Synchronize the field IDs for the server-side widget configurations with the given one.
    * 
    * @param    configs
    *           A map of configurations, having as key a class name and as value a list of field
    *           names (the order in this list gives the field ID).
    */
   public void initWidgetConfigurations(Map<String, List<String>> configs);
   
   /**
    * Accesses the SESSION:TOOLTIPS attribute for this session.
    *
    * @return   <code>true</code> if tooltips are available for this session
    *           <code>false</code> otherwise.
    */
   public boolean getSessionTooltips();
   
   /**
    * Refresh the server-side screen buffer for the given frame.
    * 
    * @param    buff
    *           The screen buffer array data to be refreshed.  May be <code>null</code>.
    */
   public void refreshBuffers(ScreenBuffer[] buff);

   /**
    * Get the Progress stack trace
    * 
    * @return the Progress stack trace
    */
   public String[] getProgressStackTrace();

   /**
    * This API is used when P2J client runs in embedded mode (i.e. the P2J screens are embedded
    * into another customer-specific application).  It allows the customer-specific to execute
    * converted 4GL code.  Only external programs, procedures and functions can be invoked.  The
    * result of this invocation (and any additional OUTPUT parameters, errors, etc) will be 
    * returned back to the customer-specific application.
    * 
    * @param    request
    *           The details about the request.
    *           
    * @return   The response for this request.
    */
   public ServerReturnValue invoke(InvocationRequestPayload request);
   
   /**
    * Delete the resource associated with the specified handle.
    * 
    * @param    handleId
    *           The handle ID of the resource.
    */
   public void deleteHandle(long handleId);

   /**
    * Create a new subscription to the given event for a specified external resource. 
    * This resource will receive a notification when the event is published.
    *
    * @param    global
    *           Flag indicating if the subscription is for global (cross-session) events.
    * @param    subscriber
    *           The resource to notify when the specified event is published. If <code>null</code>,
    *           a {@link ExternalResource} will be built and its ID return.
    * @param    event
    *           The name of the event which is being published.
    * @param    publisher
    *           The handle representing the program or resource that is the source of the event. 
    *           If specified, the named event will only cause a notification if the event source 
    *           matches this handle. <code>null</code> represents the <code>ANYWHERE</code> option
    *           which means that the events will not be filtered by source.
    * 
    * @return   The subscriber's ID.
    */
   public long subscribeExternal(boolean global, Long subscriber, String event, Long publisher);

   /**
    * Delete a subscription to the given event for the specified external resource. This method is 
    * executed in a {@link ErrorManager#silentErrorEnable()}/{@link ErrorManager#silentErrorDisable()}
    * bracket, as the <code>UNSUBSCRIBE</code> statement is executed with a default NO-ERROR 
    * clause.
    *
    * @param    global
    *           Flag indicating if the search is done in the global (cross-session) event register.
    * @param    subscriber
    *           The resource whose subscription is to be deleted.  Must not be <code>null</code>.
    * @param    event
    *           The name of the event which is being published. If this is <code>null</code>, then
    *           subscriptions to all events will be deleted.
    * @param    publisher
    *           The procedure or resource handle representing the source of the event.
    *           If specified, the subscription will only be deleted if the event source matches
    *           this handle.  <code>null</code> represents the <code>ALL</code> option which means
    *           that the events will not be filtered by source. When <code>null</code>, it means
    *           that the publisher was not set.
    */
   public void unsubscribeExternal(boolean global, Long subscriber, String event, Long publisher);
   
   /**
    * Notify all subscribers (to the given event) that the event has occurred.
    * <p>
    * All processing is done with silent error mode implicitly enabled.
    * <p>
    * The subscriptions are fired in the order they were subscribed. Although 4GL
    * documentation specifically states, that "If a named event has multiple subscribers, 
    * the order in which the AVM notifies subscribers is undefined.", the actual
    * 4GL implementation (OpenEdge 10.2B) does fire the events in the order the subscriptions
    * get subscribed.
    *
    * @param    event
    *           The name of the event which is being published.
    * @param    publisher
    *           The procedure or resource handle representing the program that is the source of  
    *           the event. Must not be <code>null</code>.
    * @param    args
    *           The parameters passed to the subscriber. May be <code>null</code> to represent no
    *           parameters. When <code>null</code>, it means that no parameters were specified for
    *           this call.
    */
   public void publishExternal(String event, long publisher, String[] args);
   
   /**
    * Create an external resource to be used by a remote, non P2J, side.
    * 
    * @return   The resource id.
    */
   public long createExternal();
}