/*
** Module   : ServerState.java
** Abstract : stores server state vars that must be sync'd with the client
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description-----------------------------
** 001 GES 20060726   @28202 Created initial version to contain a simple set of server-side state.
** 002 GES 20060728   @28220 Added data and constructors (copy and  default).
** 003 GES 20060731   @28241 Frame hiding/pausing and destroy are now deferred.
** 004 GES 20060801   @28299 Converted to Externalizable to improve efficiency because 
**                           Serializable has unneeded processing (for versioning...) that is 
**                           executed on "inflation".
** 005 SIY 20060804   @28455 Unified storage for all operations.
** 006 SIY 20080502   @38194 Inferred generic types.
** 007 SIY 20090326   @41711 Added external procedure call nesting counter.
** 008 SIY 20090701   @43024 Moved instance clearing from LogicalTerminal.
**                           Added formatCheckingId variable.
** 009 SIY 20110112          Added color mapping transfer from server to client.
** 010 HC  20140119          Implemented runtime support for the COLOR protermcap-attribute.
** 011 MAG 20140829          Add environments variable. 
** 012 VIG 20140901          Added current window ID.
** 013 CA  20151118          Added tracking of server-side dead-frames.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.io.*;
import java.util.*;

import com.goldencode.p2j.ui.ColorTable.EnvironmentColorTable;

/**
 * This class is a container for server-side state which can be easily 
 * transported to or from the client.
 */
public class ServerState
implements Externalizable
{
   /** Tracks the number of I/O interactions within the current block. */
   public int interactions = 0;
   
   /** Tracks any conditional up or down that must be applied. */
   public int conditionalUpDown = 0;
   
   /** Tracks the frame ID associated with the conditional up or down. */
   public int conditionalFrameId = 0;
   
   /** List of deferred hiding/pausing state changes. */
   public List<MarkEntry> markList = null;

   /** ID of frame which should enable hard format check. */
   public int formatCheckId = 0;
   
   /** Current procedure nesting. */
   public int nesting = 0;
   
   /** Environments */
   public List<EnvironmentColorTable> environments = null;

   /** Current window ID */
   public int currentWindow = -1;
   
   /**
    * A list of frames still registered, for which their instantiating procedure was deleted.  
    * These static frames will be de-registered when they will be made not-visible/hidden.
    */
   public int[] deadFrames = null;

   /**
    * Default constructor.
    */
   public ServerState()
   {
   }
   
   /**
    * Copy constructor.
    * 
    * @param   old
    *          The instance to duplicate.
    */
   public ServerState(ServerState old)
   {
      this.interactions       = old.interactions;
      this.conditionalUpDown  = old.conditionalUpDown;
      this.conditionalFrameId = old.conditionalFrameId;
      this.markList           = old.markList;
      this.formatCheckId      = old.formatCheckId;
      this.environments       = old.environments;
      this.currentWindow      = old.currentWindow;
      
      // deadFrames field is excluded from copy, as each time the state is sent to the client,
      // only the current still-registered dead frames are required.
   }

   /**
    * Reset state.  
    */
   public void reset()
   {
      // reset the conditional stuff as this is set externally as a
      // "one-shot" thing
      conditionalUpDown  = 0;
      conditionalFrameId = 0;
      
      // clear the hide/pause/destroy lists since they are one-time only
      markList = null;
      
      // reset format check frame ID
      formatCheckId = 0;

      environments = null;
      
      currentWindow = -1;
      
      deadFrames = null;
   }
   
   /**
    * Replacement for the default object reading method.
    * 
    * @param   in
    *          Input source from which fields will be restored.
    * 
    * @throws  IOException
    *          In case of I/O errors.
    * @throws  ClassNotFoundException
    *          If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      interactions       = in.readInt();
      conditionalUpDown  = in.readInt();
      conditionalFrameId = in.readInt();
      nesting            = in.readInt();
      formatCheckId      = in.readInt();
      currentWindow      = in.readInt();
            
      extractMarkEntries((MarkEntry[]) in.readObject());
      readEnvironment(in);
      
      deadFrames = (int[]) in.readObject();
   }

   /**
    * Replacement for the default object writing method.
    * 
    * @param   out
    *          The output destination to which fields will be saved.
    * 
    * @throws  IOException
    *          In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeInt(interactions);
      out.writeInt(conditionalUpDown);
      out.writeInt(conditionalFrameId);
      out.writeInt(nesting);
      out.writeInt(formatCheckId);
      out.writeInt(currentWindow);
      
      out.writeObject(markListToArray());
      writeEnvironment(out);
      
      out.writeObject(deadFrames);
   }

   /**
    * Convert array of mark entries to list.
    *
    * @param   markArray
    *          Array of mark entries.
    */
   private void extractMarkEntries(MarkEntry[] markArray)
   {
      if (markArray.length > 0)
      {
         markList = new ArrayList<MarkEntry>(Arrays.asList(markArray));
      }
   }

   /**
    * Convert mark entries to array.
    *
    * @return  array of mark entries.
    */
   private MarkEntry[] markListToArray()
   {
      MarkEntry[] marray;

      if (markList == null)
         marray = new MarkEntry[0];
      else
         marray = markList.toArray(new MarkEntry[markList.size()]);

      return marray;
   }
 
   /**
    * Serialize environments.
    * 
    * @param   out
    *          Output stream.
    *          
    * @throws  IOException
    *          In case of serialization error.
    */
   private void writeEnvironment(ObjectOutput out)
   throws IOException
   {
      EnvironmentColorTable[] temp = new EnvironmentColorTable[0];       
      out.writeObject(environments == null ? temp : environments.toArray(temp));
   }

   /**
    * Deserialize environments.
    * 
    * @param   in
    *          Input stream.
    *          
    * @throws  IOException
    * @throws  ClassNotFoundException
    *          In case of deserialization errors.
    */
   private void readEnvironment(ObjectInput in)
   throws IOException, 
          ClassNotFoundException
   {
      environments = Arrays.asList((EnvironmentColorTable[]) in.readObject());      
   }
}