/*
** Module   : SizeAttribute.java
** Abstract : An enumeration providing meta data for widget size attributes.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20150802 Created initial version.
** 002 HC  20160203 Improved runtime support of FRAME:VIRTUAL* attributes.
** 003 SVL 20170809 Added ROW_HEIGHT_CHARS and ROW_HEIGHT_PIXELS.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

/**
 * An enumeration providing meta data for widget size attributes.
 */
public enum SizeAttribute
{
   // the order is assumed by some of the implemented methods!
   HEIGHT_CHARS      ("HEIGHT"),
   WIDTH_CHARS       ("WIDTH"),
   MIN_HEIGHT_CHARS  ("MIN-HEIGHT"),
   MIN_WIDTH_CHARS   ("MIN-WIDTH"),
   MAX_HEIGHT_CHARS  ("MAX-HEIGHT"),
   MAX_WIDTH_CHARS   ("MAX-WIDTH"),
   ROW_HEIGHT_CHARS  ("ROW-HEIGHT-CHARS"),
   VIRT_HEIGHT_CHARS ("VIRT-HEIGHT"),
   VIRT_WIDTH_CHARS  ("VIRT-WIDTH"),
   
   HEIGHT_PIXELS      ("HEIGHT-PIXELS"),
   WIDTH_PIXELS       ("WIDTH-PIXELS"),
   MIN_HEIGHT_PIXELS  ("MIN-HEIGHT-PIXELS"),
   MIN_WIDTH_PIXELS   ("MIN-WIDTH-PIXELS"),
   MAX_HEIGHT_PIXELS  ("MAX-HEIGHT-PIXELS"),
   MAX_WIDTH_PIXELS   ("MAX-WIDTH-PIXELS"),
   ROW_HEIGHT_PIXELS  ("ROW-HEIGHT-PIXELS"),
   VIRT_HEIGHT_PIXELS ("VIRT-HEIGHT-PIXELS"),
   VIRT_WIDTH_PIXELS  ("VIRT-WIDTH-PIXELS");
   
   /** Attribute short name as recognized by 4GL. */
   private final String shortName;
   
   /**
    * Default constructor.
    * 
    * @param    shortName
    *           Attribute short name.
    */
   private SizeAttribute(String shortName)
   {
      this.shortName = shortName;
   }
   
   /**
    * Returns the attribute's short name.
    * 
    * @return   See above.
    */
   public String shortName()
   {
      return shortName;
   }

   /**
    * Returns <code>true</code> when the instance represents a character value.
    *
    * @return   See above.
    */
   public boolean isChar()
   {
      int ord = ordinal();
      return ord >= HEIGHT_CHARS.ordinal() && ord <= VIRT_WIDTH_CHARS.ordinal();
   }
   
   /**
    * Returns <code>true</code> when the instance represents HEIGHT-CHARS, WIDTH-CHARS,
    * HEIGHT-PIXELS, WIDTH-PIXELS, ROW_HEIGHT_CHARS or ROW_HEIGHT_PIXELS.
    *
    * @return   See above.
    */
   public boolean isSize()
   {
      return this == HEIGHT_CHARS  || this == WIDTH_CHARS || 
             this == HEIGHT_PIXELS || this == WIDTH_PIXELS ||
             this == ROW_HEIGHT_CHARS || this == ROW_HEIGHT_PIXELS;
   }
   
   /**
    * Returns <code>true</code> when the instance represents one of 
    * the MIN* attributes.
    *
    * @return   See above.
    */
   public boolean isMin()
   {
      return this == MIN_HEIGHT_CHARS  || this == MIN_WIDTH_CHARS || 
             this == MIN_HEIGHT_PIXELS || this == MIN_WIDTH_PIXELS; 
   }
   
   /**
    * Returns <code>true</code> when the instance represents one of 
    * the MAX* attributes.
    *
    * @return   See above.
    */
   public boolean isMax()
   {
      return this == MAX_HEIGHT_CHARS  || this == MAX_WIDTH_CHARS || 
             this == MAX_HEIGHT_PIXELS || this == MAX_WIDTH_PIXELS; 
   }
   
   /**
    * Returns <code>true</code> when the instance represents one of 
    * the VIRT* attributes.
    *
    * @return   See above.
    */
   public boolean isVirt()
   {
      return this == VIRT_HEIGHT_CHARS  || this == VIRT_WIDTH_CHARS || 
             this == VIRT_HEIGHT_PIXELS || this == VIRT_WIDTH_PIXELS; 
   }
   
   /**
    * Returns <code>true</code> when the instance represents any of the
    * attributes holding a width.
    *
    * @return   See above.
    */
   public boolean isWidth()
   {
      return this == WIDTH_CHARS      || this == MIN_WIDTH_CHARS  || 
             this == MAX_WIDTH_CHARS  || this == VIRT_WIDTH_CHARS || 
             this == WIDTH_PIXELS     || this == MIN_WIDTH_PIXELS || 
             this == MAX_WIDTH_PIXELS || this == VIRT_WIDTH_PIXELS; 
   }
}