/*
** Module   : ToggleBoxWidget.java
** Abstract : Toggle-box widget.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SVL 20101004 Created initial version.
** 002 SIY 20110112 Suppressed warning.
** 003 SIY 20111101 Updated imports.
** 004 CA  20130930 Added LegacyResource annotation, to determine resource type.
** 005 CA  20131115 Changes related to properly instantiate dynamic resources.
** 006 EVL 20140609 Adding TOGGLE-BOX specific options handling.
** 007 GES 20140627 Minor cleanup.
** 008 HC  20140911 GUI coordinate support - data types refactoring.
** 009 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 010 HC  20141021 Implemented window move/position support.
** 011 GES 20150115 Added TODO that use of SCREEN-VALUE in isChecked() is a potential problem.
** 012 IAS 20151214 LOAD-MOUSE-POINTER support.
** 013 CA  20160303 Made attribute setters be a no-op if the value doesn't change.
** 014 CA  20160429 Fixed preRealizeCheck - min 3-char width for toggle-box is enforced only in 
**                  ChUI mode.  There is no lower limit in GUI, for the width. 
** 015 CA  20161005 Fixed issues related to SCREEN-VALUE attribute (initial value or explicitly 
**                  set, when the value is empty string or unknown, or the widget is 
**                  uninitialized).
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * Toggle-box widget.
 */
@LegacyResource(resource = LegacyResource.TOGGLE_BOX)
public class ToggleBoxWidget
extends ControlEntity<ToggleBoxConfig>
implements Hoverable
{
   /**
    * Default constructor.
    */
   public ToggleBoxWidget()
   {
      this(false);
   }

   /**
    * Default constructor.
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    */
   public ToggleBoxWidget(boolean dynamic)
   {
      super(dynamic, new ToggleBoxConfig());
   }

   /**
    * Gets the CHECKED writable attribute.
    *
    * @return    The current value of the CHECKED attribute.
    */
   public logical isChecked()
   {
      // TODO: this seems wrong; it probably should not be relying upon the screen
      //       value!
      
      // obtain screen value
      character screenValue = getScreenValue();

      // check if there is a correct screen value and return it because it can differ from
      // value stored in server side toggle-box config until assign is not executed
      if (screenValue != null)
      {
         // convert character screen value to logical
         logical screenValueLogical = new logical(screenValue);
         
         // update toggle box config first but do not push screen definition
         // because it is already in the correct state 
         if (!screenValueLogical.isUnknown())
         {
            config.checked = screenValueLogical.booleanValue();
         }
      }
      
      return new logical(config.checked);
   }

   /**
    * Set CHECKED attribute value.
    * 
    * @param   checked
    *          New value for the attribute.
    */
   public void setChecked(boolean checked)
   {
      if (config.checked == checked)
      {
         return;
      }
      
      config.checked = checked;
      pushScreenDefinition();
   }

   /**
    * Set CHECKED attribute value.
    * 
    * @param   checked
    *          New value for the attribute.
    */
   public void setChecked(logical checked)
   {
      if (!checked.isUnknown())
      {
         setChecked(checked.booleanValue());
      }
      else
      {
         // create message
         String msgText = buildWidgetSpecificMessage(
            "Unable to assign UNKNOWN value to attribute CHECKED on TOGGLE-BOX ");
         
         // and display error
         ErrorManager.displayError(4083, msgText);
      }
   }

   /**
    * Reports if this widget object is valid to realize.  This is the toggle-box specific check.
    * 
    * @return   <code>true</code> if the widget can be relized.
    */
   @Override
   public boolean preRealizeCheck()
   {
      double widthRequested = config.widthChars;
      
      // safe, comparison of scaled doubles
      // this check is done only for ChUI mode - GUI toggle-box has no lower limit on the width
      if (widthRequested != 0 && widthRequested < 3 && LogicalTerminal.isChui())
      {
         // create message
         String msgText = buildWidgetSpecificMessage("Unable to realize TOGGLE-BOX ");
         
         // and display error
         ErrorManager.displayError(4025, msgText);
         
         return false;
      }
      
      return true;
   }

   /**
    * Set the MOUSE-POINTER attribute
    * 
    * @param name
    *        the attribute value
    */
   @Override
   public void setMousePointerName(String name)
   {
      config.mousePointer = name;
   }

   /**
    * Get the MOUSE-POINTER attribute
    * 
    * @return the attribute value 
    *        
    */
   @Override
   public String getMousePointerImpl()
   {
      return config.mousePointer;
   }

   /**
    * Internal worker for getting the SCREEN-VALUE on a per-widget basis.
    * 
    * @param   initialized
    *          Flag indicating if the widget is realized or not.
    * @param   bdt
    *          The current value in the frame's buffer.
    * @param   fmt
    *          The format to be applied, in case this is a formatted value.
    * @param   ignoreFormat
    *          Flag indicating if the format needs to be ignored.
    *          
    * @return  This widget's screen-value
    */
   @Override
   character getScreenValue(boolean initialized, BaseDataType bdt, String fmt, boolean ignoreFormat)
   {
      if (!initialized)
      {
         return new character("no");
      }
      
      return super.getScreenValue(initialized, bdt, fmt, ignoreFormat);
   }
   
   /**
    * Create toggle-box specific message text with widget variable name included in the final
    * text message.
    * 
    * @param   initial
    *          The initial message specific text.
    * 
    * @return   Resulting string message to display.
    */
   private String buildWidgetSpecificMessage(String initial)
   {
      StringBuilder sb = new StringBuilder();
      
      // getting the frame owning the widget
      GenericFrame parentFrame = getFrame();

      // form the warning message
      sb.append(initial);
      
      if (parentFrame != null)
      {
         // add the widget name
         sb.append(parentFrame.getOriginalName(this));
      }
      else
      {
         // if for some reasons parent frame is null request the widget label as last resort
         sb.append(getLabelStr());
      }
      
      return sb.toString();
   }
}