/*
** Module   : TriggerManager.java
** Abstract : trigger registration, deregistration, scoping and invocation
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20051202   @23598 Created initial version.
** 002 NVS 20060307   @24918 Fixed the trigger lookup method lookup().
** 003 NVS 20060612   @27116 Utilizing merge() method of WaitList class.
** 004 NVS 20060612   @27140 Fixed some NullPointerExceptions in trigger
**                           handling code for ANYWHERE triggers with no
**                           frame involved.
** 005 GES 20061017   @30453 Modifications to allow specific events to
**                           override the ANYWHERE processing and to be
**                           honored as triggers when they otherwise
**                           would not have been.
** 006 NVS 20070118   @31890 Reimplemented trigger registry management
**                           based on EventList class. WaitList class is
**                           no longer in use.
** 007 CA  20081204   @40819 ROW-DISPLAY event handling - these events
**                           will be deregistered from the client, when
**                           they are out of scope.
** 008 GES 20081217   @40926 Minor signature change.
** 009 GES 20090629   @42970 Remove most duplicates (except the last one) in
**                           the EventList instances when triggers are
**                           registered. This greatly reduces memory usage
**                           with no negative affect on processing since the
**                           removed instances cannot be reached nor can they
**                           be accessed via "revert".
** 010 SIY 20101122          Inferred generics. Small cleanups in javadoc.
** 011 CA  20130131          Implement ON ... PERSISTENT RUN triggers - as
**                           they are alive for the duration of the current
**                           session, when the scope which has defined them
**                           is ended, they are moved in the previous scope.
** 012 CA  20131013          Added new version of scopeFinished API, which explicitly merges a 
**                           registry into another one.
** 013 CA  20151029          Triggers created in a persistent procedure while a event processing
**                           loop is active need to be pushed to the event list for the nearest
**                           event processing loop, so they will be available in the UI.
** 014 CA  20151204          Fixed deregisterTrigger (ON ... REVERT) - it needs to remove the 
**                           trigger only from the current scope.
** 015 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 016 CA  20161018          Any client-to-server code invocations must maintain the trigger 
**                           registries, not just explicit trigger invocation.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.util.*;
import com.goldencode.p2j.util.*;

/**
 * This class organizes triggers registration into multiple scopes of LIFO
 * stacks.
 * <p>
 * Registrations are kept in a scoped dictionary. Every top level block creates
 * a new scope for triggers in <code>scopeStart</code> method. The registry
 * inside the scope is a <code>LinkedList</code> of <code>EventList</code>
 * objects. The list is stored as the scope extra object, so the dictionary is
 * not in use.
 * The scope's registry works as a FIFO stack for <code>EventList</code>s. The
 * most recent registration is added to the list head.
 * Deregistration may require full removal of one or more 
 * <code>EventList</code> objects
 * or editing their contents. Editing happens when only some events for a
 * widget have to be deregistered, but some others remain active.
 */
public class TriggerManager
{
   /**
    * Process a notification that a new block scope is about to be entered.
    *
    * @param registry
    *        <code>ScopedDictionary</code> that keeps track of trigger
    *        registrations for this session.
    */
   public static void scopeStart(ScopedDictionary<?,?> registry)
   {
      registry.addScope(new LinkedList<EventList>());
   }
   
   /**
    * Process a notification that a scope is about to be exited.
    * <p>
    * All non-persistent triggers registered in the scope are deregistered.
    * <p>
    * All persistent triggers registered in this scope are moved to the
    * previous scope.
    *
    * @param registry
    *        <code>ScopedDictionary</code> that keeps track of trigger
    *        registrations for this session.
    */
   @SuppressWarnings("unchecked")
   public static void scopeFinished(ScopedDictionary<?,?> registry)
   {
      LinkedList<EventList> list = (LinkedList<EventList>) registry.getScope();
      registry.deleteScope();
      LinkedList<EventList> prev = (LinkedList<EventList>) registry.getScope();
      
      scopeFinished(list, prev, true);
   }

   /**
    * Registers a trigger within the current scope. If this is a persistent
    * trigger, all prior persistent triggers defined for the same events (on
    * any scope) will be removed.
    *
    * @param    registry
    *           Tracks trigger registrations for this session.
    * @param    events
    *           Defines the events and widgets on which to execute this trigger.
    * @param    tid
    *           The trigger ID. 
    */
   @SuppressWarnings("unchecked")
   public static void register(ScopedDictionary<?,?> registry,
                               EventList             events, 
                               int                   tid)
   {
      events.setTriggerId(tid);

      // insert the latest trigger def at the front of the list. all created triggers are saved,
      // regardless if previous scope has a trigger for this or not.  a trigger can be removed 
      // from the registry only and only if its creator procedure (or resource) is deleted
      LinkedList<EventList> list = (LinkedList<EventList>) registry.getScopeAt(0);
      list.addFirst(events);
   }

   /**
    * Deregisters a trigger for specified events and widgets.
    * <p>
    * In Progress, requests to deregister (REVERT) the very first trigger
    * definition for a pair of {event, widget} are silently ignored.
    * <p>
    * Triggers are deregistered only from the current scope.  If there are definitions of this
    * trigger in previous scopes (or if this is an ANYWHERE trigger), then all definitions are
    * removed; otherwise, all are removed but one.
    * <p>
    * Trigger deregistration is complicated. Triggers are registered for a
    * list of set pairs like:
    * <code>
    *    {(event1, event2) (widget1, widget2)}          // item 1
    *    {(event1, event2) (widget1, widget2, widget3)} // item 2
    *    ...
    * </code>
    * A deregistration request may come, for example, for event2 of widget1
    * and event1 of widget2 and event10 of widget1000. Deregistration should
    * be made for the first encountered match {event, widget} and the 
    * remaining registrations should be kept. So, in that case: 
    * <code>
    *    {(event1) (widget1)}                           // item 1.1
    *    {(event2) (widget2)}                           // item 1.2
    *    {(event1, event2) (widget1, widget2, widget3)} // item 2
    *    ...
    * </code>
    * is the result.
    *
    * @param registry
    *        <code>ScopedDictionary</code> that keeps track of trigger
    *        registrations for this session.
    * @param el
    *        <code>EventList</code> object, encoding events and widgets this
    *        request has come for
    */
   @SuppressWarnings("unchecked")
   public static void deregister(ScopedDictionary<?,?> registry, EventList el)
   {
      LinkedList<EventList> registryScope = null;

      // First of all, the incoming request for deregistration has to be
      // "distilled" - all combinations of {event, widget} that refer to
      // the only existing registration have to be removed.
      
      // split the event list into (event, widgetId) and (event, resourceId) pairs, so that 
      // references can be counted.  after references are counted, all pairs with only one reference
      // are removed from the list
      List<EventDefinition> atoms = el.splitEvents();
      Map<EventDefinition, integer> refsOthers = new HashMap<>();
      Map<EventDefinition, integer> refsCurrent = new HashMap<>();
      for (EventDefinition atom : atoms)
      {
         refsOthers.put(atom, new integer(0));
         refsCurrent.put(atom, new integer(0));
      }
      
      // loop over all scopes - 1st phase
      
      // count all references in other scopes:
      // - if there are matches, then all triggers in this scope will be removed
      // - if there are no matches in previous scopes and this is NOT an ANYWHERE event, then all 
      // triggers but one in this scope will be removed
      for (int i = 1; i < registry.size(); i ++)
      {
         registryScope = (LinkedList<EventList>) registry.getScopeAt(i);
         
         // loop over all registry EventLists of the scope
         Iterator<EventList> iter = registryScope.iterator();
         EventList rel = null;
         
         while (iter.hasNext())
         {
            // get the next event list from registry
            rel = iter.next();
            
            // walk the expanded list of (event, id) pairs
            for (EventDefinition ed : refsOthers.keySet())
            {
               if (rel.intersects(ed))
               {
                  refsOthers.get(ed).increment();
               }
            }
         }
      }
      
      // check references in current scope
      {
         Iterator<EventList> iter = ((LinkedList<EventList>) registry.getScope()).iterator();
         EventList rel = null;
         
         while (iter.hasNext())
         {
            // get the next event list from registry
            rel = iter.next();
            
            // walk the expanded list of (event, id) pairs
            for (EventDefinition ed : refsCurrent.keySet())
            {
               if (rel.intersects(ed))
               {
                  refsCurrent.get(ed).increment();
               }
            }
         }
      }

      // second phase: loop through all atoms and remove them from the current scope
      LinkedList<EventList> currentScope = (LinkedList<EventList>) registry.getScope();
      for (EventDefinition atom : atoms)
      {
         boolean prevCount = refsOthers.get(atom).intValue() > 0;
         int currentCount = refsCurrent.get(atom).intValue();
         
         Iterator<EventList> iter = currentScope.iterator();
         while (iter.hasNext())
         {
            // get the next event list from current scope
            EventList rel = iter.next();
            
            if (!prevCount && currentCount == 1 && !atom.isAnywhereGlobally())
            {
               // non-ANYWHERE globally: only one reference found, do not remove it
               // ANYWHERE globally: all references in current scope are removed
               break;
            }
            
            // there are previous references or there is more than one reference in current scope,
            // attempt to remove it
            if (rel.remove(atom))
            {
               if (rel.isEmpty())
               {
                  iter.remove();
               }
               
               currentCount--;
            }
         }
      }
   }
   
   /**
    * Compute trigger differences from the specified list.  The new triggers will be saved in
    * {@link ServerReturnValue#newTriggers} and the dead triggers will be saved in 
    * {@link ServerReturnValue#deadTriggers}.
    * 
    * @param    srv
    *           The trigger's return data container, where the event lists will be saved.
    * @param    prevTriggers
    *           The trigger list before the trigger was invoked. 
    * @param    currentTriggers
    *           The trigger list after the trigger was invoked.
    */
   public static void computeTriggerDifferences(ServerReturnValue srv,
                                                EventList         prevTriggers,
                                                EventList         currentTriggers)
   {
      srv.deadTriggers = null;
      srv.newTriggers = null;
      
      // only the state of the triggers as at the end of the external program is saved.  any 
      // trigger added for a dynamic widget (for example, created in an internal procedure) is not
      // visible and dropped.

      EventList common = prevTriggers.intersection(currentTriggers);
      EventList deadList = prevTriggers.difference(common);
      EventList newList = currentTriggers.difference(common);
      
      srv.deadTriggers = deadList;
      srv.newTriggers = newList;
   }

   /**
    * Merges all trigger registry entries of all scopes and produces a
    * <code>EventList</code> instance that represents all sensitive widgets
    * with all event codes they are sensitive to.
    *
    * @param   registry
    *          <code>ScopedDictionary</code> that keeps track of trigger
    *          registrations for this session.
    *
    * @return  a merged <code>EventList</code> of all registry entries.
    */
   @SuppressWarnings("unchecked")
   public static EventList mergeTriggerRegistry(ScopedDictionary<?,?> registry) 
   {
      EventList merge = new EventList();
      LinkedList<EventList> registryScope = null;
      
      // loop over all scopes
      EventList el = null;
      for (int i = 0; i < registry.size(); i ++)
      {
         registryScope = (LinkedList<EventList>)registry.getScopeAt(i);

         // loop over all registry EventLists of the scope
         Iterator<EventList> iter = registryScope.iterator();
         while (iter.hasNext())
         {
            el = iter.next();
            merge.merge(el, false);
         }
      }

      return merge;
   }
   
   /**
    * Process a notification that a scope is about to be exited.
    * <p>
    * All non-persistent triggers registered in the scope are deregistered if current block is not
    * an external procedure.
    * <p>
    * All persistent triggers registered in this scope are moved to the previous scope.
    *
    * @param    source
    *           The source event list.
    * @param    target
    *           The target where the persistent triggers will be merged.
    * @param    cleanup
    *           Flag indicating if the triggers are cleaned up using 
    *           {@link #cleanup(EventList, Set)}.
    */
   static void scopeFinished(LinkedList<EventList> source, 
                             LinkedList<EventList> target, 
                             boolean               cleanup)
   {
      if (source.size() == 0)
      {
         return;
      }
      
      Set<Integer> wids = new HashSet<Integer>(1);
      
      for (EventList events : source)
      {
         TriggerDefinition tdef = LogicalTerminal.getTriggerDefinition(events.getTriggerId());
         
         // will survive:
         // - all persistent triggers
         // - all triggers for which their instantiating procedure still exists and were created
         // in the program's external procedure
         boolean survive = LogicalTerminal.isPersistentTrigger(events.getTriggerId()) ||
                           (tdef != null               && 
                            tdef.isFromExternalProgram() &&
                            ProcedureManager._isPersistent(tdef.getExternalProgram()));

         // prev can be null only when the global scope is ending
         if (target != null && survive)
         {
            target.addFirst(events);
         }
         else if (cleanup)
         {
            cleanup(events, wids);
         }
      }
      
      if (cleanup)
      {
         clearRowDisplay(wids);
      }
   }

   /**
    * Handle the cleanup associated with a given event list. The associated
    * trigger will be deregistered from the LogicalTerminal's trigger
    * registry. In addition, any widgets that need notifications for row
    * display events, will be removed from the client side list.
    *
    * @param    list
    *           The event list to be deregistered.
    * @param    triggers
    *           The list of widgets no longer needing row display notifications.
    *           This will be modified on output to add any widgets that need
    *           to be removed from the notification list.
    */
   private static void cleanup(EventList list, Set<Integer> triggers)
   {
      LogicalTerminal.deregisterTrigger(list.getTriggerId(), true);

      Map<Integer, Integer> rde = list.getRowDisplayEvents();
      
      if (rde != null)
      {
         for (int wid : rde.keySet())
         {
            triggers.add(rde.get(wid));
         }
      }
   }
   
   /**
    * Remove the given widgets from the client-side list of widgets which must
    * get notifications of row display events.
    *
    * @param    triggers
    *           The list of widgets no longer needing row display notifications.
    */
   private static void clearRowDisplay(Set<Integer> triggers)
   {
      if (triggers.size() > 0)
      {
         int[] triggerIds = Utils.integerCollectionToPrimitive(triggers);
         LogicalTerminal.removeRowDisplayEvents(triggerIds);
      }
   }
}