/*
** Module   : DirectPrimitives.java
** Abstract : Terminal output driver that mirrors output into direct terminal
**            and into shadow buffer. 
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20090312   @41594 Initial implementation.
** 002 SIY 20090319   @41644 Changed addArray() signature.
** 003 SIY 20090911   @43893 Now class is inherited from BasePrimitives.
** 004 SIY 20100619          Switching to own implementation of character UI.
** 005 SIY 20100804          Changes in imports, some refactorings.
** 006 SIY 20110112          Changes related to new color handling scheme.
** 007 CA  20140805          Allow direct access to all ChUI specific APIs provided by the 
**                           BasePrimitives, as this is the base class for all ChUI primitives.
** 008 HC  20140911          GUI coordinate support - data types refactoring.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;

/**
 * Direct output primitives which implement shadow buffer semantic. This
 * buffer enables access to screen content.
 */
public class DirectPrimitives
extends BasePrimitives
{
   /**
    * Original direct output primitives.
    */
   private final BasePrimitives direct;
   
   /**
    * Shadow buffer.
    */
   private final BasePrimitives buffer;

   /**
    * Constructor.
    * 
    * @param   direct
    *          Toolkit instance used for direct drawing.
    */
   public DirectPrimitives(BasePrimitives direct)
   {
      this.direct = direct;

      buffer = new TerminalBuffer(direct.screenHeight(),
                                  direct.screenWidth());
   }

   /**
    * Perform a bulk output operation for the whole screen at once with
    * output masking when desired. Set the cursor afterwards.
    *
    * @param   screen
    *          Container for whole screen content and cursor position.
    */
   @Override
   public void updateScreen(ScreenData screen)
   {
      direct.updateScreen(screen);
      buffer.updateScreen(screen);
   }

   /**
    * Get the screen bitmap for this output primitives implementation. 
    * 
    * @return   The screen bitmap.
    */
   @Override
   public ScreenBitmap getScreenBitmap()
   {
      return direct.getScreenBitmap();
   }

   /**
    * Replace the current bitmap instance with this one.
    * 
    * @param    bitmap
    *           The new bitmap instance.
    */
   @Override
   public void setScreenBitmap(ScreenBitmap bitmap)
   {
      direct.setScreenBitmap(bitmap);
   }

  /**
    * Add a character to the virtual terminal at the current cursor
    * position.
    *
    * @param    chr
    *           The character to output.
    * @param    color
    *           The number of the color-pair (foreground + background) to
    *           use for blanking the rectangle.
    */
   @Override
   public void append(int chr, Color color)
   {
      direct.append(chr, color);
      buffer.append(chr, color);
   }
   
   /**
    * Add a string to the virtual terminal at the current cursor position.
    *
    * @param    str
    *           The text to output.
    * @param    color
    *           The number of the color-pair (foreground + background) to
    *           use for blanking the rectangle.
    */
   @Override
   public void append(String str, Color color)
   {
      direct.append(str, color);
      buffer.append(str, color);
   }
   
   /**
    * Draw a horizontal line of the specified length starting at the
    * current cursor position.
    *
    * @param    length
    *           The length of the line to draw in columns.
    * @param    color
    *           The number of the color-pair (foreground + background).
    */
   @Override
   public void hLine(int length, Color color)
   {
      direct.hLine(length, color);
      buffer.hLine(length, color);
   }

   /**
    * Draw a vertical line of the specified length starting at the current
    * cursor position.
    *
    * @param    length
    *           The length of the line to draw in rows.
    * @param    color
    *           The number of the color-pair (foreground + background).
    */
   @Override
   public void vLine(int length, Color color)
   {
      direct.vLine(length, color);
      buffer.vLine(length, color);
   }
   
   /**
    * Blank out a box using the specified attributes and color pair.
    *
    * @param    left
    *           The left column of the rectangle.
    * @param    top
    *           The top row of the rectangle.
    * @param    right
    *           The right column of the rectangle.
    * @param    bottom
    *           The bottom row of the rectangle.
    * @param    color
    *           The number of the color-pair (foreground + background).
    */
   @Override
   public void clearArea(int left, int top, int right, int bottom, Color color)
   {
      direct.clearArea(left, top, right, bottom, color);
      buffer.clearArea(left, top, right, bottom, color);
   }
   
   /**
    * Draw a box using the specified dimensions and color pair.
    *
    * @param    left
    *           The left column of the rectangle.
    * @param    top
    *           The top row of the rectangle.
    * @param    right
    *           The right column of the rectangle.
    * @param    bottom
    *           The bottom row of the rectangle.
    * @param    color
    *           The number of the color-pair (foreground + background).
    */
   @Override
   public void box(int left, int top, int right, int bottom, Color color)
   {
      direct.box(left, top, right, bottom, color);
      buffer.box(left, top, right, bottom, color);
   }
   
   /**
    * Get absolute X position of the cursor.
    *
    * @return   The X position of the cursor.
    */
   @Override
   public int getX()
   {
      return direct.getX();
   }
   
   /**
    * Get absolute Y position of the cursor.
    *
    * @return   The Y position of the cursor.
    */
   @Override
   public int getY()
   {
      return direct.getY();
   }
   
   /**
    * Mark a rectangle as needing repainting.
    * Makes sense with the double buffered output only; ignored here.
    * 
    * @param    rect
    *           rectangle to mark for repainting
    */
   @Override
   public void refresh(NativeRectangle rect)
   {
      buffer.refresh(rect);
   }

   /**
    * Position the virtual cursor during the drawing.
    * 
    * @param    x
    * @param    y
    * 
    * @return   <code>true</code> if cursor positioning succeeded.
    */
   @Override
   public boolean cursorAt(int x, int y)
   {
      buffer.cursorAt(x, y);
      return direct.cursorAt(x, y);
   }
   
   /**
    * Position the real cursor.
    * 
    * @param    x
    * @param    y
    * 
    * @return   <code>true</code> if cursor positioning succeeded.
    */
   @Override
   public boolean cursorStay(int x, int y)
   {
      buffer.cursorStay(x, y);
      return direct.cursorAt(x, y);
   }

   /**
    * Gets the number of columns on the screen.
    *
    * @return   The number of columns (in characters).
    */
   @Override
   public int screenWidth()
   {
      return direct.screenWidth();
   }

   /**
    * Gets the number of rows on the screen.
    *
    * @return   The number of rows (lines of characters).
    */
   @Override
   public int screenHeight()
   {
      return direct.screenHeight();
   }
   
   /**
    * Synchronize the state of the physical terminal with the state of
    * the virtual terminal.
    */
   @Override
   public void sync()      
   {
      direct.sync();
   }

   /**
    * Clear the buffer bitmaps.
    */
   public void clear()
   {
      buffer.sync();
   }
   
   /**
    * Draw the upper-left border corner.
    * 
    * @param    col
    *           The column number where to draw.
    * @param    row
    *           The row number where to draw.
    * @param    color
    *           Color.
    */
   @Override
   public void drawULCorner(int col, int row, Color color)
   {
      direct.drawULCorner(col, row, color);
      buffer.drawULCorner(col, row, color);
   }
   
   /**
    * Draw the upper-right border corner.
    *  
    * @param    col 
    *           The column number where to draw.
    * @param    row 
    *           The row number where to draw.
    * @param    color
    *           Color.
    */
   @Override
   public void drawURCorner(int col, int row, Color color)
   {
      direct.drawULCorner(col, row, color);
      buffer.drawULCorner(col, row, color);
   }

   /**
    * Draw the lower-left border corner.
    *  
    * @param    col 
    *           The column number where to draw.
    * @param    row 
    *           The row number where to draw.
    * @param    color
    *           Color.
    */
   @Override
   public void drawLLCorner(int col, int row, Color color)
   {
      direct.drawLLCorner(col, row, color);
      buffer.drawLLCorner(col, row, color);
   }
   
   /**
    * Draw the lower-right border corner. 
    *  
    * @param    col 
    *           The column number where to draw.
    * @param    row 
    *           The row number where to draw.
    * @param    color
    *           Color.
    */
   @Override
   public void drawLRCorner(int col, int row, Color color)
   {
      direct.drawLRCorner(col, row, color);
      buffer.drawLRCorner(col, row, color);
   }
   
   /**
    * Draw the upper or lower border. 
    *  
    * @param    col
    *           The column number where to draw.
    * @param    row
    *           The row number where to draw.
    * @param    width
    *           The border width.
    * @param    color
    *           Color.
    */
   @Override
   public void drawHBorder(int col, int row, int width, Color color)
   {
      direct.drawHBorder(col, row, width, color);
      buffer.drawHBorder(col, row, width, color);
   }
   
   /**
    * Draw the left or right border. 
    *  
    * @param    col
    *           The column number where to draw.
    * @param    row
    *           The row number where to draw.
    * @param    height
    *           The border height.
    * @param    color
    *           Color.
    */
   @Override
   public void drawVBorder(int col, int row, int height, Color color)
   {
      direct.drawVBorder(col, row, height, color);
      buffer.drawVBorder(col, row, height, color);
   }

   /**
    * Check if this output primitives implementation requires border corners
    * when drawing a box. This implementation requires box border corners,
    * so this method will return <code>true</code>.
    * 
    * @return   <code>true</code>, as the box will be drawn with border 
    *           corners.
    */
   @Override
   public boolean hasBoxBorderCorners()
   {
      return direct.hasBoxBorderCorners();
   }

   /**
    * Get character at specified position.
    * 
    * @param    row
    *           Character row.
    * @param    col
    *           Character column.
    * 
    * @return   character from specified position or space if operation is not
    *           supported.
    */
   @Override
   public Cell cellAt(int row, int col)
   {
      return buffer.cellAt(row, col);
   }
}