/*
** Module   : EditorImpl.java
** Abstract : ChUI editor implementation
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20111121 Extracted CHUI-specific part from Editor.
** 002 CA  20140805 Do not allow ChUI impl classes access to APIs outside of ChuiOutputManager.
** 003 HC  20140911 GUI coordinate support - data types refactoring.
** 004 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 005 GES 20150103 Changes to match the move of some parent class state into the config class.
** 006 IAS 20150326 Fixed typo
** 007 HC  20150323 Removed business logic from config classes.
** 008 CA  20150509 Extracted ChUI-specific logic from the Editor.java class.
** 009 CA  20150616 Moved other ChUI specific logic from Editor.java class.
** 010 IAS 20150924 Added AUTO-RESIZE attribute support
** 011 EVL 20151104 Fix for '\r' line separators in incoming character data.  All these characters
**                  are removed in ChUI.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.util.*;

/**
 * CHUI-specific portion of EDITOR component.
 */
public class EditorImpl
extends Editor<ChuiOutputManager>
{
   /** Editor widget TAB size in spaces. */
   private static final int EDITOR_TAB_SIZE = 4;

   /**
    * If true then the editor is into read-only mode and INSERT key hasn't
    * been pressed. In this case we should display OFF state while actual
    * insert state may be either ON or OFF. 
    **/
   private boolean fakeInsert = false;

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public EditorImpl(WidgetId id, EditorConfig cfg)
   {
      super(id, cfg);
   }
   
   /**
    * Process KeyEvents for this widget.
    * 
    * @param   ke
    *          Key event structure.
    */
   @Override
   public void processKeyEvent(KeyInput ke)
   {
      super.processKeyEvent(ke);

      int key = ke.actionCode();

      if (key == Keyboard.KA_INSERT_MODE)
      {
         fakeInsert = false;
         insert = !getInsertMode();
         setInsertMode(insert);
         refresh();
         return;
      }
      else if (key == Keyboard.KA_EDITOR_TAB)
      {
         editorTab();
      }
      else if (key == Keyboard.KA_EDITOR_BACKTAB)
      {
         editorBackTab();
      }
      else if (key == Keyboard.KA_APPEND_LINE)
      {
         appendLine();
      }
      else if (key == Keyboard.KA_PAGE_LEFT)
      {
         if (absoluteX() < config.innerChars)
            cursorBeginOfLine();
         else
            setCursorX(absoluteX() - config.innerChars + 1);
      }
      else if (key == Keyboard.KA_PAGE_RIGHT)
      {
         if (absoluteX() + config.innerChars >= config.bufferChars - 1)
            cursorEndOfLine();
         else
            setCursorX(absoluteX() + config.innerChars - 1);
      }
      else if (key == Keyboard.KA_OPEN_LINE_ABOVE)
      {
         openLineAbove();
      }
      else if (key == Keyboard.KA_HOME)
      {
         cursorHome();
      }
      else if (key == Keyboard.KA_LEFT_END)
      {
         cursorBeginOfLine();
      }
      else if (key == Keyboard.KA_RIGHT_END)
      {
         cursorEndOfLine();
      }
      else if (key == Keyboard.KA_END)
      {
         while (absoluteY() < lastRow())
            cursorDown();
         cursorEndOfLine();
      }
      else if (key == Keyboard.KA_PREV_WORD)
      {
         wordLeft();
      }
      else if (key == Keyboard.KA_NEXT_WORD)
      {
         wordRight();
      }
      else if (key == Keyboard.KA_DELETE_END_LINE)
      {
         deleteToEndOfLine();   
      }
      else if (key == Keyboard.KA_CLEAR)
      {
         clear();
      }
      else if (key == Keyboard.KA_DELETE_LINE)
      {
         if (deleteLineWorker())
         {
            refresh();
         }
      }
      else if (key == Keyboard.KA_DELETE_WORD)
      {
         // note different handling of word separators in comparison
         // to word-left word-right operations
         deleteWord();
      }
      else if (!forceRedraw)
      {
         // ignore all unknown actions
         return; 
      }

      setInsertMode(insert);
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    *           The previous configuration.
    */
   @Override
   public void afterConfigUpdate(EditorConfig beforeUpdate)
   {
      super.afterConfigUpdate(beforeUpdate);
      
      // if no scroll bars are defined, then reset is-box flag
      if (config.scrollbarVertical || config.scrollbarHorizontal)
      {
         config.box = true;
      }
      else
      {
         config.box = false; 
      }

      int delta = config.box ? 2:0;

      // safe, comparison of scaled doubles
      if (config.widthChars == 0 && config.heightChars == 0)
      {
         // adjust chars/lines by inner chars/lines
         config.widthChars = Coordinate.scale(config.innerChars + delta);
         config.heightChars = Coordinate.scale(config.innerLines + delta);
      }
      else
      {
         // adjust inner chars/lines by chars/lines
         config.innerChars = (int) config.widthChars - delta;
         config.innerLines = (int) config.heightChars - delta;
      }
      
      // buffer-chars must be greater or equal to visible area width
      if (config.bufferChars < config.innerChars)
      {
         config.bufferChars = config.innerChars;
      }
      
      // buffer-lines, if specified, must be at least equal to visible area height 
      if (config.bufferLines > 0 && config.bufferLines < config.innerLines)
      {
         config.bufferLines = config.innerLines;
      }
   }
   
   /**
    * Assign new instance of <code>Label</code> to the fill-in widget.
    * 
    * @param   label
    *          Reference to new <code>Label</code> instance.
    */
   @Override
   public void setLabelInstance(Label<ChuiOutputManager> label)
   {
      super.setLabelInstance(label);
      
      if (label != null)
      {
         label.config().dcolor = config().dcolor;
         label.config().pfcolor = config().pfcolor;
      }
   }
   
   /**
    * Sets position of terminal's cursor unconditionally (even if this editor
    * is not in focus).
    */
   @Override
   public void drawCaretUnconditionally()
   {
      if (!isEnabled())
      {
         return;
      }

      Point origin = new Point(screenLocation());
      
      int box = (config.box) ? 1 : 0;
      
      origin = origin.offset(config.cursorX + box, config.cursorY + box);
      
      screen().at(origin.x, origin.y);
   }

   /**
    * Actual drawing routine.
    *
    * @param   color
    *          Color for drawing.
    */
   @Override
   public void draw(Color color)
   {
      boolean empty = parsedText.isEmpty() || 
                      (parsedText.size() == 1 && parsedText.get(0).length() == 0);
      boolean uninit = (getState() == ScreenBuffer.UNINITIALIZED && empty);
      Point start = screenLocation();
      
      // in chui char coordinates are always whole numbers
      int width  = (int) width();
      int height = (int) height();
      
      ChuiOutputManager screen = screen();
      
      if (config.box)
      {
         Color clr  = color.withoutUnderline();
         Color rclr = clr.withReverse();

         screen.box(start, dimension(), clr);

         if (config.scrollbarVertical && height >= 5)
         {
            int scrollbar_offset = (int) calcVpos(height - 4) + 1;

            for (int i = 0; i < (height - 2); i++)
            {
               screen.at(start.offset(width - 1, i + 1))
                     .append(' ', (i == scrollbar_offset) ? clr : rclr);            
            }
         }
         
         if (config.scrollbarHorizontal && width >= 5)
         {
            int scrollbar_offset = (int) calcHpos(width - 4) + 1;

            for (int i = 0; i < (width - 2); i++)
            {
               screen.at(start.offset(i + 1, height - 1))
                     .append(' ', (i == scrollbar_offset) ? clr : rclr);
            }
         }
         
         start = start.offset(1, 1);
         width -= 2;
         height -= 2;
      }
      
      String space = StringHelper.repeatChar(' ', (int) width);

      int line = config.scrollY;
      
      for (int i = 0; i < height; i++, line++)
      {
         String text = prepareTextRow(uninit, width, space, line);
         
         screen.at(start.offset(0, i)).append(text, color);
      }

      if (widgetFocused && isEnabled())
      {
         drawCaret();
         setInsertMode(insert);
      }
   }
   
   /**
    * Sets the cursor position in the current line
    *
    * @param    pos
    *           the new cursor position value\
    * @param    redraw
    *           request redraw/sync
    *           
    * @return   <code>true</code> if the cursor position is valid and set 
    */
   @Override
   public boolean setCursorChar(int pos, boolean redraw)
   {
      if (pos < 0 && pos > getLine().length())
      {
         return false;
      }
      
      if (pos < (config.innerChars - 1))
      {
         config.cursorX = Math.min(pos, getLine().length());
         config.scrollX = 0;
      }
      else
      {
         config.cursorX = pos % config.innerChars;
         config.scrollX = pos - config.cursorX;
      }
   
      return super.setCursorChar(pos, redraw);
   }

   /**
    * Calculate horizontal position of the scroll-bar thumb.
    * 
    * @param   size
    *          The width of the scroll-bar.
    * 
    * @return  The horizontal position of the scroll-bar thumb.
    */
   protected double calcHpos(double size)
   {
      double res = size / (config.bufferChars + 1);
      res *= absoluteX() + 0.5;
      
      return Coordinate.scale(res);
   }

   /**
    * Calculate vertical position of the scroll-bar thumb.
    * 
    * @param   size
    *          The height of the scroll-bar.
    * 
    * @return  The vertical position of the scroll-bar thumb.
    */
   protected double calcVpos(double size)
   {
      double res = size / parsedText.size();
      res *= absoluteY() + 0.5;
      
      return Coordinate.scale(res);
   }

   /**
    * Get absolute horizontal position in text.
    * 
    * @return  The horizontal position in the text.
    */
   protected int absoluteX()
   {
      return config.cursorX + config.scrollX;
   }
    
   /**
    * Check if the given text needs to output on a new line.
    * 
    * @param    subtext
    *           The text.
    * @param    placeChar
    *           flag indicating if this call is from {@link #placeChar}.
    *
    * @return   <code>true</code> if the text will output on a new line.
    */
   @Override
   protected boolean goToNewLine(String subtext, boolean placeChar)
   {
      return placeChar ? subtext.length() >= config.bufferChars 
                       : subtext.length() > config.bufferChars;
   }

   /**
    * Merge line with subsequent one(s) if possible
    * 
    * @param   row
    *          line's row
    */
   @Override
   protected void mergeLines(int row)
   {
      for(; row < parsedText.size() - 1; row++)
      {
         Line current = parsedText.get(row);
         if (current.isNewLine())
         {
            break;
         }
         
         int len = (int) config.bufferChars - current.getText().length();
         if (len <= 0)
         {
            break;
         }
         
         Line next = parsedText.get(row+1);
         String s = next.getText();
         if (s.length() <= len)
         {
            parsedText.set(row, new Line(current.getText() + s, 
                                         current.getOffset(), 
                                         next.isNewLine()));
            parsedText.remove(row + 1);
            break;
         }
         
         int pos = s.lastIndexOf(' ', len);
         if (pos < 0)
         {
            break;
         }
         
         parsedText.set(row, new Line(current.getText() + s.substring(0, pos + 1), 
                                      current.getOffset(), 
                                      false));
         
         parsedText.set(row + 1, new Line(s.substring(pos + 1), 
                                          current.getOffset() + pos, 
                                          next.isNewLine()));
      }
   }

   /**
    * Changes the insert mode state in the window.
    * 
    * @param   mode
    *          New state.
    *          <ul>
    *             <li><code>false</code>- overwrite mode,
    *             <li><code>true</code>- insert mode.
    *          </ul>
    */
   protected void setInsertMode(boolean mode)
   {
      if (config.readOnly && fakeInsert)
      {
         Window.setInsertMode(Window.INSERT_MODE_OVR, true);
      }
      else
      {
         Window.setInsertMode(mode ? Window.INSERT_MODE_INS : Window.INSERT_MODE_OVR, true);
      }
   }

   /**
    * Place new line character at current cursor position. 
    */
   @Override
   protected void placeLF()
   {
      Line current = getLine();

      if (!insert)
      {
         if (getCursorLine() >= parsedText.size())
         {
            ConfigHelper.setModified(config, true);
            if (!current.isNewLine())
            {
               current.setNewLine(true);
            }
            
            int contentLength = contentLength();
            parsedText.add(new Line(SPACE, contentLength, true));
            setCursorChar(0);
         }
         cursorDown();
         return;
      }
      
      super.placeLF();
   }

   /** 
    * Determine if a blank line needs to be inserted when RETURN key is pressed on a empty line.
    * 
    * @return   In ChUI, always <code>false</code>.
    */
   @Override
   protected boolean insertBlankLineOnLF()
   {
      return false;
   }

   /**
    * Move cursor one char right.
    */
   @Override
   protected void cursorRight()
   {
      if (config.cursorX < config.innerChars - 1)
      {
         config.cursorX++;
      }
      else
      {
         if (absoluteX() < config.bufferChars - 1)
            config.scrollX++;
      }
      
      super.cursorRight();
   }

   /**
    * Move cursor to the end of line.
    */
   @Override
   protected void cursorEndOfLine()
   {
      int len = getLine().length();
      
      if (config.scrollX < len && len < (config.scrollX + config.innerChars))
      {
         config.cursorX = len - config.scrollX;
      }
      else
      {
         if (len > config.innerChars)
         {
            config.cursorX = config.innerChars - 1;
            config.scrollX = len - config.cursorX;
         }
         else
         {
            config.cursorX = len;
            config.scrollX = 0;
         }
      }
      
      super.cursorEndOfLine();
   }

   /**
    * Move cursor one char left, if possible.
    */
   @Override
   protected void cursorLeft()
   {
      if (config.cursorX > 0)
      {
         config.cursorX--;
      }
      else
      {
         if (config.scrollX > 0)
            config.scrollX--;
      }

      super.cursorLeft();
   }

   /**
    * Calculate and set the cursor X position. If the position is not near the edge, then the
    * cursor will be positioned about at the middle of the screen.
    *
    * @param    pos
    *           New cursor position (0-based).
    */
   protected void setCursorX(int pos)
   {
      if (pos < config.scrollX || pos >= config.scrollX + config.innerChars)
      {
         // scrolling is necessary
         config.cursorX = (config.innerChars - 1)/2 - 1;
         config.scrollX = pos - config.cursorX;
         
         if (config.scrollX < 0 || (config.scrollX + config.innerChars) >= config.bufferChars)
         {
            if (config.scrollX < 0)
            {
               config.cursorX -= config.scrollX;
               config.scrollX = 0;
            }
            else
            {
               config.cursorX = config.innerChars - config.bufferChars + pos - 1;
               config.scrollX = pos - config.cursorX;
            }
         }
      }
      else
      {
         config.cursorX = pos - config.scrollX;
      }
      
      config.cursorChar = getCursorChar();
      config.cursorOffset = getCursorOffset();
   }

   /**
    * Split absolute horizontal cursor position into two components - position
    * of the first visible character and relative cursor position in the
    * visible region.
    * 
    * @param   pos
    *          The new absolute cursor position.
    */
   protected void cursorPos(int pos)
   {
      config.cursorX = pos % config.innerChars;
      
      super.cursorPos(pos);
   }

   /**
    * Get the focus listener instance.
    * 
    * @return   See above.
    */
   @Override
   protected FocusListener getFocusListener()
   {
      return new LocalChuiFocusListener();
   }
   
   /**
    * Get the next split position of this line, so that it fits the editor.
    * 
    * @param    line
    *           The line to be split.
    * @param    splitAfter
    *            minimal split position.
    *           
    * @return   An index where the split ends.
    */
   @Override
   protected int getNextSplitPosition(String line, int splitAfter)
   {
      int pos = line.lastIndexOf(SPACE, config.bufferChars - 1);
      if (pos < splitAfter || pos == line.length() - 1)
      {
         pos = config.bufferChars - 1;
      }
      pos++;
      
      return pos;
   }

   /**
    * Instructs the widget that its size may have changed. The widget updates
    * its state from the associated config instance. It is the responsibility
    * of the caller to call this method only after the config size is set with
    * the new values.
    * <p>
    * This implementation ignores the supplied config instance. Instead it relies
    * on {@link #size} and {@link #config()} to detect whether size has changed.
    * 
    * @param    beforeUpdate
    *           {@link BaseConfig} reference or <code>null</code>.
    */
   @Override
   protected <C extends WidgetConfig> void updateSizeWorker(C beforeUpdate)
   {
      super.updateSizeWorker(beforeUpdate);

      if (config.widthChars == 0 && config.heightChars == 0)
      {
         setSize(config.innerChars, config.innerLines);
      }
   }

   /**
    * Processing for unusual EOL characters in ChUI mode.  Currently removes '\r'.
    * 
    * @param   src
    *          the string to be processed.
    *    
    * @return  the string with only '\n' characters as line seperator(EOL).
    */
   @Override
   protected String processEOLChars(String src)
   {
      // the EOL to ignore in ChUI
      final String STRING_CR = "\r";
      // the empty substitution
      final String EMPTY = "";
      
      // simple check to void NPE
      if (src == null || src.isEmpty())
      {
         return src;
      }
      
      // removes '\r' symbols
      String result = src.replace(STRING_CR, EMPTY);
      
      return result;
   }
 
   /**
    * Move cursor one word left.
    */
   private void wordLeft()
   {
      if (absoluteX() == 0)
      {
         if (absoluteY() > 0)
            cursorEndOfPrevLine();
      }
      else
      {
         char[] text = getLine().getText().toCharArray();
         
         if (absoluteX() > text.length)
            cursorEndOfLine();
         
         cursorLeft();
         
         if (absoluteX() < text.length)
            while (absoluteX() >0 && text[absoluteX()] == ' ')
               cursorLeft();
         
         if (absoluteX() == 0)
         {
            if (absoluteY() > 0)
               cursorEndOfPrevLine();
         }
         else
         {
            while (absoluteX() > 0 && text[absoluteX() - 1] != ' ')
               cursorLeft();
         }
      }
   }

   /**
    * Move cursor one word right.
    */
   private void wordRight()
   {
      if (absoluteX() >= getLine().length())
      {
         if (absoluteY() < lastRow())
         {
            cursorDown();
            cursorBeginOfLine();
         }
      }
      else
      {
         char[] text = getLine().getText().toCharArray();

         cursorRight();
         
         while (absoluteX() < text.length && text[absoluteX()] != ' ')
            cursorRight();

         while (absoluteX() < text.length && text[absoluteX()] == ' ')
            cursorRight();
      }
   }

   /**
    * Get length of the first word in the next row.
    * 
    * @return  The length of the first word in the row immediately under the
    *          current.
    */
   private int getNextWordSize()
   {
      if (absoluteY() >= lastRow())
         return config.bufferChars;
      
      int sz = 0;
      
      char[] text = parsedText.get(absoluteY() + 1).getText().toCharArray();
      
      while (sz < text.length && text[sz] == ' ')
         sz++;
      
      while (sz < text.length && text[sz] != ' ')
         sz++;
      
      return sz;
   }

   /**
    * Handle EDITOR-BACKTAB event. 
    */
   private void editorBackTab()
   {
      if (absoluteX() == 0)
      {
         bell();
         return;
      }
      
      if (absoluteX() < EDITOR_TAB_SIZE)
      {
         cursorBeginOfLine();
         return;
      }
      
      int pos = absoluteX()/EDITOR_TAB_SIZE - 1;
      pos *= EDITOR_TAB_SIZE;
      
      setCursorX(pos);
   }

   /**
    * Handle EDITOR-TAB event.
    */
   private void editorTab()
   {
      if (insert)
      {
         // insert spaces
         Line current = getLine();
         if (current.length() + EDITOR_TAB_SIZE <= 
             config.bufferChars)
         {
            if (config.readOnly)
            {
               bell();
               return;
            }
            
            ConfigHelper.setModified(config, true);
            
            int pos = (absoluteX()/EDITOR_TAB_SIZE) + 1;
            pos *= EDITOR_TAB_SIZE;

            int contentLength = contentLength();
            int len = pos - absoluteX();
            
            if (absoluteX() > contentLength)
            {
               len += absoluteX() - current.length();
            }
            
            String spaces = StringHelper.repeatChar(' ', len);
            StringBuffer line = new StringBuffer(current.getText());
            line.insert(absoluteX(), spaces);
            parsedText.set(absoluteY(), 
                           new Line(line.toString(), current.getOffset(), current.isNewLine()));
            
            updateOffset();
            setCursorX(pos);
         }
         else
            bell();
      }
      else
      {
         // move cursor
         if (absoluteX() + EDITOR_TAB_SIZE <= 
             config.bufferChars)
         {
            int pos = (absoluteX()/EDITOR_TAB_SIZE) + 1;
            pos *= EDITOR_TAB_SIZE;
   
            setCursorX(pos);
         }
         else
            bell();
      }
   }

   /**
    * Handle APPEND-LINE event.
    */
   private void appendLine()
   {
      if (absoluteY() < lastRow() && (getLine().length() + getNextWordSize() < config.bufferChars))
      {
         cursorEndOfLine();
         placeChar(' ');
         deleteCharacter();
      }
   }

   /**
    * Returns the insert mode state.
    * 
    * @return  The insert mode state.
    *          <ul>
    *             <li><code>false</code>- overwrite mode,
    *             <li><code>true</code>- insert mode.
    *          </ul>
    */
   private boolean getInsertMode()
   {
      return Window.getInsertMode() == Window.INSERT_MODE_INS;
   }

   /**
    * Insert empty line above current line. 
    */
   private void openLineAbove()
   {
      if (config.readOnly)
      {
         bell();
         return;
      }
      
      ConfigHelper.setModified(config, true);
   
      parsedText.add(absoluteY(), new Line("", getLine().getOffset(), true));
      config.length = getLength();
      config.numLines = parsedText.size();
   }

   /**
    * Prepare text for display.
    * 
    * @param    uninit
    *           <code>true</code> if the widget is uninitialized.
    * @param    width
    *           Width of the displayed area.
    * @param    space
    *           Prepared string filled with spaces.
    * @param    line
    *           Index in the array of parsed strings.
    *           
    * @return   Text string ready for displaying.
    */
   private String prepareTextRow(boolean uninit, int width, String space, int line)
   {
      String text;
      
      if (line < parsedText.size())
      {
         if (parsedText.get(line).length() < config.scrollX || uninit)
         {
            text = space;
         }
         else
         {
            text = parsedText.get(line).getText() + space;
            text = text.substring(config.scrollX, config.scrollX + width);
         }
      }
      else
      {
         text = space;
      }
      
      return text;
   }
   
   /**
    * Delete the line of text in which the cursor currently resides.  Leave the cursor at the
    * leftmost position of the same line number (which will be the line previously below this
    * one if the cursor is not already on the last line) or at the leftmost position of the
    * last line (if there are no lines below this one).
    *
    * @return   <code>true</code> if anything is deleted.
    */
   private boolean deleteLineWorker()
   {
      cursorBeginOfLine();
      
      // remember if this was the last line (must be here because the deletion and re-parse
      // of the text will change the number of lines if the deleted line was there because
      // of word wrapping instead of because of an explicit newline char
      boolean wasLast = (absoluteY() >= lastRow());
      
      // these first two statements delete all the content of the current line, except for the
      // newline character
      boolean resLine = deleteToEndOfLine();
      
      // at this point if we are on the last line, the last line is empty and there is no
      // following newline char to delete but instead we must remove the line, which requires
      // that we delete the newline character from the previous line                                                                                           
      if (wasLast)                                                      
      {
         cursorEndOfPrevLine();
      }
      
      boolean resChar = false;
      
      // if there is a newline, delete it
      if (getLine().isNewLine())
      {
         resChar = deleteCharacter();
      }
      
      // in the last line case we are at the end of the previous line which is now the new
      // last line, we must move to the beginning of the line
      if (wasLast)
      {
         cursorBeginOfLine();
      }
      
      // if neither of the above operations deleted anything, return false
      return (resLine || resChar);
   }  

   /**
    * Delete word. 
    */
   private void deleteWord()
   {
      if (config.readOnly)
      {
         bell();
         return;
      }
      
      ConfigHelper.setModified(config, true);

      if (absoluteX() < getLine().length())
      {
         Line current = getLine();
         char[] text = current.getText().toCharArray();
         
         int start = absoluteX();
         int end = start;
         
         // note that order of checks in both loops is important
         
         while (end < text.length && !Character.isLetterOrDigit(text[end]))
         {
            end++;
         }
         
         while (end < text.length && Character.isLetterOrDigit(text[end]))
         {
            end++;
         }
         
         if (end > start)
         {
            StringBuffer line = new StringBuffer(getLine().getText());
            line.delete(start, end);
            mergeLines(absoluteY());
            parsedText.set(absoluteY(), 
                  new Line(line.toString(), current.getOffset(), current.isNewLine()));
            updateOffset();
         }
      }
   }

   /**
    * Some actions on focus hanging.
    */
   protected class LocalChuiFocusListener
   extends LocalFocusListener
   {
      /**
       * Setups <code>FillIn</code> as it is just activated by Tab key.
       * 
       * @param   e
       *          FocusEvent instance (ignored).
       */
      @Override
      public void onFocusGained(FocusEvent e)
      {
         if (e == null)
            return;
       
         if (config.readOnly)
         {
            fakeInsert = true;
         }

         setInsertMode(insert);

         super.onFocusGained(e);
      }

      /**
       * Reset SELECTED attribute when focus is lost.
       * 
       * @param   e
       *          FocusEvent instance (ignored).
       */
      @Override
      public void onFocusLost(FocusEvent e)
      {
         if (e == null)
            return;
         
         super.onFocusLost(e);
         Window.setInsertMode(Window.INSERT_MODE_OFF);
      }
   }
}