/*
** Module   : FillInImpl.java
** Abstract : text field editor/display implementation, CHUI-specific version.
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SIY 20111120 Extracted from FillIn.
** 002 GES 20140627 Changed color attribute method names to match naming standards.
** 003 CA  20140805 Do not allow ChUI impl classes access to APIs outside of ChuiOutputManager.
** 004 HC  20140911 GUI coordinate support - data types refactoring.
** 005 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 006 CA  20141220 Refactoring to integrate the GUI FILL-IN implementation with the ChUI logic.
** 007 CA  20150323 Some more refactoring to integrate GUI FILL-IN drawing.
** 008 HC  20150323 Removed business logic from config classes.
** 009 CA  20150805 Fixed FILL-IN cursor problems when editable width is smaller than content
**                  length.
** 010 IAS 20150806 Added PASSWORD-FIELD attribute support.
** 011 IAS 20150924 Added AUTO-RESIZE attribute support.
** 012 GES 20170104 Moved support for PASSWORD-FIELD into the parent class.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.util.*;
import com.goldencode.util.*;

/**
 * Implementation of Progress FILL-IN component for CHUI.
 */
public class FillInImpl
extends FillIn<ChuiOutputManager,FillInConfig>
implements AutoResizable<FillInConfig>
{
   /** The content resolver for TEXT, LITERAL and FILL-IN widgets, in ChUI drivers. */
   private final ContentResolver resolver;
   
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public FillInImpl(WidgetId id, FillInConfig cfg)
   {
      super(id, cfg);
      
      this.resolver = new ContentResolver(this);
   }

   /**
    * Sets new widget's size. Resets AUTO-RESIZE attribute.
    * 
    * @param   width
    *          The width component of the new size.
    * @param   height
    *          The height component of the new size.  Will be forced to 1
    *          if <code>width</code> is positive and this value is less than
    *          or equal to 0.
    */
   @Override
   public void setSize(double width, double height)
   {
      width = Coordinate.scale(width);
      height = Coordinate.scale(height);
      
      // safe, comparison of scaled doubles
      if (width > 0 && height <= 0)
      {
         height = 1;
         config.heightChars = height; // direct access
      }
      
      super.setSize(width, height);
   }
   
   /**
    * Sets position of terminal's cursor which may or may not be the same as
    * the actual cursor location within the widget.  This can be due to 
    * scrolling or due to being a right-edge widget.  All widgets which have 
    * their right-edge up against the right-edge of the frame, will draw the 
    * cursor in the last position of the widget rather than allowing the
    * cursor to move further right.  Note that the drawing position of the
    * cursor does not change the actual recorded cursor position in the
    * class' state.
    */
   @Override
   public void drawCaret()
   {
      // this check is required since the method can be called from multiple
      // drivers
      if (!hasFocus() || !isEnabled())
      {
         return;
      }

      Point origin = screenLocation();
      
      // determine if we are right or left aligned
      boolean ralign = isRightAlignedInt();
      boolean edge   = (ralign ? isLeftEdgeWidget() : isRightEdgeWidget()); 
      
      // find the "real" cursor position
      int cursor = getCursorOffset();
      int max    = config.appFormat.getScreenWidth();
      
      // input has taken us past the last char of the field 
      if (edge && cursor >= max)
      {
         // the last widget is a special case where the cursor is
         // kept inside the field (normally the cursor is allowed to
         // move further)
         cursor = max - 1;
      }
      
      // account for left padding in cursor position if we are right aligned OR
      // if this is a left aligned field with added padding on the left
      // due to an optional sign digit that was blanked out
      int target = ((ralign || resolver.leftPad != -1) ? (cursor + resolver.leftPad) : cursor);
      
      // cleared number formats that are right aligned have cursor on right
      if (ralign && cursor == 0 && resolver.cleared && justActivated)
      {
         target = max - 1;
      }
      
      // in chui char coordinates are always whole numbers
      screen().at((int) origin.x + target - editScroller.getOffset(), (int) origin.y);
   }
   
   /**
    * Actual drawing routine.
    *
    * @param   color
    *          Color for drawing.
    */
   @Override
   public void draw(Color color)
   {
      Color dc = UiUtils.getDColor(this);

      selected = !Utils.equals(color, dc);

      String display = getDrawText(' ');
      
      // obtain the initialization status
      boolean uninit = (getState() == ScreenBuffer.UNINITIALIZED);
      
      // If uninitialized and not for input and not date fill-in (in this case
      // separators should be displayed), display blanks instead.
      // TODO: probably it should be replaced by variable initialization
      //       in FrameHelper.clearComponents().
      if (uninit &&
          !(config.appFormat != null &&
            config.appFormat instanceof DateFormat &&
            isEnabled()))
      {
         int width = display.length();
         display = StringHelper.repeatChar(' ', width);
      }
      
      // this needs to be saved here, before the painter does any adjustments
      Point p = screenLocation();

      // in chui char coordinates are always whole numbers
      if ((int) p.x == -1 || (int) p.y == -1)
         return;

      ChuiOutputManager screen = screen();
      NativePoint np = screen.coordinates().pointToNative(p);
      NativePoint loc = editScroller.adjustLocation(display, np);
      // the text is already adjusted to fit the content.
      screen.at(loc).append(display, color);

      // in chui char coordinates are always whole numbers
      if ((int) height() > 1)
      {
         String spaces = DisplayFormat.generateSpaces((int) width());
         
         for (int i = 1; i < (int) height(); ++i)
         {
            p.y++;
            screen.at(p);
            
            if (isHeader() && drawLine && i == (int) (height() - 1))
               screen.hLine(width(), color);
            else
               screen.append(spaces, Color.DEFAULT);
         }
      }

      if (hasFocus() && isEnabled())
         drawCaret();
   }

   /**
    * Compute the initial cursor position, depending on the widget's implementation.
    * <p>
    * In ChUI, the initial cursor position is always on the left-side text edge.
    *  
    * @return   See above.
    */
   @Override
   public int getInitialCursorPos()
   {
      return 0;
   }
   
   /**
    * Main worker routine for the <code>FillIn</code> text accessor.
    * <p>
    * It delegates the work to the associated {@link ContentResolver feeder} instance.
    * 
    * @param   isInternal
    *          If this flag is set to <code>true</code> then alternative
    *          variable value is used when available.
    *          
    * @return  Text representation of the variable value according to all
    *          current settings, flags, format, etc.
    */
   public String getText(boolean isInternal)
   {
      return resolver.getText(isInternal);
   }
   
   /**
    * Allow the widget to adjust the cursor position, after a key was pressed.
    * <p>
    * For FILL-INs, this will adjust the anchor on which the text is scrolling, if the cursor is 
    * set on the left or right edge.
    * <p>
    * This implementation will keep the cursor on the rightmost char, if the right edge is reached.
    *  
    * @param    key
    *           The pressed key.
    * @param    pos
    *           The cursor position computed by the display format.
    * @param    max
    *           The maximum cursor position.
    *           
    * @return   The adjusted position.
    */
   @Override
   public int adjustCursor(int key, int pos, int max)
   {
      pos = super.adjustCursor(key, pos, max);
     
      return pos == max + 1 ? max : pos;
   }

   /**
    * Get the AUTO-RESIZE status
    * 
    * @return <code>true</code> is AUTO-RESIZE is enabled
    */
   @Override
   public boolean isAutoResize()
   {
      return config.autoResize;
   }

   /**
    * Update widget's size after config update
    * 
    * @param beforeUpdate
    *        previous value of the widget's config
    */
   @Override
   public void updateSize(FillInConfig beforeUpdate)
   {
      if (beforeUpdate == null)
      {
         return;
      }
      if (!StringHelper.equal(config.format, beforeUpdate.format))
      {
         updateSize();
      }
   }

   /**
    * Get the cursor position, in native units.
    * 
    * @return   See above.
    */
   @Override
   protected int getNativeCursorPos()
   {
      return getValidCursorPos();
   }

   /**
    * Create the left-edge scroller.
    * 
    * @return   See above.
    */
   @Override
   protected EditScroller createLeftAnchor()
   {
      return new LeftAnchorEditScroller();
   }

   /**
    * Create the right-edge scroller.
    * 
    * @return   See above.
    */
   @Override
   protected EditScroller createRightAnchor()
   {
      return new RightAnchorEditScroller();
   }

   /**
    * Determine the portion of the widget's editable part.  This may be different than the widget's
    * width, as determined by the WIDTH-CHARS or WIDTH-PIXELS attribute.
    * <p>
    * For ChUI widgets, this is the same as the widget's physical dimension.
    * 
    * @return   The editable dimension of this widget, in native units.
    */
   @Override
   protected NativeDimension editableNativeDimension()
   {
      return physicalDimension();
   }
   
   /**
    * Retrieve the widget's width in native units.
    * 
    * @return   See above.
    */
   @Override
   protected int nativeWidth()
   {
      return isAtField() || config.appFormat == null ? 
            config.atFormatLength : config.appFormat.getScreenWidth();
   }

   /**
    * Retrieve the widget's height in native units.
    * 
    * @return   See above.
    */
   @Override
   protected int nativeHeight()
   {
      return 1;
   }

   /**
    * Content scroller implementation when the anchor is positioned on the left-side editable edge.
    */
   private class LeftAnchorEditScroller
   extends EditScroller
   {
      /**
       * Adjust the location so that only the visible content will be displayed.
       *  
       * @param    txt
       *           The full content.
       * @param    np
       *           The current location of the content.
       *           
       * @return   The adjusted location.
       */
      @Override
      public NativePoint adjustLocation(String txt, NativePoint np)
      {
         int delta = isRightAlignedInt() ? (int) width() - txt.length() : 0;

         return new NativePoint(np.x + Math.max(delta, 0), np.y);
      }

      /**
       * Get the amount of characters which are over the left-most edge, and are not being displayed.
       * 
       * @return   See above.
       */
      @Override
      public int getDelta()
      {
         return -anchorIndex;
      }

      /**
       * Get the index of the leftmost visible character in the widget's displayable area.
       * 
       * @return   See above.
       */
      @Override
      public int getOffset()
      {
         return anchorIndex;
      }
   }

   /**
    * Content scroller implementation when the anchor is positioned on the right-side editable edge.
    */
   private class RightAnchorEditScroller
   extends EditScroller
   {
      /**
       * Adjust the location so that only the visible content will be displayed.
       *  
       * @param    txt
       *           The full content.
       * @param    np
       *           The current location of the content.
       *           
       * @return   The adjusted location.
       */
      @Override
      public NativePoint adjustLocation(String txt, NativePoint np)
      {
         int delta = isRightAlignedInt() ? (int) width() - txt.length() : getDelta();
         return new NativePoint(np.x + Math.max(delta, 0), np.y);
      }

      /**
       * Get the amount of characters which are over the left-most edge, and are not being displayed.
       * 
       * @return   See above.
       */
      @Override
      public int getDelta()
      {
         int leftPad = (int) width() - anchorIndex - 1;

         return leftPad;
      }
      
      /**
       * Get the index of the leftmost visible character in the widget's displayable area.
       * 
       * @return   See above.
       */
      @Override
      public int getOffset()
      {
         return Math.max(anchorIndex - (int) width() + 1, 0);
      }
   }
}