/*
** Module   : LabelImpl.java
** Abstract :
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SIY 20111126 Extracted from Label
** 002 SIY 20111221 Fixed constructor visibility.
** 003 GES 20140627 Changed color attribute method names to match naming standards.
** 004 CA  20140805 Do not allow ChUI impl classes access to APIs outside of ChuiOutputManager.
** 005 HC  20140911 GUI coordinate support - data types refactoring.
** 006 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 007 CA  20141220 Removed drawCaret, as is a no-op for LABEL widget.
** 008 CA  20150222 Fixed label's COLUMN and WIDTH-CHARS attributes (in ChUI).
** 009 CA  20150323 Pass the frame ID when creating a label (to properly resolve the font in GUI).
** 010 CA  20150406 Added APIs to compute the label's width/height in native units.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.format.*;

/**
 * CHUI-specific portion of Label.
 */
@SuppressWarnings("unchecked")
public class LabelImpl
extends Label<ChuiOutputManager>
{
   /**
    * Constructs label with given text.
    * 
    * @param   id
    *          The widget's ID.
    * @param   text
    *          Initial text.
    * @param   frameId
    *          The ID of the frame to which the label belongs, or -1 if unknown at the time of the 
    *          creation.
    */
   LabelImpl(WidgetId id, String text, int frameId)
   {
      super(id, text, frameId);
   }

   /**
    * Create a copy of the label.
    * 
    * @param   id
    *          The widget's ID.
    * @param   label
    *          Label to copy parameters from.
    * @param   frameId
    *          The ID of the frame to which the label belongs, or -1 if unknown at the time of the 
    *          creation.
    */
   LabelImpl(WidgetId id, Label<ChuiOutputManager> label, int frameId)
   {
      super(id, label, frameId);
   }
   
   /**
    * Retrieve the widget's width in native units.
    * 
    * @return   See above.
    */
   @Override
   protected int nativeWidth()
   {
      String lbl = name();

      String[] values = lbl.split("\n");
      int width = 0;

      for (int i = 0; i < values.length; i++)
      {
         if (width < values[i].length())
            width = values[i].length();
      }

      return width;
   }
   
   /**
    * Retrieve the widget's height in native units.
    * 
    * @return   See above.
    */
   @Override
   protected int nativeHeight()
   {
      String lbl = name();

      String[] values = lbl.split("\n");
      int height = values.length;

      if (underlined)
         height++;

      return height;
   }
   
   /**
    * Draws text. Screen value is left justified or padded to the right bound
    * with spaces. Non first lines are filled with spaces.
    */
   @Override
   public void draw()
   {
      if (parent() == null)
         return;
      
      Dimension actualSize = dimension();
      Point origin = screenLocation();
      ChuiOutputManager screen = screen();

      Color clr = UiUtils.getDColor(this);

      String[] values = name().split("\n");
      String spaces = DisplayFormat.generateSpaces((int) actualSize.width);
      
      screen.at(origin);

      // in chui char coordinates are always whole numbers
      for (int i = 0; i < (int) actualSize.height; i++, origin.y++)
      {
         String screenValue;
         
         if (i < values.length && !emptyMode)
            screenValue = DisplayFormat.formatScreenValue(values[i],
                                                          (int) actualSize.width,
                                                          0,
                                                          false,
                                                          rightAligned);
         else
            screenValue = spaces;

         if (i == (int) actualSize.height - 1 && underlined)
         {
            // draw the separator line
            screen.at(origin).hLine(actualSize.width, clr);
         }
         else
         {
            screen.at(origin.offset(offset, 0)).append(screenValue, clr);
         }
      }
      
      if (disableEmptyMode)
         emptyMode = false;
      
      disableEmptyMode = false;
   }
}