/*
** Module   : RectangleImpl.java
** Abstract : chui rectangle implementation.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140410 First release.
** 002 GES 20140627 Changed color attribute method names to match naming standards.
** 003 CA  20140805 Do not allow ChUI impl classes access to APIs outside of ChuiOutputManager.
** 004 HC  20140911 GUI coordinate support - data types refactoring.
** 005 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 006 EVL 20141106 Changing approach for setting NO-FILL option.  Now the member filled is used.
** 007 HC  20150323 Removed business logic from config classes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Specific implementation of RECTANGLE widget.
 */
public class RectangleImpl
extends Rect<ChuiOutputManager>
{
   /**
    * This constructor creates a rectangle.
    * 
    * @param   id
    *          The widget's ID.
    * @param   owner
    *          The parent container, frame which will include this
    *          rectangle.
    */
   RectangleImpl(WidgetId id, Container<ChuiOutputManager> owner) 
   {
      super(id, owner);
   }

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public RectangleImpl(WidgetId id, RectangleConfig cfg)
   {
      super(id, cfg);
   }
   
   /**
    * Draw the rectangle. Called by this rectangle's parent container.
    */
   @Override
   public void draw()
   {
      // rectangle options
      RectangleConfig config = (RectangleConfig) config();
      
      // screen manager
      ChuiOutputManager screen = screen();
      
      // get the absolute origin of this component
      Point origin = screenLocation();
               
      // get rectangle dimension
      Dimension dimension = dimension();

      // move to rectangle anchor
      screen.at(origin);
      
      // get parent color
      Color parentColor =  UiUtils.getDColor(null, this);
      
      // compute fill color
      Color fillColor = config.dcolor;
      if (fillColor == null)
      {
         fillColor = parentColor;
      }

      // compute foreground color
      Color edgeColor = config.pfcolor;
      if (edgeColor == null || Color.INPUT.equals(edgeColor))
      {
         edgeColor = parentColor.withReverse();
      }

      // border size
      int borderSize = (int) config.edgeChars;
      
      if (borderSize > 0)
      {        
         if (config.graphicEdge)
         {
            // horizontal line
            if (dimension.width > 1 && dimension.height == 1)
            {
               screen.hLine(dimension.width, edgeColor);   
               borderSize = -1;
            }
            // vertical line
            else if (dimension.width == 1 && dimension.height > 1)
            {
               screen.vLine(dimension.height, edgeColor);
               borderSize = -1;
            }
            // box
            else if (dimension.width > 1 && dimension.height > 1)
            {
               screen.box(origin, dimension, edgeColor);
               borderSize = 1;
            }
         }
         else
         {
            drawRectangleBorder(screen, origin, dimension, borderSize, edgeColor);
         }
      }

      // fill rectangle
      if (config.filled && borderSize >= 0)
      {
         fillRectangle(screen, origin, dimension, borderSize, fillColor);
      }
   }

   /**
    * Draw rectangle character border.
    * 
    * @param   screen
    *          Screen manager.
    * @param   origin
    *          Rectangle origin.
    * @param   dimension
    *          Rectangle size.
    * @param   borderSize
    *          Border size characters.
    * @param   color
    *          Border color.
    */
   private void drawRectangleBorder(ChuiOutputManager screen, 
                                    Point origin, 
                                    Dimension dimension, 
                                    int borderSize,
                                    Color color)
   {
      // in chui char coordinates are always whole numbers
      int cols = (int) dimension.width;
      int rows = (int) dimension.height;
      
      // max border values
      int maxRow = rows - borderSize;
      int maxCol = cols - borderSize;
            
      // draw rectangle border
      for (int row = 0; row < rows; row++)
      {
         for (int col = 0; col < cols; col++)
         {
            if (isBorder(row, borderSize, maxRow) || isBorder(col, borderSize, maxCol))
            {
               drawRectangleCell(screen, (int) origin.x + col, (int) origin.y + row, ' ', color);
            }
         }                     
      }      
   }

   /**
    * Check if the value is inside the border area.
    * 
    * @param   value
    *          Current value.
    * @param   minValue
    *          Minimum value.
    * @param   maxValue
    *          Maximum value.
    *          
    * @return  <code>true</code> if is inside the border area 
    *          <code>false</code> otherwise.
    */
   private boolean isBorder(int value, int minValue, int maxValue)
   {
      return !(value >= minValue && value < maxValue);
   }
   
   /**
    * Fill rectangle.
    * 
    * @param   screen
    *          Screen manager.
    * @param   origin
    *          Rectangle origin.
    * @param   dimension
    *          Rectangle size.
    * @param   borderSize
    *          Border size characters.
    * @param   color
    *          Fill color.
    */
   private void fillRectangle(ChuiOutputManager screen, 
                              Point origin, 
                              Dimension dimension, 
                              int borderSize,
                              Color color)
   {
      int cols = (int) dimension.width - borderSize;
      int rows = (int) dimension.height - borderSize;
      
      // fill rectangle
      for (int row = borderSize; row < rows; row++)
      {
         for (int col = borderSize; col < cols; col++)
         {
            // in chui char coordinates are always whole numbers
            drawRectangleCell(screen, (int) origin.x + col, (int) origin.y + row, ' ', color);
         }                     
      }      
   }

   /**
    * Draw rectangle cell.
    * 
    * @param   screen
    *          Screen manager.
    * @param   x
    *          Column position.
    * @param   y
    *          Row position.
    * @param   character
    *          Character to write
    * @param   color
    *          Cell color.
    */
   private void drawRectangleCell(ChuiOutputManager screen, 
                                  int x, 
                                  int y, 
                                  int character, 
                                  Color color)
   {
      screen.at(x, y).append(character, color);      
   }
}