/*
** Module   : TerminalBuffer.java
** Abstract : Terminal output driver that does not do output at all
**            and just collects all updates as if it was real terminal screen.
**            Initial version is derived from DoubleBufferedTerminal.java.
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SIY 20090305   @41537 Initial implementation.
** 002 SIY 20090319   @41645 Changed addArray() signature. Now characters
**                           for buffer are taken from provided array
**                           rather than from encoded data.
** 003 SIY 20090911   @43894 Now class is inherited from BasePrimitives.
** 004 SIY 20100619          Switching to own implementation of character UI.
** 005 SIY 20100804          Changes in imports, some refactorings.
** 006 SIY 20110112          Changes related to new color handling scheme.
** 007 HC  20140911          GUI coordinate support - data types refactoring.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;

/**
 * Minimal possible driver which implements all output primitives using
 * memory buffer as destination.
 */
public class TerminalBuffer
extends AbstractTerminalBuffer
{
   /**
    * Constructor.
    *
    * @param    rows
    *           Number of rows in buffer.
    * @param    cols
    *           Number of columns in buffer.
    */
   public TerminalBuffer(int rows, int cols)
   {
      super(rows, cols);
   }

   /**
    * Add a character to the virtual terminal at the current cursor
    * position.
    *
    * @param    chr
    *           The character to output.
    * @param    color
    *           Color.
    */
   @Override
   public void append(int chr, Color color)
   {
      render()[cy][cx] =  new Cell((char)chr, color);

      advanceCursor();
   }

   /**
    * Mark a rectangle as needing repainting.
    *
    * @param    rect
    *           rectangle to mark for repainting in screen coordinates
    */
   @Override
   public void refresh(NativeRectangle rect)
   {
   }

   /**
    * Synchronize the state of the physical terminal with the state of
    * the virtual terminal.
    */
   @Override
   public void sync()
   {
   }

   /**
    * Get character at specified position.
    *
    * @param    row
    *           Character row.
    * @param    col
    *           Character column.
    *
    * @return   character from specified position or space if operation is not
    *           supported.
    */
   @Override
   public Cell cellAt(int row, int col)
   {
      if (row >= rows || col >= cols)
         return super.cellAt(row, col);

      return render()[row][col];
   }

   /**
    * Perform a bulk output operation for the whole screen at once with
    * output masking when desired. Set the cursor afterwards.
    *
    * @param   screen
    *          Container for whole screen content and cursor position.
    */
   @Override
   public void updateScreen(ScreenData screen)
   {
      for (int r = 0; r < rows; r++)
      {
         for (int c = 0; c < cols; c++)
         {
            Cell cell = screen.getCell(r, c);

            if (cell != null)
               render()[r][c] = cell;
         }
      }
   }
}