/*
** Module   : DropDown.java
** Abstract : Drop down box for combo-box implementation.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100804 Created initial version
** 002 SIY 20100822 Refactored AttributeContainer support.
** 003 SIY 20100915 locateFrame() is moved to UiUtils.
** 004 SIY 20111101 Updated imports.
** 005 SIY 20111202 Refactored and moved from CHUI package.
** 006 SIY 20111221 Switched to widget factory.
** 007 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
** 008 HC  20140911 GUI coordinate support - data types refactoring.
** 009 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 010 HC  20141021 Implemented window move/position support.
** 011 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 012 CA  20150212 Allow events to be captured and processed by the widgets without server-side
**                  or main event-loop notification.  This fixes problems in the EDITING and
**                  WAIT-FOR event loops, for i.e. COMBO-BOX and EDITOR widgets.
** 013 IAS 20150214 Added getSelectedIndex() method
** 014 HC  20150323 Removed business logic from config classes.
** 015 EVL 20150427 Separation GUI related code in standalone class.  Making common methods to be
**                  public.
** 016 EVL 20150514 Adding method to adjust the visible rows value according to UI type used.
** 017 HC  20150831 Scrolling support overhaul.
** 018 HC  20151028 Fixed drop-down max-height calculation.
**         20151029 Fixed drop-down scroll list width calculation, it must not be smaller than
**                  the parent's width.
**         20151110 GUI combo-box fixes.
** 019 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 020 IAS 20160229 Change scrollpane creation.
** 021 HC  20160406 Fixed COMBO-BOX drop-down resource cleanup.
** 022 HC  20160922 Added getLegacyWidget() method.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.model.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Implements the drop-down window for a combo-box.
 * TODO &lt;SIY&gt;: this class is not managed by layout manager, review it and make
 * sure that it conforms GUI requirement, i.e. properly maintains its physical 
 * position/size. 
 */
public abstract class DropDown<O extends OutputManager<?>>
extends OuterFrame<O>
{
   /** The list instance to represent the drop-down content. */
   protected final ScrollableList<O> list;
   
   /** Wraps the list of items and handles scrolling. */
   protected ScrollPane<O> scrollpane;

   /** Reference to parent combo-box. */
   private final ComboBox<O> comboBox;

   /**
    * Create a drop-down container based on the parent combo-box and it's
    * associated data model.
    * 
    * @param    parent
    *           The combo-box object which calls this drop-down.
    */
   @SuppressWarnings("unchecked")
   public DropDown(ComboBox<O> parent)
   {
      super("");
      this.comboBox = parent;
      list = screen().getFactory().createScrollableList(parent, parent.model());

      // enable scrolling, ensure there is no extra border or spacing
      scrollpane = screen().getFactory().createScrollPane(list, this::topLevelWindow);
      super.add(scrollpane);

      // TODO: GUI combo-box layout needs improvement.
      //       The list and drop down width are calculated
      //       separately, and so any differences must be fixed
      //       in the other parts of the scattered layout logic, 
      //       see ScrollableSelectionListGuiImpl.width().
      
      // set height, width and position of the drop-down itself
      list.setColumns(parent.calcLimit(false) + 2);
      setHeightAndPos(parent, parent.model());
      
      // force the initial selection
      forceInitialSelection(parent.model());
      // set up notifications
      list.addSelectionListener(parent);

      OutputManager.instance().getRegistry().addWidget((Widget) scrollpane);
   }

   /**
    * Get the combo-box visible row adjust depending on the current UI type in use.
    * 
    * @param    visibleRows
    *           The visible rows value to check and modify.
    * @param    listSize
    *           The total size of the items in the list to handle.
    * 
    * @return  the new visible rows value.
    */
   protected abstract int fixVisibleRows(int visibleRows, int listSize);

   /**
    * Enforce the current selection.
    *
    * @param    model
    *           The data model to be used within this drop-down list.
    */
   public void forceInitialSelection(ComboBoxModel<String> model) 
   {
      int idx = model.calcSelection(comboBox.config().items);
      
      list.select(idx);
      // set up initial top row
      list.setTopRow(idx);
   }
   
   /**
    * Get the drop-down's scroll pane.
    * 
    * @return   See above.
    */
   public ScrollPane<O> getScrollPane()
   {
      return scrollpane;
   }
   
   /**
    * Get the drop-down's inner list.
    * 
    * @return   See above.
    */
   public ScrollableList<O> getList()
   {
      return list;
   }
   
   /**
    * Get the selected object from the drop-down list.
    * 
    * @return   The object which is currently selected in drop-down list
    *           or <code>null</code> if there is no selection.
    */
   public String getSelectedItem()
   {
      return list.value();
   }
   
   /**
    * Get the index of the selected object from the drop-down list.
    * 
    * @return   The index of the object which is currently selected in drop-down list
    */
   public int getSelectedIndex() 
   {
      return list.index();
   }
   
   /**
    * Calculate (and set) the height and position of the drop-down
    * container based on the parent combo-box and it's associated data 
    * model.
    * <p>
    * Precedence order (in the following list the term "max-space" is
    * the number of rows from the top of the containing combo-box's
    * frame to the bottom of the screen) for vertical size:
    * <p>
    * <pre>
    * 1. If there is an inner-lines definition AND it is smaller than the 
    *    max-space, then it is honored.
    * 2. If the number of elements in the data model is less than the
    *    number of rows in the max-space, then the number of data model
    *    elements will determine the vertical size.
    * 3. max-space is used otherwise.
    * </pre>
    * <p>
    * The drop-down box X coordinate is always the same as the X
    * coordinate of the parent combo-box in absolute screen units. For
    * the Y coordinate, the following positioning algorithm is used:
    * <p>
    * <pre>
    * 1. If the height of the drop-down is less than or equal to the space
    *    between the bottom of the combo-box and the bottom of the screen,
    *    then the drop-down T coordinate will be set to the combo-box Y +
    *    1.
    * 2. Otherwise, the drop-down Y coordinate will equal the Y coordinate
    *    of the bottom of the screen minus the height of the drop-down.
    *    This will result in the drop-down ending at the bottom of the
    *    screen and starting at or above the combo-box Y coordinate BUT
    *    never above the top of the frame containing the combo-box
    *    (because of the previously constrained height calculation above).
    * </pre>
    * 
    * @param    parent
    *           The combo-box object which calls this drop-down.
    * @param    model
    *           The data model to be used within this drop-down list.
    */
   public void setHeightAndPos(ComboBox<O> parent, ComboBoxModel<String> model)
   {
      Dimension screenSize = parent.screenDimension();

      @SuppressWarnings("unchecked")
      Frame<O> frame = (Frame<O>) UiUtils.locateFrame(parent);
      
      // calculate the maximum size available for the drop-down (the space
      // from the row just below the top of the containing frame to the 
      // bottom of the screen reserving 2 lines for the drop-down top and
      // bottom border)
      // TODO: for some reason we don't use value set by 
      //       ComboBoxConfig.setMaximumRowCount()!!!
      Point frameLoc = frame.screenLocation();
      
      CoordinatesConversion cc = screen().coordinates();
      double maxSize  = screenSize.height - frameLoc.y - cc.heightFromNative(1) - 
                        cc.heightFromNative(2);
      
      ComboBoxConfig config = parent.config();
      
      // query inner lines (explicit size request)
      int rows = config.innerLines;
      
      // override inner lines with the number of elements if not specified
      if (rows <= 0)
      {
         int numItems = model.size();
         
         // ensure that we always have at least 1 row
         rows = (numItems > 0) ? numItems : 1;
      }
      
      // if the inner lines spec or the number of elements are more than
      // the maximum space, reset to the maximum size
      if (rows > (int) maxSize)
      {
         rows = (int) maxSize;
      }
      
      // differnt UI types take differen approach here
      rows = fixVisibleRows(rows, model.size());
      
      int frameHeight = rows + 2;
      
      // set the frame height
      setHeight(frameHeight);
      
      // set the list height
      list.setVisibleRows(rows);
      
      setSize(minimumSize());
      doLayout();
      
      // get the absolute origin of the parent combo-box
      Point origin = parent.screenLocation();
      
      // calc default Y coordinate for the drop-down
      double y = origin.y + 1; 
      
      // do we have enough room below the combo-box to hold the entire
      // drop-down?
      if (Coordinate.scale(screenSize.height - y) >= frameHeight)
      {
         // yes, use the default
         origin.y = y;
      }
      else
      {
         // no, find the location above the default where we must position
         // the drop-down
         origin.y = screenSize.height - frameHeight;
      }
      
      // set the drop-down position
      setLocation(origin);
   }
   
   /**
    * Get the combo-box widget to which this drop-down belongs.
    * 
    * @return  the combo-box widget to which this drop-down is attached.
    */
   public ComboBox<O> getComboBox()
   {
      return comboBox;
   }

   /**
    * Hide this widget and remove it.
    */
   @Override
   public void hide()
   {
      super.hide();
      
      // this is a special case: if the drop-down is hidden, then its no longer available.
      WindowManager.remove(this);
   }

   /**
    * The method returns the legacy widget representing this widget.
    *
    * @return  A non-wirtual widget reference, or <code>null</code> if none found.
    */
   @Override
   public Widget<O> getLegacyWidget()
   {
      return getComboBox();
   }
}