/*
** Module   : FontDetails.java
** Abstract : a container for a font, font metrics and related state
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 GES 20150427 Created initial version based on the inner class of the FontManager.
** 002 CA  20150520 Made the class serializable, so that it can be used as a container for the 
**                  font specification, when transferring the font-table from the server to the
**                  client side.  In this case, the "font" field is not used (and it was marked
**                  transient for this reason).
** 003 CA  20150810 Added missing javadoc for "fontName".
** 004 CA  20150808 Added support for font aliases.
** 005 SBI 20160107 Added boldFontDefinition to detect the case if the font has a separate font
**                  definition for the bold style.
** 006 HC  20170104 Added toString().
** 007 EVL 20170613 Adding anti-aliasing hint flag for Swing client to use or not the mode.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.io.*;

/**
 * A container for a font, font metrics and related state.
 */
public class FontDetails<F>
implements Serializable
{
   /** A unique descriptor describing the font name, size and attributes. */
   public String key;
   
   /** The font's name. */
   public String fontName;
   
   /** The font object instance. */
   public transient F font;
   
   /** Font style. */
   public FontStyle style;
   
   /** Point size for this font. */
   public int pointSize;
   
   /** The legacy width of this font; negative value if this was not configured. */
   public int legacyWidth = -1;
   
   /** The legacy max width of this font; negative value if this was not configured. */
   public int legacyMaxWidth = -1;

   /** The legacy height of this font; negative value if this was not configured. */
   public int legacyHeight = -1;
   
   /** Byte array with the custom font definition. */
   public byte[] fontDefinition;
   
   /** Flag indicating this is a fixed-size font. */
   public boolean fixedSize;
   
   /** Flag indicating that this font name has a separate font definition for the bold style. */
   public boolean boldFontDefinition;
   
   /** Flag to inform client to use anti-aliasing rendering hint in Swing session */
   public boolean swingAntiAliasing = false;

   /**  
    * When set, this represents an alias for a legacy font; the driver will create a 
    * {@link #font} instance for the font with the specifications given by this instance, instead 
    * of the current instance.  The original font's specs will remain and will be used only for
    * code which needs the legacy specs - all drawing will be performed using the alias font.
    */ 
   public FontDetails<F> fontAlias;
   
   /**
    * Create new font details.
    */
   public FontDetails()
   {
   }
   
   /**
    * Create a new font specification.  This c'tor is used for transferring the font details from
    * the server to the client side.
    * 
    * @param    key
    *           The key for this font.
    * @param    fontName
    *           The font name.
    * @param    style
    *           The font's style.
    * @param    pointSize
    *           The font point size.
    * @param    legacyWidth
    *           The legacy width which needs to be matched by this font.  If no legacy details
    *           exist, then this argument will be negative.
    * @param    legacyMaxWidth
    *           The legacy max width needs to be matched by this font.  If no legacy details
    *           exist, then this argument will be negative.
    * @param    legacyHeight
    *           The legacy height which needs to be matched by this font.  If no legacy details
    *           exist, then this argument will be negative.
    * @param    fontDefinition
    *           The font's definition.
    * @param    fixedSize
    *           Flag indicating if this is a fixed-size font.
    * @param    swingAntiAliasing
    *           Flag indicating if we use anti-aliasing mode in Swing client.
    */
   public FontDetails(String    key,
                      String    fontName,
                      FontStyle style,
                      int       pointSize,
                      int       legacyWidth,
                      int       legacyMaxWidth,
                      int       legacyHeight,
                      byte[]    fontDefinition,
                      boolean   fixedSize,
                      boolean   swingAntiAliasing)
   {
      this.key               = key;
      this.fontName          = fontName;
      this.style             = style;
      this.pointSize         = pointSize;
      this.legacyWidth       = legacyWidth;
      this.legacyMaxWidth    = legacyMaxWidth;
      this.legacyHeight      = legacyHeight;
      this.fontDefinition    = fontDefinition;
      this.fixedSize         = fixedSize;
      this.swingAntiAliasing = swingAntiAliasing;
   }
   
   /**
    * Create a new font specification.  This c'tor is used for transferring the font details from
    * the server to the client side.
    * 
    * @param    key
    *           The key for this font.
    * @param    fontName
    *           The font name.
    * @param    style
    *           The font's style.
    * @param    pointSize
    *           The font point size.
    * @param    legacyWidth
    *           The legacy width which needs to be matched by this font.  If no legacy details
    *           exist, then this argument will be negative.
    * @param    legacyMaxWidth
    *           The legacy max width needs to be matched by this font.  If no legacy details
    *           exist, then this argument will be negative.
    * @param    legacyHeight
    *           The legacy height which needs to be matched by this font.  If no legacy details
    *           exist, then this argument will be negative.
    * @param    fontDefinition
    *           The font's definition.
    * @param    fixedSize
    *           Flag indicating if this is a fixed-size font.
    */
   public FontDetails(String    key,
                      String    fontName,
                      FontStyle style,
                      int       pointSize,
                      int       legacyWidth,
                      int       legacyMaxWidth,
                      int       legacyHeight,
                      byte[]    fontDefinition,
                      boolean   fixedSize)
   {
      this(key, fontName, style, pointSize, legacyWidth, legacyMaxWidth, legacyHeight,
           fontDefinition, fixedSize, false);
   }
   
   /**
    * Create new font details.
    * 
    * @param    font
    *           The font object instance.
    * @param    legacyWidth
    *           The legacy width of this font; negative value if this was not configured.
    * @param    legacyHeight
    *           The legacy height of this font; negative value if this was not configured.
    * @param    fontDefinition
    *           Byte array with the custom font definition.
    */
   public FontDetails(F font, int legacyWidth, int legacyHeight, byte[] fontDefinition)
   {
      this.font           = font;
      this.legacyWidth    = legacyWidth;
      this.legacyHeight   = legacyHeight;
      this.fontDefinition = fontDefinition;
   }
   
   /**
    * Return a new font details instance with the same copied state.
    *
    * @return   The copied instance.
    */
   public FontDetails<F> copy()
   {
      FontDetails<F> fd = new FontDetails<F>();
      
      fd.key            = this.key;
      fd.fontName       = this.fontName;
      fd.font           = this.font;
      fd.style          = this.style;
      fd.pointSize      = this.pointSize;
      fd.legacyWidth    = this.legacyWidth;
      fd.legacyMaxWidth = this.legacyWidth;
      fd.legacyHeight   = this.legacyHeight;
      fd.fixedSize      = this.fixedSize;
      fd.fontAlias      = (this.fontAlias == null ? null : this.fontAlias.copy());
      
      // normally we would make a copy of this array, but this is essentially a read-only
      // array of bytes and we don't want duplicates of that data
      fd.fontDefinition = this.fontDefinition;
      
      fd.boldFontDefinition = this.boldFontDefinition;
      fd.swingAntiAliasing  = this.swingAntiAliasing;
      
      return fd;
   }

   @Override
   public String toString()
   {
      return "FontDetails{" +
             "key='" + key + '\'' +
             ", fontName='" + fontName + '\'' +
             ", style=" + style +
             ", pointSize=" + pointSize +
             ", legacyWidth=" + legacyWidth +
             ", legacyMaxWidth=" + legacyMaxWidth +
             ", legacyHeight=" + legacyHeight +
             ", fixedSize=" + fixedSize +
             ", boldFontDefinition=" + boldFontDefinition +
             ", fontAlias=" + fontAlias +
             ", swingAntiAliasing=" + swingAntiAliasing +
             '}';
   }
}