/*
** Module   : FontStyle.java
** Abstract : list of font styles
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 GES 20150427 Created initial version.
** 002 CA  20150520 Added helper APIs and support for underline.
** 003 CA  20150810 Added toInt().
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

/**
 * Style constants representing the possible styles that can be applied to a font.
 */
public enum FontStyle
{
   PLAIN,
   BOLD,
   ITALIC,
   BOLD_ITALIC,
   UNDERLINE,
   UNDERLINE_BOLD,
   UNDERLINE_ITALIC,
   UNDERLINE_BOLD_ITALIC;
   
   /**
    * Check if this is a bold font.
    * 
    * @return   See above.
    */
   public boolean isBold()
   {
      return this == BOLD           ||
             this == BOLD_ITALIC    ||
             this == UNDERLINE_BOLD ||
             this == UNDERLINE_BOLD_ITALIC;
   }
   
   /**
    * Check if this is an italic font.
    * 
    * @return   See above.
    */
   public boolean isItalic()
   {
      return this == ITALIC           ||
             this == BOLD_ITALIC      ||
             this == UNDERLINE_ITALIC ||
             this == UNDERLINE_BOLD_ITALIC;
   }

   /**
    * Check if this is an underlined font.
    * 
    * @return   See above.
    */
   public boolean isUnderline()
   {
      return this == UNDERLINE        ||
             this == UNDERLINE_BOLD   ||
             this == UNDERLINE_ITALIC ||
             this == UNDERLINE_BOLD_ITALIC;
   }

   /**
    * Create a new font with the given style(s).
    * 
    * @param    isBold
    *           Flag indicating a bold font.
    * @param    isItalic
    *           Flag indicating an italic font.
    * @param    isUnderline
    *           Flag indicating an underline font.
    *
    * @return   A new {@link FontStyle}.
    */
   public static FontStyle createStyle(boolean isBold, boolean isItalic, boolean isUnderline)
   {
      final FontStyle style;
      
      if (isBold && isItalic && isUnderline)
      {
         style = FontStyle.UNDERLINE_BOLD_ITALIC;
      }
      else if (!isBold && isItalic && isUnderline)
      {
         style = FontStyle.UNDERLINE_ITALIC;
      }
      else if (isBold && !isItalic && isUnderline)
      {
         style = FontStyle.UNDERLINE_BOLD;
      }
      else if (!isBold && !isItalic && isUnderline)
      {
         style = FontStyle.UNDERLINE;
      }
      else if (isBold && isItalic && !isUnderline)
      {
         style = FontStyle.BOLD_ITALIC;
      }
      else if (!isBold && isItalic && !isUnderline)
      {
         style = FontStyle.ITALIC;
      }
      else if (isBold && !isItalic && !isUnderline)
      {
         style = FontStyle.BOLD;
      }
      else
      {
         style = FontStyle.PLAIN;
      }
      
      return style;
   }

   /**
    * Create an int representation of this style, to be used for serialization.  Each bit is set
    * as: bit 0 for bold, bit 1 for italic, bit 2 for underline.
    * 
    * @return   Int representation of this style.
    */
   public int toInt()
   {
      int result = 0;

      if (isBold())
      {
         result = result | (1 << 0);
      }

      if (isItalic())
      {
         result = result | (1 << 1);
      }

      if (isUnderline())
      {
         result = result | (1 << 2);
      }
      
      return result;
   }
}