/*
** Module   : Label.java
** Abstract : Label widget implementation.
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 SVG 20050609   @21452 Initial version.
** 003 SIY 20050831   @22506 Refactoring (see @@22394-22409).
** 004 SIY 20050907   @22640 Fixed use of @inheritDoc tag in javadoc.
** 005 SIY 20050926   @22842 Using ColorTable for drawing.
** 006 SIY 20060209   @24471 Updated to accommodate changes in xxxConfig
**                           classes. Minor cleanups.
** 007 SIY 20060323   @25222 Added copy constructor.
** 008 SIY 20060615   @27239 Do not draw label if its parent is null.
** 009 SIY 20060824   @28853 Removed getP2JType() method.
** 010 SIY 20060908   @29380 Fixed drawing.
** 011 GES 20061013   @30401 Added a delimiter flag and changed drawing in redirected output mode
**                           (if we are marked as a delimiter) to avoid the output of the line 
**                           itself (instead we only output a blank line).
** 012 SIY 20070214   @32128 Renamed getDelimiter() to isDelimiter() to match common naming 
**                           convention for boolean properties.
** 013 SIY 20071004   @35366 Rolled back #011.
** 014 NVS 20080401   @37778 Added isFocusable() method implementation.
** 015 SIY 20080401   @38798 Added support for quirks label positioning in ATTR-SPACE handling.
** 016 SIY 20100111   @44522 Fixed a rare case when drawing of empty space instead of label is 
**                           necessary. 
** 017 SIY 20100619          Switching to own implementation of character UI.
** 018 SIY 20100804          Changes in imports, some refactorings.
** 019 SIY 20100822          Refactored AttributeContainer support.
** 020 SIY 20100915          Renamed base class.
** 021 SIY 20110112          Changes related to new color handling scheme.
** 022 SIY 20111101          Updated imports.
** 023 SIY 20111202          Refactored and moved from CHUI package.
** 024 CA  20140805          Added generics to force the abstract widget classes to not use
**                           implementation-specific code: the abstract widgets will have access
**                           only to the base APIs provided by the OutputManager.
** 025 HC  20140911          GUI coordinate support - data types refactoring.
** 026 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 027 HC  20141021          Implemented window move/position support.
** 028 CA  20141220          Implemented drawCaret, as is a no-op for LABEL widget.
** 029 CA  20150105          Changes for SIDE-LABEL-HANDLE runtime support.
** 030 CA  20150114          Fixed support for changing the ROW/COLUMN attrs from server-side, 
**                           for a side-label.
** 031 CA  20150127          If label text is changed, invalidate the location.
** 032 HC  20150128          Added support for frame widget position changes induced by
**                           setting ROW/COL/X/Y attributes.
** 033 HC  20150214          Improved positioning of side labels.
** 034 CA  20150222          Fixed label's COLUMN and WIDTH-CHARS attributes (in ChUI).
** 035 CA  20150323          Pass the frame ID when creating a label (to properly resolve the font
**                           in GUI).
**                           When linking a label, preserve the label's color attributes, too.
** 036 HC  20150323          Removed business logic from config classes.
** 037 CA  20150406          Track the label's side widget (if this is a side-label), as in GUI 
**                           the side-label implicitly follows the widget's height.
** 038 CA  20150519          The label's font needs to be copied, when the label is replaced.
** 039 CA  20150722          Changes to allow frame layout to be done only once, on the frame 
**                           definition. Subsequent label changes are managed by postprocessConfig.
** 040 CA  20150728          Added getter for sideWidget field.
** 041 CA  20151112          "label does not fit in frame" error message is shown only if the label
**                           changes.
** 042 EVL 20160413          Label can not have '&' character inside.
** 043 EVL 20160823          Fix for column calculation with side-labeled widgets.
** 044 HC  20160922          Added getLegacyWidget() method.
** 045 HC  20161018          Frame layout improvements.
** 046 CA  20161101          Added support for mnemonic processing.
** 047 EVL 20170724          Direct manipulation is not valid with LABEL widget.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.util.Objects;

import com.goldencode.util.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Implementation of label for Progress components. Progress uses TEXT
 * components for labels but this seems to be incorrect. TEXT implies
 * underlying variable and complex format.
 */
public abstract class Label<O extends OutputManager<?>>
extends FixedSizeContainer<O>
implements WidgetWithConfig<LabelConfig>
{
   /** Widget's size which is set by <code>setSize</code> method. */
   private boolean forcedSize = false;
   
   /** Component attribute container. */
   protected LabelConfig config = null;

   /** Make label aligned to right boundary. */
   protected boolean rightAligned = true;

   /** Display additional line */
   protected boolean underlined = false;
   
   /** Flag the special case of use as a separator line. */
   private boolean delimiter = false;

   /** Label right shift offset. */
   protected double offset = 0;

   /** If it is <code>true</code>, empty string is displayed.*/
   protected boolean emptyMode = false;
   
   /** Prepare for disabling empty mode. */
   protected boolean disableEmptyMode = false;
   
   /** The side-widget associated with this label, if this is a side-label. */
   protected Widget<O> sideWidget = null;
   
   /**
    * Constructs label with given text.
    * 
    * @param   id
    *          The widget's ID.
    * @param   text
    *          Initial text.
    * @param   frameId
    *          The ID of the frame to which the label belongs, or -1 if unknown at the time of the 
    *          creation.
    */
   public Label(WidgetId id, String text, int frameId)
   {
      config = ConfigManager.getInstance().resolveWidgetConfig(id, LabelConfig.class);
      config.frameId = frameId;

      setName(text);
   }

   /**
    * Create a copy of the label.
    * 
    * @param   id
    *          The widget's ID.
    * @param   label
    *          Label to copy parameters from.
    * @param   frameId
    *          The ID of the frame to which the label belongs, or -1 if unknown at the time of the 
    *          creation.
    */
   public Label(WidgetId id, Label<O> label, int frameId)
   {
      config = ConfigManager.getInstance().resolveWidgetConfig(id, LabelConfig.class);
      config.frameId = frameId;

      setName(label.config.name);
      
      setUnderlined(label.isUnderlined());
      setRightAligned(label.isRightAligned());
      setWidth(label.width());
      forcedSize = label.forcedSize;
      setLocation(label.location());
      setVisible(label.isVisible());
   }
   
   /**
    * Retrieve the widget's width in native units.
    * 
    * @return   See above.
    */
   protected abstract int nativeWidth();

   /**
    * Retrieve the widget's height in native units.
    * 
    * @return   See above.
    */
   protected abstract int nativeHeight();
   
   /**
    * Set the specified widget as this side-label's {@link #sideWidget}.
    * 
    * @param    widget
    *           The side-widget for this label.
    */
   public void setSideWidget(Widget<O> widget)
   {
      this.sideWidget = widget;
      updateSize();
   }
   
   /**
    * Get the side-widget associated with this label.
    * 
    * @return   The {@link #sideWidget}.
    */
   public Widget<O> getSideWidget()
   {
      return sideWidget;
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, LabelConfig cfg)
   {
      config = ConfigManager.getInstance().resolveWidgetConfig(id, cfg);
      ConfigManager.getInstance().replaceConfig(config, cfg);
   }

   /**
    * Create a copy of the label.
    * 
    * @param   id
    *          The widget's ID.
    * @param   label
    *          Label to copy parameters from.
    */
   public Label(WidgetId id, Label<O> label)
   {
      config = ConfigManager.getInstance().resolveWidgetConfig(id, LabelConfig.class);
      
      config.name = label.config.name;
      setUnderlined(label.isUnderlined());
      setRightAligned(label.isRightAligned());
      setWidth(label.width());
      forcedSize = label.forcedSize;
      setLocation(label.location());
      setVisible(label.isVisible());
   }

   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public LabelConfig config()
   {
      return config;
   }

   /**
    * Marks this label as one used to draw a separator line for header
    * widgets.  In this case there is no label text but the width is
    * positive and the underline attribute is enabled.
    *
    * @param    delimiter
    *           <code>true</code> to set this as a separator line.
    */
   public void setDelimiter(boolean delimiter)
   {
      this.delimiter = delimiter;
   }
   
   /**
    * Gets the flag that marks this label as one used to draw a separator
    * line for header widgets.
    *
    * @return   <code>true</code> if this label is a separator line.
    */
   public boolean isDelimiter()
   {
      return delimiter;
   }

   /**
    * Draws text. Screen value is left justified or padded to the right bound
    * with spaces. Non first lines are filled with spaces.
    */
   @Override
   public abstract void draw();

   /**
    * Sets position of terminal's cursor. Really does nothing because this
    * widget will never get an input focus.
    */
   @Override
   public void drawCaret()
   {
   }

   /**
    * Sets new widget's size. After call to this method the widget will
    * not change size automatically accordingly to screen value changes.
    * 
    * @param   width
    *          New widget width.
    * @param   height
    *          New widget height.
    */
   @Override
   public void setSize(double width, double height)
   {
      forcedSize = true;
      super.setSize(width, height);
   }

   /**
    * Set widget width.
    * 
    * @param   width
    *          New widget width.
    */
   @Override
   public void setWidth(double width)
   {
      forcedSize = true;
      super.setWidth(width);
   }
   
   /**
    * Set widget height.
    * 
    * @param   height
    *          New widget height.
    */
   @Override
   public void setHeight(double height)
   {
      forcedSize = true;
      super.setHeight(height);
   }
   
   /**
    * This component will not receive an input focus when Tab or Shift-Tab is
    * pressed.
    * 
    * @return  Always <code>false</code>.
    */
   @Override
   public boolean focusTraversable()
   {
      return false;
   }

   /**
    * Checks whether this attribute container can receive focus generally.
    * 
    * @return  <code>true</code> if this container can be focused under
    *          certain conditions. <code>false</code> if, and only if, this
    *          attribute container can never receive focus.    
    */
   @Override
   public boolean isFocusable()
   {
      return false;
   }

   /**
    * Synchronizes the actual size with the default size.
    */
   protected void updateSize()
   {
      if (forcedSize)
         return;

      int nativeWith = nativeWidth();
      int nativeHeight = nativeHeight();
      
      CoordinatesConversion cc = screen().coordinates();
      double width = cc.widthFromNative(nativeWith);
      double height = cc.heightFromNative(nativeHeight);
      
      // call super to avoid setting forcedSize to true
      super.setSize(width, height);
      
      // synchronize the new size with server
      BaseConfig c = config();
      c.widthChars = width;
      c.heightChars = height;
   }

   /**
    * Set underlining mode.
    * 
    * @param   b
    *          New value for the mode.
    */
   public void setUnderlined(boolean b)
   {
      underlined = b;
      updateSize();
   }

   /**
    * Get current value of the underlining flag.
    * 
    * @return  <code>true</code> if label is underlined.
    */
   public boolean isUnderlined()
   {
      return underlined;
   }
   
   /**
    * Get current value of the alignment.
    * 
    * @return  <code>true</code> if right alignment is active.
    */
   public boolean isRightAligned()
   {
      return rightAligned;
   }
   
   /**
    * Activate or deactivate right alignment of the text.
    * 
    * @param   rightAligned
    *          <code>true</code> activate right alignment,
    *          <code>false</code>- deactivate.
    */
   public void setRightAligned(boolean rightAligned)
   {
      this.rightAligned = rightAligned;
   }

   /**
    * Sets the name of the label.
    * 
    * @param   name
    *          New name of the component.
    */
   public void setName(String name)
   {
      config.name = name;

      //Reset the forcedSize flag so updateSize() can properly adjust the size.
      forcedSize = false;
      updateSize();
   }
   
   /**
    * Get widget name.
    *
    * @return  widget name or empty string if is not set.
    */
   @Override
   public String name()
   {
      String name = super.name();
      return name == null ? "" : name;
   }

   /**
    * Set extra offset for label.
    * 
    * @param   offset
    *          Size of the shift.
    */
   public void setOffset(double offset)
   {
      this.offset = Coordinate.scale(offset);
   }

   /**
    * Disable empty label mode. Note that actual disabling happens only during
    * next {@link #draw()} call.
    */
   public void setEmptyModeOff()
   {
      disableEmptyMode = true;
   }

   /**
    * Enable empty mode.
    */
   public void setEmptyModeOn()
   {
      emptyMode = true;
   }
   
   /**
    * Process this label so that it's linked with the server-side widget represented by the 
    * <code>serverSideId</code>.
    * 
    * @param    frameId
    *           The ID of the frame to which this label belongs.
    * @param    serverSideId
    *           The server-side ID of this label.
    */
   @SuppressWarnings({ "unchecked", "rawtypes" })
   public void linkTo(int frameId, int serverSideId)
   {
      if (serverSideId == -1)
      {
         return;
      }
      
      BaseConfig lblCfg = config();
      WidgetId oldId = lblCfg.id;
      WidgetId newId = oldId.createId(serverSideId);
      
      ConfigManager cm = ConfigManager.getInstance();
      BaseConfig bc = cm.getActiveConfig(newId);
      
      if (bc != null)
      {
         Point loc = location();
        
         double x = loc.x;
         double y = loc.y;
         boolean setLoc = false;
   
         double assignedCol = ConfigHelper.getAssignedColumn(bc);
         double assignedRow = ConfigHelper.getAssignedRow(bc);
         
         if (assignedCol != BaseConfig.INV_COORD)
         {
            x = assignedCol;
            setLoc = true;
         }
         
         if (assignedRow != BaseConfig.INV_COORD)
         {
            y = assignedRow;
            setLoc = true;
         }
   
         if (setLoc)
         {
            setLocation(x, y);
         }
         
         // inherit width, but only when it is not already set on this
         if (bc.widthChars != -1 && width() == -1)
         {
            setWidth(bc.widthChars);
            lblCfg.widthChars = bc.widthChars;
         }
         
         lblCfg.align = bc.align;
         
         // restore the colors
         lblCfg.bgcolor = bc.bgcolor;
         lblCfg.fgcolor = bc.fgcolor;
         lblCfg.dcolor  = bc.dcolor;
         lblCfg.pfcolor = bc.pfcolor;
         lblCfg.sysbgcolor = bc.sysbgcolor;
         lblCfg.sysfgcolor = bc.sysfgcolor;
         
         // restore the font
         lblCfg.font = bc.font;
      }

      ClientConfigManager mgr = ConfigManager.getInstance();
      
      // switch the internal ID for this label (from virtual to real)
      mgr.switchId(oldId, newId);
      
      // mark as dirty all configs which are not the same as what the server has...
      mgr.trackAllChanges(lblCfg);
      
      OutputManager.instance().getRegistry().addWidget((Widget) this);
   }
   
   /**
    * Update the {@link #visible} flag in the configuration and in the widget's state.
    * <p>
    * If the widget has just been realized and its width has not yet been set, it will be computed
    * now.
    * 
    * @param    visible
    *           The widget's visible state.
    */
   @Override
   public void _setVisible(boolean visible)
   {
      boolean oldRealized = config.realized;
      
      super._setVisible(visible);
      
      if (!oldRealized && config.realized && config.widthChars == -1)
      {
         config.widthChars = width();
      }
   }
   
   /**
    * Get the state of the {@link #forcedSize} flag.
    * 
    * @return   See above.
    */
   public boolean isForcedSize()
   {
      return forcedSize;
   }
   
   /**
    * Set the {@link #forcedSize} flag to the specified state.
    * 
    * @param    forcedSize
    *           The new state for the {@link #forcedSize} flag.
    */
   public void setForcedSize(boolean forcedSize)
   {
      this.forcedSize = forcedSize;
   }

   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(LabelConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
      
      LabelConfig c = config();
      
      setWidth(c.widthChars);
      
      // name changed?
      if (!Objects.equals(((BaseConfig) beforeUpdate).name, c.name))
      {
         //reset the forcedSize flag so updateSize() can properly adjust the size.
         forcedSize = false;
         updateSize();
      }
   }
   
   /**
    * Additional processing required when widget configuration is updated AFTER the frame layout
    * has been performed.
    */
   @Override
   public void postprocessConfig()
   {
      super.postprocessConfig();
      
      if (sideWidget != null)
      {
         double newLblWidth = 0;
         double oldLblWidth = 0;
         boolean labelChanged = false;
         if (sideWidget instanceof LabeledDataContainer)
         {
            String newLabel = ((LabeledDataContainer) sideWidget).getLabelText() + ": ";
            labelChanged = !newLabel.equals(config.name);

            if (labelChanged)
            {
               newLblWidth = ZeroColumnLayout.labelWidth((LabeledWidget<?>) sideWidget, newLabel);
               oldLblWidth = ZeroColumnLayout.labelWidth((LabeledWidget<?>) sideWidget,
                                                         config.name);
            }
         }

         // check the left border
         if (labelChanged)
         {
            double newColumn = Coordinate.scale(ConfigHelper.getEffectiveColumn(config) +
                                                oldLblWidth - newLblWidth);

            if (newColumn < 0)
            {
               Frame<O> fr = (Frame<O>) UiUtils.locateFrame(this);

               StringBuilder err = new StringBuilder();

               if (screen().isChui())
               {
                  err.append("**Unable to set SCREEN-VALUE.  ")
                     .append("LITERAL widget does not fit in FRAME ")
                     .append(fr.shortName()).append(". (4054)");
               }
               else
               {
                  err.append("**All or part of LITERAL widget is being placed outside of FRAME ")
                     .append(fr.shortName())
                     .append(" by setting SCREEN-VALUE. (5905)");
               }

               ThinClient.getInstance().displayErrorMessage(err.toString());

               // no need to restore the label string
            }
            else
            {
               Point loc = location();
               config.column = newColumn;
               setLocation(newColumn, loc.y);

               if (sideWidget.config() instanceof ControlConfig)
               {
                  // setName will update the size, too
                  setName(((ControlConfig) sideWidget.config()).label + ": ");
               }
            }
         }
      }
   }

   /**
    * The method returns the legacy widget representing this widget.
    *
    * @return   A non-wirtual widget reference, or <code>null</code> if none found.
    */
   @Override
   public Widget<O> getLegacyWidget()
   {
      return getSideWidget();
   }

   /**
    * Additional processing required when widget configuration is updated AFTER the frame layout
    * has been performed.
    * 
    * @param    widget
    *           The widget to which this side-labe is attached.
    */
   protected void postprocessConfig(LabeledWidget<O> widget)
   {
      // check the frame first - this call is only for non-side-labels
      Frame<O> frame = (Frame<O>) UiUtils.locateFrame((Widget<O>) widget);
      if (frame.isSideLabels())
      {
         return;
      }
      
      String label = ZeroColumnLayout.getLabel(widget, frame.isSideLabels(), frame.hasHeaders());
      if (label != null && !label.equals(this.name()))
      {
         // TODO: if new width is greater than frame width, then the width remains as the 
         // current label's width - and the new label is trimmed to this width
         setName(label);
      }
   }

   /**
    * Check the direct manipulation capability of the widget.  LABEL widget does not allow direct
    * manipulation.
    * 
    * @return  <code>TRUE</code> if the widget is able to be directly manipulated,
    *          <code>FALSE</code> otherwise.
    */
   @Override
   public boolean isDirectManipulable()
   {
      return false;
   }
}