/*
** Module   : MnemonicInfo.java
** Abstract : Client object accumulating prepared label and mnemonic info.
**
** Copyright (c) 2016-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------------Description-----------------------------
** 001 VIG 20160222 Created initial version.
** 002 VIG 20160624 init() method: changed return value and renamed to prepare().
** 003 HC  20161018 Added an initial support for rendering of mnemonic labels to all labeled
**                  widgets.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

/**
 * This class accumulates label information: 
 * <ul>
 *   <li>label after processing of special character '&amp;' denoting mnemonic.</li> 
 *   <li>mnemonic character, used to call menu item, that it is assigned on.</li>
 *   <li>mnemonic index in preprocessed label for mnemonic drawing.</li>
 * </ul>
 */
public class MnemonicInfo
{
   /**
    * Preprocessed label without unnecessary '&amp;'
    */
   private String prepLabel;

   /**
    * Hot key for menu calling. Set by adding '&amp;' to label.
    * The key is the first character after '&amp;' or first 
    * character of the whole label if there is no '&amp;' in it.
    * <tt>null</tt> means not assigned.
    */
   private Character mnemonic;
   
   /**
    * Integer index in a config.label string, pointing to a mnemonic character.
    * Can be null, meaning and not drawn.
    */
   private Integer mnemIdx;
   
   /**
    * Flag, denoting using of the first character as a mnemonic if it wasn't
    * assigned manually (using correct '&amp;' sequence).
    */
   private final boolean defaultFirst;
   
   /**
    * Default constructor.
    */
   public MnemonicInfo()
   {
      this.defaultFirst = true;
   }
   
   /**
    * Constructor.
    * 
    * @param   defaultFirst
    *          <code>true</code> if first character is a default mnemonic.
    */
   public MnemonicInfo(boolean defaultFirst)
   {
      this.defaultFirst = defaultFirst;
   }

   /**
    * Initialized fields with new menu item label
    * 
    * @param   title
    *          New menu item label.
    *          
    * @return  prepared label String.
    */
   protected String prepare(String title)
   {
      String label = new String(title);
      StringBuilder result = new StringBuilder(label);

      // <tt>true</tt> means sequence of odd length was met. Wherein real mnemonic can 
      // be not assigned if that sequence ends the label. 
      boolean mnemonicAssigned = false;
      
      boolean insideSeq = false;
      int start = -1, end = -1;
      
      for (int i = result.length() - 1; i >= 0; i--)
      {
         if (label.charAt(i) == '&')
         {
            if (!insideSeq)
            {
               // sequence has started.
               start = i;
            }
            
            insideSeq = true;
         }
         
         if (label.charAt(i) != '&' || i == 0)
         {
            if (insideSeq)
            {
               // sequence has ended.
               end = (label.charAt(i) == '&' && i == 0) ? 0 : i + 1;
               
               mnemonicAssigned = processSequence(start, end, result, mnemonicAssigned);
            }
            
            insideSeq = false;
         }
      }
      
      this.prepLabel = result.toString();
      
      if (!mnemonicAssigned && !this.prepLabel.isEmpty() && defaultFirst)
      {
         this.mnemonic = this.prepLabel.charAt(0);
         this.mnemIdx = 0;
      }
      
      return this.prepLabel;
   }
   
   /**
    * Get prepared title.
    * 
    * @return  String.
    */
   public String getPrepLabel()
   {
      return prepLabel;
   }
   
   /**
    * Get mnemonic character.
    * 
    * @return  Character.
    */
   public Character getMnemonic()
   {
      return mnemonic;
   }
   
   /**
    * Set mnemonic in a prepared label.
    * 
    * @param  mnemonic
    *         New mnemonic.
    */
   protected void setMnemonic(Character mnemonic)
   {
      this.mnemonic = mnemonic;
   }
   
   /**
    * Get mnemonic index in a prepared label.
    * 
    * @return  Integer.
    */
   public Integer getMnemIdx()
   {
      return mnemIdx;
   }

   /**
    * Set mnemonic index in a prepared label.
    * 
    * @param  mnemIdx
    *         New mnemonic index.
    */
   protected void setMnemIdx(Integer mnemIdx)
   {
      this.mnemIdx = mnemIdx;
   }
   
   /**
    * Process one '&amp;' characters sequence, assigning {@link #mnemonic} and {@link #mnemIdx}.
    * 
    * @param   start
    *          the first '&amp;' index from the right.
    * @param   end
    *          the last '&amp;' index from the right.
    * @param   mnemAssnd
    *          <tt>true</tt> means that a sequence of '&amp;' of odd length was met, wherein
    *          mnemonic wasn't necessarily being assigned.
    *          
    * @param   result
    *          Resulting string buffer.
    */
   protected boolean processSequence(int start, int end, StringBuilder result, boolean mnemAssnd)
   {
      int seqLen = start - end + 1;
      
      if (seqLen % 2 != 0 && !mnemAssnd)
      {
         if (result.length() > start + 1)
         {
            mnemonic = result.charAt(start + 1);
            mnemIdx = start + 1;
         }
         
         mnemAssnd = true;
      }
      
      int newSeqLen = seqLen / 2;
      char[] f = new char[newSeqLen];
      Arrays.fill(f, '&');
      
      // If there is just one "&" - it will be deleted.
      result.replace(end, start + 1, new String(f));
      
      if (mnemIdx != null)
      { 
         // This is an index in the current version of label, so after deleting some number 
         // of '&' from that current label we have to correct current mnemonic index. 
         mnemIdx -= seqLen - newSeqLen;
      }
      
      return mnemAssnd;
   }
}
