/*
** Module   : RadioButton.java
** Abstract : RadioButton implementation as helper for CHUI Radio-set
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 EVL 20050301   @20143 Creation based on CHARVA JRadioButton class.
**                           Necessary modification to get compiled.
** 002 EVL 20050303   @20149 Adding initial data arbitrary association
**                           for radio buttons.
** 003 EVL 20050304   @20170 Handling of auto-resize feature.
** 004 EVL 20050604   @21405 Changing the inner variable names to sync
**                           with the rest of the project.
** 005 EVL 20050607   @21435 Syntax for comments checking.
** 006 EVL 20050729   @21915 Removing the ugly tabulation from source.
** 007 SIY 20050831   @22543 Formatting cleanups and spell checking.
** 008 EVL 20050914   @22713 Focus related events from single RadioSet
**                           must be combined as for the single widget.
** 009 EVL 20050916   @22734 Returned value class is changed to be the
**                           BaseDataType.
** 010 SIY 20050926   @22847 Using ColorTable for drawing.
** 011 GES 20060114   @24460 The data of the class is now a BaseDataType,
**                           there are new BDT constructors and the class
**                           now implements DataContainer.
** 012 SIY 20060209   @24461 Cleanups and fixes for new RadioSet.
** 013 NVS 20060227   @24866 Key actions KA_* are used instead of virtual
**                           keys VK_* for all editing operations.
** 014 NVS 20060322   @25198 Added dummy isModified() method since now it
**                           is part of the DataContainer interface.
** 015 SIY 20060417   @25597 Changes in event handling to prevent
**                           incorrect changing of state flags.
** 016 SIY 20060518   @26276 During drawing take into account state of the
**                           parent widget and don't highlight focused
**                           button if widget is disabled.
** 017 GES 20060724   @28155 Use forced assignment mode of BaseDataType
**                           to avoid checking the ErrorManager for a 
**                           pending error (that causes a trip to the
**                           server which is unnecessary).
** 018 NVS 20061204   @31558 Switched to using a private widgetFocused
**                           flag instead of config.selected
** 019 SIY 20061214   @31680 Added drawCaret() method.
** 020 NVS 20061214   @31691 Fixed drawing of uninitialized radio buttons.
** 021 NVS 20080320   @37614 Fixed drawing of radio buttons in focus when
**                           they are disabled.
** 022 SIY 20080411   @37922 Significant rewrite in order to accommodate 
**                           changes in focus handling.
**                           Ref: #37923
** 023 SVL 20080902   @39667 Fixed radio button drawing (only brackets
**                           should have reverse color when the button is
**                           highlighted).  
** 024 SIY 20100619          Switching to own implementation of character UI.
** 025 SIY 20100804          Changes in imports, some refactorings.
** 026 SIY 20110112          Changes related to new color handling scheme.
** 027 SIY 20111101          Updated imports.
** 028 SIY 20111202          Refactored and moved from CHUI package.
** 029 CA  20140805          Added generics to force the abstract widget classes to not use
**                           implementation-specific code: the abstract widgets will have access
**                           only to the base APIs provided by the OutputManager.
** 030 HC  20140911          GUI coordinate support - data types refactoring.
** 031 HC  20141021          Implemented window move/position support.
** 032 IAS 20150216          Added support for the EXPAND attribute
** 033 IAS 20150312          Remove 'selectable' attribute - use 'enabled' instead.
** 034 HC  20150323          Removed business logic from config classes.
** 035 CA  20151125          Fixed computing default-focus when the first widget in frame is a 
**                           radio-set.
** 036 CA  20161101          Added support for mnemonic processing.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;


import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * An implementation of a radio-button - an item that is always in one of two
 * states (SELECTED or DESELECTED) and which displays its state to the user.
 * RadioButtons are used with a ButtonGroup object to create a group of
 * buttons in which only one button at a time can be selected (Create a
 * ButtonGroup object and use its add() method to add the radio buttons to
 * it).
 */
public abstract class RadioButton<O extends OutputManager<?>>
extends AbstractContainer<O>
implements DataContainer,
           WidgetWithConfig<RadioSetConfig>,
           MnemonicTrait<O>
{
   /** The value to be returned as the RadioButton value. */
   private BaseDataType value = null;

   /** The width of the button. */
   protected double width = -1;
   
   /** Automatic size change enabling flag. */
   protected boolean autoResize = true;
   
   /** Button text. */
   private String text = "";

   /** Selection flag. */
   private boolean selected = false;

   /** Pseudo-focus flag. */
   protected boolean focused = false;

   /** button id */
   private int itemId;
   
   /** Reference to config. */
   protected RadioSetConfig config = null;

   /** Mnemonic information */
   private MnemonicInfo mnemonic = null;
   
   /**
    * Use this constructor when you want to set both the label and the value.
    * 
    * @param   config
    *          Reference to config.
    * @param   label
    *          The button's label which can clarify the button purpose.
    * @param   value
    *          The data which is the button's value.
    */
   public RadioButton(RadioSetConfig config, String label, BaseDataType value)
   {
      this.config = config;
      this.value = value;
      setText(label);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, RadioSetConfig cfg)
   {
      // no-op
   }

   /**
    * Provide access to the attributes stored in the widget.
    *
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public RadioSetConfig config()
   {
      return config;
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(RadioSetConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
   }

   /**
    * Get label text.
    * 
    * @return  label text.
    */
   @Override
   public String getText()
   {
      MnemonicInfo info = getMnemonic();
      
      return info == null ? text : info.getPrepLabel();
   }

   /**
    * Set label text.
    * 
    * @param   text
    *          Label text. 
    */
   void setText(String text)
   {
      this.text = text;
      
      updateMnemonic(text, this.text, false);
   }
   
   /**
    * Returns the BaseDataType instance representing the current value.
    *
    * @return   The value representing the current widget value.
    */
   @Override
   public BaseDataType getValue()
   {
      return value;
   }
   
   /**
    * Getting the modification state of the widget.
    * 
    * @return  always returns <code>false</code> and is never called.
    */
   @Override
   public boolean isModified()
   {
      return false;
   }

   /**
    * Sets the BaseDataType instance representing the current value.
    *
    * @param    value
    *           The value representing the current widget value.
    */
   @Override
   public void setValue(BaseDataType value)
   {
      this.value = value;
   }
   
   /**
    * Replace this radio-button with the new one.
    * 
    * @param   newButton
    *          The new radio button to be used instead of the current one.
    */
   public void replace(RadioButton<O> newButton)
   {
      setText(newButton.getText());
      value.assign(newButton.value, true);
   }
   
   /**
    * Get widget location relative to container.
    * 
    * @return  relative widget location.
    */
   @Override
   public Point location()
   {
      // Radio button shares its config with radio set.
      // Thus prevent the super method to be called which would return
      // the radio set's position.
      return new Point(location);
   }

   /**
    * Get widget physical location relative to container, expressed in units native to UI type.
    * 
    * @return  relative widget location.
    */
   @Override
   public NativePoint physicalLocation()
   {
      // Radio button shares its config with radio set.
      // Thus prevent the super method to be called which would return
      // the radio set's position.
      return new NativePoint(physicalLocation);
   }

   /**
    * Set widget relative location.
    * 
    * The x value corresponds to COLUMN while
    * y value corresponds to ROW 4GL attribute.
    * Note however that the values are expected to be
    * 0-based.
    * 
    * @param   x
    *          Widget X.
    * @param   y
    *          Widget Y.
    */
   @Override
   public void setLocation(double x, double y)
   {
      // Radio button shares its config with radio set.
      // Thus prevent the super method to be called which would return
      // the radio set's position.
      
      if (isVisible())
         repaint(); 
            
      location.x = Coordinate.scale(x);
      location.y = Coordinate.scale(y);
      
      if (isVisible())
         repaint(); 
   }

   /**
    * Set widget relative physical location.
    * 
    * @param   x
    *          Widget physical X.
    * @param   y
    *          Widget physical Y.
    */
   @Override
   public void setPhysicalLocation(int x, int y)
   {
      // Radio button shares its config with radio set.
      // Thus prevent the super method to be called which would return
      // the radio set's position.
      physicalLocation.x = x;
      physicalLocation.y = y;
   }

   /**
    * Return the width of the radio button.
    * 
    * @return  The value of the button width including the label text.
    */
   @Override
   public double width()
   {
      return width > 0 ? width : (getText().length() + 3);
   }

   /**
    * Return height of the radio button.
    * 
    * @return  The value of the button height.
    */
   @Override
   public double height()
   {
      return 1;
   }
   
   /**
    * Set up the width of the button dynamically.
    * 
    * @param   newWidth
    *          The new button width in the character units.
    */
   public void setWidthChars(double newWidth)
   {
      width = Coordinate.scale(newWidth);
      if (autoResize)
         autoResize = false;
   }

   /**
    * Called by this RadioButton's parent container.
    */
   @Override
   public abstract void draw();
   
   /**
    * Draw cursor at the correct position.
    * 
    * @see com.goldencode.p2j.ui.client.widget.Widget#drawCaret()
    */
   @Override
   public abstract void drawCaret();

   /**
    * Get current state of pseudo-focus flag.
    * 
    * @return  Current state of the flag.
    */
   public boolean isFocused()
   {
      return focused;
   }

   /**
    * Set pseudo-focus flag.
    * 
    * @param   focused
    *          New state of the flag.
    */
   public void setFocused(boolean focused)
   {
      this.focused  = focused;
   }
   
   /**
    * Reports if this widget is focusable.
    *
    * @param    exclude
    *           A widget to ignore or <code>null</code> if any focusable widget will do.
    *
    * @return   <code>true</code> if this widget is not the excluded widget.
    */
   @Override
   public boolean hasFocusable(Widget<O> exclude)
   {
      return exclude != this; 
   }

   /**
    * Sets the state of the button.
    * 
    * @param   state
    *          New state of the button.
    */
   public void setSelected(boolean state)
   {
      selected = state;
   }

   /**
    * Get current selection state.
    * 
    * @return  <code>true</code> if button is selected.
    */
   public boolean isSelected()
   {
      return selected;
   }
   
   /**
    * Main event processing entry point.
    * 
    * @param   evt
    *          Event to handle.
    */
   @Override
   public void processEvent(Event evt) 
   {
      if (evt instanceof ItemEvent)
         return;
      
      super.processEvent(evt);
   }

   /**
    * Get button id.
    * 
    * @return  button id.
    */
   public int getItemId()
   {
      return itemId;
   }

   /**
    * Set button id.
    * 
    * @param   id
    *          New button id.
    */
   public void setItemId(int id)
   {
      this.itemId = id;
   }

   /**
    * MnemonicInfo instance getter.
    * <p>
    * The implementors should simply return the value of the mnemonic class field. No
    * additional processing logic is advised.
    *
    * @return   See above.
    */
   public MnemonicInfo getMnemonic()
   {
      return mnemonic;
   }

   /**
    * MnemonicInfo instance setter.
    * <p>
    * The implementors should simply assign the parameter value to the mnemonic class
    * field. No additional processing logic is advised.
    */
   public void setMnemonic(MnemonicInfo mnemonic)
   {
      this.mnemonic = mnemonic;
   }
}