/*
** Module   : ScrollPane.java
** Abstract : ScrollPane helper class implementation for CHUI
**
** Copyright (c) 2005-2017 Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 EVL 20050203   @19739 Creation based on CHARVA JScrollPane class.
**                           Necessary modification to get compiled.
** 002 EVL 20050307   @20218 Modification of the scrollbar's slider
**                           position calculation. The new algorithm takes
**                           into account the negative coordinate value
**                           for viewport Y position.
** 003 EVL 20050408   @20672 Adding horizontal scroll-bar but not for any
**                           instances of JList. SelectionListBody and
**                           ComboBox are the such a cases for now
** 004 EVL 20050421   @20831 Changed constructor order.
** 005 EVL 20050429   @20944 Adding label creation for the widget.
** 006 EVL 20050503   @21072 Setting up the default one line border.
** 007 EVL 20050522   @21252 Adding conditions to choose the child drawing
**                           If the child is instance of the JList or
**                           parent is the Frame or DialogBox we call the
**                           super.draw(), otherwise we handle clipping in
**                           the FrameHelper and have to call only draw()
**                           method of the child.
** 008 EVL 20050531   @21337 Changing the inner variable names to sync
**                           with the rest of the project and preparing
**                           to the javadoc processing.
** 009 EVL 20050607   @21435 Syntax for comments checking.
** 010 EVL 20050729   @21915 Removing the ugly tabulation from source.
** 011 EVL 20050825   @22273 Adding method to get wrapped component
**                           inside.
** 012 SIY 20050831   @22548 Fixed formatting and spelling.
** 013 SIY 20050926   @22848 Using ColorTable for drawing.
** 014 SIY 20060209   @24451 Cleanups and fixes for new SelectionList.
** 015 SIY 20060209   @24498 Fixed scroll-bar handling.
** 016 SIY 20060222   @24641 Removed references to DialogBox.
** 017 GES 20060223   @24680 Eliminated cursor movement in no-box case.
**                           This caused problems in redirected stream
**                           output.
** 018 SIY 20060518   @26436 Added separate drawing routine for use in 
**                           down frames.
** 019 SIY 20060518   @26438 Fixed misspelled method name.
** 020 SIY 20060524   @26459 Added drawing of the border.
** 021 SIY 20060525   @26536 Fixed drawing of the child content.
** 022 NVS 20060728   @28214 Fixed drawing of the border in redirected
**                           mode.
** 023 SIY 20060824   @28857 Removed reference to DialogBox and getP2JType
**                           method.
** 024 GES 20061012   @30369 Don't draw scrollbars when in redirected
**                           mode.
** 025 NVS 20061221   @31731 Fixed drawing of contained selection lists.
** 026 SIY 20070727   @34716 Do not draw scroll bars if their size is less
**                           than 3 chars.
** 027 SIY 20080214   @37112 Fixed drawing of parts of frame which don't 
**                           fit into visible area.
**                           Ref: #37107
** 028 NVS 20080401   @37780 Added isFocusable() method implementation.
** 029 SIY 20080401   @37795 Do not draw scroll-bars if there are borders.
**                           Fix for #37745.
** 030 SIY 20080505   @38206 Draw scroll-bars for DropDown even there are
**                           no borders.
** 031 NVS 20080828   @39621 setLocation() method now calls super class
**                           as well.
** 032 SIY 20081125   @40691 Removed unnecessary scroll() method.
** 033 NVS 20090213   @41316 Dead code removal.
** 034 SIY 20100619          Switching to own implementation of character UI.
** 035 SIY 20100804          Changes in imports, some refactorings.
** 036 SIY 20100822          Refactored AttributeContainer support.
** 037 SIY 20101122          Reduced class hierarchy. Switching from KeyEvent 
**                           to KeyInput.
** 038 SIY 20110112          Changes related to new color handling scheme.
** 039 SIY 20111202          Refactored and moved from CHUI package.
** 040 CA  20140805          Added generics to force the abstract widget classes to not use
**                           implementation-specific code: the abstract widgets will have access
**                           only to the base APIs provided by the OutputManager.
** 041 HC  20140911          GUI coordinate support - data types refactoring.
** 042 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 043 MAG 20140917          Added setScrollBars method.
** 044 HC  20141021          Implemented window move/position support.
** 045 IAS 20150219          Fixed scroll() method.
** 046 HC  20150323          Removed business logic from config classes.
** 047 HC  20150831          Scrolling support overhaul.
** 048 HC  20150920          Improved scroll layout calculation for the case of native scrolling.
** 049 CA  20150929          Fixed multi-window drawing issues caused by concurrent drawing
**                           access to the GuiDriver: all access to the driver is done
**                           exclusively, each thread taking and releasing ownership for the
**                           duration of the drawing operation or other GUI API invocation which
**                           requires access to the underlying physical window.
** 050 HC  20160313          Fixed frame scroll layout calculation.
** 051 OM  20170407          Removed dead code.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.ScrollBar.Bars;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * This is the top-level component for the widget scrolling feature.
 * The class holds and manages several sub-components that together provide 
 * the scrolling support. These are a so-called viewport - {@link Viewport} -
 * scroll bars and {@link ScrollableWidget}. Scroll bars provide a visual feedback
 * of the scroll state, viewport marks the visible portion of {@link ScrollableWidget},
 * the widget being scrolled. The class also functions as the scroll event hub,
 * it makes sure the events are dispatched by the right sub-components. 
 */
public abstract class ScrollPane<O extends OutputManager<?>>
extends AbstractContainer<O>
implements WidgetWithConfig<BaseConfig>,
           ScrollListener
{
   /** The root panel, holds border, contains scroll bars and viewport. */
   protected BorderedPanel<O> rootPanel;
   
   /** View port for the visible part of the pane. */
   private Viewport<O> viewport;

   /** New label to be used with this container. */
   private String label = null;

   /** Attribute container. */
   protected BaseConfig config = null;

   /**
    * Create a ScrollPane that displays the contents of the specified widget.
    * 
    * @param   widget
    *          The widget to be displayed. This widget must implement
    *          the {@link ScrollPublisher} interface.
    */
   @SuppressWarnings("unchecked")
   public ScrollPane(ScrollableWidget<O> widget)
   {
      WidgetFactory<O> factory = screen().getFactory();
      
      viewport = new Viewport<>(widget);
      
      rootPanel = factory.createBorderedPanel();
      rootPanel.add(viewport);
      
      add(rootPanel);
      
      if (widget instanceof ScrollPublisher)
      {
         ((ScrollPublisher) widget).addScrollListener(this);
      }
      
      config = ConfigManager.getInstance().resolveWidgetConfig(WidgetId.nextID(), BaseConfig.class);
   }

   /**
    * Get the root panel.
    * 
    * @return   the {@link #rootPanel}.
    */
   public BorderedPanel<O> getRootPanel()
   {
      return rootPanel;
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, BaseConfig cfg)
   {
      // no-op
   }

   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public BaseConfig config()
   {
      return config;
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(BaseConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
   }

   /**
    * The main renderer for this container. Draws the child container and
    * vertical/horizontal scroll-bars as necessary.
    */
   @Override
   public abstract void draw();

   /**
    * The secondary renderer for this container. Draws the child container and
    * vertical/horizontal scroll-bars as necessary. Note that this routine
    * assumes that it is invoked from Frame class.
    * 
    * @param   list
    *          Array of components to draw.
    */
   public abstract void draw(Widget<O>[] list);

   /**
    * Getting the wrapped component from the scroll pane. Used for the
    * selection list widget to request the widget handle.
    * 
    * @return  The handle of the component wrapped by the scroll pane.
    */
   public Widget<O> getChild()
   {
      return viewport().getScrollWidget();
   }

   /**
    * Checks whether this attribute container can receive focus generally.
    * 
    * @return  <code>true</code> if this container can be focused under
    *          certain conditions. <code>false</code> if, and only if, this
    *          attribute container can never receive focus.    
    */
   @Override
   public boolean isFocusable()
   {
      return false;
   }

   /**
    * Get widget minimum size.
    * 
    * @return  widget minimum size.
    */
   @Override
   public Dimension minimumSize()
   {
      return rootPanel.minimumSize();
   }

   /**
    * This method is invoked when scroll pane is scrolled.
    * 
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void handleScrollEvent(ScrollEvent event)
   {
      viewport().getScrollWidget().scroll(event.newX(), event.newY());
   }

   /**
    * Setting the location of this component.
    * 
    * @param   point
    *          The Point object to be used as the new location of the
    *          container.
    */
   @Override
   public void setLocation(Point point)
   {
      if (label != null)
         point.x += label.length() + 1;
      setLocation(point.x, point.y);
   }

   /**
    * Set container size.
    * 
    * @param   width
    *          New width.
    * @param   height
    *          New height.
    */
   @Override
   public void setSize(double width, double height)
   {
      width = Coordinate.scale(width);
      height = Coordinate.scale(height);
      
      super.setSize(width, height);
      rootPanel.setSize(width, height);
   }

   /**
    * Get pane viewport.
    * 
    * @return  pane viewport.
    */
   public Viewport<O> viewport()
   {
      return viewport;
   }

   /**
    * Get the effective border implementation.
    * 
    * @return  A {@link Border} reference, never <code>null</code>. 
    */
   public Border<O> getBorder()
   {
      return rootPanel.border();
   }
   
   /**
    * Set a border implementation.
    * 
    * @param   border
    *          A {@link Border} reference or <code>null</code> to clear the border. 
    */
   public void setBorder(Border<O> border)
   {
      rootPanel.setBorder(border == null ? new EmptyBorder<O>() : border);
   }
   
   /**
    * Lay out container.
    */
   @Override
   public void doLayout()
   {
      // just forward to the root panel
      rootPanel.doLayout();
   }

   /**
    * Calculates and returns initial viewport size. The method doesn't
    * consider scroll bars and always returns the full viewport size.
    
    * @return  Initial viewport dimension.
    */
   public Dimension calcInitialViewportSize()
   {
      return dimension();
   }
   
   /**
    * Calculate horizontal thumbnail position.
    * 
    * @param   childSize
    *          The size of the child pane.
    * @param   extentSize
    *          The size of the viewport.
    * @param   viewPosition
    *          The current view position.
    * 
    * @return  The horizontal thumbnail position.
    */
   protected double calcHscroll(Dimension childSize, Dimension extentSize, Point viewPosition)
   {
      double offset = (extentSize.width - 2) * (-viewPosition.x)
                   / (childSize.width - extentSize.width);
      if (offset == 0)
         offset = 1;
      return Coordinate.scale(offset);
   }

   /**
    * Calculate vertical thumbnail position.
    * 
    * @param   childSize
    *          The size of the child pane.
    * @param   extentSize
    *          The size of the viewport.
    * @param   viewPosition
    *          The current view position.
    *          
    * @return  The vertical thumbnail position.
    */
   protected double calcVscroll(Dimension childSize, Dimension extentSize, Point viewPosition)
   {
      double offset = 1;
   
      childSize = Coordinate.scale(childSize);
      extentSize = Coordinate.scale(extentSize);
      viewPosition = Coordinate.scale(viewPosition);
      
      if (childSize.height > extentSize.height)
      {
         offset = (extentSize.height - 2) * (-viewPosition.y)
                  / (childSize.height - extentSize.height);
      }
      else if (viewport().getScrollWidget() instanceof ScrollableList)
      {
         offset = ((ScrollableList<O>) viewport().getScrollWidget()).getCurrentRow() + 1;
      }
   
      if (offset <= 0)
      {
         offset = 1;
      }

      if (offset > extentSize.height - 2)
      {
         offset = extentSize.height - 2;
      }

      return offset;
   }

   /**
    * Set container scroll bars.
    * 
    * @param   bars
    *          Scroll bars.
    */
   public void setScrollBars(Bars bars)
   {
      // Not used for ChUI. Override for GUI
   }
}
