/*
** Module   : ScrollableWidget.java
** Abstract : Interface marking a scrollable widget.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 CA  20150509 Created initial version.
** 002 HC  20150831 Scrolling support overhaul.
** 003 GES 20150923 Javadoc fix.
** 004 VIG 20160420 Added isScrollPopupSupported method.
** 005 SVL 20160412 Javadoc fix.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.client.widget.*;

/**
 * Interface which needs to be implemented widgets which require scrolling 
 * capability.
 */
public interface ScrollableWidget<O extends OutputManager<?>>
extends Widget<O>
{
   /**
    * Returns the scroll dimension of this widget. The scroll dimension is
    * the full size of the widget's content, it may be bigger than widget's
    * visible size.
    * <p>
    * The returned dimension doesn't need to be in the same units as the 
    * widget's physical size but it must be compatible with the result
    * of {@link #getVisibleDimension()} and {@link #getScrollStep()}.
    * 
    * @return   widget's scroll dimension
    */
   NativeDimension getScrollDimension();
   
   /**
    * Returns the visible dimension of this widget. The visible dimension is
    * the visible portion of the widget's content (including scrollbars area).
    * <p>
    * The returned dimension doesn't need to be in the same units as the 
    * widget's physical size but it must be compatible with the result
    * of {@link #getScrollDimension()} and {@link #getScrollStep()}.
    *
    * @return   widget's visible dimension
    */
   NativeDimension getVisibleDimension();
   
   /**
    * Returns the dimension of the base scroll step desired by the widget
    * in the X and Y directions.
    * <p>
    * The scroll step must be in the same units as the dimensions returned
    * by {@link #getScrollDimension()} and {@link #getVisibleDimension()}. 
    * 
    * @return   The dimension of the base scroll step.
    */
   NativeDimension getScrollStep();
   
   /**
    * Directs the widget to scroll to the specified X and Y positions.
    * It is the widget's responsibility to update its visuals accordingly.
    * Note that the position values are in the same units as the returned 
    * dimensions from {@link #getScrollDimension()}, {@link #getVisibleDimension()}
    * and {@link #getScrollStep()}. 
    * 
    * @param    newXPosition
    *           New X scroll position.
    * @param    newYPosition
    *           New Y scroll position.
    */
   void scroll(Integer newXPosition, Integer newYPosition);

   /**
    * Check if this widget that supports scrolling also supports scroll popup.
    *
    * @return  <code>true</code> if this widget supports scroll popup.
    */
   boolean isScrollPopupSupported();
}