/* Module   : Separator.java
** Abstract : Separator artificial widget implementation for CHUI
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 EVL 20050627   @21556 Creation initial version based on abstract Component class.
** 002 SIY 20050831   @22549 Formatting fixes.
** 003 SIY 20100619          Switching to own implementation of character UI.
** 004 SIY 20100804          Changes in imports, some refactorings.
** 005 SIY 20110112          Changes related to new color handling scheme.
** 006 SIY 20111202          Refactored and moved from CHUI package.
** 007 CA  20140805          Added generics to force the abstract widget classes to not use
**                           implementation-specific code: the abstract widgets will have access
**                           only to the base APIs provided by the OutputManager.
** 008 HC  20140911          GUI coordinate support - data types refactoring.
** 009 HC  20141021          Implemented window move/position support.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.Coordinate;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * The virtual widget to represent the horizontal line to separate the widgets
 * from it's labels for default frame layout manager.
 */
public abstract class Separator<O extends OutputManager<?>>
extends AbstractWidget<O>
{
   /** The width of the separator. */
   private double width = 1;
   
   /**
    * Constructs the separator horizontal line with the given width.
    * 
    * @param   width
    *          The width of the separator.
    */
   public Separator(double width)
   {
      super();
      
      width = Coordinate.scale(width);
      if (width > 1)
         setWidth(width);
   }
   
   /**
    * Returns whether this widget will accept the keyboard focus.
    * 
    * @return  <code>false</code> 
    */
   @Override
   public boolean focusTraversable()
   {
      return false;
   }
   
   /**
    * Sets the width of the separator widget.
    * 
    * @param   width
    *          The width of the separator.
    */
   public void setWidth(double width)
   {
      this.width = Coordinate.scale(width);
   }
   
   /**
    * Returns the current height of the separator object.
    * 
    * @return  The vertical size of the separator object, always equal to 1.
    */
   @Override
   public double height()
   {
      return 1;
   }
   
   /**
    * Returns the current width of the separator object.
    * 
    * @return  The horizontal size of the separator object.
    */
   @Override
   public double width()
   {
      return width;
   }
   
   /**
    * Getting the size of the separator object. This is the implementation of
    * the abstract methods for the CHARVA Component.
    * 
    * @return  The size of the space object.
    */
   @Override
   public Dimension dimension()
   {
      return new Dimension(width(), height());
   }
   
   /**
    * The same as the getSize() method. This is the implementation of the
    * abstract methods for the CHARVA Component.
    * 
    * @return  The size of the separator object.
    */
   @Override
   public Dimension minimumSize()
   {
      return dimension();
   }

   /**
    * Draw method to render the horizontal separator line. This is the
    * implementation of the abstract methods for the CHARVA Component.
    */
   @Override
   public abstract void draw();
}