/*
** Module   : Slider.java
** Abstract : slider widget implementation, common part.
**
** Copyright (c) 2016-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20160127 Initial version.
** 002 HC  20161018 Added an initial support for rendering of mnemonic labels to all labeled
**                  widgets.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * An implementation of SLIDER widget.
 */
@SuppressWarnings("unchecked")
public abstract class Slider<O extends OutputManager<?>>
extends AbstractWidget<O>
implements WidgetWithConfig<SliderConfig>,
           DataContainer,
           LabeledWidget<O>,
           MnemonicTrait<O>
{
   /** The increment value for slider mouse handler. */
   protected static final int INCREMENT = 4;
   
   /** Slider attributes. */
   protected SliderConfig config = null;
   
   /** Component label. */
   private Label<O> label = null;

   /** Mnemonic information */
   private MnemonicInfo mnemonic = null;

   /** Column mnemonic information */
   private MnemonicInfo columnMnemonic = null;

   /**
    * This constructor creates a rectangle.
    * 
    * @param   owner
    *          The parent container, frame which will include this
    *          rectangle.
    * @param   id
    *          The widget's ID.
    */
   public Slider(WidgetId id, Container<O> owner)
   {
      this.config = ConfigManager.getInstance().resolveWidgetConfig(id, SliderConfig.class);
   }

   /**
    * Special constructor used to restore component using given configuration.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public Slider(WidgetId id, SliderConfig cfg)
   {
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, SliderConfig cfg)
   {
      config = ConfigManager.getInstance().resolveWidgetConfig(id, cfg);
      ConfigManager.getInstance().replaceConfig(config, cfg);
   }

   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public SliderConfig config()
   {
      return config;
   }

   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(SliderConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);

      updateMnemonic(beforeUpdate.clabel, config.clabel, true);
      updateMnemonic(beforeUpdate.label, config.label, false);
   }
   
   /**
    * Get widget dimension.
    * 
    * @return  widget dimension.
    */
   @Override
   public Dimension dimension()
   {
      return new Dimension(width(), height());
   }

   /**
    * Draw the rectangle. Called by this rectangle's parent container.
    */
   @Override
   public abstract void draw();
   
   /**
    * Calculate the height of this widget.
    *
    * @return    The height in rows.
    */
   @Override
   public double height()
   {
      return config.heightChars;
   }
   
   /**
    * Get widget minimum size.
    * 
    * @return  widget minimum size.
    */
   @Override
   public Dimension minimumSize()
   {
      return dimension();
   }
   
   /**
    * Calculate the width of this widget.
    *
    * @return    The width in characters.
    */
   @Override
   public double width()
   {
      return config.widthChars;
   }
   
   /**
    * This component will not receive an input focus when Tab or Shift-Tab is pressed.
    * 
    * @return  Always <code>false</code>.
    */
   @Override
   public boolean focusTraversable()
   {
      return isEnabled() && isVisible();
   }   

   // implemenations of the LabeledWidget interface - start
   /**
    * Return current value of COLON option.
    * 
    * @return  Current value of the option column.
    */
   @Override
   public double getColon()
   {
      return ConfigHelper.getColon(config);
   }
   
   /**
    * Remember the widget was made visible at least once.
    */
   @Override
   public void setWasVisible()
   {
      return;
   }
   
   /**
    * Checking if the widget was made visible at least once.
    * 
    * @return  <code>true</code> if the widget was visible,
    *          <code>false</code> otherwise.
    */
   @Override
   public boolean wasVisible()
   {
      return true;
   }
   
   /**
    * Get current state of the NO-LABELS flag.
    * 
    * @return   The current state of the flag.
    */
   @Override
   public boolean isNoLabels()
   {
      return !config.labels;
   }

   /**
    * Force NO-LABELS flag.
    * 
    * @param    labels
    *           The new value of the flag.
    */
   @Override
   public void setNoLabels(boolean labels)
   {
      config.labels = !labels;
   }
   
   /**
    * Return text of the column label.
    * 
    * @return  text of the column label.
    */
   @Override
   public String getColumnLabelText()
   {
      return getMnemonicLabel(true).orElse(config.clabel);
   }
   
   /**
    * Return text of the label.
    * 
    * @return  text of the label.
    */
   @Override
   public String getLabelText()
   {
      return getMnemonicLabel(false).orElse(config.label);
   }

   /**
    * Return TO column.
    * 
    * @return  Column assigned via TO option.
    */
   @Override
   public double getTo()
   {
      return ConfigHelper.getTo(config);
   }

   /**
    * Get current state of flag which is forces presence of the label
    * regardless from the frame NO-LABELS option.
    * 
    * @return  Current state of the flag.
    */
   @Override
   public boolean isForceLabel()
   {
      return config().forceLabel;
   }
   
   /**
    * Return the reference to the instance of the <code>Label</code> attached
    * to the widget.
    * 
    * @return   Reference to <code>Label</code> instance.
    */
   @Override
   public Label<O> getLabelInstance()
   {
      return label;
   }
   
   /**
    * Assign new instance of <code>Label</code> to the widget.
    * 
    * @param   label
    *          Reference to new <code>Label</code> instance.
    */
   @Override
   public void setLabelInstance(Label<O> label)
   {
      if (label != this.label && this.label != null)
      {
         this.label.destroy();
      }
      
      if (label == null)
      {
         this.label = null;
         return;
      }

      int frameId = ThinClient.frameIdFromWidgetId(this.config.id.asInt());
      
      // link the label to the frame of this widget
      label.linkTo(frameId, config().sideLabelId);

      this.label = label;
      this.label.setVisible(isVisible());
      this.label.setHidden(hidden());
   }

   /**
    * Get widget alignment policy.
    * 
    * @return   <code>true</code> if the widget is formatted for right
    *           alignment.
    */
   @Override
   public boolean isRightAligned()
   {
      return false;
   }
   // implemenations of the LabeledWidget interface - end
   
   // implementation of the DataContainer - start
   /**
    * Gets the current slider value.
    *
    * @return   The integer representing the current value of the slider.
    */
   @Override
   public BaseDataType getValue()
   {
      return new int64(config.currentValue);
   }
   
   /**
    * Sets the instance representing the current value.
    *
    * @param    value
    *           The value representing the current widget value.
    */
   @Override
   @SuppressWarnings("unchecked")
   public void setValue(BaseDataType value)
   {
      if (value instanceof int64 && value != null && !value.isUnknown())
      {
         config.currentValue = ((int64)value).intValue();
      }
   }

   /**
    * Get text representation of the current slider value.
    * 
    * @return   The screen representation of the value or <code>""</code>
    *           if no screen value exists.
    */
   @Override
   public String getText()
   {
      // try to obtain the screen value in a string format
      String res = getScreenValueAsString();
      
      return res != null ? res : "";
   }
   
   /**
    * Getting the modification state of the widget.
    * 
    * @return  <code>true</code> if the widget value was modified,
    *          <code>false</code> otherwise.
    */
   @Override
   public boolean isModified()
   {
      return config.modified;
   }
   // implementation of the DataContainer - end

   /**
    * Process key events that have been generated by this component.
    * 
    * @param    keyEvent
    *           The key event to be handled.
    */
   @Override
   public void processKeyEvent(KeyInput keyEvent)
   {
      // check the key pressed
      int key = keyEvent.actionCode();

      if (key == Keyboard.KA_TAB)
      {
         // the following call must be conditional
         ThinClient.nextTabStop(this);
      }
      else if (key == Keyboard.KA_BACK_TAB)
      {
         // the following call must be conditional
         ThinClient.previousTabStop(this);
      }
      else if ((key == Keyboard.KA_CURSOR_LEFT || key == Keyboard.KA_CURSOR_RIGHT) &&
               config.horizontal)
      {
         // change the slider value by keyboard
         int delta = key == Keyboard.KA_CURSOR_RIGHT ? 1 : -1;
         // check LARGE-TO-SMALL option to invert delta
         if (config.largeToSmall)
         {
            delta *= -1;
         }
         changeScreenValue(delta);
      }
      else if ((key == Keyboard.KA_CURSOR_UP || key == Keyboard.KA_CURSOR_DOWN) &&
               !config.horizontal)
      {
         // change the slider value by keyboard
         int delta = key == Keyboard.KA_CURSOR_UP ? 1 : -1;
         // check LARGE-TO-SMALL option to invert delta
         if (config.largeToSmall)
         {
            delta *= -1;
         }
         changeScreenValue(delta);
      }
      else if (key == Keyboard.KA_HOME)
      {
         // change the slider value by keyboard, move to the left or top edge
         // both orientation and LARGE-TO-SMALL options should be taken into account
         int delta = (config.horizontal && !config.largeToSmall ||
                      !config.horizontal && config.largeToSmall)
                        ? config.minValue - config.currentValue
                        : config.maxValue - config.currentValue;
         changeScreenValue(delta);
      }
      else if (key == Keyboard.KA_END)
      {
         // change the slider value by keyboard, move to the right or bottom edge
         // both orientation and LARGE-TO-SMALL options should be taken into account
         int delta = (config.horizontal && !config.largeToSmall ||
                      !config.horizontal && config.largeToSmall)
                        ? config.maxValue - config.currentValue
                        : config.minValue - config.currentValue;
         changeScreenValue(delta);
      }
      else if (key == Keyboard.KA_PAGE_UP)
      {
         // change the slider value by keyboard, move fixed steps to the left or top edge
         // both orientation and LARGE-TO-SMALL options should be taken into account
         int delta = (config.horizontal && !config.largeToSmall ||
                      !config.horizontal && config.largeToSmall)
                        ? -INCREMENT : INCREMENT;
         changeScreenValue(delta);
      }
      else if (key == Keyboard.KA_PAGE_DOWN)
      {
         // change the slider value by keyboard, move fixed steps to the right or bottom edge
         // both orientation and LARGE-TO-SMALL options should be taken into account
         int delta = (config.horizontal && !config.largeToSmall ||
                      !config.horizontal && config.largeToSmall)
                        ? INCREMENT : -INCREMENT;
         changeScreenValue(delta);
      }
      else
      {
         super.processKeyEvent(keyEvent);
      }
   }

   /**
    * Get widget screen value as int.
    * 
    * @return  Current widget value as int number.
    */
   public int getScreenValueAsInt()
   {
      return config.currentValue;
   }
   
   /**
    * Get widget value range as int.
    * 
    * @return  Current widget value range as int number.
    */
   public int getRange()
   {
      return config.maxValue - config.minValue + 1;
   }
   
   /**
    * Get widget screen value in String format.
    * 
    * @return  Current widget screen value to display.
    */
   public String getScreenValueAsString()
   {
      return String.valueOf(config.currentValue);
   }

   /**
    * MnemonicInfo instance getter.
    * <p>
    * The implementors should simply return the value of the mnemonic class field. No
    * additional processing logic is advised.
    *
    * @return   See above.
    */
   public MnemonicInfo getMnemonic()
   {
      return mnemonic;
   }

   /**
    * MnemonicInfo instance setter.
    * <p>
    * The implementors should simply assign the parameter value to the mnemonic class
    * field. No additional processing logic is advised.
    */
   public void setMnemonic(MnemonicInfo mnemonic)
   {
      this.mnemonic = mnemonic;
   }

   /**
    * Column {@link MnemonicInfo} instance getter.
    * <p>
    * The implementors should simply return the value of the mnemonic class field. No
    * additional processing logic in the implemented method is advised.
    * <p>
    * The method will return a valid value when the implementing widget supports column label.
    *
    * @return   See above.
    */

   public MnemonicInfo getColumnMnemonic()
   {
      return columnMnemonic;
   }

   /**
    * Column {@link MnemonicInfo} instance setter.
    * <p>
    * The implementors should simply assign the parameter value to the mnemonic class
    * field. No additional processing logic in the implemented method is advised.
    */
   public void setColumnMnemonic(MnemonicInfo mnemonic)
   {
      this.columnMnemonic = mnemonic;
   }

   /**
    * Change the current value of the slider.
    * 
    * @param    delta
    *           The delta value to change the current one.
    */
   protected void changeScreenValue(int delta)
   {
      // consider the new possible value
      int newValue = config.currentValue + delta;
      // and check if it is valid
      if (newValue < config.minValue)
      {
         newValue = config.minValue;
      }
      if (newValue > config.maxValue)
      {
         newValue = config.maxValue;
      }
      if (newValue != config.currentValue)
      {
         // commit the new value
         config.currentValue = newValue;
         
         // honor any Progress triggers
         KeyInput vchng = new KeyInput(this, EventType.KEY_PRESSED,
                                       Keyboard.SE_VALUE_CHANGED, true);
         EventManager.postEvent(vchng);
         // the additional call below simulates strange 4GL behavior(loos like a bug)
         // when value changed event is coming twice to the change listener
         // may be need to remove this, do we neen to simulate a buggy logic?
         EventManager.postEvent(vchng);
      
         ConfigHelper.setModified(config, true);
         // refresh the slider
         repaint();
      }
   }
}