/*
** Module   : StaticCoordinatesConversionParams.java
** Abstract : Resolution of coordinate conversion parameters based on provided 
**            output manager instance.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 HC  20141021 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.driver.ScreenDriver;

/**
 * The class resolves coordinate conversion parameters based on provided (thus constant) 
 * {@link OutputManager} instance.
 * <p>
 * The conversion parameters are read from the {@link OutputManager} instance
 * when any of the conversion methods is called for the first time. Except for 
 * {@link #getFontWidth(WidgetId, int)} and {@link #getFontHeight(WidgetId, int)}, 
 * the read parameters are cached locally in the instance.
 */
public class StaticCoordinatesConversionParams 
extends ClientCoordinatesConversionParams
{
   /**
    * The static instance provided by this {@link CoordinatesConversionParams} implementation.
    */
   private OutputManager<?> outputManager;
   
   /**
    * Lazy-initialization flag.
    */
   private boolean inited;
   
   /** 
    * The isChui flag. This value never changes and thus can be cached.
    */
   private boolean isChui;
   
   /**
    * Pixels-per-column of the system font. This value never changes and thus can be cached.
    */
   private int pixelsPerColumn;
   
   /**
    * Pixels-per-row of the system font. This value never changes and thus can be cached.
    */
   private int pixelsPerRow;
   
   /**
    * Constructs new instance.
    * 
    * @param   outputManager
    *          The static {@link OutputManager} reference.
    */
   public StaticCoordinatesConversionParams(OutputManager<?> outputManager)
   {
      super();
      this.outputManager = outputManager;
   }
   
   /**
    * Performs lazy initialization.
    */
   private void init()
   {
      if (!inited)
      {
         this.isChui = this.outputManager.isChui();
         
         if (this.isChui)
         {
            this.pixelsPerColumn = CHUI_PIXELS_PER_COLUMN;
            this.pixelsPerRow = CHUI_PIXELS_PER_ROW;
         }
         else
         {
            ScreenDriver<?> driver = this.outputManager.getInstanceDriver();
            this.pixelsPerColumn = FontManager.getPixelsPerColumn(driver);
            this.pixelsPerRow = FontManager.getPixelsPerRow(driver);
         }
         
         inited = true;
      }
   }
   
   /**
    * Flag indicating the ChUI or GUI mode.
    * 
    * @return  Boolean value indicating the ChUI or GUI mode.
    */
   @Override
   public boolean isChui()
   {
      init();
      return isChui;
   }

   /**
    * Returns the number of pixels in each column of the display. This usually corresponds
    * to the 4GL attribute SESSION:PIXELS-PER-COLUMN.
    * 
    * @return  Number of pixels in each column.
    */
   @Override
   public int getPixelsPerColumn()
   {
      init();
      return pixelsPerColumn;
   }

   /**
    * Returns the number of pixels in each row of the display. This usually corresponds
    * to the 4GL attribute SESSION:PIXELS-PER-ROW.
    * 
    * @return  Number of pixels in each row.
    */
   @Override
   public int getPixelsPerRow()
   {
      init();
      return pixelsPerRow;
   }

   /**
    * Returns a valid reference to a static {@link OutputManager} instance.
    * The implementation returns the instance which was used during construction
    * of this object.
    *
    * @return  A valid reference to {@link OutputManager} instance.
    */
   @Override
   protected OutputManager<?> getOutputManager()
   {
      return outputManager;
   }
}