/*
** Module   : WindowManager.java
** Abstract :
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 GES 20111012 Moved static data to context-local. This allows multiple client sessions to be 
**                  hosted in the server.
** 004 SIY 20111202 Updated imports.
** 005 CA  20140805 Changes related to multi-window support and GUI windows.
** 006 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 007 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 008 CA  20141123 First pass at window clearing during client restart in GUI.  Added support 
**                  for default location and dimension, for dynamic windows.
** 009 CA  20141125 When the window is destroyed, this needs to be done outside of the synchronized
**                  block. 
** 010 CA  20150408 Added API to compute the children for a window.
**                  Keep the window's original order, so that the window tree can be identified 
**                  correctly.
** 011 HC  20150323 Removed business logic from config classes.
** 012 VIG 20150401 Fixed null child focus.
** 013 EVL 20150424 Exclude combo-box drop down from the widgow list returned.
** 014 HC  20150508 Improvements to GUI window scrolling support.
** 015 CA  20150520 Hide all windows before destroying them, on CTRL-C.
** 016 HC  20150702 Implemented window modality independently on the actual UI driver.
** 017 EVL 20150623 Windows ID array should not try to include the drop-down.
** 018 HC  20150803 Improved runtime support for WINDOW size attributes.
** 019 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
**                  Changes to allow multi-window focus management in GUI.
** 020 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 021 GES 20151015 moveToTop()/moveToBottom() forces the driver to move the specified window.
** 022 SBI 20151103 Changed clearWindowList() to dispose the default window and to clear windows
**                  following the reverse order of their creations.
**     CA  20151112 If CURRENT-WINDOW is explicitly set, resolveWindow() needs to always return 
**                  this, regardless of CURRENT-WINDOW's visible/hidden state.
** 023 HC  20151213 Improved DIALOG-BOX window parenting logic and related changes.
** 024 EVL 20151223 Adding method to get the overlay window from the current window set.
**     HC  20160113 Changed activateWindow() to also unconditionally move the activating window
**                  to top.
** 025 EVL 20160122 Adding logic to differentiate lower layer activation from higher layer. 
** 026 HC  20160203 Improved runtime support of KEEP-FRAME-Z-ORDER and TOP-ONLY attributes.
** 027 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 028 HC  20160304 Replaced Window.moveBy() with setPhysicalLocation(). Fixed unhandled
**                  exception in resolveTopLevelWindow().
** 029 SBI 20160316 Set the focused window for focused widget requests.
** 030 HC  20160406 Overhaul of window-activation logic. Removed all references to OverlayWindow
**                  and ModalWindow, instead abstract methods are queried to establish window
**                  capabilities.
** 031 EVL 20160411 Javadoc fixes.
** 032 HC  20160414 Javadoc fixes.
** 033 HC  20160424 Fixed unexpected ClassCastException in moveToTop() in ChUI.
** 034 CA  20160429 Fixed removeWindow - if the deleted window is active/current-window, then
**                  set the active/current-window reference to default-window.
** 035 HC  20160512 Prevented ClassCastException.
** 036 HC  20160607 GUI menus reworked to be displayed in dedicated top-level windows.
** 037 HC  20160628 Prevented a modal window to be moved higher in the z-order over another
**                  active modal window.
** 038 SBI 20160630 Applied HC changes to fix a next active window, 037 changes are not required,
**                  thus they are rolled back.
** 039 HC  20160701 Fixed z-order on new modal windows.
** 040 HC  20160712 Prevented an NPE when last DIALOG-BOX closed.
** 041 SBI 20160811 The alert box owner should be CURRENT-WINDOW.
** 042 HC  20160926 Removed disabling of window input events during server processing.
** 043 HC  20170202 Improved the window-activation process. This addresses complex window
**                  activation cases like multi-level menus.
**     HC  20170203 Renamed TC.invokeLater() to TC.invokeLaterOS() and introduced
**                  TC.invokeLater() to allow execution of code to be queued on the application
**                  queue.
** 044 OM  20170320 Theme support. Repainted deactivatedWindow's border and titlebar.
** 045 OM  20170502 Additional support for Windows8Theme added.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;
import java.util.function.*;
import java.util.stream.*;

import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.GuiDriver;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Window manager.
 */
public class WindowManager
{
   /** Stores context-local state variables. */
   private static final ContextLocal<WorkArea> work = 
   new ContextLocal<WorkArea>()
   {
      protected synchronized WorkArea initialValue()
      {
         return new WorkArea();
      }
   };
   
   /** Context local instance of this class. */
   private static final ContextLocal<Window<?>> defaultWnd = new ContextLocal<>();

   /** Context local instance of this class. */
   private static final ContextLocal<Window<?>> currentWnd = new ContextLocal<>();

   /** 
    * The window referenced as the ACTIVE-WINDOW, in other words the last window 
    * to receive ENTRY event. Do not use this field (or ACTIVE-WINDOW) as indication 
    * of the active OS window! It is a valid situation to have no P2J windows active 
    * at the OS while a valid window in ACTIVE-WINDOW.
    */
   private static final ContextLocal<Window<?>> activeWnd = new ContextLocal<>();

   /** Context local instance of this class. */
   private static final ContextLocal<TopLevelWindow<?>> focusWnd = new ContextLocal<>();
      
   /** The active window as reported by the driver. */
   private static final ContextLocal<TopLevelWindow<?>> activeDriverWindow = new ContextLocal<>();
   
   /**
    * Prevent creation from outside of this class.
    */
   private WindowManager()
   {
   }
   
   /**
    * Find the next visible window, in the specified direction.
    * 
    * @param    window
    *           The reference window.
    * @param    next
    *           When <code>true</code>, search to the right; otherwise, search to the left of the 
    *           reference window.
    *           
    * @return   The nearest visible window in the specified direction or <code>null</code> if 
    *           none exists.
    */
   public static Window<?> findSiblingWindow(Window<?> window, boolean next)
   {
      WorkArea wa = work.get();

      synchronized (wa.windows)
      {
         List<TitledWindow<?>> wins = new ArrayList<>();
         wins.addAll(wa.fixedOrderWindows);
         
         if (next)
         {
            Collections.reverse(wins);
         }

         Window<?> prev = null;
         for (TitledWindow w : wins)
         {
            if (w == window)
            {
               return prev;
            }
            
            if (w.isVisible())
            {
               prev = (Window<?>) w;
            }
         }
      }
      
      return null;
   }

   /**
    * Compute the child hierarchy for the specified window.
    * 
    * @param    parentId
    *           The ID of the parent window.
    * @param    reverse
    *           When <code>true</code>, the window list is walked in reverse order.
    *           
    * @return   The list of child windows, in a depth-first search.
    */
   public static List<TitledWindow<?>> findChildren(int parentId, boolean reverse)
   {
      WorkArea wa = work.get();
      List<TitledWindow<?>> result = new ArrayList<>();
      
      // this lock is used for both WorkArea.windows and WorkArea.fixedOrderWindows fields
      synchronized (wa.windows)
      {
         // use a depth-first search; we can't use the WorkArea.windows, as this does not provide
         // a stable order (as this is used for the windows' z-order). 
         // instead, the resources can be walked using their creation order, as fortunately, the
         // WINDOW's FIRST-CHILD attribute depends on this, and not on the order on which the
         // PARENT attribute is set.

         List<TitledWindow<?>> wins = new ArrayList<>();
         wins.addAll(wa.fixedOrderWindows);
         if (reverse)
         {
            Collections.reverse(wins);
         }

         for (TitledWindow<?> tw : wins)
         {
            WidgetConfig wcfg = tw.config();
            
            if (wcfg instanceof BaseConfig && ((BaseConfig) wcfg).parentId == parentId)
            {
               result.addAll(findChildren(wcfg.id.asInt(), reverse));
               result.add(tw);
            }
         }
      }
      
      return result;
   }

   /**
    * Add top level window.
    * 
    * @param   window
    *          Window to add.
    */
   @SuppressWarnings("rawtypes")
   public static void addWindow(TitledWindow window)
   {
      WorkArea wa = work.get();
      
      synchronized (wa.windows)
      {
         if (!wa.windows.contains(window))
         {
            wa.windows.add(window);
            wa.fixedOrderWindows.add(window);
         }
      }
   }
   
   /**
    * Move the specified window to the bottom, behind any other windows.
    * 
    * @param    window
    *           The window to be moved.
    */
   @SuppressWarnings("rawtypes")
   public static void moveToBottom(TitledWindow window)
   {
      WorkArea wa = work.get();
      
      if (window == null)
      {
         throw new NullPointerException("The window can not be null!");
      }
      
      synchronized (wa.windows)
      {
         wa.windows.remove(window);
         wa.windows.add(0, window);
      }
      
      GuiDriver<?,?> gd = (GuiDriver<?,?>) OutputManager.getDriver();
      gd.moveToBottom(window.getId().asInt());
   }
   
   /**
    * Move the specified window to the top, above any other windows.
    * 
    * @param    window
    *           The window to be moved.
    */
   @SuppressWarnings("rawtypes")
   public static void moveToTop(TitledWindow window)
   {
      WorkArea wa = work.get();
      
      if (window == null)
      {
         throw new NullPointerException("The window can not be null!");
      }
      
      boolean moveToTop = true;
      
      synchronized (wa.windows)
      {
         wa.windows.remove(window);
         
         int newPos = wa.windows.size();
         
         WindowConfig windowCfg = UiUtils.getConfig(window, WindowConfig.class);
         if (windowCfg != null && !windowCfg.topOnly && !window.isModal())
         {
            // iterate the registered windows from the most top to bottom...
            for (int i = newPos; i > 0; i--)
            {
               TitledWindow<?> w = wa.windows.get(i - 1);
               WindowConfig wCfg = UiUtils.getConfig(w, WindowConfig.class);
               
               // and skip visible top-only windows and 
               if (wCfg != null && w.isVisible() && wCfg.topOnly)
               {
                  moveToTop = false;
                  continue;
               }
               
               newPos = i;
               break;
            }
         }
         
         wa.windows.add(newPos, window);
      }
      
      ScreenDriver<?> driver = OutputManager.getDriver();
      
      if (driver instanceof GuiDriver<?,?>)
      {
         GuiDriver<?,?> gd = (GuiDriver<?, ?>) driver;
         if (moveToTop)
         {
            gd.moveToTop(window.getId().asInt(), false);
         }
         else
         {
            gd.stackWindows(topLevelWindowIds());
         }
      }
   }
   
   /**
    * Activates the specified window with the driver. The driver 
    * responds with a {@link WindowActivated} event.
    * 
    * @param    window
    *           The window to be activated.
    */
   public static void forceActivateWindow(TitledWindow<?> window)
   {
      if (window == null)
      {
         throw new NullPointerException("The window can not be null!");
      }
      
      if (!window.isRealized())
      {
         // window is deregistered
         return;
      }
      
      // move window to top
      WorkArea wa = work.get();
      synchronized (wa.windows)
      {
         boolean removed = wa.windows.remove(window);
         if (!removed)
         {
            // the window is not registered in the first place,
            // abort
            return;
         }
         
         wa.windows.add(window);
      }
      
      ScreenDriver<?> d = OutputManager.getDriver();
      if (d instanceof GuiDriver<?,?>)
      {
         ((GuiDriver<?,?>) d).moveToTop(window.getId().asInt(), true);
      }
   }
   
   /**
    * Activates the specified window with the driver. The driver 
    * responds with a {@link WindowActivated} event.
    * <p>
    * The method will ignore the activation request if there is
    * another visible modal window higher in the z-order.
    * 
    * @param    window
    *           The window to be activated.
    */
   public static void activateWindow(TitledWindow<?> window)
   {
      if (window == null)
      {
         throw new NullPointerException("The window can not be null!");
      }
      
      if (!window.isRealized())
      {
         // window is deregistered
         return;
      }
      
      // a consistency check - ignore the request when there are 
      // other displayed modal windows higher in the z-order
      List<TitledWindow<?>> slice = windowList().stream()
               .filter(w -> (w.isVisible() && w.isModal()) || w == window)
               .collect(Collectors.toList());
      
      int sliceSize = slice.size();
      if (sliceSize > 0 && slice.get(sliceSize - 1) != window)
      {
         // there is a visible modal window higher in the z-order
         return;
      }
      
      forceActivateWindow(window);
   }
   
   /**
    * Find the window with the given ID.
    * 
    * @param    windowId
    *           The window ID.
    */
   @SuppressWarnings({ "rawtypes", "unchecked" })
   public static <T extends TitledWindow> T findWindow(int windowId)
   {
      return (T) windowList().stream()
               .filter((TitledWindow w) -> WidgetId.equals(w.getId(), windowId))
               .findFirst()
               .orElse(null);
   }

   /**
    * Delete all windows.
    */
   public static void clearWindowList()
   {
      WorkArea wa = work.get();
      
      @SuppressWarnings("rawtypes")
      List<TitledWindow> wins = new ArrayList<>();

      synchronized (wa.windows)
      {
         wins.addAll(wa.fixedOrderWindows);
      }
      Collections.reverse(wins);
      // notify the windows are being destroyed in the reverse to their creations
      for (TitledWindow<?> w : wins)
      {
         w.hide();
         if (w instanceof Window)
         {
            remove(w);
         }
         else 
         {
            w.destroy();
         }
      }

      synchronized (wa.windows)
      {
         wa.windows.clear();
         wa.fixedOrderWindows.clear();
      }
   }

   /**
    * Return top window in Z-order.
    * 
    * @return  top window or <code>null</code> if no windows are present.
    */
   @SuppressWarnings("rawtypes")
   public static TitledWindow topWindow()
   {
      // ChUI has a peculiarity here... it requires the top window (if is a message-box or 
      // something else) regardless of visibility state.
      boolean isChui = ThinClient.getInstance().isChui();

      return Utils.reverseStream(windowList())
               .filter((TitledWindow w) -> isChui || (w.isVisible() && !(w instanceof DropDown)))
               .findFirst()
               .orElse(null);
   }

   /**
    * Get the list of registered windows.
    * 
    * @return  see above.
    */
   public static List<TitledWindow<?>> windowList()
   {
      // we cannot return a wrapper of Collections.unmodifiableList() as
      // this would cause a race condition
      WorkArea wa = work.get();
      
      synchronized (wa.windows)
      {
         return new ArrayList<>(wa.windows);
      }
   }

   /**
    * Get the list of windows that are positioned under
    * the specified <code>underWindowId</code> window in the Z-order.
    * If the window <code>underWindowId</code> is not registered with
    * {@linkplain WindowManager} the result is an empty list.
    * 
    * @return  see above.
    */
   public static List<TitledWindow<?>> windowList(int underWindowId)
   {
      // we cannot return a wrapper of Collections.unmodifiableList() as
      // this would cause a race condition
      WorkArea wa = work.get();
      
      ArrayList<TitledWindow<?>> result = null;
      int windowsSize = 0;
      
      synchronized (wa.windows)
      {
         windowsSize = wa.windows.size();
         result = new ArrayList<>(windowsSize);
         
         for (TitledWindow<?> w : wa.windows)
         {
            if (WidgetId.equals(w.getId(), underWindowId))
            {
               break;
            }
            
            result.add(w);
         }
      }
      
      // if underWindowId not in wa.windows, return an empty List
      return result.size() == windowsSize ? Collections.emptyList() : result;
   }
   
   /**
    * Remove the window instance from the client.
    * 
    * @param   window
    *          The window ID.
    */
   public static void removeWindow(final int window)
   {
      TitledWindow<?> w = (TitledWindow<?>) WindowManager.findWindow(window);

      // the default window will be deleted only when the client exits/gets reset
      if (w != null && !w.config().id.equals(WidgetId.DEFAULT_WINDOW_ID))
      {
         // dynamic window, remove it from the window manager
         WindowManager.remove(w);
         
         Window<?> aw = activeWnd.get();
         Window<?> cw = currentWnd.get();
         Window<?> dw = defaultWnd.get();
         
         if (w == aw)
         {
            activeWnd.set(dw);
         }
         if (w == cw)
         {
            currentWnd.set(dw);
         }
      }
   }
   
   /**
    * Get the main window to be reported by the <code>ACTIVE-WINDOW</code> 
    * system handle, in other words the last window to receive ENTRY event.
    * In case there is not active window yet, the method returns the default
    * window.
    * <p>
    * Do not use the result of this method (or ACTIVE-WINDOW) as indication 
    * of the active OS window! It is a valid situation to have no P2J windows 
    * active at the OS while a valid window in ACTIVE-WINDOW.
    *  
    * @return   See above.
    */
   public static Window<?> getActiveWindow()
   {
      Window<?> active = activeWnd.get();
      
      // if the top-level active window is not the main window, return its
      // main window owner
      return active == null ? getDefaultWindow() : active.window();
   }

   /**
    * Set ACTIVE-WINDOW to the specified window.
    * 
    * @param    window
    *           The new window.
    */
   public static void setActiveWindow(Window<?> window)
   {
      activeWnd.set(window);
   }
   
   /** 
    * Returns list of active top-level windows. 
    * <p>
    * Note that there may exist multiple active top-level windows at a time. 
    * Consider the COMBO-BOX drop-down window or menu window cases.
    * For more information of the concept of active windows, see 
    * {@link #forEachActiveWindow(TopLevelWindow, Predicate)}.
    * 
    * @return   See above.
    */
   public static List<TopLevelWindow<?>> getActiveWindows()
   {
      TopLevelWindow<?> activeDriverWnd = activeDriverWindow.get();
      if (activeDriverWnd == null)
      {     
         return Collections.emptyList();
      }
      
      List<TopLevelWindow<?>> result = new ArrayList<>();  
      TopLevelWindow<?> candidate = activeDriverWnd;
      while (candidate != null)
      {
         result.add(candidate);
         
         if (!(candidate instanceof GuiWindow) || 
             !((GuiWindow) candidate).isShareActivationWithOwner())
         {
            // break if the candidate window doesn't share its activation status 
            // with its owner
            break;
         }
         
         candidate = candidate.getOwner();
      }

      return result;
   }
   
   /**
    * The method returns <code>true</code> if the supplied 
    * <code>window</code> is the active window.
    * <p>
    * For more information of the concept of active windows, see 
    * {@link #forEachActiveWindow(TopLevelWindow, Predicate)}.
    * 
    * @param    window
    *           A valid top-level window.
    *           
    * @return   See above.
    */
   public static boolean isWindowActive(TopLevelWindow<?> window)
   {
      boolean[] result = new boolean[] {false};
      forEachActiveWindow(activeDriverWindow.get(), w ->
      {
         if (w == window)
         {
            result[0] = true;
            return false;
         }
         
         return true;
      });
      
      return result[0];
   }
   
   /**
    * Get the top top-level window.
    * <p>
    * Note that this window may not be the same as the focus owner window - 
    * if the active window doesn't have any enabled/focusable widgets, 
    * focus window remains on the previous owner.
    * <p>
    * The method will return <code>null</code> if no visible top-level
    * window exists.
    * 
    * @return   See above.
    */
   @SuppressWarnings("unchecked")
   public static <O extends OutputManager<?>> TopLevelWindow<O> topLevelWindow()
   {
      return (TopLevelWindow<O>) topLevelWindowOptional().orElse(null);
   }
   
   /**
    * Get the top top-level window.
    * <p>
    * Note that this window may not be the same as the focus owner window - 
    * if the active window doesn't have any enabled/focusable widgets, 
    * focus window remains on the previous owner.
    * <p>
    * The method will return an empty {@linkplain Optional} if no visible 
    * top-level window exists. 
    * 
    * @return   See above.
    */
   @SuppressWarnings("unchecked")
   public static <O extends OutputManager<?>> 
   Optional<TopLevelWindow<O>> topLevelWindowOptional()
   {
      // the top window is last in the windows list
      return Utils.reverseStream(windowList())
               .filter(w -> w.isVisible() && w instanceof TopLevelWindow)
               .map(w -> (TopLevelWindow<O>) w)
               .findFirst();
   }
   
   /**
    * Resolve the window associated with the specified widget.
    * 
    * @param    widget
    *           The widget.
    *           
    * @return   The parent window.
    */
   public static Window<?> resolveWindow(AbstractWidget<?> widget)
   {
      TitledWindow<?> window = widget.ancestor();
      
      if (window instanceof Window && ((Window<?>) window).isVisible())
      {
         return (Window<?>) window;
      }
      else
      {
         return resolveWindow();
      }
   }

   /**
    * Resolve the top-level window associated with the specified widget.
    * 
    * @param    widget
    *           The widget.
    *           
    * @return   The parent window.
    */
   public static TopLevelWindow<?> resolveTopLevelWindow(AbstractWidget<?> widget)
   {
      @SuppressWarnings("unchecked")
      TopLevelWindow<?> window = widget.parentOrSelf(TopLevelWindow.class);
      
      if (window == null)
      {
         window = resolveWindow();
      }
      
      return window;
   }
   
   /**
    * Resolve the window targeted by the current UI statements.
    * 
    * @return    Returns current window if it is visible or default window.
    *            Default window in turn can be invisible. Further logic can be different
    *            depending on situation: ignoring, showing an alert-box or some other.
    *            This logic should be implemented in concrete caller of this method.
    */
   public static Window<?> resolveWindow()
   {
      Window<?> currentWindow = getCurrentWindow();
      
      // if changed, the current window is returned regardless if is visible or not
      if (currentWindow == null)
      {
         return getDefaultWindow();
      }
      
      // TODO not sure but thinking that ACTIVE-WINDOW should be also process here.
      // because it plays its role during ENTRY processing event - can be set from the client-side.
      
      // TODO not sure how to process THIS-PROCEDURE:CURRENT-WINDOW here:
      // Can I just call ProcedureManager.getCurrentWindow() or it will be an unwanted server call?
      
      return currentWindow;
   }

   /**
    * Get the window owning the focus.
    * 
    * @return   See above.
    */
   @SuppressWarnings("unchecked")
   public static <O extends OutputManager<?>> TopLevelWindow<O> getFocusWindow()
   {
      return (TopLevelWindow<O>) focusWnd.get();
   }
   
   /**
    * Set the focus owner to the specified window.
    * 
    * @param    window
    *           The window owning the focus. Pass <code>null</code> if no
    *           window should hold the focus.
    */
   public static void setFocusWindow(TopLevelWindow<?> window)
   {
      focusWnd.set(window);
   }
   
   /**
    * Returns default window handle. Creates it if necessary.
    * 
    * @return  The handle of the default and single Window class instance.
    */
   @SuppressWarnings("unchecked")
   public static synchronized <O extends OutputManager<?>> Window<O> getDefaultWindow()
   {
      Window<?> window = defaultWnd.get();
      
      if (window == null)
      {
         OutputManager<?> om = OutputManager.instance();
         window = om.getFactory().createWindow(WidgetId.DEFAULT_WINDOW_ID);

         defaultWnd.set(window);
         currentWnd.set(window);
         
         // add it to the window manager
         WindowManager.addWindow(window);
         
         // activate the window only after it has been added to window manager
         om.activateWindow(window);

         // ensure the window is targeted by the output manager
         om.switchOutput(null);
      }

      return (Window<O>) window;
   }

   /**
    * Returns current window handle. Creates it if necessary.
    * 
    * @return  The handle of the current window.
    */
   public static synchronized <O extends OutputManager<?>> Window<O> getCurrentWindow()
   {
      return (Window<O>) currentWnd.get();
   }
   
   /**
    * Set the current window handle.
    * 
    * @param   wnd
    *          The window which is assigned to CURRENT-WINDOW.
    */
   public static synchronized void setCurrentWindow(Window<?> wnd)
   {
      currentWnd.set(wnd);
   }
   
   /**
    * Resets the default window.
    */
   public static synchronized void resetWindow()
   {
      defaultWnd.set(null);
      currentWnd.set(null);
   }

   /**
    * Check if specified window is displayed, i.e. present in the list of 
    * windows.
    * 
    * @param   window
    *          Window to check.
    *          
    * @return  <code>true</code> if window is displayed.
    */
   @SuppressWarnings("rawtypes")
   public static boolean isDisplayed(Widget window)
   {
      WorkArea wa = work.get();
      
      synchronized (wa.windows)
      {
         return wa.windows.contains(window);
      }
   }

   /**
    * Remove specified window from window list. Note that this method
    * does destroy the window by calling {@link Widget#destroy()} on
    * the supplied window reference.
    * 
    * @param   window
    *          Window to remove.
    */
   public static void remove(TitledWindow<?> window)
   {
      // this needs to be done out of lock
      window.destroy();
      
      removeNoDestroy(window);
   }

   /**
    * Remove specified window from window list. Note that this
    * method doesn't destroy the window.
    * <p>
    * If the passed in window is a modal window, it will be demodalized 
    * with {@link #demodalizeWindow(TitledWindow)}.
    * 
    * @param   window
    *          Window to remove.
    */
   public static void removeNoDestroy(TitledWindow<?> window)
   {
      WorkArea wa = work.get();
      
      synchronized (wa.windows)
      {
         // modal window should be demodalized when it is made hidden
         // this is the safety net
         if (window.isModal())
         {
            demodalizeWindow(window);
         }
         
         wa.windows.remove(window);
         wa.fixedOrderWindows.remove(window);
      }
   }
   
   /**
    * Check if specified rectangle intersects at least partially with any of 
    * the windows located above specified one in the Z-order.
    * 
    * @param   bounds
    *          Source rectangle.
    * @param   ancestor
    *          Window to check.
    *          
    * @return  <code>true</code> if there is an intersection.
    */
   @SuppressWarnings("rawtypes")
   public static boolean checkRectangle(Rectangle bounds, Widget ancestor)
   {
      List<TitledWindow<?>> list = windowList();
      
      WorkArea wa = work.get();

      synchronized (list)
      {
         int i = wa.windows.indexOf(ancestor);

         if (i < 0)
            return false;
    
         if (!ThinClient.getInstance().isChui())
         {
            // in GUI, all windows use their own canvas.  In ChUI, the canvas/screen is shared, 
            // so intersection needs to be checked.
            return false;
         }
         
         i++;

         for (; i < list.size(); i++)
         {
            if (bounds.equals(list.get(i).bounds().intersection(bounds)))
               return true;
         }
      }

      return false;
   }
   
   /**
    * Get the list of all window IDs.
    * 
    * @return   See above.
    */
   public static int[] windowIds()
   {
      // get the windows id array size excluding the drop-down
      // should have at least one main window
      return windowList().stream()
               .filter(w -> !(w instanceof DropDown))
               .mapToInt(w -> w.getId().asInt())
               .toArray();
   }
   
   /**
    * Get the list of all visible top-level window IDs.
    * 
    * @return   See above.
    */
   public static int[] topLevelWindowIds()
   {
      // get the windows id array size excluding the drop-down
      // should have at least one main window
      return windowList().stream()
               .filter(w -> w instanceof TopLevelWindow && w.isVisible())
               .mapToInt(w -> w.getId().asInt())
               .toArray();
   }
   
   /**
    * Place the new window on the screen, in its default location.
    * <p>
    * The window placement algorithm replicates the algorithm used by WIN32 default window 
    * placement.  More details can be found in note 615 of Redmine task #2252.
    * 
    * @param    window
    *           The window to be placed.
    */
   public static void placeWindow(Window<?> window)
   {
      WorkArea wa = work.get();
      
      Window<?> defw = getDefaultWindow();
      
      if (window == defw)
      {
         // default window is not moved
         return;
      }

      // advance the default position by aligning it this way:
      // vertically: just below the window's title bar
      // horizontally: just right of the window's icon
      
      // always use the default window's metrics - the window title bar always has the same
      // metrics as the default one
      
      // if the new X and Y position doesn't allow the window to be fully displayed, reset to 0
      NativeRectangle bounds = ThinClient.getInstance().getDisplayBounds(window.getId().asInt());
      NativeDimension wdim = window.physicalDimension();
      
      if (wa.x + wdim.width > bounds.right() || wa.y + wdim.height > bounds.bottom())
      {
         wa.x = 0;
         wa.y = 0;
      }
      
      // initial position is always zero, so use the actual values as deltas
      window.setPhysicalLocation(wa.x, wa.y);

      wa.x += wa.locationDx;
      wa.y += wa.locationDy;
   }   

   /**
    * Set the initial dimension for this window.  Can be overwritten by the business logic.
    * 
    * @param    window
    *           The window.
    */
   public static void setInitialDimension(Window<?> window)
   {
      WorkArea wa = work.get();

      WindowConfig c = window.config();
      
      c.widthChars = wa.defaultColumns;
      c.heightChars = wa.defaultRows;
   }

   /**
    * Set the default window dimension.
    * 
    * @param    rows
    *           The number of rows (character units).
    * @param    cols
    *           The number of columns (character units).
    */
   public static void setDefaultDimension(int rows, int cols)
   {
      WorkArea wa = work.get();
      wa.defaultRows = rows;
      wa.defaultColumns = cols;
   }
   
   /**
    * Set the window placement deltas.
    * 
    * @param   dx
    *          The X-axis advancement.
    * @param   dy
    *          The Y-axis advancement.
    */
   public static void setPlacementDeltas(int dx, int dy)
   {
      WorkArea wa = work.get();
      wa.locationDx = dx;
      wa.locationDy = dy;
      
      // set the initial non-default window position 
      wa.x = dx;
      wa.y = dy;
   }
   
   /**
    * Turn the passed-in window into a modal window. The method
    * disables event input for all application top-level windows
    * except the passed-in window.
    * 
    * @param   window
    *          The subjected window reference.
    */
   public static void modalizeWindow(TopLevelWindow<?> window)
   {
      WorkArea wa = work.get();
      
      synchronized (wa.windows)
      {
         ArrayList<Integer> toSilence = new ArrayList<>();
         
         for (TitledWindow<?> tw : wa.windows)
         {
            // we care only about top level windows, OS events can be disabled only
            // on this level also do not silence the window to be made modal
            if (!(tw instanceof TopLevelWindow) || tw == window)
               continue;
            
            // skip already silenced windows
            if (wa.silencedWindows.contains(tw))
               continue;
            
            wa.silencedWindows.add((TopLevelWindow<?>) tw);
            toSilence.add(tw.getId().asInt());
         }
         
         int[] ids = Utils.integerCollectionToPrimitive(toSilence);
         
         GuiDriver<?,?> gd = (GuiDriver<?,?>) OutputManager.getDriver();
         gd.enableEvents(ids, false);
         
         // enable in case the modal window has been previously silenced
         if (wa.silencedWindows.contains(window))  
         {
            gd.enableEvents(new int[] {window.getId().asInt()}, true);
            wa.silencedWindows.remove(window);
         }
      }
   }

   /**
    * Restores modality of the passed-in window. The method
    * enables event input for application top-level windows
    * previously disabled by {@link #modalizeWindow(TopLevelWindow)}.
    * <p>
    * When the owner window of the passed window is a modal window events will
    * be enabled only for this owner window. This is to preserve modality.
    * <p>
    * If the passed in window is not registered with {@link WindowManager} this 
    * method is a no-op.
    * <p>
    * If the passed in window is silenced this method is a no-op.
    * 
    * @param   window
    *          The subjected window reference.
    */
   public static void demodalizeWindow(TitledWindow<?> window)
   {
      WorkArea wa = work.get();
      
      synchronized (wa.windows)
      {
         // prevent state corruption
         if (!wa.windows.contains(window))
         {
            return;
         }
         
         if (wa.silencedWindows.contains(window))
         {
            return;
         }
         
         ArrayList<Integer> toDesilence = new ArrayList<>(wa.silencedWindows.size());
         
         // process in z-order from top to bottom
         ListIterator<TitledWindow<?>> it = wa.windows.listIterator(wa.windows.size());
         while (it.hasPrevious())
         {
            TitledWindow<?> tw = it.previous();
            boolean removed = wa.silencedWindows.remove(tw);
            
            // skip unsilenced windows
            if (!removed)
            {
               continue;
            }
            
            toDesilence.add(tw.getId().asInt());
            
            if (tw.isModal() && tw.isVisible())
            {
               // preserve modality of this window, do not enable events
               // for windows with lower z-order
               break;
            }
         }
         
         int[] ids = Utils.integerCollectionToPrimitive(toDesilence);
         
         GuiDriver<?,?> gd = (GuiDriver<?,?>) OutputManager.getDriver();
         gd.enableEvents(ids, true);
      }
   }
   
   /**
    * Returns the dialogs owned by the supplied window.
    * 
    * @param   ownerWindowId
    *          The owner window.
    * @param   includeInvisible
    *          The flag indicating whether invisible windows should be included
    *          in the result.
    * 
    * @return  A list of {@link DialogBoxWindow} instances or empty list if no such
    *          window exists.
    */
   public static List<DialogBoxWindow> getDialogsOwnedBy(int ownerWindowId, 
                                                         boolean includeInvisible)
   {
      // get all dialog boxes with the same owner as this frame's dialog box owner
      List<DialogBoxWindow> dialogs = WindowManager.windowList().stream()
         .filter(w -> (includeInvisible || w.isVisible()) && w instanceof DialogBoxWindow)
         .map(w -> (DialogBoxWindow) w)
         .filter(w -> WidgetId.equals(w.getOwner().getId(), ownerWindowId))
         .collect(Collectors.toList());

      return dialogs;
   }
   
   /**
    * Returns the modal windows owned by the supplied window.
    * <p>
    * "owned by" in this case means the owner relationship established during
    * frame creation, i.e. the value returned by FRAME:WINDOW rather than the 
    * actual z-ordering of the windows.
    * The side-effect of the above and the way 4GL works also means the following.
    * DIALOG-BOX is displayed while a main window is active. The actual owner 
    * (FRAME:WINDOW) of the dialog is hidden. When the previously active main window
    * is made active by the OS (for example by Alt+Tab) the dialog window is 
    * NOT activated.
    * 
    * @param   ownerWindowId
    *          The owner window.
    * @param   includeInvisible
    *          The flag indicating whether invisible windows should be included
    *          in the result.
    * 
    * @return  A list of {@link TopLevelWindow} instances or empty list if no such
    *          window exists.
    */
   public static List<TopLevelWindow<?>> getModalsOwnedBy(int ownerWindowId,
                                                          boolean includeInvisible)
   {
      // get all dialog boxes with the same owner as this frame's dialog box owner
      List<TopLevelWindow<?>> modals = WindowManager.windowList().stream()
               .filter(w -> (includeInvisible || w.isVisible()) && w.isModal())
               .map(w -> (TopLevelWindow<?>) w)
               .filter(w -> WidgetId.equals(w.getOwner().getId(), ownerWindowId))
               .collect(Collectors.toList());

      return modals;
   }
   
   /**
    * Force the list of all windows to be stacked as in the order of the currently
    * registered windows.
    */
   public static void restackWindows()
   {
      GuiDriver<?,?> gd = (GuiDriver<?,?>) OutputManager.getDriver();
      gd.stackWindows(topLevelWindowIds());
   }
   
   /**
    * Called by the UI subsystem (GUI driver for example) when a top-leve window 
    * is activated. If the GUI driver dispatches the window events asynchronously on another
    * thread, it is responsible of invoking this method with ThinClient.invokeLaterOS() to serialize
    * the call on the main thread.
    * <p>
    * For more information of the concept of active windows, see 
    * {@link #forEachActiveWindow(TopLevelWindow, Predicate)}.
    *
    * @param   windowId
    *          The window being activated.
    * @param   fromNativeWindow
    *          If <code>true</code> the window losing activated state is not a P2J window.
    */
   public static void windowActivated(int windowId, boolean fromNativeWindow)
   {
      TopLevelWindow<?> window = WindowManager.findWindow(windowId); 
      if (window == null)
      {
         // when window is hidden/destroyed, the de-activation event from
         // the driver may arrive after the window is unregistered from 
         // WindowManager, so just ignore this
         return;
      }
      
      if (window == activeDriverWindow.get())
      {
         // for now ignore activation of already activated window 
         return;
      }
      
      processWindowActivated(window, fromNativeWindow);
   }
   
   /**
    * Called by the driver when a top-leve window is deactivated. If the GUI driver dispatches
    * the window events asynchronously on another thread, it is responsible of invoking this
    * method with ThinClient.invokeLaterOS() to serialize the call on the main thread.
    * <p>
    * For more information of the concept of active windows, see 
    * {@link #forEachActiveWindow(TopLevelWindow, Predicate)}.
    *
    * @param   windowId
    *          The window being deactivated.
    * @param   toNativeWindow
    *          If <code>true</code> the window gaining activated state is not a P2J window.
    */
   public static void windowDeactivated(int windowId, boolean toNativeWindow)
   {
      TopLevelWindow<?> deactivatedWindow = WindowManager.findWindow(windowId); 
      if (deactivatedWindow == null)
      {
         // the window being deactivated may be already gone, for example
         // when window is hidden and destroyed right after
         return;
      }
      
      if (activeDriverWindow.get() != deactivatedWindow)
      {
         throw new RuntimeException(
                  "GUI driver sent deactivation event for window not previously activated!");
      }
      
      // process activation of non-p2j window
      if (toNativeWindow)
      {
         processWindowActivated(null, true);
      }
      
      if (!(deactivatedWindow instanceof OverlayWindow))
      {
         // repaint is only needed for border/titlebar of normal windows and dialogs,
         // OverlayWindow are simply hidden when deactivated
         ThinClient.getInstance().eventDrawingBracket(deactivatedWindow, 
                                                      deactivatedWindow::repaintDecorations);
      }
   }
   
   /**
    * Returns <code>true</code> when <code>ownee</code> shares its activation state with
    * <code>owner</code>. <code>owner</code> may be an indirect owner of <code>ownee</code>.
    * 
    * @param   owner
    *          The owner window.
    * @param   ownee
    *          The ownee window.
    * 
    * @return  See above.
    */
   public static boolean isSharedActivationState(TopLevelWindow<?> owner, TopLevelWindow<?> ownee)
   {
      if (owner == null || ownee == null)
      {
         throw new IllegalArgumentException();
      }
      
      if (owner == ownee)
      {
         return true;
      }
      
      TopLevelWindow<?> temp = ownee;
      while (temp != null && ((GuiWindow) temp).isShareActivationWithOwner())
      {
         temp = temp.getOwner();
         
         if (temp == owner)
         {
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * The window calculates the next main window that should eventually 
    * become active after the currently active window is deactivated. 
    * 
    * @return  See above.
    */
   public static TopLevelWindow<?> getNextActiveWindow()
   {
      for (TitledWindow<?> tw : work.get().fixedOrderWindows)
      {
         if (!tw.isVisible() || !(tw instanceof GuiWindow) || 
             ((GuiWindow) tw).isShareActivationWithOwner())
         {
            continue;
         }
         
         return (TopLevelWindow<?>) tw;
      }
      
      return null;
   }
   
   /**
    * Returns all visible modal windows in z-order with the top window
    * being the last in the returned list.
    * 
    * @param   includeInvisible
    *          Indicates whether the method should include invisible
    *          windows in its search.
    * 
    * @return  See above.
    */
   private static List<TopLevelWindow<?>> getModals(boolean includeInvisible)
   {
      // get all dialog boxes with the same owner as this frame's dialog box owner
      List<TopLevelWindow<?>> modals = WindowManager.windowList().stream()
               .filter(w -> (includeInvisible || w.isVisible()) && w.isModal())
               .map(w -> (TopLevelWindow<?>) w)
               .collect(Collectors.toList());

      return modals;
   }   
   
   /**
    * The method implements the core window activation logic. It is called
    * by {@link #windowActivated(int, boolean)} and {@link #windowDeactivated(int, boolean)}
    * as a result of driver window activation change.
    * <p>
    * When <code>activated</code> is passed <code>null</code> the method will deactivate
    * the currently active windows but will not activate any new window. 
    * Also <code>fromNative</code> is ignored in this case.
    * <p>
    * For more information of the concept of active windows, see 
    * {@link #forEachActiveWindow(TopLevelWindow, Predicate)}.
    * 
    * @param   activated
    *          The window being activated.
    * @param   fromNative
    *          If <code>true</code> the window losing activated state is not a P2J window.
    */
   private static void processWindowActivated(TopLevelWindow<?> activated, boolean fromNative)
   {
      // save the currently driver-active window
      TopLevelWindow<?> actDriverWindow = activeDriverWindow.get();
      
      // update the driver-active window before posting activation messages
      // to prepare the correct state for event dispatchers
      activeDriverWindow.set(activated);
      
      if (activated != null)
      {
         List<TopLevelWindow<?>> modals = getModals(false);
         TopLevelWindow<?> modal = modals.isEmpty() ? null : modals.get(modals.size() - 1);
         
         if (modal != null && modal != activated && !isSharedActivationState(modal, activated))
         {
            // do not allow the activation of activated because an unrelated modal window is visible; 
            // instead activate the modal 
            WindowManager.activateWindow(modals.get(modals.size() - 1));
            return;
         }
      }
      
      // only deactivate currently app-activated windows when activating a non-P2J window or
      // when the currently driver-active window is the owner of the new active window and
      // the new active window doesn't share activation state with it
      if (actDriverWindow != null &&
          (activated == null || !isSharedActivationState(actDriverWindow, activated)))
      {
         // 'actDriverWindow' represents the owner chain of the previously active windows sharing
         // the activation state,  'activated' represents the owner chain of the new active
         // windows sharing the activation.
         // The idea behind the algorithm below is to find all those windows in the
         // 'actDriverWindow' chain that are not present in the 'activated' chain and
         // deactivate those.
         forEachActiveWindow(actDriverWindow, w ->
         {
            boolean active[] = new boolean[] {false};

            forEachActiveWindow(activated, aw ->
            {
               if (aw == w)
               {
                  // stop iteration when w is found to be present in the activated window chain,
                  // w and all its owners are considered active
                  active[0] = true;
                  return false;
               }

               // continue iteration
               return true;
            });

            if (active[0])
            {
               // stop iteration when w is found to be present in the activated window chain,
               // w and all its owners are considered active
               return false;
            }
            else
            {
               // w was not found in the activated window chain, it is thus considered to be an
               // inactive window and must be deactivated
               ThinClient.getInstance().postOSEvent(new WindowActivated(w, false,
               activated == null));

               // continue iteration through the actDriverWindow window chain
               return true;
            }
         });
      }
      
      // send activation event for the new active window
      if (activated != null)
      {
         ThinClient.getInstance().postOSEvent(new WindowActivated(activated, true, fromNative));
      }
   }
   
   /**
    * The method enumerates all app-active windows and executes <code>predicate</code>
    * for each until all windows are enumerated or <code>predicate</code> returns 
    * <code>false</code>.
    * <p>
    * At the application level there may exist multiple active windows simultaneously. 
    * For example when COMBO-BOX drop-down window or menu window is visible. In these 
    * cases the main window doesn't lose its activated state (its title bar is always 
    * painted as active). In contrast, there is always one active window and the driver 
    * level. So for the example above, when a menu window is displayed the driver deactivates 
    * the main window and activates the menu window. 
    * 
    * @param   activeDriverWnd
    *          The currently driver-active window.
    * @param   predicate
    *          The predicated being executed for each app-active window.
    */
   private static void forEachActiveWindow(TopLevelWindow<?> activeDriverWnd,
                                           Predicate<TopLevelWindow<?>> predicate)
   {
      if (activeDriverWnd == null)
      {
         return;
      }

      TopLevelWindow<?> candidate = activeDriverWnd;
      while (candidate != null)
      {
         if (!predicate.test(candidate))
         {
            return;
         }

         if (!(candidate instanceof GuiWindow) ||
             !((GuiWindow) candidate).isShareActivationWithOwner())
         {
            // break if the candidate window doesn't share its activation status 
            // with its owner
            break;
         }

         candidate = candidate.getOwner();
      }
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {  
      /** Storage for windows. */
      private List<TitledWindow<?>> windows = new ArrayList<>();

      /** Storage for windows, following their creation order. */
      private List<TitledWindow<?>> fixedOrderWindows = new ArrayList<>();
      
      /** Set of silenced windows, no particular ordering is implied here. */ 
      private Set<TopLevelWindow<?>> silencedWindows = new HashSet<>();

      /** The default number of rows for new windows. */
      private int defaultRows;
      
      /** The default number of columns for new windows. */
      private int defaultColumns;
      
      /** The default X location for new windows. */
      private int x = 0;
      
      /** The default Y location for new windows. */
      private int y = 0;

      /** The X delta when placing a new window. */
      private int locationDy;

      /** The Y delta when placing a new window. */
      private int locationDx;
   }
}
