/*
** Module   : MenuChuiImpl.java
** Abstract : CHUI client implementation of Menu.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description--------------------------------
** 001 VIG 20150116 Created initial version.
** 002 HC  20150323 Removed business logic from config classes.
** 003 VIG 20150225 Added event processing and drawing methods. Implemented draw, location
**                  methods.
** 004 VIG 20150618 Moved some methods to base class.
** 005 VIG 20150905 Changed title drawing using new mnemonic processing. Added accelerators
**                  drawing support.
** 006 VIG 20150923 Fix for mnemonic drawing when the menu has focus.
** 007 VIG 20151020 Reset focus if current widow has enabled frame.
** 008 VIG 20151023 Added focusTraversable(). CHUI doesn't support navigating of disabled items.
** 009 VIG 20160222 Label class is renamed to MnemonicInfo.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.chui;

import java.util.Arrays;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * <code>Menu</code> CHUI client implementation class. 
 */
@SuppressWarnings("unchecked")
public class MenuItemChuiImpl
extends MenuItem<ChuiOutputManager>
{
   /** This menu-item left padding in a sub-menu body or in a MENUBAR in pixels. */
   private static final int LEFT_PAD = 2;

   /** This menu-item right padding in a sub-menu body or in a MENUBAR in pixels. */
   private static final int RIGHT_PAD = 2;

   /** Sub-menu arrow left padding in characters. */
   private static final int SM_ARROW_PAD = 1;

   /** menu-item accelerator left padding in a sub-menu body or in a MENUBAR in characters. */
   private static final int ACCEL_PAD = 2;
   
   /** Sub-menu arrow width (for toggle-box) in characters. */
   private static final int SM_ARROW_WIDTH = 2;
   
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public MenuItemChuiImpl(WidgetId id, MenuItemConfig cfg)
   {
      super(id, cfg);
   }

   /**
    * Draw the peer container. The methods calls the draw() method for
    * ScrollPane instance for the peer container. In addition the required
    * clipping rectangle is calculated for ChildFrame containers because they
    * must be rendered inside the parent container.
    * <p>
    * Drawing for the redirected terminal differs. Labels/headers are drawn
    * only once per page, instead of the usual once per menu.
    */
   @Override
   public void draw()
   {
      Color dColor = config().dcolor == null ? Color.NORMAL : config().dcolor;
      
      Rectangle rect = bounds().translate(parent().childrenLocation(this));
      
      drawTitle(new Point(rect.left(), rect.top()), dColor);
   }

   /**
    * Get widget dimension.
    * 
    * @return  widget dimension.
    */
   @Override
   public Dimension dimension()
   {
      return new Dimension(width(), 1);
   }

   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   @Override
   public double height()
   {
      return 1;
   }

   /**
    * Get widget minimum size.
    * 
    * @return  widget minimum size.
    */
   @Override
   public Dimension minimumSize()
   {
      return dimension();
   }

   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public double width()
   {
      MenuItemConfig cfg = config();
      
      if (!isNormal())
      {
         return SubMenuChuiImpl.getMaxItemWidth(parent(), false) - LEFT_PAD - RIGHT_PAD;
      }
      
      int w = getPreparedTitle().length();
      
      if (cfg.toggleBox)
      {
         w++;
      }
      
      if (!Menu.isMenubarElement(this)) 
      {
         String accel = cfg.accelerator;
         if (accel != null && !accel.isEmpty())
         {
            if (SubMenu.hasChildSubmenu(parent()))
            {
               w += SM_ARROW_PAD + SM_ARROW_WIDTH;
            }
            else
            {
               w = SubMenuChuiImpl.getLongestItemTitleWidth(parent(), null);
            }
            
            w += ACCEL_PAD + accel.length();
            // w += ACCEL_PAD + SubMenuChuiImpl.getLongestItemAccelWidth(parent());
         }
         
         w += LEFT_PAD + RIGHT_PAD;
      }
      
      return w;
   }

   /**
    * Check if widget is focus traversable. SKIP and RULE aren't traversable.
    * 
    * @return  <code>true</code> if widget is focus traversable.
    */
   @Override
   public boolean focusTraversable()
   {
      if (isNormal())
      {
         // menu-items are focusable even in a closed sub-menu.
         // Implementation is the same as the parents implementation, 
         // except of the parent visibility check instead of its own.
         return (parent().isVisible() && isEnabled() && !hidden());         
      }
      
      return  false;
   }

   /**
    * Trick, that lets to correct parent location in cases, when it needed,
    * for example for menus.
    * 
    * @param    parent
    *           Parent component.
    *           
    * @return   corrected parent location.
    */
   @Override
   protected Point prepParentLocation(Widget<?> parent)
   {
      Point loc = parent.location();
      if (parent instanceof SubMenu)
      {
         loc.x += 1;
         loc.y += 2;
      }
      return loc;
   }
   
   /**
    * Process an event which trigger selection of this menu-item, depending if the item is a 
    * toggle-box or not.
    */
   protected void selectMenuItem()
   {
      super.selectMenuItem();

      // reset focus if current widow has enabled frame.
      if (ThinClient.getInstance().getFocusManager().getFirstEnabledWidget() != null)
      {
         ((Window<?>) ancestor()).getMenubar().setFocus(null);
         repaint();
      }
   }
   
   /**
    * Draws menu-item title, drawing mnemonic character as underlined.
    * 
    * @param   p
    *          Title point.  
    * @param   dColor
    *          Color of the title.
    */
   private void drawTitle(Point p, Color dColor)
   {
      if (hasFocus())
      {
         dColor = dColor.withReverse();
      }
      
      // SKIP or RULE
      if (isSkip())
      {
         Character skipRuleChar = ' '; 
         StringBuffer buf = new StringBuffer();
         for(int i=0; i<width(); i++)
            buf.append(skipRuleChar);
         screen().at(p).append(buf.toString(), Color.NORMAL);
         return; 
      }

      // RULE
      if (isRule())
      {
         screen().at(p).hLine(width(), Color.NORMAL);
         return;
      }
      
      // NORMAL
      MnemonicInfo label = getPreparedLabel();
      String title = label.getPrepLabel();
      if (config().toggleBox)
      {
         title = (config().checked ? ">" : " ") + title;
      }
      
      Color mnemonicColor = dColor.withUnderline().withoutReverse();
      
      Point tp = new Point(p);
      
      // disabled menu-item is drawn differently: 
      if ((isInitDisabled() && !hasFocus()) || label.getMnemonic() == null)
      {
         // Special case, when this item label ends with odd sequence of '&' and it is a longest 
         // label between neighbors in this sub-menu body or menu. Focused item is drawn in the  
         // following way: last 2 characters including added '_' are drawn using mnemonic color.
         boolean specialCase = false;
         
         if (label.getMnemonic() == null)
         {
            title += '_';
            int longTitleWidth = SubMenuChuiImpl.getLongestItemTitleWidth(parent(), null);
            specialCase = (title.length() - 1 >= longTitleWidth) && hasFocus(); 
         }
         
         // draw
         if (specialCase)
         {
            screen().at(tp).append(title.substring(0, title.length() - 2), dColor);
            tp.x += title.length() - 2;
            screen().at(tp).append(title.substring(title.length() - 2), mnemonicColor);
            tp.x += 2;
         }
         else
         {
            screen().at(tp).append(title, dColor);  
            tp.x += title.length();
         }
      }
      else
      {
         Integer ampIdx = label.getMnemIdx();

         // if & is not specified use the first letter as drawn mnemonic
         if (ampIdx == null) ampIdx = config().toggleBox ? 1 : 0;

         String leftPart = title.substring(0, ampIdx);
         screen().at(tp).append(leftPart, dColor);

         tp.x += ampIdx;

         if (hasFocus())
         {
            mnemonicColor = mnemonicColor.withoutUnderline();
         }

         screen().at(tp).append(title.substring(ampIdx, ampIdx + 1), mnemonicColor);

         tp.x += 1;
         String rightPart = title.substring(ampIdx + 1);
         screen().at(tp).append(rightPart, dColor);
         tp.x += rightPart.length();
      }

      int longestAccelLen = SubMenuChuiImpl.getLongestItemAccelWidth(parent());
      if (!Menu.isMenubarElement(this) && longestAccelLen > 0)
      {
         // fill space between the menu-item title and accelerator with spaces.
         int longestTitleLen = SubMenuChuiImpl.getLongestItemTitleWidth(parent(), null);
         int titleLenDif = longestTitleLen - title.length();
         
         // space between title and accelerator
         int spaceLen = titleLenDif;
         if (SubMenu.hasChildSubmenu(parent()))
            spaceLen += SM_ARROW_PAD + SM_ARROW_WIDTH;
         spaceLen += ACCEL_PAD;
         char[] array = new char[spaceLen];
         Arrays.fill(array, ' ');
         
         String accel = config.accelerator; 
         int accelLen = accel == null ? 0 : accel.length();
         Color spaceColor = accelLen > 0 ? dColor : dColor.withoutReverse();
         screen().at(tp).append(new String(array), spaceColor);

         // draw accelerator
         tp.x += spaceLen;
         if (accelLen > 0)
         {
            screen().at(tp).append(accel, dColor);
         }
      }
   }
}