/*
** Module   : ChuiPrimitives.java
** Abstract : output primitives for a CHUI driver that has an emulated terminal
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 GES 20110913 Created initial version.
** 002 SIY 20111101 Updated imports.
** 003 MAG 20131114 Move the class on another package. The class has become a common base class.
** 004 GES 20150503 Reworked to use an emulated terminal that is not swing or web specific.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.chui.driver;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;

/**
 * Implementation of ChUI-specific direct output primitives using an emulated terminal
 * simulator for output.
 */
public class ChuiPrimitives
extends DriverPrimitives
{
   /** Emulated terminal simulator implemented in a driver-specific manner. */
   private EmulatedTerminalState ets = null;
   
   /**
    * Constructor.
    *
    * @param    ets
    *           Emulated terminal simulator instance used for direct drawing.
    */
   public ChuiPrimitives(EmulatedTerminalState ets)
   {
      this.ets = ets;
   }
   
   /**
    * Obtain the core worker that implements the emulated terminal.
    *
    * @return   The worker instance.
    */
   public EmulatedTerminalState getEmulatedTerminalState()
   {
      return ets;
   }

   /**
    * Add a string to the virtual terminal at the current cursor position. This
    * move the cursor after each character is output, naturally wrapping the
    * cursor to the beginning of the next row or to the beginning of the screen
    * as necessary.
    *
    * @param    str
    *           The text to output.
    * @param    color
    *           The color (includes attributes) in which to output the text.
    *           All text will be output with the same color.
    */
   @Override
   public void append(String str, Color color)
   {
      int len = str.length();
      
      Cell[] data = new Cell[len];
      
      for (int i = 0; i < len; i++)
      {
         // the color instance is passed through since it is immutable
         data[i] = new Cell(str.charAt(i), color);
      }
      
      ets.append(data);
   }

   /**
    * Position the virtual cursor during the drawing.
    *
    * @param    x
    *           New cursor X.
    * @param    y
    *           New cursor Y.
    *
    * @return   <code>true</code> if cursor positioning succeeded.
    */
   @Override
   public boolean cursorAt(int x, int y)
   {
      return ets.setCursor(x, y);
   }

   /**
    * Position the real cursor.
    *
    * @param    x
    *           New cursor X.
    * @param    y
    *           New cursor Y.
    *
    * @return   <code>true</code> if cursor positioning succeeded.
    */
   @Override
   public boolean cursorStay(int x, int y)
   {
      return ets.setCursor(x, y);
   }

   /**
    * Get absolute X position of the cursor.
    *
    * @return   The X position of the cursor.
    */
   @Override
   public int getX()
   {
      return ets.getCursorX();
   }

   /**
    * Get absolute Y position of the cursor.
    *
    * @return   The Y position of the cursor.
    */
   @Override
   public int getY()
   {
      return ets.getCursorY();
   }

   /**                                         
    * Gets the number of rows on the screen.
    *
    * @return   The number of rows (lines of characters).
    */
   @Override
   public int screenHeight()
   {
      return ets.getRows();
   }

   /**
    * Gets the number of columns on the screen.
    *
    * @return   The number of columns (in characters).
    */
   @Override
   public int screenWidth()
   {
      return ets.getColumns();
   }

   /**
    * Synchronize the state of the physical terminal with the state of
    * the virtual terminal.
    */
   @Override
   public void sync()
   {
      ets.triggerRepaint(true);
   }

   /**
    * Perform a bulk output operation for the whole screen at once with
    * output masking when desired. Set the cursor afterwards to the given
    * X and Y coordinates in the <code>screen</code> instance.
    * <p>
    * The screen container is safe for use with any UNICODE character.
    *
    * @param    screen
    *           Container for whole screen content and cursor position.
    */
   @Override
   public void updateScreen(ScreenData screen)
   {
      ets.replace(screen.asArray(), screen.getCursorX(), screen.getCursorY());
   }
}