/*
** Module   : DriverPrimitives.java
** Abstract :
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 SIY 20101223 Created initial version
** 002 SIY 20110927 Updated javadoc.
** 003 GES 20111005 Moved location to a more appropriate package.
** 004 HC  20140911 GUI coordinate support - data types refactoring.
** 005 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.chui.driver;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

// TODO: eliminate cursorStay(int, int) or provide default implementation.

/**
 * Base class for driver-level implementations of {@link OutputPrimitives}.
 * <p>
 * This class is intended to be a base class for the driver-level primitives
 * implementation. New driver implementation need to implement remaining
 * methods and/or override non-final methods. Remaining code is common for all
 * implementations and there is no need to change it.
 * <p>
 * The basic idea is to split driver implementation into two parts - completely
 * implementation dependent part and output primitives implementation. First 
 * part has no common code and should be written from scratch, but has quite 
 * small set of methods (see {@link ScreenDriver} for more details on this 
 * part). Second part - output primitives - shares most of the implementation 
 * code and uses minimal set of implementation-specific APIs. 
 * <p>
 * This class (together with its base class) provides almost all necessary 
 * shared code, so implementers of the particular driver should implement only 
 * small set of implementation-specific methods:
 * <ul>
 * <li>{@link #append(String, Color)}</li>
 * <li>{@link #cursorAt(int, int)}</li>
 * <li>{@link #cursorStay(int, int)}</li> 
 * <li>{@link #getX()}</li>
 * <li>{@link #getY()}</li>
 * <li>{@link #screenHeight()}</li>
 * <li>{@link #screenWidth()}</li>
 * <li>{@link #sync()}</li>
 * <li>{@link #updateScreen(ScreenData)}</li>
 * </ul>
 * Refer to JavaDoc of each method listed above for more details about 
 * requirements to implementation.  
 *                           
 * @author 	SIY
 * @version 1.0
 */
public abstract class DriverPrimitives
extends BasePrimitives
{
   /** The default clipping bitmap. */
   private ScreenBitmap bitmap;

   /**
    * Get color and attribute for border of specified widget.
    * <p>
    * Implementation may override this method if it need to change default 
    * behavior.
    * 
    * @param   widget
    *          Widget reference.
    *
    * @return  logical color for the widget border.
    */
   @Override
   public Color borderColor(Widget widget)
   {
      return Color.DEFAULT;
   }

   /**
    * Get color and attribute for the widget area.
    * <p>
    * Implementation may override this method if it need to change default 
    * behavior.
    *
    * @param   widget
    *          Widget reference.
    *
    * @return  Array with two elements: attribute and color.
    */
   @Override
   public Color widgetColor(Widget widget)
   {
      return Color.DEFAULT;
   }

   /**
    * Add a character to the virtual terminal at the current cursor
    * position.
    *
    * @param   chr
    *          The character to output.
    * @param   color
    *          The number of the color-pair (foreground + background) to
    *          use for blanking the rectangle.
    */
   @Override
   public final void append(int chr, Color color)
   {
      throw new UnsupportedOperationException();
   }

   /**
    * Draw a box using the specified dimensions and logical color.
    * <p>
    * The driver-level implementation does not need to override this method.
    *
    * @param   left
    *          The left column of the rectangle.
    * @param   top
    *          The top row of the rectangle.
    * @param   right
    *          The right column of the rectangle.
    * @param   bottom
    *          The bottom row of the rectangle.
    * @param   color
    *          The number of the color-pair (foreground + background).
    */
   @Override
   public final void box(int left, int top, int right, int bottom,
                      Color color)
   {
      throw new UnsupportedOperationException();
   }

   /**
    * Blank out a box using the specified logical color.
    * <p>
    * The driver-level implementation does not need to override this method.
    *
    * @param   left
    *          The left column of the rectangle.
    * @param   top
    *          The top row of the rectangle.
    * @param   right
    *          The right column of the rectangle.
    * @param   bottom
    *          The bottom row of the rectangle.
    * @param   color
    *          The number of the color-pair (foreground + background).
    */
   @Override
   public final void clearArea(int left, int top, int right,
                            int bottom, Color color)
   {
      throw new UnsupportedOperationException();
   }

   /**
    * Draw the upper or lower border.
    * <p>
    * The driver-level implementation does not need to override this method.
    *
    * @param   col
    *          The column number where to draw.
    * @param   row
    *          The row number where to draw.
    * @param   width
    *          The border width.
    * @param   color
    *          The number of the color-pair (foreground+background)
    */
   @Override
   public final void drawHBorder(int col, int row, int width,
                              Color color)
   {
      throw new UnsupportedOperationException();
   }

   /**
    * Draw the lower-left border corner.
    * <p>
    * The driver-level implementation does not need to override this method.
    *
    * @param   col
    *          The column number where to draw.
    * @param   row
    *          The row number where to draw.
    * @param   color
    *          The number of the color-pair (foreground+background) to use
    *          for blanking the rectangle.
    */
   @Override
   public final void drawLLCorner(int col, int row, Color color)
   {
      throw new UnsupportedOperationException();
   }

   /**
    * Draw the lower-right border corner.
    * <p>
    * The driver-level implementation does not need to override this method.
    *
    * @param   col
    *          The column number where to draw.
    * @param   row
    *          The row number where to draw.
    * @param   color
    *          The number of the color-pair (foreground+background) to use
    *          for blanking the rectangle.
    */
   @Override
   public final void drawLRCorner(int col, int row, Color color)
   {
      throw new UnsupportedOperationException();
   }

   /**
    * Draw the upper-left border corner.
    * <p>
    * The driver-level implementation does not need to override this method.
    *
    * @param   col
    *          The column number where to draw.
    * @param   row
    *          The row number where to draw.
    * @param   color
    *          The number of the color-pair (foreground+background) to use
    *          for blanking the rectangle.
    */
   @Override
   public final void drawULCorner(int col, int row, Color color)
   {
      throw new UnsupportedOperationException();
   }

   /**
    * Draw the upper-right border corner.
    * <p>
    * The driver-level implementation does not need to override this method.
    *
    * @param   col
    *          The column number where to draw.
    * @param   row
    *          The row number where to draw.
    * @param   color
    *          The number of the color-pair (foreground+background) to use
    *          for blanking the rectangle.
    */
   @Override
   public final void drawURCorner(int col, int row, Color color)
   {
      throw new UnsupportedOperationException();
   }

   /**
    * Draw the left or right border.
    * <p>
    * The driver-level implementation does not need to override this method.
    *
    * @param   col
    *          The column number where to draw.
    * @param   row
    *          The row number where to draw.
    * @param   height
    *          The border height.
    * @param   color
    *          The number of the color-pair (foreground+background)
    */
   @Override
   public final void drawVBorder(int col, int row, int height,
                              Color color)
   {
      throw new UnsupportedOperationException();
   }

   /**
    * Get the screen bitmap for this output primitives implementation.
    * <p>
    * The driver-level implementation does not need to override this method.
    *
    * @return  The screen bitmap.
    */
   @Override
   public final ScreenBitmap getScreenBitmap()
   {
      if (bitmap == null)
      {
         bitmap = new ScreenBitmap(screenHeight(), screenWidth());
      }
      return bitmap;
   }

   /**
    * Draw a horizontal line of the specified length starting at the
    * current cursor position.
    * <p>
    * The driver-level implementation does not need to override this method.
    *
    * @param   length
    *          The length of the line to draw in columns.
    * @param   color
    *          The number of the color-pair (foreground + background).
    */
   @Override
   public final void hLine(int length, Color color)
   {
      throw new UnsupportedOperationException();
   }

   /**
    * Mark a rectangle as needing repainting.
    * <p>
    * The driver-level implementation does not need to override or implement
    * this method.
    *
    * @param   rect
    *          rectangle to mark for repainting
    */
   @Override
   public final void refresh(NativeRectangle rect)
   {
   }

   /**
    * Replace the current bitmap instance with this one.
    * <p>
    * The driver-level implementation does not need to override this method.
    *
    * @param   bitmap
    *          The new bitmap instance.
    */
   @Override
   public final void setScreenBitmap(ScreenBitmap bitmap)
   {
      this.bitmap = bitmap;
   }

   /**
    * Draw a vertical line of the specified length starting at the current
    * cursor position.
    * <p>
    * The driver-level implementation does not need to override this method.
    *
    * @param   length
    *          The length of the line to draw in rows.
    * @param   color
    *          The number of the color-pair (foreground + background).
    */
   @Override
   public final void vLine(int length, Color color)
   {
      throw new UnsupportedOperationException();
   }
}