/*
** Module   : PopupTrigger.java
** Abstract : helper to support popup menu mouse event processing
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 GES 20110921 Created initial version.
** 002 GES 20150506 Moved to a driver package.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.driver.swing;

import java.awt.event.*;
import javax.swing.*;

/**
 * Helper to support popup menu mouse event processing. This class provides
 * a generic mouse event handler that detects when a mouse event is a popup
 * trigger and in that case it will cause the menu to appear.
 * <p>
 * The event type that is a popup trigger will vary on different platforms.
 * This helper class relies upon <code>MouseEvent.isPopupTrigger</code> to
 * detect when the popup menu should be displayed. 
 */
public class PopupTrigger
extends MouseAdapter
{
   /** The menu to be displayed. */
   private JPopupMenu popup = null;
   
   /**
    * Construct an instance.
    *
    * @param    popup
    *           The menu to be displayed when the popup trigger event occurs.
    */
   public PopupTrigger(JPopupMenu popup)
   {
      this.popup = popup;
   }
   
   /**
    * Process the event, causing the associated popup menu to be displayed
    * if the given event is a popup trigger event.
    *
    * @param    evt
    *           The event to process.
    */
   public void mousePressed(MouseEvent evt)
   {
      handlePopup(evt);
   }

   /**
    * Process the event, causing the associated popup menu to be displayed
    * if the given event is a popup trigger event.
    *
    * @param    evt
    *           The event to process.
    */
   public void mouseReleased(MouseEvent evt)
   {
      handlePopup(evt);
   }

   /**
    * Low level worker to implement event processing, causing the associated
    * popup menu to be displayed if the given event is a popup trigger event.
    *
    * @param    evt
    *           The event to process.
    */
   private void handlePopup(MouseEvent evt)
   {
      if (evt.isPopupTrigger())
      {
         popup.show(evt.getComponent(), evt.getX(), evt.getY());
      }
   }
}
