/*
** Module   : SwingHelper.java
** Abstract : useful swing helper methods
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 GES 20110921 Created initial version.
** 002 EVL 20121119 Adding method to select the firs available monospaced font.
**                  The signature and implementation of the method
**                  isMonospaced() is changed to check if the font able to draw
**                  the frame border characters.
** 003 MAG 20131206 Improve detection of monospace fonts.
** 004 GES 20150424 Added fromColorRgb() and toColorRgb() helpers. Moved to a driver package.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.driver.swing;

import java.awt.*;
import javax.swing.*;
import java.util.*;

// we don't want to import all of the ui package because some classes conflict with AWT
import com.goldencode.p2j.ui.ColorRgb;

/**
 * Provides useful Swing helper methods as static tools.
 */
public class SwingHelper
{
   /** Fixed width font list. */ 
   private static String[] monofonts = null;
   
   /** Complete font list. */
   private static String[] allfonts = null;
   
   /**
    * Constructor, which cannot be called externally.
    */
   private SwingHelper()
   {
   }
   
   /**
    * Helper to convert a generic RGB color into an AWT color.
    *
    * @param    clr
    *           The color to convert.
    *
    * @return   The AWT color instance.
    */
   public static Color fromColorRgb(ColorRgb clr)
   {
      return new Color(clr.getRed(), clr.getGreen(), clr.getBlue());
   }
   
   /**
    * Helper to convert an AWT color into a generic RGB color.
    *
    * @param    clr
    *           The AWT color to convert.
    *
    * @return   The color instance.
    */
   public static ColorRgb toColorRgb(Color clr)
   {
      return new ColorRgb(clr.getRed(), clr.getGreen(), clr.getBlue());
   }
   
   /**
    * Checks the given font and reports if that font is monospaces (has a
    * fixed width which is the same for all characters and is able to draw
    * the frame border characters).
    * <p>
    * <b>WARNING:</b> this method compares the size of multiple letters and
    * numbers which would normally be expected to be different if there was
    * no fixed width. This is not algorithmically foolproof, however it appears
    * to be reliable and correct in its results.
    *
    * @param    font
    *           The font being tested.
    *
    * @param    fm
    *           The font metrics being tested.
    *
    * @return   <code>true</code> if the given font is detected as having a
    *           fixed width.
    */
   public static boolean isMonospaced(Font font, FontMetrics fm)
   {
      if (!font.canDisplay('\u2500'))
         return false;

      // check if this font is monospaced, this is a bit of a hack but should
      // be reliable
      int msize = fm.charWidth('m');
      int isize = fm.charWidth('i');
      int osize = fm.charWidth('1');
      int nsize = fm.charWidth('9');
      int dsize = fm.charWidth('\u2500');
      int fsize = font.getSize();
      
      return (msize == isize && msize == nsize &&
              msize == osize && (osize == dsize || fsize == dsize));
   }
   
   /**
    * Obtain the list of all available font family names.
    *
    * @return   The list of font family names.
    */
   public static synchronized String[] getFontList()
   {
      if (allfonts == null)
      {
         GraphicsEnvironment ge = 
            GraphicsEnvironment.getLocalGraphicsEnvironment();
            
         allfonts = ge.getAvailableFontFamilyNames();
      }
      
      return allfonts;
   }
   
   /**
    * Obtain the list of all available monospaced font family names.
    *
    * @param    comp
    *           Used to obtain a font metrics instance for each font.
    *
    * @return   The list of font family names which have fixed width.
    */
   public static synchronized String[] getMonoFontList(Component comp)
   {
      if (monofonts == null)
      {
         String[] all = getFontList();
         
         ArrayList<String> array = new ArrayList<String>();
         
         // filter the list down to monospaced fonts
         for (int i = 0; i < all.length; i++)
         {
            Font font = Font.decode(all[i]); 
            
            if (isMonospaced(font, comp.getFontMetrics(font)))
            {
               array.add(all[i]);
            }
         }
         
         // save the list
         monofonts = (String[]) array.toArray(new String[0]);
      }
      
      return monofonts;
   }

   /**
    * Obtain the first available monospaced physical font name. The font should
    * be able to draw the frame border characters.
    *
    * @param    comp
    *           Used to obtain a monospaced font list for the component.
    *
    * @param    defaultFont
    *           The font name to be used as return value if it is not possible
    *           to get the valid physical font name from the system.
    *
    * @return   The name of the monospaced font family which has fixed width.
    */
   public static synchronized String getMonoFontName(Component comp, String defaultFont)
   {
      String[] all = getMonoFontList(comp);
      if(all.length > 0)
         return all[0];
      else
         return defaultFont;
   }

   /**
    * Finds the parent component that is a frame. This is needed for creating
    * dialogs in applets.
    *
    * @param    comp
    *           The component to inspect, starting at the component's parent.
    * 
    * @return   The parent frame.
    */
   public static Frame findParentFrame(Component comp)
   {
      comp = comp.getParent();

      while (true)
      {
         if (comp instanceof Frame)
         {
            return (Frame) comp;
         }

         comp = comp.getParent();
      }
   }
   
   /**
    * Configure the given container's <code>SpringLayout</code> with the
    * springs, sizing, padding and constraints to create a 2 dimensional
    * grid.  Each row will have the same number of columns.
    *
    * @param    outer
    *           The container that holds all of the given components, which
    *           MUST be using a <code>SpringLayout</code>.
    * @param    cells
    *           The components that are being laid out.  They must already have
    *           been added to the container.
    * @param    padding
    *           The number of pixels between rows and between columns.
    */
   public static void setupGridSprings(Container      outer,
                                       JComponent[][] cells,
                                       int            padding)
   {
      SpringLayout layout = (SpringLayout) outer.getLayout();
      
      Spring pad = Spring.constant(padding);
      Spring x   = Spring.constant(padding);
      Spring y   = Spring.constant(padding);
      
      // WARNING: assumes that all rows have the same number of columns!
      for (int col = 0; col < cells[0].length; col++)
      {
         x = setupHorizontal(layout, cells, x, pad, col);
      }
      
      for (int row = 0; row < cells.length; row++)
      {
         y = setupVertical(layout, cells, y, pad, row);
      }
      
      // setup container
      SpringLayout.Constraints current = layout.getConstraints(outer);
      current.setConstraint(SpringLayout.EAST, x);
      current.setConstraint(SpringLayout.SOUTH, y);
   }
   
   /**
    * Calculate the maximum width for the given column, force all components
    * of that column to that width and return the X coordinate of the start
    * of the following column.
    *
    * @param    layout
    *           The layout instance being edited.
    * @param    cells
    *           The components that are being laid out.  They must already have
    *           been added to the container.
    * @param    x
    *           The starting X coordinate of the column being processed.
    * @param    pad
    *           The number of pixels between columns.
    * @param    col
    *           The 0-based index of the column being processed.
    *
    * @return   The starting X coordinate of the column following the one being
    *           processed (the next column).
    */
   private static Spring setupHorizontal(SpringLayout   layout,
                                         JComponent[][] cells,
                                         Spring         x,
                                         Spring         pad,
                                         int            col)
   {
      Spring width = Spring.constant(0);
      SpringLayout.Constraints current = null;
      
      // calculate the maximum width for the column
      for (int row = 0; row < cells.length; row++)
      {
         current = layout.getConstraints(cells[row][col]);
         width   = Spring.max(width, current.getWidth());
      }
      
      // force all components in the column to the same width and x position
      for (int row = 0; row < cells.length; row++)
      {
         current = layout.getConstraints(cells[row][col]);
         current.setWidth(width);
         current.setX(x);
      }
      
      // return the new X value (x + width + pad)
      return Spring.sum(x, Spring.sum(width, pad));
   }
   
   /**
    * Calculate the maximum height for the given row, force all components
    * of that row to that height and return the Y coordinate of the start
    * of the following row.
    *
    * @param    layout
    *           The layout instance being edited.
    * @param    cells
    *           The components that are being laid out.  They must already have
    *           been added to the container.
    * @param    y
    *           The starting Y coordinate of the row being processed.
    * @param    pad
    *           The number of pixels between rows.
    * @param    row
    *           The 0-based index of the row being processed.
    *
    * @return   The starting Y coordinate of the row following the one being
    *           processed (the next row).
    */
   private static Spring setupVertical(SpringLayout   layout,
                                       JComponent[][] cells,
                                       Spring         y,
                                       Spring         pad,
                                       int            row)
   {
      Spring height = Spring.constant(0);
      SpringLayout.Constraints current = null;
      
      // calculate the maximum width for the column
      for (int col = 0; col < cells[row].length; col++)
      {
         current = layout.getConstraints(cells[row][col]);
         height   = Spring.max(height, current.getHeight());
      }
      
      // force all components in the column to the same width and x position
      for (int col = 0; col < cells[row].length; col++)
      {
         current = layout.getConstraints(cells[row][col]);
         current.setHeight(height);
         current.setY(y);
      }
      
      // return the new Y value (y + height + pad)
      return Spring.sum(y, Spring.sum(height, pad));
   }
}