/*
** Module   : WinKeyboardReader.java
** Abstract : process swing key events and convert them to 4GL keys, compatible with Windows 
**            clients.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20141029 Created initial version.
** 002 CA  20141125 Added CTRL-BREAK processing for GUI client.  On Windows, CTRL-BREAK produces
**                  the CTRL-C key (code 3); on linux, it produces the CTRL-CTRL-S key - CTRL 
**                  modifier with the CTRL-S key (code 19).
** 003 CA  20141205 Added support for the SHIFT modifier.
** 004 CA  20141207 Improved H003 - the SHIFT modifier must be treated only if the modified key 
**                  (SHIFT-<key>) ends with the same label as the unmodified key.  This is needed
**                  to distinguish keys which have a double-role (as uppercase letters or SHIFT 
**                  with a physical key which can emit two characters, depending on if SHIFT was
**                  pressed or not, like the ";:" key in a US layout keyboard).
** 005 CA  20141220 Ignore the SHIFT processing for the SHIFT-TAB key (this is treated explicitly
**                  by keyPressed).
** 006 GES 20150506 Moved to a driver package.
** 007 CA  20150615 Fixed CTRL-SHIFT-* and CTRL-DEL keys.
** 008 EVL 20150518 Removing generation of the ESC key event in keyTyped() method.  The ESC key
**                  is a END-ERROR generation and should be invoked only once per key press.  In
**                  GUI both keyPressed() and keyTyped() executed and this will generate two
**                  END-ERROR events and in case of the drop-down dismissing causes the
**                  application to stop after dismissing drop-down.
** 009 VIG 20150818 Added ALT key support for menus keys navigation mode.
** 010 SBI 20151008 Improvements to support better GUI key processing compatibility.
** 011 SBI 20151014 Auto-repeats are fixed.
** 012 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 013 SBI 20160114 Fixed the internal keyboard state to be consistent that happens due to this
**                  keys combination ALT + TAB is intercepted by the operating system.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.driver.swing;

import java.awt.AWTEvent;
import java.awt.Toolkit;
import java.awt.event.AWTEventListener;
import java.awt.event.KeyEvent;
import java.awt.event.WindowEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.event.*;

/**
 * This class represents a {@link SwingKeyboardReader} implementation for clients which require
 * Windows compatibility.
 */
public class WinKeyboardReader
extends SwingKeyboardReader
{
   /**
    * Default c'tor.  It translates specific GUI related keys.
    */
   public WinKeyboardReader()
   {
      xlate.put(KeyEvent.VK_ESCAPE, Key.VK_ESCAPE);
      
      // ALT is not processed by 4GL... => Processed as menubar activation key.
      xlate.put(KeyEvent.VK_ALT, 1024);
      // added to process OS intercepted keys
      Toolkit.getDefaultToolkit().addAWTEventListener(new AWTEventListener()
      {
         @Override
         public void eventDispatched(AWTEvent event)
         {
            if (event.getID() == WindowEvent.WINDOW_DEACTIVATED)
            {
               if (!keyDownEvents.isEmpty())
               {
                  keyDownEvents.clear();
               }
            }
         }
      }, AWTEvent.WINDOW_EVENT_MASK);

   }

   /**
    * This method is called when any key is pressed or when a multi-key combination (CTRL, SHIFT, 
    * ALT modifiers) is pressed.  Key events provided here are low level events and have not been
    * processed into a higher level code as is done in {@link #keyTyped}, with an exception: if
    * CTRL is pressed with a latin character, then the processing is deferred to {@link #keyTyped};
    * otherwise, if CTRL is pressed, bit 11 (CTRL modifier) is set and the key is sent to the
    * {@link SwingKeyboardReader#queue}, in FIFO order.
    * <p>
    * All other ALT processing is deferred to the {@link #keyTyped} method, with this exception:
    * if the 4GL key code is VK_TAB and the SHIFT modifier key is on (the shift key was pressed 
    * at the time the tab was pressed), then a VK_BACKTAB is posted to the queue.  Otherwise, 
    * the VK_TAB is ignored because it is naturally processed in {@link #keyTyped} (but the back 
    * tab is not).
    *  
    * @param    evt
    *           The high level key event to process.
    */
   public synchronized void keyPressed(KeyEvent evt)
   {
      KeyEvent last;

      if (keyDownEvents.isEmpty())
      {
         last = evt;
      }
      else
      {
         last = keyDownEvents.peek();
         if (isModifier(last.getKeyCode()))
         {
            mergeWithLastEvent(last, evt);
            return;
         }
         else if (last.getKeyCode() == evt.getKeyCode())
         { //auto repeats
           // we are here only iff typed doesn't happen.
            keyDownEvents.push(evt);
            keyReleased(evt);
            return;
         }
         else
         {
            last = evt;
         }
      }
      keyDownEvents.push(last);
   }

   /**
    * This method is called when a pressed key (or a multi-key combination) is converted to a 
    * higher level representation. As a general rule, all key codes between 0 and 127 inclusive 
    * are generated through this callback. The reason for this is that Java translates the keys or
    * multi-key combinations into the higher level representation which almost perfectly matches 
    * the 4GL key codes 0 through 127.  In other words, the keyboard inputs of the user are 
    * very compatible between the 4GL and Java for those portions of the key codes (0 through 127).
    * <p>
    * Once a valid key code is generated, that key code is added to the key input queue in FIFO 
    * order.
    * <p>
    * The non-visual values 0 through 31 are handled almost "for free". The user generates these 
    * in the following manner:
    * <p>
    * <pre>
    * Key Code  User Input               Notes
    * --------  -----------------------  --------------------------------------
    * 0         CTRL+2                   Special code in this method supports this.
    * 1         CTRL+A
    * 2         CTRL+B
    * 3         CTRL+C
    * 4         CTRL+D
    * 5         CTRL+E
    * 6         CTRL+F
    * 7         CTRL+G
    * 8         CTRL+H or VK_BACKSPACE
    * 9         CTRL+I or VK_TAB         The containing frame's focus traversal
    *                                    keys must be forced to the empty set in
    *                                    order for this to work.
    * 10        CTRL+J or VK_ENTER       NOT DELIVERED, handled in keyPressed()
    * 11        CTRL+K
    * 12        CTRL+L
    * 13        CTRL+M
    * 14        CTRL+N
    * 15        CTRL+O
    * 16        CTRL+P
    * 17        CTRL+Q
    * 18        CTRL+R
    * 19        CTRL+S
    * 20        CTRL+T
    * 21        CTRL+U
    * 22        CTRL+V
    * 23        CTRL+W
    * 24        CTRL+X
    * 25        CTRL+Y
    * 26        CTRL+Z
    * 27        VK_ESCAPE                DELIVERED; Windows mode allows this.
    * 28        CTRL+\
    * 29        CTRL+]
    * 30        CTRL+6                   Special code in this method supports this.
    * 31        CTRL+-                   Special code in this method supports this.
    * </pre>
    * <p>
    * ALT key is the modifier which sets the key code's 10th bit.  If this is determined to be
    * pressed together with the key, than bit 10 will be set and this key will be pushed to the
    * {@link SwingKeyboardReader#queue}, but only if the key code's label is not the empty string.
    * 
    * @param    evt
    *           The high level key event to process.
    */
   public synchronized void keyTyped(KeyEvent evt)
   {
      if (evt.isConsumed())
      {
         return;
      }

      boolean ctrl = evt.isControlDown();
      boolean alt  = evt.isAltDown();
      boolean shift = evt.isShiftDown();
      char    ch   = evt.getKeyChar();
      int     code = -1;

      KeyEvent lastKey = keyDownEvents.pop();
      int lastKeyCode = lastKey.getKeyCode();
      Integer mappedCode = xlate.get(lastKeyCode);
      
      if (Character.getType(ch) == Character.CONTROL)
      {
         code = (mappedCode == null) ? (int) ch : mappedCode;
         if (ctrl)
         {
            if (!Character.isAlphabetic(lastKeyCode) && !isSpecialCharacter(lastKey))
            {
               code = code | 1 << 11;
            }
         }
         if (shift)
         {
            if (code == Key.VK_TAB)
            {
               code = Key.VK_BACK_TAB;
            }
            else
            {
               code = processShift(code);
            }
         }
         if (alt)
         {
            code = processAlt(code);
         }

      }
      else
      {
         // all CTRL combinations for codes 0 through 31 inclusive are normally
         // provided by the TYPED event, except the following:
         if (ctrl)
         {
            if (ch == '2' || ch == '@')
            {
               // CTRL-@
               code = 0;
            }
            else if (ch == '6' || ch == '^')
            {
               // CTRL-^
               code = 30;
            }
            else if (ch == '-' || ch == '_')
            {
               // CTRL-_
               code = 31;
            }
            else if (ch == '|')
            {
               code = 28;
            }
            else if (ch == '}')
            {
               code = 29;
            }
         }
         // normal processing: the preprocessed character is returned directly
         if (code == -1)
         {
            // ALT/ENTER/DELETE are processed in keyPressed() and not here
            if (ch != KeyEvent.VK_ALT && ch != KeyEvent.VK_ENTER)
            {
               code = (int) ch;
               
               if (alt)
               {
                  // Remapping lowercase Latin to uppercase. This differs
                  // with Progress but seems to be more convenient to user because
                  // Ctrl+<letter> combinations are case insensitive (and even
                  // keyboard layout insensitive). Note the usage of direct
                  // calculation not Character class methods. This is
                  // because we are dealing with integer key codes not characters
                  // so this is dirty hack anyway.
                  if (code >= 'a' && code <= 'z')
                  {
                     code -= 'a' - 'A';
                  }
               }
            }
         }
         
         if (code != -1)
         {
            // honor the alt modifier if it was on previously
            if (alt)
            {
               code = processAlt(code);
            }
            
            if (shift)
            {
               code = processShift(code);
            }
         }
      }
      if (code != -1)
      {
         queue.add(new Integer(code));
      }
   }
   
   /**
    * Invoked when a key has been released. Retrieves the last keystroke from the keystrokes queue
    * if there exists and check it against the current key release event. At that moment
    * the keystrokes queue doesn't contain key down events which can produce key typed events,
    * key release event can change the order of key down events thus we need here to find
    * the corresponding down event and map it to Progress 4GL code.
    * 
    * @param    evt
    *           The key release event.
    */
   public synchronized void keyReleased(KeyEvent evt)
   {
      int code = -1;
      int key = evt.getKeyCode();
      KeyEvent lastKey = findKeyEventByKeyCode(key);
      if (lastKey != null)
      {
         int lastKeyCode = lastKey.getKeyCode();
         if (lastKeyCode == Key.VK_CTRL_C || lastKeyCode == Key.VK_CTRL_S)
         {
            // on windows , physical CTRL-BREAK emits code for the CTRL-C key (code 3);
            // on linux, physical CTRL-BREAK emits code for CTRL-CTRL-S (CTRL modifier with the 
            // CTRL-S key - code 19)
            
            // either of these need to be translated into 4GL's CTRL-BREAK (code 541)t 
            code = Key.VK_CTRL_BREAK;
            evt.consume();
         }
         else
         {
            Integer mappedCode;
            if (isModifier(lastKeyCode))
            {
               mappedCode = 0;
            }
            else
            {
               mappedCode = xlate.get(lastKeyCode);
            }
            if (mappedCode != null)
            {
               code = mappedCode.intValue(); 
               if (lastKey.isControlDown())
               {
                  code = code | 1 << 11;
               }
               if (lastKey.isAltDown())
               {
                  code = code | 1 << 10;
               }
               if (lastKey.isShiftDown())
               {
                  code = code | 1 << 9;
               }
            }
         }
      }
      if (code != -1)
      {
         queue.add(new Integer(code));
      }
   }

   /**
    * Process the given key by setting the 10th bit.
    *
    * @param    code
    *           The key code to be processed.
    *           
    * @return   The processed key code or <code>-1</code> if the ALT-key combination has no label.
    */
   private int processAlt(int code)
   {
      // modify the next character (in GUI 4GL, ALT is a modifier key)
      code = code | 1 << 10;
      
      if (Keyboard.keyLabel(code).length() == 0)
      {
         code = -1;
         queue.add(code);
      }
      
      return code;
   }
   
   /**
    * Process the given key by setting the 9th bit.
    *
    * @param    code
    *           The key code to be processed.
    *           
    * @return   The processed key code or <code>-1</code> if the SHIFT-key combination has no label.
    */
   private int processShift(int code)
   {
      // ignore tab - this is treated separately
      if (code == KeyEvent.VK_TAB)
      {
         return -1;
      }

      int orig = code;
      String lblOrig = Keyboard.keyLabel(orig);
      
      if (lblOrig.startsWith("CTRL-"))
      {
         lblOrig = lblOrig.substring("CTRL-".length());
      }
      
      // modify the next character (in GUI 4GL, SHIFT is a modifier key)
      code = code | 1 << 9;
      String lbl = Keyboard.keyLabel(code);
      
      if (lbl.length() == 0 || !lbl.endsWith("-" + lblOrig))
      {
         // return original code if there is no label or if the label doesn't end with the original
         // key
         return orig;
      }
      
      return code;
   }

   /**
    * Returns <code>true</code> if, for the given key, a {@link #keyTyped} event is also raised,
    * beside a {@link #keyPressed}, when used in conjunction with the Control key.
    * 
    * @param    key
    *           The pressed key.
    *           
    * @return   <code>true</code> if the <code>CTRL-&lt;key&gt;</code> combination will be
    *           treated by {@link #keyTyped(KeyEvent)}.
    */
   private boolean ctrlTyped(int key)
   {
      return (key >= 'a' && key <= 'z') || (key >= 'A' && key <= 'Z');
   }
}