/*
** Module   : WebClientMessageTypes.java
** Abstract : defines the message header type code for all possible web client protocol messages
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 GES 20150317 Copied from ChuiWebSimulator for shared usage.
** 002 GES 20150701 Added message types for GUI implementation.
** 003 GES 20150818 Added more GUI types including MSG_SET_CURSOR_STYLE, MSG_RESTACK_WINDOWS,
**                  MSG_PAGE_LOADED, MSG_MOUSE_EVENT, MSG_SET_WINDOW_LOC, CAPTURE_MOUSE,
**                  PROCESS_MOUSE_WIDGETS, MSG_WINDOW_ACTIVATED, ENABLE_OS_EVENTS, 
**                  SET_ICONIFICATION_STATE, MSG_WINDOW_ICONIFY.
** 004 CA  20151013 Added GUI web client support for window resize, current selection synching
**                  and z-order operations.
** 005 CA  20151024 Added support for WINDOW:SENSITIVE attribute.
** 006 IAS 20160217 Added support for LOAD-MOUSE-POINTER and SET-WAIT-STATE.
** 007 CA  20160203 Added hash management operations.  Added MSG_PARSE_EDITOR_CONTENT and messages
**                  to compute intermediate text widths.
**     SBI 20160225 Added MSG_PARTIAL, MSG_PING_PONG
** 008 SBI 20160412 Added MSG_OPEN_URL to open the target page in the browser.
**     CA  20160418 Added infrastructure for message passing from converted code to a 
**                  customer-specific application or calls from the customer-specific application 
**                  into the converted code (i.e. invoke external programs, procedures or 
**                  functions).  Can be used only when the P2J client runs embedded in a 
**                  customer-specific application.
** 009 SBI 20160615 Added MSG_INVALIDATE_SELECTION.
** 010 CA  20170203 Notify the driver that the client is ready (useful for embedded mode, to 
**                  notify the embedded app the client can handle messages).
**     CA  20170228 Added PUBLISH/SUBSCRIBE/UNSUBSCRIBE extensions for global support and for 
**                  external applications.
** 011 CA  20170330 In embedded mode, get the real iframe dimension as desktop dimension.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.driver.web;

/**
 * Defines the message header type code for all possible web client protocol messages.
 */
public interface WebClientMessageTypes
{
   /** Message that the web page has finished loading/initializing. */
   public static final byte MSG_PAGE_LOADED         = (byte) 0xFF;

   /** Progress compatible keyboard message. */
   public static final byte MSG_KEY                 = (byte) 0x01;

   /** Virtual VT100 terminal keyboard message. */
   public static final byte MSG_KEY_VT100           = (byte) 0x02;

   /** Paste message. */
   public static final byte MSG_PASTE               = (byte) 0x03;

   /** Contents from the clipboard. */
   public static final byte MSG_CLIPBOARD_CONTENTS  = (byte) 0x04;

   /** Web client is requesting the current selection text. */
   public static final byte MSG_CLIPBOARD_PREPARE   = (byte) 0x05;
   
   /** Web client is sending computed paragraph height. */
   public static final byte MSG_SENT_PAR_HEIGHT     = (byte) 0x06;

   /** Web client is sending computed text height. */
   public static final byte MSG_SENT_TEXT_HEIGHT    = (byte) 0x07;
   
   /** Web client is sending computed text width. */
   public static final byte MSG_SENT_TEXT_WIDTH     = (byte) 0x08;
   
   /** Web client is sending computed font height. */
   public static final byte MSG_SENT_FONT_HEIGHT    = (byte) 0x09;
   
   /** Web client is sending computed font widths. */
   public static final byte MSG_SENT_FONT_WIDTHS    = (byte) 0x0A;

   /** Web client is informing the font has been created. */
   public static final byte MSG_DONE_CREATE_FONT    = (byte) 0x0B;

   /** Web client is informing the font has been derived. */
   public static final byte MSG_DONE_DERIVE_FONT    = (byte) 0x0C;

   /** Web client is informing of a mouse event. */
   public static final byte MSG_MOUSE_EVENT         = (byte) 0x0D;

   /** Web client is informing of new window location. */
   public static final byte MSG_SET_WINDOW_LOC      = (byte) 0x0E;

   /** Web client is informing of new top window. */
   public static final byte MSG_WINDOW_ACTIVATED    = (byte) 0x0F;

   /** Web client is informing of new window iconification state. */
   public static final byte MSG_WINDOW_ICONIFY      = (byte) 0x10;

   /** Web client is informing of new window location/dimension after resize. */
   public static final byte MSG_WINDOW_RESIZED      = (byte) 0x11;

   /** Web client is informing it finished removing the hashes. */
   public static final byte MSG_DRAW_HASH_REMOVED   = (byte) 0x12;
   
   /** Web client is sending computed text widths. */
   public static final byte MSG_SENT_TEXT_WIDTHS    = (byte) 0x13;
   
   /** Web client is sending computed text widths - from batch. */
   public static final byte MSG_SENT_TEXT_WIDTHS_BATCH = (byte) 0x14;
   
   /** A request to (or response from) a customer-specific application into 4GL code. */
   public static final byte MSG_P2J_INVOKE          = (byte) 0x15;
   
   /** A request to (or response from) a message to delete a handle. */
   public static final byte MSG_P2J_DELETE_HANDLE   = (byte) 0x16;
   
   /** 
    * A request to (or response from) a call from 4GL code into the customer-specific 
    * application.
    */
   public static final byte MSG_REMOTE_CALL         = (byte) 0x17;

   /** FWD Client is ready  */
   public static final byte MSG_CLIENT_READY        = (byte) 0x18;

   /** 
    * A request to (or response from) a customer-specific application to subscribe to a named
    * event.
    */
   public static final byte MSG_P2J_SUBSCRIBE       = (byte) 0x19;

   /** 
    * A request to (or response from) a customer-specific application to unsubscribe from a named
    * event.
    */
   public static final byte MSG_P2J_UNSUBSCRIBE     = (byte) 0x1A;

   /** 
    * A request for (or response from) a publish event generated by the business logic and sent to
    * the remote side.
    */
   public static final byte MSG_P2J_PUBLISH         = (byte) 0x1B;

   /** 
    * A request for (or response from) a call to create a new external resource.
    */
   public static final byte MSG_P2J_CREATE_RESOURCE = (byte) 0x1C;
   
   /** Web client is sending desktop dimension. */
   public static final byte MSG_SENT_DESKTOP_DIMENSION = (byte) 0x1D;

   /** Clear screen message. */
   public static final byte MSG_CLEAR               = (byte) 0x80;

   /** Draw text on screen. */
   public static final byte MSG_DRAW                = (byte) 0x81;
   
   /** Set cursor position. */
   public static final byte MSG_CURSOR_POS          = (byte) 0x82;
   
   /** Set cursor status. */
   public static final byte MSG_CURSOR_STATUS       = (byte) 0x83;
   
   /** Beep message. */
   public static final byte MSG_BEEP                = (byte) 0x84;
   
   /** Quit message. */
   public static final byte MSG_QUIT                = (byte) 0x85;

   /** Switch input processing to virtual VT100 terminal mode. */
   public static final byte MSG_VT100               = (byte) 0x86;
   
   /** Request clipboard contents from web client. */
   public static final byte MSG_READ_CLIPBOARD      = (byte) 0x87;
   
   /** Copy the given text to the web client's clipboard. */
   public static final byte MSG_WRITE_CLIPBOARD     = (byte) 0x88;
   
   /** Create a new top-level window. */
   public static final byte MSG_CREATE_WINDOW       = (byte) 0x89;
   
   /** Create a new child window. */
   public static final byte MSG_CREATE_CHILD_WINDOW = (byte) 0x8A;
   
   /** Destroy the specified window or child window. */
   public static final byte MSG_DESTROY_WINDOW      = (byte) 0x8B;
   
   /** Change the visibility for the specified window or child window. */
   public static final byte MSG_WINDOW_VISIBILITY   = (byte) 0x8C;
   
   /** Ask the client for the paragraph height. */
   public static final byte MSG_GET_PAR_HEIGHT      = (byte) 0x8D;

   /** Ask the client for the text height. */
   public static final byte MSG_GET_TEXT_HEIGHT     = (byte) 0x8E;
   
   /** Ask the client for the text width. */
   public static final byte MSG_GET_TEXT_WIDTH      = (byte) 0x8F;
   
   /** Ask the client for the font height. */
   public static final byte MSG_GET_FONT_HEIGHT     = (byte) 0x90;
   
   /** Ask the client for the font widths. */
   public static final byte MSG_GET_FONT_WIDTHS     = (byte) 0x91;
   
   /** Ask the client to create a font. */
   public static final byte MSG_CREATE_FONT         = (byte) 0x92;
   
   /** Ask the client to derive a font. */
   public static final byte MSG_DERIVE_FONT         = (byte) 0x93;
   
   /** Set the cursor style. */
   public static final byte MSG_SET_CURSOR_STYLE    = (byte) 0x94;
   
   /** Restack the z-order of all windows. */
   public static final byte MSG_RESTACK_WINDOWS     = (byte) 0x95;

   /** Send the mouse-aware widget details. */
   public static final byte PROCESS_MOUSE_WIDGETS   = (byte) 0x96;
   
   /** Disable/enable mouse event processing. */
   public static final byte CAPTURE_MOUSE           = (byte) 0x97;
   
   /** Disable/enable OS event processing. */
   public static final byte ENABLE_OS_EVENTS        = (byte) 0x98;
   
   /** Window state changed event. */
   public static final byte SET_ICONIFICATION_STATE = (byte) 0x99;
   
   /** Resizeable widget. */
   public static final byte SET_RESIZEABLE_WINDOW   = (byte) 0x9A;
   
   /** Notifies the client about the current selection. */
   public static final byte MSG_CURRENT_SELECTION   = (byte) 0x9B;
   
   /** Requests the client to move a window to the top of the z-order. */
   public static final byte MSG_MOVE_TO_TOP         = (byte) 0x9C;
   
   /** Requests the client to move a window to the bottom of the z-order. */
   public static final byte MSG_MOVE_TO_BOTTOM      = (byte) 0x9D;
   
   /** Change the sensitivity for the specified window or child window. */
   public static final byte MSG_WINDOW_SENSITIVITY  = (byte) 0x9E;
   
   /** Register hoverable widget */
   public static final byte MSG_REGISTER_HOVERABLE_WIDGET  = (byte) 0x9F;

   /** Deregister widget */
   public static final byte MSG_DEREGISTER_WIDGET  = (byte) 0xA0;

   /** Set mouse pointer */
   public static final byte MSG_SET_MOUSE_PTR  = (byte) 0xA1;

   /** Lock mouse pointer */
   public static final byte MSG_LOCK_MOUSE_PTR  = (byte) 0xA2;

   /** Lock mouse pointer */
   public static final byte MSG_SET_WAIT_STATE  = (byte) 0xA3;
   
   /** Message to parse editor content. */
   public static final byte MSG_PARSE_EDITOR_CONTENT = (byte) 0xA4;
   
   /** Ask the client for the text's widths. */
   public static final byte MSG_GET_TEXT_WIDTHS = (byte) 0xA5;
   
   /** Ask the client for the texts' widths - in batch. */
   public static final byte MSG_GET_TEXT_WIDTHS_BATCH = (byte) 0xA6;
   
   /** Ask the client for the desktop dimension. */
   public static final byte MSG_GET_DESKTOP_DIMENSION = (byte) 0xA7;

   /** Requests the client if the target font is installed. */
   public static final byte MSG_IS_FONT_INSTALLED  = (byte) 0xF1;

   /** Inform the javascript side to remove the list of expired hashes. */
   public static final byte MSG_REMOVE_EXPIRED_HASH = (byte) 0xF2;

   /** Open the new page or resource in the browser. */
   public static final byte MSG_OPEN_URL = (byte) 0xF3;

   /** Notify that the current selection is invalidated. */
   public static final byte MSG_INVALIDATE_SELECTION = (byte) 0xF4;

   /** Wraps the large message from the client to the sequence of this partial type messages. */
   public static final byte MSG_PARTIAL = (byte) 0xFE;
   
   /** Ping/Pong messages. */
   public static final byte MSG_PING_PONG = (byte) 0xFC;
}