/*
** Module   : p2j.logger.js
** Abstract : Application logger.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 SBI  20150905 The first version of the p2j logger.
** 002 SBI  20151217 Added to turn on/off logging.
** 003 SBI  20160415 Changed log() to accept unspecified number of parameters.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

"use strict";

p2j.logger = (function()
{
   /** Generic logger. */
   var me = {};
   
   /** Storage for log entries. */
   var logstack = new Array();
   
   /** The stack capacity, the last visible history.*/
   var maxStackSize;

   /** The log is on.*/
   var isOn;
   
   /**
    * Initialize the logger with the storage capacity.
    * 
    * @param    {Number}
    *           The log stack capacity.
    * @param    {boolean}
    *           The boolean value that indicates the log is on/off.
    */
   me.init = function(capacity, on)
   {
      maxStackSize = capacity;
      isOn = on;
   };

   /**
    * Adds messages to the log stack. It supports the format string with %d place holders. The 
    * parameters are listed according their positions in the formatted message. It can be used
    * with one argument log(message) or with many arguments log(format, [parameter1, ...]).
    * In the first case the unique argument is a pure text string and in the second case the first
    * argument is a format string and the others are formats parameters.
    * 1) p2j.logger.log("pure log message");
    * 2) p2j.logger.log("log message with parameters %d and %d", parameter1, parameter2);
    */
   me.log = function()
   {
      if (arguments.length <= 0 || arguments.length >= 256)
      {
         return;
      }
      if (isOn)
      {
         var format = arguments[0];
         var parameters = [];
         for ( var i = 1 ; i < arguments.length; i++ )
         {
            parameters.push(arguments[i]);
         }
         save(format.split("%d").map(
               function(currValue, index, arr)
               {
                  if (this[index] === undefined)
                  {
                     return currValue;
                  }
                  return currValue + this[index];
               }, parameters).reduce(
                     function(prevValue, currValue)
                     {
                        return prevValue + currValue;
                     }, ""));
      }
   };
   
   /**
    * Log an error message.
    * 
    * @param    {string} text
    *           The error text to display.
    */
   me.error = function(text)
   {
      save(text);
      console.error(text);
      if (p2j.sound)
      {
         p2j.sound.beep();
      }
   };
   
   /**
    * Saves messages to the log stack.
    * 
    * @param    {String} msg
    *           The text message to log.
    */
   function save(msg)
   {
      var size = logstack.push(msg);
      
      /**
       * if the stack size exceeds 2 times the maximal stack size, then it is shifted to the left
       * and its size is reduced to the maximal stack size.
       */
      if (size > 2 * maxStackSize)
      {
         logstack.splice(0, maxStackSize);
      }
   };
   
   /**
    * Display a prompt that shows the current log output and allows it to be copied to the
    * system clipboard.
    */
   me.viewLog = function()
   {
      var text = "<!DOCTYPE html>\n"                         +
                 "<html>\n"     +
                 "   <head>\n"                               +
                 "      <title>Log Output</title>\n"         +
                 "   </head>\n"                              +
                 "   <body>\n"                               +
                 "      <textarea rows=\"10\" cols=\"60\">\n";

      for (var i = 0; i < logstack.length; i++)
      {
         text = text + "\n" + logstack[i];
      }
      
      text = text                  +
             "      </textarea>\n" +
             "   </body>\n"        +
             "</html>\n";
      
      window.open("data:text/html;charset=utf-8," + encodeURIComponent(text),
                  "",
                  "width=500,height=500");
   };
   return me;
}());