/*
** Module   : EventManager.java
** Abstract : manages events and event queues
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 SIY 20101122 Cleaned up imports after switching from KeyEvent 
**                  to KeyInput.
** 004 SIY 20101202 Removed unnecessary handling of mouse events.
** 005 SIY 20110112 Refactoring and simplification.
** 006 GES 20111012 Moved static data to context-local. This allows multiple
**                  client sessions to be hosted in the server.
** 007 CA  20130822 Added queue where to save server-side events.
** 008 CA  20130908 Added API to remove certain server events from the queue.
** 009 CA  20140805 Added mouse event support.
** 010 CA  20141021 Fixes for legacy MOUSE event processing: avoid posting mouse events if the
**                  window can not receive them (is busy with something else, i.e resize).
** 011 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 012 CA  20150415 Allow save/restoration of the OS event queue.
** 013 GES 20150507 Added a comment to clarify that AWT events usage is intentional.
** 014 CA  20150910 Removed handleMouseEvents() check.
** 015 CA  20150925 Fixed mouse entry/exit: explicit source is passed to the MouseEvt.
**                  Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
**                  Changes to allow multi-window focus management in GUI.
** 016 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 017 HC  20151024 Added support for invoking code on event thread.
** 018 HC  20151213 Improved DIALOG-BOX window parenting logic and related changes.
** 019 OM  20170623 Avoid NPE when osEvents is empty.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.event;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.*;

import java.util.*;

import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Manages a stack of event queues, allows posting events to the current queue and getting events 
 * from the current queue.  This includes both UI events and events generated by the server side
 * (like the {@link Keyboard#PROCEDURE_COMPLETE} event).
 * <p>
 * Documentation or testing showed no info which would suggest that keys/server events are 
 * processed in a certain way or chained in a certain way. When a server-event is posted, it will 
 * be chained in the event queue, but there is no guarantee where will be chained relative to  
 * other non-server events.  4GL enforces the server-event order only relative to other  
 * server-events and this only in case the events are not generated by truly asynchronous 
 * requests (like in the case of an appserver connected in Session-managed operating mode); 
 * for truly asynchronous requests (like in the case of an appserver connected in Session-free 
 * operating mode), the order of the server-events is undefined.
 * <p>
 * For this reason and for the fact that server-side events can be posted even when a WAIT-FOR 
 * loop is not in effect, the server-events are kept in a global {@link WorkArea#serverEvents 
 * event queue}.
 */
public class EventManager
{
   /** Stores context-local state variables. */
   private static final ContextLocal<WorkArea> work = 
   new ContextLocal<WorkArea>()
   {
      protected synchronized WorkArea initialValue()
      {
         return new WorkArea();
      }
   };
   
   /**
    * Hidden constructor, all usage of this class will be static.
    */
   private EventManager()
   {
   }
      
   /**
    * Return next available event or <code>null</code> if queue is empty.
    * 
    * @return  next available event.
    */
   public static Event getEvent()
   {
      WorkArea wa = work.get();
      
      Event evt = null;
      
      synchronized (wa)
      {
         evt = wa.queue.get();
      }
      
      return evt;
   }

   /**
    * Install new event queue.
    */
   public synchronized static void pushEventQueue()
   {
      WorkArea wa = work.get();
      
      synchronized (wa)
      {
         wa.stack.push(wa.queue);
         wa.queue = new EventQueue();
      }
   }

   /**
    * Restore old event queue.
    */
   public synchronized static void popEventQueue()
   {
      WorkArea wa = work.get();
      
      synchronized (wa)
      {
         if (wa.stack.isEmpty())
            throw new RuntimeException("Unbalanced PUSH/POP Event queue!!!");
         
         EventQueue save = wa.queue;
         
         wa.queue = wa.stack.pop();
         
         wa.queue.insert(save);
      }
   }

   /**
    * Install new OS event queue.
    */
   public synchronized static void pushOSEventQueue()
   {
      WorkArea wa = work.get();
      
      synchronized (wa)
      {
         wa.osStack.push(wa.osEvents);
         wa.osEvents = new EventQueue();
      }
   }

   /**
    * Restore old OS event queue.
    */
   public synchronized static void popOSEventQueue()
   {
      WorkArea wa = work.get();
      
      synchronized (wa)
      {
         if (wa.osStack.isEmpty())
            throw new RuntimeException("Unbalanced PUSH/POP Event queue!!!");
         
         wa.osEvents = wa.osStack.pop();
      }
   }

   /**
    * Saves the specified mouse event in the {@link WorkArea#osEvents} queue.
    * 
    * @param    event
    *           The mouse event.
    * @param    windowId
    *           The window where the event was captured.
    * 
    * @return   <code>true</code> if the event was posted.
    */
   public static boolean postMouseEvent(MouseEvent event, int windowId)
   {
      TitledWindow<?> window = WindowManager.findWindow(windowId);
      if (window == null)
      {
         return false;
      }

      return postOSEvent(new MouseEvt(event, window));
   }
   
   /**
    * Saves the specified mouse event in the {@link WorkArea#osEvents} queue.
    * 
    * @param    event
    *           The mouse event.
    * @param    windowId
    *           The window where the event was captured.
    * @param    sourceId
    *           The explicit ID of the widget to which the mouse event needs to be posted.
    * 
    * @return   <code>true</code> if the event was posted.
    */
   public static boolean postMouseEvent(MouseEvent event, int windowId, int sourceId)
   {
      TitledWindow<?> window = WindowManager.findWindow(windowId);
      if (window == null)
      {
         return false;
      }

      return postOSEvent(new MouseEvt(event, window, sourceId));
   }
   
   /**
    * Post an OS event (which needs to be processed by the 4GL drawing layer, but not 
    * necessarily as a legacy 4GL event) to the event queue.
    * 
    * @param    event
    *           The event to be posted.
    *           
    * @return   <code>true</code> if the event was posted.
    */
   public static boolean postOSEvent(Event event)
   {
      WorkArea wa = work.get();

      synchronized (wa.osEvents)
      {
         wa.osEvents.post(event);
      }
      
      return true;
   }

   /**
    * Check if there are pending mouse events in the {@link WorkArea#osEvents} queue.
    * 
    * @return   <code>true</code> if there are pending mouse events, <code>false</code> otherwise.
    */
   public static boolean hasOSEvents()
   {
      WorkArea wa = work.get();
      
      synchronized (wa.osEvents)
      {
         return !wa.osEvents.isEmpty();
      }
   }
   
   /**
    * Get the next available OS event from the {@link WorkArea#osEvents} queue.
    * 
    * @return   The next available SO event or <code>null</code> if the queue is empty.
    */
   public static Event getOSEvent()
   {
      WorkArea wa = work.get();
      Event evt;
      
      synchronized (wa.osEvents)
      {
         evt = wa.osEvents.get();
         
         if (evt == null)
         {
            return null; // osEvents is empty
         }
         
         if (evt instanceof InvocationEvent)
         {
            // this event has no associated source, so just return
            return evt;
         }
         
         // check if the event is for a dead window - if so, ignore it
         Widget<?> source = evt.source();
         if (source == null || !source.isDisplayed() || 
             !WindowManager.isDisplayed(source.topLevelWindow()))
         {
            return null;
         }
      }
      
      return evt;
   }

   /**
    * Convenient shortcut for posting events.
    * 
    * @param   event
    *          Event to post.
    */
   public static void postEvent(Event event)
   {
      WorkArea wa = work.get();
      
      synchronized (wa)
      {
         // do not post events to the main queue because they are just wasted;
         // there is no method in P2J that would process those events, ever,
         // because P2J always brackets events it cares about with push and pop
         if (wa.stack.size() == 0)
         {
            return;
         }
      
         wa.queue.post(event);
      }
   }

   /**
    * Saves the specified server event in the {@link WorkArea#serverEvents} queue, as server-side
    * events can be created if if a UI-blocking statement is not in effect.
    * 
    * @param    event
    *           The server event.
    */
   public static void postServerEvent(ServerEvent event)
   {
      WorkArea wa = work.get();
      
      synchronized (wa.serverEvents)
      {
         wa.serverEvents.post(event);
      }
   }

   /**
    * Remove from the server event queue all the events in the given list, identified by their id.
    * 
    * @param    eventIds
    *           The list of event IDs to be removed.
    */
   public static void removeServerEvents(int[] eventIds)
   {
      WorkArea wa = work.get();
      
      // build a set so we can ease lookup
      Set<Integer> ids = new HashSet<Integer>();
      for (int id : eventIds)
      {
         ids.add(id);
      }
      
      synchronized (wa.serverEvents)
      {
         // this will be a queue with the "clean" events in reverse order
         Deque<Event> cleanQueue = new ArrayDeque<Event>();
         
         // collect all events, excluding the unwanted events
         while (!wa.serverEvents.isEmpty())
         {
            ServerEvent event = (ServerEvent) wa.serverEvents.get();
            
            if (!ids.contains(event.id()))
            {
               cleanQueue.push(event);
            }
         }
         
         // add the clean events back to the queue, in the proper order
         while (!cleanQueue.isEmpty())
         {
            Event event = cleanQueue.pop();

            wa.serverEvents.post(event);
         }
      }
   }
   
   /**
    * Get the next available server event from the {@link WorkArea#serverEvents} queue.
    * 
    * @return   The next available server event or <code>null</code> if the queue is empty.
    */
   public static ServerEvent getServerEvent()
   {
      WorkArea wa = work.get();
      
      ServerEvent evt = null;
      
      synchronized (wa.serverEvents)
      {
         evt = (ServerEvent) wa.serverEvents.get();
      }
      
      return evt;
   }

   /**
    * Check if there are pending server events in the {@link WorkArea#serverEvents} queue.
    * 
    * @return   <code>true</code> if there are pending server events, <code>false</code> otherwise.
    */
   public static boolean hasServerEvents()
   {
      WorkArea wa = work.get();
      
      synchronized (wa.serverEvents)
      {
         return !wa.serverEvents.isEmpty();
      }
   }
   
   /**
    * Convenient shortcut for posting key events.
    * 
    * @param   key
    *          Key to post.
    */
   public static void postKey(int key)
   {
      Widget<?> focus = UiUtils.getCurrentFocus();
      if (focus == null)
      {
         // no focus found, send to active top-level window
         focus = WindowManager.topLevelWindow();
      }

      postEvent(getKeyEvent(key, focus, false));
   }

   /**
    * Convert a key code into a complete event (which includes the event's
    * source widget) by using the current widget in focus.
    *
    * @param    key
    *           The key code of the event which was read from the terminal.
    *
    * @return   The complete event including a valid source widget or
    *           <code>null</code> in the case of any error.
    */
   public static KeyInput eventFromKey(int key)
   {
      if (key == -1)
         return null;

      Widget<?> focus = UiUtils.getCurrentFocus();
      if (focus == null)
      {
         // no focus found, send to active top-level window
         focus = WindowManager.topLevelWindow();
      }

      return getKeyEvent(key, focus, true);
   }
   
   /**
    * Creates an instance of <code>keyEvent</code> for specified key.
    *
    * @param   key
    *          The key code.
    * @param   src
    *          The source widget for this event.
    * @param   isReal
    *          Flag which marks real keys pressed by user.
    *          
    * @return  New event instance generated for specified key and widget.
    */
   private static KeyInput getKeyEvent(int key, Widget<?> src, boolean isReal)
   {
      //TODO: fix character check here, at present it handles only ASCII
      //Perhaps Character.isValidCodePoint(key) + some additional check may
      //help us filter out correct keys.
      int id = (key > 255 || key < ' ') ? EventType.KEY_PRESSED :
                                          EventType.KEY_TYPED;
      
      return new KeyInput(src, id, key, false, isReal);
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {  
      /** Queue stack (saved queues). */
      private Stack<EventQueue> stack = new Stack<EventQueue>();
   
      /** Queue stack (saved OS queues). */
      private Stack<EventQueue> osStack = new Stack<EventQueue>();
   
      /** Event queue currently in use. */
      private EventQueue queue = new EventQueue();
      
      /** Queue dedicated for registered events originating from the server-side. */
      private EventQueue serverEvents = new EventQueue();
      
      /** Queue dedicated for OS events. */
      private EventQueue osEvents = new EventQueue();
   }
}
