/*
** Module   : MouseEvt.java
** Abstract : Container for a mouse event, to be processed later by the 4GL logic.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.  ThinClient is explicitly imported.
** 002 CA  20141021 Added API to translate from OS mouse event to legacy 4GL mouse event(s).
** 003 CA  20141029 Switched from event codes to event labels, when raising a portable mouse event
**                  (as the event codes are dependent on the client's compatibility mode (ChUI/GUI,
**                  Windows/Linux).
** 004 EVL 20150401 Making mouse event the extending of the KeyInput.  Moving mouse down 4GL event
**                  generation to the MOUSE_PRESSED clause.
** 005 GES 20150507 Added a comment to clarify that AWT events usage is intentional.
** 006 EVL 20150707 Adding flag and method to work with reposted mouse events.
** 007 CA  20150831 Added codes for complex events which can be processed at the driver level.
** 008 CA  20150925 Fixed mouse entry/exit: explicit source is passed to the MouseEvt.
** 009 HC  20160119 Added toString() method.
** 010 CA  20160327 Added support for LAST-EVENT:X/Y attributes: if the widget doesn't resolve a
**                  coordinate for these attributes, then do not raise a portable mouse event.
** 011 VIG 20160216 Added isLeftButton() method.
** 012 SBI 20160612 Added COPY, CUT, PASTE, MINIMIZE, MAXIMIZE and CLOSE mouse actions.
** 013 SBI 20160708 Renamed "editable" to "hoverable" due to the implemented functionalities.
** 014 HC  20160922 Improvements to mouse input.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.event;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.*;

import java.util.*;

import com.goldencode.p2j.ui.Keyboard;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.widget.*;

import static com.goldencode.p2j.ui.Keyboard.*;

/**
 * Container for mouse events.  Instances of this class are posted via 
 * {@link ThinClient#postMouseEvent} to a specialized event queue only if the business logic is
 * in a event processing loop, pause or readkey.
 */
public class MouseEvt
extends KeyInput
{
   /** The ID of mouse hoverable event, processed at the driver. */
   public static final int MOUSE_HOVERABLE = -1;

   /** The ID of mouse movable event, processed at the driver. */
   public static final int MOUSE_MOVABLE = -2;

   /** The ID of window minimized event, processed at the driver. */
   public static final int WINDOW_ICONIFIED = -3;

   /** The ID of window maximized event, processed at the driver. */
   public static final int WINDOW_MAXIMIZED= -4;

   /** The ID of window close event, processed at the driver. */
   public static final int WINDOW_CLOSED = -5;

   /** The ID of mouse popup on left click event, processed at the driver. */
   public static final int MOUSE_POPUPABLE_LEFT = -6;

   /** The ID of mouse popup on right click event, processed at the driver. */
   public static final int MOUSE_POPUPABLE_RIGHT = -7;

   /**
    * The system menu copy action id, required special processing for the web client
    * to copy the selection to the system clipboard.
    */
   public static final int COPY  = -8;

   /**
    * The system menu cut action id, required special processing for the web client
    * to cut the selection to the system clipboard.
    */
   public static final int CUT   = -9;

   /**
    * The system menu paste action id, required special processing for the web client
    * to paste from the system clipboard.
    */
   public static final int PASTE = -10;

   /**
    * The system menu minimize action id, required special processing for the web client
    * to minimize the target window.
    */
   public static final int MINIMIZE = -11;
   
   /**
    * The system menu maximize action id, required special processing for the web client
    * to maximize the target window.
    */
   public static final int MAXIMIZE = -12;
   
   /**
    * The system menu close action id, required special processing for the web client
    * to close the target window.
    */
   public static final int CLOSE = -13;
   
   /** Mouse click event. */
   public static final int MOUSE_CLICKED = MouseEvent.MOUSE_CLICKED;
   
   /** Mouse press event. */
   public static final int MOUSE_PRESSED = MouseEvent.MOUSE_PRESSED;
   
   /** Mouse release event. */
   public static final int MOUSE_RELEASED = MouseEvent.MOUSE_RELEASED;
   
   /** The captured AWT mouse event. */
   private final MouseEvent event;
   
   /** Flag indicating if the event is original or reposted. */
   private boolean original = true;

   /** The explicit ID of the widget to which the mouse event needs to be posted. */
   private Integer widgetId = null;
   
   /**
    * Create a new mouse event in the specified window.
    * 
    * @param    event
    *           The AWT mouse event.
    * @param    window
    *           The window where the event originated.
    */
   public MouseEvt(MouseEvent event, TitledWindow<?> window)
   {
      super(window, -1);
      
      this.event = event;
   }
   
   /**
    * Create a new mouse event in the specified window.
    * 
    * @param    event
    *           The AWT mouse event.
    * @param    window
    *           The window where the event originated.
    * @param    widgetId
    *           The explicit ID of the widget to which the mouse event needs to be posted.
    */
   public MouseEvt(MouseEvent event, TitledWindow<?> window, int widgetId)
   {
      this(event, window);
      
      this.widgetId = widgetId;
   }
   
   /**
    * Get the AWT mouse event.
    * 
    * @return   See above.
    */
   public MouseEvent getEvent()
   {
      return event;
   }

   /**
    * Get the nature of the mouse event.
    * 
    * @return   <CODE>TRUE</CODE> if original, <CODE>FALSE</CODE> if artificially reposted.
    */
   public boolean isOriginal()
   {
      return original;
   }
   
   /**
    * Check the button state: if used button was the left button.  
    * 
    * @return   <code>true</code> if it left button was used.
    */
   public boolean isLeftButton()
   {
      return getEvent().getButton() == MouseEvent.BUTTON1;
   }

   /**
    * Set the flag indicating the nature of the mouse event.
    * 
    * @param    flag
    *           <CODE>TRUE</CODE> if original, <CODE>FALSE</CODE> if artificially reposted.
    */
   public void setOriginal(boolean flag)
   {
      original = flag;
   }

   /**
    * Get the legacy 4GL events generated by this OS mouse event.  Note that the physical mouse
    * events (i.e. LEFT_MOUSE_*, RIGHT_MOUSE_*) are just aliases for the portable mouse events.
    * <p>
    * The rules to transform a OS mouse event into a portable mouse event are:
    * <ul>
    * 
    * <li>if the {@link MouseEvent#BUTTON1} is pressed (left-mouse-button):
    * <ol>
    * <li>{@link MouseEvent#MOUSE_PRESSED} and {@link MouseEvent#MOUSE_RELEASED} are ignored if
    *     {@link MouseEvent#getClickCount()} is not zero.</li>
    * <li>{@link MouseEvent#MOUSE_PRESSED} is transformed into a 
    *     {@link Keyboard#MOUSE_SELECT_DOWN} or {@link Keyboard#MOUSE_EXTEND_DOWN}, if CTRL
    *     is pressed.</li>
    * <li>{@link MouseEvent#MOUSE_RELEASED} is transformed into a 
    *     {@link Keyboard#MOUSE_SELECT_UP} or {@link Keyboard#MOUSE_EXTEND_UP}, if CTRL
    *     is pressed.</li>
    * <li>{@link MouseEvent#MOUSE_CLICKED} is transformed into a 
    *     {@link Keyboard#MOUSE_SELECT_DOWN} followed by a {@link Keyboard#MOUSE_SELECT_UP},
    *     {@link Keyboard#MOUSE_EXTEND_DOWN} followed by a {@link Keyboard#MOUSE_EXTEND_UP},
    *     if CTRL is pressed.</li>
    * <li>if a double-click is generated, then another {@link Keyboard#MOUSE_SELECT_UP} or
    *     {@link Keyboard#MOUSE_EXTEND_UP} are generated, this time having as function
    *     {@link Keyboard#MOUSE_SELECT_DBLCLICK} or {@link Keyboard#MOUSE_EXTEND_DBLCLICK}.
    * </ol>
    * </li>
    * 
    * <li>if the {@link MouseEvent#BUTTON3} is pressed (right-mouse-button):
    * <ol>
    * <li>{@link MouseEvent#MOUSE_PRESSED} and {@link MouseEvent#MOUSE_RELEASED} are ignored if
    *     {@link MouseEvent#getClickCount()} is not zero.</li>
    * <li>{@link MouseEvent#MOUSE_PRESSED} is transformed into a 
    *     {@link Keyboard#MOUSE_MENU_DOWN}.</li>
    * <li>{@link MouseEvent#MOUSE_RELEASED} is transformed into a 
    *     {@link Keyboard#MOUSE_MENU_UP}.</li>
    * <li>{@link MouseEvent#MOUSE_CLICKED} is transformed into a 
    *     {@link Keyboard#MOUSE_MENU_DOWN} followed by a {@link Keyboard#MOUSE_MENU_UP}.
    *     </li>
    * <li>if a double-click is generated, then another {@link Keyboard#MOUSE_MENU_UP} is
    *     generated, this time having as function {@link Keyboard#MOUSE_MENU_DBLCLICK}.
    * </ol>
    * </li>
    * 
    * <li>middle-button events ({@link MouseEvent#BUTTON2}) are not supported by 4GL, even if
    *     triggers can be defined for these events.</li>
    * 
    * <li>KEYCODE(MOUSE_MOVE_*) for portable events reports the same codes as for the 
    *     MOUSE_SELECT_* events.  These might not be real events but just actions triggered for
    *     the MOUSE_SELECT_* events (in certain circumstances), but this can't be duplicated yet.
    *     TODO: find the correct behavior for these events.</li> 
    * </ul>
    * @param    source
    *           The source widget.
    * 
    * @return   A list of {@link PortableMouseEvent events}.
    */
   public List<PortableMouseEvent> legacyEvents(Widget<?> source)
   {
      List<PortableMouseEvent> evts = new ArrayList<>();
      int button = event.getButton();
      int clicks = event.getClickCount();
      boolean ctrl = event.isControlDown();
      int x = event.getPoint().x;
      int y = event.getPoint().y;

      // portable mouse events are not raised for the window's status bar, the window border, the
      // window's title or the frame's title; let the widget compute its mouse event origin;
      // for widgets attached to frames, it's relative to the frame's top-left corner; otherwise,
      // it's relative to the window's workspace top-left corner

      NativePoint np = new NativePoint(x, y);
      NativePoint eventOrigin = ((AbstractWidget) source).getPortableMouseEventCoordinates(np);
      if (eventOrigin == null)
      {
         return Collections.emptyList();
      }

      // set the event's X/Y coordinates based on the source
      x = eventOrigin.x;
      y = eventOrigin.y;
      
      switch (event.getID())
      {
         case MouseEvent.MOUSE_CLICKED:
            switch (button)
            {
               case MouseEvent.BUTTON1:
                  // odd number of clicks trigger click events
                  if (clicks % 2 == 1)
                  {
                     evts.add(new PortableMouseEvent(source,
                                                     ctrl ? MOUSE_EXTEND_UP
                                                          : MOUSE_SELECT_UP,
                                                     ctrl ? MOUSE_EXTEND_CLICK
                                                          : MOUSE_SELECT_CLICK,
                                                     x,
                                                     y));
                  }
                  
                  // even number of clicks trigger double click events
                  if (clicks % 2 == 0)
                  {
                     evts.add(new PortableMouseEvent(source,
                                                     ctrl ? MOUSE_EXTEND_UP   
                                                          : MOUSE_SELECT_UP,
                                                     ctrl ? MOUSE_EXTEND_DBLCLICK 
                                                          : MOUSE_SELECT_DBLCLICK,
                                                     x,
                                                     y));
                  }
                  break;

               case MouseEvent.BUTTON3:
                  evts.add(new PortableMouseEvent(source, MOUSE_MENU_UP, MOUSE_MENU_CLICK, x, y));
                  
                  if (clicks == 2)
                  {
                     evts.add(new PortableMouseEvent(source,
                                                     MOUSE_MENU_UP,
                                                     MOUSE_MENU_DBLCLICK,
                                                     x,
                                                     y));
                  }
                  break;
            }
            break;

         case MouseEvent.MOUSE_PRESSED:
            switch (button)
            {
               case MouseEvent.BUTTON1:
                  // odd number of clicks trigger down events
                  if (clicks % 2 == 1)
                  {

                     evts.add(new PortableMouseEvent(source,
                                                     ctrl ? MOUSE_EXTEND_DOWN
                                                          : MOUSE_SELECT_DOWN,
                                                     -1,
                                                     x,
                                                     y));
                  }
                  break;
               case MouseEvent.BUTTON3:
                  evts.add(new PortableMouseEvent(source, MOUSE_MENU_DOWN, x, y)); 
                  break;
            }
            break;
         
         case MouseEvent.MOUSE_RELEASED:
            if (clicks != 0)
            {
               // wait for the OS mouse-click event
               break;
            }

            switch (button)
            {
               case MouseEvent.BUTTON1:
                  evts.add(new PortableMouseEvent(source, 
                                                  ctrl ? MOUSE_EXTEND_UP 
                                                       : MOUSE_SELECT_UP,
                                                  x,
                                                  y)); 
                  break;
               case MouseEvent.BUTTON3:
                  evts.add(new PortableMouseEvent(source, MOUSE_MENU_UP, x, y)); 
                  break;
            }
            break;
      }
      
      return evts;
   }
   
   /**
    * Get the explicit source to which this mouse event needs to be posted.
    * 
    * @return   The explicit ID of the widget to which the mouse event needs to be posted, or 
    *           <code>null</code> if the source needs to be computed.
    */
   public Widget<?> getExplicitSource()
   {
      return (widgetId == null ? null : ThinClient.getInstance().getWidget(widgetId));
   }
   
   /**
    * Returns the String representation of this object.
    *
    * @return text representation of this object
    */
   @Override
   public String toString() 
   {
      return ("MouseEvt:" + 
              " source=[" + source() + "]" + 
              " original=" + original +
              " widgetId=" + widgetId + 
              " mouse=[" + event + "]");
   }
}