/*
** Module   : PortableMouseEvent.java
** Abstract : Container for an event generated by client-side, having as source a mouse.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20141020 Created initial version.
** 002 CA  20141029 Added c'tors to allow specification of event labels instead of event codes
**                  (as the event codes are dependent on the client's compatibility mode (ChUI/GUI,
**                  Windows/Linux).
** 003 CA  20160327 Added support for LAST-EVENT:X/Y attributes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.event;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Container for an event generated by client-side, having as source a mouse.
 */
public class PortableMouseEvent
extends Event
{
   /** The 4GL action associated with this event. */
   private final int action;
   
   /** The value of the LAST-EVENT:X attribute. */
   private final int x;
   
   /** The value of the LAST-EVENT:Y attribute. */
   private final int y;

   /**
    * Create a new event for the specified widget, with the specified event label.
    * 
    * @param    widget
    *           The source widget.
    * @param    label
    *           The event label.
    * @param    x
    *           The value of LAST-EVENT:X attribute.
    * @param    y
    *           The value of LAST-EVENT:Y attribute.
    */
   public PortableMouseEvent(Widget<?> widget, String label, int x, int y)
   {
      this(widget, label, label, x, y);
   }

   /**
    * Create a new event for the specified widget, with the specified ID.
    * 
    * @param    widget
    *           The source widget.
    * @param    id
    *           The event ID.
    * @param    x
    *           The value of LAST-EVENT:X attribute.
    * @param    y
    *           The value of LAST-EVENT:Y attribute.
    */
   public PortableMouseEvent(Widget<?> widget, int id, int x, int y)
   {
      this(widget, id, id, x, y);
   }

   /**
    * Create a new event for the specified widget, with the specified ID.
    * 
    * @param    widget
    *           The source widget.
    * @param    label
    *           The event's label.
    * @param    action
    *           The associated action.
    * @param    x
    *           The value of LAST-EVENT:X attribute.
    * @param    y
    *           The value of LAST-EVENT:Y attribute.
    */
   public PortableMouseEvent(Widget<?> widget, String label, int action, int x, int y)
   {
      super(widget, Keyboard.keyCode(label));
      
      this.action = action;
      this.x = x;
      this.y = y;
   }

   /**
    * Create a new event for the specified widget, with the specified ID.
    * 
    * @param    widget
    *           The source widget.
    * @param    label
    *           The event's label.
    * @param    action
    *           The associated action.
    * @param    x
    *           The value of LAST-EVENT:X attribute.
    * @param    y
    *           The value of LAST-EVENT:Y attribute.
    */
   public PortableMouseEvent(Widget<?> widget, String label, String action, int x, int y)
   {
      super(widget, Keyboard.keyCode(label));
      
      this.action = Keyboard.keyCode(action);
      this.x = x;
      this.y = y;
   }

   /**
    * Create a new event for the specified widget, with the specified ID.
    * 
    * @param    widget
    *           The source widget.
    * @param    id
    *           The event ID.
    * @param    action
    *           The associated action.
    * @param    x
    *           The value of LAST-EVENT:X attribute.
    * @param    y
    *           The value of LAST-EVENT:Y attribute.
    */
   public PortableMouseEvent(Widget<?> widget, int id, int action, int x, int y)
   {
      super(widget, id);
      
      this.action = action;
      this.x = x;
      this.y = y;
   }
   
   /**
    * Get the {@link #action}.
    * 
    * @return    See above.
    */
   public int getAction()
   {
      return action;
   }

   /**
    * Get the {@link #x} coordinate for this portable mouse event.
    * 
    * @return   See above.
    */
   public int getX()
   {
      return x;
   }

   /**
    * Get the {@link #y} coordinate for this portable mouse event.
    * 
    * @return   See above.
    */
   public int getY()
   {
      return y;
   }
}