/*
** Module   : DatetimeFormat.java
** Abstract : datetime formatting and editing support
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20130327 Initial version.
** 002 EVL 20131016 Changed static call to getDisplayTimeZone() from date to SessionUtils due to
**                  method signature changes in date class.
** 003 CA  20141220 Changes related to GUI FILL-IN implementation.
** 004 CA  20150128 Added API to check if a format needs special validation.
** 005 OM  20150623 Added support for custom decimal separator for millis.
** 006 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.format;

import static com.goldencode.p2j.util.datetime.*;

import java.util.*;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager.ErrorEntry;

/**
 * Display format for <code>datetime</code> and <code>datetime-tz</code> variables.
 */
public class DatetimeFormat
extends DisplayFormat
{
   // TODO: already existing instances won't pick up the latest change
   //       when SESSION:DATE-FORMAT is called, but in the 4GL they will
   /** Order of date components for rendering. */
   protected byte[] dateCompOrder = date.getDateComponentOrder();

   /** Order of datetime(-tz) components for rendering. */
   protected byte[] datetimeCompOrder = null;

   /** 
    * Date components separator.
    * In addition to date, there are some hardcoded time separator:
    * <ul>
    *    <li><b>T</b>, <b>t</b> or one <b>space</b> that separates the date and time parts;
    *    <li><b>colon</b> that separates hours from minutes and minutes from seconds;
    *    <li><b>dot</b> that separates the seconds and milliseconds.
    *    <li><b>+</b> or <b>-</b> that prefixes the timezone components.
    * </ul>
    */
   protected char fmtSeparator;

   /* Flag for upper case AM/PM specifier */
   protected boolean amPmUppercase = false;

   /**
    * Determines whether the separators will be drawn for the unknown value
    * (applies to editing mode only). 
    */
   protected boolean drawSeparators = true;

   /** <code>true</code> for a 4-digit year. */
   protected boolean fmtFullYear = false;

   /**
    * How date components are positioned in rendered string. This array is
    * directly indexed by the component's code constants (DAY etc.).
    */
   protected DatetimeComponent[] fmtDatetimeComps;

   /** 
    * If a value cannot be correctly represented with current format a string of question marks
    * will be used instead. 
    */
   protected boolean allQMark = false;

   /** 
    * If set, SESSION:DISPLAY-TIMEZONE will be taken into consideration when displaying
    * values if the format misses the timezone component. 
    */
   private boolean isDatetimeTz = false;
   
   /**
    * Constructs a datetime format from the given Progress format string.
    * If isDatetimeTz is set then when displaying the value of a datetime-tz it will take into 
    * consideration the SESSION:DISPLAY-TIMEZONE if timezone specifier is missing from format. 
    * 
    * @param   pformat
    *          Progress format string.
    * @param   isDatetimeTz
    *          This format will be used with a datetimetz instance.
    * @throws  DisplayFormatParsingException
    *          When passed format is not a valid datetime format.
    */
   public DatetimeFormat(String pformat, boolean isDatetimeTz)
   throws DisplayFormatParsingException
   {
      super(pformat);
      this.isDatetimeTz = isDatetimeTz;
      pformat = pformat.replace('t', ' ').replace('T', ' '); // date & time separator is now space
      
      fmtDatetimeComps = new DatetimeComponent[10];
      
      // split date and time
      int dateTimeSepIdx = pformat.indexOf(' ');
      String dateSpec = (dateTimeSepIdx == -1) ?
                        pformat : // format is date only
                        pformat.substring(0, dateTimeSepIdx); // cut date part only
      
      // first handle 'no separator' case
      if (dateSpec.equals("99999999") || dateSpec.startsWith("999999"))
      {
         fmtSeparator = '\0';
         fmtFullYear = dateSpec.length() == 10;
         
         int offs = 0;
         for (byte compIndex : dateCompOrder)
         {
            int length = (compIndex == YEAR && fmtFullYear) ? 4 : 2;
            fmtDatetimeComps[compIndex] = new DatetimeComponent(offs, length);
            offs += length;
         }
      }
      else
      {
         fmtFullYear = false;
         int offs = 0;
         StringTokenizer dateFormatTok = new StringTokenizer(dateSpec, "/-.", true);
         for (int k = 0; k < 3; k++)
         {
            if (!dateFormatTok.hasMoreTokens())
            {
               throw new DisplayFormatParsingException(
                     "Invalid datetime format: " + getFormatDef());
            }
            else if (k == 1)
            {
               // this is the first occurrence  of a separator, save it
               fmtSeparator = dateFormatTok.nextToken().charAt(0);
               ++offs; // separator
            }
            else if (k == 2)
            {
               // second occurrence of a date separator, ignore it 
               dateFormatTok.nextToken();
               ++offs; // separator
            }
            
            // now a date component should come
            String tok = dateFormatTok.nextToken();
            int compIndex = dateCompOrder[k];
            int length = tok.length();
            
            // validate it, must be comprised from '9' only 
            for (int i = 0; i < length; i++)
            {
               if (tok.charAt(i) != '9')
               {
                  throw new DisplayFormatParsingException(
                        "Invalid datetime format: " + getFormatDef());
               }
            }
            if (compIndex == YEAR && length > 2)
            {
               fmtFullYear = true;
            }
            fmtDatetimeComps[compIndex] = new DatetimeComponent(offs, length);
            offs += length;
         }
      }
      fmtScreenWidth = (dateTimeSepIdx == -1) ? pformat.length() : dateTimeSepIdx;
      // the first 3 components (same as for date), are initialized now

      // if pformat consist only of date components, stop here
      if (fmtScreenWidth < pformat.length())
      {
         // parse am/pm time component
         int ampmOffset = pformat.indexOf("am", fmtScreenWidth);
         if (ampmOffset != -1)
         {
            fmtDatetimeComps[AMPM] = new DatetimeComponent(ampmOffset, 2);
         }
         else
         {
            ampmOffset = pformat.indexOf("AM", fmtScreenWidth);
            if (ampmOffset != -1)
            {
               fmtDatetimeComps[AMPM] = new DatetimeComponent(ampmOffset, 2);
               amPmUppercase = true;
            }
            else
            {
               ampmOffset = pformat.indexOf('A', fmtScreenWidth);
               if (ampmOffset != -1)
               {
                  fmtDatetimeComps[AMPM] = new DatetimeComponent(ampmOffset, 1);
                  amPmUppercase = true;
               }
               else
               {
                  ampmOffset = pformat.indexOf('a', fmtScreenWidth);
                  if (ampmOffset != -1)
                  {
                     fmtDatetimeComps[AMPM] = new DatetimeComponent(ampmOffset, 1);
                  }
               }
            }
         }
         
         // now we can simplify as next components are case insensitive
         pformat = pformat.toUpperCase();
         
         // parse timezone component:
         int tzMinutesOffset = -1;
         int tzHoursOffset = pformat.indexOf('+', fmtScreenWidth);
         if (tzHoursOffset == -1)
         {
            tzHoursOffset = pformat.indexOf('-', fmtScreenWidth);
         }
         if (tzHoursOffset != -1)
         {
            if ("HH".equals(safeSubstring(pformat, tzHoursOffset + 1, 2)))
            {
               // the component includes +/-
               fmtDatetimeComps[TZ_HOURS] = new DatetimeComponent(tzHoursOffset, 3);
            }
            else if (tzHoursOffset + 1 < pformat.length() && 
                     'H' == pformat.charAt(tzHoursOffset + 1))
            {
               // NOTE: Progress is a little idiot, it will always force 3 characters for hours,
               //       eventually overlapping the minutes
               // the component includes +/-
               fmtDatetimeComps[TZ_HOURS] = new DatetimeComponent(tzHoursOffset, 3);
            }
            
            tzMinutesOffset = pformat.indexOf(':', tzHoursOffset + 3);
            ++tzMinutesOffset;
            if (tzMinutesOffset != -1)
            {
               if ("MM".equals(safeSubstring(pformat, tzMinutesOffset, 2)))
               {
                  fmtDatetimeComps[TZ_MINS] = new DatetimeComponent(tzMinutesOffset, 2);
               }
               else if (tzMinutesOffset < pformat.length() && 
                        'M' == pformat.charAt(tzMinutesOffset))
               {
                  fmtDatetimeComps[TZ_MINS] = new DatetimeComponent(tzMinutesOffset, 1);
               }
            }
         }
         
         // now look for time specifiers in pformat between fmtScreenWidth and 
         // min(ampmOffset, tzHoursOffset)
         if (ampmOffset != -1)
         {
            pformat = safeSubstring(pformat, 0, ampmOffset).trim(); 
            // also removed the spaces at the end
         }
         else if (tzHoursOffset != -1)
         {
            pformat = safeSubstring(pformat, 0, tzHoursOffset).trim();
            // also removed the spaces at the end
         }
         
         int separator = pformat.indexOf(' ', fmtScreenWidth);
         if (separator == -1)
         {
            throw new DisplayFormatParsingException(
                  "Invalid datetime format: " + getFormatDef());
         }
         while (separator < pformat.length() && pformat.charAt(separator) == ' ')
         {
            // there can be multiple spaces
            ++separator;
         }
   
         if ((tzHoursOffset != -1 || ampmOffset != -1) && separator >= pformat.length())
         {
            // detected am/pm and timezome but no time ??
            throw new DisplayFormatParsingException(
                  "Invalid datetime format: " + getFormatDef());
         }
         
         // now should point to the first H if any
         if ("HH".equals(safeSubstring(pformat, separator, 2)))
         {
            fmtDatetimeComps[HOURS] = new DatetimeComponent(separator, 2);
            separator += 2;
         }
         else if (separator < pformat.length() && 'H' == pformat.charAt(separator))
         {
            // NOTE: Progress is a little idiot, it will always force 2 characters for hours,
            // eventually overlapping the minutes. It does not matter if in AM/PM: it will fill 
            // the 1st char of the reserved 2 with SPACE
            fmtDatetimeComps[HOURS] = new DatetimeComponent(separator, 2);
            separator += 2;
         }
         else
         {
            // no hours ?
            throw new DisplayFormatParsingException(
                  "Invalid datetime format: " + getFormatDef());
         }
         
         if (separator < pformat.length())
         {
            if (":MM".equals(safeSubstring(pformat, separator, 3)))
            {
               fmtDatetimeComps[MINUTES] = new DatetimeComponent(separator + 1, 2);
               separator += 3;
            }
            else if (":M".equals(safeSubstring(pformat, separator, 2)))
            {
               fmtDatetimeComps[MINUTES] = new DatetimeComponent(separator + 1, 1);
               separator += 2;
            }
            
            if (separator < pformat.length())
            {
               if (":SS".equals(safeSubstring(pformat, separator, 3)))
               {
                  fmtDatetimeComps[SECONDS] = new DatetimeComponent(separator + 1, 2);
                  separator += 3;
               }
               else if (":S".equals(safeSubstring(pformat, separator, 2)))
               {
                  fmtDatetimeComps[SECONDS] = new DatetimeComponent(separator + 1, 1);
                  separator += 2;
               }
               
               if (separator < pformat.length())
               {
                  if (".SSS".equals(safeSubstring(pformat, separator, 4)))
                  {
                     fmtDatetimeComps[MILLIS] = new DatetimeComponent(separator + 1, 3);
                     separator += 4;
                  }
                  else if (".SS".equals(safeSubstring(pformat, separator, 3)))
                  {
                     fmtDatetimeComps[MILLIS] = new DatetimeComponent(separator + 1, 2);
                     separator += 3;
                  }
                  else if (".S".equals(safeSubstring(pformat, separator, 2)))
                  {
                     fmtDatetimeComps[MILLIS] = new DatetimeComponent(separator + 1, 1);
                     separator += 2;
                  }
                  
                  if (separator < pformat.length())
                  {
                     // some unneeded chars ?
                     //   throw new DisplayFormatParsingException(
                     //         "Invalid datetime format: " + getFormatDef());
                  }
               }
            }
         }
      }
      fmtScreenWidth = getFormatDef().length(); // shouldn't it ?
      
      datetimeCompOrder = new byte[10];
      datetimeCompOrder[0] = dateCompOrder[0];
      datetimeCompOrder[1] = dateCompOrder[1];
      datetimeCompOrder[2] = dateCompOrder[2];
      int index = 3;
      if (fmtDatetimeComps[HOURS] != null)
      {
         datetimeCompOrder[index++] = HOURS;
      }
      if (fmtDatetimeComps[MINUTES] != null)
      {
         datetimeCompOrder[index++] = MINUTES;
      }
      if (fmtDatetimeComps[SECONDS] != null)
      {
         datetimeCompOrder[index++] = SECONDS;
      }
      if (fmtDatetimeComps[MILLIS] != null)
      {
         datetimeCompOrder[index++] = MILLIS;
      }
      if (fmtDatetimeComps[AMPM] != null)
      {
         datetimeCompOrder[index++] = AMPM;
      }
      if (fmtDatetimeComps[TZ_HOURS] != null)
      {
         datetimeCompOrder[index++] = TZ_HOURS;
      }
      if (fmtDatetimeComps[TZ_MINS] != null)
      {
         datetimeCompOrder[index++] = TZ_MINS;
      }
      for (; index < 10; index++)
      {
         datetimeCompOrder[index] = -1;
      }
   }
   
   /**
    * Verify if this format needs explicit validation.
    * 
    * @return   always <code>true</code>.
    */
   @Override
   public boolean isFormatCheck()
   {
      return true;
   }
   
   /**
    * Returns a new presentation instance with a default initialized insert mode and 
    * cursor position.
    * 
    * @param   pres
    *          A Presentation instance which was constructed by this format. May be 
    *          <code>null</code>. This presentation instance is not changed in this method.
    * 
    * @return  New presentation instance.
    */
   @Override
   public Presentation clear(Presentation pres)
   {
      drawSeparators = false;

      DatetimeBuf buf = new DatetimeBuf();
      
      buf.compBufs = null;

      if (pres != null)
      {
         buf.presInsertMode = pres.presInsertMode;
      }
      
      return buf;
   }

   /**
    * Initializes a presentation for variable's value.
    * 
    * @param   var
    *          Variable which should be compatible with format.
    * 
    * @return  Data presentation
    * 
    * @throws  DisplayFormatCheckException
    *          Variable's value is incompatible with format.
    */
   @Override
   public Presentation fromVar(BaseDataType var)
   throws DisplayFormatCheckException
   {
      return new DatetimeBuf((datetime)var);
   }

   /**
    * Forces separators for the unknown datetime value to be drawn or not in the
    * subsequent value rendering (applies to editing mode only).
    *
    * @param drawSeparators
    *        If <code>true</code> then separators will be displayed for the
    *        unknown datetime value.
    */
   public void setDrawSeparators(boolean drawSeparators)
   {
      this.drawSeparators = drawSeparators;
   }

   /**
    * Safely extract a substring from a string. The methods do the best effort to return the 
    * possible substring in case that initial string bounds are crossed.
    * 
    * @param   s
    *          The string to cut.
    * @param   start
    *          The offset to start the returning substring.
    * @param   len
    *          The maximum length of the substring to be returned.
    *          
    * @return  The substring of s starting at <code>start</code> having <code>len</code> chars.
    *          If arguments are wrong, an empty string is returned but never null object.
    */
   private static String safeSubstring(String s, int start, int len)
   {
      if (s == null || s.isEmpty() || len <= 0)
      {
         // treat exception cases first
         return "";
      }
      
      if (start < 0)
      {
         len += start; // len is lowered in fact because start is negative
         start = 0;
         
         if (len < 0)
         {
            return "";
         }
      }
      
      if (start + len > s.length())
      {
         // end offset to big, return only available chars 
         return s.substring(start);
      }
      
      return s.substring(start, start + len);
   }

   /**
    * The simple structure which records what places are occupied in a screen
    * value by the datetime component.
    */
   private static class DatetimeComponent
   {
      /** Offset of this component in rendered string. */
      private int compOffset;

      /** Number of digits to render. */
      private int compLength;

      /**
       * Constructor.
       * 
       * @param   offset
       *          Offset of this component in rendered string.
       * @param   length
       *          Number of digits to render.
       */
      public DatetimeComponent(int offset, int length)
      {
         compOffset = offset;
         compLength = length;
      }
      

      /**
       * Returns offset of this component in rendered string.
       * 
       * @return  Offset of this component in rendered string.
       */
      public int getOffset()
      {
         return compOffset;
      }

      /**
       * Returns number of digits to render.
       * 
       * @return  Number of digits to render.
       */
      public int getLength()
      {
         return compLength;
      }
   }

   /**
    * Internal form of datetime data.
    */
   private class DatetimeBuf
   extends Presentation
   {
      /** Special value for an empty digit place. */
      private static final char SPACE = ' ';

      /**
       * Buffers for datetime component digits. May be <code>null</code> which means UNKNOWN 
       * value. Array element must contain a digit or <code>SPACE</code> which will be rendered 
       * as space.
       */
      private char[][] compBufs = null;

      /**
       * Constructs default presentation used for clear operation.
       */
      public DatetimeBuf()
      {
         presCursorPos = 0;
         presInsertMode = false;
         initClear();
      }

      /**
       * Constructs presentation initialized with given <code>datetime</code>.
       * 
       * @param   varDatetime
       *          Initial value.
       * 
       * @throws  DisplayFormatCheckException
       *          This value can't be displayed with the current format.
       */
      public DatetimeBuf(datetime varDatetime)
      throws DisplayFormatCheckException
      {
         if (varDatetime.isUnknown())
         {
            return;
         }
         
         allQMark = false;
         if (isDatetimeTz && fmtDatetimeComps[TZ_HOURS] == null)
         {
            long milliOffset = 0;

            // use the SESSION:DISPLAY-TIMEZONE if set
            integer sdt = SessionUtils.getDisplayTimeZone();
            if (!sdt.isUnknown())
            {
               milliOffset = sdt.longValue() * 60 * 1000;
            }

            int tzHours   = varDatetime.getTzHours();
            int tzMinutes = varDatetime.getTzMinutes();
            milliOffset -= ((tzHours * 60) + tzMinutes) * 60 * 1000;
            
            
            varDatetime = datetime.plusMillis(varDatetime, milliOffset);
         }
         
         int year      = varDatetime.getYear();
         int month     = varDatetime.getMonth();
         int day       = varDatetime.getDay();
         int hours     = varDatetime.getHours();
         int mins      = varDatetime.getMinutes();
         int secs      = varDatetime.getSeconds();
         int millis    = varDatetime.getMilliseconds();
         int tzHours   = varDatetime.getTzHours();
         int tzMinutes = Math.abs(varDatetime.getTzMinutes()); // tzHours has the sign 
         
         if (!fmtFullYear)
         {
            if (!varDatetime.inY2KWindow())
            {
               throw new DisplayFormatCheckException(
                  "Value " +
                  varDatetime.toStringMessage() +
                  " cannot be displayed using " +
                  getFormatDef() +
                  ". Year must be in current century.");
            }
            
            year %= 100;
         }
         
         compBufs = initBuffers();
         boolean useAmPm = compBufs[AMPM] != null;
         if (useAmPm)
         {
            assignAMComponent(compBufs[AMPM], hours, amPmUppercase);
            if (hours >= 12)
            {
               hours -= 12;
            }
            if (hours == 0)
            {
               hours = 12;
            }
         }
         
         boolean noErr = assignComponent(compBufs[YEAR], year, '0', false);
         if (noErr)
         {
            noErr = assignComponent(compBufs[MONTH], month, '0', false);
            if (noErr)
            {
               noErr = assignComponent(compBufs[DAY], day, '0', false);
               if (noErr)
               {
                  noErr = assignComponent(compBufs[HOURS], hours, useAmPm ? ' ' : '0', false);
                  if (noErr)
                  {
                     noErr = assignComponent(compBufs[MINUTES], mins, '0', false);
                     if (noErr)
                     {
                        noErr = assignComponent(compBufs[SECONDS], secs, '0', false);
                        if (noErr)
                        {
                           noErr = assignComponent(compBufs[MILLIS], millis, '0', false);
                           if (noErr)
                           {
                              noErr = assignComponent(compBufs[TZ_HOURS], tzHours, '0', true);
                              if (noErr)
                              {
                                 noErr = assignComponent(compBufs[TZ_MINS], tzMinutes, '0', false);
                              }
                           }
                        }
                     }
                  }
               }
            }
         }
         
         if (!noErr)
         {
            ErrorManager.displayError(74,
                  "Value " + varDatetime.toStringMessage() +
                  " cannot be displayed using " + getFormatDef());
            allQMark = true;
         }
      }

      /**
       * Determine if this format must use the DEFAULT-FIXED-FONT, if no font is specified.
       * 
       * @return   Always <code>true</code>.
       */
      @Override
      public boolean isFixedFont()
      {
         return true;
      }

      /**
       * Checks format rules. Method finishEdit() should be called before it.
       * 
       * @throws  DisplayFormatCheckException
       *          When data is incompatible with format
       */
      @Override
      public void checkFormat()
      throws DisplayFormatCheckException
      {
         // just call this to check for exceptions, discard the return value
         toDatetime();
      }

      /**
       * Validate representation according to format rules and return error
       * message if there is an error.
       * 
       * @return   <code>null</code> if no error or error message if there is an
       *           error.
       */
      @Override
      public ErrorEntry[] validateFormat()
      {
         int day       = -1;
         int month     = -1;
         int hours     = -1;
         int minutes   = -1;
         int seconds   = -1;
         int millis    = -1;
         int tzHours   = -100;
         int tzMinutes = -100;
         
         // assume unedited datetime is valid
         if (compBufs == null)
            return null;
         
         try
         {
            day       = extractDateOrTimeComponent(compBufs[DAY],      false, -1);
            month     = extractDateOrTimeComponent(compBufs[MONTH],    false, -1);
            hours     = extractDateOrTimeComponent(compBufs[HOURS],    false,  0);
            minutes   = extractDateOrTimeComponent(compBufs[MINUTES],  false,  0);
            seconds   = extractDateOrTimeComponent(compBufs[SECONDS],  false,  0);
            millis    = extractDateOrTimeComponent(compBufs[MILLIS],   false,  0);
            tzHours   = extractDateOrTimeComponent(compBufs[TZ_HOURS], true,   0);
            tzMinutes = extractDateOrTimeComponent(compBufs[TZ_MINS],  false,  0);
         }
         catch (NumberFormatException e)
         {
            // error will be handled later
         }

         ErrorEntry res = verifyYearComponent(compBufs);
         if (res != null)
         {
            return new ErrorEntry[] { res };
         }

         try
         {
            String ampmSpec = (compBufs[AMPM] == null) ? null : new String(compBufs[AMPM]);
            int year = extractYearComponent(compBufs);

            res = datetimetz.validate(month, day, year, !fmtFullYear,
                                      hours, minutes, seconds, millis, 
                                      ampmSpec, tzHours, tzMinutes);

            return (res == null ? null : new ErrorEntry[] { res });
         }
         catch (ErrorConditionException ex)
         {
            // should not happen because we already verified data
         }
         
         return null;
      }
      
      /**
       * Returns if the presentation currently represents the 
       * <code>unknown value</code>.  If any editing has occurred, then this
       * will return <code>false</code> even if the editing is not complete.
       *
       * @return   <code>true</code> if this is <code>unknown</code>.
       */
      @Override
      public boolean isUnknown()
      {
         return (compBufs == null);
      }

      /**
       * Erases one symbol before the cursor position.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean deleteBack()
      {
         if (presCursorPos > 0)
         {
            presCursorPos--;
            return true;
         }
         
         return false;
      }

      /**
       * Erases one symbol in the cursor position.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean deleteCurrent()
      {
         return false;
      }

      /**
       * Inputs character. The result depends from the current insert mode.
       * 
       * @param   ch
       *          Input character
       * 
       * @return  <code>true</code> if input is accepted (nevertheless it may be silently 
       *          ignored in reality).
       */
      @Override
      public boolean input(char ch)
      {
         if (presCursorPos >= getScreenWidth())
         { 
            // ignore any input outside widget even '?'
            return false;
         }

         if (ch == '?')
         {
            compBufs = null;
            presCursorPos = 0;
            drawSeparators = false;
            return true;
         } 

         if (compBufs == null)
         { 
            // restart method from zero content
            initClear();
            boolean res = input(ch);
            if (!res)
            {
               compBufs = null;
            }
            return res;
         }

         int cursorPos = presCursorPos;
         
         // clear flag for last char
         setLastCharEntered(false);
         
         // iterate through datetime components
         for (int i = 0; i < datetimeCompOrder.length; i++)
         {
            byte compIndex = datetimeCompOrder[i];
            if (compIndex == -1)
            {
               //component not available in this format
               continue;
            }
            DatetimeComponent comp = fmtDatetimeComps[compIndex];
            int offset = comp.getOffset();

            // bypass components which we are not in
            if (cursorPos >= offset + comp.getLength())
            {
               continue;
            }

            byte digit = (byte) Character.digit(ch, 10);
            int pos = cursorPos - offset;
            
            if (digit >= 0 && compIndex != AMPM) //no digits in AMPM component
            {
               if (pos < 0)
               {
                  // skip separator
                  pos = 0;
                  cursorPos = offset;
               }

               if (compIndex == TZ_HOURS && pos == 0)
               {
                  // first char is reserved for the sign, moving to next char
                  ++cursorPos;
                  ++pos;
               }
               compBufs[compIndex][pos] = ch;

               // is this the last digit position of the last component?
               if (i == datetimeCompOrder.length - 1 && pos == compBufs[compIndex].length - 1)
               {
                  setLastCharEntered(true);
               }

               break;
            }
            
            // special case for TZ_HOURS which has on the fist position the sign
            if (compIndex == TZ_HOURS && pos == 0)
            {
               if (ch == '+' || ch == '-')
               {
                  compBufs[compIndex][pos] = ch;
                  setLastCharEntered(true);
                  break;
               }
            }

            // AM/PM indicator
            if (compIndex == AMPM && pos >= 0)
            {
               if (pos == 0)
               {
                  if (ch == 'a' || ch == 'A' || ch == 'p' || ch == 'P')
                  {
                     compBufs[compIndex][pos] = ch;
                     setLastCharEntered(true);
                     break;
                  }
               }
               else if (pos == 1)
               {
                  if (ch == 'm' || ch == 'M')
                  {
                     compBufs[compIndex][pos] = ch;
                     setLastCharEntered(true);
                     break;
                  }
               }
            }
            
            if (ch == '-' && compIndex == YEAR  && pos >= 0 && fmtSeparator != 0)
            {
               compBufs[compIndex][pos] = ch;

               // is this the last digit position of the last component?
               if (i == datetimeCompOrder.length - 1 && pos == compBufs[compIndex].length - 1)
               {
                  setLastCharEntered(true);
               }

               break;
            }

            // check date and time format separators 
            if (ch == fmtSeparator || ch == ':' || ch == '.' || ch == 't' || ch == 'T' || ch == ' ')
            {
               // check if separator it's on correct place:
               if (getFormatDef().charAt(cursorPos) == ch && pos < 0)
               {
                  // Separator char entered over existing separator. Only moving cursor.
                  break;
               }
               
               // this is different form date format, completing current component and moving to
               // next is not performed even if it is possible
               return true;
            }

            // invalid input rings the terminal bell
            OutputManager.instance().beep();
            
            return false;
         }

         // move the cursor right even if we go past the end by 1
         presCursorPos = cursorPos + 1;

         return true;
      }

      /**
       * Moves cursor one character to the left.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorLeft()
      {
         if (presCursorPos > 0)
         {
            presCursorPos--;
         }
         return true;
      }

      /**
       * Moves cursor one character to the right. Also updates insert mode.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorRight()
      {
         if (presCursorPos < getScreenWidth() - 1)
         {
            presCursorPos++;
         }
         return true;
      }

      /**
       * Moves cursor to the beginning of the field.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorBegin()
      {
         presCursorPos = 0;
         return true;
      }

      /**
       * Moves cursor to the end of the field.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorEnd()
      {
         presCursorPos = getScreenWidth() - 1;
         return true;
      }

      /**
       * Indicates completion of an edit. The presentation instance may fill 
       * some unfilled places.
       * 
       * @return   <code>true</code> if rendering is necessary.
       */
      @Override
      public boolean finishEdit()
      {
         if (compBufs == null)
         {
            return false;
         }

         char[][] newBufs = initBuffers();
         boolean changed = false;
         
         for (int i = 0; i < newBufs.length; i++)
         {
            if (newBufs[i] == null)
            {
               // some components may not be of interest so they are not initialized
               continue;
            }
            completeComponent(newBufs, compBufs, i);
            if (!Arrays.equals(newBufs[i], compBufs[i]))
            {
               changed = true;
            }
         }

         if (changed)
         {
            compBufs = newBufs;
         }

         return changed;
      }

      /**
       * Formats the data as a string that is ready for display.
       * 
       * @param    edit
       *           <code>true</code> if the screen value is being generated
       *           for interactive editing purposes.
       * 
       * @return   The formatted string.
       */
      @Override
      public String toScreenValue(boolean edit)
      {
         boolean tempBuf = (edit && (compBufs == null) && drawSeparators);

         if (tempBuf)
         {
            // create temporary buffers
            initClear();
         }
         
         String result = toScreenValue();
         
         if (tempBuf)
         {
            // remove temporary buffers
            compBufs = null;
         }
         
         return result;
      }

      /**
       * Formats the data as a string that is ready for display.
       * 
       * @return   The formatted string.
       */
      @Override
      public String toScreenValue()
      {
         if (compBufs == null)
         {
            return "";
         }
         else if (allQMark)
         {
            // return a string of n question marks '?', n is the length of the format 
            return new String(new char[getFormatDef().length()]).replace('\0', '?');
         }
         
         // buffer for format pattern
         char[] buf = getFormatDef().toCharArray();
         
         //iterate components in reverse order
         for (int i = fmtDatetimeComps.length - 1; i >= 0 ;i--)
         {
            if (fmtDatetimeComps[i] == null || compBufs[i] == null)
            {
               // this component does not appear in current format
               continue;
            }
            
            // replace in pattern the format specifiers with actual values from compBufs
            for (int k = 0; k < compBufs[i].length && k < fmtDatetimeComps[i].getLength(); k++)
            {
               if (buf.length <= fmtDatetimeComps[i].getOffset() + k)
               {
                  // double check buffer overflow
                  continue;
               }
               buf[fmtDatetimeComps[i].getOffset() + k] = compBufs[i][k];
            }
            if (drawSeparators && fmtDatetimeComps[i].getOffset() > 1)
            {
               char sep = 0;
               switch (i)
               {
                  case TZ_MINS:
                  case MINUTES:
                     sep = ':';
                     break;
                  case MILLIS:
                     sep = NumberType.getDecimalSeparator(); // default is '.'
                     break;
                  case YEAR:
                  case MONTH:
                  case DAY:
                     if (dateCompOrder[0] != i)
                     {
                        // if not on first position
                        sep = fmtSeparator;
                     }
                     break;
               }
               if (sep != 0)
               {
                  buf[fmtDatetimeComps[i].getOffset() - 1] = sep;
               }
            }
         }
         
         return new String(buf);
      }
      
      /**
       * Instantiates a variable that is equivalent to the presentation's
       * internal representation.
       * 
       * @param    var
       *           A variable which must be compatible with the format.
       */
      @Override
      public void toVar(BaseDataType var)
      {
         try
         {
            ((datetime) var).assign(toDatetime(), true);
         }
         catch (DisplayFormatCheckException ex)
         {
            // buffers should be already validated
         }
      }

      /**
       * Allocates datetime components buffers.
       * 
       * @return  Components buffers array.
       */
      private char[][] initBuffers()
      {
         char[][] bufs = new char[10][];
         for (int k = 0; k < 10; k++)
         {
            if (fmtDatetimeComps[k] != null)
            {
               bufs[k] = new char[fmtDatetimeComps[k].getLength()];
            }
         }
         return bufs;
      }

      /**
       * Resets content to space value.
       */
      private void initClear()
      {
         compBufs = initBuffers();
         for (int i = 0; i < compBufs.length; i++)
         {
            char[] compBuf = compBufs[i];
            if (compBuf != null)
            {
               clearComponent(compBuf, i);
            }
         }
      }

      /**
       * Fills the date component with <code>SPACE</code>.
       * 
       * @param   comp
       *          The component.
       * @param   compIdx
       *          The id of the component.
       */
      private void clearComponent(char[] comp, int compIdx)
      {
         for (int i = 0; i < comp.length; ++i)
         {
            comp[i] = SPACE;
         }
         
         char[] fmtAsChars = getFormatDef().toCharArray();
         if (compIdx == AMPM && fmtDatetimeComps[AMPM] != null)
         {
            comp[0] = fmtAsChars[fmtDatetimeComps[AMPM].compOffset];
            if (comp.length > 1 && fmtDatetimeComps[AMPM].compLength > 1)
            {
               comp[0] = fmtAsChars[fmtDatetimeComps[AMPM].compOffset + 1];
            }
         }
         else if (compIdx == TZ_HOURS                 && 
                  fmtDatetimeComps[TZ_HOURS] != null  &&
                  fmtAsChars.length > fmtDatetimeComps[TZ_HOURS].compOffset)
         {
            comp[0] = fmtAsChars[fmtDatetimeComps[TZ_HOURS].compOffset];
         }
      }

      /**
       * Fills the date/time component with the given value.
       * 
       * @param   comp
       *          The component.
       * @param   value
       *          The new value.
       * @param   leftPad
       *          Character to be used for left padding.
       * @param   sign
       *          The first character is reserved for the sign (only used for hours TZ).
       *
       * @return  <code>true</code> if the value could be assigned or if component is optional and
       *          <code>false</code> if value cannot be represented in the available space
       */
      private boolean assignComponent(char[] comp, int value, char leftPad, boolean sign)
      {
         if (comp == null)
         {
            // this component is not part of the format
            return true;
         }
         
         // the sign digit is reserved (0)
         String txt = Integer.toString(sign ? Math.abs(value) : value);
         if (txt.length() > comp.length)
         {
            return false;
         }
         
         int idx = txt.length() - 1;
         for (int i = comp.length - 1; i >= 0; i--)
         {
            if (idx >= 0)
            {
               comp[i] = txt.charAt(idx);
               idx--;
            }
            else
            {
               // left pad
               comp[i] = leftPad;
            }
         }
         if (sign)
         {
            comp[0] = value >= 0 ? '+' : '-';
         }
         return true;
      }

      /**
       * Fills the AMPM component with the computed value from current hours.
       * 
       * @param   comp
       *          The AMPM component.
       * @param   value
       *          The hours to be processed in 24H format.
       */
      private void assignAMComponent(char[] comp, int value, boolean upper)
      {
         if (upper)
         {
            comp[0] = value < 12 ? 'A' : 'P';
         }
         else 
         {
            comp[0] = value < 12 ? 'a' : 'p';
         }
         if (comp.length >= 2)
         {
            comp[1] = upper ? 'M' : 'm';
         }
      }

      /**
       * Calculates numeric value of the date component.
       * 
       * @param   comp
       *          The date or time component. May partially consist from spaces.
       * @param   includeSign
       *          The component is sign prefixed (for TZ_HOURS only).
       * @param   defVal
       *          Default value to be returned if field not found or could not be parsed 
       *
       * @return  The integer value of component of <code>defVal</code> on error.
       */
      private int extractDateOrTimeComponent(char[] comp, boolean includeSign, int defVal)
      {
         // trivial case first
         if (comp == null)
         {
            return defVal;
         }
         
         int begin = 0;

         while (begin < comp.length)
         {
            if (comp[begin] != SPACE)
            {
               break;
            }
            begin++;
         }

         if (begin == comp.length)
         {
            return defVal;
         }

         int end = begin + 1;
         while (end < comp.length)
         {
            if (comp[end] == SPACE)
            {
               break;
            }
            end++;
         }

         boolean negative = includeSign && comp[begin] == '-';
         if (includeSign && begin + 1 < end)
         {
            //skip the sign
            ++begin;
         }
         String txt = new String(comp, begin, end - begin);
         
         int val = Integer.parseInt(txt);
         return negative ? -val : val;
      }

      /**
       * Calculates numeric value of the year. Calculation can be described by
       * the following algorithm:
       * <p>
       * If format separator isn't <code>null</code> then:
       * <p>
       * <ol>
       *    <li> Input sequence is autocompleted (to a temporary buffer).
       *    <li> If the extreme right position holds '-' then it is discarded
       *         (e.g. -10- -&gt; -10)
       *    <li> Remaining characters are parsed as a usual integer.
       * </ol>
       * <p>
       * If format separator is <code>null</code> then:
       * <p>
       * <ol>
       *    <li> Only the extreme left (before the space if it presents)
       *         digits are taken into account. If the first character is a
       *         space then the whole field content is ignored (e.g. 12_3 -&gt;
       *         12; _123 -&gt; <code>null</code>)
       *    <li> If input consists of three digits then the last one is
       *         ignored (e.g. 123_ -&gt; 12).
       *    <li> Remaining characters are parsed as a usual integer. If it
       *         cannot be parsed then the current year is used.
       * </ol>
       *
       * @param   comp
       *          The input datetime buffer.
       *
       * @return  The numeric value of the year.
       *
       * @throws  ErrorConditionException
       *          If the input sequence does not represent a valid year.
       */
      private int extractYearComponent(char[][] comp)
      throws ErrorConditionException
      {
         char[][] tempBuf = new char[3][];
         tempBuf[YEAR] = Arrays.copyOf(comp[YEAR], comp[YEAR].length);
         for (char c : comp[YEAR])
         {
            if (c == SPACE)
            {
               // process autocompletion for a temporary buffer if it is required
               completeComponent(tempBuf, comp, YEAR);
               break;
            }
         }

         char[] yearBuffer = tempBuf[YEAR];

         int year = 0;
         if (fmtSeparator != 0)
         {
            int end = yearBuffer.length - 1;

            // discard extreme right '-' if present
            if (yearBuffer[end] == '-')
            {
               end--;
            }

            String txt = new String(yearBuffer);
            try
            {
               year = Integer.parseInt(txt.substring(0, end + 1));
            }
            catch (NumberFormatException ex)
            {
               if (fmtFullYear)
               {
                  ErrorManager.recordOrThrowError(85, "Invalid datetime input");
                  return (int)datetime.INVALID_DATE;
               }
            }
         }
         else // fmtSeparator == 0
         {
            // Only extreme left (before the space if it presents) digits
            // are taken into account. If the first character is a space
            // then the whole field content is ignored.
            int end = yearBuffer.length - 1;
            for (int i = 0; i < yearBuffer.length; i++)
            {
               if (yearBuffer[i] == SPACE)
               {
                  end = i - 1;
                  break;
               }
            }

            String txt = new String(yearBuffer);
            txt = txt.substring(0, end + 1);

            // if input consists of three digits then the last one is ignored
            if (txt.length() == 3)
            {
               txt = txt.substring(0, 2);
            }

            try
            {
               year = Integer.parseInt(txt);
            }
            catch (NumberFormatException ex)
            {
               year = date.today().getYear();
            }
         }

         if (year == 0 && fmtFullYear)
         {
            ErrorManager.recordOrThrowError(79, "Year is out of range of 0");
            return (int) datetime.INVALID_DATE;
         }

         return year;
      }

      /**
       * Validate year without throwing exceptions.
       * 
       * @param    comp
       *           Component to verify.
       * 
       * @return   <code>null</code> if no error found, {@link ErrorEntry}
       *           instance otherwise.
       */
      private ErrorEntry verifyYearComponent(char[][] comp)
      {
         char[][] tempBuf = new char[3][];
         tempBuf[YEAR] = Arrays.copyOf(comp[YEAR], comp[YEAR].length);
         for (char c : comp[YEAR])
         {
            if (c == SPACE)
            {
               // process autocompletion for a temporary buffer if it is required
               completeComponent(tempBuf, comp, YEAR);
               break;
            }
         }

         char[] yearBuffer = tempBuf[YEAR];

         int year = 0;
         if (fmtSeparator != 0)
         {
            int end = yearBuffer.length - 1;

            // discard extreme right '-' if present
            if (yearBuffer[end] == '-')
            {
               end--;
            }

            String txt = new String(yearBuffer);
            try
            {
               year = Integer.parseInt(txt.substring(0, end + 1));
            }
            catch (NumberFormatException ex)
            {
               if (fmtFullYear)
               {
                  return new ErrorEntry(85, "Invalid datetime input");
               }
            }
         }
         else // fmtSeparator == 0
         {
            // Only extreme left (before the space if it presents) digits
            // are taken into account. If the first character is a space
            // then the whole field content is ignored.
            int end = yearBuffer.length - 1;
            for (int i = 0; i < yearBuffer.length; i++)
            {
               if (yearBuffer[i] == SPACE)
               {
                  end = i - 1;
                  break;
               }
            }

            String txt = new String(yearBuffer);
            txt = txt.substring(0, end + 1);

            // if input consists of three digits then the last one is ignored
            if (txt.length() == 3)
            {
               txt = txt.substring(0, 2);
            }

            try
            {
               year = Integer.parseInt(txt);
            }
            catch(NumberFormatException ex)
            {
               year = date.today().getYear();
            }
         }

         if (year == 0 && fmtFullYear)
         {
            return new ErrorEntry(79, "Year is out of range of 0");
         }

         return null;
      }
      
      /**
       * Completes datetime component. Day, month, hour, minutes, seconds and milliseconds 
       * are completed with 0's.
       * <p>
       * This method uses only one old and one new buffer for component but takes all buffers to
       * simplify coding of its call.
       * <p>
       * The year completion algorithm can be described as follows:
       * <p>
       * <ol>
       *    <li> If format separator is <code>null</code> then do nothing.
       *    <li> Find the extreme right '-' and fill all empty place-holders to the left of it 
       *          with the '-' (e.g. __-_ -&gt; ---_; _1_- -&gt; -1--).
       *    <li> The extreme leftmost meaningful symbol (not SPACE) is found. After that we search
       *          for the extreme right meaningful symbol such that no spaces are present between
       *          it and the extreme left meaningful symbol. All symbols which are not between
       *          these ones are discarded (e.g. __12 -&gt; 12; -2_3 -&gt; -2; _1_3 -&gt; 1). If
       *          no meaningful symbols are present then the current year is used.
       *    <li> To the previously found sequence of meaningful symbols some leading symbol may be
       *          added depending on the sequence length and year format:
       *          <ul>
       *             <li> When length is 3 and format is full-year OR length is 1 and format is
       *                   not full-year then add a leading 0.
       *             <li> When length is 2 or 1 and format is full-year then the
       *                   <code>date.windowingImpl()</code> which takes as input the digits 
       *                   from the sequence (the one that was created on the third step)
       *                   is used (if the sequence contains no digits then 0 is used).
       *             <li> From the return value we take first 2 or 3 characters and complete the
       *                   sequence with it. (e.g. -2- -&gt; 0-2-; 99 -&gt; 1999; -- -&gt; 20--;
       *                   9 -&gt; 2009)
       *          </ul>
       * </ol>
       *
       * @param   newCompBufs
       *          Buffers for new value. May be the same with <code>oldCompBufs</code>.
       * @param   oldCompBufs
       *          Buffers for old value.
       * @param   compIndex
       *          Component's index (DAY, MONTH, YEAR, HOURS, MINUTES, SECONDS, MILLIS, AMPM,
       *          TZ_HOURS, TZ_MINS).
       */
      private void completeComponent(char[][] newCompBufs, char[][] oldCompBufs, int compIndex)
      {
         if (compIndex == AMPM && oldCompBufs[AMPM] != null)
         {
            // this is the only non-digit field, copy directly the content of the array
            System.arraycopy(oldCompBufs[AMPM], 0, newCompBufs[AMPM], 0, oldCompBufs[AMPM].length);
            return;
         }
         if (compIndex != YEAR)
         {
            int comp = extractDateOrTimeComponent(oldCompBufs[compIndex], compIndex == TZ_HOURS, 0);
            assignComponent(newCompBufs[compIndex], comp, '0', compIndex == TZ_HOURS);
            return;
         }

         // compIndex == YEAR
         if (fmtSeparator == 0)
         {
            newCompBufs[compIndex] = Arrays.copyOf(oldCompBufs[compIndex],
                                                   oldCompBufs[compIndex].length);
            return;
         }

         char[] yearBuffer = Arrays.copyOf(oldCompBufs[compIndex],
                                           oldCompBufs[compIndex].length);

         // find extreme right '-'
         int rightMinusPos = -1;
         for (int i = yearBuffer.length - 1; i >= 0; i--)
         {
            if (yearBuffer[i] == '-')
            {
               rightMinusPos = i;
               break;
            }
         }

         if (rightMinusPos != -1)
         {
            // fill all empty place-holders to the left of the extreme right '-' with '-'
            for (int i = rightMinusPos - 1; i >= 0; i--)
            {
               if (yearBuffer[i] == SPACE)
               {
                  yearBuffer[i] = '-';
               }
            }
         }

         // find extreme left meaningful symbol (not SPACE)
         int leftSymbolPos = -1;
         for (int i = 0; i < yearBuffer.length; i++)
         {
            if (yearBuffer[i] != SPACE)
            {
               leftSymbolPos = i;
               break;
            }
         }

         if (leftSymbolPos == -1)
         {
            // empty year - fill with the current one
            assignComponent(newCompBufs[compIndex], date.today().getYear(), '0', false);
            return;
         }

         // find extreme right meaningful symbol between which and the
         // extreme left meaningful symbol no spaces present
         int rightSymbolPos = leftSymbolPos;
         for (int i = leftSymbolPos + 1; i < yearBuffer.length; i++)
         {
            if (yearBuffer[i] != SPACE)
            {
               rightSymbolPos = i;
            }
            else
            {
               break;
            }
         }

         // length of the meaningful symbols sequence
         // all other symbols which are not present in this sequence
         // will not be taken into account any further
         int symbolsLength = rightSymbolPos - leftSymbolPos + 1;

         // move meaningful symbols to the right
         int writePos = yearBuffer.length - 1;
         for (int readPos = rightSymbolPos;
              readPos >= leftSymbolPos;
              readPos--)
         {
            yearBuffer[writePos--] = yearBuffer[readPos];
         }

         if (fmtFullYear)
         {
            // add leading year numbers
            if (symbolsLength == 3)
            {
               yearBuffer[0] = '0';
            }
            else if (symbolsLength == 2 || symbolsLength == 1)
            {
               // add two leading year numbers
               StringBuilder lastYearDigits = new StringBuilder("");
               if (symbolsLength == 2 && Character.isDigit(yearBuffer[2]))
               {
                  lastYearDigits.append(yearBuffer[2]);
               }
               if (Character.isDigit(yearBuffer[3]))
               {
                  lastYearDigits.append(yearBuffer[3]);
               }

               int lastYearDigitsInt = 0;
               if (lastYearDigits.length() > 0)
               {
                  lastYearDigitsInt = Integer.parseInt(lastYearDigits.toString());
               }

               String fullYear = Integer.toString(date.windowingImpl(lastYearDigitsInt));

               if (fullYear.length() == 4)
               {
                  for (int i = 0; i < yearBuffer.length - symbolsLength; i++)
                  {
                     yearBuffer[i] = fullYear.charAt(i);
                  }
               }
               else
               {
                  // just in case
                  for (int i = 0; i < yearBuffer.length - symbolsLength; i++)
                  {
                     yearBuffer[i] = '0';
                  }
               }
            }
         }
         else // not fmtFullYear
         {
            if (symbolsLength == 1)
            {
               yearBuffer[0] = '0';
            }
         }

         newCompBufs[compIndex] = Arrays.copyOf(yearBuffer, yearBuffer.length);
      }

      /**
       * Returns the current presentation's value converted to <code>datetime</code>.
       * 
       * @return  The current presentation's value converted to <code>datetime</code>.
       * 
       * @throws  DisplayFormatCheckException
       *          Value doesn't represent a correct datetime.
       */
      private datetime toDatetime()
      throws DisplayFormatCheckException
      {
         if (compBufs == null)
         {
            return new datetime();
         }

         // just extracting components. Spaces should be already eliminated
         try
         {
            int year = extractYearComponent(compBufs);
            if (!fmtFullYear)
            {
               year = date.windowingImpl(year);
            }
            int month     = extractDateOrTimeComponent(compBufs[MONTH],    false, -1);
            int day       = extractDateOrTimeComponent(compBufs[DAY],      false, -1);
            int hours     = extractDateOrTimeComponent(compBufs[HOURS],    false, 0);
            int mins      = extractDateOrTimeComponent(compBufs[MINUTES],  false, 0);
            int secs      = extractDateOrTimeComponent(compBufs[SECONDS],  false, 0);
            int millis    = extractDateOrTimeComponent(compBufs[MILLIS],   false, 0);
            int tzHours   = extractDateOrTimeComponent(compBufs[TZ_HOURS], true,  0);
            int tzMinutes = extractDateOrTimeComponent(compBufs[TZ_MINS],  false, 0);
            int tzOffset  = tzHours * 60 + (tzHours < 0 ? -tzMinutes : tzMinutes);
            
            // check AM/PM indicator
            if (compBufs[AMPM] != null && compBufs[AMPM].length > 0)
            {
               if (hours == 12)
               {
                  hours = 0;
               }
               if (compBufs[AMPM][0] == 'P' || compBufs[AMPM][0] == 'p')
               {
                  hours += 12;
               }
            }
            
            return new datetimetz(month, day, year, hours, mins, secs, millis, tzOffset);
         }
         catch (ErrorConditionException ex)
         {
            throw new DisplayFormatCheckException(ex.getMessage());
         }
      }
   }
}