/*
** Module   : StringFormat.java
** Abstract : formatting/editing helper for string data
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SVG 20050609   @21446 Initial version.
** 002 SVG 20050620   @21523 DisplayFormat related classes are redesigned.
**                           Now cursor's position and insert mode are
**                           maintained inside Presentation. This allows
**                           to not pass them by reference into
**                           Presentation edit methods. All
**                           com.goldencode.p2j.ui.chui.dataref package
**                           is obsoleted and removed.
** 003 SIY 20050831   @22551 Minor formatting fixes. 
** 004 SIY 20050907   @22652 Fixed use of @inheritDoc tag in javadoc
** 005 GES 20060508   @26018 Moved error processing into this class from
**                           FillIn to more closely duplicate the error
**                           processing of Progress.
** 006 GES 20060920   @29731 Added support for isUnknown().
** 007 NVS 20070416   @33033 checkFormat method should allow empty string
**                           and blank string to pass.
** 008 EVL 20080610   @38683 The changes were made for the CharBuffer
**                           constructor to take into account the format
**                           of the characters initially passed upon object
**                           creation.
** 009 SVL 20080914   @39799 Insert mode reset removed from the
**                           activate().
** 010 SIY 20090701   @43019 Added CharBuf.validateFormat() implementation.
** 011 SIY 20090709   @43115 Fixed validateFormat(), it must use buffer without
**                           filling characters copied from format string.
** 012 SIY 20090721   @43289 Moved activate() into base class.
** 013 SIY 20100804          Changes in imports, some refactorings.
** 014 CA  20140805          ThinClient is explicitly imported, to isolate ChUI class usage 
**                           outside of ChUI packages.
** 015 CA  20141220          Changes related to GUI FILL-IN implementation.
** 016 CA  20150128          Added API to check if a format needs special validation.
** 017 SVL 20150608          Added zap clear mode.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.format;

import java.util.*;

import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager.ErrorEntry;

/**
 * Format specialization for <code>character</code> variables.
 */
public class StringFormat
extends DisplayFormat
{
   /** The state constant for the Progress format string parser. */
   private static final int PS_FILL = 1;

   /** The state constant for the Progress format string parser. */
   private static final int PS_FILL_ESCAPE = 2;

   /** The state constant for the Progress format string parser. */
   private static final int PS_FORMAT_CHAR = 3;

   /** The state constant for the Progress format string parser. */
   private static final int PS_FORMAT_LPAREN = 4;

   /** The state constant for the Progress format string parser. */
   private static final int PS_FORMAT_DUPES = 5;

   /** Info about characters in screen value. */
   protected final CharInfo[] fmtChars;

   /** 
    * Maximum number of value chars. I.e. chars in character variable which
    * is displayed.
    */
   protected int maxPresLength;

   /**
    * Flag indicating the drawing needs to use the DEFAULT-FIXED-FONT, if no font is specified.
    * <p>
    * This can happen only if the widget's format is fully numeric.
    */
   private final boolean fixedFont;
   
   /** Flag to check if this format needs special validation. */
   private boolean formatCheck = false;
   
   /**
    * Constructs character format from the Progress format string.
    * 
    * @param   pformat
    *          Progress format string.
    * 
    * @throws  DisplayFormatParsingException
    *          Format string is invalid.
    */
   public StringFormat(String pformat)
   throws DisplayFormatParsingException
   {
      super(pformat);

      ArrayList<CharInfo> chars = new ArrayList<CharInfo>(50);

      int pflen = pformat.length();
      if (pflen == 0)
      {
         throw new DisplayFormatParsingException("Format is empty");
      }

      maxPresLength = 0;
      int formatDupes = 0; // to make compiler happy

      int pstate = PS_FILL;

      int pfi = 0;
      while (pfi < pflen)
      {
         char pfc = pformat.charAt(pfi++);
         switch (pstate)
         {
            case PS_FILL:
               if (pfc == '~')
               {
                  pstate = PS_FILL_ESCAPE;
               }
               else
               {
                  ValueCharInfo info = createValueCharInfo(pfc, maxPresLength);
                  if (info != null)
                  {
                     chars.add(info);
                     ++maxPresLength;
                     pstate = PS_FORMAT_CHAR;
                  }
                  else
                  {
                     chars.add(new FillCharInfo(pfc));
                  }
               }
               break;
            case PS_FILL_ESCAPE:
               chars.add(new FillCharInfo(pfc));
               pstate = PS_FILL;
               break;
            case PS_FORMAT_CHAR:
               if (pfc == '~')
               {
                  pstate = PS_FILL_ESCAPE;
               }
               else
               {
                  ValueCharInfo info = createValueCharInfo(pfc, maxPresLength);
                  if (info != null)
                  {
                     chars.add(info);
                     ++maxPresLength;
                  }
                  else if (pfc == '(')
                  {
                     pstate = PS_FORMAT_LPAREN;
                  }
                  else
                  {
                     chars.add(new FillCharInfo(pfc));
                     pstate = PS_FILL;
                  }
               }
               break;
            case PS_FORMAT_LPAREN:
               formatDupes = Character.digit(pfc, 10);
               if (formatDupes < 0)
               {
                  throw new DisplayFormatParsingException
                     ("Expected digit in position " + (pfi - 1) +
                      "of format " + pformat);
               }
               pstate = PS_FORMAT_DUPES;
               break;
            case PS_FORMAT_DUPES:
               if (pfc == ')')
               {
                  if (formatDupes == 0)
                  {
                     throw new DisplayFormatParsingException
                        ("Expected not zero in position " + (pfi - 1) +
                         "of format " + pformat);
                  }

                  ValueCharInfo info = 
                     (ValueCharInfo)chars.get(chars.size() - 1);
                  while (--formatDupes > 0)
                  {
                     chars.add(info.dup(maxPresLength++));
                  }
                  pstate = PS_FILL;
               }
               else
               {
                  int digit = Character.digit(pfc, 10);
                  if (digit < 0)
                  {
                     throw new DisplayFormatParsingException
                        ("Expected digit in position " + (pfi - 1) +
                         "of format " + pformat);
                  }
                  formatDupes = formatDupes * 10 + digit;
               }
               break;
         }
      }

      // format string parsed. Checking final state
      if (pstate != PS_FILL && pstate != PS_FORMAT_CHAR)
      {
         throw new DisplayFormatParsingException
            ("Format " + pformat + " incomplete");
      }

      fmtChars = new CharInfo[chars.size()];
      chars.toArray(fmtChars);
      
      boolean fixedFont = true;
      for (CharInfo ch : fmtChars)
      {
         fixedFont = fixedFont && (ch instanceof DigitCharInfo || ch instanceof FillCharInfo);
         if (!fixedFont)
         {
            break;
         }
      }
      
      this.fixedFont = fixedFont;
      this.fmtScreenWidth = fmtChars.length;
   }
   
   /**
    * Verify if this format needs explicit validation.
    * 
    * @return   The state of the {@link #formatCheck} flag.
    */
   @Override
   public boolean isFormatCheck()
   {
      return formatCheck;
   }

   /**
    * Returns a new presentation instance with a default initialized insert
    * mode and cursor position.
    * 
    * @param   pres
    *          A Presentation instance which was constructed by this format.
    *          May be <code>null</code>. This presentation instance is not 
    *          changed in this method.
    * 
    * @return  New presentation instance.
    */
   @Override
   public Presentation clear(Presentation pres)
   {
      return clear(pres, false);
   }

   /**
    * Returns a new presentation instance with a default initialized insert
    * mode and cursor position.
    *
    * @param   pres
    *          A Presentation instance which was constructed by this format.
    *          May be <code>null</code>. This presentation instance is not
    *          changed in this method.
    * @param   zapClear
    *          If <code>true</code> then presentation is cleared completely,
    *          like if zap mode is ON. If <code>false</code> then it is
    *          cleared like if CLEAR event was applied (that means it is
    *          cleared from the current cursor position).
    *
    * @return  New presentation instance.
    */
   @Override
   public Presentation clear(Presentation pres, boolean zapClear)
   {
      return (pres != null) ? ((CharBuf)pres).clear(zapClear) :
                              new CharBuf();
   }

   /**
    * Initializes a presentation for variable's value.
    * 
    * @param   var
    *          Variable which should be compatible with format.
    * 
    * @return  Data presentation
    * 
    * @throws  DisplayFormatCheckException
    *          Variable's value is incompatible with format.
    */
   @Override
   public Presentation fromVar(BaseDataType var)
   throws DisplayFormatCheckException
   {
      return new CharBuf((character)var);
   }

   /**
    * Removes final spaces from string buffer.
    * 
    * @param   sb
    *          Buffer to scan.
    */
   protected static void rightTrimStringBuf(StringBuffer sb)
   {
      int i;
      for (i = sb.length() - 1; i >= 0; --i)
      {
         if (sb.charAt(i) != ' ')
         {
            break;
         }
      }
      sb.setLength(i + 1);
   }

   /**
    * Creates <code>ValueCharInfo</code> from format char.
    * 
    * @param   ch
    *          Format char.
    * @param   valueIndex
    *          Index in value.
    * 
    * @return  <code>ValueCharInfo</code> or <code>null</code> if provided
    *          char is not for value formatting.
    */
   private ValueCharInfo createValueCharInfo(char ch, int valueIndex)
   {
      ValueCharInfo info;
      switch (ch)
      {
         case 'X':
         case 'x':
            info = new AnyCharInfo(valueIndex);
            break;
         case 'N':
         case 'n':
            info = new AlphanumCharInfo(valueIndex);
            formatCheck = true;
            break;
         case 'A':
         case 'a':
            info = new LetterCharInfo(valueIndex);
            formatCheck = true;
            break;
         case '!':
            info = new UpperLetterCharInfo(valueIndex);
            formatCheck = true;
            break;
         case '9':
            info = new DigitCharInfo(valueIndex);
            formatCheck = true;
            break;
         default:
            info = null;
      }
      return info;
   }

   /**
    * Base for format chars info. It is designed to read all info without
    * casting to more concrete char classes.
    */
   private static abstract class CharInfo
   {
      /**
       * Default constructor.
       */
      public CharInfo() {}
      
      /**
       * Returns index in the screen value corresponding to the given char.
       * 
       * @return  Index in the screen value corresponding to the given char.
       *          Equals to (-1) for fill chars
       */
      public abstract int getValueIndex();

      /**
       * Returns fill char.
       * 
       * @return  Fill char or (char)0 for replacement to empty.
       */
      public abstract char getFillChar();

      /**
       * Validates character against format rules.
       * 
       * @param   ch
       *          Character to check.
       * 
       * @return  Error message or <code>null</code> for no error.
       */
      public abstract String validateChar(char ch);

      /**
       * Get the number of the error shown if this char is invalid.
       * 
       * @return  See above.
       */
      public abstract int getErrorNum();

      /**
       * Converts input char to screen value according to format.
       * 
       * @param   ch
       *          Input character.
       * 
       * @return  Character which may be placed into screen value.
       */
      public char makeValueChar(char ch)
      {
         return ch;
      }
   }

   /**
    * Base class for info about value char. I.e. char which is taken from the
    * bound variable.
    */
   private static abstract class ValueCharInfo
   extends CharInfo
   implements Cloneable
   {
      /** Index in the screen value of this value char. */
      private int valueIndex;

      /**
       * Initializes the index of this value char.
       * 
       * @param   index
       *          Index in the screen value.
       */
      public ValueCharInfo(int index)
      {
         valueIndex = index;
      }

      /**
       * Returns index in the screen value corresponding to the given char.
       * 
       * @return  Index in the screen value corresponding to the given char.
       *          Equals to (-1) for fill chars
       */
      @Override
      public int getValueIndex()
      {
         return valueIndex;
      }

      /**
       * Return fill character.
       * 
       * @return  (char)0 indicating that this is not fill char.
       */
      @Override
      public char getFillChar()
      {
         return 0;
      }

      /**
       * Makes a clone with different value index
       * 
       * @param   index
       *          The new value index.
       * 
       * @return  Instance of the same class with the given index.
       */
      public ValueCharInfo dup(int index)
      {
         ValueCharInfo res = null;
         try
         {
            res = (ValueCharInfo)clone();
            res.valueIndex = index;
         }
         catch (CloneNotSupportedException ex)
         {
            // impossible
         }

         return res;
      }
   }

   /**
    * Info about value char which may be any char.
    */
   private static class AnyCharInfo
   extends ValueCharInfo
   {
      /**
       * Initializes the index of this value char.
       * 
       * @param   index
       *          Index in the screen value.
       */
      public AnyCharInfo(int index)
      {
         super(index);
      }

      /**
       * Validates character against format rules
       * 
       * @param   ch
       *          Character to check.
       * 
       * @return  Always <code>null</code>.
       */
      @Override
      public String validateChar(char ch)
      {
         return null; // any char is valid
      }
      
      /**
       * Get the number of the error shown if this char is invalid.
       * 
       * @return  Always <code>0</code>.
       */
      @Override
      public int getErrorNum()
      {
         // no-error
         return 0;
      }
   }

   /**
    * Info about value char which should be digit or letter.
    */
   private static class AlphanumCharInfo
   extends ValueCharInfo
   {
      /**
       * Initializes the index of this value char.
       * 
       * @param   index
       *          Index in the screen value.
       */
      public AlphanumCharInfo(int index)
      {
         super(index);
      }

      /**
       * Validates character against format rules.
       * 
       * @param   ch
       *          Character to check.
       * 
       * @return  Error message or <code>null</code> for no error.
       */
      @Override
      public String validateChar(char ch)
      {
         return (Character.isLetterOrDigit(ch)) ? null :
                                                  "digit or letter";
      }

      /**
       * Get the number of the error shown if this char is invalid.
       * 
       * @return  Always <code>632</code>.
       */
      @Override
      public int getErrorNum()
      {
         // no-error
         return 632;
      }
   }

   /**
    * Info about value char which should be letter.
    * 
    */
   private static class LetterCharInfo
   extends ValueCharInfo
   {
      /**
       * Initializes the index of this value char.
       * 
       * @param   index
       *          Index in the screen value.
       */
      public LetterCharInfo(int index)
      {
         super(index);
      }

      /**
       * Validates character against format rules.
       * 
       * @param   ch
       *          Character to check.
       * 
       * @return  Error message or <code>null</code> for no error.
       */
      @Override
      public String validateChar(char ch)
      {
         return (Character.isLetter(ch)) ? null : "letter";
      }

      /**
       * Get the number of the error shown if this char is invalid.
       * 
       * @return  Always <code>631</code>.
       */
      @Override
      public int getErrorNum()
      {
         return 631;
      }
   }

   /**
    * Info about value char which should be letter. Additionally includes
    * the uppercase converter.
    */
   private static class UpperLetterCharInfo
   extends LetterCharInfo
   {
      /**
       * Initializes the index of this value char.
       * 
       * @param   index
       *          Index in the screen value.
       */
      public UpperLetterCharInfo(int index)
      {
         super(index);
      }

      /**
       * Converts input char to screen value according to format.
       * 
       * @param   ch
       *          Input character.
       * 
       * @return  Input character converted to upper case.
       */
      @Override
      public char makeValueChar(char ch)
      {
         return Character.toUpperCase(ch);
      }

      /**
       * Get the number of the error shown if this char is invalid.
       * 
       * @return  Always <code>631</code>.
       */
      @Override
      public int getErrorNum()
      {
         return 631;
      }
   }

   /**
    * Info about value char which should be digit.
    */
   private static class DigitCharInfo
   extends ValueCharInfo
   {
      /**
       * Initializes the index of this value char.
       * 
       * @param   index
       *          Index in the screen value.
       */
      public DigitCharInfo(int index)
      {
         super(index);
      }

      /**
       * Validates character against format rules.
       * 
       * @param   ch
       *          Character to check.
       * 
       * @return  Error message or <code>null</code> for no error.
       */
      @Override
      public String validateChar(char ch)
      {
         return (Character.isDigit(ch)) ? null : "digit";
      }

      /**
       * Get the number of the error shown if this char is invalid.
       * 
       * @return  Always <code>630</code>.
       */
      @Override
      public int getErrorNum()
      {
         return 630;
      }
   }

   /**
    * Info about fill char.
    */
   private static class FillCharInfo
   extends CharInfo
   {
      /** Fill char. May be (char)0 if it is the empty fill char. */
      private char fillChar;

      /**
       * Initializes the instance with the given fill char.
       * 
       * @param   ch
       *          Fill char. May be (char)0 if it is the empty fill char.
       */
      public FillCharInfo(char ch)
      {
         fillChar = ch;
      }

      /**
       * Returns (-1) because it is fill char.
       * 
       * @return  (-1) because it is fill char.
       */
      @Override
      public int getValueIndex()
      {
         return -1;
      }

      /**
       * Returns fill char.
       * 
       * @return  Fill char.
       */
      @Override
      public char getFillChar()
      {
         return fillChar;
      }

      /**
       * Validates character against format rules. This method should not be
       * called.
       * 
       * @param   ch
       *          Character to check.
       * 
       * @return  Always throws <code>UnsupportedOperationException</code>.
       */
      @Override
      public String validateChar(char ch)
      {
         throw new UnsupportedOperationException("validateChar() is unsupported for fill chars");
      }

      /**
       * Get the number of the error shown if this char is invalid.
       * 
       * @return  Always throws <code>UnsupportedOperationException</code>.
       */
      @Override
      public int getErrorNum()
      {
         throw new UnsupportedOperationException("getErrorNum() is unsupported for fill chars");
      }
   }

   /**
    * Presentation implementation as ScreenBuffer wrapper.
    */
   private class CharBuf
   extends Presentation
   {
      /** Value chars. May be <code>null</code> which means UNKNOWN value. */
      private StringBuffer buf = null;

      /**
       * Constructs clear presentation.
       */
      public CharBuf()
      {
         buf = new StringBuffer(128);
      }

      /**
       * Constructs from variable's value.
       * 
       * @param   varChar
       *          Variable's value.
       */
      public CharBuf(character varChar)
      {
         if (!varChar.isUnknown())
         {
            // input data
            String text   = varChar.getValue();
            int    txtLen = text.length();

            // initialize our buffer of the proper maximum length
            buf = new StringBuffer(maxPresLength);

            // iterate through the format characters
            for (int i = 0; i < fmtChars.length; i++)
            {
               CharInfo info = fmtChars[i];
               int      idx  = info.getValueIndex();

               if (idx >= txtLen)
               {
                  // no more input data
                  break;
               }

               // only add data that is non-fill data (note that this
               // is safe because the maxPresLength will always be
               // equal to the number of non-fill characters in the
               // format string, so we should not ever add too much
               // data to our buffer)
               if (idx >= 0)
               {
                  // append our data, but optionally uppercase it first
                  buf.append(info.makeValueChar(text.charAt(idx)));
               }
            }
         }
      }

      /**
       * Determine if this format must use the DEFAULT-FIXED-FONT, if no font is specified.
       * 
       * @return   The state of the {@link StringFormat#fixedFont} flag. 
       */
      @Override
      public boolean isFixedFont()
      {
         return fixedFont;
      }

      /**
       * Validate representation according to format rules and return error
       * message if there is an error.
       * 
       * @return   <code>null</code> if no error or error message if there is an
       *           error.
       */
      @Override
      public ErrorEntry[] validateFormat()
      {
         // skip validation of missing buffer
         if (buf == null)
         {
            return null;
         }
         
         ErrorEntry res = null;
         boolean partialValid = true;
         
         for (int i = 0; i < fmtChars.length; ++i)
         {
            CharInfo info = fmtChars[i];
            int valueIndex = info.getValueIndex();
            
            if (valueIndex >= 0)
            {
               char ch = getChar(valueIndex);
               
               String err = info.validateChar(ch);

               if (err != null)
               {
                  if (res == null)
                  {
                     // note that error message slightly differs from one
                     // generated by checkFormat()
                     err = "Character " + ch + " at position " + (i + 1) + " must be " + err;
                     
                     res = new ErrorEntry(info.getErrorNum(), err);
                  }
                  
                  // if an invalid char (other than empty char) is encountered, then we are not
                  // "partially valid".
                  partialValid = (partialValid && ch == ' ');
               }
            }
         }

         ErrorEntry perr = 
            new ErrorEntry(629,"Data string " + buf.toString() + " does not match format");

         return res != null ? (partialValid ? new ErrorEntry[] { res, perr } 
                                            : new ErrorEntry[] { res }) 
                            : null;
      }
      
      /**
       * Checks format rules. Method finishEdit() should be called before it.
       * 
       * @throws  DisplayFormatCheckException
       *          When data is incompatible with format
       */
      @Override
      public void checkFormat()
      throws DisplayFormatCheckException
      {
         if (buf == null || buf.toString().trim().length() == 0)
         {
            return;
         }

         for (int i = 0; i < fmtChars.length; ++i)
         {
            CharInfo info = fmtChars[i];
            int valueIndex = info.getValueIndex();
            if (valueIndex >= 0)
            {
               char ch = getChar(valueIndex);
               String err = info.validateChar(ch);
               if (err != null)
               {
                  err = "Character " + ch + " in position " + (i + 1) + " must be a " + err;
                  ErrorManager.displayError(info.getErrorNum(), err);
                  
                  err = "Data string " + buf.toString() + " does not match format";
                  ErrorManager.displayError(629, err);

                  err = ErrorManager.buildErrorText(629, err);
                  throw new DisplayFormatCheckException(err);
               }
            }
         }
      }

      /**
       * Returns if the presentation currently represents the 
       * <code>unknown value</code>.  If any editing has occurred, then this
       * will return <code>false</code> even if the editing is not complete.
       *
       * @return   <code>true</code> if this is <code>unknown</code>.
       */
      @Override
      public boolean isUnknown()
      {
         return (buf == null);
      }
      
      /**
       * Erases one symbol before the cursor position.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean deleteBack()
      {
         if (buf == null)
         {
            return false;
         }

         int cursorPos = presCursorPos;
         while (cursorPos > 0)
         {
            --cursorPos;
            int valueIndex = fmtChars[cursorPos].getValueIndex();
            if (valueIndex >= 0)
            {
               if (valueIndex < buf.length())
               {
                  buf.deleteCharAt(valueIndex);
               }

               // cursor stays on position of erased char
               presCursorPos = cursorPos;
               return true; 
            }
         }

         return false; // no value char found before cursor
      }

      /**
       * Erases one symbol in the cursor position.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean deleteCurrent()
      {
         if (buf == null)
         {
            return false;
         }

         if (presCursorPos >= fmtChars.length)
         {
            return false; // cursor must be over format char
         }

         int valueIndex = fmtChars[presCursorPos].getValueIndex();
         if (valueIndex < 0)
         {
            return false; // cursor must be over value char
         }

         if (valueIndex < buf.length())
         {
            buf.deleteCharAt(valueIndex);
         }
         return true; // cursor doesn't move
      }

      /**
       * Inputs character. The result depends from the current insert mode.
       * 
       * @param   ch
       *          Input character
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean input(char ch)
      {
         if (buf == null)
         {
            return false;
         }

         if (presCursorPos == 0 && ch == '?')
         {
            buf = null;
            return true;
         }

         if (presCursorPos >= fmtChars.length)
         {
            return false; // cursor must be over format char
         }

         return (presInsertMode) ? insert(ch) : overwrite(ch);
      }

      /**
       * Moves cursor one character to the left.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorLeft()
      {
         if (presCursorPos > 0)
         {
            --presCursorPos;
         }
         return true;
      }

      /**
       * Moves cursor one character to the right. Also updates insert mode.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorRight()
      {
         if (presCursorPos < getScreenWidth() - 1)
         {
            ++presCursorPos;
         }
         return true;
      }

      /**
       * Moves cursor to the beginning of the field.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorBegin()
      {
         presCursorPos = 0;
         return true;
      }

      /**
       * Moves cursor to the end of the field.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorEnd()
      {
         presCursorPos = Math.min(toScreenValue().length(),
                                  getScreenWidth() - 1); 
         return true;
      }

      /**
       * Indicates completion of an edit. The presentation instance may fill 
       * some unfilled places.
       * 
       * @return   <code>true</code> if rendering is necessary.
       */
      @Override
      public boolean finishEdit()
      {
         return false; // dummy method
      }

      /**
       * Formats the data as a string that is ready for display.
       * 
       * @return   The formatted string.
       */
      @Override
      public String toScreenValue()
      {
         if (buf == null)
         {
            return "?";
         }

         StringBuffer display = internalScreenValue();
         
         rightTrimStringBuf(display);
         return display.toString();
      }

      /**
       * Instantiates a variable that is equivalent to the presentation's
       * internal representation.
       * 
       * @param    var
       *           A variable which must be compatible with the format.
       */
      @Override
      public void toVar(BaseDataType var)
      {
         character varChar = (character)var;
         if (buf == null)
         {
            varChar.setUnknown();
         }
         else
         {
            rightTrimStringBuf(buf);
            varChar.setValue(buf.toString());
         }
      }

      /**
       * Makes new buffer with removed all value chars after cursor. Insert
       * mode and cursor position are not changed and are the same in the both
       * buffers.
       *
       * @param   zapClear
       *          If <code>true</code> then presentation is cleared completely,
       *          like if zap mode is ON. If <code>false</code> then it is
       *          cleared like if CLEAR event was applied (that means it is
       *          cleared from the current cursor position).
       * 
       * @return  New buffer.
       */
      public CharBuf clear(boolean zapClear)
      {
         int startPos = zapClear ? 0 : presCursorPos;

         CharBuf retBuf = new CharBuf();
         retBuf.presCursorPos = startPos;
         retBuf.presInsertMode = presInsertMode;

         if (buf == null)
         {
            return retBuf;
         }

         int charsKeepCount = buf.length();
         for (int valueCursorPos = startPos;
              valueCursorPos < fmtChars.length; ++valueCursorPos)
         {
            if (fmtChars[valueCursorPos].getValueIndex() >= 0)
            {
               charsKeepCount = fmtChars[valueCursorPos].getValueIndex();
               break;
            }
         }

         for (int i = 0; i < Math.min(charsKeepCount, buf.length()); ++i)
         {
            retBuf.buf.append(buf.charAt(i));
         }

         return retBuf;
      }

      /**
       * Generate full screen representation.
       * 
       * @return   buffer of full screen value length filled with all chars,
       *           including leading and trailing spaces.
       */
      private StringBuffer internalScreenValue()
      {
         StringBuffer display = new StringBuffer(fmtChars.length);
         
         for (int i = 0; i < fmtChars.length; ++i)
         {
            CharInfo info = fmtChars[i];
            int valueIndex = info.getValueIndex();
            display.append((valueIndex < 0) ? info.getFillChar() :
                                              getChar(valueIndex));
         }
         
         return display;
      }
      
      /**
       * Inputs character with the insert mode.
       * 
       * @param   ch
       *          Input character
       * 
       * @return  Is input accepted (nevertheless it may be silently ignored in
       *          real).
       */
      private boolean insert(char ch)
      {
         int pos = presCursorPos;
         do
         {
            if (fmtChars[pos].getValueIndex() >= 0)
            {
               break;
            }
            ++pos;
         }
         while (pos < fmtChars.length);

         if (pos == fmtChars.length)
         {
            // insert works only if we have some value char after cursor
            return false;
         }

         if (fmtChars[presCursorPos].getValueIndex() < 0 &&
             fmtChars[presCursorPos].getFillChar() == ch)
         { 
            // we were over fill char and entering it. Simply moving cursor
            ++presCursorPos;
            return true;
         }

         CharInfo fc = fmtChars[pos];
         if (fc.validateChar(ch) != null)
         {
            return false; // new char fails validation
         }

         int valueIndex = fc.getValueIndex();
         char valueChar = fc.makeValueChar(ch);

         // cutting last char in screen value if necessary
         if (buf.length() >= maxPresLength)
         {
            buf.setLength(maxPresLength - 1);
         }

         if (valueIndex > buf.length())
         { 
            // insert over finalizing spaces
            setChar(valueIndex, valueChar);
         }
         else
         {
            buf.insert(valueIndex, valueChar);
         }

         if (valueIndex == maxPresLength - 1)
         {
            setLastCharEntered(true);
         }

         presCursorPos = pos + 1;
         presCursorPos = editableSource.adjustCursor(ch, presCursorPos, maxPresLength);
         
         return true;
      }

      /**
       * Inputs character with the overwrite mode.
       * 
       * @param   ch
       *          Input character
       * 
       * @return  Is input accepted (nevertheless it may be silently ignored in
       *          real).
       */
      private boolean overwrite(char ch)
      {
         if (fmtChars[presCursorPos].getValueIndex() < 0 &&
             fmtChars[presCursorPos].getFillChar() == ch)
         {
            // we were over fill char and entering it. Simply moving cursor
            ++presCursorPos;
            return true;
         }

         int pos = presCursorPos;
         do
         {
            if (fmtChars[pos].getValueIndex() >= 0)
            {
               break;
            }
            ++pos;
         }
         while (pos < fmtChars.length);

         if (pos == fmtChars.length)
         { 
            // no value char to overwrite
            return false;
         }

         CharInfo fc = fmtChars[pos];
         if (fc.validateChar(ch) != null)
         {
            return false; // new char fails validation
         }

         int valueIndex = fc.getValueIndex();
         setChar(valueIndex, fc.makeValueChar(ch));

         if (valueIndex == maxPresLength - 1)
         {
            setLastCharEntered(true);
         }

         presCursorPos = pos + 1;
         presCursorPos = editableSource.adjustCursor(ch, presCursorPos, fmtChars.length);

         return true;
      }

      /**
       * @param   pos
       *          Position.
       * 
       * @return  Char in this position making right fill with spaces.
       */
      private char getChar(int pos)
      {
         return (pos < buf.length()) ? buf.charAt(pos) : ' ';
      }

      /**
       * Places given char into given buffer position.
       * 
       * @param   pos
       *          Position.
       * @param   ch
       *          Char.
       */
      private void setChar(int pos, char ch)
      {
         int lenDiff = pos - buf.length();
         if (lenDiff < 0)
         {
            buf.setCharAt(pos, ch);
         }
         else if (ch != ' ') // all chars after buffer are spaces
         {
            while (lenDiff > 0)
            {
               buf.append(' ');
               --lenDiff;
            }
            buf.append(ch);
         }
      }
   }
}