/*
** Module   : BrowseGuiImpl.java
** Abstract : GUI browse implementation
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SVL 20151013 Initial version.
** 002 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 003 SVL 20151019 Layout procedure refactoring. Proper horizontal and vertical scrolling.
** 004 SVL 20151027 Fixed layout and re-layout (on column addition). Proper handling of scrolling
**                  keys. Added support for FIT-LAST-COLUMN.
** 005 SVL 20151202 Update column.WIDTH attribute after browse layout. Fixed right border of a
**                  fitted column.
** 006 SVL 20151230 Implemented editable browse.
** 007 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 008 IAS 20160229 LOAD-MOUSE-POINTER support.
** 009 CA  20160216 Cache the intermediate widget states.
** 010 SVL 20160314 Fixed clicking on a locked column in web version, Page Up/Down for editing
**                  browse. Fixed issues related to the case when a new edited row is out of the
**                  view. Added moveColumn.
** 011 VIG 20160107 Added parameters to horizontalScrollbar.onButtonClick() calls and to 
**                  positionUpdated() (scroll step type), fixed javadoc, fixed VerticalRowScroller.
** 012 SVL 20160412 Updated getViewportDimension, clipping rectangles. Added ColumnsScrollPane.
**                  Implemented column resizing.
** 013 SVL 20160608 VALUE-CHANGE event is fired on mouse click.
** 014 SVL 20160719 Updated translate() signature.
** 015 SVL 20160722 Implement browse multiple selection with Shift and Ctrl keys. Implemented
**                  scrolling for LONG step.
** 016 HC  20160712 Improved next/prev tab focus in nested frames.
** 017 EVL 20160705 Fix for thumb size precision loss in calculation.
** 018 EVL 20160915 Fix for vertical scrollbar enable state and thumb presenting when no data
**                  inside browse.
** 019 SBI 20160921 Fixed NPE caused by markersColumn.
** 020 SVL 20160922 Code refactoring because of switching from BrowseConfig.queryId to
**                  BrowseConfig.queryAssigned. Safety call of initQueryAssigned() in draw().
** 021 EVL 20161014 Fixing the left coordinate of the selection rectangle.  Row marker presence
**                  should include checking the browse widget config instance.
** 022 SVL 20161029 Query model initialization was removed from draw(). Fixed NPE in
**                  resolveColors().
** 023 SVL 20161031 Reverted H018.
** 024 SVL 20170126 Added support for pressed columns and START-SEARCH/END-SEARCH events.
**                  Fix for separator default color.
** 025 OM  20170320 Added support for Themes.
** 026 SVL 20170418 Sort numbers are drawn as images. Implemented search by key.
** 027 EVL 20170723 Adding browse specific installer and other changes for direct manipulation
**                  features.  Installing hover handler for browse widget must install the similar
**                  processing for browse column widgets to properly change mouse cursor.
** 028 SVL 20170814 Implemented server-side ROW-DISPLAY triggering, NO-SCROLLBAR-VERTICAL.
** 029 OM  20170713 LineBorderGuiImpl c'tor changed signature.
** 030 SVL 20170831 Support for toggle-box columns.
**     SVL 20170918 Fixed NPE in moveColumn().
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import java.awt.event.*;
import java.util.logging.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.ScrollBar.ScrollStep;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.FocusEvent;
import com.goldencode.p2j.ui.client.event.listener.FocusListener;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;

import static com.goldencode.p2j.ui.BrowseColumnWidgetType.*;

/**
 * GUI implementation of BROWSE widget.
 */
public class BrowseGuiImpl
extends Browse<GuiOutputManager>
implements Hoverable
{
   /** Top offset for text area in a cell. */
   static final int CELL_OFFSET_TOP = 1;

   /** Bottom offset for text area in a cell. */
   static final int CELL_OFFSET_BOTTOM = 2;

   /** Width of browse/column border. */
   public static final int BORDER_THICKNESS = 1;

   /** Width of the thumb of the horizontal scroll bar. */
   static final int H_BAR_THUMB_WIDTH = 16;

   /** Left offset for text area in a cell. */
   static final int DATA_OFFSET_LEFT = 1;

   /** Right offset for text area in a cell. */
   static final int DATA_OFFSET_RIGHT = 2;

   /** Width of a sort arrow displayed on column labels, in pixels. */
   public static final int SORT_ARROW_WIDTH = 9;

   /** Height of a sort arrow displayed on column labels, in pixels. */
   public static final int SORT_ARROW_HEIGHT = 5;

   /** Width of a sort number displayed on column labels, in pixels. */
   public static final int SORT_NUMBER_WIDTH = 4;

   /** Height of a sort number displayed on column labels, in pixels. */
   public static final int SORT_NUMBER_HEIGHT = 7;

   /** Minimum number of rows in a browse. */
   public static final int MIN_BROWSE_ROWS = 2;

   /**
    * Width of a resize area. See
    * {@link com.goldencode.p2j.ui.client.gui.BrowseColumnGuiImpl} enum ResizeArea for more
    * information.
    */
   static final int COLUMN_RESIZE_AREA_WIDTH = 5;

   /** Top or bottom offset for a column resize contour. */
   static final int RESIZE_LINE_VERTICAL_OFFSET = 3;

   /** Part of the cell the can be hidden at bottom. */
   private static final int ALLOWED_HIDDEN_CELL_BOTTOM_PART = 1;

   /** Default separator color. */
   private static final ColorRgb SEPARATOR_DEFAULT_COLOR = new ColorRgb(192, 192, 192);

   /**
    * Default separator color if background browse color matches {@link #SEPARATOR_DEFAULT_COLOR}.
    */
   private static final ColorRgb SEPARATOR_DEFAULT_COLOR2 = new ColorRgb(128, 128, 128);

   /** Default DOWN value for a dynamic browse. */
   private static final int DEFAULT_DYNAMIC_DOWN_SIZE = 4;

   /** Default width for a dynamic browse. */
   private static final int DEFAULT_DYNAMIC_WIDTH_CHARS = 30;

   /** Name of the sorting arrow up image. */
   private static final String SORT_ARROW_UP_IMAGE_NAME = "sort-arr-en-up";

   /** Name of the sorting arrow down image. */
   private static final String SORT_ARROW_DOWN_IMAGE_NAME = "sort-arr-en-down";

   /** Template of a sorting number image name. */
   private static final String SORT_NUMBER_IMAGE_NAME_TEMPLATE = "sort-num-%d";

   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(BrowseGuiImpl.class);

   /**
    * When the current row is out of the view below, then if we press UP key, the new current row
    * becomes <code>max(1, DOWN - UP_KEY_BOTTOM_OFFSET_ROWS)</code>th row in the viewport
    * (1-based).
    */
   private static final int UP_KEY_BOTTOM_OFFSET_ROWS = 2;

   /**
    * How many rows is scrolled per one mouse wheel rotation.
    */
   private static final int SCROLL_ROWS_PER_WHEEL_ROTATION = 3;

   /** Screen driver */
   private GuiDriver gd;

   /** Font resolver for the browse title font. */
   private GuiFontResolver titleFontResolver;

   /** Font resolver. */
   private GuiFontResolver browseFontResolver;

   /** Content pane for browse columns and data (scrollable part). */
   private ColumnSetContainer scrollableColumnSet;

   /** Content pane for browse columns and data (locked part). */
   private ColumnSetContainer lockedColumnSet;

   /** Scrollable wrapper for content pane (scrollable part). */
   private ScrollPaneGuiImpl scrollPane;

   /** Browse color resolver. */
   private GuiColorResolver colorResolver;

   /** Browse title height (without title borders). */
   private int titleHeight;

   /** Coordinates conversion. */
   private CoordinatesConversion cc;

   /** Separator color. */
   private ColorRgb separatorColor;

   /** Background color. */
   private ColorRgb backgroundColor;

   /**
    * Row height (in pixels) excluding borders and separators. <code>effectiveRowHeight =
    * config.rowHeightPixels + CELL_OFFSET_TOP + CELL_OFFSET_BOTTOM</code>.
    */
   private int effectiveRowHeight;

   /** Height of the labels area (in pixels) excluding borders. */
   private int labelsHeight;

   /** Title panel. */
   private BorderedPanelGuiImpl titlePanel;
   
   /** Total columns width. */
   private int totalColumnsWidth;

   /** Total width of the locked columns. */
   private int lockedColumnsWidth;

   /** Horizontal scrolling step for non-COLUMN-SCROLLING mode. */
   private int hScrollingStep;

   /** Markers column (leftmost). */
   private BrowseColumnGuiImpl markersColumn;

   /** Horizontal scrollbar. */
   private ScrollBarGuiImpl horizontalScrollbar;

   /** Vertical scrollbar. */
   private ScrollBarGuiImpl verticalScrollbar;

   /**
    * Maximum for horizontal scrolling in column scrolling mode. Depends on the number of last
    * columns that fit into the view.
    */
   private int horizontalColumnScrollMax = 0;

   /** If browse gained focus by mouse click rather than other ways. */
   private boolean mouseFocusGained = false;

   /** Vertical scroller. */
   private ScrollBarControllerGuiImpl verticalScroller;

   /** Horizontal scroller. */
   private ScrollBarControllerGuiImpl horizontalScroller;

   /** Reference to the column which is being dragged. */
   private BrowseColumnGuiImpl draggedColumn = null;

   /** Reference to the column which is being resized. */
   private BrowseColumnGuiImpl resizedColumn = null;

   /**
    * Offset from the left edge of the column which is being dragged to the position of the
    * pointer inside the column.
    */
   private int leftDragColumnMouseOffset = 0;

   /**
    * Offset from the left edge of the browse (or the edge of the markers column) to the left
    * edge of the column which is being dragged.
    */
   private int leftDragColumnOffset = 0;

   /**
    * Offset from the left edge of the browse to the left edge of the contour of the column which
    * is being resized.
    */
   private int leftResizeColumnOffset = 0;

   /**
    * Offset from the left edge of the browse to the right edge of the contour of the column which
    * is being resized.
    */
   private int rightResizeColumnOffset = 0;

   /** Cached sort arrow up image. */
   private ImageWrapper sortArrowUpImage = null;

   /** Cached sort arrow down image. */
   private ImageWrapper sortArrowDownImage = null;

   /** Cached images of sort numbers. */
   private ImageWrapper[] sortNumberImages = null;

   /**
    * <code>true</code> if initial reposition was done (i.e. reposition on opening of the first
    * assigned query). Used in handling of the selection state of the first row on initial display
    * of a single-selection browse.
    */
   private boolean initialRepositionDone = false;

   /**
    * Constructs the Browse widget.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    *
    * @param id
    *        The widget's ID.
    * @param cfg
    *        Browse config,
    */
   public BrowseGuiImpl(WidgetId id, BrowseConfig cfg)
   {
      super(id, cfg);

      gd = (GuiDriver) OutputManager.getDriver();
      cc = screen().coordinates();
   }

   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    *
    * @param    id
    *           The widget's ID.
    * @param    wcfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, BrowseConfig wcfg)
   {
      super.initialize(id, wcfg);
   }

   /**
    * Set container size.
    *
    * @param   width
    *          New width.
    * @param   height
    *          New height.
    */
   @Override
   public void setSize(double width, double height)
   {
      originalWidth = width;
      originalHeight = height;

      if (width == 0 || height == 0)
         return;

      // TODO set ilFirstScrollCol ?

      super.setSize(width, height);
   }

   /**
    * The method is called after the configuration associated with the
    * owner changes. The parameter points to a reference representing
    * the original configuration state. This reference can be used to
    * detect what configuration fields changed and so optimize any
    * processing related to the change.
    * <p>
    * The implemented logic must be idempotent - the method must produce
    * the same results if executed once or multiple times not relying
    * on any particular call sequence or related state. This is
    * necessary so that the method can be repeated or retried as often as
    * necessary without causing unintended effects
    *
    * @param   beforeUpdate
    *          Config reference capturing the config state before
    *          the owner's config was modified.
    */
   @Override
   public void afterConfigUpdate(BrowseConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);

      if (config.queryAssigned)
      {
         initQueryAssigned();
      }

      Window window = window();
      boolean graphicsAvail = gd != null && window != null;
      if (graphicsAvail)
      {
         updateBrowseDimensions();
      }

      setSelectionMultiple(config.multiple);
      setReadOnly(config.readOnly);

      if (!config.performLayout)
      {
         resolveFonts();
         if (colorResolver != null  // init colorResolver for the first time in doLayout
             && graphicsAvail)
         {
            gd.withSelectedWindow(window.getId().asInt(), this::resolveColors);
         }

         if (config.realized)
         {
            if (beforeUpdate.columnScrolling != config.columnScrolling)
            {
               setColumnScrolling(config.columnScrolling);
            }

            if (beforeUpdate.fitLastColumn != config.fitLastColumn ||
                beforeUpdate.noEmptySpace  != config.noEmptySpace)
            {
               fitLastColumn();
            }

            if (beforeUpdate.rowHeightPixels > 0 &&   // make sure it is row height update case
                                                      // rather than an update due to initial row
                                                      // height calculation
                beforeUpdate.rowHeightPixels != config.rowHeightPixels)
            {
               int topRow = getFirstExistingRow();
               if (topRow < 0)
               {
                  topRow = 0;
               }

               effectiveRowHeight = calcRowHeight();
               config.down = calcDown();
               scrollToRow(topRow, false, false);

               if (isEditing())
               {
                  if (isEditorFillIn())
                  {
                     FillIn editor = getActiveFillIn();
                     editor.config().heightPixels = getRowHeight();
                     editor.config().heightChars = cc.heightFromNative(editor.config().heightPixels);
                     editor.setHeight(editor.config().heightChars);
                     editor.repaint();
                  }
                  else if (isEditorToggleBox())
                  {
                     ToggleBox editor = getActiveToggleBox();
                     editor.config().y = getCurrentEditorY();
                     editor.repaint();
                  }
                  else
                  {
                     // TODO combo-box
                  }
               }
            }
         }
      }

      // GUI allows to switch labels visibility after the browse has been realized
      if (!config.performLayout && config.labels != beforeUpdate.labels)
      {
         // TODO fix minor issues
         updateLabelsVisible();
         config.down = calcDown();
         notifyColumnsUpdated();
      }
   }

   /**
    * Get attribute container for the given cell.
    *
    * @param  viewportIndex
    *         0-based row index in the viewport.
    * @param  columnIndex
    *         0-based column index.
    *
    * @return attribute container for the given cell.
    */
   public GuiCellAttributes getCellAttr(int viewportIndex, int columnIndex)
   {
      return getCellAttr(dcData, viewportIndex, columnIndex);
   }

   /**
    * Get attribute container for the given cell from the cache containing intermediate rows
    * ({@link #intermediateDcData}), if it is available, or normat cache ({@link #dcData})
    * otherwise.
    *
    * @param  viewportIndex
    *         0-based row index in the viewport.
    * @param  columnIndex
    *         0-based column index.
    *
    * @return attribute container for the given cell.
    */
   public GuiCellAttributes getCellAttrWithIntermediate(int viewportIndex, int columnIndex)
   {
      return intermediateDcData != null ?
               getCellAttr(intermediateDcData, viewportIndex, columnIndex) :
               getCellAttr(viewportIndex, columnIndex);
   }

   /**
    * Get attribute container for the given cell from the specified row cache.
    *
    * @param  rowCache
    *         Cache of rows: {@link #dcData} or {@link #intermediateDcData}.
    * @param  viewportIndex
    *         0-based row index in the viewport.
    * @param  columnIndex
    *         0-based column index.
    *
    * @return attribute container for the given cell.
    */
   public GuiCellAttributes getCellAttr(BrowseRow[] rowCache, int viewportIndex, int columnIndex)
   {
      if (rowCache == null || viewportIndex >= rowCache.length)
      {
         return null;
      }

      return (GuiCellAttributes) rowCache[viewportIndex].attr[columnIndex];
   }

   /**
    * Notify the browse that some columns have been added to it.
    */
   @Override
   public void notifyColumnsUpdated()
   {
      int scrollingPosition = 0;
      if (!config.performLayout && horizontalScrollbar.isVisible())
      {
         scrollingPosition = horizontalScroller.getPosition();
      }

      resetColumnsCache();
      setSize(originalWidth, originalHeight);

      if (!config.performLayout)
      {
         boolean wasHorizontalScrollbar = totalColumnsWidth > getViewportDimension().width;
         rebuildColumns();
         boolean needHorizontalScrollbar = totalColumnsWidth > getViewportDimension().width;

         if (!wasHorizontalScrollbar && needHorizontalScrollbar)
         {
            showHorizontalScrollbar(true);

            if (config.heightChars <= 0)
            {
               // implicit height - increase browse height
               //TODO fix drawing issue
               int barHeight = horizontalScrollbar.physicalDimension().height;
               setHeight(cc.heightFromNative(physicalDimension().height + barHeight));
               scrollPane.setHeight(
                     cc.heightFromNative(scrollPane.physicalDimension().height + barHeight));

            }
            else
            {
               // explicit height - recalculate DOWN
               config.down = calcDown();
            }

            updateColumnSetsHeight();
         }

         setColumnScrolling(config.columnScrolling);

         fitLastColumn();
      }

      refresh(false);

      if (!config.performLayout &&
          horizontalScrollbar.isVisible() &&
          scrollingPosition <= horizontalScrollbar.getMax())
      {
         horizontalScrollbar.positionUpdated(scrollingPosition, ScrollStep.CUSTOM, false);
      }
      repaint();
   }

   /**
    * Set "intermediate" rows (i.e. the rows that were already fetched in the process of getting
    * new rows by
    * {@link BrowseWidget#getRows(int, int, boolean, int, ScreenBuffer[], RowDisplayStrategy)}),
    * so these rows can be drawn if a pause occurs in ROW-DISPLAY trigger.
    *
    * @param rows
    *        Rows to set.
    */
   @Override
   public void setIntermediateRows(BrowseRow[] rows)
   {
      if (rows == null)
      {
         return;
      }

      if (currentRowDisplayStrategy == RowDisplayStrategy.INCREMENTAL_SCROLL_FWD ||
          currentRowDisplayStrategy == RowDisplayStrategy.INCREMENTAL_SCROLL_BACK)
      {
         int down = getDown();
         int dcSize = dcData == null ? 0 : dcData.length;
         intermediateDcData = new BrowseRow[Math.min(down, dcSize + rows.length)];
         int dcCopyLen = Math.min(dcSize, intermediateDcData.length - rows.length);

         if (currentRowDisplayStrategy == RowDisplayStrategy.INCREMENTAL_SCROLL_FWD)
         {
            if (dcData != null)
            {
               System.arraycopy(dcData, dcSize - dcCopyLen, intermediateDcData, 0, dcCopyLen);
            }
            System.arraycopy(rows, 0, intermediateDcData, dcCopyLen, rows.length);
         }
         else
         {
            // copy in reverse order
            for (int i = 0; i < rows.length; i++)
            {
               intermediateDcData[rows.length - i - 1] = rows[i];
            }

            if (dcData != null)
            {
               System.arraycopy(dcData, 0, intermediateDcData, rows.length, dcCopyLen);
            }
         }
      }

      //TODO other ROW-DISPLAY strategies
   }

   /**
    * Notification of a mouse clicked event occurred for this widget.
    * <p>
    * If the widget is not already focused, it forces the focus to it.  After the focus was
    * received, it computes the caret location, based on the mouse position over the widget's
    * content.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      if (!hasFocus())
      {
         requestFocus();
      }
   }

   /**
    * Notification of a mouse dragged event occurred for this widget. 
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      if (!isEnabled() || !hasFocus())
      {
         return;
      }

      if (isColumnDrag())
      {
         NativePoint point = translate(e);
         leftDragColumnOffset = point.x - leftDragColumnMouseOffset;
         
         TopLevelWindow<GuiOutputManager> wnd = topLevelWindow();

         int leftLimit = getLeftColumnContourLimit();
         if (leftDragColumnOffset < leftLimit)
         {
            leftDragColumnOffset = leftLimit;
            horizontalScrollbar.onButtonClick(ScrollBar.Position.LEFT, 
                                              ScrollStep.SHORT, 
                                              new MouseEvt(e, wnd));
         }
         else
         {
            int rightLimit = getRightColumnContourLimit();
            int colWidth = draggedColumn.physicalDimension().width;
            if (leftDragColumnOffset + colWidth > rightLimit)
            {
               leftDragColumnOffset = rightLimit - colWidth;
               horizontalScrollbar.onButtonClick(ScrollBar.Position.RIGHT, 
                                                 ScrollStep.SHORT, 
                                                 new MouseEvt(e, wnd));
            }
         }

         repaint();
      }
      else if (isColumnResize())
      {
         int pointerX = translate(e).x;
         int leftLimit = leftResizeColumnOffset + getEffectiveMinColumnWidth();

         if (pointerX < leftLimit)
         {
            pointerX = leftLimit;
         }
         else
         {
            int rightLimit = getRightColumnContourLimit();
            if (pointerX > rightLimit)
            {
               pointerX = rightLimit;
            }
         }

         rightResizeColumnOffset = pointerX;

         repaint();
      }
   }
   
   /**
    * Notification of a mouse wheel moved event occurred for this widget.  No-op by default.
    *
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseWheelMoved(MouseWheelEvent e)
   {
      if (getCacheSize() == 0)
      {
         return;
      }

      int rows = e.getWheelRotation() * SCROLL_ROWS_PER_WHEEL_ROTATION;
      if (rows == 0)
      {
         return;
      }

      if (rows > 0 && getCacheSize() < getDown())
      {
         return;
      }

      incrementalScroll(rows, false, true);
      repaint();
   }
   
   /**
    * Get the height of the title, if any.
    * 
    * @return  the title's height.
    */
   public int getTitleHeight()
   {
      return titleHeight;
   }
   
   /**
    * Lay out the browse.
    */
   @Override
   public void doLayout()
   {
      // TODO update firstEditable column

      resolveFonts();
      resolveColors();
      updateLabelsVisible();

      effectiveRowHeight = calcRowHeight();
      config.rowHeightPixels = effectiveRowHeight - CELL_OFFSET_TOP - CELL_OFFSET_BOTTOM;
      config.rowHeightChars = cc.heightFromNative(config.rowHeightPixels);
      titleHeight = calcTitleHeight();

      buildPanels();

      config.down = calcDown();

      if (config.down < MIN_BROWSE_ROWS)
      {
         ErrorManager.displayError(3520,
               String.format("BROWSE widget must have at least %d DOWN", MIN_BROWSE_ROWS), false);
         ErrorManager.displayError(4025,
               String.format("Unable to realise BROWSE %s", config.name), true);

         // TODO 4GL doesn't realise widget at this point, we leave standard row size

         config.rowHeightChars = -1;
         config.rowHeightPixels = -1;
         effectiveRowHeight = calcRowHeight();;
         config.rowHeightPixels = effectiveRowHeight - CELL_OFFSET_TOP - CELL_OFFSET_BOTTOM;
         config.rowHeightChars = cc.heightFromNative(config.rowHeightPixels);
         config.down = calcDown();
      }

      // MIN-HEIGHT-PIXELS equals the size of the browse with the same fonts and characteristics,
      // but with two rows.
      int minHeightPixels = (config.box ? 2 * BORDER_THICKNESS : 0) +
                            (hasTitle() ? titleHeight + BORDER_THICKNESS : 0) +
                            (drawLabels ? getLabelsHeight() + 2 * BORDER_THICKNESS : 0) +
                            getRowHeight() * MIN_BROWSE_ROWS +
                            BORDER_THICKNESS * (MIN_BROWSE_ROWS - 1) +
                            (horizontalScrollbar.isVisible() ?
                                 horizontalScrollbar.physicalDimension().height : 0);
      config.minHeightChars = cc.heightFromNative(minHeightPixels);

      createTitlePanel();

      setColumnScrolling(config.columnScrolling);

      // Update column.WIDTH attribute.
      int colNo = getColumnCount();
      for (int i = 0; i < colNo; i++)
      {
         BrowseColumnGuiImpl column = getColumn(i);

         int width = column.physicalDimension().width;
         width -= column.getColumnPadding();
         column.config().widthChars = cc.widthFromNative(width);
         column.config().widthPixels = width;
      }

      config.performLayout = false;
   }

   /**
    * Get row height (in pixels), excluding borders and separators.
    *
    * @return see above.
    */
   public int getRowHeight()
   {
      return effectiveRowHeight;
   }

   /**
    * Get height of the labels area (in pixels), excluding borders.
    *
    * @return see above.
    */
   public int getLabelsHeight()
   {
      return labelsHeight;
   }

   /**
    * Number of lines in the labels area (matches the number of lines in the tallest multi-line
    * label).
    *
    * @return number of lines in the labels area.
    */
   public int getLabelLinesNum()
   {
      return ilLabelsHeight;
   }

   /**
    * Sets position of terminal's cursor. Redirects this positioning to the in-place editor if it
    * is present. Otherwise does nothing because BROWSE itself has no purpose for the hardware
    * cursor.
    */
   @Override
   public void drawCaret()
   {

   }

   /**
    * Draws the widget: border, header, selection marks and line, internal FillIn etc.
    */
   @Override
   public void draw()
   {
      Window window = window();
      if (window == null)
      {
         return;
      }

      if (config.performLayout)
      {
         doLayout();
      }

      if (!config.realized || !isVisible())
      {
         return;
      }

      // colors must be resolved here, to reflect latest state of the widget (enabled/disabled,etc)
      resolveColors();
      
      NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();

      final NativeRectangle clip = clipRectangle(p, d);

      gd.draw(getId(), p, clip, () ->
      {
         if (titlePanel != null)
         {
            titlePanel.draw();
         }

         // draw browse contents
         scrollPane.draw();

         if (lockedColumnSet != null)
         {
            lockedColumnSet.draw();
         }
         
         Theme theme = ThemeManager.getCurrentTheme();
         
         if (hasFocus() && isEnabled() && !isEditing())
         {
            theme.drawBrowseRowHighlight(this, gd, 0, 0, d.width, d.height);
         }
         
         if (isColumnDrag())
         {
            NativeDimension viewport = getViewportDimension();
            theme.drawBrowseColumnDrag(this,
                                      gd,
                                      leftDragColumnOffset,
                                      hasTitle() ? titleHeight : 0,
                                      draggedColumn.physicalDimension().width,
                                      viewport.height);
         }
         else if (isColumnResize())
         {
            NativeDimension viewport = getViewportDimension();
            
            int topY = (hasTitle() ? titleHeight : 0) +
                        (config.box ? BORDER_THICKNESS : 0) + RESIZE_LINE_VERTICAL_OFFSET;
            int bottomY = viewport.height + topY - 2 * RESIZE_LINE_VERTICAL_OFFSET;
            
            theme.drawBrowseResizingColumn(this,
                                           gd,
                                           leftResizeColumnOffset,
                                           topY,
                                           rightResizeColumnOffset,
                                           bottomY);
         }
      });
   }

   /**
    * Destroy this widget.  Notify all of its components that the widget has been destroyed.
    */
   @Override
   public void destroy()
   {
      if (scrollPane != null)
      {
         OutputManager.instance().getRegistry().removeWidget(scrollPane.getId());
      }

      if (titlePanel != null)
      {
         OutputManager.instance().getRegistry().removeWidget(titlePanel.getId());
      }

      super.destroy();
   }

   /**
    * Handle situation when <code>Browse</code> widget receives the focus.
    * <p>
    * When a help string is defined for the browse it will be displayed on
    * the status line.  This method also sets the selected attribute on.
    *
    * @param    event
    *           Reference to the event.
    */
   @Override
   public void onFocusGained(FocusEvent event)
   {
      if (isEditPossible() && !mouseFocusGained)
      {
         setCurrentSingleRowSelected();
         startEdit();
      }

      super.onFocusGained(event);
   }

   /**
    * Handle situation when <code>Browse</code> widget loses the focus.
    * <p>
    * This method clears the focused flag.
    *
    * @param    event
    *           Reference to the event.
    */
   @Override
   public void onFocusLost(FocusEvent event)
   {
      leaveRow();
      super.onFocusLost(event);
   }

   /**
    * Process direct manipulation event.
    * 
    * @param   event
    *          Event to process.
    */
   @Override
   public void processDirectManipulationEvent(DirectManipulationEvent event)
   {
      int id = event.id();
      
      // row resize messages first
      if (id == Keyboard.SE_START_ROW_RESIZE)
      {
      }
      else if (id == Keyboard.SE_END_ROW_RESIZE)
      {
         int newRowHeight = effectiveRowHeight;
         switch (event.getResizeDirection())
         {
            case MouseDirectManipulation.RESIZE_DIRECTION_N:
               newRowHeight -= event.getYChange();
               break;
            case MouseDirectManipulation.RESIZE_DIRECTION_S:
               newRowHeight += event.getYChange();
               break;
         }
         // commit new height
         config.rowHeightChars = cc.heightFromNative(newRowHeight);
         // refresh browse
         doLayout();
      }
      else
      {
         super.processDirectManipulationEvent(event);
      }
   }
   
   /**
    * Enable/disable widget.
    * 
    * @param   enabled
    */
   @Override
   public void setEnabled(boolean enabled)
   {
      super.setEnabled(enabled);
      // the code below propagates the hoverable handlers install/remove for browse columns
      if (!config.performLayout)
      {
         for (Widget w : scrollableColumnSet.widgets())
         {
            w.setEnabled(enabled);
         }
      }
   }
   
   /**
    * Activates direct manipulation engine for this widget.
    */
   @Override
   protected void activateDirectManipulation()
   {
      if (!config.performLayout)
      {
         BrowseConfig brc = config(); 
         if (brc.rowResizable || brc.selectable)
         {
            super.activateDirectManipulation();
            for (Widget w : scrollableColumnSet.widgets())
            {
               w.config().selectable = brc.selectable;
               ((GuiDriver)screen().getDriver()).registerDirectManipulation(w);
            }
         }
      }
   }
   
   /**
    * Deactivates direct manipulation engine for this widget.
    */
   @Override
   protected void deactivateDirectManipulation()
   {
      BrowseConfig brc = config(); 
      if (!brc.rowResizable && !brc.selectable)
      {
         super.deactivateDirectManipulation();
         for (Widget w : scrollableColumnSet.widgets())
         {
            MouseHandler.deregisterWidget(w.getId().asInt());
         }
      }
   }
   
   /**
    * Checks if direct manipulation feature must be installed or removed in GUI mode.  If config
    * is upadting in a way that stops using direct manipulation it must be removed from system.
    * Different widgets can use different criteria to make this decision.
    * 
    * @return   <code>TRUE</code> if direct manipulation is need to be installed,
    *           <code>FALSE</code> if direct manipulation is need to be removed.
    */
   @Override
   protected boolean isDirectManipulationRequired()
   {
      BrowseConfig brc = config();
      return brc.selectable || brc.rowResizable;
   }
   
   /**
    * Get viewport index of the row which corresponds given vertical position.
    *
    * @param  y
    *         Vertical position (in native units in browse coordinates) that is supposed to be
    *         within one of the viewport rows.
    *
    * @return 0-based viewport index of the row which corresponds the given vertical position.
    *         It can be <code>-1</code> if position is higher than the first row or can be more
    *         than the index of the last row if position is lower than the last row.
    */
   public int getRowByCoordinates(int y)
   {
      if (isLabelsVisible())
      {
         y -= getLabelsHeight() + BORDER_THICKNESS * 2;
      }

      if (y < 0)
      {
         return -1;
      }
      else
      {
         return y / (getRowHeight() + BORDER_THICKNESS);
      }
   }
   
   /**
    * Get Y coordinate minus possible labels height.
    *
    * @param  y
    *         Vertical position (in native units in browse coordinates) that is supposed to be
    *         within one of the viewport rows.
    *
    * @return Vertical position without optional label height in case of labels exist or
    *         unchanged position if no labels.
    */
   public int fixupNoLabelsY(int y)
   {
      return isLabelsVisible() ? y - (getLabelsHeight() + BORDER_THICKNESS * 2) : y;
   }
   
   /**
    * Gets scrollpane container to draw something in coordinates related to it.
    *
    * @return Scrollpane containing browse columns.
    */
   public ScrollPaneGuiImpl getScrollPane()
   {
      return scrollPane;
   }
   
   /**
    * Gets vertical position of the given row related to the browse column vertical start.
    *
    * @param  rowNdx
    *         0-base row index to compute the Y coordinate.
    *
    * @return Y coordinate of the top border for the row specified.
    */
   public int getTopYforRow(int rowNdx)
   {
      int res = rowNdx * (getRowHeight() + BORDER_THICKNESS);
      
      if (isLabelsVisible())
      {
         res += getLabelsHeight() + BORDER_THICKNESS * 2;
      }
      
      return res;
   }

   /**
    * Get cached sorting arrow image.
    *
    * @param  ascending
    *         <code>true</code> to return soring arrow for ascending sorting (arrow up).
    *         <code>false</code> to return soring arrow for descending sorting (arrow down).
    *
    * @return cached sorting arrow image (up or down arrow).
    */
   public ImageWrapper getSortArrowImage(boolean ascending)
   {
      if (sortArrowUpImage == null)
      {
         sortArrowUpImage =
               VectorGraphicsHelper.getScalableImage(SORT_ARROW_UP_IMAGE_NAME,
                     SORT_ARROW_WIDTH, SORT_ARROW_HEIGHT);

         sortArrowDownImage =
               VectorGraphicsHelper.getScalableImage(SORT_ARROW_DOWN_IMAGE_NAME,
                     SORT_ARROW_WIDTH, SORT_ARROW_HEIGHT);
      }

      return ascending ? sortArrowUpImage : sortArrowDownImage;
   }

   /**
    * Get image of the given sort number.
    *
    * @param  num
    *         Sort number, from <code>1</code> to <code>9</code>.
    *
    * @return image of the given sort number.
    */
   public ImageWrapper getSortNumImage(int num)
   {
      if (num < 1 || num > 9)
      {
         throw new IllegalArgumentException("Specify a valid number (1-9).");
      }

      if (sortNumberImages == null)
      {
         sortNumberImages = new ImageWrapper[9];

         for (int i = 0; i < sortNumberImages.length; i++)
         {
            sortNumberImages[i] = VectorGraphicsHelper.getScalableImage(
                                     String.format(SORT_NUMBER_IMAGE_NAME_TEMPLATE, i + 1),
                                     SORT_NUMBER_WIDTH,
                                     SORT_NUMBER_HEIGHT);
         }
      }

      return sortNumberImages[num - 1];
   }

   /**
    * Check if the widget supports LOAD-MOUSE-POINTER() method 
    * 
    * @return <code>true</code> if the widget supports 
    *         LOAD-MOUSE-POINTER() method
    */
   @Override
   public boolean supportsCustomMousePointer()
   {
      return true;
   }

   /**
    * Returns the left limit for the resize column contour (relative to browse).
    *
    * @return see above.
    */
   protected int getLeftColumnContourLimit()
   {
      return (hasRowMarkers() ? markersColumn.physicalDimension().width : 0) -
               (config.box ? 0 : 1);
   }

   /**
    * Returns the right limit for the resize column contour (relative to browse).
    *
    * @return see above.
    */
   protected int getRightColumnContourLimit()
   {
      return getViewportDimension().width + (config.box ? 1 : 0);
   }

   /**
    * Sub-classes will override this to return a list of mouse actions on which the widget
    * responds.  This is compatible with the AWT {@link MouseEvent} codes and P2J specific
    * {@link com.goldencode.p2j.ui.client.event.MouseEvt} actions.
    *
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvent.MOUSE_CLICKED,
         MouseEvent.MOUSE_PRESSED,
         MouseEvent.MOUSE_RELEASED,
         MouseEvent.MOUSE_DRAGGED,
         MouseEvent.MOUSE_WHEEL
      };
   }

   /**
    * Handle TAB key press.
    */
   protected void handleTabKey()
   {
      if (!isEditing())
      {
         ThinClient.nextTabStop(this);
      }
      else
      {
         switchToNextCell();
      }
   }

   /**
    * Scroll to the editable row if it is out of the view.
    */
   protected void scrollToEditableRow()
   {
      int currRow = getCurrentRow();
      if (!isRowInView(currRow))
      {
         if (isNewEditingRowOutOfTheView())
         {
            scrollToRow(editingExistingRowBefore + 1);
         }
         else
         {
            int topRow = getFirstExistingRow();
            if (currRow < topRow || topRow == -1)
            {
               scrollToRow(currRow);
            }
            else
            {
               scrollToRow(currRow, true, false, false, true, true, getPageSize(), false);
            }
         }
      }
   }

   /**
    * Get browse font resolver.
    *
    * @return see above.
    */
   GuiFontResolver getBrowseFontResolver()
   {
      return browseFontResolver;
   }

   /**
    * Get browse color resolver.
    *
    * @return see above.
    */
   public GuiColorResolver getBrowseColorResolver()
   {
      return colorResolver;
   }

   /**
    * Get separator color.
    *
    * @return separator color.
    */
   public ColorRgb getSeparatorColor()
   {
      return separatorColor;
   }

   /**
    * Get top offset of the row (relatively to the place where label drawing starts).
    *
    * @param  viewportIndex
    *         0-based index of the row in the viewport.
    *
    * @return top offset of the row.
    */
   public int getTopRowOffset(int viewportIndex)
   {
      int topRowOffset = 0;
      if (isLabelsVisible())
      {
         topRowOffset += getLabelsHeight() + BORDER_THICKNESS * 2;
      }
      topRowOffset += (getRowHeight() + BORDER_THICKNESS) * viewportIndex;

      return topRowOffset;
   }

   /**
    * Check if the browse should be drawn as a focused widget.
    *
    * @return <code>true</code> is if the browse should be drawn as a focused widget.
    */
   public boolean isDrawnAsFocused()
   {
      return hasFocus() || isEditing();
   }

   /**
    * Determine if the labels area is visible.
    *
    * @return <code>true</code> if the labels area is visible.
    */
   boolean isLabelsVisible()
   {
      return drawLabels;
   }

   /**
    * Return <code>true</code> if browse layout needs to be performed.
    *
    * @return see above.
    */
   boolean isPerformLayout()
   {
      return config.performLayout;
   }

   /**
    * Get viewport dimension.
    *
    * @return viewport dimension.
    */
   public NativeDimension getViewportDimension()
   {
      NativeDimension dim = scrollableColumnSet.getVisibleDimension();

      ScrollBarGuiImpl horizontal = scrollPane.getScrollBar(ScrollBar.Orientation.HORIZONTAL);
      if (horizontal.isVisible())
      {
         dim.height -= horizontal.physicalDimension().height;
      }

      ScrollBarGuiImpl vertical = scrollPane.getScrollBar(ScrollBar.Orientation.VERTICAL);
      if (vertical.isVisible())
      {
         dim.width -= vertical.physicalDimension().width;
      }

      return dim;
   }

   /**
    * Start column dragging.
    *
    * @param column
    *        Column which is being dragged.
    * @param leftDragColumnMouseOffset
    *        Offset from the left edge of the column which is being dragged to the position of the
    *        pointer inside the column.
    */
   void startColumnDrag(BrowseColumnGuiImpl column, int leftDragColumnMouseOffset)
   {
      draggedColumn = column;
      this.leftDragColumnMouseOffset = leftDragColumnMouseOffset;
   }

   /**
    * Stop column dragging.
    */
   void stopColumnDrag(MouseEvent e)
   {
      if (draggedColumn == null)
      {
         return;
      }

      int pointerX = translate(e).x;
      int leftColumnsOffset = BrowseGuiImpl.this.config.box ? BORDER_THICKNESS : 0;
      int targetColumnIndex = -1;
      if (lockedColumnSet != null)
      {
         for (int i = 0; i < config.numLockedColumns; i++)
         {
            BrowseColumnGuiImpl column = getColumn(i);
            if (pointerX < leftColumnsOffset + column.physicalLocation().x +
                              column.physicalDimension().width)
            {
               targetColumnIndex = i;
               break;
            }
         }
      }

      int columnCount = getColumnCount();
      if (targetColumnIndex == -1)
      {
         for (int i = config.numLockedColumns; i < columnCount; i++)
         {
            BrowseColumnGuiImpl column = getColumn(i);
            if (pointerX < leftColumnsOffset + scrollableColumnSet.physicalLocation().x +
                              column.physicalLocation().x + column.physicalDimension().width)
            {
               targetColumnIndex = i;
               break;
            }
         }
      }

      if (targetColumnIndex == -1)
      {
         targetColumnIndex = columnCount - 1;
      }

      int draggedColumnIndex = draggedColumn.getColumnIndex();
      if (targetColumnIndex != draggedColumnIndex && targetColumnIndex != draggedColumnIndex - 1)
      {
         browseModel.moveColumn(draggedColumnIndex,
                                targetColumnIndex < draggedColumnIndex ? targetColumnIndex + 1 :
                                                                         targetColumnIndex);
      }

      leftDragColumnMouseOffset = 0;
      draggedColumn = null;
   }

   /**
    * Start column resizing.
    *
    * @param columnIndex
    *        Index of the column which is being resized.
    */
   void startColumnResize(int columnIndex)
   {
      resizedColumn = getColumn(columnIndex);

      int browseLeftEdge = hasRowMarkers() ? markersColumn.physicalDimension().width : 0;
      int columnLeftEdge = resizedColumn.physicalLocation().x +
                              (columnIndex < config.numLockedColumns ? 0 :
                                 resizedColumn.parent().physicalLocation().x);
      leftResizeColumnOffset = Math.max(columnLeftEdge, browseLeftEdge);
   }

   /**
    * Stop column resizing.
    */
   void stopColumnResize()
   {
      if (resizedColumn == null)
      {
         return;
      }

      // True even if the part of the column is out of the view. "+1" is 4GL quirk - resize
      // contour is narrower that the column size.
      int newWidth = rightResizeColumnOffset - leftResizeColumnOffset + 1;
      resizedColumn.setWidth(cc.widthFromNative(newWidth));
      rebuildColumns();

      int columnCount = getColumnCount();
      if ((config.fitLastColumn || config.noEmptySpace) && columnCount > 0)
      {
         // if last column can be fit (it will not be smaller than min column width) then force
         // resize the last column
         if (needsColumnFit())
         {
            fitLastColumn();

            // update total column sizes
            rebuildColumns();
         }
      }

      boolean horizontalScrollbarWasVisible = horizontalScrollbar.isVisible();
      showHorizontalScrollbar(totalColumnsWidth > getViewportDimension().width);

      // recalculate DOWN
      int oldDown = config.down;
      config.down = calcDown();

      if (oldDown < config.down)
      {
         // get more rows
         int lastRow = getLastExistingRow();
         if (lastRow != -1 && dcData.length == oldDown)
         {
            int diff = config.down - oldDown;
            BrowseRow[] serverData = getRows(lastRow + 1,
                                             diff,
                                             true,
                                             RowDisplayStrategy.NO_DISPLAY);
            if (serverData.length > 0)
            {
               BrowseRow[] oldData = dcData;
               dcData = new BrowseRow[oldData.length + serverData.length];
               System.arraycopy(oldData, 0, dcData, 0, oldData.length);
               System.arraycopy(serverData, 0, dcData, oldData.length, serverData.length);
            }
         }
      }
      else if (oldDown > config.down)
      {
         // removed extra rows
         if (dcData.length > config.down)
         {
            BrowseRow[] oldData = dcData;
            dcData = new BrowseRow[config.down];
            System.arraycopy(oldData, 0, dcData, 0, dcData.length);
         }
      }

      if (horizontalScrollbarWasVisible != horizontalScrollbar.isVisible())
      {
         if (config.columnScrolling)
         {
            updateHorizontalColumnScrollMax();
         }
         scrollableColumnSet.setPhysicalLocation(0, 0);

         updateColumnSetsHeight();
      }

      if (isEditing())
      {
         updateEditorAfterColumnModification();
      }

      leftResizeColumnOffset = 0;
      rightResizeColumnOffset = 0;
      resizedColumn = null;
      repaint();
   }

   /**
    * Returns <code>true</code> if a column is being dragged.
    *
    * @return <code>true</code> if a column is being dragged.
    */
   boolean isColumnDrag()
   {
      return draggedColumn != null;
   }

   /**
    * Returns <code>true</code> if a column is being resized.
    *
    * @return <code>true</code> if a column is being resized.
    */
   boolean isColumnResize()
   {
      return resizedColumn != null;
   }

   /**
    * Returns <code>true</code> if the mouse pointer is positioned over the resize area after the
    * right edge of the last locked column.
    *
    * @param  e
    *         Mouse event.
    *
    * @return <code>true</code> if the mouse pointer is positioned over the resize area after the
    *         right edge of the last locked column.
    */
   boolean isLastLockedColumnResizeArea(MouseEvent e)
   {
      if (config.numLockedColumns == 0)
      {
         return false;
      }

      NativePoint point = translate(e);
      int leftOffset = lockedColumnSet.physicalDimension().width +
                        (config.box ? BORDER_THICKNESS : 0);
      return point.x >= leftOffset && point.x < leftOffset + COLUMN_RESIZE_AREA_WIDTH;
   }

   /**
    * Get total width of the locked columns,
    *
    * @return see above.
    */
   int getLockedColumnsWidth()
   {
      return lockedColumnsWidth;
   }

   /**
    * Setups scrolling to fully (as possible) show the given column.
    *
    * @param cellColumn
    *        The cell's column index in model.
    */
   @Override
   protected void columnToView(int cellColumn)
   {
      //TODO implement (is it really needed?)
   }

   /**
    * Setups scrolling to fully (as possible) show the given column.
    * TODO does the same thing as columnToView in ChUI, however in ChUI columnToView is possibly
    * called in wrong places which may cause unnecessary scrolling. This should be investigated.
    *
    * @param columnIndex
    *        Column index in the model.
    */
   protected void scrollToColumn(int columnIndex)
   {
      boolean visible = true;
      boolean toTheLeft = true;

      if (columnIndex >= config.numLockedColumns)
      {
         BrowseColumnGuiImpl column = getColumn(columnIndex);
         int scrollPosition = - scrollableColumnSet.physicalLocation().x;
         int columnPosition = column.physicalLocation().x;
         int columnWidth = column.physicalDimension().width;
         int lockedWidth = lockedColumnSet != null ?
                              lockedColumnSet.physicalDimension().width : 0;
         int columnPosInViewport = columnPosition - scrollPosition - lockedWidth;
         int viewportWidth = getViewportDimension().width;
         int effectiveViewportWidth = viewportWidth - lockedWidth;

         if (columnPosInViewport < 0)
         {
            visible = false;
         }
         else
         {
            int rightColumnEdge = columnPosition + columnWidth;
            int rightViewportEdge = scrollPosition + viewportWidth;
            int allowedRightOffset = (int) (1.9 * hScrollingStep); // approximately

            if (rightColumnEdge > rightViewportEdge + allowedRightOffset)
            {
               visible = false;
               toTheLeft = false;
            }
         }

         if (!visible)
         {
            // scroll to column
            if (toTheLeft)
            {
               scrollableColumnSet.setPhysicalLocation(-getColumnOffset(columnIndex), 0);
            }
            else
            {
               if (config.columnScrolling)
               {
                  int totalWidth = 0;
                  int index;
                  for (index = columnIndex; index >= config.numLockedColumns; index--)
                  {
                     totalWidth += getColumn(index).physicalDimension().width;
                     if (totalWidth > effectiveViewportWidth)
                     {
                        break;
                     }
                  }

                  index = Math.min(index + 1, columnIndex);
                  scrollableColumnSet.setPhysicalLocation(-getColumnOffset(index), 0);
               }
               else
               {
                  int x = column.physicalLocation().x - getViewportDimension().width +
                           column.physicalDimension().width;
                  scrollableColumnSet.setPhysicalLocation(-x, 0);
               }
            }
         }
      }
      else
      {
         // reset scroll pane if a locked column is edited
         scrollableColumnSet.setPhysicalLocation(0, 0);
      }
   }

   /**
    * Creates an editor for the current cell and adds it to the browse. Adjusts selection data if
    * necessary.
    *
    * @param updateSelection
    *        For multi-selection browse: if <code>true</code>, current row becomes the only
    *        selected row.
    * @param useValue
    *        Use the specified value for the editor instead of getting it from the browse cache.
    * @param value
    *        Value to use in the editor. Can be <code>null</code> (valid value). Set
    *        <code>useValue</code> to <code>true</code> if you want to use this parameter.
    */
   @SuppressWarnings("unchecked")
   protected void startEdit(boolean updateSelection, boolean useValue, BaseDataType value)
   {
      if (!hasFocus())
      {
         return;
      }

      super.startEdit(updateSelection, useValue, value);

      int currentRow = getCurrentRow();
      if (editedRowIndex != currentRow)
      {
         editedRow = dcData[dcGetRowIndex(currentRow)];
         editedRowIndex = currentRow;
      }

      editingExistingRowBefore = null;
      if (dcData != null && isNewRow(currentRow))
      {
         int viewportRow = getViewportRow();

         // search for closest row at the top
         for (int i = viewportRow - 1; i >= 0; i--)
         {
            int modelIndex = getCachedRowNumber(i);
            if (!isNewRow(modelIndex))
            {
               editingExistingRowBefore = modelIndex;
               break;
            }
         }

         if (editingExistingRowBefore == null)
         {
            for (int i = viewportRow + 1; i < getCacheSize(); i++)
            {
               int modelIndex = getCachedRowNumber(i);
               if (!isNewRow(modelIndex))
               {
                  editingExistingRowBefore = modelIndex - 1;
                  break;
               }
            }
         }
      }
   }

   /**
    * Create editor widget for the current column. Type of editor corresponds column's widget
    * type (fill-in, toggle-box or combo-box).
    *
    * @param  useValue
    *         Use the specified value for the editor instead of getting it from the browse cache.
    * @param  value
    *         Value to use in the editor. Can be <code>null</code> (valid value). Set
    *         <code>useValue</code> to <code>true</code> if you want to use this parameter.
    *
    * @return editor widget for the current column.
    */
   @Override
   protected Widget createEditor(boolean useValue, BaseDataType value)
   {
      BrowseColumnConfig[] cols = getColumns();
      if (TOGGLE_BOX.equals(cols[currColumn].widgetType))
      {
         return createEditingToggleBox(useValue, value);
      }
      else
      {
         return createEditingFillIn(useValue, value);
      }
      // TODO combo-box
   }

   /**
    * Create toggle-box editor widget for the current column.
    *
    * @param  useValue
    *         Use the specified value for the editor instead of getting it from the browse cache.
    * @param  value
    *         Value to use in the editor. Can be <code>null</code> (valid value). Set
    *         <code>useValue</code> to <code>true</code> if you want to use this parameter.
    *
    * @return toggle-box editor widget for the current column.
    */
   protected ToggleBox createEditingToggleBox(boolean useValue, BaseDataType value)
   {
      ConfigManager mgr =         ConfigManager.getInstance();
      WidgetConfig widgetConfig =   browseModel.getEditorsConfigs()[currColumn];
      BrowseColumnConfig[] cols = getColumns();
      ToggleBoxConfig cfg =       (ToggleBoxConfig) widgetConfig;

      if (mgr.getActiveConfig(cfg.id) == null)
      {
         // Widget is unregistered when it is destroyed in stopEdit. Register it again with
         // empty backup config so all updated values are pushed to server.
         ToggleBoxConfig backupConfig = new ToggleBoxConfig();
         mgr.addWidgetConfig(cfg, backupConfig);
      }

      cfg.pfcolor = config.pfcolor;
      cfg.dcolor = config.dcolor;
      cfg.help = cols[currColumn].help;
      cfg.modified = false;
      cfg.visible = false;

      initEditorConfig(cfg);

      WidgetFactory<?> factory = screen().getFactory();
      WidgetId editorId = cfg.id;
      ToggleBox toggleBox = factory.create(ToggleBox.class, editorId, cfg);

      screen().getRegistry().addWidget(editorId, toggleBox);
      toggleBox.initialize(editorId, cfg);
      ((ToggleBoxGuiImpl) toggleBox).setBrowse(this);

      // reset underline attribute and force Browse colors
      //toggleBox.setBrowse(this);

      BaseDataType toggleValue = useValue ? value : getValue(currColumn);
      if (toggleValue != null)
      {
         toggleBox.setValue(toggleValue);
         toggleBox.setState(ScreenBuffer.CHANGED);
      }
      else
      {
         toggleBox.setState(ScreenBuffer.UNINITIALIZED);
      }

      addEditor(toggleBox);

      toggleBox.setEnabled(true);
      toggleBox.setVisible(true);
      return toggleBox;
   }

   /**
    * Initialize editor widget config.
    *
    * @param cfg
    *        Editor widget config.
    */
   @Override
   protected void initEditorConfig(BaseConfig cfg)
   {
      BrowseColumnGuiImpl column = getColumn(currColumn);
      column.initEditorConfig(cfg);
   }

   /**
    * Add editor widget to the browse.
    *
    * @param editor
    *        Editor widget.
    */
   @Override
   protected void addEditor(Widget editor)
   {
      if (currColumn < config.numLockedColumns)
      {
         lockedColumnSet.add(editor);
      }
      else
      {
         scrollableColumnSet.add(editor);
      }

      if (editor instanceof FillIn)
      {
         editor.addFocusListener(new FocusListener()
         {
            @Override
            public void onFocusGained(FocusEvent event)
            {
               if (mouseFocusGained)
               {
                  FillIn fillIn = (FillIn) editor;
                  fillIn.config().selectionStart = FillInConfig.INVALID_SELECTION;
                  fillIn.config().selectionEnd = FillInConfig.INVALID_SELECTION;
                  fillIn.setCursorOffset(0);
                  fillIn.setZapMode(false);
                  fillIn.setInsertMode(true);
               }

               mouseFocusGained = false;
            }

            @Override
            public void onFocusLost(FocusEvent event)
            {
               BrowseGuiImpl.this.onFocusLost(event);
            }
         });
      }

      scrollToColumn(currColumn);
   }

   /**
    * Clean up editor widget.
    *
    * @param editor
    *        Editor widget.
    */
   @Override
   protected void cleanupEditor(Widget editor)
   {
      gd.withSelectedWindow(window().getId().asInt(),
                            () -> gd.deregisterWidget(editor.getId().asInt()));
   }

   /**
    * Returns <code>true</code> if the current row of the single-selection browse is selected on
    * FOCUS-GAINED and some other events.
    *
    * @return see above.
    */
   protected boolean isAggressiveCurrentRowSelection()
   {
      return false;
   }

   /**
    * Get page size used for Page Up / Page Down.
    *
    * @return increment size (in rows) used for Page Up / Page Down.
    */
   @Override
   protected int getPageSize()
   {
      return getDown() - 1;
   }

   /**
    * Update height of the labels area.
    */
   @Override
   protected void updateLabelsHeight()
   {
      ilLabelsHeight = 1;
      int colNum = getColumnCount();
      for (int i = 0; i < colNum; i++)
      {
         BrowseColumnGuiImpl column = getColumn(i);
         column.updateLabel();
         String[] label = column.getLabelLines();
         ilLabelsHeight = Math.max(ilLabelsHeight, label.length);
      }

      int rowHeight = browseFontResolver.font().fontHeight;
      labelsHeight = rowHeight * ilLabelsHeight + CELL_OFFSET_TOP + CELL_OFFSET_BOTTOM;
   }

   /**
    * Notify that the browse has been scrolled (i.e. the top row has changed).
    */
   protected void notifyBrowseScrolled()
   {
      super.notifyBrowseScrolled();

      Widget editor = getActiveEditor();
      if (editor != null)
      {
         BaseConfig cfg = (BaseConfig) editor.config();
         if (isCurrentRowInView())
         {
            cfg.y = getCurrentEditorY();
         }
         else
         {
            // put the editor outside of the view so it is not visible but can receive key presses
            cfg.y = (int) (- 1.5 * editor.physicalDimension().height);
         }
      }
   }

   /**
    * Get Y coordinate for the current editor (relatively to the containing column set).
    *
    * @return see above.
    */
   protected int getCurrentEditorY()
   {
      BrowseColumnGuiImpl column = getColumn(currColumn);
      switch (column.config().widgetType)
      {
         case TOGGLE_BOX:
            int offset = ThemeManager.getCurrentTheme().getRowToggleOffset(column, true).y;
            return getTopRowOffset(getViewportRow()) + offset;
         case COMBO_BOX:
            return 0; // TODO combo-box
         case FILL_IN:
         default:
            return getTopRowOffset(getViewportRow());
      }
   }

   /**
    * Get X coordinate for the current editor (relatively to the containing column set).
    *
    * @return see above.
    */
   protected int getCurrentEditorX()
   {
      BrowseColumnGuiImpl column = getColumn(currColumn);
      switch (column.config().widgetType)
      {
         case TOGGLE_BOX:
            int offset = ThemeManager.getCurrentTheme().getRowToggleOffset(column, true).x;
            return column.physicalLocation().x + offset;
         case COMBO_BOX:
            return 0; // TODO combo-box
         case FILL_IN:
         default:
            return column.physicalLocation().x;
      }
   }

   /**
    * Determines if browse refreshed on browse widget enabling.
    *
    * @return <code>false</code>
    */
   @Override
   protected boolean isRefreshedOnEnable()
   {
      return false;
   }

   /**
    * Update value of the {@link #drawLabels} attribute which defines if the labels area is
    * visible.
    */
   private void updateLabelsVisible()
   {
      drawLabels = config.labels;

      if (!drawLabels)
      {
         for (BrowseColumnConfig column : getColumns())
         {
            if (column.forceLabel)
            {
               drawLabels = true;
               return;
            }
         }
      }
   }

   /**
    * Calculate title height.
    *
    * @return title height.
    */
   private int calcTitleHeight()
   {
      if (!hasTitle())
      {
         return 0;
      }
   
      return 20 - 2; // used the same formula as for frames but without border
   }

   /**
    * Create title panel.
    */
   private void createTitlePanel()
   {
      if (hasTitle())
      {
         NativeDimension browseDimension = physicalDimension();
         
         // create title panel
         titlePanel = new BorderedPanelGuiImpl(new LineBorderGuiImpl());
         titlePanel.setSize(cc.dimensionFromNative(
               new NativeDimension(browseDimension.width, titleHeight + 2 * BORDER_THICKNESS)));
         titlePanel.setCustomDraw(() ->
         {
            NativeDimension d = titlePanel.physicalDimension();
            ThemeManager.getCurrentTheme().drawBrowseTitlePanel(
                  this, gd, colorResolver, titleFontResolver, 0, 0, d.width - 2, d.height - 2);
         });
         
         add(titlePanel);
         OutputManager.instance().getRegistry().addWidget((Widget) titlePanel);
      }
   }

   /**
    * Create column panels and widgets.
    */
   private void buildPanels()
   {
      boolean implicitHeight = config.heightChars <= 0;
      boolean implicitWidth = config.widthChars <= 0;

      if (implicitWidth && config.dynamic)
      {
         setWidth(DEFAULT_DYNAMIC_WIDTH_CHARS);
         implicitWidth = false;
      }

      // create locked pane
      if (hasRowMarkers() || config.numLockedColumns > 0)
      {
         // create locked columns pane (markers column also belongs to this column)
         lockedColumnSet = new ColumnSetContainer(true);
         lockedColumnSet.initialize();
         lockedColumnSet.setLocation(getLockedPaneLocation());
      }

      // create scrollable columns pane it even if it is empty
      scrollableColumnSet = new ColumnSetContainer(false);
      scrollableColumnSet.initialize();

      // calc horizontal scroll step before creating scroll pane
      hScrollingStep = FontManager.getFontWidth(window(), browseFontResolver.resolveFontNum(), gd);

      // create scroll pane with borders and scrollbars
      scrollPane = new ColumnsScrollPane(scrollableColumnSet);
      scrollPane.setScrollBars(ScrollBar.Bars.BOTH);
      scrollPane.setLocation(getScrollPaneLocation());
      add(scrollPane);

      if (lockedColumnSet != null)
      {
         // proper z-order, scrollPane should be under lockedColumnSet
         add(lockedColumnSet);
      }

      OutputManager.instance().getRegistry().addWidget((Widget) scrollPane);

      // init vertical scrolling
      verticalScroller = new VerticalRowScroller();
      verticalScrollbar = scrollPane.getScrollBar(ScrollBar.Orientation.VERTICAL);
      scrollPane.setScrollBarController(ScrollBar.Orientation.VERTICAL, verticalScroller);

      // init horizontal scrolling
      horizontalScroller = new HorizontalColumnScroller();
      horizontalScrollbar = scrollPane.getScrollBar(ScrollBar.Orientation.HORIZONTAL);
      scrollPane.setScrollBarController(ScrollBar.Orientation.HORIZONTAL, horizontalScroller);
      
      if (config.box)
      {
         double x = cc.baseUnits().x * BORDER_THICKNESS;
         double y = cc.baseUnits().y * BORDER_THICKNESS;
         Border<GuiOutputManager> border =
                      new LineBorderGuiImpl(new Insets(hasTitle() ? 0 : y, x, y, x), null, this);
         scrollPane.setBorder(border);
      }
      
      updateVerticalScrollbarVisibility();
      rebuildColumns();

      int hBarHeight = getScrollbarThickness(ScrollBar.Orientation.HORIZONTAL);
      int vBarWidth = config.scrollbarVertical ?
                        getScrollbarThickness(ScrollBar.Orientation.VERTICAL) : 0;

      scrollableColumnSet.setWidth(cc.widthFromNative(totalColumnsWidth));

      // resize browse according to columns width
      if (implicitWidth)
      {
         int widthNative = totalColumnsWidth + vBarWidth +
                              (config.box ? 2 * BORDER_THICKNESS : 0);
         setWidth(cc.widthFromNative(widthNative));
      }

      int viewportWidth = physicalDimension().width - vBarWidth -
                              (config.box ? 2 * BORDER_THICKNESS : 0);
      boolean useHorizontalScrollbar = totalColumnsWidth > viewportWidth;
      showHorizontalScrollbar(useHorizontalScrollbar);

      if (implicitHeight)
      {
         int nativeHeight = calcImplicitBrowseHeightNoScrollbar();
         if (useHorizontalScrollbar)
         {
            nativeHeight += hBarHeight;
         }

         setHeight(cc.heightFromNative(nativeHeight));
      }

      // set scroll pane size
      NativeDimension scrollPaneDimension = getScrollPaneDimension();
      scrollPane.setSize(cc.dimensionFromNative(scrollPaneDimension));

      // adjust sizes of panels
      adjustLockedColumnsPanel();

      NativeDimension viewportDimension = getViewportDimension();
      scrollableColumnSet.setHeight(cc.heightFromNative(viewportDimension.height));

      if (fitLastColumn())
      {
         rebuildColumns();

         showHorizontalScrollbar(totalColumnsWidth > getViewportDimension().width);
         config.down = calcDown();

         if (config.columnScrolling)
         {
            updateHorizontalColumnScrollMax();
         }
         updateColumnSetsHeight();
      }
   }

   /**
    * Re-initiate column widgets.
    */
   private void rebuildColumns()
   {
      int colNum = getColumnCount();

      // iterate columns and get total width
      totalColumnsWidth = 0;
      lockedColumnsWidth = 0;
      if (hasRowMarkers())
      {
         createMarkersColumn();
         if (markersColumn.parent() != null)
         {
            markersColumn.parent().detach(markersColumn);
         }

         lockedColumnSet.add(markersColumn);
         int width = markersColumn.physicalDimension().width;
         totalColumnsWidth += width;
         lockedColumnsWidth += width;;
      }

      for (int i = 0; i < colNum; i++)
      {
         BrowseColumnGuiImpl column = getColumn(i);

         if (column.parent() != null)
         {
            column.parent().detach(column);
         }

         if (i < config.numLockedColumns)
         {
            lockedColumnSet.add(column);
            column.init(this, totalColumnsWidth, i);
            lockedColumnsWidth += column.physicalDimension().width;
         }
         else
         {
            scrollableColumnSet.add(column);
            column.init(this, totalColumnsWidth, i);
         }
         totalColumnsWidth += column.physicalDimension().width;
      }

      if (colNum > 0)
      {
         // right border is drawn if possible but not included in width calculations
         totalColumnsWidth -= BORDER_THICKNESS;
      }

      adjustLockedColumnsPanel();

      updateLabelsHeight();
   }
   
   /**
    * Get the total width of the columns of this browse widget.
    * 
    * @return  the total width of the columns. 
    */
   public int getTotalColumnsWidth()
   {
      return totalColumnsWidth;
   }
   
   /**
    * Get the scrollable column widget.
    * 
    * @return the scrollable column widget.
    */
   public Widget getScrollableColumnSet()
   {
      return scrollableColumnSet;
   }
   
   /**
    * Set the width of the {@link #lockedColumnSet} panel according to
    * {@link #lockedColumnsWidth}.
    */
   private void adjustLockedColumnsPanel()
   {
      // adjust panel sizes according to columns width
      if (lockedColumnSet != null)
      {
         NativeDimension viewportDimension = getViewportDimension();
         lockedColumnSet.setSize(cc.dimensionFromNative(
                                 new NativeDimension(lockedColumnsWidth, viewportDimension.height)));
      }
   }

   /**
    * Fit the last column if FIT-LAST-COLUMN attribute is set.
    *
    * @return <code>true</code> if width of the last column was updated in order to fit it in the
    *         view.
    */
   private boolean fitLastColumn()
   {
      if (!config.fitLastColumn && !config.noEmptySpace)
      {
         return false;
      }

      int columnCount = getColumnCount();
      if (columnCount == 0)
      {
         return false;
      }

      if (needsColumnFit())
      {
         BrowseColumnGuiImpl lastColumn = getColumn(columnCount - 1);
         totalColumnsWidth = getViewportDimension().width;
         int width = totalColumnsWidth - lastColumn.physicalLocation().x + BORDER_THICKNESS;
         lastColumn.setWidth(cc.widthFromNative(width));
         return true;
      }

      return false;
   }

   /**
    * Returns <code>true</code> if last column should be fit according to FIT-LAST-COLUMN or
    * NO-EMPTY-SPACE option.
    *
    * @return see above.
    */
   private boolean needsColumnFit()
   {
      if (!config.fitLastColumn && !config.noEmptySpace)
      {
         return false;
      }

      int columnCount = getColumnCount();
      BrowseColumnGuiImpl lastColumn = getColumn(columnCount - 1);
      int lastColumnWidth = lastColumn.physicalDimension().width;
      int viewportWidth = getViewportDimension().width;

      return config.fitLastColumn ?
            viewportWidth >
                     totalColumnsWidth - lastColumnWidth + getEffectiveMinColumnWidth() :
            viewportWidth > totalColumnsWidth;
   }

   /**
    * Update browse width and height without actual layout.
    */
   private void updateBrowseDimensions()
   {
      boolean implicitWidth = config.widthChars <= 0;
      boolean implicitHeight = config.heightChars <= 0;

      double width = config.widthChars;
      double height = config.heightChars;

      if (!implicitHeight && !implicitWidth)
      {
         setSize(width, height);
         return;
      }

      resolveFonts();
      updateLabelsVisible();

      // iterate columns and get total width
      int totalColumnsWidth = 0;
      if (hasRowMarkers())
      {
         createMarkersColumn();
         totalColumnsWidth += markersColumn.calcColumnWidth();
      }

      int colNum = getColumnCount();
      for (int i = 0; i < colNum; i++)
      {
         BrowseColumnGuiImpl column = getColumn(i);
         column.init(this, totalColumnsWidth, i);
         totalColumnsWidth += column.calcColumnWidth();
      }
      
      int scrollBarWidth = ThemeManager.getCurrentTheme().getScrollBarGaugeWidth();
      // rightmost vertical separator is hidden if columns width is more that viewport size
      totalColumnsWidth -= BORDER_THICKNESS;

      if (implicitWidth)
      {
         if (config.dynamic)
         {
            width = DEFAULT_DYNAMIC_WIDTH_CHARS;
         }
         else
         {
            int totalWidth = totalColumnsWidth + scrollBarWidth +
                  (config.box ? 2 * BORDER_THICKNESS : 0);

            width = cc.widthFromNative(totalWidth);
         }
      }

      if (implicitHeight)
      {
         int nativeHeight = calcImplicitBrowseHeightNoScrollbar();

         int viewportWidth = cc.widthToNative(width) - scrollBarWidth -
                                (config.box ? 2 * BORDER_THICKNESS : 0);

         if (totalColumnsWidth > viewportWidth)
         {
            nativeHeight += scrollBarWidth;
         }

         height = cc.heightFromNative(nativeHeight);
      }

      setSize(width, height);
   }

   /**
    * Get browse native height if it is not specified explicitly (based on explicit or implicit
    * DOWN value). Height is calculated as if the horizontal scrollbar is not used.
    *
    * @return see above.
    */
   private int calcImplicitBrowseHeightNoScrollbar()
   {
      int rowHeight = calcRowHeight();

      int nativeHeight = 0;
      if (hasTitle())
      {
         nativeHeight += calcTitleHeight() + 2 * BORDER_THICKNESS;
      }

      if (config.box)
      {
         nativeHeight += hasTitle() ? BORDER_THICKNESS : 2 * BORDER_THICKNESS;
      }

      if (isLabelsVisible())
      {
         nativeHeight += rowHeight + 2 * BORDER_THICKNESS;
      }

      int down = config.down > 0 ? config.down : DEFAULT_DYNAMIC_DOWN_SIZE;
      nativeHeight += down * rowHeight + (down - 1) * BORDER_THICKNESS;

      return nativeHeight;
   }
   
   /**
    * Obtain the current marker column. If if does not exist, return {@code null}. 
    * 
    * @return  the current marker column or {@code null} if not yet created. 
    */
   public BrowseColumnGuiImpl getCurrentMarkersColumn()
   {
      return markersColumn;
   }
   
   /**
    * Create markers (leftmost) column if it doesn't exist yet.
    */
   private void createMarkersColumn()
   {
      if (markersColumn == null)
      {
         markersColumn = new BrowseColumnGuiImpl(true);
         markersColumn.initMarkersColumn(this);
      }
   }
   
   /**
    * Calculate DOWN value.
    */
   private int calcDown()
   {
      return calcDown(getRowHeight() - CELL_OFFSET_TOP - CELL_OFFSET_BOTTOM);
   }

   /**
    * Calculate DOWN value for the browse basing on the given row height.
    *
    * @param  rowHeightPixels
    *         Row height in pixels.
    *
    * @return calculated DOWN value.
    */
   @Override
   public int calcDown(int rowHeightPixels)
   {
      // Calculate DOWN. It is allowed to hide bottom separator + 1 px of the last cell.
      int availableSpace = getViewportDimension().height;
      if (isLabelsVisible())
      {
         availableSpace -= getLabelsHeight() + BORDER_THICKNESS * 2;
      }

      availableSpace += ALLOWED_HIDDEN_CELL_BOTTOM_PART;

      int effectiveHeight = rowHeightPixels + CELL_OFFSET_TOP + CELL_OFFSET_BOTTOM;

      // exclude bottom border from the calculations
      return ((availableSpace - effectiveHeight) / (effectiveHeight + BORDER_THICKNESS)) + 1;
   }

   /**
    * Get scroll panel location.
    *
    * @return scroll panel location.
    */
   private Point getScrollPaneLocation()
   {
      int y = 0;
      if (hasTitle())
      {
         y += titleHeight + 2 * BORDER_THICKNESS;
      }

      return cc.pointFromNative(new NativePoint(0, y));
   }

   /**
    * Get locked panel location.
    *
    * @return locked panel location.
    */
   private Point getLockedPaneLocation()
   {
      int y = 0;
      int x = 0;
      if (config.box)
      {
         y += BORDER_THICKNESS;
         x += BORDER_THICKNESS;
      }

      if (hasTitle())
      {
         y += titleHeight + BORDER_THICKNESS;
      }

      return cc.pointFromNative(new NativePoint(x, y));
   }

   /**
    * Get scroll pane dimension.
    *
    * @return scroll pane dimension.
    */
   private NativeDimension getScrollPaneDimension()
   {
      NativeDimension dimension = physicalDimension();

      if (hasTitle())
      {
         dimension.height -= titleHeight + 2 * BORDER_THICKNESS;
      }

      return dimension;
   }

   /**
    * Calculate row height basing on the ROW-HEIGHT-CHARS or font height.
    *
    * @return row height.
    */
   private int calcRowHeight()
   {
      int rowHeight;

      if (!config.realized && config.rowHeightChars <= 0)
      {
         rowHeight = browseFontResolver.font().fontHeight;
      }
      else
      {
         rowHeight = cc.heightToNative(config.rowHeightChars);
      }

      rowHeight += CELL_OFFSET_TOP + CELL_OFFSET_BOTTOM;
      return rowHeight;
   }

   /**
    * Determines if this browse has title.
    *
    * @return <code>true</code> if this browse has title.
    */
   public boolean hasTitle()
   {
      return !config.title.isEmpty();
   }

   /**
    * Resolve colors.
    */
   private void resolveColors()
   {
      Window window = window();

      if (colorResolver == null)
      {
         colorResolver = new GuiColorResolver(gd, config, SysColor.COLOR_WINDOW,
                                                          SysColor.COLOR_WINDOWTEXT);
      }

      colorResolver.colors();
      colorResolver.titleFgColor = colorResolver.sysBgColor;
      colorResolver.titleBgColor = gd.getSysColor(SysColor.COLOR_ACTIVECAPTION);
      colorResolver.boxColor = gd.getSysColor(SysColor.COLOR_WINDOWTEXT);

      backgroundColor = colorResolver.bgColor;

      if (config().separators)
      {
         int color = config().separatorFgColor;
         if (color >= 0)
         {
            separatorColor = ColorManager.getFgColor(window, color);
         }
         else if (!SEPARATOR_DEFAULT_COLOR.equals(backgroundColor))
         {
            separatorColor = SEPARATOR_DEFAULT_COLOR;
         }
         else
         {
            separatorColor = SEPARATOR_DEFAULT_COLOR2;
         }
      }
      else
      {
         separatorColor = backgroundColor;
      }
   }

   /**
    * Resolve fonts.
    */
   private void resolveFonts()
   {
      Window window = window();
      browseFontResolver = new GuiFontResolver(window, gd, config(), false);

      if (hasTitle())
      {
         // browse title uses the window title font
         LabelConfig labelConfig = new LabelConfig();
         labelConfig.font = FontManager.WINDOW_TITLE_FONT;
         titleFontResolver = new GuiFontResolver(window, gd, labelConfig, false);
      }
   }

   /**
    * Get the browse column with the specific index.
    *
    * @param  index
    *         0-based column index (markers column in not taken in consideration).
    *
    * @return widget representing the target column.
    */
   private BrowseColumnGuiImpl getColumn(int index)
   {
      BrowseColumnConfig[] cfg = getColumns();
      if (cfg == null)
         return null;

      BrowseColumnGuiImpl col =
            (BrowseColumnGuiImpl) ThinClient.getInstance().getWidget(cfg[index].id);
      return col;
   }

   /**
    * Get thickness (width or height) of the corresponding scroll bar.
    *
    * @param  orientation
    *         Target scroll bar.
    *
    * @return thickness (width or height) of the corresponding scroll bar.
    */
   private int getScrollbarThickness(ScrollBar.Orientation orientation)
   {
      ScrollBar scrollBar = scrollPane.getScrollBar(orientation);
      NativeDimension d = scrollBar.physicalDimension();

      return orientation == ScrollBar.Orientation.VERTICAL ? d.width : d.height;
   }

   /**
    * Get text of the cell from the cache containing intermediate rows
    * ({@link #intermediateDcData}), if it is available, or normal cache ({@link #dcData})
    *
    * @param  column
    *         The column the target cell belongs to.
    * @param  viewportIndex
    *         0-based index of the target row in the viewport the target cell belongs to.
    *
    * @return text of the target cell.
    */
   public String getCellTextWithIntermediate(BrowseColumnGuiImpl column, int viewportIndex)
   {
      if (intermediateDcData != null)
      {
         return intermediateDcData[viewportIndex].text[column.getColumnIndex()];
      }

      return getCellText(viewportIndex, column.getColumnIndex());
   }

   /**
    * Get value of the cell from the cache containing intermediate rows
    * ({@link #intermediateDcData}), if it is available, or normal cache ({@link #dcData})
    *
    * @param  column
    *         The column the target cell belongs to.
    * @param  viewportIndex
    *         0-based index of the target row in the viewport the target cell belongs to.
    *
    * @return value of the target cell.
    */
   public BaseDataType getCellValueWithIntermediate(BrowseColumnGuiImpl column, int viewportIndex)
   {
      if (intermediateDcData != null)
      {
         return intermediateDcData[viewportIndex].data[column.getColumnIndex()];
      }

      return getCellValue(viewportIndex, column.getColumnIndex());
   }

   /**
    * Get value of the cell.
    *
    * @param  viewportIndex
    *         0-based index of the target row in the viewport the target cell belongs to.
    * @param  colIndex
    *         0-based index of the column the target cell belongs to.
    *
    * @return value of the target cell.
    */
   public BaseDataType getCellValue(int viewportIndex, int colIndex)
   {
      BaseDataType var = null;

      int rowIndex = getCachedRowNumber(viewportIndex);
      if (getCurrentRow() == rowIndex &&
            currRowUpdates != null &&
            currRowUpdates[colIndex] != null)
      {
         var = currRowUpdates[colIndex];
      }

      if (var != null)
      {
         return var;
      }
      else if (isEditing() &&
            rowIndex == editedRowIndex &&
            editedRow != null)
      {
         return editedRow.data[colIndex];
      }
      else
      {
         return dcData[viewportIndex].data[colIndex];
      }
   }

   /**
    * Returns <code>true</code> if the given column is the current column and editing is in
    * progress.
    *
    * @param  column
    *         Column to check.
    *
    * @return see above.
    */
   public boolean isCurrentEditingColumn(BrowseColumnGuiImpl column)
   {
      return isEditing() && column.getColumnIndex() == currColumn;
   }

   /**
    * Get text of the cell.
    *
    * @param  viewportIndex
    *         0-based index of the target row in the viewport the target cell belongs to.
    * @param  colIndex
    *         0-based index of the column the target cell belongs to.
    *
    * @return text of the target cell.
    */
   public String getCellText(int viewportIndex, int colIndex)
   {
      BaseDataType var = null;

      int rowIndex = getCachedRowNumber(viewportIndex);
      if (getCurrentRow() == rowIndex &&
          currRowUpdates != null &&
          currRowUpdates[colIndex] != null)
      {
         var = currRowUpdates[colIndex];
      }

      if (var != null)
      {
         return getCellTextNoError(getColumns()[colIndex], var);
      }
      else if (isEditing() &&
               rowIndex == editedRowIndex &&
               editedRow != null)
      {
         return editedRow.text[colIndex];
      }
      else
      {
         return dcData[viewportIndex].text[colIndex];
      }
   }

   /**
    * Determine if the current row is in the view.
    *
    * @return <code>true</code> if the current row is the view.
    */
   public boolean isCurrentRowInView()
   {
      return isRowInView(getCurrentRow());
   }

   /**
    * Handle row clicks.
    *
    * @param viewportRowIndex
    *        0-based viewport index of the clicked row.
    * @param columnIndex
    *        0-based index of the clicked column.
    * @param e
    *        Original mouse event.
    */
   void rowClicked(int viewportRowIndex, int columnIndex, MouseEvent e)
   {
      //TODO is viewportRowIndex < 0 column labels are pushed down
      mouseFocusGained = true;
      requestFocus();

      if (viewportRowIndex < getCacheSize() && viewportRowIndex >= 0)
      {
         if (isEditing())
         {
            if (viewportRowIndex == getViewportRow() && isCurrentRowInView())
            {
               stopEdit(false);
            }
            else if (!leaveRow())
            {
               mouseFocusGained = false; // in this case zap mode and selection is on
               return;
            }
         }

         int oldRow = getCurrentRow();

         int row = getCachedRowNumber(viewportRowIndex);
         setCurrentRow(row);

         if (isSelectionMultiple())
         {
            if (isShiftModifier(e))
            {
               int min = Math.min(oldRow, getCurrentRow());
               int max = Math.max(oldRow, getCurrentRow());

               selectedRowIndices.clear();
               for (int i = min; i <= max; i++)
               {
                  selectedRowIndices.add(new SelectedRow(i));
               }
            }
            else if (isCtrlModifier(e))
            {
               switchCurrentRowSelection();
            }
            else
            {
               deselectAll();
               selectFocusedRow();
            }
         }

         if (columnIndex >= 0 && isEditPossible())
         {
            BrowseColumnGuiImpl column = getColumn(columnIndex);
            if (!column.config().readOnly)
            {
               currColumn = columnIndex;
               setCurrentSingleRowSelected();

               boolean toggleClicked = false;
               if (TOGGLE_BOX.equals(column.config().widgetType))
               {
                  toggleClicked = column.isInToggleBoxArea(e);
               }

               startEdit();

               if (toggleClicked)
               {
                  // inverse-check toggle-box immediately
                  getActiveToggleBox().mouseClicked(e);
               }
            }
         }

         if (oldRow != row)
         {
            fireValueChanged();
         }

         resetPressedColumns(false);

         repaint();
      }
   }

   /**
    * Set all column in non-pressed state.
    *
    * @param trigger
    *        If <code>true</code> and a column was in pressed state, raise END-SEARCH event.
    */
   void resetPressedColumns(boolean trigger)
   {
      BrowseColumnConfig[] cols = getColumns();
      for (BrowseColumnConfig col : cols)
      {
         if (col.pressed && trigger)
         {
            ThinClient.getInstance().invokeTriggers(this, Keyboard.SE_END_SEARCH, false);
         }

         col.pressed = false;
      }
   }

   /**
    * Returns <code>true</code> if SHIFT key was held down when a mouse button was pressed.
    *
    * @param  e
    *         Mouse event.
    *
    * @return <code>true</code> if SHIFT key was held down when a mouse button was pressed.
    */
   private boolean isShiftModifier(MouseEvent e)
   {
      return (e.getModifiers()   & MouseEvent.SHIFT_MASK)      != 0 ||
             (e.getModifiersEx() & MouseEvent.SHIFT_DOWN_MASK) != 0;
   }

   /**
    * Returns <code>true</code> if CTRL key was held down when a mouse button was pressed.
    *
    * @param  e
    *         Mouse event.
    *
    * @return <code>true</code> if CTRL key was held down when a mouse button was pressed.
    */
   private boolean isCtrlModifier(MouseEvent e)
   {
      return (e.getModifiers()   & MouseEvent.CTRL_MASK)      != 0 ||
             (e.getModifiersEx() & MouseEvent.CTRL_DOWN_MASK) != 0;
   }

   /**
    * Set COLUMN-SCROLLING or non-COLUMN-SCROLLING modes.
    *
    * @param columnScrolling
    *        <code>true</code> for COLUMN-SCROLLING mode, <code>false</code> for
    *        non-COLUMN-SCROLLING mode,
    */
   private void setColumnScrolling(boolean columnScrolling)
   {
      if (columnScrolling)
      {
         updateHorizontalColumnScrollMax();
      }

      scrollableColumnSet.setPhysicalLocation(0, 0);
   }

   /**
    * Show or hide horizontal scrollbar.
    *
    * @param show
    *        <code>true</code> to show, <code>false</code> to hide.
    */
   private void showHorizontalScrollbar(boolean show)
   {
      scrollPane.setShowBars(getShowBars(show, config.scrollbarVertical));
      // updates viewport dimension
      horizontalScrollbar.setVisible(show);
   }

   /**
    * Show or hide vertical scrollbar depending on (NO-)SCROLLBAR-VERTICAL flag state.
    */
   private void updateVerticalScrollbarVisibility()
   {
      scrollPane.setShowBars(getShowBars(horizontalScrollbar.isVisible(),
                                         config.scrollbarVertical));
      // updates viewport dimension
      verticalScrollbar.setVisible(config.scrollbarVertical);
   }

   /**
    * Get {@link ScrollPaneGuiImpl.ShowBars} option which corresponds given visibility of the
    * scroll bars.
    *
    * @param  horizontal
    *         <code>true</code> if horizontal scroll bar is visible.
    * @param  vertical
    *         <code>true</code> if vertical scroll bar is visible.
    *
    * @return {@link ScrollPaneGuiImpl.ShowBars} option which corresponds given visibility of
    *         the scroll bars.
    */
   private ScrollPaneGuiImpl.ShowBars getShowBars(boolean horizontal, boolean vertical)
   {
      ScrollPaneGuiImpl.ShowBars showBars;
      if (vertical)
      {
         showBars = horizontal ? ScrollPaneGuiImpl.ShowBars.ALWAYS :
                                 ScrollPaneGuiImpl.ShowBars.VERTICAL_ALWAYS;
      }
      else
      {
         showBars = horizontal ? ScrollPaneGuiImpl.ShowBars.HORIZONTAL_ALWAYS :
                                 ScrollPaneGuiImpl.ShowBars.NEVER;
      }
      return showBars;
   }

   /**
    * Select current row for single-selection case.
    */
   private void setCurrentSingleRowSelected()
   {
      if (!isSelectionMultiple() && getCacheSize() > 0)
      {
         currRowSelected = true;
         repaint();
      }
   }

   /**
    * Perform search by key: find the row which starts with the given character. Complex rules are
    * applied, see {@link BrowseWidget#getRows(char, int, int, int, int, ScreenBuffer[])} for more
    * information.
    *
    * @param searchChar
    *        Character to be searched for.
    */
   private void searchByKey(char searchChar)
   {
      int searchColumn = getSearchColumn();
      if (searchColumn < 0)
      {
         return;
      }

      // search within the current view first
      if (isCurrentRowInView())
      {
         for (int i = getViewportRow() + 1; i < getCacheSize(); i++)
         {
            if (cellStartsWith(getCellText(i, searchColumn), searchChar))
            {
               setCurrentRow(getCachedRowNumber(i));
               repaint();
               return;
            }
         }
      }

      dcData = getRows(searchChar, searchColumn, getFirstExistingRow(), getLastExistingRow());
      setCurrentRow(browseModel.getCurrentRow());
      repaint();
   }

   /**
    * Get the search column, i.e. column which is searched when a key is pressed. If
    * ALLOW-COLUMN-SEARCHING is <code>false</code>, then it is the first column. If
    * ALLOW-COLUMN-SEARCHING is <code>true</code> then it is the pressed column.
    *
    * @return 0-based index of the search column or <code>-1</code> if there is no column to
    *         search.
    */
   private int getSearchColumn()
   {
      BrowseColumnConfig[] columns = getColumns();

      if (columns.length == 0)
      {
         return -1;
      }
      else if (!config.allowColumnSearching)
      {
         // ALLOW-COLUMN-SEARCHING is disabled
         return 0;
      }
      else
      {
         // ALLOW-COLUMN-SEARCHING is enabled

         int i = 0;
         for (BrowseColumnConfig column : columns)
         {
            if (column.pressed)
            {
               return i;
            }
            i++;
         }

         // no row is selected
         return -1;
      }
   }

   /**
    * Get ancestor Window for this widget.
    *
    * @return  ancestor Window or <code>null</code> if the widget is not yet attached to a Window.
    */
   @Override
   public Window window()
   {
      Window window = parent(Window.class);

      if (window == null && (config.frameId != -1))
      {
         window = ThinClient.getInstance().getFrame(config.frameId).window();
      }

      return window;
   }

   /**
    * Get value of the given column in current row.
    *
    * @param   ordinal
    *          The column ordinal.
    *
    * @return  The value of the given column in current row.
    */
   public BaseDataType getColumnValue(int ordinal)
   {
      if (!isCurrentRowInView())
      {
         return new character();
      }

      return super.getColumnValue(ordinal);
   }

   /**
    * Process key event.
    *
    * @param   event
    *          Reference to keyboard event.
    */
   @Override
   public void onKeyPressed(KeyInput event)
   {
      int key = event.actionCode();

      if (key == Keyboard.KA_CURSOR_RIGHT)
      {
         event.consume();
         if (horizontalScrollbar.isVisible())
         {
            horizontalScrollbar.onButtonClick(ScrollBar.Position.RIGHT, ScrollStep.SHORT, null);
         }
      }
      else if (key == Keyboard.KA_CURSOR_LEFT)
      {
         event.consume();
         if (horizontalScrollbar.isVisible())
         {
            horizontalScrollbar.onButtonClick(ScrollBar.Position.LEFT, ScrollStep.SHORT, null);
         }
      }
      else if (key == Keyboard.KA_CURSOR_DOWN &&
               !isSelectionMultiple() &&
               !currRowSelected &&
               getCacheSize() > 0 &&
               getCurrentRow() == 0)
      {
         // TODO handle the case when the first row is deleted

         // Do not change current row, just select the first row.
         setCurrentSingleRowSelected();

         if (!isCurrentRowInView())
         {
            scrollToRow(0);
         }

         event.consume();
         return;
      }
      else if ((key == Keyboard.KA_CURSOR_DOWN || key == Keyboard.KA_CURSOR_UP) &&
               !isCurrentRowInView() &&
               getCacheSize() > 0)
      {
         event.consume();
         boolean upper = getCurrentRow() < getFirstExistingRow();

         // TODO check if current row is a new row

         if (key == Keyboard.KA_CURSOR_DOWN)
         {
            int oldRow = getCurrentRow();
            int newRow = oldRow + 1;
            if (isRowInView(newRow))
            {
               setCurrentRow(newRow);
               repaint();
            }
            else if (upper)
            {
               scrollToRow(newRow, true);
            }
            else
            {
               scrollToRow(newRow, true, false, false, true, true, getPageSize(), false);
            }

            if (oldRow != getCurrentRow())
            {
               setCurrentSingleRowSelected();
            }
         }
         else // Keyboard.KA_CURSOR_UP
         {
            int newRow = Math.max(0, getCurrentRow() - 1);
            if (upper)
            {
               scrollToRow(newRow, true);
               if (!isCurrentRowInView() && getCacheSize() > 0)
               {
                  setCurrentRow(getCachedRowNumber(0));
                  repaint();
               }
            }
            else
            {
               int repoRow = Math.max(0, getPageSize() - UP_KEY_BOTTOM_OFFSET_ROWS);
               scrollToRow(newRow, true, false, false, true, true, repoRow, false);
            }

            // UP always selects row
            setCurrentSingleRowSelected();
         }

         return;
      }
      else if (key == Keyboard.KA_PAGE_DOWN)
      {
         // TODO check for new rows
         event.consume();

         boolean editing = isEditing();
         if (editing && !leaveRow())
         {
            return;
         }

         int oldRow = getCurrentRow();
         if (isCurrentRowInView())
         {
            int newViewportRow = getViewportRow() + getPageSize();
            if (getCacheSize() < getDown())
            {
               newViewportRow = Math.min(newViewportRow, getCacheSize() - 1);
            }

            if (newViewportRow <= getCacheSize() - 1)
            {
               setCurrentRow(getCachedRowNumber(newViewportRow));
               if (oldRow != getCurrentRow())
               {
                  setCurrentSingleRowSelected();
               }

               if (editing)
               {
                  startEdit();
               }

               repaint();
               return;
            }
         }

         int newRow = getCurrentRow() + getPageSize();
         // TODO if there is overlap, it should be iterative scrolling
         // or may be even it should be iterative scrolling only
         scrollToRow(newRow, true, false, false, true, true, getPageSize(), false);
         if (oldRow != getCurrentRow())
         {
            setCurrentSingleRowSelected();
         }

         if (editing)
         {
            startEdit();
         }
         repaint();
         return;
      }
      else if (key == Keyboard.KA_PAGE_UP)
      {
         // TODO check for new rows
         event.consume();

         boolean editing = isEditing();
         if (editing && !leaveRow())
         {
            return;
         }

         int oldRow = getCurrentRow();
         if (isCurrentRowInView())
         {
            int newViewportRow = getViewportRow() - getPageSize();
            if (newViewportRow >= 0)
            {
               setCurrentRow(getCachedRowNumber(newViewportRow));
               if (oldRow != getCurrentRow())
               {
                  setCurrentSingleRowSelected();
               }

               if (editing)
               {
                  startEdit();
               }
               repaint();
               return;
            }
         }

         int newRow = Math.max(0, getCurrentRow() - getPageSize());
         scrollToRow(newRow, true);
         if (oldRow != getCurrentRow())
         {
            setCurrentSingleRowSelected();
         }

         if (editing)
         {
            startEdit();
         }
         repaint();
         return;
      }
      else if (!isEditing() &&
               !event.isTriggerFired() &&
               Keyboard.isPrintableKey(key) &&
               key != ' ')
      {
         searchByKey((char) key);
         return;
      }

      int oldRow = getCurrentRow();
      super.onKeyPressed(event);

      if (key == Keyboard.KA_CURSOR_UP)
      {
         setCurrentSingleRowSelected();
      }
      else if ((key == Keyboard.KA_HOME || key == Keyboard.KA_END ||
                key == Keyboard.KA_LEFT_END || key == Keyboard.KA_RIGHT_END) &&
               (oldRow != getCurrentRow()))
      {
         setCurrentSingleRowSelected();
      }
   }

   /**
    * Intercept special events. Using regular keyTyped/keyPressed may not work
    * if component is not in focus.
    *
    * @see  Widget#processEvent(Event)
    */
   @Override
   public void processEvent(Event evt)
   {
      if (evt instanceof KeyInput)
      {
         KeyInput ki = (KeyInput) evt;
         if (ki.actionCode() == Keyboard.SE_END_SEARCH)
         {
            if (!ki.isConsumed())
            {
               resetPressedColumns(false);
               repaint();
            }
            return;
         }
      }

      super.processEvent(evt);
   }

   /**
    * Notify the browse that the underlying query has been repositioned..
    *
    * @param   repositionError
    *          For reposition mode - if an error has happened during reposition.
    * @param   targetRepositionRow
    *          For reposition mode - row used as the target for reposition. If it was deleted,
    *          reposition with fetching may end up on a different row.
    */
   @Override
   public void queryRepositioned(boolean repositionError, int targetRepositionRow)
   {
      super.queryRepositioned(repositionError, targetRepositionRow);

      if (!initialRepositionDone)
      {
         initialRepositionDone = true;
         if (isVisible())
         {
            setCurrentSingleRowSelected();
         }
      }
      else
      {
         setCurrentSingleRowSelected();
      }
   }

   /**
    * Set the MOUSE-POINTER attribute
    * 
    * @param name
    *        the attribute value
    */
   @Override
   public void setMousePointerName(String name)
   {
      config.mousePointer = name;
   }

   /**
    * Get the MOUSE-POINTER attribute
    * 
    * @return the attribute value 
    *        
    */
   @Override
   public String getMousePointerImpl()
   {
      return config.mousePointer;
   }

   /**
    * Get the current effective custom mouse pointer
    * 
    * @return the current effective custom mouse pointer  
    *        
    */
   @Override
   public MousePtrWrapper getEffectiveMousePointer()
   {
      return MousePtrWrapper.DEFAULT;
   }

   /**
    * Implements the client side of the MOVE-COLUMN() widget method.<p>
    * Repositions a column in a browse widget.
    *
    * @param  srcIndex
    *         0-based position of the column to be moved.
    * @param  destIndex
    *         0-based position to which the column is moved.
    */
   @Override
   public void moveColumn(int srcIndex, int destIndex)
   {
      if (config.performLayout)
      {
         return;
      }

      resetColumnsCache();
      rebuildColumns();

      // in dcData the first column is the index of the row in the browse model
      moveColumnInArray(dcData, srcIndex + 1, destIndex + 1);
      if (editedRow != null)
      {
         moveColumnInArray(editedRow.data, srcIndex, destIndex);
         moveColumnInArray(editedRow.text, srcIndex, destIndex);
      }
      moveColumnInArray(currRowUpdates, srcIndex, destIndex);

      if (config.columnScrolling)
      {
         // 4GL quirk - scroll to the first column only if COLUMN-SCROLLING is on
         scrollToColumn(0);
      }

      // update currColumn
      if (srcIndex == currColumn)
      {
         currColumn = destIndex;
      }
      else if (srcIndex < currColumn && destIndex >= currColumn)
      {
         currColumn--;
      }
      else if (srcIndex > currColumn && destIndex <= currColumn)
      {
         currColumn++;
      }

      if (isEditing())
      {
         updateEditorAfterColumnModification();
      }

      repaint();
   }

   /**
    * Repositions the element in the array.
    *
    * @param array
    *        Target array.
    * @param srcIndex
    *        0-based position of the element to be moved.
    * @param destIndex
    *        0-based position to which the element is moved.
    */
   private void moveColumnInArray(Object[] array, int srcIndex, int destIndex)
   {
      if (array == null)
      {
         return;
      }

      Object edgeElement = array[srcIndex];
      if (destIndex > srcIndex)
      {
         System.arraycopy(array, srcIndex + 1, array, srcIndex, destIndex - srcIndex);
      }
      else
      {
         System.arraycopy(array, destIndex, array, destIndex + 1, srcIndex - destIndex);
      }
      array[destIndex] = edgeElement;
   }

   /**
    * Set of arranged columns. In browse there are two sets: the set of locked columns (including
    * row markers column) and the set of scrollable columns.
    */
   private class ColumnSetContainer
   extends NativeScrollContainer<GuiOutputManager>
   {
      /**
       * <code>true</code> for locked column set. <code>false</code> for scrollable column set.
       */
      private final boolean locked;

      /**
       * Constructor.
       *
       * @param locked
       *        <code>true</code> for locked column set. <code>false</code> for scrollable column
       *        set.
       */
      ColumnSetContainer(boolean locked)
      {
         this.locked = locked;
      }

      /**
       * Initialize the widget.
       */
      public void initialize()
      {
         config = new BaseConfig();
         config.id = WidgetId.nextID();
         OutputManager.instance().getRegistry().addWidget((Widget) this);
      }

      /**
       * Draw the widget.
       */
      @Override
      public void draw()
      {
         final NativePoint p = this.physicalLocation();
         final NativeDimension d = this.physicalDimension();
         final NativeDimension clipDimension = new NativeDimension(d.width, d.height);
         if (locked &&
             BrowseGuiImpl.this.config.separators &&
             BrowseGuiImpl.this.config.numLockedColumns < getColumnCount())
         {
            clipDimension.width += BORDER_THICKNESS; // for separator
         }
         final NativeRectangle clip = clipRectangle(p, clipDimension);
         
         gd.draw(p, clip, () ->
         {
            // draw the background. It will only be visible the the extreme right and bottom where
            // cells might not overlap it
            gd.setColor(backgroundColor);
            gd.fillRect(0, 0, d.width, d.height);
            
            ColumnSetContainer.super.draw();
         });
      }

      /**
       * Notification of a mouse clicked event occurred for this widget.
       * <p>
       * In GUI mode, if the widget has a popup-menu attached, it will display it.
       *
       * @param    e
       *           The mouse event.
       */
      @Override
      public void mouseClicked(MouseEvent e)
      {
         if (!BrowseGuiImpl.this.hasFocus())
         {
            BrowseGuiImpl.this.requestFocus();
         }

         if (BrowseGuiImpl.this.isEnabled() && e.getID() == MouseEvent.MOUSE_CLICKED)
         {
            NativePoint point = translate(e);
            int rowIndex = getRowByCoordinates(point.y);
            int columnIndex = getColumnCount() - 1; // the click was to the right of the
                                                    // rightmost column
            rowClicked(rowIndex, columnIndex, e);
         }
      }

      /**
       * Destroy this widget.  Notify all of its components that the widget has been destroyed.
       */
      @Override
      public void destroy()
      {
         OutputManager.instance().getRegistry().removeWidget(config.id);
         super.destroy();
      }

      /**
       * Directs the widget to scroll to the specified X and Y positions.
       * It is the widget's responsibility to update its visuals accordingly.
       * Note that the position values are in the same units as the returned
       * dimensions from {@link #getScrollDimension()}, {@link #getVisibleDimension()}
       * and {@link #getScrollStep()}.
       *
       * @param    newXPosition
       *           New X scroll position.
       * @param    newYPosition
       *           New Y scroll position.
       */
      @Override
      public void scroll(Integer newXPosition, Integer newYPosition)
      {
         if (newXPosition != null)
         {
            horizontalScroller.positionUpdated(newXPosition, ScrollStep.CUSTOM, false);
         }

         if (newYPosition != null)
         {
            verticalScroller.positionUpdated(newYPosition, ScrollStep.CUSTOM, false);
         }
      }

      /**
       * Sub-classes will override this to return a list of mouse actions on which the widget
       * responds.  This is compatible with the AWT {@link MouseEvent} codes and P2J specific
       * {@link com.goldencode.p2j.ui.client.event.MouseEvt} actions.
       *
       * @return   See above.
       */
      @Override
      protected int[] mouseActions()
      {
         return new int[]
         {
            MouseEvent.MOUSE_CLICKED,
            MouseEvent.MOUSE_PRESSED
         };
      }
   }

   /**
    * Get offset of the column from the left edge of the viewport or from the right edge of
    * the last locked column if there are locked columns (if the is scroll pane is on its initial
    * position). I.e. first scrollable column always has offset 0 (regardless of presence of
    * locked columns).
    *
    * @param  columnIndex
    *         0-based index of the column in the set of columns.
    *
    * @return see above.
    */
   private int getColumnOffset(int columnIndex)
   {
      BrowseColumnGuiImpl column = getColumn(columnIndex);
      int offset = column.physicalLocation().x;
      offset -= lockedColumnSet != null ? lockedColumnSet.physicalDimension().width : 0;
      return offset;
   }

   /**
    * Update height of scrollable and locked column sets in order to to match the viewport height.
    */
   private void updateColumnSetsHeight()
   {
      double height = cc.heightFromNative(getViewportDimension().height);
      scrollableColumnSet.setHeight(height);
      if (lockedColumnSet != null)
      {
         lockedColumnSet.setHeight(height);
      }
   }

   /**
    * Update editor after column moving or resizing.
    */
   private void updateEditorAfterColumnModification()
   {
      BaseDataType value = getEditorValue();
      stopEdit(false);

      // Just copy value instead of querying browse cache because the editor can be out of the
      // view.
      startEdit(true, true, value);

      // update editor position
      notifyBrowseScrolled();
   }

   /**
    * Update maximum for horizontal scrolling in column scrolling mode. See
    * {@link #horizontalColumnScrollMax}.
    */
   private void updateHorizontalColumnScrollMax()
   {
      int colCount = getColumnCount();

      // how many full columns can fit on the last page?
      int availWidth = getViewportDimension().width;
      if (lockedColumnSet != null)
      {
         availWidth -= lockedColumnSet.physicalDimension().width;
      }

      int lastPageColumns = 0;
      int lastColumnsWidth = 0;
      for (int i = colCount - 1; i >= config.numLockedColumns; i--)
      {
         BrowseColumnGuiImpl column = getColumn(i);
         lastColumnsWidth += column.physicalDimension().width;

         if (lastColumnsWidth > availWidth)
         {
            break;
         }

         lastPageColumns++;
      }

      lastPageColumns = Math.max(0, --lastPageColumns);

      horizontalColumnScrollMax = Math.max(0, colCount - config.numLockedColumns - lastPageColumns - 1);

      if (horizontalScroller.getPosition() > horizontalColumnScrollMax)
      {
         horizontalScroller.positionUpdated(horizontalColumnScrollMax, ScrollStep.CUSTOM, false);
      }
   }

   /**
    * Get minimum visual width of a column (data width + offsets).
    *
    * @return minimum visual width of a column.
    */
   private int getEffectiveMinColumnWidth()
   {
      return config.minColumnWidthPixels + DATA_OFFSET_LEFT + DATA_OFFSET_RIGHT;
   }

   /**
    * Class for handling horizontal scrolling.
    */
   private class HorizontalColumnScroller
   extends ScrollBarControllerGuiImpl
   implements ScrollBarController
   {
      /**
       * Get maximum value for the scroll bar.
       *
       * @return maximum value for the scroll bar.
       */
      @Override
      public int getMax()
      {
         return isColumnScrolling() ? horizontalColumnScrollMax :
                                      totalColumnsWidth - getViewportDimension().width;
      }

      /**
       * Get step for the scroll bar.
       *
       * @return step for the scroll bar.
       */
      @Override
      public int getStep()
      {
         return isColumnScrolling() ? 1 : hScrollingStep;
      }

      /**
       * Get visibility state for the scroll bar.
       *
       * @return <code>true</code> if scroll bar is visible.
       */
      @Override
      public boolean isVisible()
      {
         return isEnabled();
      }

      /**
       * Get enable state for the scroll bar.
       *
       * @return <code>true</code> if scroll bar is enabled.
       */
      @Override
      public boolean isEnabled()
      {
         return isColumnScrolling() ? totalColumnsWidth > getViewportDimension().width:
                                      super.isEnabled();
      }

      /**
       * Get position for the scroll bar.
       *
       * @return position for the scroll bar.
       */
      @Override
      public int getPosition()
      {
         int offset = -scrollableColumnSet.physicalLocation().x;

         if (isColumnScrolling())
         {
            int colNum = getColumnCount();
            for (int i = config.numLockedColumns; i < colNum; i++)
            {
               //TODO drag calculations for the last column are more tolerant, so normally
               //     you shouldn't drag thumb to the very end to get to the last column
               if (offset <= getColumnOffset(i))
               {
                  return i - config.numLockedColumns;
               }
            }

            return 0;
         }
         else
         {
            return offset;
         }
      }

      /**
       * Get thumb size for the scroll bar.
       *
       * @return thumb size for the scroll bar (in pixels).
       */
      @Override
      public int getThumbSize()
      {
         return H_BAR_THUMB_WIDTH;
      }

      /**
       * Get thumb position for the scroll bar.
       *
       * @return thumb position for the scroll bar (distance between left/top button and thumb's
       * left/top edge).
       */
      @Override
      public int getThumbPosition()
      {
         if (drag)
         {
            return dragPos;
         }

         if (scrollBar == null || totalColumnsWidth == 0 || getMax() == 0)
         {
            return 0;
         }

         int availSize = scrollBar.getTravelSize() - getThumbSize();
         int tPos = Coordinate.toInt(Coordinate.scale(availSize * getPosition() / getMax()));
         tPos = Math.max(0, tPos);
         tPos = Math.min(tPos, availSize);
         return tPos;
      }

      /**
       * Notify controller that position has been updated.
       *
       * @param   position 
       *          New position of the scroll bar.
       * @param   scrollStep 
       *          Scroll step type.
       * @param   drag     
       *          <code>true</code> if position was changed because of thumb dragging.
       */
      @Override
      public void positionUpdated(int position, ScrollStep scrollStep, boolean drag)
      {
         int colCount = getColumnCount();
         int oldPosition = getPosition();
         if (position == oldPosition || colCount == 0 || config.numLockedColumns >= colCount)
         {
            return;
         }

         // Mouse click between the thumb and a scroll button.
         if (scrollStep.equals(ScrollStep.LONG))
         {
            int viewportWidth = getViewportDimension().width;

            if (isColumnScrolling())
            {
               // Absolute value of the new position can be incorrect, but it is comparable with the
               // old position. Here we must calculate actual position value.
               if (position > oldPosition)
               {
                  // Scroll right - to the last column in view, or at least to the next column.
                  position = Math.max(oldPosition + 1, getIntersectedColumn(viewportWidth));
               }
               else if (position < oldPosition)
               {
                  // Scroll left by (viewport width - locked column width -
                  // width of the first scrollable column in the view), or at least to the previous
                  // column.
                  int oldColumn = oldPosition + config.numLockedColumns;

                  if (oldColumn + 1 == getColumnCount())
                  {
                     // The last scrollable column is the single scrollable row in the view- scroll
                     // by a single row.
                     position = oldPosition - 1;
                  }
                  else
                  {
                     int secondColumnOffset = getColumnOffset(oldColumn + 1);
                     int leftOffset = secondColumnOffset + lockedColumnsWidth +
                           scrollableColumnSet.physicalLocation().x;
                     if (leftOffset > viewportWidth)
                     {
                        // There is the single scrollable column in the view.
                        position = oldPosition - 1;
                     }
                     else
                     {
                        int rightOffset = viewportWidth - leftOffset;
                        position = Math.min(oldPosition - 1,
                              getIntersectedColumn(-rightOffset + lockedColumnsWidth));
                     }
                  }
               }
            }
            else
            {
               int effectiveViewportWidth = Math.max(0, viewportWidth - lockedColumnsWidth);
               position = position < oldPosition ? position - effectiveViewportWidth :
                                                   position + effectiveViewportWidth;
            }
         }

         position = Math.max(0, position);
         position = Math.min(position, getMax());

         if (isColumnScrolling())
         {
            scrollableColumnSet.setPhysicalLocation(
                  -getColumnOffset(config.numLockedColumns + position), 0);
         }
         else
         {
            scrollableColumnSet.setPhysicalLocation(-position, 0);
         }
         repaint();
      }

      /**
       * Notify controller that thumb drag has ended.
       */
      @Override
      public void stopThumbDrag()
      {
         if (drag)
         {
            super.stopThumbDrag();

            int availSize = scrollBar.getTravelSize() - getThumbSize();

            int position = availSize > 0 ?
                              Coordinate.toInt(Coordinate.scale(dragPos * getMax() / availSize)) :
                              0;

            positionUpdated(position, ScrollStep.CUSTOM, true);
         }
      }

      /**
       * Return value of the COLUMN-SCROLLING attribute.
       *
       * @return see above
       */
      private boolean isColumnScrolling()
      {
         return config.columnScrolling;
      }

      /**
       * Get the column which body intersects given x coordinate.
       *
       * @param  x
       *         X coordinate, offset relatively to left viewport edge (regardless
       *         presence/absence of locked columns).
       *
       * @return 0-based index of the target column or <code>-1</code> if there is no such column.
       */
      private int getIntersectedColumn(int x)
      {
         int lockedWidth = getLockedColumnsWidth();
         int colNo = getColumnCount();
         for (int i = config.numLockedColumns; i < colNo; i++)
         {
            int offset = getColumnOffset(i) + lockedWidth +
                  scrollableColumnSet.physicalLocation().x;
            if (offset > x)
            {
               return i - config.numLockedColumns - 1;
            }
         }

         return -1;
      }
   }

   /**
    * Class for handling vertical scrolling.
    */
   private class VerticalRowScroller
   extends ScrollBarControllerGuiImpl
   {
      /**
       * Constructor.
       */
      public VerticalRowScroller()
      {
         super();
      }

      /**
       * Notify controller that position has been updated.
       *
       * @param   position
       *          New position of the scroll bar.
       * @param   scrollStep
       *          Scroll step type.       
       * @param   drag
       *          <code>true</code> if position was changed because of thumb dragging.
       *          <code>false</code> if position was changed because a scroll button was pressed.
       */
      @Override
      public void positionUpdated(int position, ScrollStep scrollStep, boolean drag)
      {
         if (getCacheSize() > 0)
         {
            if (!drag &&
                isNewEditingRowOutOfTheView() &&
                position < 0 &&
                editingExistingRowBefore < 0)
            {
               // special case - first row is a new row and is out of the view
               incrementalScroll(-1);
               repaint();
               return;
            }

            int oldRow = getPosition();
            if (position < 0)
            {
               position = 0;
            }

            if (oldRow == position)
            {
               repaint();
               return;
            }

            if (drag || ScrollStep.LONG.equals(scrollStep))
            {
               int oldPosition = getPosition();

               if (position == oldPosition)
               {
                  return;
               }
               else
               {
                  if (ScrollStep.LONG.equals(scrollStep))
                  {
                     if (position < oldPosition)
                     {
                        position = oldPosition - (getDown() - 1);
                     }
                     else
                     {
                        position = oldPosition + (getDown() - 1);
                     }
                  }

                  position += getDown() - 1;
                  scrollToRow(position, false, false, false, true, true, getDown() - 1, false);
               }
            }
            else
            {
               if (oldRow < position)
               {
                  if (getCacheSize() == getDown())
                  {
                     if (ScrollStep.MAX.equals(scrollStep))
                     {
                        EventManager.postEvent(new KeyInput(BrowseGuiImpl.this,
                                                            EventType.KEY_PRESSED,
                                                            Keyboard.KA_END));
                     }
                     else if (ScrollStep.LONG.equals(scrollStep))
                     {
                        if (isEditing() && !leaveRow())
                        {
                           return;
                        }
                        
                        int newRow = getCurrentRow() + getPageSize();
                        scrollToRow(newRow, true, false, false, true, true, getPageSize(), false);
                     }
                     else
                     {
                        incrementalScroll(1);
                     }
                  }
               }
               else
               {
                  if (ScrollStep.MIN.equals(scrollStep))
                  {
                     EventManager.postEvent(new KeyInput(BrowseGuiImpl.this,
                                                         EventType.KEY_PRESSED,
                                                         Keyboard.KA_HOME));
                  }
                  else if (ScrollStep.LONG.equals(scrollStep))
                  {
                     if (isEditing() && !leaveRow())
                     {
                        return;
                     }
                     
                     int newRow = getCurrentRow() - getPageSize();
                     scrollToRow(newRow, true, false, false, true, true, getPageSize(), false);
                  }
                  else
                  {
                     incrementalScroll(-1);
                  }
               }
            }
            repaint();
         }
      }

      /**
       * Get step for the scroll bar.
       *
       * @return step for the scroll bar.
       */
      @Override
      public int getStep()
      {
         return 1;
      }

      /**
       * Get maximum value for the scroll bar.
       *
       * @return maximum value for the scroll bar.
       */
      @Override
      public int getMax()
      {
         return Math.max(0, config.maxDataGuess - getDown());
      }

      /**
       * Get visibility state for the scroll bar.
       *
       * @return <code>true</code> if scroll bar is visible.
       */
      @Override
      public boolean isVisible()
      {
         return config.scrollbarVertical;
      }

      /**
       * Get position for the scroll bar.
       *
       * @return position for the scroll bar.
       */
      @Override
      public int getPosition()
      {
         int row = getFirstExistingRow();
         if (row < 0)
         {
            if (editingExistingRowBefore != null)
            {
               row = editingExistingRowBefore;
            }
            else
            {
               row = 0; // safe option
            }

         }
         return row;
      }

      /**
       * Notify controller that thumb drag has ended.
       */
      @Override
      public void stopThumbDrag()
      {
         if (drag)
         {
            super.stopThumbDrag();

            int scrollSize = getAvailableTravelSize() - getThumbSize();
            int maxVal = getMax();
            int position = Coordinate.toInt(Coordinate.scale(dragPos * maxVal / scrollSize));
            position = Math.max(0, position);
            position = Math.min(position, maxVal);

            positionUpdated(position, ScrollStep.CUSTOM, true);
         }
      }

      /**
       * Get thumb size for the scroll bar.
       *
       * @return thumb size for the scroll bar (in pixels).
       */
      @Override
      public int getThumbSize()
      {
         int thumbMinSize = ThemeManager.getCurrentTheme().getScrollBarThumbMinSize();
         if (config.maxDataGuess == 0)
         {
            return thumbMinSize;
         }

         int availSize = getAvailableTravelSize();

         int tHeight = Coordinate.toInt(Coordinate.scale(
                           (double)availSize * (double)getDown() / (double)config.maxDataGuess));

         tHeight = Math.max(tHeight, thumbMinSize);
         tHeight = Math.min(tHeight, availSize);
         return tHeight;
      }

      /**
       * Get thumb position for the scroll bar.
       *
       * @return thumb position for the scroll bar (distance between left/top button and thumb's
       *         left/top edge).
       */
      @Override
      public int getThumbPosition()
      {
         if (drag)
         {
            return dragPos;
         }

         if (config.maxDataGuess - getDown() <= 0)
         {
            return 0;
         }

         int thumbSize = getThumbSize();
         int availSize = getAvailableTravelSize() - thumbSize;

         int pos = Coordinate.toInt(Coordinate.scale(
               availSize * getPosition() / (config.maxDataGuess - getDown())));
         pos = Math.min(pos, availSize);

         return pos;
      }

      /**
       * Get size of the travel area available for thumb scrolling, excluding the bottom gap which
       * is used to indicate scrolling towards the end of the result set.
       *
       * @return size of travel area, in pixels.
       */
      private int getAvailableTravelSize()
      {
         if (scrollBar == null)
         {
            return 0;
         }

         int travelSize = scrollBar.getTravelSize();

         // real data, no estimation
         if (config.maxDataReached)
         {
            return travelSize;
         }

         // default bottom gap
         int gap = Coordinate.toInt(Coordinate.scale(travelSize * 0.01));

         // have we moved past row 100?
         if (config.maxDataGuess > BrowseConfig.DATA_GUESS_DEFAULT_VALUE)
         {
            int topRow = getPosition();
            int shrinkBy = 0;
            int calcRow = topRow + getDown() - BrowseConfig.DATA_GUESS_DEFAULT_VALUE + 1;
            if (calcRow > 0)
            {
               shrinkBy = Coordinate.toInt(Coordinate.scale((gap / 3) * Math.log10(calcRow)));
            }
            gap = Math.max(0, gap - shrinkBy);
         }

         return travelSize - gap;
      }
   }

   /**
    * Scroll pane for the {@link #scrollableColumnSet}. The purpose of this class is to increase
    * the width of the {@link #scrollableColumnSet} if there should be an background-colored empty
    * area with after the last column. This can happen only in column-scrolling mode.
    */
   private class ColumnsScrollPane extends ScrollPaneGuiImpl
   {
      /**
       * Constructor.
       *
       * @param widget Child widget.
       */
      ColumnsScrollPane(ScrollableWidget<GuiOutputManager> widget)
      {
         super(widget);
      }

      /**
       * Adjusts viewport size, scroll bar visibility and positions. Additionally increases
       * the width of the {@link #scrollableColumnSet} if there should be an background-colored
       * empty area with after the last column. This can happen only in column-scrolling mode.
       */
      @Override
      protected void adjustScrollLayoutImpl()
      {
         super.adjustScrollLayoutImpl();

         int columnSetWidth = scrollableColumnSet.physicalDimension().width;
         int columnSetLocation = scrollableColumnSet.physicalLocation().x;
         int viewportWidth = getViewportDimension().width;
         if (columnSetLocation + columnSetWidth < viewportWidth)
         {
            scrollableColumnSet.setWidth(viewportWidth - columnSetLocation);
         }
      }
   }
}
