/*
** Module   : DropDownGuiImpl.java
** Abstract : Drop down box for combo-box GUI specific implementation.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20150428 Created initial version.
** 002 EVL 20150514 Adding method to adjust the visible rows value according to GUI needs.
**                  The scrollbar approach changed from ALWAYS to default(AS_NEEDED).  Several UI
**                  painting related fixes.
** 003 EVL 20150723 Small fix for draw clipping.
** 004 EVL 20150814 Changing height calculation approach.  Fixed drop-down Y coordinate.
** 005 HC  20150831 Scrolling support overhaul.
** 006 CA  20150918 Fix location to be relative to parent control.
** 007 CA  20151006 Fixed window resolution when attaching the drop-down.
** 008 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 009 HC  20151103 Improved GUI COMBO-BOX layout and fixes.
** 010 EVL 20151217 Drop-down approach rework.  Now the set height and pos API can be
**                  significantly simplified.  The overlay window behind drop-down must be able
**                  to receive all focus related events after it is shown.
** 011 EVL 20160118 Implements special interface common for every overlay child.  Adding code to
**                  compute the initial drop-down location depending on window position and screen
**                  height.
** 012 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 013 IAS 20160229 Catch potential exception on baseWnd.hide(), override setEnabled() method
** 014 CA  20160216 Cache the intermediate widget states.
** 015 HC  20160406 COMBO-BOX drop-down resources cleanup. Various drop-down related fixes.
** 016 HC  20160511 Fixed COMBO-BOX focusing issues and related changes.
** 017 SBI 20161004 Fixed visible items for this drop down.
** 018 EVL 20161108 Fix for clipping area.
** 019 CA  20161101 Fixed drawin right-side border for the drop-down.
** 020 OM  20170320 Added support for Themes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.model.*;
import com.goldencode.p2j.ui.client.ScrollBar.Bars;

/**
 * Implements the drop-down window for a combo-box in GUI mode.
 * TODO: this class is not managed by layout manager, review it and make sure that it conforms GUI
 * requirement, i.e. properly maintains its physical position/size. 
 */
public class DropDownGuiImpl
extends DropDown<GuiOutputManager>
{
   /** Inner lines minimum. */   
   public static final int INNER_LINES_MIN = 3;
   
   /** Flag to conrol whether the drop-dow be painting or not. */
   private boolean drawEnabled = false;
   
   /** Screen driver */
   private GuiDriver gd;
   
   /**
    * Create a drop-down container based on the parent combo-box and it's
    * associated data model.
    * 
    * @param    parent
    *           The combo-box object which calls this drop-down.
    */
   @SuppressWarnings("unchecked")
   public DropDownGuiImpl(ComboBox<GuiOutputManager> parent)
   {
      super(parent);
      // disable layout manager
      setLayout(null);
      // set up overlay window
      new OverlayWindow(this, parent.topLevelWindow());
      topLevelWindow().setSize(width(), height());
   }

   /**
    * Calculate (and set) the height and position of the drop-down
    * container based on the parent combo-box and it's associated data 
    * model.
    * <p>
    * Precedence order (in the following list the term "max-space" is
    * the number of rows from the top of the containing combo-box's
    * frame to the bottom of the screen) for vertical size:
    * <p>
    * <pre>
    * 1. If there is an inner-lines definition AND it is smaller than the 
    *    max-space, then it is honored.
    * 2. If the number of elements in the data model is less than the
    *    number of rows in the max-space, then the number of data model
    *    elements will determine the vertical size.
    * 3. max-space is used otherwise.
    * </pre>
    * <p>
    * The drop-down box X coordinate is always the same as the X
    * coordinate of the parent combo-box in absolute screen units. For
    * the Y coordinate, the following positioning algorithm is used:
    * <p>
    * <pre>
    * 1. If the height of the drop-down is less than or equal to the space
    *    between the bottom of the combo-box and the bottom of the screen,
    *    then the drop-down T coordinate will be set to the combo-box Y +
    *    1.
    * 2. Otherwise, the drop-down Y coordinate will equal the Y coordinate
    *    of the bottom of the screen minus the height of the drop-down.
    *    This will result in the drop-down ending at the bottom of the
    *    screen and starting at or above the combo-box Y coordinate BUT
    *    never above the top of the frame containing the combo-box
    *    (because of the previously constrained height calculation above).
    * </pre>
    * 
    * @param    parent
    *           The combo-box object which calls this drop-down.
    * @param    model
    *           The data model to be used within this drop-down list.
    */
   @Override
   public void setHeightAndPos(ComboBox<GuiOutputManager> parent, ComboBoxModel<String> model)
   {
      // screen driver to be initialized early enough
      gd = (GuiDriver) OutputManager.getDriver();
      
      // the border is handled by scrollpane
      setInsets(new Insets(0, 0, 0, 0));
      
      int innerLines  = getComboBox().config().innerLines;
      int listSize    = model.size();
      
      int visibleRows = (innerLines > 0) ? innerLines : listSize;
      
      // set the list height
      list.setVisibleRows(fixVisibleRows(visibleRows, listSize));
      
      // init scrollbars, visible on demand
      ScrollPaneGuiImpl scrollPaneGui = (ScrollPaneGuiImpl) scrollpane;
      scrollPaneGui.setBorder(new LineBorderGuiImpl());
      
      // set list width, account for the border
      list.setColumns(width() - scrollPaneGui.getBorder().insets().widthDelta());

      // start GUI specifics
      // get the absolute coord of the combo-box this drop-down belongs to
      NativePoint cbbPhysLoc = parent.screenPhysicalLocation();
      
      // set the coordinates relative to parent combo-box widget
      //
      // The combo-box and drop-down both belong to different frames. The parent frame
      // for drop-down is border-less and header-less invisible frame located inside Window
      // workspaces. So when we get the physical location of the combo-box we have location
      // related to frame the combo-box belongs to. If we use this coords to place drop-down we
      // have XY shift according to combo-box's frame placement and decoration. That's why we can
      // not just use combo-box location plus combo box height here.
      //
      // The approach is to get absolute coordinates of the combo-box, get the coordinate of the
      // workspace, subtract second from first and add the combo-box height.
      
      // the combo-box height plus some extra pixels to match 4GL look and feel
      int cbbHeightPlus = screen().coordinates().heightToNative(parent.height()) + 
                          ThemeManager.getCurrentTheme().getComboBoxDropDownOffset();
      NativePoint ddPhysLoc = new NativePoint(cbbPhysLoc.x, cbbPhysLoc.y);
      
      // check if the drop-down will fit the current display dimension
      // get combo-box window ID
      int mainWndId = getComboBox().topLevelWindow().getId().asInt();
      // get combo-box absolute coordinates
      NativePoint cbbDispLoc = parent.displayPhysicalLocation();
      // total screen and drop-down dimension
      NativeRectangle nr = gd.getDisplayBounds(mainWndId);
      NativeDimension nDim = physicalDimension();
      
      // calculate final drop-down y coordinate (to display the full list above or below the
      // parent) based on space available, but the preference will be below if possible
      if (cbbDispLoc.y + cbbHeightPlus + nDim.height > nr.bottom() &&
          cbbDispLoc.y - nDim.height > nr.top())
      {
         ddPhysLoc.y -= nDim.height;
      }
      else
      {
         ddPhysLoc.y += cbbHeightPlus;
      }
      
      // set up drop-down location below or above the combo-box depending on above calculations
      setPhysicalLocation(ddPhysLoc.x, ddPhysLoc.y);
      
      scrollPaneGui.setScrollBars(Bars.VERTICAL);
   }

   /**
    * Check if widget is displayed.
    * 
    * @return  <code>true</code> if widget is displayed.
    */
   @Override
   public boolean isDisplayed()
   {
      // do not use topLevelWindow() here as it throws when no top-level window
      TopLevelWindow<?> wnd = parent(TopLevelWindow.class);
      return wnd != null && wnd.isDisplayed();
   }

   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   @Override
   public double height()
   {
      ScrollPaneGuiImpl scrollPaneGui = (ScrollPaneGuiImpl) scrollpane;
      int rowVisible = ((ScrollableSelectionListGuiImpl)scrollPaneGui.getChild()).visibleRows();
      
      // add 2px - the borders size - so that the rowVsible rows fit fully,
      // this is also needed because ScrollableSelectionListGuiImpl explicitly calculates its size
      // as rowheight * rows + 2
      int nativeH = ((ComboBoxGuiImpl)getComboBox()).getRowHeight() * rowVisible + 2 + 2;
      
      return screen().coordinates().heightFromNative(nativeH);
   }
   
   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public double width()
   {
      return getComboBox().width();
   }

   /**
    * Show widget.
    */
   @Override
   public void show()
   {
      _setVisible(true);
      topLevelWindow().show();
   }
   
   /**
    * Hide this widget and remove it.
    */
   @Override
   public void hide()
   {
      topLevelWindow().close();
      super.hide();
   }

   /**
    * Draw widget.  Filters superclass for special cases.
    */
   @Override
   public void draw()
   {
      // drawaing is enabling late enough to prevent screen distortion 
      if (drawEnabled)
      {
         // set up clipping rectangle
         NativePoint p = physicalLocation();
         NativeDimension d = physicalDimension();
         NativeRectangle clip = clipRectangle(p, new NativeDimension(d.width, d.height));
      
         // clippable drawing start 
         gd.draw(getId(), p, clip, this::drawInt);
      }
   }

   /**
    * Enable/disable widget.
    * 
    * @param   enabled
    */
   @Override
   public void setEnabled(boolean enabled)
   {
      scrollpane.setEnabled(enabled);
      super.setEnabled(enabled);
   }

   /**
    * Enables drawing for this drop-down.  This is ideally the last call before starting event
    * processing loop.
    */
   public void enableDraw()
   {
      drawEnabled = true;
   }
   
   /**
    * Determine the location of the drop-down, relative to the parent window.
    * 
    * @return   See above.
    */
   @Override
   public Point screenLocation()
   {
      if (parent() == null) {
         throw new IllegalStateException("widget is not in the container");
      }

      return Coordinate.scale(parent().screenLocation().translate(location()));
   }
   
   /**
    * Process key event.
    * 
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      processKeyListeners(event);
      
      // do not call processKeyEvent() on the combo box, this would
      // lead to recursive call
      if (!(event.source() instanceof ComboBoxGuiImpl))
      {
         event.source().processKeyEvent(event);
      }
      else
      {
         int key = event.actionCode();
         
         // additional action for TAB/BACK-TAB to connect the trigger defined
         if (key == Keyboard.KA_TAB || key == Keyboard.KA_BACK_TAB)
         {
            KeyInput ki = new KeyInput(list, EventType.KEY_PRESSED, event.keyCode());
            EventManager.postEvent(ki);
         }
         
         // the processing should go through scrollable list
         ThinClient.getInstance().independentEventDrawingBracket(this, () ->
         {
            list.processKeyEvent(event);
            repaint();
         });
      }
   }   
   
   /**
    * Get the combo-box GUI specific visible row adjust.
    * 
    * @param    visibleRows
    *           The visible rows value to check and modify.
    * @param    listSize
    *           The total size of the items in the list to handle.
    * 
    * @return  the new visible rows value.
    */
   @Override
   protected int fixVisibleRows(int visibleRows, int listSize)
   {
      int rowsToReturn = visibleRows;
     
      // check the items size
      if (rowsToReturn > listSize)
      {
         rowsToReturn = listSize;
      }
      
      // but there is a minimum for inner lines 
      if (rowsToReturn < INNER_LINES_MIN)
      {
         rowsToReturn = INNER_LINES_MIN;
      }
      
      return rowsToReturn;
   }
}