/*
** Module   : EditorPopupGuiImpl.java
** Abstract : GUI client implementation of Editor system popup.
**
** Copyright (c) 2016-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------------Description-------------------------------------
** 001 VIG 20160216 Created initial version.
** 002 HC  20160607 GUI menus reworked to be displayed in dedicated top-level windows.
** 003 SBI 20160613 Changed to register copy, cut, paste system actions within web client
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import java.awt.event.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * GUI client implementation class of system popups. There are 3 possible syste popup types:
 * <ul>
 *    <li>Editor popup, containing clipboard operations such as CRUD</li>
 *    <li>Window title popup, containing window operations: MIN,MAX,CLOSE ..</li>
 *    <li>Scroll popup, containing scrolling operations for scrollbars</li>
 * </ul> 
 */
public class EditorPopupGuiImpl 
extends MenuGuiImpl
{
   /** CUT items from editor system popup */
   private MenuItemGuiImpl cutOpt;

   /** COPY items from editor system popup */
   private MenuItemGuiImpl copyOpt;

   /** PASTE items from editor system popup */
   private MenuItemGuiImpl pasteOpt;

   /** SELECT_ALL items from editor system popup */
   private MenuItemGuiImpl selectAllOpt;

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param    owner
    *           A widget that will own the popup.
    */
   private EditorPopupGuiImpl(Widget<?> owner)
   {
      super(null, new MenuConfig());
      config().popupOnly = true;
      config().ownerId = owner.getId().asInt();
   }
   
   /**
    * Create a system editor popup menu.
    * 
    * @param   owner
    *          The widget owning created system menu.
    * @param   textEmpty
    *          <code>true</code> editor field selection empty.
    *   
    * @return   A popup widget.
    */
   public static EditorPopupGuiImpl create(Widget<?> owner, boolean textEmpty)
   {
      final EditorPopupGuiImpl menu = new EditorPopupGuiImpl(owner);
      menu.initialize(WidgetId.nextID(), menu.config());
      
      // Creating items
      
      MenuItemGuiImpl undo = MenuItemGuiImpl.createDynamicItem("Undo", false, new MouseAdapter()
      {
         @Override
         public void mouseClicked(MouseEvent e)
         {
            Widget<?> owner = (Widget<?>) 
                     menu.screen().getRegistry().getComponent(menu.config().ownerId);
            
            EventManager.postEvent(new KeyInput(owner, EventType.KEY_PRESSED, Key.VK_CTRL_Z));
         }
      });
      
      MenuItemGuiImpl rule1 = MenuItemGuiImpl.createDynamicRule();

      menu.cutOpt = MenuItemGuiImpl.createDynamicItem("Cut", !textEmpty, new MouseAdapter()
      {
         @Override
         public void mouseClicked(MouseEvent e)
         {
            Widget<?> owner = (Widget<?>) 
                     menu.screen().getRegistry().getComponent(menu.config().ownerId);
            
            EventManager.postEvent(new KeyInput(owner, EventType.KEY_PRESSED, Keyboard.KA_CUT));
         }
      });
      int menuId = menu.cutOpt.getId().asInt();
      MouseHandler.registerWidgetAction(
               menuId,
               new SystemAction(menuId, MouseEvt.CUT, null));

      menu.copyOpt = MenuItemGuiImpl.createDynamicItem("Copy", !textEmpty, new MouseAdapter()
      {
         @Override
         public void mouseClicked(MouseEvent e)
         {
            Widget<?> owner = (Widget<?>) 
                     menu.screen().getRegistry().getComponent(menu.config().ownerId);
            
            EventManager.postEvent(new KeyInput(owner, EventType.KEY_PRESSED, Keyboard.KA_COPY));
         }
      });
      menuId = menu.copyOpt.getId().asInt();
      MouseHandler.registerWidgetAction(
               menuId,
               new SystemAction(menuId, MouseEvt.COPY, null));
      
      menu.pasteOpt = MenuItemGuiImpl.createDynamicItem("Paste", false, new MouseAdapter()
      {
         @Override
         public void mouseClicked(MouseEvent e)
         {
            Widget<?> owner = (Widget<?>) 
                     menu.screen().getRegistry().getComponent(menu.config().ownerId);
            
            EventManager.postEvent(new KeyInput(owner, EventType.KEY_PRESSED, Keyboard.KA_PASTE));
         }
      });
      menuId = menu.pasteOpt.getId().asInt();
      MouseHandler.registerWidgetAction(
               menuId,
               new SystemAction(menuId, MouseEvt.PASTE, null));

      MenuItemGuiImpl rule2 = MenuItemGuiImpl.createDynamicRule();
      
      menu.selectAllOpt = MenuItemGuiImpl.createDynamicItem("Select All", textEmpty, new MouseAdapter()
      {
         @Override
         public void mouseClicked(MouseEvent e)
         {
            Widget<?> owner = (Widget<?>) 
                     menu.screen().getRegistry().getComponent(menu.config().ownerId);
            
            // TODO make an event
            if (owner instanceof EditorGuiImpl)
            {
               EditorGuiImpl editor = (EditorGuiImpl) owner;
               String text = editor.getText();
               editor.setSelection(1, text.length() + 1);
            }
            else if (owner instanceof FillInGuiImpl)
            {
               FillInGuiImpl fillIn = (FillInGuiImpl) owner;
               fillIn.config().selectionStart = 0;
               String txt = fillIn.getText(true);
               fillIn.config().selectionEnd = txt.length();
               
               // To draw selection FillIn must be have focus.
               fillIn.requestFocus();
            }
         }
      });
      
      menu.add(undo);
      menu.add(rule1);
      menu.add(menu.cutOpt);
      menu.add(menu.copyOpt);
      menu.add(menu.pasteOpt);
      menu.add(rule2);
      menu.add(menu.selectAllOpt);
      
      menu.initEditorItems(textEmpty);
      
      return menu;
   }
   
   /**
    * Initialized items of editor menu.
    * 
    * @param   textEmpty
    *          <code>true</code> editor field selection empty.
    */
   private void initEditorItems(boolean textEmpty)
   {
      cutOpt.setEnabled(!textEmpty);
      copyOpt.setEnabled(!textEmpty);
      
      String clipboardContents = ((GuiDriver) OutputManager.getDriver()).clipboardContents();
      pasteOpt.setEnabled(clipboardContents != null && !clipboardContents.isEmpty());
      
      selectAllOpt.setEnabled(textEmpty);
   }
}