/*
** Module   : FileDialogGuiImpl.java
** Abstract : a dialog that allows the user to select a file form local disk in GUI mode.
**
** Copyright (c) 2011-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20170815 Initial commit. Just a stub with mandatory methods. Work in progress. 
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * This class represents a dialog that allows the user to select a file form local disk, in GUI
 * client. The dialog is composed from six rows that expand horizontally. The first two and last
 * three are fixed in height, the third can expand vertically if the dialog is resized. The
 * content of each row (pane) is individually laid out by specialized inner classes.
 * <p>
 * All dimensions are expressed in pixels, rather than character units.
 */
public class FileDialogGuiImpl
extends FileDialogImpl 
{
   /** Constant height of the first (upper) pane (navigation). */
   private final int NAV_PANE_HEIGHT = 37;

   /** Constant height of the second pane (tools). */
   private final int TOOLS_PANE_HEIGHT = 31;

   /** Constant height of the fourth pane (filename direct input). */
   private final int FILENAME_PANE_HEIGHT = 34;

   /** Constant height of the fourth pane (file type selection). */
   private final int SAVEAS_PANE_HEIGHT = 31;

   /** Constant height of the last pane (dialog closing pane). */
   private final int BUTTONS_PANE_HEIGHT = 41;

   /** The bottom inset. */
   private final int BOTTOM_INSET = 11;
   
   /** The alignment for file and type labels (on last two rows). */
   private final int FILE_TYPE_ALIGN = 125;
   
   /** Cached driver. */
   protected GuiDriver gd = null;
   
   /** Cached coordinates. */
   private CoordinatesConversion cc;
   
   /** The navigation pane. */
   private NavigationPane navigationPane = null;
   private ButtonGuiImpl back;
   private ButtonGuiImpl forward;
   private ButtonGuiImpl recentLocations;
   private ButtonGuiImpl folderUp;
   private FillInGuiImpl address;
   private ButtonGuiImpl refresh;
   private FillInGuiImpl search;
   
   /** The tools pane. */
   private ToolsPane toolsPane = null;
   private ButtonGuiImpl organize;
   private ButtonGuiImpl newFolder;
   private ButtonGuiImpl view;
   private ButtonGuiImpl help;
   
   /** The content pane. */
   private ContentPane contentPane = null;
   
   /** The filename input pane. */
   private FilenamePane filenamePane = null;
   private LabelGuiImpl filenameLabel;
   private ComboBoxGuiImpl filename;
   
   /** The file type selection pane (if available). */
   private SaveasPane saveasPane = null;
   private LabelGuiImpl saveasLabel;
   private ComboBoxGuiImpl saveasType;
   
   /** The dialog close pane. */
   private ButtonsPane buttonsPane = null;
   private ButtonGuiImpl hideFolders;
   private ButtonGuiImpl accept;
   private ButtonGuiImpl reject;
   
   /**
    * The constructor build an object by forwarding the required parameters to super class. 
    *
    * @param   title
    *          The title of the dialog, if known.
    * @param   parentWindow
    *          The window id of the parent frame of this dialog.
    */
   public FileDialogGuiImpl(String title, int parentWindow)
   {
      super(title, parentWindow);
   }
   
   /**
    * Initialize the dialog by performing tasks that was not possible in the constructor because
    * it was not fully created.
    */
   protected void init()
   {
      if (gd != null)
      {
         return; // already initialized
      }
      
      gd = (GuiDriver<?,?>) OutputManager.getDriver();
      cc = screen().coordinates();
      realize();
      
      // minim size:
      setSize(cc.dimensionFromNative(new NativeDimension(625, 472)));
      
      Container<GuiOutputManager> content = getContentPane();
      
      navigationPane = new NavigationPane();
      content.add(navigationPane);
      back = createButton("<-", navigationPane, 24, 24, true, false);
      navigationPane.add(back);
      forward = createButton("->", navigationPane, 24, 24, true, false);
      navigationPane.add(forward);
      recentLocations = createButton("v", navigationPane, 24, 24, true, false);
      navigationPane.add(recentLocations);
      folderUp = createButton("^", navigationPane, 24, 24, true, false);
      navigationPane.add(folderUp);
         WidgetId addressId = WidgetId.nextID();
         FillInConfig addressCfg = new FillInConfig();
      address = new FillInGuiImpl(addressId, addressCfg);
      address.initialize(addressId, addressCfg);
      navigationPane.add(address);
      refresh = createButton("r", navigationPane, 22, 22, true, false);
      navigationPane.add(refresh);
         WidgetId searchId = WidgetId.nextID();
         FillInConfig searchCfg = new FillInConfig();
      search = new FillInGuiImpl(searchId, searchCfg);
      search.initialize(searchId, searchCfg);
      navigationPane.add(search);
      
      toolsPane = new ToolsPane();
      content.add(toolsPane);
      organize = createButton("Organize", toolsPane, 91, 26, true, false);
      toolsPane.add(organize);
      newFolder = createButton("New Folder", toolsPane, 88, 26, true, false);
      toolsPane.add(newFolder);
      view = createButton("[-] v", toolsPane, 88, 26, false, false);
      toolsPane.add(view);
      help = createButton("(?)", toolsPane, 88, 26, true, false);
      toolsPane.add(help);
      
      contentPane = new ContentPane();
      content.add(contentPane);
      // TODO: add this pane content
      
      WidgetId frameId = WidgetId.nextID();
      FrameConfig frameConfig = new FrameConfig();
      filenamePane = new FilenamePane(frameId, frameConfig);
      content.add(filenamePane);
      filenameLabel = new LabelGuiImpl(WidgetId.nextID(), "File &name:", frameId.asInt());
      filenamePane.add(filenameLabel);
      filename = new ComboBoxGuiImpl(WidgetId.nextID(), new ComboBoxConfig());
      filenamePane.add(filename);
      
      frameId = WidgetId.nextID();     // reuse var
      frameConfig = new FrameConfig(); // reuse var
      saveasPane = new SaveasPane(frameId, frameConfig);
      content.add(saveasPane);
      saveasLabel = new LabelGuiImpl(WidgetId.nextID(), "Save as &type:", frameId.asInt());
      saveasPane.add(saveasLabel);
      saveasType = new ComboBoxGuiImpl(WidgetId.nextID(), new ComboBoxConfig());
      saveasPane.add(saveasType);
      
      buttonsPane = new ButtonsPane();
      content.add(buttonsPane);
      hideFolders = createButton("Hide Folders", buttonsPane, 92, 21, true, false);
      buttonsPane.add(hideFolders);
      accept = createButton("OK", buttonsPane, 86, 24, false, true);
      buttonsPane.add(accept);
      reject = createButton("Cancel", buttonsPane, 86, 24, false, false);
      buttonsPane.add(reject);
   }
   
   /**
    * Get widget width.
    *
    * @return  widget width.
    */
   @Override
   public double width()
   {
      double supW = super.width();
      double minW = cc.widthFromNative(609);
      return Math.max(supW, minW);
   }
   
   /**
    * Updates the layout of this dialog. Calls {@code doLayout()} on all panes/rows.
    */
   @Override
   public void doLayout()
   {
      double thisWidth = width();
      double thisHeight = height();
      int natHeight = cc.heightToNative(thisHeight);
      int natWidth = cc.widthToNative(thisWidth);
      
      setWidgetLocation(navigationPane, 0, 0);
      setWidgetLocation(toolsPane, 0, NAV_PANE_HEIGHT);
      setWidgetLocation(contentPane, 0, NAV_PANE_HEIGHT + TOOLS_PANE_HEIGHT); // expanding vertically
      setWidgetLocation(filenamePane, 0, natHeight - 
            (BOTTOM_INSET + BUTTONS_PANE_HEIGHT + SAVEAS_PANE_HEIGHT + FILENAME_PANE_HEIGHT));
      setWidgetLocation(saveasPane, 0, natHeight - 
            (BOTTOM_INSET + BUTTONS_PANE_HEIGHT + SAVEAS_PANE_HEIGHT));
      setWidgetLocation(buttonsPane, 0, natHeight - (BOTTOM_INSET + BUTTONS_PANE_HEIGHT));
      // 11 pixels in bottom inset (paints the 'resize' grip in the right corner)
   
      // [contentPane] expands so it occupy all remaining vertical space:
      contentPane.setHeight(thisHeight - cc.heightFromNative(
            NAV_PANE_HEIGHT + TOOLS_PANE_HEIGHT + FILENAME_PANE_HEIGHT + 
            SAVEAS_PANE_HEIGHT + BUTTONS_PANE_HEIGHT + BOTTOM_INSET));
      
      // all panels are dynamically resizing to fit all available width: 
      navigationPane.setWidth(thisWidth);
      toolsPane.setWidth(thisWidth);
      contentPane.setWidth(thisWidth);
      filenamePane.setWidth(thisWidth);
      saveasPane.setWidth(thisWidth);
      buttonsPane.setWidth(thisWidth);
   
      navigationPane.doLayout();
      toolsPane.doLayout();
      contentPane.doLayout();
      filenamePane.doLayout();
      saveasPane.doLayout();
      buttonsPane.doLayout();
   }
   
   /**
    * This method is invoked when action event is generated.
    *
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void onAction(ActionEvent event)
   {
      
   }
   
   /**
    * Contains the following widgets:
    *    NAME_____________ TYPE_______ PROPERTIES____ ACTION______________________________________
    *    HistoryBack       button      round          Loads the previous folder
    *    HistoryForward    button      round          Loads the forward folder 
    *    RecentLocations   button                     Opens a popup-menu with recent locations
    *    FolderUp          button                     Loads the parent folder
    *    Address           editor      expandable     Allows user to type in a location
    *    Refresh           button                     Reload the current location
    *                      /splitter/                 Changes ratio between Address and Search
    *    Search            editor      expandable     Searches sub-folders for specific pattern
    */
   private class NavigationPane
   extends AbstractContainer<GuiOutputManager>
   {
      /**
       * Obtain the constant height of this pane.
       * 
       * @return  the constant height of this pane.
       */
      @Override
      public double height()
      {
         return cc.heightFromNative(NAV_PANE_HEIGHT);
      }
   
      /**
       * Lays out this pane/row.
       */
      @Override
      public void doLayout()
      {
         int natWidth = cc.widthToNative(this.width());
      
         setWidgetLocation(back, 10, 6);
         setWidgetLocation(forward, 10 + 24, 6);
         setWidgetLocation(recentLocations, 10 + 24 +24, 6);
         setWidgetLocation(folderUp, 10 + 24 + 24 + 24,  6);
         setWidgetLocation(address, 10 + 24 + 24 + 24 + 24, 6);
         setWidgetLocation(search, natWidth - 216, 6);
      }
   }
   
   /**
    * Contains the following widgets:
    *    NAME_____________ TYPE_______ PROPERTIES____ ACTION______________________________________
    *    Organize          button                     Opens a contextual popup-menu
    *    NewFolder         button                     Creates a new folder at current location
    *                      /spacer/    expandable     Nothing (fills in empty space)
    *    MoreOptions       button                     Opens a popup-menu with different view modes
    *    Help              button                     Opens Help page
    */
   private class ToolsPane
   extends AbstractContainer<GuiOutputManager>
   {
      /**
       * Obtain the constant height of this pane.
       *
       * @return  the constant height of this pane.
       */
      @Override
      public double height()
      {
         return cc.heightFromNative(TOOLS_PANE_HEIGHT);
      }
      
      /**
       * Lays out this pane/row.
       */
      @Override
      public void doLayout()
      {
         int natWidth = cc.widthToNative(this.width());
      
         setWidgetLocation(organize, 3, 2);
         setWidgetLocation(newFolder, 94, 2);
         setWidgetLocation(view, natWidth - (10 + 26 + 26 + 26), 2);
         setWidgetLocation(help, natWidth - (10 + 26), 2);
      }
   }
   
   /**
    * Contains the following widgets:
    *    NAME_____________ TYPE_______ PROPERTIES____ ACTION______________________________________
    *    FoldersTree       treeview                   Contains 3 trees: Favorites/This PC/Network
    *                      /splitter/                 Changes ratio between Folders and Contents
    *    FolderContent     browse/list                List the files and folder at current location
    */
   private class ContentPane
   extends AbstractContainer<GuiOutputManager>
   {
   }
   
   /**
    * Contains the following widgets:
    *    NAME_____________ TYPE_______ PROPERTIES____ ACTION______________________________________
    *    FileName          label       fixed          Nothing (right-aligned)
    *    FileNameCb        combobox    expandable     Allows user to type in a filename 
    */
   private class FilenamePane
   extends FrameGuiImpl
   {
      /**
       * Special constructor used to restore component using given configuration.
       * <p>
       * Call {@link #initialize} to finish the widget initialization.
       *
       * @param   id
       *          The widget's ID.
       * @param   cfg
       *          Configuration data.
       */
      public FilenamePane(WidgetId id, FrameConfig cfg)
      {
         super(id, cfg);
         initialize(id, cfg);
         screen().getRegistry().addWidget(this);
      }
   
      /**
       * Obtain the constant height of this pane.
       *
       * @return  the constant height of this pane.
       */
      @Override
      public double height()
      {
         return cc.heightFromNative(FILENAME_PANE_HEIGHT);
      }
      
      /**
       * Lays out this pane/row.
       */
      @Override
      public void doLayout()
      {
         int natWidth = cc.widthToNative(this.width());
      
         setWidgetLocation(filenameLabel, 10, 10);
         setWidgetLocation(filename, FILE_TYPE_ALIGN, 10);
         filenameLabel.resizeTo(new NativePoint(10, 10), 
                                new NativeDimension(FILE_TYPE_ALIGN - 10, 24), 
                                false);
         filename.resizeTo(new NativePoint(FILE_TYPE_ALIGN, 10), 
                           new NativeDimension(natWidth - FILE_TYPE_ALIGN - 5, 24), 
                           false);
      }
   }
   
   /**
    * Contains the following widgets:
    *    NAME_____________ TYPE_______ PROPERTIES____ ACTION______________________________________
    *    SaveType          label       fixed          Nothing (right-aligned)
    *    SaveAsCb          combobox    expandable     Allows user to select a filter/type
    */
   private class SaveasPane
   extends FrameGuiImpl
   {
      /**
       * Special constructor used to restore component using given configuration.
       * <p>
       * Call {@link #initialize} to finish the widget initialization.
       *
       * @param   id
       *          The widget's ID.
       * @param   cfg
       *          Configuration data.
       */
      public SaveasPane(WidgetId id, FrameConfig cfg)
      {
         super(id, cfg);
         initialize(id, cfg);
         screen().getRegistry().addWidget(this);
      }
   
      /**
       * Obtain the constant height of this pane.
       *
       * @return  the constant height of this pane.
       */
      @Override
      public double height()
      {
         return cc.heightFromNative(SAVEAS_PANE_HEIGHT);
      }
      
      /**
       * Lays out this pane/row.
       */
      @Override
      public void doLayout()
      {
         int natWidth = cc.widthToNative(this.width());
      
         setWidgetLocation(saveasLabel, 10, 1);
         setWidgetLocation(saveasType, FILE_TYPE_ALIGN, 1);
         saveasLabel.resizeTo(new NativePoint(10, 1),
                              new NativeDimension(FILE_TYPE_ALIGN - 10, 24),
                              false);
         saveasType.resizeTo(new NativePoint(FILE_TYPE_ALIGN, 1),
                             new NativeDimension(natWidth - FILE_TYPE_ALIGN - 5, 24),
                             false);
      }
   }
   
   /**
    * Contains the following widgets:
    *    NAME_____________ TYPE_______ PROPERTIES____ ACTION______________________________________
    *    HideFolders       button                     Toggles visibility of Tools and Content panes
    *                      /spacer/    expandable     Nothing (fills in empty space)
    *    OK                button                     Accepts the current selection
    *    Cancel            button                     Rejects the dialog
    */
   private class ButtonsPane
   extends AbstractContainer<GuiOutputManager>
   {
      /**
       * Obtain the constant height of this pane.
       *
       * @return  the constant height of this pane.
       */
      @Override
      public double height()
      {
         return cc.heightFromNative(BUTTONS_PANE_HEIGHT);
      }
   
      /**
       * Lays out this pane/row.
       */
      @Override
      public void doLayout()
      {
         int natWidth = cc.widthToNative(this.width());
         
         setWidgetLocation(hideFolders, 11, 20);
         setWidgetLocation(accept, natWidth - 204, 16);
         setWidgetLocation(reject, natWidth - 104, 16);
      }
   }
   
   /**
    * Creates and initializes a button widget.
    *
    * @return   A reference to {@link Button} instance.
    */
   private ButtonGuiImpl createButton(String text,
                                      Container container,
                                      int width,
                                      int height,
                                      boolean flat,
                                      boolean def)
   {
      ButtonGuiImpl b = (ButtonGuiImpl) widgetFactory.createButton(WidgetId.nextID(),
                                                                   container,
                                                                   text);
      b.initialize(b.getId(), null);
      b.setSupportsMnemonic();
      b.setTextLabel(text);
      b.setCommand(text);
      b.setEnabled(true);
      b.addListener(this);
      
      ButtonConfig btnConfig = b.config();
      btnConfig.widthPixels = width;
      btnConfig.heightPixels = height;
      btnConfig.isDefault = def;
      btnConfig.flatButton = flat;
      btnConfig.fixedHeight = true;
      btnConfig.fixedWidth = true;
      
      b.resizeTo(new NativePoint(0, 0), new NativeDimension(width, height), false);
      return b;
   }
   
   private void setWidgetLocation(AbstractWidget w, int x, int y) 
   {
      w.setLocation(cc.columnFromNative(x), cc.rowFromNative(y));
      w.setPhysicalLocation(x, y);
   }
}
