/* 
** Module   : GuiWindowCommons.java
** Abstract : The class with common logic for GUI windows.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20151016 Initial version.
** 002 HC  20151020 Added 3D border to ALERT-BOX and DIALOG-BOX.
** 003 HC  20151106 Restoring owner's focus when modal window closed. 
** 004 HC  20151213 Improved DIALOG-BOX window parenting logic and related changes.
** 005 CA  20160129 When drawing the border, check if the invalidation rectangles registered in
**                  ScreenBitmap intersect it or not. 
** 006 HC  20160406 Overhaul of window-activation logic. Fixed the algorithm to determine the
**                  next active main window after another window is hidden.
** 007 HC  20160413 Fixed focusing of main window when overlay window closes.
** 008 HC  20160428 GUI window border drawing changed to use the Border interface mechanism. Also
**                  removed hardcoded assumptions of the border size.
** 009 HC  20160502 Fixed selection of next active window when modal window with no visible owner
**                  is being hidden.
** 010 HC  20160628 Prevented a modal window to be moved higher in the z-order over another
**                  active modal window.
** 011 SBI 20160630 Applied HC changes to fix a next active window, 010 changes are not required,
**                  thus they are rolled back.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import java.util.*;
import java.util.function.*;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * The class holds logic common to all GUI windows - types implementing
 * the interface {@link GuiWindow}.
 */
public class GuiWindowCommons
{
   /**
    * Hide the supplied window.
    * 
    * @param   window
    *          A reference to a {@link GuiWindow} instance, 
    *          must not be <code>null</code>.  
    */
   public static void hide(GuiWindow window)
   {
      if (WindowManager.isWindowActive((TopLevelWindow<?>) window))
      {
         TopLevelWindow<?> nextActive = null;
         
         if (window instanceof WindowGuiImpl)
         {
            // ENTRY/LEAVE events are not sent; the new top is computed as this:
            nextActive = WindowManager.getNextActiveWindow();
         }
         else if (window instanceof ModalWindow)
         {
            // honor modality - find a visible modal window just under 
            // the window being made invisible
            java.util.stream.Stream<TitledWindow<?>> s = 
                     Utils.reverseStream(WindowManager.windowList(window.getId().asInt()));
            
            nextActive = s.filter(w -> w instanceof TopLevelWindow)
                          .map(w -> (TopLevelWindow<?>) w)
                          .filter(w -> w.isVisible() && w.isModal())
                          .findFirst()
                          .orElse(null);
            
            // if no visible modal window was found, use the owner as the next active
            if (nextActive == null)
            {
               // hiding the owner window should also hide its owned modal
               // window so the owner should be always visible at this point 
               nextActive = ((ModalWindow) window).getOwner();
            }
            
            // no other modal window is visible or there is no visible owner of this modal?
            if (nextActive == null || !nextActive.isVisible())
            {
               nextActive = WindowManager.getNextActiveWindow();
            }
            
            WindowManager.setFocusWindow(nextActive);
         }
         
         if (nextActive != null)
         {
            // activation state must be changed before gd.setWindowVisible(false)
            // otherwise Swing doesn't send focus-lost/focus-gained events! 
            WindowManager.activateWindow(nextActive);
         }
      }
      
      GuiDriver<?, ?> gd = (GuiDriver<?, ?>) window.screen().getInstanceDriver();

      List<TitledWindow<?>> children = WindowManager.findChildren(window.getId().asInt(), false);
      for (TitledWindow<?> w : children)
      {
         w.hide();
         
         // ensure the driver is notified - as only the root parent is actually painted by P2J
         // event drawing bracket
         gd.selectWindow(w.config().id.asInt());
         gd.setWindowVisible(false);
         gd.releaseWindow();
      }
      
      // hide all visible dialog boxes owned by this main window
      if (window instanceof WindowGuiImpl)
      {
         int wndId = window.getId().asInt();
         List<DialogBoxWindow> dialogs = WindowManager.getDialogsOwnedBy(wndId, false);
         for (DialogBoxWindow d : dialogs)
         {
            gd.selectWindow(d.getId().asInt());
            d._setVisible(false);
            gd.setWindowVisible(false);
            gd.releaseWindow();
            
            // set the visibility flag on the frame so that it is reflected on the server
            d.getFrame()._setVisible(false);
         }
         
         if (!dialogs.isEmpty())
         {
            WindowManager.demodalizeWindow(dialogs.get(dialogs.size() - 1));
         }
      }
      
      // hide us now
      gd.selectWindow(window.getId().asInt());
      gd.setWindowVisible(false);
      gd.releaseWindow();
   }
   
   /**
    * Check if the specified point is inside the window's top, left, bottom or right border.
    * 
    * @param    window
    *           The window to check.
    * @param    point
    *           The point to check.
    * 
    * @return   <code>true</code> if the point is inside the window's border.
    */
   public static boolean insideBorder(GuiWindow window, NativePoint point)
   {
      Border<?> borderRef = window.getBorder();
      if (borderRef == null)
      {
         return false;
      }
      
      Function<NativeRectangle, Boolean> checkBorder = (nrect) -> (nrect.contains(point));
      
      NativeInsets border = window.screen().coordinates().insetsToNative(borderRef.insets());
      NativeDimension d = window.physicalDimension();

      NativeRectangle topBorder    = new NativeRectangle(0, 0,
                                                         border.top - 1, d.width - 1);
      NativeRectangle bottomBorder = new NativeRectangle(d.height - border.bottom, 0, 
                                                         d.height - 1, d.width - 1);
      NativeRectangle leftBorder   = new NativeRectangle(0, 0,
                                                         d.height - 1, border.left - 1);
      NativeRectangle rightBorder  = new NativeRectangle(0, d.width - border.right, 
                                                         d.height - 1, d.width - 1);
      
      if (checkBorder.apply(topBorder)    ||
          checkBorder.apply(bottomBorder) ||
          checkBorder.apply(leftBorder)   ||
          checkBorder.apply(rightBorder))
      {
         return true;
      }
      
      return false;
   }
}