/*
** Module   : LineBorderGuiImpl.java
** Abstract : Line border GUI implementation.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20150831 Created initial version.
** 002 HC  20150922 Fixed border drawing.
** 003 OM  20170320 Added support for Themes.
** 004 OM  20170320 Added support for different rendering in Windows 8/10 Themes.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Simple GUI line border. 
 * <p>
 * This border implementation will fill the whole border area with the background color
 * {@link SysColor#COLOR_WINDOW} or the background color supplied in the
 * config instance passed to the constructor.
 */
public class LineBorderGuiImpl
implements Border<GuiOutputManager> 
{
   /** Insets. **/
   private Insets insets = null;
   
   /** Native insets. */
   private NativeInsets ninsets = null;
   
   /** Cached driver. */
   private GuiDriver<?, ?> gd;
   
   /** Color resolver. */
   private GuiColorResolver gc;
   
   /** Coordinate conversion reference. */
   private CoordinatesConversion c;
   
   /** The legacy widget of this line border. */
   private AbstractWidget legacyWidget = null;
   
   /**
    * Default constructor.
    */
   public LineBorderGuiImpl()
   {
      this(null, null, null);
   }

   /**
    * Creates a new instance with a specific insets and color resolver.
    * <p>
    * The resolver is used to resolve colors for the border line and for the border background, 
    * in particular {@link GuiColorResolver#fgColor} and {@link GuiColorResolver#bgColor}
    * are used. If {@code null} is given then default foreground/background colors will be used, 
    * {@link SysColor#COLOR_WINDOW} and {@link SysColor#COLOR_WINDOWTEXT}.
    *  
    * @param    insets
    *           The parameter defines the border thickness. May be {@code null}.
    * @param    resolver
    *           Color resolver instance, see above. May be {@code null}.
    * @param    legacyWidget
    *           The legacy widget of this border. On this flag depends how it will be painted.
    */
   public LineBorderGuiImpl(Insets insets, GuiColorResolver resolver, AbstractWidget legacyWidget)
   {
      GuiOutputManager o = (GuiOutputManager) OutputManager.instance();
      gd = (GuiDriver<?, ?>) o.getInstanceDriver();
      c = o.coordinates();
      this.legacyWidget = legacyWidget;
      
      if (resolver == null)
      {
         // GuiColorResolver doesn't accept null config
         gc = new GuiColorResolver(gd,
                                   new BaseConfig(),
                                   SysColor.COLOR_WINDOW,
                                   SysColor.COLOR_WINDOWTEXT);
      }
      else
      {
         gc = resolver;
      }
      
      if (insets == null)
      {
         Point p = c.baseUnits();
         insets = new Insets(p.y, p.x, p.y, p.x);
      }
      
      this.insets = new Insets(insets.top, insets.left, insets.bottom, insets.right);
      this.ninsets = new NativeInsets(c.rowToNative(insets.top), 
                                      c.columnToNative(insets.left), 
                                      c.rowToNative(insets.bottom), 
                                      c.columnToNative(insets.right));
   }
   
   /**
    * Get insets introduced by the border.
    * 
    * @return border insets.
    */
   @Override
   public Insets insets()
   {
      return new Insets(insets.top, insets.left, insets.bottom, insets.right);
   }
   
   /**
    * Get native insets introduced by the border.
    * 
    * @return  border insets in native units.
    */
   public NativeInsets getNativeInsets()
   {
      return new NativeInsets(ninsets.top, ninsets.left, ninsets.bottom, ninsets.right);
   }
   
   /**
    * Draw border.
    * 
    * @param   widget
    *          Base widget. 
    * @param   size
    *          Border dimensions.
    */
   @Override
   public void drawBorder(Widget<GuiOutputManager> widget, Dimension size)
   {
      gc.colors();
      NativeDimension nsize = c.dimensionToNative(size);
      
      ThemeManager.getCurrentTheme().drawLineBorder(
            this, gd, gc, 0, 0, nsize.width, nsize.height);
   }
   
   /**
    * Obtain the legacy widget for this line border.
    *
    * @return  The legacy widget of the border.
    */
   public AbstractWidget getLegacyWidget()
   {
      return legacyWidget;
   }
}
