/*
** Module   : RectangleGuiImpl.java
** Abstract : GUI rectangle implementation
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20141114 Initial version.
** 002 EVL 20141120 Fixed origin coords inside painting worker.  Changed double line shaded
**                  rectangle drawing approach.  Also when GROUP-BOX attribute is true the filling
**                  become ignored.
** 003 CA  20141205 Avoid unnecessary conversion, as the physical location/dimension can be 
**                  determined from existing APIs.
** 004 EVL 20150409 Turning on clipping tool.
** 005 HC  20150323 Removed business logic from config classes.
** 006 GES 20150425 Eliminated AWT dependencies.
** 007 IAS 20150926 MOVE-TO-TOP()/MOVE-TO_BUTTON() support.
** 008 EVL 20151030 Fix for 1 pixel size issue and incorrect filled rounded rectangle draw.  Fix
**                  for wrong mode with TREE-D defined owning frame.  Added border size capability
**                  for 3D shaded rectangle.
** 009 CA  20160216 Cache the intermediate widget states.
** 010 EVL 20160420 Fix for 3D shaded shape and size, not a simple two rectangles.
** 011 CA  20170130 Fixed border drawing when EDGE-PIXELS > 1.
** 012 OM  20170406 Added support for Themes.
** 013 EVL 20170629 Enable mouse events for widget to be selected with direct manipulation.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Specific GUI implementation of RECTANGLE widget.
 */
public class RectangleGuiImpl
extends Rect<GuiOutputManager>
{
   /** Screen driver. */
   private GuiDriver gd;

   /** Color resolver. */
   private GuiColorResolver gc;
   
   /**
    * This constructor creates a rectangle.
    * 
    * @param    id
    *           The widget's ID.
    * @param    owner
    *           The parent container, frame which will include this
    *           rectangle.
    */
   RectangleGuiImpl(WidgetId id, Container<GuiOutputManager> owner) 
   {
      super(id, owner);
   }

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public RectangleGuiImpl(WidgetId id, RectangleConfig cfg)
   {
      super(id, cfg);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    wcfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, RectangleConfig wcfg)
   {
      super.initialize(id, wcfg);
      
      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
   }
   
   /**
    * Draw the rectangle. Called by this rectangle's parent container.
    */
   @Override
   public void draw()
   {
      // rectangle options need to be refreshed before every painting
      RectangleConfig config = config();
      gc = new GuiColorResolver(gd, config, SysColor.COLOR_WINDOW, SysColor.COLOR_WINDOWTEXT);
      
      // determine colors to use
      gc.colors();
      
      // get location and size
      NativePoint p = physicalLocation();
      NativeDimension d = physicalDimension();
      
      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, new NativeDimension(d.width, d.height));
      
      // clippable drawing start 
      gd.draw(getId(), p, clip, () ->
            ThemeManager.getCurrentTheme().drawRectangle(this, gd, gc, 0, 0, d.width, d.height));
   }

   /**
    * Get the Z-Order class
    * 
    * @return Z-Order class of the widget
    */
   @Override
   public ZOrderClass getZOrderClass()
   {
      return ZOrderClass.DECORATION;
   }

   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvent.MOUSE_CLICKED, 
         MouseEvent.MOUSE_PRESSED, 
         MouseEvent.MOUSE_RELEASED
      };
   }
}
