/*
** Module   : ScrollBarGuiImpl.java
** Abstract : scroll bar implementation. 
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140917 First version.
** 002 MAG 20141013 Fixed system colors. Handle mouse events.
** 003 HC  20141021 Implemented window move/position support.
** 004 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 005 CA  20141123 Fixed scrollbar button dimensions.
** 006 CA  20150323 The GuiFontResolver needs to know if the expected font is for a fixed-width
**                  widget.
** 007 CA  20150509 Changed to allow double-value step for scrolling.
** 008 GES 20150425 Eliminated AWT dependencies.
** 009 EVL 20150511 Consider scrollpane box when computing the scrollbar coordinates.  Scrollbar
**                  thumb size/position changing approach.  Adding support for disabled state.
** 010 CA  20150617 Allow drawing of scrollbars without a thumb.
** 011 EVL 20150724 Fix for scrollbar position in case of the borderless parent.  Coord is 0
**                  based.  Using only dimension to compute the coordinate gives 1 based coord.
**                  So we have to subtract 1 pixel to get correct coordinate.  One pixel shift for
**                  right scroll button and borderless parent.
** 012 GES 20150730 Modified GUI driver translate() usage to match API changes.
** 013 HC  20150831 Scrolling support overhaul.
** 014 CA  20150910 Added mouse actions. Removed unnecessary call to setLocation.
** 015 CA  20150925 Fixed scrollbar dimension when attached to a bordered panel.
**                  Scrollbar size no longer depends on font size, it is a constant predefined 
**                  value. Layout fixes when insets are involved.
** 016 SVL 20150828 Javadoc fix.
** 017 EVL 20151010 Fix for thumb drawing issues.  Fix for scrolling via thumb dragging. 
**     SVL 20151019 Added ignorePanelSize and sendNonTrimmedEvents.
** 018 SVL 20151029 Added support for scroll bar controllers.
**     EVL 20151111 Fix for scrollbar thumb size calculation. The minimum size is 8.  When nothing
**                  to scroll - it is not required to draw the thumb.  Fix for background drawing.
** 019 EVL 20151125 Adding executor shutdown to prevent resource leak.  Changing static executor
**                  to be instance based.
** 020 CA  20160328 Do not raise legacy mouse events for scroll bar.
** 021 SBI 20151223 Changed createAutoScrollEvent() to create only valid auto scroll events.
**         20160216 Cache the intermediate widget states.
** 022 VIG 20160127 Added border cases check to createAutoScrollEvent() (HC change).
**                  Added isRMBEntrySupported() and getOwningWidget() methods.
**                  Changed onButtonClick(), system scroll popup support, changed showPopupMenu().
** 023 VIG 20160417 Fixed thumb drug. Added support of isScrollPopupSupported() method to ignore
**                  scroll popup for specific widgets.
** 024 EVL 20160419 Sometimes we need more effects to implement for 3d mode.
** 025 HC  20160607 GUI menus reworked to be displayed in dedicated top-level windows.
** 026 SVL 20160719 Updated translate() signature.
** 027 OM  20170320 Added support for Themes.
** 018 OM  20170407 Fixed height of scrollbars (left/top insets were substracted twice).
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.*;
import java.util.concurrent.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * A GUI scroll bar implementation. 
 * TODO 
 * REPEAT_MILLIS constant is used to simulate mouse press repeated events. 
 * This parameter should be configurable for different customers and users.
 */
@SuppressWarnings("unchecked")
public class ScrollBarGuiImpl 
extends ScrollBar<GuiOutputManager> 
implements ScrollListener
{
   /** Executor for handling mouse input. */
   private ExecutorService mouseInputExecutor = Executors.newSingleThreadExecutor();

   /** Screen driver */
   private GuiDriver gd;

   /** Color resolver */
   private GuiColorResolver gc;
   
   /** The related viewport */
   private Viewport<GuiOutputManager> viewport;

   /** Left button */
   private ScrollBarGuiButton lButt;

   /** Right button */
   private ScrollBarGuiButton rButt;

   /** Thumb button position */
   private NativePoint thumbPos;

   /** Thumb button size */
   private NativeDimension thumbDim;

   /** Width correction (right part) */
   double wOff = 0;

   /** Height correction (bottom part) */
   double hOff = 0;

   /** 
    * Mouse pressed indicator. 
    */
   volatile boolean pressed = false;
   
   /** 
    * Mouse pressed indicator for thumb. If {@code true} then the mouse clicked the thumb and
    * probably is dragging it.
    */
   private boolean thumbPressed = false;
   
   /**
    * Mouse hovers the thumb. If {@code true} then the mouse if located over the thumb. Some 
    * {@code Theme}s draw the thumb different in this case. 
    */
   private boolean thumbHover = false;

   /** The offset of mouse-press event relative to thumb x-location. */
   private int thumbOffset;

   /**
    * Default constructor.
    * 
    * @param   id
    *          The widget's ID.
    * @param   viewport
    *          The related viewport.
    */
   public ScrollBarGuiImpl(WidgetId id, Viewport<GuiOutputManager> viewport)
   {
      super(id);
      this.viewport = viewport;
      init();
   }

   /**
    * Constructor.
    * 
    * @param   id
    *          The widget's ID.
    * @param   orientation
    *          Scroll bar orientation.
    * @param   viewport
    *          The related viewport.
    */
   ScrollBarGuiImpl(WidgetId id, Orientation orientation, Viewport<GuiOutputManager> viewport)
   {
      super(id, orientation);
      this.viewport = viewport;
      init();
   }

   /**
    * Get widget width.
    *
    * @return  widget width.
    */
   @Override
   public double width()
   {
      // note: super.width() already substracts X offset (location.x)
      return super.width() - wOff;
   }

   /**
    * Get widget height.
    *
    * @return  widget height.
    */
   @Override
   public double height()
   {
      // note: super.height() already substracts Y offset (location.y)
      return super.height() - hOff;
   }
   
   /**
    * Layout container.
    */
   @Override
   public void doLayout()
   {
      Container<GuiOutputManager> parent = parent();
      
      if (parent == null)
      {
         return;
      }
      
      boolean isThreeDBorder = ((BorderedPanelGuiImpl) parent).is3DBorder();
      int offset3dBorder = ThemeManager.getCurrentTheme().get3DBorderOffset();
      if (orientation == Orientation.HORIZONTAL && isThreeDBorder && wOff == 0)
      {
         wOff = screen().coordinates().widthFromNative(offset3dBorder);
      }
      else if (orientation == Orientation.VERTICAL && isThreeDBorder && hOff == 0)
      {
         hOff = screen().coordinates().heightFromNative(offset3dBorder);
      }
      
      NativeDimension pd = parent.physicalDimension();
      NativeDimension d = physicalDimension();

      int x = 0;
      int y = 0;
      
      NativeInsets ins = screen().coordinates().insetsToNative(parent.insets());

      // determine the bar location, account for the parent's insets
      if (orientation == Orientation.HORIZONTAL)
      {
         y = pd.height - d.height - ins.bottom - ins.top;
         if (isThreeDBorder)
         {
            x += offset3dBorder;
            y -= offset3dBorder;
         }
      }
      else
      {
         x = pd.width - d.width - ins.right - ins.left;
         if (isThreeDBorder)
         {
            x -= offset3dBorder;
            y += offset3dBorder;
         }
      }
      
      setPhysicalLocation(x, y);
   }

   /**
    * Request widget repainting.  Also perform widget layout.
    */
   @Override
   public void repaint()
   {
      doLayout();
      super.repaint();
   }
   
   /**
    * Widget drawing method.
    */
   @Override
   public void draw()
   {
      // draw only when widget is visible
      if (!isVisible())
      {
         return;
      }
      
      Theme theme = ThemeManager.getCurrentTheme();
      
      gc.colors();
      
      doLayout();
      
      // get scroll bar coordinates
      final NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();
      final NativeRectangle clip = clipRectangle(p, d);
      
      gd.draw(getId(), p, clip, () ->
      {
         // draw scroll bar area
         theme.drawScrollBarGauge(gd, gc, 0, 0, d.width, d.height,
                                  orientation == Orientation.HORIZONTAL); // fill background
         
         // compute the size of the buttons
         int h, w;
         NativePoint np = null;
         boolean drawThumb;
         
         if (orientation == Orientation.HORIZONTAL)
         {
            // compute buttons size
            w = Math.min(d.width / 2, theme.getScrollBarButtonHeight());
            h = theme.getScrollBarButtonWidth();
            drawThumb = d.width - 2 * w > 0;
         }
         else
         {
            // compute buttons size
            w = theme.getScrollBarButtonWidth();
            h = Math.min(d.height / 2, theme.getScrollBarButtonHeight());
            drawThumb = d.height - 2 * h > 0;
         }
         
         // draw left/top button
         lButt.config().widthPixels = w;
         lButt.config().heightPixels = h;
         if (orientation == Orientation.HORIZONTAL)
         {
            np = new NativePoint(0, (d.height - h) / 2);
         }
         else
         {
            np = new NativePoint((d.width - w) / 2, 0);
         }
         lButt.setPhysicalLocation(np.x, np.y);
         lButt.draw();
         
         // draw right/bottom button
         rButt.config().widthPixels = w;
         rButt.config().heightPixels = h;
         if (orientation == Orientation.HORIZONTAL)
         {
            np = new NativePoint(d.width - w, (d.height - h) / 2);
         }
         else
         {
            np = new NativePoint((d.width - w) / 2, d.height - h);
         }
         rButt.setPhysicalLocation(np.x, np.y);
         rButt.draw();

         // draw scroll bar thumb only if need
         if (drawThumb && (isControlled() ? controller.isEnabled() : 0 != max))
         {
            double range = max;
            int thumbMinSize = theme.getScrollBarThumbMinSize();

            // we allow position to be out of min/max values 
            int effectivePos = position;
            if (effectivePos < 0)
            {
               effectivePos = 0;
            }
            if (effectivePos > max)
            {
               effectivePos = max;
            }

            if (orientation == Orientation.HORIZONTAL)
            {
               int visibleRange = viewport.getScrollWidget().getVisibleDimension().width;
               int travelSize = d.width - 2 * w;
               
               double tWidth;
               
               if (isControlled())
               {
                  tWidth = controller.getThumbSize();
               }
               else
               {
                  double unitWidth = (double) d.width / visibleRange;
                  double totalWidth = (viewport.getScrollWidget().getScrollDimension().width *
                        unitWidth);
                  tWidth = Coordinate.scale((double) travelSize * d.width / totalWidth);
               }
               
               thumbDim.width = Coordinate.toInt(tWidth);
               thumbDim.height = theme.getScrollBarThumbWidth();
               
               // thumb travel-width is the total scrollbar width excluding buttons
               // and the width of the thumb itself (the thumb travel is limited
               // by its left and right edges)
               int maxPosX = d.width - 2 * w - thumbDim.width;
               
               // draw the thumb centered on gauge
               thumbPos.y = (d.height - thumbDim.height) / 2;
               
               // the condition below means the total area <= visible one and we have no thumb
               // to draw, or travel size is smaller than min thumb size, must return here
               if (travelSize < thumbMinSize || thumbDim.width >= travelSize)
               {
                  return;
               }
               
               if (isControlled())
               {
                  thumbPos.x = w + controller.getThumbPosition();
                  theme.drawScrollBarThumb(gd, gc, 
                        thumbPos.x, thumbPos.y, thumbDim.width, thumbDim.height,
                        thumbPressed, thumbHover, orientation == Orientation.HORIZONTAL);
               }
               else
               {
                  thumbPos.x = w + Coordinate.toInt(maxPosX * effectivePos / range);
                  
                  // if thumb size is small use default size
                  if (thumbDim.width < thumbMinSize)
                  {
                     int test = travelSize - thumbMinSize;
                     if (test > 0)
                     {
                        tWidth = test / range;
                        thumbPos.x = w + Coordinate.toInt(tWidth * effectivePos);
                        thumbDim.width = thumbMinSize;
                     }
                  }
                  
                  if (thumbDim.width >= thumbMinSize)
                  {
                     theme.drawScrollBarThumb(gd, gc,
                           thumbPos.x, thumbPos.y, thumbDim.width, thumbDim.height,
                           thumbPressed, thumbHover, orientation == Orientation.HORIZONTAL);
                  }
               }
            }
            else // VERTICAL scrollbar
            {
               int visibleRange = viewport.getScrollWidget().getVisibleDimension().height;
   
               // draw scroll bar thumb
               int travelSize = d.height - 2 * h;
   
               double tHeight;
   
               if (isControlled())
               {
                  tHeight = controller.getThumbSize();
               }
               else
               {
                  double unitHeight = (double) d.height / visibleRange;
                  double totalHeight = (viewport.getScrollWidget().getScrollDimension().height *
                        unitHeight);
      
                  tHeight = Coordinate.scale((double) travelSize * d.height / totalHeight);
               }
   
               thumbDim.width = theme.getScrollBarThumbWidth();
               thumbDim.height = Coordinate.toInt(tHeight);
   
               // thumb travel-height is the total scrollbar height excluding buttons
               // and the height of the thumb itself (the thumb travel is limited
               // by its top and bottom edges)
               int maxPosY = d.height - 2 * h - thumbDim.height;
   
               // draw the thumb centered on gauge
               thumbPos.x = (d.width - thumbDim.width) / 2;
               
               // the condition below means the total area <= visible one and we have no
               // thumb to draw, or travel size is smaller than min thumb size, must return here
               if (travelSize < thumbMinSize || thumbDim.height >= travelSize)
               {
                  return;
               }
               
               if (isControlled())
               {
                  thumbPos.y = h + controller.getThumbPosition();
                  theme.drawScrollBarThumb(gd, gc,
                        thumbPos.x, thumbPos.y, thumbDim.width, thumbDim.height,
                        thumbPressed, thumbHover, orientation == Orientation.HORIZONTAL);
               }
               else
               {
                  thumbPos.y = h + Coordinate.toInt(maxPosY * effectivePos / range);
                  
                  // if thumb size is small use default size
                  if (thumbDim.height < thumbMinSize)
                  {
                     int test = travelSize - thumbMinSize;
                     if (test > 0)
                     {
                        tHeight = test / range;
                        thumbPos.y = h + Coordinate.toInt(tHeight * effectivePos);
                        thumbDim.height = thumbMinSize;
                     }
                  }
                  
                  if (thumbDim.height >= thumbMinSize)
                  {
                     theme.drawScrollBarThumb(gd, gc,
                           thumbPos.x, thumbPos.y, thumbDim.width, thumbDim.height,
                           thumbPressed, thumbHover, orientation == Orientation.HORIZONTAL);
                  }
               }
            }
         }
      });
   }

   /**
      // TODO: this triggers repaint on each mouse move over it... needs to be moved in some 
      // layout/initialization code
    * Notification of a mouse clicked event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      // check if the widget is enabled
      if (isEnabled())
      {
         synchronized (e)
         {
            if (!isOverThumbButton(e) && e.getButton() != getMenuMouse())
            {
               Position direction = scrollDirection(e);
               onButtonClick(direction, ScrollStep.LONG, new MouseEvt(e, ancestor()));
               refresh();
            }
            else if (e.getButton() == getMenuMouse())
            {
               ScrollableWidget<?> scrollWidget = 
                     parent(ScrollPaneGuiImpl.class).viewport().getScrollWidget();
               TopLevelWindow<GuiOutputManager> wnd = parent(TopLevelWindow.class);
               if (wnd != null && scrollWidget.isScrollPopupSupported())
               {
                  showPopup(e);
               }
               e.consume();
            }
         }
      }
   }

   /**
    * Notification of a mouse dragged event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      // check if the widget is enabled
      // we will use only left button to drag the thumb
      int dragEventMask = MouseEvent.BUTTON1_DOWN_MASK;
      if (isEnabled() && 
          (e.getModifiersEx() & dragEventMask) == dragEventMask && 
           e.getButton() != getMenuMouse())
      {
         pressed = false;
         if (thumbPressed)
         {
            thumbHover = isOverThumbButton(e);
         }
         
         synchronized (e)
         {
            int pos = position(e);
            onDrag(pos);
            refresh();
         }
      }
   }
   
   /**
    * Notification of a mouse pressed event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(final MouseEvent e)
   {
      // check if the widget is enabled
      if (!isEnabled() || e.getButton() != MouseEvent.BUTTON1)
      {
         return;
      }
      NativePoint mousePos = translate(e);
      
      // keep the position of the mouse-press relative to the thumb
      // x location, this will be needed during mouse drag
      if (orientation == Orientation.HORIZONTAL)
      {
         thumbOffset = mousePos.x - thumbPos.x;
      }
      else
      {
         thumbOffset = mousePos.y - thumbPos.y;
      }
      
      pressed = true;
      thumbHover = isOverThumbButton(e);
      thumbPressed = thumbHover; 
      
      if (!isOverThumbButton(e))
      {
         mouseInputExecutor.execute(() ->
         {
            ThinClient tc = ThinClient.getInstance(); 
            while (pressed)
            {
               try
               {
                  // wait to see if not a click event
                  Thread.sleep(GuiConstants.REPEAT_MILLIS);

                  // fire mouse clicked
                  if (pressed)
                  {
                     Position clickPos = scrollDirection(e);
                     int pos = position;

                     if (clickPos == Position.LEFT)
                     {
                        pos -= longStep;
                     }
                     else
                     {
                        pos += longStep;
                     }

                     ScrollEvent event = createScrollEvent(pos);

                     tc.postOSEvent(event);
                  }
               }
               catch (InterruptedException ex)
               {
                  // nop
               }
            }
         });
      }
      else
      {
         if (controller != null)
         {
            controller.startThumbDrag();
         }
      }
   }
   
   /**
    * Notification of a mouse released event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      // check if the widget is enabled
      if (!isEnabled())
      {
         return;
      }
      
      thumbHover = false;
      thumbPressed = false;
      pressed = false;
      
      if (controller != null)
      {
         controller.stopThumbDrag();
      }
      
      repaint(); // restore neutral UI
   }
   
   /**
    * Notification of a mouse moved event occurred for this widget.
    *
    * @param   e
    *          The mouse event.
    */
   @Override
   public void mouseMoved(MouseEvent e)
   {
      boolean wasHoveringThumb = thumbHover;
      thumbHover = isOverThumbButton(e);
      
      if (wasHoveringThumb != thumbHover)
      {
         repaint();
      }
   }
   
   /**
    * Notification of a mouse entered event occurred for this widget.
    *
    * @param   e
    *          The mouse event.
    */
   @Override
   public void mouseEntered(MouseEvent e)
   {
      if (!isEnabled())
      {
         return;
      }
      
      thumbHover = isOverThumbButton(e);
      if (thumbHover && ThemeManager.getCurrentTheme().isMouseHoverSensitive())
      {
         repaint();
      }
   }
   
   /**
    * Notification of a mouse exited event occurred for this widget.
    *
    * @param   e
    *          The mouse event.
    */
   @Override
   public void mouseExited(MouseEvent e)
   {
      if (thumbHover)
      {
         thumbHover = false;
         if (ThemeManager.getCurrentTheme().isMouseHoverSensitive())
         {
            repaint();
         }
      }
   }
   
   /**
    * Notification of a mouse wheel moved event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseWheelMoved(MouseWheelEvent e)
   {
      // check if the widget is enabled
      if (isEnabled())
      {
         Position direction;
         int rotation = e.getWheelRotation();

         // determine scroll direction
         if (orientation == Orientation.HORIZONTAL)
         {
            direction = (rotation < 0) ? Position.LEFT : Position.RIGHT;
         }
         else
         {
            direction = (rotation < 0) ? Position.LEFT : Position.RIGHT;
         }

         // do scroll
         for (int i = 0; i < Math.abs(rotation); i++)
         {
            onButtonClick(direction, ScrollStep.LONG, new MouseEvt(e, ancestor()));
            refresh();
         }
      }
   }

   /**
    * Check if the widget implementation can raise legacy, 4GL-style, mouse events.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   public boolean raiseLegacyMouseEvents()
   {
      return false;
   }
   
   /**
    * Destroy this widget.  Cleans up all including widgets before destruction.
    */
   @Override
   public void destroy()
   {
      // stop the mouse related thread if it is active
      if (!mouseInputExecutor.isShutdown())
      {
         mouseInputExecutor.shutdownNow();
      }

      // call superclass for main destruction
      super.destroy();
   }
   
   /**
    * Check if this widget gains ENTRY event when right/middle mouse button is clicked.
    *
    * @return  <code>true</code> if this widget gains ENTRY event
    *          when right/middle mouse button is clicked.
    */
   @Override
   public boolean isRMBEntrySupported()
   {
      return false;
   }
   
   /**
    * Gets the widget, that owns this scrollbar. For message area or window workspace
    * return the Window.
    * 
    * @return   Widget. 
    *           MessageAreaGuiImpl and WindowWorkspace are subclasses of ScrollPane,
    *           therefore when scrollbar is owned by message area or workspace - 
    *           this method returns Window.
    */
   public Widget<GuiOutputManager> getOwningWidget()
   {
      Widget<GuiOutputManager> parent = this.parent();
      
      while(parent != null && (parent instanceof ScrollPane<?> || parent instanceof BorderedPanel))
      {
         parent = parent.parent();
      }
      
      return parent;
   }
   
   /**
    * Create GUI specific auto-scroll event if the changed position is valid, otherwise it returns
    * null indicating that it is not required to send scroll events due to the current scroller's
    * position reaches its bounds. 
    * 
    * @param    btn
    *           The source button that caused the event.
    * 
    * @return   Returns the valid scroll event if the current scroller's position doesn't reach
    *           its bounds, otherwise null.
    */
   public ScrollEvent createAutoScrollEvent(Position btn)
   {
      if (position <= 0 || position >= max)
      {
         return null;
      }
      
      // start with the current position
      int newPos = position;
      
      if (btn == ScrollBar.Position.RIGHT)
      {
         newPos += shortStep;
         if (newPos > max)
         {
            newPos = max;
         }
      }
      else
      {
         newPos -= shortStep;
         if (newPos < 0)
         {
            newPos = 0;
         }
      }
      
      return createScrollEvent(newPos);
   }

   /**
    * Notification on scroll bar button click.
    *  
    * @param   button
    *          Clicked button.
    * @param   scrollStep
    *          The indicator of the desired scroll step. How far
    *          we scroll depends on the location of the mouse-click.
    *          For example, scroll bar buttons generate a short scroll
    *          versus the location between the scroll button and scroll
    *          thumb generates a long scroll.
    * @param   e
    *          The mouse event. <code>null</code> means KeyEvent
    */
   public void onButtonClick(Position button, ScrollStep scrollStep, MouseEvt e)
   {
      if (e != null && e.getEvent().getButton() == getMenuMouse())
      {
         ScrollableWidget<?> scrollWidget = parent(ScrollPaneGuiImpl.class).viewport().getScrollWidget();
         TopLevelWindow<GuiOutputManager> wnd = parent(TopLevelWindow.class);
         if (wnd != null && scrollWidget.isScrollPopupSupported())
         {
            showPopup(e.getEvent());
         }
         e.consume();
      }
      else
      {
         super.onButtonClick(button, scrollStep, e);
      }
   }
   
   /**
    * Get thumb size for the scroll bar.
    *
    * @return thumb size for the scroll bar (in pixels).
    */
   public NativeDimension getThumbDim()
   {
      return thumbDim;
   }

   /**
    * Create UI specific scroll event.
    * 
    * @param   newPosition
    *          New scroll position.
    */
   @Override
   protected ScrollEvent createScrollEvent(int newPosition)
   {
      if (orientation == Orientation.HORIZONTAL)
      {
         return new ScrollEvent(this, newPosition, null);
      }
      else
      {
         return new ScrollEvent(this, null, newPosition);
      }
   }

   /**
    * Scroll bar initialization.
    */
   private void init()
   {
      gd = (GuiDriver) OutputManager.getDriver();
      // set default BGCOLOR and FGCOLOR
      gc = new GuiColorResolver(gd, config(), SysColor.COLOR_SCROLLBAR, SysColor.COLOR_WINDOWTEXT);
      gc.thumbColor = gd.getSysColor(SysColor.COLOR_3DFACE);
      // buttons
      thumbPos = new NativePoint(0, 0);
      thumbDim = new NativeDimension(0, 0);
      lButt = new ScrollBarGuiButton(WidgetId.nextID(), this, ScrollBar.Position.LEFT);
      rButt = new ScrollBarGuiButton(WidgetId.nextID(), this, ScrollBar.Position.RIGHT);
      add(lButt);
      add(rButt);
      
      Theme theme = ThemeManager.getCurrentTheme();
      CoordinatesConversion cc = screen().coordinates();
      
      if (orientation == Orientation.HORIZONTAL)
      {
         setHeight(cc.heightFromNative(theme.getScrollBarGaugeWidth()));
      }
      else
      {
         setWidth(cc.widthFromNative(theme.getScrollBarGaugeWidth()));
      }
   }

   /**
    * Check if mouse is over thumb button.
    * 
    * @param   event
    *          Mouse event.
    *          
    * @return  <code>true</code> if mouse under the thumb button <code>false</code> otherwise. 
    */
   private boolean isOverThumbButton(MouseEvent event)
   {
      NativeRectangle thumbRect = new NativeRectangle(thumbPos, thumbDim);
      NativePoint mousePos = translate(event);
      return thumbRect.contains(mousePos);
   }

   /**
    * Determine the equivalent scroll bar position which correspond to actual mouse position. 
    * 
    * @param   e
    *          Mouse event.
    * 
    * @return  see above.
    */
   public int position(MouseEvent e)
   {
      NativePoint mousePos = translate(e);
      return position(mousePos);
   }

   /**
    * Determine the equivalent scroll bar position which correspond to actual mouse position. 
    * 
    * @param   mousePos
    *          The point where mouse is clicked.
    * 
    * @return  see above.
    */
   public int position(NativePoint mousePos)
   {
      NativeDimension d = physicalDimension();

      double ratio = 0;
      int result = 0;

      int thumbPos;
      
      // first determine the ratio of the mouse event location,
      // this is the mouse event location offset by the left button size,
      // divided by the total size of the thumb travel path. 
      if (orientation == Orientation.HORIZONTAL)
      {
         int lButtW = lButt.physicalDimension().width;
         int rButtW = rButt.physicalDimension().width;

         thumbPos = mousePos.x - lButtW - thumbOffset;
         ratio = (double) thumbPos / (d.width - lButtW - rButtW - thumbDim.width);
      }
      else
      {
         int lButtH = lButt.physicalDimension().height;
         int rButtH = rButt.physicalDimension().height;
         
         thumbPos = mousePos.y - lButtH - thumbOffset;
         ratio = (double) thumbPos / (d.height - lButtH - rButtH - thumbDim.height);
      }
      
      // translate ratio to scroll range
      result = (int) Math.round(ratio * max);

      if (controller != null)
      {
         controller.thumbMoved(thumbPos);
      }

      if (result < 0)
      {
         result = 0;
      }

      if (result > max)
      {
         result = max;
      }

      return result;
   }

   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvent.MOUSE_CLICKED,
         MouseEvent.MOUSE_DRAGGED,
         MouseEvent.MOUSE_MOVED,
         MouseEvent.MOUSE_ENTERED,
         MouseEvent.MOUSE_EXITED,
         MouseEvent.MOUSE_PRESSED,
         MouseEvent.MOUSE_RELEASED,
         MouseEvent.MOUSE_WHEEL
      };
   }
   
   /**
    * Get size of the travel area available for thumb scrolling.
    *
    * @return size of travel area, in pixels.
    */
   int getTravelSize()
   {
      boolean horizontal = orientation == Orientation.HORIZONTAL;
      int lButtSize = horizontal ? lButt.physicalDimension().width :
                                   lButt.physicalDimension().height;
      int rButtSize = horizontal ? rButt.physicalDimension().width :
                                   rButt.physicalDimension().height;
      int totalSize = horizontal ? physicalDimension().width :
                                   physicalDimension().height;
      return totalSize - rButtSize - lButtSize;
   }

   /**
    * Get size of the button.
    *
    * @return size of one of the buttons in pixels.
    */
   int getButtonSize()
   {
      return orientation == Orientation.HORIZONTAL ? lButt.physicalDimension().width :
                                                     lButt.physicalDimension().height;
   }
   
   /**
    * Get thumb offset (mouse click point relative to thumb position).
    * 
    * @return   thumbOffset
    */
   int getThumbOffset()
   {
      return thumbOffset;
   }
   
   /**
    * 
    * @param   e
    *          Mouse event.
    *          
    * @return  see above.
    */
   Position scrollDirection(MouseEvent e)
   {
      NativePoint n = translate(e);

      if (orientation == Orientation.HORIZONTAL)
      {
         return (n.x < thumbPos.x) ? Position.LEFT : Position.RIGHT;
      }
      else
      {
         return (n.y < thumbPos.y) ? Position.LEFT : Position.RIGHT;
      }
   }

   /**
    * Invoke repaint for this widget.
    */
   private void refresh()
   {
      ThinClient.getInstance().eventDrawingBracket(this, false, true, this::repaint);
   }
   
   /**
    * Shows the scroll bar popup menu.
    * 
    * @param    e
    *           The mouse event caused the menu to be shown.
    */
   private void showPopup(MouseEvent e)
   {
      boolean vertical = ScrollBar.Orientation.VERTICAL.equals(this.orientation());
      ScrollPopupGuiImpl menu = ScrollPopupGuiImpl.create(this, vertical);
      menu.showPopup(this, new NativePoint(e.getX(), e.getY()));
   }
}
