/*
** Module   : ScrollableListGuiImpl.java
** Abstract : scrollable list implementation. 
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140917 First version.
** 002 MAG 20141016 Improve text drawing and color management. Added mouse wheel handler. 
** 003 CA  20141123 Added scroll support via cursor keys.
** 004 CA  20150323 The GuiFontResolver needs to know if the expected font is for a fixed-width
**                  widget.
** 005 HC  20150323 Removed business logic from config classes.
** 006 CA  20150507 Draw text centered vertically.
** 007 GES 20150425 Eliminated AWT dependencies.
** 008 GES 20150730 Modified GUI driver translate() usage to match API changes.
** 009 HC  20150831 Scrolling support overhaul.
** 010 GES 20150828 Honor the right and bottom insets when drawing. If non-zero we used to draw
**                  outside of our intended area. Added mouse actions.
** 011 EVL 20151013 Fix for item highlight text.  Also text position and background area.
** 012 CA  20160216 Cache the intermediate widget states.
** 013 SVL 20160719 Updated translate() signature.
** 014 OM  20170320 Added support for Themes.
** 015 OM  20170713 Added dotted rectangle on focus.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.Window;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.model.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Scrollable list GUI implementation. 
 */
public class ScrollableListGuiImpl
extends DefaultList<String, GuiOutputManager>
implements WidgetWithConfig<BaseConfig>
{
   /** Screen driver */
   private GuiDriver gd;

   /** Widget configuration */
   private BaseConfig config;

   /** Color resolver */
   private GuiColorResolver gc;
   
   /** Font resolver */
   private GuiFontResolver gf;
   
   /** Row height */
   private int rowHeight;

   /** Flag indicating if the list elements can be unselected. */
   private final boolean allowDeselect;
   
   /** Inherit parent width */
   private boolean inheritWidth = false;

   /** Inherit parent height */
   private boolean inheritHeight = false;
   
   /**
    * Constructor.
    * 
    * @param   id
    *          The widget's ID.
    * @param   model
    *          List model.
    * @param   allowDeselect
    *          <code>true</code> if the list elements can be unselected.
    */
   public ScrollableListGuiImpl(WidgetId id, ListModel<String> model, boolean allowDeselect)
   {
      super(model);
      this.allowDeselect = allowDeselect;
      gd = (GuiDriver) OutputManager.getDriver();
      config = ConfigManager.getInstance().resolveWidgetConfig(id, BaseConfig.class);
      config.visible = true;
      // set default BGCOLOR and FGCOLOR
      gc = new GuiColorResolver(gd, config, SysColor.COLOR_3DFACE, SysColor.COLOR_WINDOWTEXT);
      gc.selectionBgColor = gd.getSysColor(SysColor.COLOR_HIGHLIGHT);
      gc.selectionFgColor = gd.getSysColor(SysColor.COLOR_HIGHLIGHTTEXT);
      // margins
      nativeInsets = new NativeInsets(0, 3, 0, 1);
   } 
         
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, BaseConfig cfg)
   {
      // no-op
   }

   /**
    * Get the widget config.
    * 
    * @return   See above.
    */
   @SuppressWarnings("unchecked")
   @Override
   public BaseConfig config()
   {
      return config;
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param   beforeUpdate
    *          Config reference capturing the config state before the owner's config was modified.
    */
   @Override
   public void afterConfigUpdate(BaseConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
   }
   
   /**
    * Set if width is inherited from parent widget.
    * 
    * @param   inheritWidth
    *          <code>true</code> if width is inherited from parent. 
    */
   public void setInheritWidth(boolean inheritWidth)
   {
      this.inheritWidth = inheritWidth;
   }
   
   /**
    * Set if width is inherited from parent widget.
    * 
    * @param   inheritHeight
    *          <code>true</code> if height is inherited from parent. 
    */
   public void setInheritHeight(boolean inheritHeight)
   {
      this.inheritHeight = inheritHeight;
   }
   
   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public double width()
   {
      boolean parentWidth = (inheritWidth && parent() != null);
      return (parentWidth) ? parent().width() : super.width();
   }

   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   @Override
   public double height()
   {
      boolean parentHeight = (inheritHeight && parent() != null);
      return (parentHeight) ? parent().height() : super.height();
   }

   /**
    * Draw widget.
    */
   @Override
   public void draw()
   {
      // draw only when widget is visible
      if (!isVisible())
      {
         return;
      }
      
      // resolve font
      final FontCache cached = font();
      
      // resolve colors
      gc.colors();
      
      // get container coordinates
      final NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();
      final NativeRectangle clip = clipRectangle(p, d);

      // display rows
      final int row  = this.currentRow();
      final int rows = this.visibleRows();
      
      // model size
      final int modelSize = model().size();

      gd.draw(getId(), p, clip, () ->
      {
         // draw list
         // TODO: this code fills an empty selection list - for message area - with black
         // TODO: if other scrollable widget(s) experience errors, test a message area with 
         // no messages
         /*
         gd.fillRect(nativeInsets.left,
                     nativeInsets.top,
                     d.width - nativeInsets.right,
                     d.height - nativeInsets.bottom);
         */
         if (modelSize > 0 && cached.font != null)
         {
            gd.setGuiFont(cached.font);
            gd.setFontStyle(cached.font.style);

            for (int i = row, l = 0; i < (row + rows); i++, l++)
            { 
               if (i >= 0 && i < modelSize)
               {
                  int x = 0;
                  int y = rowHeight * l;
                  int dx = d.width;
                  int dy = rowHeight;
                  
                  if (selected(i))
                  {
                     gd.setColor(gc.selectionBgColor);
                     gd.fillRect(x, y, dx, dy);
                     gd.setColor(gc.selectionFgColor);
                  }
                  else
                  {
                     gd.setColor(gc.bgColor);
                     gd.fillRect(x, y, dx, dy);
                     gd.setColor(gc.fgColor);
                  }
                  
                  final String text = model().get(i);
                  
                  // draw item text
                  if (text != null)
                  {
                     NativePoint     pt = new NativePoint(x, y);
                     NativeRectangle cl = new NativeRectangle(0, 0, dy, dx); 
                     
                     gd.draw(pt, cl, () -> 
                        gd.drawStringCentered(text, nativeInsets.left - 1, cl.height() / 2 - 1));
                  }
                  
                  if (selected(i) && hasFocus())
                  {
                     ThemeManager.getCurrentTheme().drawSelectionFocus(gd, x, y, dx - 1, dy - 1);
                  }
               }
            }
         }
      });
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      synchronized (this)
      {
         if (model().size() > 0 && rowHeight > 0)
         {
            NativePoint n = translate(e);
            
            // calculate clicked line
            int row = n.y / rowHeight + currentRow();
            
            if (selected(row))
            {
               if (allowDeselect && selectionMode() == ListSelectionModel.SINGLE_SELECTION)
               {
                  deselect(row, row);
               }
            }
            else
            {
               select(row);
            }
            
            repaint(); 
         }
      }

      ThinClient.getInstance().getFocusManager().setFocusOn(this);
   }

   /**
    * Notification of a mouse wheel moved event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseWheelMoved(MouseWheelEvent e)
   {
      // total items in the list
      int items = model().size(); 
      
      if (items > 0)
      {
         int item = currentRow() + e.getWheelRotation();
         int max = items - visibleRows();

         if (item >= max )
         {
            item = max;
         }
         else if (item < 0)
         {
            item = 0;
         }

         // scroll list
         ensureRowVisible(item);
         
         repaint();
      }
   }
   
   /**
    * Process key event.
    * 
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      int[] selection = selection();
      
      if (selection.length == 0)
      {
         // nothing to do, go with default
         super.processKeyEvent(event);
         return;
      }
      
      if (selection.length > 1)
      {
         UnimplementedFeature.partial(
            "Key received for a scrollable list with multiple selected elements");
      }
      
      int item = selection[0];
      int oldRow = currentRow();
      
      switch (event.keyCode())
      {
         case Key.VK_CURSOR_UP:
            item = item - 1;
            break;
         
         case Key.VK_CURSOR_DOWN:
            item = item + 1;
            break;
            
         default:
            // for other keys, go with default
            super.processKeyEvent(event);
            return;
      }
      
      if (item < 0 || item > model().size() - 1)
      {
         // reached top/bottom limit
         event.consume();
         return;
      }
      
      select(item);
      
      if (item < oldRow || item > oldRow + visibleRows() - 1)
      {
         // simulate a knob drag so that the currently selected item is brought to the visible
         // area
         int current = item < oldRow ? item : item - visibleRows() + 1;
         ensureRowVisible(current);
      }
      
      repaint();
   }

   /**
    * Add new item to list.
    * 
    * @param   item
    *          Item to add.
    *          
    * @return  First visible row. Could be negative.
    */
   public int addItem(String item)
   {
      model().add(item);
      int firstVisibleRow = model().size() - visibleRows();
      ensureRowVisible(firstVisibleRow);
      return firstVisibleRow;
   }
   
   /**
    * Delete all items.
    */
   public void deleteItems()
   {
      model().clear();
      ensureRowVisible(0);
   }

   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[] { MouseEvent.MOUSE_CLICKED, MouseEvent.MOUSE_WHEEL };
   }
   
   /**
    * Resolve font. Because the font could be changed dynamically whenever we draw the widget we
    * have to check if font was changed and rebuild font object if true.
    */
   private FontCache font()
   {
      if (gf == null)
      {
         Window<?> window = (Window<?>) ancestor();
         gf = new GuiFontResolver(window, gd, config, false);
      }
      
      FontCache cached = gf.font();
      rowHeight = cached.fontHeight + nativeInsets.top + nativeInsets.bottom;
      return cached;
   }
}
