/*
** Module   : SliderGuiImpl.java
** Abstract : GUI slider implementation
**
** Copyright (c) 2016-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20160127 Initial version.
** 002 EVL 20160217 LOAD-MOUSE-POINTER support.
** 003 SVL 20160719 Updated translate() signature.
** 004 OM  20170320 VectorGraphicsHelper.getScalableImage() signature change.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Specific GUI implementation of SLIDER widget.
 */
public class SliderGuiImpl
extends Slider<GuiOutputManager>
{
   /** The size is used to compute different positions for parts. */
   private static final int FIXED_SIZE = 16;
   
   /** The size of the internal slider area for none ticmarks defined. */
   private static final int TM_NONE_SIZE = 18;
   
   /** The size of the internal slider area for both ticmarks mode defined. */
   private static final int TM_BOTH_SIZE = 40;
   
   /** The size of the internal slider area for single side ticmarks mode defined. */
   private static final int TM_SINGLE_SIZE = 35;
   
   /** Working tic-marks mode. */
   private static enum TicMarks
   {
      UNKNOWN, NONE, TOP, BOTTOM, LEFT, RIGHT, BOTH
   }
   
   /** Cached tic marks mode in  format. */
   private TicMarks tmMode = TicMarks.UNKNOWN;
   
   /** Screen driver. */
   private GuiDriver gd;

   /** Output manager. */
   private GuiOutputManager screen;
   
   /** Coordinate converter. */
   private CoordinatesConversion cc;

   /** Color resolver. */
   private GuiColorResolver gc;
   
   /** Font resolver */
   private GuiFontResolver gf;
      
   /** Light shadow box color. */
   private ColorRgb lightShadow;

   /** Dark shadow box color. */
   private ColorRgb darkShadow;
   
   /** Dark shadow box color. */
   private ColorRgb veryDarkShadow;
   
   /** Mode dependent thumb dimension of the slider. */
   private NativeDimension thumbDim;
   
   /** Thumb X position of the slider. */
   private int thumbX = -1;
   
   /** Thumb Y position of the slider. */
   private int thumbY = -1;
   
   /** The rectangle where the mouse events must be handled. */
   private NativeRectangle mouseArea;
   
   /** Mouse drag indicator. */
   private boolean dragInProgress = false;
   
   /** Flag indicating internals has been completed. */
   private boolean initCompleted = false;
   
   /** The image object for poiner draw. */
   private ImageWrapper<?> img;
   
   /**
    * This constructor creates a slider.
    * 
    * @param    id
    *           The widget's ID.
    * @param    owner
    *           The parent container, frame which will include this slider.
    */
   SliderGuiImpl(WidgetId id, Container<GuiOutputManager> owner) 
   {
      super(id, owner);
   }

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public SliderGuiImpl(WidgetId id, SliderConfig cfg)
   {
      super(id, cfg);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    wcfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, SliderConfig wcfg)
   {
      super.initialize(id, wcfg);
      
      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      
      // output manager
      screen = screen();
      cc = screen.coordinates();
      
      // font resolver
      gf = new GuiFontResolver(WindowManager.resolveWindow(this), gd, config, false);
      
      // initialize the shadow box colors
      lightShadow = gd.getSysColor(SysColor.COLOR_3DFACE);
      darkShadow  = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      veryDarkShadow = gd.getSysColor(SysColor.COLOR_3DDKSHADOW);
      
      // adding widget to the registry
      screen.getRegistry().addWidget(this);
   }
   
   /**
    * Draw the slider. Called by this rectangle's parent slider.
    */
   @Override
   public void draw()
   {
      // rectangle options need to be refreshed before every painting
      final SliderConfig config = (SliderConfig) config();
      gc = new GuiColorResolver(gd, config, SysColor.COLOR_WINDOW, SysColor.COLOR_WINDOWTEXT);
      
      // determine colors to use
      gc.colors();
   
      // get location and size
      NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();
      
      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, new NativeDimension(d.width, d.height));
      
      // initialize internal variables
      initInternals();
      
      // check if we need to load additional image to draw pointer
      if (img == null && tmMode != TicMarks.BOTH)
      {
         initThumbImage(tmMode, config.horizontal);
      }

      // clippable drawing start 
      gd.draw(p, clip, new Runnable()
      {

         @Override
         public void run()
         {
            // fill the background first
            gd.setColor((isThreeD() ? lightShadow : gc.bgColor));
            gd.fillRect(0, 0, d.width, d.height);
            
            // inside this call widget origin is always (0,0)
            int nativeLine = cc.heightToNative(1.0) * 3 / 4;
            // starting point to the area of dotted focused frame 
            NativePoint po = new NativePoint(0, config.horizontal
               ? tmMode == TicMarks.TOP || tmMode == TicMarks.BOTTOM ||
                 tmMode == TicMarks.BOTH ? nativeLine : nativeLine * 2 : 0);
            
            // width and heigth depend on the oriantation and tic mark mode
            int w = config.horizontal
                       ? d.width
                       : tmMode == TicMarks.NONE
                          ? TM_NONE_SIZE
                          : tmMode == TicMarks.BOTH
                             ? TM_BOTH_SIZE : TM_SINGLE_SIZE;

            int h = config.horizontal
                       ? tmMode == TicMarks.NONE
                          ? TM_NONE_SIZE
                          : tmMode == TicMarks.BOTH
                             ? TM_BOTH_SIZE : TM_SINGLE_SIZE
                       : d.height;
                       
            NativeDimension dim =  new NativeDimension(w, h);
                       
            // draw the slider components
            drawBase(po, dim, config.horizontal);
            // check if the widget is highlighted and enabled
            if (highlighted() && isEnabled())
            {
               drawFrame(po, dim, config.horizontal);
            }
            // check if the current value must be visible, note:
            // NO-CURRENT-VALUE == TRUE means to display the current value
            if (config.noCurrentValue)
            {
               drawCurrentValue(po, dim, config.horizontal);
            }
         }
         
      });
   }

   /**
    * Get the Z-Order class
    * 
    * @return Z-Order class of the widget
    */
   @Override
   public ZOrderClass getZOrderClass()
   {
      return ZOrderClass.DECORATION;
   }
   
   /**
    * Notification of a mouse pressed event occurred for this widget.  Moves the slider pointer
    * and changes the current screen value according the mouse coordinate.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      // we consider only for enabled widget
      if (!isEnabled())
      {
         return;
      }

      synchronized (this)
      {
         NativePoint n = translate(e);
         // check if the mouse is over the pointer
         if (isOverThumb(n.x, n.y, config.horizontal))
         {
            dragInProgress = true;
         }
         else
         {
            // check if the mouse coords is inside the value change area and get the suggested
            // value
            int newPos = getNewSliderValue(n.x, n.y);
            // -1 means the mouse is not in meaning area, do nothing for this
            if (newPos != -1 && newPos != config.currentValue)
            {
               // change the slider value by mouse
               changeScreenValue((newPos > config.currentValue ? INCREMENT : -INCREMENT));
            }
         }
      }
   }
   
   /**
    * Notification of a mouse released event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      // check if the widget is enabled
      if (isEnabled())
      {
         // here we can stop the drag
         synchronized (this)
         {
            dragInProgress = false;
         }
      }
   }

   /**
    * Notification of a mouse dragged event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      // check if the widget is enabled
      // we will use only left button to drag the thumb
      int dragEventMask = MouseEvent.BUTTON1_DOWN_MASK;
      if (dragInProgress && isEnabled() && (e.getModifiersEx() & dragEventMask) == dragEventMask)
      {
         synchronized (e)
         {
            NativePoint n = translate(e);
            // get new slider position
            int newPos = getNewSliderValue(n.x, n.y);
            // -1 means the mouse is not in meaning area, do nothing for this
            if (newPos != -1)
            {
               // change the slider value by mouse drag
               changeScreenValue(newPos - config.currentValue);
            }
         }
      }
   }
   
   /**
    * Check if the widget supports LOAD-MOUSE-POINTER() method 
    * 
    * @return <code>true</code> if the widget supports 
    *         LOAD-MOUSE-POINTER() method
    */
   @Override
   public boolean supportsCustomMousePointer()
   {
      return true;
   }

   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvent.MOUSE_CLICKED, 
         MouseEvent.MOUSE_PRESSED, 
         MouseEvent.MOUSE_RELEASED,
         MouseEvent.MOUSE_ENTERED,
         MouseEvent.MOUSE_EXITED,
         MouseEvent.MOUSE_DRAGGED
      };
   }
   
   /**
    * Check if the given X,Y coordinates fall into value changed area and calculate the new
    * suggested one.
    * 
    * @param    x
    *           The X coordinate to check.
    * @param    y
    *           The Y coordinate to check.
    * 
    * @return   The new suggested value for slider or -1 if the coordinate is not in area of
    *           changes.
    */
   private int getNewSliderValue(int x, int y)
   {
      int result = -1;
      // check if the mouse inside work area
      if (mouseArea != null && mouseArea.contains(x, y))
      {
         // take into account LARGE-TO-SMALL attribute too
         int newPx = config.horizontal
                        ? config.largeToSmall ? mouseArea.right() - x : x - mouseArea.left()
                        : config.largeToSmall ? y - mouseArea.top() : mouseArea.bottom() - y;
         // area length depending on orientation
         int size  = config.horizontal ? mouseArea.width() : mouseArea.height();
         // new slider screen value
         result = 1 + newPx * (getRange() - 1) / size;
      }
      
      return result;
   }
   
   /**
    * Defines the area the mouse events should be handled.
    * 
    * @param    xbase
    *           The X coordinate of the main ruler.
    * @param    ybase
    *           The Y coordinate of the main ruler.
    * @param    lbase
    *           The length of the main ruler.
    * @param    horiz
    *           The flag indicating if the orientation is horizontal or not.
    */
   private void setupMouseEventsArea(int xbase, int ybase, int lbase, boolean horiz)
   {
      int left   = horiz ? xbase : xbase - thumbDim.height/2;
      int top    = horiz ? ybase - thumbDim.height/2 : ybase;
      int right  = horiz ? xbase + lbase : xbase + thumbDim.height/2;
      int bottom = horiz ? ybase + thumbDim.height/2 : ybase + lbase;
      
      mouseArea = new NativeRectangle(top, left, bottom, right);
   }
   
   /**
    * Check if the given X,Y coordinates fall into slider pointer rectangle.
    * 
    * @param    x
    *           The X coordinate to check.
    * @param    y
    *           The Y coordinate to check.
    * @param    horiz
    *           The flag indicating if the orientation is horizontal or not.
    * 
    * @return   <code>TRUE</code> if X,Y is over pointer, <code>FALSE</code> otherwise.
    */
   private boolean isOverThumb(int x, int y, boolean horiz)
   {
      // check the pointer area
      int left   = horiz ? thumbX - thumbDim.width/2  : thumbX - thumbDim.height/2;
      int top    = horiz ? thumbY - thumbDim.height/2 : thumbY - thumbDim.width/2;
      int right  = horiz ? thumbX + thumbDim.width/2  : thumbX + thumbDim.height/2;
      int bottom = horiz ? thumbY + thumbDim.height/2 : thumbY + thumbDim.width/2;

      // true if inside this
      return x >= left && x <= right && y >= top && y <= bottom;
   }
   
   /**
    * Draws the slider trackbar and pointer.
    * 
    * @param    orig
    *           The widget's starting point.
    * @param    dim
    *           The widget's size point.
    * @param    horiz
    *           The flag indicating if the orientation is horizontal or not.
    */
   private void drawBase(NativePoint orig, NativeDimension dim, boolean horiz)
   {
      // horizontal left edge
      int x = horiz ? orig.x + (tmMode == TicMarks.NONE ? FIXED_SIZE/2 - 2 : FIXED_SIZE/2)
                    : tmMode == TicMarks.BOTH ? orig.x + dim.width/2 - 2
                       : tmMode == TicMarks.LEFT ? orig.x + 19
                          : tmMode == TicMarks.RIGHT ? orig.x + 9 : orig.x + 5;
      // vertical top edge
      int y = horiz ? tmMode == TicMarks.BOTH ? orig.y + dim.height/2 - 2
                    : tmMode == TicMarks.TOP
                       ? orig.y + 19 : tmMode == TicMarks.BOTTOM ? orig.y + 9 : orig.y + 5
                          : (tmMode == TicMarks.NONE ? FIXED_SIZE/2 - 2 : FIXED_SIZE/2);
      // vertical or horizontal length
      int l = horiz ? dim.width - (tmMode == TicMarks.NONE ? FIXED_SIZE - 4 : FIXED_SIZE)
                    : dim.height - (tmMode == TicMarks.NONE ? FIXED_SIZE - 4 : FIXED_SIZE);
      
      // store the mouse sensitive area
      if (mouseArea == null)
      {
         setupMouseEventsArea(x, y, l, horiz);
      }

      // the ruler drawing depends on orientation, simulate 3D effect
      if (horiz)
      {
         gd.setColor(veryDarkShadow);
         gd.drawLine(x + 1, y + 1, x + l - 3, y + 1);

         gd.setColor(darkShadow);
         gd.drawLine(x, y, x + l - 2, y);
         gd.drawLine(x, y + 1, x, y + 2);

         gd.setColor(lightShadow);
         gd.drawLine(x + 1, y + 2, x + l - 2, y + 2);
         gd.drawLine(x + l - 2, y + 1, x + l - 2, y + 1);

         gd.setColor(gc.bgColor);
         gd.drawLine(x, y + 3, x + l - 1, y + 3);
         gd.drawLine(x + l - 1, y, x + l - 1, y + 2);
      }
      else
      {
         gd.setColor(veryDarkShadow);
         gd.drawLine(x + 1, y + 1, x + 1, y + l - 3);

         gd.setColor(darkShadow);
         gd.drawLine(x, y, x, y + l - 1);
         gd.drawLine(x + 1, y, x + 2, y);

         gd.setColor(lightShadow);
         gd.drawLine(x + 2, y + 1, x + 2, y + l - 2);
         gd.drawLine(x + 1, y + l - 2, x + 1, y + l - 2);

         gd.setColor(gc.bgColor);
         gd.drawLine(x + 3, y, x + 3, y + l - 1);
         gd.drawLine(x, y + l - 1, x + 2, y + l - 1);
      }

      // the processing below needs the coord and length to be modified
      if (horiz)
      {
         x += thumbDim.width/2;
      }
      else
      {
         y += thumbDim.width/2;
      }
      // adjust length
      l -= thumbDim.width;
      
      // draw slider tic marks
      drawTicMarks(x, y, l);
      
      // draw slider pointer
      drawThumb(x, y, l, horiz);
   }
   
   /**
    * Draws the slider thumb.
    * 
    * @param    xst
    *           The X start coordinate of the trackbar.
    * @param    yst
    *           The Y start coordinate of the trackbar.
    * @param    length
    *           The trackbar length.
    * @param    horiz
    *           The flag indicating if the orientation is horizontal or not.
    */
   private void drawThumb(int xst, int yst, int length, boolean horiz)
   {
      // calculate current position depending on widget value
      // base is the X or Y coord for the ruler start depending on orientation
      int base = horiz ? xst : yst;
      // pixel based thumb position within the ruler line
      int pos = indexToCoord(base, length, horiz, getScreenValueAsInt());
      
      // new thumb coordinates
      thumbX = horiz ? pos : xst;
      thumbY = horiz ? yst : pos;

      // the pointer drawing depends on orientation and tic marks mode
      if (tmMode == TicMarks.BOTH)
      {
         int x1 = horiz ? thumbX - thumbDim.width/2 : thumbX - thumbDim.height/2 + 2;
         int x2 = horiz ? x1 + thumbDim.width - 1: x1 + thumbDim.height - 1;
         int y1 = horiz ? thumbY - thumbDim.height/2 + 2 : thumbY - thumbDim.width/2;
         int y2 = horiz ? y1 + thumbDim.height - 1: y1 + thumbDim.width - 1;
         // 3d frame
         gd.setColor(gc.bgColor);
         gd.drawLine(x1, y1, x1, y2 - 1);
         gd.drawLine(x1 + 1, y1, x2 - 1, y1);
         gd.setColor(darkShadow);
         gd.drawLine(x2 - 1, y1 + 1, x2 - 1, y2 - 1);
         gd.drawLine(x1 + 1, y2 - 1, x2 - 2, y2 - 1);
         gd.setColor(veryDarkShadow);
         gd.drawLine(x2, y1, x2, y2);
         gd.drawLine(x1, y2, x2 - 1, y2);
         // and body
         gd.setColor(lightShadow);
         gd.fillRect(x1 + 1, y1 + 1, x2 - x1 - 2, y2 - y1 - 2);
      }
      else if (img != null)
      {
         // calculate top-left thumb coordinates depending on
         // orientation and thumb size
         int x = horiz ? thumbX - thumbDim.width/2 : thumbX - thumbDim.height/2;
         int y = horiz ? thumbY - thumbDim.height/2 : thumbY - thumbDim.width/2;
         // adding extra pixels depending on the tic marks mode
         if (tmMode == TicMarks.BOTTOM && horiz)
         {
            y += 3;
         }
         else if (tmMode == TicMarks.RIGHT && !horiz)
         {
            x += 3;
         }
         else if (tmMode == TicMarks.TOP && horiz)
         {
            y += 1;
         }
         else if (tmMode == TicMarks.LEFT && !horiz)
         {
            x += 1;
         }
         else if (tmMode == TicMarks.NONE)
         {
            if (horiz)
            {
               y += 3;
            }
            else
            {
               x += 3;
            }
         }
         // and draw the image loaded before
         gd.drawImage(img, x, y,
                      img.getWidth(), img.getHeight(), 0, 0, false, false, false, false);
      }
   }
   
   /**
    * Draws the slider tic marks.
    * 
    * @param    xst
    *           The X start coordinate of the trackbar.
    * @param    yst
    *           The Y start coordinate of the trackbar.
    * @param    length
    *           The trackbar length.
    */
   private void drawTicMarks(int xst, int yst, int length)
   {
      // get slider config
      final SliderConfig config = (SliderConfig) config();
      
      // tic marks not possible, do nothing
      if (tmMode == TicMarks.NONE || tmMode == TicMarks.UNKNOWN)
      {
         return;
      }
      
      // the length of the regular marks
      final int TM_LENGTH = 2;
      final int START_SHIFT = 11;
      
      // using the background color
      gd.setColor(gc.fgColor);
      
      // distance between rule and tic mark line start, for Y dimension in horizontal slider and
      // for X dimension in vertical slider respectively
      int shift = tmMode == TicMarks.BOTH
                     ? START_SHIFT - 1
                     : tmMode == TicMarks.BOTTOM || tmMode == TicMarks.RIGHT
                        ? START_SHIFT + 4 : START_SHIFT;
                           
      // draw tic marks lines depends on orientation
      if (config.horizontal)
      {
         // tics above ruler 
         if (tmMode == TicMarks.TOP || tmMode == TicMarks.BOTH)
         {
            // extra length marks on edges
            gd.drawLine(xst, yst - shift, xst, yst - shift - TM_LENGTH - 1);
            gd.drawLine(xst + length, yst - shift, xst + length, yst - shift - TM_LENGTH - 1);
            // regular tics between edges
            // when frequency == 0 there must be no tics between edges
            if (config.frequency > 0)
            {
               for (int i = config.minValue + config.frequency;
                    i < config.maxValue; i += config.frequency)
               {
                  int xt = indexToCoord(xst, length, true, i);
                  gd.drawLine(xt, yst - shift, xt, yst - shift - TM_LENGTH);
               }
            }
            // both mode need to change the shift
            if (tmMode == TicMarks.BOTH)
            {
               shift += 4;
            }
         }
         // tics below ruler
         if (tmMode == TicMarks.BOTTOM || tmMode == TicMarks.BOTH)
         {
            // extra length marks on edges
            gd.drawLine(xst, yst + shift, xst, yst + shift + TM_LENGTH + 1);
            gd.drawLine(xst + length, yst + shift, xst + length, yst + shift + TM_LENGTH + 1);
            // regular tics between edges
            // when frequency == 0 there must be no tics between edges
            if (config.frequency > 0)
            {
               for (int i = config.minValue + config.frequency;
                    i < config.maxValue; i += config.frequency)
               {
                  int xt = indexToCoord(xst, length, true, i);
                  gd.drawLine(xt, yst + shift, xt, yst + shift + TM_LENGTH);
               }
            }
         }
      }
      else
      {
         // tics left to the ruler
         if (tmMode == TicMarks.LEFT || tmMode == TicMarks.BOTH)
         {
            // extra length marks on edges
            gd.drawLine(xst - shift, yst, xst - shift - TM_LENGTH - 1, yst);
            gd.drawLine(xst - shift, yst + length, xst - shift - TM_LENGTH - 1, yst + length);
            // regular tics between edges
            // when frequency == 0 there must be no tics between edges
            if (config.frequency > 0)
            {
               for (int i = config.minValue + config.frequency;
                    i < config.maxValue; i += config.frequency)
               {
                  int yt = indexToCoord(yst, length, false, i);
                  gd.drawLine(xst - shift, yt, xst - shift - TM_LENGTH, yt);
               }
            }
            // both mode need to change the shift
            if (tmMode == TicMarks.BOTH)
            {
               shift += 4;
            }
         }
         // tics right ruler
         if (tmMode == TicMarks.RIGHT || tmMode == TicMarks.BOTH)
         {
            // extra length marks on edges
            gd.drawLine(xst + shift, yst, xst + shift + TM_LENGTH + 1, yst);
            gd.drawLine(xst + shift, yst + length, xst + shift + TM_LENGTH + 1, yst + length);
            // regular tics between edges
            // when frequency == 0 there must be no tics between edges
            if (config.frequency > 0)
            {
               for (int i = config.minValue + config.frequency;
                    i < config.maxValue; i += config.frequency)
               {
                  int yt = indexToCoord(yst, length, false, i);
                  gd.drawLine(xst + shift, yt, xst + shift + TM_LENGTH, yt);
               }
            }
         }
      }
   }
   
   /**
    * Draws the dotted frame border.
    * 
    * @param    orig
    *           The widget's starting point.
    * @param    dim
    *           The widget's size point.
    * @param    horiz
    *           The flag indicating if the orientation is horizontal or not.
    */
   private void drawFrame(NativePoint orig, NativeDimension dim, boolean horiz)
   {
      // set dotted line style
      gd.setLineStroke(LineStroke.DOTTED_SMALL);
      gd.setColor(gc.fgColor);
      gd.drawRect(orig.x, orig.y, dim.width - 1, dim.height - 1);
      // restore default line style
      gd.setLineStroke(LineStroke.DEFAULT);
   }
   
   /**
    * Draws the current widget screen value.
    * 
    * @param    orig
    *           The widget's starting point.
    * @param    dim
    *           The widget's size point.
    * @param    horiz
    *           The flag indicating if the orientation is horizontal or not.
    */
   private void drawCurrentValue(NativePoint orig, NativeDimension dim, boolean horiz)
   {
      // obtain the current value
      String current = getScreenValueAsString();
      if (current != null && !current.isEmpty())
      {
         // compute the text coordinates
         int xt = (horiz ? orig.x : orig.x + dim.width) + 1;
         int yt = (horiz ? orig.y : orig.y + dim.height) - 4;
         // set up font style
         FontDetails fntDetails = gf.font().font;
         gd.setGuiFont(fntDetails);
         gd.setFontStyle(fntDetails.style);
         // draw the current value
         gd.setColor(gc.fgColor);
         gd.drawString(current, xt, yt);
      }
   }
   
   /**
    * Loads scalable graphic image to draw the slider thumb shape.
    * 
    * @param    ticMarksMode
    *           The tic marks mode value.
    * @param    horiz
    *           The flag indicating if the orientation is horizontal or not.
    */
   private void initThumbImage(TicMarks ticMarksMode, boolean horiz)
   {
      // the name of the image, the base
      StringBuffer sbName = new StringBuffer("slider-tm");
      // different images are storing in different files
      if (ticMarksMode == TicMarks.TOP)
      {
         sbName.append("-top");
      }
      else if (ticMarksMode == TicMarks.BOTTOM)
      {
         sbName.append("-bottom");
      }
      else if (ticMarksMode == TicMarks.LEFT)
      {
         sbName.append("-left");
      }
      else if (ticMarksMode == TicMarks.RIGHT)
      {
         sbName.append("-right");
      }
      else if (horiz)
      {
         sbName.append("-none-bottom");
      }
      else
      {
         sbName.append("-none-right");
      }
      // check the 3d aware version
      if (isThreeD())
      {
         sbName.append("-3d");
      }
      // calculate image width and height
      int w = horiz ? thumbDim.width : thumbDim.height;
      int h = horiz ? thumbDim.height : thumbDim.width;
      img = VectorGraphicsHelper.getScalableImage(sbName.toString(), w, h);
   }
   
   /**
    * Calculate X or Y coordinate for given slider index tha must be in ramge from slider
    * min value to slider max value.
    * 
    * @param    start
    *           The X or Y coordinate to use as the base.
    * @param    length
    *           The length in pixels of the slider ruler.
    * @param    horizontal
    *           The flag indicating if the orientation is horizontal or not.
    * @param    ndx
    *           The value for the slider intenal position to transform to coord.
    * 
    * @return   The new coordinate, in X or Y dimension for the given slider index value.
    */
   private int indexToCoord(int start, int length, boolean horizontal, int ndx)
   {
      // preparing to precise calculations
      float baseDbl = (float) start;
      float lengthDbl = (float) length;
      
      // main proportion considering orientation and LARGE-TO-SMALL option
      float res = horizontal && !config.largeToSmall || !horizontal && config.largeToSmall  
         ? baseDbl + lengthDbl * (float)(ndx - config.minValue) / (float)(getRange() - 1)
         : baseDbl + lengthDbl - lengthDbl * (float)(ndx - config.minValue) /
                                             (float)(getRange() - 1);
      
      // perform honest rounding to get the result
      return Math.round(res);
   }
   
   /**
    * Initialize internal variables.
    */
   private void initInternals()
   {
      // already done, just leave
      if (initCompleted)
      {
         return;
      }
      
      // setting up the current tic marks mode
      if (config.ticMarks.equals(SliderConfig.TM_NONE))
      {
         tmMode = TicMarks.NONE;
      }
      else if (config.ticMarks.equals(SliderConfig.TM_TOP))
      {
         tmMode = TicMarks.TOP;
      }
      else if (config.ticMarks.equals(SliderConfig.TM_BOTTOM))
      {
         tmMode = TicMarks.BOTTOM;
      }
      else if (config.ticMarks.equals(SliderConfig.TM_LEFT))
      {
         tmMode = TicMarks.LEFT;
      }
      else if (config.ticMarks.equals(SliderConfig.TM_RIGHT))
      {
         tmMode = TicMarks.RIGHT;
      }
      else if (config.ticMarks.equals(SliderConfig.TM_BOTH))
      {
         tmMode = TicMarks.BOTH;
      }
      
      // setting up the thumb dimension depending on the current mode
      // no tic marks available
      if (tmMode == TicMarks.NONE || tmMode == TicMarks.UNKNOWN)
      {
         thumbDim = new NativeDimension(7, 12);
      }
      // special thumb size when tic marks are available
      else
      {
         thumbDim = new NativeDimension(11, 21);
      }
      
      // to avoid double init when already done
      initCompleted = true;
   }
}