/* 
** Module   : TextGuiImpl.java
** Abstract : static text implementation for GUI
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20141220 Created initial version.
** 002 CA  20150323 Fixed calculations of widget dimensions in GUI.
** 003 HC  20150323 Removed business logic from config classes.
** 004 CA  20150406 Fixed text height in non-3d mode.
** 005 CA  20150507 Draw text centered vertically.
** 006 GES 20150425 Eliminated AWT dependencies.
** 007 IAS 20150924 AUTO-RESIZE attribute support
** 008 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 009 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 010 HC  20151112 Implemented THREE-D runtime support.
** 011 CA  20160216 Cache the intermediate widget states.
** 012 HC  20161018 Added an initial support for rendering of mnemonic labels to all labeled
**                  widgets.
** 013 CA  20161101 Added support for mnemonic drawing.  Fixed issues related to TEXT's width.
** 014 OM  20170320 Added support for Themes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.Text;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.util.*;

/**
 * GUI-specific static text widget ({@link Text}) implementation.
 */
public class TextGuiImpl
extends Text<GuiOutputManager>
implements AutoResizable<TextConfig>
{
   /** Font resolver - for container purposes. */
   private GuiFontResolver gfd;

   /** Font resolver - for display purposes. */
   private GuiFontResolver gfi;

   /** Screen driver */
   private GuiDriver gd;

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public TextGuiImpl(WidgetId id, TextConfig cfg)
   {
      super(id, cfg, new GuiContentMnemonic());
   }

   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, TextConfig cfg)
   {
      super.initialize(id, cfg);

      boolean fixedFontDim = !config.flagStatic && config.appDataPres.isFixedFont();
      boolean fixedFontDisp = !config.flagStatic       && 
                              !config.explicitViewAs   && 
                              config.appDataPres.isFixedFont();

      gd = (GuiDriver) OutputManager.getDriver();

      gfd = new GuiFontResolver(WindowManager.resolveWindow(this), gd, config(), fixedFontDim);
      gfi = new GuiFontResolver(WindowManager.resolveWindow(this), gd, config(), fixedFontDisp);
      
      if (cfg.frameDefValue != null)
      {
         setValue(cfg.frameDefValue);
         
         // use only once
         cfg.frameDefValue = null;
      }

      updateSize();
   }

   /**
    * Process the given key, if this widget's mnemonic matches the key.
    * 
    * @param    key
    *           The pressed mnemonic key code.
    *           
    * @return   Always <code>false</code>, as TEXT widgets can't process the key.
    */
   @Override
   public boolean processMnemonic(int key)
   {
      return false;
   }
   
   /**
    * Draw this text.
    */
   @Override
   public void draw(Color color)
   {
      if (assigned && getState() == ScreenBuffer.UNINITIALIZED)
         setState(ScreenBuffer.UNCHANGED);
      
      Color dc = UiUtils.getDColor(this);

      selected = !Utils.equals(color, dc);
      
      final FontCache fc = gfi.font();
      final FontDetails<?> font = fc.font;

      int bgColor = isThreeD() ? SysColor.COLOR_3DFACE : SysColor.COLOR_FILLIN_BG;
      final GuiColorResolver gc = new GuiColorResolver(gd, config, 
                                                       bgColor,
                                                       SysColor.COLOR_FILLIN_FG);
      gc.colors();

      NativePoint p = physicalLocation();
      final NativeDimension dim = physicalDimension();
      NativeRectangle clip = clipRectangle(p, dim);
      
      final String txt = contentMnemonic.getMnemonic().getPrepLabel();
      
      gd.draw(getId(), p, clip, new Runnable()
      {
         @Override
         public void run()
         {
            // clear the area
            gd.setColor(gc.bgColor);
            gd.fillRect(0, 0, dim.width, dim.height);

            // draw full text
            gd.setGuiFont(font);
            gd.setColor(gc.fgColor);

            if (contentMnemonic != null               && 
                contentMnemonic.getMnemonic() != null &&
                contentMnemonic.getMnemonic().getMnemIdx() != null)
            {
               NativePoint tp = new NativePoint(0, dim.height);
               gd.drawStringWithMnemonic(txt, tp, contentMnemonic.getMnemonic(), gfi,
                                         TextGuiImpl.this, true);
            }
            else
            {
               gd.drawStringCentered(txt, 0, dim.height / 2);
            }
         }
      });
   }

   /**
    * Main worker routine for the text accessor.
    * 
    * @param   isInternal
    *          If this flag is set to <code>true</code> then alternative variable value is used 
    *          when available.
    *          
    * @return  Text representation of the variable value according to all current settings, 
    *          flags, format, etc.
    */
   @Override
   public String getText(boolean isInternal)
   {
      String display = getScreenValue(isInternal);

      if (isInternal && !config.flagStatic)
      {
         // obtain the initialization status
         boolean uninit = (getState() == ScreenBuffer.UNINITIALIZED);
         
         // If uninitialized and not for input and not date fill-in (in this case
         // separators should be displayed), display blanks instead.
         // TODO: probably it should be replaced by variable initialization
         //       in FrameHelper.clearComponents().
         if (uninit &&
             !(config.appFormat != null &&
               config.appFormat instanceof DateFormat &&
               isEnabled()))
         {
            display = "";
         }
      }
      
      return display;
   }
   
   /**
    * Get the AUTO-RESIZE status
    * 
    * @return <code>true</code> is AUTO-RESIZE is enabled
    */
   @Override
   public boolean isAutoResize()
   {
      return config.autoResize;
   }

   /**
    * Update widget's size after config update
    * 
    * @param beforeUpdate
    *        previous value of the widget's config
    */
   @Override
   public void updateSize(TextConfig beforeUpdate)
   {
      if (beforeUpdate != null && config.font != beforeUpdate.font)
      {
         updateSize();
      }
   }

   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public TextConfig config()
   {
      return (TextConfig) super.config();
   }

   /**
    * Get widget width.
    * 
    * @return  Widget width.
    */
   @Override
   public double width()
   {
      return screen().coordinates().widthFromNative(nativeWidth());
   }
   
   /**
    * Get widget height.
    * 
    * @return  Widget height.
    */
   @Override
   public double height()
   {
      return screen().coordinates().heightFromNative(nativeHeight());
   }
   
   /**
    * Retrieve the widget's width in native units.
    * 
    * @return   See above.
    */
   @Override
   protected int nativeWidth()
   {
      if (config.fixedWidth)
      {
         return config.widthPixels;
      }

      return nativeWidth(false);
   }
   
   /**
    * Retrieve the widget's height in native units.
    * 
    * @return   See above.
    */
   @Override
   protected int nativeHeight()
   {
      if (config.fixedHeight)
      {
         return config.heightPixels;
      }

      return nativeHeight(false);
   }

   /**
    * Retrieve the widget's width in native units.
    * 
    * @param    display
    *           <code>true</code> if this is called during drawing.
    *           
    * @return   See above.
    */
   private int nativeWidth(boolean display)
   {
      Window<?> window = WindowManager.resolveWindow(this);
      ScreenDriver<?> driver = OutputManager.getDriver();
      GuiFontResolver gf = (display ? gfi : gfd);

      if (gf == null || config == null || config.appDataPres == null)
         return 0;

      int font = gf.resolveFontNum();
      
      // literals case
      String txt = getText(true);
      boolean atField = isAtField();
      int width = 0;
      
      if (display || (!atField && config.flagStatic))
      {
         width = getTextWidth(txt, font);
      }
      else if (config.explicitViewAs || config.dynamic)
      {
         width = FillInGuiImpl.nativeWidth(this, gd, gf, true);
      }
      else
      {
         int charNo = atField ? config.atFormatLength : config.appFormat.getScreenWidth();
         
         width = FontManager.getFontWidth(window, font, driver) * charNo;
      }
      
      return width;
   }
   
   /**
    * Retrieve the widget's height in native units.
    * 
    * @param    display
    *           <code>true</code> if this is called during drawing.
    *           
    * @return   See above.
    */
   private int nativeHeight(boolean display)
   {
      Window<?> window = WindowManager.resolveWindow(this);
      ScreenDriver<?> driver = OutputManager.getDriver();
      GuiFontResolver gf = (display ? gfi : gfd);
      
      if (gf == null || config == null || config.appDataPres == null)
         return 0;

      int font = gf.resolveFontNum();
      // for text-widget height, always use the DEFAULT-FONT
      if (font == FontManager.DEFAULT_FIXED_FONT)
      {
         font = FontManager.DEFAULT_FONT;
      }
      String txt = getText(true);
      boolean threeD = FontManager.isThreeD();

      if (display)
      {
         // literals case
         return getTextHeight(txt, font);
      }
      else
      {
         return config.flagStatic && !config.explicitViewAs
                   ? screen().coordinates().heightToNative(1) - (threeD ? 2 : 0)
                   : FontManager.getFontHeight(window, font, driver);
      }
   }
   
   /**
    * Container class for the TEXT's content mnemonic.
    */
   protected static class GuiContentMnemonic
   extends ContentMnemonic<GuiOutputManager>
   {
      /**
       * Creates new {@link MnemonicInfoGuiImpl} instance.
       *
       * @return   {@link MnemonicInfoGuiImpl} instance.
       */
      @Override
      public MnemonicInfo createMnemonic()
      {
         return new MnemonicInfoGuiImpl(false);
      }
   }
}