/*
** Module   : ToggleBoxGuiImpl.java
** Abstract : GUI toggle-box implementation
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 VIG 20150210 Initial version.
** 002 GES 20150508 Eliminated AWT dependencies.
** 003 VIG 20150618 Extracted drawCheckmark method.
** 004 CA  20150826 Added mouse actions.
** 005 IAS 20150924 AUTO-RESIZE attribute support
** 006 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 007 VIG 20151005 Added isToggleKey method, defining keys, able to change the value.
**                  Added background drawing for PRESSED state, added parameter to check(..) call.
** 008 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 009 EVL 20151022 Fix for wrong size calculation for explicitly sized widget.
** 010 VIG 20151023 Fixed disposition of the label and highlight rectangle relative to check-box 
**                  and highlight rectangle coloring using XOR.
**     HC  20151112 Implemented THREE-D runtime support.
** 011 IAS 20151214 LOAD-MOUSE-POINTER support.
** 012 CA  20160216 Cache the intermediate widget states.
** 013 VIG 20160122 Highlight height is aligned corresponding to original, removed debug code.
** 014 HC  20161018 Fixed non-sensitive appearance of GUI toggle-box widget.
** 015 CA  20161228 Do not draw the highlight rectangle if there is no text.
** 016 CA  20161101 Added support for mnemonic processing and dawing.
** 017 OM  20170320 Added support for Themes.
** 018 OM  20170505 Fixed widget's width.
** 019 SVL 20170831 Can now be used as in-browse editor.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.util.*;

/**
 * Specific GUI implementation of TOGGLE-BOX widget.
 */
@SuppressWarnings("unchecked")
public class ToggleBoxGuiImpl
extends ToggleBox<GuiOutputManager>
implements AutoResizable<ToggleBoxConfig>
{
   /**
    * Color which is XORed with background to draw current row highlighting (dotted rectangle).
    */
   private final static ColorRgb CURRENT_ROW_XOR_COLOR = new ColorRgb(192, 192, 192);
   
   /** Color resolver */
   private GuiColorResolver gc;
   
   /** Screen driver */
   private GuiDriver gd;
   
   /** Font resolver */
   private GuiFontResolver gf;
   
   /** Pressed button state flag. */
   private boolean pressed = false;

   /** Browse widget to notify. */
   private Browse browse = null;
   
   /**
    * Constructor.
    *
    * @param   id
    *          The widget's ID.
    * @param   owner
    *          The parent container, frame or dialog which will include this toggle-box.
    * @param   text
    *          The string to be used as the toggle-box text.
    */
   public ToggleBoxGuiImpl(WidgetId id, Container<GuiOutputManager> owner, String text)
   {
      super(id, owner, text);
   }

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           Configuration data.
    */
   public ToggleBoxGuiImpl(WidgetId id, WidgetConfig cfg)
   {
      super(id, (ToggleBoxConfig) cfg);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    wcfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, ToggleBoxConfig wcfg)
   {
      super.initialize(id, wcfg);
      
      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      
      // font resolver
      gf = new GuiFontResolver(WindowManager.resolveWindow(this), gd, config());
   }

   /**
    * Calculate the width of this widget.
    *
    * @return    The width in characters.
    */
   @Override
   public double width()
   {
      Theme theme = ThemeManager.getCurrentTheme();
      // check if the width was set explicitly, return it instead
      if (config.widthChars > 0.0)
      {
         return config.widthChars;
      }
      
      Window<?> window = WindowManager.resolveWindow(this);
      int font = gf.resolveFontNum();
      int nativeW = theme.getCheckBoxSize() + 
                    theme.getCheckBoxSpace() +
                    FontManager.getTextWidth(textLabel(), window, font, gd) + 
                    3; /* space at the end, to allow the selection dotted line to be drawn */
      
      // update widget width if necessary
      if (config.widthPixels == 0.0 || config.widthPixels != nativeW)
      {
         config.widthPixels = nativeW; 
      }
      
      return window.screen().coordinates().widthFromNative(nativeW);
   }

   /**
    * Calculate the height of this widget.
    * 
    * @return    The height in rows.
    */
   @Override
   public double height()
   {
      // check if the height was set explicitly, return it instead
      if (config.heightChars > 0.0)
      {
         return config.heightChars;
      }
      
      Window<?> window = WindowManager.resolveWindow(this);
      int fontHeight = FontManager.getFontHeight(window, gf.resolveFontNum(), gd);
      int nativeH = Math.max(fontHeight, ThemeManager.getCurrentTheme().getCheckBoxSize());
      nativeH += 4; // Insets
      
      // update widget height if necessary
      if (config.heightPixels == 0.0 || config.heightPixels != nativeH)
      {
         config.heightPixels = nativeH; 
      }

      return window.screen().coordinates().heightFromNative(nativeH);
   }
   
   /**
    * Set the new text for the component.
    * 
    * @param   text
    *          New text value.
    */
   @Override
   public void setText(String text)
   {
      setTextLabel(text);

      repaint();
   }
   
   /**
    * Draw the toggle-box implementation. 
    */
   @Override
   public void draw()
   {
      int bgColor = isThreeD() ? SysColor.COLOR_3DFACE : SysColor.COLOR_FILLIN_BG;
      gc = new GuiColorResolver(gd, config, bgColor, SysColor.COLOR_BTNTEXT);
      
      // determine colors to use
      gc.colors();
      
      // get location and size
      final NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();

      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, new NativeDimension(d.width, d.height));
      
      gd.draw(getId(), p, clip, () ->
            ThemeManager.getCurrentTheme().drawToggleBox(
                  this, gd, gc, gf, 0, 0, d.width, d.height, pressed, false));
   }

   /**
    * Recalculates WIDTH attribute using actual label.
    */
   public void setWidthInt()
   {
      // method calls server-side to get legacy text metrics, so current changes could be sent.
      int nativeWidth = getTextWidth(config.label, gf.resolveFontNum());
      CoordinatesConversion cc = screen().coordinates();
      config.widthChars = cc.widthFromNative(nativeWidth) + 5;
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.  Simulate ENTER key typing.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      // simulate enter key
      if (isEnabled() && e.getButton() == MouseEvent.BUTTON1)
      {
         check(e);
         e.consume(); 
         this.requestFocus();
      }
   }
   
   /**
    * Notification of a mouse dragged event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      // TODO implement disabling during dragging
   }
   
   /**
    * Notification of a mouse pressed event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      boolean oldVal = pressed; 
      pressed = true;
      
      if (!oldVal)
      {
         repaint();
      }
   }

   /**
    * Notification of a mouse released event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      boolean oldVal = pressed; 
      pressed = false;
      
      if (oldVal)
      {
         repaint();
      }
   }

   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    *           Previous configuration values.
    */
   @Override
   public void afterConfigUpdate(ToggleBoxConfig beforeUpdate)
   {
      if (!StringHelper.equal(beforeUpdate.label, config().label) && config.autoResize)
      {
         setTextLabel(config().label);
      }
      
      super.afterConfigUpdateBase(beforeUpdate);
   }

   /**
    * Check if the widget supports LOAD-MOUSE-POINTER() method 
    * 
    * @return <code>true</code> if the widget supports 
    *         LOAD-MOUSE-POINTER() method
    */
   @Override
   public boolean supportsCustomMousePointer()
   {
      return true;
   }

   /**
    * Get the AUTO-RESIZE status
    * 
    * @return <code>true</code> is AUTO-RESIZE is enabled
    */
   @Override
   public boolean isAutoResize()
   {
      return config.autoResize;
   }

   /**
    * Update widget's size after config update
    * 
    * @param beforeUpdate
    *        previous value of the widget's config
    */
   @Override
   public void updateSize(ToggleBoxConfig beforeUpdate)
   {
      if (!StringHelper.equal(config.label, beforeUpdate.label) || 
          config.font != beforeUpdate.font)
      {
         width();
      }
   }
   
   /**
    * Process the given key, if this widget's mnemonic matches the key.
    * 
    * @param    key
    *           The pressed mnemonic key code.
    *           
    * @return   <code>true</code> if this widget accepted this mnemonic key.  Not enabled or not
    *           visible widget will not accept it.
    */
   @Override
   public boolean processMnemonic(int key)
   {
      if (processMnemonicImpl(key, this))
      {
         // send a value-changed event and process it
         config.checked = !config.checked;
         refresh();

         EventManager.postEvent(new KeyInput(this, EventType.KEY_TYPED, Keyboard.SE_VALUE_CHANGED));
         
         return true;
      }
      
      return false;
   }

   /**
    * Processes key events occurring on this object by dispatching them
    * to any registered KeyListener objects.
    *
    * @param    keyEvent
    *           The event to be handled.
    */
   @Override
   public void processKeyEvent(KeyInput keyEvent)
   {
      if (browse != null)
      {
         browse.notifyKeyPressedInEditor(keyEvent);

         int key = keyEvent.actionCode();
         if (key == Keyboard.KA_TAB)
         {
            browse.processKeyEvent(keyEvent);
            return;
         }
      }

      super.processKeyEventCommon(keyEvent);
   }

   /**
    * Link editor with the <code>Browse</code> widget.
    *
    * @param   browse
    *          The instance of the <code>Browse</code> to link with.
    */
   public void setBrowse(Browse browse)
   {
      this.browse = browse;
   }

   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvent.MOUSE_CLICKED, 
         MouseEvent.MOUSE_DRAGGED, 
         MouseEvent.MOUSE_PRESSED,
         MouseEvent.MOUSE_RELEASED
      };
   }

   /**
    * In GUI mode, the drawing needs to be delayed.
    */
   @Override
   protected void refresh()
   {
      repaint();
   }
   
   /**
    * Return <code>true</code> if specified key can change toggle-box value.
    * 
    * TODO: Toggle key (SPACE) PRESSED/RELEASE events work as corresponding MOUSE
    *       events: we must draw check box background as inactive when the key is pressed 
    *       and until it will be released. So we need to add support of RELEASE event 
    *       for keys and implement that drawing.
    *      
    * @see     com.goldencode.p2j.ui.client.driver.swing.SwingKeyboardReader#keyReleased
    * 
    * @param   key
    *          Key code being processed.
    * 
    * @return  <code>true</code> if specified key is a toggle key. 
    */
   protected boolean isToggleKey(int key)
   {
      return key == Key.VK_SPACE;
   }
}
