/*
** Module   : WindowTitleBar.java
** Abstract : GUI window title bar implementation
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.
** 002 MAG 20140902 Use GUI driver system color API to get default colors for widgets drawing.
** 003 HC  20140911 GUI coordinate support - data types refactoring.
** 004 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 005 MAG 20140929 Make class public.
** 006 CA  20141021 Finished support for the caption buttons, drawing, and associated system menu.
** 007 MAG 20140929 Eliminated call to AbstarctContainer.draw.
** 008 HC  20141021 Implemented window move/position support.
** 009 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 010 CA  20141123 Fixed drawing of window title bar, to reflect if the window is the one with
**                  current focus or not.  Changes in drawing of the caption buttons/icon/title to
**                  reflect the OS drawing.
** 011 EVL 20141204 Adding title bar icon changing capability.
** 012 EVL 20150108 Taking into loaded icon size difference from either LOAD-SMALL-ICON or
**                  LOAD-ICON sources.  Fix the default application icon to be set as gclogo.ico
**                  embedded into runtime jar file.  Rework for icon/image implementation.
** 013 CA  20150323 Pass the title-bar's ID when constructing the window title, so the window can 
**                  be resolved.
** 014 CA  20150415 If the title is not set, use the default title.
** 015 HC  20150323 Removed business logic from config classes.
** 016 CA  20150406 Implemented nativeWidth and nativeHeight for WindowTitle class (change required
**                  after adding new Label APIs).
** 017 CA  20150705 Draw title centered vertically.
** 018 HC  20150510 Improved drawing of caption buttons - they are now more Windows alike. 
** 019 GES 20150425 Eliminated AWT dependencies.
** 020 HC  20150612 Implemented support for modal windows.
** 021 HC  20150702 Improved positioning of top-level windows on systems with multiple displays.
** 022 CA  20150826 Added mouse actions.
** 023 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
**                  Changes to allow multi-window focus management in GUI.
**                  Fix for title bar size and painting issues.
** 024 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 025 HC  20151021 Added the ability to enable or disable caption buttons.
** 026 HC  20151213 Improved DIALOG-BOX window parenting logic and related changes.
** 027 EVL 20160122 Change call from driver to window manager to find out if the window is active.
** 028 HC  20160304 Replaced Window.moveBy() with setPhysicalLocation(). Window layout
**                  improvements.
** 029 CA  20160216 Cache the intermediate widget states.
** 030 VIG 20151229 Added tooltips initialization for caption buttons.
**                  Code cleanup and changed showPopupMenu() call.
** 031 HC  20160418 Fixed ALERT-BOX title bar close button to close the box with correct result.
** 032 HC  20160607 GUI menus reworked to be displayed in dedicated top-level windows.
** 033 SBI 20160610 Added registerWindowTitlePopup() to register mouse popupable regions for the
**                  Web client and triggerTitlePopup(AbstractWidget holder, MouseEvent e) to
**                  trigger an action to open a window title popup on a left mouse click.
** 034 HC  20160712 Improved next/prev tab focus in nested frames.
** 035 HC  20161018 Fixed abends in image processing and drawing when image offset was involved.
** 036 OM  20170320 Added support for Themes.
** 037 OM  20170623 Fixed title management.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.*;
import java.io.*;
import java.util.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.theme.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Implementation of a GUI window title bar.  This consists of (from left to right):
 * <ul>
 *    <li>an {@link #icon}</li>
 *    <li>a {@link #title}</li>
 *    <li>caption {@link #buttons}</li>
 * </ul>
 * Each element is a specific widget instance.
 */
public class WindowTitleBar
extends AbstractContainer<GuiOutputManager>
implements WidgetWithConfig<BaseConfig>,
           MouseListener,
           MouseMotionListener
{
   /** The window's icon. */
   WindowIcon icon;
   
   /** The window title. */
   WindowTitle title;

   /** The graphics driver. */
   private GuiDriver gd;

   /** The caption buttons. */
   private CaptionButton[] buttons;

   /** This widget configuration instance. */
   private BaseConfig config;
   
   /**
    * Basic c'tor.
    * 
    * @param    id
    *           The widget's ID.
    * @param    windowTitle
    *           The window title.
    */
   public WindowTitleBar(WidgetId id, String windowTitle)
   {
      this(id, true, new CaptionButtonType[]
         { 
            CaptionButtonType.ICONIFY, 
            CaptionButtonType.MAXIMIZE, 
            CaptionButtonType.CLOSE
         },
         windowTitle);
   }
   
   /**
    * Custom c'tor.
    * 
    * @param    id
    *           The widget's ID.
    * @param    showWindowIcon
    *           If <code>true</code> the window's window icon will be shown.
    * @param    buttonTypes
    *           Array of button types to create on the bar.
    * @param    windowTitle
    *           The window title.
    */
   public WindowTitleBar(WidgetId id,
                         boolean showWindowIcon,
                         CaptionButtonType[] buttonTypes,
                         String windowTitle)
   {
      gd = (GuiDriver) OutputManager.getDriver();
      
      config = ConfigManager.getInstance().resolveWidgetConfig(id, BaseConfig.class);
      
      if (showWindowIcon)
      {
         icon = new WindowIcon(WidgetId.nextID());
         
         byte[] appIcon = Window.getWindowDefaultIcon();
         
         // found the application icon defined on the server side
         if (appIcon != null)
         {
            // at this time we have either application icon or default icon loaded
            icon.setIconData(appIcon);
         }
            
         this.add(icon);
      }

      title = new WindowTitle(WidgetId.nextID());
      title.setTitle(windowTitle);
      this.add(title);

      buttons = new CaptionButton[buttonTypes.length];
      for (int i = 0; i < buttonTypes.length; i++)
      {
         CaptionButton btn = new CaptionButton(WidgetId.nextID(), buttonTypes[i]);
         this.add(btn);
         buttons[i] = btn;
         OutputManager.instance().getRegistry().addWidget((Widget) btn);
         ThinClient.setupTooltip(btn);
      }
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, BaseConfig cfg)
   {
      // no-op
   }

   /**
    * Get the widget config.
    * 
    * @return   See above.
    */
   @SuppressWarnings("unchecked")
   @Override
   public BaseConfig config()
   {
      return config;
   }

   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(BaseConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
   }

   /**
    * Layout the elements of this title.   Everything is done in pixel units.
    */
   public void doLayout()
   {
      double height = ThemeManager.getCurrentTheme().layoutTitleBar(
            topLevelWindow(),
            icon,
            title,
            buttons,
            this.width(),
            screen().coordinates()
      );
      setHeight(height);
   }
   
   /**
    * Register widgets regions for a right mouse click action and a left mouse click action
    * to open a window title popup widget. For the Web client it adds required popupable
    * mouse widgets regions and it does nothing for the Swing GUI client.
    */
   void registerWindowTitlePopup()
   {
      MouseHandler.registerWidgetAction(
               this.getId().asInt(),
               new SystemAction(this.getId().asInt(), MouseEvt.MOUSE_POPUPABLE_RIGHT, null));
      MouseHandler.registerWidgetAction(
               this.title.getId().asInt(),
               new SystemAction(this.title.getId().asInt(), MouseEvt.MOUSE_POPUPABLE_RIGHT, null));
      // check if the icon is present
      if (this.icon != null && this.icon.getId() != null)
      {
         MouseHandler.registerWidgetAction(
            this.icon.getId().asInt(),
            new SystemAction(this.icon.getId().asInt(), MouseEvt.MOUSE_POPUPABLE_LEFT, null));
      }
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.
    * <p>
    * If the widget is not already focused, it forces the focus to it.  After the focus was
    * received, it computes the caret location, based on the mouse position over the widget's
    * content.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      triggerTitlePopup(this, e);
   }
   
   /**
    * Trigger an action to open a window title popup.
    * 
    * @param    holder
    *           A widget that handles a mouse click event
    * @param    e
    *           The mouse event.
    */
   private void triggerTitlePopup(AbstractWidget holder, MouseEvent e)
   {
      if (e.getButton() == holder.getMenuMouse())
      {
         TopLevelWindow<GuiOutputManager> wnd = topLevelWindow();
         if (wnd != null)
         {
            Menu<GuiOutputManager> menu = WindowTitlePopupGuiImpl.create(wnd);
            menu.showPopup(this, new NativePoint(e.getX(), e.getY()));
         }
      }
   }
   
   /**
    * Find the widget positioned just bellow the specified mouse position (in physical units).
    * 
    * @param    p
    *           The mouse physical location.
    * 
    * @return   A widget over which the mouse is positioned or <code>null</code> if none found.
    */
   @Override
   public Widget<GuiOutputManager> findMouseSource(NativePoint p)
   {
      Widget<GuiOutputManager> w = super.findMouseSource(p);
      
      return (w != null ? w : this);
   }
   
   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[]
      {
         MouseEvt.MOUSE_MOVABLE,
         MouseEvt.MOUSE_CLICKED
      };
   }
   
   /**
    * Draw the title's background.
    */
   @Override
   public void draw()
   {
      Theme theme = ThemeManager.getCurrentTheme();
      
      // drawing is done only once, after each (re)layout
      WindowTitleBar bar = WindowTitleBar.this;
      NativePoint origin = bar.physicalLocation();
      final NativeDimension dim = bar.physicalDimension();
      final NativeRectangle clip = clipRectangle(origin, dim);
      
      // fill title bar
      gd.draw(getId(), origin, clip, () ->
            theme.drawWindowTitleBar(gd, 0, 0, dim.width, dim.height, isActiveWindow()));
      
      // draw contained widgets
      for (Widget<?> widget : widgets())
      {
         widget.draw();
      }
   }
   
   /**
    * Check if the widget implementation can raise legacy, 4GL-style, mouse events.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   public boolean raiseLegacyMouseEvents()
   {
      return false;
   }
   
   /**
    * Activates tooltips for caption buttons.
    */
   public void activateTooltips()
   {
      if (buttons != null)
      {
         for (CaptionButton btn : buttons)
         {
            ThinClient.getInstance().activateTooltipForWidget(btn.getId().asInt());
         }
      }
   }
   
   /**
    * Enabled or disables a caption button.
    * 
    * @param    button
    *           The button to enable/disable.
    * @param    enabled
    *           The new enabled state.
    */
   public void setCaptionButtonEnabled(CaptionButtonType button, boolean enabled)
   {
      for (CaptionButton b : buttons)
      {
         if (b.getType() == button)
         {
            b.setEnabled(enabled);
            break;
         }
      }
   }
   
   /**
    * Sets a click handler for any of the title bar buttons. 
    * 
    * @param button
    *        Button type.
    * @param handler
    *        Click handler.
    */
   public void setClickHandler(CaptionButtonType button, Runnable handler)
   {
      for (CaptionButton b : buttons)
      {
         if (b.getType() == button)
         {
            b.setClickHandler(handler);
            break;
         }
      }
   }
   
   /**
    * This component will not receive focus when Tab or Shift-Tab is pressed.
    * 
    * @return  <code>false</code>
    */
   @Override
   public boolean focusTraversable()
   {
      return false;
   }
   
   /**
    * Determine if the window associated with this bar is the active window.
    * 
    * @return   See above.
    */
   private boolean isActiveWindow()
   {
      TopLevelWindow<GuiOutputManager> w = topLevelWindow();
      
      // note that a top window is not necessarily active
      return WindowManager.isWindowActive(w);
   }
   
   /**
    * The window's icon.
    */
   class WindowIcon
   extends ImageGuiImpl
   implements MouseListener
   {
      /** Flag indicating the icon was loaded by LOAD-SMALL-ICON(). */
      private boolean useSmallIconSize = false;
      
      /**
       * The boolean value indicates that the icon is changed. The flag is used
       * to track setIconData usages.
       */
      private boolean isIconChanged;

      /**
       * Create the window icon.
       * 
       * @param    id
       *           The widget's ID.
       */
      public WindowIcon(WidgetId id)
      {
         super(id);
         // popup menu is opened on the left mouse click
         config.menuMouse = MouseEvent.BUTTON1;
      }
      
      /**
       * Get widget height.
       *
       * @return  widget height.
       */
      @Override
      public double height()
      {
         return screen().coordinates().heightFromNative(16);
      }
      
      /**
       * Get widget width.
       *
       * @return  widget width.
       */
      @Override
      public double width()
      {
         return screen().coordinates().widthFromNative(16);
      }
      
      /**
       * Check if the widget implementation can raise legacy, 4GL-style, mouse events.
       * 
       * @return   Always <code>false</code>.
       */
      @Override
      public boolean raiseLegacyMouseEvents()
      {
         return false;
      }
      
      /**
       * Sets or replaces the window icon data with provided array data.
       * 
       * @param    newIcon
       *           A new icon image array data.
       *
       * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
       */
      public boolean setIconData(byte[] newIcon)
      {
         // the new image candidate
         ImageWrapper<?> imgNew = gd.createIcon(newIcon, useSmallIconSize);
         
         // image loaded, replace current with new one 
         if (imgNew != null)
         {
            img = imgNew;
            isIconChanged = true;
            
            // repaint changed icon
            repaint();

            return true;
         }
         else
         {
            // if error while data processing - do nothing and return false
            return false;
         }
      }

      /**
       * Loads the new application icon for given window from the client side filesystem.
       * 
       * @param    small
       *           Flag indicating usage of the small icon case(16x16 in size).
       * @param    name
       *           The fully qualified file name for the icon to be loaded.
       *
       * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
       */
      public boolean loadIcon(boolean small, String name)
      {
         InputStream is = null;
         byte[] icon = null;
         
         // find the icon on local file system
         try
         {
            is = new FileInputStream(name);
         }
         catch (FileNotFoundException ex)
         {
            is = null;
         }
         
         // the file is found, try to read
         if (is != null)
         {
            icon = UiUtils.getData(is);
         }
         
         // if having new icon data - set it
         if (icon != null)
         {
            useSmallIconSize = small;
            return setIconData(icon);
         }
         else
         {
            return false;
         }
      }

      /**
       * Sets the window icon, if the icon is changed and then draws it on the graphic device.
       */
      @Override
      public void draw()
      {
         if (isIconChanged && img != null)
         {
            isIconChanged = false;
            Object[] drawingParameters = calculateDrawingParameters();
            int iw = (Integer) drawingParameters[0];
            int ih = (Integer) drawingParameters[1];
            int offsetX = (Integer) drawingParameters[2];
            int offsetY = (Integer) drawingParameters[3];
            boolean transparent  = (Boolean) drawingParameters[4];
            boolean stretchToFit = (Boolean) drawingParameters[5];
            boolean retainShape  = (Boolean) drawingParameters[6];
            boolean convert3D    = (Boolean) drawingParameters[7];
            /*
             * Sets the window icon. The drawing parameters are due to the drawing operations are
             * performed on the server side for the web rendering engine implementation.
             */
            gd.setIcon(img, 0, 0, iw, ih, offsetX, offsetY, transparent,
                      stretchToFit, retainShape, convert3D);
         }

         super.draw();
      }

      /**
       * Open a title popup menu.
       * 
       * @param    e
       *           A mouse click event that triggers an action to open a window title popup menu.
       */
      @Override
      public void mouseClicked(MouseEvent e)
      {
         WindowTitleBar.this.triggerTitlePopup(this, e);
      }

      /**
       * The method returns a value indicating whether this image is supposed to be stretched-to-fit.
       *
       * @return   See above.
       */
      @Override
      protected boolean isStretchToFit()
      {
         // we want the icon to always fit into the designated area
         return true;
      }
   }
   
   /**
    * The window's title.
    */
   private class WindowTitle
   extends Label<GuiOutputManager>
   {
      /** The cached window title */
      private String title;
      
      /** True value indicates that the title is changed and the graphics device must be notified.*/
      private boolean titleIsChanged;
      
      /**
       * Basic c'tor.
       * 
       * @param    id
       *           The widget's ID.
       */
      public WindowTitle(WidgetId id)
      {
         super(id, "", WindowTitleBar.this.getId().asInt());
      }

      /**
       * Returns the window title.
       * 
       * @return   The title
       */
      public String getTitle()
      {
         return title;
      }

      /**
       * Sets the window title.
       * 
       * @param    title
       *           The title to set
       */
      public void setTitle(String title)
      {
         this.titleIsChanged = !Objects.equals(this.title, title);
         this.title = title;
      }

      /**
       * Draw this window title.
       */
      @Override
      public void draw()
      {
         Theme theme = ThemeManager.getCurrentTheme();
         
         TopLevelWindow<GuiOutputManager> window = topLevelWindow();
         String windowTitle = window.title();
         if (!Objects.equals(this.title, windowTitle))
         {
            // detected a change in window title, we must update and display it 
            setTitle(windowTitle);
         }
         
         String title = (windowTitle == null ? WindowGuiImpl.DEFAULT_TITLE : windowTitle);
         
         if (titleIsChanged || !title.equals(getTitle()))
         {
            // notify client about new title
            gd.setTitle(title);
            titleIsChanged = false;
         }
         NativePoint origin = physicalLocation();
         final NativeDimension d = physicalDimension();
         final NativeRectangle clip = clipRectangle(origin, d);
         
         config.font = window.resolveTitleFont();
         final GuiFontResolver gf = new GuiFontResolver(window.window(), gd, config);
         
         gd.draw(getId(), origin, clip, () ->
               theme.drawWindowTitle(gd, gf, 0, 0,  d.width, d.height, title, isActiveWindow()));
      }
      
      /**
       * Draw caret - this is a no-op.
       */
      @Override
      public void drawCaret()
      {
         // no-op
      }
      
      /**
       * Find the widget to which the given mouse event can be applied.
       * 
       * @param    p
       *           The mouse physical location.
       * 
       * @return   A widget over which the mouse is positioned or the associated 
       *           {@link WindowTitleBar} if none found.
       */
      @Override
      public Widget<GuiOutputManager> findMouseSource(NativePoint p)
      {
         Widget<GuiOutputManager> w = super.findMouseSource(p);
         
         return (w == null ? WindowTitleBar.this : w);
      }
      
      /**
       * Check if the widget implementation can raise legacy, 4GL-style, mouse events.
       * 
       * @return   Always <code>false</code>.
       */
      @Override
      public boolean raiseLegacyMouseEvents()
      {
         return false;
      }
      
      /**
       * Retrieve the widget's width in native units.
       * 
       * @return   See above.
       */
      @Override
      protected int nativeWidth()
      {
         return screen().coordinates().widthToNative(width());
      }
      
      /**
       * Retrieve the widget's height in native units.
       * 
       * @return   See above.
       */
      @Override
      protected int nativeHeight()
      {
         return screen().coordinates().heightToNative(height());
      }
   }
}
