/*
** Module   : AbstractGuiDriver.java
** Abstract : simple base class for GUI screen driver implementation
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20150401 Created initial version.
** 002 CA  20150520 Changed signature for scaleFont; also, it was made final.
** 003 HC  20150612 Implemented support for modal windows and paragraph layout.
** 004 HC  20150624 Improved Widget Browser GUI highlight. Restoring the highlighted area was
**                  causing artifacts. Also the highlighted area was frozen for draw updates.
** 005 HC  20150702 Implemented window modality independently on the actual UI driver. Improved
**                  positioning of top-level windows on systems with multiple displays.
** 006 GES 20150730 Removed translatedOrigin() since it isn't actually used. translate() was
**                  converted to translatePush()/translatePop() to make it clear that they need
**                  to have matched pairs. A stack of changes is kept to reduce mistakes (the
**                  pop always knows the values to reverse). Added caching for text and paragraph
**                  metrics.
** 007 HC  20150831 Scrolling support overhaul.
** 008 GES 20150817 Removed isOnTop() since it is unnecessary.  Added an abstract method for
**                  window restacking.  Editable and Movable actions are moved at the driver level.
**                  deriveFont() now is driver specific - it scales the font depending on the DPI.
**                  Implemented setIcon(). All windows listen to any mouse actions, at any time.
**                  Mouse actions can not be disabled during drawing, as on slower machines it
**                  might mean ignoring posted mouse released/clicked events, as P2J is busy
**                  drawing/processing the mouse pressed event.
** 009 GES 20150925 Eliminate an abend that occurred when we tried to switch windows during a
**                  drawing batch. Fixed multi-window drawing issues caused by concurrent drawing
**                  access to the GuiDriver: all access to the driver is done exclusively, each
**                  thread taking and releasing ownership for the duration of the drawing
**                  operation or other GUI API invocation which requires access to the underlying
**                  physical window.
** 010 SVL 20151013 Added setXORComposite and resetComposite.
** 011 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 012 CA  20151013 Added support for window resize, z-order operations for the GUI web client.
** 013 CA  20151024 Added support for WINDOW:SENSITIVE attribute.  APIs using String for drawing
**                  or measuring are a no-op if the string is null or empty.
**     EVL 20151106 Adding grey image creation for the given original.
** 014 HC  20151213 Improved DIALOG-BOX window parenting logic and related changes.
** 015 EVL 20160107 Adding overlay window positioning.
** 016 IAS 20160107 LOAD-MOUSE-POINTER and SET-WAIT-STATE support
** 017 CA  20160216 Added optional widget ID parameter to draw, which will let the driver record
**                  and cache the state of this widget.
**                  Added font and text-widths APIs.  Fixed window deadlock in [de]iconifyWindow() 
**                  APIs.
**     SBI 20160324 Added setDesktopBgColor(ColorRgb color).
** 018 HC  20160407 Overhaul of window-activation logic. Removed setOverlayWindowLocation(), the
**                  method was unnecessary and was causing the COMBO-BOX drop-down location
**                  incorrect.
** 019 SBI 20160506 Fixed NPE if a user minimizes a target web window.
** 020 VIG 20160620 Fixed clipping rectangle rclip in draw().
** 021 HC  20160628 Prevented unhandled exceptions when window/frame closed/destroyed.
** 022 VIG 20160707 Added NPE check to enableEvents().
** 023 CA  20170104 Added flag to track window's modal state.
** 024 EVL 20170131 Fix for multi-image icons loading.
**     HC  20170202 Implemented runtime support for WINDOW-STATE. The implementation is missing
**                  support for WINDOW-DELAYED-MINIMIZE.
**     CA  20161101 Added API for drawing text with mnemonic.
**     HC  20170203 Prevented unexpected exception when a previous window emulator is being
**                  restored as a result of releaseWindow() call and the window has been
**                  destroyed since.
**     HC  20170223 Improved clipping to allow to clip with arbitrary set of overlapping or non-
**                  overlapping regions.
** 025 OM  20170215 Moved window resize and location messages to Swing dispatch thread.
** 026 OM  20170502 Dropped getUiTheme() (dead code).
** 027 OM  20170518 Small javadoc fixes.
** 028 EVL 20170608 Adding function to register direct manipulation handler and method to draw
**                  or clean selection box..
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;

// WARNING: these AWT/ImageIO dependencies are being intentionally used in this generic GUI code
//          because there is no better alternative for non-Swing drivers and this code works
//          fine in headless mode
import java.awt.image.BufferedImage;

import javax.imageio.*;
import javax.imageio.stream.*;
import java.io.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.logging.*;
import com.goldencode.cache.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Simple base class for GUI-specific screen driver implementation features. ChUI-specific
 * features are not to be implemented here.  This class is intended to be specific to a given
 * UI-modality (GUI in this case). However, no driver-specific implementation classes or features
 * should be present.  For example, no Swing-specific feature usage is allowed.
 * <p>
 * The only exceptions to this no-Swing rule are for events and image loading.  We are using
 * the AWT event classes as simple containers to eliminate the need to create our own versions
 * of the same thing. For image loading, there is no better solution than to use ImageIO and
 * AWT. 
 */
public abstract class AbstractGuiDriver<F>
extends AbstractDriver<GuiPrimitives, GuiOutputManager>
implements GuiDriver<F, BufferedImage>
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(AbstractGuiDriver.class.getName());
   
   /** Currently selected window emulator. */
   protected EmulatedWindowState ews = null;
   
   /** The real location of the translated origin. */
   protected NativePoint translatedOrigin = new NativePoint(0, 0);

   /** All current translate() operations in proper nested order. */
   private Deque<NativePoint> translations = new ConcurrentLinkedDeque<>();
   
   /** Pending tooltip workers to be registered with mouse handler. */
   private Map<Widget, MouseWidgetAction> pendingTtWorkers = new HashMap<>();
   
   /** System colors cache. */
   private Map<Integer, ColorRgb> systemColorsCache = new HashMap<>();

   /** The currently used theme for GUI widgets. */
   private String theme = null;
   
   /** A cache of resolved text metrics. */
   private LFUAgingCache<TextKey, TextMetrics> textMetricsCache = new LFUAgingCache<>(8192);
   
   /** A cache of resolved paragraph metrics. */
   private LFUAgingCache<ParagraphKey, TextMetrics> 
                                                paragraphMetricsCache = new LFUAgingCache<>(8192);

   /** Track batch mode, to not allow window switching while batch mode is active. */
   private Deque<Boolean> batch = new ArrayDeque<>();
   
   /**
    * The last posted clipping rectangles. If next clipping rectangles are the same as these
    * ones, then they will not be posted. The list gets reset when the origin is translated.
    */
   private Set<NativeRectangle> lastClips = null;

   /**
    * The {@link Thread} having exclusive access to the GUI driver primitives.  If this is 
    * <code>null</code>, any thread can gain exclusive access immediately in {@link #selectWindow}.
    * Otherwise, it will wait for the owner to finish.
    */
   private Thread drawingOwner = null;
   
   /** 
    * Stack indicating if last call to {@link #selectWindow} ended in a window switch.  
    * This indicates when {@link #releaseWindow} will release the GUI driver exclusive access hold
    * by the current {@link #drawingOwner}.
    */
   private Deque<Boolean> windowSwitched = new ArrayDeque<>();

   /** The selected window stack by the {@link #drawingOwner}. */
   private Deque<EmulatedWindowState<F, ?>> lastWindow = new ArrayDeque<>();

   /**
    * Construct a new instance.
    *
    * @param    direct
    *           The output primitives.
    */
   public AbstractGuiDriver(GuiPrimitives direct)
   {
      super(new GuiWidgetFactory(), direct, false, "win3");
      
      batch.push(false);
   }
   
   /**
    * Get the {@link OutputManager} instance associated with this driver.
    * 
    * @param    cfg
    *           The initial configuration for this driver.
    *
    * @return   A {@link GuiOutputManager} instance.
    */
   @Override
   public GuiOutputManager createOutputManager(BootstrapConfig cfg)
   {
      return new GuiOutputManager(this, cfg);
   }
   
   /**
    * Clear the entire screen, with the default window color.
    */
   @Override
   public void clear()
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.CLEAR);
      ps.x = 0;
      ps.y = 0;
      ps.width = ews.getWidth();
      ps.height = ews.getHeight();
      ps.color = null; // TODO: color???
      
      ews.offer(ps);
   }

   /**
    * Create a native image from the raw bytes and return a generic wrapper that can be used
    * for operations.
    *
    * @param    bytes
    *           The image as persisted in the file system or a JAR file.
    *
    * @return   The generic wrapper representing the given image or <code>null</code> if the
    *           image cannot be created (e.g. if the bytes are not a valid image type).
    */
   public static ImageWrapper<BufferedImage> createImage(byte[] bytes)
   {
      // the new image candidate
      BufferedImage        img  = null;
      ByteArrayInputStream bais = null;

      // try creating new image
      try
      {
         bais = new ByteArrayInputStream(bytes);
         img  = ImageIO.read(bais);
      }
      catch (IOException e)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, "Could not load image from byte array...", e);
         }
      }
      finally
      {
         if (img == null && bais != null)
         {
            try
            {
               bais.close();
            }
            catch (IOException ioe)
            {
               // ignore, don't care
            }
         }
      }
      
      return img == null ? null : new ImageWrapper<>(img, img.getHeight(), img.getWidth());
   }
   
   /**
    * Create image from the binary data
    * 
    * @param data
    *        image data
    * @return
    *        image
    *        
    * @throws IOException
    */
   public BufferedImage makeImage(byte[] data)
   throws IOException
   {
      return ImageIO.read(ImageIO.createImageInputStream(data)); 
   }

   /**
    * Create a native grey image from the original initial image to be used for disabled control
    * states.
    *
    * @param    imageIni
    *           The image to be used as source for grey transformation.
    *
    * @return   The generic wrapper representing the given image or <code>null</code> if the
    *           image cannot be created (e.g. if the bytes are not a valid image type).
    */
   public ImageWrapper<BufferedImage> createGreyImage(BufferedImage imageIni)
   {
      BufferedImage        img = null;

      // get the image helper
      BufferedImageDrawHelper bidh = ews.getBufferedImageDrawHelper();
      if (bidh != null)
      {
         img = bidh.createDisabledGreyImage(imageIni,
                                            getSysColor(SysColor.COLOR_3DSHADOW).getGuiRgb(),
                                            getSysColor(SysColor.COLOR_WINDOW).getGuiRgb());
      }
      
      return img == null ? null
                         : new ImageWrapper<BufferedImage>(img, img.getHeight(), img.getWidth());
   }
   
   /**
    * Create a native icon from the raw bytes and return a generic wrapper that can be used
    * for operations.
    *
    * @param    byts
    *           The image as persisted in the file system or a JAR file.
    * @param    small
    *           <code>true</code> if the small icon size is to be loaded.
    *
    * @return   The generic wrapper representing the given icon or <code>null</code> if the
    *           icon cannot be created (e.g. if the bytes are not a valid icon type).
    */
   @Override
   public ImageWrapper<BufferedImage> createIcon(byte[] byts, boolean small)
   {
      BufferedImage        img = null;
      ByteArrayInputStream bais   = null;
      ImageReader          reader = null;
      int targetBppQuality = 0;
      
      try
      {
         bais = new ByteArrayInputStream(byts);
         
         ImageInputStream iis = ImageIO.createImageInputStream(bais);
         
         // get the list of readers
         Iterator<ImageReader> readers = ImageIO.getImageReaders(iis);

         if (readers.hasNext())
         {
            // the selected image number to load
            int iconNdx = -1;
            
            // get the reader, we only ever use the first one
            reader = readers.next();

            // connect reader to input stream
            reader.setInput(iis);
            
            // get the icon numbers in a stream allowing internal scanning
            int numIcons = reader.getNumImages(true);
            
            // calculate the size of the icon we are searching for
            int targetSize = small ? WindowConfig.SMALL_ICON_SIZE : WindowConfig.ICON_SIZE;
            
            // check all icons
            for (int i = 0; i < numIcons; i++)
            {
               int iconSize = Math.max(reader.getWidth(i), reader.getHeight(i));
               int bppQuality = reader.getRawImageType(i).getColorModel().getPixelSize();
               
               if (iconSize == targetSize && bppQuality > targetBppQuality)
               {
                  // found the required image, read it and exit icon searching loop
                  iconNdx = i;
                  targetBppQuality = bppQuality;
               }
            }
            
            // if image found, read the data
            if (iconNdx != -1)
            {
               img = reader.read(iconNdx);
            }
         }

         // have not found required icon size or not icon, use first available image 
         if (img == null)
         {
            if (bais != null)
            {
               bais.close();
            }
            
            bais = new ByteArrayInputStream(byts);
            
            // if this succeeds, there is no need to close the stream, but if this returns
            // null then the stream needs closing
            img  = ImageIO.read(bais);
         }
      }
      
      catch (IOException e)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, "Could not set icon from array...", e);
         }

         img = null;
      }
      
      finally
      {
         if (reader != null)
         {
            reader.dispose();
         }
         
         if (bais != null)
         {
            try
            {
               bais.close();
            }
            
            catch (IOException ioe)
            {
               // ignore, don't care
            }
         }
      }
      
      return img == null ? null
                         : new ImageWrapper<BufferedImage>(img, img.getHeight(), img.getWidth());
   }
   
   /**
    * Get driver-specific child process factory.
    * 
    * @return   The reference to the factory.
    */
   @Override
   public ChildProcessFactory getChildProcessFactory()
   {
      return new ChildProcessFactory()
      {
         /**
          * Obtains a new instance of an interactive child process.
          *
          * @return   The new child process instance.
          */
         public InteractiveChildProcess getChildProcess()
         {
            // TODO: this requires a ChUI screen driver, to execute the emulator in it
            return null;
         }
      };
   }

   /**
    * Get BGCOLOR color. Get a RGB value from COLOR-TABLE at the specified index.
    * 
    * @param    index
    *           Color number.
    *           
    * @return   instance of ColorRgb
    */
   @Override
   public ColorRgb getBgColor(int index)
   {
      ColorRgb rgb = ColorManager.DEFAULT_RGB;
      if (ews != null)
      {
         int windowId = ews.getWindowId();
         Window<?> window = WindowManager.findWindow(windowId).window();
         
         if (window != null)
         {
            rgb = ColorManager.getBgColor(window, index);
         }
      }
      
      return rgb;
   }

   /**
    * Get FGCOLOR color. Get a RGB value from COLOR-TABLE at the specified index.
    * 
    * @param   index
    *          Color number.
    *          
    * @return  instance of ColorRgb
    */
   @Override
   public ColorRgb getFgColor(int index)
   {
      ColorRgb rgb = ColorManager.DEFAULT_RGB;
      
      if (ews != null)
      {
         int windowId = ews.getWindowId();
         Window<?> window = (Window<?>) WindowManager.findWindow(windowId).window();
         
         if (window != null)
         {
            rgb = ColorManager.getFgColor(window, index);
         }
      }
      
      return rgb;
   }
 
   /**
    * Get a system color.  Because system colors cannot be change at runtime a cache is used in
    * order to optimize the creation of <code>ColorRgb</code> instances.  
    * 
    * @param   index
    *          Color index.
    *          
    * @return  see above.
    */
   @Override
   public ColorRgb getSysColor(int index)
   {
      // first lookup in cache
      ColorRgb color = systemColorsCache.get(index);
      
      if (color == null)
      {
         // if not found in cache get from system colors table.
         color = ColorManager.getSysColor(index);
         
         // cache value
         systemColorsCache.put(index, color);
      }
      
      return color;
   }

   /**
    * Get foreground color for a standard color like NORMAL, INPUT or MESSAGES.
    * 
    * @param   stdColor
    *          Color name.
    *          
    * @return  see above.
    */
   @Override
   public ColorRgb getFgStdColor(String stdColor)
   {
      ColorRgb rgb = ColorManager.DEFAULT_RGB;
      
      int windowId = ews.getWindowId();
      Window<?> window = (Window<?>) WindowManager.findWindow(windowId).window();
      
      if (window != null)
      {
         rgb = ColorManager.getStdColor(window, stdColor, true);
      }
      
      return rgb;
   }

   /**
    * Get background color for a standard color like NORMAL, INPUT or MESSAGES.
    * 
    * @param    stdColor
    *           Color name.
    *           
    * @return   RGB value.
    */
   @Override
   public ColorRgb getBgStdColor(String stdColor)
   {
      ColorRgb rgb = ColorManager.DEFAULT_RGB;
      
      int windowId = ews.getWindowId();
      Window<?> window = (Window<?>) WindowManager.findWindow(windowId).window();
      
      if (window != null)
      {
         rgb = ColorManager.getStdColor(window, stdColor, false);
      }
      
      return rgb;
   }
   
   /**
    * Translates the origin of the graphics context to the point (<i>x</i>,&nbsp;<i>y</i>) in the
    * current coordinate system.  The modified context will have its new origin corresponding to
    * the point (<i>x</i>,&nbsp;<i>y</i>) in the context's former coordinate system.  All  
    * coordinates used in subsequent rendering operations on this graphics context are relative
    * to this new origin.
    * <p>
    * All usage of this method MUST be paired with a subsequent {@link #translatePop}. A stack
    * of translations is maintained, so that translate calls can be nested.
    *
    * @param    x 
    *           the specified x coordinate
    * @param    y 
    *           the specified y coordinate
    */
   @Override
   public void translatePush(int x, int y)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.TRANSLATE_PUSH);
      ps.x = x;
      ps.y = y;
      
      ews.offer(ps);
      
      // track the current operation so that it is easy to reverse
      translations.push(new NativePoint(x, y));

      // this is mutable...
      translatedOrigin.translate(x, y);

      // let the next CLIP process, as we are translating.
      lastClips = null;
   }
   
   /**
    * Reverses the translation at the top of the stack.
    */
   public void translatePop()
   {
      NativePoint op = translations.isEmpty() ? null : translations.pop();
      
      if (op == null)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, "Unbalanced translate() operations!", new RuntimeException());
         }
         
         return;
      }
      
      PaintStructure ps = new PaintStructure(PaintPrimitives.TRANSLATE_POP);
      ps.x = -op.x;
      ps.y = -op.y;
      
      ews.offer(ps);
      
      // this is mutable...
      translatedOrigin.translate(-op.x, -op.y);
   }
   
   /**
    * Draw a string with a mnemonic.
    * 
    * @param    text
    *           The full text, which needs to be drawn.
    * @param    tp
    *           The text's coordinates.   If <code>centered</code> parameter is set to 
    *           <code>true</code>, then the {@link NativePoint#y} will represent the height on 
    *           which centering will be done.
    * @param    info
    *           The mnemonic details.
    * @param    gf
    *           The widget's font details.
    * @param    widget
    *           The widget instance to which the text belongs.
    * @param    centered
    *           Flag indicating if the text needs to be centered vertically.
    */
   @Override
   public void drawStringWithMnemonic(String                            text,
                                      NativePoint                       tp,
                                      MnemonicInfo                      info,
                                      GuiFontResolver                   gf,
                                      AbstractWidget<GuiOutputManager>  widget,
                                      boolean                           centered)
   {
      NativePoint p = new NativePoint(tp);
      FontCache fc = gf.font();
      
      Integer ampIdx = info.getMnemIdx();
      
      String leftPart = text.substring(0, ampIdx);
      if (!leftPart.isEmpty())
      {
         if (centered)
         {
            drawStringCentered(leftPart, p.x, p.y);
         }
         else
         {
            drawString(leftPart, p.x, p.y);
         }
         p.x += widget.getTextWidth(leftPart, gf.resolveFontNum());
      }
      
      String mnem = text.substring(ampIdx, ampIdx + 1);
      
      // set temporarily underlined font
      setFontStyle(FontStyle.createStyle(fc.font.style.isBold(), fc.font.style.isItalic(), true));
      if (centered)
      {
         drawStringCentered(mnem, p.x, p.y);
      }
      else
      {
         drawString(mnem, p.x, p.y);
      }
      p.x += widget.getTextWidth(mnem, gf.resolveFontNum());
       
      setGuiFont((FontDetails<F>) fc.font);
      setFontStyle(fc.font.style);
      
      String rightPart = text.substring(ampIdx + 1);
      if (!rightPart.isEmpty())
      {
         if (centered)
         {
            drawStringCentered(rightPart, p.x, p.y);
         }
         else
         {
            drawString(rightPart, p.x, p.y);
         }
      }
   }
   
   /**
    * Draw a string at the specified location.
    * 
    * @param    text
    *           text string to draw.
    * @param    x
    *           left position.
    * @param    y
    *           top position.
    */
   @Override
   public void drawString(String text, int x, int y)
   {
      if (isEmptyText(text))
      {
         return;
      }
      
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_STRING);
      ps.text = text;
      ps.x = x;
      ps.y = y;
      
      ews.offer(ps);
   }
   
   /**
    * Draw a string using the current font, which will be scaled so that the final string will
    * have the specified legacy width and height.
    * <p>
    * The string will be drawn at the specified location.
    * 
    * @param    text
    *           text string to draw.
    * @param    legacyWidth
    *           The legacy width, which needs to be matched when drawing.
    * @param    legacyHeight
    *           The legacy width, which needs to be matched when drawing.
    * @param    x
    *           left position.
    * @param    y
    *           top position.
    */
   @Override
   public void drawStringScaled(String text, int legacyWidth, int legacyHeight, int x, int y)
   {
      if (isEmptyText(text))
      {
         return;
      }

      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_STRING_SCALED);
      ps.text = text;
      ps.x = x;
      ps.y = y;
      ps.width = legacyWidth;
      ps.height = legacyHeight;
      
      ews.offer(ps);
   }

   /**
    * Draw a string using the current font, which will be scaled so that the final string will
    * have the specified legacy width and height.
    * <p>
    * The string will be drawn at the specified location.
    * 
    * @param   text
    *          text string to draw.
    * @param   legacyWidth
    *          The legacy width, which needs to be matched when drawing.
    * @param   legacyHeight
    *          The legacy width, which needs to be matched when drawing.
    * @param   x
    *          left position.
    * @param   height
    *          the height to center on.
    */
   @Override
   public void drawStringCentered(String text, int legacyWidth, int legacyHeight, int x, int height)
   {
      if (isEmptyText(text))
      {
         return;
      }

      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_STRING_SCALED);
      ps.text = text;
      ps.x = x;
      ps.y = height;
      ps.width = legacyWidth;
      ps.height = legacyHeight;
      ps.centered = true;
      
      ews.offer(ps);
   }
   
   /**
    * The method performs a layout operation on the supplied text and 
    * returns the resulting paragraph height while maintaining the
    * supplied maximum width.
    * <p>
    * Note that the algorithm for text layout is compatible with 
    * {@link #drawParagraph(String, FontDetails, int, int, int)} and so
    * can be used in conjunction with this method. 
    * 
    * @param    text
    *           The text to measure.
    * @param    font
    *           The font to use during layout.
    * @param    maxWidth
    *           The maximum paragraph width.
    */
   @Override
   public int getParagraphHeight(String text, FontDetails<F> font, int maxWidth)
   {
      if (text == null)
      {
         return 0;
      }

      TextMetrics metrics = resolveParagraphMetrics(text, font, maxWidth);

      return metrics.height;
   }
   
   /**
    * Draws a multiline text (paragraph) at the supplied position and 
    * with the specified font. The paragraph width will not exceed the
    * supplied maximum width. The text will be properly broken into
    * multiple lines.
    * <p>
    * Note that the algorithm for text layout is compatible with 
    * {@link #getParagraphHeight(String, FontDetails, int)} and so
    * can be used in conjunction with this method.
    *
    * @param    text
    *           The text to draw.
    * @param    font
    *           The drawing font.
    * @param    x
    *           The x position.
    * @param    y
    *           The y position.
    * @param    maxWidth
    *           The maximum paragraph width.
    */
   @Override
   public void drawParagraph(String text, FontDetails<F> font, int x, int y, int maxWidth)
   {
      if (isEmptyText(text))
      {
         return;
      }

      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_PARAGRAPH);
      ps.text = text;
      ps.font = font;
      ps.x = x;
      ps.y = y;
      ps.width = maxWidth;
      
      ews.offer(ps);
   }

   /**
    * Draw a string at the specified location.
    * 
    * @param   text
    *          text string to draw.
    * @param   x
    *          left position.
    * @param   height
    *          the height to center on.
    */
   @Override
   public void drawStringCentered(String text, int x, int height)
   {
      if (isEmptyText(text))
      {
         return;
      }

      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_STRING);
      ps.text = text;
      ps.x = x;
      ps.y = height;
      ps.centered = true;
      
      ews.offer(ps);
   }
   
   /**
    * Draws a line, using the current color, between the points
    * <code>(x1,&nbsp;y1)</code> and <code>(x2,&nbsp;y2)</code>
    * in this graphics context's coordinate system.
    * 
    * @param    x1  
    *           the first point's <i>x</i> coordinate.
    * @param    y1  
    *           the first point's <i>y</i> coordinate.
    * @param    x2  
    *           the second point's <i>x</i> coordinate.
    * @param    y2  
    *           the second point's <i>y</i> coordinate.
    */
   @Override
   public void drawLine(int x1, int y1, int x2, int y2)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_LINE);
      ps.x = x1;
      ps.y = y1;
      ps.width = x2;
      ps.height = y2;
      
      ews.offer(ps);
   }      

   /**
    * Draw a rectangle.
    * 
    * @param    x
    *           left position
    * @param    y
    *           top position
    * @param    width
    *           rectangle width.
    * @param    height
    *           rectangle height.
    */   
   @Override
   public void drawRect(int x, int y, int width, int height)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_RECT);
      ps.x = x;
      ps.y = y;
      ps.width = width;
      ps.height = height;
      
      ews.offer(ps);
   }      

   /**
    * Draw a rounded rectangle.
    * 
    * @param    x
    *           left position
    * @param    y
    *           top position
    * @param    width
    *           rectangle width.
    * @param    height
    *           rectangle height.
    * @param    arcDiameter
    *           diameter of the arc at corners.
    */
   @Override
   public void drawRoundRect(int x, int y, int width, int height, int arcDiameter)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_ROUND_RECT);
      ps.x = x;
      ps.y = y;
      ps.width = width;
      ps.height = height;
      ps.arcDiameter = arcDiameter;
      
      ews.offer(ps);
   }      

   /**
    * Fill a rectangle.
    * 
    * @param   x
    *          left position
    * @param   y
    *          top position
    * @param   width
    *          rectangle width.
    * @param   height
    *          rectangle height.
    */   
   @Override
   public void fillRect(int x, int y, int width, int height)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.FILL_RECT);
      ps.x = x;
      ps.y = y;
      ps.width = width;
      ps.height = height;
      
      ews.offer(ps);
   }      

   /**
    * Fill a rounded rectangle.
    * 
    * @param    x
    *           left position
    * @param    y
    *           top position
    * @param    width
    *           rectangle width.
    * @param    height
    *           rectangle height.
    * @param    arcDiameter
    *           diameter of the arc at corners.
    */   
   @Override
   public void fillRoundRect(int x, int y, int width, int height, int arcDiameter)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.FILL_ROUND_RECT);
      ps.x = x;
      ps.y = y;
      ps.width = width;
      ps.height = height;
      ps.arcDiameter = arcDiameter;
      
      ews.offer(ps);
   }      

   /**
    * Draws a 3-D highlighted outline of the specified rectangle.
    * The edges of the rectangle are highlighted so that they
    * appear to be beveled and lit from the upper left corner.
    * <p>
    * The colors used for the highlighting effect are determined
    * based on the current color.
    * The resulting rectangle covers an area that is
    * <code>width&nbsp;+&nbsp;1</code> pixels wide
    * by <code>height&nbsp;+&nbsp;1</code> pixels tall.  This method
    * uses the current <code>ColorRgb</code> exclusively and ignores
    * the current <code>Paint</code>.
    * 
    * @param    x 
    *           the x coordinate of the rectangle to be drawn.
    * @param    y 
    *           the y coordinate of the rectangle to be drawn.
    * @param    width 
    *           the width of the rectangle to be drawn.
    * @param    height 
    *           the height of the rectangle to be drawn.
    * @param    raised 
    *           a boolean that determines whether the rectangle appears 
    *           to be raised above the surface or sunk into the surface.
    */
   @Override
   public void draw3DRect(int x, int y, int width, int height, boolean raised)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_3D_RECT);
      ps.x = x;
      ps.y = y;
      ps.width = width;
      ps.height = height;
      ps.raised = raised;
      
      ews.offer(ps);
   }      

   /**
    * Paints a 3-D highlighted rectangle filled with the current color.
    * The edges of the rectangle are highlighted so that it appears
    * as if the edges were beveled and lit from the upper left corner.
    * The colors used for the highlighting effect and for filling are
    * determined from the current <code>ColorRgb</code>.  This method uses
    * the current <code>ColorRgb</code> exclusively and ignores the current
    * <code>Paint</code>.
    * 
    * @param    x 
    *           the x coordinate of the rectangle to be filled.
    * @param    y 
    *           the y coordinate of the rectangle to be filled.
    * @param    width
    *           the width of the rectangle to be filled.
    * @param    height 
    *           the height of the rectangle to be filled.
    * @param    raised 
    *           a boolean value that determines whether the rectangle appears 
    *           to be raised above the surface or etched into the surface.
    */   
   @Override
   public void fill3DRect(int x, int y, int width, int height, boolean raised)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.FILL_3D_RECT);
      ps.x = x;
      ps.y = y;
      ps.width = width;
      ps.height = height;
      ps.raised = raised;
      
      ews.offer(ps);
   }      
   
   /**
    * Draw a selection box rectangular area with current color.  Swing specific implementation.
    * 
    * @param    x
    *           left position
    * @param    y
    *           top position
    * @param    width
    *           rectangle width.
    * @param    height
    *           rectangle height.
    * @param    clear
    *           <code>TRUE</code> means clean up area instead of drawing selections.
    */
   public void drawSelection(int x, int y, int width, int height, boolean clear)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_SELECTION);
      ps.x = x;
      ps.y = y;
      ps.width = width;
      ps.height = height;
      ps.clear = clear;
      
      ews.offer(ps);
   }
   
   /**
    * Fill polygon.
    * 
    * @param    xPoints
    *           X coordinates.
    * @param    yPoints
    *           Y coordinates.
    */   
   @Override
   public void fillPolygon(int[] xPoints, int[] yPoints)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.FILL_POLYGON);
      ps.xPoints = xPoints;
      ps.yPoints = yPoints;
      ps.nPoints = Math.min(xPoints.length, yPoints.length);
      
      ews.offer(ps);
   }
   
   /**
    * Draw a image at the specified location.
    * 
    * @param    img
    *           Image to draw.
    * @param    x
    *           left location.
    * @param    y
    *           top location.
    * @param    w
    *           the image width.
    * @param    h
    *           the image height.
    * @param    xoff
    *           x offset inside image to start drawing.
    * @param    yoff
    *           y offset inside image to start drawing.
    * @param    transparent
    *           Flag indicating to use lower-left pixel as color marker to draw part of image as
    *           transparent.
    * @param    stretch
    *           Flag indicating stretch or shrink the image to fit display area.
    * @param    retain
    *           Flag indicating keep original image aspect ratio when stretching the image.
    * @param    convert3D
    *           Flag indicating autoconvert some image colors to the corresponding system wide 3D
    *           colors.
    */
   @Override
   public void drawImage(ImageWrapper img,
                         int          x,
                         int          y,
                         int          w,
                         int          h,
                         int          xoff,
                         int          yoff,
                         boolean      transparent,
                         boolean      stretch,
                         boolean      retain,
                         boolean      convert3D)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_IMAGE);
      fillPaintStructure(ps, img, x, y, w, h, xoff, yoff, transparent, stretch, retain, convert3D);

      ews.offer(ps);
   }

   /**
    * Draws a highlighted rectangular area of the specified color and bounds.
    * The color is alpha-blended with the content of the screen buffer.
    * 
    * @param    x
    *           left position
    * @param    y
    *           top position
    * @param    width
    *           area width
    * @param    height
    *           area height
    * @param    color
    *           The highlight color.
    */
   @Override
   public void drawHighlight(int x, int y, int width, int height, ColorRgb color)
   {
      if (width <= 0 || height <= 0)
      {
         return;
      }
      
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_HIGHLIGHT);
      ps.x = x;
      ps.y = y;
      ps.width = width;
      ps.height = height;
      ps.color = color;
      
      ews.offer(ps);
   }

   /**
   * Set current color used on draw operations.
   * 
   * @param    color
   *           Drawing color.
   */
   @Override
   public void setColor(ColorRgb color)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_COLOR);
      ps.color = color;

      ews.offer(ps);
   }
   
   /**
   * Set current color used on draw operations.
   * 
   * @param    color
   *           Drawing color.
   */
   @Override
   public void setColor(java.awt.Color color)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_COLOR);
      ps.color = new ColorRgb(color);

      ews.offer(ps);
   }
   
   /**
    * Set the desktop background color.
    * 
    * @param    color
    *           The new desktop background color.
    */
   @Override
   public void setDesktopBgColor(ColorRgb color)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_DESKTOP_BGCOLOR);
      ps.color = color;

      ews.offer(ps);
   }

   /**
    * Set "XOR composite" mode: color of the drawn elements is calculated as XOR (for each RGB
    * component) of their color with color of underlying pixels.
    */
   @Override
   public void setXORComposite()
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_XOR_COMPOSITE);

      ews.offer(ps);
   }

   /**
    * Set XOR mode painting allowing to invert the destination color with current drawing
    * primitive color.  Double painting of the same figure at the same location should restore
    * original picture.
    * 
    * @param    color
    *           XOR alternation color.
    */
   public void setXORMode(ColorRgb color)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_XOR_MODE);
      ps.color = color;

      ews.offer(ps);
   }

   /**
    * Reset composite mode and return to normal drawing mode.
    */
   @Override
   public void resetComposite()
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.RESET_COMPOSITE);

      ews.offer(ps);
   }
   
   /**
    * Set the window title at the OS level.
    * 
    * @param    title
    *           The window's title.
    */
   @Override
   public void setTitle(String title)
   {
      if (title == null)
      {
         title = "";
      }
      
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_TITLE);
      ps.title = title;

      ews.offer(ps);
   }
   
   /**
    * Sets the image to the window to be displayed as the window icon representation
    * for the system task bar, the title window panel or the system tray icons panel. 
    * 
    * @param    icon
    *           The wrapped icon.
    * @param    x
    *           The icon left location.
    * @param    y
    *           The icon top location.
    * @param    width
    *           The icon width.
    * @param    height
    *           The icon height.
    * @param    xoff
    *           x offset inside image of the icon to select drawing region.
    * @param    yoff
    *           y offset inside image of the icon to select drawing region.
    * @param    transparent
    *           Flag indicating to use lower-left pixel as color marker to draw part of image as
    *           transparent.
    * @param    stretch
    *           Flag indicating stretch or shrink the image to fit display area.
    * @param    retain
    *           Flag indicating keep original image aspect ratio when stretching the image.
    * @param    convert3D
    *           Flag indicating autoconvert some image colors to the corresponding system wide 3D
    *           colors.
    */
   @Override
   public void setIcon(ImageWrapper<BufferedImage> icon,
                         int             x,
                         int             y,
                         int             width,
                         int             height,
                         int             xoff,
                         int             yoff,
                         boolean         transparent,
                         boolean         stretch,
                         boolean         retain,
                         boolean         convert3D)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_ICON);
      fillPaintStructure(ps, icon, x, y, width, height, xoff, yoff, transparent, stretch, retain,
               convert3D);

      ews.offer(ps);
   }
   
   /**
    * Fills the helper paint structure with target values.
    * 
    * @param    ps
    *           The paint structure to be filled with the provided values.
    * @param    img
    *           The wrapped image.
    * @param    x
    *           The image left location.
    * @param    y
    *           The image top location.
    * @param    w
    *           The image width.
    * @param    h
    *           The image height.
    * @param    xoff
    *           x offset inside image to select the drawing region.
    * @param    yoff
    *           y offset inside image to select the drawing region.
    * @param    transparent
    *           Flag indicating to use lower-left pixel as color marker to draw part of image as
    *           transparent.
    * @param    stretch
    *           Flag indicating stretch or shrink the image to fit display area.
    * @param    retain
    *           Flag indicating keep original image aspect ratio when stretching the image.
    * @param    convert3D
    *           Flag indicating autoconvert some image colors to the corresponding system wide 3D
    *           colors.
    */
   private static void fillPaintStructure(
                        PaintStructure ps,
                        ImageWrapper<BufferedImage> img,
                        int             x,
                        int             y,
                        int             w,
                        int             h,
                        int             xoff,
                        int             yoff,
                        boolean         transparent,
                        boolean         stretch,
                        boolean         retain,
                        boolean         convert3D)
   {
      ps.x = x;
      ps.y = y;
      ps.width = w;
      ps.height = h;
      ps.xOffset = xoff;
      ps.yOffset = yoff;
      ps.transparent = transparent;
      ps.stretchToFit = stretch;
      ps.retainShape = retain;
      ps.convert3D = convert3D;
      ps.img = img;
   }

   /**
    * Set current font used on draw operations.
    * 
    * @param    font
    *           Drawing font.
    */
   @Override
   public void setGuiFont(FontDetails<F> font)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_FONT);
      ps.font = font;

      ews.offer(ps);
   }

   /**
    * Get the width of the specified font.
    * 
    * @param    fd
    *           The font details.
    *           
    * @return   The font's width.
    */
   @Override
   public int getFontWidth(FontDetails<F> fd)
   {
      FontMetricsHelper fm = ews.getFontMetrics(fd);
      
      int avgWidth = 0;
      int charno = 255;
      
      for (char ch = 1; ch <= 255; ch++)
      {
         avgWidth += fm.charWidth(ch);
      }
      avgWidth = (int) Math.ceil((avgWidth + 1.0) / (double) charno);

      return avgWidth;
   }

   /**
    * Get the maximum font width, for the given font.
    * 
    * @param    fd
    *           The font details.
    *           
    * @return   The font's maximum width.
    */
   @Override
   public int getFontMaxWidth(FontDetails<F> fd)
   {
      FontMetricsHelper fm = ews.getFontMetrics(fd);
      
      int[] widths = fm.getWidths();
      int maxWidth = 0;
      
      for (int w : widths)
      {
         maxWidth = Math.max(w,  maxWidth);
      }
      
      return maxWidth;
   }
   
   /**
    * Report the advance widths of the first 256 characters in the font.  Advance width measures 
    * the length of the character's baseline.
    * 
    * @param    fd
    *           The font details.
    *           
    * @return   The font widths.
    */
   @Override
   public int[] getFontWidths(FontDetails<F> fd)
   {
      return ews.getFontMetrics(fd).getWidths();
   }
   
   /**
    * Get the height of the specified font.
    * 
    * @param    fd
    *           The font details.
    *           
    * @return   The fonts's height.
    */
   @Override
   public int getFontHeight(FontDetails<F> fd)
   {
      FontMetricsHelper fm = ews.getFontMetrics(fd);
      
      return fm.getHeight();
   }

   /**
    * Get the width of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    fd
    *           The font details used to draw and measure the text.
    *           
    * @return   The text's width.
    */
   @Override
   public int getTextWidth(String text, FontDetails<F> fd)
   {
      if (isEmptyText(text))
      {
         return 0;
      }
      
      TextMetrics metrics = resolveTextMetrics(text, fd);
      
      return metrics.width;
   }

   /**
    * Get the widths of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    fd
    *           The font details used to draw and measure the text.
    *           
    * @return   The text's widths.
    */
   @Override
   public int[] getTextWidths(String text, FontDetails<F> fd)
   {
      if (isEmptyText(text))
      {
         return null;
      }
      
      TextMetrics metrics = resolveTextMetrics(text, fd);
      if (metrics.widths == null)
      {
         FontMetricsHelper fm = ews.getFontMetrics(fd);
         metrics.widths = fm.stringWidths(text);
      }
      
      return metrics.getWidths();
   }

   /**
    * Get the widths of the specified texts, using the given font.
    * 
    * @param    texts
    *           The texts.
    * @param    fd
    *           The font details used to draw and measure the text.
    *           
    * @return   The texts' widths.
    */
   @Override
   public int[][] getTextWidths(String[] texts, FontDetails<F> fd)
   {
      int[][] res = new int[texts.length][];
      for (int i = 0; i < texts.length; i++)
      {
         res[i] = getTextWidths(texts[i], fd);
      }
      
      return res;
   }

   /**
    * Get the height of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    fd
    *           The font details used to draw and measure the text.
    *           
    * @return   The text's height.
    */
   @Override
   public int getTextHeight(String text, FontDetails<F> fd)
   {
      if (isEmptyText(text))
      {
         return 0;
      }

      TextMetrics metrics = resolveTextMetrics(text, fd);
      
      return metrics.height;
   }
   
   /**
    * Scale the font so its width and height match the legacy values.
    * 
    * @param    font
    *           The font to be scaled.  This will be modified on return.
    */
   @Override
   public final void scaleFont(FontDetails<F> font)
   {
      if (font.legacyHeight < 0 || font.legacyWidth < 0)
      {
         // do not scale, no legacy metrics are present!
         return;
      }
      
      FontMetricsHelper fm = ews.getFontMetrics(font);
   
      int avgWidth = 0;
      for (int ch = 0x20; ch <= 0xff; ch++)
      {
         avgWidth += fm.charWidth((char) ch);
      }
      avgWidth = (int) Math.ceil((double) avgWidth / (double) (0xff - 0x20 + 1));
      int height = fm.getHeight();
      
      if (avgWidth != font.legacyWidth || height != font.legacyHeight)
      {
         deriveFont(font);
      }
   }
   
   /**
    * Scale the given font by the specified X and Y factors.
    *
    * @param    fd
    *           Font to scale.  This instance's contained font will be modified. 
    */
   protected abstract void deriveFont(FontDetails<F> fd);
   
   /**
    * Set the style of the current font used on draw operations.
    * 
    * @param    style
    *           Font style.
    */
   @Override
   public void setFontStyle(FontStyle style)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_FONT_STYLE);
      ps.style = style;
      
      ews.offer(ps);
   }
   
   /**
    * Set the style of the current line drawing primitives.
    * 
    * @param    stroke
    *           Line style.
    */
   @Override
   public void setLineStroke(LineStroke stroke)
   {
      setLineStroke(stroke, 0.0F);
   }
   
   /**
    * Set the style of the current line drawing primitives.
    * 
    * @param    stroke
    *           Line style.
    * @param    width
    *           Width value for the line.
    */
   @Override
   public void setLineStroke(LineStroke stroke, float width)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_LINE_STROKE);
      ps.stroke = stroke;
      ps.strokeWidth = width;
      
      ews.offer(ps);
   }
   
   /**
    * Resize window.
    * 
    * @param    width
    *           new width.
    * @param    height
    *           new height
    */
   @Override
   public void resizeWindow(int width, int height)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.RESIZE_WINDOW);
      ps.width = width;
      ps.height = height;
      
      ews.offer(ps);
      ews.resizeWindow(width, height);
   }

   /**
    * Set window bounds.
    * 
    * @param    x
    *           new x
    * @param    y
    *           new y
    * @param    width
    *           new width.
    * @param    height
    *           new height
    */
   @Override
   public void setWindowBounds(int x, int y, int width, int height)
   {
      if (ews != null)
      {
         // make sure the window is moved before showing it or it will flicker (shown at (0, 0)
         // then moved to this location) then post the repaint event
         ews.setWindowBounds(x, y, width, height);
      }
   }
   
   /**
    * Set window location.
    * 
    * @param    left
    *           left coordinate
    * @param    top
    *           top coordinate
    */
   @Override
   public void setWindowLocation(int left, int top)
   {
      if (ews != null)
      {
         // make sure the window is moved before showing it or it will flicker (shown at (0, 0)
         // then moved to this location) then post the repaint event
         ews.setWindowLocation(left, top);
      }
   }
   
   /**
    * Check if window has an emulator for the given window id.
    * 
    * @param    windowId
    *           Window ID.
    *           
    * @return   <code>true</code> if window has an associated emulator, <code>true</code>
    *           otherwise.
    */
   @Override
   public boolean hasWindowEmulator(int windowId)
   {
      return direct.hasWindowEmulator(windowId);
   }

   /**
    * Register a new window with the driver.
    * 
    * @param    windowId
    *           The window ID.
    */
   @Override
   public void registerWindow(int windowId)
   {
      EmulatedWindowState ews = newEmulator(windowId);
      
      direct.registerWindow(windowId, ews);
      captureMouseEvents(true);
   }

   /**
    * Register new child window with the driver.
    * 
    * @param    ownerId
    *           The owning window ID.
    * @param    windowId
    *           The window ID.
    * @param    windowTitle
    *           Title of the window emulator.
    * @param    resizable
    *           <code>true</code> if the window should be created as resizable.
    * @param    isShareActivationWithOwner
    *           Indicates whether the window shares activation state with its owner.
    * @param    modal
    *           <code>true</code> if this is a modal window.
    */
   @Override
   public void registerChildWindow(int ownerId, int windowId, String windowTitle,
                                   boolean resizable, boolean isShareActivationWithOwner,
                                   boolean modal)
   {
      EmulatedWindowState parentWindow = direct.getWindowEmulator(ownerId);
      
      EmulatedWindowState ews = newChildWindowEmulator(windowId, parentWindow, windowTitle, 
                                                       resizable, isShareActivationWithOwner,
                                                       modal);
      
      direct.registerWindow(windowId, ews);
      captureMouseEvents(true);
   }
   
   /**
    * Deregister this window from the driver.  This needs to be called only when the 
    * window widget is being destroyed.
    * 
    * @param    windowId
    *           The window ID.
    */
   @Override
   public void deregisterWindow(int windowId)
   {
      direct.deregisterWindow(windowId);
   }
   
   /**
    * Create a new window emulator.
    * 
    * @param    windowId
    *           The window ID.
    *
    * @return   The newly instantiated window emulator.
    */
   public abstract EmulatedWindowState newEmulator(int windowId);
   
   /**
    * Create a new child window emulator.
    * 
    * @param    windowId
    *           The window ID.
    * @param    owner
    *           The owning window.
    * @param    windowTitle
    *           Title of the window emulator.
    * @param    resizable
    *           <code>true</code> if the window is resizable.
    * @param    isShareActivationWithOwner
    *           Indicates whether the window shares activation state with its owner.
    * @param    modal
    *           <code>true</code> if this is a modal window.
    *
    * @return   The newly instantiated window emulator.
    */
   public abstract EmulatedWindowState newChildWindowEmulator(
                                        int                 windowId,
                                        EmulatedWindowState owner, 
                                        String              windowTitle,
                                        boolean             resizable,
                                        boolean             isShareActivationWithOwner,
                                        boolean             modal);
   
   /**
    * Force the list of all GUI windows to be stacked as in the specified order.
    * 
    * @param    windowIds
    *           The stack order to be enforced.
    */
   public abstract void restackWindows(int[] windowIds);
   
   /**
    * Move the specified window to the top of the z-order, at the OS level.
    *
    * @param    windowId
    *           The window to be moved.
    * @param    focus
    *           If <code>true</code>, the window will become the focused window 
    *           and its state changed to active. 
    */
   public abstract void moveToTop(int windowId, boolean focus);
   
   /**
    * Move the specified window to the bottom of the z-order, at the OS level.
    *
    * @param    windowId
    *           The window to be moved.
    */
   public abstract void moveToBottom(int windowId);
   
   /**
    * On a GUI interface multiple windows could be created. Each window has it own renderer.
    * When a window is draw first the window must call this method to setup it renderer as 
    * the current graphical canvas.  
    * <p>
    * The GUI driver primitives can be accessed by only a single thread at a time.  This requires
    * all GUI driver API calls to be bracketed in {@link #selectWindow} and {@link #releaseWindow}
    * calls.
    * <p>
    * If the {@link #drawingOwner} is <code>null</code>, the current thread will gain exclusive
    * access immediately and will hold it until has finished work.
    * 
    * @param   windowId
    *          Window ID.
    *          
    * @throws  IllegalStateException
    *          If the window is switched by the current thread while batch mode is on or the wait
    *          is interrupted.
    */
   @Override
   public void selectWindow(int windowId)
   {
      synchronized (this)
      {
         Thread current = Thread.currentThread();
         String msg = "Can not change window to " + windowId + " within the same thread, " +
                      "if a window is already selected!";
         
         boolean switched = false;
         
         // check the owner
         while (true)
         {
            switched = (ews == null || ews.getWindowId() != windowId);

            if (drawingOwner == null)
            {
               drawingOwner = current;
               break;
            }
            else if (drawingOwner == current)
            {
               if (switched && batch.peek())
               {
                  // same thread, can not switch windows if batch mode started
                  throw new IllegalStateException(msg);
               }
               
               break;
            }

            // different threads, wait for the window to be released regardless if switched or not
            try
            {
               wait();
            }
            catch (InterruptedException e)
            {
               throw new IllegalStateException("Waiting interrupted:" + msg, e);
            }
         }
         
         if (switched)
         {
            direct.selectWindow(windowId);
            ews = direct.getWindowEmulator(windowId);
            
            // save the last switched window
            lastWindow.push(ews);
         }

         windowSwitched.push(switched);
      }
   }
   
   /**
    * Release the last window used by the current thread.  This can be called only by the same
    * thread as the {@link #drawingOwner thread} which acquired the exclusive drawing access.
    * <p>
    * If no more windows are in use by the current thread, then the exclusive drawing access is
    * released and other other threads waiting in {@link #selectWindow} can acquire it.
    * 
    * @throws   NullPointerException
    *           If {@link #drawingOwner} is <code>null</code>
    * @throws   IllegalStateException
    *           If called from other thread than the {@link #drawingOwner}.
    */
   @Override
   public void releaseWindow()
   {
      synchronized (this)
      {
         if (drawingOwner == null)
         {
            throw new NullPointerException(
               "The drawing owner is not set. Unbalanced select/release window calls?");
         }
         
         if (drawingOwner != Thread.currentThread())
         {
            throw new IllegalStateException("The drawing window can be released only by the owner!");
         }
         
         boolean switched = windowSwitched.pop();
         if (switched)
         {
            lastWindow.pop();
            
            EmulatedWindowState last = lastWindow.peek();
            
            // the window can be released only if it has been switched by the previous 
            // selectWindow call (this indicates the owner thread has finished all current work 
            // with that window)

            if (last != null)
            {
               ews = last;
               // the emulator may have been destroyed since the switch
               if (direct.hasWindowEmulator(ews.windowId))
               {
                  direct.selectWindow(ews.windowId);
               }
               else
               {
                  direct.releaseWindow();
               }
            }
            else
            {
               drawingOwner = null;
               ews = null;
               direct.releaseWindow();

               if (!windowSwitched.isEmpty())
               {
                  throw new IllegalStateException(
                     "Releasing exclusive lock while other windows are still processed!");
               }

               notifyAll();
            }
         }
      }
   }
   
   /**
    * Save the current drawing state and release the drawing lock, so that other threads can 
    * acquire it.  This is useful in cases when the current thread needs to wait an indefinite
    * amount of time (i.e. for keyboard/mouse events).
    * 
    * @return   An object encoding the drawing state.
    */
   @Override
   public Object saveDrawingState()
   {
      synchronized (this)
      {
         if (drawingOwner != null && Thread.currentThread() != drawingOwner)
         {
            // different threads, wait for the window to be released regardless if switched or not
            try
            {
               wait();
            }
            catch (InterruptedException e)
            {
               throw new IllegalStateException("Waiting interrupted while restoring drawing state");
            }
         }
         
         try
         {
            return new Object[] { windowSwitched, batch, ews, lastWindow };
         }
         finally
         {
            windowSwitched = new ArrayDeque<>();
            batch = new ArrayDeque<>();
            batch.push(false);
            lastWindow = new ArrayDeque<>();
            ews = null;
            drawingOwner = null;
            direct.releaseWindow();
            
            notifyAll();
         }
      }
   }
   
   /**
    * Restore the drawing state to the specified state.  If the current thread is not currently 
    * the {@link #drawingOwner}, then wait for it to relinquish the exclusive lock, before 
    * restoring the state.
    * 
    * @param    state
    *           The drawing state.
    */
   @Override
   public void restoreDrawingState(Object state)
   {
      if (state == null)
      {
         throw new NullPointerException("Null state can not be processed.");
      }

      synchronized (this)
      {
         if (drawingOwner != null && Thread.currentThread() != drawingOwner)
         {
            // different threads, wait for the window to be released regardless if switched or not
            try
            {
               wait();
            }
            catch (InterruptedException e)
            {
               throw new IllegalStateException("Waiting interrupted while restoring drawing state");
            }
         }
         
         Object[] ostate = (Object[]) state;
         windowSwitched = (Deque<Boolean>) ostate[0];
         batch = (Deque<Boolean>) ostate[1];
         ews = (EmulatedWindowState) ostate[2];
         lastWindow = (Deque<EmulatedWindowState<F, ?>>) ostate[3];
         
         if (ews != null)
         {
            // the emulator may have been destroyed since the state saved
            if (direct.hasWindowEmulator(ews.windowId))
            {
               direct.selectWindow(ews.windowId);
            }
            else
            {
               direct.releaseWindow();
            }
            
            drawingOwner = Thread.currentThread();
         }
         else
         {
            direct.releaseWindow();
            drawingOwner = null;
         }
      }
   }
   
   /**
    * Selects the target window and then minimizes it.
    * 
    * @param    sourceId
    *           The root ID of the sub-tree which was iconified, and triggered iconification of 
    *           this window.
    * @param    postEvent
    *           When {@code true} the method posts the WINDOW-MINIMIZED legacy event,
    *           {@code false} means no event is posted.
    */
   @Override
   public void iconifyWindow(int sourceId, boolean postEvent)
   {
      EmulatedWindowState ews = getPrimitives().getWindowEmulator(sourceId);
      ews.iconifyWindow(sourceId, postEvent);
   }
   
   /**
    * Selects the target window and then restores it.
    * 
    * @param    sourceId
    *           The root ID of the sub-tree which was deiconified, and triggered restoration of 
    *           this window.
    */
   @Override
   public void deiconifyWindow(int sourceId)
   {
      EmulatedWindowState ews = getPrimitives().getWindowEmulator(sourceId);
      ews.deiconifyWindow(sourceId);
   }

   /**
    * Check if the specified window is iconified or not.
    * 
    * @param     windowId
    *            The window ID.
    *            
    * @return    See above.
    */
   @Override
   public boolean isIconified(int windowId)
   {
      return direct.getWindowEmulator(windowId).isIconified();
   }
   
   /**
    * Execute a widget drawing bracket.  This includes translating the origin of the graphics
    * context to the specified point, execute the drawing and restore the origin of the graphics
    * context.
    * <p>
    * When the widget ID is specified, all the drawing operations for this widget will be
    * bracketed in {@link PaintPrimitives#START_RECORDING} and
    * {@link PaintPrimitives#STOP_RECORDING}.  If the driver implementation handles these
    * primitives, it is expected to hash the resulted and either save the resulted image in a
    * cache or, if the image hash is already cached, replace the drawing operations with a
    * {@link PaintPrimitives#DRAW_CACHED} operation, so that the cached image will be drawn
    * instead.
    * <p>
    * The method intersects the supplied <code>clip</code> with the screen bitmap rectangles
    * retrieved by {@link ScreenBitmap#getClippings(NativeRectangle)}.
    *
    * @param    id
    *           When not-null, it caches the drawing result of this widget.
    * @param    origin
    *           The drawing origin.
    * @param    clip
    *           Clip rectangle.
    * @param    core
    *           The drawing worker.
    */
   @Override
   public void draw(WidgetId id, NativePoint origin, NativeRectangle clip, Runnable core)
   {
      PaintStructure record = null;
      Set<NativeRectangle> useClips = null;

      NativePoint tempOrig = new NativePoint(translatedOrigin);
      tempOrig.translate(origin.x, origin.y);

      if (clip != null)
      {
         // intersect the clipping rectangle with screen bitmap.
         NativeRectangle rclip = clip.translate(tempOrig);
         ScreenBitmap sb = ews.getScreenBitmap();
         useClips = new HashSet<>(sb.getClippings(rclip));

         // if the intersection is null/empty, do not draw anything
         if (useClips.isEmpty())
         {
            return;
         }

         boolean clipContainedFully = false;

         // move it back into translated coordinates (for origin)
         Set<NativeRectangle> tempClips = new HashSet<>(useClips.size());
         for (NativeRectangle nr : useClips)
         {
            tempClips.add(nr.translate(-tempOrig.x, -tempOrig.y));
            if (!clipContainedFully && nr.contains(clip))
            {
               clipContainedFully = true;
            }
         }
         useClips = tempClips;

         if (id != null && clipContainedFully)
         {
            // determine if a per-widget hash can be accommodated.  this will require that the
            // entire widget can be drawn (is not affected by clipping).
            record = new PaintStructure(PaintPrimitives.START_RECORDING);
            record.x = rclip.left();
            record.y = rclip.top();
            record.width = clip.width();
            record.height = clip.height();
         }
      }
      else
      {
         // TODO: enabling this code breaks the web client, but the Swing client works fine.  The
         // issue can be seen in drawing the EDITOR caret, which passes a NULL clipping rectangle.
         // It may have something to do with negative clipping coordinates, which the JS code may
         // not handle well (they are transferred OK, but the issue may be lower in the canvas
         // drawing/clipping/translation APIs)
         /*
         // the clip is the bitmap's invalidation rectangle, relative to physical origin
         clip = ews.getScreenBitmap().getOuterRectangle();
         if (clip != null)
         {
            clip = clip.translate(-tempOrig.x, -tempOrig.y);
         }
         */
      }

      boolean doTranslate = (origin.x != 0 || origin.y != 0);

      if (doTranslate)
      {
         translatePush(origin.x, origin.y);

         // save the last clipping rectangles, if we are translating
         lastClips = useClips;
      }
      else
      {
         // if we are NOT translating AND the clips are not null AND they are the same as the last
         // clips, then do not clip again
         if (useClips != null && useClips.equals(lastClips))
         {
            useClips = null;
         }
         else
         {
            // save the last clipping rectangles
            lastClips = useClips;
         }
      }

      try
      {
         if (record != null)
         {
            // record all drawing, except translation - as we want the recorded image to be
            // generic, so it can be applied to other widgets which look the same
            ews.offer(record);

            // clipping is included in the hash
         }

         if (useClips != null)
         {
            setDrawingAreas(useClips);
         }

         core.run();
      }

      finally
      {
         if (useClips != null)
         {
            setDrawingArea(null);
            lastClips = null;
         }

         if (record != null)
         {
            // stop the recording, if we've started it
            ews.offer(new PaintStructure(PaintPrimitives.STOP_RECORDING));
         }

         // restore context
         if (doTranslate)
         {
            translatePop();
         }
      }
   }

   /**
    * Execute a widget drawing bracket.  This includes translating the origin of the graphics
    * context to the specified point, execute the drawing and restore the origin of the graphics
    * context.
    * <p>
    * The method intersects the supplied <code>clip</code> with the screen bitmap rectangles
    * retrieved by {@link ScreenBitmap#getClippings(NativeRectangle)}.
    *
    * @param    origin
    *           The drawing origin.
    * @param    clip
    *           Clip rectangle.
    * @param    core
    *           The drawing worker.
    */
   @Override
   public void draw(NativePoint origin, NativeRectangle clip, Runnable core)
   {
      draw(null, origin, clip, core);
   }

   /**
    * Execute a widget drawing bracket.  This includes translating the origin of the graphics
    * context to the window's real (0,0) origin (top-left corner) and then translating to the
    * specified location, before executing the drawing.
    * <p>
    * The translated origin (as before this was called) needs to be restored, before this call
    * ends.
    * 
    * @param    origin
    *           The drawing origin (relative to window's real origin).
    * @param    clip
    *           Clip rectangle.
    * @param    core
    *           The drawing worker.
    */
   @Override
   public void drawFromOrigin(NativePoint origin, NativeRectangle clip, Runnable core)
   {
      NativePoint save = new NativePoint(this.translatedOrigin);
      boolean doTranslate = (save.x != 0 || save.y != 0);
      
      if (doTranslate)
      {
         translatePush(-save.x, -save.y);
      }
      
      try
      {
         draw(origin, clip, core);
      }
      
      finally
      {
         // translate to the saved origin
         if (doTranslate)
         {
            translatePop();
         }
      }
   }
   
   /**
    * Limit the available drawing area to the specified boundaries provided in the
    * <code>bounds</code> parameter. The individual rectangular areas may or may not overlap.
    * 
    * @param    bounds
    *           The bounds for the drawing area.
    */
   @Override
   public void setDrawingAreas(Set<NativeRectangle> bounds)
   {
      boolean clip = bounds != null && !bounds.isEmpty();

      PaintStructure ps = new PaintStructure(clip ? PaintPrimitives.CLIP
                                                  : PaintPrimitives.NO_CLIP);
      if (clip)
      {
         ps.clip = bounds;
      }
      else
      {
         // no clip, clear the last clipping rectangle
         lastClips = null;
      }
      
      ews.offer(ps);
   }

   /**
    * Limit the available drawing area to the specified boundaries.
    *
    * @param    bounds
    *           The bounds for the drawing area.
    */
   @Override
   public void setDrawingArea(NativeRectangle bounds)
   {
      setDrawingAreas(bounds == null ? null : Collections.singleton(bounds));
   }

   /**
    * Start a batch drawing routine.
    */
   @Override
   public void startBatch()
   {
      batch.push(true);
      PaintStructure ps = new PaintStructure(PaintPrimitives.START_BATCH);
      ews.offer(ps);
   }

   /**
    * End a batch drawing routine.
    */
   @Override
   public void endBatch()
   {
      batch.pop();
      PaintStructure ps = new PaintStructure(PaintPrimitives.END_BATCH);
      ews.offer(ps);
      
      ews.endBatch();
   }

   /**
    * Force screen buffer draw.
    */
   @Override
   public void flush()
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.FLUSH);
      
      ews.offer(ps);
   }
      
   /**
    * Hide or show the window, depending on the {@code visible} flag.
    * 
    * @param    visible
    *           The window visibility state.
    */
   @Override
   public void setWindowVisible(boolean visible)
   {
      if (ews != null)
      {
         ews.setWindowVisible(visible);
      }
   }
   
   /**
    * Set the window' sensitive state, depending on the {@code enable} flag.
    * 
    * @param    enable
    *           The window sensitive state.
    */
   @Override
   public void setWindowEnabled(boolean enable)
   {
      if (ews != null)
      {
         ews.setWindowEnabled(enable);
      }
   }

   /**
    * Get the window location on screen.
    * 
    * @return   See above.
    */
   @Override
   public NativePoint getWindowLocation()
   {
      return ews.getWindowLocation();
   }

   /**
    * Get the physical window dimension, in pixel units.
    * 
    * @return   See above.
    */
   @Override
   public NativeDimension getWindowDimension()
   {
      return new NativeDimension(ews.getWidth(), ews.getHeight());
   }
   
   /**
    * Force the list of all GUI windows to be stacked as in the specified order.
    * 
    * @param    windowIds
    *           The stack order to be enforced.
    */
   @Override
   public void stackWindows(int[] windowIds)
   {
      restackWindows(windowIds);
   }

   /**
    * Enable or Disable OS-level events for the specified windows.
    * 
    * @param    windowIds
    *           The window IDs which need to be processed.
    * @param    capture
    *           Flag indicating OS level events are captured (when <code>true</code>).
    */
   @Override
   public void enableEvents(int[] windowIds, boolean capture)
   {
      for (int wid : windowIds)
      {
         // Do NOT call selectWindow() here, since this code can be called from TC.getChanges()
         // which can occur within a single drawing batch. Each drawing batch should be specific
         // to a single window and thus we don't allow windows to be switched during a batch.
         // Instead, we directly access each window to change its capturing state.
         EmulatedWindowState emulator = direct.getWindowEmulator(wid);
         if (emulator != null)
         {
            emulator.enableEvents(capture);
         }
      }
   }
   
   /**
    * Set cursor status (ON/OFF).
    * <p>
    * For GUI, this does nothing.
    * 
    * @param   on
    *          <code>true</code> - show cursor, <code>false</code> - hide cursor.
    *
    * @return  Always <code>false</code>.
    */
   @Override
   public boolean setCursorStatus(boolean on)
   {
      return false;
   }

   /**
    * Reset current screen mode.
    * <p>
    * Upon exit from this method P2J assumes that screen is in correct state
    * and can be used for interaction with the user.
    */
   @Override
   public void resetMode()
   {
      // NOP for now, what should this do?
   }
   
    /**
     * Resume normal driver operation. If implementation is terminal-based,
     * they terminal should be acquired again and driver should be ready to 
     * perform regular screen I/O operations.
     */
   @Override
   public void resume()
   {
      // NOP for now, what should this do?
   }

   /**
    * Suspend all screen I/O operations. If implementation is terminal-based, 
    * then terminal should be released to allow other application executed 
    * from P2J access terminal.
    */
   @Override
   public void suspend()
   {
      // NOP for now, what should this do?
   }
   
   /**
    * Perform driver shutdown.
    */
   @Override
   public void shutdown()
   {
      // destroy all emulators
      direct.shutdown();
   }
   
   /**
    * Returns bounds of a physical screen the passed in window is positioned on.
    * The bounds are in virtual device coordinate system. For more information
    * about the virtual device coordinate system see {@link TopLevelWindow}.
    * <p>
    * For single-display setups the top-left corner of the returned rectangle is always 
    * 0,0. For multiple-display setups it can be 0,0 or positive.
    * 
    * @param    windowId
    *           A valid window id.
    *
    * @return   The physical screen bounds.
    */
   @Override
   public NativeRectangle getDisplayBounds(int windowId)
   {
      // this will never be null because an exception will be thrown if window id is invalid
      EmulatedWindowState ews = direct.getWindowEmulator(windowId);
      
      return ews.getDisplayBounds();
   }

   /**
    * Returns an application-usable display area. For example, a task bar is
    * not counted as the usable area.
    * 
    * For multiple-display environments the display is selected based on the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    *
    * @return   The area in pixels.
    */
   @Override
   public NativeRectangle getDisplayWorkArea(int windowId)
   {
      // this will never be null because an exception will be thrown if window id is invalid
      EmulatedWindowState ews = direct.getWindowEmulator(windowId);
      
      return ews.getDisplayWorkArea();
   }

   /**
    * Check if the currently selected window is visible.
    * 
    * @return   See above.
    */
   @Override
   public boolean isWindowVisible()
   {
      return ews.isWindowVisible();
   }

   /**
    * Set status text message of current window.
    * 
    * @param    message
    *           Text message.
    */
   @Override
   public void setStatusText(String message)
   {
      Window<?> window = (Window<?>) WindowManager.findWindow(ews.getWindowId());
      setStatusText(window, message);
   }

   /**
    * Set status text message of specified window.
    * 
    * @param    message
    *           Text message.
    * @param    window
    *           Window instance.
    */
   @Override
   public void setStatusText(Window<?> window, String message)
   {
      if (window != null)
      {
         window.getStatusArea().setText(message);
      }      
   }
   
   /**
    * Enable or disable capture of mouse events.
    * 
    * @param    state
    *           {@code true} to capture mouse events, {@code false} to ignore them.
    *           
    * @return   The previous state.
    */
   @Override
   public boolean captureMouseEvents(boolean state)
   {
      return MouseHandler.captureMouseEvents(state);
   }

   /**
    * Handle the specified event, in the context of the specified widget.
    * 
    * @param    sourceID
    *           The source widget ID.
    * @param    evt
    *           The mouse event.
    */
   @Override
   public void handleMouseEvent(int sourceID, MouseEvent evt)
   {
      MouseHandler.handleMouseEvent(sourceID, evt);
   }
      
   /**
    * Notify the widget of any mouse events, generated after the given timestamp (in millis).
    * 
    * @param    wid
    *           The widget to be notified.
    * @param    when
    *           Only events generated after this timestamp will be notified.
    */
   @Override
   public void listenAnyMouseEvent(int wid, long when)
   {
      MouseHandler.listenAnyMouseEvent(wid, when);
   }

   /**
    * Register the specified widget for popup events, via mouse.
    * <p>
    * The popup with the given ID will be shown when the specified mouse button is pressed inside
    * the specified widget.
    * 
    * @param    wid
    *           The widget id.
    * @param    btn
    *           The button to be pressed, when showing the popup.
    * @param    pid
    *           The popup widget ID.
    */
   @Override
   public void registerWidgetPopup(int wid, int btn, int pid)
   {
      MouseHandler.registerWidgetAction(wid, new MousePopupable(wid, btn, pid));
   }

   /**
    * Register this widget as having tooltip functionality reacting on mouse entering and exiting
    * events.  The registration is pending if tooltip engine is not available for session or
    * tooltip text is empty.
    * 
    * @param    widget
    *           The widget object to register.
    * @param    fnt
    *           The font to draw tooltip text.
    * @param    now
    *           The flag indicating whether the action be registered immediately or later.
    */
   @Override
   public void registerTooltipWorker(Widget<GuiOutputManager> widget,
                                     int                      fnt,
                                     boolean                  now)
   {
      int wid = widget.getId().asInt();
      MouseTooltipWorker worker = new MouseTooltipWorker(wid, fnt);
      
      // now or delayed?
      if (now)
      {
         MouseHandler.registerWidgetAction(wid, worker);
      }
      else
      {
         pendingTtWorkers.put(widget, worker);
      }
   }

   /**
    * Register the tooltip workers if this process was dalayed with regular registration.
    * 
    * @param    wid
    *           The widget id. 0 means register all pending workers. 
    */
   @Override
   public void registerPendingTooltipWorkers(int wid)
   {
      // if tooltip feature is turned off for session nothing to do
      if (!ToolTip.getSessionTooltips())
      {
         return;
      }
      
      // handle pending workers
      if (wid == 0)
      {
         // all workers mode enumerate and register all valid
         Iterator<Widget> iter = pendingTtWorkers.keySet().iterator();
         while (iter.hasNext())
         {
            // get the widget object
            Widget w = iter.next();
            // handle single worker
            if (registerPendingTooltipWorkerInt(w))
            {
               iter.remove();
            }
         }
      }
      else
      {
         // handle single worker
         Widget<GuiOutputManager> w = (Widget<GuiOutputManager>) OutputManager.resolveWidget(wid);
         if (registerPendingTooltipWorkerInt(w))
         {
            pendingTtWorkers.remove(w);
         }
      }
   }
   
   /**
    * Deregister the tooltip workers already registered with mouse handler.
    * 
    * @param    wid
    *           The widget id. 0 means deregister all registered workers. 
    */
   @Override
   public void deregisterTooltipWorkers(int wid)
   {
      if (wid == 0)
      {
         // deregister all widgets mode
         Integer[] widList = MouseHandler.getWidgetListForTooltipWorkers();
         // enumerate all widget ID's if there is
         if (widList != null)
         {
            for (int i = 0; i < widList.length; i++)
            {
               deregisterTooltipWorkerInt(widList[i].intValue());
            }
         }
      }
      else
      {
         deregisterTooltipWorkerInt(wid);
      }
   }
   
   /**
    * Deregister the specified widget for mouse events.
    * 
    * @param    wid
    *           The widget ID.
    */
   @Override
   public void deregisterWidget(int wid)
   {
      MouseHandler.deregisterWidget(wid);
      // clean the pending worker too but for real widgets only
      if (wid > 0)
      {
         pendingTtWorkers.remove((Widget<GuiOutputManager>) OutputManager.resolveWidget(wid));
      }
   }
   
   /**
    * Register all the mouse-aware widgets (in all Windows) with the driver.
    * <p>
    * No-op by default.
    */
   @Override
   public void registerMouseWidgets()
   {
      // no-op by default
   }
   
   /**
    * Register the specified widget as a container able to use direct manipulation via mouse.
    * 
    * @param    widget
    *           The widget object to register.
    */
   @Override
   public void registerDirectManipulation(Widget<GuiOutputManager> widget)
   {
      int wid = widget.getId() != null ? widget.getId().asInt() : -1;
      MouseDirectManipulation worker = new MouseDirectManipulation(widget);
      
      MouseHandler.registerWidgetAction(wid, worker);
   }
   
   /**
    * Register the tooltip worker if this process was dalayed with regular registration.  Internal
    * version.
    * 
    * @param    widget
    *           The widget object. Null means register all pending workers. 
    *           
    * @return   <code>TRUE</code> if worker registered, <code>FALSE</code> otherwise.
    */
   private boolean registerPendingTooltipWorkerInt(Widget<GuiOutputManager> widget)
   {
      // simple validity check
      if (widget == null)
      {
         // not valid, exit
         return false;
      }
      // verify the text
      String text = UiUtils.getConfig(widget, BaseConfig.class).tooltip;
      if (text != null && text.length() > 0)
      {
         // remove worker from list of pendings
         MouseWidgetAction worker = pendingTtWorkers.get(widget);
         if (worker != null)
         {
            MouseHandler.registerWidgetAction(widget.getId().asInt(), worker);
            return true;
         }
      }
      // worker was not registered
      return false;
   }
   
   /**
    * Deregister the tooltip workers already registered with mouse handler.  Internal wersion.
    * 
    * @param    wid
    *           The widget id.
    */
   private void deregisterTooltipWorkerInt(int wid)
   {
      // extract handler for this widget
      MouseWidgetAction worker = MouseHandler.deregisterTooltipWorker(wid);

      // determine the widget object for this widget ID
      Widget<GuiOutputManager> widget =
         (Widget<GuiOutputManager>) OutputManager.resolveWidget(wid);
         
      // if widget or handler are null we do nothing 
      if (widget == null || worker == null)
      {
         // something is wrong, exit
         return;
      }
      
      // save the deregistered worker in pending map to be registered later 
      pendingTtWorkers.put(widget, worker);
   }
   
   /**
    * Resolve the metrics for the given text and font.
    * <p>
    * If the {@link #textMetricsCache} doesn't already contain it, they are resolved now. 
    * 
    * @param    text
    *           The text for which metrics are needed.
    * @param    fd
    *           The font details.
    * 
    * @return   The text metrics.
    */
   protected TextMetrics resolveTextMetrics(String text, FontDetails<F> fd)
   {
      TextKey<F> key = new TextKey(text, fd.font);
      
      TextMetrics metrics = textMetricsCache.get(key);
      if (metrics == null)
      {
         FontMetricsHelper fm = ews.getFontMetrics(fd);
         int width = fm.stringWidth(text);
         int height = ews.getTextHeight(text, fd);
         
         metrics = new TextMetrics(width, height);
         
         textMetricsCache.put(key, metrics);
      }
      
      return metrics;
   }
  
   /**
    * Get the metrics for the given text and font from the {@link #textMetricsCache}.
    * 
    * @param    text
    *           The text for which metrics are needed.
    * @param    fd
    *           The font details.
    * 
    * @return   The text metrics or <code>null</code> if they don't yet exist.
    */
   protected TextMetrics getTextMetrics(String text, FontDetails<F> fd)
   {
      TextKey<F> key = new TextKey(text, fd.font);
      return textMetricsCache.get(key);
   }
   
   /**
    * Set the metrics for the given text and font to the {@link #textMetricsCache}.
    * 
    * @param    text
    *           The text for which metrics are needed.
    * @param    fd
    *           The font details.
    * @param    metrics
    *           The metrics.
    */
   protected void setTextMetrics(String text, FontDetails<F> fd, TextMetrics metrics)
   {
      TextKey<F> key = new TextKey(text, fd.font);
      textMetricsCache.put(key, metrics);
   }
   
   /**
    * Resolve the metrics for the given paragraph, font and max width.
    * <p>
    * If the {@link #paragraphMetricsCache} doesn't already contain it, they are resolved now. 
    * 
    * @param    text
    *           The text for which metrics are needed.
    * @param    fd
    *           The font details.
    * @param    maxWidth
    *           The maximum width in which the paragraph must fit.
    *           
    * @return   The paragraph metrics.
    */
   private TextMetrics resolveParagraphMetrics(String text, FontDetails<F> fd, int maxWidth)
   {
      ParagraphKey<F> key = new ParagraphKey(text, fd.font, maxWidth);
      
      TextMetrics metrics = paragraphMetricsCache.get(key);
      if (metrics == null)
      {
         FontMetricsHelper fm = ews.getFontMetrics(fd);
         int width = maxWidth; // the width is already known, as it needs to fit in it
         int height = ews.getParagraphHeight(text, fd, maxWidth);
         
         metrics = new TextMetrics(width, height);
         
         paragraphMetricsCache.put(key, metrics);
      }
      
      return metrics;
   }
   
   /**
    * Check if the specified text is empty.
    * 
    * @param    text
    *           The text to check.
    *           
    * @return   <code>true</code> if the text is <code>null</code> or empty string.
    */
   protected boolean isEmptyText(String text)
   {
      return text == null || text.length() == 0;
   }
   
   /**
    * A simple container for text metrics.
    */
   protected static class TextMetrics
   {
      /** The text's width. */
      private final int width;
      
      /** The text's height. */
      private final int height;
      
      /** The widths for this text, for each substring starting from index 0 and ending on 
       * indices 1 to text's length - 1.
       */
      private int[] widths = null;
      
      /**
       * Constructor.
       * 
       * @param   width
       *          The text's width.
       * @param   height
       *          The text's height.
       */
      public TextMetrics(int width, int height)
      {
         this.width = width;
         this.height = height;
      }
      
      /**
       * Get the intermediate widths.
       * 
       * @return   See above.
       */
      public int[] getWidths()
      {
         return widths == null ? null : Arrays.copyOf(widths, widths.length);
      }
      
      /**
       * Set the intermediate widths.
       * 
       * @param    widths
       *           The intermediate widths.
       */
      public void setWidths(int[] widths)
      {
         this.widths = (widths == null ? null : Arrays.copyOf(widths, widths.length));
      }
   }
   
   /**
    * Identify the text metrics in the cache.
    */
   private static class TextKey<F>
   {
      /** The text. */
      private final String text;
      
      /** The font with which the metrics were computed. */
      private final F font;

      /**
       * Constructor.
       * 
       * @param    text
       *           The text.
       * @param    font
       *           The font with which the metrics were computed.
       */
      public TextKey(String text, F font)
      {
         this.text = text;
         this.font = font;
      }
      
      /**
       * Compute an unique hash key for this text and font combination.
       * 
       * @return   See above.
       */
      @Override
      public int hashCode()
      {
         return font.hashCode() * 37 + text.hashCode();
      }
      
      /**
       * Check if this key is equal with the specified object.
       * 
       * @param    obj
       *           The object to check.
       * 
       * @return   See above.
       */
      @Override
      public boolean equals(Object obj)
      {
         if (!(obj instanceof TextKey))
         {
            return false;
         }
         
         TextKey tk = (TextKey) obj;
         
         return font.equals(tk.font) && text.equals(tk.text);
      }
   }
   
   /**
    * Identify the paragraph metrics in the cache.
    */
   private static class ParagraphKey<F>
   extends TextKey<F>
   {
      /** The width in which to fit the paragraph. */
      private final int maxWidth;

      /**
       * Constructor.
       * 
       * @param    text
       *           The text.
       * @param    font
       *           The font with which the metrics were computed.
       * @param    maxWidth
       *           The width in which to fit the paragraph.
       */
     public ParagraphKey(String text, F font, int maxWidth)
      {
         super(text, font);
         this.maxWidth = maxWidth;
      }
      
      /**
       * Compute an unique hash key for this paragraph and font combination.
       * 
       * @return   See above.
       */
      @Override
      public int hashCode()
      {
         return super.hashCode() * 37 + maxWidth;
      }
      
      /**
       * Check if this key is equal with the specified object.
       * 
       * @param    obj
       *           The object to check.
       * 
       * @return   See above.
       */
      @Override
      public boolean equals(Object obj)
      {
         if (!(obj instanceof ParagraphKey))
         {
            return false;
         }
         
         ParagraphKey tk = (ParagraphKey) obj;
         
         return tk.maxWidth == maxWidth && super.equals(obj);
      }
   }
}
