/*
** Module   : EmulatedWindowState.java
** Abstract : generic code for the core GUI window emulator
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20150429 First version with most functionality moved from GuiSimulator. There is a
**                  one-to-one relationship between instances of this class and a 4GL window.
** 002 HC  20150612 Implemented support for paragraph layout.
** 003 HC  20150702 Improved positioning of top-level windows on systems with multiple displays.
** 004 GES 20150817 Removed isOnTop() since it is unnecessary. The saveFocusListeners(),
**                  restoreFocusListeners() and moveToTop() are only needed for Swing usage and
**                  have been removed from this abstract base class.
**                  Changed the mouse handler to be driver-specific.
** 005 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 006 SBI 20151014 Added temporary logging.
** 007 CA  20151024 Added support for WINDOW:SENSITIVE attribute.
**     EVL 20151106 Adding method to obtain the BufferedImageDrawHelper instance.
** 008 IAS 20160107 LOAD-MOUSE-POINTER and SET-WAIT-STATE support
** 009 HC  20160413 Converted System.err debug output to java.util.logging.
** 010 HC  20170202 Implemented runtime support for WINDOW-STATE. The implementation is missing
**                  support for WINDOW-DELAYED-MINIMIZE.
** 011 OM  20170215 Moved window resize and location messages to Swing dispatch thread.
** 012 OM  20170412 Made LOG protected, to be available in subclasses.
** 013 OM  20170517 Small javadoc fixes.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver;

import java.util.*;
import java.util.concurrent.*;
import java.util.logging.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.util.*;

/**
 * This is generic code for the core GUI window emulator. A pool of window-specific drawing
 * operations is used to queue drawing primitives. The operations are enqueued in the pool
 * one at a time and then are drawn into a buffered image which is provided by the sub-class.
 * When the pool is empty, the screen is updated with the image from the stored buffered image.
 * <p>
 * There should not be any Swing/AWT-specific or Web-specific code in this class.
 * <p>
 * The actual GUI processing does not heavily utilize the pluggable <code>OutputPrimitives</code>
 * interface.  In other words, the interactive UI processing GUI is not switched in and out like
 * in ChUI.  For this reason, the core output processing in ChUI is driven using the primitives
 * approach, but in GUI we simply honor the minimum implementation in this class.  The majority
 * of this class is really about providing access to and management of the core window structure.
 * The primitives implementation is secondary, but the functionality is best implemented here.
 */
public abstract class EmulatedWindowState<F, I>
{
   /** Logger. */
   protected static final Logger LOG = LogHelper.getLogger(EmulatedWindowState.class);
   
   /** The window ID associated with this emulated window. */
   protected int windowId;
   
   /** The registered mouse handler. */
   protected MouseHandler mouseHandler;
   
   /** Track the window's iconified state (which can be set at the OS or P2J level). */
   private boolean iconified = false;
   
   /** Draw operations pools */
   private Deque<PaintStructure<F, I>> graphicPool = new ConcurrentLinkedDeque<>();

   /** The bitmap associated with this simulator. */
   private ScreenBitmap bitmap;
   
   /** Track clipping operations. */
   private int clips = 0;

   /** Track translate operations. */
   private int trans = 0;

   /**
    * Constructor.
    */
   public EmulatedWindowState(MouseHandler mouseHandler)
   {
      this.mouseHandler = mouseHandler;
   }
   
   /**
    * Get the window's mouse handler.
    * 
    * @return   The {@link #mouseHandler}.
    */
   public MouseHandler getMouseHandler()
   {
      return mouseHandler;
   }
   
   /**
    * Get the screen width, in character units.
    * 
    * @return    See above.
    */
   public int screenWidth()
   {
      return bitmap.cols();
   }

   /**
    * Get the screen height, in character units.
    * 
    * @return    See above.
    */
   public int screenHeight()
   {
      return bitmap.rows();
   }
   
   /**
    * Set the active screen bitmap for the current window.
    * 
    * @param    bitmap
    *           The screen bitmap.
    */
   public void setScreenBitmap(ScreenBitmap bitmap)
   {
      ScreenBitmap prev = this.bitmap;
      
      this.bitmap = bitmap;
      
      if (prev == null || bitmap.cols() != prev.cols() || bitmap.rows() != prev.rows())
      {
         resizeWindow(bitmap.cols(), bitmap.rows());
      }
   }

   /**
    * Get the screen bitmap for the current window.
    * 
    * @return   The current window's screen bitmap.
    */
   public ScreenBitmap getScreenBitmap()
   {
      return bitmap;
   }

   /**
    * Set the window ID for this simulator.
    * 
    * @param   windowId
    *          The window ID.
    */
   public void setWindowId(int windowId)
   {
      this.windowId = windowId;
   }

   /**
    * Get the window ID for this simulator.
    */
   public int getWindowId()
   {
      return windowId;
   }

   /**
    * Add a new <code>PaintStructure</code> to the drawing pool.
    * The item is enqueued at the end of the pool.
    * 
    * @param   ps
    *          A new PaintStructure instance.
    */      
   public void offer(PaintStructure<F, I> ps)
   {
      // to prevent deadlocks, synchronize on the frame's layout lock
      synchronized (lock())
      {
         graphicPool.offerLast(ps);

         lock().notifyAll();
      }
   }

   /**
    * Resize window.
    * 
    * @param    width
    *           new width.
    * @param    height
    *           new height
    */
   public void resizeWindow(int width, int height)
   {
      repaint();

      this.bitmap.setRows(height);
      this.bitmap.setColumns(width);
   }
   
   /**
    * Resize window.
    * 
    * @param    width
    *           new width.
    * @param    height
    *           new height
    */
   public void setWindowBounds(int left, int top, int width, int height)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_WINDOW_BOUNDS);
      ps.x = left;
      ps.y = top;
      ps.width = width;
      ps.height = height;
      
      offer(ps);
      resizeWindow(width, height);
   }
   
   /**
    * Set window location.
    *
    * @param   left
    *          left coordinate
    * @param   top
    *          top coordinate
    */
   public void setWindowLocation(int left, int top)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_LOCATION);
      ps.x = left;
      ps.y = top;
      
      offer(ps);
      repaint();
   }
   
   /**
    * Manage any window state related to iconification.
    * 
    * @param    iconify
    *           <code>true</code> if this window is being iconified, <code>false</code> for
    *           de-iconified.
    */
   public void setIconificationState(boolean iconify)
   {
      this.iconified = iconify;
   }

   /**
    * Iconify the window.
    * 
    * @param    sourceId
    *           The root ID of the sub-tree which was iconified, and triggered iconification of 
    *           this window.
    * @param    postEvent
    *           When {@code true} the method posts the WINDOW-MINIMIZED legacy event,
    *           {@code false} means no event is posted.
    */
   public void iconifyWindow(int sourceId, boolean postEvent)
   {
      if (iconified)
      {
         return;
      }
      
      setIconificationState(true);

      if (postEvent)
      {
         // raise the event here
         WindowGuiImpl window = (WindowGuiImpl) ThinClient.getInstance().getWidget(windowId);
         WindowEvent evt = new WindowEvent(window, Keyboard.SE_WINDOW_MINIMIZED);
         evt.setOrigin(sourceId);
         ThinClient.getInstance().postOSEvent(evt);
      }
   }

   /**
    * De-iconify the window.
    * 
    * @param    sourceId
    *           The root ID of the sub-tree which was deiconified, and triggered restoration of 
    *           this window.
    */
   public void deiconifyWindow(int sourceId)
   {
      if (!iconified)
      {
         return;
      }
      
      setIconificationState(false);
      
      // some window systems don't always generate a window deiconified event, so post one
      // manually... setting the "iconified" flag prevents a double-event generation, in case one
      // is raised by the window system too
      WindowGuiImpl window = (WindowGuiImpl) ThinClient.getInstance().getWidget(windowId);
      WindowEvent evt = new WindowEvent(window, Keyboard.SE_WINDOW_RESTORED);
      evt.setOrigin(sourceId);
      ThinClient.getInstance().postOSEvent(evt);
   }
   
   /**
    * Check if this window is iconified or not.
    * 
    * @return    See above.
    */
   public boolean isIconified()
   {
      return iconified;
   }
   
   /**
    * End a batch drawing routine.
    */
   public void endBatch()
   {
      repaint();
      
      synchronized (lock())
      {
         lock().notifyAll();
      }
   }
   
   /**
    * Enable or Disable OS-level events for the specified windows.
    * 
    * @param    capture
    *           Flag indicating OS level events are captured (when <code>true</code>).
    */
   public abstract void enableEvents(boolean capture);
   
   /**
    * Returns bounds of a physical screen the top-level window this instance 
    * represents is positioned on.
    * The bounds are in virtual device coordinate system. For more information
    * about the virtual device coordinate system see {@link TopLevelWindow}.
    * <p>
    * For single-display setups the top-left corner of the returned rectangle is always 
    * 0,0. For multiple-display setups it can be 0,0 or positive.
    * 
    * @return   The physical screen bounds.
    */
   public abstract NativeRectangle getDisplayBounds();

   /**
    * Returns an application-usable display area. For example, a task bar is
    * not counted as the usable area.
    * 
    * @return   The area in pixels.
    */
   public abstract NativeRectangle getDisplayWorkArea();
   
   /**
    * Get the window location on screen.
    * 
    * @return   This window's location.
    */
   public abstract NativePoint getWindowLocation();
   
   /**
    * Get the metrics of the given font.
    *  
    * @param    font
    *           The font for which the font metrics are to be obtained.
    *           
    * @return   The font metrics.
    */
   public abstract FontMetricsHelper getFontMetrics(FontDetails<F> font);

   /**
    * Get the height of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    font
    *           The font used to draw and measure the text.
    *           
    * @return   The text's height.
    */
   public abstract int getTextHeight(String text, FontDetails<F> font);
   
   /**
    * The method performs a layout operation on the supplied text and 
    * returns the resulting paragraph height while maintaining the
    * supplied maximum width.
    * <p>
    * Note that the algorithm for text layout is compatible with 
    * {@link PaintPrimitives#DRAW_PARAGRAPH} and so
    * can be used in conjunction with this method. 
    * 
    * @param    text
    *           The text to measure.
    * @param    font
    *           The font to use during layout.
    * @param    maxWidth
    *           The maximum paragraph width.
    */
   public abstract int getParagraphHeight(String text, FontDetails<F> font, int maxWidth);
   
   /**
    * Get the window height.
    * 
    * @return   This window's width in pixels.
    */
   public abstract int getHeight();
   
   /**
    * Get the window width.
    * 
    * @return   This window's width in pixels.
    */
   public abstract int getWidth();
   
   /**
    * Render the given paint operation.
    * 
    * @param    ps
    *           Paint operation details.
    */
   public abstract void draw(PaintStructure<F, I> ps);
   
   /**
    * Cause repaint processing to occur.
    */
   public abstract void repaint();
   
   /**
    * Hide or show the window, depending on the {@code visible} flag.
    * 
    * @param    visible
    *           The window visibility state.
    */
   public abstract void setWindowVisible(boolean visible);
   
   /**
    * Set the window' sensitive state, depending on the {@code enable} flag.
    * 
    * @param    enable
    *           The window sensitive state.
    */
   public abstract void setWindowEnabled(boolean enable);
   
   /**
    * Check if the simulated window is visible.
    * 
    * @return   See above.
    */
   public abstract boolean isWindowVisible();
   
   /**
    * Render the given paint operation.  Any graphics context or other resources needed must
    * have already been setup/stored by the caller.
    */
   public abstract void flush();
   
   /**
    * Release resources as the window is being terminated.
    */
   public abstract void quit();
   
   /**
    * Gets the instanve of the helper to work with buffered images.
    * 
    * @return   See above.
    */
   public abstract BufferedImageDrawHelper getBufferedImageDrawHelper();
   
   /**
    * Obtain the lock object that can be used to synchronize drawing operations.
    *
    * @return   The shared lock object.
    */
   protected abstract Object lock();
   
   /**
    * Update screen by iterating over the pool and render each pending operation. Any graphics
    * context or other resources needed must have already been setup/stored by the caller.
    */
   protected void drawOnScreenBitmap()
   {
      // ensure the drawing will be performed by only one thread at a time
      synchronized (lock())
      {
         int waitForBatch = 0;
         
         LOG.log(Level.FINEST, "START DRAWING CYCLE FOR WINDOW " + windowId);
         
         while (!graphicPool.isEmpty() || waitForBatch != 0)
         {
            // if a batch was started, wait for it to end.
            if (graphicPool.isEmpty() && waitForBatch != 0)
            {
               try
               {
                  lock().wait();
               }
               catch (InterruptedException e)
               {
                  // terminate painting
                  break;
               }
            }
            
            PaintStructure<F, I> ps = graphicPool.pollFirst();
            
            if (ps != null)
            {
               String log = ps.toString();
               switch (ps.id)
               {
                  // clipping primitives:
                  case CLIP:
                     clips++;
                     log += " " + clips;
                     break;
                  case NO_CLIP:
                     log += " " + clips;
                     clips--;
                     break;
                  // translation primitives:
                  case TRANSLATE_PUSH:
                     trans++;
                     log += " " + trans;
                     break;
                  case TRANSLATE_POP:
                     log += " " + trans;
                     trans--;
                     break;
                  // batch bracketing:
                  case START_BATCH:
                     waitForBatch++;
                     log += " " + waitForBatch;
                     break;
                  case END_BATCH:
                     log += " " + waitForBatch;
                     waitForBatch--;
                     break;
               }
               
               LOG.log(Level.FINEST, "PaintPrimitives." + log + "\n");
               boolean DEBUG = false;
               if (DEBUG)
               {
                  System.out.println(log);
               }
               
               switch (ps.id)
               {
                  case FLUSH:
                     flush();
                     break;
                  
                  case START_BATCH:
                  case END_BATCH:
                     break;
                  
                  default:
                     draw(ps);
                     break;
               }
            }
         }
         
         flush();
         
         LOG.log(Level.FINEST, "END DRAWING CYCLE FOR WINDOW " + windowId);
      }
   }
}
