/*
** Module   : GuiOutputManager.java
** Abstract : Provides a generic interface to low level GUI primitives.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 CA  20140805 Created initial version.  ThinClient is imported explicitly.
** 002 VIG 20140826 Changed default window ID constant to dynamic ID definition.
** 003 MAG 20140912 Implement GUI status bar.
** 004 HC  20140911 GUI coordinate support - data types refactoring.
** 005 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 006 MAG 20140925 Check if window has renderer when process default window.
** 007 CA  20141021 Removed the repaint(Widget) API: all drawing needs to be done via the common
**                  event processing.   Added translatedOrigin, to inform the caller about the 
**                  real location of the current origin point.
** 008 HC  20141021 Implemented window move/position support.
** 009 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 010 CA  20141123 Added API to check if a certain key can raise a STOP condition.
** 011 CA  20141125 Added API to conditionally enable/disable OS-level events.
** 012 EVL 20150107 Fix for dynamic windows renderer missing exception.
** 013 HC  20150323 WidgetId.getId() renamed to asInt().
** 014 GES 20150425 Match refactoring changes elsewhere.
** 015 HC  20150514 Support for modal windows.
** 016 HC  20150702 Implemented window modality independently on the actual UI driver.
** 017 GES 20150730 Removed translatedOrigin() since it isn't actually used.
** 018 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 019 HC  20151013 enableOSEvents() moved to WindowManager.
** 020 IAS 20160217 Special handling of the overlay windows.
** 021 HC  20160926 Removed disabling of window input events during server processing.
** 022 CA  20170104 Added flag to track window's modal state.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Provides a generic interface to low level graphic user interface primitives.
 */
public class GuiOutputManager
extends OutputManager<GuiPrimitives>
{
   /** Widget registry. */
   private WidgetRegistry<GuiOutputManager> registry = new WidgetRegistry<GuiOutputManager>();

   /** The initial configuration of this manager. */
   private final BootstrapConfig cfg;
   
   /**
    * Create a new output manager instance, with the specified driver.
    *
    * @param    driver
    *           Implements low level input/output primitives.  Must NOT be <code>null</code>.
    * @param    cfg
    *           The initial configuration.
    */
   public GuiOutputManager(GuiDriver driver, BootstrapConfig cfg)
   {
      super(driver);
      
      this.cfg = cfg;
   }
   
   /**
    * Get the context-specific instance, if it exists. In a single-client
    * JVM environment, this is equivalent to a singleton.
    * 
    * @return   The instance or <code>null</code> if none exists.
    */
   public static synchronized GuiOutputManager instance()
   {
      return (GuiOutputManager) defaultTerminal.get();
   }

   /**
    * Process the default window, depending on the output manager type.
    * 
    * @param    window
    *           A window.
    */
   @Override
   public void activateWindow(TopLevelWindow<?> window)
   {
      int windowId = window.getId().asInt();
  
      // this part of code need to be called for dynamically created windows not yet active to
      // register simulator
      realizeWindow(windowId);
   }

   /**
    * Realize new window, depending on the output manager type.  This is not default window but
    * one is created dynamically.
    * 
    * @param    windowId
    *           A window ID to be realized.
    */
   @Override
   public void realizeWindow(int windowId)
   {
      GuiDriver driver = (GuiDriver) this.driver;
      
      if (!driver.hasWindowEmulator(windowId))
      {
         driver.registerWindow(windowId);
      }
   }
   
   /**
    * Realize new child window.
    * 
    * @param    ownerId
    *           The owning window ID.
    * @param    windowId
    *           A window ID to be realized.
    * @param    windowTitle
    *           Title of the realized window.
    * @param    resizable
    *           <code>true</code> if the window is resizable.
    * @param    isShareActivationWithOwner
    *           Indicates whether the window shares activation state with its owner.
    */
   public void realizeChildWindow(int ownerId, 
                                  int windowId, 
                                  String windowTitle, 
                                  boolean resizable, 
                                  boolean isShareActivationWithOwner)
   {
      realizeChildWindow(ownerId, windowId, windowTitle, 
                         resizable, isShareActivationWithOwner, false);
   }
   
   /**
    * Realize new child window.
    * 
    * @param    ownerId
    *           The owning window ID.
    * @param    windowId
    *           A window ID to be realized.
    * @param    windowTitle
    *           Title of the realized window.
    * @param    resizable
    *           <code>true</code> if the window is resizable.
    * @param    isShareActivationWithOwner
    *           Indicates whether the window shares activation state with its owner.
    * @param    modal
    *           <code>true</code> if this is a modal window.
    */
   public void realizeChildWindow(int ownerId, 
                                  int windowId, 
                                  String windowTitle, 
                                  boolean resizable, 
                                  boolean isShareActivationWithOwner,
                                  boolean modal)
   {
      GuiDriver driver = (GuiDriver) this.driver;
      
      if (!driver.hasWindowEmulator(windowId))
      {
         driver.registerChildWindow(ownerId, windowId, windowTitle, resizable, 
                                    isShareActivationWithOwner, modal);
      }
   }

   /**
    * Limit the available drawing area to the specified boundaries.
    * 
    * @param    bounds
    *           The bounds for the drawing area.
    */
   public void setDrawingArea(Rectangle bounds)
   {
      NativeRectangle nbounds = bounds == null ? null : coordinates().rectangleToNative(bounds);
      ((GuiDriver) driver).setDrawingArea(nbounds);
   }
   
   /**
    * Clear area occupied by specified widget.
    * 
    * @param   widget
    *          Reference widget. 
    */
   @Override
   public void clear(Widget<?> widget)
   {
      if (widget == null || isRedirected())
         return;

      // TODO: which color?
      int windowId = widget.topLevelWindow().getId().asInt();
      
      clear(windowId, widget.physicalLocation(), widget.physicalDimension(), null);
   }

   /**
    * Access driver-specific widget factory.
    * 
    * @return  reference to widget factory.
    */
   @Override
   public GuiWidgetFactory getFactory()
   {
      return (GuiWidgetFactory) driver.getFactory();
   }

   /**
    * Get the widget registry for this output manager.
    * 
    * @return   See above.
    */
   @Override
   public WidgetRegistry<GuiOutputManager> getRegistry()
   {
      return registry;
   }

   /**
    * Set status text message.
    * 
    * @param   message
    *          Text message.
    */
   public void setStatusText(String message)
   {
      GuiDriver screen = (GuiDriver) driver;
      screen.setStatusText(message);
   }   
   
   /**
    * Set status text message of specified window.
    * 
    * @param   message
    *          Text message.
    * @param   window
    *          Window instance.
    */
   public void setStatusText(Window<?> window, String message)
   {
      GuiDriver screen = (GuiDriver) driver;
      screen.setStatusText(window, message);      
   }   

   /**
    * Check if the given key can raise the STOP condition.
    * 
    * @param    key
    *           The key code to be checked.
    * 
    * @return   <code>true</code> if the key is the {@link Key#VK_CTRL_BREAK} key.
    */
   @Override
   protected boolean raiseStopCondition(int key)
   {
      return key == Key.VK_CTRL_BREAK;
   }
   
   /**
    * Check if the window being drawn needs to be overridden.
    * <p>
    * In GUI, the <code>window</code> parameter is returned.
    *  
    * @param    window
    *           The currently active window.
    *           
    * @return   See above.
    */
   protected TitledWindow<?> overrideWindow(TopLevelWindow<?> window)
   {
      return window;
   }
   
   /**
    * On a device which allows multiple windows to be created, this allows selecting an explicit
    * window before using the primitives or before drawing a widget.
    * 
    * @param    windowId
    *           Window ID.
    */
   @Override
   public void selectWindow(int windowId)
   {
      ((GuiDriver) driver).selectWindow(windowId);
   }
   
   /**
    * On a device which allows multiple windows to be created, this allows releasing the window to
    * allow drawing from other threads.
    */
   @Override
   public void releaseWindow()
   {
      ((GuiDriver) driver).releaseWindow();
   }
}