/*
** Module   : ImageWrapper.java
** Abstract : contains the state for a native image
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.

**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20150420 First version.
** 002 SBI 20151012 Added unique ids support.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver;

import java.util.concurrent.atomic.*;

/**
 * Contains the state for a native image which can be accessed in a general purpose manner.
 */
public class ImageWrapper<I>
{
   /** Represents a images sequence that holds the next free sequence id. */
   private final static AtomicInteger IMAGES_SEQUENCE = new AtomicInteger(1);
   
   
   /**
    * Returns the next unique image id within this class loader.
    * 
    * @return   The next unigue image id.
    */
   private static final int getImageUniqueId()
   {
      return IMAGES_SEQUENCE.getAndIncrement();
   }
   
   /** The image unique id.*/
   private final int uniqueId;
   
   /** Contained native image. */
   private I image = null;
   
   /** Image height in pixels. */
   private int height = 0;
   
   /** Image width in pixels. */
   private int width = 0;
   
   /**
    * Construct an instance.
    *
    * @param    image
    *           The contained native image.
    * @param    height
    *           Image height in pixels.
    * @param    width
    *           Image width in pixels.
    */
   public ImageWrapper(I image, int height, int width)
   {
      this.uniqueId = getImageUniqueId();
      this.image  = image;
      this.height = height;
      this.width  = width;
   }

   /**
    * Obtain the contained native image instance.
    *
    * @return   The contained image.
    */
   public I getImage()
   {
      return image;
   }
   
   /**
    * Obtain the height of the image in pixels.
    *
    * @return   The height in pixels.
    */
   public int getHeight()
   {
      return height;
   }
   
   /**
    * Obtain the width of the image in pixels.
    *
    * @return   The width in pixels.
    */
   public int getWidth()
   {
      return width;
   }

   /**
    * Returns the image unique id.
    * 
    * @return The image unique id.
    */
   public int getUniqueId()
   {
      return uniqueId;
   }
}