/*
** Module   : MouseResizeable.java
** Abstract : Implementation of a mouse resizeable widget.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.
** 002 HC  20140911 GUI coordinate support - data types refactoring.
** 003 CA  20141021 Removed the window-specific code.  This class is now generic enough to be used
**                  by any widget.
** 004 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 005 HC  20150323 WidgetId.getId() renamed to asInt().
** 006 GES 20150501 Removed AWT dependencies and moved to a more generic package.
** 007 CA  20150731 When terminating a resize, mouseReleased must simulate a mouse move, to adjust
**                  the mouse cursor.
** 008 HC  20150831 Fixed cursor position to be stable in respect to the widget corner while
**                  dragging.
** 009 CA  20150918 Mouse processing improvements.
** 010 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 011 CA  20151013 This class is Swing specific, moved to the swing driver.
** 012 IAS 20151214 LOAD-MOUSE-POINTER support
** 013 HC  20160428 GUI window border drawing changed to use the Border interface mechanism. Also
**                  removed hardcoded assumptions of the border size.
** 014 SBI 20160712 Changed to fix the mouse pointer cursor if the frame cursor is set.
** 015 SBI 20160719 Fixed NPE if the mouse source is not found.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver.swing;
                                   
// this usage of AWT events (and Point and Rectangle) is intentional, they are being used
// separately from the AWT/Swing event model; no other AWT/Swing usage is occurring in this class
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.*;
import java.util.*;

import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Listener for mouse actions which can trigger a widget to be resized.
 * <p>
 * Note that when a widget is registered to be mouse resizable, only its
 * top-level window can be really resized. This limits the class to just
 * one use case - resizable window border. 
 */
public class MouseResizeable
extends MouseWidgetAction                               
{
   /** The resizable window. */
   private GuiWindow window;
   
   /** Flag indicating if resize is currently active. */
   private boolean canResize = false;
   
   /** The active mouse cursor type. */
   private MousePtrWrapper activeCursor;

   /** The initial resize bounds. */
   private NativeRectangle initialBounds;

   /** Cached initial resize event. */
   private MouseEvent resizeEvent;
   
   /**
    * Create a new mouse resizeable action, for the specified window.
    * 
    * @param    wid
    *           The widget ID.
    */
   public MouseResizeable(int wid)
   {
      super(wid, Integer.MIN_VALUE);
      this.window = (GuiWindow) widget.topLevelWindow();
   }

   /**
    * Notification of a mouse pressed event for this {@link MouseWidgetAction#widget}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      if (canResize && e.getButton() == MouseEvent.BUTTON1)
      {
         // dragging has started - save the position
         resizeEvent = e;
         
         initialBounds = window.physicalBounds();
      }
   }

   /**
    * Notification of a mouse released event for this {@link MouseWidgetAction#widget}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      resizeEvent = null;
      
      // simulate mouse move to restore/set cursor
      mouseMoved(e);
   }

   /**
    * Notification of a mouse dragged event for this {@link MouseWidgetAction#widget}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {                                                  
      if (resizeEvent == null)
      {
         return;
      }
      
      if (canResize)
      {
         // check which cursor is in use
         boolean setLocationX = activeCursor == MousePtrWrapper.NW_RESIZE ||
                                activeCursor == MousePtrWrapper.W_RESIZE  ||
                                activeCursor == MousePtrWrapper.SW_RESIZE;
         boolean setLocationY = activeCursor == MousePtrWrapper.NW_RESIZE  ||
                                activeCursor == MousePtrWrapper.N_RESIZE   ||
                                activeCursor == MousePtrWrapper.NE_RESIZE;
         
         // depending on the dragging point, it will resize x axis, y axis or both
         boolean resizeX = activeCursor == MousePtrWrapper.NW_RESIZE || 
                           activeCursor == MousePtrWrapper.NE_RESIZE ||
                           activeCursor == MousePtrWrapper.SW_RESIZE ||
                           activeCursor == MousePtrWrapper.SE_RESIZE ||
                           activeCursor == MousePtrWrapper.E_RESIZE  ||
                           activeCursor == MousePtrWrapper.W_RESIZE;
         boolean resizeY = activeCursor == MousePtrWrapper.NW_RESIZE || 
                           activeCursor == MousePtrWrapper.NE_RESIZE ||
                           activeCursor == MousePtrWrapper.SW_RESIZE ||
                           activeCursor == MousePtrWrapper.SE_RESIZE ||
                           activeCursor == MousePtrWrapper.N_RESIZE  ||
                           activeCursor == MousePtrWrapper.S_RESIZE;
         int signX = activeCursor == MousePtrWrapper.W_RESIZE ||
                     activeCursor == MousePtrWrapper.NW_RESIZE ||
                     activeCursor == MousePtrWrapper.SW_RESIZE ? 1 : -1;
         int signY = activeCursor == MousePtrWrapper.N_RESIZE ||
                     activeCursor == MousePtrWrapper.NW_RESIZE ||
                     activeCursor == MousePtrWrapper.NE_RESIZE ? 1 : -1;
   
         // determine the new origin
         NativePoint origin = null;
         
         if (setLocationX || setLocationY)
         {
            if (!(setLocationX && setLocationY))
            {
               origin = window.physicalLocation();
            }
            else
            {
               origin = new NativePoint(0, 0);
            }
            
            if (setLocationX)
            {
               origin.x = e.getXOnScreen();
            }
            
            if (setLocationY)
            {
               origin.y = e.getYOnScreen();
            }
         }
         
         // determine the new size
         int dx = initialBounds.width();
         int dy = initialBounds.height();
         
         if (resizeX)
         {
            dx -= signX * (e.getXOnScreen() - resizeEvent.getXOnScreen());
         }
         
         if (resizeY)
         {
            dy -= signY * (e.getYOnScreen() - resizeEvent.getYOnScreen());
         }
         
         gd.selectWindow(window.getId().asInt());

         gd.startBatch();
         
         try
         {
            window.resizeTo(origin, new NativeDimension(dx, dy), true);
         }
         finally
         {
            gd.endBatch();
            gd.releaseWindow();
         }
      }
   }

   /**
    * Notification of a mouse exited event for this {@link MouseWidgetAction#widget}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseExited(MouseEvent e)
   {
      if (resizeEvent == null)
      {
         gd.selectWindow(window.getId().asInt());
         
         // find a widget under the mouse pointer
         Widget<GuiOutputManager> widget = window.findMouseSource(new NativePoint(e.getX(), e.getY()));
         
         Optional<MousePtrWrapper> frameCursor;
         
         if (widget != null)
         {
            frameCursor = widget.getCustomMousePointer();
         }
         else
         {
            frameCursor = Optional.empty();
         }
         
         MousePtrWrapper windowCursor = window.getEffectiveMousePointer();
         
         if (frameCursor.isPresent())
         {
            activeCursor = frameCursor.get();
         }
         else
         {
            activeCursor = windowCursor;
         }
         
         gd.setCursor(activeCursor);
         canResize = false;

         gd.releaseWindow();
      }
   }

   /**
    * Notification of a mouse moved event for this {@link MouseWidgetAction#widget}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseMoved(MouseEvent e)
   {
      Point p = e.getPoint();
      com.goldencode.p2j.ui.client.NativeRectangle bounds = window.physicalBounds();
      
      NativeInsets nborder = NativeInsets.ZERO_INSETS;
      
      // if the window has border, get its insets
      Border<?> border = window.getBorder();
      if (border != null)
      {  
         nborder = window.screen().coordinates().insetsToNative(border.insets());
      }
      
      // we have 8 cases; corners have precedence
      MousePtrWrapper[] cursors =
      {
         MousePtrWrapper.NW_RESIZE,
         MousePtrWrapper.NE_RESIZE,
         MousePtrWrapper.SE_RESIZE,
         MousePtrWrapper.SW_RESIZE,
         MousePtrWrapper.N_RESIZE,
         MousePtrWrapper.E_RESIZE,
         MousePtrWrapper.S_RESIZE,
         MousePtrWrapper.W_RESIZE,
      };
      int[] xcoords1 =                         
      {
         bounds.left(),                  // NW
         bounds.right() - nborder.right, // NE
         bounds.right() - nborder.right, // SE
         bounds.left(),                  // SW
         bounds.left(),                  // N
         bounds.right() - nborder.right, // E
         bounds.left(),                  // S
         bounds.left(),                  // W
      };
      int[] xcoords2 =
      {
         bounds.left() + nborder.left,  // NW
         bounds.right(),                // NE
         bounds.right(),                // SE
         bounds.left() + nborder.left,  // SW
         bounds.right(),                // N
         bounds.right(),                // E
         bounds.right(),                // S
         bounds.left() + nborder.left,  // W
      };
      int[] ycoords1 =
      {
         bounds.top(),                     // NW
         bounds.top(),                     // NE
         bounds.bottom() - nborder.bottom, // SE
         bounds.bottom() - nborder.bottom, // SW
         bounds.top(),                     // N
         bounds.top(),                     // E
         bounds.bottom() - nborder.bottom, // S
         bounds.top(),                     // W
      };
      int[] ycoords2 =
      {
         bounds.top() + nborder.top,    // NW
         bounds.top() + nborder.top,    // NE
         bounds.bottom(),               // SE
         bounds.bottom(),               // SW
         bounds.top() + nborder.top,    // N
         bounds.bottom(),               // E
         bounds.bottom(),               // S
         bounds.bottom(),               // W
      };
      
      MousePtrWrapper cursor = MousePtrWrapper.DEFAULT;
      canResize = false;
      for (int i = 0; i < cursors.length; i++)
      {
         int x1 = xcoords1[i];
         int y1 = ycoords1[i];
         int x2 = xcoords2[i];
         int y2 = ycoords2[i];
         
         Rectangle r = new Rectangle(x1, y1, x2 - x1 + 1, y2 - y1 + 1);
         if (r.contains(p))
         {
            cursor = cursors[i];
            canResize = true;
            break;
         }
      }
      
      cursor = ThinClient.lockedMousePointer().orElse(cursor);
      
      gd.selectWindow(window.getId().asInt());

      if (canResize)
      {
         gd.setCursor(cursor);
         activeCursor = cursor;
      }
      else if (resizeEvent == null)
      {
         cursor = window.getEffectiveMousePointer();
         gd.setCursor(cursor);
         activeCursor = cursor;
      }

      gd.releaseWindow();
   }
}