/*
** Module   : SwingMouseHandler.java
** Abstract : Provides mouse event support for the swing GUI driver.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20150826 Created initial version.
** 002 CA  20150925 Fixed mouse entry/exit events for P2J widgets; fixed window resize cursor and
**                  abend. Fixed multi-window drawing issues caused by concurrent drawing access
**                  to the GuiDriver: all access to the driver is done exclusively, each thread
**                  taking and releasing ownership for the duration of the drawing operation or
**                  other GUI API invocation which requires access to the underlying physical
**                  window.
** 003 HC  20151013 NPE fix.
** 004 HC  20151024 Fixed race condition when dispatching mouse events.
** 005 SBI 20160310 Changed processAction/findMouseSource to deliver mouse events to the closest
**                  registered widget in the source mouse widget parent's hierarchy.
** 006 SVL 20160419 Fixed mouseMoved for destroyed widgets.
** 007 HC  20160428 Fixed MOUSE_EXITED event sometime not triggered, this fixes window resize
**                  cursor not always appearing when at window border. ALso MOUSE_EXITED is
**                  processed only after mouse dragging has finished.
** 008 HC  20160607 Prevented unexpected behavior when MOUSE_EXITED event sent to a destroyed
**                  widget. 
** 009 HC  20170203 Renamed TC.invokeLater() to TC.invokeLaterOS() and introduced
**                  TC.invokeLater() to allow execution of code to be queued on the application
**                  queue.
** 010 OM  20170407 Propagated mouse events to scroll buttons. Replaced anonymous classes with
**                  lambdas. Dispatched MouseEnter and MouseExit to widgets.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver.swing;

import java.awt.Component;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.WidgetId;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.EditorGuiImpl;
import com.goldencode.p2j.ui.client.gui.LabelGuiImpl;
import com.goldencode.p2j.ui.client.gui.ScrollPaneGuiImpl;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Provide mouse event support for the swing GUI driver.
 */
public class SwingMouseHandler
extends MouseHandler
{
   /** Context-local data for this mouse-handler. */
   private static final ContextLocal<WorkArea> local = new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      {
         return new WorkArea();
      };
   };
   
   /** The last widget over which the mouse was moved. */
   private Widget<?> lastHoveredWidget;
   
   /** Pressed flag. */
   private boolean pressed = false;
   
   /** Delayed mouseExited event. */
   private MouseEvent delayedMouseExited;

   /**
    * Notification of a mouse clicked event.  If the mouse event can't be processed by the driver
    * directly, it will post it to the event manager.
    * 
    * @param    e
    *           The mouse event.
    */
   public void mouseClicked(MouseEvent e)
   {
      if (!canProcessMouse())
      {
         return;
      }
      
      // serialize this call on the OS event queue to prevent race condition
      tc.invokeLaterOS(() ->
      {
         if (!processAction(e))
         {
            SwingMouseHandler.super.mouseClicked(e);
         }
      });
   }

   /**
    * Notification of a mouse pressed event.  If the mouse event can't be processed by the driver
    * directly, it will post it to the event manager.
    * 
    * @param    e
    *           The mouse event.
    */
   public void mousePressed(MouseEvent e)
   {
      if (!canProcessMouse())
      {
         return;
      }
      
      // serialize this call on the OS event queue to prevent race condition
      tc.invokeLaterOS(() ->
      {
         // mouseDragged is invoked AFTER mouseExited and so dragging state cannot be used
         // to delay mouseExited
         pressed = true;
         delayedMouseExited = null;
         
         if (!processAction(e))
         {
            SwingMouseHandler.super.mousePressed(e);
         }
      });
   }

   /**
    * Notification of a mouse released event.  If the mouse event can't be processed by the driver
    * directly, it will post it to the event manager.
    * 
    * @param    e
    *           The mouse event.
    */
   public void mouseReleased(MouseEvent e)
   {
      if (!canProcessMouse())
      {
         return;
      }
      
      // serialize this call on the OS event queue to prevent race condition
      tc.invokeLaterOS(() ->
      {
         pressed = false;
         
         if (delayedMouseExited != null)
         {
            mouseExited(delayedMouseExited);
            delayedMouseExited = null;
         }
         
         if (!processAction(e))
         {
            SwingMouseHandler.super.mouseReleased(e);
         }
      });
   }

   /**
    * Notification of a mouse entered event.  If the mouse event can't be processed by the driver
    * directly, it will post it to the event manager.
    * 
    * @param    e
    *           The mouse event.
    */
   public void mouseEntered(MouseEvent e)
   {
      if (!canProcessMouse())
      {
         return;
      }
      
      // serialize this call on the OS event queue to prevent race condition
      tc.invokeLaterOS(() ->
      {
         Widget<?> mouseSource = findMouseSource(e);
         lastHoveredWidget = mouseSource;
         
         if (mouseSource != null)
         {
            processAction(mouseSource, e); // calls mouseSource.mouseEntered(e);
            // call this unconditionally, as the widget needs to be notified that it was exited...
            tc.postMouseEvent(e, ews.getWindowId(), UiUtils.getWidgetIdAsInt(mouseSource));
         }
         
         if (!processAction(e))
         {
            SwingMouseHandler.super.mouseEntered(e);
         }
      });
   }
   
   /**
    * Notification of a mouse exited event.  If the mouse event can't be processed by the driver
    * directly, it will post it to the event manager.
    * 
    * @param    e
    *           The mouse event.
    */
   public void mouseExited(MouseEvent e)
   {
      if (!canProcessMouse())
      {
         return;
      }
      
      // serialize this call on the OS event queue to prevent race condition
      tc.invokeLaterOS(() ->
      {
         // make sure mouseExited is processed AFTER mouse-dragging is finished
         if (pressed)
         {
            delayedMouseExited = e;
            return;
         }
         
         Widget<?> mouseSource = findMouseSource(e);
         if (mouseSource != null)
         {
            processAction(mouseSource, e); // calls mouseSource.mouseExited(e);
            // call this unconditionally, as the widget needs to be notified that it was exited...
            tc.postMouseEvent(e, ews.getWindowId(), UiUtils.getWidgetIdAsInt(mouseSource));
         }
         
         lastHoveredWidget = null;
         
         if (!processAction(e))
         {
            SwingMouseHandler.super.mouseExited(e);
         }
      });
   }

   /**
    * Notification of a mouse wheel event.  If the mouse event can't be processed by the driver
    * directly, it will post it to the event manager.
    * 
    * @param    e
    *           The mouse event.
    */
   public void mouseWheelMoved(MouseWheelEvent e)
   {
      if (!canProcessMouse())
      {
         return;
      }
      
      // serialize this call on the OS event queue to prevent race condition
      tc.invokeLaterOS(() ->
      {
         if (!processAction(e))
         {
            SwingMouseHandler.super.mouseWheelMoved(e);
         }
      });
   }
   
   /**
    * Notification of a mouse dragged event.  If the mouse event can't be processed by the driver
    * directly, it will post it to the event manager.
    * 
    * @param    e
    *           The mouse event.
    */
   public void mouseDragged(MouseEvent e)
   {
      if (!canProcessMouse())
      {
         return;
      }
      
      // serialize this call on the OS event queue to prevent race condition
      tc.invokeLaterOS(() ->
      {
         if (!processAction(e))
         {
            SwingMouseHandler.super.mouseDragged(e);
         }
      });
   }

   /**
    * Notification of a mouse moved event.  If the mouse event can't be processed by the driver
    * directly, it will post it to the event manager.
    * <p>
    * If the hovered widget has changed, it will raise {@link MouseEvent#MOUSE_EXITED} and
    * {@link MouseEvent#MOUSE_ENTERED} events instead of {@link MouseEvent#MOUSE_MOVED}.
    * 
    * @param    e
    *           The mouse event.
    */
   public void mouseMoved(MouseEvent e)
   {
      if (!canProcessMouse())
      {
         return;
      }
      
      // serialize this call on the OS event queue to prevent race condition
      tc.invokeLaterOS(() ->
      {
         Widget<?> mouseSource = findMouseSource(e);
         
         Runnable exit = () ->
         {
            Component src = (Component) e.getSource();
            
            if (src.isDisplayable())
            {
               MouseEvent exite = new MouseEvent(src,
                                                 MouseEvent.MOUSE_EXITED,
                                                 e.getWhen(),
                                                 e.getModifiers(),
                                                 e.getX(),
                                                 e.getY(),
                                                 e.getClickCount(),
                                                 false,
                                                 e.getButton());
               // lastHoveredWidget.mouseExited(exite);
               processAction(lastHoveredWidget, exite);
               // call this unconditionally, as the widget needs to be notified that it was exit...
               tc.postMouseEvent(exite, ews.getWindowId(), UiUtils.getWidgetIdAsInt(lastHoveredWidget));
            }
            lastHoveredWidget = null;
         };
         
         // if the widget has changed, send EXITED/ENTERED events
         if (lastHoveredWidget != null && 
             (!lastHoveredWidget.isDisplayed() ||
              !lastHoveredWidget.getActualBounds().contains(e.getX(), e.getY())))
         {
            exit.run();
         }
         
         if (lastHoveredWidget != mouseSource &&
             mouseSource!= null               &&
             mouseSource.getActualBounds().contains(e.getX(), e.getY()))
         {
            if (lastHoveredWidget != null)
            {
               exit.run();
            }
            
            MouseEvent entere = new MouseEvent((Component) e.getSource(),
                                               MouseEvent.MOUSE_ENTERED,
                                               e.getWhen(),
                                               e.getModifiers(),
                                               e.getX(),
                                               e.getY(),
                                               e.getClickCount(),
                                               false,
                                               e.getButton());
            // mouseSource.mouseEntered(entere);
            processAction(mouseSource, entere);
            // call this unconditionally, as the widget needs to be notified that it was entered...
            tc.postMouseEvent(entere, ews.getWindowId(), UiUtils.getWidgetIdAsInt(mouseSource));
            
            lastHoveredWidget = mouseSource;
         }
         else
         {
            if (!processAction(e))
            {
               SwingMouseHandler.super.mouseMoved(e);
            }
         }
      });
   }

   /**
    * Deregister the complex actions for the specified widget ID.
    * 
    * @param    wid
    *           The widget ID.
    */
   void deregisterAction(int wid)
   {
      local.get().actions.remove(wid);
   }

   /**
    * Register a complex action for the specified widget ID.
    * <p>
    * Use this method (instead of the 
    * {@link MouseHandler#registerWidgetAction(int, MouseWidgetAction)}) to register
    * actions to Swing GUI-specific action map, i.e. the actions will be ignored
    * by other drivers.
    * 
    * @param    wid
    *           The widget ID.
    * @param    action
    *           The mouse action.
    */
   void registerAction(int wid, MouseWidgetAction action)
   {
      WorkArea wa = local.get();
      
      List<MouseWidgetAction> actionList = wa.actions.get(wid);
      if (actionList == null)
      {
         wa.actions.put(wid, actionList = new ArrayList<>());
      }
            
      actionList.add(action);
   }

   /**
    * Find the mouse source on which the specified event will act, based on coordinates.
    * 
    * @param    e
    *           The event.
    *           
    * @return   The mouse source, as returned by {@link Window#findMouseSource}.
    */
   private Widget<?> findMouseSource(MouseEvent e)
   {
      TopLevelWindow<?> window = (TopLevelWindow<?>) WindowManager.findWindow(ews.getWindowId());
      if (window == null)
      {
         // the window may have been gone when the event is generated
         // during window destroy
         return null;
      }
      
      Widget<?> mouseSource = window.findMouseSource(new MouseEvt(e, window));
      
      // find widget in its ancestor's hierarchy that has an id 
      Widget parent = WidgetRegistry.findAncestor(mouseSource, t ->
      {
         // TODO: to improve/simplify the mouse pointer is over a widget work around to skip labels
         return UiUtils.getWidgetId(t) != null && !(LabelGuiImpl.class.equals(t.getClass()));
      });
      
      // TODO: move this widget-specific logic into methods that are implemented in the widget
      // hierarchy so that we do not have direct references to widget classes in this common code
      if (parent instanceof ScrollPaneGuiImpl && parent.parent() instanceof EditorGuiImpl)
      {
         parent = parent.parent();
      }

      return parent;
   }

   /**
    * Determine the widget which will react on the specified event and process any actions
    * associated with it.
    * 
    * @param    e
    *           The mouse event.
    *           
    * @return   See above.
    */
   private boolean processAction(MouseEvent e)
   {
      return processAction(null, e);
   }

   /**
    * Process any actions for the given widget and event.
    * 
    * @param    widget
    *           The widget.  If <code>null</code>, it will be determined by 
    *           {@link #findMouseSource}.
    * @param    e
    *           The mouse event.
    * 
    * @return   See above.
    */
   private boolean processAction(Widget<?> widget, MouseEvent e)
   {
      if (widget == null)
      {
         widget = findMouseSource(e);
      }

      if (widget == null || !(widget.isVisible() && widget.isDisplayed()))
      {
         return false;
      }
      
      WidgetId wid = UiUtils.getWidgetId(widget);
      
      if (wid == null)
      {
         return false;
      }
      
      List<MouseWidgetAction> actionList = local.get().actions.get(wid.asInt());
         
      if (actionList == null)
      {
         return false;
      }
      
      for (MouseWidgetAction action : actionList)
      {
         if (!action.hasEvent(e))
         {
            continue;
         }
         
         // process all actions for this widget
         switch (e.getID())
         {
            case MouseEvent.MOUSE_CLICKED:
               action.mouseClicked(e);
               break;
            case MouseEvent.MOUSE_PRESSED:
               action.mousePressed(e);
               break;
            case MouseEvent.MOUSE_RELEASED:
               action.mouseReleased(e);
               break;
            case MouseEvent.MOUSE_ENTERED:
               action.mouseEntered(e);
               break;
            case MouseEvent.MOUSE_EXITED:
               action.mouseExited(e);
               break;
            case MouseEvent.MOUSE_DRAGGED:
               action.mouseDragged(e);
               break;
            case MouseEvent.MOUSE_MOVED:
               action.mouseMoved(e);
               break;
            case MouseEvent.MOUSE_WHEEL:
               action.mouseWheelMoved((MouseWheelEvent) e);
               break;
         }
         
         return true;
      }
      
      return false;
   }
   
   /**
    * Container for mouse-related data.
    */
   private static class WorkArea
   {
      /** Map of widgets with complex mouse actions which can be processed at the driver. */
      private final Map<Integer, List<MouseWidgetAction>> actions = new HashMap<>();
   }
}
