/*
** Module   : GuiWebDriver.java
** Abstract : WebSocket-based GUI driver
**
** Copyright (c) 2013-2017 Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 GES 20150325 Created initial version patterned off of the ChuiWebDriver.
** 002 CA  20150520 Changed signature for createFont.
** 003 HC  20150612 Compatibility changes related to support of modal windows.
** 004 HC  20150702 Implemented window modality independently on the actual UI driver.
** 005 GES 20150701 Added GUI implementation for previously stubbed-out methods,
**                  including image drawing, font and metrics-related APIs.
** 006 GES 20150818 Reworked z-order implementation and finished remaining stubbed methods.
**                  Added waitInitialization, which waits for a notification from the web page
**                  that it has finished initialized. Added mouse support. deriveFont() is
**                  now driver specific. It scales the font depending on the DPI.
** 007 CA  20150925 Fixed mouse entry/exit: explicit source is passed to the MouseEvt. Z-order
**                  fixes: the full z-order widget list must always be sent if there are changes
**                  in the widget tree (adjusting the new/dead widget list can not provide
**                  correct z-order information). Fixed multi-window drawing issues caused by
**                  concurrent drawing access to the GuiDriver: all access to the driver is done
**                  exclusively, each thread taking and releasing ownership for the duration of
**                  the drawing operation or other GUI API invocation which requires access to
**                  to the underlying physical window.
** 008 CA  20151013 Added support for window resize, z-order operations for the GUI web client.
** 009 SBI 20151108 Changed moveToTop, moveToBottom in order to check if the target window is
**                  visible. Added a mouse event modifier for left, middle and right mouse
**                  buttons.
** 010 SBI 20151124 Changed/added: a widget region is considered to be new if there doesn't exist
**                  in the last widgets cache or its bounded rectangle is changed.
**     SBI 20151201 Changed to provide processMouseWidgets with calculated widgets bounds, added
**                  the special EditorGuiImpl case to consider its dimension as its viewport size
**                  in order to fix a mouse cursor type over its scroll bars.
** 011 HC  20160104 Added the capability to identify active window in gui drivers.
** 012 EVL 20160106 Adding overlay window support.
** 013 EVL 20160122 Change the logic for setting active window.  Overlay window can not be active
**                  from the higher layer activation point of view.  Removing isWindowActive()
**                  method.  Logic is now in WindowManager. 
** 014 IAS 20160130 Renamed registerEditableWidget to registerHoverableWidget
**                  LOAD-MOUSE-POINTER and SET-WAIT-STATE support
** 015 EVL 20160301 Small javadoc fix.
** 016 SBI 20151208 Added new parameters, scroll amount and scroll unit for wheel events and
**                  a mouse modifiers mask for all mouse events.
**     SBI 20160321 Fixed an inconsistent state of the web page handlers to process requests by
**                  the dedicated web application handler, implemented new method
**                  isWindowDecorationSupported().
**     CA  20160323 Added parseEditorContent() and getTextWidths().
** 017 HC  20160407 Overhaul of window-activation logic. Removed setOverlayWindowLocation(), the
**                  method was unnecessary and was causing the COMBO-BOX drop-down location
**                  incorrect.
** 018 HC  20160413 Fixed top-level windows to always receive focus event when activated, also
**                  deduplicated processing of focus events.
** 019 SBI 20160418 Added openURL(String path) to open the target page in the browser. Fixed
**                  maximal window's width and height values if they are not set by the server.
**     CA  20160418 Added infrastructure for message passing from converted code to a 
**                  customer-specific application or calls from the customer-specific application 
**                  into the converted code (i.e. invoke external programs, procedures or 
**                  functions).  Can be used only when the P2J client runs embedded in a 
**                  customer-specific application.
** 020 SBI 20160427 Added graphicsCached parameter to turn off the graphics cache for testing.
**     SBI 20160506 Fixed RuntimeException if window id is -1.
** 021 SBI 20160708 Renamed "editable" to "hoverable" due to the implemented functionalities.
** 022 EVL 20161028 Fixed the case when visible widget is inside invisible frame.  In this case
**                  the mouse events should not be enabled and/or registered.
** 023 CA  20170104 Added flag to track window's modal state.
** 024 HC  20170202 Fixed potential race condition between GUI and WindowManager.
**     HC  20170203 Renamed TC.invokeLater() to TC.invokeLaterOS() and introduced
**                  TC.invokeLater() to allow execution of code to be queued on the application
**                  queue.
**     CA  20170203 Notify the driver that the client is ready (useful for embedded mode, to 
**                  notify the embedded app the client can handle messages).
**     CA  20170228 Added PUBLISH/SUBSCRIBE/UNSUBSCRIBE extensions for global support and for
**                  external applications.
** 025 CA  20170330 In embedded mode, get the real iframe dimension as desktop dimension.
** 026 EVL 20170811 Adding resetSelection() method to clean up web client after selection boxes
**                  drawing completed.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver.web;

/// TODO: remove this, it is there to allow compilation
import java.awt.Component;
import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.net.*;
import java.util.function.*;
// don't import the entire logging package as it has naming conflicts with jetty
import java.util.logging.Logger;
import java.util.stream.*;
import java.util.*;

import org.eclipse.jetty.util.resource.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.ui.client.driver.web.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.Utils.InstrumentedMap;
// don't import the entire logging package as it has naming conflicts with jetty

/**
 * GUI driver using WebSockets to communicate with a very thin javascript client hosted in
 * a browser. This class creates and uses an embedded web server to accomplish this.
 */
public class GuiWebDriver
extends AbstractGuiDriver<Integer>
implements EmbeddedWebServer,
           ClientProtocolHooks,
           EmbeddedClient
{
   /** Taskbar height in pixels when it exists. */
   public static final int TASKBAR_HEIGHT = 32;
   
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(GuiWebDriver.class.getName());

   /** Main page root target. */
   private static final String MAIN_TARGET = "/";

   /** Welcome page. */
   private static final String WELCOME_PAGE = "/index.html";
   
   /** A bogus AWT component used to raise mouse events, on notification from the JS code. */
   private static final Component BOGUS_AWT_COMPONENT = new Component()
   {
   };
   
   /** Actions for hoverable widgets */
   private static final Set<Integer> HOVER_ACTIONS =
         Collections.unmodifiableSet(
                  new HashSet<Integer>() {{
                     add(MouseEvent.MOUSE_ENTERED);
                     add(MouseEvent.MOUSE_EXITED);
                     add(MouseEvt.MOUSE_HOVERABLE);
                  }}
               );

   /** Configuration. */
   private BootstrapConfig config;
   
   /** Delgated implementation for the embedded web server. */
   private EmbeddedWebServerImpl websrv;
   
   /** Web-Socket which provides a GUI-specific protocol for communication with the browser. */
   private GuiWebSocket websock;
   
   /** Lock object for shared syncrhonization of the web socket. */
   private Object lock = new Object();
   
   /** Desktop bounds. */ 
   private NativeDimension desktop;
   
   /** Work-Area bounds. */ 
   private NativeDimension workarea;
   
   /** Virtual screen for drawing in a memory */
   private VirtualScreen virtualScreen;

   /** Maps the image seal to the unique image identity. */
   private final Map<List<Integer>, Integer > imageSeals;

   /**
    * Image usages is a mapping of an image loaded by a client to its owner windows.
    * The loaded image is represented by its image seal 6-tuple integer array.
    **/
   private final Map<List<Integer>, Set<Integer> > imageUsages;
   
   /** Owned images is a mapping of a window to its owned images seals. */
   private final Map<Integer, Set<List<Integer>> > ownedImages;
   
   /** A cache of font metrics objects, by their remote font ID. */
   private Map<Integer, FontMetricsHelper> fontMetricsCache = new HashMap<>();
   
   /** A counter for font IDs to be mapped on the js-side. */
   private int nextFontId = 0;
   
   /** 
    * A map with the last widgets sent to the JS side in [ windowid : [widgetId : actions] ]
    * form.
    */
   private Map<Integer, Map<Integer, Set<Integer>>> lastMouseWidgets = new HashMap<>();
   
   /** 
    * A map with the last widgets listening for "any mouse action" sent to the JS side in 
    * [ windowid : [widgetId(s) ... ] ] form.
    */
   private Map<Integer, Set<Integer>> lastAnyWidgets = new HashMap<>();
   
   /** Collects the mouse widgets belonging to a window. */
   private Map<Integer, Map<Integer, NativeRectangle>> widgetsRegions = new HashMap<>();

   /** 
    * A queue of pending repaint requests which needs to be processed by the 
    * {@link RepaintThread} thread.
    */
   private LinkedList<GuiWebEmulatedWindow> repaints = new LinkedList<>();
   
   /** The active window. */
   private int activeWindowId = -1;
   
   /** Indicates if the web client embedded mode is on or off. */
   private final boolean embedded;

   /** Flag indicating if the virtual desktop taskbar is shown. */
   private boolean taskbar = true;
   
   /**
    * Constructor.
    * 
    * @param   config
    *          The <code>BootstrapConfig</code> configuration.
    */
   public GuiWebDriver(BootstrapConfig config)
   {
      super(new GuiPrimitivesImpl()
      {
         @Override
         public void sync()
         {
            ews.repaint();
         };
      });
      
      this.config = config;
      this.embedded = config.getBoolean("client", "web", "embedded", false);
      this.imageUsages = new HashMap<List<Integer>, Set<Integer>>();
      this.ownedImages = new HashMap<Integer, Set<List<Integer>>>();
      this.imageSeals  = new HashMap<List<Integer>, Integer>();
      
      new RepaintThread().start();
   }
   
   /**
    * Initialize driver.
    */
   @Override
   public void init()
   {
      long timeout = -1;
      long wdtimeout = -1;
      int  height = -1;
      int  width  = -1;
      
      this.taskbar = true;
      boolean graphicsCached = true;
      
      if (config != null)
      {
         
         timeout   = config.getInt("client", "web", "socketTimeout", -1);
         wdtimeout = config.getInt("client", "web", "watchdogTimeout", -1);
         height    = config.getInt("client", "gui", "desktopHeight", height);
         width     = config.getInt("client", "gui", "desktopWidth", width);
         taskbar   = config.getBoolean("client", "gui", "taskbar", taskbar);
         graphicsCached = config.getBoolean("client", "gui", "graphicsCached", true);
      }
      
      if (!embedded)
      {
         // for non-embedded mode, leave the old way

         if (height == -1 || width == -1)
         {
            height = 768;
            width = 1024;
         }

         // initialize now
         desktop  = new NativeDimension(width, height);
         workarea = new NativeDimension(width, taskbar ? height - TASKBAR_HEIGHT : height);
         virtualScreen = new VirtualScreenImpl(width, height);
      }
      else
      {
         // these will be initialized when the clientReady is executed, to get the real dimension
         // dimension
         desktop = null;
         workarea = null;
         virtualScreen = null;
      }

      // calculate our resource base
      Resource cpres = EmbeddedWebServerImpl.calcResourceBase(GuiWebDriver.class, "res");
     
      // calculate our welcome page
      String welcomeUrl = getWelcomePageUrl(MAIN_TARGET, WELCOME_PAGE);
      
      // create our handlers
      GuiWebPageHandler[] hdlrs = new GuiWebPageHandler[1]; 
      
      hdlrs[0] = new GuiWebPageHandler(config, WELCOME_PAGE, welcomeUrl);
      
      // create our web socket
      websock = new GuiWebSocket(lock, this, timeout, wdtimeout, graphicsCached);
      
      // create the embedded server instance
      websrv = new EmbeddedWebServerImpl("GuiWebDriver",
                                         new GenericWebSocketCreator<WebClientProtocol>(websock),
                                         cpres,
                                         hdlrs);
   }
   
   /**
    * Notify the driver that the client is ready (useful for embedded mode, to notify the embedded
    * app the client can handle messages).
    */
   @Override
   public void clientReady()
   {
      websock.clientReady();
      
      // ensure these are initialized
      getDisplayWorkArea();
      getVirtualScreen();
   }

   /**
    * Called each time a connect occurs, this allows caller-controlled initialization of the
    * remote client.  There can only be 1 connection open at a time, but it is possible for
    * the connection to be opened, closed and then re-opened.
    * <p>
    * Whenever the main page is loaded or re-loaded, the WebSocket will connect and we must
    * put the virtual terminal into a known state.
    */
   @Override
   public void initRemoteClient()
   {
   }

   /**
    * Wait for the page to notify that it has fully initialized its JS code.
    */
   @Override
   public void waitInitialization()
   {
      websrv.waitInitialization(websock);
   }
   
   /**
    * This is a special API which is used to delegate a call from converted 4GL code 
    * (via _P2J_REMOTE_CALL_ functions) to the remote side.  Usually, this is used when P2J client
    * runs in embedded mode, i.e. the P2J screens are embedded in another customer-specific 
    * application.  The call will be remoted into the customer-specific application.
    * <p>
    * This API is expected to block waiting for a response.
    * 
    * @param    request
    *           The request data incoming from 4GL.
    * 
    * @return   The response from the customer-specific application, which will be sent back to 
    *           the converted 4GL code.
    */
   @Override
   public UnformattedPayload remoteCall(UnformattedPayload request)
   {
      return websock.remoteCall(request);
   }
   
   /**
    * Send a PUBLISH notification to a remote application. 
    * 
    * @param    subId
    *           The subscriber's ID, known by the remote application.
    * @param    event
    *           The published event name.
    * @param    param
    *           The parameters for this event.
    */
   @Override
   public void publish(long subId, String event, String[] param)
   {
      websock.publish(subId, event, param);
   }

   /**
    * Input key events when in VT100 mode.
    * 
    * @param    keyCode
    *           The key code from the input event.
    * @param    charCode
    *           The character code from the input event.
    */
   @Override
   public void injectVT100Key(int keyCode, char charCode)
   {
   }
   
   /**
    * Since web interface is inherently a single-display setup the method 
    * effectively returns a rectangle with top-level corner equal to 0,0.
    * 
    * @return   The physical screen bounds.
    */
   public NativeRectangle getDisplayBounds()
   {
      if (desktop == null)
      {
         desktop = websock.getDesktopDimension();
         
         if (desktop == null)
         {
            // if it couldn't be computed for any reason...
            desktop = new NativeDimension(1024, 768);
         }
      }
      
      return new NativeRectangle(new NativePoint(0, 0), desktop); 
   }
   
   /**
    * Returns its virtual screen that represents the physical screen in a memory
    * using a sRGB pixel model.
    * 
    * @return The virtual screen to draw images in a memory.
    */
   public VirtualScreen getVirtualScreen()
   {
      if (virtualScreen == null)
      {
         NativeRectangle bounds = getDisplayBounds();
         
         virtualScreen = new VirtualScreenImpl(bounds.width(), bounds.height());
      }
      
      return virtualScreen;
   }
   
   /**
    * Register a usage of the given image by the given window.
    * 
    * @param    imageHash
    *           The image hash id.
    * @param    windowId
    *           The window id.
    * 
    * @return   True iff it is the first usage of the given image.
    */
   public boolean addImageUsage(List<Integer> imageHash, int windowId)
   {
      boolean newLoadedImage = false;
      
      Set<Integer> usages = imageUsages.get(imageHash);
      if (usages == null)
      {
         usages = new HashSet<Integer>();
         imageUsages.put(imageHash, usages);
         newLoadedImage = true;
      }
      usages.add(windowId);
      
      Set<List<Integer>> images = ownedImages.get(windowId);
      if (images == null)
      {
         images = new HashSet<List<Integer>>();
         ownedImages.put(windowId, images);
      }
      images.add(imageHash);
      
      return newLoadedImage;
   }
   
   /**
    * Returns all unused images. Removes the given window from all image usages.
    * 
    * @param    windowId
    *           The window id.
    * 
    * @return The set of all unused images.
    */
   public Set<Integer> removeUnusedImages(int windowId)
   {
      Set<Integer> imagesToRemove = new HashSet<Integer>();
      Set<List<Integer>> images = ownedImages.remove(windowId);
      if (images != null && !images.isEmpty())
      {
         for (List<Integer> imageSeal : images)
         {
            Set<Integer> usages = imageUsages.get(imageSeal);
            if (usages.remove(windowId))
            {
               if (usages.isEmpty())
               {
                  Integer imageUniqueId = imageSeals.remove(imageSeal);
                  imagesToRemove.add(imageUniqueId);
                  imageUsages.remove(imageSeal);
               }
            }
         }
      }
      
      return imagesToRemove;
   }
   
   /**
    * Maps the image identity array to the unique image id.
    * 
    * @param    imageSeal
    *           The image identity array.
    * @param    uniqueId
    *           The image unique id.
    */
   public void mapSealToUniqueId(List<Integer> imageSeal, Integer uniqueId)
   {
      imageSeals.put(imageSeal, uniqueId);
   }
   
   /**
    * Gets the unique image id by its identity array.
    * 
    * @param    imageSeal
    *           The image identity array.
    * 
    * @return   The image unique id.
    */
   public Integer getUniqueIdForSeal(List<Integer> imageSeal)
   {
      return imageSeals.get(imageSeal);
   }
   
   /**
    * Returns an application-usable display area. For example, a task bar is
    * not counted as the usable area and it is always at the bottom of the
    * virtual screen (when present).
    * 
    * @return   The area in pixels.
    */
   public NativeRectangle getDisplayWorkArea()
   {
      if (workarea == null)
      {
         NativeRectangle bounds = getDisplayBounds();

         int width = bounds.width();
         int height = bounds.height();

         workarea = new NativeDimension(width, taskbar ? height - TASKBAR_HEIGHT : height);
      }
      
      return new NativeRectangle(new NativePoint(0, 0), workarea);
   }
   
   /**
    * Provide an audible feedback to user by generating a beep signal.
    */
   @Override
   public void beep()
   {
      if (websock != null)
      {
         websock.beep();
      }
   }

   /**
    * Read single keystroke from keyboard. Returned code should be either proper
    * Progress 4GL/P2J key code or -1 if environment-specific input cannot be 
    * properly translated.
    * 
    * @return   The pressed key code or -1 if no keys were pressed since last call.
    */
   @Override
   public int readKey()                      
   {
      return websock.readKey();
   }

   /**
    * Create a new window emulator.
    * 
    * @param    windowId
    *           The window ID.
    *
    * @return   The newly instantiated window emulator.
    */
   @Override
   public EmulatedWindowState newEmulator(int windowId)
   {
      GuiWebEmulatedWindow ews = new GuiWebEmulatedWindow(lock, websock, this);
         
      // the real window size is unknown at this time; it will be set later, when the layout
      // is performed
      ews.setScreenBitmap(new ScreenBitmap(0, 0));
      
      // remote setup
      websock.createWindow(windowId);
      
      return ews;
   }
   
   /**
    * Create a new child window emulator.
    * 
    * @param    windowId
    *           The window ID.
    * @param    parentWindow
    *           The parent window.
    * @param    windowTitle
    *           Title of the window emulator.
    * @param    resizable
    *           <code>true</code> if the window is resizable.
    * @param    isShareActivationWithOwner
    *           Indicates whether the window shares activation state with its owner.
    * @param    modal
    *           <code>true</code> if this is a modal window.
    *           
    * @return   The newly instantiated window emulator.
    */
   @Override
   public EmulatedWindowState newChildWindowEmulator(
                               int                 windowId,
                               EmulatedWindowState parentWindow, 
                               String              windowTitle,
                               boolean             resizable,
                               boolean             isShareActivationWithOwner,
                               boolean             modal)
   {
      // TODO: we may need to remember the window type (child vs top-level) in the GWEW
      GuiWebEmulatedWindow ews = new GuiWebEmulatedWindow(lock, websock, this);
         
      // the real window size is unknown at this time; it will be set later, when the layout
      // is performed
      ews.setScreenBitmap(new ScreenBitmap(0, 0));
      
      // remote setup
      websock.createChildWindow(windowId, parentWindow.getWindowId(), windowTitle, resizable, 
                                isShareActivationWithOwner, modal);
      
      return ews;
   }
   
   /**
    * Create a font with the given details and register it with the driver.
    * 
    * @param    details
    *           The font specifications.
    * 
    * @return   A font instance.
    */
   @Override
   public FontDetails<Integer> createFont(FontDetails<Integer> details)
   {
      FontDetails<Integer> res = (details.fontAlias == null ? details : details.fontAlias);
      
      // get the ID for this font
      res.font = ++nextFontId;
      
      int jsFont = websock.createFont(res.font,
                                      res.fontName,
                                      res.pointSize,
                                      res.style.toInt(),
                                      res.fontDefinition);
      
      res = details.copy();
      res.font = jsFont;
      res.pointSize = details.pointSize;
      res.fontName = details.fontName;
      
      return res;
   }
   
   /**
    * Scale the given font by the specified X and Y factors.
    *
    * @param    fd
    *           Font to scale.  This instance's contained font will be modified. 
    */
   @Override
   public void deriveFont(FontDetails<Integer> fd)
   {
      // this is done in place, as is expected for deriveFont to be called always after createFont
      // is called
      websock.deriveFont(fd.font);
   }

   /**
    * Reports the screen resolution in dots per inch (DPI).
    *
    * @return   The DPI of the screen.
    */
   @Override
   public int getScreenResolution()
   {
      // web browsers use a "CSS pixel" which is defined as 96 DPI; this may differ from the
      // actual number of pixels on the device; in other words the CSS pixel is device
      // independent and is the basis for everything we do; the only output that may not
      // be handled as well in such an environment is bitmap rendering, which may not scale
      // well to the device-specific resolution
      return 96;
   }
   
   /**
    * Copy the given text to the system clipboard.
    * 
    * @param    txt
    *           The text to be copied.
    */
   @Override
   public void copyToClipboard(String txt)
   {
      websock.writeClipboard(txt);
   }
   
   /**
    * Get the current contents from the OS clipboard.
    * 
    * @return   The clipboard contents.
    */
   @Override
   public String clipboardContents()
   {
      return websock.readClipboard();
   }
   
   /**
    * Force the list of all GUI windows to be stacked as in the specified order.
    * 
    * @param    windowIds
    *           The stack order to be enforced.
    */
   @Override
   public void restackWindows(int[] windowIds)
   {
      websock.restackWindows(windowIds);
   }
   
   /**
    * Move the specified window to the top of the z-order, at the OS level.
    *
    * @param    windowId
    *           The window to be moved.
    * @param    focus
    *           If <code>true</code>, the window will become the focused window 
    *           and its state changed to active. 
    */
   @Override
   public void moveToTop(int windowId, boolean focus)
   {
      EmulatedWindowState<?, ?> window = direct.getWindowEmulator(windowId);
      if (!window.isWindowVisible())
      {
         return;
      }
      websock.moveToTop(windowId);
      
      if (focus)
      {
         // activate the window and notify the upper layers;
         // do not wait for the JS layer to notify us back as
         // the focus request always succeeds
         windowActivated(activeWindowId, windowId);
      }
   }
   
   /**
    * Move the specified window to the bottom of the z-order, at the OS level.
    *
    * @param    windowId
    *           The window to be moved.
    */
   @Override
   public void moveToBottom(int windowId)
   {
      EmulatedWindowState window = direct.getWindowEmulator(windowId);
      if (!window.isWindowVisible())
      {
         return;
      }
      websock.moveToBottom(windowId);
   }
   
   /**
    * Perform driver shutdown.
    */
   @Override
   public void shutdown()
   {
      // close the emulated windows first
      super.shutdown();
      
      // notify the remote client about application exit    
      websock.quit();
      
      // since the application will exit shutdown the server
      shutdownServer();
   }
   
   /**
    * Startup the embedded server using the server's key store to build the web server SSL
    * context.
    * <p>
    * There are two page handlers.  One is for the root ("/") target and the other is for
    * the index ("/index.html") target.  Both load the same page index.html.
    * <p>
    * When the user is first redirected the URL is <code>https://server:port/</code> and the
    * target is "/". The index.html page is loaded for this target.
    * <p>
    * After redirection and the main page is loaded, the URL in the browser address bar will
    * become <code>https://server:port/index.html</code> and the corresponding target is now
    * "/index.html".
    * <p>
    * When the user reloads the main page within the browser environment the "/index.html"
    * target is sent to the web server and the same main page index.html is loaded. 
    * 
    * @param    keyStore
    *           The exported server KeyStore. 
    * @param    config
    *           Client configuration.
    * @param    port
    *           An explicit port to start the web server or 0 to let the OS assign one.
    * @param    host
    *           An explicit host to start the web server or <code>null</code> to use the server's
    *           host.
    */
   @Override
   public String startupServer(ServerKeyStore keyStore, BootstrapConfig config, String host, int port)
   {
      return websrv.startupServer(keyStore, config, host, port);
   }
   
   /**
    * Shutdown the embedded web server.
    */
   @Override
   public void shutdownServer()
   {
      websrv.shutdownServer();
   }
   
   /**
    * Get the embedded web server's URI.
    * 
    * @return   A string representing the server's URI.
    */
   @Override
   public String getServerURI()
   {
      return websrv.getServerURI();
   }
   
   /**
    * Enable or disable capture of mouse events.
    * <p>
    * It also notifies the JS side to enable/disable mouse events.
    *  
    * @param    state
    *           {@code true} to capture mouse events, {@code false} to ignore them.
    *           
    * @return   The previous state.
    */
   @Override
   public boolean captureMouseEvents(boolean state)
   {
      boolean prev = super.captureMouseEvents(state);
      
      websock.captureMouseEvents(state);
      
      return prev;
   }
   
   /**
    * Raise a mouse event.
    * 
    * @param    windowId
    *           The window ID.
    * @param    mouseOp
    *           The mouse event code.
    * @param    tstamp
    *           The event timestamp.
    * @param    mouseX
    *           The absolute mouse X for this event.
    * @param    mouseY
    *           The absolute mouse Y for this event.
    * @param    button
    *           The button which was pressed.
    * @param    clickCount
    *           The click count.
    * @param    modifiersMask
    *           The bit mask that represents the modifiers keys are pressed. The 0-bit corresponds
    *           to the shift key, the 1-bit to the ctrl key, the 2-bit to the meta key and
    *           the 3-bit to the alt key.
    * @param    wheelParameters
    *           The array that represents the wheel rotation amount and its unit.
    * @param    widgetId
    *           The explicit widget ID to which this event needs to be posted.
    */
   @Override
   public void raiseMouseEvent(int   windowId,
                               int   mouseOp, 
                               long  tstamp, 
                               int   mouseX, 
                               int   mouseY, 
                               int   button,
                               int   clickCount, 
                               int   modifiersMask,
                               int[] wheelParameters,
                               int   widgetId)
   {
      MouseOps mop = MouseOps.fromCode(mouseOp);
      
      int modifiers = getMouseModifiers(button, modifiersMask);
      
      MouseEvent event;
      if (mop.isWheel())
      {
         // the scroll unit
         int unit = MouseWheelEvent.WHEEL_UNIT_SCROLL;
         // the scroll direction, 1 means down or right and -1 means up or left 
         int direction = 1;
         // the scroll amount in the scroll unit
         int scrollAmount = 0;
         if (wheelParameters != null && wheelParameters.length == 2)
         {
            direction    = (wheelParameters[0] > 0) ? 1 : -1;
            scrollAmount = Math.abs(wheelParameters[0]);
            if (wheelParameters[1] == 2)
            {
               unit = MouseWheelEvent.WHEEL_BLOCK_SCROLL;
            }
         }
         event = new MouseWheelEvent(BOGUS_AWT_COMPONENT, 
                                     mop.awtEventCode, 
                                     tstamp, 
                                     modifiers, 
                                     mouseX, 
                                     mouseY, 
                                     mouseX, 
                                     mouseY, 
                                     clickCount, 
                                     false, 
                                     unit, 
                                     scrollAmount, 
                                     direction);
      }
      else
      {
         event = new MouseEvent(BOGUS_AWT_COMPONENT, 
                                mop.awtEventCode, 
                                tstamp, 
                                modifiers, 
                                mouseX, 
                                mouseY, 
                                mouseX, 
                                mouseY,
                                clickCount, 
                                false,
                                button);
      }
      
      EmulatedWindowState window = direct.getWindowEmulator(windowId);
      ((WebMouseHandler) window.getMouseHandler()).raiseMouseEvent(event, widgetId);
   }

   /**
    * Gets the mouse modifiers mask for this event.
    * 
    * @param    button
    *           The mouse button that is currently pressed
    * @param    modifiersMask
    *           The bit mask that represents the modifiers keys are pressed. The 0-bit corresponds
    *           to the shift key, the 1-bit to the ctrl key, the 2-bit to the meta key and
    *           the 3-bit to the alt key.
    * 
    * @return   The key's modifiers bit mask for this mouse event.
    */
   private int getMouseModifiers(int button, int modifiersMask)
   {
      int modifiers = 0;
      // button codes from JS (same as from AWT) 
      // 0 : No button
      // 1 : Left mouse button
      // 2 : Wheel button or middle button (if present)
      // 3 : Right mouse button
      switch (button)
      {
         case 1:
            modifiers = InputEvent.BUTTON1_MASK;
            break;
         case 2:
            modifiers = InputEvent.BUTTON2_MASK;
            break;
         case 3:
            modifiers = InputEvent.BUTTON3_MASK;
            break;
         default:
      }
      if ((modifiersMask & MouseEvent.SHIFT_MASK) != 0)
      {
         modifiers |= MouseEvent.SHIFT_DOWN_MASK; 
      }
      if ((modifiersMask & MouseEvent.CTRL_MASK) != 0)
      {
         modifiers |= MouseEvent.CTRL_DOWN_MASK; 
      }
      if ((modifiersMask & MouseEvent.META_MASK) != 0)
      {
         modifiers |= MouseEvent.META_DOWN_MASK; 
      }
      if ((modifiersMask & MouseEvent.ALT_MASK) != 0)
      {
         modifiers |= MouseEvent.ALT_DOWN_MASK; 
      }
      
      return modifiers;
   }
   
   /**
    * Set the location for the specified window.
    * 
    * @param    windowId
    *           The window ID.
    * @param    x
    *           The X coordinate of the top-left corner.
    * @param    y
    *           The Y coordinate of the top-left corner.
    */
   @Override
   public void setWindowLocation(int windowId, int x, int y)
   {
      TopLevelWindow<?> w = WindowManager.findWindow(windowId);
      ThinClient.getInstance().postOSEvent(new WindowMoved(w, x, y));
   }

   /**
    * Set the dimension for the specified window.
    * 
    * @param    windowId
    *           The window ID.
    * @param    width
    *           The window width.
    * @param    height
    *           The window height.
    */
   @Override
   public void setWindowDimension(int windowId, int width, int height)
   {
      synchronized (this)
      {
         // invalidate all widget locations
         lastMouseWidgets.clear();
         lastAnyWidgets.clear();
      }

      TopLevelWindow<?> w = WindowManager.findWindow(windowId);
      ThinClient.getInstance().postOSEvent(new WindowResized(w, width, height));
   }
   
   /**
    * Raise an event indicating that the target window has been activated or deactivated.
    * 
    * @param    deactivatingWindowId
    *           Id of the window being deactivated, may be negative.
    * @param    activatingWindowId
    *           Id of the window being activated, always valid.
    */
   @Override
   public void windowActivated(int deactivatingWindowId, int activatingWindowId)
   {
      if (activatingWindowId == activeWindowId)
      {
         return;
      }
      
      activeWindowId = activatingWindowId;
      if (deactivatingWindowId != -1)
      {
         // this always executes on the main thread no need to synchronize through
         // ThinClient.invokeLaterOS()
         WindowManager.windowDeactivated(deactivatingWindowId, false);
      }
      if (activatingWindowId != -1)
      {
         // this always executes on the main thread no need to synchronize through
         // ThinClient.invokeLaterOS()
         WindowManager.windowActivated(activatingWindowId, false);
      }
   }

   /**
    * Raise an event as the window was (de)iconified.
    * 
    * @param    windowId
    *           The window ID.
    * @param    state
    *           <code>true</code> if the window was iconified.
    */
   @Override
   public void windowIconified(int windowId, boolean state)
   {
      if (state)
      {
         iconifyWindow(windowId, true);
      }
      else
      {
         deiconifyWindow(windowId);
      }
   }
   
   /**
    * Register all the mouse-aware widgets (in all Windows) with the JS side.
    * <p>
    * To register a widget, it must be visible, displayed and enabled, and part of a visible
    * window.
    */
   @Override
   public synchronized void registerMouseWidgets()
   {
      // track widgets registered for complex mouse actions
      InstrumentedMap<Integer, Map<Integer, Set<Integer>>> allWidgets = 
                                                new InstrumentedMap<>(LinkedHashMap::new);
      InstrumentedMap<Integer, Map<Integer, Set<Integer>>> newWidgets = 
                                                new InstrumentedMap<>(LinkedHashMap::new);
      // track widgets bounds
      InstrumentedMap<Integer, Map<Integer, NativeRectangle>> newWidgetsBounds = 
                                                new InstrumentedMap<>(LinkedHashMap::new);
      // track widgets registered for "any mouse action"
      InstrumentedMap<Integer, Set<Integer>> newAnyWidgets = new InstrumentedMap<>(TreeSet::new);
      InstrumentedMap<Integer, Set<Integer>> allAnyWidgets = new InstrumentedMap<>(TreeSet::new);
      // track existed widgets: mouse widgets and any widgets
      Map<Integer,List<Integer>> existingMouseWidgets = new HashMap<>();
      Map<Integer,List<Integer>> existingAnyWidgets = new HashMap<>();
      
      Consumer<Widget> actionsCollector = widget ->
      {
         int widgetId = widget.getId().asInt();
         Set<Integer> webActions = getWebActions(widget);

         if (!webActions.isEmpty())
         {
            Integer windowId = widget.topLevelWindow().getId().asInt();
            allWidgets.modifyValue(windowId, actions -> {actions.put(widgetId, webActions);});
            
            // determine if the widget was previously registered with the web side
            // or it changed its bounds.
            boolean isExisted = isInLastMouseWidgets(widgetId, windowId);
            if (isExisted)
            {
               existingMouseWidgets.get(windowId).add(widgetId);
            }
            boolean isSizeChanged = updateCachedWidgetBounds(windowId, widget);
            if (!isExisted || isSizeChanged)
            {
               // new widget, track it
               newWidgets.modifyValue(windowId, actions -> {actions.put(widgetId, webActions);});

               // track its bounds
               newWidgetsBounds.modifyValue(windowId, 
                    bounds -> {bounds.put(widgetId, widgetsRegions.get(windowId).get(widgetId));}
               );
            }
         }

         boolean hasAnyAction = MouseHandler.hasAnyAction(widgetId);
         if (hasAnyAction)
         {
            Integer windowId = widget.topLevelWindow().getId().asInt();
            allAnyWidgets.modifyValue(windowId, actions -> {actions.add(widgetId);});
            
            if (!isInLastAnyWidgets(widgetId, windowId))
            {
               newAnyWidgets.modifyValue(windowId,  actions -> {actions.add(widgetId);});
            }
            else
            {
               // added the given any widget to the existed any widget map
               existingAnyWidgets.get(windowId).add(widgetId);
            }
         }
      };
      
      // go through all windows which are active/shown/etc and collect widgets,
      // title bars, message area widgets
      // collect all enabled, visible widgets.  the title bar widgets
      // are collected regardless if they are enabled or not
      // the visible widget inside invisible frame is considering as invisible
      // and is not including into result
     
      Predicate<Widget> isVisibleNonVirtual = widget -> 
                                                   UiUtils.getWidgetId(widget) != null      &&
                                                   widget.isVisible()                       &&
                                                   isInsideVisibleFrameOrStandalone(widget) &&
                                                   widget.isDisplayed()                     &&
                                                   !widget.isTotallyObscured();

      WindowManager.windowList().stream().filter(window -> 
                                                      window.isVisible()   &&
                                                      window.isDisplayed() &&
                                                      window instanceof TopLevelWindow
                                          ).
            forEach(
               window -> 
                  {
                     Integer windowId = window.getId().asInt();
                     Predicate<Widget> isValidChild = isVisibleNonVirtual.and(
                                           w -> windowId == w.topLevelWindow().getId().asInt()
                                        );
                           
                     existingMouseWidgets.put(windowId, new ArrayList<>());
                     existingAnyWidgets.put(windowId, new ArrayList<>());

                     // window border is managed by the JS side
                     if (GuiWindow.class.isInstance(window))
                     {
                        GuiWindow gw = (GuiWindow) window;
                        collectWidgets(gw.getTitleBar(), 
                              isValidChild, actionsCollector);
                        
                        collectWidgets((MessageAreaGuiImpl) gw.getMessageArea(), 
                              isValidChild, actionsCollector);
                     }
                     
                     collectWidgets(window, 
                           isValidChild.and(Widget::isEnabled), actionsCollector);
                  }
      );

      // remove existing mouse widgets from the last mouse widgets map
      //    to determine dead mouse widgets.
      existingMouseWidgets.keySet().stream().forEach(
               windowId -> lastMouseWidgets.getOrDefault(windowId, Collections.emptyMap()).
                  keySet().
                  removeAll(existingMouseWidgets.getOrDefault(windowId, Collections.EMPTY_LIST))
            );
      // remove existing any widgets from the last any widgets map
      //    to determine dead any widgets.
      existingAnyWidgets.keySet().stream().forEach(
            windowId -> lastMouseWidgets.getOrDefault(windowId, Collections.emptyMap()).
               keySet().
               removeAll(existingMouseWidgets.getOrDefault(windowId, Collections.EMPTY_LIST))
         );
      // remove all empty window maps from dead widgets
      Utils.removeAllIf(lastMouseWidgets.keySet(), wid -> lastMouseWidgets.get(wid).isEmpty());
      
      // register the new widgets and delete the widgets which are no longer in the set
      websock.processMouseWidgets(allWidgets, newWidgets, newWidgetsBounds, lastMouseWidgets,
               newAnyWidgets, lastAnyWidgets);
      
      // save the entire set
      lastMouseWidgets.clear();
      lastMouseWidgets.putAll(allWidgets);
      
      lastAnyWidgets.clear();
      lastAnyWidgets.putAll(allAnyWidgets);
   }

   /**
    * Split the text for a editor widget, so that it fits the specified maximum width.  The 
    * <code>delta</code> parameter represents the start position of this text, in the editor's 
    * content; when the full content is parsed, this will be zero.
    *  
    * @param    widgetId
    *           The editor's ID.
    * @param    windowId
    *           The editor's window ID.
    * @param    maxWidth
    *           The width for the content to fit (the editor's editable area).
    * @param    text
    *           The text to split into lines.
    * @param    delta
    *           The start position of this text in the editor's full content.
    * @param    font
    *           The editor's font.
    *           
    * @return   Two arrays: one with the split indexes and one with the split indexes which were
    *           determined to be caused by a new line.
    */
   @Override
   public int[][] parseEditorContent(int                  widgetId,
                                     int                  windowId,
                                     int                  maxWidth,
                                     String               text,
                                     int                  delta,
                                     FontDetails<Integer> font)
   {
      return websock.parseEditorContent(widgetId,
                                        windowId,
                                        maxWidth,
                                        text,
                                        delta,
                                        font.font);
   }
   
   /**
    * Reset selection boxes painting mode and return to normal drawing mode.
    */
   @Override
   public void resetSelection()
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.RESET_SELECTION);

      ews.offer(ps);
   }
   
   /**
    * Updates the cached widgets bounds and returns true iff the cached widgets bounds has been
    * updated.
    * 
    * @param    windowId
    *           The target window id.
    * @param    widget
    *           The target widget with its cached bounds.
    * 
    * @return   true value indicates that the widget's bounds has been changed and must be sent
    *           to the JS client.
    */
   private boolean updateCachedWidgetBounds(int windowId, Widget widget)
   {
      boolean updated = false;
      Integer widgetId = widget.getId().asInt();
      NativeDimension dim = widget.physicalDimension();
      if (widget instanceof EditorGuiImpl)
      {
         Widget scrollPane = WidgetRegistry.findDescendant(widget, ScrollPaneGuiImpl.class);
         dim = ((ScrollPaneGuiImpl) scrollPane).viewport().physicalDimension();
      }
      NativeRectangle actualBounds = new NativeRectangle(widget.screenPhysicalLocation(), dim);
      Map<Integer, NativeRectangle> cachedRegions = widgetsRegions.get(windowId);

      if (cachedRegions == null)
      {
         widgetsRegions.put(windowId, cachedRegions = new LinkedHashMap<>());
         updated = true;
      }
      else
      {
         NativeRectangle bounds = cachedRegions.get(widgetId);
         if (bounds == null)
         {
            updated = true;
         }
         else
         {
            updated = !bounds.equals(actualBounds);
         }
      }

      if (updated)
      {
         cachedRegions.put(widgetId, actualBounds);
      }

      return updated;
   }

   /**
    * Returns the set of mouse actions handled by this widget.
    * 
    * @param    widget
    *           The collected mouse widget.
    * @return   The web actions that are handled by the given widget.
    */
   private Set<Integer> getWebActions(Widget widget)
   {
      Set<Integer> mouseActions = new TreeSet<>();
      mouseActions.addAll(widget.getMouseActions());
      mouseActions.addAll(MouseHandler.getWidgetActions(widget.getId().asInt()));
      if (widget.supportsCustomMousePointer())
      {
         mouseActions.addAll(HOVER_ACTIONS);
      }
      
      return Arrays.stream(MouseOps.values()).
                     filter(op -> mouseActions.contains(op.awtEventCode)).
                     map(op -> op.code).collect(Collectors.toSet());
   }
   
   /**
    * Tests if the given mouse widget is in the last mouse widgets map.
    * 
    * @param    widgetId
    *           The widget id.
    * @param    windowId
    *           The window id.
    * 
    * @return   Returns true iff the given widget is in the target map.
    */
   private boolean isInLastMouseWidgets(Integer widgetId, Integer windowId)
   {
      // determine if the widget was previously registered with the web side
      return lastMouseWidgets.getOrDefault(windowId, 
            Collections.emptyMap()).containsKey(widgetId);
   }
   
   /**
    * Tests if the given any mouse widget is in the last any mouse widgets map.
    * 
    * @param    widgetId
    *           The widget id.
    * @param    windowId
    *           The window id.
    * 
    * @return   Returns true iff the given widget is in the target map.
    */
   private boolean isInLastAnyWidgets(Integer widgetId, Integer windowId)
   {
      return lastAnyWidgets.getOrDefault(windowId, Collections.emptySet()).contains(widgetId);
   }
   
   /**
    * Register the specified widget as a moveable widget, via mouse.
    * <p>
    * This is a no-op for the web client: the work is done by {@link #registerMouseWidgets()}.
    *  
    * @param    wid
    *           The widget ID.
    */
   @Override
   public void registerMoveableWidget(int wid)
   {
      // no-op
   }

   /**
    * Register this widget as being editable, so that the mouse cursor will adjust properly when
    * the widget is entered.
    * <p>
    * This is a no-op for the web client: the work is done by {@link #registerMouseWidgets()}.
    * 
    * @param    action
    *           The widget mouse action.
    */
   @Override
   public void registerHoverableWidget(MouseHoverAction action)
   {
      Widget<?> widget = action.widget();
      Widget<?> parent = widget.enclosingFrame().orElse(null);

      if (parent == null && (widget instanceof Frame))
      {
         parent = ((Frame)widget).window();
      }

      websock.registerHoverableWidget(
         widget.topLevelWindow().getId().asInt(), 
         widget.getId().asInt(), 
         parent != null ? parent.getId().asInt() : -1, 
         widget.getActualBounds(),
         widget.getLoadedMousePointer().orElse(null),
         widget.defaultMousePtr().orElse(MousePtrWrapper.DEFAULT),
         widget.enforceDefaultMousePtr()
      );
   }

   /**
    * Propagate wait state
    * 
    * @param isWaitState
    *        wait state
    */
   public void propagateWaitState(boolean isWaitState)
   {
      websock.propagateWaitState(isWaitState);
   }

   /**
    * Register the specified widget as a resizeable widget, via mouse.  Only WINDOW widgets are
    * supported.
    * <p>
    * This is a no-op for the web client: the work is done by {@link #registerMouseWidgets()}.
    * 
    * @param    wid
    *           The widget ID.
    */
   @Override
   public void registerResizeableWidget(int wid)
   {
      // no-op
   }

   /**
    * Deregister the specified widget for mouse events.
    * 
    * @param    wid
    *           The widget ID.
    */
   @Override
   public void deregisterWidget(int wid)
   {
      super.deregisterWidget(wid);
      websock.deregisterWidget(ews.getWindowId(), wid);
   }

   /**
    * Set the window details for resize; this includes the minimum and maximun dimensions, and 
    * also the window's resizeable state.
    * 
    * @param    resize
    *           Flag indicating if the window can be resized or not.
    * @param    minWidth
    *           The window's minimum width.
    * @param    minHeight
    *           The window's minimum height.
    * @param    maxWidth
    *           The window's maximum width.
    * @param    maxHeight
    *           The window's maximum height.
    */
   @Override
   public void setResizeableWindow(boolean resize,
                                   int     minWidth, 
                                   int     minHeight, 
                                   int     maxWidth, 
                                   int     maxHeight)
   {
      // only windows are resized
      Window<?> window = WindowManager.findWindow(ews.getWindowId());
      
      if (window == null)
      {
         return;
      }
      
      WindowConfig wcfg = window.config();
      // get actual size
      NativeDimension size = window.physicalDimension();
      
      // if the actual size is greater than the configuration size, then let the maximal size be
      // equal to its actual size.
      int maxWidthPixels  = Math.max(ConfigHelper.getMaxWidthPixels(wcfg), size.width);
      
      int maxHeightPixels = Math.max(ConfigHelper.getMaxHeightPixels(wcfg), size.height);
      
      int minWidthPixels  = ConfigHelper.getMinWidthPixels(wcfg);
      
      int minHeightPixels = ConfigHelper.getMinHeightPixels(wcfg);

      websock.setResizeableWindow(ews.getWindowId(), 
                                  wcfg.resize,
                                  minWidthPixels,
                                  minHeightPixels,
                                  maxWidthPixels,
                                  maxHeightPixels);
   }

   /**
    * Opens the target page or resource in a browser's window.
    * 
    * @param    path
    *           The url path to the target resource.
    */
   @Override
   public void openURL(String path)
   {
      try
      {
         URI uri = new URI(path).normalize();
         websock.sendBinaryMessage(WebClientMessageTypes.MSG_OPEN_URL, uri.toASCIIString());
      }
      catch (URISyntaxException ex)
      {
         // TODO: handle exception
         ex.printStackTrace();
      }
   }

   /**
    * Resolved the font metrics for the given font.
    * 
    * @param    font
    *           The font identifier, on the js-side font table.
    * 
    * @return   The font metrics.
    */
   protected FontMetricsHelper resolveFontMetrics(int font)
   {
      FontMetricsHelper metrics = fontMetricsCache.get(font);
      
      if (metrics == null)
      {
         metrics = new WebFontMetricsHelper(font, websock);
         
         fontMetricsCache.put(font, metrics);
      }
      
      return metrics;
   }
   
   /**
    * Post a new repaint for the specified window to the {@link #repaints} queue.  This will
    * be processed by the {@link RepaintThread}.
    * 
    * @param    ews
    *           The window requiring repaint.
    */
   protected void postRepaint(GuiWebEmulatedWindow ews)
   {
      synchronized (repaints)
      {
         repaints.addLast(ews);
         repaints.notifyAll();
      }
   }
   
   /**
    * Add the widget (and all its possible children) in the specified set, maintaining z-order in
    * a left-to-right (topmost to bottommost) order.
    * 
    * @param    w
    *           The parent widget.
    * @param    filter
    *           The condition predicate that must be evaluated for collected widgets.
    * @param    actionsCollector
    *           Collects a widget's action if the target is selected by the filter condition.
    */
   private void collectWidgets(Widget            w,
                               Predicate<Widget> filter, 
                               Consumer<Widget>  actionsCollector)
   {
      if (w == null)
      {
         return;
      }
   
      if (w instanceof AbstractContainer)
      {
         Widget[] children = ((AbstractContainer) w).widgets();

         // the javascript expects them in z-order, but JS keeps them left-to-right not
         // right-to-left
         for (int i = children.length - 1; i >= 0; i--)
         {
            collectWidgets(children[i], filter, actionsCollector);
         }
      }

      if (filter.test(w))
      {
         actionsCollector.accept(w);
      }
   }

   /**
    * Build the name of the main page as a resource locator string.
    * 
    * @param   pageContext
    *          Page context root.
    * @param   pageFile
    *          Main page HTML file.
    * 
    * @return  Resource name for the main page.
    */
   private String getWelcomePageUrl(String pageContext, String pageFile)
   {
      StringBuilder sb = new StringBuilder(pageContext);
      String packageName = WebPageHandler.class.getPackage().getName();
      return sb.append(packageName.replace('.', '/')).append(pageFile).toString();
   }
   
   /**
    * Checks if the given widget is belonging to visible frame or not inside any frame, like
    * window decorations.
    * 
    * @param   wid
    *          Widget to check.
    * 
    * @return  <code>TRUE</code> if the widget is inside visible frame or standalone,
    *          <code>FALSE</code> otherwise.
    */
   private boolean isInsideVisibleFrameOrStandalone(Widget wid)
   {
      // get the nclosing frame
      Frame fr = UiUtils.locateFrame(wid);
      
      // and return if it exists and is visible or does not exist
      return fr == null || fr.isVisible();
   }
   
   /**
    * Dedicated thread responsible of processing repaint requests.  For the Web driver, this 
    * requires calling {@link GuiWebEmulatedWindow#drawOnScreenBitmap()} for each window in the
    * {@link GuiWebDriver#repaints} queue.
    */
   private class RepaintThread
   extends Thread
   {
      /**
       * Initialize the repaint thread and start it.
       * <p>
       * The thread's name will be set to <code>WebRepaint</code> and is set as daemon.  It will
       * be alive for the entire life of the client instance.
       */
      public RepaintThread()
      {
         setName("WebRepaint");
         setDaemon(true);
      }
      
      /**
       * Listening loop for repaint requests.
       */
      @Override
      public void run()
      {
         while (true)
         {
            GuiWebEmulatedWindow ews = null;

            synchronized (repaints)
            {
               if (repaints.isEmpty())
               {
                  try
                  {
                     repaints.wait();
                  }
                  catch (InterruptedException e)
                  {
                     break;
                  }
               }
               
               if (!repaints.isEmpty())
               {
                  ews = repaints.removeFirst();
               }
               else
               {
                  continue;
               }
            }
            
            ews.drawOnScreenBitmap();
         }
      }
   }

   /**
    * Sets the current editors selection.
    * 
    * @param txt
    *        The current editors selection.
    */
   @Override
   public void setCurrentSelection(String txt)
   {
      websock.setSelection(txt);
   }

   /**
    * Load mouse pointer from the InputStream
    * 
    * @param data
    *        cursor data
    * @param ptrName
    *        pointer name
    *        
    * @return
    *        mouse pointer or <code>null</code>
    */
   @Override
   public MousePtrWrapper createMousePointer(byte[] data, String ptrName)
   {
      return data == null ? null : 
                           MousePtrWrapper.createStaticWebInstance(
                              data, 
                              Utils.mimeType(ptrName)
                           );
   }

   /**
    * Lock the mouse pointer (suppress widget-specific changes) or unlock it
    * 
    * @param widgetId
    *          the id of the active DropDown which caused the pointer lock or -1
    * @param ptr
    *          the locked mouse pointer or <code>null</code> for unlock
    *           
    */
   @Override
   public void lockMousePointer(int widgetId, MousePtrWrapper ptr)
   {
      websock.lockMousePointer(widgetId, ptr);
   }

   /**
    * Push information about the loaded mouse pointer for widget 
    * 
    * @param windowId
    *           to level window id
    * @param wid
    *           widgetId
    * @param ptr
    *           mouse pointer
    */
   public void propagate(int windowId, int wid, MousePtrWrapper ptr)
   {
      websock.setMousePtr(windowId, wid, ptr); 
   }

   /**
    * Returns the boolean value if the target font is installed or not.
    * 
    * @param    fontName
    *           The target font name.
    * 
    * @return   True if the target font is installed, otherwise false.
    */
   @Override
   public boolean isFontInstalled(String fontName)
   {
      return websock.isFontInstalled(fontName);
   }

   /**
    * Enable or Disable OS-level events for the specified windows.
    * 
    * @param    windowIds
    *           The window IDs which need to be processed.
    * @param    capture
    *           Flag indicating OS level events are captured (when <code>true</code>).
    */
   @Override
   public void enableEvents(int[] windowIds, boolean capture)
   {
      websock.enableOsEvents(windowIds, capture);
   }

   /**
    * Returns true if a window can be decorated in this graphics environment, otherwise false.
    * 
    * @return   true if a window can have its title, caption buttons and border,
    *           otherwise false.
    */
   @Override
   public boolean isWindowDecorationSupported()
   {
      // In the web client embedded mode all windows must be drawn without decorations.
      return !embedded;
   }
   
   /**
    * Returns true if this client runs in embedded mode.
    * 
    * @return   See above.
    */
   @Override
   public boolean isEmbeddedClient()
   {
      return embedded;
   }
   
   /**
    * Get the widths of the specified texts, using the given font, in batch.
    * 
    * @param    texts
    *           The texts.
    * @param    fd
    *           The font details used to draw and measure the text.
    *           
    * @return   The texts' widths.
    */
   @Override
   public int[][] getTextWidths(String[] texts, FontDetails<Integer> fd)
   {
      int[][] res = new int[texts.length][];
      
      List<Integer> indices = new ArrayList<>();
      List<String> resolve = new ArrayList<>();
      List<TextMetrics> tms = new ArrayList<>();
      
      for (int i = 0; i < texts.length; i++)
      {
         String text = texts[i];
         
         if (isEmptyText(text))
         {
            res[i] = new int[0];
            continue;
         }
         
         TextMetrics metrics = getTextMetrics(text, fd);
         int[] widths = (metrics == null ? null : metrics.getWidths());
         if (widths != null)
         {
            res[i] = widths;
            continue;
         }
         
         // batch it...
         resolve.add(text);
         indices.add(i);
         tms.add(metrics);
      }
      
      if (resolve.isEmpty())
      {
         return res;
      }
      
      int[][] widths = websock.getTextWidths(resolve.toArray(new String[0]), fd.font.intValue());
      for (int i = 0; i < widths.length; i++)
      {
         TextMetrics tm = tms.get(i);
         
         if (tm != null)
         {
            tm.setWidths(widths[i]);
         }
         else
         {
            String text = resolve.get(i);
            
            FontMetricsHelper fm = ews.getFontMetrics(fd);
            int height = ews.getTextHeight(text, fd);
            int width = widths[i][text.length() - 1];
            
            TextMetrics metrics = new TextMetrics(width, height);
            metrics.setWidths(widths[i]);
            
            setTextMetrics(text, fd, metrics);
         }
         
         res[indices.get(i)] = widths[i];
      }
      
      return res;
   }
}