/*
** Module   : GuiWebEmulatedWindow.java
** Abstract : core GUI renderer for web client
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20150504 First version which is just stubbed out.
** 002 HC  20150612 Compatibility changes related to support of paragraph layout.
** 003 HC  20150702 Improved positioning of top-level windows on systems with multiple displays.
** 004 GES 20150701 Added GUI implementation for previously stubbed-out methods,
**                  including image drawing, font and text metrics, text drawing and 
**                  font creation/selection.
** 005 GES 20150817 Removed isOnTop() since it is unnecessary. The saveFocusListeners(),
**                  restoreFocusListeners() and moveToTop() are only needed for Swing usage and
**                  have been removed from this class, added line strokes.
**                  Improvements for mouse support: added dedicated mouse handler. Implemented
**                  SET_ICON drawing primitive.
** 006 CA  20150929 Fixed multi-window drawing issues caused by concurrent drawing access to the
**                  GuiDriver: all access to the driver is done exclusively, each thread taking
**                  and releasing ownership for the duration of the drawing operation or other
**                  GUI API invocation which requires access to the underlying physical window.
** 007 SVL 20151013 Handled SET_XOR_COMPOSITE and RESET_COMPOSITE.
** 008 SBI 20151011 Fixed images caching.
** 009 CA  20151024 Added support for WINDOW:SENSITIVE attribute.
**     SBI 20151026 Fixed mouse processing for modal windows.
**     EVL 20151106 Adding method to obtain the BufferedImageDrawHelper instance.
** 010 EVL 20160113 Merge changes from SBI to fix initial window location.
** 011 IAS 20160130 LOAD-MOUSE-POINTER and SET-WAIT-STATE support
** 012 CA  20160323 In getTextHeight(), return the font's height, as p2j.fonts.js computes the 
**                  text's height in the same way.
**     SBI 20160324 Implemented SET_DESKTOP_BGCOLOR.
** 013 EVL 20161021 Fix to prevent requesting rectangle with negative coordinates from RawEncoder. 
** 014 HC  20170223 Improved clipping to allow to clip with arbitrary set of overlapping or non-
**                  overlapping regions.
** 015 EVL 20170706 Adding XOR mode handling for web client, DRAW_SELECTION web specific handling.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver.web;

//TODO: this dependency should be eliminated
import java.awt.image.BufferedImage;

import java.util.*;


import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;

/**
 * The core web GUI window emulator.
 */
public class GuiWebEmulatedWindow 
extends EmulatedWindowState<Integer, BufferedImage>
{
   /** Lock object for shared syncrhonization of the web socket. */
   private final Object lock;
   
   /** Web-Socket which provides a GUI-specific protocol for communication with the browser. */
   private GuiWebSocket websock;
   
   /** Web driver instance. */ 
   private GuiWebDriver webdriver;
   
   /** Locally cached visibility state flag. */
   private boolean visible = false;
   
   /** Locally cached sensitivity state flag. */
   private boolean enabled = true;
   
   /** Locally cached X coordinate (pixels). */
   private int x = 0;
   
   /** Locally cached Y coordinate (pixels). */
   private int y = 0;
   
   /** Locally cached height (pixels). */
   private int height = -1;
   
   /** Locally cached width (pixels). */
   private int width = -1;
   
   /** Locally cached title text. */
   private String title = null;
   
   /** Helper to generate image wrappers. */
   private final BufferedImageDrawHelper drawHelper;

   /** Flag indicating that the initial location should be initialized. */ 
   private boolean initLocation = true;

   /**
    * Constructor.
    *
    * @param    lock
    *           Object to use for locking. By passing this in, the caller can share the
    *           locking instance with this code.  Use <code>null</code> if an internal
    *           lock object should be created.
    * @param    websock
    *           Web-Socket which provides a GUI-specific protocol for communication with the
    *           browser.
    * @param    webdriver
    *           Web driver instance.
    */
   public GuiWebEmulatedWindow(Object lock, GuiWebSocket websock, GuiWebDriver webdriver)
   {
      super(new WebMouseHandler());
      
      this.mouseHandler.setWindow(this);
      this.lock      = (lock != null) ? lock : new Object();
      this.websock   = websock;
      this.webdriver = webdriver;
      this.drawHelper = new BufferedImageDrawHelper();
   }

   /**
    * Cause the outer frame to be disposed, as the window is being terminated.
    */
   @Override
   public void quit()
   {
      Set<Integer> imagesToRemove = webdriver.removeUnusedImages(windowId);
      
      websock.destroyWindow(windowId, imagesToRemove.toArray(new Integer[imagesToRemove.size()]));
   }

   /**
    * Cause repaint processing to occur.
    */
   @Override
   public void repaint()
   {
      webdriver.postRepaint(this);
   }

   /**
    * Update screen by iterating over the pool and render each pending operation. Any graphics
    * context or other resources needed must have already been setup/stored by the caller.
    * <p>
    * The API is made public to be used by the {@link GuiWebDriver}'s repaint thread.
    */
   @Override
   public void drawOnScreenBitmap()
   {
      super.drawOnScreenBitmap();
   }
   
   /**
    * Hide or show the window, depending on the {@code visible} flag.
    * 
    * @param    visible
    *           The window visibility state.
    */
   @Override
   public void setWindowVisible(boolean visible)
   {
      // we only send this to the client when there is a change in visibility
      if (this.visible != visible)
      {
         websock.setWindowVisible(windowId, visible);
         
         this.visible = visible;
      }
   }
   
   /**
    * Set the window's sensitive state, depending on the {@code enable} flag.
    * 
    * @param    enable
    *           The window sensitive state.
    */
   @Override
   public void setWindowEnabled(boolean enable)
   {
      // we only send this to the client when there is a change in visibility
      if (this.enabled != enable)
      {
         websock.setWindowEnabled(windowId, enable);
         
         this.enabled = enable;
      }
   }
   
   /**
    * Check if the simulated window is visible.
    * 
    * @return   See above.
    */
   @Override
   public boolean isWindowVisible()
   {
      return visible;
   }

   /**
    * Get the window location on screen.
    * 
    * @return   This window's location.
    */
   @Override
   public NativePoint getWindowLocation()
   {
      // until we implement multiple screen support for the web client, the stored coordinates
      // can be returned directly (with no translation)
      return new NativePoint(x, y);
   }

   /**
    * Get the window height.
    * 
    * @return   This window's width in pixels.
    */
   @Override
   public int getHeight()
   {
      return height;
   }
   
   /**
    * Get the window width.
    * 
    * @return   This window's width in pixels.
    */
   @Override
   public int getWidth()
   {
      return width;
   }
   
   /**
    * Manage any window state related to iconification.
    * 
    * @param    iconify
    *           <code>true</code> if this window is being iconified, <code>false</code> for
    *           de-iconified.
    */
   @Override
   public void setIconificationState(boolean iconify)
   {
      super.setIconificationState(iconify);
      
      websock.setIconificationState(this.windowId, iconify);
   }

   /**
    * Enable or Disable OS-level events for the specified windows.
    * 
    * @param    capture
    *           Flag indicating OS level events are captured (when <code>true</code>).
    */
   @Override
   public void enableEvents(boolean capture)
   {
      // this is a no-op, it is done in batch
   }

   /**
    * Get the metrics of the given font.
    *  
    * @param    f
    *           The font for which the font metrics are to be obtained.
    *           
    * @return   The font metrics.
    */
   @Override
   public FontMetricsHelper getFontMetrics(FontDetails<Integer> f)
   {
      synchronized (lock())
      {
         return webdriver.resolveFontMetrics(f.font);
      }
   }

   /**
    * Get the height of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    f
    *           The font used to draw and measure the text.
    *           
    * @return   The text's height.
    */
   @Override
   public int getTextHeight(String text, FontDetails<Integer> f)
   {
      // in p2j.fonts.js, the text height is already computed as "heighest possible char" (same
      // as font's height). so return directly that
      FontMetricsHelper fm = getFontMetrics(f);
      return fm.getHeight();
   }
   
   /**
    * The method performs a layout operation on the supplied text and 
    * returns the resulting paragraph height while maintaining the
    * supplied maximum width.
    * 
    * @param    text
    *           The text to measure.
    * @param    font
    *           The font to use during layout.
    * @param    maxWidth
    *           The maximum paragraph width.
    */
   @Override
   public int getParagraphHeight(String text, FontDetails<Integer> font, int maxWidth)
   {
      synchronized (lock())
      {
         return websock.getParagraphHeight(text, font.font, maxWidth);
      }
   }
   
   /**
    * Draw using a paint operation.
    * 
    * @param    ps
    *           Paint operation parameters.
    */
   @Override
   public void draw(PaintStructure<Integer, BufferedImage> ps)
   {
      switch (ps.id)
      {
         case DRAW_STRING:
            websock.drawString(ps.text, ps.centered, ps.x, ps.y);
            break;
         case DRAW_STRING_SCALED:
            websock.drawStringScaled(ps.text, ps.centered, ps.x, ps.y, ps.width, ps.height);
            break;
         case DRAW_PARAGRAPH:
            websock.drawParagraph(ps.text, ps.font.font, ps.x, ps.y, ps.width);
            break;
         case DRAW_LINE:
            websock.drawLine(ps.x, ps.y, ps.width, ps.height);
            break;
         case DRAW_RECT:
            websock.drawRect(ps.x, ps.y, ps.width, ps.height);
            break;
         case DRAW_ROUND_RECT:
            websock.drawRoundRect(ps.x, ps.y, ps.width, ps.height, ps.arcDiameter);
            break;
         case DRAW_IMAGE:
            {
               Object[] imageEncodedPacket = encodeImage(ps);
               int imageId = (Integer) imageEncodedPacket[0];
               ImageEncoding encoding = (ImageEncoding) imageEncodedPacket[1];
               byte[] encodedImage = (byte[]) imageEncodedPacket[2];
               Integer imageWidth  = (Integer) imageEncodedPacket[3];
               Integer imageHeight = (Integer) imageEncodedPacket[4];
               websock.drawImage(ps.x, ps.y, imageWidth, imageHeight,
                        encoding, imageId, encodedImage);
            }
            break;
         case FILL_RECT:
            websock.fillRect(ps.x, ps.y, ps.width, ps.height);
            break;
         case FILL_ROUND_RECT:
            websock.fillRoundRect(ps.x, ps.y, ps.width, ps.height, ps.arcDiameter);
            break;
         case FILL_POLYGON:
            websock.fillPolygon(ps.xPoints, ps.yPoints, ps.nPoints);
            break;
         case DRAW_3D_RECT:
            websock.draw3DRect(ps.x, ps.y, ps.width, ps.height, ps.raised);
            break;
         case FILL_3D_RECT:
            websock.fill3DRect(ps.x, ps.y, ps.width, ps.height, ps.raised);
            break;
         case DRAW_SELECTION:
            // draw selection box in clear or draw modes
            websock.drawSelection(ps.x, ps.y, ps.width, ps.height, ps.clear);
            break;
         case RESET_SELECTION:
            // clean up web client after selectin completed
            websock.resetSelection();
            break;
         case SET_COLOR:
            websock.setColor(ps.color.getRed(), ps.color.getGreen(), ps.color.getBlue());
            break;
         case SET_XOR_COMPOSITE:
            websock.setXORComposite();
            break;
         case SET_XOR_MODE:
            websock.setXORMode(ps.color.getRed(), ps.color.getGreen(), ps.color.getBlue());
            break;
         case RESET_COMPOSITE:
            websock.resetComposite();
            break;
         case SET_LINE_STROKE:
            websock.setLineStroke(ps.stroke.ordinal(), Math.round(ps.strokeWidth));
            break;
         case SET_FONT:
            websock.setFont(ps.font.font);
            break;
         case SET_FONT_STYLE:
            websock.setFontStyle(ps.style.toInt());
            break;
         case TRANSLATE_PUSH:
            websock.translate(true, ps.x, ps.y);
            break;
         case TRANSLATE_POP:
            websock.translate(false, ps.x, ps.y);
            break;
         case CLEAR_AREA:
         case CLEAR:
            if (ps.color != null)
            {
               websock.setColor(ps.color.getRed(), ps.color.getGreen(), ps.color.getBlue());
            }
            websock.fillRect(ps.x, ps.y, ps.width, ps.height);
            break;
         case CLIP:
            websock.clip(ps.clip);
            break;
         case NO_CLIP:
            websock.unclip();
            break;
         case SET_LOCATION:
            if (x != ps.x || y != ps.y || initLocation)
            {
               initLocation  = false;
               x = ps.x;
               y = ps.y;
               websock.setWindowLocation(x, y);
            }
            break;
         case SET_WINDOW_BOUNDS:
            if (x != ps.x || y != ps.y || width != ps.width || height != ps.height || initLocation)
            {
               initLocation  = false;
               x = ps.x;
               y = ps.y;
               width  = ps.width;
               height = ps.height;
               websock.setWindowBounds(x, y, width, height);
            }
            break;
         case RESIZE_WINDOW:
            if (width != ps.width || height != ps.height)
            {
               width  = ps.width;
               height = ps.height;
               websock.resizeWindow(width, height);
            }
            break;
         case DRAW_HIGHLIGHT:
            // TODO
            break;
         case CLEAR_HIGHLIGHT:
            // TODO
            break;
         case SET_TITLE:
            if (title != ps.title)
            {
               title = ps.title;
               websock.setWindowTitle(title);
            }
            break;
         case SET_ICON:
            {
               Object[] imageEncodedPacket = encodeImage(ps);
               int imageId = (Integer) imageEncodedPacket[0];
               ImageEncoding encoding = (ImageEncoding) imageEncodedPacket[1];
               byte[] encodedImage = (byte[]) imageEncodedPacket[2];
               Integer imageWidth  = (Integer) imageEncodedPacket[3];
               Integer imageHeight = (Integer) imageEncodedPacket[4];
               websock.setIconImage(imageWidth, imageHeight, encoding,
                        imageId, encodedImage);
            }
            break;
         case START_RECORDING:
            websock.startRecording(ps.x, ps.y, ps.width, ps.height);
            break;
         case STOP_RECORDING:
            websock.stopRecording();
            break;
         case SET_DESKTOP_BGCOLOR:
            websock.setDesktopBgColor(ps.color.getRed(), ps.color.getGreen(), ps.color.getBlue());
            break;
         default:
            break;
      }      
   }
   
   /**
    * Encodes the image if it isn't already loaded and returns the 3-elements array that holds
    * the image hash in its first element, the image encoding in its second element
    * and the encoded image or null in its third element depending on its loaded state.
    * 
    * @param    ps
    *           The paint structure filled with the target image parameters. 
    * 
    * @return   The 5-elements array that holds the image unique id in its first element, the image
    *           encoding in its second element, the bytes array of the encoded image or null
    *           in its third element, the width and the height of the embedded image in its forth
    *           element and its fifth element respectively.
    */
   private final Object[] encodeImage(PaintStructure<Integer, BufferedImage> ps)
   {
      EmbeddedImageStructure<BufferedImage> embeddedImageStructure = EmbeddedImageStructure.create(ps);
      List<Integer> imageSeal = embeddedImageStructure.getObjectSeal();
      VirtualScreen virtualScreen = webdriver.getVirtualScreen();
      ImageEncoding encoding;
      byte[] encodedImage;
      Integer imageWidth;
      Integer imageHeight;
      Integer imageId;
      if (webdriver.addImageUsage(imageSeal, windowId))
      {
         int x = ps.x;
         int y = ps.y;
         ImageWrapper<BufferedImage> embeddedImage = drawHelper.processImage(embeddedImageStructure);
         imageId = embeddedImage.getUniqueId();
         webdriver.mapSealToUniqueId(imageSeal, imageId);
         imageWidth  = embeddedImage.getWidth();
         imageHeight = embeddedImage.getHeight();
         encoding = ImageEncoding.RAW;
         virtualScreen.drawImage(embeddedImage, 0, 0);
         encodedImage = new RawEncoder().packToBinaries(virtualScreen, 0, 0,
                                                        imageWidth, imageHeight);
      }
      else
      {
         imageId = webdriver.getUniqueIdForSeal(imageSeal);
         encoding = ImageEncoding.HASH;
         encodedImage = null;
         Integer[] imageBounds = drawHelper.calculateImageBounds(embeddedImageStructure);
         imageWidth  = imageBounds[0];
         imageHeight = imageBounds[1];
      }
      
      return new Object[] {imageId, encoding, encodedImage, imageWidth, imageHeight};
   }
   
   /**
    * Render the given paint operation.  Any graphics context or other resources needed must
    * have already been setup/stored by the caller.
    */
   @Override
   public void flush()
   {
      websock.sendPendingDrawingOps(windowId);
   }
   
   /**
    * Since web interface is inherently a single-display setup the method 
    * effectively returns a rectangle with top-level corner equal to 0,0.
    * 
    * @return   The physical screen bounds.
    */
   @Override
   public NativeRectangle getDisplayBounds()
   {
      return webdriver.getDisplayBounds(); 
   }

   /**
    * Returns an application-usable display area. For example, a task bar is
    * not counted as the usable area and it is always at the bottom of the
    * virtual screen (when present).
    * 
    * @return   The area in pixels.
    */
   @Override
   public NativeRectangle getDisplayWorkArea()
   {
      return webdriver.getDisplayWorkArea();
   }
   
   /**
    * Gets the instance of the helper to work with buffered images.
    * 
    * @return   See above.
    */
   @Override
   public BufferedImageDrawHelper getBufferedImageDrawHelper()
   {
      return drawHelper;
   }
   
   /**
    * Obtain the lock object that can be used to synchronize drawing operations.
    *
    * @return   The shared lock object.
    */
   @Override
   protected Object lock()
   {
      return lock;
   }
}