/*
** Module   : VirtualScreenImpl.java
** Abstract : Implements virtual screen functionalities.
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 SBI 20150810 First version which implements a virtual screen based on sRGB model.
** 002 SBI 20150822 Expanded image support and fixed issues.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.driver.web;

// AWT is used here intentionally, to allow rendering of images on the Java side; only the "raw"
// pre-rendered set of pixels is sent to the JS side
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.color.ColorSpace;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferInt;
import java.awt.image.DirectColorModel;
import java.awt.image.Raster;
import java.awt.image.SampleModel;
import java.awt.image.SinglePixelPackedSampleModel;
import java.awt.image.WritableRaster;

import com.goldencode.p2j.ui.client.gui.driver.ImageWrapper;
import com.goldencode.p2j.ui.client.gui.driver.VirtualScreen;

/**
 * VirtualScreenImpl is an implementation of VirtualScreen drawings operations.
 */
class VirtualScreenImpl
implements VirtualScreen
{
   /** The color model that defines pixels representation */
   private final DirectColorModel colorModel;
   
   /** The screen width. */
   private final int width;
   
   /** The screen height. */
   private final int height;
   
   /** The pixels array.*/
   private final int[] pixels;

   /** The pixels blank array.*/
   private final int[] blanks;
   
   /** The canvas in a memory to draw images. */
   private final BufferedImage screen;

   /**
    * Constructs a virtual screen with the default color model model as sRGB 32bits per a pixel.
    * 
    * @param    width
    *           The screen width.
    * @param    height
    *           The screen height.
    */
   public VirtualScreenImpl(int width, int height)
   {
      this(new DirectColorModel(ColorSpace.getInstance(ColorSpace.CS_sRGB), 32, 0xFF, 0xFF00,
               0xFF0000, 0xFF000000, false, DataBuffer.TYPE_INT), width, height);
   }

   /**
    * Constructs a virtual screen of the given width and height based on the given pixels model.
    * 
    * @param    colorModel
    *           The pixels model based on sRGB.
    * @param    width
    *           The screen width.
    * @param    height
    *           The screen height.
    */
   public VirtualScreenImpl(DirectColorModel colorModel, int width, int height)
   {
      this.colorModel = colorModel;
      this.width = width;
      this.height = height;
      this.pixels = new int[width * height];
      this.blanks = new int[width * height];
      DataBuffer data = new DataBufferInt(pixels, pixels.length);
      SampleModel sampleModel = new SinglePixelPackedSampleModel(DataBuffer.TYPE_INT,
               width, height, colorModel.getMasks());
      WritableRaster raster = Raster.createWritableRaster(sampleModel, data, null);
      this.screen = new BufferedImage(colorModel, raster, true, null);
   }

   /**
    * Returns RGBA model via pixels, each pixel is encoded as an integer.
    * 
    * @return   The array of pixels. 
    */
   @Override
   public int[] getPixels()
   {
      return pixels;
   }

   /**
    * Gets a screen width.
    * 
    * @return   The screen width.
    */
   @Override
   public int getScreenWidth()
   {
      return width;
   }

   /**
    * Gets a screen height.
    * 
    * @return   The screen height
    */
   @Override
   public int getScreenHeight()
   {
      return height;
   }

   /**
    * Describes a pixel format used for drawing.
    * 
    * @return The color model.
    */
   @Override
   public DirectColorModel getColorModel()
   {
      return colorModel;
   }

   /**
    * Draw an image at the target position on the screen.
    * 
    * @param    image
    *           The target image.
    * @param    x
    *           The x-coordinate of the target image position. 
    * @param    y
    *           The y-coordinate of the target image position.
    */
   @Override
   public void drawImage(ImageWrapper image, int x, int y)
   {
      // do blank screen
      System.arraycopy(blanks, 0, pixels, 0, blanks.length);
      Graphics2D g2 = screen.createGraphics();
      g2.setColor(new Color(0, 0, 0, 0));
      g2.fillRect(0, 0, getScreenWidth(), getScreenHeight());
      g2.drawImage((Image) image.getImage(), x, y, image.getWidth(), image.getHeight(), null);
      g2.dispose();
   }
}