/*
** Module   : p2j.embedded.js
** Abstract : The embedded application controller that is responsible to extend embedded dialogs
**            modality to the external web application.
**
** Copyright (c) 2016-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -------------------------------Description-----------------------------------
** 001 SBI 20160330 Added the embedded application module to be loaded by the DOJO loader.
** 002 CA  20160425 Added the remote APIs which can be used by customer-specific applications, 
**                  to call into the converted code.
** 003 CA  20160429 Added a missing 'return' when treating the SOCKET_DISCONNECTED request in 
**                  receiveRemoteMessage
** 004 CA  20170105 Fixed overlay panel state with nested modal windows.
** 005 CA  20170228 Added PUBLISH/SUBSCRIBE/UNSUBSCRIBE extensions for global support and for
**                  external applications.
** 006 CA  20170825 Enhanced to allow an automatic way to drive the embedded windows via the 
**                  fwd-embedded-driver.p program.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

define(function ()
{
   "use strict";
   
   var me = {};
      
   me.REQUEST         = "request";
   me.RESPONSE        = "response";
   me.REMOTE_CALL     = "remote-call";
   me.INVOKE          = "invoke";
   me.DELETE_HANDLE   = "delete-handle";
   me.SUBSCRIBE       = "subscribe";
   me.UNSUBSCRIBE     = "unsubscribe";
   me.PUBLISH         = "publish";
   me.CREATE_RESOURCE = "create-resource";

   me.EXTERNAL_PROGRAM = 0;
   me.PROCEDURE        = 1;
   me.FUNCTION         = 2;

   me.PAGE_LOADED         = -1;
   me.PAGE_CLOSED         = -2;
   me.SOCKET_DISCONNECTED = -3;
   me.SOCKET_RECONNECTED  = -4;

   /** Flag indicating if the remote side can start receiving messages. */
   me.handleMessages = false;

   /** The data for embedded windows.*/
   var admWindows = {};
   
   /** The current ADM window index. */
   var currentADMWindow;
   
   /** The current page in the current embedded window. */
   var currentADMWindowPage;
   
   /** The handle ID for the FWD embedded driver. */
   var fwdEmbeddedDriverHandle;
   
   /** A callback to initialize/populate the embedded window list in the UI. */
   var populateWindowsFunc;
   
   /** A callback to initialize/populate the pages for a embedded window. */
   var populatePagesFunc;
   
   /**
    * A registry of callbacks, to be executed when a response is received for a message.  If a
    * callback is not set for a message ID, then the global callback is executed.
    */
   var callbacks = {};

   /** Flag indicating if the API calls have the NO-ERROR mode enabled. */
   var noError = false;

   /** The generator for the message IDs, on the remote side (always positive). */ 
   var nextMessageId = 0;
   
   /**
    * A register with all subscriptions.  This acts like a map with the following structure:
    * - each key is a resource ID
    * - the value is an associative array with "event" key representing the event name and 
    *   the "method" key representing the function callback to be executed for the event.
    */
   var subscriptions = {};
   
   /**
    * A register with the pending subscriptions.  Acts the same as the <code>subscriptions</code>
    * object, except the keys in this object represent message IDs.
    * <p>
    * Once a response is received for the SUBSCRIBE request, the subscription will be registered
    * in the <code>subscriptions</code> register.
    */
   var pendingSubscriptions = {};
   
   /**
    * A register with the pending unsubscriptions.  Acts like a map with the following structure:
    * - each key is the message ID
    * - each value is an associative array with "subscriber-id" the ID of the subscriber which
    *   initiated the UNSUBSCRIBE and the "event" key the event name.
    * <p>
    * Once a response is received for the UNSUBSCRIBE request, the subscription will be removed
    * from the <code>subscriptions</code> register.
    */
   var pendingUnsubscriptions = {};

   /**
    * It is responsible to create the overlay panel and to put it over all page elements except
    * the embedded frame that must be on the top. The method must be invoked on the page load
    * event to be sure that all page elements have been loaded.
    * 
    * @param    {String} frameId
    *           The embedded frame unique identifier.
    * @param    {String} overlayId
    *           The overlay panel unique identifier.
    * @param    cls
    *           The overlay panel css name.
    */
   me.init = function(frameId, overlayId, cls)
   {
      var frame  = document.getElementById(frameId);
      var origin = getUrlOrigin(frame.src);
      // calculates the maximal zIndex over the page elements
      var zIndex  = getMaxZIndex();
      // supposed that IFRAME is wrapped by DIV that is used to place
      // the embedded content within the page
      var frameParent = frame.parentElement;
      // create the overlay panel that must cover the displayed page
      // let the overlay panel be a sibling of the frame parent - frameParent
      var overlayPanel = createOverlayPanel(overlayId, zIndex, frameParent.parentElement, cls);
      // let the frame be on the top of the overlay panel
      while (frameParent != overlayPanel.parentElement)
      {
         frameParent.style.zIndex = zIndex + 1;
         frameParent = frameParent.parentElement;
         zIndex = zIndex + 1;
      }
      frame.style.zIndex = zIndex;
      var overlayPanelStates = [];
      overlayPanelStates.push("none");
      
      window.addEventListener(
                              "message",
                              function (event)
                              {
                                 if (event.origin != origin)
                                 {
                                    console.error("event.origin " + event.origin +
                                          " but expected " + origin);
                                    return;
                                 }
                                 // expected to get { target : "OverlayPanel", visible: [true|false]}.
                                 var msg = event.data;
                                 if (msg.target === "OverlayPanel" && msg.visible === true)
                                 {
                                    overlayPanelStates.push(overlayPanel.style.display);
                                    overlayPanel.style.display = "block";
                                    frame.focus();
                                 }
                                 else if (msg.target === "OverlayPanel" && msg.visible === false)
                                 {
                                    var state = overlayPanelStates.pop();
                                    overlayPanel.style.display = state;
                                 }
                                 else if (msg.target === "Frame" && msg.bgcolor)
                                 {
                                    frame.style.background = msg.bgcolor;
                                 }
                              },
                              false);

      this.targetWindow = window.frames[frameId];
      this.targetOrigin = "*";

      if (window.addEventListener)
      {
         window.addEventListener("message", receiveRemoteMessage, false);
      }
      else
      {
         window.attachEvent("onmessage", receiveRemoteMessage);
      }
   };

   /**
    * Check if the NO-ERROR mode is enabled for messages.
    */
   me.isSilentError = function()
   {
      return noError;
   };
   
   /**
    * Enable NO-ERROR (silent error) mode for all future requests.  This will add a 
    * <code>msg.noError = true</code> flag into the message.
    */
   me.silentErrorEnable = function()
   {
      noError = true;
   };
   
   /**
    * Disable NO-ERROR (silent error) mode for all future requests.  This will add a 
    * <code>msg.noError = true</code> flag into the message.
    */
   me.silentErrorDisable = function()
   {
      noError = false;
   };
   
   /**
     * Send a message to the target window.
     */
   me.send = function(msg)
   {
      this.targetWindow.postMessage(msg, this.targetOrigin);
   };

   /**
    * Get the the callback handler for the specified message ID.
    *
    * @param   {number} msgId
    *          The message ID.
    *
    * @return  The message-specific callback or the global callback (if defined).
    */
   me.getHandler = function(msgId)
   {
      var msgCallback = msgId ? callbacks[msgId] : undefined;
      if (msgCallback != undefined)
      {
         delete callbacks[msgId];
      }
      else if (p2j.embedded.messageCallback)
      {
         msgCallback = p2j.embedded.messageCallback;
      }
      
      return msgCallback
   };

   /**
    * Set a global callback to be executed when a message is received.
     *
     * @param    {function} callback
     *           The global callback.
    */
   me.setMessageCallback = function(callback)
   {
      this.messageCallback = callback;
   };
   
   /**
    * Set a callback to be executed when a PAGE_LOADED message is received.
    *
    * @param    {function} callback
    *           The page-loaded callback.
    */
   me.setPageLoadedCallback = function(callback)
   {
      this.pageLoadedCallback = callback;
   };
   
   /**
    * Set a callback to be executed when a PAGE_CLOSED message is received.
    *
    * @param    {function} callback
    *           The page-closed callback.
    */
   me.setPageClosedCallback = function(callback)
   {
      this.pageClosedCallback = callback;
   };
   
   /**
    * Set a callback to be executed when a SOCKET_DISCONNECTED message is received.
    *
    * @param    {function} callback
    *           The socket-disconnected callback.
    */
   me.setSocketDisconnectedCallback = function(callback)
   {
      this.socketDisconnectedCallback = callback;
   };
      
   /**
    * Set a callback to be executed when a SOCKET_RECONNECTED message is received.
    *
    * @param    {function} callback
    *           The socket-reconnected callback.
    */
   me.setSocketReconnectedCallback = function(callback)
   {
      this.socketReconnectedCallback = callback;
   };
   
   /**
    * This API creates a DELETE_HANDLE message request and sends it to the target window.
    *
    * @param   {function} callback
    *          The callback to be executed for the response.  If <code>null</code>, the global
    *          callback will be executed.
    * @param   {string} handleId
    *          A string with a hex representation of the handleId. 
    */
   me.deleteHandle = function(callback, handleId)
   {
      var msgId = getNextMessageId();
      
      if (callback)
      {
         callbacks[msgId] = callback;
      }
      
      var msg = {};
      msg["type"]    = this.DELETE_HANDLE;
      msg["mode"]    = this.REQUEST;
      msg["id"]      = msgId;
      msg["payload"] = handleId;
      msg["noError"] = noError;
      
      this.send(msg);
   };

   /**
    * Execute a SUBSCRIBE request.  If the <code>subscriber</code> is undefined, then it will be 
    * built and its id can be saved via the <code>callback</code> function.
    *
    * @param   {function} callback
    *          A callback with the response for the SUBSCRIBE request. 
    * @param   {boolean} global
    *          Flag indicating if this SUBSCRIBE is for notifications from all contexts.
    * @param   {string} subscriber
    *          The hex representation of this subscriber resource ID, or undefined if it is not 
    *          yet built.
    * @param   {string} event
    *          The event for which to register.
    * @param   {string} publisher
    *          The hex representation of an external resource or program ID.  If undefined or
    *          empty string, acts as ANYWHERE legacy option.
    * @param   {function} method
    *          The callback which will be executed when the event is raised via PUBLISH.
    */
   me.subscribe = function(callback, global, subscriber, event, publisher, method)
   {
      var msgId = getNextMessageId();
      
      if (subscriber == undefined)
      {
         subscriber = "";
      }
      
      if (event == undefined)
      {
         event = "";
      }
      
      if (publisher == undefined)
      {
         publisher = "";
      }
      
      // 'callback' is used to save the subscriber ID (if is unknown here) 
      if (callback)
      {
         callbacks[msgId] = callback;
      }
      
      pendingSubscriptions[msgId] = 
      {
         "event"  : event,
         "method" : method
      };
      
      var payload = 
      {
         "global"     : global,
         "subscriber" : subscriber,
         "event"      : event,
         "publisher"  : publisher
      };

      var msg = {};
      msg["type"]    = this.SUBSCRIBE;
      msg["mode"]    = this.REQUEST;
      msg["id"]      = msgId;
      msg["noError"] = noError;
      msg["payload"] = payload;

      this.send(msg);
   };

   /**
    * Execute a UNSUBSCRIBE request.
    *
    * @param   {function} callback
    *          A callback with the response for the UNSUBSCRIBE request. 
    * @param   {boolean} global
    *          Flag indicating if this UNSUBSCRIBE is for global registrations or for local ones.
    * @param   {string} subscriber
    *          The hex representation of this subscriber resource ID.
    * @param   {string} event
    *          The event the subscription was registered.  undefined or empty string will 
    *          act as null on the remote side.
    * @param   {string} publisher
    *          The hex representation of an external resource or program ID.  If undefined or
    *          empty string, acts as ANYWHERE legacy option.
    */
   me.unsubscribe = function(callback, global, subscriber, event, publisher)
   {
      var msgId = getNextMessageId();
      
      if (callback)
      {
         callbacks[msgId] = callback;
      }
            
      pendingUnsubscriptions[msgId] = 
      {
         "event"         : event,
         "subscriber-id" : subscriber
      };
      
      var payload = 
      {
         "global"     : global,
         "subscriber" : subscriber,
         "event"      : event,
         "publisher"  : publisher
      };
      
      var msg = {};
      msg["type"]    = this.UNSUBSCRIBE;
      msg["mode"]    = this.REQUEST;
      msg["id"]      = msgId;
      msg["noError"] = noError;
      msg["payload"] = payload;

      this.send(msg);
   };

   /**
    * Execute a PUBLISH request.
    *
    * @param   {string} event
    *          The event name to PUBLISH.
    *          act as null on the remote side.
    * @param   {string} publisher
    *          The hex representation of an external resource or program ID.
    * @param   {array} args
    *          An array with the arguments for this event.
    */
   me.publish = function(event, publisher, args)
   {
      // no callbacks for this api, as no response is sent back
      var msgId = getNextMessageId();
      
      if (args == undefined)
      {
         args = [];
      }

      var payload = 
      {
         "event"      : event,
         "publisher"  : publisher,
         "args"       : args
      };
      
      var msg = {};
      msg["type"]    = this.PUBLISH;
      msg["mode"]    = this.REQUEST;
      msg["id"]      = msgId;
      msg["payload"] = payload;

      this.send(msg);
   };

   /**
    * Execute a CREATE-RESOURCE request.
    *
    * @param   {function} callback
    *          A callback with the response for the CREATE-RESOURCE request. 
    */
   me.createResource = function(callback)
   {
      // no callbacks for this api, as no response is sent back
      var msgId = getNextMessageId();
      
      // 'callback' is used to save the resource ID 
      if (callback)
      {
         callbacks[msgId] = callback;
      }

      var msg = {};
      msg["type"]    = this.CREATE_RESOURCE;
      msg["mode"]    = this.REQUEST;
      msg["id"]      = msgId;

      this.send(msg);
   };

   /**
    * This API creates a INVOKE message request for an external program, and sends it to the
    * target window.
    *
    * @param   {function} callback
    *          The callback to be executed for the response.  If <code>null</code>, the global
    *          callback will be executed.
    * @param   {string} name
    *          An external program name. 
    * @param   {object} args
    *          An associative array with the arguments.  The arg's value must be a string. 
    */
   me.invokeExternalProgram = function(callback, name, args)
   {
      invoke(callback, name, this.EXTERNAL_PROGRAM, args, null, false);
   };

   /**
    * This API creates a INVOKE message request for an external program in persistent mode, and 
    * sends it to the target window.
    *
    * @param   {function} callback
    *          The callback to be executed for the response.  If <code>null</code>, the global
    *          callback will be executed.
    * @param   {string} name
    *          An external program name. 
    * @param   {object} args
    *          An associative array with the arguments.  The arg's value must be a string. 
    */
   me.invokeExternalProgramPersistent = function(callback, name, args)
   {
      invoke(callback, name, this.EXTERNAL_PROGRAM, args, null, true);
   };

   /**
    * This API creates a INVOKE message request for a function, and sends it to the target window.
    *
    * @param   {function} callback
    *          The callback to be executed for the response.  If <code>null</code>, the global
    *          callback will be executed.
    * @param   {string} name
    *          A function name. 
    * @param   {object} args
    *          An associative array with the arguments.  The arg's value must be a string. 
    * @param   {string} handle
    *          A hex value with the handle ID in which to execute the function.  The client must
    *          have "remoteentrypoint" rights on the external program for this handle. 
    */
   me.invokeFunction = function(callback, name, args, handle)
   {
      invoke(callback, name, this.FUNCTION, args, handle, false);
   };

   /**
    * This API creates a INVOKE message request for an internal procedure, and sends it to the 
    * target window.
    *
    * @param   {function} callback
    *          The callback to be executed for the response.  If <code>null</code>, the global
    *          callback will be executed.
    * @param   {string} name
    *          An internal procedure name. 
    * @param   {object} args
    *          An associative array with the arguments.  The arg's value must be a string. 
    * @param   {string} handle
    *          A hex value with the handle ID in which to execute the function.  The client must
    *          have "remoteentrypoint" rights on the external program for this handle. 
    */
   me.invokeProcedure = function(callback, name, args, handle)
   {
      invoke(callback, name, this.PROCEDURE, args, handle, false);
   };

   /**
     * Send a response for a REMOTE_CALL request, to the target window.
     *
     * @param   {number} msgId
     *          The request message ID.
     * @param   {string} payload
     *          The payload to send for this response.
     */
   me.sendRemoteCallResponse = function(msgId, payload)
   {
      // this is a notification that a request 4GL-to-remote has finished, and the response needs
      // to be sent back, with this payload.
      
      var msg = {};
      msg["type"]    = this.REMOTE_CALL;
      msg["mode"]    = this.RESPONSE;
      msg["id"]      = msgId;
      msg["payload"] = payload;
      
      this.send(msg);
   };
   
   /**
    * Initialize the embedded windows by retrieveing the data from adm_windows.json.
    *
    * @param    {function} initCallback
    *           Callback executed when all initialization has been done.
    * @param    {function} populateWindows
    *           Callback to populate the ADM windows in the UI.
    * @param    {function} populatePages
    *           Callback to populate the pages for a ADM window in the UI.
    */
   me.initilizeEmbeddedWindows = function(initCallback, populateWindows, populatePages)
   {
      $.ajax({
         url      : "adm_windows.json",
         type     : "GET",
         async    : true,
         dataType : "json",
         
         success  : 
         function(data)
         {
            loadEmbeddedWindows(data, initCallback, populateWindows, populatePages);
         },
         
         error    : 
         function (xhr, textStatus, errorThrown)
         {
            console.error("ERROR (ajax get): " + xhr.status + " - " + errorThrown);
         },
      });
   };
   
   /**
    * Logout from the application.
    *
    * @param   {function} callback
    *          Callback executed when the API returns.
    */
   me.logout = function(callback)
   {
      p2j.embedded.invokeProcedure(callback, "inMsg_logout", {}, fwdEmbeddedDriverHandle);
   };

   /**
    * Show the specified window.
    *
    * @param   {int} windowNum
    *          The window index in <code>admWindows</code> array.
    * @param   {function} callback
    *          Callback executed when the API returns.
    */
   me.showWindow = function(windowNum, callback)
   {
      currentADMWindow = windowNum;
      currentADMWindowPage = 1;
      
      var programName = admWindows[currentADMWindow]["program-name"];
      var pages = admWindows[currentADMWindow]["pages"];

      populatePagesFunc(programName, pages);
   
      p2j.embedded.showWindowPage(currentADMWindowPage, callback);
   };

   /**
    * Show the specified page in the current ADM window.
    *
    * @param   {int} windowPageNum
    *          The page number.
    * @param   {function} callback
    *          Callback executed when the API returns.
    */
   me.showWindowPage = function(windowPageNum, callback)
   {
      currentADMWindowPage = windowPageNum;

      var args = 
      {
         "pfile"         : admWindows[currentADMWindow]["program-name"],
         "pnum"          : currentADMWindowPage
      };
      
      if (admWindows[currentADMWindow]["hidden-embedded-widgets"])
      {
         args["hiddenWidgets"] = admWindows[currentADMWindow]["hidden-embedded-widgets"];
      }
      else
      {
         args["hiddenWidgets"] = "";
      }

      p2j.embedded.invokeProcedure(callback, "inMsg_fwdShowWindow", args, fwdEmbeddedDriverHandle);
   };
   
   /**
    * Get the current ADM window.
    *
    * @return   The <code>currentADMWindow</code>.
    */
   function getCurrentWindow()
   {
      return currentADMWindow;
   };
   
   /**
    * Get the current page in the current ADM window.
    *
    * @return   The <code>currentADMWindowPage</code>.
    */
   function getCurrentPage()
   {
      return currentADMWindowPage;
   };
   
   /**
    * Initialize the embedded driver and load the embedded windows.  
    * API executed once the remote side sends the ADM window data.
    *
    * @param    {array} data
    *           The ADM windows data.
    * @param    {function} initCallback
    *           Callback executed when all initialization has been done.
    * @param    {function} populateWindows
    *           Callback to populate the ADM windows in the UI.
    * @param    {function} populatePages
    *           Callback to populate the pages for a ADM window in the UI.
    */
   function loadEmbeddedWindows(data, initCallback, populateWindows, populatePages)
   {
      populateWindowsFunc = populateWindows;
      populatePagesFunc = populatePages;

      admWindows = data;

      currentADMWindow = 0;
      currentADMWindowPage = 1;
      
      p2j.embedded.invokeExternalProgramPersistent(
         function(msg)
         {
            fwdEmbeddedDriverHandle = msg.payload.handle;
            
            var windowTitle = admWindows[currentADMWindow]["window-title"];
            var programName = admWindows[currentADMWindow]["program-name"];
            var pages = admWindows[currentADMWindow]["pages"];

            populateWindowsFunc(admWindows);
            populatePagesFunc(programName, pages);
   
            p2j.embedded.showWindowPage(currentADMWindowPage, initCallback);
         },
         "fwd-embedded-driver.p",
         {});
   };

   /**
     * Generate a new message ID for a request into the P2J client.
     *
     * @return   a message ID: always a positive value.
     */
   function getNextMessageId()
   {
      // the messages incoming from P2J have a negative value; the messages from other windows
      // must have a positive value
      var msgId = nextMessageId + 1;
      
      if (msgId < 0)
      {
         msgId = 1;
      }
      
      nextMessageId = msgId;
      
      return msgId;
   };

   /**
    * Creates the overlay panel that must cover the displayed page.
    * 
    * @param    {String} id
    *           The element unique identifier.
    * @param    {Number} zIndex
    *           The maximal z-index calculated for the current page.
    * @param    {HtmlElement} container
    *           The container element.
    * @param    cls
    *           The overlay panel css name.
    * 
    * @return   The overlay panel.
    */
   function createOverlayPanel(id, zIndex, container, cls)
   {
      var overlayPanel = document.getElementById(id);

      if (!overlayPanel)
      {
         overlayPanel    = document.createElement("div");
         overlayPanel.id = id;
         overlayPanel.className = cls;
         container.appendChild(overlayPanel);
      }

      overlayPanel.style.position = "fixed";
      overlayPanel.style.top      = "0px";
      overlayPanel.style.left     = "0px";
      overlayPanel.style.width    = "100%";
      overlayPanel.style.height   = "100%";

      overlayPanel.style.background = "black";
      overlayPanel.style.opacity    = 0.5;
      overlayPanel.style.display    = "none";
      overlayPanel.style.zIndex     = zIndex;

      return overlayPanel;
   };

   /**
    * Calculates the maximal z-index on the current page.
    * 
    * @return   The maximal z-index on the current page.
    */
   function getMaxZIndex()
   {
      var zIndex;
      var maxZIndex = 1;
      var all  = document.getElementsByTagName('*');
      for (var i = 0, n = all.length; i < n; i++)
      {
          zIndex = window.getComputedStyle(all[i], null).getPropertyValue("z-index");
          zIndex = parseInt(zIndex, 10);
          if (zIndex != null && isFinite(zIndex))
          {
             maxZIndex = Math.max(maxZIndex, zIndex);
          }
      }
      
      return maxZIndex;
   };
   
   /**
    * Calculates the url origin. It is supposed that url has "https" protocol part. It retrieves
    * the url domain part and forms the origin. If the url doesn't satisfy this assumption, then
    * it returns "*" as an origin indicating any origin is accepted.
    * 
    * @param    {String} url
    *           The url path.
    * 
    * @return   The url origin.
    */
   function getUrlOrigin(url)
   {
      var protocol = "https";
      if (url != null && url.indexOf(protocol) >= 0)
      {
         var components = url.substr(8).split("/");
         return protocol + "://" + components[0];
      }
      else
      {
         return "*";
      }
   };
   
   /**
    * This API creates a INVOKE message request and sends it to the target window.
    *
    * @param   {function} callback
    *          The callback to be executed for the response.  If <code>null</code>, the global
    *          callback will be executed.
    * @param   {string} name
    *          The legacy name of the target external program, function or procedure. 
    * @param   {number} type
    *          The target's type: EXTERNAL_PROGRAM - (0), PROCEDURE - (1) or FUNCTION - (2) 
    * @param   {object} args
    *          An associative array with the arguments.  The arg's value must be a string. 
    * @param   {string} handle
    *          A hex value with the handle ID in which to execute the function.  The client must
    *          have "remoteentrypoint" rights on the external program for this handle.
    *          Only used by functions or internal procedure calls.
    * @param   {boolean} persistent
    *          Only used when invoking external programs; it indicates if it is ran persistent.
    */
   function invoke(callback, name, type, args, handle, persistent)
   {
      var msgId = getNextMessageId();
      var vargs = {};
      
      if (callback)
      {
         callbacks[msgId] = callback;
      }
      
      for (var arg in args)
      {
         vargs[arg] = "" + args[arg];
      }
      
      var payload = {};
      payload["type"]       = type;
      payload["name"]       = name;
      payload["persistent"] = persistent;
      payload["handle"]     = (handle ? handle : "");
      payload["arguments"]  = vargs;
      payload["noError"]    = noError;
      
      var msg = {};
      msg["type"]    = p2j.embedded.INVOKE;
      msg["mode"]    = p2j.embedded.REQUEST;
      msg["id"]      = msgId;
      msg["payload"] = payload;
      
      p2j.embedded.send(msg);
      
      return msgId;
   };

   /**
    * Handler for message events.
    *
    * @param   {object} event
    *          The event object.
    */
   function receiveRemoteMessage(event)
   {
      /**
        * Worker to handle the specified message.
        * <p>
        * If a custom callback is set for this message ID in the <code>p2j.embedded.callbacks</code>
        * map, then that will be executed.  Otherwise, the global 
        * <code>p2j.embedded.messageCallback</code> callback will be executed.  If not set, a message
        * is logged in the JS console.
        *
        * @param   {object} msg
        *          The message to handle.
        */
      function handleMessage(msg)
      {
         var msgId = msg.id;
         var msgCallback = p2j.embedded.getHandler(msgId);

         if (msgCallback)
         {
            msgCallback(msg);
         }
         else
         {
            var smsg = JSON.stringify(msg);
            console.log("No callback is set for the message: " + smsg);
         }
      };
      
      // WARNING! here "this" will be the Window object (not "me"!)

      // For Chrome, the origin property is in the event.originalEvent object.
      var origin = event.origin || event.originalEvent.origin;

      var msg = event.data;

      // if the message is a request for PAGE_LOADED, then check that the source is the same as
      // the targetWindow and the secret
      if (msg.type == p2j.embedded.PAGE_LOADED)
      {
         if (p2j.embedded.handleMessages)
         {
            var smsg = JSON.stringify(msg);
            console.log("p2j.embedded received a PAGE_LOADED request after  it was initialized: " +
                        smsg);
            return;
         }

         if (msg.mode == p2j.embedded.REQUEST)
         {
            if (event.source == p2j.embedded.targetWindow)
            {
               p2j.embedded.handleMessages = true;

               var secret = msg.payload;
               
               // send back a notification with this page's uri
               p2j.embedded.targetOrigin = origin;
               
               var msg = {};
               msg["type"] = p2j.embedded.PAGE_LOADED;
               msg["mode"] = p2j.embedded.RESPONSE;
               msg["payload"] = secret;
               
               p2j.embedded.send(msg);
               
               if (p2j.embedded.pageLoadedCallback)
               {
                  p2j.embedded.pageLoadedCallback(origin);
               }
            }
            else
            {
               console.log("Received a PAGE_LOADED message from incorrect window: " + origin);
            }
         }
         else
         {
            var msgMode = (msg.mode ? msg.mode : "undefined");
            console.log("Received a PAGE_LOADED message with " + msgMode + 
                        " mode - remote side can handle only RESPONSE modes!");
         }
         
         return;
      }
      else if (msg.type == p2j.embedded.PAGE_CLOSED)
      {
         if (msg.mode == p2j.embedded.REQUEST)
         {
            p2j.embedded.targetOrigin = "*";
            p2j.embedded.handleMessages = false;

            // this is a request for a PAGE_CLOSED notification (no response is required)
            if (p2j.embedded.pageClosedCallback)
            {
               p2j.embedded.pageClosedCallback(origin);
            }
            else
            {
               handleMessage(msg);
            }
         }
         else
         {
            var msgMode = (msg.mode ? msg.mode : "undefined");
            console.log("Received a PAGE_CLOSED message with " + msgMode + 
                        " mode - remote side can handle only REQUEST modes!");
         }
         
         return;
      }
      else if (msg.type == p2j.embedded.SOCKET_RECONNECTED)
      {
         if (msg.mode == p2j.embedded.REQUEST)
         {
            p2j.embedded.handleMessages = true;

            // this is a request for a SOCKET_RECONNECTED notification (no response is required)
            if (p2j.embedded.socketReconnectedCallback)
            {
               p2j.embedded.socketReconnectedCallback();
            }
            else
            {
               handleMessage(msg);
            }
         }
         else
         {
            var msgMode = (msg.mode ? msg.mode : "undefined");
            console.log("Received a SOCKET_RECONNECTED message with " + msgMode + 
                        " mode - remote side can handle only REQUEST modes!");
         }

         return;
      }
      else if (msg.type == p2j.embedded.SOCKET_DISCONNECTED)
      {
         if (msg.mode == p2j.embedded.REQUEST)
         {
            p2j.embedded.handleMessages = false;

            // this is a request for a SOCKET_DISCONNECTED notification (no response is required)
            if (p2j.embedded.socketDisconnectedCallback)
            {
               p2j.embedded.socketDisconnectedCallback();
            }
            else
            {
               handleMessage(msg);
            }
         }
         else
         {
            var msgMode = (msg.mode ? msg.mode : "undefined");
            console.log("Received a SOCKET_DISCONNECTED message with " + msgMode + 
                        " mode - remote side can handle only REQUEST modes!");
         }
         
         return;
      }
      
      if (!p2j.embedded.handleMessages)
      {
         var smsg = JSON.stringify(msg);
         console.log("p2j.embedded received a message before it was initialized: " + smsg);
         return;
      }

      if (origin !== p2j.embedded.targetOrigin)
      {
         var smsg = JSON.stringify(msg);

         console.log("Received message " + smsg + " from unexpected window origin: " + origin +
                     "(expected origin is " + p2j.embedded.targetOrigin + ")");
         return;
      }
      
      var types = 
      [
         p2j.embedded.REMOTE_CALL,    // 0
         p2j.embedded.INVOKE,         // 1
         p2j.embedded.DELETE_HANDLE,  // 2
         p2j.embedded.SUBSCRIBE,      // 3
         p2j.embedded.UNSUBSCRIBE,    // 4
         p2j.embedded.PUBLISH,        // 5
         p2j.embedded.CREATE_RESOURCE // 6
      ];
      var modes =
      [
         p2j.embedded.REQUEST,  // 0
         p2j.embedded.RESPONSE, // 1
         p2j.embedded.RESPONSE, // 2
         p2j.embedded.RESPONSE, // 3
         p2j.embedded.RESPONSE, // 4
         p2j.embedded.REQUEST,  // 5
         p2j.embedded.RESPONSE  // 6
      ];
      var customCallbacks = 
      [
         undefined,           // 0
         undefined,           // 1
         undefined,           // 2
         subscribeCallback,   // 3
         unsubscribeCallback, // 4
         publishCallback,     // 5
         undefined            // 6
      ];
      
      var found = false;
      
      // handle other message types
      for (var i = 0; i < types.length; i++)
      {
         if (msg.type == types[i])
         {
            if (msg.mode == modes[i])
            {
               if (customCallbacks[i] != undefined)
               {
                  customCallbacks[i](msg);
               }

               handleMessage(msg);
            }
            else
            {
               var msgMode = (msg.mode ? msg.mode : "undefined");
               console.log("Received a " + types[i] + " message with " + msgMode + 
                           " mode - remote side can handle only " + modes[i] + " modes!");
            }
            
            found = true;
            break;
         }
      }
      
      if (!found)
      {
         // unknown payload, discard
         var data = JSON.stringify(msg);
         console.log("remote side has received an unknown message, and will be discarded: " + data);
      }
   };
   
   /**
    * Callback executed when a UNSUBSCRIBE response is received.
    * <p>
    * If "msg.payload.error == true", then it will not remove the event.  Otherwise, it will
    * be removed from the "subscriptions" register.
    * 
    * @param    {object} msg
    *           The response message.
    */
   function unsubscribeCallback(msg)
   {
      var payload = msg["payload"];
   
      var msgId = msg["id"];
      
      var ps = pendingUnsubscriptions[msgId];
      
      delete pendingUnsubscriptions[msgId];

      if (payload["error"])
      {
         // unsubscribe failed, do not remove
         return;
      }

      var event = ps["event"];
      var subId = ps["subscriber-id"];

      var subs = subscriptions[subId];
      delete subs[event];
   };
   
   /**
    * Callback executed when a SUBSCRIBE response is received.
    * <p>
    * If "msg.payload.error == true", then it will not register the event.  Otherwise, it will
    * be added to the "subscriptions" register.
    * 
    * @param    {object} msg
    *           The response message.
    */
   function subscribeCallback(msg)
   {
      var payload = msg["payload"];

      var msgId = msg["id"];
      
      var ps = pendingSubscriptions[msgId];
      
      delete pendingSubscriptions[msgId];

      if (payload["error"])
      {
         // subscription failed, do not register
         return;
      }

      var method = ps["method"];
      var event = ps["event"];

      var subId = payload["subscriber-id"];
      if (subscriptions[subId] == undefined)
      {
         subscriptions[subId] = {};
      }
      
      subscriptions[subId][event] = method;
   };
   
   /**
    * Callback executed when the PUBLISH request is received.
    * If there is a subscription for the "msg.payload.subscriber-id" and "msg.payload.event"
    * in the "subscriptions" register, then that callback will be executed.
    * 
    * @param    {object} msg
    *           The request message.
    */
    function publishCallback(msg)
    {
      var payload = msg["payload"];
      var event   = payload["event"];
      var subId   = payload["subscriber-id"];
      var params  = payload["parameters"];
      
      if (subscriptions[subId])
      {
         var subscriberCallback = subscriptions[subId][event];
         if (subscriberCallback)
         {
            subscriberCallback(params);
         }
      }
   };

   me.getNextMessageId = getNextMessageId;

   return me;
});