/*
** Module   : p2j.mouse.js
** Abstract : web-specific mouse management
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 CA  20150826 First version.
** 002 CA  20150925 Fixed mouse enter/exit support; fixed cursor restoration on window border.
** 003 CA  20151013 Added support for window resize for the GUI web client.
** 004 CA  20151024 Added support for WINDOW:SENSITIVE attribute.
**     SBI 20151026 Fixed mouse processing for modal windows.
**     SBI 20151108 Added add and remove events listeners methods.
** 005 EVL 20160113 Merged changes from SBI to keep track on window moving.  Fix for overlay
**                  early incorrect dismiss, no need to attach focus listener to click if we
**                  already have it for press.  This confuses the overlay window.
** 006 IAS 20160217 LOAD-MOUSE-POINTER ans SET-WAIT-STATE support
** 007 SBI 20151217 Changed to resize the offscreen canvas first and then to blast the offscreen
**                  image to the original canvas.
**     SBI 20160310 Added getEvents() to MouseMovable, fixed undeclared variable that was hidden
**                  by the variable "parent".
** 008 HC  20160406 Overhaul of window-activation logic.
** 009 SBI 20160505 Fixed MousePopupable to catch "click" and "contextmenu" events depending on
**                  the popup activation button.
**     SBI 20160603 Added system action listeners to handle mouse clicks on the copy, cut, paste
**                  system menu items, added Safari copy, cut and paste support.
**     SBI 20160613 Changed to handle the case if the target window is minimized from the system
**                  menu.
** 010 SBI 20160708 Changed to send a mouse exit and mouse enter to all involved widgets and
**                  renamed "editable" to "hoverable" due to the implemented functionalities.
**     SBI 20160713 Changed to set the effective parent's mouse pointer if the mouse leaves
**                  the target widget.
** 011 EVL 20170811 Small syntax error fixes in JS code.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

"use strict";

/**
 * Define complex mouse management operations.   If a mouse action is not processed by any event
 * defined here, it is deferred to the Window which may or may not decide to send it to the java
 * side.
 */
function MouseHandler(win)
{
   /** The last widget which was hovered in the window. */
   var lastMouseWidget = undefined;

   // these need to be registered first
   for (var op in p2j.screen.mouseOps)
   {
      win.canvas.addEventListener(op, checkWindowDisable());
   }
   
   // complex mouse listeners
   
   // window listeners
   win.canvas.addEventListener("mousedown", processWindowFocus("mousedown", this));
   
   // mouse move: if the mouse cursor hasn't exited the last hovered area, cancel
   win.canvas.addEventListener("mousemove",  processMouseMove(this, "mousemove"));
   win.canvas.addEventListener("mouseenter", processMouseMove(this, "mouseenter"));
   win.canvas.addEventListener("mouseleave", processMouseMove(this, "mouseleave"));
   win.canvas.addEventListener("mouseout",   processMouseMove(this, "mouseout"));
   win.canvas.addEventListener("mouseover",  processMouseMove(this, "mouseover"));

   /**
    * Check if the window is disabled.  In this case, consume the event.
    */
   function checkWindowDisable()
   {
      return function(evt)
      {
         if (win.canvas.disabled)
         {
            // consume the event, the window is disabled
            evt.preventDefault();
            return;
         }
      };
   };

   /**
    * Determine the mouse location relative to the window's canvas.
    *
    * @param    {Event} evt
    *           The raised mouse event.
    *
    * @return   The mouse position, in an associative array.
    */
   this.getMousePos = function(evt)
   {
      var rect = win.canvas.getBoundingClientRect();
      var pos = 
      {
         "x" : evt.clientX - rect.left,
         "y" : evt.clientY - rect.top
      };
      
      return pos;
   };
   
   /**
    * Listen for events which may trigger window activation.
    *
    * @param    {String} mouseOp
    *           The mouse event listening on.
    * @param    {Object} mThis
    *           The <code>MouseHandler</code> instance.
    */
   function processWindowFocus(mouseOp, mThis)
   {
      return function(evt)
      {
         if (!p2j.screen.processMouse ||
             evt.defaultPrevented     ||
             !win.processOsEvents     ||
             p2j.isDialogDisplayed())
         {
            return;
         }

         // to request the input focus to this window
         win.canvas.focus();

         var topWindowId = p2j.screen.topWindowId();
         
         if (topWindowId == win.id)
         {
            return;
         }
         
         // move the window to top
         win.moveToTop();

         if (!topWindowId)
         {
            topWindowId = -1;
         }
         
         p2j.socket.sendWindowActivated(topWindowId, win.id);

         // consume the event
         evt.preventDefault();
      };
   }
   
   /**
    * Send mouse leave event to a widget
    * 
    *  @param wid 
    *         widget Id
    *  @param evt
    *         source mouse event
    */
   function sendMouseExit(wid, evt)
   {
      if (wid != undefined)
      {
         sendMouseEvent(wid, evt, "mouseleave", "mouseExit");
      }
   };

   /**
    * Send mouse leave enter to a widget
    * 
    *  @param wid 
    *         widget Id
    *  @param evt
    *         source mouse event
    */
   function sendMouseEnter(wid, evt)
   {
      if (wid != undefined)
      {
         sendMouseEvent(wid, evt, "mouseenter", "mouseEnter");
      }
   };

   /**
    * Send mouse leave to a widget
    * 
    *  @param wid 
    *         widget Id
    *  @param evt
    *         source mouse event
    *  @param mouseOp
    *         operation name
    *  @param handler
    *         the name of the handler method
    */
   function sendMouseEvent(wid, evt, mouseOp, handler)
   {
      // send an exit event to last widget
      var exitOpCode = p2j.screen.mouseOps[mouseOp] & 0xff;
      
      var osEvent = p2j.screen.getOsEvent(evt, 
                                          win, 
                                          mouseOp, 
                                          wid);
      if (osEvent && (typeof osEvent[handler] == "function"))
      {
         // let the OS event process it
         osEvent[handler](evt);
      }
      else
      {
         win.sendExplicitMouseEvent(
               wid, 
               evt, 
               win.getMousePos(evt), 
               win, 
               exitOpCode);
      }
   };

   /**
    * Listen for mouse move events.   This listener is intended to discard any mouse movement
    * inside the same widget.
    *
    * TODO: CA: test how it interacts with drag (i.e. in EDITOR).
    *
    * @param    {Object} mThis
    *           The <code>MouseHandler</code> instance.
    * @param    {String} mop
    *           The mouse event associated with this callback.
    */
   function processMouseMove(mThis, mop)
   {
      return function(evt)
      {
         if (!p2j.screen.canProcessOsEvent(evt, win))
         {
            return;
         }

         p2j.screen.setMousePos(evt);
         var mouseOp = mop;
         var wid = win.findMouseSource(evt, win.mouseAwareWidgets[mouseOp]);
         var mousePos = mThis.getMousePos(evt);

         if ((mouseOp == "mousemove" || mouseOp == "mouseover")) 
         {
            if (wid == undefined || wid != lastMouseWidget)
            {
               wid = win.findMouseSource(evt, win.mouseAwareWidgets["mouseenter"]);
               if (wid != lastMouseWidget)
               {
//                  console.log("*** Entered %s -> %s", lastMouseWidget, wid);
                  mouseOp = "mouseenter";
               }
            }
         }

         if (wid == undefined)
         {
            if (lastMouseWidget != undefined)
            {
               sendMouseExit(lastMouseWidget, evt);
            }

            lastMouseWidget = undefined;

            return;
         }
         
         var opCode = p2j.screen.mouseOps[mouseOp] & 0xff;

         var sendCurrent = false;
      
         // determine if the widget has changed
         if (lastMouseWidget != undefined && lastMouseWidget == wid)
         {
            // if the widget has not changed but has a "mousemove" op, then let it process
            // on the server-side
            if (mouseOp == "mousemove" || mouseOp == "mouseover")
            {
               return;
            }
            
            // otherwise, if widget has not changed, do nothing
         }
         else
         {
            if (mouseOp == "mouseenter")
            {
               // send an enter event to the previous widget
               sendMouseExit(lastMouseWidget, evt);
               // send an enter event to this widget
               sendMouseEnter(wid, evt);
            }
            else if (mouseOp == "mouseleave" || mouseOp == "mouseout" || mouseOp == "mousemove" || mouseOp == "mouseover")
            {
               // exiting the window
               sendCurrent = true;
            }
         
            // save the last widget the mouse was over
            lastMouseWidget = wid;
            
            if (sendCurrent)
            {
               win.sendMouseEvent(evt, mousePos, win, opCode, wid);
            }
         }
         
         // event was processed, cancel other listeners
         evt.preventDefault();
      };
   };
}

/**
 * Dedicated object to implement a widget resize action.  Currently, this is hardcoded for the 
 * WINDOW widget.
 */
function MouseResizeable()
{
   var BORDER_SIZE = 4;
   var rops = [ "mousemove", "mouseenter", "mouseleave", "mouseout", "mouseover"];
   //The peer canvas needs to animate the resize operation 
   var canvas = document.createElement('canvas');
   canvas.id = "resizable_peer_canvas";
   canvas.style.visibility  = "hidden";
   canvas.style["position"] = "fixed";
   var ctx = canvas.getContext("2d");
   
   // register default mouse listeners last
   for (var idx in rops)
   {
      // window border
      var mouseOp = rops[idx];
      document.addEventListener(mouseOp, processWindowBorder(mouseOp));
   }
   
   // dragging operations
   document.addEventListener("click",       processMouseDragStop ("click"));
   document.addEventListener("contextmenu", processMouseDragStop ("contextmenu"));
   document.addEventListener("dblclick",    processMouseDragStop ("dblclick"));
   document.addEventListener("mousedown",   processMouseDragStart("mousedown"));
   document.addEventListener("mouseup",     processMouseDragStop ("mouseup"));
   document.addEventListener("mousemove",   processMouseDrag     ("mousemove"));
   
   /**
    * Resizes the peer canvas according to given area.
    * 
    * @param    {Object} area
    *           The new size and position to be assigned to the peer canvas 
    */
   function resizePeerCanvas(area)
   {
      canvas.height = area.height;
      canvas.width  = area.width;
      canvas.style.top  = area.top  + "px";
      canvas.style.left = area.left + "px";
      ctx.fillStyle = "white";
      ctx.fillRect(0, 0, area.width, area.height);
      ctx.strokeStyle = "black";
      ctx.strokeRect(0, 0, area.width, area.height);

   }
   
   /**
    * Shows the resized peer canvas that represents the resize of the original window.
    * 
    * @param    {Window} win
    *           The original window to be resized.
    * @param    {String} resizeCursor
    *           The cursor style.
    */
   function displayResizedPeerCanvas(win, resizeCursor)
   {
      var parent = win.canvas.parentNode;
      if (parent)
      {
         canvas.height = win.canvas.height;
         canvas.width  = win.canvas.width;
         canvas.style.top  = win.canvas.style.top;
         canvas.style.left = win.canvas.style.left;
         canvas.style.zIndex = win.canvas.style.zIndex + 1;
         canvas.style.visibility  = "visible";
         canvas.style.opacity = 0.5;
         parent.appendChild(canvas);

         ctx.fillStyle = "white";
         ctx.fillRect(0, 0, canvas.width, canvas.height);
         ctx.strokeStyle = "black";
         ctx.strokeRect(0, 0, canvas.width, canvas.height);
         
         canvas.style.cursor = resizeCursor;
      }
   }
   
   /**
    * Calculates the resize cursor style.
    * 
    * @param    {Object} bounds
    *           The current client area given by its sides coordinates {top, left, right, bottom}
    * @param    {Object} mousePos
    *           The mouse position relative to the current client area given by its coordinates
    *           {x, y}.
    */
   function calculateResizeCursorStyle(bounds, mousePos)
   {
      // check the window borders
      if (mousePos.x <= BORDER_SIZE  || 
          mousePos.x >= bounds.right - bounds.left - BORDER_SIZE ||
          mousePos.y <= BORDER_SIZE   ||
          mousePos.y >= bounds.bottom - bounds.top - BORDER_SIZE)
      {
         // we have 8 cases; corners have precedence
         var cursors =
         [
            "nw-resize",
            "ne-resize",
            "se-resize",
            "sw-resize",
            "n-resize",
            "e-resize",
            "s-resize",
            "w-resize"
         ];
         var xcoords1 =
         [
            bounds.left,  // NW
            bounds.right, // NE
            bounds.right, // SE
            bounds.left,  // SW
            bounds.left,  // N
            bounds.right, // E
            bounds.left,  // S
            bounds.left   // W
         ];
         var xcoords2 =
         [
            bounds.left,  // NW
            bounds.right, // NE
            bounds.right, // SE
            bounds.left,  // SW
            bounds.right, // N
            bounds.right, // E
            bounds.right, // S
            bounds.left   // W
         ];
         var ycoords1 =
         [
            bounds.top,    // NW
            bounds.top,    // NE
            bounds.bottom, // SE
            bounds.bottom, // SW
            bounds.top,    // N
            bounds.top,    // E
            bounds.bottom, // S
            bounds.top     // W
         ];
         var ycoords2 =
         [
            bounds.top,    // NW
            bounds.top,    // NE
            bounds.bottom, // SE
            bounds.bottom, // SW
            bounds.top,    // N
            bounds.bottom, // E
            bounds.bottom, // S
            bounds.bottom  // W
         ];
         
         for (var i = 0; i < cursors.length; i++)
         {
            var x1 = xcoords1[i] - BORDER_SIZE - bounds.left;
            var y1 = ycoords1[i] - BORDER_SIZE - bounds.top;
            var x2 = xcoords2[i] + BORDER_SIZE - bounds.left;
            var y2 = ycoords2[i] + BORDER_SIZE - bounds.top;
            var w = x2 - x1 + 1;
            var h = y2 - y1 + 1;
            
            if (mousePos.x >= x1 && mousePos.x <= x2 && mousePos.y >= y1 && mousePos.y <= y2)
            {
               return cursors[i];
            }
         }
      }
      return null;
   };

   // leave/out listeners are not required for dragging purposes.
   /**
    * If the event acts on the window border, send it to the Java side immediately.
    * 
    * @param    {Object} mThis
    *           The <code>MouseHandler</code> instance.
    * @param    {String} mouseOp
    *           The string representation of the event type.
    */
   function processWindowBorder(mouseOp)
   {
      var opCode = p2j.screen.mouseOps[mouseOp] & 0xff;
      
      return function(evt)
      {
         // TODO: this doesn't seem right, resize cursor is also shown for other
         //       visible windows when hovered over, not just for the top one.
         
         // determine the top-most visible window
         var win = p2j.screen.topVisibleWindow();
         if (win == undefined || win.isResizable)
         {
            // no window is shown
            return;
         }
//         console.log("%s:processWindowBorder(%s)", win.id, mouseOp);
         var mThis = win.mouseHandler;
         
         if (mThis.mouseDrag || !p2j.screen.canProcessOsEvent(evt, win))
         {
            return;
         }

         var bounds = win.canvas.getBoundingClientRect();
         var mousePos = mThis.getMousePos(evt);

         var w = win.getHoverableWidget(win.id);
         var wstyle = w ? w.ptr : "default";

         var cursor = calculateResizeCursorStyle(bounds, mousePos);
         if (!cursor) // ouside the controlled area
         {
            p2j.screen.refreshHovered();
            return;
         }
         if (cursor !== wstyle)
         {
            if (cursor != win.getCursorStyle())
            {
//               console.log("processWindowBorder(%s): set window cursor to %s", mouseOp, cursor);
               win.setCursorStyle(cursor);
            }
         }
         else
         {
            // restore mouse, if current is set to a "resize" style
            var style = win.getCursorStyle();
            if (style.lastIndexOf("resize") >= 0 && !mThis.mouseDrag)
            {
//               console.log("processWindowBorder(%s): reset window cursor to default", mouseOp);
               win.setCursorStyle("default");
            }
         }
      };
   };
   
   function processMouseDragStop(mouseOp)
   {
      return function(evt)
      {
         // determine the top-most visible window
         var win = p2j.screen.topVisibleWindow();
         if (win == undefined)
         {
            // no window is shown
            return;
         }
         var mThis = win.mouseHandler;

         if (!win.resizeable || !mThis.mouseDrag || !mThis.resizeArea)
         {
            return;
         }
         // resize the offscreen canvas
         win.resize(mThis.resizeArea.width, mThis.resizeArea.height);
         
         // set dimension for the window canvas
         win.canvas.width  = mThis.resizeArea.width;
         win.canvas.height = mThis.resizeArea.height;
         
         // blast the offscreen image to the window canvas
         win.ctx.drawImage(win.offscreenCanvas, 0, 0);

         // set new location
         win.canvas.style.left = mThis.resizeArea.left + "px";
         win.canvas.style.top  = mThis.resizeArea.top + "px";
         
         var bounds = win.canvas.getBoundingClientRect();
         var h = bounds.bottom - bounds.top  + 1;
         var w = bounds.right - bounds.left + 1;

         // send the window is resized to the java side
         p2j.socket.sendWindowResized(win.id, bounds.left, bounds.top, w, h);
         
         mThis.mouseDrag = false;
         mThis.initialBounds = undefined;
         mThis.resizeEvent = undefined;
         mThis.dragOwner = undefined;
         canvas.style.cursor = "default";
         canvas.style.visibility  = "hidden";
         if (canvas.parentNode)
         {
            canvas.parentNode.removeChild(canvas);
         }
         mThis.resizeArea = undefined;

         // event was processed, cancel other listeners
         evt.preventDefault();
      };
   };
   
   function processMouseDragStart(mouseOp)
   {
      return function(evt)
      {
         // determine the top-most visible window
         var win = p2j.screen.topVisibleWindow();
         if (win == undefined)
         {
            // no window is shown
            return;
         }
         var mThis = win.mouseHandler;

         // only left button can trigger move
         if (!win.resizeable || evt.which != 1)
         {
            return;
         }
         
         if (mThis.mouseDrag)
         {
            evt.preventDefault();
            return;
         }

         var bounds = win.canvas.getBoundingClientRect();
         var mousePos = mThis.getMousePos(evt);
         var resizeCursor = calculateResizeCursorStyle(bounds, mousePos);
         var w = bounds.right - bounds.left;
         var h = bounds.bottom - bounds.top;
         if (!(inside(mousePos, BORDER_SIZE, BORDER_SIZE, w - BORDER_SIZE, h - BORDER_SIZE))
               && inside(mousePos, -BORDER_SIZE, -BORDER_SIZE, w + BORDER_SIZE, h + BORDER_SIZE))
         {
            mThis.dragOwner = win.id;
            mThis.mouseDrag = true;
            // save the location
            mThis.initialBounds = win.canvas.getBoundingClientRect();
            mThis.resizeEvent = evt;

            mThis.resizeCursor = resizeCursor;
            displayResizedPeerCanvas(win, resizeCursor);
            // event was processed, cancel other listeners
            evt.preventDefault();
         }
      };
   };
   
   function processMouseDrag(mouseOp)
   {
      return function(evt)
      {
         // determine the top-most visible window
         var win = p2j.screen.topVisibleWindow();
         if (win == undefined)
         {
            // no window is shown
            return;
         }
         var mThis = win.mouseHandler;

         if (mThis.dragOwner != win.id ||
             !win.resizeable           ||
             !mThis.mouseDrag          ||
             !p2j.screen.canProcessOsEvent(evt, win))
         {
            return;
         }

         var mousePos = win.getMousePos(evt);
         var bounds = win.canvas.getBoundingClientRect();

         var activeCursor = mThis.resizeCursor;

         // check which cursor is in use
         var setLocationX = (activeCursor == "nw-resize" ||
                             activeCursor == "w-resize"  ||
                             activeCursor == "sw-resize");
         var setLocationY = (activeCursor == "nw-resize"  ||
                             activeCursor == "n-resize"   ||
                             activeCursor == "ne-resize");
         
         // depending on the dragging point, it will resize x axis, y axis or both
         var resizeX = (activeCursor == "nw-resize" || 
                        activeCursor == "ne-resize" ||
                        activeCursor == "sw-resize" ||
                        activeCursor == "se-resize" ||
                        activeCursor == "e-resize"  ||
                        activeCursor == "w-resize");
         var resizeY = (activeCursor == "nw-resize" || 
                        activeCursor == "ne-resize" ||
                        activeCursor == "sw-resize" ||
                        activeCursor == "se-resize" ||
                        activeCursor == "n-resize"  ||
                        activeCursor == "s-resize");
         var signX = (activeCursor == "w-resize" ||
                      activeCursor == "nw-resize" ||
                      activeCursor == "sw-resize" ? 1 : -1);
         var signY = (activeCursor == "n-resize" ||
                      activeCursor == "nw-resize" ||
                      activeCursor == "ne-resize" ? 1 : -1);
   
         // determine the new origin
         var origin = {};
         origin.x = bounds.left;
         origin.y = bounds.top;
         
         
         // determine the new size
         var dx = mThis.initialBounds.right - mThis.initialBounds.left + 1;
         var dy = mThis.initialBounds.bottom - mThis.initialBounds.top + 1;
         var new_dx = dx;
         var new_dy = dy;
         if (resizeX)
         {
            new_dx -= signX * (evt.clientX - mThis.resizeEvent.clientX);
         }
         
         if (resizeY)
         {
            new_dy -= signY * (evt.clientY - mThis.resizeEvent.clientY);
         }
         
         var maxWidth  = Math.max( win.canvas.width,  win.maxWidth);
         var maxHeight = Math.max( win.canvas.height, win.maxHeight);
         // 
         new_dx = Math.min(Math.max(new_dx, win.minWidth), maxWidth);
         new_dy = Math.min(Math.max(new_dy, win.minHeight), maxHeight);
         //recalculate origin 
         if (setLocationX || setLocationY)
         {
            if (!(setLocationX && setLocationY))
            {
               origin.x = bounds.left;
               origin.y = bounds.top;
            }
            else
            {
               origin.x = 0;
               origin.y = 0;
            }
            
            if (setLocationX)
            {
               origin.x = mThis.resizeEvent.clientX + signX * (dx - new_dx);
            }
            
            if (setLocationY)
            {
               origin.y = mThis.resizeEvent.clientY + signY * (dy - new_dy);
            }
         }
         mThis.resizeArea = {left : origin.x, top : origin.y, width : new_dx, height : new_dy};
         resizePeerCanvas(mThis.resizeArea);
         
         // event was processed, cancel other listeners
         evt.preventDefault();
      };
   };

}

/**
 * Dedicated object to implement a complex mouse action: when entering a hoverable widget,
 * it sets the loadded cursor without any trips to the Java side.
 *
 * @param    {Object} win
 *           The window where to listen for events.
 * @param    {Number} wid
 *           The ID of the hoverable widget.
 * @param    {Number} pid
 *           The parent widget ID.
 * @param    {Number} x
 *           The left-corner position of the widget.
 * @param    {Number} y
 *           The top-corner position of the widget.
 * @param    {Number} width
 *           The width of the widget.
 * @param    {Number} height
 *           The height of the widget.
 * @param    {Object} ptr
 *           Loaded mouse pointer
 * @param    {Object} defaultPtr
 *           Default mouse pointer
 * @param    {boolean} forceDefault
 *           Use default pointer only
 */
function MouseHoverable(win, wid, pid,  x, y, width, height, ptr, defaultPtr, forceDefault)
{
   var lastMouseWidget = undefined;
   var ma = { "x" : x, "y" : y, "width" : width, "height" : height };
   this.wid = wid;
   this.pid = pid;
   this.ptr = ptr;
   this.defaultPtr = defaultPtr; 
   this.forceDefault = forceDefault;
   this.active = false;
   
   this.mouseExit = function(evt)
   {
      this.active = false;
      
      // sets the effective parent mouse pointer
      var parentHoverable = win.getHoverableWidget(pid);
      win.setHovered(parentHoverable);
      if (parentHoverable)
      {
         parentHoverable.setCursorPtr();
      }
      
      var exitOpCode = p2j.screen.mouseOps["mouseleave"] & 0xff;
      win.sendExplicitMouseEvent(wid, 
            evt, 
            win.getMousePos(evt), 
            win, 
            exitOpCode);
//      console.log("mouseExit(%s); parent: %s", wid, pid);
      evt.preventDefault();
   };

   this.mouseEnter = function(evt)
   {
      this.active = true;
      win.setHovered(this);
      
      p2j.screen.setActiveWindow(win);
//      console.log("mouseEnter(%s:%s): %s; %s", win.id, wid, this.ptr, this.forceDefault);
      win.setCursorStyle(this.effectivePtr(), this.wid == p2j.screen.getActiveDropDown());
      var exitOpCode = p2j.screen.mouseOps["mouseenter"] & 0xff;
      win.sendExplicitMouseEvent(wid, 
            evt, 
            win.getMousePos(evt), 
            win, 
            exitOpCode);
      evt.preventDefault();
   };

   this.setCursorPtr = function()
   {
//      console.log("%s: setCursorPtr: %s", wid, this.ptr);
      win.setCursorStyle(this.effectivePtr());
   }

   this.customPtr = function()
   {
      if (this.ptr != null)
      {
         return this.ptr;
      }
      var p = win.getHoverableWidget(pid);
      return p ? p.customPtr() : null;
   };

   this.effectivePtr = function()
   {
      if (this.forceDefault)
      {
         return this.defaultPtr;
      }
      var p = this.customPtr();
      return p != null ? p : this.defaultPtr;
   }

   this.deregisterListeners = function()
   {
      // no-op
   };


   this.setPtr = function(ptr)
   {
//      console.log("%s.setPtr(%s)", wid, ptr);
      this.ptr = ptr;
      if (this.active)
      {
         win.setCursorStyle(this.effectivePtr());
      }
   };

   this.getEvents = function()
   {
      return { "mouseenter" : true, "mouseleave" : true};
   };
};

/**
 * Dedicated object to implement a widget move action.  Currently, this is hardcoded for the 
 * WINDOW widget.
 *
 * @param    {Object} win
 *           The window where to listen for events.
 * @param    {Number} wid
 *           The ID of the editable widget.
 * @param    {Number} x
 *           The left-corner position of the widget.
 * @param    {Number} y
 *           The top-corner position of the widget.
 * @param    {Number} width
 *           The width of the widget.
 * @param    {Number} height
 *           The height of the widget.
 */
function MouseMovable(win, wid, x, y, width, height)
{
   var mouseDrag = false;
   var ma = { "x" : x, "y" : y, "width" : width, "height" : height };
   // register the listeners
   // 1. drag: this is computed from three events:
   // a. mouse down: enables possible drag
   // b. mouse move: if drag start, it raises a drag event
   // c. mouse up: disables the drag
   var mops = ["click", "contextmenu", "dblclick", "mousedown", "mouseup", "mousemove", "mouseleave", "mouseout"];
   var listeners = [processMouseDragStop (this, "click"), processMouseDragStop (this, "contextmenu"),
                    processMouseDragStop (this, "dblclick"), processMouseDragStart(this, "mousedown"),
                    processMouseDragStop (this, "mouseup"), processMouseDrag     (this, "mousemove"),
                    processMouseDragStop (this, "mouseleave"), processMouseDragStop (this, "mouseout")];
   // win.canvas.addEventListener("drag", processMouseDrag(this, "drag"));
   addListeners(win, mops, listeners);
   
   this.deregisterListeners = function()
   {
      removeListeners(win, mops, listeners);
   };

   this.getEvents = function()
   {
      var events = {};
      for (var i = 0; i < mops.length; i++)
      {
         events[mops[i]] = true;
      }
      
      return events;
   };

   function processMouseDragStop(mThis, mouseOp)
   {
      return function(evt)
      {
         if (!mThis.mouseDrag || !p2j.screen.canProcessOsEvent(evt, win))
         {
            return;
         }

         // send the window location
         p2j.socket.sendWindowLocation(win.id, mThis.screenLoc.x, mThis.screenLoc.y);

         mThis.mouseDrag = false;
         mThis.screenLoc = undefined;
         mThis.lastMouseEvent = undefined;
         mThis.dragOwner = undefined;

         // event was processed, cancel other listeners
         evt.preventDefault();
      };
   };
   
   function processMouseDragStart(mThis, mouseOp)
   {
      return function(evt)
      {
         // only left button can trigger move
         if (evt.which != 1                     || 
             !p2j.screen.canProcessOsEvent(evt, win) || 
             !win.canProcessWidget(wid, evt, mouseOp))
         {
            return;
         }
         
         if (mThis.mouseDrag)
         {
            evt.preventDefault();
            return;
         }

         var mousePos = win.getMousePos(evt);
   
         if (mousePos.x >= ma.x            &&
             mousePos.x <= ma.x + ma.width &&
             mousePos.y >= ma.y            &&
             mousePos.y <= ma.y + ma.height)
         {
            mThis.mouseDrag = true;
           
            // save the location
            var rect = win.canvas.getBoundingClientRect();
            mThis.screenLoc = { "x" : rect.left, "y" : rect.top };
            mThis.lastMouseEvent = evt;
            mThis.dragOwner = wid;
           
            // event was processed, cancel other listeners
            evt.preventDefault();
         }
      };
   };
   
   function processMouseDrag(mThis, mouseOp)
   {
      return function(evt)
      {
         if (mThis.dragOwner != wid || !mThis.mouseDrag || !p2j.screen.canProcessOsEvent(evt, win))
         {
            return;
         }

         var mousePos = win.getMousePos(evt);

         var dx = evt.clientX - mThis.lastMouseEvent.clientX;
         var dy = evt.clientY - mThis.lastMouseEvent.clientY;

         var rect = win.canvas.getBoundingClientRect();
         
         var x = dx + rect.left;
         var y = dy + rect.top;
         
         win.canvas.style.left = x.toString() + "px";
         win.canvas.style.top  = y.toString() + "px";

         // !!! set the current location in order to send the actual data to java web client side.
         mThis.screenLoc = { "x" : x, "y" : y };

         mThis.lastMouseEvent = evt;

         // event was processed, cancel other listeners
         evt.preventDefault();
      };
   };
};

/**
 * Dedicated worker to implement a window state change action.
 *
 * @param    {Object} win
 *           The window where to listen for events.
 * @param    {Number} wid
 *           The ID of the caption button.
 * @param    {Number} x
 *           The left-corner position of the widget.
 * @param    {Number} y
 *           The top-corner position of the widget.
 * @param    {Number} width
 *           The width of the widget.
 * @param    {Number} height
 *           The height of the widget.
 * @param    {Function} stateChangedCallback
 *           A callback function for this window state change action.
 */
function WindowStateChangedWorker(win, wid, x, y, width, height, stateChangedCallback)
{
   var ma = { "x" : x, "y" : y, "width" : width, "height" : height };
   var mops = ["dblclick", "click", "mousedown", "mouseup"];
   var listeners = [];
   for (var i = 0; i < mops.length; i++)
   {
      listeners[i] = processWindowState(this, mops[i]);
   }
   addListeners(win, mops, listeners);
   
   function processWindowState(mThis, mouseOp)
   {
      var opCode = p2j.screen.mouseOps[mouseOp] & 0xff;

      return function(evt)
      {
         if (evt.which != 1 || !p2j.screen.canProcessOsEvent(evt, win))
         {
            return;
         }

         var mousePos = win.getMousePos(evt);
         
         if (mousePos.x >= ma.x            &&
             mousePos.x <= ma.x + ma.width &&
             mousePos.y >= ma.y            &&
             mousePos.y <= ma.y + ma.height)
         {
            stateChangedCallback(win, evt, mousePos, opCode);
            
            evt.preventDefault();
         }
      };
   };

   this.getEvents = function()
   {
      return { "dblclick" : true, "click" : true, "mousedown" : true, "mouseup" : true };
   };

   this.deregisterListeners = function()
   {
      removeListeners(win, mops, listeners);
   };
};

/**
 * Dedicated object to implement a window close action.
 *
 * @param    {Object} win
 *           The window where to listen for events.
 * @param    {Number} wid
 *           The ID of the caption button.
 * @param    {Number} x
 *           The left-corner position of the widget.
 * @param    {Number} y
 *           The top-corner position of the widget.
 * @param    {Number} width
 *           The width of the widget.
 * @param    {Number} height
 *           The height of the widget.
 */
function WindowClose(win, wid, x, y, width, height)
{
   this.wsc = new WindowStateChangedWorker(win, wid, x, y, width, height, windowCloseCallback);

   function windowCloseCallback(win, evt, mousePos, opCode)
   {
      win.close();

      // let the event be processed by java side
      win.sendMouseEvent(evt, mousePos, win, opCode, wid);
   };

   this.getEvents = function()
   {
      return this.wsc.getEvents();
   };

   this.deregisterListeners = function()
   {
      this.wsc.deregisterListeners();
   };
};

/**
 * Dedicated object to implement a window iconify action.
 *
 * @param    {Object} win
 *           The window where to listen for events.
 * @param    {Number} wid
 *           The ID of the caption button.
 * @param    {Number} x
 *           The left-corner position of the widget.
 * @param    {Number} y
 *           The top-corner position of the widget.
 * @param    {Number} width
 *           The width of the widget.
 * @param    {Number} height
 *           The height of the widget.
 */
function WindowIconify(win, wid, x, y, width, height)
{
   this.wsc = new WindowStateChangedWorker(win, wid, x, y, width, height, windowIconifyCallback);

   function windowIconifyCallback(win, evt, mousePos, opCode)
   {
      if (evt.type === "mouseup")
      {
         win.iconify();
         
         p2j.socket.sendWindowIconState(win.id, true);
      }
   };

   this.getEvents = function()
   {
      return this.wsc.getEvents();
   };

   this.deregisterListeners = function()
   {
      this.wsc.deregisterListeners();
   };
};

/**
 * Dedicated object to implement a window maximize/restore action.
 *
 * @param    {Object} win
 *           The window where to listen for events.
 * @param    {Number} wid
 *           The ID of the caption button.
 * @param    {Number} x
 *           The left-corner position of the widget.
 * @param    {Number} y
 *           The top-corner position of the widget.
 * @param    {Number} width
 *           The width of the widget.
 * @param    {Number} height
 *           The height of the widget.
 */
function WindowMaximize(win, wid, x, y, width, height)
{
   this.wsc = new WindowStateChangedWorker(win, wid, x, y, width, height, windowMaximizeCallback);

   function windowMaximizeCallback(win, evt, mousePos, opCode)
   {
      win.maximize();

      // let the event be processed by java side
      win.sendMouseEvent(evt, mousePos, win, opCode, wid);
   };

   this.getEvents = function()
   {
      return this.wsc.getEvents();
   };

   this.deregisterListeners = function()
   {
      this.wsc.deregisterListeners();
   };
};

/**
 * Dedicated object to listen for a popup menu.
 * <p>
 * This implementation delegates the work to the Java side, and consumes the event.
 *
 * @param    {Object} win
 *           The window where to listen for events.
 * @param    {Number} wid
 *           The ID of the caption button.
 * @param    {Number} x
 *           The left-corner position of the widget.
 * @param    {Number} y
 *           The top-corner position of the widget.
 * @param    {Number} width
 *           The width of the widget.
 * @param    {Number} height
 *           The height of the widget.
 * @param    {Number} btnId
 *           The button which triggers the popup menu.  1 for left, 3 for right.
 */
function MousePopupable(win, wid, x, y, width, height, btnId)
{
   var ma = { "x" : x, "y" : y, "width" : width, "height" : height };
   var events = {};
   var mop;
   if (btnId === 1)
   {
      // left mouse triggers the popup
      mop = "click";
   }
   else if (btnId === 3)
   {
      // right mouse triggers the popup
      mop = "contextmenu";
   }
   else
   {
      return;
   }
   var listener = processPopup(this, mop);
   win.canvas.addEventListener(mop, listener);
   events[mop] = true;
   
   function processPopup(mThis, mouseOp)
   {
      var opCode = p2j.screen.mouseOps[mouseOp] & 0xff;

      return function(evt)
      {
         if (evt.which != btnId || !p2j.screen.canProcessOsEvent(evt, win))
         {
            return;
         }

         var mousePos = win.getMousePos(evt);

         if (mousePos.x >= ma.x            &&
             mousePos.x <= ma.x + ma.width &&
             mousePos.y >= ma.y            &&
             mousePos.y <= ma.y + ma.height)
         {
            // send the event and consume it
            win.sendMouseEvent(evt, mousePos, win, opCode, wid);

            evt.preventDefault();
         }
         
         // let the event be processed by other widgets
      };
   };

   this.deregisterListeners = function()
   {
      win.canvas.removeEventListener(mop, listener);
   };

   this.getEvents = function()
   {
      return events;
   };
};

/**
 * Dedicated object to listen for a sytem action.
 *
 * @param    {Object} win
 *           The window where to listen for events.
 * @param    {Number} wid
 *           The ID of the caption button.
 * @param    {Number} x
 *           The left-corner position of the widget.
 * @param    {Number} y
 *           The top-corner position of the widget.
 * @param    {Number} width
 *           The width of the widget.
 * @param    {Number} height
 *           The height of the widget.
 * @param    {String} osOpName
 *           The system operation name.
 */
function SystemAction(win, wid, x, y, width, height, osOpName)
{
   var ma = { "x" : x, "y" : y, "width" : width, "height" : height };
   var events = {};
   var mop = "click";
   var btnId = 1;
   var listener = processOSSystemAction(this, mop, osOpName);
   win.canvas.addEventListener(mop, listener);
   events[mop] = true;
   
   function processOSSystemAction(mThis, mouseOp, osOp)
   {
      var opCode = p2j.screen.mouseOps[mouseOp] & 0xff;

      return function(evt)
      {
         if (evt.which != btnId || !p2j.screen.canProcessOsEvent(evt, win))
         {
            return;
         }

         var mousePos = win.getMousePos(evt);

         if (mousePos.x >= ma.x            &&
             mousePos.x <= ma.x + ma.width &&
             mousePos.y >= ma.y            &&
             mousePos.y <= ma.y + ma.height)
         {
            var onSystemMenuAction = function()
            {
               // send the event and consume it
               win.sendMouseEvent(evt, mousePos, win, opCode, wid);
            }
            
            if (osOp === "copy" || osOp === "cut")
            {
               p2j.clipboard.notifyAboutUserInitiatedAction(onSystemMenuAction);
               p2j.clipboard.prepareCopy();
               if (p2j.commandSupported(osOp) && document.execCommand(osOp))
               {
                  console.debug("document.execCommand(" + osOp + ")");
               }
               else
               {
                  var dialogCloseHandler;
                  if (osOp === "cut")
                  {
                     dialogCloseHandler = function(evt)
                     {
                        if ((evt.metaKey  && (evt.keyCode === keys.X) && p2j.isSafari) ||
                              (((evt.shiftKey && (evt.keyCode === keys.DELETE)) ||
                                ( evt.ctrlKey && (evt.keyCode === keys.X))
                               ) && !p2j.isSafari))
                        {
                           p2j.closeDialog();
                           p2j.clipboard.prepareCopy();
                        }
                     }
                  }
                  else
                  {
                     dialogCloseHandler = function(evt)
                     {
                        if ((evt.metaKey && (evt.keyCode === keys.C) && p2j.isSafari) ||
                            (evt.ctrlKey &&
                             (evt.keyCode === keys.C || evt.keyCode === keys.INSERT) &&
                                    !p2j.isSafari))
                        {
                           p2j.closeDialog();
                           p2j.clipboard.prepareCopy();
                        }
                     }
                  }
                  var maxZIndex = p2j.getMaxZIndex();
                  p2j.createDialog(
                        "writeClipboard",
                        1000 + maxZIndex,
                        document.body,
                        p2j.prompts(osOp),
                        p2j.hints(osOp),
                        dialogCloseHandler);
               }
            }
            else if (osOp === "paste")
            {
               p2j.clipboard.notifyAboutUserInitiatedAction(onSystemMenuAction);
               p2j.clipboard.preparePaste();
               if (!p2j.isIe && p2j.commandSupported(osOp) && document.execCommand(osOp))
               {
                  console.debug("document.execCommand(" + osOp + ")");
               }
               else
               {
                  var maxZIndex = p2j.getMaxZIndex();
                  p2j.createDialog(
                        "readClipboard",
                        1000 + maxZIndex,
                        document.body,
                        p2j.prompts("paste"),
                        p2j.hints("paste"),
                        function (evt)
                        {
                           if ((evt.metaKey  && (evt.keyCode === keys.V) && p2j.isSafari) ||
                               (((evt.shiftKey && (evt.keyCode === keys.INSERT)) ||
                                 ( evt.ctrlKey && (evt.keyCode === keys.V))
                                ) && !p2j.isSafari))
                           {
                              p2j.closeDialog();
                              p2j.clipboard.preparePaste();
                           }
                        });
               }
            }
            else if (osOp === "minimize")
            {
               win.cleanupCallback = function()
               {
                  // win is an overlay window, but the target window to minimize is its owner.
                  var target = p2j.screen.getWindow(win.owner);
                  target.iconify();
               };
               onSystemMenuAction();
            }
            p2j.consumeEvent(evt);
         }
      };
   };

   this.deregisterListeners = function()
   {
      win.canvas.removeEventListener(mop, listener);
   };

   this.getEvents = function()
   {
      return events;
   };
};

/**
 * Returns true if the point is inside the rectangle given by its sides coordinates: left,
 * top, right and bottom.
 * 
 * @param    {Object} point
 *           The point {x: x, y : y} to test.
 * @param    {Number} left
 *           The coord of the left side of the target rectangle.
 * @param    {Number} top
 *           The coord of the top side of the target rectangle.
 * @param    {Number} right
 *           The coord of the right side of the target rectangle.
 * @param    {Number} bottom
 *           The coord of the bottom side of the target rectangle.
 * 
 * @return   {Boolean}
 *           True iff the point is inside the target.
 */
function inside(point, left, top, right, bottom)
{
   return (point.x > left && point.x < right &&  point.y > top  && point.y < bottom);
}

/**
 * Adds window events listeners.
 * 
 * @param    {Window} win
 *           The window for which the target events listeners must be added.
 * @param    {Array} mops
 *           The handled events.
 * @param    {Array} listeners
 *           The target listeners corresponding to their handled events
 */
function addListeners(win, mops, listeners)
{
  for(var i = 0; i < listeners.length; i++)
  {
     win.canvas.addEventListener(mops[i], listeners[i]);
  }
}

/**
 * Removes window events listeners.
 * 
 * @param    {Window} win
 *           The window for which the target events handlers must be removed.
 * @param    {Array} mops
 *           The handled events to remove.
 * @param    {Array} listeners
 *           The target listeners corresponding to their handled events
 */
function removeListeners(win, mops, listeners)
{
   for(var i = 0; i < listeners.length; i++)
   {
      win.canvas.removeEventListener(mops[i], listeners[i]);
   }
}