/*
** Module   : ThemeManager.java
** Abstract : Theme manager responsible with loading and retrieving the current Theme for users. 
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20170309 First version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.gui.theme;

import com.goldencode.p2j.ui.*;

/**
 * This manager is the central point for obtaining and setting the current {@code Theme}. The
 * current is defined for the entire client process. This class is client-side only. It will
 * request the configuration from server, but it will use that as a hint only. The class
 * configured in server directory might no contain a proper theme/class name. 
 */
public class ThemeManager
{
   /** 
    * Current theme. May remain {@code null} after calls to {@link #setTheme(String)}. Any call to
    * {@link #getCurrentTheme()} will make sure this field is initialized with the default value
    * before returning it.
    */
   private static Theme crtTheme = null;
   
   /**
    * Obtain the current {@code Theme}. If not already initialized the default theme is
    * initialized and returned.
    *
    * @return  the current {@code Theme}. Always not {@code null}.
    */
   public static Theme getCurrentTheme()
   {
      if (crtTheme == null)
      {
         crtTheme = new Windows10Theme();
      }
      
      return crtTheme;
   }
   
   /**
    * Sets a new {@link Theme}. Before setting the new theme, a check if performed to test if the
    * theme is already active. If already active, the theme is not reloaded.
    * <p>
    * The old theme is not removed and its resources released until the new {@code Theme} was
    * successfully initialized.
    * <p> 
    * TODO: when a new theme is set, all resources need to be re-requested from server (fonts, 
    *       font-sizes and color palette). 
    *
    * @param   name
    *          The name of the new {@code Theme} to be activated. The built-in 
    *          {@link ClassicTheme} and {@link Windows10Theme} are recognized by their short
    *          names.
    *
    * @return  {@code true} if the old {@code Theme} was replaced with the new one. 
    */
   @SuppressWarnings("unchecked")
   public static boolean setTheme(String name)
   {
      if (name == null || name.isEmpty())
      {
         // parameter validation
         return false;
      }
      
      if (crtTheme != null &&
          (crtTheme.getName().equals(name) || crtTheme.getClass().getName().equals(name)))
      {
         return false; // same theme already installed
      }
      
      // support for nicknaming builtin Themes:
      if (name.equalsIgnoreCase("classic")   ||
          name.equals("ClassicTheme")        ||
          name.equals("com.goldencode.p2j.ui.client.gui.theme.ClassicTheme"))
      {
         return setTheme(ClassicTheme.class);
      }
      else if (name.equalsIgnoreCase("win10") ||
               name.equals("Windows10Theme")  ||
               name.equals("com.goldencode.p2j.ui.client.gui.theme.Windows10Theme"))
      {
         return setTheme(Windows10Theme.class);
      }
      else
      {
         try
         {
            //  parse and load the class
            Class<?> ufoClass = Class.forName(name);
            if (!Theme.class.isAssignableFrom(ufoClass))
            {
               System.err.println("[" + name + "] is not a Theme manager.");
               return false;
            }
         
            return setTheme((Class<Theme>) ufoClass); // warning disabled by @SuppressWarnings
         }
         catch (ClassNotFoundException e)
         {
            System.err.println("Could not load [" + name + "] theme.");
            return false;
         }
      }
   }
   
   /**
    * Sets a new {@link Theme}. The old theme is not removed and its resources released until
    * the new {@code Theme} was successfully initialized.
    *
    * @param   newThemeClass
    *          The new class to be instantiated and the new object to be set as new active
    *          theme.
    *
    * @return  {@code true} if the old {@code Theme} was replaced with the new one. 
    */
   private static boolean setTheme(Class<? extends Theme> newThemeClass)
   {
      Theme newTheme;
      try
      {
         newTheme = newThemeClass.newInstance();
      }
      catch (InstantiationException | IllegalAccessException e)
      {
         System.err.println("Could not instantiate [" + newThemeClass + "] theme.");
         return false;
      }
      
      // if everything went fine up to here, we can replace the old theme with the new one
      if (crtTheme != null)
      {
         // after releasing any resources occupied by the old theme
         crtTheme.releaseResources();
      }
      crtTheme = newTheme;
      
      System.err.println("UI Theme successfully changed to '" + newTheme.getName() + "'");
      // TODO: notify listeners that current theme has changed and they need to do their layout
      
      return true;
   }
   
   /**
    * Initialize the {@code ThemeManager} using negotiation with the server. If a theme is not
    * requested in command line (passed in by {@code preferredTheme} parameter), it will request
    * the configuration from server directory, but it will use that as a hint only. The class
    * configured in server directory might no contain a proper theme/class name. If no definition
    * is found in directory, or the string obtained does not represent a name or full class of a
    * theme accessible from client path, the theme remains uninitialized.
    * It will be lazily instantiated when accessed the first time.
    *
    * @param   server
    *          the {@link ServerExports} object for communication with server.
    * @param   preferredTheme
    *          the preferred theme from command-line.
    */
   public static boolean init(ServerExports server, String preferredTheme)
   {
      if (preferredTheme == null || preferredTheme.isEmpty())
      {
         // request a configuration from server
         preferredTheme = server.getUiTheme();
      }
      
      // we attempt to set this theme. If call fails, the default theme will automatically be set,
      // using lazy initialization, at first call to ThemeManager.getCurrentTheme() 
      return setTheme(preferredTheme);
   }
   
   /** Flag for debugging GUI. Set this to true to enable debug code. */
   private static boolean DEBUG = false;
   
   /**
    * Execute some debug code if debug is enabled.
    *
    * @param   debugCode
    *          The code to be executed.
    */
   public static void debug(Runnable debugCode)
   {
      if (DEBUG)
      {
         debugCode.run();
      }
   }
}