/*
** Module   : BorderedPanel.java
** Abstract : Bordered panel base class.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 SIY 20100822 Added javadoc.
** 004 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
** 005 HC  20150831 Border panel refactored to make it suitable for composition.
** 006 SBI 20151022 Move draw upper to subclasses due to the resize issue inside drawing cycles
**                  in the case of the web client.
** 007 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.widget;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.layout.*;

/**
 * Bordered panel base class.
 */
public abstract class BorderedPanel<O extends OutputManager<?>>
extends AbstractContainer<O>
{
   /** Border around the panel. */
   protected Border<O> border = null;
   
   /** When true, the panel doesn't draw its background. */
   private boolean transparent;

   /**
    * Default constructor.
    */
   public BorderedPanel()
   {
   }

   /**
    * Constructor.
    * 
    * @param   border
    *          Border for this widget.
    */
   public BorderedPanel(Border<O> border)
   {
      this.border = border;
   }
   
   /**
    * Constructor.
    * 
    * @param   mgr
    *          Layout manager.
    * @param   border
    *          Border for this widget.
    */
   public BorderedPanel(LayoutManager<O> mgr, Border<O> border)
   {
      setLayout(mgr);
      this.border = border;
   }

   /**
    * Get border reference.
    * 
    * @return  border reference.
    */
   public Border<O> border()
   {
      return border;
   }
   
   /**
    * Set border reference.
    * 
    * @param   border reference
    */
   public void setBorder(Border<O> border)
   {
      this.border = border;
   }

   /**
    * Get panel insets according to current border.
    *  
    * @return  container insets.
    */
   @Override
   public Insets insets()
   {
      if (border != null)
         return border.insets();

      return super.insets();
   }
   
   /**
    * Returns the transparent flag for this panel.
    * 
    * @return  See above.
    */
   public boolean isTransparent()
   {
      return transparent;
   }

   /**
    * Sets the transparent flag. When <code>true</code>, the panel
    * doesn't draw its background.
    * 
    * @param   value
    *          The new value.
    */
   public void setTransparent(boolean value)
   {
      transparent = value;
   }
}