/*
** Module   : TitledWindow.java
** Abstract : Window with title.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 SIY 20100822 Added javadoc.
** 004 SIY 20101122 Switching from KeyEvent to KeyInput.
** 005 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
**                  Added processing of mouse events.
** 006 HC  20140911 GUI coordinate support - data types refactoring.
** 007 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 008 MAG 20141006 Fix NPE on mouse events.
** 009 CA  20141021 Added API to check if the window can raise MOUSE events.  In some cases (like
**                  when the window is being moved/resized/drawing), capturing of mouse events 
**                  needs to be disabled. 
** 010 HC  20141021 Implemented window move/position support.
** 011 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 012 CA  20141123 Fixed processing of window gained/lost focus events.
** 013 CA  20141205 Fixed a problem when processing mouse events: go up the hierarchy until a 
**                  parent with a non-null config is found (to avoid virtual widgets/containers).
** 014 HC  20150323 Removed business logic from config classes.
** 015 HC  20150612 Implemented support for modal windows.
** 016 HC  20150622 Fixed GUI frame to be properly centered in its parent window.
** 017 HC  20150626 Fixed calculation of workspace size in redirected mode.
** 018 HC  20150702 Improved positioning of top-level windows on systems with multiple displays.
**                  Improved parent-child relationship of top-level windows. Implemented window
**                  modality independently on the actual UI driver.
** 019 HC  20150831 Widgets are automatically detached from their parents when destroyed.
** 020 CA  20150910 Removed handleMouseEvents, as the window must collect all mouse events, 
**                  regardless of what the current thread is doing.
** 021 IAS 20150926 MOVE-TO-TOP()/MOVE-TO-BOTTOM() support
** 022 CA  20151006 Changes to allow multi-window focus management in GUI.
** 023 HC  20160104 Implemented toString().
** 024 EVL 20160124 Adding method to detect if the window is overlay type.
** 025 HC  20160406 Removed focus request from draw() method.
** 026 CA  20160915 Added destroyWidgets(), to allow subclasses to selectively destroy the
**                  children.
** 027 HC  20160922 Prevented an NPE.
** 028 OM  20170222 Improved performance by caching in local variable.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui.client.widget;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.util.*;

/**
 * Simple window with title.
 */
public class TitledWindow<O extends OutputManager<?>>
extends AbstractContainer<O>
{
   /** Window title. */
   private String title;

   /**
    * Constructor.
    * 
    * @param   title
    *          Window title.
    * @param   insets
    *          Window insets.
    */
   public TitledWindow(String title, Insets insets)
   {
      setLayout(new BorderLayout<O>());
      forceVisibility(false);
      setTitle(title);
      setInsets(insets);
   }

   /**
    * Check if widget is displayed.
    * 
    * @return  <code>true</code> if widget is displayed.
    */
   @Override
   public boolean isDisplayed()
   {
      return WindowManager.isDisplayed(this);
   }

   /**
    * Check if widget is focus traversable.
    * 
    * @return  <code>true</code> if widget is focus traversable.
    */
   @Override
   public boolean focusTraversable()
   {
      return false;
   }

   /**
    * Check if widget has focus.
    * 
    * @return  <code>true</code> if widget has focus.
    */
   @Override
   public boolean hasFocus()
   {
      return true;
   }

   /**
    * Hide widget.
    */
   @Override
   public void hide()
   {
      if (!isVisible())
         return;

      repaint();
      
      // the window can not be removed here.  the fact that is hidden does not mean that the 
      // window no longer exists.  for it to be removed, it must be destroyed
      
      forceVisibility(false);
   }

   /**
    * Process event and dispatch it to specific method.
    * 
    * @param   event
    *          Event to process.
    */
   @Override
   public void processEvent(Event event)
   {
      @SuppressWarnings("unchecked")
      Widget<O> source = event.source();

      if (event instanceof PaintEvent)
      {
         // prevent calling isTotallyObscured() when source is not displayed,
         // isTotallyObscured() could fail in this case
         if (!source.isDisplayed() || !source.isTotallyObscured())
         {
            // accumulating invalidation rectangle or ignoring the event
            // invalidation rectangle is accumulated in screen coordinates
            if (screen().isInvalidate())
               clipWith(((PaintEvent) event).getUpdateRect());
         }
         return;
      }
      
      if (event instanceof SyncEvent)
      {
         sync();
         return;
      }

      if (event instanceof ScrollEvent)
      {
         @SuppressWarnings("unchecked")
         ScrollPane<O> pane = source.parent(ScrollPane.class);
         
         // pane can be null in case the scroll pane has been discarded,
         // this happens in some rare cases when combo box drop down
         // is closed on focus lost
         if (pane != null)
         {
            pane.handleScrollEvent((ScrollEvent) event);
         }
         
         return;
      }

      if (source == this)
      {
         if (event instanceof FocusEvent)
         {
            processFocusEvent((FocusEvent) event);
         }
         
         return;
      }
      
      if (source != null)
      {
         source.processEvent(event);
      }
   }

   /**
    * Find the widget to which the given mouse event can be applied.
    * 
    * @param    evt
    *           The mouse event.
    * 
    * @return   A widget over which the mouse is positioned or {@code null} if none found.
    *           By default, this method returns {@code null}.
    */
   public Widget<O> findMouseSource(MouseEvt evt)
   {
      // no-op by default
      return null;
   }
   
   /**
    * Process key event.
    * 
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      processKeyListeners(event);

      if (event.source() != null)
         event.source().processKeyEvent(event);
   }

   /**
    * Get absolute location of widget in screen coordinates.
    * 
    * @return  absolute widget location.
    */
   @Override
   public Point screenLocation()
   {
      return location();
   }

   /**
    * Set window title.
    * 
    * @param   title
    *          Title to set.
    */
   public void setTitle(String title)
   {
      this.title = title;
   }

   /**
    * Show widget.
    */
   @Override
   public void show()
   {
      if (isVisible())
      {
         return;
      }
      
      forceVisibility(true);
      WindowManager.addWindow(this);
      
      // all events triggered during relayout must be ignored
      ThinClient.getInstance().eventBracket(true, this::doLayout);
      
      repaint();
      
      // TODO: Remove fixCol/fixRow and push the logic down to
      //       the window types that require it. Top level windows
      //       certainly don't need this kind of location fix, they
      //       can be positioned outside of visible display bounds.
      //       Also top level windows operate in virtual screen coordinates
      //       but fixCol/fixRow expect coordinates relative to the parent.
      Point loc = location();
      if (this instanceof TopLevelWindow)
      {
         setLocation(loc.x, loc.y);
      }
      else
      {
         setLocation(fixCol(loc.x), fixRow(loc.y));
      }
      
      sync();
   }

   /**
    * Get window title.
    *  
    * @return  window title.
    */
   public String title()
   {
      return title;
   }

   /**
    * Check if window has non-empty title.
    * 
    * @return  <code>true</code> if window has title.
    */
   public boolean hasTitle()
   {
      return title() != null && title().length() > 0;
   }
   
   /**
    * Check if window has title bar.  This should return true for GUI window even if the
    * title length is zero.  For ChUI code returns false.
    * 
    * @return  <code>true</code> if window has title to repaint.
    */
   public boolean hasTitlebar()
   {
      return false;
   }
   
   /**
    * Get the size of the window workspace area. This default implementation
    * returns the window dimension.
    * 
    * @return  See above.
    */
   public Dimension getWorkspaceSize()
   {
      return screenDimension();
   }
   
   /**
    * The method returns <code>true</code> if this represents a modal window. 
    * That is a window that sits on top of other windows and takes exclusive user 
    * input.
    * <p>
    * This implementation always returns <code>false</code>.
    * 
    * @return  See above.
    */
   public boolean isModal()
   {
      return false;
   }
   
   /**
    * Fix column to make widget fit into screen.
    * 
    * @param   x
    *          Source column.
    *          
    * @return  fixed column.
    */
   private double fixCol(double x)
   {
      int wndId = topLevelWindow().getId().asInt();
      int displayWidth = ThinClient.getInstance().getDisplayWidth(wndId);
      double displayColumns = screen().coordinates().widthFromNative(displayWidth); 
      
      double deltaX = Coordinate.scale(x + width() - displayColumns);
      
      // safe, comparison of scaled doubles
      if (deltaX > 0)
         x = Coordinate.scale(x - deltaX);

      if (x < 0)
         x = 0;
      
      return x;
   }

   /**
    * Fix row to make widget fit into screen.
    * 
    * @param   y
    *          Source row.
    *          
    * @return  fixed row.
    */
   private double fixRow(double y)
   {
      int wndId = topLevelWindow().getId().asInt();
      int displayHeight = ThinClient.getInstance().getDisplayHeight(wndId);

      double deltaY = Coordinate.scale(y + height() - displayHeight);
      
      // safe, comparison of scaled doubles
      if (deltaY > 0)
         y = Coordinate.scale(y - deltaY);

      if (y < 0)
         y = 0;
      
      return y;
   }

   
   /**
    * Destroy this widget.  Notify all of its components that the widget has been destroyed.
    */
   @Override
   public void destroy()
   {
      destroyWidgets();
      
      super.destroy();
   }

   /**
    * Get the Z-Order class
    * 
    * @return Z-Order class of the widget
    */
   @Override
   public ZOrderClass getZOrderClass()
   {
      return ZOrderClass.WINDOW;
   }
   
   /**
    * Returns a string representation of the object.
    * 
    * @return   A string representation of the object.
    */
   @Override
   public String toString()
   {
      return getClass().getSimpleName() + " [id=" + Utils.nonNull(getId(), "<null>") +
                                          ", title=" + Utils.nonNull(title, "") + "]";
   }
   
   /**
    * Destroy all the widgtets in this window.
    */
   protected void destroyWidgets()
   {
      screen().getRegistry().removeWidgetsRecursive((Widget) this);
   }
}