/*
** Module   : ArrayAssigner.java
** Abstract : automates the assignment of each element of a BDT array
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description----------------------------
** 001 GES 20051104   @23229 Provides a simple helper to automate the assignment of each element
**                           of a BaseDataType array based on an array input.
** 002 GES 20060410   @25444 All elements can be set to the same scalar value (new signature).
** 003 GES 20060412   @25486 Added a setUnknown() method.
** 004 GES 20060425   @25707 Added a new assign() signature.
** 005 ECF 20060428   @25851 Added safety code. If target array element is null, duplicate
**                           corresponding source to initialize it.
** 006 GES 20080911   @39793 Added signatures to allow primitives to be assigned from.
** 007 GES 20090625   @42956 Fix and helpers to avoid ArrayIndexOutOfBounds. This implements a
**                           new delegation model to hide all array variable dereferencing (for
**                           assignment or simple access). The helpers can thus translate or
**                           avoid Java level AIOOB exceptions and behave the way that the 4GL
**                           requires.
** 008 CA  20130209          Added support for the EXTENT statement and output EXTENT parameters.
** 009 CA  20130315          Added missing assignSingle and assignMulti APIs, for to the cases
**                           when the rvalue is a resource, as in some cases the :HANDLE attribute
**                           conversion emits the WrappedResource instance instead of handle.
** 010 OM  20130418          Switched to long/int64 to support 64-bit wide indexes and values. 
** 011 CA  20131013          Added no-op scopeDeleted() method, required by the changes in 
**                           Scopeable interface.
** 012 GES 20140116          Made isDynamicArray() public.
** 013 HC  20140324          Fixed extent-to-extent assignment.
** 014 HC  20140413          Fixed extent parameter initialization.
** 015 HC  20140430          Fixed decimal precision for extent variables. 
** 016 HC  20140613          Improved extent parameter support - removed some of 
**                           the code being emitted and introduced multiple wrapper classes 
**                           depending on the parameter type and the l-value type (variable vs.
**                           field).
** 017 CA  20151214          When resizing a shared extent variable, update the reference in the
**                           SharedVariableManager registries.
** 018 OM  20160706          Small fix in subscriptWorker() method.
** 019 CA  20160728          Fixed undoable support for dynamic-extent arrays: in this case, when
**                           the array has its extent set, all the elements must register for undo
**                           depending on the scope where the array was defined.
** 020 CA  20160812          Performance improvements for H019: use IdentityHashMap when using
**                           ext prog instances or BDT[] as keys, pass the tx level when 
**                           registering dynamic extent arrays.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.security.*;
import java.lang.reflect.*;
import java.util.*;

/**
 * Provides a simple helper to automate the assignment of each element of a 
 * {@link BaseDataType} array based on 2 array inputs.
 */
public class ArrayAssigner
{
   /** Context-local data container. */
   private static ContextLocal<WorkArea> local = new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      {
         WorkArea wa = new WorkArea();
         TransactionManager.registerScopeable(wa);
         return wa;
      };
   };

   /**
    * Ensure all the context-local data has initialized (to be able to listen for scope 
    * notifications).
    */
   static void load()
   {
      // needed to ensure the variable is loaded, so it listens for scope notifications
      local.get();
   }

   /**
    * Assigns (mutates the content of) each matching element of a target
    * array based on the corresponding element of the source array.  The
    * source and target arrays must be assignable to each other (must be
    * assignment-compatible). The copy operation will stop as soon as the end
    * of either array is encountered. If either array is longer than the other,
    * then the copy will only occur for the overlapping portion. The copy
    * always starts at element 0 and finishes at the lesser of source length or
    * target length.
    * <p>
    * In case the assignment target represents an indeterminate extent variable
    * and the source is a determinate extent or a fixed indeterminate extent
    * the target is reallocated to the size of the source extent. The allocated 
    * reference is returned as the result from the function.
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    source
    *           The array from which to assign. Must not be <code>null</code>.
    *           
    * @return   The reference to the target array or a reference to a newly
    *           allocated array representing the assignment target.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static BaseDataType[] assignMulti(BaseDataType[] target, BaseDataType[] source) 
   throws ErrorConditionException,
          NullPointerException
   {
      boolean dynamicTarget = isDynamicArray(target);

      if (dynamicTarget     && 
          target.length > 0 && 
          (source.length == 0 ||
           source.length != target.length))
      {
         final String err = "Indeterminate extent is already fixed to a dimension of %d";
         String errMsg = String.format(err, target.length);
         ErrorManager.recordOrThrowError(13738, errMsg, false);
         
         return target;
      }
      if (!dynamicTarget)
      {
         if (source.length == 0)
         {
            final String errMsg = "Uninitialized array used as source of assignment";
            ErrorManager.recordOrThrowError(14906, errMsg, false);
            
            return target;
         }
         else if (source.length != target.length)
         {
            final String errMsg = 
               "Whole-array assignment target and source must have the same extent unless " +
               "the target is indeterminate";
            ErrorManager.recordOrThrowError(14905, errMsg, false);
            
            return target;
         }
      }
      
      if (target.length == 0 && source.length > 0)
      {
         target = resize(target, source.length);
      }

      for (int i = 0; (i < target.length && i < source.length); i++)
      {
         if (target[i] == null)
         {
            target[i] = source[i].duplicate();
         }
         else
         {
            (target[i]).assign(source[i]);
         }
      }
      
      return target;
   }
   
   /**
    * Assigns (mutates the content of) all elements of a target array from
    * the same source value.  The source and target types must be
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, long source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, new int64(source));
   }   
   
   /**
    * Assigns (mutates the content of) all elements of a target array from
    * the same source value.  The source and target types must be
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, double source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, new decimal(source));
   }   
   
   /**
    * Assigns (mutates the content of) all elements of a target array from
    * the same source value.  The source and target types must be
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, String source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, new character(source));
   }   
   
   /**
    * Assigns (mutates the content of) all elements of a target array from
    * the same source value.  The source and target types must be
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, boolean source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, new logical(source));
   }   
   
   /**
    * Assigns (mutates the content of) all elements of a target array from
    * the same source value.  The source and target types must be
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, BaseDataType source) 
   throws ErrorConditionException,
          NullPointerException
   {
      if (target.length == 0 && TransactionManager.blockInitialized())
      {
         // TODO: for some reason, this error message doesn't end in a dot in 4GL
         String msg = "Invalid assignment to an unfixed indeterminate extent";
         ErrorManager.recordOrThrowError(11389, msg, false);
         
         return;
      }
      
      for (int i = 0; i < target.length; i++)
      {
         if (target[i] == null)
         {
            target[i] = source.duplicate();
         }
         else
         {
            (target[i]).assign(source);
         }
      }
   }   
   
   /**
    * Assigns (mutates the content of) all elements of a target array from
    * the same source value.  The source and target types must be
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, WrappedResource source) 
   throws ErrorConditionException,
          NullPointerException
   {
      if (target.length == 0 && TransactionManager.blockInitialized())
      {
         // TODO: for some reason, this error message doesn't end in a dot in 4GL
         String msg = "Invalid assignment to an unfixed indeterminate extent";
         ErrorManager.recordOrThrowError(11389, msg, false);
         
         return;
      }
      
      for (int i = 0; i < target.length; i++)
      {
         if (target[i] == null)
         {
            target[i] = new handle(source);
         }
         else
         {
            (target[i]).assign(new handle(source));
         }
      }
   }   
   
   /**
    * Assigns (mutates the content of) all elements of a target array starting
    * at the given index through the end of the array, using the same source
    * value.  The source and target types must be assignable to each other 
    * (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    idx
    *           Starting index (0-based) for the assignments.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, int idx, long source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, idx, new int64(source));
   }
   
   /**
    * Assigns (mutates the content of) all elements of a target array starting
    * at the given index through the end of the array, using the same source
    * value.  The source and target types must be assignable to each other 
    * (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    idx
    *           Starting index (0-based) for the assignments.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, int idx, double source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, idx, new decimal(source));
   }
   
   /**
    * Assigns (mutates the content of) all elements of a target array starting
    * at the given index through the end of the array, using the same source
    * value.  The source and target types must be assignable to each other 
    * (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    idx
    *           Starting index (0-based) for the assignments.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, int idx, String source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, idx, new character(source));
   }
   
   /**
    * Assigns (mutates the content of) all elements of a target array starting
    * at the given index through the end of the array, using the same source
    * value.  The source and target types must be assignable to each other 
    * (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    idx
    *           Starting index (0-based) for the assignments.
    * @param    source
    *           The value from which to assign.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, int idx, boolean source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignMulti(target, idx, new logical(source));
   }
   
   /**           
    * Assigns (mutates the content of) all elements of a target array starting
    * at the given index through the end of the array, using the same source
    * value.  The source and target types must be assignable to each other 
    * (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    idx
    *           Starting index (0-based) for the assignments.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[] target, int idx, BaseDataType source)
   throws ErrorConditionException,
          NullPointerException
   {
      for (int i = idx; i < target.length; i++)
      {
         if (target[i] == null)
         {
            target[i] = source.duplicate();
         }
         else
         {
            (target[i]).assign(source);
         }
      }
   }   
   
   /**           
    * Assigns (mutates the content of) all elements of a target array starting
    * at the given index through the end of the array, using the same source
    * value.  The source and target types must be assignable to each other 
    * (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    idx
    *           Starting index (0-based) for the assignments.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure of an individual assignment.
    * @throws   NullPointerException
    *           If either the source or the target arrays are 
    *           <code>null</code>.
    */
   public static void assignMulti(BaseDataType[]  target, int idx, WrappedResource source) 
   throws ErrorConditionException,
          NullPointerException
   {
      for (int i = idx; i < target.length; i++)
      {
         if (target[i] == null)
         {
            target[i] = new handle(source);
         }
         else
         {
            (target[i]).assign(new handle(source));
         }
      }
   }   
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, long index, long source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, new int64(index), new int64(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, long index, double source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, new int64(index), new decimal(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, long index, boolean source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, new int64(index), new logical(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, long index, String source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, new int64(index), new character(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, long index, BaseDataType source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, new int64(index), source);
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, long index, WrappedResource source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, new int64(index), new handle(source));
   }
      
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, NumberType index, long source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, index, new int64(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, NumberType index, double source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, index, new decimal(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, NumberType index, boolean source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, index, new logical(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, NumberType index, String source) 
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingle(target, index, new character(source));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, NumberType index, BaseDataType source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingleWorker(target, index, source);
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingle(BaseDataType[] target, NumberType index, WrappedResource source)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingleWorker(target, index, new handle(source));
   }
   
   /**
    * Dereference the given array using the given 1-based index.
    *
    * @param    array
    *           The array to dereference.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    */
   public static <B extends BaseDataType> B subscript(B[] array, long index)
   {
      int idx = subscriptHelper(array, index);
      return subscriptWorker(array, idx);
   }
   
   /**
    * Dereference the given array using the given 1-based index.
    *
    * @param    array
    *           The array to dereference.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    */
   public static <B extends BaseDataType> B subscript(B[] array, NumberType index)
   {
      int idx = subscriptHelper(array, index, false);
      return subscriptWorker(array, idx);
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the unknown value.
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingleUnknown(BaseDataType[] target, long index)
   throws ErrorConditionException,
          NullPointerException
   {
      assignSingleUnknown(target, new int64(index));
   }
   
   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the unknown value.
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   public static void assignSingleUnknown(BaseDataType[] target, NumberType index)
   throws ErrorConditionException,
          NullPointerException
   {
      Class<? extends BaseDataType> cls = 
         (Class<? extends BaseDataType>) target.getClass().getComponentType();

      assignSingleWorker(target, index, BaseDataType.generateUnknown(cls));
   }
   
   /**                     
    * Sets all elements of a target array to the <code>unknown value</code>.
    *
    * @param    target
    *           The array to be modified. Must not be <code>null</code>.
    */
   public static void assignUnknown(BaseDataType[] target)
   {
      if (target.length == 0 && TransactionManager.blockInitialized())
      {
         // TODO: for some reason, this error message doesn't end in a dot in 4GL
         String msg = "Invalid assignment to an unfixed indeterminate extent";
         ErrorManager.recordOrThrowError(11389, msg, false);
         
         return;
      }
      
      for (int i = 0; i < target.length; i++)
      {
         target[i].setUnknown();
      }
   }
   
   /**
    * Implement the EXTENT statement. This resizes the given array to the new size and returns
    * the new instance.
    * <p>
    * If the given array is not of size zero, then an error is raised.  Else, all the elements
    * of the array will be set to default value.
    * 
    * @param    array
    *           The dynamic-extent array to be resized.
    * @param    newSize
    *           The array's new size.
    * 
    * @return   The resized array.
    */
   public static <T extends BaseDataType> T[] resize(T[] array, int newSize)
   {
      return resize(array, new integer(newSize), false);
   }
   
   /**
    * Implement the EXTENT statement. This resizes the given array to the new size and returns
    * the new instance.
    * <p>
    * If the given array is not of size zero, then an error is raised.  Else, all the elements
    * of the array will be set to default value.
    * 
    * @param    array
    *           The dynamic-extent array to be resized.
    * @param    newSize
    *           The array's new size.
    * 
    * @return   The resized array.
    */
   public static <T extends BaseDataType> T[] resize(T[] array, int64 newSize)
   {
      return resize(array, newSize, false);
   }

   /**
    * Implement the EXTENT statement. This resizes the given array to the new size and returns
    * the new instance.
    * <p>
    * If the given array is not of size zero, then an error is raised.  Else, all the elements
    * of the array will be set to unknown or default value depending on the <code>initToUnknown</code> 
    * parameter.
    * 
    * @param    array
    *           The dynamic-extent array to be resized.
    * @param    newSize
    *           The array's new size.
    * @param    initToUnknown
    *           Initialization flag, see above for more info.
    * 
    * @return   The resized array.
    */
   public static <T extends BaseDataType> T[] resize(T[] array, int newSize, boolean initToUnknown)
   {
      return resize(array, new integer(newSize), initToUnknown);
   }

   /**
    * Implement the EXTENT statement. This resizes the given array to the new size and returns
    * the new instance.
    * <p>
    * If the given array is not of size zero (and the newSize is not unknown), then an error is 
    * raised. Else, the array will be resized and all the elements will be set to unknown or 
    * the default value depending on the <code>initToUnknown</code> parameter.
    * <p>
    * The new extent value must be from the [1..28000] interval.
    * 
    * @param    array
    *           The dynamic-extent array to be resized. Must not be null. If the array was not
    *           resized yet, must be 0-length.
    * @param    newSize
    *           The array's new size.
    * @param    initToUnknown
    *           Initialization flag, see above for more info.
    * 
    * @return   The resized array.
    */
   @SuppressWarnings("unchecked")
   public static <T extends BaseDataType> T[] resize(T[] array, int64 newSize, boolean initToUnknown)
   {
      long size = newSize.isUnknown() ? 0 : newSize.longValue();
      if (!newSize.isUnknown())
      {
         if (array.length != 0)
         {
            final String msg = "Indeterminate extent is already fixed to a dimension of %d";
            ErrorManager.recordOrThrowError(13738, String.format(msg, array.length), false);
            return array;
         }
   
         if (size <= 0 || size > 28000)
         {
            //size < 1:  Invalid value specified for the EXTENT statement. (13737)
            //size > 28000: Array extent may not be greater than 28000 (3350)
            final String msg = "Indeterminate extent dimension %d is out of range 1 to 28000";
            ErrorManager.recordOrThrowError(11385, String.format(msg, size), false);
            return array;
         }
      }
      
      Class<? extends BaseDataType> cls = 
         (Class<? extends BaseDataType>) array.getClass().getComponentType();
      T[] copy = (T[]) Array.newInstance(cls, (int) size);
      for (int i = 0; i < copy.length; i++)
      {
         if (initToUnknown)
         {
            copy[i] = (T) BaseDataType.generateUnknown(cls);
         }
         else
         {
            copy[i] = (T) BaseDataType.generateDefault(cls);
         }
      }
      
      // re-register it - replace the old array with the new one
      WorkArea wa = local.get();
      wa.reRegister(array, copy);
      
      // preserve precision of the source array
      copyDecimalPrecision(array, copy);
      
      // if this array belongs to an extent parameter, update it
      AbstractExtentParameter<BaseDataType> extentParam = wa.getExtentParameter(array);
      if (extentParam != null)
      {
         extentParam.setParameter(copy);
      }
      
      // if shared, update the shared reference (global or local)
      SharedVariableManager.updateExtentVar(array, copy);
      
      // if this is an undoable extent var, we need to register all instances as undoable!
      Object referent = ProcedureManager._thisProcedure();
      Deque<Map<BaseDataType[], Integer>> undoableExts = wa.dynamicUndoableArrays.get(referent);
      Iterator<Map<BaseDataType[], Integer>> iter = undoableExts.iterator();
      while (iter.hasNext())
      {
         Map<BaseDataType[], Integer> undoables = iter.next();
         
         Integer transLevel = undoables.remove(array);

         if (transLevel != null)
         {
            // if this is the ext program scope and the proc is ran persistent
            boolean persistent = !iter.hasNext() && ProcedureManager._isPersistent(referent);
            
            //  the entry must be registered at the scope where the extent was defined
            TransactionManager.registerAt(transLevel, persistent, copy);
            
            break;
         }
      }
      
      return copy;
   }

   /**
    * Get a copy of the given array.
    * 
    * @param    array
    *           The array for which a copy is needed.
    * 
    * @return   The copy.
    */
   public static <T extends BaseDataType> T[] copyOf(T[] array)
   {
      Class<? extends BaseDataType> cls = 
               (Class<? extends BaseDataType>) array.getClass().getComponentType();
   
      T[] copy = (T[]) Array.newInstance(cls, array.length);
      for (int i = 0; i < copy.length; i++)
      {
         copy[i] = (T) BaseDataType.generateUnknown(cls);
         copy[i].assign(array[i]);
      }
      
      WorkArea wa = local.get();
      if (wa.isRegistered(array))
      {
         // if the original array is registered as a dynamic array
         // register the copy as well.
         // for example, when assigning the output parameter back to its 
         // original reference, a copy is made
         wa.register(copy);
      }
      
      // preserve precision of the source array
      copyDecimalPrecision(array, copy);
      
      return copy;
   }

   /**
    * Get the length of the given dynamic-extent variable.
    * 
    * @param    array
    *           The dynamic-extent variable.
    *           
    * @return   If the array's length is zero, return unknown. Else, the actual size.
    */
   public static integer lengthOf(BaseDataType[] array)
   {
      return array.length == 0 ? new integer() : new integer(array.length);
   }

   /**
    * Register the given array as a dynamic-array.
    * 
    * @param    array
    *           The dynamic-array to register.
    */
   public static void registerDynamicArray(BaseDataType[] array)
   {
      local.get().register(array);
   }

   /**
    * Register the set of given arrays as dynamic extent.
    * 
    * @param    arrays
    *           The set of dynamic arrays to register.
    * @param    transLevel
    *           The transaction nesting level where this instance was created.
    */
   public static void registerDynamicUndoableArray(Set<BaseDataType[]> arrays, int transLevel)
   {
      local.get().registerDynamicUndoable(arrays, transLevel);
   }
   
   /**
    * This method returns true for any real dynamic-extent array (regardless if it had its
    * size set).
    * 
    * @param    array
    *           The array to be checked.
    *           
    * @return   See above.
    */
   public static boolean isDynamicArray(BaseDataType[] array)
   {
      return local.get().isRegistered(array);
   }
   
    /**
     * Sets decimal precision for the array argument.
     * 
     * @param     array
     *            The array to be set with the supplied precision.
     * @param     precision
     *            Decimal precision to be set in the supplied array.
     */
   public static void setDecimalPrecision(decimal[] array, int precision)
   {
      if (array.length == 0)
      {
         // for dynamic extents, precision is stored in the context
         WorkArea wa = local.get();
         wa.setDecimalPrecision(array, precision);
      }
      else
      {
         // precision is stored in the components for non-dynamic extents
         for (decimal d : array)
         {
            d.setPrecision(precision);
         }
      }
   }
   
   /**
    * Returns decimal precision of the supplied array. If the array represents a
    * dynamic-extent, the precision must have been previously set, otherwise <code>null</code> is
    * returned.
    * 
    * @param    array
    *           Array representing an extent variable.
    * 
    * @return   The decimal precision value or <code>null</code>, see above.      
    */
   public static Integer getDecimalPrecision(decimal[] array)
   {
      if (array.length == 0)
      {
         // for dynamic extents, precision is stored in the context
         WorkArea wa = local.get();
         return wa.getDecimalPrecision(array);
      }
      else 
      {
         // precision is stored in the components for non-dynamic extents
         return array[0].getPrecision();
      }
   }
   
   /**
    * Registers an unfixed indeterminate extent parameter instance.
    * The registration is required to keep the extent parameter instance
    * up-to-date in case its parameter reference changes (for example
    * when the parameter value is resized by the EXTENT statement).
    * 
    * @param   extentParameter
    *          An instance representing an input or input-output extent parameter. 
    * 
    * @throws  IllegalArgumentException
    *          When the passed in instance does not represent an unfixed 
    *          indeterminate extent.
    */
   static <T extends BaseDataType> void registerExtentParameter(
         AbstractExtentParameter<T> extentParameter)
   {
      T[] param = extentParameter.getParameter();
      if (param.length > 0)
      {
         throw new IllegalArgumentException(
               "Unfixed indeterminate extent for extentParameter argument is expected.");
      }
      
      WorkArea wa = local.get();
      wa.registerExtentParameter(extentParameter);
   }

   /**
    * Dereference the given array using the given 0-based index.
    *
    * @param    array
    *           The array to dereference.
    * @param    idx
    *           The 0-based index into the array. No bounds checking is done.
    */
   private static <B extends BaseDataType> B subscriptWorker(B[] array, int idx)
   {
      if (idx == -2)
      {
         // this case means the index is unknown, so we need to return unknown
         return (B) BaseDataType.generateUnknown(array.getClass().getComponentType());
      }
      
      if (idx == -1)
      {
         // if an invalid index was passed and [subscriptHelper()] returned -1 then the 
         // [ErrorManager.isPendingError()] is already set to true anyway.
         // However, if [ErrorManager.isPendingError()] was set because ANY OTHER reason and [idx]
         // is valid, then we need to return the correct dereference, array[idx].
         Class<?> cls = array.getClass().getComponentType();
         
         // silently return unknown value (safest thing)
         return (B) BaseDataType.generateUnknown(cls);
      }
      
      return array[idx];
   }
   
   /**
    * Generate a valid 0-based integral subscript from the 1-based numeric 
    * value given. Bounds checking is done based on the target array length.
    *
    * @param    array
    *           The array to dereference.
    * @param    _idx
    *           The 1-based index to be checked and modified.
    * @param    assign
    *           <code>true</code> if the extent var in on the left-side of the assign.
    *
    * @return   The 0-based integral subscript, -2 if the index is the
    *           <code>unknown value</code> or -1 if the index is out of bounds
    *           in silent error mode.
    *
    * @throws   ErrorConditionException
    *           If the index is out of bounds.
    */
   private static int subscriptHelper(BaseDataType[] array, NumberType _idx, boolean assign)
   throws ErrorConditionException
   { 
      if (_idx.isUnknown())
      {
         // no error is generated if the index is unknown, but unknown needs to be returned
         return -2;
      }

      // 4GL lets the the 33 MSB overflow, computing the array index using last 31 bits
      int idx = (int)(_idx.longValue() & 0x7FFFFFFFL);
      if (idx < 0)
      {
         String spec   = "Array subscript %d is out of range";
         String errmsg = String.format(spec, idx);
         
         ErrorManager.recordOrThrowError(26, errmsg);
         return idx - 1;
      }

      int len = array.length;
      if (len == 0)
      {
         // dynamic-extent, not initialized
         String msg = "Invalid runtime use of indeterminate extent.  " +
                      "It must be set to a fixed dimension";
         
         // TODO: if we are in message mode, it seems like the error is shown, but not thrown
         ErrorManager.recordOrThrowError(11387, msg, false);

         // we need to return -1
         return -1;
      }
      else if (idx <= 0 || idx - 1 >= len)
      {
         // error out!
         if (isDynamicArray(array) && (!assign || idx - 1 >= len))
         {
            String spec = 
               "Array subscript %d is out of range.  The indeterminate extent is fixed to a " +
               "dimension of %d";
            String errmsg = String.format(spec, idx, len);
            
            ErrorManager.recordOrThrowError(11388, errmsg, false);
         }
         else
         {
            String spec   = "Array subscript %d is out of range";
            String errmsg = String.format(spec, idx);
            
            ErrorManager.recordOrThrowError(26, errmsg);
         }

         // we need to return -1
         return -1;
      }
      
      return idx - 1;
   }
   
   /**
    * Generate a valid 0-based integral subscript from the 1-based numeric 
    * value given. Bounds checking is done based on the target array length.
    *
    * @param    array
    *           The array to dereference.
    * @param    idx
    *           The 1-based index to be checked and modified.
    *
    * @return   The 0-based integral subscript or -1 if the index is out
    *           of bounds in silent error mode.
    *
    * @throws   ErrorConditionException
    *           If the index is out of bounds.
    */
   private static int subscriptHelper(BaseDataType[] array, long idx)
   throws ErrorConditionException
   {
      return subscriptHelper(array, new int64(idx), false);
   }

   /**
    * Assigns (or mutates the content of) the indexed element of a target array
    * to the given source value.  The source and target types must be 
    * assignable to each other (must be assignment-compatible).
    *
    * @param    target
    *           The array to be assigned. Must not be <code>null</code>.
    * @param    index
    *           The 1-based index into the array. This will be decremented by
    *           1 and then used if it is in bounds.
    * @param    source
    *           The value from which to assign. Must not be <code>null</code>.
    *
    * @throws   ErrorConditionException
    *           On any failure in the assignment or array index out of bounds.
    * @throws   NullPointerException
    *           If either the source or the target arrays are <code>null</code>.
    */
   private static void assignSingleWorker(BaseDataType[] target,
                                          NumberType     index,
                                          BaseDataType   source)
   {

      boolean unknown = index.isUnknown();
      boolean dynamic = isDynamicArray(target);
      if (!unknown && target.length == 0)
      {
         // dynamic-extent, not initialized, and tried to assign an element
         if (index.intValue() <= 0)
         {
            String spec   = "Array subscript %d is out of range";
            String errmsg = String.format(spec, index.intValue());
            
            ErrorManager.recordOrThrowError(26, errmsg);
         }
         else
         {
            // TODO: for some reason, this error message doesn't end in a dot in 4GL
            String msg = "Invalid assignment to an unfixed indeterminate extent";
            ErrorManager.recordOrThrowError(11389, msg, false);
         }
         return;
      }
      
      int idx = -1;
      
      try
      {
         idx = subscriptHelper(target, index, true);
         if (idx == -2)
         {
            // in case the index is unknown, show only this message
            ErrorManager.recordOrThrowError(26, "Array subscript 0 is out of range");
            return;
         }
      }
      
      catch (ErrorConditionException ece)
      {
         // re-throw the error if in dynamic mode or if the index is less than zero
         if (dynamic || (!index.isUnknown() && index.intValue() < 0))
            throw ece;
         
         // eat the first error, we will rethrow shortly
      }
      
      if (idx == -1 || ErrorManager.isPendingError())
      {
         if (!dynamic)
         {
            // only non-dynamic arrays show this error (double space before Field is intented)
            ErrorManager.recordOrThrowError(142, "Unable to update  Field");
         }
         
         // silently return
         return;
      }
      
      if (target[idx] == null)
      {
         target[idx] = source.duplicate();
      }
      else
      {
         (target[idx]).assign(source);
      }
   }
   
   /**
    * Copies decimal precision from <code>source</code> to <code>destination</code> if 
    * decimal array is passed in. For non-decimal arrays, this method is effectively a no-op.
    * 
    * @param    source
    *           An array instance, the source of decimal precision.
    * @param    destination
    *           An array instance, the destination for decimal precision.
    * 
    * @throws   IllegalArgumentException
    *           When component types of the source and destination array are not equal.
    */
   private static void copyDecimalPrecision(BaseDataType[] source, BaseDataType[] destination)
   {
      Class<?> sourceComponentClass = source.getClass().getComponentType();
      Class<?> destinationComponentClass = destination.getClass().getComponentType();
      
      if (!sourceComponentClass.equals(destinationComponentClass))
      {
         throw new IllegalArgumentException("Source and destination component types must be equal.");
      }
      
      // only for decimal arrays
      if (decimal.class.isAssignableFrom(source.getClass().getComponentType()))
      {
         // copy the precision
         Integer precision = getDecimalPrecision((decimal[]) source);
         if (precision != null)
         {
            setDecimalPrecision((decimal[]) destination, precision);
         }
      }
   }

   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   implements Scopeable
   {
      /** Stack of dynamic-extent arrays, per each scope.*/
      private Deque<Set<BaseDataType[]>> dynamicArrays = new ArrayDeque<Set<BaseDataType[]>>();
      
      /** Stack of undoable dynamic-extent arrays, per each scope and each external program.*/
      private Map<Object, Deque<Map<BaseDataType[], Integer>>> 
      dynamicUndoableArrays = new IdentityHashMap<>();

      /** Map of decimal dynamic array precisions. */
      private Map<BaseDataType[], Integer> dynamicArrayPrecisions = new IdentityHashMap<>();
      
      /** Map of extent parameter instances */
      private Deque<Map<BaseDataType[], AbstractExtentParameter<?>>> extentParameters = 
            new ArrayDeque<>();
      
      @Override
      public void scopeStart()
      {
         dynamicArrays.push(new HashSet<>());
         extentParameters.push(new IdentityHashMap<>());
         Object referent = ProcedureManager._thisProcedure();
         Deque<Map<BaseDataType[], Integer>> undoableExtStack = dynamicUndoableArrays.get(referent);
         if (undoableExtStack == null)
         {
            dynamicUndoableArrays.put(referent, undoableExtStack = new ArrayDeque<>());
         }
         undoableExtStack.push(new IdentityHashMap<>());
      }

      @Override
      public void scopeDeleted()
      {
         Object referent = ProcedureManager.getProcessedProcedure();
         if (referent != null)
         {
            dynamicUndoableArrays.remove(referent);
         }
      }

      @Override
      public void scopeFinished()
      {
         if (!dynamicArrays.isEmpty())
         {
            // remove scope-registered arrays
            Set<BaseDataType[]> arrays = dynamicArrays.pop();
            
            // remove precisions and output extent param registrations 
            // for the arrays being removed
            for (BaseDataType[] array : arrays)
            {
               dynamicArrayPrecisions.remove(array);
            }
         }
         
         if (!extentParameters.isEmpty())
         {
            extentParameters.pop();
         }
         
         Object referent = ProcedureManager._thisProcedure();
         Deque<Map<BaseDataType[], Integer>> undoableExtStack = dynamicUndoableArrays.get(referent);
         if (undoableExtStack != null && undoableExtStack.size() > 1)
         {
            // the top-most scope (for the ext program) remains saved until the external program
            // gets deleted
            undoableExtStack.pop();
         }
      }
      
      /**
       * Returns <code>true</code> if the supplied array argument has been
       * previously registered, <code>false</code> otherwise.
       * 
       * @param    array
       *           A valid array instance.
       *        
       * @return   See above.
       */
      public boolean isRegistered(BaseDataType[] array)
      {
          for (Set<BaseDataType[]> scopeArrays : dynamicArrays)
          {
             if (scopeArrays.contains(array))
             {
                return true;
             }
          }
          
          return false;
      }
      
      /**
       * Register the given <code>array</code> as a dynamic-extent array. If the array is already
       * registered the method doesn't register it again and simply returns.
       * 
       * @param    array
       *           The dynamic-extent array reference, must not be <code>null</code>.
       */
      public void register(BaseDataType[] array)
      {
         // register only to one scope
         if (!isRegistered(array))
         {
            dynamicArrays.peek().add(array);
         }
      }
      
      /**
       * Register the set of given arrays as dynamic extent.
       * 
       * @param    arrays
       *           The set of dynamic arrays to register.
       * @param    transLevel
       *           The transaction nesting level where this instance was created.
       */
      public void registerDynamicUndoable(Set<BaseDataType[]> arrays, int transLevel)
      {
         Object referent = ProcedureManager._thisProcedure();
         Map<BaseDataType[], Integer> refs = dynamicUndoableArrays.get(referent).peek();
         for (BaseDataType[] array : arrays)
         {
            refs.put(array, transLevel);
         }
      }

      /**
       * Deregisters <code>oldArray</code> and registers <code>newArray</code> into the scope
       * previously occupied by <code>oldArray</code>.
       * 
       * @param    oldArray
       *           A valid array instance.
       * @param    newArray
       *           A valid array instance.
       */
      public void reRegister(BaseDataType[] oldArray, BaseDataType[] newArray)
      {
         // go through all scopes
         for (Set<BaseDataType[]> scopeArrays : dynamicArrays)
         {
            if (scopeArrays.remove(oldArray))
            {
                // oldArray exists in the scope, replace with newArray
                scopeArrays.add(newArray);
                
                // arrays are registered only in the first-level scope
                break;
            }
         }
      }
      
      /**
       * Retrieve the decimal precision of the supplied array or <code>null</code> if precision
       * was not previously set for the array argument.
       * 
       * @param    array
       *           Array representing an extent variable.
       * 
       * @return   The decimal precision value or <code>null</code>, see above.
       */
      public Integer getDecimalPrecision(decimal[] array)
      {
         return dynamicArrayPrecisions.get(array);
      }
      
      /**
       * Sets decimal precision for the array argument.
       * 
       * @param    array
       *           The array to be set with the supplied precision.
       * @param    precision
       *           Decimal precision to be set for the supplied array.
       */
      public void setDecimalPrecision(decimal[] array, int precision)
      {
         // make sure the array is registered, so it's properly cleaned up
         // on scope unrolling
         // it's actually a valid case precision being set before registration
         register(array);
         dynamicArrayPrecisions.put(array, precision);
      }
      
      /**
       * Registers an unfixed indeterminate extent parameter instance into 
       * the current scope. The registered instance will be automatically unregistered 
       * when the scope is finished. The method will use the return value of
       * {@link AbstractExtentParameter#getParameter()} as the registration key.
       * To get the registered instance back, use that reference in the call to
       * {@link #getExtentParameter(BaseDataType[])}. 
       *  
       * @param    extentParameter
       *           An extent parameter instance. 
       */
      public <T extends BaseDataType> void registerExtentParameter(
            AbstractExtentParameter<T> extentParameter)
      {
         T[] parameter = extentParameter.getParameter();
         extentParameters.peek().put(parameter, extentParameter);
      }
      
      /**
       * Returns a previously registered extent parameter instance, 
       * see {@link #registerExtentParameter(AbstractExtentParameter)} for more info.
       * 
       * @param    array
       *           A valid dynamic array instance.
       * 
       * @return   An extent parameter instance.
       */
      public <T extends BaseDataType> AbstractExtentParameter<T> getExtentParameter(
            BaseDataType[] array)
      {
         // go through all scopes
         for (Map<BaseDataType[], AbstractExtentParameter<?>> scopeParams : extentParameters)
         {
            @SuppressWarnings("unchecked")
            AbstractExtentParameter<T> extentParam = (AbstractExtentParameter<T>) scopeParams.get(array);
            if (extentParam != null)
            {
               return extentParam;
            }
         }
         
         return null;
      }
   }
}