/*
** Module   : Commitable.java
** Abstract : the interface for objects whose changes can be externally 
**            committed
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20050811   @22108 Created initial version, providing an 
**                           interface for objects whose changes must be
**                           externally committed (saved) on successful
**                           completion of a block.
** 002 GES 20051007   @23006 Added a rollback() notification method
**                           which occurs after all undo processing is
**                           complete and which allows custom internal
**                           state to be maintained.
** 003 CA  20080826   @39566 Added rollbackPending() notification method
**                           which notifies the nested blocks when the 
**                           target for an UNDO statement is one of its
**                           parent block.
** 004 GES 20090729   @43441 Added validate() notification which occurs just
**                           before any normal exit from a block.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

/**
 * An interface for objects whose changes can be externally committed (saved)
 * on successful completion of a block or rolled back (undone) on failure.
 */
public interface Commitable
{
   /**
    * Saves any temporary state to the persistent form or storage that is
    * associated with this class.
    *
    * @param    transaction
    *           <code>true</code> if this is a full transaction and
    *           <code>false</code> if this is only a sub-transaction (a
    *           nested scope with transaction support).
    */
   public void commit(boolean transaction);
   
   /**
    * Notifies that a rollback has just occurred (all state maintained
    * through the {@link Undoable} interface has been already rolled back).
    * This notification can be used to implement cleanup of any custom
    * state that needs to be maintained in this case.  It can also be used
    * to implement a rollback that is completely independent of the
    * <code>Undoable</code> (externally driven) interface.
    *
    * @param    transaction
    *           <code>true</code> if this is a full transaction and
    *           <code>false</code> if this is only a sub-transaction (a
    *           nested scope with transaction support).
    */
   public void rollback(boolean transaction);

   /**
    * Notifies that a pending rollback has just occurred (i.e. a UNDO targets
    * a parent block). This can be used to rollup any information up the
    * parent stack, so the targeted block will be able to process.
    */
   public void rollbackPending();
   
   /**
    * Notifies that a block is about to be normally exited or normally
    * iterated and that any deferred validation processing should be executed.
    * A normal exit or iteration is one that occurs naturally by hitting the
    * end of the block (or loop). A <code>LEAVE</code>, <code>NEXT</code> and 
    * <code>RETURN</code> are also considered normal exits or iterations so
    * long as they were not preceded by an <code>UNDO</code>. Likewise, a
    * <code>QUIT</code> is a normal exit so long as it is not inside a block
    * that has an <code>ON QUIT UNDO, &lt;action&gt;</code>.
    *
    * @param    transaction
    *           <code>true</code> if this is a full transaction and
    *           <code>false</code> if this is only a sub-transaction (a
    *           nested scope with transaction support).
    *
    * @throws   ErrorConditionException
    *           If validation fails.
    */
   public void validate(boolean transaction)
   throws ErrorConditionException;
}