/*
** Module   : CommonHandleChain.java
** Abstract : Defines common APIs used by handles which allow chaining.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.

**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 CA  20130112 Created initial version.
** 002 CA  20130221 Added setters for NAME attribute.
** 003 EVL 20130911 Adding legacy attributes and methods annotation.
** 004 SVL 20131128 NAME attribute moved to Nameable interface.
** 005 GES 20141124 Javadoc fix.
** 006 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.ui.GenericWidget;

/**
 * Exposes the following methods or attributes, common for the window, widget,
 * frame and procedure handles:
 * <table summary="">
 * <tr>
 *    <th>attribute or method</th>
 *    <th>API</th>
 * </tr>
 * <tr>
 *    <td>NAME</td>
 *    <td>{@link #name}</td>
 * </tr>
 * <tr>
 *    <td>PREV-SIBLING</td>
 *    <td>{@link #getPrevSibling()}</td>
 * </tr>
 * <tr>
 *    <td>NEXT-SIBLING</td>
 *    <td>{@link #getNextSibling()}</td>
 * </tr>
 * <tr>
 *    <td>PRIVATE-DATA</td>
 *    <td>{@link #getPrivateData()}, {@link #setPrivateData(character)} and
 *        {@link #setPrivateData(String)}
 *    </td>
 * </tr>
 * </table>
 * This interface is used as a returned type by the 
 * {@link handle#unwrap} API, which, depending on the 
 * handle's state and type, it either returns an 
 * {@link ExternalProgramWrapper} instance (for procedure handles) or an 
 * {@link GenericWidget} instance (for widget, window and frame handles), or  
 * a dynamic proxy using {@link handle#invalidAttrAccessProxy}.
 * <p>
 * All implementations which do not support one or more of the NAME, PRIVATE-DATA, NEXT-SIBLING
 * or PREV-SIBLING attributes should call {@link handle#invalidAttribute}. If an attribute is
 * supported only in read-only mode or the attribute is not supported but its setter is called, 
 * then its setter implementation must explicitly call {@link handle#readOnlyError}.
 * <p>
 * This interface must be implemented by the following resources: BUFFER, BUFFER-FIELD, PROCEDURE,
 * QUERY, SAX-ATTRIBUTES, SAX-READER, SAX-WRITER, SERVER, SERVER-SOCKET, SOAP-HEADER, 
 * SOAP-HEADER-ENTRYREF, SOCKET, TEMP-TABLE, widgets, X-DOCUMENT, X-NODEREF. Following is a list
 * of attributes not supported by certain resources:
 * <ul>
 * <li>NEXT-SIBLING: BUFFER-FIELD, SAX-ATTRIBUTES, SAX-READER, SAX-WRITER, TEMP-TABLE, X-DOCUMENT,
 *                   X-NODEREF</li>
 * <li>PREV-SIBLING: BUFFER, BUFFER-FIELD, SAX-ATTRIBUTES, SAX-READER, SAX-WRITER, TEMP-TABLE, 
 *                   X-DOCUMENT, X-NODEREF</li>
 * <li>NAME: SAX-WRITER; writable only for: TEMP-TABLE, widgets</li>
 * 
 * </ul>
 */
public interface CommonHandleChain
extends CommonHandle,
        Nameable
{
   /**
    * Get the <code>next-sibling</code> attribute of this handle.
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "NEXT-SIBLING")
   public handle getNextSibling();

   /**
    * Get the <code>prev-sibling</code> attribute of this handle.
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "PREV-SIBLING")
   public handle getPrevSibling();

   /**
    * Get the <code>private-data</code> attribute of this handle.
    * 
    * @return   See above.
    */
   @LegacyAttribute(name = "PRIVATE-DATA")
   public character getPrivateData();

   /**
    * Set the <code>private-data</code> attribute of this handle.
    * 
    * @param    data
    *           The new private-data value. 
    */
   @LegacyAttribute(name = "PRIVATE-DATA", setter = true)
   public void setPrivateData(character data);

   /**
    * Set the <code>private-data</code> attribute of this handle.
    * 
    * @param    data
    *           The new private-data value. 
    */
   @LegacyAttribute(name = "PRIVATE-DATA", setter = true)
   public void setPrivateData(String data);

}