/*
** Module : ErrorManager.java
** Public : provides ERROR-STATUS, NO-ERROR and error message support
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 GES 20050823   @21210 Created initial version which provides
**                           ERROR-STATUS and NO-ERROR support.
** 002 GES 20050829   @22362 Minor mods to more closely duplicate how
**                           Progress handles error text. There are
**                           cases where error text is added but no
**                           error condition is raised (even in a
**                           non-silent mode situation).
** 003 ECF 20051104   @23262 Added capability to extract error data from
**                           Java exceptions. This includes both formatted
**                           error texts as well as recording ErrorEntry
**                           objects. Overloaded recordOrThrowError for
**                           this purpose and added a set of overloaded
**                           errorTexts() methods.
** 004 ECF 20051205   @23698 Added recordOrShowError method variants.
**                           Displays an error message in an alert box and
**                           logs the error.
** 005 ECF 20051219   @23752 Added warning mode. When in warning mode,
**                           errors are reported to the UI using an alert
**                           box, rather than an exception being thrown.
** 006 ECF 20060117   @23959 Enabled show error code. Removed temporary
**                           code which used stderr and replaced with
**                           calls to LogicalTerminal.message().
** 007 GES 20060123   @24021 Moved to context-local storage.
** 008 GES 20060208   @24389 Errors are now displayed via message when
**                           not in silent error mode.
** 009 NVS 20060306   @24903 Errors occurring on the client are now passed
**                           to the server for centralized logging.
** 010 NVS 20060313   @24994 Error messages are displayed using Logical
**                           Terminal methods if on the server, and Client
**                           Exports directly otherwise.
** 011 GES 20060313   @24995 Use new logging infrastructure.
** 012 GES 20060626   @27569 Implemented scoped context local data which
**                           makes it safe to use silent error mode in
**                           a nested manner.
** 013 ECF 20060719   @28122 Changed logging level for silent mode errors
**                           from WARNING to FINE.
** 014 GES 20060807   @28454 Match new signature in ClientExports.
** 015 GES 20060927   @29966 Exposed displayError() as public.
** 016 GES 20061004   @30134 Added support for nested errors (errors
**                           generated during processing of another
**                           error).
** 017 NVS 20061017   @30470 Implemented ability to restart the session
**                           without exiting the process.
** 018 ECF 20071217   @36499 Added headless mode. Intended to support
**                           environments with no UI and limited logging.
**                           Implementation at this time is limited.
** 019 CA  20080530   @38463 Added methods related to displaying an error
**                           message before the application stops.
** 020 GES 20080619   @38886 Added a convenience method to format and
**                           display an error without throwing any
**                           exception, w/o being affected by silent
**                           error mode and without recording anything.
** 021 GES 20080624   @38912 Added an ignore flag to NOP all display,
**                           recording or exception generation. Cleaned
**                           up the API to meet coding standards.
** 022 CA  20080815   @39447 Added displayAbend - it displays an abend
**                           message, but no actual abend is triggered.
** 023 SVL 20081208   @40826 Match client exports API change.
** 024 ECF 20081215   @40913 Added throwError() variants. Unconditionally
**                           displays an error message and throws an error
**                           condition exception, regardless of silent
**                           error mode. Headless mode is still honored.
** 025 GES 20090422   @41908 Converted to standard string formatting.
** 026 SVL 20090612   @42669 showErrorAndAbend() can be executed on the client
**                           side.
** 027 ECF 20090622   @42987 Added new display option for errors. New variants
**                           of displayError() and buildErrorText() allow the
**                           standard message prefix to be omitted.
** 028 SIY 20090701   @43012 Made ErrorEntry public for use by other classes.
**                           Inferred generics. Minor cleanups.
** 029 SVL 20090803   @43476 showErrorAndAbend(int[],String[]) has been added.
** 030 GES 20090831   @43802 Removed stacking from the error-status data
**                           members. The 4GL doesn't stack the error list or
**                           the error flag. The previous stacking was an
**                           over-compensation for saving state across top-
**                           level block boundaries. The transaction mgr now
**                           handles saving/restoring the silent error mode
**                           flag in the necessary cases. This class now can
**                           be greatly simplified which also fixes some issues
**                           with the error-status system handle.
** 031 SVL 20090908   @43847 recordOrThrowError() saves error code in the
**                           ErrorConditionException.
** 032 GES 20090924   @44010 Completed headless support. Both the global flag
**                           and a context-specific flag are honored.
** 033 SIY 20110112          Minor refactorings and cleanups.
** 034 CA  20130112          Added asHandle, getType, isValid and isUnknown 
**                           methods, accessed by the ERROR-STATUS handle.
**                           Added new recordOrThrowError APIs.
** 035 CA  20130216          Added support for ERROR-OBJECT-DETAIL attribute.
** 036 CA  20130221          getType getter for TYPE attribute was renamed to getResourceType.
** 037 CS  20130221          Added conversion support for _msg() function
** 038 CA  20130222          Added missing setter for ERROR attribute.
** 039 OM  20130304          Refactored isValid and isUnknown of WrappedResource to valid 
**                           and unknown.
** 040 VMN 20130411          Added boolean parameter modal to recordOrShowError() method.
** 041 EVK 20130413          Implemented method getErrorNumberAtIndex
** 042 CA  20130612          ErrorConditionException's generated in recordOrThrowError will carry
**                           as cause a NumberedException identifying the error number and the
**                           associated message.
** 043 OM  20130606          Handle for soapFault set to null for the moment. This is a bug that
**                           manifests with pl/java and should be fixed (with lazy initialization ?)
** 044 EVL 20130718          Adding new method to display message and abend. Also it silently
**                           returns if NO-ERROPR option is specified.
** 045 VMN 20130731          Adding asMsg parameter to recordOrThrowError() method.
** 046 CA  20130909          Added "prefix" parameter to recordOrShowError, to allow error 
**                           messages without double-asterix prefix.
**                           Added "isError" parameter to recordOrShowError, to allow messages to 
**                           be recorded without setting the pending error flag.
** 047 CA  20130920          Fixed H046 - if ERROR attribute is set to false by recordOrShowError,
**                           than do not clean the error message list.
** 048 EVK 20130922          Added new addError version which will conditionally add the "**" to 
**                           the error message, based on the new prefix parameter.
** 049 CA  20130927          Static proxies need the unknown() API. Resource type is determined 
**                           from LegacyResource constants.
** 050 CA  20131219          Added setDontClean static implementation. Fixed implementation of
**                           isErrorFlag.
** 051 EVL 20131119          Separating error output from client implementation.
** 052 CA  20140107          Made setErrorObjectDetail package private.  The WorkArea.soapFault
**                           must be initialized to null by default, to allow ErrorManager usage
**                           from pl/java (for this reason, setErrorObjectDetail will create a new
**                           handle and assign it to WorkArea.soapFault).
** 053 GES 20140121          Added noRecordOrThrowError().
** 054 CA  20140513          Added a weight for the context-local var, to ensure predetermined 
**                           order during context reset.
** 055 VIG 20140701          Added separate error statuses for the current and pending statuses.
**                           Added class ErrorStatus, that corresponds to ERROR-STATUS handle.
**                           Removed clearList and ServerDataAccess.setError methods.
** 056 SVL 20140709          Prefix field was added to ErrorEntry.
** 057 VIG 20140709          Changed implementation of getElementAtIndex methods (_MSG),
**                           Added raisedConditions - list of error numbers and corresp. methods.
** 058 OM  20141015          Eliminate NPE in addRaisedCondition() when argument is null.
**                           Fixed postprocessOnAbend(String[] text) message.
** 059 SVL 20141209          Added ability to display error messages without finishing dot.
** 060 CA  20150308          Added support for client-side error data storage.  This allows 
**                           temporary disabling the error-management without a server-side trip.
** 061 CA  20150502          The resource ID needs to be kept at the resource, so it can be 
**                           accessed after the resource is deleted.
** 062 OM  20151004          Added setter stub for ERROR-STATUS:NUM-MESSAGES attribute. 
** 063 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 064 IAS 20160328          Client log support
** 065 HC  20160722          Added recordOrShowWarning() stub forwarding calls to
**                           recordOrShowError().
** 066 ECF 20160827          Minor refactoring to avoid expense of getting context-local work
**                           area multiple times in some API calls.
** 067 ECF 20160912          Minor logging change.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.util.*;
import java.util.logging.*;
import com.goldencode.p2j.*;
import com.goldencode.p2j.security.*;
import com.goldencode.proxy.*;

/**
 * A context-local class providing static methods to support the 
 * <code>ERROR-STATUS</code> and <code>NO-ERROR</code> Progress features.
 * <p>
 * The following is the mapping of Progress features to Java:
 * <p>
 * <pre>
 * Progress Feature             Type          Java Feature
 * ---------------------------  ------------  ----------------------------
 * error-status:error           attribute     {@link #isError}         
 * error-status:num-messages    attribute     {@link #numErrors}
 * error-status:get-message     method        {@link #getErrorText}
 * error-status:get-number      method        {@link #getErrorNumber}
 * </pre>
 * <p>
 * In order to properly use this class, all processing which must be handled
 * on a <code>NO-ERROR</code> basis must be enclosed within a matching pair
 * of calls to {@link #silentErrorEnable} and {@link #silentErrorDisable}
 * silent error mode respectively.  These calls must be handled from the
 * client application and all called code that honors the
 * <code>NO-ERROR</code> must honor the silent mode.  This can be done via
 * the following methods (in a library):
 * <p>
 * <ol>
 *    <li> If a simple return from a library is easily possible, delegate
 *         processing to this class using {@link #recordOrThrowError}. After
 *         calling this method, return silently from the library.
 *    <li> Do it yourself.
 *    <ul>
 *       <li> Use {@link #isSilentError} to determine if the silent mode is  
 *            enabled.
 *       <li> Use {@link #addError} to add an error to the list if silent
 *            mode is enabled.
 *       <li> Return silently after the error is added to the list.
 *       <li> If silent mode is not enabled, the
 *            {@link ErrorConditionException} can be thrown directly.
 *       <li> This may be useful if a simple silent return is not easy to do
 *            as one can use the {@link ErrorUnwindException} to unwind the
 *            stack to the point in the library where a silent return is
 *            possible.  This exception must be caught in the library code
 *            and will not be caught in the client application.  Use of
 *            this exception should be completely hidden from the client
 *            application.
 *    </ul>
 * </ol>
 * <p>
 * On the server, the silent error mode flag is saved at every top level
 * scope (see {@link TransactionManager#pushScope}) and it is restored at
 * the end of that scope (see {@link TransactionManager#popScope}).
 * This is done so that "nested" NO-ERROR conditions (e.g. during a RUN
 * statement on an internal or external procedure or an assignment statement
 * with a call to a user defined function) are maintained properly.  In
 * particular, the called procedure or function does NOT execute as if in
 * silent mode unless the code there is so bracketed.  However, the one
 * exception to this is that the RETURN ERROR behavior implemented by the
 * <code>TransactionManager</code> is aware of whether or not silent error
 * mode was enabled in the caller.
 * <p>
 * It must be noted that Progress seems to provide a single global instance of
 * ERROR-STATUS handle and all its data. Thus this data can be seen across
 * call/return boundaries and is only modified at very specific points. for
 * example, the error flag is only changed at the end of a NO-ERROR statement.
 * Likewise, the previous list of errors can be read until the first error
 * occurs after a NO-ERROR statement starts. This deferred clearing of the
 * error list is implemented using the pending error flag. When that flag is
 * set, the error list is cleared. If no errors occur in the NO-ERROR statement
 * then the error list is cleared at the same time the error flag is cleared
 * at the end of the NO-ERROR statement.
 */
public class ErrorManager
{
   /** Capacity of error conditions stack for _MSG function */
   private static final int ERR_CONDITIONS_CAPACITY = 25;
   
   /** Logger (safe as a JVM-wide value rather than as context-local). */
   private static final Logger LOG = LogHelper.getLogger(ErrorManager.class.getName());
   
   /** Stores context-local state variables. */
   private static ContextContainer work = new ContextContainer(); 
   
   /** Error data access interface between client and server. */
   private static RemoteErrorData da = new ServerDataAccess();
   
   /** Error data access to local storage. */
   private static final RemoteErrorData local = new LocalDataAccess();
   
   /** Headless mode */
   private static boolean headless = false;
   
   /** Error writer object instance for client. */
   private static ErrorWriter errorWriter = null;
   
   /** Does this instance belong to the server side part. */
   private static boolean serverSide = true;
   
   /**
    * Get the {@link RemoteErrorData} implementation which uses local storage.
    * 
    * Can be used on client-side only.
    * 
    * @return   See above.
    */
   public static RemoteErrorData getLocalAccess()
   {
      return local;
   }
   
   /**
    * Switch the error storage to the specified instance.
    * 
    * @param    instance
    *           The {@link RemoteErrorData} implementation.
    *           
    * @return   The previous {@link #da} reference.
    */
   public static RemoteErrorData switchErrorStorage(RemoteErrorData instance)
   {
      RemoteErrorData old = da;
      da = instance;

      return old;
   }
   
   /**
    * Get a the instance for the ERROR-STATUS system handle.  Is obtained using
    * a call to the {@link StaticProxy#obtain(Class, Class[])}, using the
    * {@link CommonErrorStatus} interface and its methods implemented by these
    * classes: {@link ErrorManager}.
    * 
    * @return   See above.
    */
   public static handle asHandle()
   {
      WorkArea wa = work.get();

      Class[] classes = { ErrorManager.class };
      CommonErrorStatus proxy = StaticProxy.obtain(CommonErrorStatus.class, classes);

      if (wa.id == null)
      {
         wa.id = handle.resourceId(proxy);
      }

      return new handle(proxy);
   }   
   
   /**
    * Get this resource's ID.
    * 
    * @return   The resource's ID.
    */
   public static Long id()
   {
      WorkArea wa = work.get();
      
      return wa.id;
   }
   
   /**
    * Set this resource's ID.
    * <p>
    * This is a no-op for system handles.
    *  
    * @param    id
    *           The resource's ID.
    */
   public static void id(long id)
   {
      // no-op
   }

   /**
    * Implementation for the {@link WrappedResource#valid()} API.
    * 
    * @return   Always true.
    */
   public static boolean valid()
   {
      return true;
   }
   
   /**
    * Implementation for the {@link WrappedResource#unknown()} API.
    * 
    * @return   Always false.
    */
   public static boolean unknown()
   {
      return false;
   }
   
   /**
    * API needed to implement read-only attribute assignment (a 4GL 
    * "feature").
    * 
    * @param    attribute
    *           The attribute's name.
    *           
    * @see      handle#readOnlyError(handle, String)
    */
   public static void readOnlyError(String attribute)
   {
      handle.readOnlyError(asHandle(), attribute);
   }
   
   /**
    * Initialize <code>ErrorManager</code> for headless mode.
    */
   public static void setHeadless()
   {
      headless = true;
   }
   
   /**
    * Sets silent mode during CAN-FIND calls. The semaphore keep tracks of the nested CAN-FIND
    * calls. 
    * If the counter is positive <code>ErrorManager</code> will run similar to headless mode 
    * and will silently discard error messages.
    * <p> 
    * <strong>Note</strong>Always call this method in pairs (one with enter = true before 
    * executing the query and, one with enter = false, immediately after) to assure the 
    * correct management of the headless context flag. 
    * 
    * @param   enter
    *          If <code>true</code> will increment the nested CAN-FIND count.
    */
   public static void setCanFindBracket(boolean enter)
   {
      if (enter)
      {
         // push a CAN-FIND level
         work.obtain().canFindLevel++;
      }
      else
      {
         // pop put CAN-FIND level:
         work.obtain().canFindLevel--;
      }
   }
   
   /**
    * Check whether this context is in a CAN-FIND bracket. In a CAN-FIND bracket all 4GL errors
    * are not printed 
    * 
    * @return  <code>true</code> if currently a CAN-FIND is executed.
    */
   public static boolean isInCanFindBracket()
   {
      return work.obtain().canFindLevel > 0;
   }
   
   /**
    * Initializes the proper error data access on the client.
    * 
    * @param   errorServer
    *          The remote error interface 
    */
   public static void initClient(RemoteErrorData errorServer)
   {
      da = errorServer;
      serverSide = false;
   }
      
   /**
    * Initializes the proper error writer on the client.
    * 
    * @param    errorWriterInitial
    *           Error writer implementation for client side.
    */
   public static void initErrorWriter(ErrorWriter errorWriterInitial)
   {
      errorWriter = errorWriterInitial;
   }
      
   /**
    * Checks if the client is running in batch mode.
    * 
    * @return  <code>true</code> for batch mode, <code>false</code> otherwise.
    * 
    */
   public static boolean isInBatchMode()
   {
      return (errorWriter != null) ? errorWriter.isInBatchMode() : false;
   }

   /**
    * Stops the associated client session by generating UnstoppableExitException.
    */
   public static void stopBatchSession()
   {
      if (errorWriter != null)
      {
         errorWriter.stopBatchSession();
      }
   }
      
   /**
    * Reverts the error data access on the client.
    */
   public static void resetClient()
   {
      da = new ServerDataAccess();
   }
      
   /**
    * Reports if silent error mode has been enabled.
    *
    * @return   <code>true</code> if silent mode is enabled, otherwise
    *           <code>false</code>.
    */
   public static boolean isSilentError()
   {
      return work.obtain().silent;
   }
   
   /**
    * Reports if warning mode has been enabled.  When in warning mode, error
    * messages are reported to the UI using an alert box.
    *
    * @return   <code>true</code> if warning mode is enabled, otherwise
    *           <code>false</code>.
    */
   public static boolean isWarningMode()
   {
      return work.obtain().warning;
   }
   
   /**
    * Enables silent error mode, clears the current error list but leaves the
    * previous error status alone.  The previous error status is assigned
    * in {@link #silentErrorDisable} based on whether any errors were added
    * in the interim.
    * <p>
    * Each call to this method must be matched with a corresponding call  
    * to {@link #silentErrorDisable}.
    */
   public static void silentErrorEnable()
   {
      // enable silent error processing
      da.setSilent(true);
   }
   
   /**
    * Disables silent error mode, assigns the error status based on whether
    * any errors were added to the error list since the last call to
    * {@link #silentErrorDisable}.
    * <p>
    * Each call to this method must be matched with a corresponding previous  
    * call to {@link #silentErrorEnable}.
    */
   public static void silentErrorDisable()
   {
      // disable further silent error processing
      da.setSilent(false);
      
      da.forwardPending();
   }
   
   /**
    * Enables warning mode, during which errors are non-fatal, but are
    * reported to the UI of the current context via a call to
    * <code>LogicalTerminal.message()</code>.
    * <p>
    * Each call to this method must be matched with a corresponding call  
    * to {@link #warningModeDisable}.
    */
   public static void warningModeEnable()
   {
      // enable warning processing
      work.obtain().warning = true;
   }
   
   /**
    * Disables warning mode.
    * <p>
    * Each call to this method must be matched with a corresponding, previous
    * call to {@link #warningModeEnable}.
    */
   public static void warningModeDisable()
   {
      // disable further warning processing
      work.obtain().warning = false;
   }
   
   /**
    * Manufacture error text based on the given parameters, if silent error mode is enabled,
    * then this error number and the generated error text will be stored in the error list,
    * otherwise this generated error text will be used to display modal or modeless box in
    * the UI.
    * <p>
    * The error message(s) will be prefixed with <code>**</code> and the <code>pending</code>
    * error flag will be set to <code>true</code>.
    * 
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * @param    modal
    *           <code>true</code> if message will be shown in modal dialog.
    */
   public static void recordOrShowError(int num, String text, boolean modal)
   {
      recordOrShowError(new int[] { num }, new String[] { text }, modal, true, true);
   }
   
   /**
    * Manufacture error text based on the given parameters, if silent error mode is enabled,
    * then this error number and the generated error text will be stored in the error list,
    * otherwise this generated error text will be used to display modal or modeless box in
    * the UI.
    * <p>
    * The error message(s) will be prefixed with <code>**</code> and the <code>pending</code>
    * error flag will be set to <code>true</code>.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * @param    modal
    *           <code>true</code> if message will be shown in modal dialog.
    */
   public static void recordOrShowError(int[] num, String[] text, boolean modal)
   {
      recordOrShowError(num, text, modal, true, true);
   }
    
   /**
    * Manufacture error text based on the given parameters, if silent error mode is enabled,
    * then this error number and the generated error text will be stored in the error list,
    * otherwise this generated error text will be used to display modal or modeless box in
    * the UI.
    * <p>
    * The error message(s) will be prefixed with <code>**</code> depending on the 
    * <code>prefix</code> parameter and the <code>pending</code> error flag will be set to 
    * <code>true</code>.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * @param    modal
    *           <code>true</code> if message will be shown in modal dialog.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the formatted error 
    *           message, <code>false</code> to omit this text.
    */
   public static void recordOrShowError(int num, String text, boolean modal, boolean prefix)
   {
      recordOrShowError(new int[] { num }, new String[] { text }, modal, prefix, true);
   }
   
   /**
    * Manufacture error text based on the given parameters, if silent error mode is enabled,
    * then this error number and the generated error text will be stored in the error list,
    * otherwise this generated error text will be used to display modal or modeless box in
    * the UI.
    * <p>
    * The error message(s) will be prefixed with <code>**</code> depending on the 
    * <code>prefix</code> parameter and the <code>pending</code> error flag will be set to 
    * <code>true</code>.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * @param    modal
    *           <code>true</code> if message will be shown in modal dialog.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the formatted error 
    *           message, <code>false</code> to omit this text.
    */
   public static void recordOrShowError(int[] num, String[] text, boolean modal, boolean prefix)
   {
      recordOrShowError(num, text, modal, prefix, true);
   }

   /**
    * Manufacture one or more error texts based on the given parameters, if silent error mode is
    * enabled, then these error numbers and the respective generated error text messages will be
    * stored in the error list, otherwise this generated error texts will be used to display
    * modal or modeless box in the UI.
    * <p>
    * The error message(s) will be prefixed with <code>**</code> depending on the 
    * <code>prefix</code> parameter and the <code>pending</code> error flag will be set to the
    * value specified by the <code>isError</code> flag.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * @param    modal
    *           <code>true</code> if message will be shown in modal dialog.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the formatted error 
    *           message, <code>false</code> to omit this text.
    * @param    isError
    *           The new value for the <code>pending</code> error flag.
    */
   public static void recordOrShowError(int     num,
                                        String  text,
                                        boolean modal,
                                        boolean prefix,
                                        boolean isError)
   {
      recordOrShowError(num, text, modal, prefix, isError, false);
   }

   /**
    * Manufacture one or more error texts based on the given parameters, if silent error mode is
    * enabled, then these error numbers and the respective generated error text messages will be
    * stored in the error list, otherwise this generated error texts will be used to display
    * modal or modeless box in the UI.
    * <p>
    * The error message(s) will be prefixed with <code>**</code> depending on the 
    * <code>prefix</code> parameter and the <code>pending</code> error flag will be set to the
    * value specified by the <code>isError</code> flag.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * @param    modal
    *           <code>true</code> if message will be shown in modal dialog.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the formatted error 
    *           message, <code>false</code> to omit this text.
    * @param    isError
    *           The new value for the <code>pending</code> error flag.
    * @param    asMsg
    *           <code>true</code> to use the given text as the message text else
    *           <code>false</code> to use buildErrorText.
    */
   public static void recordOrShowError(int     num,
                                        String  text,
                                        boolean modal,
                                        boolean prefix,
                                        boolean isError,
                                        boolean asMsg)
   {
      recordOrShowError(new int[] { num }, new String[] { text }, modal, prefix, isError, asMsg);
   }

   /**
    * Manufacture one or more error texts based on the given parameters, if silent error mode is
    * enabled, then these error numbers and the respective generated error text messages will be
    * stored in the error list, otherwise this generated error texts will be used to display
    * modal or modeless box in the UI.
    * <p>
    * The error message(s) will be prefixed with <code>**</code> depending on the 
    * <code>prefix</code> parameter and the <code>pending</code> error flag will be set to the
    * value specified by the <code>isError</code> flag.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * @param    modal
    *           <code>true</code> if message will be shown in modal dialog.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the formatted error 
    *           message, <code>false</code> to omit this text.
    * @param    isError
    *           The new value for the <code>pending</code> error flag.
    */
   public static void recordOrShowError(int[]    num,
                                        String[] text,
                                        boolean  modal,
                                        boolean  prefix,
                                        boolean  isError)
   {
      recordOrShowError(num, text, modal, prefix, isError, false);
   }

   /**
    * Manufacture one or more error texts based on the given parameters, if silent error mode is
    * enabled, then these error numbers and the respective generated error text messages will be
    * stored in the error list, otherwise this generated error texts will be used to display
    * modal or modeless box in the UI.
    * <p>
    * The error message(s) will be prefixed with <code>**</code> depending on the
    * <code>prefix</code> parameter and the <code>pending</code> error flag will be set to the
    * value specified by the <code>isError</code> flag.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * @param    modal
    *           <code>true</code> if message will be shown in modal dialog.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the formatted error
    *           message, <code>false</code> to omit this text.
    * @param    isError
    *           The new value for the <code>pending</code> error flag.
    * @param    asMsg
    *           <code>true</code> to use the given text as the message text else
    *           <code>false</code> to use buildErrorText.
    */
   public static void recordOrShowError(int[]    num,
                                        String[] text,
                                        boolean  modal,
                                        boolean  prefix,
                                        boolean  isError,
                                        boolean  asMsg)
   {
      recordOrShowError(num, text, modal, prefix, isError, asMsg, true);
   }
   
   /**
    * Manufacture one or more error texts based on the given parameters, if silent error mode is
    * enabled, then these error numbers and the respective generated error text messages will be
    * stored in the error list, otherwise this generated error texts will be used to display
    * modal or modeless box in the UI.
    * <p>
    * The error message(s) will be prefixed with <code>**</code> depending on the 
    * <code>prefix</code> parameter and the <code>pending</code> error flag will be set to the
    * value specified by the <code>isError</code> flag.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * @param    modal
    *           <code>true</code> if message will be shown in modal dialog.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the formatted error 
    *           message, <code>false</code> to omit this text.
    * @param    isError
    *           The new value for the <code>pending</code> error flag.
    * @param    asMsg
    *           <code>true</code> to use the given text as the message text else
    *           <code>false</code> to use buildErrorText.
    * @param    addDot
    *           <code>true</code> if dot(s) are added to the end of the message(s).
    */
   public static void recordOrShowError(int[]    num,
                                        String[] text,
                                        boolean  modal,
                                        boolean  prefix,
                                        boolean  isError,
                                        boolean  asMsg,
                                        boolean  addDot)
   {
      if (da.isIgnore())
         return;
      
      if (num.length != text.length)
      {
         throw new IllegalArgumentException(
            "Unbalanced error number and text message pairs");
      }
      
      StringBuilder buf = new StringBuilder();
      String sep = System.getProperty("line.separator");
      boolean silent = isSilentError();
      
      for (int i = 0; i < num.length; i++)
      {
         String errmsg = asMsg ? text[i] : buildErrorText(num[i], text[i], prefix, addDot);
         
         if (i > 0)
         {
            buf.append(sep);
         }
         buf.append(errmsg);
         
         if (silent)
         {
            if (i == 0)
            {
               // this may reset the error list so it must be done BEFORE we
               // add the new error(s) to the list; this only needs to be done
               // the first time through this loop
               da.setPending(isError);
            }
            
            da.addRecord(new ErrorEntry(num[i], errmsg));
         }
      }
      
      String masterMsg = buf.toString();
      
      if (SessionUtils.isDebugAlert().booleanValue() && SessionUtils.isClientlog())
      {
         if (errorWriter != null)
         {
            errorWriter.logStackTrace(masterMsg);
            
         }
      }

      log(masterMsg, silent);
      
      if (!silent && !isHeadless())
      {
         if (modal)
         {
            if (errorWriter != null)
            {
               errorWriter.messageBox(masterMsg, "Error");
            }
         }
         else
         {
            if (errorWriter != null)
            {
               errorWriter.message(masterMsg);
            }
         }
         
         addRaisedCondition(num);
      }
   }
   
   /**
    * Manufacture one or more warning texts based on the given parameters. 
    * <p>
    * The error message(s) will be prefixed with <code>**</code> depending on the 
    * <code>prefix</code> parameter and the <code>pending</code> error flag will be set to the
    * value specified by the <code>isError</code> flag.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the formatted error 
    *           message, <code>false</code> to omit this text.
    * @param    isError
    *           The new value for the <code>pending</code> error flag.
    * @param    asMsg
    *           <code>true</code> to use the given text as the message text else
    *           <code>false</code> to use buildErrorText.
    * @param    addDot
    *           <code>true</code> if dot(s) are added to the end of the message(s).
    */
   public static void recordOrShowWarning(int      num,
                                          String   text,
                                          boolean  prefix,
                                          boolean  isError,
                                          boolean  asMsg,
                                          boolean  addDot)
   {
      // TODO: Currently this method just forwards the call to recordOrShowError().
      //       Depending on the expected semantics of the method, its name may change to 
      //       showWarning.
      recordOrShowError(new int[] {num}, new String[] {text}, true, prefix, isError, asMsg, addDot);
   }

   /**
    * Manufacture error text based on the given parameters; after displaying
    * the error text on the client, a pause is forced before stopping the
    * application.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * 
    * @throws   StopConditionException
    *           always, unless the error manager is in ignore mode.
    */
   public static void showErrorAndAbend(int num, String text)
   throws StopConditionException
   {
      if (da.isIgnore())
         return;
      
      String errmsg = buildAbendText(num, text);
      try
      {
         if (!isHeadless())
         {
            // display the error message (because of pause processing, an
            // endkey, error... even stop condition can be generated here)
            displayError(errmsg);
   
            if (errorWriter != null)
            {                                  
               errorWriter.pause();
            }
         }
      }
      finally
      {
         postprocessOnAbend(errmsg);
      }
   }

   /**
    * Manufacture error text based on the given parameters; after displaying the error text on
    * the client, the application is stopping.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * 
    * @throws   StopConditionException
    *           always, unless the error manager is in ignore mode.
    */
   public static void conditionalShowErrorAndAbend(int num, String text)
   throws StopConditionException
   {
      // If the NO-ERROR option was specified this method just silently retuns doing nothing
      if (isSilentError())
      {
         return;
      }
      
      if (da.isIgnore())
      {
         return;
      }
      
      String errmsg = buildAbendText(num, text);
      try
      {
         if (!isHeadless())
         {
            // display the error message (because of pause processing, an
            // endkey, error... even stop condition can be generated here)
            displayError(errmsg);
         }
      }
      finally
      {
         postprocessOnAbend(errmsg);
      }
   }

   /**
    * Perform postprocessing for {@link #showErrorAndAbend(int, String)}.
    *
    * @param   errmsg
    *          Message text.
    */
   private static void postprocessOnAbend(String errmsg)
   {
      throw new StopConditionException(errmsg);
   }

   /**
    * Manufacture one or more error texts based on the given parameters, if
    * silent error mode is enabled, then these error numbers and the
    * respective generated error text messages will be stored in the error
    * list, otherwise this generated error texts will be used to display a
    * modal, alert (error) box in the UI.<p>
    * After error messages has been processed, a STOP condition is raised.
    *
    * @param    num
    *           The error numbers.
    * @param    text
    *           The text of the error messages.
    *
    * @throws   StopConditionException
    *           always, unless the error manager is in ignore mode.
    */
   public static void showErrorAndAbend(int[] num, String[] text)
   throws StopConditionException
   {
      if (da.isIgnore())
      {
         return;
      }
      
      try
      {
         recordOrShowError(num, text, true);
      }
      finally
      {
         postprocessOnAbend(text);
      }
   }

   /**
    * Perform postprocessing for {@link #showErrorAndAbend(int[], String[])}.
    *
    * @param   text
    *          Message text.
    */
   private static void postprocessOnAbend(String[] text)
   {
      StringBuilder sb = new StringBuilder();
      String sep = System.getProperty("line.separator");
      
      for (int i = 0; i < text.length; i++)
      {
         sb.append(text[i]);
         
         if (i != text.length - 1)
         {
            sb.append(sep);
         }
      }
      
      throw new StopConditionException(sb.toString());
   }

   /**
    * Manufacture error text based on the given parameters, if silent error
    * mode is enabled, then this error number and the generated error text
    * will be stored in the error list, otherwise this generated error text
    * will be used to throw a new instance of {@link ErrorConditionException}.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    *           
    * @throws   ErrorConditionException 
    */
   public static void recordOrThrowError(int num, String text)
   throws ErrorConditionException
   {
      recordOrThrowError(num, text, true);
   }
   
   /**
    * Manufacture error text based on the given parameters, if silent error
    * mode is enabled, then this error number and the generated error text
    * will be stored in the error list, otherwise this generated error text
    * will be used to throw a new instance of {@link ErrorConditionException}.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the
    *           formatted error message, <code>false</code> to omit this text.
    *           
    * @throws   ErrorConditionException 
    */
   public static void recordOrThrowError(int num, String text, boolean prefix)
   {
      recordOrThrowError(new int[] { num }, new String[] { text }, prefix, false);
   }

   /**
    * Manufacture error text based on the given parameters, if silent error
    * mode is enabled, then this error number and the generated error text
    * will be stored in the error list, otherwise this generated error text
    * will be used to throw a new instance of {@link ErrorConditionException}.
    *
    * @param    nums
    *           The error numbers.
    * @param    texts
    *           The text of the error messages.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the
    *           formatted error message, <code>false</code> to omit this text.
    *
    * @throws   ErrorConditionException
    */
   public static void recordOrThrowError(int[] nums, String[] texts, boolean prefix)
   throws ErrorConditionException
   {
      recordOrThrowError(nums, texts, prefix, false);
   }

   /**
    * Manufacture error text based on the given parameters, if silent error
    * mode is enabled, then this error number and the generated error text
    * will be stored in the error list, otherwise this generated error text
    * will be used to throw a new instance of {@link ErrorConditionException}.
    *
    * @param    nums
    *           The error numbers.
    * @param    texts
    *           The text of the error messages.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the
    *           formatted error message, <code>false</code> to omit this text.
    * @param    asMsg
    *           <code>true</code> to use the given text as the message text else
    *           <code>false</code> to use buildErrorText.
    *
    * @throws   ErrorConditionException 
    */
   public static void recordOrThrowError(int[] nums,
                                         String[] texts, 
                                         boolean prefix,
                                         boolean asMsg)
   throws ErrorConditionException
   {
      if (da.isIgnore())
         return;
      
      boolean silent = isSilentError();

      for (int i = 0; i < nums.length; i++)
      {
         int num = nums[i];
         String text = texts[i];
         boolean last = i == nums.length - 1;

         String errmsg = asMsg ? text : buildErrorText(num, text, prefix);
         
         log(errmsg, silent);
         
         if (silent)
         {
            // this may reset the error list so it must be done BEFORE we add
            // the new error to the list
            da.setPending(true);
            
            da.addRecord(new ErrorEntry(num, errmsg));
         }
         else if (isHeadless())
         {
            if (last)
            {
               throw new ErrorConditionException(num, errmsg, new NumberedException(text, num));
            }
         }
         else
         {
            try
            {
               // display the error message (because of pause processing, an
               // endkey, error... even stop condition can be generated here)
               displayError(errmsg);
            }
            
            catch (ConditionException err)
            {
               if (serverSide && TransactionManager.isNestedMode())
               {
                  // honor nested mode here
                  TransactionManager.saveNestedError(err);
               }
               else
               {
                  throw err;
               }
            }
            
            if (last && !isWarningMode())
            {
               // raise the error condition
               throw new ErrorConditionException(num, errmsg, new NumberedException(text, nums));
            }
         }
      }
   }

   /**
    * Manufacture error text based on the given text, if silent error
    * mode is enabled, then this error number and the generated error text
    * will be stored in the error list, otherwise this error text
    * will be used to throw a new instance of {@link ErrorConditionException}.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the
    *           formatted error message, <code>false</code> to omit this text.
    * @param    asMsg
    *           <code>true</code> to use the given text as the message text else
    *           <code>false</code> to use buildErrorText.
    *
    * @throws   ErrorConditionException
    */
   public static void recordOrThrowError(int num, String text, boolean prefix, boolean asMsg)
   {
      recordOrThrowError(new int[] { num }, new String[] { text }, prefix, asMsg);
   }

   /**
    * Record a chain of errors or throw an exception based upon the given
    * exception (which of these two is done depends upon the state of the
    * silent mode feature).  If the former, the error messages and numbers
    * are extracted from <code>error</code> and any other instances of
    * <code>NumberedException</code> present in the chain of root cause
    * exceptions.  The first error entry recorded is derived from the given
    * error number and text, however.
    *
    * @param   num
    *          The error number.
    * @param   text
    *          The text of the error message.
    * @param   error
    *          Exception containing an error number, message, and optionally
    *          a root cause chain.
    */
   public static void recordOrThrowError(int num,
                                         String text,
                                         NumberedException error)
   {
      recordOrThrowError(new NumberedException(text, num, error));
   }
   
   /**
    * Record a chain of errors or throw an exception based upon the given
    * exception (which of these two is done depends upon the state of the
    * silent mode feature).  If the former, the error messages and numbers
    * are extracted from <code>error</code> and any other instances of
    * <code>NumberedException</code> present in the chain of root cause
    * exceptions.
    *
    * @param   error
    *          Exception containing an error number, message, and optionally
    *          a root cause chain.
    */
   public static void recordOrThrowError(NumberedException error)
   {
      if (da.isIgnore())
      {
         return;
      }
      
      boolean silent = isSilentError();
      
      log("stack trace follows", error, silent);
      
      if (silent)
      {
         // this may reset the error list so it must be done BEFORE we add
         // the new error(s) to the list
         da.setPending(true);
         
         for (ErrorEntry errorEntry : compileErrorEntries(error))
         {
            da.addRecord(errorEntry);
         }
      }
      else
      {
         throwError(error);
      }
   }
   
   /**
    * Throws an {@link ErrorConditionException} or sets the <code>pending</code> flag to
    * <code>true</code>, if we are in silent error mode.  NO error information (messages or
    * error numbers) will be recorded in either case (i.e. silent mode or not).  In silent
    * mode the throwing the exception is suppressed and since the pending flag is set, the
    * <code>ERROR-STATUS:ERROR</code> will be <code>true</code> but the other "records" of
    * the error that normally exist, will not exist. If the error is thrown, no message will
    * be displayed.  Because no message is ever displayed or recorded, the normal error
    * number/message processing is not executed.
    * 
    * @param    msg
    *           The text to use in throwing the error exception.
    */
   public static void noRecordOrThrowError(String msg)
   {
      if (isSilentError())
      {
         // ERROR-STATUS:ERROR flag is set, but no error is recorded
         setPending(true);
      }
      else
      {
         // just throw an ERROR condition
         throw new ErrorConditionException(msg);
      }
   }
   
   /**
    * Display error text and throw an exception based upon the given
    * exception.  The error messages and numbers are extracted from
    * <code>error</code> and any other instances of
    * <code>NumberedException</code> present in the chain of root cause
    * exceptions.  The first error entry recorded is derived from the given
    * error number and text, however.
    *
    * @param   num
    *          The error number.
    * @param   text
    *          The text of the error message.
    * @param   error
    *          Exception containing an error number, message, and optionally
    *          a root cause chain.
    */
   public static void throwError(int num, String text, NumberedException error)
   {
      throwError(new NumberedException(text, num, error));
   }
   
   /**
    * Display error text and throw an exception based upon the given
    * exception.  The error messages and numbers are extracted from
    * <code>error</code> and any other instances of
    * <code>NumberedException</code> present in the chain of root cause
    * exceptions.
    *
    * @param   error
    *          Exception containing an error number, message, and optionally
    *          a root cause chain.
    */
   public static void throwError(NumberedException error)
   {
      if (!isHeadless())
      {
         // display the messages
         Iterator<String> iter = errorTexts(error);
         while (iter.hasNext())
         {
            try
            {
               // because of possible pause processing, an endkey, error...
               // even stop condition can be generated here
               displayError(iter.next());
            }
            catch (ConditionException err)
            {
               if (serverSide && TransactionManager.isNestedMode())
               {
                  // honor nested mode here
                  TransactionManager.saveNestedError(err);
               }
               else
               {
                  throw err;
               }
            }
         }
      }
      
      if (!isWarningMode())
      {
         // raise the error condition
         throw new ErrorConditionException(error);
      }
      else
      {
         addRaisedCondition(error.getNumbers());
      }
   }
   
   /**
    * Manufactures Progress compatible formatted error text based on the
    * given error data.  The given data is substituted into the following
    * string "** &#37;s. (&#37;i)".
    *
    * @param    num
    *           The error number substituted for &#37;i in the format string.
    * @param    text
    *           The text of the error message substituted for &#37;s in the
    *           format string.
    *
    * @return   The formatted text.
    */
   public static String buildErrorText(int num, String text)
   {
      return buildErrorText(num, text, true);
   }

   /**
    * Manufactures Progress compatible formatted error text based on the
    * given error data.  The given data is substituted into the following
    * string "** &#37;s. (&#37;i)".
    *
    * @param    num
    *           The error number substituted for &#37;i in the format string.
    * @param    text
    *           The text of the error message substituted for &#37;s in the
    *           format string.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the
    *           formatted error message, <code>false</code> to omit this text.
    *
    * @return   The formatted text.
    */
   public static String buildErrorText(int num, String text, boolean prefix)
   {
      return buildErrorText(num, text, prefix, true);
   }

   /**
    * Manufactures Progress compatible formatted error text based on the
    * given error data.  The given data is substituted into the following
    * string "** &#37;s. (&#37;i)".
    *
    * @param    num
    *           The error number substituted for &#37;i in the format string.
    * @param    text
    *           The text of the error message substituted for &#37;s in the
    *           format string.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the
    *           formatted error message, <code>false</code> to omit this text.
    * @param    addDot
    *           <code>true</code> if a dot is added to the end of the message.
    *
    * @return   The formatted text.
    */
   public static String buildErrorText(int num, String text, boolean prefix, boolean addDot)
   {
      String spec = addDot ? (prefix ? "** %s. (%d)" : "%s. (%d)") :
                             (prefix ? "** %s (%d)" : "%s (%d)");
      
      return String.format(spec, text, num);      
   }
   
   /**
    * Manufactures Progress compatible formatted abend error text based on the
    * given error data.  The given data is substituted into the following
    * string "&#37;s. (&#37;i)".
    *
    * @param    num
    *           The error number substituted for &#37;i in the format string.
    * @param    text
    *           The text of the error message substituted for &#37;s in the
    *           format string.
    *
    * @return   The formatted text.
    */
   public static String buildAbendText(int num, String text)
   {
      return String.format("%s. (%d)", text, num);
   }

   /**
    * Adds an error entry to the error list.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * @param    errFlag
    *           If <code>false</code> this should add the number and text to
    *           the error list but not set the error flag.  If
    *           <code>true</code> the error flag should also be set.
    */
   public static void addError(int num, String text, boolean errFlag)
   {
      addError(num, text, errFlag, true);
   }

   /**
    * Adds an error entry to the error list.
    *
    * @param    num
    *           The error number.
    * @param    text
    *           The text of the error message.
    * @param    errFlag
    *           If <code>false</code> this should add the number and text to
    *           the error list but not set the error flag.  If
    *           <code>true</code> the error flag should also be set.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the
    *           formatted error message, <code>false</code> to omit this text.
    */
   private static void addError(int num, String text, boolean errFlag, boolean prefix)
   {
      if (errFlag)
      {
         // this may reset the error list so it must be done BEFORE we add
         // the new error to the list
         da.setPending(true);
      }

      String errmsg = buildErrorText(num, text, prefix);

      // manually using the work area is OK in this case since this method
      // is only used on the server
      work.obtain().pendingErrorStatus.errorList.add(new ErrorEntry(num, errmsg));
      
      // add error number to _MSG stack.
      addRaisedCondition(new int[] { num });
   }
   
   /**
    * Returns the state of the current error flag which denotes if any error
    * has occurred between the current bracketed calls to the methods
    * {@link #silentErrorEnable} and {@link #silentErrorDisable}. If silent
    * error mode is not active, then this method always will return
    * <code>false</code>.
    *
    * @return    State of the current error flag.
    */
   public static boolean isPendingError()
   {
      return da.isSilent() ? da.isPending() : false;
   }
   
   /**
    * Returns the state of the error flag which was set at the last call to
    * {@link #silentErrorDisable} or if {@link #silentErrorEnable} has never
    * been called, this will return <code>false</code>.
    *
    * @return    Current state of the error flag.
    */
   public static logical isError()
   {
      return new logical(work.obtain().errorStatus.error);
   }
   
   /**
    * Returns the number of the errors (recorded in the error list) which
    * have occurred since the last call to {@link #silentErrorEnable}.
    *
    * @return    Current number of errors in the error list if silent mode
    *            has been enabled. This will be 0 if no errors have occurred  
    *            since the last <code>silentErrorEnable</code> call OR if  
    *            <code>silentErrorEnable</code> has never been called.
    */
   public static integer numErrors()
   {
      return new integer(work.obtain().errorStatus.errorList.size());
   }
   
   /**
    * Undocumented feature encountered in customer code. After all, NUM-MESSAGES (KW_NUM_MSG)
    * attribute seems to be writable.
    * 
    * TODO: investigate if any number can be set or if there are some constraints:
    *    Ex: only 0 can be assigned, meaning some kind of clear error status.
    * 
    * @param   errCount
    *          New value.
    */
   @LegacyAttribute(name = "NUM-MESSAGES", setter = true)
   public static void setNumErrors(int64 errCount)
   {
      UnimplementedFeature.missing("ERROR-STATUS:NUM-MESSAGES not implemented.");
   }
   
   /**
    * Returns the Progress compatible error number associated with the error
    * entry at the given index in the error list.  If the list is empty or
    * if the index is invalid, 0 is returned.
    *
    * @param    index
    *           Position in the error list from which to return the entry.
    *           This is 1-based list index, where 1 is the oldest error number
    *           and highest number is the most recent error number.
    *
    * @return   The error number at that index or 0 if the list is empty or
    *           if the index is invalid.
    */
   public static integer getErrorNumber(integer index)
   {
      if (index.isUnknown())
      {
         return new integer();
      }
      
      return getErrorNumber(index.intValue());
   }
   
   /**
    * Returns the Progress compatible error number associated with the error
    * entry at the given index in the error list.  If the list is empty or
    * if the index is invalid, 0 is returned.
    *
    * @param    index
    *           Position in the error list from which to return the entry.
    *           This is 1-based list index, where 1 is the oldest error number
    *           and highest number is the most recent error number.
    *
    * @return   The error number at that index or 0 if the list is empty or
    *           if the index is invalid.
    */
   public static integer getErrorNumber(int index)
   {
      int result = 0;
      
      if (index > 0 && index <= da.getSize())
      {
         ErrorEntry ee = (ErrorEntry) da.getRecord(--index);
         
         result = ee.num;
      }
      
      return new integer(result);
   }
   
   /**
    * Java implementation for the _msg() function.
    * <p>
    * This will return the error number found on the error stack at the given position, 
    * or 0 if there isn't any error in the stack at the given position.
    * <p>
    * This implementation always will return 0 for any non-positive number (0 or negative). 
    * <p>
    * In the 4GL, some negative numbers will return what appears to be a random number. 
    * This is most likely a flaw in the 4GL implementation where there is improper boundary 
    * testing and an invalid index value is used to dereference memory that is not associated 
    * with the error stack.
    * <p> 
    * As such the behavior is non-deterministic and also is a likely security flaw.
    * <p> 
    * This implementation is deliberately not duplicating this bad behavior and it seems unlikely
    * that an application would be coded to rely upon it.
    * 
    * @param  index
    *         The position of the stack from which we want to get the value error number.
    *         This should be greater than 0 for valid results.
    *         This is 1-based stack offset, where 1 is the most recent error number
    *         and highest number is oldest error number in the stack.
    * 
    * @return The error number found at the given position in the error stack, or 0 if there
    *         isn't any error number at the specified position
    */
   public static integer getErrorNumberAtIndex(long index)
   {
      return getErrorNumberAtIndex(new int64(index));
   }
   
   /**
    * Java implementation for the _msg() function.
    * <p>
    * This will return the error number found on the error stack at the given position, 
    * or 0 if there isn't any error in the stack at the given position.
    * <p>
    * This implementation always will return 0 for any non-positive number (0 or negative). 
    * <p>
    * In the 4GL, some negative numbers will return what appears to be a random number. 
    * This is most likely a flaw in the 4GL implementation where there is improper boundary 
    * testing and an invalid index value is used to dereference memory that is not associated 
    * with the error stack.
    * <p> 
    * As such the behavior is non-deterministic and also is a likely security flaw.
    * <p> 
    * This implementation is deliberately not duplicating this bad behavior and it seems unlikely
    * that an application would be coded to rely upon it.
    * 
    * @param  index
    *         The position of the stack from which we want to get the value error number.
    *         This should be greater than 0 for valid results.
    *         This is 1-based stack offset, where 1 is the most recent error number
    *         and highest number is oldest error number in the stack.
    * 
    * @return The error number found at the given position in the error stack, or 0 if there
    *         isn't any error number at the specified position
    */
   public static integer getErrorNumberAtIndex(integer index)
   {
      return getErrorNumberAtIndex(new int64(index));
   }
   
   /**
    * Java implementation for the _msg() function.
    * <p>
    * This will return the error number found on the error stack at the given position, 
    * or 0 if there isn't any error in the stack at the given position.
    * <p>
    * This implementation always will return 0 for any non-positive number (0 or negative). 
    * <p>
    * In the 4GL, some negative numbers will return what appears to be a random number. 
    * This is most likely a flaw in the 4GL implementation where there is improper boundary 
    * testing and an invalid index value is used to dereference memory that is not associated 
    * with the error stack.
    * <p> 
    * As such the behavior is non-deterministic and also is a likely security flaw.
    * <p> 
    * This implementation is deliberately not duplicating this bad behavior and it seems unlikely
    * that an application would be coded to rely upon it.
    * 
    * @param  index
    *         The position of the stack from which we want to get the value error number.
    *         This should be greater than 0 for valid results.
    *         This is 1-based stack offset, where 1 is the most recent error number
    *         and highest number is oldest error number in the stack.
    * 
    * @return The error number found at the given position in the error stack, or 0 if there
    *         isn't any error number at the specified position
    */
   public static integer getErrorNumberAtIndex(int64 index)
   {
      // default value for incorrect index or not existent value for it is 0
      integer result = new integer(0);
      
      int intIdx = index.intValue();
      
      // check if index is valid
      if (intIdx > 0 && intIdx <= ERR_CONDITIONS_CAPACITY)
      {
         if (intIdx <= getRaisedConditions().size())
         {
            Integer errorNumber = getRaisedConditions().get(intIdx - 1);
            
            // check if there are error number exists for specified index
            if (errorNumber != null)
            {
               result = new integer(errorNumber);
            }
         }
      }

      return result;
   }
   
   /**
    * Gets err nums for raised errors.
    * 
    * @return   stack of errors numbers
    */
   public static LinkedList<Integer> getRaisedConditions() 
   {
      return da.getRaisedConditions();
   }
   
   /**
    * Adds error numbers to raised error numbers stack (for _MSG function).
    * 
    * @param    errNums
    *           errors numbers.
    */
   static void addRaisedCondition(int[] errNums)
   {
      if (isSilent() || errNums == null)
      {
         return;
      }
      
      LinkedList<Integer> raisedConditions = da.getRaisedConditions();
      for (int errNum : errNums)
      {
         raisedConditions.addFirst(errNum);
         if (raisedConditions.size() > ERR_CONDITIONS_CAPACITY)
         {
            raisedConditions.removeLast();
         }
      }
   }
   
   /**
    * Returns the Progress compatible error text associated with the error
    * entry at the given index in the error list.  If the list is empty or
    * if the index is invalid, an empty string "" is returned.
    *
    * @param    index
    *           Position in the error list from which to return the entry.
    *
    * @return   The error text at that index or the empty string "" if the
    *           list is empty or if the index is invalid.
    */
   public static character getErrorText(integer index)
   {
      if (index.isUnknown())
      {
         return new character();
      }
      
      return getErrorText(index.intValue());
   }
   
   /**
    * Returns the Progress compatible error text associated with the error
    * entry at the given index in the error list.  If the list is empty or
    * if the index is invalid, an empty string "" is returned.
    *
    * @param    index
    *           Position in the error list from which to return the entry.
    *
    * @return   The error text at that index or the empty string "" if the
    *           list is empty or if the index is invalid.
    */
   public static character getErrorText(int index)
   {
      String result = "";
      
      if (index > 0 && index <= da.getSize())
      {
         ErrorEntry ee = (ErrorEntry) da.getRecord(--index);
         
         result = ee.text;
      }
      
      return new character(result);
   }   
   
   /**
    * Queries the state of <code>silent</code> variable.
    *
    * @return   <code>true</code> if silent mode is enabled, otherwise
    *           <code>false</code>.
    */
   public static boolean isSilent()
   {
      return da.isSilent();
   }

   /**
    * Sets the <code>silent</code> variable.
    *
    * @param    value
    *           The new state of the variable.
    */
   public static void setSilent(boolean value)
   {
      da.setSilent(value);
   }

   /**
    * Queries the state of <code>ignore</code> variable.
    *
    * @return   <code>true</code> if ignore mode is enabled, otherwise
    *           <code>false</code>.
    */
   public static boolean isIgnore()
   {
      return da.isIgnore();
   }

   /**
    * Sets the <code>ignore</code> variable.
    *
    * @param    value
    *           The new state of the variable.
    */
   public static void setIgnore(boolean value)
   {
      da.setIgnore(value);
   }

   /**
    * Queries the state of <code>error</code> variable.
    *
    * @return   <code>true</code> if error flag is turned on, otherwise
    *           <code>false</code>.
    */
   public static boolean isErrorFlag()
   {
      return da.isErrorFlag();
   }
   
   /**
    * Sets the error variable.  This is a no-op, as the ERROR-STATUS:ERROR attribute
    * is read-only.
    *
    * @param    value
    *           The new state of the variable.
    */
   @LegacyAttribute(name = "ERROR", setter = true, ignore = true)
   public static void setError(boolean value)    
   {
      // no-op
   }
   
   /**
    * Sets the error variable.  This is a no-op, as the ERROR-STATUS:ERROR attribute
    * is read-only.
    *
    * @param    value
    *           The new state of the variable.
    */
   @LegacyAttribute(name = "ERROR", setter = true, ignore = true)
   public static void setError(logical value)    
   {
      // no-op
   }

   /**
    * Queries the state of <code>pendingError</code> variable.
    *
    * @return   <code>true</code> if pending error flag is turned on, otherwise
    *           <code>false</code>.
    */
   public static boolean isPending()
   {
      return da.isPending();
   }

   /**
    * Sets the <code>pendingError</code> variable.
    *
    * @param    value
    *           The new state of the variable.
    */
   public static void setPending(boolean value)    
   {
      da.setPending(value);
   }

   /**
    * Determines if headless mode should be supported. In headless mode, no
    * client-side output will occur.
    *
    * @return   <code>true</code> if the headless flag is turned on, otherwise <code>false</code>.
    */
   public static boolean isHeadless()
   {
      return headless || da.isHeadless() || isInCanFindBracket();
   }

   /**
    * Gets the size of the list of error records.
    *
    * @return   The size of the list.
    */
   public static int getSize()    
   {
      return da.getSize();
   }

   /**
    * Adds a record to the list of error records.
    *
    * @param    record
    *           The object to add to the list of error records.
    */
   public static void addRecord(Object record)    
   {
      da.addRecord(record);
   }

   /**
    * Gets a record from the list of error records.
    *
    * @param    index
    *           The record index.
    *
    * @return   The object associated with the given index in the list.
    */
   public static Object getRecord(int index)    
   {
      return da.getRecord(index);
   }

   /**
    * Makes pending error status current. 
    */
   public static void forwardPending()    
   {
      da.forwardPending();
   }
   
   /**
    * Format the text of the error message and then display that text on
    * on the terminal. Adds the error number to _MSG stack.
    *
    * @param    num
    *           Error number.
    * @param    msg
    *           The text to display in the message line.
    */
   public static void displayError(int num, String msg)
   {
      displayError(num, msg, true);
   }
   
   /**
    * Format the text of the error message and then display that text on
    * on the terminal. Adds the error number to _MSG stack.
    *
    * @param    num
    *           Error number.
    * @param    msg
    *           The text to display in the message line.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the
    *           formatted error message, <code>false</code> to omit this text.
    */
   public static void displayError(int num, String msg, boolean prefix)
   {
      displayError(buildErrorText(num, msg, prefix));
      addRaisedCondition(new int[num]);
   }
   
   /**
    * Format the text of the error message and then display that text on the terminal.  This is
    * the redirected version of the displayError.  If the output is redirected to file the message
    * is writing to the file. Adds the error number to _MSG stack.
    *
    * @param    num
    *           Error number.
    * @param    msg
    *           The text to display in the message line.
    * @param    prefix
    *           <code>true</code> to prepend a double asterisk prefix to the
    *           formatted error message, <code>false</code> to omit this text.
    */
   public static void displayErrorRedirected(int num, String msg, boolean prefix)
   {
      displayErrorRedirected(buildErrorText(num, msg, prefix));
      addRaisedCondition(new int[] {num});
   }

   /**
    * Format the text of the error message and then display that text on the terminal.  This is
    * the redirected version of the displayError.  If the output is redirected to file the message
    * is writing to the file.
    *
    * @param    msg
    *           The text to display in the message line.
    */
   public static void displayErrorRedirected(String msg)
   {
      if (errorWriter != null)
      {
         errorWriter.displayErrorRedirected(msg);
      }
   }
   
   /**
    * Format the text of the abend message and then display that text on
    * on the terminal.
    *
    * @param    num
    *           Error number.
    * @param    msg
    *           The text to display in the message line.
    */
   public static void displayAbend(int num, String msg)
   {
      displayError(buildAbendText(num, msg));
      addRaisedCondition(new int[]{num});
   }

   /**
    * Display the given error message on the terminal.
    * This method does not add a record to _MSG stack.
    *
    * @param    errmsg
    *           The text to display in the message line.
    */
   public static void displayError(String errmsg)
   {
      if (da.isIgnore() || isHeadless())
         return;
      
      if (errorWriter != null)
      {
         errorWriter.displayError(errmsg);
      }
   }
   
   /**
    * Get the type of its associated handle.
    * 
    * @return   Always return the PSEUDO-WIDGET value.
    */
   public static character getResourceType()
   {
      return new character(LegacyResource.PSEUDO_WIDGET);
   }
   
   /**
    * Returns a handle for the ERROR-OBJECT-DETAIL attribute, set in cases of SOAP-related faults.
    * 
    * @return   See above.
    */
   public static handle getErrorObjectDetail()
   {
      WorkArea wa = work.get();
      
      return (wa.errorStatus.soapFault == null)
             ? new handle()
             : new handle(wa.errorStatus.soapFault);
   }
   
   /**
    * Set the ERROR-OBJECT-DETAIL attribute, a handle of type SOAP-FAULT.
    * <p>
    * If the handle is valid and the referred resource is not a {@link SOAPFault} instance, then
    * this will throw a {@link RuntimeException}.
    * <p>
    * If the handle is {@code null} or invalid, the {@link WorkArea#pendingErrorStatus} 's 
    * <code>soapFault</code> field will be set to {@code null}.
    * 
    * @param   fault
    *          The resource representing the soap-fault.
    */
   static void setErrorObjectDetail(SOAPFaultImpl fault)
   {
      handle hFault = new handle();
      // we need to force assignment, as passing the resource at the c'tor will not save it,
      // because at this step we are in no-error mode.
      hFault.set(fault);

      work.get().pendingErrorStatus.soapFault = hFault;
   }

   /**
    * Get an iterator on Progress-style error text strings, derived from the
    * error information stored in <code>error</code> (including its possible
    * chain of root causes).
    *
    * @param   error
    *          A throwable which is an instance of
    *          <code>NumberedException</code> or which contains instances
    *          thereof.
    *
    * @return  Iterator on error text strings.
    */
   static Iterator<String> errorTexts(Throwable error)
   {
      return errorTexts(compileErrorEntries(error));
   }
   
   /**
    * Get an iterator on Progress-style error text strings, derived from the
    * error information stored in <code>error</code> (including its possible
    * chain of root causes), and from the given error number and text.
    *
    * @param   num
    *          Error number of the first message to occur in the resulting
    *          iteration.
    * @param   text
    *          Error text of the first message to occur in the resulting
    *          iteration.
    * @param   error
    *          A throwable which is an instance of
    *          <code>NumberedException</code> or which contains instances
    *          thereof.
    *
    * @return  Iterator on error text strings.
    */
   static Iterator<String> errorTexts(int num, String text, Throwable error)
   {
      List<ErrorEntry> list = compileErrorEntries(error);
      ErrorEntry entry = new ErrorEntry(num, text);
      list.add(entry);
      
      return errorTexts(list);
   }
   
   /**
    * Get an iterator on Progress-style error text strings, based upon the
    * <code>ErrorEntry</code> objects stored in a list.
    *
    * @param   entries
    *          List of error entries.
    *
    * @return  Iterator on error text strings.
    */
   private static Iterator<String> errorTexts(final List<ErrorEntry> entries)
   {
      return new Iterator<String>()
      {
         private final Iterator<ErrorEntry> iter = entries.iterator();

         @Override
         public boolean hasNext()
         {
            return iter.hasNext();
         }

         @Override
         public String next()
         {
            ErrorEntry entry = iter.next();
            
            return buildErrorText(entry.num, entry.text, entry.prefix);
         }

         @Override
         public void remove()
         {
            throw new UnsupportedOperationException();
         }
      };
   }
   
   /**
    * Extract from <code>error</code> and any chained root cause exceptions
    * of type {@link com.goldencode.p2j.NumberedException NumberedException}
    * error messages and numbers with which to compile a list of error
    * entries.  This list is compiled in reverse order of the root cause
    * chain traversal, such that the most specific errors appear earliest in
    * the list.  If no instances of <code>NumberedException</code> are
    * detected during the traversal, the list will be empty.
    *
    * @param   error
    *          Error from which to extract error messages and numbers.
    *
    * @return  List of <code>ErrorEntry</code> objects.
    */
   private static List<ErrorEntry> compileErrorEntries(Throwable error)
   {
      List<ErrorEntry> list = new LinkedList<>();
      
      Throwable next = error;
      do
      {
         if (next instanceof NumberedException)
         {
            NumberedException exc = (NumberedException) next;
            int num        = exc.getNumber();
            String msg     = exc.getMessage();
            boolean prefix = exc.isPrefix();
            ErrorEntry entry = new ErrorEntry(num, msg, prefix);
            list.add(0, entry);
         }
         next = next.getCause();
      }
      while (next != null);
      
      return list;
   }
   
   /**
    * Write a formatted log entry with the error text.  In silent error mode
    * generate a warning and in normal mode generate a severe entry. 
    *
    * @param    message
    *           The text to insert into the log.
    * @param    silent
    *           <code>true</code> if silent error mode is active.
    */
   private static void log(String message, boolean silent)
   {
      log(message, null, silent);
   }
   
   /**
    * Write a formatted log entry with the error text.  In silent error mode
    * generate a warning and in normal mode generate a severe entry. 
    *
    * @param    message
    *           The text to insert into the log.
    * @param    trw
    *           The cause of the failure.
    * @param    silent
    *           <code>true</code> if silent error mode is active.
    */
   private static void log(String message, Throwable trw, boolean silent)
   {
      if (isHeadless())
      {
         return;
      }
      
      Level lvl = silent ? Level.FINE : Level.SEVERE;
      
      if (LOG.isLoggable(lvl))
      {
         String err = LogHelper.generate("%s%s", 
                                         (silent ? "(silent mode) " : ""),
                                         message);
         if (trw == null)
         {
            LOG.logp(lvl, "ErrorManager", "", err);
         }
         else
         {
            LOG.logp(lvl, "ErrorManager", "", err, trw);
         }
      }
   }
   
   /**
    * Stores the data related to a single error.
    */
   public static class ErrorEntry
   {
      /** Error number. */
      public final int num;
      
      /** Error text. */
      public final String text;

      /** Determines if the prefix should be displayed for the error message. */
      public final boolean prefix;
      
      /**
       * Construct an instance.
       *
       * @param    num
       *           The error number.
       * @param    text
       *           The text of the error message.
       */
      public ErrorEntry(int num, String text)
      {
         this.num  = num;
         this.text = text;
         this.prefix = true;
      }

      /**
       * Construct an instance.
       *
       * @param    num
       *           The error number.
       * @param    text
       *           The text of the error message.
       * @param    prefix
       *           Determines if the prefix should be displayed for the error message.
       */
      public ErrorEntry(int num, String text, boolean prefix)
      {
         this.num  = num;
         this.text = text;
         this.prefix = prefix;
      }
   }
   
   /**
    * Error storage local to client-side.
    */
   private static class LocalDataAccess
   extends ErrorDataAccess
   {
      /**
       * Determines if headless mode should be supported.
       * 
       * 
       * @return   always <code>false</code>.
       */
      @Override
      public boolean isHeadless()
      {
         return false;
      }

      /**
       * Makes pending error status current.
       * 
       * This is a no-op.
       */
      @Override
      public void forwardPending()
      {
         // no-op
      }
   }
   
   /**
    * Provides access to the error data when running on the server.
    */
   private static class ServerDataAccess
   extends ErrorDataAccess
   {
      /**
       * Sets the <code>silent</code> variable.
       *
       * @param   value
       *          The new state of the variable.
       */
      @Override
      public void setSilent(boolean value)
      {
         WorkArea wa = work.obtain();
         
         super.setSilent(wa, value);
         
         if (value)
         {
            // we can't delete the soapFault handle here, as this might be a call for accessing
            // some soap-fault attribute/method with the NO-ERROR clause set!
         }
         else
         {
            if (wa.errorStatus.soapFault != null && wa.errorStatus.soapFault._isValid())
            {
               // delete it only if we haven't deleted it already
               HandleOps.delete(wa.errorStatus.soapFault);
            }
            
            wa.errorStatus.soapFault = wa.pendingErrorStatus.soapFault;
            wa.pendingErrorStatus.soapFault = null;
         }
      }
      
      /**
       * Determines if headless mode should be supported. In headless mode, no
       * client-side output will occur.
       *
       * @return   <code>true</code> if the headless flag is turned on,
       *           otherwise <code>false</code>.
       */
      @Override
      public boolean isHeadless()
      {
         return TransactionManager.isHeadless();
      }
      
      /**
       * Makes pending error status current. 
       */
      @Override
      public void forwardPending() 
      {
         WorkArea wa = work.obtain();

         wa.errorStatus.set(wa.pendingErrorStatus);
         wa.pendingErrorStatus.initialize();
      }
   }

   /**
    * Provides access to the error data when running on the server.
    */
   private abstract static class ErrorDataAccess
   implements RemoteErrorData
   {
      /**
       * Queries the state of <code>silent</code> variable.
       *
       * @return   <code>true</code> if silent mode is enabled, otherwise
       *           <code>false</code>.
       */
      @Override
      public boolean isSilent()
      {
         return work.obtain().silent;
      }
   
      /**
       * Sets the <code>silent</code> variable.
       *
       * @param   value
       *          The new state of the variable.
       */
      @Override
      public void setSilent(boolean value)
      {
         work.obtain().silent = value;
      }
   
      /**
       * Queries the state of <code>ignore</code> variable.
       *
       * @return   <code>true</code> if silent mode is enabled, otherwise
       *           <code>false</code>.
       */
      @Override
      public boolean isIgnore()
      {
         return work.obtain().ignore;
      }
   
      /**
       * Sets the <code>ignore</code> variable.
       *
       * @param    value
       *           The new state of the variable.
       */
      @Override
      public void setIgnore(boolean value)
      {
         work.obtain().ignore = value;
      }
   
      /**
       * Queries the state of <code>error</code> variable.
       *
       * @return   <code>true</code> if error flag is turned on, otherwise
       *           <code>false</code>.
       */
      @Override
      public boolean isErrorFlag()
      {
         return work.obtain().errorStatus.error;
      }
      
      /**
       * Sets the <code>pendingError</code> variable.
       *
       * @param    value
       *           The new state of the variable.
       */
      @Override
      public void setPending(boolean value)
      {
         work.obtain().pendingErrorStatus.error = value;  
      }
      
      /**
       * Gets the size of the list of error records.
       *
       * @return   The size of the list.
       */
      @Override
      public int getSize()
      {
         return work.obtain().errorStatus.errorList.size();
      }
   
      /**
       * Adds a record to the list of error records.
       *
       * @param    record
       *           The object to add to the list of error records.
       */
      @Override
      public void addRecord(Object record)
      {
         work.obtain().pendingErrorStatus.errorList.add(record);
      }
   
      /**
       * Gets a record from the list of error records.
       *
       * @param    index
       *           The record index.
       *
       * @return   The object associated with the given index in the list.
       */
      @Override
      public Object getRecord(int index)
      {
         return work.obtain().errorStatus.errorList.get(index);
      }

      /**
       * Queries the state of <code>pendingError</code> variable.
       *
       * @return   <code>true</code> if pending error flag is turned on, otherwise
       *           <code>false</code>.
       */
      @Override
      public boolean isPending()
      {
         return work.obtain().pendingErrorStatus.error; 
      }

      /**
       * Gets err nums for raised errors.
       * 
       * @return   stack of errors numbers
       */
      @Override
      public LinkedList<Integer> getRaisedConditions()
      {
         return work.obtain().raisedConditions;
      }
      
      /**
       * Sets the <code>silent</code> variable.
       *
       * @param   value
       *          The new state of the variable.
       */
      protected void setSilent(WorkArea wa, boolean value)
      {
         wa.silent = value;
      }
   }
   
   /**
    * Stores global data relating to the state of the current context's
    * transactions.
    */
   private static class WorkArea
   {  
      /** The resource's ID. */
      private Long id = null;

      /**
       * Current error status value. 
       */
      private ErrorStatus errorStatus = new ErrorStatus();
      
      /**
       * Pending error status value. 
       */
      private ErrorStatus pendingErrorStatus = new ErrorStatus();
      
      /** 
       * No error flag (if <code>true</code> exceptions are silently recorded
       * rather than thrown).
       */ 
      private boolean silent = false;
      
      /**
       * Flags when a query is executed in a CAN-FIND function.
       * In a CAN-FIND, errors are "swallowed" similar to headless mode. 
       */
      private int canFindLevel = 0;
      
      /**
       * <code>true</code> to NO-OP all error recording, display and
       * exception generation.
       */
      private boolean ignore = false;
      
      /** Whether exceptions are processed as non-fatal, warning messages. */
      private boolean warning = false;
      
      /** 
       * Raised errors numbers stack. Used for undocumented _MSG function implementation.
       * Can contain &le;25 elements, indexed starting from 1.
       */
      private LinkedList<Integer> raisedConditions = new LinkedList<>();
   }
   
   /**
    * Simple container that stores and returns a context-local instance of
    * the global work area.
    */
   private static class ContextContainer
   extends ContextLocal<WorkArea>
   {
      /**
       * Get the weight of this context-local var.
       * 
       * @return   Always {@link WeightFactor#LAST}.
       */
      @Override
      public WeightFactor getWeight()
      {
         return WeightFactor.LAST;
      }
      
      /**
       * Obtains the context-local instance of the contained global work
       * area.
       *
       * @return   The work area associated with this context.
       */
      public WorkArea obtain()
      {
         return this.get();
      }   
      
      /**
       * Initializes the work area, the first time it is requested within a
       * new context.
       *
       * @return   The newly instantiated work area.
       */
      @Override
      protected synchronized WorkArea initialValue()
      {
         return new WorkArea();
      }   
   }

   /**
    * This class corresponds ERROR-STATUS p2j handle, that can contain supressed errors 
    * information in case of NO-ERROR expression used in P4GL statement. 
    */
   private static class ErrorStatus 
   {
      /** List of errors. */
      private List<Object> errorList = new ArrayList<>();
      
      /** Stores whether an error condition has occurred previously. */
      private boolean error = false;
      
      /**
       * Handle containing the value of the ERROR-OBJECT-DETAIL attribute set by the last statement
       * executed with a NO-ERROR clause.
       */
      private handle soapFault = null;

      /**
       * Sets all fields to initial state.
       */
      private void initialize()
      {
         this.errorList.clear();
         this.error = false;
         this.soapFault = null;
      }
      
      /**
       * Sets all fields with values from specified errStatus.
       * 
       * @param   errStatus
       *          source error status.
       */
      private void set(ErrorStatus errStatus)
      {
         this.errorList.clear();
         this.errorList.addAll(errStatus.errorList);
         this.error = errStatus.error;
         this.soapFault = errStatus.soapFault;
      }
   }

}