/*
** Module   : HandleOps.java
** Abstract : Defines APIs to implement handle-specific operations.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130112 Created initial version.
** 002 CA  20130215 Added support for DELETE WIDGET. All deletes are done at the resource, as they
**                  need to implement Deletable.
** 003 CA  20130221 getType was renamed to getResourceType.
** 004 GES 20130302 Added CAN-QUERY and CAN-SET builtin function support.
** 005 CA  20130529 Removed explicit handling of procedure deletion, now all is done in a generic
**                  approach.
** 006 CA  20130918 Resource cleanup must be done only if the resource deleted. 
** 007 EVL 20130911 Adding implementation for CAN-QUERY/CAN-SET functions.
** 008 EVL 20131009 Fix for getting the attribute getters and setters.
** 009 EVL 20131014 Add scan the static methods to CAN-QUERY/CAN-SET functions for possible
**                  ignored attributes.
** 010 CA  20140218 Error messages use the program name as it was used at the RUN call.
** 011 CA  20140223 Fixed deletion of persistent procedures via DELETE OBJECT <system-proc-handle>.
** 012 CA  20140526 handle unwrap APIs were switched to instance method calls.
** 013 CA  20150122 Added isSystemHandle API.
** 014 CA  20151222 Added logging to invalidDelete(handle).
**     OM  20160128 Added 4140 error condition for deleting DEFAULT-WINDOW.
** 015 SBI 20160608 Added CLIPBOARD handle to isSystemHandle.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.ui.*;
import com.goldencode.proxy.*;
import java.util.*;
import java.util.logging.*;
import java.lang.reflect.*;

/**
 * Provides implementation for these handle-specific operations: 
 * <ul>
 *    <li>DELETE OBJECT</li>
 *    <li>DELETE PROCEDURE</li>
 *    <li>CAN-QUERY() built-in function</li>
 *    <li>CAN-SET() built-in function</li>
 * </ul>
 * 
 * Current implementation provides support only for SESSION, FILE-INFO, 
 * LAST-EVENT, ERROR-STATUS and procedure handles.
 */
public class HandleOps
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(HandleOps.class.getName());
   
   /**
    * Check if the given resource represents a system handle.  The procedure system handles 
    * (THIS-PROCEDURE, SOURCE-PROCEDURE, TARGET-PROCEDURE) and the window system handles 
    * (DEFAULT-WINDOW, ACTIVE-WINDOW and CURRENT-wINDOW) are not considered system handles by 
    * this API.
    * 
    * @param     res
    *            The resource to check.
    *            
    * @return    <code>true</code> if the resource represents a system handle as defined above.
    */
   public static boolean isSystemHandle(WrappedResource res)
   {
      return res.valid() && isSystemHandle(new handle(res));
   }
   
   /**
    * Check if the given resource represents a system handle.  The procedure system handles 
    * (THIS-PROCEDURE, SOURCE-PROCEDURE, TARGET-PROCEDURE) and the window system handles 
    * (DEFAULT-WINDOW, ACTIVE-WINDOW and CURRENT-wINDOW) are not considered system handles by 
    * this API.
    * 
    * @param     h
    *            The handle to check.
    *            
    * @return    <code>true</code> if the handle references a system handle.
    */
   public static boolean isSystemHandle(handle h)
   {
      return h._isValid() && (h.equals(SessionUtils.asHandle())  ||
                              h.equals(FileSystemOps.asHandle()) ||
                              h.equals(KeyReader.asHandle())     ||
                              h.equals(ErrorManager.asHandle())  ||
                              h.equals(FontTable.asHandle())     ||
                              h.equals(ColorTable.asHandle())    ||
                              h.equals(ClipboardManager.asHandle()));
   }
   
   /**
    * Deletes the given handles. If multiple handles refer the same resource, then the resource is
    * deleted only once (the first time is encountered).
    * <p>
    * In cases when the handle is a system handle, an {@link #invalidDelete()}
    * is called (to raise an error).
    * 
    * @param    handles
    *           The list of handles to be deleted.
    */
   public static void delete(handle... handles)
   {
      Set<Object> unique = new HashSet<>();
      for (handle h : handles)
      {
         if (h == null)
         {
            invalidDelete();
            break;
         }

         Object referent = h.get();
         if (unique.contains(referent))
            continue;
         
         unique.add(referent);
         
         if (!h._isValid())
         {
            // this is shown for all invalid handles.
            String msg = "Invalid handle.  Not initialized or points to a deleted object";
            ErrorManager.displayError(3135, msg);
            continue;
         }
         
         // DELETE WIDGET can't be protected with no error, so if something fails here, the error
         // will propagate
         delete(h);
      }
   }

   /**
    * Deletes the given handle. The current implementation allows only procedure handles to be
    * deleted.
    * <p>
    * In cases when the handle is a system handle, an {@link #invalidDelete()} is called (to raise
    * an error).
    *
    * @param   h
    *          The handle to be deleted.
    */
   public static void delete(handle h)
   {
      if (h != null && h._isValid())
      {
         Object referent = h.get();
         if (isSystemHandle(h) || 
             (ProcedureManager.systemHandle(referent)) && 
              !ProcedureManager._isPersistent(referent))
         {
            // persistent system handle procedure can be deleted!
            invalidDelete(h);
            return;
         }
         
         WrappedResource resource = h.getResource();
         if (resource instanceof WindowWidget && 
             resource == LogicalTerminal.defaultWindow().getResource())
         {
            // the default window is automatically created at startup and it cannot be destroyed.
            // Only in GUI mode, other WINDOWS can created and destroyed.
            ErrorManager.recordOrShowError(4140, "Cannot DESTROY default window", false, false);
            return;
         }
         
         if (resource instanceof Deletable)
         {
            ((Deletable) resource).delete();
            // do not perform any cleanup here. delete() doesn't ensure the resource was deleted
            // (an error might be raised, with NO-ERROR clause present), so cleanup must be done
            // either in HandleChain.delete (if the resource extends HandleChain) or in resource's
            // delete implementation, if the resource can be deleted
            return;
         }
         
         throw new RuntimeException(
            "Delete implementation not available for handle " + 
            h.unwrapType().getResourceType().toStringMessage());
      }
      
      invalidDelete(h);
   }

   /**
    * Reports if the given attribute can be read from the given handle. The 4GL documentation
    * suggests this can only work on handles that are widgets, but this can be called on any
    * handle.  This corresponds to the <code>CAN-QUERY()</code> built-in function.
    * 
    * @param    h
    *           The handle to be tested.
    * @param    attr
    *           The name of the attribute to be checked.
    *
    * @return   <code>true</code> if the resource in this handle supports reading the named
    *           attribute, <code>false</code> if the resource is valid but doesn't support
    *           the named attribute and <code>unknown</code> if either input is 
    *           <code>unknown</code>. 
    */
   public static logical canQuery(handle h, character attr)
   {
      if (attr == null || attr.isUnknown())
      {
         return new logical();
      }
      
      return canQuery(h, attr.toStringMessage());
   }
   
   /**
    * Reports if the given attribute can be read from the given handle. The 4GL documentation
    * suggests this can only work on handles that are widgets, but this can be called on any
    * handle.  This corresponds to the <code>CAN-QUERY()</code> built-in function.
    * 
    * @param    h
    *           The handle to be tested.
    * @param    attr
    *           The name of the attribute to be checked.
    *
    * @return   <code>true</code> if the resource in this handle supports reading the named
    *           attribute, <code>false</code> if the resource is valid but doesn't support
    *           the named attribute and <code>unknown</code> if either input is 
    *           <code>unknown</code>. 
    */
   public static logical canQuery(handle h, String attr)
   {
      if (h == null || h.isUnknown())
      {
         return new logical();
      }
      
      LegacyAttribute legacyAnnotation = getLegacyAttribute(h, attr, false);
      
      if (isHandleAttribute(attr) ||
          (legacyAnnotation != null && !legacyAnnotation.ignore()))
      {
         return new logical(true);
      }

      return new logical(false);
   }
   
   /**
    * Reports if the given attribute can be written to the given handle. The 4GL documentation
    * suggests this can only work on handles that are widgets, but this can be called on any
    * handle.  This corresponds to the <code>CAN-SET()</code> built-in function.
    * 
    * @param    h
    *           The handle to be tested.
    * @param    attr
    *           The name of the attribute to be checked.
    *
    * @return   <code>true</code> if the resource in this handle supports writing the named
    *           attribute, <code>false</code> if the resource is valid but doesn't support
    *           the named attribute and <code>unknown</code> if either input is 
    *           <code>unknown</code>. 
    */
   public static logical canSet(handle h, character attr)
   {
      if (attr == null || attr.isUnknown())
      {
         return new logical();
      }
      
      return canSet(h, attr.toStringMessage());
   }
   
   /**
    * Reports if the given attribute can be written to the given handle. The 4GL documentation
    * suggests this can only work on handles that are widgets, but this can be called on any
    * handle.  This corresponds to the <code>CAN-SET()</code> built-in function.
    * 
    * @param    h
    *           The handle to be tested.
    * @param    attr
    *           The name of the attribute to be checked.
    *
    * @return   <code>true</code> if the resource in this handle supports writing the named
    *           attribute, <code>false</code> if the resource is valid but doesn't support
    *           the named attribute and <code>unknown</code> if either input is 
    *           <code>unknown</code>. 
    */
   public static logical canSet(handle h, String attr)
   {
      if (h == null || h.isUnknown())
      {
         return new logical();
      }
      
      LegacyAttribute legacyAnnotation = getLegacyAttribute(h, attr, true);
      
      if (!isHandleAttribute(attr) &&
          (legacyAnnotation != null && !legacyAnnotation.ignore()))
      {
         return new logical(true);
      }

      return new logical(false);
   }
   
   /**
    * Obtains the legacy attribute annotation instance for the given method  and given legacy
    * attribute name.
    * 
    * @param    meth_check
    *           The method to be tested.
    * @param    attr
    *           The name of the legacy attribute to be checked.
    * @param    setter
    *           The flag indicating if we need to find the setter or getter annotation.
    *
    * @return   The instance of the legacy attribute annotation if it exists for the given class
    *           or null if not found. 
    */
   private static LegacyAttribute getLegacyAttributeInMethod(Method meth_check,
                                                             String attr,
                                                             boolean setter)
   {
      LegacyAttribute legacyAttrAnno = meth_check.getAnnotation(LegacyAttribute.class);
      // Check annotation for name match(ignoring letter case)
      if (legacyAttrAnno != null && legacyAttrAnno.name().equalsIgnoreCase(attr))
      {
         boolean settable = legacyAttrAnno.setter(); 
         if ((!setter && !settable) || (setter && settable))
         {
            return legacyAttrAnno;
         }
      }
      
      return null;
   }
   
   /**
    * Obtains the legacy attribute annotation instance for the given implementation or interface
    * class object and given legacy attribute name.
    * 
    * @param    class_check
    *           The interface or implementation class object to be tested.
    * @param    attr
    *           The name of the legacy attribute to be checked.
    * @param    setter
    *           The flag indicating if we need to find the setter or getter annotation.
    *
    * @return   The instance of the legacy attribute annotation if it exists for the given class
    *           or null if not found. 
    */
   private static LegacyAttribute getLegacyAttributeInClass(Class class_check,
                                                            String attr,
                                                            boolean setter)
   {
      // For each interface implementation get the methods and check each one
      Method[] methods = class_check.getMethods();
      if (methods != null)
      {
         for (Method method : methods)
         {
            // Check every method for match
            LegacyAttribute legacyAttrAnno = getLegacyAttributeInMethod(method, attr, setter);
            // Check if we found acceptable annotation
            if (legacyAttrAnno != null)
            {
               return legacyAttrAnno;
            }
         }
      }
      
      return null;
   }
   
   /**
    * Obtains the legacy attribute annotation instance for the given static class object and
    * given legacy attribute name.
    * 
    * @param    obj_check
    *           The class object to be tested.
    * @param    attr
    *           The name of the legacy attribute to be checked.
    * @param    setter
    *           The flag indicating if we need to find the setter or getter annotation.
    *
    * @return   The instance of the legacy attribute annotation if it exists for the given class
    *           or null if not found. 
    */
   private static LegacyAttribute getLegacyAttributeInStatic(Object obj_check,
                                                             String attr,
                                                             boolean setter)
   {
      // For each interface implementation get the methods and check each one
      List<Method> staticMethods = StaticProxy.getStaticMethods(obj_check);
      if (staticMethods != null)
      {
         for (Method method : staticMethods)
         {
            // Check every method for match
            LegacyAttribute legacyAttrAnno = getLegacyAttributeInMethod(method, attr, setter);
            // Check if we found acceptable annotation
            if (legacyAttrAnno != null)
            {
               return legacyAttrAnno;
            }
         }
      }
      
      return null;
   }
   
   /**
    * Obtains the legacy attribute annotation instance for the given handle object and given
    * legacy attribute name.
    * <p>
    * The search algorithm for this method uses reflection to read annotations from various
    * sources in the following precedence order:
    * <ol>
    *   <li> The proxy static methods of the resource itself if exist.
    *   <li> The class of the resource itself.
    *   <li> The superclass of the resource or the superclass of the superclass... the entire
    *        parent hierarchy is walked from the resource to the root.  The first match found ends
    *        the upward search.
    *   <li> All interfaces of the resource itself.
    *   <li> All interfaces of the superclass of the resource or the superclass of the
    *        superclass... the entire parent hierarchy is walked from the resource to the root.
    *        The first superclass with an interface that has a matching attribute will end the
    *        upward search.
    * </ol>
    * <p>
    * Where possible, we put annotations on the interface, not the class for the resource.  But
    * there are some attributes where the configuration is not universal for all resources
    * implementing the same attribute.  For example, the following attributes are readable for
    * certain resources even though for most other resources the same attribute is writable:
    * <ul>
    *   <li> LABEL for LAST-EVENT
    *   <li> MODIFIED for browse columns
    *   <li> COLUMN for browse cells, field groups, and LAST-EVENT
    *   <li> NO-VALIDATE for both static and dynamic browses (it is writeable only for dynamic
    *        browses)
    * </ul>
    * <p>
    * Keeping the above precedence order is critical to allow attribute configurations of
    * more-specific resources to properly override the more general configuration in the parent
    * hierarchy or the interface hierarchy.
    * 
    * @param    h
    *           The handle object to be tested.
    * @param    attr
    *           The name of the legacy attribute to be checked.
    * @param    setter
    *           The flag indicating if we need to find the setter or getter annotation.
    *
    * @return   The instance of the legacy attribute annotation if it exists for the given handle
    *           or null if not found. 
    */
   private static LegacyAttribute getLegacyAttribute(handle h, String attr, boolean setter)
   {
      LegacyAttribute legacyAnnotationRes = null;
      Object objFromHandle = h.get();
      
      if (objFromHandle != null)
      {
         // Checking the static proxy first
         legacyAnnotationRes = getLegacyAttributeInStatic(objFromHandle, attr, setter);
         // Found result in static methods
         if (legacyAnnotationRes != null)
         {
            return legacyAnnotationRes;
         }
         
         // Getting the base class
         Class current_base = objFromHandle.getClass(); 
         // Getting implementation methods from current class first
         while (current_base != null)
         {
            // For each implementation get the methods and check each one
            legacyAnnotationRes = getLegacyAttributeInClass(current_base, attr, setter);
            // Found result in implementation
            if (legacyAnnotationRes != null)
            {
               return legacyAnnotationRes;
            }
            // If we are here the annotation is not found, try superclass if exists
            current_base = current_base.getSuperclass(); 
         }

         // Restore the search from the base class
         current_base = objFromHandle.getClass();
         // Getting iterfaces from current class
         while (current_base != null)
         {
            Class[] classes = current_base.getInterfaces();
            if (classes != null)
            {
               for (Class iface_curr : classes)
               {
                  // For each interface implementation get the methods and check each one
                  legacyAnnotationRes = getLegacyAttributeInClass(iface_curr, attr, setter);
                  // Found result in implementation
                  if (legacyAnnotationRes != null)
                  {
                     return legacyAnnotationRes;
                  }
               }
            }
            // If we are here the annotation is not found, try superclass if exists
            current_base = current_base.getSuperclass(); 
         }
      }
      
      return null;
   }
   
   /**
    * Checks if the given attribute name is HANDLE.  This is used to bypass reflection based code
    * to find out if the attribute can be set or queried.  We know this is read-only and always
    * exists for valid handle we use in CAN-SET and CAN-QUERY calls.
    * 
    * @param    attrName
    *           The name of the attribute to be checked.
    *
    * @return   <code>true</code> if the attribute name is HANDLE, <code>false</code> otherwise.
    */
   private static boolean isHandleAttribute(String attrName)
   {
      return attrName.equalsIgnoreCase("HANDLE");
   }
   
   /**
    * Invoked when the handle is invalid or a system (non-procedure) handle. 
    * Shows appropriate error messages and raises an error. 
    * Calls {@link #invalidDelete(handle) invalidDelete(null)}.
    */
   private static void invalidDelete()
   {
      invalidDelete(null);
   }

   /**
    * Invoked when the handle is invalid or a system (non-procedure) handle. 
    * Shows appropriate error messages and raises an error.
    * 
    * @param    h
    *           The handle for which the delete failed. May be 
    *           <code>null</code> if it was decide this is a non-procedure 
    *           handle.
    */
   private static void invalidDelete(handle h)
   {
      boolean procedure = (h != null && h._isValid() && ProcedureManager.isProcedure(h));

      final String msg1 = "Invalid or inappropriate handle value given to " +
                          "DELETE OBJECT or DELETE PROCEDURE statement";
      final int id1 = 5425;

      final String msg2 = "Invalid or inappropriate handle value given to " +
                          "DELETE PROCEDURE statement. Procedure '%s':%d";
      final int id2 = 1428;
      // this can't be computed, is the legacy line num from the 4GL file.
      final int linenum = 0;
      
      int id = -1;
      String msg = null;
      
      if (procedure)
      {
         String pname = ProcedureManager.getRelativeName(h.get());
         msg = String.format(msg2, pname, linenum);
         id = id2;
      }
      else
      {
         msg = msg1;
         id = id1;
      }

      if (LOG.isLoggable(Level.FINE))
      {
         String emsg = (h == null ? "with null handle" 
                                  : (h.get() == null ? "with null resource" 
                                                     : "with resource " + h.get().getClass()));
         String noError = ErrorManager.isSilentError() ? "(with NO-ERROR) " : "";
         emsg = msg + ". (" + id + ") - " + noError + emsg;
         
         LOG.log(Level.FINE, emsg, new ErrorConditionException(emsg));
      }
      
      ErrorManager.recordOrThrowError(id, msg, false);
   }
}