/*
** Module   : HandleResource.java
** Abstract : Implementation of common attributes and APIs for all resources. 
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130221 Created initial version.
** 002 OM  20130304 Refactored isValid and isUnknown of WrappedResource to valid and unknown.
** 003 EVK 20130903 Added method type().
** 004 CA  20130927 All resources are known by default; special resources which need to be unknown
**                  will override the unknown() API. The resource type is determined based on the
**                  LegacyResource annotation: an annotation at the class has precedence over an
**                  annotation at the interface.
** 005 CA  20131113 Added INSTANTIATING-PROCEDURE support.
**                  Changes related to properly instantiate dynamic resources.
** 006 CA  20140326 Fixed implicitDeletion in case is called from an Agent, outside of an external
**                  procedure (when the remote persistent procedure is deleted by the Agent).
** 007 CA  20150502 The resource ID needs to be kept at the resource, so it can be accessed after
**                  the resource is deleted.
**                  Track the resources created by each procedure, as dynamic resources don't get  
**                  deleted when the creator proc gets deleted, but their INSTANTIATING-PROCEDURE 
**                  attr needs to be unknown.
** 008 EVL 20160224 Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 009 CA  20160327 getResourceType() invokes type(), so sub-classes can override it.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.util.*;

import com.goldencode.proxy.*;

/**
 * This acts as a superclass which implements:
 * <table summary="">
 * <tr>
 *    <th>attribute or base method</th>
 *    <th>API</th>
 * </tr>
 * <tr>
 *    <td>TYPE</td>
 *    <td>{@link #getResourceType()} and {@link #type()}</td>
 * </tr>
 * <tr>
 *    <td>INSTANTIATING-PROCEDURE</td>
 *    <td>{@link #instantiatingProcedure()}</td>
 * </tr>
 * <tr>
 *    <td>Unknown status</td>
 *    <td>{@link #unknown()}</td>
 * </tr>
 * <tr>
 *    <td>Read only field error</td>
 *    <td>{@link #readOnlyError(String)}</td>
 * </tr>
 * <tr>
 *    <td>Invalid attribute access</td>
 *    <td>{@link #invalidAttribute(String)}</td>
 * </tr>
 * </table>
 * <p>
 * The abstract class also provides setters for the implemented attributes 
 * which are not defined inside the interfaces to be used for initializing of
 * those attributes
 */
public abstract class HandleResource
implements CommonHandle,
           InstantiatingProcedure
{
   /** Handle referencing the instantiating procedure. */
   protected final handle instProc = new handle();

   /** Variable for holding the type attribute */
   private String type = null;
   
   /**
    * The resource's ID.  Remains <code>null</code> until the ID is requested via 
    * {@link handle#resourceId}
    */
   private Long id = null;
   
   /**
    * Default c'tor.
    */
   public HandleResource()
   {
      this(false);
   }
   
   /**
    * Initialize this resource by determining its type.
    * <p>
    * The resource's type is determined from a {@link LegacyResource} annotation, which is 
    * searched in these locations, in this order:
    * <ol>
    * <li>The class implementing the resource in P2J is checked for an annotation. If found, this
    *     is used to determine the resource type.</li>
    * <li>The entire interface hierarchy is checked. There should be only one interface with a
    *     {@link LegacyResource} annotation; if there is more than annotated interface, then n
    *     {@link IllegalStateException} is thrown.</li>
    * </ol>
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    * 
    * @throws   IllegalStateException
    *           If the resource type could not be determined from a {@link LegacyResource} 
    *           annotation, after checking the interface hierarchy, or more than one interface
    *           is annotated.
    */
   protected HandleResource(boolean dynamic)
   {
      // check the class first
      Class<?> cls = getClass();
      LegacyResource lr = cls.getAnnotation(LegacyResource.class);
      if (lr != null)
      {
         type = lr.resource();
      }
      else
      {
         // else, determine it from the interface. the annotated interface in most cases is a 
         // directly implemented interface, but ensure we reach all interfaces, to cover all
         // bases

         Deque<Class<?>> ifcs = new ArrayDeque<Class<?>>();
         // add all the interfaces from the extended classes first
         Class<?> superCls = cls;
         while (superCls != null)
         {
            ifcs.addAll(Arrays.asList(superCls.getInterfaces()));
            
            superCls = superCls.getSuperclass();
         }
         
         Set<Class<?>> checked = new HashSet<Class<?>>();
         
         while (!ifcs.isEmpty())
         {
            Class<?> ifc = ifcs.remove();
            
            if (checked.contains(ifc))
               continue;
            
            // keep track of which interface we've visited, as the same interface may appear more
            // than once in the hierarchy
            checked.add(ifc);
            
            lr = ifc.getAnnotation(LegacyResource.class);
            if (lr != null)
            {
               if (type != null)
               {
                  // if more than one interface is annotated, then we are ambiguous
                  final String msg = 
                     "More than one LegacyResource annotation in the implemented interfaces " +
                     "for class %s.";
                  throw new IllegalStateException(String.format(msg, cls.getName()));
               }

               type = lr.resource();
            }
            
            ifcs.addAll(Arrays.asList(ifc.getInterfaces()));
         }
      }

      if (type == null)
      {
         final String msg = "Could not determine the legacy resource type for class %s.";
         throw new IllegalStateException(String.format(msg, cls.getName()));
      }
      
      // this will ensure the INSTANTIATING-PROCEDURE attribute is setup accordingly and also that
      // static resources will be registered with the correct top-level block
      if (!ProxyFactory.isProxyClass(this.getClass()))
      {
         // for proxy cases, explicit registration is needed
         ProcedureManager.processResource(this, dynamic);
      }
   }
   
   /**
    * Get this resource's ID, if is already set.
    * 
    * @return   The resource's ID or <code>null</code> if not set.
    */
   @Override
   public Long id()
   {
      return id;
   }
   
   /**
    * Set this resource's ID.
    *  
    * @param    id
    *           The resource's ID.
    */
   @Override
   public void id(long id)
   {
      this.id = id;
   }

   /**
    * Get the type of this handle.
    * 
    * @return   See above.
    */
   public character getResourceType()
   {
      return new character(type());
   }

   /**
    * A newly-created resource is always not-unknown. If implementations need to return something
    * different, then override this method.
    * 
    * @return   always <code>false</code>
    */
   @Override
   public boolean unknown()
   {
      return false;
   }
   
   /**
    * To be used only in cases when a setter for a read-only attribute or a not-supported  
    * attribute is forced to be implemented by the resource. It displays appropriate error 
    * message.
    * 
    * @param    attr
    *           The read-only attribute.
    */
   public void readOnlyError(String attr)
   {
      handle.readOnlyError(new handle(this), attr);
   }

   /**
    * Get the type of this handle as type of String.
    *
    * @return   See above.
    */
   public String type()
   {
      return type;
   }
   
   /**
    * Getter for the INSTANTIATING-PROCEDURE attribute.
    * <p>
    * The returned handle will refer a valid external procedure (persistent or not) or will be
    * unknown, in case its instantiating procedure was persistent and has been deleted.
    * 
    * @return   See above.
    */
   @Override
   public handle instantiatingProcedure()
   {
      return ProcedureManager.resolveInstantiatingProcedure(instProc);
   }
   
   /**
    * To be used only in cases a getter for a not-supported attribute is forced to be implemented
    * by the resource. It displays appropriate error message.
    * 
    * @param    attr
    *           The not-supported attribute attribute.
    */
   protected void invalidAttribute(String attr)
   {
      handle.invalidAttribute(attr, new handle(this));
   }
   
   /**
    * Check if the resource may be implicitly deleted. 
    * 
    * @return   <code>true</code> if the INSTANTIATING-PROCEDURE attribute is unknown (i.e. the 
    *           instantiating procedure has been deleted) and we are during finalizable or
    *           scopeable notification.
    */
   protected boolean implicitDeletion()
   {
      return ProcedureManager.inNotifications() &&
             (TransactionManager.getNestingLevel() == 0 || 
              !TransactionManager.isExternalBlock()     || 
              instantiatingProcedure().isUnknown());
   }

   /**
    * Set the instantiating procedure for this resource.
    * 
    * @param    hproc
    *           The procedure where the resource was instantiated.
    */
   protected void setInstantiatingProcedure(handle hproc)
   {
      this.instProc.assign(hproc);
      
      if (hproc._isValid())
      {
         // register the resource with the procedure, too
         ProcedureManager.registerResource(hproc.get(), this);
      }
   }
}