/*
** Module   : InteractiveChildProcess.java
** Abstract : API for starting interactive child processes
**
** Copyright (c) 2015-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 GES 20150409 First version. This interface helps isolate process launching from UI.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

/**
 * Provides an API to launch and manage an interactive child process which is independent from
 * any specific UI modality. 
 */
public interface InteractiveChildProcess
{
   /**
    * Reports if the client is in batch mode.
    *
    * @return   <code>true</code> if in batch mode.
    */
   public boolean inBatchMode();
   
   /**
    * Reports if the client is missing an already initialized native terminal session.
    *
    * @return   <code>true</code> if this is an artificial terminal (there is no native
    *           terminal).
    */
   public boolean isArtificialTerm();
   
   /**
    * Launch the specified command line as an interactive child process.
    *
    * @param    cmdline
    *           The command line to execute.  The first element must be the program name.
    * @param    silent
    *           <code>false</code> for interactive mode (there the terminal is expected to be
    *           shared by the child process and the user is allowed to directly interact with
    *           the child process.
    *
    * @return   The operating system process ID (pid) of the child process.
    */
   public int launch(String[] cmdline, boolean silent);
   
   /**
    * Block execution until the interactive child process has exited. This will ensure that
    * any operating system resources are released before returning.
    *
    * @param    pid
    *           The operating system process ID of the child process on which to wait.
    * @param    silent
    *           <code>false</code> for interactive mode (there the terminal is expected to be
    *           shared by the child process and the user is allowed to directly interact with
    *           the child process.
    */
   public void waitForExit(int pid, boolean silent)
   throws InterruptedException;
   
   /**
    * Asynchronously cleanup any operating system resources when the interactive child process
    * has exited.  This processing should be done an a secondary thread and this method will
    * return immediately after starting that thread.
    *
    * @param    pid
    *           The operating system process ID of the child process on which to wait.
    */
   public void asynchCleanup(int pid);
   
   /**
    * Wait for the user to press a key.
    *
    * @return   The key code of the key pressed to clear the pause or -1 on any failure.
    */
   public int pauseForUser();

   /**
    * Temporarily halts the client UI processing (key reading, normal drawing
    * mode)  After entering suspend mode, the current contents of the
    * terminal are cleared.  Use {@link #resume} to return to normal mode.
    */
   public void suspend();
   
   /**
    * Re-enables client UI processing (key reading, normal drawing mode).
    * Before resuming the current contents of the terminal are cleared.  This
    * is the mechanism to return to normal processing mode after having
    * called {@link #suspend}.
    */
   public void resume();   
}