/*
** Module   : LegacyLogManagerImpl.java
** Abstract : Implementation for the LOG-MANAGER and DSLOG-MANAGER system handles.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 GES 20141028 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

/**
 * Implementation for the LOG-MANAGER and DSLOG-MANAGER system handles.  The two resources
 * have virtually identical interfaces and the minor differences will be managed inside this
 * class instead of making the overall solution more complicated.
 */
@LegacyResource(resource = LegacyResource.PSEUDO_WIDGET)
public class LegacyLogManagerImpl
extends HandleResource
implements LegacyLogManager
{
   /** Default subsystem description for WRITE-MESSAGE(). */
   private static final String DEFAULT_SUBSYSTEM = "APPL";
   
   /** If <code>true</code>, implement DSLOG-MANAGER otherwise this is the LOG-MANAGER. */
   private boolean isDataServer = false;
   
   /**
    * Package private constructor to create an instance that is either a LOG-MANAGER or a
    * DSLOG-MANAGER instance.
    *
    * @param    isDataServer
    *           <code>true</code> if this should be a DSLOG-MANAGER, <code>false</code> for a
    *           LOG-MANAGER instance.
    */
   LegacyLogManagerImpl(boolean isDataServer)
   {
      this.isDataServer = isDataServer;
   }
   
   /**
    * Provides the comma-separated list of currently valid entry types that can be assigned to
    * the {@link #setLogEntryTypes}.
    *
    * @return   The list of valid log entry types.
    */
   public character getEntryTypesList()
   {
      UnimplementedFeature.missing("ENTRY-TYPES-LIST attribute getter");
      return new character();
   }
   
   /**
    * Provides the comma-separated list of entry types that are being actively used for logging.
    *
    * @return   The list of active log entry types.
    */
   public character getLogEntryTypes()
   {
      UnimplementedFeature.missing("LOG-ENTRY-TYPES attribute getter");
      return new character();
   }
   
   /**
    * Modifies the list of log entry types that are being actively used for logging.
    *
    * @param    types
    *           The comma-separated list of active entry types.
    */
   public void setLogEntryTypes(String types)
   {
      UnimplementedFeature.missing("LOG-ENTRY-TYPES attribute setter");
   }
   
   /**
    * Modifies the list of log entry types that are being actively used for logging.
    *
    * @param    types
    *           The comma-separated list of active entry types.
    */
   public void setLogEntryTypes(character types)
   {
      if (types == null || types.isUnknown())
      {
         UnimplementedFeature.todo("Confirm unknown behavior is correct.");
         return;
      }
      
      setLogEntryTypes(types.toStringMessage());
   }
   
   /**
    * Obtains the filename to which log entries and stack traces are written.
    *
    * @return   The filename.
    */
   public character getLogFileName()
   {
      UnimplementedFeature.missing("LOGFILE-NAME attribute getter");
      return new character();
   }
   
   /**
    * Assigns the filename to which log entries and stack traces are written.
    *
    * @param    name
    *           The log file name.
    */
   public void setLogFileName(String name)
   {
      UnimplementedFeature.missing("LOGFILE-NAME attribute setter");
   }
   
   /**
    * Assigns the filename to which log entries and stack traces are written.
    *
    * @param    name
    *           The log file name.
    */
   public void setLogFileName(character name)
   {
      if (name == null || name.isUnknown())
      {
         UnimplementedFeature.todo("Confirm unknown behavior is correct.");
         return;
      }
      
      setLogFileName(name.toStringMessage());
   }
   
   /**
    * Obtains the default logging level for all logging categories.
    *
    * @return   The logging level.
    */
   public integer getLoggingLevel()
   {
      UnimplementedFeature.missing("LOGGING-LEVEL attribute getter");
      return new integer();
   }
   
   /**
    * Assigns the default logging level for all logging categories.
    * <p>
    * There are 5 valid levels:
    * <p>
    * <ul>
    *   <li> 0 - no logging
    *   <li> 1 - only errors
    *   <li> 2 - basic, each entry type determines its own output but all types will generate
    *            some output
    *   <li> 3 - verbose
    *   <li> 4 - extended
    * </ul>
    *
    * @param    level
    *           The logging level to use by default.
    */
   public void setLoggingLevel(int level)
   {
      UnimplementedFeature.missing("LOGGING-LEVEL attribute setter");
   }
   
   /**
    * Assigns the default logging level for all logging categories.
    * <p>
    * There are 5 valid levels:
    * <p>
    * <ul>
    *   <li> 0 - no logging
    *   <li> 1 - only errors
    *   <li> 2 - basic, each entry type determines its own output but all types will generate
    *            some output
    *   <li> 3 - verbose
    *   <li> 4 - extended
    * </ul>
    *
    * @param    level
    *           The logging level to use by default.
    */
   public void setLoggingLevel(integer level)
   {
      if (level == null || level.isUnknown())
      {
         UnimplementedFeature.todo("Confirm unknown behavior is correct.");
         return;
      }
      
      setLoggingLevel(level.intValue());
   }
   
   /**
    * Obtains the configured log file size in bytes at which the log file will be renamed and a
    * new log file started.
    *
    * @return   The maximum log file size.
    */
   public integer getLogThreshold()
   {
      UnimplementedFeature.missing("LOG-THRESHOLD attribute getter");
      return new integer();
   }
   
   /**
    * Obtains the configured number of log files that will be stored in total (including both
    * backed up log files and the current log).
    *
    * @return   The maximum number of log files.
    */
   public integer getNumLogFiles()
   {
      UnimplementedFeature.missing("NUM-LOG-FILES attribute getter");
      return new integer();
   }
   
   /**
    * Obtains the logical database name of database that the DSLOG-MANAGER is generating log
    * entries.
    *
    * @return   The logical database name or <code>unknown</code> if no database name has been
    *           assigned.
    */
   public character getDBContext()
   {
      UnimplementedFeature.missing("DB-CONTEXT attribute getter");
      return new character();
   }
   
   /**
    * Assigns the logical database name of database that the DSLOG-MANAGER is generating log
    * entries.
    *
    * @param    ldbname
    *           The logical database name.
    */
   public void setDBContext(String ldbname)
   {
      UnimplementedFeature.missing("DB-CONTEXT attribute setter");
   }
   
   /**
    * Assigns the logical database name of database that the DSLOG-MANAGER is generating log
    * entries.
    *
    * @param    ldbname
    *           The logical database name.
    */
   public void setDBContext(character ldbname)
   {
      if (ldbname == null || ldbname.isUnknown())
      {
         UnimplementedFeature.todo("Confirm unknown behavior is correct.");
         return;
      }
      
      setDBContext(ldbname.toStringMessage());
   }
   
   /**
    * Delete all messages from the currently open log file.
    *  
    * @return   <code>true</code> if the operation succeeded, or <code>false</code>
    *           otherwise.
    */
   public logical clearLog()
   {
      UnimplementedFeature.missing("CLEAR-LOG() method");
      return new logical();
   }
   
   /**
    * Close the currently open log file.  Additional logging will not be possible after this
    * operation.
    *  
    * @return   <code>true</code> if the operation succeeded, or <code>false</code>
    *           otherwise.
    */
   public logical closeLog()
   {
      UnimplementedFeature.missing("CLOSE-LOG() method");
      return new logical();
   }
   
   /**
    * Write a user defined message to the currently open log file using the default subsystem
    * name "APPL".
    *
    * @param    msg
    *           The message to write.
    *  
    * @return   <code>true</code> if the operation succeeded, or <code>false</code>
    *           otherwise.
    */
   public logical writeMessage(String msg)
   {
      return writeMessage(msg, DEFAULT_SUBSYSTEM);
   }
   
   /**
    * Write a user defined message to the currently open log file using the default subsystem
    * name "APPL".
    *
    * @param    msg
    *           The message to write.
    *  
    * @return   <code>true</code> if the operation succeeded, or <code>false</code>
    *           otherwise.
    */
   public logical writeMessage(character msg)
   {
      return writeMessage(msg, DEFAULT_SUBSYSTEM);
   }
   
   /**
    * Write a user defined message to the currently open log file using the given subsystem
    * name.
    *
    * @param    msg
    *           The message to write.
    * @param    subsys
    *           Subsystem name (up to 10 characters will be used).
    *  
    * @return   <code>true</code> if the operation succeeded, or <code>false</code>
    *           otherwise.
    */
   public logical writeMessage(String msg, String subsys)
   {
      UnimplementedFeature.missing("WRITE-MESSAGE() method");
      return new logical();
   }
   
   /**
    * Write a user defined message to the currently open log file using the given subsystem
    * name.
    *
    * @param    msg
    *           The message to write.
    * @param    subsys
    *           Subsystem name (up to 10 characters will be used).
    *  
    * @return   <code>true</code> if the operation succeeded, or <code>false</code>
    *           otherwise.
    */
   public logical writeMessage(character msg, String subsys)
   {
      if (msg == null || msg.isUnknown() || subsys == null)
      {
         UnimplementedFeature.todo("Confirm unknown behavior is correct.");
         return new logical();
      }
      
      return writeMessage(msg.toStringMessage(), subsys);
   }
   
   /**
    * Write a user defined message to the currently open log file using the given subsystem
    * name.
    *
    * @param    msg
    *           The message to write.
    * @param    subsys
    *           Subsystem name (up to 10 characters will be used).
    *  
    * @return   <code>true</code> if the operation succeeded, or <code>false</code>
    *           otherwise.
    */
   public logical writeMessage(String msg, character subsys)
   {
      if (msg == null || subsys == null || subsys.isUnknown())
      {
         UnimplementedFeature.todo("Confirm unknown behavior is correct.");
         return new logical();
      }
      
      return writeMessage(msg, subsys.toStringMessage());
   }
   
   /**
    * Write a user defined message to the currently open log file using the given subsystem
    * name.
    *
    * @param    msg
    *           The message to write.
    * @param    subsys
    *           Subsystem name (up to 10 characters will be used).
    *  
    * @return   <code>true</code> if the operation succeeded, or <code>false</code>
    *           otherwise.
    */
   public logical writeMessage(character msg, character subsys)
   {
      if (msg == null || msg.isUnknown() || subsys == null || subsys.isUnknown())
      {
         UnimplementedFeature.todo("Confirm unknown behavior is correct.");
         return new logical();
      }
      
      return writeMessage(msg.toStringMessage(), subsys.toStringMessage());
   }
   
   /**
    * Reports if this object is valid for use.  
    *
    * @return   Always <code>true</code>.
    */
   public boolean valid()
   {
      UnimplementedFeature.todo("Check is there are cases where this should return false " +
                                "(such as after the log file is closed).");
      return true;
   }
}