/*
** Module   : LowLevelSocket.java
** Abstract : An interface defining APIs specific to a client SOCKET.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CA  20130908 Created initial version.
** 002 HC  20131031 Changed resource ID data type to Long, see issue #2183.
** 003 OM  20151104 Added support for SSL-connection related errors.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.net.*;
import java.security.*;

/**
 * Interface defining APIs to connect or use a client socket in P2J. This interface is exported
 * as a service during P2J client startup.
 */
public interface LowLevelSocket
{
   /**
    * Establish a connection to the given host and port.
    * 
    * @param    resourceId
    *           The ID of the associated {@link SocketImpl} resource on server side.
    * @param    host
    *           The IP address or name of the target host.
    * @param    port
    *           The port number.
    * 
    * @return   The ID of the associated socket, if the connection was established.
    *           
    * @throws   IOException
    *           In case of errors while establishing the connection.
    */
   public int connect(long resourceId, String host, int port)
   throws IOException;

   /**
    * Establish a SSL connection to the given host and port.
    * 
    * @param    resourceId
    *           The ID of the associated {@link SocketImpl} resource on server side.
    * @param    host
    *           The IP address or name of the target host.
    * @param    port
    *           The port number.
    * @param    nosessionreuse
    *           Flag indicating that the ssl session ID must not be reused.
    * @param    nohostverify
    *           Flag indicating that the target host name must not be validated against the
    *           certificate's principal name.
    * 
    * @return   The ID of the associated socket, if the connection was established.
    *           
    * @throws   IOException
    *           In case of errors while establishing the connection.
    * @throws   GeneralSecurityException
    *           In case of errors while establishing the SSL connection.
    */
   public int connectSSL(long     resourceId, 
                         String  host,
                         int     port,
                         boolean nosessionreuse,
                         boolean nohostverify)
   throws IOException, 
          GeneralSecurityException;

   /**
    * Check if the java socket is still connected.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *           
    * @return   <code>true</code> if the java socket is still connected.
    */
   public boolean isConnected(int id);
   
   /**
    * Disconnect this socket.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *           
    * @return   <code>true</code> if no errors were encountered during disconnect.
    */
   public boolean disconnect(int id);
   
   /**
    * Get the available bytes to be read.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *
    * @return   See above.
    */
   public long getBytesAvailable(int id);
   
   /**
    * Java implementation for the READ() SOCKET method, this will read the specified number of 
    * bytes from the available data on the socket, and put it inside the memory buffer 
    * argument at the specified position.
    * <p>
    * Optionally the read mode can be set by the following values:
    * <ul>
    *    <li> 1 (READ-AVAILABLE)
    *    <li> 2 (READ-EXACT-NUM)
    * </ul>
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    * @param    buffer
    *           The memory buffer where read() puts the data.
    * @param    pos
    *           The position in the memory buffer where the data is added.
    * @param    len
    *           The number of bytes to read from the socket.
    * @param    mode
    *           The mode to set for the reading operation.
    *        
    * @return   The number of read bytes. 0 means a timeout during read, -1 meas the stream 
    *           reached EOF (socket closed).
    *           
    * @throws   IOException
    *           In case IO errors were encoubtered during read.
    */
   public long read(int id, memptr buffer, long pos, long len, int mode)
   throws IOException;

   /**
    * Java implementation for the WRITE() SOCKET method, this will write data from the specified 
    * buffer location (position and number of bytes) to the socket. 
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    * @param    buffer
    *           The memory buffer from which the data will be written to the socket.
    * @param    pos
    *           The position in the memory buffer from where the data to be taken starts.
    * @param    len
    *           The number of bytes to take from the memory buffer.
    *        
    * @return   The number of written bytes.
    *           
    * @throws   IOException
    *           In case IO errors were encoubtered during write.
    */
   public long write(int id, memptr buffer, long pos, long len)
   throws IOException;
   
   /**
    * In case of a SSL connection, get the digital certificate subject name of the server for
    * the current SSL session; else, return <code>null</code>.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    */
   public String getSslServerName(int id);
   
   /**
    * Get the TCP_NODELAY option value for this socket. 
    * See {@link java.net.Socket#getTcpNoDelay} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    * 
    * @throws   SocketException
    *           If the socket option could not be interrogated.
    */
   public boolean getTcpNoDelay(int id)
   throws SocketException;

   /**
    * Get the SO_LINGER option value for this socket.
    * See {@link java.net.Socket#getSoLinger} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    * 
    * @throws   SocketException
    *           If the socket option could not be interrogated.
    */
   public int getSoLinger(int id)
   throws SocketException;

   /**
    * Get the SO_KEEPALIVE option value for this socket.
    * See {@link java.net.Socket#getKeepAlive} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    * 
    * @throws   SocketException
    *           If the socket option could not be interrogated.
    */
   public boolean getKeepAlive(int id)
   throws SocketException;

   /**
    * Get the SO_REUSEADDR option value for this socket.
    * See {@link java.net.Socket#getReuseAddress} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    * 
    * @throws   SocketException
    *           If the socket option could not be interrogated.
    */
   public boolean getReuseAddress(int id)
   throws SocketException;

   /**
    * Get the SO_RCVBUF option value for this socket.
    * See {@link java.net.Socket#getReceiveBufferSize} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    * 
    * @throws   SocketException
    *           If the socket option could not be interrogated.
    */
   public int getReceiveBufferSize(int id)
   throws SocketException;

   /**
    * Get the SO_SNDBUF option value for this socket.
    * See {@link java.net.Socket#getSendBufferSize} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    * 
    * @throws   SocketException
    *           If the socket option could not be interrogated.
    */
   public int getSendBufferSize(int id)
   throws SocketException;

   /**
    * Get the SO_RCVTIMEO option value for this socket.
    * See {@link java.net.Socket#getSoTimeout} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    *            
    * @return   See above.
    * 
    * @throws   SocketException
    *           If the socket option could not be interrogated.
    */
   public int getSoTimeout(int id)
   throws SocketException;

   /**
    * Set the TCP_NODELAY option for this socket.
    * See {@link java.net.Socket#setTcpNoDelay} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side. 
    * @param    on
    *           The TCP_NODELAY value.
    * 
    * @throws   SocketException
    *           If the socket option could not be set.
    */
   public void setTcpNoDelay(int id, boolean on)
   throws SocketException;

   /**
    * Set the SO_LINGER option for this socket.
    * See {@link java.net.Socket#setSoLinger} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side. 
    * @param    on
    *           <code>true</code> to turn on the SO_LINGER option.
    * @param    linger
    *           The linger time.
    * 
    * @throws   SocketException
    *           If the socket option could not be set.
    */
   public void setSoLinger(int id, boolean on, int linger)
   throws SocketException;

   /**
    * Set the SO_KEEPALIVE option for this socket.
    * See {@link java.net.Socket#setKeepAlive} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side. 
    * @param    on
    *           The SO_KEEPALIVE value.
    * 
    * @throws   SocketException
    *           If the socket option could not be set.
    */
   public void setKeepAlive(int id, boolean on)
   throws SocketException;

   /**
    * Set the SO_REUSEADDR option for this socket.
    * See {@link java.net.Socket#setReuseAddress} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side. 
    * @param    on
    *           The SO_REUSEADDR value.
    * 
    * @throws   SocketException
    *           If the socket option could not be set.
    */
   public void setReuseAddress(int id, boolean on)
   throws SocketException;

   /**
    * Set the SO_RCVBUF value (the receive buffer size) for this socket.
    * See {@link java.net.Socket#setReceiveBufferSize} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side. 
    * @param    size
    *           The SO_RCVBUF value.
    * 
    * @throws   SocketException
    *           If the socket option could not be set.
    */
   public void setReceiveBufferSize(int id, int size)
   throws SocketException;

   /**
    * Set the SO_SNDBUF value (the send buffer size) for this socket.
    * See {@link java.net.Socket#setSendBufferSize} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side. 
    * @param    size
    *           The SO_SNDBUF value.
    * 
    * @throws   SocketException
    *           If the socket option could not be set.
    */
   public void setSendBufferSize(int id, int size)
   throws SocketException;

   /**
    * Set the SO_RCVTIMEO value (the number of seconds the socket waits to receive data before 
    * timing out) for this socket.
    * See {@link java.net.Socket#setSoTimeout} for more details.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side. 
    * @param    timeout
    *           The SO_RCVTIMEO value.
    * 
    * @throws   SocketException
    *           If the socket option could not be set.
    */
   public void setSoTimeout(int id, int timeout)
   throws SocketException;

   /**
    * Get the IP address of this side of this connection.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    * 
    * @return   See above.
    */
   public String getLocalHost(int id);

   /**
    * Get the port number of this side of this connection.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    * 
    * @return   See above.
    */
   public int getLocalPort(int id);

   /**
    * Get the host IP address of the remote side of this connection.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    * 
    * @return   See above.
    */
   public String getRemoteHost(int id);

   /**
    * Get the port number of the remote side of this connection.
    * 
    * @param    id
    *           The ID of the associated socket on P2J Client side.
    * 
    * @return   See above.
    */
   public int getRemotePort(int id);
}