/*
** Module : LowLevelStream.java
** Public : low level stream primitives in a form that is suitable for use remotely
**
** Copyright (c) 2006-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 GES 20060211   @24551 Created initial version which provides low level stream
**                           primitives in a form that is suitable for use remotely.
** 002 GES 20060217   @24676 Added writeByte().
** 003 GES 20060226   @24754 Added methods to remote the paging state.
** 004 GES 20060817   @28658 Redirected cleanup processing to the client.
** 005 NVS 20070725   @34684 Added isIn() and isOut() methods.
** 006 NVS 20070731   @34753 Added advancePage() method.
** 007 NVS 20070802   @34774 Added flush() method.
** 008 SIY 20070928   @35321 Added more methods. Ref: #35314
** 009 SIY 20071009   @35405 Added setDelimiter methods. Ref: #35404
** 010 SIY 20071017   @35487 Added setControl() method. Ref: #35486
** 011 CA  20071228   @36615 Added method putWorkerUnformatted.
** 012 CA  20080304   @37293 Added method setUnnamed.
** 013 GES 20080412   @38016 Added method setConvert().
** 014 GES 20080425   @38121 Refactored to eliminate round trips to the client by passing more
**                           data in a single invocation. This improves performance.
** 015 SVL 20090612   @42675 Added setUnbuffered() method. 
** 016 SIY 20090925   @44030 Added getNextLineNum() method.
** 017 SVL 20110403          Added write(byte[]) function.
** 018 GES 20110812          Added readChunk() method.
** 019 GES 20141030          Added setConvertSource/Target() methods.
** 020 OM  20170613          Added setBinary() method. Added peekCh() method.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.io.*;

/**
 * An interface providing low level stream primitives in a form that is
 * suitable for use remotely.  In particular, all methods take an integral
 * identifier which specifies the stream that is the target of the given
 * operation.
 *
 * @author    GES
 */
public interface LowLevelStream
{
   /**
    * The number of bytes available to be immediately read without blocking.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    *
    * @return   The number of available bytes.
    * 
    * @throws   IOException
    *           if an I/O error occurs. 
    */
   public long available(int id)
   throws IOException;
   
   /**
    * The 0-based offset into the stream at which the next read or write will 
    * occur.  This does not work for "streams" that require sequential
    * access.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    *
    * @return   The current position in the stream.
    * 
    * @throws   UnsupportedOperationException 
    *           If the requested operation is not supported.
    * @throws   IOException
    *           If an I/O error occurs. 
    */
   public long getPos(int id)
   throws UnsupportedOperationException,   
          IOException;
   
   /**
    * Moves the current read/write position to the specified absolute 0-based
    * offset.  
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    * @param    pos
    *           The new read/write position in the stream.
    * 
    * @throws   UnsupportedOperationException 
    *           If the requested operation is not supported.
    * @throws   IOException
    *           If an I/O error occurs. 
    */
   public void setPos(int id, long pos)
   throws UnsupportedOperationException,   
          IOException;
   
   /**
    * The length of the stream in bytes.  This does not work for "streams"
    * that require sequential access.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    *
    * @return   The length of the stream.
    * 
    * @throws   UnsupportedOperationException 
    *           If the requested operation is not supported.
    * @throws   IOException
    *           If an I/O error occurs. 
    */
   public long getLen(int id)
   throws UnsupportedOperationException,   
          IOException;
   
   /**
    * Truncates or extends the stream to the specified length if this stream
    * supports such an operation. If truncation is requested, all data
    * located after this point in the file is discarded.  If extending the
    * file is requested, the values of the data in the extended portion of
    * the file is undefined. 
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    * @param    len
    *           The new length of the file.
    * 
    * @throws   UnsupportedOperationException 
    *           If the requested operation is not supported.
    * @throws   IOException
    *           If an I/O error occurs. 
    */
   public void setLen(int id, long len)
   throws UnsupportedOperationException,   
          IOException;
   
   /**
    * Write the given character to the output stream.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    * @param    ch
    *           The character to be written.
    * 
    * @throws   IOException
    *           If an I/O error occurs. 
    */
   public void writeCh(int id, char ch)
   throws IOException;
   
   /**
    * Write the given byte to the output stream.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    * @param    b
    *           The byte to be written.
    * 
    * @throws   IOException
    *           If an I/O error occurs. 
    */
   public void writeByte(int id, byte b)
   throws IOException;
   
   /**
    * Write the given string to the output stream.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    * @param    data
    *           The data to be written.
    * 
    * @throws   IOException
    *           If an I/O error occurs. 
    */
   public void write(int id, String data)
   throws IOException;

   /**
    * Write the given byte array to the output stream.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    * @param    data
    *           The data to be written.
    *
    * @throws   IOException
    *           If an I/O error occurs.
    */
   public void write(int id, byte[] data)
   throws IOException;
   
   /**
    * Peeks at the character from the current read position in the stream (reads a character from
    * the current read position in the stream without incrementing stream read position. The next
    * {@code peekCh()} and {@code readCh()} will return the same value).
    * <p>
    * The underlying stream subclass determines the content of the result. Byte oriented streams
    * such as pipes or files will return a byte while streams that generate keystrokes or
    * characters may return a DBCS or Unicode character.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    *
    * @return  The next character read from the stream, -1 on any failure and -2 upon an
    *          {@code EOF}.
    */
   public int peekCh(int id);
   
   /**
    * Read a character from the current read position in the stream.
    * <p>
    * The underlying stream subclass determines the content of the result. 
    * Byte oriented streams such as pipes or files will return a byte while
    * streams that generate keystrokes or characters may return a DBCS or
    * Unicode character.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    *
    * @return   The next character read from the stream, -1 on any failure
    *           and -2 upon an <code>EOF</code>.
    */
   public int readCh(int id);
   
   /**
    * Read all characters from the current read position in the stream to the
    * next line separator (as determined by the <code>File.separator</code>
    * or to the <code>EOF</code>. Any line separator character(s) and the
    * <code>EOF</code> character are not returned.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    *
    * @return   The next line read from the stream.
    * 
    * @throws   IOException
    *           If an I/O error occurs. 
    * @throws   EOFException 
    *           If this input stream reaches the end before reading all the 
    *           bytes.
    * @throws   InterruptedException
    *           If any thread interrupted the current thread before or while 
    *           the current thread was waiting for a notification. 
    */
   public String readLn(int id)
   throws EOFException,   
          IOException,
          InterruptedException;
   
   /**
    * Read a chunk of bytes from the current read position in the stream to the
    * offset calculated by given length or to the <code>EOF</code>, whichever
    * comes first. The <code>EOF</code> character is not returned.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    * @param    len
    *           The number of bytes to read.
    *
    * @return   The chunk read from the stream or <code>null</code> if there
    *           is nothing to read (the <code>EOF</code> has been reached). The
    *           length of the returned array will be the smaller of the
    *           <code>len</code> parameter or the actual bytes left before the
    *           <code>EOF</code>.
    * 
    * @throws   IOException
    *           If an I/O error occurs. 
    */
   public byte[] readChunk(int id, int len)
   throws IOException;
   
   /**
    * Closes the input stream and releases OS resources associated with it.
    * Any output stream that is open is unaffected.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    */
   public void closeIn(int id);
   
   /**
    * Closes the output stream and releases OS resources associated with it.
    * Any input stream that is open is unaffected.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    */
   public void closeOut(int id);
   
   /**
    * State of the input side of the stream.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    *
    * @return   <code>true</code> if the input side of the stream is active.
    */
   public boolean isIn(int id);
   
   /**
    * State of the output side of the stream.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    *
    * @return   <code>true</code> if the output side of the stream is active.
    */
   public boolean isOut(int id);

   /**
    * Closes the stream and releases OS resources associated with it.  After
    * completion, further calls may not be made to use this instance, until
    * an <code>assign</code> is used to "re-open" the stream. This is safe to
    * call when no stream is open (it will do nothing in that case).
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    */
   public void close(int id);   
   
   /**
    * Inserts a page break if the stream is paged and if the next write
    * position is not on the first line of a new page.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    *           
    * @throws   ErrorConditionException
    *           On any error during I/O.
    */
   public void advancePage(int id)
   throws ErrorConditionException;

   /**
    * Flushes the stream.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    */
   public void flush(int id);
   
   /**
    * Detects when a page break is needed.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    *
    * @return   <code>true</code> if a page break should be output.
    */
   public boolean needsPageBreak(int id);
   
   /**
    * Gets the current page size in lines.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    *
    * @return   The number of lines per page.
    */
   public int rawGetPageSize(int id);
   
   /**
    * Sets the current page size in lines.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    * @param    sz
    *           The number of lines per page.
    */
   public void rawSetPageSize(int id, int sz);
   
   /**
    * Gets the current page number.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    *
    * @return   The current page number.
    */
   public int rawGetPageNum(int id);
   
   /**
    * Increments the current page number.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    */
   public void rawIncrementPageNum(int id);
   
   /**
    * Gets the current page's next line number.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    *
    * @return   The next line number.
    */
   public int rawGetNextLineNum(int id);

   /**
    * Gets the current page's next logical line number.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    *
    * @return   The next line number.
    */
   public int getNextLineNum(int id);
   
   /**
    * Increments the current page's next line number.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    */
   public void rawIncrementNextLineNum(int id);
   
   /**
    * Resets the current page's next line number to 1.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    */
   public void rawResetNextLineNum(int id);
   
   /**
    * Perform common close <b>preprocessing</b> for all streams.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    * @param    in
    *           <code>true</code> if this was triggered during a call to 
    *           {@link #closeIn} or {@link #close}.  Input resources are
    *           to be cleaned up.
    * @param    out
    *           <code>true</code> if this was triggered during a call to 
    *           {@link #closeOut} or {@link #close}.  Output resources are
    *           to be cleaned up.
    *           
    * @throws   IOException
    *           On any error during I/O.           
    */
   public void processCleanup(int id, boolean in, boolean out)
   throws IOException;

   /**
    * Write the current output buffer to the stream. This flushes all 
    * buffered output of the {@link Stream#putField}, {@link Stream#putSpace},
    * {@link Stream#putLineEnd} and {@link Stream#writeField} methods to the 
    * stream. This method also resets the current delimiter to the default.
    * <p>
    * This method does nothing if the stream is in <code>control</code> mode.
    * 
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    * @param    appendNewLine
    *           Automatically add new line at the end of output.
    *           
    * @throws   ErrorConditionException
    *           On any error during I/O.
    * @throws   StopConditionException
    *           If any interruption occurs during I/O processing.
    */
   public void endOutput(int id, boolean appendNewLine)
   throws ErrorConditionException,
          StopConditionException;

   /**
    * Output a list of fields to the stream. 
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    * @param    data
    *           The list of fields to be output.
    * @param    mode
    *           <code>MODE_PUT_NORMAL</code>, <code>MODE_PUT_CONTROL</code>,
    *           <code>MODE_PUT_UNFORMATTED</code> or <code>MODE_EXPORT</code>.
    * @param    delim
    *           The delimiter to use in EXPORT mode.
    *           
    * @throws   ErrorConditionException
    *           On any error during I/O.
    */
   public void putWorker(int id, FieldEntry[] data, int mode, char delim)
   throws ErrorConditionException;

   /**
    * Set the unnamed status for the specified stream.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    * @param    unnamed
    *           <code>true</code> if it is an unnamed stream.
    */
   public void setUnnamed(int id, boolean unnamed);

   /**
    * Set the specified stream UNBUFFERED.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    */
   public void setUnbuffered(int id);
   
   /**
    * Sets the stream's character conversion mode.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    * @param    convert
    *           The new character conversion mode.  <code>true</code> to
    *           enable character conversion, <code>false</code> to disable
    *           conversion.
    */
   public void setConvert(int id, boolean convert);
   
   /**
    * Sets the stream's source codepage and activates character conversion mode.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    * @param    cp
    *           The source codepage for character conversion mode.
    */
   public void setConvertSource(int id, String cp);
   
   /**
    * Sets the stream's target codepage and activates character conversion mode.
    *
    * @param    id
    *           Specifies the stream on which this operation is to execute.
    * @param    cp
    *           The target codepage for character conversion mode.
    */
   public void setConvertTarget(int id, String cp);
   
   /**
    * Sets the stream into binary mode which disables character translations during I/O
    * operations. It is important to note that field and line reading (<code>import</code>,
    * <code>set</code>, <code>update</code> <code>prompt-for</code> and <code>import
    * unformatted</code>) are unaffected by this setting (this is based on Progress testcases).
    *
    * @param   id
    *          Specifies the stream on which this operation is to execute.
    */
   public void setBinary(int id);
}
