/*
** Module : RemoteErrorData.java
** Public : provides remote access to ErrorManager data
**
** Copyright (c) 2005-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description-----------------------------
** 001 NVS 20060303   @24884 Created initial version.
** 002 GES 20080624   @38915 Added support for an ignore flag. Changed
**                           some method names to match coding standards.
** 003 GES 20090924   @44009 Added headless mode flag.
** 004 CA  20130920          if ERROR attribute is set to false by ErrorManager.recordOrShowError,
**                           than do not clean the error message list.
** 005 CA  20131219          Renamed isError to isErrorFlag.
** 006 VIG 20140701          Removed setDontClean method and added forwardPending method. 
**                           Removed clearList and setError methods.
** 007 VIG 20140709          Added getRaisedConditions method. 
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.util.*;

/**
 * Describes methods available for the use in the ErrorManager to access
 * the data residing on the server from both the server and client uniformly.
 *
 * @author NVS
 */
public interface RemoteErrorData
{
   /**
    * Queries the state of <code>silent</code> variable.
    *
    * @return   <code>true</code> if silent mode is enabled, otherwise
    *           <code>false</code>.
    */
   public boolean isSilent();    

   /**
    * Sets the <code>silent</code> variable.
    *
    * @param    value
    *           The new state of the variable.
    */
   public void setSilent(boolean value);    

   /**
    * Queries the state of <code>ignore</code> variable.
    *
    * @return   <code>true</code> if silent mode is enabled, otherwise
    *           <code>false</code>.
    */
   public boolean isIgnore();    

   /**
    * Sets the <code>ignore</code> variable.
    *
    * @param    value
    *           The new state of the variable.
    */
   public void setIgnore(boolean value);    

   /**
    * Queries the state of <code>error</code> variable.
    *
    * @return   <code>true</code> if error flag is turned on, otherwise
    *           <code>false</code>.
    */
   public boolean isErrorFlag();
   
   /**
    * Queries the state of <code>pendingError</code> variable.
    *
    * @return   <code>true</code> if pending error flag is turned on, otherwise
    *           <code>false</code>.
    */
   public boolean isPending();    

   /**
    * Sets the <code>pendingError</code> variable.
    *
    * @param    value
    *           The new state of the variable.
    */
   public void setPending(boolean value);

   /**
    * Determines if headless mode should be supported. In headless mode, no
    * client-side output will occur.
    *
    * @return   <code>true</code> if the headless flag is turned on, otherwise
    *           <code>false</code>.
    */
   public boolean isHeadless();

   /**
    * Gets the size of the list of error records.
    *
    * @return   The size of the list.
    */
   public int getSize();    

   /**
    * Adds a record to the list of error records.
    *
    * @param    record
    *           The object to add to the list of error records.
    */
   public void addRecord(Object record);    

   /**
    * Gets a record from the list of error records.
    *
    * @param    index
    *           The record index.
    *
    * @return   The object associated with the given index in the list.
    */
   public Object getRecord(int index);
   
   /**
    * Makes pending error status current. 
    */
   public void forwardPending();
   
   /**
    * Gets err nums for raised errors.
    * 
    * @return   stack of errors numbers
    */
   public LinkedList<Integer> getRaisedConditions();
}