/*
** Module   : SOAPFaultImpl.java
** Abstract : Implementation for the SOAP-FAULT resource.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CA  20130216 Created initial version.
** 002 CA  20130221 Added Deletable support. Fixed NAME and TYPE support.
** 003 OM  20130304 Refactored isValid and isUnknown of WrappedResource to valid and unknown.
** 004 CA  20130927 Removed setResourceType, as the type is determined from LegacyResource 
**                  annotation.
** 005 CA  20140104 Added runtime implementation for this resource.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

/**
 * A class for implementing the SOAP-FAULT method and attributes.
 */
public class SOAPFaultImpl
extends HandleResource
implements SOAPFault,
           Deletable
{
   /** The URI of the Web service actor that caused this SOAP-FAULT. */
   private final String faultActor;
   
   /** The SOAP-FAULT-CODE of this SOAP-FAULT object. */
   private final String faultCode;
   
   /** The SOAP-FAULT-STRING of this SOAP-FAULT object. */
   private final String faultString;
   
   /** The SOAP-FAULT-DETAIL associated with this SOAP-FAULT object. */
   private final SOAPFaultDetailImpl faultDetail;
   
   /** Flag indicating this resource was deleted or not. */
   private boolean deleted = false;
   
   /**
    * Create a new SOAP Fault instance.
    * 
    * @param    faultActor
    *           The URI of the Web service actor that caused this SOAP-FAULT.
    * @param    faultCode
    *           The SOAP-FAULT-CODE of this SOAP-FAULT object.
    * @param    faultString
    *           The SOAP-FAULT-STRING of this SOAP-FAULT object.
    * @param    faultDetail
    *           The SOAP-FAULT-STRING of this SOAP-FAULT object.
    */
   SOAPFaultImpl(String              faultActor,
                 String              faultCode,
                 String              faultString,
                 SOAPFaultDetailImpl faultDetail)
   {
      this.faultActor = faultActor;
      this.faultCode = faultCode;
      this.faultString = faultString;
      this.faultDetail = faultDetail;
   }
   
   /**
    * The URI of the Web service actor that caused this SOAP-FAULT.
    * 
    * @return   See above.
    */
   public character getFaultActor()
   {
      return new character(faultActor);
   }

   /**
    * The SOAP-FAULT-CODE of this SOAP-FAULT object.
    * 
    * @return   See above.
    */
   public character getFaultCode()
   {
      return new character(faultCode);
   }

   /**
    * Returns the handle of the SOAP-FAULT-DETAIL associated with this SOAP-FAULT object.
    * 
    * @return   See above.
    */
   public handle getFaultDetail()
   {
      return new handle(faultDetail);
   }

   /**
    * The SOAP-FAULT-STRING of this SOAP-FAULT object.
    * 
    * @return   See above.
    */
   public character getFaultString()
   {
      return new character(faultString);
   }
   
   /**
    * Reports if this object is valid for use.  
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   public boolean valid()
   {
      return !deleted;
   }

   /**
    * Delete this resource. This just sets the {@link #deleted} flag to {@code true}.
    * <p>
    * The referenced {@link #faultDetail} will be deleted too.
    */
   @Override
   public void delete()
   {
      if (deleted)
      {
         return;
      }

      deleted = true;
      
      if (faultDetail != null)
      {
         // the fault detail is deleted if the parent soap fault is deleted.
         faultDetail.delete();
      }
   }
}