/*
** Module   : TableParameter.java
** Abstract : Container for table handle and calling-side APPEND option.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------Description---------------------------------------------
** 001 SVL 20130226  Created initial version.
** 002 SVL 20140106  Added table handle variable.
** 003 CA  20140110  Added support for usage from java side (when there is no DMO passed).
** 004 SVL 20140414  Added support for remote parameters.
** 005 SVL 20140611  Do not allow tableHandle to be null.
** 006 SVL 20140709  Added parameterIndex.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.persist.*;

/**
 * Container for table handle and calling-side APPEND option. Serves two purposes: 1. passes
 * calling-side APPEND option to the called side; 2. implements interchangeability between TABLE
 * and TABLE-HANDLE parameters (In 4GL you can pass TABLE parameter to a procedure that takes
 * TABLE-HANDLE and vice versa).
 */
public class TableParameter
{
   /** Buffer of the wrapped table. */
   private final Temporary table;

   /** Table handle. */
   private final handle tableHandle;

   /** Corresponds calling-side APPEND option. */
   private final boolean append;

   /** Calling- or called-side result set (for remote parameters). */
   private TableWrapper resultSet = null;

   /** Determines if this is a remote (passed to appserver) parameter. */
   private boolean remoteParameter = false;

   /**
    * 1-based index of the table parameter (in declaration of the function). <code>0</code>
    * if it is not defined.
    */
   private int parameterIndex = 0;

   /**
    * Create an instance using a buffer.
    *
    * @param table
    *        Buffer of the wrapped table.
    */
   public TableParameter(Temporary table)
   {
      this(table, false);
   }

   /**
    * Create an instance using a buffer.
    *
    * @param table
    *        Buffer of the wrapped table.
    * @param append
    *        Determines if APPEND option was specified on the calling side.
    */
   public TableParameter(Temporary table, boolean append)
   {
      this.table = table;
      this.tableHandle = (table == null) ? new handle() 
                                         : ((BufferReference) table).buffer().tableHandle();
      this.append = append;
   }

   /**
    * Create an instance using a table handle.
    *
    * @param tableHandle
    *        Handle to the wrapped table.
    */
   public TableParameter(handle tableHandle)
   {
      this(tableHandle, false);
   }

   /**
    * Create an instance using a table handle.
    *
    * @param tableHandle
    *        Handle to the wrapped table.
    * @param append
    *        Determines if APPEND option was specified on the calling side.
    */
   public TableParameter(handle tableHandle, boolean append)
   {
      this.table = TemporaryBuffer.getDefaultBuffer(tableHandle);
      this.tableHandle = tableHandle;
      this.append = append;
   }

   /**
    * Create an instance using a result set (for remote parameters).
    *
    * @param resultSet
    *        Result set containing input data.
    * @param append
    *        Determines if APPEND option was specified on the calling side.
    */
   public TableParameter(TableWrapper resultSet, boolean append)
   {
      this.table = null;
      this.tableHandle = new handle();
      this.resultSet = resultSet;
      this.append = append;
      this.remoteParameter = true;
   }

   /**
    * Determines if this is a remote parameter.
    *
    * @return <code>true</code> if this is a remote parameter, <code>false</code> if this is a
    *         local parameter.
    */
   public boolean isRemoteParameter()
   {
      return remoteParameter;
   }

   /**
    * Get result set (for remote parameters).
    *
    * @return Result set.
    */
   public TableWrapper getResultSet()
   {
      return resultSet;
   }

   /**
    * Set result set (for remote parameters).
    *
    * @param resultSet
    *        Result set, normally containing output data.
    */
   public void setResultSet(TableWrapper resultSet)
   {
      this.resultSet = resultSet;
   }

   /**
    * Get the handle which references the wrapped table.
    *
    * @return  handle to the wrapped table.
    */
   public handle getTableHandle()
   {
      return tableHandle;
   }

   /**
    * Get the buffer of the wrapped table.
    *
    * @return  buffer of the wrapped table.
    */
   public Temporary getTable()
   {
      if (table == null)
      {
         AbstractTempTable tt = (AbstractTempTable) tableHandle.getResource();
         return (tt == null) ? null : (Temporary) tt.defaultBufferHandle().getResource();
      }
      else
      {
         return table;
      }
   }

   /**
    * Determines if the APPEND option was specified on the calling side.
    *
    * @return <code>true</code> if the APPEND option was specified on the calling side.
    */
   public boolean isAppend()
   {
      return append;
   }

   /**
    * Get 1-based index of the table parameter (in declaration of the function).
    *
    * @return 1-based index of the table parameter or <code>0</code> if it is not defined.
    */
   public int getParameterIndex()
   {
      return parameterIndex;
   }

   /**
    * Set 1-based index of the table parameter (in declaration of the function).
    *
    * @param parameterIndex
    *        1-based index of the table parameter.
    */
   public void setParameterIndex(int parameterIndex)
   {
      this.parameterIndex = parameterIndex;
   }
}