/*
** Module   : TypeFactory.java
** Abstract : defines APIs to create 4GL-compatible instances and initialize parameters.
**
** Copyright (c) 2016-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20160721 Created initial version.  Added APIs to create 4GL-compatible instances, with
**                  or without initializers, with or without extent clauses.  Also, there are
**                  APIs to initialize function/procedure parameters.
** 002 SVL 20170126 Added initializer for COM handle type.
** 003 OM  20170807 Added missing methods related to comhandles. 
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.util.function.Function;

/**
 * Provides APIs to create or initialize or 4GL-compatbile instance.
 */
public class TypeFactory
{
   /** Function to create and initialize a {@link date} instance. */
   private static Function<date, date> DATE_INIT = (d) -> (new date(d));

   /** Function to create and initialize a {@link datetime} instance. */
   private static Function<datetime, datetime> DATETIME_INIT = (d) -> (new datetime(d));

   /** Function to create and initialize a {@link datetimetz} instance. */
   private static Function<datetimetz, datetimetz> DATETIMETZ_INIT = (d) -> (new datetimetz(d));

   /** Function to create and initialize a {@link logical} instance. */
   private static Function<Boolean, logical> BOOLEAN_INIT = (d) -> (d == null ? new logical() 
                                                                              : new logical(d));

   /** Function to create and initialize a {@link decimal} instance. */
   private static Function<decimal, decimal> DECIMAL_INIT = (d) -> (new decimal(d));

   /** Function to create and initialize a {@link integer} instance. */
   private static Function<Long, integer> INTEGER_INIT = (d) -> (d == null ? new integer() 
                                                                           : new integer(d));

   /** Function to create and initialize a {@link int64} instance. */
   private static Function<Long, int64> INT64_INIT = (d) -> (d == null ? new int64() 
                                                                       : new int64(d));

   /** Function to create and initialize a {@link character} instance. */
   private static Function<String, character> CHARACTER_INIT = (d) -> (d == null ? new character() 
                                                                                 : new character(d));

   /** Function to create and initialize a {@link longchar} instance. */
   private static Function<String, longchar> LONGCHAR_INIT = (d) -> (d == null ? new longchar() 
                                                                               : new longchar(d));

   /** Function to create and initialize a {@link recid} instance. */
   private static Function<Long, recid> RECID_INIT = (d) -> (d == null ? new recid() 
                                                                       : new recid(d));
   
   /**
    * Initialize an <code>INPUT</code> parameter by creating a deep copy of the specified instance.
    * 
    * @param    var
    *           The original instance receive as an argument.
    *           
    * @return   A copy of the given variable.
    * 
    * @see BaseDataType#deepCopy()
    */
   public static <T extends BaseDataType> T initInput(T var)
   {
      T newVar = (T) var.deepCopy();
      
      return newVar;
   }

   /**
    * Initialize an <code>INPUT</code> extent parameter by creating a deep copy of the specified 
    * array.
    * 
    * @param    var
    *           The original array instance receive as an argument.
    *           
    * @return   A copy of the given array.
    * 
    * @see ArrayAssigner#copyOf(BaseDataType[])
    */
   public static <T extends BaseDataType> T[] initInput(T[] var)
   {
      T[] newVar = ArrayAssigner.copyOf(var);
      
      return newVar;
   }
   
   /**
    * Initialize an <code>OUTPUT</code> parameter by setting it to its default value.
    * 
    * @param    var
    *           The original instance receive as an argument.
    */
   public static <T extends BaseDataType> void initOutput(T var)
   {
      initOutput(true, var);
   }
   
   /**
    * Initialize an <code>OUTPUT</code> parameter by setting it either to its default value or
    * to <code>unknown</code>, depending on the state of the <code>initDefault</code> flag.
    * 
    * @param    initDefault
    *           When <code>true</code>, initialize to default init value, otherwise to unknown.
    * @param    var
    *           The original instance receive as an argument.
    */
   public static <T extends BaseDataType> void initOutput(boolean initDefault, T var)
   {
      if (initDefault)
      {
         var.assign(var.instantiateDefault());
      }
      else
      {
         var.setUnknown();
      }
   }
   
   /**
    * Initialize an <code>OUTPUT</code> parameter by setting it to the specified init value.
    * 
    * @param    var
    *           The original instance receive as an argument.
    * @param    init
    *           The init instance.
    */
   public static <T extends BaseDataType> void initOutput(T var, T init)
   {
      var.assign(init);
   }

   /**
    * Initialize an <code>OUTPUT</code> extent parameter.
    * 
    * @param    var
    *           The original instance receive as an argument.
    *           
    * @return   The associated array.
    * 
    * @see OutputExtentParameter#initParameter()
    */
   public static <T extends BaseDataType> T[] initOutput(OutputExtentParameter<T> var)
   {
      T[] newVar = var.initParameter();
      
      return newVar;
   }

   /**
    * Initialize an <code>OUTPUT</code> extent parameter using the specified initializer and
    * extent value.
    * 
    * @param    var
    *           The original instance receive as an argument.
    * @param    extent
    *           The array length.
    * @param    init
    *           The initializer array.
    *           
    * @return   The associated array.
    * 
    * @see #initExtent(BaseDataType[], Object[], Function)
    */
   public static <T extends BaseDataType> T[] initOutput(OutputExtentParameter<T> var, 
                                                         int                      extent, 
                                                         T...                     init)
   {
      T[] newVar = var.initParameter(extent);
      
      if (extent == 0)
      {
         return newVar;
      }

      if (init == null || init.length == 0)
      {
         BaseDataType.initializeDefaultExtent(newVar);
      }
      else
      {
         T last = init[init.length - 1];
         for (int i = 0; i < Math.min(extent, init.length); i++)
         {
            newVar[i].assign(init[i]);
         }
         for (int i = init.length; i < extent; i++)
         {
            newVar[i].assign(last);
         }
      }
      
      return newVar;
   }
   
   /**
    * Initialize an <code>INPUT-OUTPUT</code> extent parameter.
    * 
    * @param    var
    *           The original instance receive as an argument.
    *           
    * @return   The associated array.
    * 
    * @see InputOutputExtentParameter#initParameter()
    */
   public static <T extends BaseDataType> T[] initInputOutput(InputOutputExtentParameter<T> var)
   {
      T[] newVar = var.initParameter();
      
      return newVar;
   }
   
   /**
    * Initialize an <code>INPUT-OUTPUT</code> extent parameter.
    * 
    * @param    var
    *           The original instance receive as an argument.
    * @param    extent
    *           The array length.
    *           
    * @return   The associated array.
    * 
    * @see InputOutputExtentParameter#initParameter(int)
    */
   public static <T extends BaseDataType> T[] initInputOutput(InputOutputExtentParameter<T> var, 
                                                              int                           extent)
   {
      T[] newVar = var.initParameter(extent);
      
      return newVar;
   }

   /**
    * Create a {@link memptr} instance initialized to the default value.
    * 
    * @return   See above.
    */
   public static memptr memptr()
   {
      memptr res = new memptr();
      
      return res;
   }
   
   /**
    * Create a {@link memptr} array of the given extent and initialize it with the default value.
    * 
    * @param    extent
    *           The array length.
    * 
    * @return   See above.
    */
   public static memptr[] memptrExtent(int extent)
   {
      memptr[] res = new memptr[extent];
      
      BaseDataType.initializeDefaultExtent(res);
      
      return res;
   }
   
   /**
    * Create a 0-length {@link memptr} array.
    * 
    * @return   See above.
    */
   public static memptr[] memptrExtent()
   {
      memptr[] res = new memptr[0];
      
      return res;
   }

   /**
    * Create a {@link raw} instance initialized to the default value.
    * 
    * @return   See above.
    */
   public static raw raw()
   {
      raw res = new raw();
      
      return res;
   }

   /**
    * Create a {@link raw} array of the given extent and initialize it with the default value.
    * 
    * @param    extent
    *           The array length.
    * 
    * @return   See above.
    */
   public static raw[] rawExtent(int extent)
   {
      raw[] res = new raw[extent];
      
      BaseDataType.initializeDefaultExtent(res);
      
      return res;
   }

   /**
    * Create a 0-length {@link raw} array.
    * 
    * @return   See above.
    */
   public static raw[] rawExtent()
   {
      raw[] res = new raw[0];
      
      return res;
   }

   /**
    * Create a {@link date} instance initialized to the default value.
    * 
    * @return   See above.
    */
   public static date date()
   {
      date res = new date();
      
      return res;
   }

   /**
    * Create a {@link date} array of the given extent.  If there is data in the 
    * <code>init</code> array for a specified index, that will be used to initialize it.  
    * Otherwise, if the init array is empty, default initializer will be used; if the init array 
    * is not empty, all indices which don't have a counterpart will use the last entry in 
    * this array.
    * 
    * @param    extent
    *           The array length.
    * @param    init
    *           The initializer array.
    * 
    * @return   See above.
    */
   public static date[] dateExtent(int extent, date... init)
   {
      date[] res = new date[extent];
      
      initExtent(res, init, DATE_INIT);
      
      return res;
   }

   /**
    * Create a {@link date} extent using as initial values the data in the <code>init</code> 
    * array.  If the init array is empty, a non-initialized (0-extent) array will be returned.
    * 
    * @param    init
    *           The initializer array.
    * 
    * @return   See above.
    */
   public static date[] dateExtent(date... init)
   {
      date[] res = new date[init.length];
      
      initExtent(res, init, DATE_INIT);
      
      return res;
   }

   /**
    * Create a {@link date} instance initialized to the init value.
    * 
    * @param    init
    *           The init value.
    *           
    * @return   See above.
    */
   public static date date(date init)
   {
      date res = new date(init);
      
      return res;
   }
   
   /**
    * Create a {@link datetime} instance initialized to the default value.
    * 
    * @return   See above.
    */
   public static datetime datetime()
   {
      datetime res = new datetime();
      
      return res;
   }
   
   /**
    * Create a {@link datetime} array of the given extent.  If there is data in the 
    * <code>init</code> array for a specified index, that will be used to initialize it.  
    * Otherwise, if the init array is empty, default initializer will be used; if the init array 
    * is not empty, all indices which don't have a counterpart will use the last entry in 
    * this array.
    * 
    * @param    extent
    *           The array length.
    * @param    init
    *           The initializer array.
    * 
    * @return   See above.
    */
   public static datetime[] datetimeExtent(int extent, datetime... init)
   {
      datetime[] res = new datetime[extent];
      
      initExtent(res, init, DATETIME_INIT);
      
      return res;
   }
   
   /**
    * Create a {@link datetime} extent using as initial values the data in the <code>init</code> 
    * array.  If the init array is empty, a non-initialized (0-extent) array will be returned.
    * 
    * @param    init
    *           The initializer array.
    * 
    * @return   See above.
    */
   public static datetime[] datetimeExtent(datetime... init)
   {
      datetime[] res = new datetime[init.length];
      
      initExtent(res, init, DATETIME_INIT);
      
      return res;
   }
   
   /**
    * Create a {@link datetime} instance initialized to the init value.
    * 
    * @param    init
    *           The init value.
    *           
    * @return   See above.
    */
   public static datetime datetime(datetime init)
   {
      datetime res = new datetime(init);
      
      return res;
   }

   /**
    * Create a {@link datetimetz} instance initialized to the default value.
    * 
    * @return   See above.
    */
   public static datetimetz datetimetz()
   {
      datetimetz res = new datetimetz();
      
      return res;
   }
   
   /**
    * Create a {@link datetimetz} array of the given extent.  If there is data in the 
    * <code>init</code> array for a specified index, that will be used to initialize it.  
    * Otherwise, if the init array is empty, default initializer will be used; if the init array 
    * is not empty, all indices which don't have a counterpart will use the last entry in 
    * this array.
    * 
    * @param    extent
    *           The array length.
    * @param    init
    *           The initializer array.
    * 
    * @return   See above.
    */
   public static datetimetz[] datetimetzExtent(int extent, datetimetz... init)
   {
      datetimetz[] res = new datetimetz[extent];
      
      initExtent(res, init, DATETIMETZ_INIT);
      
      return res;
   }
   
   /**
    * Create a {@link datetimetz} extent using as initial values the data in the <code>init</code> 
    * array.  If the init array is empty, a non-initialized (0-extent) array will be returned.
    * 
    * @param    init
    *           The initializer array.
    * 
    * @return   See above.
    */
   public static datetimetz[] datetimetzExtent(datetimetz... init)
   {
      datetimetz[] res = new datetimetz[init.length];
      
      initExtent(res, init, DATETIMETZ_INIT);
      
      return res;
   }
   
   /**
    * Create a {@link datetimetz} instance initialized to the init value.
    * 
    * @param    init
    *           The init value.
    *           
    * @return   See above.
    */
   public static datetimetz datetimetz(datetimetz init)
   {
      datetimetz res = new datetimetz(init);
      
      return res;
   }
   
   /**
    * Create a {@link handle} instance initialized to the default value.
    * 
    * @return   See above.
    */
   public static handle handle()
   {
      handle res = new handle();
      
      return res;
   }
   
   /**
    * Create a 0-length {@link handle} array.
    *
    * @return   See above.
    */
   public static handle[] handleExtent()
   {
      handle[] res = new handle[0];
      
      return res;
   }
   
   /**
    * Create a {@link handle} array of the given extent and initialize it with the default value.
    * 
    * @param    extent
    *           The array length.
    * 
    * @return   See above.
    */
   public static handle[] handleExtent(int extent)
   {
      handle[] res = new handle[extent];
      
      BaseDataType.initializeDefaultExtent(res);
      
      return res;
   }
   
   /**
    * Create a {@link comhandle} instance initialized to the default value.
    *
    * @return   See above.
    */
   public static comhandle comhandle()
   {
      return new comhandle();
   }
   
   /**
    * Create a 0-length {@link comhandle} array.
    *
    * @return   See above.
    */
   public static comhandle[] comhandleExtent()
   {
      return new comhandle[0];
   }
   
   /**
    * Create a 0-length {@link comhandle} array.
    *
    * @return   See above.
    */
   public static comhandle[] comhandleExtent(int extent)
   {
      comhandle[] res = new comhandle[extent];
      
      BaseDataType.initializeDefaultExtent(res);
      
      return res;
   }
   
   /**
    * Create a {@link logical} instance initialized to the default value.
    * 
    * @return   See above.
    */
   public static logical logical()
   {
      logical res = new logical(false);
      
      return res;
   }
   
   /**
    * Create a {@link logical} array of the given extent.  If there is data in the 
    * <code>init</code> array for a specified index, that will be used to initialize it.  
    * Otherwise, if the init array is empty, default initializer will be used; if the init array 
    * is not empty, all indices which don't have a counterpart will use the last entry in 
    * this array.
    * 
    * @param    extent
    *           The array length.
    * @param    init
    *           The initializer array.
    * 
    * @return   See above.
    */
   public static logical[] logicalExtent(int extent, Boolean... init)
   {
      logical[] res = new logical[extent];
      
      initExtent(res, init, BOOLEAN_INIT);
      
      return res;
   }
   
   /**
    * Create a {@link logical} extent using as initial values the data in the <code>init</code> 
    * array.  If the init array is empty, a non-initialized (0-extent) array will be returned.
    * 
    * @param    init
    *           The initializer array.
    * 
    * @return   See above.
    */
   public static logical[] logicalExtent(Boolean... init)
   {
      logical[] res = new logical[init.length];
      
      initExtent(res, init, BOOLEAN_INIT);
      
      return res;
   }
   
   /**
    * Create a {@link logical} instance.  If the <code>init</code> is null, initialize it to
    * <code>unknown</code>.  Otherwise, initialize it to the init value.
    * 
    * @param    init
    *           The init value.
    *           
    * @return   See above.
    */
   public static logical logical(Boolean init)
   {
      logical res = (init == null ? new logical() : new logical(init));
      
      return res;
   }
   
   /**
    * Create a {@link decimal} instance initialized to the default value.
    * 
    * @return   See above.
    */
   public static decimal decimal()
   {
      decimal res = new decimal(0);
      
      return res;
   }
   
   /**
    * Create a {@link decimal} array of the given extent.  If there is data in the 
    * <code>init</code> array for a specified index, that will be used to initialize it.  
    * Otherwise, if the init array is empty, default initializer will be used; if the init array 
    * is not empty, all indices which don't have a counterpart will use the last entry in 
    * this array.
    * 
    * @param    extent
    *           The array length.
    * @param    init
    *           The initializer array.
    * 
    * @return   See above.
    */
   public static decimal[] decimalExtent(int extent, decimal... init)
   {
      decimal[] res = new decimal[extent];
      
      initExtent(res, init, DECIMAL_INIT);
      
      return res;
   }
   
   /**
    * Create a {@link decimal} extent using as initial values the data in the <code>init</code> 
    * array.  If the init array is empty, a non-initialized (0-extent) array will be returned.
    * 
    * @param    init
    *           The initializer array.
    * 
    * @return   See above.
    */
   public static decimal[] decimalExtent(decimal... init)
   {
      decimal[] res = new decimal[init.length];
      
      initExtent(res, init, DECIMAL_INIT);
      
      return res;
   }
   
   /**
    * Create a {@link decimal} instance initialized to the init value.
    * 
    * @param    init
    *           The init value.
    *           
    * @return   See above.
    */
   public static decimal decimal(decimal init)
   {
      decimal res = new decimal(init);
      
      return res;
   }
   
   /**
    * Create a {@link integer} instance initialized to the default value.
    * 
    * @return   See above.
    */
   public static integer integer()
   {
      integer res = new integer(0);
      
      return res;
   }
   
   /**
    * Create a {@link integer} array of the given extent.  If there is data in the 
    * <code>init</code> array for a specified index, that will be used to initialize it.  
    * Otherwise, if the init array is empty, default initializer will be used; if the init array 
    * is not empty, all indices which don't have a counterpart will use the last entry in 
    * this array.
    * 
    * @param    extent
    *           The array length.
    * @param    init
    *           The initializer array.
    * 
    * @return   See above.
    */
   public static integer[] integerExtent(int extent, Long... init)
   {
      integer[] res = new integer[extent];
      
      initExtent(res, init, INTEGER_INIT);
      
      return res;
   }
   
   /**
    * Create a {@link integer} extent using as initial values the data in the <code>init</code> 
    * array.  If the init array is empty, a non-initialized (0-extent) array will be returned.
    * 
    * @param    init
    *           The initializer array.
    * 
    * @return   See above.
    */
   public static integer[] integerExtent(Long... init)
   {
      integer[] res = new integer[init.length];
      
      initExtent(res, init, INTEGER_INIT);
      
      return res;
   }

   /**
    * Create a {@link integer} instance.  If the <code>init</code> is null, initialize it to
    * <code>unknown</code>.  Otherwise, initialize it to the init value.
    * 
    * @param    init
    *           The init value.
    *           
    * @return   See above.
    */
   public static integer integer(Long init)
   {
      integer res = (init == null ? new integer() : new integer(init));
      
      return res;
   }
   
   /**
    * Create a {@link int64} instance initialized to the default value.
    * 
    * @return   See above.
    */
   public static int64 int64()
   {
      int64 res = new int64(0);
      
      return res;
   }
   
   /**
    * Create a {@link int64} array of the given extent.  If there is data in the 
    * <code>init</code> array for a specified index, that will be used to initialize it.  
    * Otherwise, if the init array is empty, default initializer will be used; if the init array 
    * is not empty, all indices which don't have a counterpart will use the last entry in 
    * this array.
    * 
    * @param    extent
    *           The array length.
    * @param    init
    *           The initializer array.
    * 
    * @return   See above.
    */
   public static int64[] int64Extent(int extent, Long... init)
   {
      int64[] res = new int64[extent];
      
      initExtent(res, init, INT64_INIT);

      return res;
   }

   /**
    * Create a {@link int64} extent using as initial values the data in the <code>init</code> 
    * array.  If the init array is empty, a non-initialized (0-extent) array will be returned.
    * 
    * @param    init
    *           The initializer array.
    * 
    * @return   See above.
    */
   public static int64[] int64Extent(Long... init)
   {
      int64[] res = new int64[init.length];
      
      initExtent(res, init, INT64_INIT);
      
      return res;
   }
   
   /**
    * Create a {@link int64} instance.  If the <code>init</code> is null, initialize it to
    * <code>unknown</code>.  Otherwise, initialize it to the init value.
    * 
    * @param    init
    *           The init value.
    *           
    * @return   See above.
    */
   public static int64 int64(Long init)
   {
      int64 res = (init == null ? new int64() : new int64(init));
      
      return res;
   }
   
   /**
    * Create a {@link rowid} instance initialized to the default value.
    * 
    * @return   See above.
    */
   public static rowid rowid()
   {
      rowid res = new rowid();
      
      return res;
   }
   
   /**
    * Create a {@link rowid} array of the given extent and initialize it with the default value.
    * 
    * @param    extent
    *           The array length.
    * 
    * @return   See above.
    */
   public static rowid[] rowidExtent(int extent)
   {
      rowid[] res = new rowid[extent];
      
      BaseDataType.initializeDefaultExtent(res);
      
      return res;
   }
   
   /**
    * Create a 0-length {@link rowid} array.
    * 
    * @return   See above.
    */
   public static rowid[] rowidExtent()
   {
      rowid[] res = new rowid[0];
      
      return res;
   }
   
   /**
    * Create a {@link recid} instance initialized to the default value.
    * 
    * @return   See above.
    */
   public static recid recid()
   {
      recid res = new recid();
      
      return res;
   }
   
   /**
    * Create a {@link recid} array of the given extent.  If there is data in the 
    * <code>init</code> array for a specified index, that will be used to initialize it.  
    * Otherwise, if the init array is empty, default initializer will be used; if the init array 
    * is not empty, all indices which don't have a counterpart will use the last entry in 
    * this array.
    * 
    * @param    extent
    *           The array length.
    * @param    init
    *           The initializer array.
    * 
    * @return   See above.
    */
   public static recid[] recidExtent(int extent, Long... init)
   {
      recid[] res = new recid[extent];
      
      initExtent(res, init, RECID_INIT);
      
      return res;
   }
   
   /**
    * Create a {@link recid} extent using as initial values the data in the <code>init</code> 
    * array.  If the init array is empty, a non-initialized (0-extent) array will be returned.
    * 
    * @param    init
    *           The initializer array.
    * 
    * @return   See above.
    */
   public static recid[] recidExtent(Long... init)
   {
      recid[] res = new recid[init.length];
      
      initExtent(res, init, RECID_INIT);
      
      return res;
   }
   
   /**
    * Create a {@link recid} instance.  If the <code>init</code> is null, initialize it to
    * <code>unknown</code>.  Otherwise, initialize it to the init value.
    * 
    * @param    init
    *           The init value.
    *           
    * @return   See above.
    */
   public static recid recid(Long init)
   {
      recid res = (init == null ? new recid() : new recid(init));
      
      return res;
   }

   /**
    * Create a {@link character} instance initialized to the default value.
    * 
    * @return   See above.
    */
   public static character character()
   {
      character res = new character("");
      
      return res;
   }

   /**
    * Create a {@link character} array of the given extent.  If there is data in the 
    * <code>init</code> array for a specified index, that will be used to initialize it.  
    * Otherwise, if the init array is empty, default initializer will be used; if the init array 
    * is not empty, all indices which don't have a counterpart will use the last entry in 
    * this array.
    * 
    * @param    extent
    *           The array length.
    * @param    init
    *           The initializer array.
    * 
    * @return   See above.
    */
   public static character[] characterExtent(int extent, String... init)
   {
      character[] res = new character[extent];
      
      initExtent(res, init, CHARACTER_INIT);
      
      return res;
   }

   /**
    * Create a {@link character} extent using as initial values the data in the <code>init</code> 
    * array.  If the init array is empty, a non-initialized (0-extent) array will be returned.
    * 
    * @param    init
    *           The initializer array.
    * 
    * @return   See above.
    */
   public static character[] characterExtent(String... init)
   {
      character[] res = new character[init.length];
      
      initExtent(res, init, CHARACTER_INIT);
      
      return res;
   }

   /**
    * Create a {@link character} instance.  If the <code>init</code> is null, initialize it to
    * <code>unknown</code>.  Otherwise, initialize it to the init value.
    * 
    * @param    init
    *           The init value.
    *           
    * @return   See above.
    */
   public static character character(String init)
   {
      character res = (init == null ? new character() : new character(init));
      
      return res;
   }
   
   /**
    * Create a {@link longchar} instance initialized to the default value.
    * 
    * @return   See above.
    */
   public static longchar longchar()
   {
      longchar res = new longchar("");
      
      return res;
   }
   
   /**
    * Create a {@link longchar} array of the given extent.  If there is data in the 
    * <code>init</code> array for a specified index, that will be used to initialize it.  
    * Otherwise, if the init array is empty, default initializer will be used; if the init array 
    * is not empty, all indices which don't have a counterpart will use the last entry in 
    * this array.
    * 
    * @param    extent
    *           The array length.
    * @param    init
    *           The initializer array.
    * 
    * @return   See above.
    */
   public static longchar[] longcharExtent(int extent, String... init)
   {
      longchar[] res = new longchar[extent];
      
      initExtent(res, init, LONGCHAR_INIT);
      
      return res;
   }
   
   
   /**
    * Create a {@link longchar} extent using as initial values the data in the <code>init</code> 
    * array.  If the init array is empty, a non-initialized (0-extent) array will be returned.
    * 
    * @param    init
    *           The initializer array.
    * 
    * @return   See above.
    */
   public static longchar[] longcharExtent(String... init)
   {
      longchar[] res = new longchar[init.length];
      
      initExtent(res, init, LONGCHAR_INIT);
      
      return res;
   }
   
   /**
    * Create a {@link longchar} instance.  If the <code>init</code> is null, initialize it to
    * <code>unknown</code>.  Otherwise, initialize it to the init value.
    * 
    * @param    init
    *           The init value.
    *           
    * @return   See above.
    */
   public static longchar longchar(String init)
   {
      longchar res = (init == null ? new longchar() : new longchar(init));
      
      return res;
   }
   
   /**
    * Initialize the specified array using; if there is data in the <code>init</code> array for
    * a specified index, that will be used to initialize it.  Otherwise, if the init array is 
    * empty, default initializer will be used; if the init array is not empty, all indices which
    * don't have a counterpart will use the last entry in this array.
    * 
    * @param    res
    *           The array to be initialized.
    * @param    init
    *           The initializer.
    * @param    finit
    *           Function which will create either a default value or an initialized value, 
    *           depending on the given argument.
    */
   private static <T extends BaseDataType, I> void initExtent(BaseDataType[] res, 
                                                              I[]            init, 
                                                              Function<I, T> finit)
   {
      int extent = res.length;
      if (extent == 0)
      {
         return;
      }

      if (init == null || init.length == 0)
      {
         BaseDataType.initializeDefaultExtent(res);
      }
      else
      {
         I last = init[init.length - 1];
         for (int i = 0; i < Math.min(extent, init.length); i++)
         {
            res[i] = finit.apply(init[i]);
         }
         for (int i = init.length; i < extent; i++)
         {
            res[i] = finit.apply(last);
         }
      }
   }
}
