/*
** Module   : WebService.java
** Abstract : An interface defining APIs specific to a client-side access to Web Services.
**
** Copyright (c) 2016-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20160323 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

/**
 * Interface defining APIs to connect and invoke Web Services in P2J. This interface is exported
 * as a service during P2J client startup.
 */
public interface WebService
{
   /**  Error code. Specified WSDL document failed to load. */
   public int ERR_LOADING_WSDL_DOCUMENT = -1;
   
   /**  Error code. Parameters specified in connection string failed validation. */
   public int ERR_INVALID_PARAMS_IN_CONNECTION_STRING = -2;
   
   /**  Error code. The Service is not in the WSDL document. */
   public int ERR_SERVICE_NOT_IN_WSDL_DOCUMENT = -3;
   
   /**  Error code. No service. */
   public int ERR_NO_SERVICE_ERROR = -4;
   
   /**  Error code. The Port  is not in the WSDL document. */
   public int ERR_PORT_NOT_IN_WSDL_DOCUMENT = -5;
   
   /**  Error code. The Binding is not supported, it must be a SOAP 1.1 binding. */
   public int ERR_BINDING_NOT_SUPPORTED = -6;
   
   /**  Error code. The Binding cannot be found in the WSDL document. */
   public int ERR_BINDING_NOT_IN_WSDL_DOCUMENT = -7;
   
   /**  Error code. Internal error locating Service and Port for Binding. */
   public int ERR_SERVICE_AND_PORT_NOT_FOUND = -8;
   
   /**  Error code. A specific Service name must be specified from the WSDL document. */
   public int ERR_NO_SERVICE_SPECIFIED = -9;
   
   /**  Error code. The TargetNamespace option  does not match the WSDL's target namespace. */
   public int ERR_TARGETNAMESPACE_NOT_MATCH_WDSL = -10;
   
   /**
    * Validate parameters and establish a new connection using the given options. The returned id
    * will be used in further references to this connection until the connection is disconnected.
    *
    * @param   options
    *          A collection of all parameters used for this connection.
    * 
    * @return  A positive (possible 0) integer representing the id of this connection to
    *          web-service server and a negative value representing an error code if any error
    *          occurred during validation and initialization of this connection.
    */
   public int connect(WebServiceConnectOptions options);
   
   /**
    * Get the name of the PortType of the specified connection.
    *
    * @param   id
    *          The id of the connection.
    *
    * @return  the name of the PortType of specified connection or empty string if there is no
    *          connection with requested id.
    */
   public String getPortTypeName(int id);
   
   /**
    * Disconnect the web service.
    *
    * @param   id
    *          The ID of the associated web service on P2J Client side.
    *
    * @return  {@code true} if no errors were encountered during disconnect.
    */
   public boolean disconnect(int id);
   
   /**
    * Clears the temporary state of an invocation. Does not close open connections. It is called
    * when the result of the invocation was obtained on server side or when an error occurred. If
    * {@code id} is invalid nothing happens.
    *
    * @param   id
    *          The id of an open connection.
    */
   public void cleanupInvocation(int id);
   
   /**
    * Prepare the connection for an invocation of an operation on the Web Service. The parameters
    * are validated and, on success, the internal state of the call is temporarily saved.
    *
    * @param   id
    *          The id of an open connection.
    * @param   portTypeName
    *          The name of the port type to be invoked. 
    * @param   operationName
    *          The name of the operation to be invoked on the Web Service. 
    * @param   modes
    *          A string representation of the modes of each parameter. May not be {@code null}.
    * @param   args
    *          The operation's arguments.
    *
    * @return  {@code true} if validation was successful and {@code false} if some errors occurred
    *          and invocation is not available. 
    */
   public boolean prepareInvoke(int       id,
                                String    portTypeName,
                                String    operationName,
                                String    modes,
                                Object... args);
   
   /**
    * Do the actual Web Service invocation. The call uses an already open and prepared connection
    * identified by {@code id}. 
    *
    * @param   id
    *          The id of an open connection.
    * @param   operationName
    *          The name of the operation to be invoked on the Web Service. 
    * @param   portTypeName
    *          The name of the port type to be invoked. 
    * @param   requestHeader
    *          The request header if any was obtained form a call to a configured header callback.
    *
    * @return  {@code true} if the service operation was successful and {@code false} if 
    *          {@code id} does not represent a valid connection or other error occurred.
    */
   public boolean invoke(int    id,
                         String operationName,
                         String portTypeName,
                         String requestHeader);
   
   /**
    * Obtain the values of output parameters after a successful Web Service invocation. 
    *
    * @param   id
    *          The id of an open connection.
    *
    * @return  An array with output values returned by the last web service operation.
    */
   public Object[] getOutputValues(int id);
   
   /**
    * Get the wsdl namespace of this connection.
    *
    * @param   id
    *          The id of an open connection.
    *
    * @return  the wsdl namespace of this connection or {@code null} if {@code id} does not
    *          represent the identifier of a valid open connection.
    */
   public String getNamespace(int id);
   
   /**
    * Get the header entries of an already open and prepared connection.
    *
    * @param   id
    *          The id of an open connection.
    *
    * @return  the header entries of the specified connection ot {@code null} if the {@code id} is
    *          not the identifier of an open and prepared connection.
    */
   public String[] getHeaderEntries(int id);
}