/*
** Module   : DojoToolkitHandler.java
** Abstract : delivers Dojo toolkit modules by loading them from the classpath as resources
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 MAG 20140109 First version.
** 002 MAG 20140701 Deliver multiple content types (MIME).
** 003 MAG 20140730 Change Dojo toolkit resource loading. Eliminate any logging messages.
** 004 GES 20150308 Moved to a more generic package location.
** 005 HC  20170110 Updated dojo resource directory.
** 006 SBI 20170720 Delegated general methods to AbstractModulesHandler in order to be used to
**                  deliver js zipped modules.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.web;

import javax.servlet.http.HttpServletRequest;

import org.eclipse.jetty.server.Request;



/**
 * Jetty Handler which delivers Dojo toolkit modules.  The Dojo toolkit distribution zip file
 * should be located in the JVM CLASSPATH and modules are accessed as resources. The Dojo module
 * target request should start with DOJO_TOOLKIT_ROOT.
 * <p>
 * Starting from the target the Dojo module resource name is constructed by replacing the 
 * DOJO_TOOLKIT_ROOT with DOJO_TOOLKIT. Then the module resource is loaded and sent as response
 * to the requester.
 * <p>
 * IMPORTANT: Don't use logging anywhere inside this class. When used in a web client, this runs
 * in conjunction with an authentication plugin during session establishment with the P2J server
 * and using logging the application will deadlock.
*/
public class DojoToolkitHandler
extends AbstractModulesHandler
{
   /** Dojo toolkit package. */
   private static final String DOJO_TOOLKIT = "dojo-1.10.0";

   /** Dojo toolkit root path. */
   private static final String DOJO_TOOLKIT_ROOT = "/dojo-toolkit";

   /**
    * Returns true if it is a request for a dojo js module or a dojo html resource, otherwise
    * false.
    * 
    * 
    * @param    target 
    *           The target of the request - either a URI or a name.
    * @param    base
    *           The base request.
    * @param    request
    *           The object or a wrapper of the request.
    * 
    * @return   true if it is a request for a dojo js module or a dojo html resource
    */
   @Override
   public boolean isModuleRequest(String target, Request base, HttpServletRequest request)
   {
      // check if this is a Dojo module request
      
      return target.toLowerCase().startsWith(DOJO_TOOLKIT_ROOT);
   }

   /**
    * Maps the target of the request to its resource name.
    * 
    * @param    target
    *           The target of the request - either a URI or a name.
    * 
    * @return   The target resource name
    */
   @Override
   public String mapToResource(String target)
   {
      return target.replace(DOJO_TOOLKIT_ROOT, DOJO_TOOLKIT);
   }
}