/*
** Module   : GenericWebServer.java
** Abstract : generic design for an embedded web server.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 MAG 20131105 First version based on jetty 9.1
** 002 MAG 20131129 Import server KeyStore.
** 003 AB  20150515 Removed configWsContext() and now directly use the handler collection.
** 004 SBI 20160220 Changed to set the websocket parameters from the server provided configuration.
** 005 HC  20170612 Changes related to implementation of new GWT-based Admin client.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.web;

import java.io.*;
import java.net.*;
import java.security.*;
import java.util.*;
import java.util.logging.*;

import javax.net.ssl.*;

import org.eclipse.jetty.http.*;
import org.eclipse.jetty.server.*;
import org.eclipse.jetty.server.Handler;
import org.eclipse.jetty.server.Server;
import org.eclipse.jetty.server.handler.*;
import org.eclipse.jetty.util.ssl.*;
import org.eclipse.jetty.util.thread.*;
import org.eclipse.jetty.websocket.servlet.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.util.*;

/**
 * A generic design for an embedded web server having support for WebSocket's.
 * 
 * @author mag
 * 
 */
public class GenericWebServer
{
   /** Logger. */
   protected static final Logger LOG = LogHelper.getLogger(GenericWebServer.class.getName());
   
   /** Exported server key store */
   private ServerKeyStore serverKeyStore;
   
   /** The bootstrap configuration */
   private BootstrapConfig config;
   
   /** Embedded jetty server */
   private Server server;
   
   /** Context Handlers */
   private HandlerCollection handlerCollection = new HandlerCollection();
   
   /** Context path */
   private String contextPath = "/";
   
   /** Server port */
   private int port;
   
   /** Server host name */
   private String host;
   
   /**
    * Startup the embedded server. 
    * The sever will listen on all interfaces. 
    * The port number is dynamically allocated by the OS.
    * 
    * @throws  Exception
    *          If an error condition occur during the server startup.
    */
   public void startup() 
   throws Exception
   {
      this.startup(null, 0);
   }
   
   /**
    * Startup server.
    * 
    * @param   host
    *          Host name or IP address. If <b>null</b> server will listen on
    *          all interfaces.
    * @param   port
    *          Port number to listen. If <b>0</b> the port is OS dynamically
    *          allocated.
    * 
    * @throws  Exception
    *          If an error condition occur during the server startup.
    */
   public void startup(String host, int port) 
   throws Exception
   {
      // SSL Context Factory for HTTPS
      SslContextFactory sslContextFactory = getSslContextFactory();
      
      QueuedThreadPool pool = new QueuedThreadPool();
      pool.setDaemon(true);
      
      server = new Server(pool);
      server.manage(pool);
      server.addConnector(configSslConnector(server, 
                                             host, 
                                             port,
                                             sslContextFactory));
      server.setHandler(handlerCollection);
      server.setStopAtShutdown(true);
      server.start();
      
      // get host and port after server is starting
      getRuntimeParams();
      
      LOG.logp(Level.INFO, 
               "GenericWebServer.startup", 
               "", 
               LogHelper.generate("Server URL: https://%s:%d%s", 
                                  getHost(), 
                                  getPort(),
                                  contextPath));
   }
   
   /**
    * Shutdown server.
    * 
    * @throws  Exception
    *          If the server cannot be stopped.
    */
   public void shutdown() 
   throws Exception
   {
      if (server != null && server.isRunning())
      {
         server.stop();
      }
   }
   
   /**
    * Waits for server thread pool to die.
    * 
    * @throws  Exception
    *          If any error condition occur.
    */
   public void join()
   throws Exception
   {
      if (server != null && server.isRunning())
      {
         server.join();
      }
   }
   
   /**
    * Find out if server is running.
    * 
    * @return  <code>true</code> if server is running <code>false</code> otherwise.
    */
   public boolean isRunning()
   {
      return server != null && server.isRunning();
   }
   
   /**
    * Set the bootstrap configuration structure if any.
    * 
    * @param   config
    *          The <code>BootstrapConfig</code> instance or <b>null</b>.
    */
   public void setBootstrapConfig(BootstrapConfig config)
   {
      this.config = config;
   }
   
   /**
    * Get the server runtime host and port number.
    * 
    */
   private void getRuntimeParams()
   {
      if (server.isRunning())
      {
         Connector[] connectors = server.getConnectors();
         ServerConnector connector = (ServerConnector) connectors[0];
         port = connector.getLocalPort();
         host = connector.getHost();
      }
   }
   
   /**
    * Register a list of handlers.
    * 
    * @param   handlers
    *          A list of handlers to be registered.
    */
   public void addHandler(Handler[] handlers)
   {
      for (Handler handler : handlers)
      {
         addHandler(handler);
      }
   }
   
   /**
    * Register a new handler.
    * 
    * @param   handler
    *          A Handler instance.
    */
   public void addHandler(Handler handler)
   {
      if (handler != null)
      {
         handlerCollection.addHandler(handler);
      }
   }
   
   /**
    * Register a new WebSocket.
    * 
    * @param   target
    *          The WebSocket target.
    * @param   webSocket
    *          A POJO decorated with <code>@WebSocket</code> annotation.
    */
   public void addWebSocketHandler(String target, Class<?> webSocket)
   {
      Handler wsHandler = new PojoWebSocketHandler(target, webSocket, new WebSocketConfig(config));
      handlerCollection.addHandler(wsHandler);
   }
   
   /**
    * Register a new WebSocket by hand using a custom WebSocket creator.
    * 
    * @param   target
    *          The WebSocket target.
    * @param   creator
    *          A custom <code>WebSocketCreator</code> instance.
    */
   public void addWebSocketHandler(String target, WebSocketCreator creator)
   {
      Handler wsHandler = new PojoWebSocketHandler(target, creator, new WebSocketConfig(config));
      handlerCollection.addHandler(wsHandler);
   }
   
   /**
    * Construct a SSL Context Factory for HTTPS.
    * 
    * @return  The <code>SslContextFactory</code> instance.
    * 
    * @throws  Exception
    *          If any error conditions occur.
    */
   protected SslContextFactory getSslContextFactory()
   throws Exception
   {
      SecurityManager sm = SecurityManager.getInstance();
      SslContextFactory sslContextFactory = new SslContextFactory();
      if (config == null)
      {
         SSLContext sslContext = sm.getSecureSocketContext();
         sslContextFactory.setSslContext(sslContext);
      }
      else
      {
         if (serverKeyStore != null && serverKeyStore.isValid())
         {
            // Import sever SSL context
            ByteArrayInputStream stream = new ByteArrayInputStream(serverKeyStore.getKeyStore());
            KeyStore keyStore = KeyStore.getInstance("JKS");
            keyStore.load(stream, serverKeyStore.getKeyStorePassword().toCharArray());
            sslContextFactory.setKeyStore(keyStore);
            sslContextFactory.setKeyManagerPassword(serverKeyStore.getKeyManagerPassword());
         }
         else
         {
            SSLContext sslContext = sm.getSecureSocketContext(config);
            sslContextFactory.setSslContext(sslContext);
         }
      }
      
      return sslContextFactory;
   }
   
   /**
    * Create and configure a SSL server connector.
    * 
    * @param   server
    *          The server instance.
    * @param   host
    *          The server host name or IP address.
    * @param   port
    *          The port number to listen.
    * @param   sslContextFactory
    *          The <code>SslContextFactory</code> instance.
    * 
    * @return  The SSL server connector
    */
   private ServerConnector configSslConnector(Server server, 
                                              String host,
                                              int port,
                                              SslContextFactory sslContextFactory)
   {
      String protocol = HttpVersion.HTTP_1_1.asString();
      ServerConnector sslConnector = new ServerConnector(server,
         // for GWT development comment out SslConnectionFactory,
         // GTW's superdevmode doesn't run with SSL
         new SslConnectionFactory(sslContextFactory, protocol),
         new HttpConnectionFactory(new HttpConfiguration()));
      sslConnector.setHost(host);
      sslConnector.setPort(port);
      
      return sslConnector;
   }
   
   /**
    * Get host site local address.
    * 
    * @return  Site local host address if any or <b>localhost</b> if none.
    */
   private String getHostLocalAddress()
   {
      String hostAddress = "localhost";
      try
      {
         InetAddress address = getSiteLocalAddress();
         if (address != null)
         {
            hostAddress = address.getHostAddress();
         }
      }
      catch (Exception e)
      {
         LOG.logp(Level.SEVERE,
                  "GenericWebServer.getHostLocalAddress()",
                  "",
                  LogHelper.generate("Error!"),
                  e);
      }
      return hostAddress;
   }
   
   /**
    * Search on the network interfaces of this machine for
    * <code>InetAddress</code> which is a site local address.
    * 
    * @return  The InetAddress of the first site local address.
    * 
    * @throws  Exception
    *          If an error condition occur.
    */
   private InetAddress getSiteLocalAddress()
   throws Exception
   {
      // enumerate network interfaces
      Enumeration<NetworkInterface> netInterfaces =
         NetworkInterface.getNetworkInterfaces();
      if (netInterfaces != null)
      {
         while (netInterfaces.hasMoreElements())
         {
            NetworkInterface netInterface = netInterfaces.nextElement();
            if (netInterface.isUp())
            {
               // enumerate addresses bound to this network interface.
               Enumeration<InetAddress> addresses = 
                  netInterface.getInetAddresses();
               while (addresses.hasMoreElements())
               {
                  InetAddress address = addresses.nextElement();
                  if (address.isSiteLocalAddress())
                  {
                     return address;
                  }
               }
            }
         }
      }
      return null;
   }
   
   /**
    * Set the server context path root.
    * 
    * @param   contextPath
    *          A string representing the context path root.
    */
   public void setContextPath(String contextPath)
   {
      this.contextPath = contextPath;
   }
   
   /**
    * Get the server context path root.
    * 
    * @return  A string representing the context path root.
    */
   public String getContextPath()
   {
      return contextPath;
   }
   
   /**
    * Get server runtime port number.
    * 
    * @return  The server runtime port number.
    */
   public int getPort()
   {
      return port;
   }
   
   /**
    * Get the server runtime host.
    * 
    * @return  The server runtime host.
    * 
    */
   public String getHost()
   {
      return (host == null) ? getHostLocalAddress() : host;
   }
   
   /**
    * Set imported key store
    * 
    * @param   keyStore
    *          Imported key store if any or <code>null</code>
    */
   public void setKeyStore(ServerKeyStore keyStore)
   {
      this.serverKeyStore = keyStore;
   }
}