/*
** Module   : HtmlResourceHandler.java
** Abstract : delivers main page to AJAX clients
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ----------------------Description--------------------------
** 001 MAG 20131105 First version based on jetty 9.1
** 002 MAG 20140206 Set Cache-Control header.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.web;

import java.io.*;
import java.util.logging.*;

import javax.servlet.*;
import javax.servlet.http.*;

import org.eclipse.jetty.http.*;
import org.eclipse.jetty.server.handler.*;
import org.eclipse.jetty.server.*;

import com.goldencode.p2j.util.*;

/**
 * jetty Handler which delivers HTML resource files.
 * This class provide a method to parse the HTML text and
 * substitute placeholder's. 
 */
public class HtmlResourceHandler
extends AbstractHandler
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(HtmlResourceHandler.class.getName());

   /** The root portion of the target request name (part of the URL). */
   private String targetRoot = null;
   
   /** The html skeleton page. */
   private String skeletonPage = null;
     
   /**
    * Constructor.
    *
    * @param    targetRoot
    *           Prefix for all URLs relating to this page.
    * @param    skeletonPage
    *           Fully qualified resource name of the HTML file which should be used.
    */
   public HtmlResourceHandler(String targetRoot, String skeletonPage)
   {
      this.targetRoot   = targetRoot;
      this.skeletonPage = skeletonPage;  
   }
   
   /**
    * Return the main page for AJAX clients.
    *    
    * @param   target 
    *          The target of the request - either a URI or a name.
    * @param   base
    *          The base request.
    * @param   request
    *          The object or a wrapper of the request.
    * @param   response 
    *          The object or a wrapper of the response. 
    */
   public void handle(String              target, 
                      Request             base,
                      HttpServletRequest  request, 
                      HttpServletResponse response)
   throws IOException, 
          ServletException
   {
      if (!base.isHandled() && target.equalsIgnoreCase(targetRoot))
      {
         generateMainPage(base, response);
      }    
   }
      
   /**
    * Use the skeleton main page to generate a response.
    *
    * @param   base
    *          The details and state the request.
    * @param   response 
    *          The details and state of the response.
    *           
    * @throws  IOException
    *          ServletException
    *          If any error conditions occur.
    */
   private void generateMainPage(Request base, HttpServletResponse response)
   throws IOException, 
          ServletException
   {
      // produce the page contents
      response.setContentType(MimeTypes.Type.TEXT_HTML.asString());
      response.setStatus(HttpServletResponse.SC_OK);
      response.setHeader(HttpHeader.CACHE_CONTROL.asString(), 
                         "no-cache, no-store, must-revalidate");
      response.setDateHeader(HttpHeader.EXPIRES.asString(), 0); // Proxies.      
      
      PrintWriter writer = response.getWriter();
      
      try
      {
         InputStream is = this.getClass().getResourceAsStream(skeletonPage);
         BufferedReader reader = new BufferedReader(new InputStreamReader(is));
         
         String str = null;
         
         // read the skeleton page and translate %substitutions
         // on the fly
         while (null != (str = reader.readLine()))
         {
            writer.println(handleReplacements(base, str));
         }
         
         reader.close();
      }
      catch (IOException ioe)
      {
         LOG.logp(Level.SEVERE,
                  "HtmlResourceHandler.generateMainPage()",
                  "",
                  LogHelper.generate("IOException!"),
                  ioe);
      }
      finally
      {
         // mark the request as handled
         base.setHandled(true);
      }
   }
      
   /**
    * Process the given string and make any replacements of parameter values
    * as needed. The replacement placeholder follow the JEE expression
    * language (EL) format for immediate evaluation. 
    * <p>
    * Example: ${expression}
    * <p>
    * Extends this class and override this method to provide your custom code
    * for placeholder's replacement.
    *
    * @param   base
    *          The details and state the request.
    * @param   text
    *          The text line that may need modifications.
    *
    * @return  The new text if a modification occurred or the original text
    *          if no modification was needed.
    */
    protected String handleReplacements(Request base, String text)
    {
       return text;
    }
}