/*
** Module   : VT100Renderer.java
** Abstract : VT-100 renderer contract.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.

**
** -#- -I- --Date-- ----------------Description-----------------
** 001 MAG 20140318 First version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.terminal;

/**
 * VT-100 renderer contract.
 * 
 * @author mag
 *
 */
public interface VT100Renderer
{   
   /**
    * Set cursor position.
    *  
    * @param   row
    *          Line number.
    * @param   col
    *          Column number.
    */
   public void setCursorPos(int row, int col);

   /**
    * Move cursor position with the amount of rows and columns.
    *  
    * @param   nRows
    *          Number of rows.
    * @param   nCols
    *          Number of columns.
    */
   public void moveCursorPos(int nRows, int nCols);

   /**
    * Set cell text attributes and color.
    * 
    * @param   text
    *          Character that should be displayed.
    * @param   attribute
    *          Cell attributes.
    * @param   color
    *          Call color FG/BG
    */
   public void setCellText(char text, int attribute, int color);
   
   /**
    * Play a message beep.
    */   
   public void beep();
   
   /**
    * Clear lines on screen.
    * 
    * @param   mode
    *          An integer value representing a screen region:
    *          0 - Erase from the active position to the end of the screen, inclusive (default)
    *          1 - Erase from start of the screen to the active position, inclusive
    *          2 - Erase all of the display -- all lines are erased, changed to single-width, 
    *              and the cursor does not move.
    */
   public void eraseInDisplay(int mode);   

   /**
    * Clear characters in current line.
    * 
    * @param   mode
    *          An integer value representing a line region:
    *          0 - Erase from the active position to the end of the line, inclusive (default)
    *          1 - Erase from the start of the line to the active position, inclusive
    *          2 - Erase all of the line, inclusive
    */
   public void eraseInLine(int mode);   
   
   /**
    * Set cursor at column.
    * 
    * @param   column
    *          Column number.
    */
   public void setCursorCol(int column);
   
   /**
    * NEL -- Next Line ESC E
    * 
    * This sequence causes the active position to move to the first position on the next line
    * downward. If the active position is at the bottom margin, a scroll up is performed.
    */
   public void nextLine();
   
   /**
    * IND -- Index ESC D
    * 
    * This sequence causes the active position to move downward one line without changing the 
    * column position. If the active position is at the bottom margin, a scroll up is performed.    
    */
   public void index();

   /**
    * RI -- Reverse Index ESC M
    * 
    * Move the active position to the same horizontal position on the preceding line.
    * If the active position is at the top margin, a scroll down is performed.
    */
   public void reverseIndex();

   /**
    * Set cursor visibility.
    * 
    * @param   showCursor
    * 
    */
   public void setCursorVisible(boolean showCursor);

   /**
    * Set auto wrap mode.
    * 
    * @param   autoWrap
    *          
    */
   public void setAutoWrap(boolean autoWrap);
      
   /**
    * Set scroll window.
    * 
    * @param   begLine
    *          Start line.
    * @param   endLine
    *          End line.
    */
   public void setScrollRegion(int begLine, int endLine);
   
   /**
    * Get cursor current row position.
    *  
    * @return  row number
    */
   public int getCursorRow();
   
   /**
    * Get cursor current column position.
    * 
    * @return  column number
    */
   public int getCursorCol();

   /**
    * Get number of columns.
    * 
    * @return  Number of columns
    */
   public int getColumns();
   
   /**
    * Get attribute mask.
    * 
    * @param   id
    *          Attribute id.
    * <p>
    *          0 - NORMAL 
    *          1 - BLINK 
    *          2 - BOLD 
    *          3 - UNDERLINE 
    *          4 - REVERSE
    *          
    * @return  attribute mask.
    */
   public int attributeMask(int id);
}