/*
** Module   : NestedWalker.java
** Abstract : merges a list of tree walkers into a single walker
**
** Copyright (c) 2009-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20090525   @42536 Merges a list of tree walkers into a single walker.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.trpl;

import java.util.*;
 
/**
 * Merges a list of tree walkers into a single walker. During tree processing,
 * for each hook, the list of walkers is iterated and that hook is invoked on
 * each one in order.
 */
public class NestedWalker
implements TreeWalker
{
   /** Tree walkers to be merged. */
   private List<TreeWalker> list = null;
   
   /**
    * Default constructor.
    */
   private NestedWalker()
   {
   }
   
   /**
    * Construct an instance using the given list.
    *
    * @param    list
    *           The list of walkers to merge in order of execution.
    */
   public NestedWalker(List<TreeWalker> list)
   {
      this.list = list;
   }
   
   /**
    * Called once per tree, before any other hook is invoked. At the time this
    * is called, the tree's root node will be the current node.
    * <p>
    * This is useful for allocating resources or calculating/initializing any
    * values that should be handled once per tree.
    *
    * @param    state
    *           Provides tree processing state including which node is being
    *           processed and the current status of the tree processing.
    */
   public void pre(WalkState state)
   {
      for (TreeWalker walker : list)
      {
         walker.pre(state);
      }
   }
   
   /**
    * Called once per node, this notifies the registered code that the node
    * is being visited. This will first be called for the root node, after
    * {@link #pre} completes. From there it will be called for each of the
    * nodes of the tree in a depth first traversal.
    * <p>
    * At the time this is called, the current node will be set to the node
    * being "walked" or visited.
    * <p>
    * Most logic is inserted in these hooks. Anything that does not depend upon
    * the order of child node processing is usually placed here.
    *
    * @param    state
    *           Provides tree processing state including which node is being
    *           processed and the current status of the tree processing.
    */
   public void walk(WalkState state)
   {
      for (TreeWalker walker : list)
      {
         walker.walk(state);
      }
   }
   
   /**
    * Called whenever a transition from a parent node to its first child
    * takes place during an AST walk. This will be called <em>before</em> the
    * next AST node (the child node) in the tree traversal is visited with
    * {@link #walk}. This call never occurs for nodes that have no children.
    * <p>
    * At the time this is called, the parent node will be the current node.
    * <p>
    * This is very useful for implementing scope-aware data structures (such as
    * a stack).  Scopes are normally pushed (added) on descent.
    *
    * @param    state
    *           Provides tree processing state including which node is being
    *           processed and the current status of the tree processing.
    */
   public void descent(WalkState state)
   {
      for (TreeWalker walker : list)
      {
         walker.descent(state);
      }
   }
   
   /**
    * Called whenever a transition from a child node (the "left sibling") to
    * its next right sibling takes place during tree traversal. This will be
    * called <em>before</em> the next sibling node in the tree traversal is
    * visited with {@link #walk}. This will be called <em>after</em> all the
    * child nodes (if any) of the left sibling have been processed. If the
    * left sibling has child nodes, there will be an {@link #ascent} call with
    * the left sibling in scope just before the next child call is executed.
    * This never occurs for nodes that have no children.
    * <p>
    * The parent node of the children is in scope at the time of the
    * notification. This effectively mimics recursion by notifying the
    * listener when one child node has been processed, before the next is
    * visited. This allows the listener to perform parent-level processing
    * between visits to child nodes.
    * <p>
    * This method is <em>not</em> invoked after the last child is visited,
    * since there is no next child to which to move. To obtain a callback
    * after the last child has processed, use the <code>ascent</code> method.
    * <p>
    * This callback is highly useful for implementing differential processing
    * based on the index of the child being processed.  For example, in
    * converting the tree form of a non-static Java method call (assume that
    * the first child is the referent upon which the method is called) into a
    * source code representation, one might need to output a ".methodName("
    * after the first child has been processed but before all other children
    * are processed.
    * <p>
    * This is generally useful in any case where processing needs to occur
    * between the processing of any two children.
    *
    * @param    state
    *           Provides tree processing state including which node is being
    *           processed and the current status of the tree processing.
    * @param    index
    *           The 0-based index of the child which is about to be processed.
    */
   public void nextChild(WalkState state, int index)
   {
      for (TreeWalker walker : list)
      {
         walker.nextChild(state, index);
      }
   }
   
   /**
    * Called whenever a transition from the last child node to its parent takes
    * place during tree traversal. This will be called <em>after</em> the
    * last child node in the tree traversal is visited with {@link #walk} AND
    * <em>after</em> all of the last child node's children are processed. This
    * never occurs for nodes that have no children.
    * <p>
    * At the time this is called, the parent node will be the current node.
    * <p>
    * This is very useful for implementing scope-aware data structures (such as
    * a stack).  Scopes are normally popped (deleted) on ascent.
    * <p>
    * In code that converts a tree to a human readable form, these hooks are
    * highly useful for closing processing after all children are processed.
    * For example, in converting the tree form of a function call (assume that
    * the children each represent a function call parameter) into a source code
    * version, one might need to output a closing ")" after all parameters are
    * emitted.
    *
    * @param    state
    *           Provides tree processing state including which node is being
    *           processed and the current status of the tree processing.
    */
   public void ascent(WalkState state)
   {
      for (TreeWalker walker : list)
      {
         walker.ascent(state);
      }
   }
   
   /**
    * Called once per tree, after all other hooks have been invoked. At the
    * time this is called, the tree's root node will be the current node.
    * <p>
    * This is useful for cleanup or termination processing. For example, this
    * is the common location for clearing or saving any values or state that
    * should be handled once per tree.
    *
    * @param    state
    *           Provides tree processing state including which node is being
    *           processed and the current status of the tree processing.
    */
   public void post(WalkState state)
   {
      for (TreeWalker walker : list)
      {
         walker.post(state);
      }
   }
   
   /**
    * Make a deep copy of this instance and return that new instance.
    *
    * @return   The copy of this instance.
    */
   public TreeWalker duplicate()
   {
      NestedWalker copy = new NestedWalker();
      
      copy.list = new ArrayList();
      
      for (TreeWalker walker : list)
      {
         copy.list.add(walker.duplicate());
      }
      
      return copy;
   }
}   