/*
** Module   : FastSiblingDeque.java
** Abstract : A linked deque (that works both as stack & queue) which exposes fast access getter
**             to prev/next siblings of elements.
**
** Copyright (c) 2016-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 OM  20160901 Created initial version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
package com.goldencode.util;

import java.util.*;

/**
 * A {@link Deque} with support for access of siblings of a particular element. It is logically
 * that, to uniquely identify the elements from this {@link Collection} for querying its sibling
 * there must NOT be duplicated entries. The implementation prevents this by throwing
 * exceptions or returning false at any attempt to duplicate an entry.
 * 
 * @param   <E>
 *          The type of the objects stored in this structure.
 */
public class FastSiblingDeque<E>
implements Deque<E>
{
   /** Internal data structure is a map for fast direct access to independent elements. */
   private Map<E, FsdNode<E>> data = new HashMap<>();
   
   /** The head of the deque. */
   private FsdNode<E> first = null;
   
   /** The tail of the deque. */
   private FsdNode<E> last = null;
   
   /**
    * Get the next element after {@code e}. If the collection is empty or the element is the last
    * one, then {@code null} is returned.
    * 
    * @param   e
    *          An element from collection to obtain its next sibling. 
    *
    * @return  The next element to follow {@code e} when iterating this collection.
    */
   public E getNext(E e)
   {
      final FsdNode<E> toReturn = data.get(e);
   
      if (toReturn == null || toReturn.next == null)
      {
         return null;
      }
      return toReturn.next.val;
   }
   
   /**
    * Get the previous element before {@code e}. If the collection is empty or the element is the
    * first one, then {@code null} is returned.
    * 
    * @param   e
    *          An element from collection to obtain its previous sibling. 
    *
    * @return  The next element to follow {@code e} when reverse iterating this collection.
    */
   public E getPrevious(E e)
   {
      final FsdNode<E> toReturn = data.get(e);
   
      if (toReturn == null || toReturn.prev == null)
      {
         return null;
      }
      return toReturn.prev.val;
   }
   
   @Override
   public void addFirst(E e)
   {
      if (data.containsKey(e))
      {
         throw new IllegalArgumentException("Elements must be unique!");
      }

      FsdNode<E> newLink = new FsdNode<>();
      newLink.val = e;
      newLink.next = first;
      
      // make connection to old first item 
      if (first != null)
      {
         first.prev = newLink;
      }
      first = newLink;
      
      if (last == null)
      {
         last = newLink; // this is the single link
      }
      
      data.put(e, newLink);
   }
   
   @Override
   public void addLast(E e)
   {
      if (data.containsKey(e))
      {
         throw new IllegalArgumentException("Elements must be unique!");
      }
      
      FsdNode<E> newLink = new FsdNode<>();
      newLink.val = e;
      newLink.prev = last;
      
      // make connection to old first item 
      if (last != null)
      {
         last.next = newLink;
      }
      last = newLink;
      
      if (first == null)
      {
         first = newLink; // this is the single link
      }
      
      data.put(e, newLink);
   }
   
   @Override
   public boolean offerFirst(E e)
   {
      if (data.containsKey(e))
      {
         return false;
      }

      addFirst(e);
      return true;
   }
   
   @Override
   public boolean offerLast(E e)
   {
      if (data.containsKey(e))
      {
         return false;
      }
      
      addLast(e);
      return true;
   }
   
   @Override
   public E removeFirst()
   {
      if (first == null)
      {
         throw new NoSuchElementException();
      }
      
      FsdNode<E> toReturn = first;
      
      if (last == first)
      {
         last = null;
      }
      else
      {
         first.next.prev = null;
      }
      first = first.next; // update the head
      
      data.remove(toReturn.val);
      return toReturn.val;
   }
   
   @Override
   public E removeLast()
   {
      if (last == null)
      {
         throw new NoSuchElementException();
      }
   
      FsdNode<E> toReturn = last;
   
      if (first == last)
      {
         first = null;
      }
      else
      {
         last.prev.next = null;
      }
      last = last.prev; // update the tail
   
      data.remove(toReturn.val);
      return toReturn.val;
   }
   
   @Override
   public E pollFirst()
   {
      if (first == null)
      {
         return null;
      }
      
      return removeFirst();
   }
   
   @Override
   public E pollLast()
   {
      if (last == null)
      {
         return null;
      }
      
      return removeLast();
   }
   
   @Override
   public E getFirst()
   {
      if (first == null)
      {
         throw new NoSuchElementException();
      }
      return first.val;
   }
   
   @Override
   public E getLast()
   {
      if (last == null)
      {
         throw new NoSuchElementException();
      }
      return last.val;
   }
   
   /**
    * Retrieves, but does not remove, the first element of this deque, or returns {@code null} if
    * this deque is empty.
    *
    * @return the head of this deque, or {@code null} if this deque is empty
    */
   @Override
   public E peekFirst()
   {
      return first == null ? null : first.val;
   }
   
   /**
    * Retrieves, but does not remove, the last element of this deque, or returns {@code null} if
    * this deque is empty.
    *
    * @return the tail of this deque, or {@code null} if this deque is empty
    */
   @Override
   public E peekLast()
   {
      return last == null ? null : last.val;
   }
   
   @Override
   public boolean removeFirstOccurrence(Object o)
   {
      FsdNode<E> toReturn = data.get(o);
      if (toReturn == null)
      {
         return false;
      }
      
      if (toReturn != last) // toReturn.next != null
      {
         toReturn.next.prev = toReturn.prev;
      }
      else
      {
         last = toReturn.prev;
      }
      
      if (toReturn != first) // toReturn.prev != null
      {
         toReturn.prev.next = toReturn.next;
      }
      else
      {
         first = toReturn.next;
      }
      
      data.remove(o);
      return true;
   }
   
   @Override
   public boolean removeLastOccurrence(Object o)
   {
      return removeFirstOccurrence(o); // objects are unique in this structure
   }
   
   /**
    * Inserts the specified element into the queue represented by this deque (in other words, at
    * the tail of this deque) if it is possible to do so immediately without violating capacity
    * restrictions, returning {@code true} upon success and throwing an
    * {@code IllegalStateException} if no space is currently available.
    *
    * <p>This method is equivalent to {@link #addLast}.
    *
    * @param   e
    *          the element to add
    *
    * @return  {@code true} (as specified by {@link Collection#add})
    *
    * @throws  ClassCastException
    *          if the class of the specified element prevents it from being added to this deque
    */
   @Override
   public boolean add(E e)
   {
      addLast(e);
      return true;
   }
   
   @Override
   public boolean offer(E e)
   {
      return offerLast(e);
   }
   
   @Override
   public E remove()
   {
      return removeFirst();
   }
   
   @Override
   public E poll()
   {
      return pollFirst();
   }
   
   @Override
   public E element()
   {
      return getFirst();
   }
   
   @Override
   public E peek()
   {
      return peekFirst();
   }
   
   @Override
   public void push(E e)
   {
      addFirst(e);
   }
   
   @Override
   public E pop()
   {
      return removeFirst();
   }
   
   /**
    * Removes the first occurrence of the specified element from this deque. If the deque does not
    * contain the element, it is unchanged.
    * Returns {@code true} if this deque contained the specified element (or equivalently, if this
    * deque changed as a result of the call).
    *
    * @param   o
    *          element to be removed from this deque, if present
    *
    * @return  {@code true} if an element was removed as a result of this call
    *
    * @throws  ClassCastException if the class of the specified element is incompatible with this
    *          structure
    */
   @Override
   public boolean remove(Object o)
   {
      return removeFirstOccurrence(o);
   }
   
   @Override
   public boolean containsAll(Collection<?> c)
   {
      for (Object o : c)
      {
         if (!data.containsKey(o))
         {
            return false;
         }
      }
      return true;
   }
   
   @Override
   public boolean addAll(Collection<? extends E> c)
   {
      boolean modified = false;
      for (E e : c)
      {
         if (offerLast(e))
         {
            modified = true;
         }
      }
      return modified;
   }
   
   @Override
   public boolean removeAll(Collection<?> c)
   {
      // TODO: to be optimized
      boolean modified = false;
      Iterator<?> it = iterator();
      while (it.hasNext())
      {
         if (c.contains(it.next()))
         {
            it.remove();
            modified = true;
         }
      }
      return modified;
   }
   
   @Override
   public boolean retainAll(Collection<?> c)
   {
      // TODO: to be optimized
      boolean modified = false;
      Iterator<E> it = iterator();
      while (it.hasNext())
      {
         if (!c.contains(it.next()))
         {
            it.remove();
            modified = true;
         }
      }
      return modified;
   }
   
   @Override
   public void clear()
   {
      first = null;
      last = null;
      data = new HashMap<>();
   }
   
   /**
    * Returns {@code true} if this deque contains the specified element.
    *
    * @param   o
    *          element whose presence in this deque is to be tested
    *
    * @return  {@code true} if this deque contains the specified element
    *
    * @throws  ClassCastException
    *          if the type of the specified element is incompatible with this deque
    */
   @Override
   public boolean contains(Object o)
   {
      return data.containsKey(o);
   }
   
   @Override
   public int size()
   {
      return data.size();
   }
   
   /**
    * Returns  {@code true} if this collection contains no elements.
    *
    * @return  {@code true} if this collection contains no elements
    */
   @Override
   public boolean isEmpty()
   {
      return first == null;
   }
   
   @Override
   public Iterator<E> iterator()
   {
      return new Iterator<E>()
      {
         private FsdNode<E> crt = null; // this means 'before first' 
         
         @Override
         public boolean hasNext()
         {
            return (crt != null) ? (crt.next != null) : (first != null);
         }
   
         @Override
         public E next()
         {
            if (crt == null)
            {
               if (first == null)
               {
                  throw new NoSuchElementException();
               }
               else
               {
                  crt = first;
               }
            }
            else 
            {
               if (crt.next == null)
               {
                  throw new NoSuchElementException();
               }
               else
               {
                  crt = crt.next;
               }
            }
            
            return crt.val;
         }
         
         @Override
         public void remove()
         {
            FastSiblingDeque.this.remove(crt.val);
         }
      };
   }
   
   @Override
   public Object[] toArray()
   {
      Object[] ret = new Object[data.size()];
      return toArray(ret);
   }
   
   @Override
   @SuppressWarnings("unchecked")
   public <T> T[] toArray(T[] a)
   {
      int size = data.size();
      if (a.length < size)
      {
         a = (T[]) java.lang.reflect.Array.newInstance(a.getClass().getComponentType(), size);
      }
      
      FsdNode<E> crt = first;
      for (int k = 0; k < a.length; k++)
      {
         a[k] = (T) crt.val;
         crt = crt.next;
      }
   
      if (a.length > size)
      {
         a[size] = null;
      }
      
      return a;
   }
   
   @Override
   public Iterator<E> descendingIterator()
   {
      return new Iterator<E>()
      {
         private FsdNode<E> crt = null; // this means 'after last'
      
         @Override
         public boolean hasNext()
         {
            return (crt != null) ? (crt.prev != null) : (last != null);
         }
      
         @Override
         public E next()
         {
            if (crt == null)
            {
               if (last == null)
               {
                  throw new NoSuchElementException();
               }
               else
               {
                  crt = last;
               }
            }
            else
            {
               if (crt.prev == null)
               {
                  throw new NoSuchElementException();
               }
               else
               {
                  crt = crt.prev;
               }
            }
            
            return crt.val;
         }
         
         @Override
         public void remove()
         {
            FastSiblingDeque.this.remove(crt.val);
         }
      };
   }
   
   /**
    * A node in double-linked list of {@link FastSiblingDeque}.
    * 
    * @param   <E>
    *          The type of the objects stored in parent class collection.
    */
   private static class FsdNode<E>
   {
      /** The current node value. */
      E val;
      
      /** Pointer to previous element in linked list, if any. */
      FsdNode<E> prev = null;
   
      /** Pointer to next element in linked list, if any. */
      FsdNode<E> next = null;
   }
   
   /** Quick testing. */
   public static void main(String[] args)
   {
      FastSiblingDeque<Integer> deck = new FastSiblingDeque<>();
      assert deck.isEmpty();
      assert deck.size() == 0;
   
      assert deck.addAll(Arrays.asList(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 ));
      assert deck.size() == 11;
   
      assert deck.remove(3);
      assert deck.size() == 10;
      assert deck.getNext(2) == 4;
      assert deck.getPrevious(4) == 2;
   
      assert deck.add(3);
      assert deck.size() == 11;
      assert deck.getFirst() == 0;
      assert deck.getLast() == 3;
      assert deck.getNext(10) == 3;
      assert deck.getNext(3) == null;
      
      System.out.println("Current collection: " + Arrays.toString(deck.toArray()));
      System.out.print("Ascending iteration: {");
      for (Integer i : deck)
      {
         if (deck.getNext(i) != null)
         {
            System.out.print(i + ", ");
         }
         else
         {
            System.out.println(i + "}");
         }
      }
      System.out.print("Descending iteration: {");
      final Iterator<Integer> revIt = deck.descendingIterator();
      while (revIt.hasNext())
      {
         Integer i = revIt.next();
         if (deck.getPrevious(i) != null)
         {
            System.out.print(i + ", ");
         }
         else
         {
            System.out.println(i + "}");
         }
      }
      
      assert deck.removeAll(Arrays.asList(1, 3, 5, 7, 9, 11, 13));
      assert deck.size() == 6;
      System.out.println("Current collection: " + Arrays.toString(deck.toArray()));
      
      assert deck.retainAll(Arrays.asList(0, 1, 2, 3, 4, 5));
      assert deck.size() == 3;
      System.out.println("Current collection: " + Arrays.toString(deck.toArray()));
      
      deck.clear();
      assert (deck.isEmpty());
      assert deck.add(99);
      assert deck.getFirst() == 99;
      assert deck.getLast() == 99;
      assert deck.getNext(99) == null;
      assert deck.getPrevious(99) == null;
      
      assert deck.offerFirst(-99);
      System.out.println("Current collection: " + Arrays.toString(deck.toArray()));
      assert deck.size() == 2;
      assert deck.getFirst() == -99;
      assert deck.getLast() == 99;
      assert deck.getNext(99) == null;
      assert deck.getPrevious(99) == -99;
      assert deck.getNext(-99) == 99;
      assert deck.getPrevious(-99) == null;
      
      assert deck.removeLast() == 99;
      assert deck.size() == 1;
      assert deck.getFirst() == -99;
      assert deck.getLast() == -99;
      assert deck.getNext(-99) == null;
      assert deck.getPrevious(-99) == null;
      
      assert deck.remove(-99);
      assert deck.isEmpty();
   }
}