/*
** Module   : GenericSAXHandler.java
** Abstract : General purpose SAX handler for parsing arbitrary XML data into Java objects.
**
** Copyright (c) 2010-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 LMR 20110103 Created the initial version; supports recursive composite structures and
**                  context sensitive type mapping (tags that map to different Java types
**                  depending on where in the XML tree they appear).
** 002 ECF 20150715 Replace StringBuffer with StringBuilder.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.util;

import java.util.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import com.goldencode.p2j.cfg.*;

/**
 * Generic SAX handler for building Java objects from data stored as XML. The
 * end result is a list of top-level objects that are built and populated as 
 * XML content is parsed. When parsing is done, objects can be retrieved from 
 * the list by the name of the XML node they were built from, by calling the
 * {@link #getGlobalObjects(String)} method.
 * <p>
 * The handler can be configured to produce objects of arbitrary types by
 * supplying instances of custom <code>Mapper</code> subclasses to it: each 
 * such "mapper" is responsible for loading the XML data under a particular 
 * XML element (attributes and child elements) into a "payload" Java object of
 * the right type.
 * <p>
 * The default behavior provided by <code>Mapper</code> base class is to 
 * attempt to relate XML attributes to setter methods in the Java class, on a
 * best effort basis, following the Java bean conventions. Attempts are made 
 * to convert <code>String</code> attribute values read from XML to the
 * respective properties' own types. In cases where this default approach does
 * not work, a custom property setter can be defined by <code>Mapper</code> 
 * subclasses, overriding the one in the base ({@link
 * Mapper#setProperty(Object, String, String)}).
 * <p>
 * As the nested structure of XML nodes reflect the composite nature of Java
 * objects, mappers also have the duty of controlling how objects that spring
 * from a certain sub-tree of the XML document are "consumed" by the parent
 * object, corresponding to its ancestor node in the XML tree. Unlike setting
 * properties, no assumptions can be made here regarding how an object is to 
 * be integrated in its container, therefore the <code>Mapper</code> subclass
 * for the parent object would have to override method {@link 
 * Mapper#mergeSubObj(Object, Object)}.
 * <p>
 * Recursive composite structures (types that contain instances of their own 
 * type in the object structure) are supported by mappers by default. The same
 * mapper instance can be used simultaneously with any number of objects in 
 * the process of being constructed at a given time and <code>Mapper</code>
 * subclasses can learn about the context where they're operating by using the
 * convenience methods provided by <code>Mapper</code> base class ({@link 
 * Mapper#getXmlPath()}, {@link Mapper#getParentTag()}).
 * <p>
 * Mapping XML nodes to Java types is also sensitive to context. The same XML
 * tag can be mapped to different types, depending on where in the XML 
 * tree it appears. If a mapper is installed for a more specific context 
 * (e.g. "cfg/schema/namespace/dialect-specific/parameter"), it takes 
 * precedence over a general purpose mapper installed for the simple tag 
 * name (in this case "parameter").
 * <p>
 * Constraints can be added for data that is read from XML: for any element, 
 * sets of mandatory and optional attributes as well as required and optional 
 * child elements can be specified with {@link 
 * GenericSAXHandler#setConstraints(String, String[], String[], String[], 
 * String[])}.
 * <p>
 * The handler can be instructed to regard particular XML tag names as
 * "unhandled" (or "transparent"). XML elements having those names are not
 * considered for translating into Java objects. Child elements of such a node
 * are perceived as direct children of the node's first normal 
 * (non-transparent) ancestor. In case all nodes up to the root are unhandled,
 * the respective child objects land as top-level objects in the global
 * end-result object list. This dismisses the constraint of having a 1:1 
 * correspondence between Java types and XML tag names, which increases 
 * flexibility.
 * <p>
 * This class is not synchronized.
 * 
 * @see Mapper 
 */
public class GenericSAXHandler extends DefaultHandler
{
   /** Separator for paths in the XML tree. */
   public static char XMLPATH_SEPARATOR = '/';
   /**
    * Container for all <code>Mapper</code> objects, indexed by the name of 
    * the XML tag they process.
    */
   private Map<String, Mapper<?>> mappers = new HashMap<String, Mapper<?>>();

   /**
    * The stack of objects currently being processed (populated), each by its
    * own kind of mapper.
    */
   private Stack<ActiveNode<?>> inProcess = new Stack<ActiveNode<?>>();

   /** XML tags don't relate to mappers 1:1 so we keep track of tags here. */
   private Stack<String> tagsInProcess = new Stack<String>();

   /**
    * The set of "unhandled" tags that this handler was configured with.
    * Unhandled tags are "transparent", in that they are not considered for
    * encapsulation but have their content silently unfolded and processed as
    * if they belonged to their parent.
    */
   private Set<String> unhandledTags = new HashSet<String>();

   /**
    * The container for top-level objects that were constructed by this 
    * handler during the parsing process. The objects are mapped by the name 
    * of the XML tag they were built from.
    * 
    */
   private Map<String, List<Object>> globalObjs = 
      new HashMap<String, List<Object>>();

   /**
    * Notification of beginning of an XML element.
    * 
    * @param  uri
    *         The Namespace URI, or the empty string if the element has no
    *         Namespace URI or if Namespace processing is not being performed.
    * @param  localName
    *         The local name (without prefix), or the empty string if
    *         Namespace processing is not being performed.
    * @param  qName
    *         The qualified name (with prefix), or the empty string if
    *         qualified names are not available.
    * @param  attrs
    *         The attributes attached to the element. If there are no
    *         attributes, it shall be an empty Attributes object.
    * 
    * @throws org.xml.sax.SAXException
    *         On any SAX errors, possibly wrapping another exception.
    * @throws RuntimeException
    *         wrapping a <code>ConfigurationException</code> if any error 
    *         occurs while the objects are being populated with data read 
    *         from XML. This includes file format constraint violations 
    *         detected during parsing.
    *         
    * @see    org.xml.sax.ContentHandler#startElement
    */
   @Override
   public void startElement(String uri, 
                            String localName, 
                            String qName,
                            org.xml.sax.Attributes attrs) 
   throws SAXException, 
          RuntimeException
   {
      StringBuilder sb = new StringBuilder(getCurrentPath());
      sb.append(XMLPATH_SEPARATOR);
      sb.append(qName);
      String path = sb.toString();
      String mapperKey = null;
      
      tagsInProcess.push(qName);
      
      if (unhandledTags.contains(qName))
      {
         // don't push anything on stack
         return;
      }
      
      // first look for a specific mapper for the current context
      mapperKey = path;
      Mapper<?> m = mappers.get(mapperKey);
      
      if (m == null)
      {
         mapperKey = qName;
         m = mappers.get(mapperKey);
         if (m == null)
         {
            throw new RuntimeException(new ConfigurationException("No " +
               "mapper configured for XML tag <" + qName + ">, appearing "+
               "at: <" + path + ">"));
         }
      }

      m.setXmlPath(path);
      
      // this also creates the payload obj
      ActiveNode<?> pair = m.createActiveNode(mapperKey);
      
      try
      {
         pair.runApplyAttributes(attrs);

         // check for unexpected children early:
         if (inProcess.size() > 0)
         {
            inProcess.peek().checkChild(qName);
         }
      }
      catch (ConfigurationException e)
      {
         throw new RuntimeException(e);
      }
      
      inProcess.push(pair);
   }

   /**
    * Notification of the end of an XML element.
    * 
    * @param  uri
    *         The Namespace URI, or the empty string if the element has no
    *         Namespace URI or if Namespace processing is not being performed.
    * @param  localName
    *         The local name (without prefix), or the empty string if
    *         Namespace processing is not being performed.
    * @param  qName
    *         The qualified name (with prefix), or the empty string if
    *         qualified names are not available.
    * 
    * @throws org.xml.sax.SAXException
    *         On any SAX error, possibly wrapping another exception.
    * @throws RuntimeException
    *         wrapping a <code>ConfigurationException</code> if any error 
    *         occurs while the objects are being populated with data read 
    *         from XML. This includes file format constraint violations 
    *         detected during parsing.
    *         
    * @see org.xml.sax.ContentHandler#endElement
    */
   @Override
   public void endElement(String uri, String localName, String qName)
   throws SAXException, 
          RuntimeException   
   {
      String cpath = getCurrentPath();
      String tag = tagsInProcess.pop();
      assert(qName.equals(tag));
      
      if (unhandledTags.contains(qName))
      {
         return;
      }
      assert (inProcess.size() > 0);
      //this tag's validity was checked in startElement()
      ActiveNode<?> c = inProcess.pop();

      //check for missing attrs/children:
      try
      {
         c.checkMissing();
      }
      catch (ConfigurationException e)
      {
         throw new RuntimeException(e);
      }
      
      c.mapper.setXmlPath(cpath);
      if (inProcess.size() > 0)
      {
         ActiveNode<?> p = inProcess.peek();
         p.runMergeSubObj(c.payloadObj);
      }
      else
      {
         //save away global objects:
         saveGlobalObject(c.mapperKey, c.payloadObj);
      }
   }

   /**
    * Provides the list of top-level objects that were constructed from
    * top-level XML nodes having a specific name.
    * 
    * @param  name
    *         The XML tag / XML path that was used to install the mapper that
    *         built the object.
    *         
    * @return The list of all objects that built from XML nodes having the 
    *         given name.
    */
   public List<Object> getGlobalObjects(String name)
   {
      List<Object> ret = globalObjs.get(name);
      if (ret != null)
      {
         return ret;
      }
      return new ArrayList<Object>();
   }

   /**
    * Configures the handler with a specific <code>Mapper</code> for handling
    * XML nodes.
    * 
    * @param xmlKey
    *        A string identifying the XML nodes that will be handled by the 
    *        provided mapper. Normally, for general purpose mappers this is a
    *        simple tag name, but it can also be a full path in the XML tree
    *        denoting the specific context where the mapper is going to be 
    *        employed. The path uses {@link #XMLPATH_SEPARATOR} as separator, 
    *        defined as '/'.
    *         
    *        
    * @param m
    *        The concrete mapper instance to be used for handling the 
    *        designated XML nodes.
    */
   public void installMapper(String xmlKey, Mapper<?> m)
   {
      mappers.put(xmlKey, m);
   }

   /**
    * Adds constraints for a specific kind of XML nodes, in the form of 
    * mandatory and optional attributes and/or child elements. The sets of 
    * mandatory and optional items have different meaning with respect to 
    * empty set and <code>null</code> value:
    * <p>
    * If the set of mandatory attributes/children is empty or <code>null
    * </code>, the respective constraint is disabled: there are no required
    * attributes or child elements, respectively. The meaning for the sets of
    * optional items is different: <code>null</code> means the constraint is
    * disabled (any elements can appear), while the empty set indicates that
    * there are no optional elements.
    * <p>
    * The way the XML node name that the constraints apply to is specified is
    * similar to how mappers are installed for specific nodes.  
    * 
    * @param  xmlKey
    *         The name of the XML node that the constraints apply to. Can be a
    *         simple tag name or a specific path in the XML tree, using 
    *         {@link #XMLPATH_SEPARATOR} as separator (defined to '/'). 
    * @param  reqAttrs
    *         Array containing mandatory attribute names.
    * @param  optAttrs
    *         Array containing optional attribute names.
    * @param  reqChld
    *         Array containing mandatory child names.
    * @param  optChld
    *         Array containing optional attribute names.
    * 
    * @throws IllegalArgumentException
    *         If the tag name is invalid.
    */
   public void setConstraints(String xmlKey, 
                              String[] reqAttrs,
                              String[] optAttrs, 
                              String[] reqChld, 
                              String[] optChld)
   throws IllegalArgumentException
   {
      if (!mappers.containsKey(xmlKey))
      {
         if (unhandledTags.contains(xmlKey))
         {
            throw new IllegalArgumentException("XML tag <" + xmlKey
                  + "> is unhandled, it can't have constraints added to it.");
         }
         throw new IllegalArgumentException("Bad XML tag name <" + xmlKey + ">");
      }
      mappers.get(xmlKey).setConstraints(reqAttrs, 
                                         optAttrs, 
                                         reqChld, 
                                         optChld);
   }

   /**
    * Resets the state built during processing; leaves configured data in 
    * place (installed mappers, constraints, unhandled tags).
    */
   public void resetState()
   {
      inProcess.clear();
      tagsInProcess.clear();
      globalObjs.clear();
   }

   /**
    * Resets handler state and also discards all configuration data (installed
    * mappers, constraints, unhandled tags).
    */
   public void reset()
   {
      resetState();
      mappers.clear();
      unhandledTags.clear();
   }
   
   /**
    * Unhandled tags are "transparent", in that they are not considered for
    * encapsulation but have their content silently unfolded and processed as
    * if it belonged to their parent.
    * 
    * @param tags
    *        The set of unhandled XML tags.
    */
   public void setUnhandledTags(String[] tags)
   {
      unhandledTags.clear();
      if (tags != null)
      {
         unhandledTags.addAll(Arrays.asList(tags));
      }
   }

   /**
    * Save away (as end-result) a top-level object that was constructed from
    * XML.
    * 
    * @param name
    *        The name of the XML node corresponding to the constructed
    *        object.
    * @param g
    *        The constructed object.
    */
   private void saveGlobalObject(String name, Object g)
   {
      List<Object> objs = globalObjs.get(name);
      if (objs == null)
      {
         objs = new ArrayList<Object>();
         globalObjs.put(name, objs);
      }
      objs.add(g);
   }

   /**
    * Gives the full path in the XML tree from the root to (and including) the
    * current node being processed. The path uses {@link #XMLPATH_SEPARATOR} 
    * as separator, defined as '/'.
    *
    * @return The path from root to current node as a <code>String</code>.
    */
   private String getCurrentPath()
   {
      StringBuilder ret = new StringBuilder("");
      Iterator<String> it = tagsInProcess.iterator();
      while (it.hasNext())
      {
         ret.append(it.next());
         if (it.hasNext())
         {
            ret.append(XMLPATH_SEPARATOR);
         }
      }
      return ret.toString();
   }
   
   /**
    * Simple convenience container to decouple the mapper from the object it
    * constructs, without breaking generic type safety. 
    *  
    * @param <T>
    *        The type of the end-result payload object.
    */
   static final class ActiveNode<T>
   {
      /** The mapper that constructs the payload object. */
      private Mapper<T> mapper;
      
      /** The payload object being constructed and populated.*/
      private T payloadObj;
      
      /**
       * The key that was used to install and find this mapper - can be a tag
       * name or an XML path, using {@link #XMLPATH_SEPARATOR} as separator
       * (defined as '/'). The object produced by the mapper will be
       * retrieved from the global registry of objects by the same key.   
       */
      private String mapperKey;

      /**
       * Mandatory attributes which are found to be missing from the XML node 
       * handled by this mapper.
       */
      private Set<String> missingAttrs = new HashSet<String>();

      /**
       * Mandatory child elements that are found to be missing from the XML 
       * node handled by this mapper.
       */
      private Set<String> missingChld = new HashSet<String>();
      
      /**
       * Builds an <code>ActiveNode</code> container instance.
       * 
       * @param m
       *        The mapper.
       * @param t
       *        The payload object, that will be populated by the mapper.
       * @param mapKey
       *        The key identifying the mapper, the same that was used when
       *        the mapper was installed. 
       */
      ActiveNode(Mapper<T> m, T t, String mapKey)
      {
         if (m == null || t == null || mapKey == null)
         {
            throw new IllegalArgumentException("Null argument(s)");
         }
         mapper = m;
         payloadObj = t;
         mapperKey = mapKey;
         resetConstraintsCheckState();
      }
      
      /**
       * Convenience method to capture the wildcard in the caller and use the
       * type consistently. At the same time this updates the constraint 
       * check state. 
       * 
       * @param  attrs
       *         The attributes to set into the payload object.
       *         
       * @throws ConfigurationException
       *         Thrown by {@link Mapper#applyAttributes(Attributes, Object)}.
       * @throws IllegalStateException 
       *         Thrown by {@link Mapper#applyAttributes(Attributes, Object)}.
       * 
       * @see    Mapper#applyAttributes(Attributes, Object)
       */
      private final void runApplyAttributes(Attributes attrs)
      throws ConfigurationException,
             IllegalStateException
      {
         mapper.applyAttributes(attrs, payloadObj);
         int n = attrs.getLength();
         for (int i = 0; i<n; i++)
         {
            missingAttrs.remove(attrs.getQName(i));
         }
      }
      
      /**
       * Utility method to capture the wildcard in the caller and use the
       * type consistently.
       * 
       * @param c 
       *        The object to be merged into this node's payload object.
       *        
       * @see   Mapper#mergeSubObj(Object, Object)       
       */
      private final void runMergeSubObj(Object c)
      {
         mapper.mergeSubObj(payloadObj, c);
      }

      /**
       * Checks if any mandatory attributes and child names are missing and 
       * throws a <code>ConfigurationException</code> if it's the case. 
       *  
       * @throws ConfigurationException
       *         If mandatory attributes/child elements are found to be 
       *         missing.
       */
      private final void checkMissing() 
      throws ConfigurationException
      {
         if (missingAttrs.size() > 0)
         {
            throw new ConfigurationException("The following mandatory " +
               "properties were not set in element <" + mapper.getXmlPath() + 
               ">: " + missingAttrs);
         }
         if (missingChld.size() > 0)
         {
            throw new ConfigurationException("The following mandatory " +
               "chidren are not present in element <" + mapper.getXmlPath() +
               ">: " + missingChld);
         }
      }      
      
      /**
       * During parsing, we keep track of each node's mandatory attributes and
       * child elements that have not yet been encountered. This  method
       * (re)initializes this state information.
       */
      private final void resetConstraintsCheckState()
      {
         missingAttrs.clear();
         missingChld.clear();
         if (mapper.requiredAttrs != null)
         {
            missingAttrs.addAll(mapper.requiredAttrs);
         }
         if (mapper.requiredChld != null)
         {
            missingChld.addAll(mapper.requiredChld);
         }
      }   
      
      /**
       * Checks if a child element can appear within the node processed by 
       * the current mapper: if constraints are active (defined) and the child
       * is neither mandatory nor optional, a <code>ConfigurationException
       * </code> is thrown, otherwise the constraint check state is updated
       * to reflect this child's existence.
       *   
       * @param  chld
       *         The name of the XML child element that is being checked.
       *        
       * @throws ConfigurationException
       *         If the child is unexpected.
       */
      private final void checkChild(String chld) 
      throws ConfigurationException
      {
         mapper.checkUnexpectedChild(chld);
         missingChld.remove(chld);
      }
   }
}