/*
** Module   : NativeTypeSerializer.java
** Abstract : Helper class to serialize native values, when the value's type is unknown at compile
**            time.
**
** Copyright (c) 2014-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20141002 Created initial version.
** 002 IAS 20150702 Added cautious null processing.
*/ 
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.util;

import java.io.*;
import java.lang.reflect.*;

/**
 * Helper class to serialize native values, useful in cases when the value's type is unknown at
 * compile time (i.e. reflection is used to get values from a {@link Field}).
 * <p>
 * When using this class, all writing needs to be done using the {@link #writeValue} API and it
 * must be read using the {@link #readValueWorker} API.
 * <p>
 * All Java 7 native types are supported, plus {@link String}.  If the class passed to 
 * {@link #getWorker} is not a native type or {@link String}, then it will fallback to serializing
 * the value as a plain {@link Object}.
 */
public abstract class NativeTypeSerializer
{
   /** A serializer for native <code>byte</code> values. */
   public static final ByteSerializer BYTE_SERIALIZER = new ByteSerializer();
  
   /** A serializer for native <code>short</code> values. */
   public static final ShortSerializer SHORT_SERIALIZER = new ShortSerializer();
   
   /** A serializer for native <code>int</code> values. */
   public static final IntSerializer INT_SERIALIZER = new IntSerializer();
   
   /** A serializer for native <code>long</code> values. */
   public static final LongSerializer LONG_SERIALIZER = new LongSerializer();
   
   /** A serializer for native <code>float</code> values. */
   public static final FloatSerializer FLOAT_SERIALIZER = new FloatSerializer();
   
   /** A serializer for native <code>double</code> values. */
   public static final DoubleSerializer DOUBLE_SERIALIZER = new DoubleSerializer();
   
   /** A serializer for native <code>boolean</code> values. */
   public static final BooleanSerializer BOOLEAN_SERIALIZER = new BooleanSerializer();
   
   /** A serializer for native <code>char</code> values. */
   public static final CharSerializer CHAR_SERIALIZER = new CharSerializer();
   
   /** A serializer for {@link String} values. */
   public static final StringSerializer STRING_SERIALIZER = new StringSerializer();
   
   /** Fallback serializer for other {@link Object} instances. */
   public static final ObjectSerializer OBJECT_SERIALIZER = new ObjectSerializer();
   
   /** An array of workers, identified by their {@link #mark}. */
   private static final NativeTypeSerializer[] workers = new NativeTypeSerializer[]
   {
      BYTE_SERIALIZER,
      SHORT_SERIALIZER,
      INT_SERIALIZER,
      LONG_SERIALIZER,
      FLOAT_SERIALIZER,
      DOUBLE_SERIALIZER,
      BOOLEAN_SERIALIZER,
      CHAR_SERIALIZER,
      STRING_SERIALIZER,
      OBJECT_SERIALIZER
   };

   /** A special mark to identify the type of the written value, when reading. */
   protected final byte mark;

   /**
    * Create a new serializer and set its mark.
    * 
    * @param    mark
    *           The serializer's mark.
    */
   protected NativeTypeSerializer(int mark)
   {
      this.mark = (byte) mark;
   }
   
   /**
    * Write the given value using the specified {@link NativeTypeSerializer worker}, to the 
    * specified destination.
    * <p>
    * The first written byte will be the worker's {@link #mark}, followed by the serialized value.
    * 
    * @param    value
    *           The value to be written.
    * @param    worker
    *           The worker to write the value.
    * @param    out
    *           The destination.
    *           
    * @throws   IOException
    *           In case of I/O errors.
    */
   public static void writeValue(Object value, NativeTypeSerializer worker, ObjectOutput out)
   throws IOException
   {
      if (value == null)
      {
         out.writeByte(OBJECT_SERIALIZER.mark);
         OBJECT_SERIALIZER.writeValueWorker(value, out);
         return;
      }
      out.writeByte(worker.mark);
      worker.writeValueWorker(value, out);
   }

   /**
    * Read the given value using the specified {@link NativeTypeSerializer worker}, from the 
    * specified destination.
    * <p>
    * The first read byte will be the worker's {@link #mark}, followed by the serialized value.
    * 
    * @param    in
    *           The source.
    * 
    * @return   The read value.
    * 
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If the worker is {@link ObjectSerializer} and the object's class can not be found.
    */
   public static Object readValue(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      byte worker = in.readByte();
      NativeTypeSerializer cfg = workers[worker];
      return cfg.readValueWorker(in);
   }

   /**
    * Get the {@link NativeTypeSerializer} associated with the given {@link Class}.
    * 
    * @param    ftype
    *           The class.
    *           
    * @return   See above.
    */
   public static NativeTypeSerializer getWorker(Class<?> ftype)
   {
      if (ftype == byte.class)
      {
         return BYTE_SERIALIZER;
      }
      else if (ftype == short.class)
      {
         return SHORT_SERIALIZER;
      }
      else if (ftype == int.class)
      {
         return INT_SERIALIZER;
      }
      else if (ftype == long.class)
      {
         return LONG_SERIALIZER;
      }
      else if (ftype == float.class)
      {
         return FLOAT_SERIALIZER;
      }
      else if (ftype == double.class)
      {
         return DOUBLE_SERIALIZER;
      }
      else if (ftype == boolean.class)
      {
         return BOOLEAN_SERIALIZER;
      }
      else if (ftype == char.class)
      {
         return CHAR_SERIALIZER;
      }
      else if (ftype == String.class)
      {
         return STRING_SERIALIZER;
      }
      else
      {
         return OBJECT_SERIALIZER;
      }
   }
   
   /**
    * A worker to write the value to the specified destination.
    * 
    * @param    value
    *           The value to be written.
    * @param    out
    *           The destination.
    *           
    * @throws   IOException
    *           In case of I/O errors.
    */
   protected abstract void writeValueWorker(Object value, ObjectOutput out)
   throws IOException;

   /**
    * A worker to read the value from the specified source.
    * 
    * @param    in
    *           The source.
    *           
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If the worker is {@link ObjectSerializer} and the object's class can not be found.
    */
   protected abstract Object readValueWorker(ObjectInput in)
   throws IOException,
          ClassNotFoundException;

   /**
    * A serializer for native <code>byte</code> values.
    */
   private static class ByteSerializer
   extends NativeTypeSerializer
   {
      /**
       * Create serializer for native <code>byte</code> values and set its 
       * {@link NativeTypeSerializer#mark} to <code>0</code>.
       */
      protected ByteSerializer()
      {
         super(0);
      }
      
      /**
       * Worker to write the value to the specified destination, as a native <code>byte</code>.
       * 
       * @param    value
       *           The value to be written.
       * @param    out
       *           The destination.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected void writeValueWorker(Object value, ObjectOutput out)
      throws IOException
      {
         out.writeByte((byte) value);
      }

      /**
       * A worker to read the value from the specified source, as a native <code>byte</code>.
       * 
       * @param    in
       *           The source.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected Byte readValueWorker(ObjectInput in)
      throws IOException
      {
         return in.readByte();
      }
   }
   
   /**
    * A serializer for native <code>short</code> values.
    */
   private static class ShortSerializer
   extends NativeTypeSerializer
   {
      /**
       * Create serializer for native <code>short</code> values and set its 
       * {@link NativeTypeSerializer#mark} to <code>1</code>.
       */
      protected ShortSerializer()
      {
         super(1);
      }

      /**
       * Worker to write the value to the specified destination, as a native <code>short</code>.
       * 
       * @param    value
       *           The value to be written.
       * @param    out
       *           The destination.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected void writeValueWorker(Object value, ObjectOutput out)
      throws IOException
      {
         out.writeShort((short) value);
      }

      /**
       * A worker to read the value from the specified source, as a native <code>short</code>.
       * 
       * @param    in
       *           The source.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected Short readValueWorker(ObjectInput in)
      throws IOException
      {
         return in.readShort();
      }
   }
   
   /**
    * A serializer for native <code>int</code> values.
    */
   private static class IntSerializer
   extends NativeTypeSerializer
   {
      /**
       * Create serializer for native <code>int</code> values and set its 
       * {@link NativeTypeSerializer#mark} to <code>2</code>.
       */
      protected IntSerializer()
      {
         super(2);
      }

      /**
       * Worker to write the value to the specified destination, as a native <code>int</code>.
       * 
       * @param    value
       *           The value to be written.
       * @param    out
       *           The destination.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected void writeValueWorker(Object value, ObjectOutput out)
      throws IOException
      {
         out.writeInt((int) value);
      }

      /**
       * A worker to read the value from the specified source, as a native <code>int</code>.
       * 
       * @param    in
       *           The source.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected Integer readValueWorker(ObjectInput in)
      throws IOException
      {
         return in.readInt();
      }
   }
   
   /**
    * A serializer for native <code>long</code> values.
    */
   private static class LongSerializer
   extends NativeTypeSerializer
   {
      /**
       * Create serializer for native <code>long</code> values and set its 
       * {@link NativeTypeSerializer#mark} to <code>3</code>.
       */
      protected LongSerializer()
      {
         super(3);
      }

      /**
       * Worker to write the value to the specified destination, as a native <code>long</code>.
       * 
       * @param    value
       *           The value to be written.
       * @param    out
       *           The destination.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected void writeValueWorker(Object value, ObjectOutput out)
      throws IOException
      {
         out.writeLong((long) value);
      }

      /**
       * A worker to read the value from the specified source, as a native <code>long</code>.
       * 
       * @param    in
       *           The source.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected Long readValueWorker(ObjectInput in)
      throws IOException
      {
         return in.readLong();
      }
   }
   
   /**
    * A serializer for native <code>float</code> values.
    */
   private static class FloatSerializer
   extends NativeTypeSerializer
   {
      /**
       * Create serializer for native <code>float</code> values and set its 
       * {@link NativeTypeSerializer#mark} to <code>4</code>.
       */
      protected FloatSerializer()
      {
         super(4);
      }

      /**
       * Worker to write the value to the specified destination, as a native <code>float</code>.
       * 
       * @param    value
       *           The value to be written.
       * @param    out
       *           The destination.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected void writeValueWorker(Object value, ObjectOutput out)
      throws IOException
      {
         out.writeFloat((float) value);
      }

      /**
       * A worker to read the value from the specified source, as a native <code>float</code>.
       * 
       * @param    in
       *           The source.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected Float readValueWorker(ObjectInput in)
      throws IOException
      {
         return in.readFloat();
      }
   }
   
   /**
    * A serializer for native <code>double</code> values.
    */
   private static class DoubleSerializer
   extends NativeTypeSerializer
   {
      /**
       * Create serializer for native <code>double</code> values and set its 
       * {@link NativeTypeSerializer#mark} to <code>5</code>.
       */
      protected DoubleSerializer()
      {
         super(5);
      }

      /**
       * Worker to write the value to the specified destination, as a native <code>double</code>.
       * 
       * @param    value
       *           The value to be written.
       * @param    out
       *           The destination.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected void writeValueWorker(Object value, ObjectOutput out)
      throws IOException
      {
         out.writeDouble((double) value);
      }

      /**
       * A worker to read the value from the specified source, as a native <code>double</code>.
       * 
       * @param    in
       *           The source.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected Double readValueWorker(ObjectInput in)
      throws IOException
      {
         return in.readDouble();
      }
   }
   
   /**
    * A serializer for native <code>boolean</code> values.
    */
   private static class BooleanSerializer
   extends NativeTypeSerializer
   {
      /**
       * Create serializer for native <code>boolean</code> values and set its 
       * {@link NativeTypeSerializer#mark} to <code>6</code>.
       */
      protected BooleanSerializer()
      {
         super(6);
      }

      /**
       * Worker to write the value to the specified destination, as a native <code>boolean</code>.
       * 
       * @param    value
       *           The value to be written.
       * @param    out
       *           The destination.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected void writeValueWorker(Object value, ObjectOutput out)
      throws IOException
      {
         out.writeBoolean((boolean) value);
      }

      /**
       * A worker to read the value from the specified source, as a native <code>boolean</code>.
       * 
       * @param    in
       *           The source.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected Boolean readValueWorker(ObjectInput in)
      throws IOException
      {
         return in.readBoolean();
      }
   }
   
   /**
    * A serializer for native <code>char</code> values.
    */
   private static class CharSerializer
   extends NativeTypeSerializer
   {
      /**
       * Create serializer for native <code>char</code> values and set its 
       * {@link NativeTypeSerializer#mark} to <code>7</code>.
       */
      protected CharSerializer()
      {
         super(7);
      }

      /**
       * Worker to write the value to the specified destination, as a native <code>char</code>.
       * 
       * @param    value
       *           The value to be written.
       * @param    out
       *           The destination.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected void writeValueWorker(Object value, ObjectOutput out)
      throws IOException
      {
         out.writeChar((char) value);
      }

      /**
       * A worker to read the value from the specified source, as a native <code>char</code>.
       * 
       * @param    in
       *           The source.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected Character readValueWorker(ObjectInput in)
      throws IOException
      {
         return in.readChar();
      }
   }
   
   /**
    * A serializer for {@link String} values.
    */
   private static class StringSerializer
   extends NativeTypeSerializer
   {
      /**
       * Create serializer for {@link String} values and set its {@link NativeTypeSerializer#mark}
       * to <code>8</code>.
       */
      protected StringSerializer()
      {
         super(8);
      }

      /**
       * Worker to write the value to the specified destination, as a {@link String} value.
       * 
       * @param    value
       *           The value to be written.
       * @param    out
       *           The destination.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected void writeValueWorker(Object value, ObjectOutput out)
      throws IOException
      {
         out.writeUTF((String) value);
      }

      /**
       * A worker to read the value from the specified source, as a {@link String} value.
       * 
       * @param    in
       *           The source.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected String readValueWorker(ObjectInput in)
      throws IOException
      {
         return in.readUTF();
      }
   }
   
   /**
    * A fallback serializer for {@link Object} instances.
    */
   private static class ObjectSerializer
   extends NativeTypeSerializer
   {
      /**
       * Create serializer for {@link Object} instances and set its 
       * {@link NativeTypeSerializer#mark} to <code>9</code>.
       */
      protected ObjectSerializer()
      {
         super(9);
      }

      /**
       * Worker to write the value to the specified destination, as a {@link Object} instance.
       * 
       * @param    value
       *           The value to be written.
       * @param    out
       *           The destination.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       */
      @Override
      protected void writeValueWorker(Object value, ObjectOutput out)
      throws IOException
      {
         out.writeObject(value);
      }

      /**
       * A worker to read the value from the specified source, as a {@link Object} instance.
       * 
       * @param    in
       *           The source.
       *           
       * @throws   IOException
       *           In case of I/O errors.
       * @throws   ClassNotFoundException
       *           If the object's class can not be found.
       */
      @Override
      protected Object readValueWorker(ObjectInput in)
      throws IOException,
             ClassNotFoundException
      {
         return in.readObject();
      }
   }
}