/*
** Module   : PasswordHelper.java
** Abstract : Encrypt and authenticate passwords.
**
** Copyright (c) 2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20170706 Created first version.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.util;

import java.security.*;
import java.security.spec.*;
import java.util.*;
import javax.crypto.*;
import javax.crypto.spec.*;

/**
 * Helper class to generate salt, encrypt a password with the PBKDF2WithHmacSHA1 algorithm, and
 * to authenticate passwords by encrypting them with the same algorithm and comparing results.
 * <p>
 * Based on an example published by Jerry Orr at
 * http://blog.jerryorr.com/2012/05/secure-password-storage-lots-of-donts.html
 */
public final class PasswordHelper
{
   /**
    * Encrypt a clear text password candidate and compare it with the given encrypted password
    * to determine if they match.
    * 
    * @param   clearPassword
    *          Clear text password candidate.
    * @param   encryptedPassword
    *          Previously encrypted password.
    * @param   salt
    *          Salt to use with encryption algorithm.
    *          
    * @return  {@code True} if the candidate represents the correct password, else {@code false}.
    * 
    * @throws  NoSuchAlgorithmException
    *          if the requested encryption algorithm is not available in the current environment.
    * @throws  InvalidKeySpecException
    *          if the specifications for the cryptographic key are invalid.
    */
   public boolean authenticate(String clearPassword, byte[] encryptedPassword, byte[] salt)
   throws NoSuchAlgorithmException,
          InvalidKeySpecException
   {
      return Arrays.equals(encryptedPassword, encryptPassword(clearPassword, salt));
   }
   
   /**
    * Encrypt a password using the &quot;PBKDF2WithHmacSHA1&quot; algorithm.
    * 
    * @param   password
    *          Clear text password to encrypt.
    * @param   salt
    *          Salt to use with encryption algorithm.
    * 
    * @return  Encrypted password as a byte array.
    * 
    * @throws  NoSuchAlgorithmException
    *          if the requested encryption algorithm is not available in the current environment.
    * @throws  InvalidKeySpecException
    *          if the specifications for the cryptographic key are invalid.
    */
   public byte[] encryptPassword(String password, byte[] salt)
   throws NoSuchAlgorithmException,
          InvalidKeySpecException
   {
      String algorithm = "PBKDF2WithHmacSHA1";
      int keyLength = 160;
      int iterations = 40000;
      
      KeySpec keySpec = new PBEKeySpec(password.toCharArray(), salt, iterations, keyLength);
      
      return SecretKeyFactory.getInstance(algorithm).generateSecret(keySpec).getEncoded();
   }
   
   /**
    * Generate a random, 8-byte salt value for use when encrypting a password.
    * 
    * @return  Salt value as a byte array.
    * 
    * @throws  NoSuchAlgorithmException
    *          if the requested encryption algorithm is not available in the current environment.
    */
   public byte[] salt()
   throws NoSuchAlgorithmException
   {
      byte[] salt = new byte[8];
      SecureRandom.getInstance("SHA1PRNG").nextBytes(salt);
      
      return salt;
   }
}
