/*
** Module   : StringHelper.java
** Abstract : provides a set of useful helper methods
**
** Copyright (c) 2004-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 GES 20041109   @18684 Built as a subset of some utility functions
**                           written for the GCD Storefront application.
**                           This provides string formatting and
**                           manipulation functions.  These were
**                           written and tested by ECF but integrated
**                           (and in somecases documented) by GES.
** 002 GES 20041118   @18691 Added safeTrimLeading and safeTrimTailing
**                           methods.  These delete leading or trailing
**                           space characters from a String.
** 003 NVS 20050221   @19893 Added isRepeatedChar method, which verifies
**                           whether the specified string buffer is made 
**                           of copies of a single character.
** 004 ECF 20050304   @20172 Added sprintf method. Works the same way as
**                           the C runtime library function of the same
**                           name, except that the variable argument list
**                           is handled as an array of objects.
** 005 GES 20050321   @20450 Added a processEscapes() method to convert
**                           a string with escape sequences into a string
**                           with the proper resulting characters.
** 006 GES 20050421   @20625 Fixed sprintf() to handle embedded % chars
**                           in replacement strings.
** 007 ECF 20050412   @20702 Added stripEnclosing method.
** 008 GES 20050426   @20870 Changed processEscapes() method to fix a
**                           bug in unicode char parsing and to add octal
**                           escape sequence support.
** 009 NVS 20050510   @21448 Added countChar method which returns number
**                           of occurences of a char in a string.
** 010 GES 20050601   @21364 Updated sprintf() to handle all primitive
**                           types AND to convert all other Objects using
**                           toString() allowing any object to be passed
**                           as an argument as long as one uses %s as the
**                           specifier for such generic/unrecognized
**                           Objects.  The primitive types all match up
**                           with the specifiers as expected.
** 011 ECF 20050603   @21404 Modified 010 (@21364) to use String.valueOf()
**                           instead of toString(). This safely handles
**                           null references.
** 012 GES 20050617   @21508 Added method sweep() to replace tokens in
**                           a target string based on a map of key and
**                           value replacements.
** 013 ECF 20050628   @21602 More robust replacement algorithm inside
**                           sweep method. Escapes backslash and dollar
**                           sign in replacement value. Also added
**                           exception handler to report more useful
**                           diagnostics.
** 014 ECF 20051013   @23029 Added changeCase() method. Capitalizes or
**                           decapitalizes the first letter of a name.
** 015 GES 20051028   @23140 Added protection logic to changeCase().
** 016 ECF 20060113   @23914 Moved generateBeanName() from NameConverter.
**                           Method is needed at runtime; NameConverter is
**                           a conversion-time class only.
** 017 ECF 20060712   @27996 Changed safeTrimTrailing() to behave more
**                           like Progress. In addition to spaces, it now
**                           removes tabs, line feeds, and carriage
**                           returns as well.
** 018 GES 20060720   @28171 Added methods to render primative arrays to
**                           a text form.
** 019 GES 20060821   @28694 Added an alternate word wrap helper.
** 020 GES 20060925   @29868 Added removeChar().
** 021 ECF 20080305   @37464 Made slight optimizations to sweep() method.
** 022 CA  20090408   @41734 Fixed edge case for nolossWrapText(...) - 
**                           remove any white space after each line is 
**                           extracted.
** 023 CA  20090413   @41771 Javadoc change in nolossWrapText.
** 024 GES 20090422   @41919 Removed sprintf() method as all callers now use the
**                           standard formatting support.
** 025 GES 20090424   @41977 Package change.
** 026 EVL 20121102          Adding helper functions to double backslashes in
**                           string under Windows environment. Adding helper
**                           functions to substitute the filename in Windows
**                           environment if it contains the Linux file
**                           separator. The char '/' replaced with '\' if
**                           the current OS is Windows.
** 027 EVL 20130108          Shifting the Windows detection code to separate
**                           method in com.goldencode.util.PlatformHelper
**                           class. Call this static to detect the OS here.
** 028 VMN 20130411          Helper function parseTokens(...) is added
** 029 GES 20140228          Changed word wrapping to honor embedded newlines.
** 030 ECF 20140301          Minor cleanup of parseTokens method.
** 031 GES 20150124          Fixed octal escape sequence processing which was profoundly broken.
** 032 ECF 20150715          Replace StringBuffer with StringBuilder.
** 033 IAS 20150926          Added equals() method.
** 034 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 035 ECF 20160409          Rewrote sweep, which was horribly slow.
** 036 GES 20160309          Added a variant of stripEnclosing().
** 037 OM  20160603          Fixed rtrimming to space character only.
** 038 GES 20170302          Added stripMatchedComments().
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.util;

import java.util.*;

/**
 * Provides a set of methods to help in processing and formatting
 * <code>String</code> objects.  This is intended to augment the limited
 * set of formatting and processing methods in the <code>String</code> 
 * class.  Methods are provided to left pad, right pad, center, trim and
 * truncate among other features.
 * <p>
 * Most (if not all) of these methods are stateless and thus are implemented
 * as static methods. 
 */
public class StringHelper
{
   /** Starting text for a slash-star style comment. */
   private static final String SLASH_STAR = "/*";
   
   /** Ending text for a slash-star style comment. */
   private static final String STAR_SLASH = "*/";
   
   /** Platform-specific line separator. */
   private static final String LINE_SEP = System.getProperty( "line.separator", "\n" );
      
   /**
    * Substitute all instances of <code>${key_name}</code> (where
    * <code>key_name</code> identifies a valid replacement parameter)
    * in the specified text with the associated parameter value. Replacements 
    * are specified as key and value pairs in the parameter map. Variables
    * are substituted in the order in which the key set is returned, which
    * may be undefined for some <code>Map</code> implementations and will
    * be fixed and controllable by the caller for other <code>Map</code>
    * implementations.
    * <p>
    * <strong>Note:</strong> if a particular replacement key name cannot be
    * found in the map of parameters, it is left untouched in the output
    * string.
    *
    * @param    text
    *           String upon which variable substitutions are to be performed.
    * @param    parmMap
    *           A list of key and value pairs specifying the replacements
    *           to be made.
    *
    * @return   <code>text</code>, with all variables substituted for their
    *           parameter values.
    */
   public static String sweep(String text, Map<String, String> parmMap)
   {
      int len;
      if (text == null || (len = text.length()) == 0 || parmMap.isEmpty())
      {
         // garbage in, garbage out
         return text;
      }
      
      // buffer for output text
      StringBuilder buf = new StringBuilder();
      
      char last = 0;
      
      // loop through input text searching for and replacing tokens with replacement parameters
      outer:
      for (int i = 0; i < len; i++)
      {
         char c = text.charAt(i);
         switch (c)
         {
            case '$':
               
               if (last == '$')
               {
                  buf.append('$');
               }
               
               break;
               
            case '{':
               
               if (last != '$')
               {
                  buf.append(c);
                  break;
               }
               
               // scan for token containing key
               StringBuilder keyBuf = new StringBuilder();
               String key = null;
               for (int j = i + 1; j < len; j++)
               {
                  char x = text.charAt(j);
                  if (x == '}')
                  {
                     key = keyBuf.toString();
                     String replacement = parmMap.get(key);
                     
                     if (replacement == null)
                     {
                        // leave token containing key untouched if not found in parm map
                        buf.append("${");
                        buf.append(key);
                        buf.append(x);
                     }
                     else
                     {
                        // replace token with value from parameter map
                        buf.append(replacement);
                     }
                     
                     // skip past token to sweep remaining input text
                     i = j;
                     
                     // last is set to '$' at this point, so reset it
                     last = 0;
                     
                     continue outer;
                  }
                  else
                  {
                     keyBuf.append(x);
                  }
               }
               
               if (key == null)
               {
                  buf.append("${");
               }
               
               break;
            
            default:
               
               if (last == '$')
               {
                  buf.append('$');
               }
               
               buf.append(c);
               
               break;
         }
         
         last = c;
      }
      
      return buf.toString();
   }
   
   /**
    * Create a string representation of <code>o</code> and indent each line
    * of that representation with <code>padSize</code> whitespace characters.
    * The platform-specific line separator as returned by
    * <code>System.getProperty</code> is used for parsing the input and in
    * generating the output.
    *
    * @param      o
    *             Object to be rendered as a string.
    * @param      padSize
    *             Number of whitespace characters to indent each line of the
    *             output.
    *
    * @return     <code>o</code> represented as a string, with each line
    *             indented.
    */
   public static String indent(Object o, int padSize)
   {
      return indent(o, padSize, new StringBuilder());
   }
   
   /**
    * Create a string representation of <code>o</code> and indent each line
    * of that representation with <code>padSize</code> whitespace characters.
    * The platform-specific line separator as returned by
    * <code>System.getProperty</code> is used for parsing the input and in
    * generating the output.
    *
    * @param      o
    *             Object to be rendered as a string.
    * @param      padSize
    *             Number of whitespace characters to indent each line of the
    *             output.
    * @param      buf
    *             Target buffer.
    *
    * @return     <code>o</code> represented as a string, with each line
    *             indented.
    */
   public static String indent(Object o, int padSize, StringBuilder buf)
   {
      int start = buf.length();
      
      String pad = repeatChar(' ', padSize);
      String text = String.valueOf(o);
      String[] lines = text.split(LINE_SEP, -1);
      int last = lines.length - 1;
      for (int i = 0; i <= last; i++)
      {
         buf.append(pad).append(lines[i]);
         if (i < last)
         {
            buf.append(LINE_SEP);
         }
      }
      int end = buf.length();
      
      return buf.substring(start, end);
   }
   
   /**
    * Create a new <code>String</code> with the content of the passed
    * argument, centered in a space padded <code>String</code> of the
    * user-specified width.
    *
    * @param   text
    *          Text to be centered in the output <code>String</code>. Note
    *          that this text must not contain the platform-specific line
    *          separator character(s).
    * @param   width
    *          Total number of characters in the resulting
    *          <code>String</code>.  This must be greater than the current
    *          size of the input text.
    * @return  The centered text.
    * @throws  IllegalArgumentException
    */
   public static String centerText(String text, int width)
   throws IllegalArgumentException
   {
      return centerText(text, width, new StringBuilder());
   }
   
   /**
    * Create a new <code>String</code> in a user-created
    * <code>StringBuilder</code> with the content of the passed argument,
    * centered in a space padded <code>String</code> of the user-specified
    * width.
    *
    * @param   text
    *          Text to be centered in the output <code>String</code>. Note
    *          that this text must not contain the platform-specific line
    *          separator character(s).
    * @param   width
    *          Total number of characters in the resulting
    *          <code>String</code>.  This must be greater than the current
    *          size of the input text.
    * @param   buf
    *          Buffer in which to create the resulting text.
    * @return  The centered text.
    * @throws  IllegalArgumentException
    */
   public static String centerText(String text, int width, StringBuilder buf)
   throws IllegalArgumentException
   {
      int len = text.length();
      if (len > width)
      {
         throw new IllegalArgumentException("Text too wide to format");
      }
      if (text.indexOf(LINE_SEP) >= 0)
      {
         throw new IllegalArgumentException("Text contains line separator");
      }
      
      int leading = (width - len) / 2;
      int trailing = width - len - leading;
      
      repeatChar(' ', leading, buf);
      buf.append(text);
      repeatChar(' ', trailing, buf);
      
      int end = buf.length();
      
      return buf.substring(end - width, end);
   }
   
   /**
    * Create a new <code>String</code> with the content of the passed
    * argument, right-aligned and padded on the left with spaces up to the
    * user-specified width.
    *
    * @param   text
    *          Text to be right aligned in the output <code>String</code>.
    *          This text must not contain the platform-specific line
    *          separator character(s).
    * @param   width
    *          Total number of characters in the resulting
    *          <code>String</code>.  This must be greater than the current
    *          size of the input text.
    * @return  The right-aligned text.
    * @throws  IllegalArgumentException
    */
   public static String rightAlignText(String text, int width)
   throws IllegalArgumentException
   {
      return rightAlignText(text, width, new StringBuilder());
   }
   
   /**
    * Create a new <code>String</code> in a user-created
    * <code>StringBuilder</code> with the content of the passed argument,
    * right-aligned and padded on the left with spaces up to the
    * user-specified width.
    *
    * @param   text
    *          Text to be right aligned in the output <code>String</code>.
    *          This text must not contain the platform-specific line
    *          separator character(s).
    * @param   width
    *          Total number of characters in the resulting
    *          <code>String</code>.  This must be greater than the current
    *          size of the input text.
    * @param   buf
    *          Buffer in which to create the resulting text.
    * @return  The right-aligned text.
    * @throws  IllegalArgumentException
    */
   public static String rightAlignText(String text, int width, StringBuilder buf)
   throws IllegalArgumentException
   {
      int len = text.length();
      if (len > width)
      {
         throw new IllegalArgumentException("Text too wide to format");
      }
      if (text.indexOf(LINE_SEP) >= 0)
      {
         throw new IllegalArgumentException("Text contains line separator");
      }
      
      int leading = width - len;
      
      repeatChar(' ', leading, buf);
      buf.append(text);
      
      int end = buf.length();
      
      return buf.substring(end - width, end);
   }
   
   /**
    * Create a new <code>String</code> with the content of the passed
    * argument, left-aligned and padded on the right with spaces up to the
    * user-specified width.
    *
    * @param   text
    *          Text to be left aligned in the output <code>String</code>.
    *          This text must not contain the platform-specific line
    *          separator character(s).
    * @param   width
    *          Total number of characters in the resulting
    *          <code>String</code>.  This must be greater than the current
    *          size of the input text.
    * @return  The left-aligned text.
    * @throws  IllegalArgumentException
    */
   public static String leftAlignText(String text, int width)
   throws IllegalArgumentException
   {
      return leftAlignText(text, width, new StringBuilder());
   }
   
   /**
    * Create a new <code>String</code> in a user-created
    * <code>StringBuilder</code> with the content of the passed argument,
    * left-aligned and padded on the right with spaces up to the
    * user-specified width.
    *
    * @param   text
    *          Text to be left aligned in the output <code>String</code>.
    *          This text must not contain the platform-specific line
    *          separator character(s).
    * @param   width
    *          Total number of characters in the resulting
    *          <code>String</code>.  This must be greater than the current
    *          size of the input text.
    * @param   buf
    *          Buffer in which to create the resulting text.
    * @return  The left-aligned text.
    * @throws  IllegalArgumentException
    */
   public static String leftAlignText(String text, int width, StringBuilder buf)
   throws IllegalArgumentException
   {
      int len = text.length();
      if (len > width)
      {
         throw new IllegalArgumentException("Text too wide to format");
      }
      if (text.indexOf(LINE_SEP) >= 0)
      {
         throw new IllegalArgumentException("Text contains line separator");
      }
      
      int trailing = width - len;
      
      buf.append(text);
      repeatChar(' ', trailing, buf);
      
      int end = buf.length();
      
      return buf.substring(end - width, end);
   }
   
   /**
    * A simple algorithm to wrap a text string to fit into one or more lines
    * of the specified width. Expects the space character to delimit the
    * words in the input string. Returns an array of as many strings as are
    * necessary to meet the per-line width requirement. Single words that
    * are wider than <code>width</code> are <em>not</em> split, but will
    * appear in their own line. All lines are left-aligned within the
    * alloted width.
    * <p>
    * Warning: this algorithm will convert any sequential set of multiple
    * spaces between words into a single space between words.  If this is 
    * not wanted, then use {@link #nolossWrapText} instead.
    *
    * @param   text
    *          Text to be split.
    * @param   width
    *          Per-line width in characters.
    *
    * @return  An array of strings representing <code>text</code> split into
    *          lines of the specified width.
    */
   public static String[] wrapText(String text, int width)
   {
      List lines = new ArrayList();
      StringTokenizer tok = new StringTokenizer(text);
      StringBuilder buf = new StringBuilder();
      while (tok.hasMoreTokens())
      {
         String word = tok.nextToken();
         
         if (buf.length() + word.length() >= width)
         {
            repeatChar(' ', width - buf.length(), buf);
            lines.add(buf.toString());
            buf = new StringBuilder();
         }
         
         if (buf.length() > 0)
         {
            buf.append(' ');
         }
         
         buf.append(word);
      }
      
      if (buf.length() > 0)
      {
         repeatChar(' ', width - buf.length(), buf);
         lines.add(buf.toString());
      }
      
      return (String[]) lines.toArray(new String[lines.size()]);
   }
   
   /**
    * A simple algorithm to wrap a text string to fit into one or more lines
    * of the specified width. Word breaks are detected using whitespace
    * characters in the input string but the whitespace is not converted or
    * modified in any way. Returns an array of as many strings as are
    * necessary to meet the per-line width requirement. Single words that
    * are wider than <code>width</code> are <em>not</em> split, but will
    * appear in their own line.  If the text must be split into multiple lines
    * only the first line has its white-space preserved; all subsequent lines
    * lines will have their leading white-space removed.
    * <p>
    * This will honor embedded newlines as places to split a line.  Then any lines that are
    * still longer than the width will be wrapped as noted above.
    *
    * @param   text
    *          Text to be split.
    * @param   width
    *          Per-line width in characters.
    *
    * @return  An array of strings representing <code>text</code> split into
    *          lines of the specified width.
    */
   public static String[] nolossWrapText(String text, int width)
   {
      List lines = new ArrayList();
      
      String[] presplit = text.split("\n|\r\n|\r");
      
      for (String next : presplit)
      {
         if (next.length() == 0)
         {
            lines.add(next);
            continue;
         }
         
         while (next.length() > 0)
         {
            if (next.length() <= width)
            {
               lines.add(next);
               break;
            }
            else
            {
               int idx = -1;
               
               // look backwards from the wrap point for a word boundary
               for (int i = width; i >= 0; i--)
               {
                  if (Character.isSpaceChar(next.charAt(i)))
                  {
                     idx = i;
                     break;
                  }
               }
               
               // special case where the line is a single word that is larger 
               // than our width, search forward from the break point to find
               // the end of the word
               if (idx == -1)
               {
                  for (int j = width; j < next.length(); j++)
                  {
                     if (Character.isSpaceChar(next.charAt(j)))
                     {
                        idx = j;
                        break;
                     }
                  }
               }
               
               if (idx == -1)
               {
                  // there are no words on which to break
                  lines.add(next);
                  break;
               }
               else
               {
                  // save our line
                  lines.add(next.substring(0, idx));
                  
                  // reduce our input string and eliminate any white-space
                  next = next.substring(idx).trim();
               }
            }
         }
      }
      
      return (String[]) lines.toArray(new String[lines.size()]);
   }
   
   /**
    * Truncate input from <code>text</code> to <code>width</code> characters
    * leaving behind an ellipses ("...") in cases where a truncation has
    * occurred.  If the size of <code>text</code> is not greater than the
    * requested size, then no truncation will occur.  Note that when the
    * ellipses is added, the total size (in characters) of the text will
    * not exceed the requested size.
    *
    * @param      text
    *             Text string to be trimmed.
    * @param      width
    *             Total number of characters in the resulting output.
    * @return     Truncated <code>String</code> object with the contents of
    *             the input <code>text</code> and an ellipses when truncation
    *             has occurred.
    */
   public static String truncate(String text, int width)
   {
      int len = text.length();
      if (len <= width)
      {
         return text;
      }
      
      StringBuilder buf = new StringBuilder();
      buf.append(text.substring(0, width - 3))
         .append("...");
      
      return buf.toString();
   }
   
   /**
    * Create a string composed of <code>c</code> repeated <code>len</code>
    * times.
    *
    * @param      c
    *             Character from which to compose string.
    * @param      len
    *             Length of string to be created.
    *
    * @return     Resulting string.
    */
   public static String repeatChar(char c, int len)
   {
      return repeatChar(c, len, new StringBuilder());
   }
   
   /**
    * Append <code>c</code> to the specified string buffer <code>len</code>
    * times.
    *
    * @param      c
    *             Character to append to <code>buf</code>.
    * @param      len
    *             Number of times to append <code>c</code>.
    * @param      buf
    *             Target buffer.
    *
    * @return     Substring of <code>buf</code> which represents only the
    *             new content appended (i.e., <code>c</code> repeated
    *             <code>len</code> times.
    */
   public static String repeatChar(char c, int len, StringBuilder buf)
   {
      if (len <= 0)
      {
         return "";
      }
      
      for (int i = 0; i < len; i++)
      {
         buf.append(c);
      }
      int end = buf.length();
      
      return buf.substring(end - len, end);
   }
   
   /**
    * Verifies whether the specified string buffer is made of copies of 
    * a single character.
    *
    * @param      buf
    *             String buffer to verify
    * @param      c
    *             Character to verify the buffer with
    *
    * @return     <code>yes</code> if the string buffer contains only
    *             specified character and is non-empty.
    */
   public static boolean isRepeatedChar(StringBuilder buf, char c)
   {
      if (buf.length() == 0)
         return false;
      
      for (int i = 0; i < buf.length(); i++)
      {
         if (buf.charAt(i) != c)
            return false;
      }
      
      return true;
   }
   
   /**
    * Trim trailing spaces from {@code text}, without triggering a {@link NullPointerException} if
    * {@code text} is {@code null}. The tab, line feed, and carriage return are NOT trimmed.
    *
    * @param      text
    *             Text string to be trimmed.
    *
    * @return     {@code text} with trailing spaces removed, or {@code null} if {@code text} is 
    *             {@code null}.
    */
   public static String safeTrimTrailingSpaces(String text)
   {
      if (text == null)
      {
         return text;
      }
      
      int end = text.length();
      
      for (int i = end - 1; i >= 0; i--)
      {
         char c = text.charAt(i);
         if (c == ' ')
         {
            end--;
         }
         else
         {
            // first non-space exits the loop
            break;
         }
      }
      
      return text.substring(0, end);
   }
   
   /**
    * Trim trailing whitespace from {@code text}, without triggering a
    * {@code NullPointerException} if {@code text} is {@code null}. Whitespace includes space,
    * tab, line feed, and carriage return.
    *
    * @param   text
    *          Text string to be trimmed.
    *
    * @return  {@code text} with trailing whitespace removed, or {@code null} if {@code text}
    *          is {@code null}.
    */
   public static String safeTrimTrailing(String text)
   {
      if (text == null)
      {
         return text;
      }
      
      int end = text.length();
      
      for (int i = end - 1; i >= 0; i--)
      {
         char c = text.charAt(i);
         if (c == ' ' || c == '\t' || c == '\n' || c == '\r')
         {
            end--;
         }
         else
         {
            // first non-whitespace exits the loop
            break;
         }
      }
      
      return text.substring(0, end);
   }
   
   /**
    * Trim leading whitespace from <code>text</code>, without triggering
    * a <code>NullPointerException</code> if <code>text</code> is
    * <code>null</code>.
    *
    * @param      text
    *             Text string to be trimmed.
    *
    * @return     <code>text</code> with leading whitespace removed, or 
    *             <code>null</code> if <code>text</code> is 
    *             <code>null</code>.
    */
   public static String safeTrimLeading(String text)
   {
      if (text == null)
         return text;
      
      int end = text.length();
      int i;
      
      for (i = 0; i < end; i++)
      {
         if (!Character.isSpaceChar(text.charAt(i)))
         {
            // first non-space exits the loop
            break;
         }
      }
      
      return text.substring(i, end);
   }
   
   /**
    * Trim leading and trailing whitespace from <code>text</code>, without
    * triggering a <code>NullPointerException</code> if <code>text</code> is
    * <code>null</code>.
    *
    * @param      text
    *             Text string to be trimmed.
    *
    * @return     <code>text</code> with leading and trailing whitespace
    *             removed, or <code>null</code> if <code>text</code> is
    *             <code>null</code>.
    */
   public static String safeTrim(String text)
   {
      return (text == null ? text : text.trim());
   }
   
   /**
    * Converts a string with valid Java escape sequences into a string in
    * which those sequences are converted to the corresponding character.
    * This method supports the following escape sequences (all the escapes
    * that the javac compiler supports):
    * <p>
    * <ul>
    *    <li> '\b' - backspace 
    *    <li> '\t' - horizontal tab
    *    <li> '\n' - linefeed
    *    <li> '\f' - formfeed
    *    <li> '\r' - carriage return
    *    <li> '\&quot;' - double quote 
    *    <li> '\&#x27;' - single quote
    *    <li> '\\' - backslash
    *    <li> &#x27;&#092;uxxxx&#x27; where xxxx represents 4 hexidecimal
    *         digits - unicode
    *    <li> '\ooo' - an octal escape char where ooo represents up to 3
    *         octal digits with a maximum value of &#x27;&#092;377&#x27;
    * </ul>
    * <p>
    * Note that in javac, the unicode conversion occurs prior to the escape
    * sequence processing, which disallows certain unicode characters from
    * being used inside a Java string literal.  This method makes no such
    * distinction so it accepts a superset of the valid Java syntax.
    *
    * @param    input
    *           The string to be converted.
    *
    * @return   The string with escape sequences converted to the
    *           corresponding character.
    */
   public static String processEscapes(String input)
   throws IllegalArgumentException,
          NumberFormatException,
          IndexOutOfBoundsException
   {
      StringBuilder sb = new StringBuilder();
      int begin = 0;
      int next  = input.indexOf('\\');
      
      // as long as there are backslashes, loop
      while (next != -1)
      {
         // copy the chars up to the backslash
         sb.append(input.substring(begin, next));
         
         // look at the char after the backslash
         char esc = input.charAt(next + 1);
         
         // initialize our default escape sequence size in chars
         int fwd = 2;
         
         // the char MUST be one of the following
         switch (esc)
         {
            case 'b':
               sb.append("\b");    // backspace
               break;
            case 't':
               sb.append("\t");    // horizontal tab
               break;
            case 'n':
               sb.append("\n");    // linefeed
               break;
            case 'f':
               sb.append("\f");    // formfeed
               break;
            case 'r':
               sb.append("\r");    // carriage return
               break;
            case '\"':
               sb.append("\"");    // double quote
               break;
            case '\'':
               sb.append("\'");    // single quote
               break;
            case '\\':
               sb.append("\\");    // backslash
               break;
            case 'u':              // unicode
               int unicode = Integer.parseInt(input.substring(next+1, next+4), 16);
               sb.append((char) unicode);
               fwd += 4;
               break;
            case '0':              // octal
            case '1':
            case '2':
            case '3':
            case '4':
            case '5':
            case '6':
            case '7':
               StringBuilder otxt = new StringBuilder();
               
               otxt.append(esc);
               
               // calculate the extent of the octal chars (the largest octal
               // char literal is \377 but it is also possible to specify 
               // single or double digits - \7 or \26 for example)
               int size = 1;
               
               int two = next + 2;
               
               if (two < input.length())
               {
                  char second = input.charAt(two);
                  
                  if (second >= '0' && second <= '7')
                  {
                     otxt.append(second);
                     fwd++;
                     
                     int three = next + 3;
                     
                     if (three < input.length())
                     {
                        char third = input.charAt(three);
                        
                        if (third >= '0' && third <= '7')
                        {
                           otxt.append(third);
                           fwd++;
                        }
                     }
                  }
               }
               
               int octal = Integer.parseInt(otxt.toString(), 8);
               
               sb.append((char) octal);
               break;
            default:
               throw new IllegalArgumentException("Unrecognized escape sequence \\" + esc + ".");
         }
         
         // "eat" the escaped portion of the string
         begin = next + fwd;
         
         // look for the our next escape sequence
         next = input.indexOf('\\', begin);
      }
      
      // copy the rest of the string into our output
      sb.append(input.substring(begin));
      
      return sb.toString();
   }
   
   /**
    * Strip a matched pair of slash-star comments, including any nested matched pairs of comments
    * and return the result.
    * <p>
    * This code assumes that there will always be a matched set of start and end comment
    * sequences.
    *
    * @param    text
    *           Text from which to remove the comments.
    *
    * @return   The adjusted string.
    *
    * 
    */
   public static String stripMatchedComments(String text)
   {
      if (text == null || safeTrim(text).length() == 0)
      {
         // nothing to do
         return text;
      }
      
      int start = text.indexOf(SLASH_STAR);
      
      if (start < 0)
      {
         // quick out, no starting comment sequence
         return text;
      }
      
      int end = text.indexOf(STAR_SLASH, start);
      
      if (end < 0)
      {
         // this should not happen, it is illegal
         throw new IllegalArgumentException("Missing end comment sequence */.");
      }
      
      StringBuilder sb = new StringBuilder();
      
      int len     = text.length();
      int idx     = 0;
      
      int nesting = 0;
      
      // at this point there is supposed to be at least 1 comment to remove
      while (idx < len && (start >= 0 || nesting > 0))
      {
         if (nesting == 0)
         {
            // save off the next portion
            sb.append(text.substring(idx, start));
         }
         
         if (end < 0)
         {
            // this should not happen, it is illegal
            throw new IllegalArgumentException("Missing end comment sequence */.");
         }
         else if (start < 0 || start > end)
         {
            // close off the currently open comment by moving our index ahead
            idx = end + STAR_SLASH.length();
            
            nesting--;
         }
         else
         {
            // move our index forward
            idx = start + SLASH_STAR.length();
            
            // now we are inside a comment
            nesting++;
         }
         
         start = text.indexOf(SLASH_STAR, idx);
         end   = text.indexOf(STAR_SLASH, idx);
      }
      
      if (nesting > 0)
      {
         // this should not happen, it is illegal
         throw new IllegalArgumentException("Missing end comment sequence */.");
      }
      
      // save off the last portion
      sb.append(text.substring(idx));
      
      return sb.toString();
   }

   /**
    * Strip a matched pair of the specified enclosing character, if it exists, from the
    * beginning and end of a string. A character is only removed if it is is both the first and
    * last character of <code>text</code>.
    *
    * @param   text
    *          Text from which to remove character.
    * @param   c
    *          Character to be removed.
    *
    * @return  The adjusted string.
    */
   public static String stripEnclosing(String text, char c)
   {
      int begin = text.indexOf(c);
      int end = text.lastIndexOf(c);
      if (begin == 0 && end == text.length() - 1 && begin < end)
      {
         text = text.substring(begin + 1, end);
      }
      
      return text;
   }

   /**
    * Strip a matched pair of any of the specified enclosing characters, if they exist, from the
    * beginning and end of a string. A character is only removed if it is is both the first and
    * last character of <code>text</code>.
    *
    * @param   text
    *          Text from which to remove character.
    * @param   ch
    *          List of characters to be removed.
    *
    * @return  The adjusted string.
    */
   public static String stripEnclosing(String text, char[] ch)
   {
      for (char c : ch)
      {
         text = stripEnclosing(text, c);
      }
      
      return text;
   }

   /**
    * Counts the number of occurrences of the specified character in a string.
    *
    * @param   text
    *          string to inspect.
    * @param   c
    *          Character to be counted.
    *
    * @return  number of occurrences
    */
   public static int countChar(String text, char c)
   {
      int count = 0;
      int occ = text.indexOf(c);
      
      while (occ >= 0)
      {
         count ++;
         occ = text.indexOf(c, occ + 1);
      }
      
      return count;
   }
   
   /**
    * Removes all occurrences of the specified character in a string.
    *
    * @param    text
    *           The string to modify.
    * @param    ch
    *           Character to be removed.
    *
    * @return   The modified string.
    */
   public static String removeChar(String text, char ch)
   {
      StringBuilder sb = new StringBuilder();
      
      boolean found = false;
      
      int last = 0;
      int idx  = text.indexOf(ch);
      
      while (idx >= 0)
      {
         found = true;
         sb.append(text.substring(last, idx));
         last = idx + 1; 
         idx  = text.indexOf(ch, last);
      }
      
      // copy the last portion of the string if needed
      if (found && last < text.length())
      {
         sb.append(text.substring(last));
      }
      
      return found ? sb.toString() : text;
   }
   
   /**
    * Capitalize or decapitalize the given name by changing the case of its
    * first character. This unconditionally changes case of the first
    * character, which departs slightly from the Java-beans standard.
    *
    * @param   name
    *          Name to be capitalized/decapitalized.
    * @param   upper
    *          <code>true</code> to uppercase the first character of
    *          <code>name</code>; <code>false</code> to lowercase it.
    *
    * @return  A new string which represents <code>name</code> with the
    *          requested modification.  If <code>name</code> is empty
    *          or <code>null</code>, the empty string is returned.
    */
   public static String changeCase(String name, boolean upper)
   {
      if (name == null || name.length() == 0)
      {
         // nothing to do
         return "";
      }
      
      // the string has at least 1 char
      StringBuilder buf = new StringBuilder();
      char first = name.charAt(0);
      if (upper)
      {
         buf.append(Character.toUpperCase(first));
      }
      else
      {
         buf.append(Character.toLowerCase(first));
      }
      
      // handle after the 1st char
      if (name.length() > 1)
      {
         // append the rest
         buf.append(name.substring(1));
      }
      
      return buf.toString();
   }
   
   /**
    * Create a (mostly) Java-bean compatible getter or setter name for a
    * given field.  If the field is a <code>boolean</code> and a getter is
    * requested, then the prefix will be "is", otherwise getters are prefixed 
    * with "get".  All setters are prefixed with "set".  The name of the
    * field always has its first character uppercased which is slightly
    * different than how Sun defines the standard.
    *
    * @param    field
    *           The name to use as the basis, must be non-null and non-empty.
    * @param    getter
    *           If <code>true</code> the resulting bean name will be a
    *           getter.
    * @param    bool
    *           If <code>getter</code> is <code>true</code> and this is
    *           <code>true</code> then the resulting bean name will prefixed
    *           with "is" instead of "get".  Ignored for setters.
    *
    * @return   The resulting getter or setter name.
    *
    * @throws   IllegalArgumentException
    *           If the field name is <code>null</code> or is the empty
    *           string.
    */
   public static String generateBeanName(String field, boolean getter, boolean bool)
   throws IllegalArgumentException
   {
      if (field == null || field.length() == 0)
      {
         String errmsg = "Empty string or null input for field name!";
         throw new IllegalArgumentException(errmsg);
      }
      
      StringBuilder sb = new StringBuilder();
      sb.append(getter ? (bool ? "is" : "get") : "set");
      sb.append(changeCase(field, true));
      
      return sb.toString();
   }
   
   /**
    * Convert the given array into a string of comma separated elements which
    * starts with a '[' and ends with a ']'.  This is meant to replace the 
    * poor default approach provided by J2SE <code>Object.toString</code>
    * (which does not emit any of the contents of an array). 
    * <p>
    * Note that for <code>byte[]</code> and <code>char[]</code> the rendering
    * is direct to a <code>String</code> rather than using the comma separated
    * format (for bytes a hexidecimal format is used for each byte and only
    * the first 1024 bytes are processed).
    *
    * @param    array
    *           The array to render as text.
    *
    * @return   The rendered text or <code>null</code> if the parameter is
    *           <code>null</code>.
    */
   public static String render(Object array)
   {
      if (array == null)
      {
         return null; 
      }
      
      StringBuilder sb = new StringBuilder();
      render(array, sb);
      
      return sb.toString();
   }
   
   /**
    * Convert the given array into a string of comma separated elements which
    * starts with a '[' and ends with a ']'.  This is meant to replace the 
    * poor default approach provided by J2SE <code>Object.toString</code>
    * (which does not emit any of the contents of an array). 
    * <p>
    * Note that for <code>byte[]</code> and <code>char[]</code> the rendering
    * is direct to a <code>String</code> rather than using the comma separated
    * format (for bytes a hexidecimal format is used for each byte and only
    * the first 1024 bytes are processed).
    *
    * @param    array
    *           The array to render as text.
    * @param    buf
    *           The output buffer.  Must not be <code>null</code>.
    */
   public static void render(Object array, StringBuilder buf)
   {
      if (array instanceof byte[])
      {
         render((byte[]) array, buf);
      }
      else if (array instanceof char[])
      {
         buf.append(new String((char[]) array));
      }
      else if (array instanceof short[])
      {
         render((short[]) array, buf);
      }
      else if (array instanceof int[])
      {
         render((int[]) array, buf);
      }
      else if (array instanceof long[])
      {
         render((long[]) array, buf);
      }
      else if (array instanceof float[])
      {
         render((float[]) array, buf);
      }
      else if (array instanceof double[])
      {
         render((double[]) array, buf);
      }
      else if (array instanceof boolean[])
      {
         render((boolean[]) array, buf);
      }   
   }
   
   /**
    * Convert the given array into a hexidecimal format for each byte, start
    * with a '[' and end with a ']'.  Only the first 1024 bytes are
    * processed.  This is meant to replace the poor default approach provided
    * by J2SE <code>Object.toString</code> (which does not emit any of the
    * contents of an array). 
    *
    * @param    array
    *           The array to render as text.
    * @param    buf
    *           The output buffer.  Must not be <code>null</code>.
    */
   public static void render(byte[] array, StringBuilder buf)
   {
      if (array == null)
      {
         // nothing to do
         return;
      }
      
      buf.append('[');
      
      int len = (int) Math.min(array.length, 1024);
      
      for (int j = 0; j < len; j++)
      {
         String hex = Integer.toHexString((int) array[j]);
         hex = hex.length() > 2 ? hex.substring(hex.length() - 2) : hex;
         buf.append(hex);
      }
      
      if (len != array.length)
      {
         buf.append("... truncated!");
      }
      
      buf.append(']');
   }
   
   /**
    * Convert the given array into a string of comma separated elements which
    * starts with a '[' and ends with a ']'.  This is meant to replace the 
    * poor default approach provided by J2SE <code>Object.toString</code>
    * (which does not emit any of the contents of an array). 
    *
    * @param    array
    *           The array to render as text.
    * @param    buf
    *           The output buffer.  Must not be <code>null</code>.
    */
   public static void render(short[] array, StringBuilder buf)
   {
      if (array == null)
      {
         // nothing to do
         return;
      }
      
      buf.append('[');
      
      for (int j = 0; j < array.length; j++)
      {
         if (j != 0)
         {
            buf.append(", ");
         }
         
         buf.append(Short.toString(array[j]));
      }
      
      buf.append(']');
   }
   
   /**
    * Convert the given array into a string of comma separated elements which
    * starts with a '[' and ends with a ']'.  This is meant to replace the 
    * poor default approach provided by J2SE <code>Object.toString</code>
    * (which does not emit any of the contents of an array). 
    *
    * @param    array
    *           The array to render as text.
    * @param    buf
    *           The output buffer.  Must not be <code>null</code>.
    */
   public static void render(int[] array, StringBuilder buf)
   {
      if (array == null)
      {
         // nothing to do
         return;
      }
      
      buf.append('[');
      
      for (int j = 0; j < array.length; j++)
      {
         if (j != 0)
         {
            buf.append(", ");
         }
         
         buf.append(Integer.toString(array[j]));
      }
      
      buf.append(']');
   }
   
   /**
    * Convert the given array into a string of comma separated elements which
    * starts with a '[' and ends with a ']'.  This is meant to replace the 
    * poor default approach provided by J2SE <code>Object.toString</code>
    * (which does not emit any of the contents of an array). 
    *
    * @param    array
    *           The array to render as text.
    * @param    buf
    *           The output buffer.  Must not be <code>null</code>.
    */
   public static void render(long[] array, StringBuilder buf)
   {
      if (array == null)
      {
         // nothing to do
         return;
      }
      
      buf.append('[');
      
      for (int j = 0; j < array.length; j++)
      {
         if (j != 0)
         {
            buf.append(", ");
         }
         
         buf.append(Long.toString(array[j]));
      }
      
      buf.append(']');
   }
   
   /**
    * Convert the given array into a string of comma separated elements which
    * starts with a '[' and ends with a ']'.  This is meant to replace the 
    * poor default approach provided by J2SE <code>Object.toString</code>
    * (which does not emit any of the contents of an array). 
    *
    * @param    array
    *           The array to render as text.
    * @param    buf
    *           The output buffer.  Must not be <code>null</code>.
    */
   public static void render(float[] array, StringBuilder buf)
   {
      if (array == null)
      {
         // nothing to do
         return;
      }
      
      buf.append('[');
      
      for (int j = 0; j < array.length; j++)
      {
         if (j != 0)
         {
            buf.append(", ");
         }
         
         buf.append(Float.toString(array[j]));
      }
      
      buf.append(']');
   }
   
   /**
    * Convert the given array into a string of comma separated elements which
    * starts with a '[' and ends with a ']'.  This is meant to replace the 
    * poor default approach provided by J2SE <code>Object.toString</code>
    * (which does not emit any of the contents of an array). 
    *
    * @param    array
    *           The array to render as text.
    * @param    buf
    *           The output buffer.  Must not be <code>null</code>.
    */
   public static void render(double[] array, StringBuilder buf)
   {
      if (array == null)
      {
         // nothing to do
         return;
      }
      
      buf.append('[');
      
      for (int j = 0; j < array.length; j++)
      {
         if (j != 0)
         {
            buf.append(", ");
         }
         
         buf.append(Double.toString(array[j]));
      }
      
      buf.append(']');
   }
   
   /**
    * Convert the given array into a string of comma separated elements which
    * starts with a '[' and ends with a ']'.  This is meant to replace the 
    * poor default approach provided by J2SE <code>Object.toString</code>
    * (which does not emit any of the contents of an array). 
    *
    * @param    array
    *           The array to render as text.
    * @param    buf
    *           The output buffer.  Must not be <code>null</code>.
    */
   public static void render(boolean[] array, StringBuilder buf)
   {
      if (array == null)
      {
         // nothing to do
         return;
      }
      
      buf.append('[');
      
      for (int j = 0; j < array.length; j++)
      {
         if (j != 0)
         {
            buf.append(", ");
         }
         
         buf.append(Boolean.toString(array[j]));
      }
      
      buf.append(']');
   }
   
   /**
    * Fixing the regular expression string for special Windows requirement.
    * Now it solves the backslash issue in Windows Java API call involving
    * regular expression string. Does nothing if OS other than Windows.
    *
    * @param   text
    *          Regular expression string to be adjusted acccording to
    *          OS special requirements.
    *
    * @return  Specially modified regular expression string. Now doubles
    *          the backslash character containing in the input string.
    */
   public static String fixupRegex(String text)
   {
      StringBuilder sb = new StringBuilder();
      
      if (PlatformHelper.isUnderWindowsFamily())
         for (int i = 0; i < text.length(); i++ )
         {
            sb.append(text.charAt(i));
            if (text.charAt(i) == '\\')
               sb.append(text.charAt(i));
         }
      else
         sb.append(text);
      
      return sb.toString();
   }

   /**
    * Fixing the file name for special Windows requirement. Now it
    * substitutes the possible '/' character with backslash '\' to
    * accept to Windows filename requirements. Does nothing if OS other
    * than Windows.
    *
    * @param   text
    *          The string representing the filename with path separator.
    *
    * @return  Specially modified file name string. The Linux hardcoded
    *          '/' character replaced with '\' if current OS is Windows.
    */
   public static String fixupFileName(String text)
   {
      StringBuilder sb = new StringBuilder();
      
      if (PlatformHelper.isUnderWindowsFamily())
         for (int i = 0; i < text.length(); i++ )
            if (text.charAt(i) == '/')
               sb.append('\\');
            else
               sb.append(text.charAt(i));
      else
         sb.append(text);
      
      return sb.toString();
   }

   /**
    * Get tokens from string.
    *
    * @param   str
    *          The string to be tokenized.
    * @param   delimiter
    *          Delimiter.
    * @param   addIfStartsWithDelimiter
    *          If <code>true</code> and <code>str</code> starts with delimiter
    *          then empty token will be added.
    * @param   addIfEndsWithDelimiter
    *          If <code>true</code> and <code>str</code> ends with delimiter
    *          then empty token will be added.
    * @param   trim
    *          If <code>true</code> tokens will be trimmed
    *
    * @return  <code>null</code> if str equals <code>null</code> otherwise list of tokens.
    */
   public static List<String> parseTokens(String str,
                                          String delimiter,
                                          boolean addIfStartsWithDelimiter,
                                          boolean addIfEndsWithDelimiter,
                                          boolean trim)
   {
      List<String> tokens = null;
      if (str != null)
      {
         tokens = new ArrayList<String>();
         if (addIfStartsWithDelimiter && str.startsWith(delimiter))
         {
            tokens.add("");
         }
         for (StringTokenizer stringTokenizer =
            new StringTokenizer(str, delimiter); stringTokenizer.hasMoreTokens();)
         {
            String token = stringTokenizer.nextToken();
            if (trim)
            {
               token = token.trim();
            }
            tokens.add(token);
         }
         if (addIfEndsWithDelimiter && str.endsWith(delimiter))
         {
            tokens.add("");
         }
      }
      return tokens;
   }
   
   /**
    * Compare two (possibly <code>null</code>) strings
    * 
    * @param    s1 
    *           the first string
    * @param    s2
    *           the second string
    *          
    * @return   <code>true</code> if both strings are <code>null</code> or equal
    */
   public static boolean equal(String s1, String s2)
   {
      return s1 == null ? s2 == null : s1.equals(s2);
   }
}