/*
** Module   : filesys_linux.c
** Abstract : implementation for detection of simbolic links, non-FIFO pipes and special devices.
**            This is used by filelys.c
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 EVL 20130329 Created initial version to implement Linux specifics for file system
**                  operations.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/
#include "filesys.h"

/**
* The helper function to get symlink status in Linux.
*
* @param  env 
*         pointer to a structure that provides access to JNI environment.
* @param  fname 
*         The file name to be checked passed from FileChecker.checkFileType().
* @param  file_path 
*         The pointer to file path string.
* @param  status_initial 
*         The current status value for the given file.
*
* @return An integer modified status code of the file.
*/
int getSymLinkStatus(JNIEnv *env, const char *file_path, int status_initial)
{
   // Start with currently available information
   int status_return = status_initial;

   struct stat buf;
   //get symbolic link status
   //we can't use fstat for detecting symbolic links
   
   int lnk_stat = lstat(file_path, &buf); 
   
   //if link status is -1 return -1
   if (lnk_stat==-1) return -1;
   
   // check for special devices
   if (S_ISCHR(buf.st_mode) || S_ISBLK(buf.st_mode))
   { 
      // status code for devices is standalone
      status_return |= DEVICE;
   }
   // check for symbolic links
   // NOTE: A symbolic link that points to a directory or file will not be
   // detected as file or directory by the logic present here so this status 
   //cannot be made inclusive with file or directory
   if (S_ISLNK(buf.st_mode))
   {
      //status code for symbolic link is additive
      status_return |= SYMLINK;
   }
   
   return status_return;
}

/**
* The helper function to open file or pipe in Linux.
*
* @param  filename 
*         The name of the file to open.
* @param  flags 
*         The open mode flags common across all operating systems.
*
* @return An integer file fescriptor or -1 in case of error.
*/
int open_native(const char * filename, int flags)
{
   return open(filename, flags | O_NONBLOCK);
}