/*
** Module   : registry.c
** Abstract : implementation for Windows registry access functions.
**
** Copyright (c) 2013-2017, Golden Code Development Corporation.
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 EVL 20130618 Created initial version to implement Windows specifics for registry related
**                  environment access operations.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

#include <windows.h>

#include "com_goldencode_p2j_util_Registry.h"

// Registry name length limits according to Win32 API documentation
#define MAX_KEY_LENGTH 255
#define MAX_VALUE_NAME 16383

/**
* Helper function that will return the full name of the registry key for given section, base and
* key names. 
* 
* @param  env 
*         pointer to a structure that provides access to JNI environment
* @param  base 
*         The name of the registry base section to scan(HKEY_* hives level).
* @param  section 
*         The name of the registry section to scan.
* @param  key 
*         The name of the registry key within given section.
* @param  isWrite 
*         Flag to indicate whether we need to open key for writing or reading only.
* @param  pHKey
*         Pointer to the variable that will receive the handle value of the registry key.
* 
* @return A string representing the registry key name and the key handle value.
*/
char * getKeyNameAndHandle(JNIEnv* env, jstring base, jstring section, jstring key, BOOL isWrite,
                           HKEY * pHKey)
{
   char * name = NULL;
   char * keyPath = NULL;
   HKEY hBaseKey = 0;

   // Construct initial parameters for Windows API call
   const char *base_name = (*env)->GetStringUTFChars(env, base, NULL);
   const char *sect_name = (*env)->GetStringUTFChars(env, section, NULL);
   const char *key_name  = (*env)->GetStringUTFChars(env, key, NULL);
   
   // Reconstruct the handle to the registry hive
   hBaseKey = HKEY_CURRENT_USER;
   
   int keyLength = 0;
   // Check the memory size to hold the full key path name
   if (sect_name != NULL && strlen(sect_name) > 0)
   {
      keyLength += strlen(sect_name);
   }
   // One more char for final null terminator string
   keyLength += 1;
   
   // Allocate memory to registry path name
   keyPath = malloc(sizeof(char) * keyLength);
   // And initialize with empty
   strcpy(keyPath, "");
   // Form the final key path
   if (sect_name != NULL && strlen(sect_name) > 0)
   {
      strcat(keyPath, sect_name);
   }   

   // Extract the key name itself
   keyLength = 0;
   if (key_name != NULL && strlen(key_name) > 0)
   {
      keyLength += strlen(key_name);
   }
   // One more char for final null terminator string
   keyLength += 1;
   
   // Allocate memory to registry key name
   name = malloc(sizeof(char) * keyLength);
   // And initialize with empty
   strcpy(name, "");
   if (key_name != NULL && strlen(key_name) > 0)
   {
      strcat(name, key_name);
   }

   DWORD rc;
   if (isWrite) // Create or reopen subkey to write access
   {
      DWORD dwDisposition;
      rc = RegCreateKeyEx(hBaseKey, keyPath, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL,
                          pHKey, &dwDisposition);
   }
   else         // Open subkey to read only access
   {
      rc = RegOpenKeyEx(hBaseKey, keyPath, 0, KEY_READ, pHKey);
   }
   
   // Release Java memory
   (*env)->ReleaseStringUTFChars(env, key, key_name);
   (*env)->ReleaseStringUTFChars(env, section, sect_name);
   (*env)->ReleaseStringUTFChars(env, base, base_name);
   
   // Free local memory
   free(keyPath);
   // The registry key was created or opened
   if (rc == ERROR_SUCCESS)
   {
      // Remember to release allocated memory in caller in case of success
      return name;
   }
   else // Creation or opening failure, clean up resources   
   {
      // No name value will be returned
      free(name);
      RegCloseKey(*pHKey);
      return NULL;
   }
}

/**
* Helper function that will return the handle of the registry subkey for given section and base. 
* 
* @param  env 
*         pointer to a structure that provides access to JNI environment
* @param  base 
*         The name of the registry base section to scan(HKEY_* hives level).
* @param  section 
*         The name of the registry section to find out.
* @param  samOption 
*         The security access mode for open registry key.
* 
* @return A handle of the registry key requested or 0 in the case of error.
*/
HKEY getSubKeyHandle(JNIEnv* env, jstring base, jstring section, REGSAM samOption)
{
   char * keyPath = NULL;
   HKEY hBaseKey = 0;
   HKEY hSubKeyResult = 0;

   // Construct initial parameters for Windows API call
   const char *base_name = (*env)->GetStringUTFChars(env, base, NULL);
   const char *sect_name = (*env)->GetStringUTFChars(env, section, NULL);
   
   // Reconstruct the handle to the registry hive
   hBaseKey = HKEY_CURRENT_USER;
   
   int keyLength = 0;
   // Check the memory size to hold the full key path name
   if (sect_name != NULL && strlen(sect_name) > 0)
   {
      keyLength += strlen(sect_name);
   }
   // One more char for final null terminator string
   keyLength += 1;
   
   // Allocate memory to registry path name
   keyPath = malloc(sizeof(char) * keyLength);
   // And initialize with empty
   strcpy(keyPath, "");
   // Form the final key path
   if (sect_name != NULL && strlen(sect_name) > 0)
   {
      strcat(keyPath, sect_name);
   }   

   DWORD rc;
   
   rc = RegOpenKeyEx(hBaseKey, keyPath, 0, samOption, &hSubKeyResult);
   
   // Release Java memory
   (*env)->ReleaseStringUTFChars(env, section, sect_name);
   (*env)->ReleaseStringUTFChars(env, base, base_name);
   
   // Free local memory
   free(keyPath);
   // The registry key was created or opened
   if (rc == ERROR_SUCCESS)
   {
      // Remember to release allocated memory in caller in case of success
      return hSubKeyResult;
   }
   else // Creation or opening failure, clean up resources   
   {
      // No name value will be returned
      RegCloseKey(hSubKeyResult);
      return 0;
   }
}

/**
* Utility method that will return the value of the registry key for given section name. 
* 
* @param  env 
*         pointer to a structure that provides access to JNI environment
* @param  cls 
*         reference to the Java class who called this function
*         (from static context)
* @param  base 
*         The name of the registry base section to scan(HKEY_* hives level).
* @param  section 
*         The name of the registry section to scan.
* @param  key 
*         The name of the registry key within given section whose value to get.
* 
* @return A string representing the registry key value if exists or NULL pointer if the key does
*         not exist in the Windows registry.
*/
JNIEXPORT jstring JNICALL
   Java_com_goldencode_p2j_util_Registry_getKeyVal(JNIEnv* env,
                                                   jclass  cls,
                                                   jstring base,
                                                   jstring section,
                                                   jstring key)
{
   char * result = NULL;

   // Reconstruct the handle to the registry hive and full key name
   HKEY hKey = 0;
   // Get registry key for reading
   char * keyName = getKeyNameAndHandle(env, base, section, key, FALSE, &hKey);
   
   DWORD rc;
   DWORD cbData = 0;
   
   // Find out how much memory we need to store the result
   rc = RegQueryValueEx(hKey, keyName, NULL, NULL, NULL, &cbData);
   if (rc == ERROR_SUCCESS && cbData > 0 && keyName != NULL)
   {
      result = malloc(sizeof(char) * cbData);
      if (result != NULL)
      {
         rc = RegQueryValueEx(hKey, keyName, NULL, NULL, (LPBYTE)result, &cbData);
      }
   }

   // Key name was valid
   if (keyName != NULL)
   {
      // Release local memory
      free(keyName);
      // And close opened registry key
      RegCloseKey(hKey);
   }

   if (result != NULL)
   {
      // We need this additional variable to be able to free local memory allocated for
      // Windows API call and to prevent memory leak
      jstring res = (*env)->NewStringUTF(env, result);
      free(result);
      return res;
   }
   else
   {
      return NULL;
   }   
}

/**
* Utility method that will set the new value for the registry key for given section name. 
* 
* @param  env 
*         pointer to a structure that provides access to JNI environment
* @param  cls 
*         reference to the Java class who called this function
*         (from static context)
* @param  base 
*         The name of the registry base section to scan(HKEY_* hives level).
* @param  section 
*         The name of the registry section to scan.
* @param  key 
*         The name of the registry key within given section whose value to change.
* @param  value 
*         The new value for the registry entry to set for given section and given key.
* 
* @return TRUE in case of success, FALSE otherwise.
*/
JNIEXPORT jboolean JNICALL
   Java_com_goldencode_p2j_util_Registry_setKeyVal(JNIEnv* env,
                                                   jclass  cls,
                                                   jstring base,
                                                   jstring section,
                                                   jstring key,
                                                   jstring value)
{
   jboolean res = JNI_FALSE;
   // Construct initial parameters for Windows API call
   const char *key_value = (*env)->GetStringUTFChars(env, value, NULL);

   // Reconstruct the handle to the registry hive and full key name
   HKEY hKey = 0;
   // Get registry key for writing
   char * keyName = getKeyNameAndHandle(env, base, section, key, TRUE, &hKey);
   
   DWORD rc;
   DWORD cbData = 0;
   
   if (key_value != NULL && strlen(key_value) > 0 && keyName != NULL)
   {
      // The size of data to be stored should include null terminator for string data
      cbData = strlen(key_value) + 1;
      rc = RegSetValueEx(hKey, keyName, 0, REG_SZ, (LPBYTE)key_value, cbData);
      if (rc == ERROR_SUCCESS)
      {
         res = JNI_TRUE;
      }
   }
   
   // Key name was valid
   if (keyName != NULL)
   {
      // Release local memory
      free(keyName);
      // And close opened registry key
      RegCloseKey(hKey);
   }
   // Release Java memory
   (*env)->ReleaseStringUTFChars(env, value, key_value);
   
   return res;
}

/**
* Utility method that removes the subkey from registry with it's content for given section name. 
* 
* @param  env 
*         pointer to a structure that provides access to JNI environment
* @param  cls 
*         reference to the Java class who called this function
*         (from static context)
* @param  base 
*         The name of the registry base section to scan(HKEY_* hives level).
* @param  section 
*         The name of the registry section(subkey) to remove.
* 
* @return TRUE in case of success, FALSE otherwise.
*/
JNIEXPORT jboolean JNICALL
   Java_com_goldencode_p2j_util_Registry_removeSubKey(JNIEnv* env,
                                                      jclass  cls,
                                                      jstring base,
                                                      jstring section)
{
   jboolean res = JNI_FALSE;
   // Reconstruct the handle to the registry hive and full key name
   HKEY hKey = 0;
   
   DWORD rc;
   HKEY hBaseKey = HKEY_CURRENT_USER;

   // Construct initial parameters for Windows API call
   const char *section_value = (*env)->GetStringUTFChars(env, section, NULL);
   
   rc = RegOpenKeyEx(hBaseKey, NULL, 0, KEY_WRITE, &hKey);
   if (rc == ERROR_SUCCESS)
   {
      rc = RegDeleteKey(hKey, section_value);
      if (rc == ERROR_SUCCESS)
      {
         res = JNI_TRUE;
      }
      // And close opened registry key
      RegCloseKey(hKey);
   }
   
   // Release Java memory
   (*env)->ReleaseStringUTFChars(env, section, section_value);
   
   return res;
}

/**
* Utility method that removes the subkey from registry with it's content for given section name. 
* 
* @param  env 
*         pointer to a structure that provides access to JNI environment
* @param  cls 
*         reference to the Java class who called this function
*         (from static context)
* @param  base 
*         The name of the registry base section to scan(HKEY_* hives level).
* @param  section 
*         The name of the registry section(subkey) to remove the key.
* @param  key 
*         The name of the registry key(value name) to remove within the value.
* 
* @return TRUE in case of success, FALSE otherwise.
*/
JNIEXPORT jboolean JNICALL
   Java_com_goldencode_p2j_util_Registry_removeValueName(JNIEnv* env,
                                                         jclass  cls,
                                                         jstring base,
                                                         jstring section,
                                                         jstring key)
{
   jboolean res = JNI_FALSE;
   
   // Construct initial parameters for Windows API call
   const char *key_value = (*env)->GetStringUTFChars(env, key, NULL);

   // Reconstruct the handle to the registry subkey
   HKEY hSubKey = getSubKeyHandle(env, base, section, KEY_SET_VALUE);
   
   DWORD rc = 0;
   
   if (hSubKey != 0)
   {
      rc = RegDeleteValue(hSubKey, key_value);
      if (rc == ERROR_SUCCESS)
      {
         res = JNI_TRUE;
      }
      // And close opened registry key
      RegCloseKey(hSubKey);
   }
   
   // Release Java memory
   (*env)->ReleaseStringUTFChars(env, key, key_value);

   return res;
}

/**
* Utility method that will return the comma separated list of the value-namws of the registry for
* given section name. 
* 
* @param  env 
*         pointer to a structure that provides access to JNI environment
* @param  cls 
*         reference to the Java class who called this function
*         (from static context)
* @param  base 
*         The name of the registry base section to scan(HKEY_* hives level).
* @param  section 
*         The name of the registry section to scan.
* 
* @return A string representing the list of the registry key names or NULL pointer if the section
*         does not exist in the Windows registry.
*/
JNIEXPORT jstring JNICALL
   Java_com_goldencode_p2j_util_Registry_getValueNames(JNIEnv* env,
                                                       jclass  cls,
                                                       jstring base,
                                                       jstring section)
{
   char * result = NULL;

   // Reconstruct the handle to the registry subkey
   HKEY hSubKey = getSubKeyHandle(env, base, section, KEY_QUERY_VALUE);
   
   if (hSubKey != 0)
   {
      char chValueName[MAX_VALUE_NAME];
      DWORD dwBuffSize = MAX_VALUE_NAME;
      int listLength = 0;
      DWORD rc = 0;
      int i;
      // First pass, determine the size of the buffer required
      for (i = 0, rc = ERROR_SUCCESS; rc != ERROR_NO_MORE_ITEMS; i++)
      {
         chValueName[0] = '\0';
         dwBuffSize = MAX_VALUE_NAME;
         rc = RegEnumValue(hSubKey, i, chValueName, &dwBuffSize, NULL, NULL, NULL, NULL);
         if (rc == ERROR_SUCCESS)
         {
            listLength += dwBuffSize;
            if (i > 0)
            {
               // Space for comma separator
               listLength += 1;
            }
         }
      }
      // Final null terminator char
      listLength += 1;
      // Allocate memory for result
      result = malloc(sizeof(char) * listLength);
      if (result != NULL)
      {
         // And initialize with empty
         strcpy(result, "");
         // Second pass, construct final list buffer
         for (i = 0, rc = ERROR_SUCCESS; rc != ERROR_NO_MORE_ITEMS; i++)
         {
            chValueName[0] = '\0';
            dwBuffSize = MAX_VALUE_NAME;
            rc = RegEnumValue(hSubKey, i, chValueName, &dwBuffSize, NULL, NULL, NULL, NULL);
            if (rc == ERROR_SUCCESS)
            {
               if (i > 0)
               {
                  // Comma separator adding
                  strcat(result, ",");
               }
               strcat(result, chValueName);
            }
         }
      }
      
      // And close opened registry key
      RegCloseKey(hSubKey);
   }
   
   if (result != NULL)
   {
      // We need this additional variable to be able to free local memory allocated for
      // Windows API call and to prevent memory leak
      jstring res = (*env)->NewStringUTF(env, result);
      free(result);
      return res;
   }
   else
   {
      return NULL;
   }   
}

/**
* Utility method that will return the comma separated list of the subkey@value_name pair for
* given environment name. 
* 
* @param  env 
*         pointer to a structure that provides access to JNI environment
* @param  cls 
*         reference to the Java class who called this function
*         (from static context)
* @param  base 
*         The name of the registry base section to scan(HKEY_* hives level).
* @param  section 
*         The name of the registry section to scan.
* 
* @return A string representing the list of the registry key names or NULL pointer if the section
*         does not exist in the Windows registry.
*/
JNIEXPORT jstring JNICALL
   Java_com_goldencode_p2j_util_Registry_getSubKeys(JNIEnv* env,
                                                    jclass  cls,
                                                    jstring base,
                                                    jstring section)
{
   char * result = NULL;

   // Reconstruct the handle to the registry subkey
   HKEY hKey = getSubKeyHandle(env, base, section, KEY_ENUMERATE_SUB_KEYS);
   
   if (hKey != 0)
   {
      char chKeyName[MAX_KEY_LENGTH];
      DWORD dwKeyBuffSize = 0;
      int listLength = 0;
      DWORD rc = 0;
      int i;
      // First pass to detect the memory required for result
      for (i = 0, rc = ERROR_SUCCESS; rc != ERROR_NO_MORE_ITEMS; i++)
      {
         chKeyName[0] = '\0';
         dwKeyBuffSize = MAX_KEY_LENGTH;
         rc = RegEnumKeyEx(hKey, i, chKeyName, &dwKeyBuffSize, NULL, NULL, NULL, NULL);
         if (rc == ERROR_SUCCESS)
         {
            // The entry format is subkey,subkey
            listLength += dwKeyBuffSize;
            if (i > 0)
            {
               // Space for comma separator
               listLength += 1;
            }
         }
      }
      // Final null terminator char
      listLength += 1;
      // Allocate memory for result
      result = malloc(sizeof(char) * listLength);
      if (result != NULL)
      {
         // And initialize with empty
         strcpy(result, "");
         // Second pass, construct final list buffer
         for (i = 0, rc = ERROR_SUCCESS; rc != ERROR_NO_MORE_ITEMS; i++)
         {
            chKeyName[0] = '\0';
            dwKeyBuffSize = MAX_KEY_LENGTH;
            rc = RegEnumKeyEx(hKey, i, chKeyName, &dwKeyBuffSize, NULL, NULL, NULL, NULL);
            if (rc == ERROR_SUCCESS)
            {
               // Start enumerating the value-names for the given subkey
               if (i > 0)
               {
                  // Comma separator adding
                  strcat(result, ",");
               }
               // The entry format is subkey@value_name
               strcat(result, chKeyName);
            }
         }
      }

      // And close opened registry key
      RegCloseKey(hKey);
   }   
      
   if (result != NULL)
   {
      // We need this additional variable to be able to free local memory allocated for
      // Windows API call and to prevent memory leak
      jstring res = (*env)->NewStringUTF(env, result);
      free(result);
      return res;
   }
   else
   {
      return NULL;
   }   
}
